/***************************************************************************
 *   Copyright (C) 2018 by Staud   *
 *   michael.staud@uni-ulm.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "UtiArchiv.h"
#include<math.h>
#include <assert.h>
#include <iostream>
#include <fstream>

#include <randomnumber.h>
#include "utikeycodes.h"
#include <tuple>
#include "curves.h"
#include "syntheticrule.h"
#include "simplecpp.h"
#include "utisystem.h"

int main (int argc, char *argv[])
{

   std::wcout << "C to PDDL Compiler - V 0.1" << std::endl;
   std::wcout << "Copyright (C) 2019  Michael Staud" << std::endl;
   std::wcout << "This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'." << std::endl;
   std::wcout << "This is free software, and you are welcome to redistribute it" << std::endl;
   std::wcout << "under certain conditions" << std::endl;
   CUtiString inputfile;
   std::string inputfile_str;
    int argindex = 1;
    bool param_version = false;
    CUtiString domainfile, problemfile;

    simplecpp::DUI dui;
    while (argindex < argc) {
        std::string param = argv[argindex];
        std::string next;
        if (argindex+1 < argc) {
            next = argv[argindex+1];
        }
        argindex++;
        if (param == "--version") {
            param_version = true;
        }else if (param == "-D") {
            dui.defines.push_back(next.c_str());
        }else if (param == "-U") {
            dui.undefined.insert(next.c_str());
        }else if (param == "-I") {
            dui.includePaths.push_back(next.c_str());
        }else if (param.rfind("-include=", 0) == 0) {
            std::string include = param.substr(9);
            dui.includes.push_back(include.c_str());
        }else if (param == "--v") {
            param_version = true;
        }else if (param == "--o") {
            domainfile =UtiFramework::converttowide(next.c_str());
            argindex++;
        }else if (param == "--op") {
            problemfile =UtiFramework::converttowide(next.c_str());

            argindex++;
        }else {
           inputfile = UtiFramework::converttowide(param.c_str());
           inputfile_str = param;
        }

    }

    if (!inputfile.size()) {
       std::wcout << "No input file!" << std::endl;
       std::wcout << "Syntax:" << std::endl;
        std::wcout << "ctopddl [options] filename" << std::endl;
        std::wcout << "  -o              Output Domain File Name." << std::endl;
        std::wcout << "  -op             Output Problem File Name." << std::endl;
        std::wcout << "  -DNAME          Define NAME." << std::endl;
        std::wcout << "  -IPATH          Include path." << std::endl;
        std::wcout << "  -include=FILE   Include FILE." << std::endl;
        std::wcout << "  -UNAME          Undefine NAME." << std::endl;

       return 0;
    }

    simplecpp::OutputList outputList;
    std::vector<std::string> files;
    std::ifstream f(inputfile_str);
    simplecpp::TokenList rawtokens(f,files,inputfile_str,&outputList);
    rawtokens.removeComments();
    std::map<std::string, simplecpp::TokenList*> included = simplecpp::load(rawtokens, files, dui, &outputList);
    for (std::pair<std::string, simplecpp::TokenList *> i : included)
        i.second->removeComments();
    simplecpp::TokenList outputTokens(files);
    simplecpp::preprocess(outputTokens, rawtokens, files, included, dui, &outputList);

    // Output
    std::wcout << "After Preprocessor: " << std::endl;
    std::wcout << outputTokens.stringify().c_str() << std::endl;
    std::string outputdata = outputTokens.stringify();
    for (const simplecpp::Output &output : outputList) {
        std::wcout << output.location.file().c_str() << ':' << output.location.line << ": ";
        switch (output.type) {
        case simplecpp::Output::ERROR:
            std::wcout << "#error: ";
            break;
        case simplecpp::Output::WARNING:
            std::wcout << "#warning: ";
            break;
        case simplecpp::Output::MISSING_HEADER:
            std::wcout << "missing header: ";
            break;
        case simplecpp::Output::INCLUDE_NESTED_TOO_DEEPLY:
            std::wcout << "include nested too deeply: ";
            break;
        case simplecpp::Output::SYNTAX_ERROR:
            std::wcout << "syntax error: ";
            break;
        case simplecpp::Output::PORTABILITY_BACKSLASH:
            std::wcout << "portability: ";
            break;
        case simplecpp::Output::UNHANDLED_CHAR_ERROR:
            std::wcout << "unhandled char error: ";
            break;
        }
        std::wcout << output.msg.c_str() << std::endl;
    }

    // cleanup included tokenlists
    simplecpp::cleanup(included);

    CRProgram baseprogram;
    baseprogram.loadFromString(inputfile, outputdata);
    CRActionContainer* action=NULL;
    if (domainfile.size() == 0) {
       domainfile = inputfile;
       domainfile.ExtractFileNameWithoutExt();
       domainfile += UTISTR(".pddl");
    }
    baseprogram.compileTo(action, domainfile);
    if (problemfile.size() == 0) {
       problemfile = inputfile;
       problemfile.ExtractFileNameWithoutExt();
       problemfile += UTISTR("_problem.pddl");
    }
    baseprogram.executeProgram(problemfile);
    return 0;
}

