/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "Uti3DCore.h"
#include "Uti3DAnimations.h"
#include "Uti3DMath.h"
//#include "Uti3D.h"

#ifdef EDITOR
#include "UtiDEdit.h"
#include "UtiApp.h"
#endif
//#include "MainEngine.h"
//#include "SoftEngine.h"
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <algorithm>
#include "easyprofiler.h"



CDECLAREANITYPE_ALL_IMPL(CLinearInterpolationPoint, 2100);
CDECLAREANITYPE_ALL_IMPL(CLinearInterpolationPointNoWatch, 12110);
CDECLAREANITYPE_ALL_IMPL(CFixedPoint, 12120);
CDECLAREANITYPE_ALL_IMPL(CFixedPointNoWatch, 12130);
CDECLAREANITYPE_ALL_IMPL(CBezierInterpolationPointNoWatch, 12140);

CDECLAREANITYPE_ALL_CLASS_IMPL(CAnimationValueClass, 2200);

CDECLAREANITYPE_ALL_IMPL(CAnimationDataTemplate, 12280);



CAniPoint::CAniPoint()
{
  //$extern::begin constructor(name="CAniPoint")
   m_Time = 0.0f;
   ;
   ;
      //$extern::end

//   typinfo = UTI_ANIPOINT;
}

void ReadOldTime(CStream& Stream, int Flags, int version, TIME &time)
{
  int mytime;
	Stream.ReadInteger(mytime);
	time = mytime / 1000.0;
}

void WriteOldTime(CStream& Stream, int Flags, int version, TIME time)
{
  Stream.WriteInteger((int)round(time*1000.0));
}

HResult CAniPoint::Read(CStream & Stream, int Flags, int version)
{


    ReadOldTime(Stream, Flags, version, m_Time);
	if (Flags & 2) {
       int f;
	   Stream.ReadInteger(f);
	   float m;
	   Stream.ReadFloat(m);
	   Stream.ReadFloat(m);
	   Stream.ReadFloat(m);
	   Stream.ReadFloat(m);
	}
	return(UTI_OK);
}

HResult CAniPoint::Write(CStream & Stream, int Flags, int version)
{
    WriteOldTime(Stream, Flags, version,m_Time);
	return(UTI_OK);
}


void CAnimationSubPart::evaluate(const SAnimationInput& input, SAnimationResult* output) const
{
    SAnimationInput in;
    in.m_time = input.m_time-m_Start;
    if (in.m_time < 0) in.m_time = 0.0f;
    if (in.m_time > m_Length) in.m_time = m_Length;
    m_Data->evaluate(in, output);
}

CAniDoublePoint::CAniDoublePoint() : CAniPoint()
{
  //$extern::begin constructor(name="CAniDoublePoint")
   m_Value = 0.0f;
   ;
   ;
      //$extern::end

//   typinfo = UTI_ANIDOUBLEPOINT;
}

HResult CAniDoublePoint::Read(CStream & Stream, int Flags, int version)
{
	double v;
	Stream.ReadDouble(v);
    m_Value = v;
	CAniPoint::Read(Stream, Flags, version);
	return(UTI_OK);
}

HResult CAniDoublePoint::Write(CStream & Stream, int Flags, int version)
{
	Stream.WriteDouble(m_Value);
	CAniPoint::Write(Stream, Flags, version);
	return(UTI_OK);
}

CAniOrderPoint::CAniOrderPoint() : CAniPoint()
{
   value = 0;
//   typinfo = UTI_ANIORDERPOINT;
}

HResult CAniOrderPoint::Read(CStream & Stream, int Flags, int version)
{
	int v;
	Stream.ReadInteger(v);
value = v;
	CAniPoint::Read(Stream, Flags, version);
	return(UTI_OK);
}

HResult CAniOrderPoint::Write(CStream & Stream, int Flags, int version)
{
	Stream.WriteInteger(value);
	CAniPoint::Write(Stream, Flags, version);
	return(UTI_OK);
}



CAnimation::CAnimation(TIME newTime, SCode newCode)
{
     //$extern::begin constructor(name="CAnimation")
   m_CurrentPosition = 0.0f;
   m_TimeUsed = 0.0f;
   m_Data = NULL;
   m_Code = NULLCODE;
   m_Flags = 0;
   ;
   ;
      //$extern::end

   m_Code = newCode;
//   typinfo = UTI_ANIMATION;
   m_CurrentPosition = 0; m_TimeUsed = newTime; m_Flags = 0;
//   points.sorted = true;


}

CAnimation::CAnimation()
{
  //$extern::begin constructor(name="CAnimation")
   m_CurrentPosition = 0.0f;
   m_TimeUsed = 0.0f;
   m_Data = NULL;
   m_Code = NULLCODE;
   m_Flags = 0;
   ;
   ;
      //$extern::end

}

void CAnimation::createStdAnimation()
{
  m_AnimationClass->createStdAnimation(this);
}

/*int pointCompare(const void* a, const void* b)
{
  CAniPoint* pa = ((CAniPoint**) a);
  CAniPoint* pb = ((CAniPoint**) b);
  return (pa->getTime() > pb->getTime()) - (pa->getTime() < pb->getTime());
};
*/



void CAnimationDataPoints::checkPoint(CAniPoint* p)
{
    int index = m_Points.indexOf(p);
    utiassert(index != -1);
    if (index != 0){
        TIME last = m_Points[index-1].getTime();
        if (p->getTime() < last) {
           p->setTime(last);
        };
    }
    if (index != m_Points.size()-1){
        TIME next = m_Points[index+1].getTime();
        if (p->getTime()> next) {
           p->setTime(next);
        };
    }

}

bool compareRefPoints(const CAniPoint* p1, const CAniPoint* p2)
{
    return p1->getTime() < p2->getTime();
}

void CAnimationDataPoints::sortPoints()
{
   std::sort(m_Points.begin_pointer(), m_Points.end_pointer(), compareRefPoints);
}


CAniPoint* CAnimationDataPoints::getAniPointByTime(TIME time, TIME tolerance)
{
    for (int i = 0; i < m_Points.size(); i++) {
        CAniPoint* p = (CAniPoint*)&m_Points[i];
        if (fabs(p->getTime() - time) < tolerance) {
            return p;
        }
    }
    return NULL;
}

void CAnimationDataPoints::updatePointPosition(CAniPoint* p)
{
  if (m_Points.size() == 1) return;
  int index = m_Points.indexOf(p);
#define CHECK_ANIMATIONS for (int i = 0; i < m_Points.size()-1; i++) {\
  utiassert(m_Points[i].getTime() <= m_Points[i+1].getTime());\
}

  utiassert(index != -1);
  if (index == 0 && (m_Points.size() == 1 || m_Points[1].getTime() >= p->getTime())){
      CHECK_ANIMATIONS;
    return;
  };
  if (index == m_Points.size()-1 && (m_Points.size() == 1 || m_Points[m_Points.size()-2].getTime() <= p->getTime())) {

      CHECK_ANIMATIONS;
    return;
  };
  if (index != 0 && index != m_Points.size()-1) {
    if (m_Points[index-1].getTime() <= p->getTime() && p->getTime() <= m_Points[index+1].getTime()) {

        CHECK_ANIMATIONS;
      return;
    };
  };
  if (m_Points[0].getTime() >= p->getTime()) {

    m_Points.move(index, 0);
    CHECK_ANIMATIONS;
    return;
  };
  sortPoints();
   CHECK_ANIMATIONS;
 /* if (m_Points[m_Points.size()-1].getTime() <= p->getTime()) {


    m_Points.move(index, m_Points.size()-1);
    CHECK_ANIMATIONS;
    return;
  };
  TIME t = p->getTime();
  for (int i = 0; i < m_Points.size()-1; i++) {
    if (m_Points[i].getTime() <= t && m_Points[i+1].getTime() >= t) {
      m_Points.move(index, i+1);

      CHECK_ANIMATIONS;
    return;
    };
  };*/

  //utiassert(false);
}

#ifdef EDITOR
void CDoubleClass::AddTreeItem(void* view, UINT parent, U_CHAR *c)
{
   CUtiTreeView *tree = (CUtiTreeView*) view;
   if (TreeItem == 0) {
	   CUtiOB::AddTreeItem(view, parent, c);
	   CUtiString number;
	   number.SetFloat(Wert, 2);
	   CUtiString info = c;
	   info += UTEXT(": ");

	   info += number;
       TreeItem = tree->Insert(parent, info.getString(), false, this);
   }
}

void C3DTriple::AddTreeItem(void* view, UINT parent, U_CHAR *c)
{
   CUtiTreeView *tree = (CUtiTreeView*) view;
   if (TreeItem == 0) {
	   CUtiOB::AddTreeItem(view, parent, c);
	   TreeItem = tree->Insert(parent, c, true, this);
	   x.AddTreeItem(view, TreeItem, UTEXT("X"));
	   y.AddTreeItem(view, TreeItem, UTEXT("Y"));
	   z.AddTreeItem(view, TreeItem, UTEXT("Z"));
   }
}
#endif

HResult CAnimation::Write(CStream &Stream, int Flags, int version)
{
CAnimationDataPoints* datapoint = check_cast<CAnimationDataPoints*>(getData());
    WriteCode(Stream, m_Code, version);
   Stream.WriteData(&m_Flags, 4);
   WriteOldTime(Stream, Flags, version,m_CurrentPosition);
   WriteOldTime(Stream, Flags, version,m_TimeUsed);
   int c = datapoint->getPointsCount();
   Stream.WriteInteger(c);
   CAniPoint *p = NULL;
   for (int i = 0; i < c; i++) {

      p = datapoint->getPoints(i);
      if (dynamic_cast<CLinearInterpolationPoint<double>*>(p)) {
          CLinearInterpolationPoint<double>* p2 = (CLinearInterpolationPoint<double>*)p;
          double v = p2->getValue();
          Stream.WriteDouble(v);
          WriteOldTime(Stream, Flags, version, p2->m_Time);
          if ((m_Flags & 2) !=  0) {
             int f=0;
             Stream.WriteInteger(f);
             float m=0.0f;
             Stream.WriteFloat(m);
             Stream.WriteFloat(m);
             Stream.WriteFloat(m);
             Stream.WriteFloat(m);
          }
      } else {
          p->Write(Stream, m_Flags, version);
      }
      //datapoint->addPoints(p);
   }
  /*  //Stream.WriteData(&m_Code, sizeof CAniDoublePoint);
WriteCode(Stream, m_Code, version);
	Stream.WriteData(&m_Flags, sizeof m_Flags);
	WriteOldTime(Stream, Flags, version,m_CurrentPosition);
	WriteOldTime(Stream, Flags, version,m_TimeUsed);
	//Stream.WriteInteger(m_CurrentPosition);
	//Stream.WriteInteger(m_TimeUsed);
	Stream.WriteInteger(m_Points.size());
	for (int i = 0; i < m_Points.size(); i++) {
	  m_Points[i].Write(Stream, Flags, version);
	};
   // points.EnumItems(&Stream, WritefuncCache);*/
	return(UTI_OK);
}

HResult CAnimation::Read(CStream &Stream, int Flags, int version)
{
//	m_Points.clear();
    CAnimationDataPoints* datapoint = m_AnimationClass->createAnimationData();
    setData(datapoint);
//	 Stream.ReadData(&CAniDoublePoint, sizeof CAniDoublePoint);
     ReadCode(Stream, m_Code, version);
	Stream.ReadData(&m_Flags, 4);
    ReadOldTime(Stream, Flags, version,m_CurrentPosition);
    ReadOldTime(Stream, Flags, version,m_TimeUsed);
	int c = 0;
	Stream.ReadInteger(c);
	CAniPoint *p = NULL;
	for (int i = 0; i < c; i++) {

       p = m_AnimationClass->createAniPoint();
       if (dynamic_cast<CLinearInterpolationPoint<double>*>(p)) {
           CLinearInterpolationPoint<double>* p2 = (CLinearInterpolationPoint<double>*)p;
           double v;
           Stream.ReadDouble(v);
           p2->setValue(v);
           ReadOldTime(Stream, Flags, version, p2->m_Time);
           if ((m_Flags & 2) !=  0) {
              int f;
              Stream.ReadInteger(f);
              float m;
              Stream.ReadFloat(m);
              Stream.ReadFloat(m);
              Stream.ReadFloat(m);
              Stream.ReadFloat(m);
           }
       } else {
           p->Read(Stream, m_Flags, version);
       }
       datapoint->addPoints(p);
	}
/*	if (m_AnimationClass->normalAnimation() && m_Points.size() != 0) {
	   m_Points[ m_Points.size()-1].setTime(m_TimeUsed);
    };*/
   return(UTI_OK);
}

CAnimationInfo::CAnimationInfo()
{
  //$extern::begin constructor(name="CAnimationInfo")
   ;
   ;
      //$extern::end
//	typinfo =UTI_ANIMATIONINFO;

}

CAnimation* CAnimationInfo::addAnimation(TIME newTime, SCode newCode)
{
  CAnimation* ani = new CAnimation(newTime, newCode);
  ani->setAnimationClass(getAnimationClass());
  m_Animations.push_back(ani);
  return ani;
}

CAnimation* CAnimationInfo::findAnimation(const SCode &code)
{
  for (int i= 0;i < m_Animations.size(); i++) {
     if (m_Animations[i].getCode() == code) return &m_Animations[i];
  };

  return NULL;
}

HResult CAnimationInfo::Write(CStream &Stream, int Flags, int version)
{

  double Grundwert=0.0;
   bool SetGrundwert=false;
   int aniIndex=m_Animations.indexOf(m_Active.get());
	Stream.WriteDouble(Grundwert);

	Stream.WriteBoolean(SetGrundwert);
	Stream.WriteInteger(aniIndex);
	int c = m_Animations.size();
	Stream.WriteInteger(c);
   // Animations.EnumItems(&Stream, WritefuncCache);
      for (int i = 0; i < m_Animations.size(); i++) {
	  m_Animations[i].Write(Stream, Flags, version);
      };
	return (UTI_OK);
}

HResult CAnimationInfo::Read(CStream &Stream, int Flags, int version)
{	//double v

    EASY_FUNCTION(profiler::colors::Silver);
double Grundwert;
      int aniIndex=0;
    Stream.ReadDouble(Grundwert);//Grundwert = v;
    bool SetGrundwert;
	Stream.ReadBoolean(SetGrundwert);
	Stream.ReadInteger(aniIndex);
	int i, c;
	Stream.ReadInteger(c);
	m_Active = NULL;
    CAnimation* newAni = NULL;
	for (i = 0; i < c; i++) {
		newAni = addAnimation(0, NULLCODE);

	   newAni->Read(Stream, Flags, version);
      // m_Animations.push_back(newAni);
       if (i == aniIndex) {
           m_Active = newAni;
       }
	}
	return (UTI_OK);
}

CAnimationClass::CAnimationClass()
{
  //$extern::begin constructor(name="CAnimationClass")
   ;
   ;
      //$extern::end

}



CAnimationClass::~CAnimationClass()
{

}


CAniPoint* CAnimationClass::createAniPoint()
{
  return new CAniPoint();
}

CAnimation *CAnimationClass::CreateNewAnimation(TIME newTime, SCode newCode, bool activate)
{
 setAnimated(true);

	//m_AniInfo->aniIndex = m_AniInfo->Animations.count;

	//CAnimation *newIt = /*cmem->CreateAnimation();//*/ new CAnimation(newTime, newCode);
	//newIt->m_TimeUsed = newTime;
	//newIt->code = newCode;

	CAnimation* animation = m_AniInfo->addAnimation(newTime, newCode);
        animation->createStdAnimation();
        if (activate) {
            activateTrace(animation);
        }

	return(animation);
};




/*void CDoubleClass::createStdAnimation(CAnimation* ani)
{
  ani->clearPoints();
   CAniDoublePoint *newPoint = NULL;

	newPoint = new CAniDoublePoint();
	newPoint->setTime(0);
	newPoint->setValue(Wert);
	ani->addPoints(newPoint);

	newPoint = new CAniDoublePoint();
	newPoint->setTime(ani->getTimeUsed());
	newPoint->setValue(Wert);
	ani->addPoints(newPoint);


}

CDoubleClass::CDoubleClass() : CAnimationClass()
{

   Wert = 0;
//   typinfo = UTI_DOUBLECLASS;
   #ifdef EDITOR
	   editfeld = NULL;
	   dedit = NULL;
    #endif

}

CAniPoint* CDoubleClass::createAniPoint()
{
    return new CAniDoublePoint();
}


CDoubleClass::~CDoubleClass()
{
#ifdef EDITOR
	if (dedit)
   delete dedit;
#endif
}*/
/*
bool searchItems(CUtiOB* ob, void* data)
{
	CAnimation *ani = NULL;
	ani = (CAnimation*)ob;
	if (ani == NULL) return true;
	_anires *rec = (_anires*) data;

	if (compareCode(ani->m_Code, rec->m_Code)) {
        rec->result = ani;
		return false;
	}
	return true;
}
*/
CAnimation* CAnimationClass::findAnimation(const SCode &code)
{
   if (m_AniInfo == NULL) return NULL;
   return m_AniInfo->findAnimation(code);
   /*_anires rec;
   rec.result = NULL;
   rec.code = code;
   m_AniInfo->Animations.EnumItems(&rec, &searchItems); //nT(' Pointer auf n') Pointer
   return rec.result;*/

}

void CAnimationClass::FirstPosition()
{
   setTrackPosition(0, 0);
}

void CAnimationClass::LastPosition()
{
   if (m_AniInfo == NULL) return;
   if (m_AniInfo->getActive() == NULL) return;
   setTrackPosition(m_AniInfo->getActive()->getTimeUsed(), 0);
}

void CAnimationClass::setTrackPosition(TIME time, int flags)
{

}

CAnimation* CAnimationClass::activateTrace(const SCode& code)
{
    CAnimation* ani = findAnimation(code);
    if (ani == NULL) return (ani);
   // m_AniInfo->aniIndex = m_AniInfo->Animations.IndexOf(ani);
	m_AniInfo->setActive( ani);
    FirstPosition();
	return(ani);
}
/*
void CDoubleClass::addIndices(const SCode &code)
{
    CAnimation* ani = findAnimation(code);
    if (ani == NULL) {
        m_AniInfo->indices.Add(ani);
        return;
    };
    utiassert(m_AniInfo->indices.IndexOf(ani) == -1);
    m_AniInfo->indices.Add(ani);
};

float CDoubleClass::getAnimationValueByIndices(int index, TIME time)
{
    if (m_AniInfo == NULL) return Wert;
    utiassert(!(index < 0 || index >= m_AniInfo->indices.count));
    if (index < 0 || index >= m_AniInfo->indices.count) {
        return m_AniInfo->getBaseValue();
    };
    CAnimation* ani = (CAnimation*)m_AniInfo->indices[index];
    if (ani == NULL) return m_AniInfo->getBaseValue();
    return getAnimationValue(ani, time);
}
*/
CAnimation* CAnimationClass::activateTrace(CAnimation* ani)
{
   // m_AniInfo->aniIndex = m_AniInfo->Animations.IndexOf(ani);
	m_AniInfo->setActive(ani);
    FirstPosition();
	return(ani);
}
/*
int CAnimationClass::getLength()
{
    if (m_AniInfo == NULL) return(0);
	if (m_AniInfo->active == NULL) return(0);
    return(m_AniInfo->active->getTimeUsed());
}
*/
/*bool setLengthItems(CUtiOB* ob, void* data)
{
	CAniPoint *ani = NULL;
	ani = (CAniPoint*)ob;
	int length = *((TIME*) data);
    if (ani->m_Time > length-1) {
		ani->m_Time = length-1;
	}

	return true;
}*/

/*void CAnimationClass::setLength(TIME length)
{
    if (m_AniInfo == NULL) return;
	if (m_AniInfo->getActive() == NULL) return;
	m_AniInfo->getActive()->setLength(length);

}*/

/*void CAnimationClass::setLength(TIME length, CAnimation* ani)
{
    ani->points.EnumItems(&length, setLengthItems);
	ani->m_TimeUsed = length;

	CAniPoint* p = (CAniPoint*)ani->points[ani->points.count-1];
//    int i = UTI_DOUBLECLASS;
    if (dynamic_cast<CAniDoublePoint*>(p))
	    p->m_Time = length-1;
}*/
/*
void CAnimation::setLength(TIME time)
{
  for (int i = 0; i < m_Points.size(); i++) {
     if (m_Points[i].getTime() > time) {
       m_Points[i].setTime(time);
     };
  };
  if (m_AnimationClass->normalAnimation()) {
    m_Points[m_Points.size()-1].setTime(time);
  };
}*/

/*SCode CAnimationClass::getCode()
{
	SCode n;
	n = NULLCODE;
    if (m_AniInfo == NULL) return(n);
	if (m_AniInfo->getActive() == NULL) return(n);
	return(m_AniInfo->getActive()->getCode());
}

void CAnimationClass::setCode()
{
	if (m_AniInfo == NULL) return;
	if (m_AniInfo->getActive() == NULL) return;
	m_AniInfo->getActive()->setCode(code);
}*/

/*int CAnimationClass::getPosition()
{
    if (m_AniInfo == NULL) return(-1);
	if (m_AniInfo->getActive() == NULL) return(-1);
	return(m_AniInfo->getActive()->getCurrentPosition());
}
*/
bool CAnimationClass::getAnimated()
{
   if (m_AniInfo == NULL) return(false);
   if (m_AniInfo->m_Animations.size() == 0) return(false);
   return(true);
}

void CAnimationClass::setAnimated(bool animated)
{
  if (animated) {
    if (m_AniInfo.get() != NULL) return;
    m_AniInfo = new CAnimationInfo();
    m_AniInfo->setAnimationClass(this);
  } else {
    if (m_AniInfo.get() == NULL) return;
    m_AniInfo = NULL;
  };


}

CAnimation* CAnimationClass::getActiveAnimation()
{
   if (m_AniInfo == NULL) return NULL;
   return m_AniInfo->getActive();
}
/*
bool CAnimationClass::getRepeat()
{
  	if (m_AniInfo == NULL) return(false);
	if (m_AniInfo->active == NULL) return(false);
	return (m_AniInfo->active->flag == CAnimation::ofRepeat);
}

void CAnimationClass::setRepeat(bool r)
{
    if (m_AniInfo == NULL) return;
	if (m_AniInfo->active == NULL) return;
	if (r) m_AniInfo->active->flag = CAnimation::ofRepeat;
		else m_AniInfo->active->flag = 0;
}*/

bool CAnimation::getRepeat()
{
   return m_Flags == CAnimation::ofRepeat;
}

void CAnimation::setRepeat(bool r)
{
   if (r) m_Flags = CAnimation::ofRepeat;
		else m_Flags = 0;
}

/*
float CDoubleClass::getBaseValue()
{
    if (m_AniInfo.get() == NULL) return 0;
    return m_AniInfo->m_BaseValue;
}

float CDoubleClass::getAnimationValue(CAnimation* ani, TIME time)
{
   utiassert(ani != NULL);
   utiassert(m_AniInfo.get()  != NULL);

	if (time > ani->m_TimeUsed){
		//return;
     time = ani->m_TimeUsed;
	}
	if (time < 0) return m_AniInfo->m_BaseValue;

	CAniDoublePoint* p1 = NULL;
	CAniDoublePoint* p2 = NULL;
    CReferenceList<CAniPoint>& l = ani->m_Points.getInternData();
    if (l.size() < 2) return m_AniInfo->m_BaseValue;
    //Binary Suche
    int min = 0;
    int max = l.size()-1;
    #define POINT(i) ((CAniDoublePoint*)&l[i])
    while (!(POINT(min)->m_Time.get()  <= time && POINT(min+1)->m_Time.get()  >= time) && min != max-1) {
       int middle = (min+max) / 2;
       if (POINT(middle)->m_Time.get()  > time) {
          max = middle;
       } else {
          min = middle;
       };
    };
    p1 = POINT(min);
    p2 = POINT(min+1);
    utiassert(p1->m_Time.get() <= time+0.01 && p2->m_Time.get() >= time-0.01);
    utiassert(min >= 0);
    utiassert(min < l.size()-1);

    #undef POINT

	if ((p1 == NULL) && (p2 == NULL)) {
		//Fehler(UTEXT("Fehler beim setzen der Spurposition"));
		return m_AniInfo->m_BaseValue.get() ;
	}
    if ((p1 == NULL) && (p2 != NULL)) {
		return p2->m_Value;
    }
	if ((p1 != NULL) && (p1->m_Time.get()  == time)) {
		return p1->m_Value;
	}
    if ((p2 != NULL) && (p2->m_Time.get()  == time)) {
		return p2->m_Value;

	}
	float y1 = p1->m_Value;
	float y2 = p2->m_Value;
	TIME x1 = p1->m_Time.get() ;
	TIME x2 = p2->m_Time.get() ;

	float deltax = (float)(x1-x2);
    float m;
	if (deltax == 0) {
		m = (float)1E+10;
	} else {
		m = (float)((y1-y2)/deltax);
	}
	float c = (float)(y1 - m * x1);
	return (float)(m*time + c);

};

void CDoubleClass::setTrackPosition(TIME time, int flags)
{
    if (m_AniInfo.get()  == NULL) return;
	if (m_AniInfo->m_Active == NULL) return;
	if (time > m_AniInfo->m_Active->m_TimeUsed){
		//return;
     time = m_AniInfo->m_Active->m_TimeUsed;
	}
	if (time < 0) return;
    m_AniInfo->m_Active->m_CurrentPosition = time;
    Wert = getAnimationValue(m_AniInfo->m_Active.get(), time);

}

void CDoubleClass::setWert(float d)
{
    Wert = d;

}

HResult CDoubleClass::Read(CStream &Stream, int Flags, int version)
{
    bool IsAnimated;

	if (m_AniInfo.get() != NULL) m_AniInfo = NULL;
//MWorld->
	Stream.ReadData(&IsAnimated, 1);
	double w;
	Stream.ReadData(&w, 8);
Wert  = w;

	if (IsAnimated) {
        //m_AniInfo = new CAnimationInfo;
	      setAnimated(true);
		m_AniInfo->m_BaseValue = w;
		return(m_AniInfo->Read(Stream, Flags, version));
    }
	return(UTI_OK);
}*/
void CAnimationClass::PaintWorld()
{

}
/*
int CDoubleClass::GetMessageUti(CUtiOB *Sender, int msg, int lParam, void* wParam, bool &handled)
{
#ifdef EDITOR
	if ((msg ==OnChange) &&(dynamic_cast<CUtiEdit*>(Sender))){
       CUtiEdit* edit = (CUtiEdit*)Sender;
	   CUtiString text;

	   edit->getText(text);
	   setWert(text.ToFloat());
	   PaintWorld();


       return(0);
	};
	if ((msg ==OnClick) &&(dynamic_cast<CUtiButton*>(Sender)))
	{
       CUtiButton* button = (CUtiButton*)Sender;
       CUtiEdit* edit = (CUtiEdit*)button->tagOB;
	   if (button->tag != 0) {
	      setWert(Wert+button->tag);
          setEditText(edit);
	      //PaintWorld();
	   } else {
		 // #ifdef EDITOR
		   if (!dedit) {
			   dedit = new CDoubleEditClass(edit->Application);
			   ((CDoubleEditClass*)dedit)->Init(this);
		   };
		  ((CDoubleEditClass*)dedit)->Dialog->Show();
 	//#endif

	   };
	};
	#endif
	return(0);
}
#ifdef EDITOR
void CDoubleClass::setEditText(CUtiEdit *edit)
{
	if (!edit) return;
	U_CHAR c[256];
	float f = (float)Wert;
	swprintf(c, 256, UTEXT("%f"), f);
    CUtiString number=c;

		edit->setText(number);
}

void CDoubleClass::CreateEdit(U_CHAR*c, CUtiControl *parent, int x, int y, int width, int height, int textw)
{
    CUtiEdit* edit = new CUtiEdit(parent->Application, parent, 1);
	CUtiLabel* label = new CUtiLabel(parent->Application, parent, -1);
	label->setClientEdge(true);
	label->Create(c, x-textw, y, textw-1, height-1);

	int buttonsize = (int)(height / 2);
    edit->Create(UTEXT(""), x, y, width-height, height);
	setEditText(edit);
	edit->Eventhandler = this;
	#ifdef EDITOR
	editfeld = edit;
    #endif

	CUtiButton* button = new CUtiButton(parent->Application, parent);
	button->Create(UTEXT("+"), x+width-height, y, buttonsize, height);
	button->tag = 1;
	button->tagOB = edit;
	button->Eventhandler = this;

	button = new CUtiButton(parent->Application, parent);
	button->Create(UTEXT("-"), x+width-height+buttonsize, y, buttonsize, height);
	button->tag = -1;
	button->tagOB = edit;
	button->Eventhandler = this;

	button = new CUtiButton(parent->Application, parent);
	button->Create(UTEXT("E"), x+width-height+buttonsize*2, y, buttonsize, height);
	button->tag = 0;
	button->tagOB = edit;
	button->Eventhandler = this;

}
#endif
HResult CDoubleClass::Write(CStream &Stream, int Flags, int version)
{
   bool IsAnimated;

   IsAnimated = m_AniInfo.get() != NULL;

   Stream.WriteBoolean(IsAnimated);
   double w = Wert;
   Stream.WriteDouble(w);
   if (IsAnimated) {
       return(m_AniInfo->Write(Stream, Flags, version));
   }
   return(UTI_OK);
}
*/
/*
typ
   ctCountAnimations:
   l: Anzahl der Animationene (kein Pointer)

   ctUpdateCodes
   l: Pointer auf UpdateCodes Struktur

   ctSetCode
   Aktiviert einen Spur
   l: Pointer auf SCODE Struktur

   ctSetCodeZero
   Deaktiviert die aktuelle Spur

   ctFrametime                              )  Wenn l = REPEATALWAYS wird die Spur
   l: Zeit die hinzuaddiert wird            )  immer wiederholt
   time: Zeit in Sekunden                   )
   ctSetTime                                )
   l: Zeit die gesetzt wird                 )

  ctSetController
  l: Pointer zur _con Struktur
  ctAfterRead
*/

void CAnimationClass::ProzessUpdate(SProzess* Update)
{

	CUpdateCodes* Codes;
    CAnimation* Ani;
    TIME newTime;
	if (Update->type == CAnimationClass::ctCountAnimations) {
		if (m_AniInfo.get() != NULL) {
			Update->longdata = Update->longdata + m_AniInfo->getAnimationsCount();
		}
		return;
	}
	if (m_AniInfo.get() == NULL) return;
    TIME time = Update->time;
    //UINT l = Update->longdata;
	int typ = Update->type;
	void* data = Update->data;

	if (typ == CAnimationClass::ctUpdateCodes) {
        assert(false);
        Codes = NULL;//(CUpdateCodes*)(void*)l;
		Ani = findAnimation(Codes->oldCode);
		if (Ani == NULL) return;
		Ani->setCode(Codes->newCode);

		if (Codes->setFlags) Ani->setFlags(Codes->newFlags);
		if (Codes->setLength) {
            assert(false);
//		  Ani->setLength(Codes->newLength);
         //  setLength(Codes->newLength, Ani);
		}
		return;
	}
	if (typ == CAnimationClass::ctSetCode) {
      if (data) {
		Ani = findAnimation(*((SCode*)data));
      } else {
        Ani = NULL;
        if (m_AniInfo->m_Animations.size()) Ani = m_AniInfo->m_Animations.get(0);
      };
		if (Ani == NULL) return;
		if ((m_AniInfo->m_Active == Ani) && (absf(Ani->m_CurrentPosition - Ani->m_TimeUsed) < 0.001)) return;

		activateTrace(Ani);
		return;
	}
	if (typ == CAnimationClass::ctIndex) {
       // if (dynamic_cast<CDoubleClass*>(this)) {
       addIndices(*((SCode*)data));
        //};
		return;
	}
	if (typ == CAnimationClass::ctSetCodeZero) {
		m_AniInfo->setActive(NULL);
		//m_AniInfo->aniIndex = -1;
		return;
	}

	if (m_AniInfo->getActive() == NULL) return;

    newTime = 0;
    Ani = m_AniInfo->getActive();
	if ((typ == CAnimationClass::ctFrametime) || (typ == CAnimationClass::ctSetTime)) {
        if (typ == CAnimationClass::ctFrametime)
			newTime = Ani->m_CurrentPosition + time;
        if (typ == CAnimationClass::ctSetTime)
			newTime = time;
		if (Ani->m_TimeUsed <= newTime) {

            if (Ani->m_Flags == CAnimation::ofRepeat ) {
				newTime = newTime - Ani->m_TimeUsed;
				setTrackPosition(newTime, 1);
			} else {
				newTime = Ani->m_TimeUsed;
				setTrackPosition(newTime, 1);
				m_AniInfo->setActive(NULL);

			}
		} else {
			setTrackPosition(newTime, 1);
		}
	}
}

CVector4 CVector4Class::getVec4()
{
    CVector4 s;  s.x = (utifloat)x.getValue();
    s.y = (utifloat)y.getValue();
    s.z = (utifloat)z.getValue();
    s.w = (utifloat)a.getValue();
	return s;
}


void CVector4Class::setVec4(const CVector4& v)
{
    x.setValue( v.x);
   y.setValue( v.y);
   z.setValue( v.z);
   a.setValue( v.w);
}


void CVector4Class::Set(utifloat _x, utifloat _y, utifloat _z, utifloat _a)
{
   x.setValue( _x);
   y.setValue( _y);
   z.setValue( _z);
   a.setValue( _a);
}

void CVector4Class::Set(UINT rgbvalue)
{
   unsigned char* c = (unsigned char*)&rgbvalue;
   x.setValue( ((utifloat)c[0]) / 255);
   y.setValue( ((utifloat)c[1]) / 255);
   z.setValue( ((utifloat)c[2]) / 255);
   a.setValue(  ((utifloat)c[3]) / 255);

}



void CAnimationClass::setParent(CUtiOB *p, int info)
{
//   Parent = p;
   m_ValueInfo = info;
}

void CVector4Class::setParent(CUtiOB *p, int info)
{
   x.setParent(p, info);
   y.setParent(p, info);
   z.setParent(p, info);
   a.setParent(p, info);

}

/*procedure ReadVector4(v: TVector4Class; Stream: TStream; version: integer);
var
   IsAnimate: Boolean;
   v1: TVector4;
begin
    Stream.ReadBuffer(IsAnimate, sizeof(IsAnimate));
    if IsAnimate then begin
       ReadDouble(v.x, Stream, version);
      ReadDouble(v.y, Stream, version);
      ReadDouble(v.z, Stream, version);
      ReadDouble(v.w, Stream, version);
    end else begin
       Stream.ReadBuffer(v1, sizeof(v1));
       v.setWert(v1);
    end;
end;*/


HResult CVector4Class::ReadStream(CStream & Stream, int Flags, int version)
{
    x.Read(Stream, Flags, version);
        y.Read(Stream, Flags, version);
    z.Read(Stream, Flags, version);
        a.Read(Stream, Flags, version);
        if ((x.getAnimated()) || (y.getAnimated()) || (z.getAnimated()) || (a.getAnimated())) {
           animated = true;
        } else {
            animated = false;
        }
        return 0;
}
HResult CVector4Class::WriteStream(CStream & Stream, int Flags, int version)
{
    x.Write(Stream, Flags, version);
        y.Write(Stream, Flags, version);
        z.Write(Stream, Flags, version);
   a.Write(Stream, Flags, version);
   return 0;
}

HResult CVector4Class::Read(CStream & Stream, int Flags, int version)
{
   bool b=false;
   Stream.ReadData(&b, 1);
   if (!b) {
      SVector4 v;
	  Stream.ReadData(&v, 16);
      /*x.setAnimated(false);
	  y.setAnimated(false);
	  z.setAnimated(false);
      a.setAnimated(false);*/
          if (!(utiisnan(v.x) == false && utiisfinite(v.x))) {
              v.x = 0;
              assert(false);
          }
          if (!(utiisnan(v.y) == false && utiisfinite(v.y))) {
              v.y = 0;
              assert(false);
          }
          if (!(utiisnan(v.z) == false && utiisfinite(v.z))) {
              v.z = 0;
              assert(false);
          }
          if (!(utiisnan(v.a) == false && utiisfinite(v.a))) {
              v.a = 0;
              assert(false);
          }
      x.setValue(v.x);
      y.setValue(v.y);
      z.setValue(v.z);
      a.setValue(v.a);
   } else {
    x.Read(Stream, Flags, version);
	y.Read(Stream, Flags, version);
    z.Read(Stream, Flags, version);
	a.Read(Stream, Flags, version);
   };
   if ((x.getAnimated()) || (y.getAnimated()) || (z.getAnimated()) || (a.getAnimated())) {
      animated = true;
   } else {
       animated = false;
   }
   return(UTI_OK);
}

/*
 * procedure WriteVector4(v: TVector4Class; Stream: TStream; version: integer);
var
   IsAnimate: Boolean;
   v1: TVector4;
begin
   IsAnimate := v.x.Animated OR  v.Y.Animated or  v.z.Animated or  v.w.Animated;
   Stream.Write(IsAnimate, sizeof(IsAnimate));
   if (IsAnimate) then begin
      WriteDouble(v.x, Stream, version);
      WriteDouble(v.y, Stream, version);
      WriteDouble(v.z, Stream, version);
      WriteDouble(v.w, Stream, version);
   end else begin
      v1 := v.Wert;
      Stream.Write(v1, sizeof(v1));
   end;

end;
*/

HResult CVector4Class::Write(CStream & Stream, int Flags, int version)
{
    if ((x.getAnimated()) || (y.getAnimated()) || (z.getAnimated()) || (a.getAnimated())) {
       animated = true;
    } else {
        animated = false;
    }
   // bool IsAnimate := animated;
//Stream.Write(IsAnimate, sizeof(IsAnimate));
    bool b = animated;
    Stream.WriteBoolean(b);
   if (b) {
      SVector4 v;
      v.x = (float)x.getValue();
      v.y = (float)y.getValue();
      v.z = (float)z.getValue();
      v.a = (float)a.getValue();
	  Stream.WriteData(&v, 16);

   } else {
    x.Write(Stream, Flags, version);
	y.Write(Stream, Flags, version);
	z.Write(Stream, Flags, version);
   a.Write(Stream, Flags, version);
   };
   return(UTI_OK);
}

void CVector4Class::ProzessUpdate(SProzess* Update)
{
	if (!animated) return;
   x.ProzessUpdate(Update);
   y.ProzessUpdate(Update);
   z.ProzessUpdate(Update);
   a.ProzessUpdate(Update);
}

void CVector4Class::Get(S3DPoint* p)
{
    p->x = (float)x.getValue();
    p->y = (float)y.getValue();
    p->z = (float)z.getValue();
}

void CVector4Class::Get(CVector* p)
{
    p->x = (utifloat)x.getValue();
    p->y = (utifloat)y.getValue();
    p->z = (utifloat)z.getValue();
}

void CVector4Class::getWert(SVector4 &s)
{
    s.x = (float)x.getValue();
    s.y = (float)y.getValue();
    s.z = (float)z.getValue();
    s.a = (float)a.getValue();
}

void C3DTriple::setParent(CUtiOB* Parent)
{
   x.setParent(Parent, 0);
   y.setParent(Parent, 0);
   z.setParent(Parent, 0);
}

void C3DTriple::VectorToAngle(CVector v)
{
//   C3DObject* obj = (C3DObject*)x.Parent;
 //  C3DEngine* eng = obj->getMain()->SoftEngine;
   utifloat w, l;
   l = (utifloat)sqrt(v.x*v.x + v.z*v.z);
   if (l != 0) {
           w = (utifloat)asin(v.z/l)/(PI / 180);
	   if (v.x < 0) {
		   w = 180-w;
	   };
	   w = 90-w;
       y.setValue(w);
   } else {
       y.setValue(0);
   };

   CMatrix m;
   m.LoadIdentity();
   m.Rotate(0, (utifloat)y.getValue(), 0);
   m.Transform(&v);
   /*eng->PushMatrix();
   eng->LoadIdentity();
   eng->Rotate(0, (float)y.Wert, 0);
   eng->UnTransform(&v);
   eng->PopMatrix();*/
   v.normalize();
   w = (utifloat)acos(v.z)/(PI/180);
   if (v.y < 0) {
	   w = 360-w;
   };
   x.setValue(-w);
   z.setValue(0);
};
/*
void C3DTriple::setVec3(const CVector3& v)
{
  x.setValue(v.x);
  y.setValue(v.y);
  z.setValue(v.z);
}
*//*
CVector3 C3DTriple::getVec3()
{
    return CVector3(x.getValue(), y.getValue(), z.getValue());
}
*/
HResult C3DTriple::Read(CStream & Stream, int Flags, int version)
{
   x.Read(Stream, Flags, version);
   y.Read(Stream, Flags, version);
   z.Read(Stream, Flags, version);
   if ((x.getAnimated()) || (y.getAnimated()) || (z.getAnimated())) {
      animated = true;
   };
   return(0);
}

HResult C3DTriple::Write(CStream & Stream, int Flags, int version)
{
   x.Write(Stream, Flags, version);
   y.Write(Stream, Flags, version);
   z.Write(Stream, Flags, version);
   return(0);
}

void C3DTriple::ProzessUpdate(SProzess* Update)
{
	if (!animated) return;
   x.ProzessUpdate(Update);
   y.ProzessUpdate(Update);
   z.ProzessUpdate(Update);
}

void C3DTriple::set(utifloat _x, utifloat _y, utifloat _z)
{
    x.setValue(_x);
    y.setValue(_y);
    z.setValue(_z);
}

void C3DTriple::set(const CVector& v)
{
  x.setValue(v.x);
    y.setValue(v.y);
    z.setValue(v.z);
};

void C3DTriple::set(S3DPoint *p)
{
    x.setValue(p->x);
    y.setValue(p->y);
    z.setValue(p->z);
}

CAnimationFilter::CAnimationFilter()
{
	komma = 0;
	faktor = 1;
}

void CAnimationFilter::Filter(SProzess* pnew, SProzess* pold)
{
	//float t = ((float)pold->l*faktor)+komma;
	//int t1 = (int)t;
	//komma = t-(float)t1;
	pnew->longdata = pold->longdata;
	pnew->time = pold->time*faktor;
	pnew->type = pold->type;
}

COrderClass::COrderClass() : CAnimationClass()
{
	Orders.deleteItems = false;
	Orders.capacity = 20;
}

COrderClass::~COrderClass()
{

}

CAniPoint* COrderClass::createAniPoint()
{
    return new CAniOrderPoint();
}


HResult COrderClass::Write(CStream & Stream, int Flags, int version)
{
	bool IsAnimated;

   IsAnimated = m_AniInfo.get() != NULL;

   if ((Flags & SAVE_NO_ANIMATIONS) != 0) {
       IsAnimated = false;
   }

   Stream.WriteBoolean(IsAnimated);
   if (IsAnimated) {
       return(m_AniInfo->Write(Stream, Flags, version));
   }
   return(UTI_OK);
}


HResult COrderClass::Read(CStream & Stream, int Flags, int version)
{
  bool IsAnimated;

	if (m_AniInfo.get() != NULL) m_AniInfo=NULL;

	Stream.ReadData(&IsAnimated, 1);
	Orders.Clear();
	if (IsAnimated) {
		setAnimated(true);
		return(m_AniInfo->Read(Stream, Flags, version));
	}
	return(0);

}

void COrderClass::setWert(double d)
{
   //nix
}

void COrderClass::setTrackPosition(TIME time, int flags)
{
   	TIME Start, Ende;
	Orders.Clear();
   if (!m_AniInfo.get()) return;
   if (!m_AniInfo->m_Active.get()) return;
 //  dout << UTEXT("Order: ") << time << UTEXT("f") << flags << UTEXT("\n");
	Start = m_AniInfo->m_Active->m_CurrentPosition;
	Ende = time;
	m_AniInfo->m_Active->m_CurrentPosition = time;
	if (flags == 0) return;
	if (Start <= Ende) {
	   AddSegment(Start, Ende);
	} else {
	   AddSegment(Start, m_AniInfo->m_Active->m_TimeUsed);//nicht minus 1 !!!!
	   AddSegment(0, Ende);
	};
}

void COrderClass::AddSegment(TIME Start, TIME Ende)
{
	CAnimation* a = m_AniInfo->m_Active.get();
    CAnimationDataPoints* data = check_cast<CAnimationDataPoints*>(a->getData());
    for (int i = 0; i < data->m_Points.size(); i++) {
      CAniOrderPoint* p = (CAniOrderPoint*)&data->m_Points[i];
      if (p->getTime() >= Start && p->getTime() < Ende) {
	      Orders.Add(p);
	   };
	};

}
/*
CAnimation *COrderClass::CreateNewAnimation(int newTime, SCode newCode)
{
	if (m_AniInfo == NULL) {
		m_AniInfo = new CAnimationInfo;
		m_AniInfo->usedouble = false;
	}
	m_AniInfo->aniIndex = m_AniInfo->Animations.count;
	CAnimation *newIt =
	new CAnimation(newTime, newCode);
	newIt->TimeUsed = newTime;
	newIt->m_Code = newCode;
	newIt->usedouble = false;
	m_AniInfo->Animations.Add(newIt);
	return(newIt);

}*/

//$cppoutput::begin
CUtiString CAniPoint::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CAniPoint");
}
TIME CAniPoint::getTime() const
{
  return m_Time;
}
void CAniPoint::setTime(const TIME& value0)
{
  CHECKEXISTANCE;
   m_Time = value0;
}
const SCode CAniPoint::TYPECODE =
 {
    2003, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CAniPoint::GetTypeCode()
{
    return CAniPoint::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CAniPoint);
void CAniPoint::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xread(n, wri);
   uti_readValue(m_Time, n, wri, UTISTR("Time"), 0.0f);
   ;
}
void CAniPoint::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xwrite(n, wri);
   uti_writeValue(m_Time, n, wri, UTISTR("Time"));
}
CUtiString CAniDoublePoint::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CAniDoublePoint");
}
utifloat CAniDoublePoint::getValue() const
{
CHECKEXISTANCE;   return m_Value;
}
void CAniDoublePoint::setValue(const utifloat& value0)
{
  CHECKEXISTANCE;
   m_Value = value0;
}
const SCode CAniDoublePoint::TYPECODE =
 {
    2002, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CAniDoublePoint::GetTypeCode()
{
    return CAniDoublePoint::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CAniDoublePoint);
void CAniDoublePoint::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   CAniPoint::xread(n, wri);
   uti_readValue(m_Value, n, wri, UTISTR("Value"), 0.0f);
   ;
}
void CAniDoublePoint::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   CAniPoint::xwrite(n, wri);
   uti_writeValue(m_Value, n, wri, UTISTR("Value"));
}
CUtiString CAnimationDataSource::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CAnimationDataSource");
}
CAniPoint* CAnimationDataSource::getPoints(int index)
{
CHECKEXISTANCE;    return &m_Points[index];
}
int CAnimationDataSource::getPointsCount()
{
 CHECKEXISTANCE;   return m_Points.size();
}
void CAnimationDataSource::addPoints(CAniPoint* value)
{
 CHECKEXISTANCE;   m_Points.push_back(value);
}
void CAnimationDataSource::insertPoints(int index, CAniPoint* value)
{
 CHECKEXISTANCE;   m_Points.insert(index, value);
}
void CAnimationDataSource::clearPoints()
{ CHECKEXISTANCE; 
   return m_Points.clear();
}
void CAnimationDataSource::deletePoints(int index)
{ CHECKEXISTANCE; 
   m_Points.deleteItem(index);
}
void CAnimationDataSource::removePoints(CAniPoint* value)
{ CHECKEXISTANCE; 
   m_Points.remove(value);
}
 int CAnimationDataSource::indexOfPoints(CAniPoint* value)
{ CHECKEXISTANCE; 
   return m_Points.indexOf(value);
}
const SCode CAnimationDataSource::TYPECODE =
 {
    2340, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CAnimationDataSource::GetTypeCode()
{
    return CAnimationDataSource::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CAnimationDataSource);
void CAnimationDataSource::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xread(n, wri);
   uti_readValue(new CListReferenceAdapter<CAniPoint>(m_Points), n, wri, UTISTR("Points"));
   ;
}
void CAnimationDataSource::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xwrite(n, wri);
   uti_writeValue(new CListReferenceAdapter<CAniPoint>(m_Points), n, wri, UTISTR("Points"));
}
CAnimationDataPoints::CAnimationDataPoints()
{
   ;
   ;
}
CAnimationDataPoints::~CAnimationDataPoints()
{
   ;
}
CUtiString CAnimationDataPoints::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CAnimationDataPoints");
}
CAniPoint* CAnimationDataPoints::getPoints(int index)
{
CHECKEXISTANCE;    return &m_Points[index];
}
int CAnimationDataPoints::getPointsCount()
{
 CHECKEXISTANCE;   return m_Points.size();
}
void CAnimationDataPoints::addPoints(CAniPoint* value)
{
 CHECKEXISTANCE;   m_Points.push_back(value);
}
void CAnimationDataPoints::insertPoints(int index, CAniPoint* value)
{
 CHECKEXISTANCE;   m_Points.insert(index, value);
}
void CAnimationDataPoints::clearPoints()
{ CHECKEXISTANCE; 
   return m_Points.clear();
}
void CAnimationDataPoints::deletePoints(int index)
{ CHECKEXISTANCE; 
   m_Points.deleteItem(index);
}
void CAnimationDataPoints::removePoints(CAniPoint* value)
{ CHECKEXISTANCE; 
   m_Points.remove(value);
}
 int CAnimationDataPoints::indexOfPoints(CAniPoint* value)
{ CHECKEXISTANCE; 
   return m_Points.indexOf(value);
}
const SCode CAnimationDataPoints::TYPECODE =
 {
    2341, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CAnimationDataPoints::GetTypeCode()
{
    return CAnimationDataPoints::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CAnimationDataPoints);
void CAnimationDataPoints::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   CAnimationDataSource::xread(n, wri);
   uti_readValue(new CListReferenceAdapter<CAniPoint>(m_Points), n, wri, UTISTR("Points"));
   ;
}
void CAnimationDataPoints::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   CAnimationDataSource::xwrite(n, wri);
   uti_writeValue(new CListReferenceAdapter<CAniPoint>(m_Points), n, wri, UTISTR("Points"));
}
CAnimationSubPart::CAnimationSubPart()
{
   m_Data = NULL;
   m_Start = 0.0f;
   m_Length = 1.0f;
   ;
   ;
}
CAnimationSubPart::~CAnimationSubPart()
{
   ;
}
CUtiString CAnimationSubPart::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CAnimationSubPart");
}
CAnimationDataSource* CAnimationSubPart::getData()
{
 CHECKEXISTANCE;   return m_Data.get();
}
const CAnimationDataSource* CAnimationSubPart::getData() const
{
 CHECKEXISTANCE;   return m_Data.get();
}
void CAnimationSubPart::setData(CAnimationDataSource* value0)
{
  CHECKEXISTANCE;
   m_Data = value0;
}
TIME CAnimationSubPart::getStart() const
{
CHECKEXISTANCE;   return m_Start;
}
void CAnimationSubPart::setStart(const TIME& value0)
{
  CHECKEXISTANCE;
   m_Start = value0;
}
TIME CAnimationSubPart::getLength() const
{
CHECKEXISTANCE;   return m_Length;
}
void CAnimationSubPart::setLength(const TIME& value0)
{
  CHECKEXISTANCE;
   m_Length = value0;
}
const SCode CAnimationSubPart::TYPECODE =
 {
    2342, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CAnimationSubPart::GetTypeCode()
{
    return CAnimationSubPart::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CAnimationSubPart);
void CAnimationSubPart::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   CAnimationDataSource::xread(n, wri);
   uti_readValue(new CReferenceAdapter<CAnimationDataSource>(m_Data), n, wri, UTISTR("Data"), NULL);
   uti_readValue(m_Start, n, wri, UTISTR("Start"), 0.0f);
   uti_readValue(m_Length, n, wri, UTISTR("Length"), 1.0f);
   ;
}
void CAnimationSubPart::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   CAnimationDataSource::xwrite(n, wri);
   uti_writeValue(new CReferenceAdapter<CAnimationDataSource>(m_Data), n, wri, UTISTR("Data"));
   uti_writeValue(m_Start, n, wri, UTISTR("Start"));
   uti_writeValue(m_Length, n, wri, UTISTR("Length"));
}
CUtiString CAnimation::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CAnimation");
}
TIME CAnimation::getCurrentPosition() const
{
CHECKEXISTANCE;   return m_CurrentPosition;
}
void CAnimation::setCurrentPosition(const TIME& value0)
{
  CHECKEXISTANCE;
   m_CurrentPosition = value0;
}
TIME CAnimation::getTimeUsed() const
{
CHECKEXISTANCE;   return m_TimeUsed;
}
void CAnimation::setTimeUsed(const TIME& value0)
{
  CHECKEXISTANCE;
   m_TimeUsed = value0;
}
CAnimationDataSource* CAnimation::getData()
{
 CHECKEXISTANCE;   return m_Data.get();
}
const CAnimationDataSource* CAnimation::getData() const
{
 CHECKEXISTANCE;   return m_Data.get();
}
void CAnimation::setData(CAnimationDataSource* value0)
{
  CHECKEXISTANCE;
   m_Data = value0;
}
CAnimationClass* CAnimation::getAnimationClass()
{
 CHECKEXISTANCE;   return m_AnimationClass;
}
const CAnimationClass* CAnimation::getAnimationClass() const
{
 CHECKEXISTANCE;  return m_AnimationClass;
}
void CAnimation::setAnimationClass(CAnimationClass* value0)
{  CHECKEXISTANCE;

   m_AnimationClass = value0;
}
SCode CAnimation::getCode() const
{
CHECKEXISTANCE;   return m_Code;
}
void CAnimation::setCode(const SCode& value0)
{
  CHECKEXISTANCE;
   m_Code = value0;
}
int CAnimation::getFlags() const
{
CHECKEXISTANCE;   return m_Flags;
}
void CAnimation::setFlags(const int& value0)
{
  CHECKEXISTANCE;
   m_Flags = value0;
}
const SCode CAnimation::TYPECODE =
 {
    2001, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CAnimation::GetTypeCode()
{
    return CAnimation::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CAnimation);
void CAnimation::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xread(n, wri);
   uti_readValue(m_CurrentPosition, n, wri, UTISTR("CurrentPosition"), 0.0f);
   uti_readValue(m_TimeUsed, n, wri, UTISTR("TimeUsed"), 0.0f);
   uti_readValue(new CReferenceAdapter<CAnimationDataSource>(m_Data), n, wri, UTISTR("Data"), NULL);
   uti_readValue((CUtiOB**)&m_AnimationClass, n, wri, UTISTR("AnimationClass"));
   uti_readValue(m_Code, n, wri, UTISTR("Code"), NULLCODE);
   uti_readValue(m_Flags, n, wri, UTISTR("Flags"), 0);
   ;
}
void CAnimation::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xwrite(n, wri);
   uti_writeValue(m_CurrentPosition, n, wri, UTISTR("CurrentPosition"));
   uti_writeValue(m_TimeUsed, n, wri, UTISTR("TimeUsed"));
   uti_writeValue(new CReferenceAdapter<CAnimationDataSource>(m_Data), n, wri, UTISTR("Data"));
   uti_writeValue((CUtiOB**)&m_AnimationClass, n, wri, UTISTR("AnimationClass"));
   uti_writeValue(m_Code, n, wri, UTISTR("Code"));
   uti_writeValue(m_Flags, n, wri, UTISTR("Flags"));
}
CUtiString CAnimationInfo::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CAnimationInfo");
}
CAnimation* CAnimationInfo::getAnimations(int index)
{
CHECKEXISTANCE;    return &m_Animations[index];
}
int CAnimationInfo::getAnimationsCount()
{
 CHECKEXISTANCE;   return m_Animations.size();
}
void CAnimationInfo::addAnimations(CAnimation* value)
{
 CHECKEXISTANCE;   m_Animations.push_back(value);
}
void CAnimationInfo::insertAnimations(int index, CAnimation* value)
{
 CHECKEXISTANCE;   m_Animations.insert(index, value);
}
void CAnimationInfo::clearAnimations()
{ CHECKEXISTANCE; 
   return m_Animations.clear();
}
void CAnimationInfo::deleteAnimations(int index)
{ CHECKEXISTANCE; 
   m_Animations.deleteItem(index);
}
void CAnimationInfo::removeAnimations(CAnimation* value)
{ CHECKEXISTANCE; 
   m_Animations.remove(value);
}
 int CAnimationInfo::indexOfAnimations(CAnimation* value)
{ CHECKEXISTANCE; 
   return m_Animations.indexOf(value);
}
CAnimationClass* CAnimationInfo::getAnimationClass()
{
 CHECKEXISTANCE;   return m_AnimationClass;
}
const CAnimationClass* CAnimationInfo::getAnimationClass() const
{
 CHECKEXISTANCE;  return m_AnimationClass;
}
void CAnimationInfo::setAnimationClass(CAnimationClass* value0)
{  CHECKEXISTANCE;

   m_AnimationClass = value0;
}
CAnimation* CAnimationInfo::getActive()
{
 CHECKEXISTANCE;   return m_Active.get();
}
const CAnimation* CAnimationInfo::getActive() const
{
 CHECKEXISTANCE;   return m_Active.get();
}
void CAnimationInfo::setActive(CAnimation* value0)
{
  CHECKEXISTANCE;
   m_Active = value0;
}
const SCode CAnimationInfo::TYPECODE =
 {
    2004, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CAnimationInfo::GetTypeCode()
{
    return CAnimationInfo::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CAnimationInfo);
void CAnimationInfo::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xread(n, wri);
   uti_readValue(new CListReferenceAdapter<CAnimation>(m_Animations), n, wri, UTISTR("Animations"));
   uti_readValue((CUtiOB**)&m_AnimationClass, n, wri, UTISTR("AnimationClass"));
   uti_readValue(new CReferenceAdapter<CAnimation>(m_Active), n, wri, UTISTR("Active"), NULL);
   ;
}
void CAnimationInfo::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xwrite(n, wri);
   uti_writeValue(new CListReferenceAdapter<CAnimation>(m_Animations), n, wri, UTISTR("Animations"));
   uti_writeValue((CUtiOB**)&m_AnimationClass, n, wri, UTISTR("AnimationClass"));
   uti_writeValue(new CReferenceAdapter<CAnimation>(m_Active), n, wri, UTISTR("Active"));
}
CUtiString CAnimationClass::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CAnimationClass");
}
CAnimation* CAnimationClass::getAnimations(int index)
{
CHECKEXISTANCE;    return &m_Animations[index];
}
int CAnimationClass::getAnimationsCount()
{
 CHECKEXISTANCE;   return m_Animations.size();
}
void CAnimationClass::addAnimations(CAnimation* value)
{
 CHECKEXISTANCE;   m_Animations.push_back(value);
}
void CAnimationClass::insertAnimations(int index, CAnimation* value)
{
 CHECKEXISTANCE;   m_Animations.insert(index, value);
}
void CAnimationClass::clearAnimations()
{ CHECKEXISTANCE; 
   return m_Animations.clear();
}
void CAnimationClass::deleteAnimations(int index)
{ CHECKEXISTANCE; 
   m_Animations.deleteItem(index);
}
void CAnimationClass::removeAnimations(CAnimation* value)
{ CHECKEXISTANCE; 
   m_Animations.remove(value);
}
 int CAnimationClass::indexOfAnimations(CAnimation* value)
{ CHECKEXISTANCE; 
   return m_Animations.indexOf(value);
}
int CAnimationClass::getValueInfo() const
{
CHECKEXISTANCE;   return m_ValueInfo;
}
void CAnimationClass::setValueInfo(const int& value0)
{
  CHECKEXISTANCE;
   m_ValueInfo = value0;
}
CAnimationInfo* CAnimationClass::getAniInfo()
{
 CHECKEXISTANCE;   return m_AniInfo.get();
}
const CAnimationInfo* CAnimationClass::getAniInfo() const
{
 CHECKEXISTANCE;   return m_AniInfo.get();
}
void CAnimationClass::setAniInfo(CAnimationInfo* value0)
{
  CHECKEXISTANCE;
   m_AniInfo = value0;
}
const SCode CAnimationClass::TYPECODE =
 {
    2005, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CAnimationClass::GetTypeCode()
{
    return CAnimationClass::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CAnimationClass);
void CAnimationClass::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xread(n, wri);
   uti_readValue(new CListReferenceAdapter<CAnimation>(m_Animations), n, wri, UTISTR("Animations"));
   uti_readValue(m_ValueInfo, n, wri, UTISTR("ValueInfo"));
   uti_readValue(new CReferenceAdapter<CAnimationInfo>(m_AniInfo), n, wri, UTISTR("AniInfo"), NULL);
   ;
}
void CAnimationClass::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xwrite(n, wri);
   uti_writeValue(new CListReferenceAdapter<CAnimation>(m_Animations), n, wri, UTISTR("Animations"));
   uti_writeValue(m_ValueInfo, n, wri, UTISTR("ValueInfo"));
   uti_writeValue(new CReferenceAdapter<CAnimationInfo>(m_AniInfo), n, wri, UTISTR("AniInfo"));
}
//$cppoutput::end
