/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef __UTI3DANIMATIONS__
#define __UTI3DANIMATIONS__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "UtiType.h"
#include "UtiTypeHandlers.h"
#include "utiadvarray.h"
//#include "undosystem.h"
//#include "undosystemlist.h"

struct SProzess
{   int type;
    TIME time;
    UINT longdata;
	void* data;
	SProzess() {data = NULL; longdata = 0; time = 0.0f; type = 0;};
};

//#include "UtiApp.h"

//#define EDITOR
#ifdef EDITOR
class CUtiControl;
class CUtiEdit;
#endif
const int REPEATALWAYS = 1;
const int chColor = 100;
const int chShininess = 101;
const int chYesNo = 102;
const int chPosition = 103;
const int chAngle = 104;


const int SAVE_NO_ANIMATIONS = 1 << 5;

struct SAniOrderPoint
{
	int value;
	int flags;
	int time;
	int info;
    int d1, d2, d3, d4;
};

struct CUpdateCodes
{
	SCode oldCode;
	SCode newCode;
	bool setFlags;
    unsigned short int newFlags;
    bool setLength;
	TIME newLength;
};

class CAnimationInfo;
class CAnimationClass;

class CAniPoint : public ReferenceUtiOB
{
  friend class CDoubleClass;
   //$classbody::begin CAniPoint(ancestor="ReferenceUtiOB" serialize="2003")

       //$property Time(type="TIME" init="0.0f" public)
   //$classbody::middle
public: CUtiString getClassName();
public: TIME m_Time;
public: TIME getTime() const;
public: void setTime(const TIME& value0);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end
public:

	CAniPoint();

	HResult Read(CStream & Stream, int Flags, int version);
	HResult Write(CStream & Stream, int Flags, int version);
};

class CAniDoublePoint : public CAniPoint
{
  friend class CDoubleClass;
   //$classbody::begin CAniDoublePoint(ancestor="CAniPoint" serialize="2002")

       //$property Value(type="double" init="0.0f")
   //$classbody::middle
public: CUtiString getClassName();
protected: utifloat m_Value;
public: utifloat getValue() const;
public: void setValue(const utifloat& value0);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end
public:

	CAniDoublePoint();
	HResult Read(CStream & Stream, int Flags, int version);
	HResult Write(CStream & Stream, int Flags, int version);
};

template<class T> class CAniBasePoint : public CAniPoint
{
  public:
    CAniBasePoint() : CAniPoint()
    {

    };
    virtual T getStartPoint() const
    {
       return T();
    }
    virtual T getStartTangent(CAniBasePoint<T>& next) const
    {
       return T();
    }
    virtual T getValue(TIME t, CAniBasePoint<T>& next) const
    {
       return T();
    }
};

template<class T> class CLinearInterpolationPoint : public CAniBasePoint<T>
{
  protected:

    T m_value;

  public:
    CLinearInterpolationPoint() : CAniBasePoint<T>()
    {
      m_value = T();
    }
    virtual T getStartPoint() const
    {
       return m_value;
    }
    virtual T getStartTangent(CAniBasePoint<T>& next) const
    {

       return T(0);
    }
    inline T getValue() const
    {
       return m_value;
    }
    inline void setValue(const T &time)
    {
      m_value = time;
    }
    virtual T getValue(TIME t, CAniBasePoint<T>& next) const
    {
        T nextvalue = next.getStartPoint();
	TIME time = (next.getTime()-this->getTime());
	utiassert(time >= 0);
	TIME factor = t / time;
       return m_value*(1.0-factor)+nextvalue*factor;
    }
    public: virtual void xread(CXNode* n, CWriteReadInterface* wri)
    {
       CAniBasePoint<T>::xread(n, wri);
       uti_readValue(m_value, n, wri, UTISTR("Value"), T(0));
    }
    public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri)
    {
      CAniBasePoint<T>::xwrite(n, wri);
      uti_writeValue(m_value, n, wri, UTISTR("Value"));
    }
};


template<class T> class CLinearInterpolationPointNoWatch : public CAniBasePoint<T>
{
  protected:

    T m_value;

  public:
    CLinearInterpolationPointNoWatch() : CAniBasePoint<T>()
    {
      m_value = T();
    }
    virtual T getStartPoint() const
    {
       return m_value;
    }
    virtual T getStartTangent(CAniBasePoint<T>& next) const
    {

       return T(0);
    }
    inline T getValue() const
    {
       return m_value;
    }
    inline void setValue(const T &time)
    {
      m_value = time;
    }
    virtual T getValue(TIME t, CAniBasePoint<T>& next) const
    {
        T nextvalue = next.getStartPoint();
    TIME time = (next.getTime()-this->getTime());
    utiassert(time >= 0);
    TIME factor = t / time;
       return m_value*(1.0-factor)+nextvalue*factor;
    }
    public: virtual void xread(CXNode* n, CWriteReadInterface* wri)
    {
       CAniBasePoint<T>::xread(n, wri);
       uti_readValue(m_value, n, wri, UTISTR("Value"), T(0));
    }
    public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri)
    {
      CAniBasePoint<T>::xwrite(n, wri);
      uti_writeValue(m_value, n, wri, UTISTR("Value"));
    }
};


template<class T> class CBezierInterpolationPointNoWatch : public CAniBasePoint<T>
{
  protected:

    T m_value;
    TIME m_tangenttime1;
    T m_tangentvalue1;
    TIME m_tangenttime2;
    T m_tangentvalue2;

  public:
    CBezierInterpolationPointNoWatch() : CAniBasePoint<T>()
    {
      m_value = T();
    }
    virtual T getStartPoint() const
    {
       return m_value;
    }
    virtual T getStartTangent(CAniBasePoint<T>& next) const
    {

       return T(0);
    }
    inline void setTangent1(const CVector2&v)
    {
        m_tangenttime1 = v.x;
        m_tangentvalue1 = v.y;
    }

    inline void setTangent2(const CVector2&v)
    {
        m_tangenttime2 = v.x;
        m_tangentvalue2 = v.y;
    }

    inline T getValue() const
    {
       return m_value;
    }
    inline void setValue(const T &time)
    {
      m_value = time;
    }
    virtual T getValue(TIME t, CAniBasePoint<T>& next) const
    {
        T nextvalue = next.getStartPoint();
    TIME time = (next.getTime()-this->getTime());
    utiassert(time >= 0);
    TIME factor = t / time;
       return m_value*(1.0-factor)+nextvalue*factor;
    }
    public: virtual void xread(CXNode* n, CWriteReadInterface* wri)
    {
       CAniBasePoint<T>::xread(n, wri);
       uti_readValue(m_value, n, wri, UTISTR("Value"), T(0));
       uti_readValue(m_tangentvalue1, n, wri, UTISTR("TangentValue1"), T(0));
       uti_readValue(m_tangentvalue2, n, wri, UTISTR("TangentValue2"), T(0));
        uti_readValue(m_tangenttime1, n, wri, UTISTR("TangentTime1"), 0.0f);
        uti_readValue(m_tangenttime2, n, wri, UTISTR("TangentTime2"), 0.0f);
    }
    public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri)
    {
      CAniBasePoint<T>::xwrite(n, wri);
      uti_writeValue(m_value, n, wri, UTISTR("Value"));
      uti_writeValue(m_tangentvalue1, n, wri, UTISTR("TangentValue1"));
      uti_writeValue(m_tangentvalue2, n, wri, UTISTR("TangentValue2"));
      uti_writeValue(m_tangenttime1, n, wri, UTISTR("TangentTime1"));
      uti_writeValue(m_tangenttime2, n, wri, UTISTR("TangentTime2"));
    }
};

template<class T> class CFixedPointNoWatch : public CAniBasePoint<T>
{
  protected:

    T m_value;

  public:
    CFixedPointNoWatch() : CAniBasePoint<T>()
    {
      m_value = T();
    }
    virtual T getStartPoint() const
    {
       return m_value;
    }
    virtual T getStartTangent(CAniBasePoint<T>& next) const
    {

       return T(0);
    }
    inline T getValue() const
    {
       return m_value;
    }
    inline void setValue(const T &time)
    {
      m_value = time;
    }
    virtual T getValue(TIME t, CAniBasePoint<T>& next) const
    {
       return m_value;
    }
    public: virtual void xread(CXNode* n, CWriteReadInterface* wri)
    {
       CAniBasePoint<T>::xread(n, wri);
       uti_readValue(m_value, n, wri, UTISTR("Value"), T(0));
    }
    public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri)
    {
      CAniBasePoint<T>::xwrite(n, wri);
      uti_writeValue(m_value, n, wri, UTISTR("Value"));
    }
};

template<class T> class CFixedPoint : public CAniBasePoint<T>
{
  protected:

    T m_value;

  public:
    CFixedPoint() : CAniBasePoint<T>()
    {
      m_value = T();
    }
    virtual T getStartPoint() const
    {
       return m_value;
    }
    virtual T getStartTangent(CAniBasePoint<T>& next) const
    {

       return T(0);
    }
    inline T getValue() const
    {
       return m_value;
    }
    inline void setValue(const T &time)
    {
      m_value = time;
    }
    virtual T getValue(TIME t, CAniBasePoint<T>& next) const
    {

       return m_value;
    }
    public: virtual void xread(CXNode* n, CWriteReadInterface* wri)
    {
       CAniBasePoint<T>::xread(n, wri);
       uti_readValue(m_value, n, wri, UTISTR("Value"), T(0));
    }
    public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri)
    {
      CAniBasePoint<T>::xwrite(n, wri);
      uti_writeValue(m_value, n, wri, UTISTR("Value"));
    }
};



#define CDECLAREANITYPE(BaseClass,Type,serial)\
  class BaseClass ## _ ## Type : public BaseClass<Type> \
  {\
   public:\
     BaseClass ## _ ## Type() : BaseClass<Type>() {\
       \
     }\
     public: virtual const SCode GetTypeCode();\
     public: static const SCode TYPECODE;\
  };

#define CDECLAREANITYPE_IMPL(BaseClass,Type,serial)\
  const SCode BaseClass ## _ ## Type::TYPECODE =\
 {\
    serial, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }\
};\
const SCode BaseClass ## _ ## Type::GetTypeCode()\
{\
    return BaseClass ## _ ## Type::TYPECODE;\
}\
IMPLEMENT_CLASS_CREATORTYPENOPARENT(BaseClass ## _ ## Type);


  #define CDECLAREANITYPE_ALL(BaseClass, baseserial)\
  CDECLAREANITYPE(BaseClass, double, baseserial+0)\
  CDECLAREANITYPE(BaseClass, CVector2, baseserial+2)\
  CDECLAREANITYPE(BaseClass, CVector3, baseserial+3)\
  CDECLAREANITYPE(BaseClass, CVector4, baseserial+4)

  #define CDECLAREANITYPE_ALL_IMPL(BaseClass, baseserial)\
  CDECLAREANITYPE_IMPL(BaseClass, double, baseserial+0)\
  CDECLAREANITYPE_IMPL(BaseClass, CVector2, baseserial+2)\
  CDECLAREANITYPE_IMPL(BaseClass, CVector3, baseserial+3)\
  CDECLAREANITYPE_IMPL(BaseClass, CVector4, baseserial+4)

CDECLAREANITYPE_ALL(CLinearInterpolationPoint, 2100);
CDECLAREANITYPE_ALL(CLinearInterpolationPointNoWatch, 12110);
CDECLAREANITYPE_ALL(CFixedPoint, 12120);
CDECLAREANITYPE_ALL(CFixedPointNoWatch, 12130);
CDECLAREANITYPE_ALL(CBezierInterpolationPointNoWatch, 12140);

class CAniOrderPoint : public CAniPoint
{
public:
	int value;
	CAniOrderPoint();
	HResult Read(CStream & Stream, int Flags, int version);
	HResult Write(CStream & Stream, int Flags, int version);

};

struct SAnimationInput
{
    TIME m_time;
};

struct SAnimationResult {

};

template<class T> struct SAnimationResultType : public SAnimationResult
{
    T m_value;
};

class CAnimationDataSource : public ReferenceUtiOB
{
    //$classbody::begin CAnimationDataSource(ancestor="ReferenceUtiOB" serialize="2340")
         //$property Points(type="CAniPoint" init="NULL" mode="array_ref" public)
     //$classbody::middle
public: CUtiString getClassName();
public:  CReferenceList<CAniPoint> m_Points;
public: CAniPoint* getPoints(int index);
public: int getPointsCount();
public: void addPoints(CAniPoint* value);
public: void insertPoints(int index, CAniPoint* value);
public: void removePoints(CAniPoint* value);
public: void clearPoints();
public: void deletePoints(int index);
public: int indexOfPoints(CAniPoint* value);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end
public:
    CAnimationDataSource() {};
    virtual void evaluate(const SAnimationInput& input, SAnimationResult* output) const{assert(false);};
};

class CAnimationDataPoints : public CAnimationDataSource
{
public:
    //$classbody::begin CAnimationDataPoints(ancestor="CAnimationDataSource" serialize="2341"stdconde)
         //$property Points(type="CAniPoint" init="NULL" mode="array_ref" public)

     //$classbody::middle
public: CAnimationDataPoints();
public: ~CAnimationDataPoints();
public: CUtiString getClassName();
public:  CReferenceList<CAniPoint> m_Points;
public: CAniPoint* getPoints(int index);
public: int getPointsCount();
public: void addPoints(CAniPoint* value);
public: void insertPoints(int index, CAniPoint* value);
public: void removePoints(CAniPoint* value);
public: void clearPoints();
public: void deletePoints(int index);
public: int indexOfPoints(CAniPoint* value);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end
public:
    void checkPoint(CAniPoint* p);
    void updatePointPosition(CAniPoint* p);
    CAniPoint* getAniPointByTime(TIME time, TIME tolerance);
    void sortPoints();
};

class CAnimationSubPart : public CAnimationDataSource
{
    //$classbody::begin CAnimationSubPart(ancestor="CAnimationDataSource" serialize="2342" stdconde)
          //$property Data(type="CAnimationDataSource" init="NULL" mode="rw_ref" public)
          //$property Start(type="TIME" init="0.0f")
          //$property Length(type="TIME" init="1.0f")
     //$classbody::middle
public: CAnimationSubPart();
public: ~CAnimationSubPart();
public: CUtiString getClassName();
public: uti_refptr<CAnimationDataSource> m_Data;
public: CAnimationDataSource* getData();
public: const CAnimationDataSource* getData() const;
public: void setData(CAnimationDataSource* value0);
protected: TIME m_Start;
public: TIME getStart() const;
public: void setStart(const TIME& value0);
protected: TIME m_Length;
public: TIME getLength() const;
public: void setLength(const TIME& value0);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end
 public:
    void evaluate(const SAnimationInput& input, SAnimationResult* output) const;
};

const double ANIMATION_ELEMENT_PER_SECOND = 30;
const double ANIMATION_ELEMENT_LENGTH = 1.0/ANIMATION_ELEMENT_PER_SECOND;

enum TAnimationDataState {adsUndecided, adsSpeedUpData, adsNoSpeedUp, adsNull};

template <class T> class CAnimationDataTemplate : public CAnimationDataPoints
{
protected:
    TAnimationDataState m_state;
    Array<T> m_speedupdata;
    TIME getMaxTime() const
    {
        CReferenceList<CAniPoint>& l = const_cast<CAnimationDataTemplate<T>*>( this)->m_Points;
        if (l.size() == 0) {
            return 1.0f;
        }
        return ((CPointClass*)&l[l.size()-1])->m_Time;
    }

    void createSpeedUpData()
    {
        //CReferenceList<CAniPoint>& l = const_cast<CAnimationDataTemplate<T>*>( this)->m_Points;

        TIME maxtime = getMaxTime();
        int c = ANIMATION_ELEMENT_PER_SECOND*maxtime+3.0;
        m_speedupdata.setSize(c);
        SAnimationResultType<T> res;
        SAnimationInput input;
        for (int i = 0; i < c; i++) {
            input.m_time = ANIMATION_ELEMENT_LENGTH*double(i);
            evaluate_intern(input, &res);
            m_speedupdata[i] = res.m_value;

        }
    }

public:
    CAnimationDataTemplate()
    {

        m_state = adsUndecided;
    }
     typedef CAniBasePoint<T> CPointClass;



    void evaluate_intern(const SAnimationInput& input, SAnimationResult* output) const
    {
        SAnimationResultType<T>* res = (SAnimationResultType<T>*)output;


        CReferenceList<CAniPoint>& l = const_cast<CAnimationDataTemplate<T>*>( this)->m_Points;
      CPointClass* p1 = NULL;
      CPointClass* p2 = NULL;

      if (l.size() == 0) {
          res->m_value =  T(0);
          return;
      }


      #define POINT(i) ((CPointClass*)&l[i])
      p1 = POINT(l.size()-1);
      TIME time = input.m_time;
      if (time > p1->m_Time) {

       res->m_value =  p1->getStartPoint();
       return;
      }
       p1 = POINT(0);
      if (time < p1->m_Time) {

          res->m_value =  p1->getStartPoint();
          return;
      }


          if (l.size() < 2) {
              p1 = POINT(0);
              res->m_value =  p1->getStartPoint();
              return;

          }
         int min = 0;
         int max = l.size()-1;

         while (!(POINT(min)->m_Time  <= time && POINT(min+1)->m_Time  >= time) && min != max-1) {
            int middle = (min+max) / 2;
            if (POINT(middle)->m_Time  > time) {
                max = middle;
            } else {
               min = middle;
            };
         };
      p1 = POINT(min);
         p2 = POINT(min+1);
         utiassert(p1->m_Time <= time && p2->m_Time >= time);
         utiassert(min >= 0);
         utiassert((size_t)min < l.size()-1);

        // p1 = POINT(min);
        // p2 = POINT(min+1);
         #undef POINT

    if ((p1 == NULL) && (p2 == NULL)) {
         res->m_value =  T() ;
          return;
    }
        if ((p1 == NULL) && (p2 != NULL)) {
            res->m_value =  p2->getStartPoint();
            return;
        }
    if ((p1 != NULL) && (p1->m_Time  == time)) {
        res->m_value =  p1->getStartPoint();
        return;
    }
    if ((p2 != NULL) && (p2->m_Time  == time)) {
        res->m_value =  p2->getStartPoint();
        return;
    }
    res->m_value =  p1->getValue(time-p1->getTime(), *p2);
    return;
    }

    virtual void evaluate(const SAnimationInput& input, SAnimationResult* output) const
    {
        SAnimationResultType<T>* res = (SAnimationResultType<T>*)output;

        switch (m_state) {
        case adsUndecided: {

            CReferenceList<CAniPoint>& l = const_cast<CAnimationDataTemplate<T>*>( this)->m_Points;

          if (l.size() == 0) {
             const_cast<CAnimationDataTemplate*>(this)->m_state = adsNull;

          }else if (getMaxTime() < 8.0f) {
              const_cast<CAnimationDataTemplate*>(this)->createSpeedUpData();
              const_cast<CAnimationDataTemplate*>(this)->m_state = adsSpeedUpData;

          } else
          {
              const_cast<CAnimationDataTemplate*>(this)->m_state = adsNoSpeedUp;
          }
            evaluate_intern(input, output);
        }; break;
        case adsSpeedUpData: {
            double intime = input.m_time*ANIMATION_ELEMENT_PER_SECOND;
            int myindex = bintrunc(intime);
            double subtime = intime - myindex;
            int s = m_speedupdata.size()-1;
            if (myindex < 0 || myindex >= s) {
                subtime = 0;
                if (myindex < 0) myindex =0;
                if (myindex >= s) myindex = s-1;
            }
            T resvalue = m_speedupdata[myindex]*(1-subtime)+ m_speedupdata[myindex+1]*subtime;

/*#ifndef NDEBUG

            evaluate_intern(input, output);
            assert(absf(res->m_value-resvalue) < 0.2f);
#endif*/

            res->m_value = resvalue;
        }; break;
        case adsNoSpeedUp: {

            evaluate_intern(input, output);
        }; break;
        case adsNull: {
             res->m_value =  T(0);
        }; break;
        }

    }
};


CDECLAREANITYPE_ALL(CAnimationDataTemplate, 12280);

class CAnimation : public ReferenceUtiOB
{
  friend class CDoubleClass;
  friend class CAnimationClass;
  friend class COrderClass;
  //$classbody::begin CAnimation(ancestor="ReferenceUtiOB" serialize="2001")

       //$property CurrentPosition(type="TIME" init="0.0f")
       //$property TimeUsed(type="TIME" init="0.0f")

       //$property Data(type="CAnimationDataSource" init="NULL" mode="rw_ref" public)
       //$property AnimationClass(type="CAnimationClass" mode="rw_ptr" noundo)
       //$property Code(type="SCode" init="NULLCODE")
       //$property Flags(type="int" init="0")
   //$classbody::middle
public: CUtiString getClassName();
protected: TIME m_CurrentPosition;
public: TIME getCurrentPosition() const;
public: void setCurrentPosition(const TIME& value0);
protected: TIME m_TimeUsed;
public: TIME getTimeUsed() const;
public: void setTimeUsed(const TIME& value0);
public: uti_refptr<CAnimationDataSource> m_Data;
public: CAnimationDataSource* getData();
public: const CAnimationDataSource* getData() const;
public: void setData(CAnimationDataSource* value0);
protected: CAnimationClass* m_AnimationClass;
public: CAnimationClass* getAnimationClass();
public: const CAnimationClass* getAnimationClass() const;
public: void setAnimationClass(CAnimationClass* value0);
protected: SCode m_Code;
public: SCode getCode() const;
public: void setCode(const SCode& value0);
protected: int m_Flags;
public: int getFlags() const;
public: void setFlags(const int& value0);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end

public:
	HResult Read(CStream & Stream, int Flags, int version);
	HResult Write(CStream & Stream, int Flags, int version);
    CAnimation(TIME newTime, SCode newCode);
	CAnimation();

	//CCacheList points;
	//SCode code;
	//UINT flag;
	//bool usedouble;
	//TIME CurrentPosition, TimeUsed;
	void createStdAnimation();
    //void setLength(TIME time);
	void setRepeat(bool r);
	bool getRepeat();
    //void sortPoints();

    enum {ofRepeat = 1};

};

class CAnimationInfo : public ReferenceUtiOB
{
  friend class CDoubleClass;
  friend class CAnimationClass;
  friend class COrderClass;
  //$classbody::begin CAnimationInfo(ancestor="ReferenceUtiOB" serialize="2004")


       //$property Animations(type="CAnimation"  mode="array_ref")
       //$property AnimationClass(type="CAnimationClass" mode="rw_ptr" noundo)
       //$property Active(type="CAnimation" mode="rw_ref")
       //property BaseValue(type="float" init="0.0f" mode="rw")

   //$classbody::middle
public: CUtiString getClassName();
protected:  CReferenceList<CAnimation> m_Animations;
public: CAnimation* getAnimations(int index);
public: int getAnimationsCount();
public: void addAnimations(CAnimation* value);
public: void insertAnimations(int index, CAnimation* value);
public: void removeAnimations(CAnimation* value);
public: void clearAnimations();
public: void deleteAnimations(int index);
public: int indexOfAnimations(CAnimation* value);
protected: CAnimationClass* m_AnimationClass;
public: CAnimationClass* getAnimationClass();
public: const CAnimationClass* getAnimationClass() const;
public: void setAnimationClass(CAnimationClass* value0);
protected: uti_refptr<CAnimation> m_Active;
public: CAnimation* getActive();
public: const CAnimation* getActive() const;
public: void setActive(CAnimation* value0);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end
public:
	HResult Read(CStream & Stream, int Flags, int version);
	HResult Write(CStream & Stream, int Flags, int version);
	CAnimationInfo();
	//CCacheList Animations;
    DynArray1D<CAnimation*> indices;
	CAnimation* addAnimation(TIME newTime, SCode newCode);
	CAnimation* findAnimation(const SCode &code);
	//CAnimation* active;
	//float Grundwert;
	//float BaseValue;
	//bool SetGrundwert;
	//int aniIndex;
	//bool usedouble;
};

struct _anires
{
	CAnimation *result;
    SCode code;
};

struct _con
{
    SCode s;
	CUtiOB* con;
};

#define MSG_SEARCHMANAGER 500
#define MSG_CREATEMANAGER 501
#define MSG_SEARCHMANAGERSTRING 502

class CAnimationClass : public ReferenceUtiOB
{
protected:
  //$classbody::begin CAnimationClass(ancestor="ReferenceUtiOB" serialize="2005")


       //$property Animations(type="CAnimation"  mode="array_ref")
       //$property ValueInfo(type="int" mode="rw")
       //$property AniInfo(type="CAnimationInfo" mode="rw_ref")

   //$classbody::middle
public: CUtiString getClassName();
protected:  CReferenceList<CAnimation> m_Animations;
public: CAnimation* getAnimations(int index);
public: int getAnimationsCount();
public: void addAnimations(CAnimation* value);
public: void insertAnimations(int index, CAnimation* value);
public: void removeAnimations(CAnimation* value);
public: void clearAnimations();
public: void deleteAnimations(int index);
public: int indexOfAnimations(CAnimation* value);
protected: int m_ValueInfo;
public: int getValueInfo() const;
public: void setValueInfo(const int& value0);
protected: uti_refptr<CAnimationInfo> m_AniInfo;
public: CAnimationInfo* getAniInfo();
public: const CAnimationInfo* getAniInfo() const;
public: void setAniInfo(CAnimationInfo* value0);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end
   void PaintWorld();
public:
       virtual CAniPoint* createAniPoint();
       virtual CAnimationDataPoints* createAnimationData(){return NULL;};
       virtual bool normalAnimation() {return true;};
   virtual void createStdAnimation(CAnimation* ani) {assert(false);};
	void setParent(CUtiOB* p, int info);
	void ProzessUpdate(SProzess* Update);
    virtual void addIndices(const SCode &code){};
	//void setRepeat(bool r);
	//bool getRepeat();
	void setAnimated(bool animated);
	bool getAnimated();
	//int getPosition();
	//void setCode(SCode code);
	//SCode getCode();
	//void setLength(TIME length);
	//void setLength(int length, CAnimation* ani);
	//int getLength();
	CAnimation* getActiveAnimation();
	CAnimation* activateTrace(const SCode &code);
	CAnimation* activateTrace(CAnimation* ani);
	virtual void setTrackPosition(TIME time, int flags);
	void LastPosition();
	void FirstPosition();
	CAnimation* findAnimation(const SCode &code);
    virtual CAnimation * CreateNewAnimation(TIME newTime, SCode newCode, bool activate=false);
	CAnimationClass();
	~CAnimationClass();

	//int valueInfo;
	//CAnimationInfo* Aniinfo;
	enum {ctFrametime, ctSetTime, ctSetCode, ctSetCodeZero, ctCountAnimations, ctUpdateCodes, ctSetController, ctAfterRead, ctAfterRead2, ctIndex};
};
/*
class CDoubleClass : public CAnimationClass
{

public:
#ifdef EDITOR
	CUtiEdit* editfeld;
	CUtiOB* dedit;
#endif
      virtual CAniPoint* createAniPoint();
      void createStdAnimation(CAnimation* ani);
	HResult Write(CStream & Stream, int Flags, int version);
	HResult Read(CStream & Stream, int Flags, int version);
	void setWert(float d);
	void setTrackPosition(TIME time, int flags);
	CDoubleClass();
	~CDoubleClass();
   // CAnimation *CreateNewAnimation(TIME newTime, SCode newCode);
    float getBaseValue();
	float Wert;
	void addIndices(const SCode &code);
	float getAnimationValue(CAnimation* ani, TIME time);
	float getAnimationValueByIndices(int index, TIME time);
    int GetMessageUti(CUtiOB *Sender, int msg, int lParam, void* wParam, bool &handled);
    #ifdef EDITOR
	void CreateEdit(U_CHAR *c, CUtiControl *parent, int x, int y, int width, int height, int textw);
	void setEditText(CUtiEdit *edit);
#endif
#ifdef EDITOR
	void AddTreeItem(void* view, UINT parent, U_CHAR *c);
#endif
};*/

template<class T> class CAnimationValueClass : public CAnimationClass
{
protected:

    T m_basevalue;
    T m_value;

public:
      typedef CLinearInterpolationPoint<T> CPointClass;
      typedef CAnimationDataTemplate<T> CDataClass;

      virtual CAniPoint* createAniPoint()
      {
	 return new CPointClass();
      }

    virtual CAnimationDataPoints* createAnimationData()
    {
   return new CDataClass();
    }

      void createStdAnimation(CAnimation* ani)
      {
          CDataClass* dataclass = (CDataClass*)createAnimationData();

     ani->setData(dataclass);
         CPointClass *newPoint = NULL;

	newPoint = (CPointClass*)createAniPoint();
	newPoint->setTime(0);
	newPoint->setValue(m_basevalue);
    dataclass->addPoints(newPoint);

	newPoint = (CPointClass*)createAniPoint();
	newPoint->setTime(ani->getTimeUsed());
	newPoint->setValue(m_basevalue);
    dataclass->addPoints(newPoint);
      }
      CAnimationValueClass()
      {
	setAnimated(true);
    m_basevalue = T();
    m_value = T();
      }

      ~CAnimationValueClass()
      {

      }
   // CAnimation *CreateNewAnimation(TIME newTime, SCode newCode);
       T getBaseValue() const
       {
	  return m_basevalue;
       }
       void setBaseValue(const T& value)
       {
	  m_basevalue = value;
       }
       T getValue() const
       {
           return m_value;
       }
       void setValue(const T& v)
       {
           m_value = v;
       }

public: virtual void xread(CXNode* n, CWriteReadInterface* wri)
{
   CAnimationClass::xread(n, wri);
   uti_readValue(m_value, n, wri, UTISTR("Value"), T(0));
   uti_readValue(m_basevalue, n, wri, UTISTR("BaseValue"), T(0));
}
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri)
{
  CAnimationClass::xwrite(n, wri);
  uti_writeValue(m_value, n, wri, UTISTR("Value"));
  uti_writeValue(m_basevalue, n, wri, UTISTR("BaseValue"));
}
	void addIndices(const SCode &code)
	{
	   CAnimation* ani = findAnimation(code);
           utiassert(ani == NULL || m_AniInfo->indices.indexOf(ani) == -1);
           m_AniInfo->indices.push_back(ani);
	}
    T getAnimationValue(CAnimation* ani, TIME time)const
	{
      if (ani == NULL) return m_basevalue;
          utiassert(m_AniInfo.get()  != NULL);
          SAnimationInput input;
          input.m_time = time;
          SAnimationResultType<T> output;
          output.m_value = T();
          ani->getData()->evaluate(input, &output);
          return output.m_value;
	}
    T getAnimationValueByIndices(int index, TIME time)const
	{
	   utiassert(m_AniInfo.get() != NULL);
           utiassert(!(index < 0 || index >= m_AniInfo->indices.size()));
           CAnimation* ani = (CAnimation*)m_AniInfo->indices[index];
           if (ani == NULL) return m_basevalue.get() ;
           return getAnimationValue(ani, time);
	}

    HResult Write(CStream & Stream, int Flags, int version)
    {
        return(UTI_OK);
    }

    HResult Read(CStream & Stream, int Flags, int version)
    {
        return(UTI_OK);
    }
};


template<> class CAnimationValueClass<double> : public CAnimationClass
{
protected:


public:
    double m_basevalue;
    double m_value;
      typedef CLinearInterpolationPoint<double> CPointClass;
      typedef CAnimationDataTemplate<double> CDataClass;

      virtual CAniPoint* createAniPoint()
      {
     return new CPointClass();
      }

    virtual CAnimationDataPoints* createAnimationData()
    {
   return new CDataClass();
    }

      void createStdAnimation(CAnimation* ani)
      {
          CDataClass* dataclass = (CDataClass*)createAnimationData();

     ani->setData(dataclass);
         CPointClass *newPoint = NULL;

    newPoint = (CPointClass*)createAniPoint();
    newPoint->setTime(0);
    newPoint->setValue(m_basevalue);
    dataclass->addPoints(newPoint);

    newPoint = (CPointClass*)createAniPoint();
    newPoint->setTime(ani->getTimeUsed());
    newPoint->setValue(m_basevalue);
    dataclass->addPoints(newPoint);
      }
      CAnimationValueClass()
      {
    setAnimated(true);

    m_basevalue = 0.0f;
    m_value = 0.0f;

      }

      ~CAnimationValueClass()
      {

      }
   // CAnimation *CreateNewAnimation(TIME newTime, SCode newCode);
      inline double getBaseValue() const
       {
      return m_basevalue;
       }
       void setBaseValue(const double& value)
       {
           assert(utiisnan(value) == false && utiisfinite(value));
      m_basevalue = value;
       }
       inline double getValue() const
       {
           return m_value;
       }
       inline void setValue(const double& v)
       {
           assert(utiisnan(v) == false && utiisfinite(v));
           m_value = v;
       }
       void setValue(TIME time, const double& v)
       {
           assert(utiisnan(v) == false && utiisfinite(v));

           assert(utiisnan(time) == false && utiisfinite(time));
           CAnimation* animation = getActiveAnimation();
           assert(animation != NULL);
           CDataClass* dataclass = (CDataClass*)animation->getData();

           CPointClass *newPoint = dynamic_cast<CPointClass*>(dataclass->getAniPointByTime(time, 0.001));
           if (newPoint) {
                newPoint->setValue(v);
                return;
           }
           newPoint = (CPointClass*)createAniPoint();
           newPoint->setTime(time);
           newPoint->setValue(v);
           dataclass->addPoints(newPoint);
           dataclass->updatePointPosition(newPoint);


       }

public: virtual void xread(CXNode* n, CWriteReadInterface* wri)
{
   CAnimationClass::xread(n, wri);
   uti_readValue(m_value, n, wri, UTISTR("Value"),float(0));
   uti_readValue(m_basevalue, n, wri, UTISTR("BaseValue"), float(0));
}
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri)
{
  CAnimationClass::xwrite(n, wri);
  uti_writeValue(m_value, n, wri, UTISTR("Value"));
  uti_writeValue(m_basevalue, n, wri, UTISTR("BaseValue"));
}
    void addIndices(const SCode &code)
    {
       CAnimation* ani = findAnimation(code);
           utiassert(ani == NULL || m_AniInfo->indices.indexOf(ani) == -1);
           m_AniInfo->indices.push_back(ani);
    }
    double getAnimationValue(CAnimation* ani, TIME time)const
    {
        assert(utiisnan(time) == false && utiisfinite(time));
         if (m_AniInfo.get() == NULL || m_AniInfo->getAnimationsCount() == 0) return m_basevalue;
      if (ani == NULL) return m_basevalue ;
          utiassert(m_AniInfo.get()  != NULL);
          SAnimationInput input;
          input.m_time = time;
          SAnimationResultType<double> output;
          output.m_value = double();
          ani->m_Data->evaluate(input, &output);
          assert(utiisnan(output.m_value) == false && utiisfinite(output.m_value));
          return output.m_value;
    }
    double getAnimationValueByIndices(int index, TIME time)const
    {
        assert(utiisnan(time) == false && utiisfinite(time));
        if (m_AniInfo.get() == NULL || m_AniInfo->getAnimationsCount() == 0) return m_basevalue;
       utiassert(m_AniInfo.get() != NULL);
           //utiassert(!(index < 0 || index >= m_AniInfo->indices.count));
           if (index >= m_AniInfo->indices.size()) {
               //assert(false);
               return m_basevalue;
           }
           CAnimation* ani = (CAnimation*)m_AniInfo->indices[index];
           if (ani == NULL) return m_basevalue ;
           double res =  getAnimationValue(ani, time);
           assert(utiisnan(res) == false && utiisfinite(res));
           return res;
    }

    HResult Write(CStream & Stream, int Flags, int version)
    {
        bool IsAnimated;

       IsAnimated = (m_AniInfo.get() != NULL);
    //MWorld->
        if ((Flags & SAVE_NO_ANIMATIONS) != 0) {
            IsAnimated = false;
        }

        Stream.WriteData(&IsAnimated, 1);
        double w = getValue();
        Stream.WriteData(&w, 8);


        if (IsAnimated) {
            //m_AniInfo = new CAnimationInfo;

             // m_AniInfo->setBaseValue(w);

            m_AniInfo->Write(Stream, Flags, version);
           /* if (m_AniInfo->getActive()) {
                setValue( getAnimationValue(m_AniInfo->getActive(), 0));
            }*/
        }
        return(UTI_OK);
    }

    HResult Read(CStream & Stream, int Flags, int version)
    {
        bool IsAnimated;

        if (m_AniInfo.get() != NULL) m_AniInfo = NULL;
    //MWorld->
        Stream.ReadData(&IsAnimated, 1);
        double w;
        Stream.ReadData(&w, 8);
    setValue(w);
    setBaseValue(w);

        if (IsAnimated) {
            //m_AniInfo = new CAnimationInfo;
              setAnimated(true);
           assert(m_AniInfo.get() != NULL);
             // m_AniInfo->setBaseValue(w);

            m_AniInfo->Read(Stream, Flags, version);
           /* if (m_AniInfo->getActive()) {
                setValue( getAnimationValue(m_AniInfo->getActive(), 0));
            }*/
        }
        return(UTI_OK);
    }
};







#define CDECLAREANITYPE_CLASS(BaseClass,Type,serial, pointclass, dataclass)\
  class BaseClass ## _ ## Type : public BaseClass<Type> \
  {\
   public:\
     BaseClass ## _ ## Type() : BaseClass<Type>() {\
       \
     }\
     virtual CAniPoint* createAniPoint()\
      {\
	 return new pointclass();\
      }\
    virtual CAnimationDataPoints* createAnimationData()\
    {\
   return new dataclass();\
    }\
     public: virtual const SCode GetTypeCode();\
     public: static const SCode TYPECODE;\
  };

#define CDECLAREANITYPE_CLASS_IMPL(BaseClass,Type,serial, pointclass, dataclass)\
  const SCode BaseClass ## _ ## Type::TYPECODE =\
 {\
    serial, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }\
};\
const SCode BaseClass ## _ ## Type::GetTypeCode()\
{\
    return BaseClass ## _ ## Type::TYPECODE;\
}\
IMPLEMENT_CLASS_CREATORTYPENOPARENT(BaseClass ## _ ## Type);

#define CDECLAREANITYPE_ALL_CLASS(BaseClass, baseserial)\
  CDECLAREANITYPE_CLASS(BaseClass, double, baseserial+0, CLinearInterpolationPoint_double, CAnimationDataTemplate_double)\
  CDECLAREANITYPE_CLASS(BaseClass, CVector2, baseserial+2, CLinearInterpolationPoint_CVector2, CAnimationDataTemplate_CVector2)\
  CDECLAREANITYPE_CLASS(BaseClass, CVector3, baseserial+3, CLinearInterpolationPoint_CVector3, CAnimationDataTemplate_CVector3)\
  CDECLAREANITYPE_CLASS(BaseClass, CVector4, baseserial+4, CLinearInterpolationPoint_CVector4, CAnimationDataTemplate_CVector4)

  #define CDECLAREANITYPE_ALL_CLASS_IMPL(BaseClass, baseserial)\
  CDECLAREANITYPE_CLASS_IMPL(BaseClass, double, baseserial+0, CLinearInterpolationPoint_double, CAnimationDataTemplate_double)\
  CDECLAREANITYPE_CLASS_IMPL(BaseClass, CVector2, baseserial+2, CLinearInterpolationPoint_CVector2, CAnimationDataTemplate_CVector2)\
  CDECLAREANITYPE_CLASS_IMPL(BaseClass, CVector3, baseserial+3, CLinearInterpolationPoint_CVector3, CAnimationDataTemplate_CVector3)\
  CDECLAREANITYPE_CLASS_IMPL(BaseClass, CVector4, baseserial+4, CLinearInterpolationPoint_CVector4, CAnimationDataTemplate_CVector4)

CDECLAREANITYPE_ALL_CLASS(CAnimationValueClass, 2200);

class COrderClass : public CAnimationClass
{
public:
     virtual CAniPoint* createAniPoint();
     virtual bool normalAnimation() {return false;};
	CCacheList Orders;
	COrderClass();
	~COrderClass();
	HResult Write(CStream & Stream, int Flags, int version);
	HResult Read(CStream & Stream, int Flags, int version);
        void setWert(double d);
	void setTrackPosition(TIME time, int flags);
	//CAnimation *CreateNewAnimation(int newTime, SCode newCode);
    void AddSegment(TIME Start, TIME Ende);
};

class CVector4Class : public ReferenceUtiOB
{
public:
	void getWert(SVector4 &s);
	void setParent(CUtiOB* p, int info);
        void Set(utifloat _x, utifloat _y, utifloat _z, utifloat _a);
	void setVec4(const CVector4& v);
	CVector4 getVec4();
	void Set(UINT rgbvalue);
	void Get(S3DPoint* p);
	void Get(CVector* p);
    void Get(utifloat *p){p[0] = (utifloat)x.getValue();p[1] = (utifloat)y.getValue();p[2] = (utifloat)z.getValue();p[3] = (utifloat)a.getValue();};
    //void Add(CVector4Class *v){x.getValue() += v->x.Wert;y.Wert += v->y.Wert;z.Wert += v->z.Wert;};
	void ProzessUpdate(SProzess* Update);
	bool animated;
    CVector4Class() {animated = false;/*typinfo = UTI_4DQUATRE;*/};
	HResult Read(CStream & Stream, int Flags, int version);
	HResult Write(CStream & Stream, int Flags, int version);
    HResult ReadStream(CStream & Stream, int Flags, int version);
    HResult WriteStream(CStream & Stream, int Flags, int version);
    CAnimationValueClass_double x;
    CAnimationValueClass_double y;
    CAnimationValueClass_double z;
    CAnimationValueClass_double a;
};

class C3DTriple : public ReferenceUtiOB
{
public:
	void set(S3DPoint *p);
        void set(utifloat _x, utifloat _y, utifloat _z);
	void set(const CVector& v);
    CAnimationValueClass_double x;
    CAnimationValueClass_double y;
    CAnimationValueClass_double z;
	bool animated;
    inline void setVec3(const CVector3& v)
    {
      x.setValue(v.x);
      y.setValue(v.y);
      z.setValue(v.z);
    }
    CVector3 getVec3()
    {
        return CVector3(x.getValue(), y.getValue(), z.getValue());
    }
    C3DTriple()  {animated = false;/*typinfo = UTI_3DTRIPLE;*/};
	HResult Read(CStream & Stream, int Flags, int version);
	HResult Write(CStream & Stream, int Flags, int version);
	void setParent(CUtiOB* Parent);
    //void Get3DPoint(S3DPoint *p) const {p->x = (float)x.Wert; p->y = (float)y.Wert; p->z = (float)z.Wert;};
    void GetVector(CVector *v)const  {v->x = (utifloat)x.getValue(); v->y = (utifloat)y.getValue(); v->z = (utifloat)z.getValue();};
    CVector getVector()const {return CVector((utifloat)x.getValue(), (utifloat)y.getValue(), (utifloat)z.getValue());};
    //void Set3DPoint(S3DPoint *p) {x.Wert = p->x; y.Wert = p->y; z.Wert = p->z;};
   inline void SetVector(const CVector *p) {x.setValue(p->x); y.setValue(p->y); z.setValue(p->z);};
    //void Get(float* v)const  {v[0] = (float)x.Wert; v[1] = (float)y.Wert; v[2] = (float)z.Wert;};
	void ProzessUpdate(SProzess* Update);
	void VectorToAngle(CVector v);
#ifdef EDITOR
	void AddTreeItem(void* view, UINT parent, U_CHAR *c);
#endif
};

class CAnimationFilter : public CUtiOB
{
protected:
        utifloat komma;
public:
        utifloat faktor;
	CAnimationFilter();
	void Filter(SProzess* pnew, SProzess* pold);
};

#endif
