#pragma once
//Uti3DMath
/***************************************************************************
 *   Copyright (C) 2006 by Michael Staud   *
 *   michael.staud@uni-ulm.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file */


#ifndef __UTI3DMATH__
#define __UTI3DMATH__

#include "Uti3DCore.h"

#include <stdlib.h>
#include <cmath>
#include <iostream>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

const int MAX_POINTS =20;
const float EPSILON =0.05f;

const int CP_FRONT =1;
const int CP_BACK =-1;
const int CP_ZERO =0;

//#define USE_DOUBLEtrunc()




//Datentypen

const int TBGL[16] = {0, 4, 8, 12,
                                1, 5, 9, 13,
                                2, 6, 10, 14,
                                3, 7, 11, 15};

struct Matrix3
{
   utifloat m[3][3];
};

#ifdef _MSC_VER
inline bool finite(float f){return isfinite(f); };
#endif

void Perform_OpenGL_matrix_inverse(float* _in,float* _out);

class C3DPoint;
//struct CVector;
struct CPlane;

const double pi180 =0.0174532925199432957692369076848861; //pi/180



struct S3DVERT
{
    U_SINGLE x, y, z;
	U_SINGLE u, v;
	U_SINGLE u2, v2;
	U_SINGLE s;
};

struct S3DUNLIT
{
    U_SINGLE x, y, z;
	U_SINGLE u, v, r;
	UINT color;
};

/*typedef struct S2DPoint
{
    U_SINGLE x, y;
} *P2DPoint;

S2DPoint set2DPoint(U_SINGLE x, U_SINGLE y);
struct S2DRect
{
	S2DPoint p1;
	S2DPoint p2;
};*/

struct CBox;
struct CVector3;
struct CVector2;


class CLineDrawingDebugOutputInterface : public CUtiOB
{
public:
    CLineDrawingDebugOutputInterface(){};
    virtual void debug_Line(const CVector3& start, const CVector3& end, const CVector3& color, float duration)=0;
    virtual void debug_Ray(const CVector3& start, const CVector3& dir, const CVector3& color, float duration)=0;
    virtual void debug_Box(const CBox& box, const CVector3& color, float duration)=0;
    virtual void debug_Point(const CVector3& pos, const CVector3& color, float duration)=0;
    virtual void debug_String(const CVector3& pos, const CVector3& color, float duration, const CUtiString& html)=0;
    virtual void debug_StringXY(const CVector2& pos, const CVector3& color, float duration, const CUtiString& html)=0;
};


void setLineDrawingDebugInterface(CLineDrawingDebugOutputInterface* debug);
void debug_Line(const CVector3& start, const CVector3& end, const CVector3& color, float duration);
void debug_Ray(const CVector3& start, const CVector3& dir, const CVector3& color, float duration);
void debug_Box(const CBox& box, const CVector3& color, float duration);
void debug_Point(const CVector3& pos, const CVector3& color, float duration);
void debug_String(const CVector3& pos, const CVector3& color, float duration, const CUtiString& html);
void debug_StringXY(const CVector2& pos, const CVector3& color, float duration, const CUtiString& html);

struct CRect;

struct SRect
{
   float x1, y1, x2, y2;
   inline void move(float x, float y)
   {
       x1 += x;
	   y1 += y;
	   x2 += x;
	   y2 += y;
   };
};


/*
struct SPoint
{
  SPoint() {x = 0; y = 0;};
  SPoint(int _x, int _y) {x = _x; y = _y;};
   int x, y;
};*/


int maxf(utifloat x, utifloat y, utifloat z);

/*

void Add(S3DPoint* result, S3DPoint* sum1, S3DPoint* sum2);
void Add(S3DPoint* result, S3DPoint* sum1, CVector* sum2);
void Add(S3DPoint* result, CVector* sum1, S3DPoint* sum2);
void Add(S3DPoint* result, CVector* sum1, CVector* sum2);
void AddV(S3DPoint* result, S3DPoint* sum1, S3DPoint* sum2);
void AddV(S3DPoint* result, S3DPoint* sum1, CVector* sum2);
void AddV(S3DPoint* result, CVector* sum1, S3DPoint* sum2);
void AddV(S3DPoint* result, CVector* sum1, CVector* sum2);*/

template< class T1, class T2, class T3> inline void Point_Add(T1 *sum, T2* s1, T3* s2, int count)
{
for (int i = 0; i < count; i++) {
    sum[i] = s1[i] + s2[i];
}
}

/*void Point_Sub(float *diff, float *sub, float *min, int count);*/
template< class T1, class T2, class T3> inline void Point_Mul(T1 *p, T2 *f1, T3 *f2, int count){
    for (int i = 0; i < count; i++) {
        p[i] = f1[i] * f2[i];
    }
}
/*void Point_Div(float *quot, float *div, float *divisor, int count);
void Point_Add_Value(float *sum, float*sum1, float value, int count);
void Point_Sub_Value(float *diff, float*sub1, float min, int count);
void Value_Sub_Point(float *diff, float sub1, float* min, int count);*/
template< class T1, class T2, class T3> inline void Point_Mul_Value(T1 *pro, T2*f1, T3 value, int count)
{
    for (int i = 0; i < count; i++) {
        pro[i] = f1[i] * value;
    }
}
/*void Point_Div_Value(float *quot, float*div, float divisor, int count);
void Value_Div_Point(float *quot, float div, float*divisor, int count);*/

inline float absfloat(float x)
{
    if (x < 0)
        return(-x);
    else
        return(x);
};


inline utifloat absf(utifloat x)
{
    if (x < 0)
		return(-x);
	else
		return(x);
};

inline int intpow(int base, int pot)
{
  int r = 1;
  for (int i = 0; i < pot; i++) {
    r *= base;
  };
  return r;
};



#include "utivector.h"


inline unsigned int UtiRGBA(const CVector4& color)
{
    return UtiRGBA((unsigned int)(color.x*255), (unsigned int)(color.y*255),(unsigned int) (color.z*255),(unsigned int) (color.w*255));
}

inline unsigned int UtiRGBA(const CVector4f& color)
{
    return UtiRGBA((unsigned int)(color.x*255), (unsigned int)(color.y*255),(unsigned int) (color.z*255),(unsigned int) (color.w*255));
}


CVector3 utireflect(const CVector3&p, const CVector3& Norm);
CVector3f utireflect(const CVector3f&p, const CVector3f& Norm);

bool isvectorok(const CVector2& v);
bool isvectorok(const CVector3& v);
bool isvectorok(const CVector4& v);

std::wostream & operator<<(std::wostream & os, const CVector2 & str);
std::ostream & operator<<(std::ostream & os, const CVector2 & str);

std::wostream & operator<<(std::wostream & os, const CVector3 & str);
std::ostream & operator<<(std::ostream & os, const CVector3 & str);

std::wostream & operator<<(std::wostream & os, const CVector4 & str);
std::ostream & operator<<(std::ostream & os, const CVector4 & str);


inline int getMinorAxis(const CVector3& v)
{
float current = fabs(v.x);
int index = 0;
if (current > fabs(v.y)) {
index = 1;
current = fabs(v.y);
};
if (current > fabs(v.z)) {
index = 2;
current = fabs(v.z);
};
return index;
}
/*
#ifdef _MSC_VER
inline float round(float v)
{
	return (float)( (v >= 0) ? (int)(v + 0.5) : (int)(v - 0.5));
}
inline double exp2(double arg)
{
    return pow(2.0, arg);
}
inline double log2( double n )
{
    // log(n)/log(2) is log2.
    return log( n ) / log( 2.0 );
}
inline double rint(double x)
{
      return floor(x + 0.5);
}
inline  double lrint(double x)
{
     return (floor((x)+((x)>0) ? 0.5 : -0.5));
}

inline double trunc (double x)
{
  return (((x) < 0) ? ceil((x)) : floor((x)));
}
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#endif*/


struct S3DPointf
{
    U_SINGLE x, y, z;
    U_SINGLE nx, ny, nz;
    U_SINGLE u, v, r;
    int flags;
    U_UINT32 color;
    inline CVector3 getPosition() const {return CVector3(x,y,z);};
    inline void setPosition(const CVector3& v1) {x = v1.x; y = v1.y; z = v1.z;};
    inline CVector3 getNormal() const {return CVector3(nx,ny,nz);};
    inline void setNormal(const CVector3& v1) {nx = v1.x; ny = v1.y; nz = v1.z;};
    inline CVector3 getUVR() const {return CVector3(u,v,r);};
    inline void setUVR(const CVector3& v1) {u= v1.x; v = v1.y; r = v1.z;};

    inline CVector3f getPositionf() const {return CVector3f(x,y,z);};
    inline void setPosition(const CVector3f& v1) {x = v1.x; y = v1.y; z = v1.z;};
    inline CVector3f getNormalf() const {return CVector3f(nx,ny,nz);};
    inline void setNormal(const CVector3f& v1) {nx = v1.x; ny = v1.y; nz = v1.z;};
    inline CVector3f getUVRf() const {return CVector3f(u,v,r);};
    inline void setUVR(const CVector3f& v1) {u= v1.x; v = v1.y; r = v1.z;};
};


typedef struct S3DPoint
{
    double x, y, z;
    double nx, ny, nz;
    double u, v, r;
	int flags;
	U_UINT32 color;
    inline CVector3 getPosition() const {return CVector3(x,y,z);};
	inline void setPosition(const CVector3& v1) {x = v1.x; y = v1.y; z = v1.z;};
    inline CVector3 getNormal() const {return CVector3(nx,ny,nz);};
	inline void setNormal(const CVector3& v1) {nx = v1.x; ny = v1.y; nz = v1.z;};
    inline CVector3 getUVR() const {return CVector3(u,v,r);};
	inline void setUVR(const CVector3& v1) {u= v1.x; v = v1.y; r = v1.z;};

    inline CVector3f getPositionf() const {return CVector3f(x,y,z);};
    inline void setPosition(const CVector3f& v1) {x = v1.x; y = v1.y; z = v1.z;};
    inline CVector3f getNormalf() const {return CVector3f(nx,ny,nz);};
    inline void setNormal(const CVector3f& v1) {nx = v1.x; ny = v1.y; nz = v1.z;};
    inline CVector3f getUVRf() const {return CVector3f(u,v,r);};
    inline void setUVR(const CVector3f& v1) {u= v1.x; v = v1.y; r = v1.z;};
} T3DPoint;

class CUtiPHMaterial;



typedef CVector3 CVector;
typedef CVector SVector;
typedef CVector TVector;
typedef CVector* PVector;

void Set3DPoint(S3DPoint* p, float x, float y, float z, float nx=0, float ny=0, float nz=0, float tx=0, float ty=0, float tz = 0, UINT color = 0);
//S3DPoint Set3DPoint(float x, float y, float z, S3DPoint* p1);

void Set3DPoint(float x, float y, float z, S3DPoint*p);


void GetRandomPointInRect(S3DPoint* p, CBox* r);

inline void vectorToPoint(S3DPoint& res, const CVector& source)
{
   res.x = source.x;
   res.y = source.y;
   res.z = source.z;
}

inline void pointToVector(CVector& res, const S3DPoint& source)
{
   res.x = source.x;
   res.y = source.y;
   res.z = source.z;
}


/*
inline bool operator ==(const CIVector2 &v1, const CIVector2 &v2)
{
   return (v1.x == v2.x) && (v1.y == v2.y);
}
inline bool operator ==(const CIVector3 &v1, const CIVector3 &v2)
{
   return (v1.x == v2.x) && (v1.y == v2.y) && (v1.z == v2.z);
}
inline bool operator ==(const CIVector4 &v1, const CIVector4 &v2)
{
   return (v1.x == v2.x) && (v1.y == v2.y) && (v1.z == v2.z) && (v1.w == v2.w);
}*/

struct CBVector2 {
   bool x;
   bool y;
   inline CBVector2(){};
   inline CBVector2(bool _x, bool _y){x = _x; y = _y;};
};

struct CBVector3 {
   bool x;
   bool y;
   bool z;
   inline CBVector3(){};
   inline CBVector3(bool _x, bool _y, bool _z){x = _x; y = _y; z = _z;};
};

struct CBVector4 {
   bool x;
   bool y;
   bool z;
   bool w;
   inline CBVector4(){};
   inline CBVector4(bool _x, bool _y, bool _z, bool _w){x = _x; y = _y; z = _z; w = _w;};
};



typedef CVector2 S2DPoint;
typedef S2DPoint *P2DPoint;

S2DPoint set2DPoint(U_SINGLE x, U_SINGLE y);
struct S2DRect
{
	S2DPoint p1;
	S2DPoint p2;
};

const CVector setVector(utifloat x, utifloat y, utifloat z);
const CVector3f setVectorf(utifloat x, utifloat y, utifloat z);
CVector crossProduct(const CVector& v1, const CVector& v2);
CVector3f crossProduct(const CVector3f& v1, const CVector3f& v2);

bool IntersectLinePlaneV(utifloat d, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis);
bool IntersectLinePlaneV(utifloat d, CVector3f *v, S3DPoint *line, S3DPoint *result, utifloat *dis);
void Get3DRectClipPlanes(CPlane *planes, CBox* rect);
void GetRectCenter(const CBox* rect, CVector *p);
void MoveRect(CBox* rect, S3DPoint* t);
int GetRectPoints(CBox* rectm, S3DPoint* points, int count, CVector* dir);
void getBoxPoints(const CBox& rectm, CVector* points);
bool PointInRect(const CBox* rect, const S3DPoint *p);
bool PointInRect(const CBox* rect, const CVector *p);
bool PtInRect3D(const S3DPoint *p,  const CBox* rect2);
bool IntersectRect3D(const CBox* rect1, const CBox* rect2);
bool IntersectRay(const S3DPoint* vector, const CBox* b, utifloat &dis);
bool IntersectRayB(const S3DPoint* vector, const CBox* b, utifloat &dis, bool &in);
bool IntersectNormRay(const S3DPoint* vector, const CBox* b, utifloat &dis, CVector* normal);
void InitBounds(CBox &Bounds);
bool IntersectLinePlaneV(utifloat d, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis);
bool IntersectLineSphere(const CVector3& origin, const CVector3& dir,
                         const CVector& position, utifloat radius,
                         utifloat& nearest,  CVector& res);
bool IntersectLineCircle(const CVector2& origin, const CVector2& dir,
                         const CVector2& position, utifloat radius,
                         utifloat& nearest,  CVector2& res);
struct CRect;


template <class T> inline void utiswap(T &f1, T &f2) {T temp= f1; f1 = f2; f2 = temp;};


/*
struct CBox {
  static CBox INIT() {CBox r; r.init(); return r;};
    static CBox ADVANCEINIT() {CBox r; r.m_min=CVector3(1000000, 1000000,1000000);
       r.m_max=CVector3(-1000000, -1000000, -1000000);return r;};
    CVector3 m_min;
    CVector3 m_max;
    inline CBox() {};
    inline CBox(float x1, float y1, float x2, float y2) {m_min.x = x1; m_min.y = y1; m_max.x = x2; m_max.y = y2;};
    inline CBox(const CVector3& v1, const CVector3&v2) {m_min = v1; m_max=v2;};
    inline void init() {m_min = CVector3(0.0, 0.0, 0.0); m_max = CVector3(0.0,0.0, 0.0);};
    inline CVector3 getCenter() const {return (m_max+m_min)/2.0f;};
    inline float getWidth() const {return m_max.x-m_min.x;};
    inline float getHeight()const  {return m_max.y-m_min.y;};
    inline float getDepth() const {return m_max.z-m_min.z;};
    inline void setWidth(float v) {m_max.x = m_min.x+v;};
    inline void setHeight(float v) {m_max.y = m_min.y+v;};
    inline void setDepth(float v) {m_max.z = m_min.z+v;};
    inline float getLeft()const  {return m_min.x;};
    inline float getTop() const {return m_min.y;};
    inline float getFront() const {return m_min.z;};
    inline void move(const CVector3& v) { m_min+= v; m_max += v;};
    inline void setLeft(float v) {move(CVector3(v - m_min.x, 0, 0));};
    inline void setTop(float v) {move(CVector3(0, v - m_min.y, 0));};
    inline void setFront(float v) {move(CVector3(0, 0, v-m_min.z));};
    inline void correct() {if (m_min.x > m_max.x) {utiswap(m_min.x, m_max.x);};
                           if (m_min.y > m_max.y) {utiswap(m_min.y, m_max.y);};
                           if (m_min.z > m_max.z) {utiswap(m_min.z, m_max.z);};}
    inline bool intersect(const CVector3& position) const {
       return (m_min.x <= position.x ) && (m_max.x >= position.x)
       && (m_min.y <= position.y) && (m_max.y >= position.y)
       && (m_min.z <= position.y) && (m_max.z >= position.z);
    };
    inline void expand(const CBox& rect) {
      if (m_min.x > rect.m_min.x)m_min.x = rect.m_min.x;

      if (m_max.x < rect.m_max.x) m_max.x = rect.m_max.x;
      if (m_min.y > rect.m_min.y)m_min.y = rect.m_min.y;

      if (m_max.y < rect.m_max.y) m_max.y = rect.m_max.y;
      if (m_min.z > rect.m_min.z) m_min.z = rect.m_min.z;

      if (m_max.z < rect.m_max.z) m_max.z = rect.m_max.z;
    };
    inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y < m_max.y && m_min.z < m_max.z;
    }
    inline int getMaximumAxis()
    {
        int res = 0;
	float size = getWidth();
	if (getHeight() > size) {size = getHeight(); res = 1;};
	if (getDepth() > size) {size = getDepth(); res =2;};
	return res;
    };
    inline bool intersect(const CBox& rect) const {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      if (m_min.x> rect.m_max.x) return false;
      if (m_min.y> rect.m_max.y) return false;
      if (m_min.z> rect.m_max.z) return false;
      if (m_max.x< rect.m_min.x) return false;
      if (m_max.y< rect.m_min.y) return false;
      if (m_max.z< rect.m_min.z) return false;
      return true;
    };
};
*/

struct TBounds
{
    TVector p1;
	TVector p2;
};


class C3DPoint
{
public:
	float x;
	float y;
	float z;
	C3DPoint();
	C3DPoint(float _x, float _y, float _z);
	virtual ~C3DPoint();
	void print();
};

CVector3 hexRGBtoVec3(unsigned int color);

inline CVector4 getQuaternionIdentity() {return CVector4(0.0f, 0.0f, 0.0f, 1.0f);};
CVector4 getQuaternionAxisAngle(const CVector& axis, utifloat angle);
CVector4 quaternionMultiply(const CVector4& a, const CVector4& b);
CVector4f quaternionMultiply(const CVector4f& a, const CVector4f& b);
CVector4 quaternionMultiplyInvA(const CVector4& a, const CVector4& b);
CVector4 quaternionMultiplyInvB(const CVector4& a, const CVector4& b);
CVector4 quaternionMultiplyInvAB(const CVector4& a, const CVector4& b);
CVector4 eulerToQuaternion(const CVector3 & a);
CVector4f eulerToQuaternion(const CVector3f & a);
//CVector4 eulerToQuaternionOld(const CVector3 & a);
bool isQuaternionEqual(const CVector4& quat1, const CVector4& quat2, float threshold);
inline CVector4 inverseQuaternion(const CVector4& q) {return CVector4(-q.x, -q.y, -q.z, q.w);};
inline CVector4f inverseQuaternion(const CVector4f& q) {return CVector4f(-q.x, -q.y, -q.z, q.w);};
CVector4 farthestQuaternion(const CVector4& self, const CVector4& other);
CVector4 nearestQuaternion(const CVector4& self, const CVector4& other);
CVector4 slerpQuaternion(const CVector4& self, const CVector4&q, float t);
CVector3 rotateViaQuaternion(const CVector4& self, const CVector3& p);
CVector3f rotateViaQuaternion(const CVector4f& self, const CVector3f& p);
CVector3 rotateViaEuler(const CVector3& euler, const CVector3& p);
CVector3 rotateViaEuler2(const CVector3& euler, const CVector3& p);

struct NetRotationInfo;

struct CMatrixf
{
public:

   float m[4][4];
   static void inverse(const CMatrixf &in, CMatrixf &out);
   CMatrixf()
   {
      m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1;
       m[1][0] = m[2][0] = m[3][0] = 0;
       m[0][1] = m[2][1] = m[3][1] = 0;
       m[0][2] = m[1][2] = m[3][2] = 0;
       m[0][3] = m[1][3] = m[2][3] = 0;
   }

 /*  void CMatrixf::multiply(CMatrixf & in, CMatrixf& out)
   {

   }
*/
   inline void LoadIdentity()
  {
      m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1;
      m[1][0] = m[2][0] = m[3][0] = 0;
      m[0][1] = m[2][1] = m[3][1] = 0;
      m[0][2] = m[1][2] = m[3][2] = 0;
      m[0][3] = m[1][3] = m[2][3] = 0;
   }

   inline float& operator()(int i, int j)
   {
       utiassert(i >= 0 && i < 4);
       utiassert(j >= 0 && j < 4);
       return m[i][j];

   }
   inline float operator()(int i, int j) const
   {
       utiassert(i >= 0 && i < 4);
       utiassert(j >= 0 && j < 4);
       return m[i][j];

   }

   inline void set(int i, int j, float f)
   {
       utiassert(i >= 0 && i < 4);
       utiassert(j >= 0 && j < 4);
       m[i][j] = f;

   }
   inline float get(int i, int j) const
   {
       utiassert(i >= 0 && i < 4);
       utiassert(j >= 0 && j < 4);
       return m[i][j];

   }
};

struct CMatrix
{
public:
	static void inverse(const CMatrix &in, CMatrix &out);

   utifloat m[4][4];
   CMatrix()
   {
      m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1;
	   m[1][0] = m[2][0] = m[3][0] = 0;
	   m[0][1] = m[2][1] = m[3][1] = 0;
	   m[0][2] = m[1][2] = m[3][2] = 0;
	   m[0][3] = m[1][3] = m[2][3] = 0;
   }
	//~CMatrix();
	//void LoadIdentity();
	inline void LoadIdentity()
   {
	   m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1;
	   m[1][0] = m[2][0] = m[3][0] = 0;
	   m[0][1] = m[2][1] = m[3][1] = 0;
	   m[0][2] = m[1][2] = m[3][2] = 0;
	   m[0][3] = m[1][3] = m[2][3] = 0;
    }
    static CMatrix getIdentity();
    void loadOpenGLMatrix(const float* mi);
    void storeOpenGLMatrix(float* mi);
   inline void LoadZero()
   {
     m[0][0] = m[1][1] = m[2][2] = m[3][3] = 0;
	   m[1][0] = m[2][0] = m[3][0] = 0;
	   m[0][1] = m[2][1] = m[3][1] = 0;
	   m[0][2] = m[1][2] = m[3][2] = 0;
	   m[0][3] = m[1][3] = m[2][3] = 0;
   };
   bool isValid() const;
    void SetXRotation(utifloat value);
    void SetYRotation(utifloat value);
    void SetZRotation(utifloat value);
    void SetTranslate(utifloat _x, utifloat _y, utifloat _z);
    void SetScaling(utifloat _x, utifloat _y, utifloat _z);
    void Rotate(const NetRotationInfo& info);
	inline void Rotate(const CVector3& v) {Rotate(v.x, v.y, v.z);};
    inline void RotateInv(const CVector3& v) {RotateInv(v.x, v.y, v.z);};
    void Rotate(utifloat _x, utifloat _y, utifloat _z);
    void RotateInv(utifloat _x, utifloat _y, utifloat _z);
    void ShearXfromY(utifloat x);
    void ShearYfromX(utifloat x);

	void RotateQuaternion(const CVector4 & quat);
    void SetRotateQuaternion(const CVector4 & quat2);
	CVector4 GetRotationAsQuaternion() const;
    CVector4f GetRotationAsQuaternionf() const;
    void SetEulerRotation(utifloat x, utifloat y, utifloat z);
    void SetEulerRotationOld(utifloat x, utifloat y, utifloat z);
    void EulerRotation(utifloat x, utifloat y, utifloat z);
    void EulerRotation(const CVector3& v);
    void Rotation(const NetRotationInfo& rot);
    void Translation(utifloat _x, utifloat _y, utifloat _z);
    static CVector rotateEuler(const CVector& rot, const CVector& point);
    void Scal(utifloat x, utifloat y, utifloat z);
    void Scal(const CVector3& v);
    void RotateAroundPoint(const CVector3& point, utifloat angle, int axis);
	void ScalAroundPoint(const CVector3& point, const CVector3& scal);
    void ShearAroundPoint(const CVector3& point, utifloat amount, int axis);
    static CMatrix TranslationMatrix(utifloat x, utifloat y, utifloat z);
	inline void Translation(const CVector& v) { Translation(v.x, v.y, v.z);};
    void Scaling(utifloat _x, utifloat _y, utifloat _z);
        void SetFrustum(utifloat l, utifloat r, utifloat b, utifloat t, utifloat n, utifloat f);
        void SetOrtho(utifloat l, utifloat r, utifloat b, utifloat t, utifloat n, utifloat f);
        void SetPerspective(utifloat fovy, utifloat aspect, utifloat near_plane, utifloat far_plane);
        void Frustum(utifloat l, utifloat r, utifloat b, utifloat t, utifloat n, utifloat f);
        void Ortho(utifloat l, utifloat r, utifloat b, utifloat t, utifloat n, utifloat f);
        void Perspective(utifloat fovy, utifloat aspect, utifloat near_plane, utifloat far_plane);
	void multipyMatrix(const CMatrix & b);
	void print()const ;
	C3DPoint Transform(const C3DPoint & b)const ;
	//void Transform(S3DPoint *p);
	//void Transform(CVector *p);
	//void TransformP(CVector *p);
	inline CVector TransformNormal(const CVector& p) const
	{
	  CVector r;
	  r.x = p.x * m[0][0] + p.y * m[0][1] + p.z * m[0][2];
          r.y = p.x * m[1][0] + p.y * m[1][1] + p.z * m[1][2];
           r.z = p.x * m[2][0] + p.y * m[2][1] + p.z * m[2][2];
	   return r;
	};
    inline CVector3f TransformNormal(const CVector3f& p) const
    {
      CVector3f r;
      r.x = p.x * m[0][0] + p.y * m[0][1] + p.z * m[0][2];
          r.y = p.x * m[1][0] + p.y * m[1][1] + p.z * m[1][2];
           r.z = p.x * m[2][0] + p.y * m[2][1] + p.z * m[2][2];
       return r;
    };
	inline CVector2 transform2(const CVector2& p) const
	{
	  CVector2 r;
	  r.x = p.x * m[0][0] + p.y * m[0][1] + 0 * m[0][2]+ m[0][3];
          r.y = p.x * m[1][0] + p.y * m[1][1] + 0* m[1][2]+ m[1][3];
          // r.z = p.x * m[2][0] + p.y * m[2][1] + p.z * m[2][2];
	   return r;
	};
    inline CVector2f transform2(const CVector2f& p) const
    {
      CVector2f r;
      r.x = p.x * m[0][0] + p.y * m[0][1] + 0 * m[0][2]+ m[0][3];
          r.y = p.x * m[1][0] + p.y * m[1][1] + 0* m[1][2]+ m[1][3];
          // r.z = p.x * m[2][0] + p.y * m[2][1] + p.z * m[2][2];
       return r;
    };
   inline void Transform(CVector *p)const
   {
      utifloat x, y, z;
      x = p->x * m[0][0] + p->y * m[0][1] + p->z * m[0][2];
      y = p->x * m[1][0] + p->y * m[1][1] + p->z * m[1][2];
      z = p->x * m[2][0] + p->y * m[2][1] + p->z * m[2][2];
      p->x = x;
      p->y = y;
      p->z = z;
   }


   inline void Transform(CVector3f *p)const
   {
      utifloat x, y, z;
      x = p->x * m[0][0] + p->y * m[0][1] + p->z * m[0][2];
      y = p->x * m[1][0] + p->y * m[1][1] + p->z * m[1][2];
      z = p->x * m[2][0] + p->y * m[2][1] + p->z * m[2][2];
      p->x = x;
      p->y = y;
      p->z = z;
   }
   inline CVector getPosition() const
   {
        return CVector(m[0][3], m[1][3], m[2][3]);
   };
   inline CVector3f getPositionf() const
   {
        return CVector3f(m[0][3], m[1][3], m[2][3]);
   };
   inline void setPosition(const CVector& v)
   {
       m[0][3] = v.x;
       m[1][3] = v.y;
       m[2][3] = v.z;
   }
   inline CVector getScaling() const
   {
       return CVector(CVector(m[0][0], m[1][0],m[2][0]).length(),
               CVector(m[0][1], m[1][1],m[2][1]).length(),
               CVector(m[0][2], m[1][2],m[2][2]).length());
   }

   inline void TransformP(CVector *p)const
   {
      utifloat x, y, z;
      x = p->x * m[0][0] + p->y * m[0][1] + p->z * m[0][2] + m[0][3];
      y = p->x * m[1][0] + p->y * m[1][1] + p->z * m[1][2] + m[1][3];
      z = p->x * m[2][0] + p->y * m[2][1] + p->z * m[2][2] + m[2][3];

      p->x = x;
      p->y = y;
      p->z = z;
   }
   inline void TransformNormal(CVector3f *p)const
   {
	   utifloat x, y, z;
	   x = p->x * m[0][0] + p->y * m[0][1] + p->z * m[0][2] ;
	   y = p->x * m[1][0] + p->y * m[1][1] + p->z * m[1][2];
	   z = p->x * m[2][0] + p->y * m[2][1] + p->z * m[2][2];

	   p->x = x;
	   p->y = y;
	   p->z = z;
   }

   inline CVector TransformP(const CVector &p)const
   {
     CVector res;

      res.x = p.x * m[0][0] + p.y * m[0][1] + p.z * m[0][2] + m[0][3];
      res.y = p.x * m[1][0] + p.y * m[1][1] + p.z * m[1][2] + m[1][3];
      res.z = p.x * m[2][0] + p.y * m[2][1] + p.z * m[2][2] + m[2][3];

      return res;
   }
   inline CVector3f TransformP(const CVector3f &p)const
   {
     CVector3f res;

      res.x = p.x * m[0][0] + p.y * m[0][1] + p.z * m[0][2] + m[0][3];
      res.y = p.x * m[1][0] + p.y * m[1][1] + p.z * m[1][2] + m[1][3];
      res.z = p.x * m[2][0] + p.y * m[2][1] + p.z * m[2][2] + m[2][3];

      return res;
   }

   inline void TransformP(CVector3f *p)const
   {
      utifloat x, y, z;
      x = p->x * m[0][0] + p->y * m[0][1] + p->z * m[0][2] + m[0][3];
      y = p->x * m[1][0] + p->y * m[1][1] + p->z * m[1][2] + m[1][3];
      z = p->x * m[2][0] + p->y * m[2][1] + p->z * m[2][2] + m[2][3];

      p->x = x;
      p->y = y;
      p->z = z;
   }


   inline CVector getXAxis() const
   {
       return CVector(m[0][0], m[1][0], m[2][0]);
   }
   inline CVector getYAxis() const
   {
       return CVector(m[0][1], m[1][1], m[2][1]);
   }
   inline CVector getZAxis() const
   {
       return CVector(m[0][2], m[1][2], m[2][2]);
   }

   inline void setXAxis(const CVector3& v)
   {
       m[0][0] = v.x;
       m[1][0] = v.y;
       m[2][0] = v.z;
   }
   inline void setYAxis(const CVector3& v)
   {
       m[0][1] = v.x;
       m[1][1] = v.y;
       m[2][1] = v.z;
   }
   inline void setZAxis(const CVector3& v)
   {
       m[0][2] = v.x;
       m[1][2] = v.y;
       m[2][2] = v.z;
   }

   CBox TransformP(const CBox& box) const;
   inline CVector TransformProjection(const CVector &p)const
   {
     CVector res;

      res.x = p.x * m[0][0] + p.y * m[0][1] + p.z * m[0][2] + m[0][3];
      res.y = p.x * m[1][0] + p.y * m[1][1] + p.z * m[1][2] + m[1][3];
      res.z = p.x * m[2][0] + p.y * m[2][1] + p.z * m[2][2] + m[2][3];
      utifloat w = p.x * m[3][0] + p.y * m[3][1] + p.z * m[3][2] + m[3][3];
     // float ab = w;
      if (w < 0.0001f && w > -0.0001f) {w = 0.0001f;}
      res /= w;
      return res;
   }

   inline void Transform(S3DPoint *p)const
   {
      utifloat x, y, z;
      x = p->x * m[0][0] + p->y * m[0][1] + p->z * m[0][2] + m[0][3];
      y = p->x * m[1][0] + p->y * m[1][1] + p->z * m[1][2] + m[1][3];
      z = p->x * m[2][0] + p->y * m[2][1] + p->z * m[2][2] + m[2][3];
      p->x = x;
      p->y = y;
      p->z = z;
    }

   inline void Transform(S3DPointf *p)const
   {
      utifloat x, y, z;
      x = p->x * m[0][0] + p->y * m[0][1] + p->z * m[0][2] + m[0][3];
      y = p->x * m[1][0] + p->y * m[1][1] + p->z * m[1][2] + m[1][3];
      z = p->x * m[2][0] + p->y * m[2][1] + p->z * m[2][2] + m[2][3];
      p->x = x;
      p->y = y;
      p->z = z;
    }
    inline utifloat& operator()(int i, int j)
    {
        utiassert(i >= 0 && i < 4);
        utiassert(j >= 0 && j < 4);
        return m[i][j];

    }
    inline utifloat operator()(int i, int j) const
    {
        utiassert(i >= 0 && i < 4);
        utiassert(j >= 0 && j < 4);
        return m[i][j];

    }

    inline void set(int i, int j, utifloat f)
    {
        utiassert(i >= 0 && i < 4);
        utiassert(j >= 0 && j < 4);
        m[i][j] = f;

    }
    inline utifloat get(int i, int j) const
    {
        utiassert(i >= 0 && i < 4);
        utiassert(j >= 0 && j < 4);
        return m[i][j];

    }

	void TransformNorm(S3DPoint *p)const ;
	void TransformAll(S3DPoint *p)const ;
	void Transpose(CMatrix* target);
	CMatrix operator*(float value) const
	{
		CMatrix res;
		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {

				res.m[i][j] = m[i][j] * value;
			}
		}
		return res;
	}
	CMatrix operator+(const CMatrix & b) const
	{
		CMatrix res;
		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {

				res.m[i][j] = m[i][j] + b.m[i][j];
			}
		}
		return res;
	}
	CMatrix operator*(const CMatrix & b) const;
	bool operator==(const CMatrix & b);
    bool numericalCompare(const CMatrix & b, utifloat threshold);
    utifloat numericalDistance(const CMatrix & b);
	CVector4 getRow(int index) const;
	void setRow(int index, const CVector4& v);
	CVector4 getColumn(int index)const;
	void setColumn(int index, const CVector4& v);
	inline void SetMatrix(const CMatrix* m)
    {
		*this = *m;/*
	   for (int i = 0; i<4; i++) {
		//for (int j = 0; j<4; j++) {
           this->m[i][0] = m->m[i][0];
		   this->m[i][1] = m->m[i][1];
		   this->m[i][2] = m->m[i][2];
		   this->m[i][3] = m->m[i][3];
		//}
	   };*/
    }
    CMatrix getInverse() const;
    Matrix3 getRotationMatrix();
};

inline CMatrixf toFloat(const CMatrix& r)
{
    CMatrixf res;
    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            res.m[i][j] = r.m[i][j];

        }

    }
    return res;
}

inline CMatrix toDouble(const CMatrixf& r)
{
    CMatrix res;
    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            res.m[i][j] = r.m[i][j];

        }

    }
    return res;
}

CVector rotateEuler(const CVector& point, const CVector& rot);

void utianalysematrix(Matrix3* m, utifloat &a, utifloat &b, utifloat &c);

typedef utifloat TUtiMatrix[16];

struct SClipResult {
   LPVOID oldpoly;
   UINT count;
   S3DPoint points[20];
};

#define BSP_FRONT 1
#define BSP_BACK -1
#define BSP_CUT 2
#define BSP_SAME 0

struct uvector
{
    float x,y,z;
};
/*
struct CVector2 {
   float x;
   float y;
};

struct CVector4 {
   float x;
   float y;
   float z;
   float w;
};

typedef CVector CVector3;*/

void ReadVector(CVector* v,CStream & Stream);
void WriteVector(CVector* v, CStream & Stream);

struct CVertex
{
public:
    int Flags;
    float weight;
    CVector3f pos;
    CVector3f normal;
};

struct SHitResult;

struct CPlanef;

struct CPlane
{
public:
	int ClassifyPolygon(S3DPoint* list, int count); //Gibt BSP_FRONT, BSP_BACK, BSP_CUT, BSP_SAME zurck
	int ClassifyPolygon(UINT* list, CVertex* v, int count); //Gibt BSP_FRONT, BSP_BACK, BSP_CUT, BSP_SAME zurck
   void cliptoplane(S3DPoint *in, int countin, S3DPoint *out, int &countout, int maxcount);
    void cliptoplane(CVector *in, int countin, CVector *out, int &countout, int maxcount);
    void cliptoplane(S3DPointf *in, int countin, S3DPointf *out, int &countout, int maxcount);
     void cliptoplane(CVector3f *in, int countin, CVector3f *out, int &countout, int maxcount);
	int getPlanePos(const C3DPoint & p);
	int getPlanePos(const S3DPoint & p);
	int getPlanePos(const CVector &p);
    int getPlanePos(const CVector3f &p);
    static bool IntersectTriangle(S3DPoint* p1, S3DPoint* p2, S3DPoint* p3, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);
    static bool IntersectTriangle(S3DUNLIT* p1, S3DUNLIT* p2, S3DUNLIT* p3, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);
    static bool IntersectTriangleWithV(S3DPoint* p1, S3DPoint* p2, S3DPoint* p3, CPlane *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);
    static bool IntersectTriangleWithV(CVector* p1, CVector* p2, CVector* p3, CPlane *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);
    static bool IntersectTriangleWithV(CVector3f* p1, CVector3f* p2, CVector3f* p3, CPlanef *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);
    static bool IntersectTriangleWithV(CVector** points, CPlane *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);
    static bool IntersectTriangle(const CVector* p1, const CVector* p2, const CVector* p3, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);
    static bool IntersectTriangle(const CVector& pos, const CVector& direction, const CVector& p1, const CVector& p2, const CVector& p3, utifloat &distance);
    static bool IntersectTriangleWithV(CVector3f** points, CPlanef *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);


    static bool IntersectTriangle(const CVector3f* p1, const CVector3f* p2, const CVector3f* p3, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec);

    static bool PointInTriangle(const CVector* p1, const CVector* p2, const CVector* p3, const CVector3* point);
    static CPlane createPlane(const CVector3& position, const CVector3& normal);
    CVector3 cvVector;
    utifloat d;
	//CPlane();

	void GetNormal(S3DUNLIT *p1, S3DUNLIT *p2, S3DUNLIT *p3);
	//float PlaneDistance(C3DPoint & p);
	//float PlaneDistance(S3DPoint & p);
	//float PlaneDistance(CVector & p);
	/*inline float CPlane::PlaneDistance(C3DPoint & p)
    {
       return(cvVector.dotProduct(p)-d);
    }*/

    inline utifloat PlaneDistance(const S3DPoint & p)
    {
       return(cvVector.dotProduct(*((CVector*)&p.x))-d);
	}
    inline utifloat PlaneDistance(const CVector &p)
{
   return(cvVector.dotProduct(p)-d);
}
    inline utifloat PlaneDistance(const CVector3f &p)
{
   return((cvVector.x* p.x+cvVector.y* p.y+cvVector.z* p.z)-d);
}
    bool IntersectLine(CVector* Start, CVector* Dir, utifloat &dis);
	int getBoundingPos(S3DPoint *p, int count);
	void GetNormal(S3DPoint *p);
	void GetNormal(S3DPoint *p1, S3DPoint *p2, S3DPoint *p3);
   void GetNormal(const CVector *p1, const CVector *p2,const CVector *p3);
   void GetNormal(const CVector3f *p1, const CVector3f *p2,const CVector3f *p3);
   void getPlane(const CVector &p1, const CVector &p2,const CVector &p3);
   CVector3 pointOnPlane(const CVector3& v);
      static CVector getNormal(const CVector &p1, const CVector &p2,const CVector &p3);
      static CVector3f getNormal(const CVector3f &p1, const CVector3f &p2,const CVector3f &p3);
	static const S3DPoint GetNormal(S3DPoint p1, S3DPoint p2, S3DPoint p3, S3DPoint *res);
};

struct CPlanef
{
public:

    CVector3f cvVector;
    float d;

    inline utifloat PlaneDistance(const CVector3f &p)
{
   return((cvVector.x* p.x+cvVector.y* p.y+cvVector.z* p.z)-d);
}
    inline utifloat PlaneDistance(const S3DPoint & p)
    {
       return((cvVector.x* p.x+cvVector.y* p.y+cvVector.z* p.z)-d);
    }
    void GetNormal(const CVector3f *p1, const CVector3f *p2,const CVector3f *p3);
    void getPlane(const CVector3f &p1, const CVector3f &p2,const CVector3f &p3);
    void cliptoplane(S3DPointf *in, int countin, S3DPointf *out, int &countout, int maxcount);
};

inline CPlanef toFloat(const CPlane& p) {
    CPlanef res;
    res.cvVector.setVector(p.cvVector.x, p.cvVector.y, p.cvVector.z);
    res.d = p.d;
    return res;
}


inline CPlane toDouble(const CPlanef& p) {
    CPlane res;
    res.cvVector.setVector(p.cvVector.x, p.cvVector.y, p.cvVector.z);
    res.d = p.d;
    return res;
}
struct SEntity
{
	SEntity(){valid = false;added=false;Group=0;Parent=0;index=0;};
    SVector Position;
	SVector Rotation;
	UINT Group;
	bool valid;
	UINT Parent;
	int index;
	bool added;
};

class CEntity : public CUtiOB
{
private:

public:
	CEntity(){valid = false;Position.setVector(0, 0, 0);Rotation.setVector(0, 0, 0);};
	CVector Position;
	CVector Rotation;
	UINT Group;
	bool valid;
	CEntity& operator= (const SEntity &a);
};




bool Has2DPointZ(S3DPoint *points, int npol, S3DPoint *point);
bool Has2DPointY(S3DPoint *points, int npol, S3DPoint *point);
bool Has2DPointX(S3DPoint *points, int npol, S3DPoint *point);

bool Has2DPointZ(CVector **points, int npol, S3DPoint *point);
bool Has2DPointY(CVector **points, int npol, S3DPoint *point);
bool Has2DPointX(CVector **points, int npol, S3DPoint *point);

bool Has2DPointZ(CVector3f **points, int npol, S3DPoint *point);
bool Has2DPointY(CVector3f **points, int npol, S3DPoint *point);
bool Has2DPointX(CVector3f **points, int npol, S3DPoint *point);


bool Has2DPoint2D(const Array<CVector2> &points, int npol, const CVector2 &point);
bool Has2DPoint2D(const Array<CVector2f>& points, int npol, const CVector2f &point);

utifloat SearchRoundedAreaWidthZ(S3DPoint *points, int npol);
utifloat SearchRoundedAreaWidthY(S3DPoint *points, int npol);
utifloat SearchRoundedAreaWidthX(S3DPoint *points, int npol);

void operator += (S3DPoint &a, const S3DPoint &b);
void operator += (S3DPoint &a, const CVector &b);
void operator *= ( S3DPoint &a, const float b );
S3DPoint operator * (const S3DPoint &a, const float b);
S3DPoint operator - ( const S3DPoint &a, const S3DPoint &b );
S3DPoint operator - ( const S3DPoint &a, const CVector &b )	;
S3DPoint operator + ( const S3DPoint &a, const S3DPoint &b );
S3DPoint operator + ( const S3DPoint &a, const CVector &b );
S3DPoint operator ^ ( const S3DPoint &a, const S3DPoint &b );
S3DPoint operator ^ ( const S3DPoint &a, const CVector &b );
float operator * ( const S3DPoint &a, const S3DPoint &b );
float operator * ( const S3DPoint &a, const CVector &b );


//float Interpolatef(float a, float b, float value);
//float InterpolateCosinef(float a, float b, float x);

float normalize(float*v);
double normalize(double*v);
float Length(float*v);
double Length(double*v);
int ClassifyPoints(S3DPoint* points, int count, CPlane*plane);
//float Interpolate(float a, float b, float x);
//float InterpolateRectangle(float aa, float ba, float ab, float bb, float x, float y);

void IntersectQuaderRayMinMax(const S3DPoint& vector, const CBox& b, utifloat &t_min, utifloat &t_max);
void GetRandomPointInRect(CVector* p, CBox* r);
void GetRandomPointInRect(S3DPoint* p, CBox* r);

CVector2 solveLinearEquation2(const CVector2&v1, const CVector2&v2, const CVector2& c, bool* hassolution);

CVector2 intersectLines(const CVector2&p1, const CVector2&dir1, const CVector2& p2, const CVector2& dir2, bool* hassolution);
utifloat angleBetween(const CVector2&v1, const CVector2&v2);

#define UTI_INLINE_DOT(a, b) ((a).x*(b).x+(a).y*(b).y+(a).z*(b).z)
#define UTI_INLINE_INIT(a) a.x = a.y = a.z = 0;

void calcBumpTangent(const CVector3f &P, const CVector3f &Q, const CVector3f &N, float s1, float s2, float t1, float t2, CVector3f &tangent, float &m);

int inline fasterfloor( const float x )
{
   /* return x < 0 ? ((int) x == x ?
                        (int) x :
                        (int) x -1)
                 : (int) x;*/
     return (int) x - (x < (int) x);
}

int inline fasterfloor( const double x )
{
     return (int) x - (x < (int) x);
  /*  return x < 0 ? ((int) x == x ?
                        (int) x :
                        (int) x -1)
                 : (int) x;*/
}

CVector3 rgbToHSV(const CVector3& v);
CVector3 rgbToHSV(const CVector3& v, const CVector3& oldhsv);
CVector3 hsvToRGB(const CVector3& v);


CVector3f rgbToHSV(const CVector3f& v);
CVector3f rgbToHSV(const CVector3f& v, const CVector3f& oldhsv);
CVector3f hsvToRGB(const CVector3f& v);



void transformTranspose(const CMatrix& m, const CVector4 &v, CVector4 &res);

utifloat rangecalc(utifloat value, utifloat min, utifloat max, utifloat bordermin, utifloat bordermax);
inline float uclamp(float value, float min, float max) {if (value < min){ value = min; }if (value > max){ value = max; }return value;};
inline double uclamp(double value, double min, double max) {if (value < min) {value = min; }if (value > max) {value = max;} return value;};
inline int uclamp(int value, int min, int max) {if (value < min) {value = min;} if (value > max){ value = max;} return value;};
inline CVector4 uclamp(CVector4 value, const CVector4& min, const CVector4& max) {if (value.x < min.x){ value.x = min.x;} if (value.x > max.x){ value.x = max.x;}
if (value.y < min.y) {value.y = min.y;} if (value.y > max.y) {value.y = max.y;}
if (value.z < min.z){ value.z = min.z; }if (value.z > max.z) {value.z = max.z;}
if (value.w < min.w) {value.w = min.w;} if (value.w > max.w) {value.w = max.w;}
return value;};
inline CVector3 uclamp(CVector3 value, const CVector3& min, const CVector3& max) {if (value.x < min.x) {value.x = min.x; }if (value.x > max.x){ value.x = max.x;}
if (value.y < min.y) {value.y = min.y; }if (value.y > max.y){ value.y = max.y;}
if (value.z < min.z) {value.z = min.z; }if (value.z > max.z){ value.z = max.z;}
return value;};
inline CVector4f uclamp(CVector4f value, const CVector4f& min, const CVector4f& max) {if (value.x < min.x){ value.x = min.x;} if (value.x > max.x) {value.x = max.x;}
if (value.y < min.y) {value.y = min.y;} if (value.y > max.y){ value.y = max.y;}
if (value.z < min.z) {value.z = min.z;} if (value.z > max.z){ value.z = max.z;}
if (value.w < min.w) {value.w = min.w;} if (value.w > max.w){ value.w = max.w;}
return value;};
inline CVector3f uclamp(CVector3f value, const CVector3f& min, const CVector3f& max) {if (value.x < min.x){ value.x = min.x; }if (value.x > max.x) {value.x = max.x;}
if (value.y < min.y) {value.y = min.y;} if (value.y > max.y) {value.y = max.y;}
if (value.z < min.z) {value.z = min.z; }if (value.z > max.z) {value.z = max.z;}
return value;};


inline int binround(utifloat f)
{
  if (f >=0){
  return (int)(f+0.5f);
  } else {
    return (int)(f-0.5f);
  };
  return 0;
};

union i_cast { double d; int i[2]; };
#define double2int_lua(i, d, t)  \
    {volatile union i_cast u; u.d = double(d) + 6755399441055744.0; \
    (i) = (t)u.i[ENDIANLOC];}

#define double2int_simpler(i, d) \
    {double t = ((d) + 6755399441055744.0); i = *((int *)(&t));}

inline int double2int(double d)
{
	d += 6755399441055744.0;
	return reinterpret_cast<int&>(d);
}

int expandTo2(int w);


int bintrunc(double f);
int bintruncd(double f);

inline int bindivide(int a, int b)
{
  utiassert(b >= 1);

  int res = a / b;
    if ((a < 0) && (a % b != 0)) res--;
    return res;
};



inline float Interpolatef(float a, float b, float value)
{
   return(a+(b-a)*value);
}

inline double Interpolatef(double a, double b, double value)
{
   return(a+(b-a)*value);
}


inline CVector2f Interpolatef(const CVector2f &a, const CVector2f &b, float value)
{
   return(a+(b-a)*value);
}

inline CVector2 Interpolatef(const CVector2 &a, const CVector2 &b, double value)
{
   return(a+(b-a)*value);
}


inline CVector3f Interpolatef(const CVector3f &a, const CVector3f &b, float value)
{
   return(a+(b-a)*value);
}

inline CVector3 Interpolatef(const CVector3 &a, const CVector3 &b, double value)
{
   return(a+(b-a)*value);
}

inline CVector4f Interpolatef(const CVector4f &a, const CVector4f &b, float value)
{
   return(a+(b-a)*value);
}

inline CVector4 Interpolatef(const CVector4 &a, const CVector4 &b, double value)
{
   return(a+(b-a)*value);
}


inline float sin_low_precision(float x)
{
 //compute cosine: sin(x + PI/2) = cos(x)
if (x < -3.14159265)
    x += 6.28318531;
else
if (x >  3.14159265)
    x -= 6.28318531;

//compute sine
if (x < 0)
    return 1.27323954 * x + .405284735 * x * x;
else
    return 1.27323954 * x - 0.405284735 * x * x;

}

inline float cos_low_precision(float x)
{
 //compute cosine: sin(x + PI/2) = cos(x)
x += 1.57079632;
if (x >  3.14159265)
    x -= 6.28318531;

if (x < 0)
    return 1.27323954 * x + 0.405284735 * x * x;
else
    return 1.27323954 * x - 0.405284735 * x * x;

}
inline float sin_high_precision(float x)
{
//always wrap input angle to -PI..PI
if (x < -3.14159265)
    x += 6.28318531;
else
if (x >  3.14159265)
    x -= 6.28318531;
float s;
//compute sine
if (x < 0)
{
    s = 1.27323954 * x + .405284735 * x * x;

    if (s < 0)
        s = .225 * (s *-s - s) + s;
    else
        s = .225 * (s * s - s) + s;
}
else
{
    s = 1.27323954 * x - 0.405284735 * x * x;

    if (s < 0)
        s = .225 * (s *-s - s) + s;
    else
        s = .225 * (s * s - s) + s;

}
return s;
}

inline float cos_high_precision(float x)
{

//compute cosine: sin(x + PI/2) = cos(x)
x += 1.57079632;
if (x >  3.14159265)
    x -= 6.28318531;
float c;
if (x < 0)
{
    c = 1.27323954 * x + 0.405284735 * x * x;

    if (c < 0)
        c = .225 * (c *-c - c) + c;
    else
        c = .225 * (c * c - c) + c;
}
else
{
    c = 1.27323954 * x - 0.405284735 * x * x;

    if (c < 0)
        c = .225 * (c *-c - c) + c;
    else
        c = .225 * (c * c - c) + c;
}
return c;
}

inline float cos_taylor(float x)
{
     float x2 = x*x;
     float x4 = x*x*x*x;
     return 1-x2*0.5f+x4*0.0416666666666667-x2*x4*0.00138888888888888888888888888889;
}

inline float cos_taylor8(float x)
{
     float x2 = x*x;
     float x4 = x*x*x*x;
     float x6 = x2*x4;
     return 1-x2*0.5f+x4*0.0416666666666667-x6*0.00138888888888888888888888888889+x6*x2*2.4801587301587301587301587301587e-5;
}

inline float InterpolateCosinef(float a, float b, float x)
{
   /*float value = (1.0f - (float)cos(x*PI))*0.5f;
   return(a*(1-value)+b*value);*/
    assert(x > -0.01 && x <= 1.01);
    if (x < 0.5f) {
         float ft = x* M_PI;

         ft = (1-cos_taylor8(ft))*0.5;
         float res;
         res = a*(1-ft)+b*ft;
         return res;
         } else {
         float ft = (1.0f-x)* M_PI;

         ft = (1+cos_taylor8(ft))*0.5;
         float res;
         res = a*(1-ft)+b*ft;
         return res;
                }
};

inline double InterpolateCosinef(double a, double b, double x)
{
   /*float value = (1.0f - (float)cos(x*PI))*0.5f;
   return(a*(1-value)+b*value);*/
    assert(x > -0.01 && x <= 1.01);
    if (x < 0.5f) {
         double ft = x* M_PI;

         ft = (1-cos_taylor8(ft))*0.5;
         double res;
         res = a*(1-ft)+b*ft;
         return res;
         } else {
         double ft = (1.0f-x)* M_PI;

         ft = (1+cos_taylor8(ft))*0.5;
         double res;
         res = a*(1-ft)+b*ft;
         return res;
                }
};
/*
inline float InterpolateCosinef(float a, float b, float x)
{
   float value = (1.0f - (float)cos(x*PI))*0.5f;
   return(a*(1-value)+b*value);
};*/
inline float Interpolate(float a, float b, float x)
{
   return a + (b-a)*x;
};

inline double Interpolate(double a, double b, double x)
{
   return a + (b-a)*x;
};

inline float InterpolateRectangle(float aa, float ba, float ab, float bb, float x, float y)
{
   float a = Interpolate(aa, ba, x);
   float b = Interpolate(ab, bb, x);
   return Interpolate(a, b, y);
};


inline double InterpolateRectangle(double aa, double ba, double ab, double bb, double x, double y)
{
   double a = Interpolate(aa, ba, x);
   double b = Interpolate(ab, bb, x);
   return Interpolate(a, b, y);
};


inline void uti_moddiv(int value, int divider, int &mod, int &res)
{
  res = value / divider;
  if ((value < 0) && ((value % divider) != 0)) res--;
  mod = (value+(abs(res) +3)*divider) % divider;
  utiassert(mod >= 0);
  utiassert(mod < divider);
  utiassert(res*divider+mod == value);
};

inline void uti_moddiv(long value, long divider, long &mod, long &res)
{
	res = value / divider;
	if ((value < 0) && ((value % divider) != 0)) res--;
    mod = (value + (labs(res) + 3)*divider) % divider;
	utiassert(mod >= 0);
	utiassert(mod < divider);
	utiassert(res*divider + mod == value);
};



inline int uti_div(int value, int divider)
{   int res;

    res = value / divider;
    if ((value < 0) && ((value % divider) != 0)) res--;
    return res;
}

inline int uti_mod(int value, int divider)
{   int res;
    int mod;

    res = value / divider;
    if ((value < 0) && ((value % divider) != 0)) res--;
    mod = (value + (abs(res) + 3)*divider) % divider;

    utiassert(mod >= 0);
    utiassert(mod < divider);
    utiassert(res*divider + mod == value);
    return mod;
}


inline bool hasNan(const CVector& v)
{
  return utiisnan(v.x) || utiisnan(v.y) || utiisnan(v.z);
};

inline bool utiIsInf(float v)
{
	return !utiisnan(v) &&utiisnan(v - v);
}


static inline unsigned int intlog2(const unsigned int x) {
  unsigned int y; //Die Typen waren hier alle uint32_t, checke auf Fehler
#ifdef _MSC_VER
 y = 0;
  int index2 = x;
  while (index2 >>= 1) ++y;
 
#else
  //Works only with power of two
 /* asm ( "\tbsr %1, %0\n"
      : "=r"(y)
      : "r" (x)
  );*/
  y = 0;
   int index2 = x;
   while (index2 >>= 1) ++y;
#endif
#ifndef NDEBUG
  int targetlevel = 0;
  int index = x;
  while (index >>= 1) ++targetlevel;
  assert(y == targetlevel);
#endif

  return y;
}


unsigned int UtiRGBA(const CVector4& color);

void getLocalCoordinateSystemByVector(const CVector& input, CVector &xp, CVector &yp, CVector &zp);
CVector2 rotateVector(const CVector2& p, utifloat degree);
CVector2f rotateVector(const CVector2f& p, float degree);

float distanceToRectangle(const CRect& rect, const CVector2& point);
float distanceToRotatedRectangle(const CRect& rect, CVector2& origin, float rotation, const CVector2& point);

template<class T> inline T distanceToValue(T v1, T v2)
{
    v1 -= v2;
    if (v1 < 0) v1 = -v1;
    return v1;
}


CVector uVectorToAngle(const CVector& v2);

/** \addtogroup mathgroup
 *  @{
 */



/**
 * @brief The rotation type
 */
enum TNetRotationType {nrtEuler, nrtQuaternion};
/**
 * @brief The NetRotationInfo struct
 */
struct NetRotationInfo
{
    TNetRotationType m_rotationtype;
    CVector4 m_values;
    inline NetRotationInfo()
    {
        m_rotationtype = nrtEuler;
        m_values.setVector(0,0,0,0);
    }
    /**
     * @brief Returns the current rotation
     * @return
     */
    inline CVector4 getValues()
    {
        return m_values;
    }
    /**
     * @brief Sets the euler rotation or the quaternion
     * @param v A vector4
     */
    inline void setValues(const CVector4& v)
    {
        m_values = v;
    }
    /**
     * @brief Returns the rotation type
     * @return
     */
    inline TNetRotationType getRotationType()
    {
        return m_rotationtype;
    }
    /**
     * @brief setRotationType
     * @param r
     */
    inline void setRotationType(TNetRotationType r)
    {
        m_rotationtype = r;
    }

    /**
     * @brief NetRotationInfo
     * @param p
     */
    explicit inline NetRotationInfo(const CVector3& p){
        m_values = CVector4(p, 0);
        m_rotationtype = nrtEuler;
    }
    /**
     * @brief NetRotationInfo
     * @param p
     */
    explicit inline NetRotationInfo(const CVector4& p) {
        m_values = p;
        m_rotationtype = nrtQuaternion;
    }
    /**
     * @brief Returns the rotation as quaternion
     * @return A quaternion
     */
    inline CVector4 toQuaternion() const
    {
        if (m_rotationtype == nrtEuler) {
            return eulerToQuaternion(m_values.xyz());
        } else {
            return m_values;

        }
    }
    /**
     * @brief Returns the rotation as euler angles.
     * @return
     */
    inline CVector3 toEuler() const
    {
        assert(m_rotationtype == nrtEuler);
        return m_values.xyz();
    }
    /**
     * @brief Only returns the y rotation
     * @return
     */
    inline utifloat getYRotation() const
    {
        assert(m_rotationtype == nrtEuler);
        return m_values.y;
    }
    inline utifloat distanceTo(const NetRotationInfo& info)const
    {
        return m_values.distanceTo(info.m_values)+((info.m_rotationtype != info.m_rotationtype)  ? 100.0 : 0.0);
    }
    /**
     * @brief isEuler
     * @return
     */
    inline bool isEuler() const
    {
        return m_rotationtype == nrtEuler;
    }

    /**
     * @brief isQuaternion
     * @return
     */
    inline bool isQuaternion() const
    {
        return m_rotationtype == nrtQuaternion;
    }
    inline bool operator ==(const NetRotationInfo& info)
    {
        return m_rotationtype == info.m_rotationtype && m_values == info.m_values;
    }
};


/**
 * @brief Converts a vector to euler rotation
 * @param vector A vector
 * @return The euler rotation
 */
CVector vectorToEuler(const CVector& vector);

/**
 * @brief Converts a vector to a rotation
 * @param vector The direction vector
 * @return The rotation
 */
NetRotationInfo vectorToRotation(const CVector& vector);

/**
 * @brief convertToString
 * @param info
 * @return
 */
CUtiString convertToString(const NetRotationInfo& info);

/** @}*/


bool calcUmkreis(const CVector2& p1, const CVector2& p2, const CVector2& p3, utifloat &R, CVector2 &center);

inline CVector2f toFloat(const CVector2& v)
{
    return CVector2f(v.x, v.y);
}
inline CVector2 toDouble(const CVector2f& v)
{
    return CVector2(v.x, v.y);

}

inline CVector3f toFloat(const CVector3& v)
{
    return CVector3f(v.x, v.y, v.z);
}
inline CVector3 toDouble(const CVector3f& v)
{
    return CVector3(v.x, v.y, v.z);

}

inline CVector4f toFloat(const CVector4& v)
{
    return CVector4f(v.x, v.y, v.z, v.w);
}
inline CVector4 toDouble(const CVector4f& v)
{
    return CVector4(v.x, v.y, v.z, v.w);

}

inline float spike(float value, float center, float radius)
{
    float minv = center-radius;
    float maxv = center+radius;
    if (value < minv) return 0;
    if (value > maxv) return 0;
    if (value < center) {
        return (value-minv)/radius;
    } else {
        return 1.0f-(value-center)/radius;
    }
}


//quakeC similar functions

/**
 * \defgroup mathgroup Math functions
 * This module contains all math functions which are implemented in the game engine
 */

/** \addtogroup mathgroup
 *  @{
 */

/**
 * @brief This function changes the angle so that it lies into the range 0 .. 360 without changing the direction.
 * @param angle The input angle
 * @return The output angle in the range 0 .. 360
 */
double anglemod(double angle);
/**
 * @brief This function converts a double value to a string
 * @param value The input value.
 * @return A string which contains the number
 */
CUtiString ftos(double value);
//double rint(double val);
/**
 * @brief Takes euler angles, creates a rotation matrix and returns the axis of the transformation.
 * @param angles Euler angles of the rotation
 * @param forward The vector which points forward
 * @param up The vector which points upwards
 * @param right The vector which points to the right
 */
void makevectors(const CVector3& angles, CVector3& forward, CVector3& up, CVector3& right);
/**
 * @brief Takes a vector as input and returns the euler angles. Can be used to orient an entity along a direction.
 * @param v The input vector. Does not need to be normalized
 * @return A vector which contains an euler angle
 */
CVector vectoangles(const CVector3& v);
/**
 * @brief Takes a vector as input and returns the y angle of the euler angles.
 * @param v The input vector. Does not need to be normalized
 * @return A float which contains the y angle.
 */
float vectoyaw(const CVector3& v);
/**
 * @brief Converts a vector to a string
 * @param v The input value in form of a vector
 * @return A string which contains the vector in form of (x, y, z)
 */
CUtiString vtos(const CVector3& v);

/**
 * @brief Converts an angle to a 2D vector
 * @param angle The angle in degrees
 * @return The result vector
 */
CVector2 angleToVector(utifloat angle);
/**
 * @brief Converts euler angles to a 3D vector
 * @param v The euler angle in degrees
 * @return The result vector
 */
CVector3 angleToVector(const CVector3 & v);

/**
 * @brief Converts euler angles to a 3D vector and then adds the origin
 * @param v The euler angle in degrees
 * @param origin The origin vector
 * @return  The result vector
 */
CVector3 angleToVector(const CVector3 & v, const CVector3& origin);

/**
 * @brief Returns the position on a line
 * @param s The origin vector of the line
 * @param normvector The direction vector of the line
 * @param point The test point
 * @return The result position
 */
utifloat positionOnRay(const CVector3& s, const CVector3& normvector, const CVector3& point);

/**
 * @brief Returns the distance between a vector and a line
 * @param s The origin vector of the line
 * @param t The direction vector of the line
 * @param point The test point
 * @return The result position
 */
utifloat distanceToLine(const CVector3& s, const CVector3& t, const CVector3& point);

/**
 * @brief Returns the position on a ray
 * @param s The origin vector of the line
 * @param normvector The direction vector of the line
 * @param point The test point
 * @return The result position
 */
utifloat positionOnRay2(const CVector2& s, const CVector2& normvector, const CVector2& point);

/**
 * @brief Returns the distance between a vector and a line
 * @param s The origin vector of the line
 * @param t The direction vector of the line
 * @param point The test point
 * @return The result position
 */
utifloat distanceToLine2(const CVector2& s, const CVector2& t, const CVector2& point);



/** @}*/

#endif

