/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Uti3DNode.h"
#include "utimathbox.h"

#include<assert.h>

MAKE_SINGLETON_IMPL(CMetaClassRegistry);

CUtiString intBoxToString(const CIntBox& box)
{
  CUtiString res;
  res += UTISTR("(");
  res += box.m_min.x;
  res += UTISTR(", ");
  res += box.m_min.y;
  res += UTISTR(", ");
  res += box.m_min.z;
  res += UTISTR(") - (");
  res += box.m_max.x;
  res += UTISTR(", ");
  res += box.m_max.y;
  res += UTISTR(", ");
  res += box.m_max.z;
  res += UTISTR(")");
  
  return res;
};

CNodeItem::CNodeItem(const CUtiString& _name, int _objid)
{
   name = CUtiIDString(_name);
   objid = _objid;
   position = 0;
   parent = NULL;
}

CNodeItem::~CNodeItem()
{

}

void CNodeItem::create(CNode* node)
{

}

void CNodeItem::destroy(CNode* node)
{

}

void CNodeItem::xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
{
  
}

void CNodeItem::xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
{
  
}


int CNodeItem::getSize()
{
   return 0;
}

void CNodeItem::setPosition(int pos)
{
	utiassert(pos >= 0);
    position = pos;
};

void CNodeItem::setParent(CNodeDescription* _parent)
{
	parent = _parent;
}

CNodeDescription* CNodeItem::getParent()
{
   return parent;
}

CNodeDescription::CNodeDescription()
{
   instances = 0;
   datasize = 0;
}

CNodeDescription::~CNodeDescription()
{
   utiassert(instances == 0);
}

CUtiIDString CNodeItem::getName()
{
   return name;
}


void CNodeItem::setName(const CUtiIDString& str)
{
    name = str;
}

CNodeItem* CNodeDescription::addItem(CNodeItem* item)
{
	CHECKEXISTANCEOF(item);
   utiassert(names.find(item->getName()) == names.end());
   names.insert(item->getName());
 //  utiassert(instances == 0);
   utiassert(datasize >= 0);
   item->setPosition(datasize);
   utiassert(item->getParent() == NULL);
   item->setParent(this);
   int s = item->getSize();
   utiassert(s > 0);
   while (s % 4 != 0) s++;
   datasize += s;
   Items.push_back(item);
   return item;
};


void CNodeDescription::createRange(CNode* node, int start)
{
    CHECKEXISTANCE;
    CHECKEXISTANCEOF(node);
 //  instances++;
   for (int i = start; i < Items.size(); i++) {
       Items.get(i)->create(node);
   };
}

void CNodeDescription::create(CNode* node)
{
    CHECKEXISTANCE;
    CHECKEXISTANCEOF(node);
   instances++;
   for (int i = 0; i < Items.size(); i++) {
       Items[i].create(node);
   };
};
void CNodeDescription::destroy(CNode* node)
{
    CHECKEXISTANCE;
    CHECKEXISTANCEOF(node);
    for (int i = 0; i < Items.size(); i++) {
       Items[i].destroy(node);
   };
	instances--;
}

void CNodeDescription::xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
{
  for (int i = 0; i < Items.size(); i++) {
       Items[i].xreadNode(node, n, wri);
   };
}

void CNodeDescription::xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
{
  for (int i = 0; i < Items.size(); i++) {
       Items[i].xwriteNode(node, n, wri);
   };
}

int CNodeDescription::getDataSize()
{
   return datasize;
};

CNode::CNode(CNodeDescription* _description, CUtiOB *parent)
{
        CHECKEXISTANCEOF(_description);
        m_parent = parent;
   description = _description;
   datasize = description->getDataSize();
   data.setSize(umax(4,description->getDataSize()));

   m_lastpropertycount = description->getNodeCount();
  // data = new unsigned char[datasize];
   description->create(this);
}

CNode::~CNode()
{
    CHECKEXISTANCEOF(description.get());
    description->destroy(this);
    // delete [] data;
};
CNodeDescription* CNode::getDescription()
{
    return description.get();
}


void CNode::updateStore()
{
    assert(m_lastpropertycount <= description->getNodeCount());
    while (data.size() < description->getDataSize()) {
        data.setSizeCopy(data.size()*2);
    }
    datasize = description->getDataSize();
    description->createRange(this, m_lastpropertycount);
    //Init
    m_lastpropertycount = description->getNodeCount();
}

void CNode::xread(CXNode* n, CWriteReadInterface* wri)
{
  description->xreadNode(this, n, wri);
}

void CNode::xwrite(CXNode* n, CWriteReadInterface* wri)
{
  description->xwriteNode(this, n, wri);
}

void encodeRotation(const NetRotationInfo& info, CNode*node, CVectorItem* rot, CVectorItem* rot2)
{
    if (info.m_rotationtype == nrtEuler) {
        rot->set(node, info.m_values.xyz());
        rot2->set(node, CVector(info.m_values.w, 0,0));
    } else {

        rot->set(node, info.m_values.xyz());
        rot2->set(node, CVector(info.m_values.w, 1,0));
    }
}

void decodeRotation(NetRotationInfo& info, CNode* node, CVectorItem* rot, CVectorItem* rot2)
{
    CVector r1 = rot->get(node);
    CVector r2 = rot2->get(node);
    info.m_values.x = r1.x;
    info.m_values.y = r1.y;
    info.m_values.z = r1.z;
    info.m_values.w = r2.x;
    if (r2.y > 0.5f) {
        info.m_rotationtype = nrtQuaternion;
    } else {
        info.m_rotationtype = nrtEuler;
    }
}

void encodeRotation(const NetRotationInfo& info,  CVector3& rot, CVector3& rot2)
{
    if (info.m_rotationtype == nrtEuler) {
        rot = info.m_values.xyz();
        rot2 = CVector(info.m_values.w, 0,0);
    } else {

        rot = info.m_values.xyz();
        rot2 = CVector(info.m_values.w, 1,0);
    }
}

void decodeRotation(NetRotationInfo& info, const CVector3& rot, const CVector3& rot2)
{

    info.m_values.x = rot.x;
    info.m_values.y = rot.y;
    info.m_values.z = rot.z;
    info.m_values.w = rot2.x;
    if (rot2.y > 0.5f) {
        info.m_rotationtype = nrtQuaternion;
    } else {
        info.m_rotationtype = nrtEuler;
    }
}

NetRotationInfo decodeRotation(const CVector3& rot, const CVector3& rot2)
{

    NetRotationInfo info;
    info.m_values.x = rot.x;
    info.m_values.y = rot.y;
    info.m_values.z = rot.z;
    info.m_values.w = rot2.x;
    if (rot2.y > 0.5f) {
        info.m_rotationtype = nrtQuaternion;
    } else {
        info.m_rotationtype = nrtEuler;
    }
    return info;
}

CTableItem::CTableItem(const CUtiString& _name, int _objid) : CNodeItem(_name, _objid)
{
    description = new CNodeDescription();
}

CTableItem::~CTableItem()
{
}

CNodeDescription* CTableItem::getNodeDescription()
{
    return description.get();
}

void CTableItem::create(CNode* node)
{
	CCacheList* l = new CCacheList();
	l->deleteItems = true;
   *((CCacheList**)&node->data[position]) = l;
   
}

void CTableItem::destroy(CNode* node)
{
   delete *((CCacheList**)&node->data[position]);
}

void CTableItem::xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
{
  CXNode* newnode = n->getNodeByName(name.str());
   CCacheList* l = *((CCacheList**)&node->data[position]);
   l->Clear();
   int c = newnode->getChildCount();
   for (int i = 0; i < c; i++) {
     CXNode* n = newnode->getChild(i);
     CNode* newn = addItem(node);
     description->xreadNode(newn, n, wri);
   };
   
	
}

void CTableItem::xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
{
  CXNode* newnode = n->addNode(name.str());
  CCacheList* l = *((CCacheList**)&node->data[position]);
  for (int i = 0; i < l->count; i++) {
    CXNode* n = newnode->addNode(UTISTR("tablechild"));
    CNode* savenode = getItem(node, i);
    description->xwriteNode(savenode, n, wri);
  };
}

int CTableItem::getSize()
{
   return sizeof(CCacheList*);
}
int CTableItem::getCount(CNode* node)
{
	utiassert(node->getDescription() == parent);
    CCacheList* l = *((CCacheList**)&node->data[position]);
	return l->count;
};
CNode* CTableItem::addItem(CNode* node)
{
	utiassert(node->getDescription() == parent);
    CCacheList* l = *((CCacheList**)&node->data[position]);
    CNode* n = new CNode(description.get(), node->getEntity());
	l->Add(n);
	return n;
};
CNode* CTableItem::getItem(CNode* node, int pos)
{
    utiassert(node->getDescription() == parent);
    CCacheList* l = *((CCacheList**)&node->data[position]);
	return (CNode*)(*l)[pos];
};
void CTableItem::deleteItem(CNode* node, int itempos)
{
	utiassert(node->getDescription() == parent);
    CCacheList* l = *((CCacheList**)&node->data[position]);
	l->Delete(itempos);
	
};
void CTableItem::removeItem(CNode* node, CNode* item)
{
	utiassert(node->getDescription() == parent);
    CCacheList* l = *((CCacheList**)&node->data[position]);
	l->Remove(item);
};


CMetaType::CMetaType()
{

}

void CMetaType::setName(const CUtiString& name)
{
    m_name = name;
}

CUtiString CMetaType::getName()
{
    return m_name;
}

TMetaBaseType CMetaBaseType::getBaseType()
{
    return m_type;
}


void CMetaBaseType::setBaseType(TMetaBaseType t)
{
    m_type = t;
}

CMetaBaseType::CMetaBaseType(TMetaBaseType t, const CUtiString& name)
{
    m_type = t;
    m_name = name;
}

CMetaBaseType::~CMetaBaseType()
{

}

CMetaEnumType::CMetaEnumType(int size)
{
   m_sizeinbytes = size;
}

CMetaEnumType::~CMetaEnumType()
{

}

void CMetaEnumType::addEnum(const CUtiString & d, int value)
{
    m_enums.push_back(d);
    m_enumvalues.push_back(value);
}

int CMetaEnumType::getEnumCount() const
{
    assert(m_enums.size() == m_enumvalues.size());
    return m_enums.size();
}
CUtiString CMetaEnumType::getEnum(int i)
{
    return m_enums.at(i);
}

int CMetaEnumType::getEnumValue(int i)
{
    return m_enumvalues.at(i);
}

CMetaClassRegistry::CMetaClassRegistry()
{

    m_int = new CMetaBaseType(mbtInteger, UTISTR("int"));
    m_float = new CMetaBaseType(mbtFloat, UTISTR("float"));
    m_string = new CMetaBaseType(mbtString, UTISTR("CUtiString"));
    m_double = new CMetaBaseType(mbtDouble, UTISTR("double"));
    m_bool = new CMetaBaseType(mbtBool, UTISTR("bool"));
    registerType(m_int);
    registerType(m_float);
    registerType(m_string);
    registerType(m_double);
    registerType(m_bool);
}

void CMetaClassRegistry::registerType(CMetaType* cls)
{
   assert(m_nametotype.find(cls->getName()) == m_nametotype.end());
   m_nametotype.insert(std::make_pair(cls->getName(), cls));
}

CMetaType* CMetaClassRegistry::t_type(const CUtiString& name)
{
    std::map<CUtiString, CMetaType*>::iterator it = m_nametotype.find(name);
    if (it == m_nametotype.end()) {
        return NULL;
    }
    return it->second;
}

CMetaBaseType* CMetaClassRegistry::t_int()
{
    return m_int;
}

CMetaBaseType* CMetaClassRegistry::t_float()
{
   return m_float;
}

CMetaBaseType* CMetaClassRegistry::t_string()
{
   return m_string;
}

CMetaBaseType* CMetaClassRegistry::t_double()
{
   return m_double;
}

CMetaBaseType* CMetaClassRegistry::t_bool()
{
    return m_bool;
}



CMetaBase::CMetaBase()
{
    m_name = EMPTYSTRING;
}

CUtiString CMetaBase::getName()
{
   return m_name;
}

CMetaProperty::CMetaProperty(CMetaType* type, const CUtiString & name)
{
    m_type = type;
    m_name = name;
}

CMetaType* CMetaProperty::getType()
{
   return m_type;
}

uti_refptr<ReferenceUtiOB> CMetaProperty::getData(void* object)
{
   uti_refptr<ReferenceUtiOB> data;
   return data;
}

void CMetaProperty::setData(void* object, ReferenceUtiOB* value)
{

}

bool CMetaProperty::getBool(void* object)
{
    uti_refptr<ReferenceUtiOB> data;
    data = getData(object);
    assert(data.get() != NULL);
    return check_cast<CMetaBaseValueReturn<bool>*>(data.get())->getValue();
}

void CMetaProperty::setBool(void* object, bool b)
{
    uti_refptr<CMetaBaseValueReturn<bool> > data(new CMetaBaseValueReturn<bool>());
    data->setValue(b);
    setData(object, data.get());


}

int CMetaProperty::getInteger(void* object)
{
    uti_refptr<ReferenceUtiOB> data;
    data = getData(object);
    assert(data.get() != NULL);
    return check_cast<CMetaBaseValueReturn<int>*>(data.get())->getValue();
}

void CMetaProperty::setInteger(void* object, int v)
{
    uti_refptr<CMetaBaseValueReturn<int> > data(new CMetaBaseValueReturn<int>());
    data->setValue(v);
    setData(object, data.get());


}
float CMetaProperty::getFloat(void* object)
{
    uti_refptr<ReferenceUtiOB> data;
    data = getData(object);
    assert(data.get() != NULL);
    return check_cast<CMetaBaseValueReturn<float>*>(data.get())->getValue();
}

void CMetaProperty::setFloat(void* object, float f)
{
    uti_refptr<CMetaBaseValueReturn<float> > data(new CMetaBaseValueReturn<float>());
    data->setValue(f);
    setData(object, data.get());


}
double CMetaProperty::getDouble(void* object)
{
    uti_refptr<ReferenceUtiOB> data;
    data = getData(object);
    assert(data.get() != NULL);
    return check_cast<CMetaBaseValueReturn<double>*>(data.get())->getValue();
}

void CMetaProperty::setDouble(void* object, double f)
{
    uti_refptr<CMetaBaseValueReturn<double> > data(new CMetaBaseValueReturn<double>());
    data->setValue(f);
    setData(object, data.get());


}
CUtiString CMetaProperty::getString(void* object)
{
    uti_refptr<ReferenceUtiOB> data;
    data = getData(object);
    assert(data.get() != NULL);
    return check_cast<CMetaBaseValueReturn<CUtiString>*>(data.get())->getValue();
}

void CMetaProperty::setString(void* object, const CUtiString& str)
{
        uti_refptr<CMetaBaseValueReturn<CUtiString> > data(new CMetaBaseValueReturn<CUtiString>());
        data->setValue(str);
        setData(object, data.get());
}

CMetaPointerProperty::CMetaPointerProperty(CMetaType* type, const CUtiString&name, void* instance, void* property) : CMetaProperty(type, name)
{
   unsigned char* ch = (unsigned char*)instance;
   unsigned char* prop = (unsigned char*)property;
   m_baseshift = prop-ch;
}

/*ReferenceUtiOB* CMetaPointerProperty::getData(void* object)
{

}

void CMetaPointerProperty::setData(void* object, ReferenceUtiOB* value)
{

}*/


CMetaMethod::CMetaMethod()
{

}
