#pragma once
#ifndef __UTI3DNODE__
#define __UTI3DNODE__
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "UtiTypeHandlers.h"

#include <assert.h>
#include <map>
#include <set>



typedef int NodeProperty;

class CNodeDescription;
class CNode;


class CNode : public ReferenceUtiOB
{
public:
    Array<unsigned char> data;
    int m_lastpropertycount;
    int datasize;
    CUtiOB* m_parent;

    uti_refptr<CNodeDescription> description;
public:
    CNode(CNodeDescription* _description, CUtiOB* parent);
	~CNode();
    CUtiOB* getEntity() {return m_parent;}
	virtual CNodeDescription* getDescription();
    virtual void updateStore();
	virtual void xread(CXNode* n, CWriteReadInterface* wri);
    virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
};

class CNodeItem : public ReferenceUtiOB
{
protected:
    CUtiIDString name;
	int objid;
	int position;
    CNodeDescription* parent;
public:
	CNodeItem(const CUtiString& _name, int _objid);
	virtual ~CNodeItem();
    virtual CUtiIDString getName();
    virtual void setName(const CUtiIDString& str);
	virtual void create(CNode* node);
	virtual void destroy(CNode* node);
    virtual int getSize();
	virtual void setPosition(int pos);
	virtual void setParent(CNodeDescription* _parent);
	virtual CNodeDescription* getParent();
    virtual void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri);
        virtual void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri);
};

template <class T> class CDefaultItem : public CNodeItem
{
protected:
	T def;
public:
	CDefaultItem(const CUtiString& _name, int _objid, const T& _def): CNodeItem(_name, _objid)
	{
       def = _def;
	};

	void create(CNode* node)
	{
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		T*p = new(&node->data[position]) T();
		*p = def;
	};
	void destroy(CNode* node)
	{

	};
    int getSize()
	{
        return sizeof(T);
	};
	void set(CNode* node, const T& value) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
       *((T*)&node->data[position]) = value;
	};

	T get(CNode* node) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
       return *((T*)&node->data[position]);
	};
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xreadNode(node, n, wri);
	  T value = def;
      uti_readValue(value, n, wri,name.str(), def);
	  set(node, value);

	}
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xwriteNode(node, n, wri);
	  T value = get(node);
      uti_writeValue(value,  n, wri, name.str());
	}
};

template <> class CDefaultItem<void*> : public CNodeItem
{
protected:
	void* def;
public:
	CDefaultItem(const CUtiString& _name, int _objid, void*const & _def): CNodeItem(_name, _objid)
	{
       def = NULL;
	};

	void create(CNode* node)
	{
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		void**p = ((void**)&node->data[position]);
		*p = def;
	};
	void destroy(CNode* node)
	{

	};
    int getSize()
	{
        return sizeof(void*);
	};
	void set(CNode* node, void* const & value) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
       *((void**)&node->data[position]) = value;
	};

	void* get(CNode* node) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
       return *((void**)&node->data[position]);
	};
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xreadNode(node, n, wri);
	//  T value = def;
	  //uti_readValue(value, n, wri,name, def);
	  set(node, NULL);

	}
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xwriteNode(node, n, wri);
	  //T value = get(node);
	 // uti_writeValue(value,  n, wri, name);
	}
};



template <class T> class CReferenceItem : public CNodeItem
{
public:
	CReferenceItem(const CUtiString& _name, int _objid): CNodeItem(_name, _objid)
	{
    
	};

	void create(CNode* node)
	{
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		T**p = ((T**)&node->data[position]);
		*p = new T();
		(*p)->addRef();
	};
	void destroy(CNode* node)
	{
              T**p = ((T**)&node->data[position]);
	      if (*p) {
		(*p)->decRef();
	      };
	};
    int getSize()
	{
        return sizeof(T*);
	};
	void set(CNode* node, T* const & value) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
		T**p = ((T**)&node->data[position]);
		if (*p == value)return;
		if (*p) {
		  (*p)->decRef();
		};
		*p = value;
		if (*p) {
		  (*p)->addRef();
		};
		
	};

	T* get(CNode* node) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
       return *((T**)&node->data[position]);
	};
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xreadNode(node, n, wri);
	//  T value = def;
	  //uti_readValue(value, n, wri,name, def);
	  T**p = ((T**)&node->data[position]);
	  utiassert(*p != NULL);
	  (*p)->xread(n, wri);

	}
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xwriteNode(node, n, wri);
	  T**p = ((T**)&node->data[position]);
	  utiassert(*p != NULL);
	  (*p)->xwrite(n, wri);
	}
};


template <class T> class CReferenceItemSubSystem : public CNodeItem
{
public:
    CReferenceItemSubSystem(const CUtiString& _name, int _objid): CNodeItem(_name, _objid)
    {

    };

    void create(CNode* node)
    {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
        T**p = ((T**)&node->data[position]);
        *p = new T();
        (*p)->addRef();
    };
    void destroy(CNode* node)
    {
              T**p = ((T**)&node->data[position]);
          if (*p) {
        (*p)->decRef();
          };
    };
    int getSize()
    {
        return sizeof(T*);
    };
    void set(CNode* node, T* const & value) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
        utiassert(node->getDescription() == parent);
        T**p = ((T**)&node->data[position]);
        if (*p == value)return;
        if (*p) {
          (*p)->decRef();
        };
        *p = value;
        if (*p) {
          (*p)->addRef();
        };

    };

    T* get(CNode* node) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
        utiassert(node->getDescription() == parent);
       return *((T**)&node->data[position]);
    };
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
    {
      CNodeItem::xreadNode(node, n, wri);
    //  T value = def;
      //uti_readValue(value, n, wri,name, def);
      /*T**p = ((T**)&node->data[position]);
      utiassert(*p != NULL);
      (*p)->xread(n, wri);*/
      CUtiOB* result=NULL;
      XLoadFromNode(n, NULL, result);
      if (result != NULL) {
      set(node, check_cast<T*>(result));
      }

    }
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
    {
      CNodeItem::xwriteNode(node, n, wri);
      T**p = ((T**)&node->data[position]);
      utiassert(*p != NULL);
      //(*p)->xwrite(n, wri);
      XSaveToNode(n, (*p));
    }
};

template <class T> class CReferenceItemNoSave : public CNodeItem
{
public:
	CReferenceItemNoSave(const CUtiString& _name, int _objid): CNodeItem(_name, _objid)
	{
    
	};

	void create(CNode* node)
	{
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		T**p = ((T**)&node->data[position]);
		*p = NULL;
		//(*p)->addRef();
	};
	void destroy(CNode* node)
	{
              T**p = ((T**)&node->data[position]);
	      if (*p) {
		(*p)->decRef();
	      };
	};
    int getSize()
	{
        return sizeof(T*);
	};
	void set(CNode* node, T* const & value) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
		T**p = ((T**)&node->data[position]);
		if (*p == value)return;
		if (*p) {
		  (*p)->decRef();
		};
		*p = value;
		if (*p) {
		  (*p)->addRef();
		};
		
	};

	T* get(CNode* node) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
       return *((T**)&node->data[position]);
	};
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xreadNode(node, n, wri);
	//  T value = def;
	  //uti_readValue(value, n, wri,name, def);
	  set(node,NULL);
	 /* T**p = ((T**)&node->data[position]);
	  utiassert(*p != NULL);
	  (*p)->xread(n, wri);*/

	}
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xwriteNode(node, n, wri);
	  /*T**p = ((T**)&node->data[position]);
	  utiassert(*p != NULL);
	  (*p)->xwrite(n, wri);*/
	}
};

template <class T> class CReferenceItemNull : public CNodeItem
{
public:
	CReferenceItemNull(const CUtiString& _name, int _objid): CNodeItem(_name, _objid)
	{
    
	};

	void create(CNode* node)
	{
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		T**p = ((T**)&node->data[position]);
		*p = NULL;
		//(*p)->addRef();
	};
	void destroy(CNode* node)
	{
              T**p = ((T**)&node->data[position]);
	      if (*p) {
		(*p)->decRef();
	      };
	};
    int getSize()
	{
        return sizeof(T*);
	};
	void set(CNode* node, T* const & value) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
		T**p = ((T**)&node->data[position]);
		if (*p == value)return;
		if (*p) {
		  (*p)->decRef();
		};
		*p = value;
		if (*p) {
		  (*p)->addRef();
		};
		
	};

	T* get(CNode* node) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
       return *((T**)&node->data[position]);
	};
	
	/*void xread(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
	  CNodeItem::xread(node, n, wri);
	//  T value = def;
	  //uti_readValue(value, n, wri,name, def);
	  T**p = ((T**)&node->data[position]);
	  utiassert(*p != NULL);
	  (*p)->xread(n, wri);

	}
        void xwrite(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
	  CNodeItem::xwrite(node, n, wri);
	  T**p = ((T**)&node->data[position]);
	  utiassert(*p != NULL);
	  (*p)->xwrite(n, wri);
	}*/
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xreadNode(node, n, wri);
	//  T value = def;
	  //uti_readValue(value, n, wri,name, def);
//	  T**p = ((T**)&node->data[position]);
    /*  utiassert(*p != NULL);
      (*p)->xread(n, wri);*/

	}
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xwriteNode(node, n, wri);
    //  T**p = ((T**)&node->data[position]);
    /*  utiassert(*p != NULL);
      (*p)->xwrite(n, wri);*/
	}
};

template <class T> class CReferenceItemNullSave : public CNodeItem
{
public:
	CReferenceItemNullSave(const CUtiString& _name, int _objid): CNodeItem(_name, _objid)
	{
    
	};

	void create(CNode* node)
	{
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		T**p = ((T**)&node->data[position]);
		*p = NULL;
		//(*p)->addRef();
	};
	void destroy(CNode* node)
	{
              T**p = ((T**)&node->data[position]);
	      if (*p) {
		(*p)->decRef();
	      };
	};
    int getSize()
	{
        return sizeof(T*);
	};
	void set(CNode* node, T* const & value) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
		T**p = ((T**)&node->data[position]);
		if (*p == value)return;
		if (*p) {
		  (*p)->decRef();
		};
		*p = value;
		if (*p) {
		  (*p)->addRef();
		};
		
	};

	T* get(CNode* node) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
		utiassert(node->getDescription() == parent);
       return *((T**)&node->data[position]);
	};
	
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xreadNode(node, n, wri);
	//  T value = def;
	  //uti_readValue(value, n, wri,name, def);
	  T**p = ((T**)&node->data[position]);
	  utiassert(*p != NULL);
	  (*p)->xread(n, wri);

	}
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
	{
      CNodeItem::xwriteNode(node, n, wri);
	  T**p = ((T**)&node->data[position]);
	  utiassert(*p != NULL);
	  (*p)->xwrite(n, wri);
	}
	
};


template <class T> class CReferenceADRItemNullSave : public CNodeItem
{
public:
    CReferenceADRItemNullSave(const CUtiString& _name, int _objid): CNodeItem(_name, _objid)
    {

    };

    void create(CNode* node)
    {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
        T**p = ((T**)&node->data[position]);
        *p = NULL;
        //(*p)->addRef();
    };
    void destroy(CNode* node)
    {
              T**p = ((T**)&node->data[position]);
          if (*p) {
        (*p)->cpp_decRef();
          };
    };
    int getSize()
    {
        return sizeof(T*);
    };
    void set(CNode* node, T* const & value) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
        utiassert(node->getDescription() == parent);
        T**p = ((T**)&node->data[position]);
        if (*p == value)return;
        if (*p) {
          (*p)->cpp_decRef();
        };
        *p = value;
        if (*p) {
          (*p)->cpp_addRef();
        };

    };

    T* get(CNode* node) {
        utiassert(position >= 0 && position + getSize() <= node->datasize);
        utiassert(node->getDescription() == parent);
       return *((T**)&node->data[position]);
    };

    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
    {
      CNodeItem::xreadNode(node, n, wri);
    //  T value = def;
      //uti_readValue(value, n, wri,name, def);
      T**p = ((T**)&node->data[position]);
      utiassert(*p != NULL);
      (*p)->xread(n, wri);

    }
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
    {
      CNodeItem::xwriteNode(node, n, wri);
      T**p = ((T**)&node->data[position]);
      utiassert(*p != NULL);
      (*p)->xwrite(n, wri);
    }

};

typedef CDefaultItem<int> CIntItem;
typedef CDefaultItem<UTI_UINT64> CUInt64Item;
typedef CDefaultItem<float> CFloatItem;
typedef CDefaultItem<CUtiString> CStringItem;
typedef CDefaultItem<CVector> CVectorItem;
typedef CDefaultItem<TIME> CTimeItem;
typedef CDefaultItem<bool> CBoolItem;

void encodeRotation(const NetRotationInfo& info, CNode*node, CVectorItem* rot, CVectorItem* rot2);
void decodeRotation(NetRotationInfo& info, CNode* node, CVectorItem* rot, CVectorItem* rot2);

void encodeRotation(const NetRotationInfo& info,  CVector3& rot, CVector3& rot2);
void decodeRotation(NetRotationInfo& info, const CVector3& rot, const CVector3& rot2);
NetRotationInfo decodeRotation(const CVector3& rot, const CVector3& rot2);

class CTableItem : public CNodeItem
{
protected:
    uti_refptr<CNodeDescription> description;
public:
	CTableItem(const CUtiString& _name, int _objid);
	~CTableItem();
	CNodeDescription* getNodeDescription();
	void create(CNode* node);
	void destroy(CNode* node);
    int getSize();
	int getCount(CNode* node);
    CNode* addItem(CNode* node);
	CNode* getItem(CNode* node, int ipos);
	void deleteItem(CNode* node, int itempos);
	void removeItem(CNode* node, CNode* item);
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri);
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri);
};


class CNodeDescription : public ReferenceUtiOB
{
protected:
    CReferenceList<CNodeItem> Items;
	int datasize;
	int instances;
        std::set<CUtiIDString> names;
public:
	CNodeDescription();
	~CNodeDescription();
	CNodeItem* addItem(CNodeItem* item);
    inline CNodeItem* getNodeItem(int n) {return Items.get(n);}
    inline int getNodeCount() {return Items.size();}
	void create(CNode* node);
    void createRange(CNode* node, int start);
	void destroy(CNode* node);
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri);
        void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri);
	int getDataSize();
};

template<class T> class CDictonaryItem : public CNodeItem
{
protected:
    typedef T MyValue;
    typedef std::map<MyValue, CNode*> DictonaryMap;
	typedef typename DictonaryMap::iterator myIterator;
    uti_refptr<CNodeDescription> description;
public:
    CDictonaryItem(const CUtiString& _name, int _objid) : CNodeItem(_name, _objid)
    {
       description = new CNodeDescription();
    }

    ~CDictonaryItem()
    {
      delete description;
    }

    CNodeDescription* getNodeDescription()
    {
      return description.get();
    }
	void create(CNode* node)
    {
	    DictonaryMap* l = new DictonaryMap();
        *((DictonaryMap**)&node->data[position]) = l;
    }

    void destroy(CNode* node)
    {
	   DictonaryMap & m = **((DictonaryMap**)&node->data[position]);
	 //  DictonaryMap::iterator it = n.begin();
	   for (myIterator it = m.begin(); it != m.end(); it++) {
          delete it->second;
	   };
       delete &m;
    }
    int getSize()
    {
       return sizeof(DictonaryMap*);
    }
    int getCount(CNode* node)
    {
        DictonaryMap * m = *((DictonaryMap**)&node->data[position]);
	    return m->size();
    };
    CNode* addItem(CNode* node, const T& item)
    {
		utiassert(node->getDescription() == parent);
       DictonaryMap * m = *((DictonaryMap**)&node->data[position]);

       CNode* n = new CNode(description.get(), node->getEntity());
	   m->insert(make_pair(item, n));
	   return n;
    };
    CNode* getItem(CNode* node, const T& item)
    {
		utiassert(node->getDescription() == parent);
       DictonaryMap * m = *((DictonaryMap**)&node->data[position]);

	   myIterator it = m->find(item);

	   if (it == m->end()) return NULL;
	   return it->second;
    };
    void deleteItem(CNode* node, const T& item)
    {
		utiassert(node->getDescription() == parent);
        DictonaryMap * m = *((DictonaryMap**)&node->data[position]);
	    utiassert(m->find(item) != m->end());
	    m->erase(item);
    };
    void xreadNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
    {
      CXNode* newnode = n->getNodeByName(name.str());
      DictonaryMap * m = *((DictonaryMap**)&node->data[position]);
      m->clear();
      int c = newnode->getChildCount();
      for (int i = 0; i < c; i++) {
        CXNode* mainn = newnode->getChild(i);
	T value = T();
	uti_readValue(value, n, wri,UTISTR("value"));
	CXNode* datan = mainn->getNodeByName(UTISTR("data"));
        CNode* newn = addItem(node, value);
        description->xreadNode(newn, datan, wri);
      };
   
    }
    void xwriteNode(CNode* node, CXNode* n, CWriteReadInterface* wri)
    {
      CXNode* newnode = n->addNode(name.str());
       DictonaryMap * m = *((DictonaryMap**)&node->data[position]);
       for (typename DictonaryMap::iterator it = m->begin(); it != m->end(); it++) {
          CXNode* n = newnode->addNode(UTISTR("dictonarychild"));
       T value = it->first;
       uti_writeValue(value, n, wri,UTISTR("value"));
       CXNode* datan = n->addNode(UTISTR("data"));
          CNode* savenode = it->second;
          description->xwriteNode(savenode, datan, wri);
       };
    }
};

class CMetaType : public ReferenceUtiOB
{
protected:
    CUtiString m_name;
public:
    CMetaType();
    void setName(const CUtiString& name);
    CUtiString getName();
};

enum TMetaBaseType {mbtInteger, mbtBool, mbtFloat, mbtDouble, mbtString/*, mbtVector2, mbtVector3, mbtVector4, mbtIVector2, mbtIVector3, mbtIVector4, mbtRect, mbtIntRect, mbtBox, mbtIntBox*/};

class CMetaBaseType : public CMetaType
{
protected:
    TMetaBaseType m_type;
public:
    TMetaBaseType getBaseType();
    void setBaseType(TMetaBaseType t);
    CMetaBaseType(TMetaBaseType t=mbtInteger, const CUtiString& name=EMPTYSTRING);
    ~CMetaBaseType();
};

class CMetaEnumType : public CMetaType
{
protected:
    int m_sizeinbytes;
    std::vector<CUtiString> m_enums;

    std::vector<int> m_enumvalues;

public:
    CMetaEnumType(int size);
    ~CMetaEnumType();
    void addEnum(const CUtiString & d, int value);
    int getEnumCount() const;
    CUtiString getEnum(int i);
    int getEnumValue(int i);
};

class CMetaProperty;
class CMetaMethod;

class CMetaClassCreationInfo : public ReferenceUtiOB
{
public:
    CMetaClassCreationInfo() {}
};

class CParentMetaClassCreationInfo : public CMetaClassCreationInfo
{
public:
    CParentMetaClassCreationInfo() {}
};

class CMetaClassCreator : public ReferenceUtiOB
{
public:
    virtual void* createInstance(CMetaClassCreationInfo* parent)=0;
    virtual void destroyInstance(void* instance)=0;
};

class CMetaClass: public CMetaType
{
protected:
    CReferenceList<CMetaProperty> m_properties;
    CReferenceList<CMetaMethod> m_methods;
    uti_refptr<CMetaClassCreator> m_classcreator;
    CMetaClass* m_parentclass;
public:
    CMetaClass(const CUtiString& name, CMetaClassCreator* creator, CMetaClass* parent);
    void* createInstance(CMetaClassCreationInfo* parent);
    void destroyInstance(void* inst);
    void setParentClass(CMetaClass* parent);
    CMetaClass* getParentClass();
    int getPropertyCount() const;
    CMetaProperty* getProperty(int i);
    int getMethodCount() const;
    CMetaMethod* getMethod(int i);
    void addInternProperty(CMetaProperty* p);
    void addInternMethod(CMetaMethod* p);

};

class CMetaDerivered : public CMetaType
{
protected:
    CMetaType* m_type;
public:
    CMetaDerivered(CMetaType* t);
    CMetaType* getType();
};

class CMetaBasePointerType : public CMetaDerivered
{
public:
    CMetaBasePointerType(CMetaType* t) : CMetaDerivered(t){};
    virtual void* getData(void* data) {return *((void**)data);}
    virtual void setData(void* data, void* d) {*((void**)data) = d;}
};



class CMetaPointerType : public CMetaBasePointerType

{
protected:

public:
    CMetaPointerType(CMetaType* t2) : CMetaBasePointerType(t2){};
};

class CMetaStaticType : public CMetaBasePointerType

{
protected:

public:
    CMetaStaticType(CMetaType* t2) : CMetaBasePointerType(t2){};
    virtual void* getData(void* data) {return data;};
    virtual void setData(void* data, void* d) {assert(false);};
};

class CMetaReferenceType : public CMetaBasePointerType
{
protected:

public:
    CMetaReferenceType(CMetaType* t2) : CMetaBasePointerType(t2){};
    virtual void* getData(void* data) {return (*((uti_refptr<ReferenceUtiOB>*)data)).get();}
    virtual void setData(void* data, void* d) {(*((uti_refptr<ReferenceUtiOB>*)data)) = ((ReferenceUtiOB*)d);}
};

template<class T> class CMetaReferenceTypeTemplate : public CMetaReferenceType
{
public:
    CMetaReferenceTypeTemplate(CMetaType* t2) : CMetaReferenceType(t2){};
    virtual void* getData(void* data) {return (*((uti_refptr<T>*)data)).get();}
    virtual void setData(void* data, void* d) {(*((uti_refptr<T>*)data)) = ((T*)d);}
};


class CMetaList : public CMetaDerivered
{
protected:

public:
    CMetaList(CMetaType* t2) : CMetaDerivered(t2)
    {

    }

    virtual int getListCount(void* data)=0;
    virtual void* getListData(void* data, int index)=0;
    virtual void setListData(void* data, int index, void* d)=0;
    virtual void addList(void* data, void* d)=0;
    virtual void clearList(void* data)=0;
    virtual void moveList(void* data, int t1, int t2)=0;
    virtual void deleteItem(void* data, int index)=0;
};

template<class T> class CMetaReferenceList : public CMetaList
{
public:
     CMetaReferenceList(CMetaType* t2)
     {

     }
     int getListCount(void* data)
     {
         CReferenceList<T>* d = (CReferenceList<T>*)data;
         return d->size();

     }

     void* getListData(void* data, int index)
     {
         CReferenceList<T>* d = (CReferenceList<T>*)data;
         return &((*d)[index]);
     }

     void setListData(void* data, int index, void* d)
     {
         CReferenceList<T>* d2 = (CReferenceList<T>*)data;
         CUtiOB* obj = (CUtiOB*)d;

         ((*d2)[index]) = check_cast<T*>(obj);
     }

     void addList(void* data, void* d)
     {
         CReferenceList<T>* d2 = (CReferenceList<T>*)data;
         CUtiOB* obj = (CUtiOB*)d;

         d2->push_back(check_cast<T*>(obj));
     }

     void clearList(void* data)
     {
         CReferenceList<T>* d2 = (CReferenceList<T>*)data;

         d2->clear();
     }
     void moveList(void* data, int t1, int t2)
     {
         CReferenceList<T>* d2 = (CReferenceList<T>*)data;
         d2->move(t1, t2);

     }

     void deleteItem(void* data, int index)
     {
         CReferenceList<T>* d2 = (CReferenceList<T>*)data;
         d2->deleteItem(index);
     }

};

class CMetaIValueType : public CMetaBasePointerType
{
public:
    CMetaIValueType(CMetaType* t);
};


class CMetaIValueSourceType : public CMetaBasePointerType
{

public:
    CMetaIValueSourceType(CMetaType* t);
};

class CMetaClassRegistry : public CUtiOB
{
protected:
     MAKE_SINGLETON(CMetaClassRegistry);
     CMetaBaseType* m_int;
     CMetaBaseType* m_float;
     CMetaBaseType* m_string;
     CMetaBaseType* m_double;
     CMetaBaseType* m_bool;
     std::map<CUtiString, CMetaType*> m_nametotype;
public:
    CMetaClassRegistry();
    void registerType(CMetaType* cls);
    CMetaType* t_type(const CUtiString& name);
    CMetaBaseType* t_int();
    CMetaBaseType* t_float();
    CMetaBaseType* t_string();
    CMetaBaseType* t_double();
    CMetaBaseType* t_bool();
};

class CMetaBase : public ReferenceUtiOB
{
protected:
    CUtiString m_name;
public:
    CMetaBase();
    CUtiString getName();
};

template<class T> class CMetaBaseValueReturn : public ReferenceUtiOB
{
public:
    T m_value;
    CMetaBaseValueReturn(){}
    ~CMetaBaseValueReturn(){}
    inline void setValue(const T& v) {m_value = v;};
    inline T getValue() const {return m_value;};

};

class CMetaProperty : public CMetaBase
{
protected:
    CMetaType* m_type;
public:
    CMetaProperty(CMetaType* type, const CUtiString & name);
    virtual CMetaType* getType();
    virtual uti_refptr<ReferenceUtiOB> getData(void* object);
    virtual void setData(void* object, ReferenceUtiOB* value);
    virtual bool getBool(void* object);
    virtual void setBool(void* object, bool b);
    virtual int getInteger(void* object);
    virtual void setInteger(void* object, int v);
    virtual float getFloat(void* object);
    virtual void setFloat(void* object, float f);
    virtual double getDouble(void* object);
    virtual void setDouble(void* object, double f);
    virtual CUtiString getString(void* object);
    virtual void setString(void* object, const CUtiString& str);
    /*virtual CVector2 getVector2(CUtiOB* object);
    virtual void setVector2(CUtiOB* object, const CVector2& v);
    virtual CVector3 getVector3(CUtiOB* object);
    virtual void setVector3(CUtiOB* object, const CVector3& v);
    virtual CVector4 getVector4(CUtiOB* object);
    virtual void setVector4(CUtiOB* object, const CVector4& v);
    virtual CIVector2 getIVector2(CUtiOB* object);
    virtual void setIVector2(CUtiOB* object, const CIVector2& v);
    virtual CIVector3 getIVector3(CUtiOB* object);
    virtual void setIVector3(CUtiOB* object, const CIVector3& v);
    virtual CIVector4 getIVector4(CUtiOB* object);
    virtual void setIVector4(CUtiOB* object, const CIVector4& v);*/
};

class CMetaPointerProperty : public CMetaProperty
{
protected:
    int m_baseshift;
public:
    CMetaPointerProperty(CMetaType* type,  const CUtiString&name, void* instance, void* property);
    //virtual ReferenceUtiOB* getData(void* object);
    //virtual void setData(void* object, ReferenceUtiOB* value);
};

class CMetaMethod : public CMetaBase
{
public:
    CMetaMethod();

};

#endif
