/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Uti3DThreads.h"
#include "utisystem.h"
#include "utistream.h"

#include "ziparchiv.h"
#include "easyprofiler.h"


/*CReferenceObjectAtomic::~CReferenceObjectAtomic()
{
   utiassert(m_references <= 1);
}
void CReferenceObjectAtomic::destroyReferenceObject()
{
   delete this;
}*/

CBaseJob::CBaseJob()
{
  m_finished = false;
  m_listcount = 0;
}

bool CBaseJob::satisfydepends()
{
  return true;
}

CBaseJob::~CBaseJob()
{
    assert(m_listcount == 0);
}

void CBaseJob::addList()
{
    m_listcount++;
    assert(m_listcount == 1);//sonst gehört es mehreren threads
}

void CBaseJob::removeList()
{
    m_listcount--;
    assert(m_listcount == 0); //sonst gehört es mehreren threads
}


void CBaseJob::executeJob(int startjob, int endjob)
{

}

void CBaseJob::finaliseMainThread()
{
  
}

void CBaseJob::waitForFinished()
{
  while (!isFinished()) {
   UtiFramework::utisleep(1);
  };
}

int CBaseJob::getID()
{
    return 0;
}

int CBaseJob::getJobCount()
{
   return 1;
}

CInvalidObject* CBaseJob::getInvalidObject()
{
    return NULL; 
}

WORKTIME CBaseJob::getJobWorkTime()
{CHECKEXISTANCE;
   return 1;
}

bool CBaseJob::isFinished()
{
  CHECKEXISTANCE;
   return m_finished && m_listcount == 0;
}

void CBaseJob::setFinished()
{CHECKEXISTANCE;
   m_finished = true;
}

bool CBaseJob::deleteJob()
{CHECKEXISTANCE;
  return false;
}

CBaseThreadObject::CBaseThreadObject()
{

}

CBaseThreadObject::~CBaseThreadObject()
{

}

TIME CBaseThreadObject::executeJob(int amount)
{
return 0.0;
}

WORKTIME CBaseThreadObject::getJobWorkTime()
{
  return 1;
}

bool CBaseThreadObject::isFinished()
{
   return false;
}



CBaseJobQueue::CBaseJobQueue()
{

}

CBaseJobQueue::~CBaseJobQueue()
{

}

void CBaseJobQueue::addJob(CBaseJob* job)
{

}

CSimpleReferenceJobQueue::CSimpleReferenceJobQueue()
{

}

CSimpleReferenceJobQueue::~CSimpleReferenceJobQueue()
{

}

void CSimpleReferenceJobQueue::addJob(CBaseJob* job)
{
   m_jobqueue.push_back(job);
}

void CSimpleReferenceJobQueue::executeJobs()
{
   while (m_jobqueue.size() != 0) {
      CBaseJob* job = &m_jobqueue.front();
      job->executeJob(0, job->getJobCount());
      job->setFinished();
      m_jobqueue.pop_front();
   };
}


void CSimpleReferenceJobQueue::executeJobs(int totalcount)
{

    EASY_FUNCTION(profiler::colors::Grey);
    CHECKEXISTANCE;
    int jobcount = 0;
    while (m_jobqueue.size() != 0 && jobcount < totalcount) {
       CBaseJob* job = &m_jobqueue.front();
       jobcount += job->getJobCount();
       job->executeJob(0, job->getJobCount());

       job->setFinished();
       m_jobqueue.pop_front();
    };
}



void CSimpleReferenceJobQueue::executeJobsFinished(int totalcount)
{

    EASY_FUNCTION(profiler::colors::Grey);
    CHECKEXISTANCE;
    int jobcount = 0;
    while (m_jobqueue.size() != 0 && jobcount < totalcount) {
       CBaseJob* job = &m_jobqueue.front();
       jobcount += job->getJobCount();
       job->executeJob(0, job->getJobCount());

       if (!job->isFinished())return;
       m_jobqueue.pop_front();
    };
}



CQueueThread::CQueueThread(CThreadNormalJobQueue* queue): CThread()
{
  m_queue = queue;
}

void* CQueueThread::Execute()
{
    return m_queue->Execute();
}

CThreadNormalJobQueue::CThreadNormalJobQueue(int count)
{

    m_currentactivejobs = 0;
  m_threads.setSize(count);
  for (int i = 0; i < m_threads.size(); i++) {
    m_threads[i] = new CQueueThread(this);
    m_threads[i]->Run();
  };
 // m_thread.Run();
}

CThreadNormalJobQueue::~CThreadNormalJobQueue()
{
  //Warnung es wird nur der erste Thread auf Terminierungsgrund geprüft
  for (int i = 0; i < m_threads.size(); i++) {
    m_threads[i]->Terminate();
  };
  m_signal.Broadcast();
   for (int i = 0; i < m_threads.size(); i++) {
    m_threads[i]->WaitFor();
  };
  for (CNormalLinkedList<CBaseJob, 0>::iterator it = m_jobqueue.begin(); it != m_jobqueue.end(); it++) {
      it->removeList();
  }
}

void CThreadNormalJobQueue::signal()
{
   CCSControl control(cs_accessjob);
   if (m_jobqueue.size() != 0)
   m_signal.Signal();
}


int CThreadNormalJobQueue::getJobCount()
{
    CCSControl control(cs_accessjob);
    return m_jobqueue.size()+m_currentactivejobs;

}

void CThreadNormalJobQueue::addJob(CBaseJob* job)
{
  CCSControl control(cs_accessjob);
  
  m_jobqueue.push_back(job);
  job->addList();
 // wcout <<"addJob(" << m_jobqueue.size() << ")" << endl;
  m_signal.Signal();
}

void CThreadNormalJobQueue::removeDeleteMeObjects()
{
  CCSControl control(cs_accessjob);
  CNormalLinkedList<CBaseJob, 0>::iterator it, next;
  for (it = m_jobqueue.begin(); it != m_jobqueue.end(); ) {
    next = it;
    next++;
    if (it->deleteJob()) {
        it->removeList();
      it->setFinished();
      m_jobqueue.remove(&(*it));
    };
    it = next;
  };
}

void* CThreadNormalJobQueue::Execute()
{
#ifdef UTI_ACCESS_SINGLE_OBJECT
   CNormalLinkedList<CBaseJob, 0> m_newjobs;
  // bool showstring = true;
   while (!m_thread.ToTerminate) {
     m_newjobs.clear();
     {
        CCSControl control(cs_accessjob);
	bool blocked= false;
	while (m_jobqueue.size() != 0) {
          CBaseJob* job = &m_jobqueue.front();
	 /* if (showstring && m_jobqueue.size() > 1){
	    wcout <<"Execute(" << m_jobqueue.size() << ")" << endl;
	    
	  }*/
	  if (job->satisfydepends() == false) {
	    if (m_newjobs.size() != 0) {
	      break;
	    }; 
	   // UtiFramework::utisleep(0);
	   // showstring = false;
	    blocked = true;
	   break;
	  };
	  //showstring = true;
	  m_jobqueue.pop_front();
	  m_newjobs.push_back(job);
	  //Im Moment nur ein einziger Job
	  break;
	 // wcout <<" Took Job" << endl;
          
       };
       if (m_newjobs.size() == 0 && blocked == false) {
	 m_signal.WaitFor(&cs_accessjob);
	
       };
     }
     if (m_newjobs.size() == 0) {
       UtiFramework::utisleep(0);

     } else {
       while (m_newjobs.size() != 0) {
          CBaseJob* job = &m_newjobs.front();
          job->executeJob(0, job->getJobCount());
          job->setFinished();
          m_newjobs.pop_front();
       };
     };
   };
   return NULL;
#else
  //CBaseJob* newjob = NULL;
  // bool showstring = true;
   while (!m_threads[0]->ToTerminate) {
     CBaseJob* newjob = NULL;
     {
        CCSControl control(cs_accessjob);
    //bool blocked= true;
	CNormalLinkedList<CBaseJob, 0>::iterator it;
        if (m_jobqueue.size() == 0) {
      //blocked = false;
	};
        for (it = m_jobqueue.begin(); it != m_jobqueue.end(); it++) {
	  CBaseJob* job = &(*it);
	  if (job->satisfydepends() == true) {
	    utiassert(newjob == NULL);
	     newjob = job;
	     m_jobqueue.remove(job);
         m_currentactivejobs++;
	     break;
	  };
	};
	
        if (newjob == NULL/* && blocked == false*/) {
	 m_signal.WaitFor(&cs_accessjob);
	
        };
     }
     if (newjob == NULL) {
//       utiassert(false);
       UtiFramework::utisleep(0);

     } else {CHECKEXISTANCEOF(newjob);
          newjob->executeJob(0, newjob->getJobCount());CHECKEXISTANCEOF(newjob);
          newjob->setFinished();
          newjob->removeList();
          {
             CCSControl control(cs_accessjob);
m_currentactivejobs--;
          }
	  m_signal.Broadcast();

     };
   };
   return NULL;
#endif
}



CReadStreamJob::CReadStreamJob(const CUtiString& f)
{
  m_filename = f;
  m_memorystream = new CMemoryStream();
}

void CReadStreamJob::executeJob(int startjob, int endjob)
{
    uti_refptr<CZipMemoryStream> zipstream;
    uti_refptr<CFileStream> f;

    bool b = FileExists(m_filename.getString());

      if (!b)
      {

          m_memorystream = NULL;
          setFinished();
        return;

      }

       //UTI_INT64 Size = FileSize(m_filename.c_str());
        f = new CFileStream(m_filename, S_READ);
       if (f.get() == NULL) {

           m_memorystream = NULL;
           setFinished();
         return;

       }
      // uti_refptr<CBufferedReadStream> buffer_f(new CBufferedReadStream(f.get(), 16384, FileSize(m_filename)));

       zipstream = new CZipMemoryStream(f.get());

       zipstream->beginRead();
       m_memorystream->setAllowWrite(true);
       Array<unsigned char> databuffer(65536);

       while (true) {
          int readcount = zipstream->ReadData(databuffer.array, databuffer.size());
          m_memorystream->WriteData(databuffer.array, readcount);
          if (readcount < databuffer.size()) {
              break;
          }

       }
       m_memorystream->setAllowWrite(false);
       m_memorystream->setPosition(0, CStream::START);


       zipstream->endRead();
       zipstream = NULL;
      f = NULL;
      setFinished();
}


CWriteStreamJob::CWriteStreamJob(const CUtiString& f, CMemoryStream* mem)
{
  m_filename = f;
  m_memorystream = mem;
}

void CWriteStreamJob::executeJob(int startjob, int endjob)
{
    m_memorystream->SaveToFile(m_filename);
       setFinished();
}
