/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "UtiArchiv.h"
#include "UtiString.h"
#include <fstream>
//#include <windows.h>
#include "utisystem.h"
#include <math.h>
#include <assert.h>
#include "objectfactory.h"
#include "utiarray.h"
#include <vector>
#include <algorithm>

//#include <iostream>

#include "md5.c"

using namespace UtiFramework;

using namespace std;

MAKE_SINGLETON_IMPL(CFileSystem);

SCode UtiMD5(const CUtiString& source)
{
	FILE * sourcefile = fopen(converttomulti(source).c_str(), "rb");

	MD5_CTX md5ex;

	const int BFRSIZE = 65536;
	char c[BFRSIZE];

	int buffersize = BFRSIZE;
	int readcount = 0;
	int segmente = 0;
	int count = 0;

	MD5_Init(&md5ex);

	do {
		readcount = fread(c, 1, buffersize, sourcefile);
		count += readcount;
		//fwrite(c, 1, readcount, destfile);

		MD5_Update(&md5ex, c, readcount);
		
		segmente++;
	} while (readcount == buffersize);

	fclose(sourcefile);

	SCode result;
	assert(sizeof(result) == 16);
	MD5_Final((unsigned char*)&result, &md5ex);
	return result;
}



void UtiCopyFile(const CUtiString& source, const CUtiString& dest, bool visible)
{
   if (visible) {
      wcout << L"Copy " << source << L" to " << dest << endl;
   }
/*#ifdef WIN32
   CopyFile(source.c_str(), dest.c_str(), false);

#else*/
   FILE * sourcefile = fopen(converttomulti(source).c_str(), "rb");
   FILE * destfile = fopen(converttomulti(dest).c_str(), "wb");
   if (!destfile || !sourcefile) {
      wcout << L"Kopieren nicht möglich" << endl;
      return;
   }
   const int BFRSIZE=65536;
   char c[BFRSIZE];
   int buffersize = BFRSIZE;
   int readcount=0;
   int segmente =0;
   int count=0;
   do {
      readcount = fread(c, 1, buffersize, sourcefile);
      count += readcount;
      fwrite(c, 1, readcount, destfile);
      if (visible) {
         if (segmente % 256 == 0) wcout << L" ";
         if (segmente % 16==0) wcout << L"*";
      }
      segmente++;
   }  while (readcount == buffersize);

     if (visible) {
   wcout << count << L" Bytes kopiert" << endl;
     }
   fclose(sourcefile);
   fclose(destfile);
//#endif
}


CUtiString UtiGetResolvedFileName(const CUtiString& str)
{
    return CFileSystem::getInstance()->resolveFile(str);
}

CStream* UtiOpenFile(const CUtiString& str, TOpenType t)
{
    // MessageBox(0, UTEXT("n"), name, 0);
   return CFileSystem::getInstance()->openFile(str, t);
   
}

UTI_INT64 UtiFileSize(const CUtiString& str)
{
   return CFileSystem::getInstance()->fileSize(str);
}

bool UtiFileExists(const CUtiString& str)
{
   return CFileSystem::getInstance()->fileExists(str);
}

CRegisterClassSystem<CFileSystem, CFileSystemArchiv> registerFileSystem(UTEXT("filesystem"));
CRegisterClassSystem<CFileSystem, CUtiArchiv> registerUtiArchiv(UTEXT("pak"));

CMemoryFile::CMemoryFile()
{

}

CMemoryDataFile::CMemoryDataFile(SInlineFile* f)
{
  m_file = f;
}

CUtiString CMemoryDataFile::getFileName()
{
   return m_file->getFileName();
}


bool CMemoryDataFile::isDirectory()
{
   return false;
}

void CMemoryDirectory::addFile(CMemoryFile* file)
{
    CUtiString name = file->getFileName();
    utiassert(m_nametofile.find(name) == m_nametofile.end());
    m_nametofile.insert(make_pair(name, file));
    m_directory.push_back(file);
    
}

CMemoryDirectory::CMemoryDirectory(const CUtiString& name): CMemoryFile()
{
  m_name = name;
}

bool CMemoryDirectory::isDirectory()
{
    return true;
}

CUtiString CMemoryDirectory::getFileName()
{
   return m_name;
}


CMemoryDataFile* CMemoryDirectory::getFile(const CUtiString& str)
{
  std::map<CUtiString, CMemoryFile*>::iterator it = m_nametofile.find(str);
  if (it == m_nametofile.end()) return NULL;
  CMemoryFile* f = it->second;
  if (f->isDirectory()) return NULL;
  return (CMemoryDataFile*) f;
}

CMemoryDirectory* CMemoryDirectory::getDir(const CUtiString& str)
{
  std::map<CUtiString, CMemoryFile*>::iterator it = m_nametofile.find(str);
  if (it == m_nametofile.end()) return NULL;
  CMemoryFile* f = it->second;
  if (!f->isDirectory()) return NULL;
  return (CMemoryDirectory*) f;
}

CFileSystem::CFileSystem()
{
   Archives.deleteItems = true;
   factory = new CObjectFactory();
}

CFileSystem::~CFileSystem()
{
  CCSControl control(cs_accessjob);
   delete factory;
   Archives.Clear();
}

void CFileSystem::registerClass(const CUtiString& str, CClassCreator* creator)
{
  CCSControl control(cs_accessjob);
   factory->registerClass(str, creator);
}

CBaseArchiv* CFileSystem::initArchiv(const CUtiString& str, const CUtiString& type)
{
  CCSControl control(cs_accessjob);
   CBaseArchiv* MainArchive = dynamic_cast<CBaseArchiv*>(factory->createObject(type));
   utiassert(MainArchive != NULL);
   MainArchive->init(str);
   Archives.Add(MainArchive);
   return MainArchive;
}

CBaseArchiv* CFileSystem::initArchivFirst(const CUtiString& str, const CUtiString& type)
{
  CCSControl control(cs_accessjob);
   CBaseArchiv* MainArchive = dynamic_cast<CBaseArchiv*>(factory->createObject(type));
   utiassert(MainArchive != NULL);
   MainArchive->init(str);
   Archives.Insert(MainArchive, 0);
   return MainArchive;
}


void CFileSystem::initMemoryDirectory(CMemoryDirectory* dir)
{
  CCSControl control(cs_accessjob);
  CMemorySystemArchiv* archiv =new CMemorySystemArchiv(dir);
  
  Archives.Add(archiv);
}

void CFileSystem::freeArchiv(const CUtiString& str)
{
  CCSControl control(cs_accessjob);
   for (int i = 0; i < Archives.count; i++) {
      CBaseArchiv* MainArchive = (CBaseArchiv*)Archives[i];
      if (MainArchive->getArchivName() == str)
      {
         Archives.Delete(i);
         return;
      };
   };
   utiassert(false);
}


CUtiString getParagraph()
{
    CUtiString str;
    str = UTISTR(" ");
    str[0] = UTI_PARAGRAPH;
    return str;
}

CUtiString CFileSystem::resolveFile(const CUtiString& str)
{

    CCSControl control(cs_accessjob);
    if (str.size() > 0 && (str[0] == L'§' || str[0] == 167)) {

         return str;

    }
    CUtiString strout;
     for (int i = 0; i < Archives.count; i++) {
        CBaseArchiv* MainArchive = (CBaseArchiv*)Archives[i];
        if (MainArchive->resolveFile(str, strout))
        {

           return strout;
        };
     };
     return str;
}


void CFileSystem::resolveAllFiles(const CUtiString& str, CStringList& list)
{
    CCSControl control(cs_accessjob);
    if (str.size() > 0 && (str[0] == L'§' || str[0] == 167)) {

         return;

    }
    CUtiString strout;
     for (int i = 0; i < Archives.count; i++) {
        CBaseArchiv* MainArchive = (CBaseArchiv*)Archives[i];
        if (MainArchive->resolveFile(str, strout))
        {

           list.add(strout);
        };
     };
}

CStream* CFileSystem::openFile(const CUtiString& str, TOpenType ot)
{
  CCSControl control(cs_accessjob);

  if (str.size() > 0 && (str[0] == L'§' || str[0] == 167)) {
   CUtiString substr;
   str.Copy(substr, 1, str.size()-1);
   if (FileExists(substr)) {
       switch (ot) {
          case otRead: {
             return new CFileStream(substr, S_READ);
          }; break;
          case otAppend:{
             return new CFileStream(substr, S_WRITE, false);
          }; break;
          case otReset:{
             return new CFileStream(substr, S_WRITE, true);
          }; break;
       };


   } else {
       return NULL;
   }
  }
   for (int i = 0; i < Archives.count; i++) {
      CBaseArchiv* MainArchive = (CBaseArchiv*)Archives[i];
      if (MainArchive->fileExists(str))
      {
        
         CStream* s = MainArchive->openFile(str, ot);
         errorassert(s != NULL, CUtiString(L"Could not open file '") + str + CUtiString(L"'"));
         return s;
      };
   };
   errorassert(false, CUtiString(L"Could not find file '") + str + CUtiString(L"'"));
   return NULL;
}

bool CFileSystem::fileExists(const CUtiString& str)
{
  CCSControl control(cs_accessjob);

  if (str.size() > 0) {
	  wchar_t ch = str[0];
	  //wchar_t* ch2 = L"§";
	  if (ch == L'§' || ch == 167) {
		  CUtiString substr;
		  str.Copy(substr, 1, str.size() - 1);
		  return FileExists(substr);
	  }
  }
   for (int i = 0; i < Archives.count; i++) {
      CBaseArchiv* MainArchive = (CBaseArchiv*)Archives[i];
      if (MainArchive->fileExists(str))
      {
        
         return true;
      };
   };
   return false;
}
UTI_INT64 CFileSystem::fileSize(const CUtiString& str)
{
  CCSControl control(cs_accessjob);
  if (str.size() > 0) {
	  wchar_t ch = str[0];
	  //wchar_t* ch2 = L"§";
	  if (ch == L'§' || ch == 167) {
		  CUtiString substr;
		  str.Copy(substr, 1, str.size() - 1);
		  return FileSize(substr);
	  }
  }
   for (int i = 0; i < Archives.count; i++) {
      CBaseArchiv* MainArchive = (CBaseArchiv*)Archives[i];
      if (MainArchive->fileExists(str))
      {
        
         return MainArchive->fileSize(str);
      };
   };
   errorassert(false, CUtiString(L"Could not find file '") + str + CUtiString(L"'"));
   return 0;
}
void CFileSystem::deleteFile(const CUtiString& str)
{
   CCSControl control(cs_accessjob);
}
void CFileSystem::getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result, bool relative, bool searchdirs)
{
  CCSControl control(cs_accessjob);
   for (int i = 0; i < Archives.count; i++) {
      CBaseArchiv* MainArchive = (CBaseArchiv*)Archives[i];
      MainArchive->getFilesInDir(dir, type, result, relative, searchdirs);
   };
}

CBaseArchiv::CBaseArchiv()
{
  
}

CBaseArchiv::~CBaseArchiv()
{
   
}

void CBaseArchiv::init(const CUtiString& _name)
{
    archivname = _name;
}

CUtiString CBaseArchiv::getArchivName()
{
   return archivname;
}
void CBaseArchiv::deleteFile(const CUtiString& str)
{
  
}

bool CBaseArchiv::fileExists(const CUtiString& name)
{
  return false;
}
UTI_INT64 CBaseArchiv::fileSize(const CUtiString& name)
{
   return -1;
};
CStream* CBaseArchiv::openFile(const CUtiString& name, TOpenType ot)
{
   return NULL;
}

void CBaseArchiv::getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result, bool relativ, bool searchdirs)
{
   
}

CFileSystemArchiv::CFileSystemArchiv()
{
   m_allowNormal = false;
}
CFileSystemArchiv::~CFileSystemArchiv()
{
   
}

void CFileSystemArchiv::setArchivNormal(bool normal)
{
  m_allowNormal = normal; 
}

void CFileSystemArchiv::init(const CUtiString& _name)
{
   archivname = _name;
}


void CFileSystemArchiv::deleteFile(const CUtiString& str)
{
   
}


bool CFileSystemArchiv::resolveFile(const CUtiString& input, CUtiString& output)
{
    if (input.size() == 0) return false;
    CUtiString RealPos = archivname + input;
    bool b = FileExists(RealPos.getString());
   // if (!b) {
   // std::wcout << RealPos.getString() << std::endl;
   // }
    output = RealPos;
    if (!b && m_allowNormal) {
        output = input;
      return FileExists(input.getString());

    };
    return b;
}

bool CFileSystemArchiv::fileExists(const CUtiString& name)
{
  if (name.size() == 0) return false;

   CUtiString RealPos = archivname + name;
   bool b = FileExists(RealPos.getString());
   if (!b && m_allowNormal) {
     return FileExists(name.getString());

   };
  return b;
}
UTI_INT64 CFileSystemArchiv::fileSize(const CUtiString& name)
{
   CUtiString RealPos = archivname + name;
   bool b = FileExists(RealPos.getString());

   if (b)
   {
      return FileSize(RealPos.c_str());
   } else if (m_allowNormal && FileExists(name.getString())) {
     return FileSize(name.c_str());
   } else {
      return -1;
   };
}
CStream* CFileSystemArchiv::openFile(const CUtiString& name, TOpenType ot)
{
   CUtiString RealPos = archivname + name;
   bool b = FileExists(RealPos.getString());

   if (b)
   {
      // std::wcout << "Loaded File: " << RealPos << std::endl;
      switch (ot) {
         case otRead: {
            return new CFileStream(RealPos, S_READ);
         }; break;
         case otAppend:{
            return new CFileStream(RealPos, S_WRITE, false);
         }; break;
         case otReset:{
            return new CFileStream(RealPos, S_WRITE, true);
         }; break;
      };
         
     return NULL;
      
   } else if (m_allowNormal && FileExists(name.c_str())) {
     switch (ot) {
         case otRead: {
            return new CFileStream(name, S_READ);
         }; break;
         case otAppend:{
            return new CFileStream(name, S_WRITE, false);
         }; break;
         case otReset:{
            return new CFileStream(name, S_WRITE, true);
         }; break;
      };
         
     return NULL;
   } else {
      return NULL;
   };
}

void findFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result)
{
    CUtiString RealPos;
    RealPos =  dir;
    Dir_FileInfo data;
    bool h = FindFirst(RealPos.getString(), &data);
    //CUtiString* name;

    if (h != false)
    {
       CUtiString k;
       if (!data.dir){
       k = data.name;
       k.ExtractFileExt();
       k= UTEXT('.')+k;
       if (k == type || type.size() == 0)
       {
          k = data.name;
          if (result.find(CFileSystem_Info(k)) == result.end()) {
             result.insert(CFileSystem_Info(k, data.size, data.dir));
          };
       }
       }
       while(FindNext(&data))
       {
          CUtiString k;
          if (!data.dir){
          k = data.name;
          k.ExtractFileExt();
          k= UTEXT('.')+k;
          if (k == type || type.size() == 0)
          {
             k = data.name;
             if (result.find(CFileSystem_Info(k)) == result.end()) {
                result.insert(CFileSystem_Info(k, data.size, data.dir));
             };
          }
          }

       }
    } else {
        return;
    }
    FindClose(&data);
}

void findDirsInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result)
{
    CUtiString RealPos;
    RealPos =  dir;
    Dir_FileInfo data;
    bool h = FindFirst(RealPos.getString(), &data);
    //CUtiString* name;

    if (h != false)
    {
       CUtiString k;
       k = data.name;
       if (data.dir) {
       k.ExtractFileExt();
       k= UTEXT('.')+k;
       if (k == type || type.size() == 0)
       {
          k = data.name;
          if (result.find(CFileSystem_Info(k)) == result.end()) {
             result.insert(CFileSystem_Info(k, data.size, data.dir));
          };
       }
       }
       while(FindNext(&data))
       {
          CUtiString k;
          if (data.dir) {
          k = data.name;
          k.ExtractFileExt();
          k= UTEXT('.')+k;
          if (k == type || type.size() == 0)
          {
             k = data.name;
             if (result.find(CFileSystem_Info(k)) == result.end()) {
                result.insert(CFileSystem_Info(k, data.size, data.dir));
             };
          }
          }

       }
    } else {
        return;
    }
    FindClose(&data);
}

void CFileSystemArchiv::getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result, bool relativ, bool searchdirs)
{
   CUtiString RealPos;
   if (relativ) {
   RealPos = archivname + dir;
   } else {
      RealPos = dir;
   }
   if (searchdirs) {
      findDirsInDir(RealPos, type, result);
   } else {
       findFilesInDir(RealPos, type, result);
   }
   /*
   Dir_FileInfo data;
   bool h = FindFirst(RealPos.getString(), &data);
   //CUtiString* name;

   if (h != false)
   {
      CUtiString k;
      k = data.name;
      k.ExtractFileExt();
      k= UTEXT('.')+k;
      if (k == type)
      {
         k = data.name;
         if (result.find(k) == result.end()) {
            result.insert(k);
         };
      }
      while(FindNext(&data))
      {
         CUtiString k;
         k = data.name;
         k.ExtractFileExt();
         k= UTEXT('.')+k;
         if (k == type)
         {
            k = data.name;
            if (result.find(k) == result.end()) {
               result.insert(k);
            };
         }

      }
   } else {
       return;
   }
   FindClose(&data);*/
}

CUtiArchiv::CUtiArchiv() :CBaseArchiv()
{

   
    //	rseed = time();
}


void CUtiArchiv::init(const CUtiString& _name)
{
   CBaseArchiv::init(_name); 
   Files.deleteItems = true;
    Directories.deleteItems= true;
    //str = NULL;
    if (FileExists(archivname.getString()))
    {
        Load();
    }
    else
    {
        Init();
    };
    //	FileList = NULL;
    LastExists = NULL;
    StartPath = UTEXT("");
}


CUtiArchiv::~CUtiArchiv()
{
    //if (str) delete str;
}

CMinMaxFileStream::~CMinMaxFileStream()
{}

bool CMinMaxFileStream::eof()
{
    fstream* str = (fstream*)strm;
    return(str->eof()!=0|| getReadPosition() >= Max);
};

CMinMaxFileStream::CMinMaxFileStream(const U_CHAR *name, int i, bool trunc, UTI_INT64 Min,UTI_INT64 Max) : CFileStream(name,i,trunc)
{
	utiassert(i == S_READ);
    this->Min = Min;
    this->Max = Max;
    setPosition(0, START);
}

void CMinMaxFileStream::setPosition(off_t i, TPositionType postype)
{
    if (postype == START) {
       CFileStream::setPosition(i+Min, START);
    } else if (postype == CURRENT) {
        CFileStream::setPosition(i, CURRENT);
    } else if (postype == END) {
        CFileStream::setPosition(Max-i, START);
	};
}

size_t CMinMaxFileStream::ReadData(void* data, size_t size)
{
	
    int pos = getReadPosition();
    utiassert((pos >= Min));
    utiassert(pos+size <= Max);
    return CFileStream::ReadData(data, size);
}

UTI_INT64 CUtiArchiv::GetFileSize()
{
    UTI_INT64 s=0;
    for (int i = 0; i < Files.count; i++)
    {
        CUtiAFile* d = (CUtiAFile*)Files[i];
        s += d->Info.OriginalSize;

    };
    return s;
}

int CUtiArchiv::GetFileCount()
{
    return Header.FileCount;
}

int CUtiArchiv::GetDirCount()
{
    return Header.DirCount;
}

void CUtiAFile::GetFileName(CUtiString &fn)
{
    fn = UTEXT("");
    CUtiADirectory* p = Parent;
    while (p)
    {
        fn += p->Name + UTEXT("/");
        p = p->Parent;
    };
    fn += Name.getString();
}
/*
void CUtiArchiv::AddFileToList(CUtiAFile* f)
{
   if (!FileList) return;
   CUtiString fn;
   f->GetFileName(fn);
   FileList->Add(fn.getString());
}*/

/*void CUtiArchiv::FillList()
{
   if (!FileList) return;
   for (int i = 0; i < Files.count; i++) {
	   CUtiAFile* d = (CUtiAFile*)Files[i];
	   AddFileToList(d);
 
   };
}*/

CUtiADirectory* CUtiArchiv::FindDir(UTI_INT64 ID)
{
    if (ID == -1)
        return NULL;
    for (int i = 0; i < Directories.count; i++)
    {
        CUtiADirectory* d = (CUtiADirectory*)Directories[i];
        if (d->Info.DirID == ID)
        {
            return d;

        };
    };
    return NULL;
}

CUtiADirectory* CUtiArchiv::GetDir(const CUtiString &_Path)
{
    CUtiString Path = _Path;
   if (Path == UTEXT(""))
    {
        return NULL;
    };
    //CUtiString paths[16];
    UTI_DECL_ARRAY(CUtiString, paths, 16);
    
    int p = Path.Pos(UTEXT("/"));
    int c = 0;
    if (p != -1)
    {
        while (p != -1)
        {
            Path.Copy(paths[c], 0, p);
            c++;
            Path.Delete(0, p+1);
            p = Path.Pos(UTEXT("/"));
        }
        paths[c] = Path;
        c++;

        if (c >= paths.size()) {return NULL;};
    }
    else
    {
        c = 1;
        paths[0] = Path;
    };
    for (int i = 0; i < Directories.count; i++)
    {
        CUtiADirectory* d = (CUtiADirectory*)Directories[i];
        if (paths[c-1] != d->Name)
            continue;
        CUtiADirectory* p = d->Parent;
        bool fail= false;
        for (int j = c-2; j >= 0; j--)
        {
            if (p == NULL)
            {
                fail = true;
                break;
            };
            if (paths[j] != d->Name)
            {
                fail = true;
                break;
            };
            p = p->Parent;

        };
        if (fail)
            continue;
        return d;

    };
    return NULL;
}

CStream* CUtiArchiv::GetReadStream()
{
    return new CFileStream(archivname.getString(), S_READ);
}

CStream* CUtiArchiv::GetWriteStream()
{
    CStream* str;
    if (FileExists(archivname.getString()))
    {
       str = new CFileStream(archivname.getString(), S_WRITE, false);
    }
    else
    {
       str = new CFileStream(archivname.getString(), S_WRITE);
    }
    return str;
}

void CUtiArchiv::DeleteStream(CStream* *str)
{
    delete *(CFileStream**)str;
}

void CUtiArchiv::Init()
{
    CFileStream* str = (CFileStream*)GetWriteStream();
    //CFileStream* str = new CFileStream(Name.getString(), S_WRITE);

    Directories.Clear();
    Files.Clear();
    Header.DirCount = 0;
    Header.FileCount = 0;
    Header.FilterCount = 0;
    Header.Flags = 0;
    Header.ID =0xA046B476;
    Header.MajorVersion = 1;
    Header.MinorVersion = 0;
    str->WriteData(&Header, sizeof Header);
    //delete str;
    DeleteStream((CStream**)&str);
}

void CUtiArchiv::Load()
{
    CStream* str = GetReadStream();
    str->setPosition(0, CStream::START);
    Directories.Clear();
    Files.Clear();
    str->ReadData(&Header, sizeof Header);
    ArchivDirectory Dir;
    ArchivFile File;
    int pos;
    int count = Header.DirCount + Header.FileCount;
    for (int i = 0; i < count; i++)
    {
        UINT id;
        pos = str->getReadPosition();
        str->ReadData(&id, sizeof(UINT));
        str->setPosition(str->getReadPosition()-sizeof(UINT), CStream::START);
        if (id == 1 || id == 3)
        {
            File.ID = id;
            str->ReadData(&File, sizeof File);
            if (id == 1)
            {
                CUtiAFile* f = new CUtiAFile();
                f->Info = File;
                f->Position = pos;
                f->Name = File.name;
                Files.Add(f);
            };
            str->setPosition(str->getReadPosition()+File.Size, CStream::START);
        }
        if (id == 2 || id == 4)
        {
            Dir.ID = id;
            str->ReadData(&Dir, sizeof Dir);
            if (id == 2)
            {
                CUtiADirectory* d = new CUtiADirectory();
                d->Info = Dir;
                d->Position = pos;
                d->Name = Dir.name;
                Directories.Add(d);

            }
        }
    }
    DeleteStream((CStream**)&str);
    for (int i = 0; i < Directories.count; i++)
    {
        CUtiADirectory* d = (CUtiADirectory*)Directories[i];
        d->Parent = FindDir(d->Info.Directory);

    };
    for (int i = 0; i < Files.count; i++)
    {
        CUtiAFile* d = (CUtiAFile*)Files[i];
        d->Parent = FindDir(d->Info.Directory);

    };
}

CUtiADirectory* CUtiArchiv::CreateDir(CUtiADirectory* Pos, const CUtiString& name)
{
    CStream* str = GetWriteStream();
    ArchivDirectory Dir;
    CUtiADirectory* d = new CUtiADirectory();
    d->Position = str->getWritePosition();
    d->Parent = Pos;
    if (d->Parent)
    {
        Dir.Directory = d->Parent->Info.DirID;
        d->Parent->Info.ChildrenCount++;
    }
    else
        Dir.Directory = -1;
    Dir.ChildrenCount = 0;
    Dir.ID = 2;
    d->Name = name;
    wcscpy(Dir.name, name.c_str());

    Dir.DirID = (UTI_UINT64)urandom(65536) | (UTI_UINT64)urandom(65536) << 16 |(UTI_UINT64)urandom(65536) << 32;
    Dir.Size = 0;
    d->Info = Dir;
    Directories.Add(d);
    str->WriteData(&Dir, sizeof Dir);
    //Header
    str->setPosition(0, CStream::START);

    Header.DirCount++;
    str->WriteData(&Header, sizeof Header);
    //Parentdir;
    if (d->Parent)
    {
        str->setPosition(d->Parent->Position, CStream::START);
        str->WriteData(&d->Parent->Info, sizeof d->Parent->Info);
    }
    DeleteStream((CStream**)&str);
    return d;
}

void CUtiArchiv::AddFile(const CUtiString&  FileName, const CUtiString&  Name, CUtiADirectory* Pos)
{
    if (!FileExists(FileName.c_str()))
        return;
    CStream* str = GetWriteStream();
    ArchivFile File;
    CUtiAFile* f = new CUtiAFile();
    f->Position = str->getWritePosition();
    f->Parent = Pos;
    if (f->Parent)
    {
        File.Directory = f->Parent->Info.DirID;
        f->Parent->Info.ChildrenCount++;
    }
    else
        File.Directory = -1;
    File.Filter = 0;
    File.ID = 1;

    f->Name = Name;
    wcscpy(File.name, Name.c_str());

    File.Size = FileSize(FileName.c_str());
    File.OriginalSize = File.Size;
    //Gr�e ermitteln!

    f->Info = File;
    Files.Add(f);
    str->WriteData(&File, sizeof File);

    //Datei schreiben
    CFileStream* sf = new CFileStream(FileName, S_READ);
    UTI_INT64 size = f->Info.Size;
    UTI_INT64 PufferSize = 65536;
    U_CHAR* Puffer = new U_CHAR[PufferSize];
    while (size > 0)
    {
        sf->ReadData(Puffer, min(size, PufferSize));
        str->WriteData(Puffer, min(size, PufferSize));

        size -= PufferSize;
    };
    delete []Puffer;
    delete sf;
    //Header
    str->setPosition(0, CStream::START);

    Header.FileCount++;
    str->WriteData(&Header, sizeof Header);
    //Parentdir;
    if (f->Parent)
    {
        str->setPosition(f->Parent->Position, CStream::START);
        str->WriteData(&f->Parent->Info, sizeof f->Parent->Info);
    }
    DeleteStream((CStream**)&str);
    // AddFileToList(f);
}

void CUtiArchiv::deleteFile(const CUtiString &name)
{}

void CUtiArchiv::Pack()
{}

CUtiAFile* CUtiArchiv::GetFile(const CUtiString &Path)
{
    CUtiString fn;
    CUtiAFile* f = LastExists;
    if (f)
    {
        f->GetFileName(fn);
        if (fn == Path)
        {
            return f;
        }
    }
    CUtiString name;
    name = Path;
    name.ExtractFileName();
    for (int i = 0; i < Files.count; i++)
    {
        f = (CUtiAFile*)Files[i];
        if (f->Name == name)
        {
            f->GetFileName(fn);
            if (fn == Path)
            {
                LastExists = f;
                return f;
            }
        }

    };
    return NULL;
}

void CUtiArchiv::AddRecursiv(const CUtiString&  dir, CUtiADirectory* Subdir,const CUtiString&  PathSubdir, const CUtiString&  types)
{
    CUtiString Dir(dir);
    Dir += PathSubdir;
    Dir += UTEXT("/");
    Dir += types;
    Dir_FileInfo FindFileData;

    if (!FindFirst(Dir, &FindFileData))
        return; //Da ist nichts

    do
    {
        if (FindFileData.name==UTEXT(".."))
            continue;
        if (FindFileData.name==UTEXT("."))
            continue;
        if (FindFileData.dir)
        {
            CUtiString sdir(PathSubdir);
            sdir += UTEXT("/");
            sdir += FindFileData.name;


            CUtiADirectory* newsub = CreateDir(Subdir, FindFileData.name.getString());
            AddRecursiv(dir, newsub, sdir.getString(), types);

        }
        else
        {
            CUtiString fname(dir);
            fname += PathSubdir;
            fname += UTEXT("/");
            fname += FindFileData.name;
            AddFile(fname.getString(), FindFileData.name.getString(), Subdir);

        }

    }
    while(FindNext(&FindFileData));
    FindClose(&FindFileData);
}

void CUtiArchiv::AddDir(const CUtiString&  dir, const CUtiString&  types, bool recursiv)
{

    CUtiString Dir(dir);
    Dir += types;

    Dir_FileInfo FindFileData;
    if (!FindFirst(Dir, &FindFileData))
        return; //Da ist nichts

    do
    {
        if (FindFileData.name==UTEXT(".."))
            continue;
        if (FindFileData.name==UTEXT("."))
            continue;
        if (FindFileData.dir)
        {
            if (recursiv)
            {
                CUtiADirectory* newsub = CreateDir(NULL, FindFileData.name.getString());
                AddRecursiv(dir, newsub, FindFileData.name.getString(), types);
            };
        }
        else
        {
            CUtiString fname(dir);
            fname += FindFileData.name;
            AddFile(fname.getString(), FindFileData.name.getString(), NULL);

        }

    }
    while(FindNext(&FindFileData));
    FindClose(&FindFileData);
}

bool CUtiArchiv::fileExists(const CUtiString&  name)
{
   if (name.size() == 0) return false;
    CUtiString RealPos;
    RealPos = StartPath+name;
   /* bool b = FileExists(RealPos.getString());

    if (b)
        return true;*/

    CUtiAFile* f = GetFile(name);
    if (f)
    {
        return true;
    }
    else
    {
        // MessageBox(0, name, UTEXT("Datei nicht gefunden"), 0);
        return false;
    }
}

UTI_INT64 CUtiArchiv::fileSize(const CUtiString&  name)
{
    CUtiString RealPos;
    RealPos = StartPath+name;
    /*bool b = FileExists(RealPos.getString());

    if (b)
    {
        return FileSize(name);
    }
*/
    CUtiAFile* f = GetFile(name);
    if (f)
    {
        return f->Info.OriginalSize;
    }
    else
    {
        // MessageBox(0, name, UTEXT("Datei nicht gefunden"), 0);
        return 0;
    }
}

CStream* CUtiArchiv::openFile(const CUtiString& name, TOpenType ot)
{
    //MessageBox(0, name, UTEXT("Datei nicht gefunden"), 0);
   errorassert(ot == otRead, CUtiString(L"A file in an archiv can only be opened: ")+name);
    CUtiString RealPos;
    RealPos = StartPath+name;
    /*bool b = FileExists(RealPos.getString());

    if (b)
    {
        return new CFileStream(name, S_READ);
    }*/
    CUtiAFile* f = GetFile(name);
    if (f)
    {
        return new CMinMaxFileStream(archivname.getString(), S_READ, true, f->Position+sizeof(ArchivFile),f->Position+sizeof(ArchivFile)+f->Info.Size);
    }
    else
    {
        // MessageBox(0, name, UTEXT("Datei nicht gefunden"), 0);
        return NULL;

    }

    /*CUtiAFile* f = GetFile(name);
      if (f) {
       return new CMinMaxFileStream(Name.getString(), S_READ, true, f->Position+sizeof(ArchivFile),f->Position+sizeof(ArchivFile)+f->Info.Size);
      } else {
       CUtiString RealPos;
    RealPos = StartPath+name;
    bool b = FileExists(RealPos.getString());
    if (b) {
    	return new CFileStream(name, S_READ);
    }
       return NULL;
      }*/

}

void CUtiArchiv::ExtractFileDirect(CUtiAFile* f, const CUtiString&  target)
{

    if (!f)
        return;
    CUtiString fpath(target);
    CUtiString subpath;
    f->GetFileName(subpath);
    fpath += UTEXT("/");
    fpath += subpath;
    fpath.ExtractFilePath();
    if (!FileExists(fpath.getString()))
    {
        ForceDir(fpath.getString());
    };

    CStream* str = GetReadStream();
    ArchivFile file;
    str->setPosition(f->Position, CStream::START);
    str->ReadData(&file, sizeof file);

    //Schreiben
    CUtiString FileName(target);
    FileName +=UTEXT("/");
    FileName += subpath;
    /*FileName +=UTEXT("/");
       FileName += name;*/
    CFileStream* sf = new CFileStream(FileName.getString(), S_WRITE);
    UTI_INT64 size = f->Info.Size;
    UTI_INT64 PufferSize = 65536;
    U_CHAR* Puffer = new U_CHAR[PufferSize];
    while (size > 0)
    {
        str->ReadData(Puffer, min(size, PufferSize));
        sf->WriteData(Puffer, min(size, PufferSize));
        size -= PufferSize;


    };
    delete []Puffer;
    delete sf;


    DeleteStream((CStream**)&str);
}

void CUtiArchiv::getFilesInDir(const CUtiString&  dir, const CUtiString&  type, set<CFileSystem_Info> & result, bool relativ, bool searchdirs)
{
   if (!relativ)return;
    CUtiADirectory* Parent = GetDir(dir);
   
    if (searchdirs) {
        for (int i = 0; i < Directories.count; i++)
        {
            CUtiADirectory* f = (CUtiADirectory*)Directories[i];
            if (f->Parent != Parent)
                continue;
          CUtiString k;
            k = f->Name;
            if (result.find(CFileSystem_Info(k)) == result.end()) {
               k.ExtractFileExt();
               k= UTEXT('.')+k;
               if (k == type || type.size() == 0)
               {

                result.insert(CFileSystem_Info(f->Name, 0, true));
               }
            };


        };
    }else {
    for (int i = 0; i < Files.count; i++)
    {
        CUtiAFile* f = (CUtiAFile*)Files[i];
        if (f->Parent != Parent)
            continue;
      CUtiString k;
        k = f->Name;
        if (result.find(CFileSystem_Info(k)) == result.end()) {
           k.ExtractFileExt();
           k= UTEXT('.')+k;
           if (k == type || type.size() == 0)
           {
            result.insert(CFileSystem_Info(f->Name,0, false));
           }
        };
       
        
    };
    };
    
}

void CUtiArchiv::ExtractFile(const CUtiString& name, const CUtiString&  target)
{
    CUtiAFile* f = GetFile(name);
    ExtractFileDirect(f, target);
}

CMemorySystemArchiv::CMemorySystemArchiv(CMemoryDirectory* rootdir): CBaseArchiv()
{
   m_rootdir = rootdir;
}

CMemorySystemArchiv::~CMemorySystemArchiv()
{

}

void CMemorySystemArchiv::deleteFile(const CUtiString& str)
{
   
}

bool CMemorySystemArchiv::fileExists(const CUtiString& name)
{
    return GetFile(name) != NULL;
}

UTI_INT64 CMemorySystemArchiv::fileSize(const CUtiString& name)
{
    CMemoryDataFile* f = GetFile(name);
    if (!f) return -1;
    return f->m_file->getFileSize();
}

void CMemorySystemArchiv::getFilesInDir(const CUtiString& dir, const CUtiString& type, set< CFileSystem_Info >& result, bool relativ, bool searchdirs)
{ if (!relativ)return;
    CMemoryDirectory* Parent = GetDir(dir);
    if (Parent == NULL)return;
    if (searchdirs) {
        for (int i = 0; i < Parent->m_directory.size(); i++)
        {
          CMemoryFile* f = &Parent->m_directory[i];
          if (!f->isDirectory()) continue;

CUtiString k;
            k = f->getFileName();
            if (result.find(CFileSystem_Info(k)) == result.end()) {
               k.ExtractFileExt();
               k= UTEXT('.')+k;
               if (k == type || type.size() == 0)
               {
                result.insert(CFileSystem_Info(f->getFileName(), 0, f->isDirectory()));
               }
            };


        };
    }else {
    for (int i = 0; i < Parent->m_directory.size(); i++)
    {
      CMemoryFile* f = &Parent->m_directory[i];
      if (f->isDirectory()) continue;
      CMemoryDataFile* fd = check_cast<CMemoryDataFile*>(f);
     
       
      CUtiString k;
        k = f->getFileName();
        if (result.find(CFileSystem_Info(k)) == result.end()) {
           k.ExtractFileExt();
           k= UTEXT('.')+k;
           if (k == type || type.size() == 0)
           {
            result.insert(CFileSystem_Info(f->getFileName(), fd->m_file->getFileSize(), f->isDirectory()));
           }
        };
       
        
    };
    }
    
}

void CMemorySystemArchiv::init(const CUtiString& _name)
{
    CBaseArchiv::init(_name);
}

CStream* CMemorySystemArchiv::openFile(const CUtiString& name, TOpenType ot)
{
  CMemoryDataFile* f = GetFile(name);
    if (!f) return NULL;
    CMemoryStream* mem = new CMemoryStream(f->m_file->getFileData(), f->m_file->getFileSize(), false);
    return mem;
}

CMemoryDirectory* CMemorySystemArchiv::GetDir(const CUtiString& _Path)
{
   CUtiString name = _Path;
    CUtiString p = _Path;
    CUtiString temp;
    name.ExtractFileName();
    p.ExtractFilePath();
	   std::vector<CUtiString> paths;
	   while (p.size() > 0) {
	     temp = p;
	     temp.ExtractFileName();
	     paths.push_back(temp);
	      p.ExtractFilePath();
	   };
    CMemoryDirectory* currentdir = m_rootdir;
    for (int i = paths.size()-1; i>= 0; i--) {   
      CMemoryDirectory* d = currentdir->getDir(paths.at(i));
      if (d == NULL) return NULL;
      currentdir = d;
    };
    CMemoryDirectory* df = currentdir->getDir(name);
    return df;
}

CMemoryDataFile* CMemorySystemArchiv::GetFile(const CUtiString& Path)
{

    CUtiString name = Path;
    CUtiString p = Path;
    CUtiString temp;
    name.ExtractFileName();
    p.ExtractFilePath();
	   std::vector<CUtiString> paths;
	   while (p.size() > 0) {
	     temp = p;
	     temp.ExtractFileName();
	     paths.push_back(temp);
	      p.ExtractFilePath();
	   };
    CMemoryDirectory* currentdir = m_rootdir;
    for (int i = paths.size()-1; i>= 0; i--) {   
      CMemoryDirectory* d = currentdir->getDir(paths.at(i));
      if (d == NULL) return NULL;
      currentdir = d;
    };
    CMemoryDataFile* df = currentdir->getFile(name);
    return df;
}


