#ifndef __UTIARCHIV__
#define __UTIARCHIV__
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "utilists.h"
#include "utistream.h"

#include <set>
#include <map>

class CUtiArchiv;
class CObjectFactory;
class CClassCreator;
class CBaseArchiv;



//extern CUtiArchiv* MainArchive;
//extern bool uti_inarchiv;

//void InitArchive(const U_CHAR* name);
//void FreeArchive();
enum TOpenType {otRead, otAppend, otReset};

CStream* UtiOpenFile(const CUtiString& str, TOpenType t);
bool UtiFileExists(const CUtiString& str);
UTI_INT64 UtiFileSize(const CUtiString& str);
CUtiString UtiGetResolvedFileName(const CUtiString& str);

SCode UtiMD5(const CUtiString& source);
void UtiCopyFile(const CUtiString& source, const CUtiString& dest, bool visible);

struct SInlineFile {
  virtual unsigned int getFileSize()=0;
  virtual unsigned char* getFileData()=0;
  virtual CUtiString getFileName()=0;
};

class CMemoryFile : public CUtiOB 
{
public:
  CMemoryFile();
  virtual CUtiString getFileName()=0;
  virtual bool isDirectory()=0;
  
};

class CMemoryDataFile : public CMemoryFile
{
public:
  SInlineFile* m_file;
  CMemoryDataFile(SInlineFile* f);
  CUtiString getFileName();
  bool isDirectory();
};

class CMemoryDirectory : public CMemoryFile
{
public:
  CUtiString m_name;
  CNormalList<CMemoryFile> m_directory;
  std::map<CUtiString, CMemoryFile*> m_nametofile;
  CMemoryDirectory(const CUtiString& name);
  void addFile(CMemoryFile* file);
  CUtiString getFileName();
  bool isDirectory();
  CMemoryDataFile* getFile(const CUtiString& str);
  CMemoryDirectory* getDir(const CUtiString& str);
  
};

struct CFileSystem_Info
{
    CUtiString m_name;
    UTI_INT64 m_size;
    bool m_directory;
    inline CFileSystem_Info()
    {
        m_size = -1;
        m_directory = false;
    }

    inline CFileSystem_Info(const CUtiString& n, UTI_INT64 s=-1, bool d=false)
    {
        m_name= n;
        m_size = s;
        m_directory = d;
    }

    inline bool operator<(const CFileSystem_Info& info) const
    {
        if (!m_name.equal_lowerCase(info.m_name)) {
        return m_name.less_lowerCase(info.m_name);
        } else {
         return m_name < info.m_name;
        }
    }
    inline bool operator==(const CFileSystem_Info& info) const
    {
        return m_name == info.m_name;
    }
};


const int UTI_PARAGRAPH = 167;
CUtiString getParagraph();

class CFileSystem : public CUtiOB
{
   protected:
      MAKE_SINGLETON(CFileSystem);
      CCacheList Archives;
      CObjectFactory* factory;
      CCriticalSection cs_accessjob;
   public:
      char* m_argv;
      CFileSystem();
      ~CFileSystem();
      inline void setArgv(char* c){m_argv = c;}
      inline char* getArgv(){return m_argv;}
      void registerClass(const CUtiString& str, CClassCreator* creator);
      CBaseArchiv* initArchiv(const CUtiString& str, const CUtiString& type);
      CBaseArchiv* initArchivFirst(const CUtiString& str, const CUtiString& type);
      void initMemoryDirectory(CMemoryDirectory* dir);
      void freeArchiv(const CUtiString& str);
      CStream* openFile(const CUtiString& str, TOpenType ot);
      CUtiString resolveFile(const CUtiString& str);
      void resolveAllFiles(const CUtiString& str, CStringList& list);
      bool fileExists(const CUtiString& str);
      UTI_INT64 fileSize(const CUtiString& str);
      void deleteFile(const CUtiString& str);
      //Enum Dir!
      void getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result, bool relative=true, bool searchdirs=false);
};

class CBaseArchiv : public CUtiOB
{
   protected:
      CUtiString archivname;
   public:
      CBaseArchiv();
      ~CBaseArchiv();
      virtual void init(const CUtiString& _name);
      virtual CUtiString getArchivName();
      virtual bool resolveFile(const CUtiString& input, CUtiString& output){return false;};
      virtual void deleteFile(const CUtiString& str);
      //virtual int getDirectorySize(const CUtiString& name);
      virtual bool fileExists(const CUtiString& name);
      virtual UTI_INT64 fileSize(const CUtiString& name);
      virtual CStream* openFile(const CUtiString& name, TOpenType ot);
      virtual void getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result, bool relativ, bool searchdirs);
};

class CFileSystemArchiv : public CBaseArchiv
{
protected:
    bool m_allowNormal;
   public:
      CFileSystemArchiv();
      ~CFileSystemArchiv();
      void setArchivNormal(bool normal);
      void init(const CUtiString& _name);
      virtual void deleteFile(const CUtiString& str);
      //virtual int getDirectorySize(const CUtiString& name);
      virtual bool fileExists(const CUtiString& name);

      virtual bool resolveFile(const CUtiString& input, CUtiString& output);
      virtual UTI_INT64 fileSize(const CUtiString& name);
      virtual CStream* openFile(const CUtiString& name, TOpenType ot);
      virtual void getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result, bool relativ, bool searchdirs);
};

void findFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result);

struct ArchivHeader
{
    UINT ID;
    USHORT MajorVersion;
    USHORT MinorVersion;
    UINT FilterCount;
    UINT FileCount;
    UINT DirCount;
    UINT Flags;
};

struct ArchivFile
{
    UINT ID; //Dateisignatur:1 //3:
    UTI_INT64 Size;
    UTI_INT64 OriginalSize;
    UTI_INT64 Directory;
    UTI_INT64 Filter;
    U_CHAR name[64];
};

struct ArchivDirectory
{
    UINT ID; //Directorysignatur:2//4:
    UTI_INT64 Size;
    UTI_INT64 DirID;
    UTI_INT64 Directory;
    UTI_INT64 ChildrenCount;
    U_CHAR name[64];
};


class CMinMaxFileStream : public CFileStream
{
public:
    ~CMinMaxFileStream();
    CMinMaxFileStream(const U_CHAR *name, int i, bool trunc=true, UTI_INT64 Min=0,UTI_INT64 Max=0xFFFFFFF);
    UTI_INT64 Min;
    UTI_INT64 Max;
    bool eof();

    size_t ReadData(void* data, size_t size);
    void setPosition(off_t i, TPositionType postype);
};

class CUtiADirectory;

class CUtiAFile : public CUtiOB
{
public:
    ArchivFile Info;
    CUtiString Name;
    UTI_INT64 Position;
    CUtiADirectory* Parent;
    void GetFileName(CUtiString &fn);
};

class CUtiADirectory : public CUtiOB
{
public:
    ArchivDirectory Info;
    CUtiString Name;
    UTI_INT64 Position;
    CUtiADirectory* Parent;
};

class CUtiArchiv : public CBaseArchiv
{
protected:
    ArchivHeader Header;
    CUtiAFile* LastExists;
    CCacheList Directories;
    void Load();
    void Init();
    CStream* GetReadStream();
    CStream* GetWriteStream();
    void DeleteStream(CStream* *str);
    CUtiADirectory* FindDir(UTI_INT64 ID);
    CUtiADirectory* GetDir(const CUtiString &_Path);
    CUtiAFile* GetFile(const CUtiString &Path);
    //void AddFileToList(CUtiAFile* f);
    void AddRecursiv(const CUtiString& dir, CUtiADirectory* Subdir, const CUtiString& PathSubdir,const CUtiString& types);
public:
    CCacheList Files;

    CUtiString StartPath;
    //CUtiListBox* FileList;
    //void FillList();
    UTI_INT64 GetFileSize();
    int GetFileCount();
    int GetDirCount();
    CUtiArchiv();
    ~CUtiArchiv();
    void init(const CUtiString& _name);

    CUtiADirectory* CreateDir(CUtiADirectory* Pos, const CUtiString& name);
    void AddFile(const CUtiString& FileName, const CUtiString& Name, CUtiADirectory* Pos);
    void AddDir(const CUtiString& dir, const CUtiString& types, bool recursiv);
    void deleteFile(const CUtiString& name);
    void Pack();
    void ExtractFile(const CUtiString& name, const CUtiString& target);
    void ExtractFileDirect(CUtiAFile* f, const CUtiString& target);
    int getDirectorySize(const CUtiString& name);
    virtual bool fileExists(const CUtiString& name);
    virtual UTI_INT64 fileSize(const CUtiString& name);
    virtual CStream* openFile(const CUtiString& name, TOpenType ot);
    virtual void getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result, bool relativ, bool searchdirs);
};

class CMemorySystemArchiv : public CBaseArchiv
{
protected:
    CMemoryDirectory* m_rootdir;
    CMemoryDirectory* GetDir(const CUtiString &_Path);
    CMemoryDataFile* GetFile(const CUtiString &Path);
   public:
      CMemorySystemArchiv(CMemoryDirectory* rootdir);
      ~CMemorySystemArchiv();
      void init(const CUtiString& _name);
      virtual void deleteFile(const CUtiString& str);
      //virtual int getDirectorySize(const CUtiString& name);
      virtual bool fileExists(const CUtiString& name);
      virtual UTI_INT64 fileSize(const CUtiString& name);
      virtual CStream* openFile(const CUtiString& name, TOpenType ot);
      virtual void getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set<CFileSystem_Info> & result, bool relativ, bool searchdirs);
};


#endif
