/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "UtiCacheStructures.h"
#include "ziparchiv.h"

#include "utisystem.h"

CStream* zip_streamCompress(CStream* compress)
{
   CZipMemoryStream* zipstream;

            zipstream = new CZipMemoryStream(compress);

            zipstream->beginWrite();
   return zipstream;
}

void zip_streamEndCompress(CStream* compress)
{
   ((CZipMemoryStream*)compress)->endWrite();
}

CStream* zip_streamUnCompress(CStream* compress)
{
    CZipMemoryStream* zipstream;
            zipstream = new CZipMemoryStream(compress);
            zipstream->beginRead();
	    return zipstream;
}

void zip_streamEndUnCompress(CStream* compress)
{
    ((CZipMemoryStream*)compress)->endRead();
}


CGeneralResourceManager::CGeneralResourceManager(CResourceInterface* resint): ReferenceUtiOB()
{
  m_resourceinterface = resint;
}

CGeneralResourceManager::~CGeneralResourceManager()
{

}

void CGeneralResourceManager::addJobQueue(CBaseJobQueue* jobqueue)
{  CHECKEXISTANCE;
  m_jobqueue.push_back(jobqueue);
}


void CGeneralResourceManager::notifyDestroy(ReferenceUtiOB* res)
{
CHECKEXISTANCEOF(res);
    setState(res, resNotLoaded, -1, -1, NULL, true);
    if (m_interestedcount.find(res) != m_interestedcount.end())
    {
        m_interestedcount.erase(res);
    }
    if (m_interested.find(res) != m_interested.end())
    {
        m_interested.erase(res);
    }
    if (m_loading.find(res) != m_loading.end())
    {
        m_loading.erase(res);
    }
    if (m_freeing.find(res) != m_freeing.end())
    {
        m_freeing.erase(res);
    }
    if (m_normal.find(res) != m_normal.end())
    {
        m_normal.erase(res);
    }
    if (m_notinterested.find(res) != m_notinterested.end())
    {
        m_notinterested.erase(res);
    }

    if (m_objectstates.find(res) != m_objectstates.end())
    {
        m_objectstates.erase(res);
    }
   assert(m_interestedcount.find(res) == m_interestedcount.end());
   assert(m_interested.find(res) == m_interested.end());
   assert(m_loading.find(res) == m_loading.end());
   assert(m_freeing.find(res) == m_freeing.end());
   assert(m_normal.find(res) == m_normal.end());
   assert(m_notinterested.find(res) == m_notinterested.end());
   assert(m_objectstates.find(res) == m_objectstates.end());
}

int CGeneralResourceManager::getInterest(ReferenceUtiOB* res)
{  CHECKEXISTANCE;CHECKEXISTANCEOF(res);
  std::map<ReferenceUtiOB*, SResourceData>::iterator it = m_objectstates.find(res);
   if (it == m_objectstates.end()) {
     return 0;
   };
   return it->second.m_interest;
}

SResourceDataInfo CGeneralResourceManager::getResourceInfo(ReferenceUtiOB* res)
{  CHECKEXISTANCE;
    std::map<ReferenceUtiOB*, SResourceData>::iterator it = m_objectstates.find(res);
     if (it == m_objectstates.end()) {
       SResourceDataInfo info;
       info.m_interest = -255;
       info.m_priority = 0;
       info.m_state = resDoNotCare;
       return info;
     };
     SResourceDataInfo info;
     info.m_interest = it->second.m_interest;
     info.m_priority = it->second.m_priority;
     info.m_state = it->second.m_state;
     return info;
}

void CGeneralResourceManager::increaseInterest(ReferenceUtiOB* res, int priority)
{  CHECKEXISTANCE;
   int interest = getInterest(res);

   if (m_interestedcount.find(res) == m_interestedcount.end()) {
       m_interestedcount.insert(std::make_pair(res, 1));
   } else {
       int v = m_interestedcount[res];
       assert(v == interest);
       m_interestedcount.erase(m_interestedcount.find(res));
       m_interestedcount.insert(std::make_pair(res, interest+1));
   }

   if (interest == 0) {
     setState(res, resInterested, -1, 1, NULL, false);
    // dealInterested(res);
   } else {
     setState(res, resDoNotCare, -1, interest+1, NULL, false);
   };
}



void CGeneralResourceManager::decreaseInterest(ReferenceUtiOB* res, int priority)
{  CHECKEXISTANCE;
   int interest = getInterest(res);
   assert(m_interestedcount.find(res) != m_interestedcount.end());
   int v = m_interestedcount[res];
   assert(v == interest);
   m_interestedcount.erase(m_interestedcount.find(res));
   if (interest-1 != 0) {
   m_interestedcount.insert(std::make_pair(res, interest-1));
   }


   if (interest > 1) {

     setState(res, resDoNotCare, -1, interest-1, NULL, false);
   } else {
       TResourceState s = getState(res);
       assert(interest == 1);
       if (s != resLoading && s != resFreeing && s != resNotLoaded) {
     setState(res, resNotInterested, -1, interest-1, NULL, false);
       } else  {
           setState(res, resDoNotCare, -1, interest-1, NULL, false);
       }
    //dealFreeing(res);
   };
}

ReferenceUtiOB* CGeneralResourceManager::getResource(ReferenceUtiOB* res)
{  CHECKEXISTANCE;
  return m_resourceinterface->getResource(res);
}

void CGeneralResourceManager::invalidResource(ReferenceUtiOB* res)
{  CHECKEXISTANCE;
  /*TResourceState r =*/ getState(res);
}

TResourceState CGeneralResourceManager::getState(ReferenceUtiOB* obj)
{  CHECKEXISTANCE;
   std::map<ReferenceUtiOB*, SResourceData>::iterator it = m_objectstates.find(obj);
   if (it == m_objectstates.end()) {
     return resNotLoaded;
   };
   return it->second.m_state;
}


void CGeneralResourceManager::setState(ReferenceUtiOB* obj, TResourceState state, int priority, int interest, CBaseJob* job, bool deletejob)
{  CHECKEXISTANCE;
  TResourceState st;
  std::map<ReferenceUtiOB*, SResourceData>::iterator it = m_objectstates.find(obj);
   if (it == m_objectstates.end()) {
     st = resNotLoaded;
   } else {
      st = it->second.m_state;
   }
   /* if (st == state && interest > 0 && priority > 0) {

      return;
    };*/
    switch (st) {
      case resNotLoaded: {

      }; break;
      case resInterested: {
	utiassert(m_interested.find(obj) != m_interested.end());
        m_resourceinterface->checkType(obj);
	m_interested.erase(obj);
      }; break;
      case resLoading: {
	utiassert(m_loading.find(obj) != m_loading.end());
        m_resourceinterface->checkType(obj);
	m_loading.erase(obj);
      }; break;
      case resNormal: {
	utiassert(m_normal.find(obj) != m_normal.end());
        m_resourceinterface->checkType(obj);
	m_normal.erase(obj);
      }; break;
      case resNotInterested: {
	utiassert(m_notinterested.find(obj) != m_notinterested.end());
        m_resourceinterface->checkType(obj);
	m_notinterested.erase(obj);
      }; break;
      case resFreeing: {
	utiassert(m_freeing.find(obj) != m_freeing.end());
        m_resourceinterface->checkType(obj);
	m_freeing.erase(obj);
      }; break;
      default: {
	utiassert(false);
      }; break;

    };

    if (state == resDoNotCare) {
     state = st;
    }

    SResourceData data;
    if (it != m_objectstates.end()) {
      data = it->second;
      m_objectstates.erase(it);
    } else {
     data.m_priority = 1;
     data.m_interest = 0;

    }
    if (priority >= 0) {
      data.m_priority = priority;
    };
    if (interest >= 0) {
      data.m_interest = interest;
    };
    if (deletejob)  {
    data.m_job = NULL;
    }
    if (job) {
      data.m_job = job;
    };
    data.m_state = state;
    if (state != resNotLoaded) {
      m_objectstates.insert(std::make_pair(obj, data));
    };
   switch (state) {
      case resNotLoaded: {

      }; break;
      case resInterested: {
        utiassert(m_interested.find(obj) == m_interested.end());
        m_resourceinterface->checkType(obj);
	m_interested.insert(obj);
      }; break;
      case resLoading: {
	utiassert(m_loading.find(obj) == m_loading.end());
        m_resourceinterface->checkType(obj);
	m_loading.insert(obj);
      }; break;
      case resNormal: {
	utiassert(m_normal.find(obj) == m_normal.end());
        m_resourceinterface->checkType(obj);
	m_normal.insert(obj);
      }; break;
      case resNotInterested: {
	utiassert(m_notinterested.find(obj) == m_notinterested.end());
        m_resourceinterface->checkType(obj);
	m_notinterested.insert(obj);
      }; break;
      case resFreeing: {
	utiassert(m_freeing.find(obj) == m_freeing.end());
        m_resourceinterface->checkType(obj);
	m_freeing.insert(obj);
      }; break;
      default: {
	utiassert(false);
      }; break;

    };


}

bool CGeneralResourceManager::isFinished()
{
    for ( std::map<ReferenceUtiOB*, SResourceData>::iterator objinfo = m_objectstates.begin(); objinfo != m_objectstates.end(); objinfo++) {
       if (objinfo->second.m_job.get() != NULL) {
          if (objinfo->second.m_job->isFinished() == false) {
       return false;
          }
       }
    };
    return true;
}

void CGeneralResourceManager::waitForFinish()
{  CHECKEXISTANCE;
   bool doexit;
  do {
    doexit = true;
    for ( std::map<ReferenceUtiOB*, SResourceData>::iterator objinfo = m_objectstates.begin(); objinfo != m_objectstates.end(); objinfo++) {
       if (objinfo->second.m_job.get() != NULL) {
          if (objinfo->second.m_job->isFinished() == false) {
	    doexit = false;
	    break;
          } else {
	    objinfo->second.m_job->finaliseMainThread();
        objinfo->second.m_job = NULL;
	    setState(objinfo->first, resDoNotCare, -1, -1, NULL, true);
        //Iterator is now invalid
        doexit = false;
	    break;
	  }
       }
    };
     UtiFramework::utisleep(10);
  } while (!doexit);
  //m_jobqueue->waitForFinish();
}

void CGeneralResourceManager::dealNotInterested(ReferenceUtiOB* ref)
{  CHECKEXISTANCE;
    EASY_FUNCTION(profiler::colors::Blue);
  utiassert(m_notinterested.find(ref) != m_notinterested.end());
  CBaseJob* job = m_resourceinterface->destroyResource(ref);
  if (job) {
    setState(ref, resFreeing, -1, -1, job,true);
     m_jobqueue[urandom(m_jobqueue.size())].addJob(job);
  } else {
    setState(ref, resNotLoaded, -1, -1, NULL,true);
  };
}

void CGeneralResourceManager::dealInterested(ReferenceUtiOB* ref)
{  CHECKEXISTANCE;
    EASY_FUNCTION(profiler::colors::Blue);
   utiassert(m_interested.find(ref) != m_interested.end() || m_normal.find(ref) != m_normal.end());

  CBaseJob* job = m_resourceinterface->loadResource(ref);
  m_resourceinterface->resetValid(ref);
  if (job) {
    setState(ref, resLoading, -1, -1, job,true);
    m_jobqueue[urandom(m_jobqueue.size())].addJob(job);
  } else {
    setState(ref, resNormal, -1, -1, NULL,true);
  };
}

void CGeneralResourceManager::think(TIME time)
{  CHECKEXISTANCE;

    EASY_FUNCTION(profiler::colors::Blue);
    std::set<ReferenceUtiOB*>::iterator it, next;
     std::map<ReferenceUtiOB*, SResourceData>::iterator objinfo;
    //Delete jobs

    it = m_notinterested.begin();
    while (it != m_notinterested.end()) {
      next = it;
      next++;
      ReferenceUtiOB* ref = *it;
      objinfo = m_objectstates.find(ref);
      utiassert(objinfo != m_objectstates.end());
      if (objinfo->second.m_job.get() != NULL && objinfo->second.m_job->isFinished()) {
         objinfo->second.m_job->finaliseMainThread();
          setState(ref, resDoNotCare, -1, -1, NULL, true);

      };
      it = next;
    };


    it = m_interested.begin();
    while (it != m_interested.end()) {
      next = it;
      next++;
      ReferenceUtiOB* ref = *it;
      objinfo = m_objectstates.find(ref);
      utiassert(objinfo != m_objectstates.end());
      if (objinfo->second.m_job.get() != NULL && objinfo->second.m_job->isFinished()) {
         objinfo->second.m_job->finaliseMainThread();
          setState(ref, resDoNotCare, -1, -1, NULL, true);

      };
      it = next;
    };



   if (m_notinterested.size() > m_interested.size()) {
     /*if (m_notinterested.size() != 0) {
         ReferenceUtiOB* ref = *m_notinterested.begin();
         if (m_resourceinterface->isReadyUnload(ref)) {
              dealNotInterested(ref);
         }
     };*/
     for (it = m_notinterested.begin(); it != m_notinterested.end(); it++) {
          ReferenceUtiOB* ref = *it;
          if (m_resourceinterface->isReadyUnload(ref)) {
               dealNotInterested(ref);
               break;
          }
     }
   } else {
     for (it = m_interested.begin(); it != m_interested.end(); it++) {
         ReferenceUtiOB* ref = *it;CHECKEXISTANCEOF(ref);

	 if (m_resourceinterface->isReady(ref)) {
	 dealInterested(ref);
	 break;
	 };
     };
   };

   it = m_loading.begin();
   while (it != m_loading.end()) {
     next = it;
     next++;
     ReferenceUtiOB* ref = *it;
     objinfo = m_objectstates.find(ref);
     utiassert(objinfo != m_objectstates.end());
     if (objinfo->second.m_job.get() == NULL || objinfo->second.m_job->isFinished()) {
      if (objinfo->second.m_job.get()) objinfo->second.m_job->finaliseMainThread();
       if (objinfo->second.m_interest != 0) {
           setState(ref, resNormal, -1, -1, NULL, true);
       }else {
	 setState(ref, resNotInterested, -1, -1, NULL, true);
       };
     };
     it = next;
   };

   it = m_freeing.begin();
   while (it != m_freeing.end()) {
     next = it;
     next++;
     ReferenceUtiOB* ref = *it;
     objinfo = m_objectstates.find(ref);
     utiassert(objinfo != m_objectstates.end());
     if (objinfo->second.m_job.get() == NULL || objinfo->second.m_job->isFinished()) {
       if (objinfo->second.m_job.get()) objinfo->second.m_job->finaliseMainThread();
        if (objinfo->second.m_interest == 0) {
           setState(ref, resNotLoaded, -1, -1, NULL, true);
	} else {
	   setState(ref, resInterested, -1, -1, NULL, true);
	}
     };
     it = next;
   };

   for (it = m_normal.begin(); it != m_normal.end(); it++) {
     if (m_resourceinterface->isValid(*it) == false && m_resourceinterface->isReady(*it)) {
       dealInterested(*it);
       break;
     }
   }
}

void CGeneralResourceManager::setPriority(ReferenceUtiOB* res, int priority)
{  CHECKEXISTANCE;
  setState(res, resDoNotCare, priority, -1, NULL, false);
}

void CGeneralResourceManager::updatedResource(ReferenceUtiOB* obj)
{

}


C2DResourceController::C2DResourceController(CGeneralResourceManager* resmanager, C2DResourceInterface* p_interface)
{
  m_resmanager =resmanager;
  m_interface = p_interface;
}

CUtiString SResourceDataInfo::toString()
{
    return UTISTR("(")+resourceStateToString(m_state)+UTISTR(", ")+m_interest+UTISTR(", ")+m_priority+UTISTR(")");
}


CUtiString resourceStateToString(TResourceState res)
{
    switch (res) {
    case resNotLoaded: {
        return UTISTR("resNotLoaded");
    }; break;
    case resInterested: {
        return UTISTR("resInterested");

    }; break;
    case resLoading: {
        return UTISTR("resLoading");

    }; break;
    case resNormal: {
        return UTISTR("resNormal");

    }; break;
    case resNotInterested: {
        return UTISTR("resNotInterested");

    }; break;
    case resFreeing: {
        return UTISTR("resFreeing");

    }; break;
    case resDoNotCare: {
        return UTISTR("resDoNotCare");

    }; break;
    default: {
        assert(false);

    }; break;
    }
	return UTISTR("Unknown");
}

C2DResourceInterface* C2DResourceController::getInterface()
{
    CHECKEXISTANCE;
    return m_interface.get();
};

SResourceDataInfo C2DResourceController::getResourceInfo(const CIVector2& v)
{
    CHECKEXISTANCE;
      ReferenceUtiOB* ref = m_interface->getResourceDescription(v);
      return m_resmanager->getResourceInfo(ref);
}

ReferenceUtiOB* C2DResourceController::getResource(const CIVector2& v)
{  CHECKEXISTANCE;
  ReferenceUtiOB* ref = m_interface->getResourceDescription(v);
  return m_resmanager->getResource(ref);
}

void C2DResourceController::modifyInterest(const CIntRect& rect, int p)
{  CHECKEXISTANCE;
  utiassert(abs(p) == 1);
  for (int j = rect.m_min.y;j < rect.m_max.y; j++) {
     for (int i = rect.m_min.x;i < rect.m_max.x; i++) {
         ReferenceUtiOB* ref = m_interface->getResourceDescription(CIVector2(i,j));
	 if (p > 0) {
             m_resmanager->increaseInterest(ref);
	 } else {
	   m_resmanager->decreaseInterest(ref);
	 }
     };
  };
}

int C2DResourceController::getInterest(const CIVector2& node)
{

    ReferenceUtiOB* ref = m_interface->getResourceDescription(node);
    return m_resmanager->getInterest(ref);
}

CUtiString getRandomString(const std::vector<CUtiString>& str)
{
    int v = urandom(str.size());
    return str.at(v);
}

