#pragma once
#ifndef __UTICACHESTRUCTURES__
#define __UTICACHESTRUCTURES__
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Uti3DCore.h"
#include "Uti3DThreads.h"
#include "utiarray.h"
#include "utimathbox.h"
#include "utistream.h"
#include <map>
#include <vector>

template <class T>
class Array3DY
{
   protected:
      bool dynamic;
   public:
      class Array3DIterator {
      protected:
          T* m_array;

          int m_newline;
          int m_pagesize;
#ifdef ITERATORDEBUG
          CIVector3 m_currentposition;
          CIVector3 m_size;
#endif
      public:
          inline Array3DIterator(const Array3DIterator& v)
          {
              m_array = v.m_array;
              m_newline = v.m_newline;
              m_pagesize = v.m_pagesize;
#ifdef ITERATORDEBUG
              m_currentposition = v.m_currentposition;
              m_size = v.m_size;
#endif
          }
          inline Array3DIterator()
          {
              m_array = NULL;
              m_newline = 0;
              m_pagesize = 0;
#ifdef ITERATORDEBUG
              m_currentposition = CIVector3(0,0,0);
              m_size = CIVector3(0,0,0);
#endif
          }

          inline explicit Array3DIterator(T* a,  int newline, int pagesize, const CIVector3& size,const CIVector3& start)
          {
              m_array = a;
              m_newline = newline;
              m_pagesize = pagesize;
#ifdef ITERATORDEBUG
              m_currentposition = start;
              m_size = size;
#endif
          }
          inline Array3DIterator& operator=(const Array3DIterator& v)
          {
              m_array = v.m_array;
              m_newline = v.m_newline;
              m_pagesize = v.m_pagesize;
#ifdef ITERATORDEBUG
              m_currentposition = v.m_currentposition;
              m_size = v.m_size;
#endif
              return *this;
          }

          inline const T& get() const  {
#ifdef ITERATORDEBUG
assert(m_currentposition.x >= 0 && m_currentposition.x < m_size.x);
assert(m_currentposition.y >= 0 && m_currentposition.y < m_size.y);
assert(m_currentposition.z >= 0 && m_currentposition.z < m_size.z);
#endif
              return *m_array;
          }
          inline T& getPointer()   {
#ifdef ITERATORDEBUG
assert(m_currentposition.x >= 0 && m_currentposition.x < m_size.x);
assert(m_currentposition.y >= 0 && m_currentposition.y < m_size.y);
assert(m_currentposition.z >= 0 && m_currentposition.z < m_size.z);
#endif
              return *m_array;
          }
          inline void set(const T& v)
          {
#ifdef ITERATORDEBUG
assert(m_currentposition.x >= 0 && m_currentposition.x < m_size.x);
assert(m_currentposition.y >= 0 && m_currentposition.y < m_size.y);
assert(m_currentposition.z >= 0 && m_currentposition.z < m_size.z);
#endif
              *m_array = v;
          }
          inline void incX(int amount = 1) {
              m_array += amount*m_newline;
#ifdef ITERATORDEBUG
              m_currentposition.x += amount;

#endif
          }
          inline void incY(int amount = 1) {
              m_array += amount;
#ifdef ITERATORDEBUG
              m_currentposition.y += amount;

#endif
          }
          inline void incZ(int amount = 1) {
              m_array += amount*m_pagesize;
#ifdef ITERATORDEBUG
              m_currentposition.z+= amount;

#endif
          }
		  inline void inc_x(int amount = 1) {
			  m_array += amount*m_newline;
#ifdef ITERATORDEBUG
			  m_currentposition.x += amount;

#endif
		  }
		  inline void inc_y(int amount = 1) {
			  m_array += amount;
#ifdef ITERATORDEBUG
			  m_currentposition.y += amount;

#endif
		  }
		  inline void inc_z(int amount = 1) {
			  m_array += amount*m_pagesize;
#ifdef ITERATORDEBUG
			  m_currentposition.z += amount;

#endif
		  }
          inline void move(const CIVector3& v)
          {
              m_array += v.x*m_newline+v.y+ v.z*m_pagesize;
#ifdef ITERATORDEBUG
              m_currentposition+= v;

#endif

          }
#ifdef ITERATORDEBUG
          inline bool checkDebugPosition(const CIVector3& v)
          {

              return v == m_currentposition;
          }
#else
          inline bool checkDebugPosition(const CIVector3& v)
          {

              return true;
          }
#endif
      };
      T* array;
      int array_width;
      int array_height;
      int array_depth;
      int array_pagesize;
      int array_newline;
      typedef Array3DIterator iterator;

      inline Array3DIterator getIterator()
      {
          return Array3DIterator(array, array_newline, array_pagesize, CIVector3(array_width, array_height, array_depth), CIVector3(0,0,0));
      }
      inline Array3DIterator getConst_Iterator() const
      {
          return Array3DIterator(const_cast<T*>(array), array_newline, array_pagesize, CIVector3(array_width, array_height, array_depth), CIVector3(0,0,0));
      }

      inline Array3DIterator getIterator(const CIVector3& pos)
      {
          return Array3DIterator(&(*this)(pos), array_newline, array_pagesize, CIVector3(array_width, array_height, array_depth), pos);
      }
      inline Array3DIterator getConst_Iterator(const CIVector3& pos) const
      {
          return Array3DIterator(const_cast<T*>( &(*this)(pos)), array_newline, array_pagesize, CIVector3(array_width, array_height, array_depth), pos);
      }

      bool isDynamic() const
      {
         return dynamic;
      }
      inline int width() const
      {
         return array_width;
      }
      inline int height() const
      {
         return array_height;
      }
      inline int depth() const
      {
	return array_depth;
      };
      inline CIVector3 size() const {
	return CIVector3(array_width, array_height, array_depth);
      };
      Array3DY(int w, int h, int d)
      {
         dynamic = true;
         array_width = w;
         array_height = h;
	 array_depth = d;
         array_newline = h;
	 array_pagesize = array_newline*array_width;
         array = new T[array_pagesize*array_depth];
      }
      Array3DY(T* ar, int w, int h, int d)
      {
         dynamic = false;
         array_width = w;
         array_height = h;
	 array_depth = d;
         array_newline = h;
	 array_pagesize = array_newline*array_width;
         array = ar;
      }

      Array3DY()
      {
         array_width = 0;
         array_height = 0;
	 array_depth = 0;
         array_newline = 0;
	 array_pagesize = 0;

         array = NULL;
	 dynamic = false;
      };
      Array3DY(const Array3DY &narray);
      ~Array3DY()
      {
         if (dynamic)
            delete [] array;
      }
      size_t getDataSize()
      {
	return (size_t)array_width*(size_t)array_height*(size_t)array_depth*sizeof(T);
      };
      inline CIntBox getBox() const
      {
          return CIntBox(0,0,0, array_width, array_height, array_depth);
      }
      void setSize(const CIVector3& v) {
	setSize(v.x, v.y, v.z);
      };
      void setSize(int w, int h, int d) {
          if (array != NULL) {
              delete [] array;
              array = NULL;
          }
         utiassert(array == NULL);
         dynamic = true;
         array_width = w;
         array_height = h;
	 array_depth = d;
         array_newline = h;
	 array_pagesize = array_newline*array_width;
         array = new T[array_pagesize*array_depth];
      };
      /*void ReferenzTo(const Array2D &a)
      {
         dynamic = false;
         array_size = a.array_size;
         array = a.array;
      }*/
	  template<class UnaryPredicate> bool isEqualCompare(const CIntBox& box, const UnaryPredicate& v) const
	  {
		  ENUMBOX_BEGIN(box, pos)
			  if (!v(operator()(pos))) {
				  return false;
			  }
		  ENUMBOX_END
			return true;
	  }
      inline T &operator()(unsigned int x, unsigned int y, unsigned int z)
      {
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
	 utiassert(z >= 0 && z < (unsigned int)array_depth);
         return array[x*array_newline+y+z*array_pagesize];

      }
      inline const T &operator()(unsigned int x, unsigned int y, unsigned int z) const
      {
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
	 utiassert(z >= 0 && z < (unsigned int)array_depth);
         return array[x*array_newline+y+z*array_pagesize];

      }
      inline T &operator()(const CIVector3& v)
      {
         utiassert(array != NULL);
         utiassert(v.x >= 0 && v.x < array_width);
         utiassert(v.y >= 0 && v.y < array_height);
     utiassert(v.z >= 0 && v.z < array_depth);
         return array[v.x*array_newline+v.y+v.z*array_pagesize];

      }
      inline const T &operator()(const CIVector3& v) const
      {
         utiassert(array != NULL);
         utiassert(v.x >= 0 && v.x < array_width);
         utiassert(v.y >= 0 && v.y < array_height);
     utiassert(v.z >= 0 && v.z < array_depth);
         return array[v.x*array_newline+v.y+v.z*array_pagesize];

      }
      void fill(const T& value)
      {
	  int s = array_pagesize*array_depth;
          for (int i = 0; i < s; i++) array[i] = value;
       }
       void clear()
       {
	 if (array != NULL)
	 memset(array, 0, array_pagesize*array_depth*sizeof(T));
       };
       inline void set(unsigned int x, unsigned int y, unsigned int z, const T& v)
       {
          utiassert(array != NULL);
          utiassert(x >= 0 && x < (unsigned int)array_width);
          utiassert(y >= 0 && y < (unsigned int)array_height);
          utiassert(z >= 0 && z < (unsigned int)array_depth);
          array[x*array_newline+y+z*array_pagesize] = v;

       }
       inline T get(unsigned int x, unsigned int y, unsigned int z) const
       {
          utiassert(array != NULL);
          utiassert(x >= 0 && x < (unsigned int)array_width);
          utiassert(y >= 0 && y < (unsigned int)array_height);
          utiassert(z >= 0 && z < (unsigned int)array_depth);
          return array[x*array_newline+y+z*array_pagesize];

       }
       inline void set(const CIVector3& v, const T& value)
       {
          utiassert(array != NULL);
          utiassert(v.x >= 0 && v.x < array_width);
          utiassert(v.y >= 0 && v.y < array_height);
          utiassert(v.z >= 0 && v.z < array_depth);
          array[v.x*array_newline+v.y+v.z*array_pagesize] = value;

       }
       inline T get(const CIVector3& v) const
       {
          utiassert(array != NULL);
          utiassert(v.x >= 0 && v.x < array_width);
          utiassert(v.y >= 0 && v.y < array_height);
          utiassert(v.z >= 0 && v.z < array_depth);
          return array[v.x*array_newline+v.y+v.z*array_pagesize];

       }
};

template <class T>
      Array3DY<T>::Array3DY(const Array3DY &narray)
{
   dynamic = true;
   array_width = narray.width();
   array_height = narray.height();
   array_depth = narray.depth();
   array_newline = array_height;
   array_pagesize = array_newline*array_width;
   array = new T[array_depth*array_pagesize];
//   wcout << "Warnung: Array wird kopiert"<<endl;
   for (int k = 0; k < array_depth; k++)
   {

      for (int i = 0; i < array_width; i++)
      {
	for (int j = 0; j < array_height; j++)
        {
          (*this)(i, j, k) = narray(i, j, k);
        }

      }

   }

}

CStream* zip_streamCompress(CStream* compress);
void zip_streamEndCompress(CStream* compress);

CStream* zip_streamUnCompress(CStream* compress);
void zip_streamEndUnCompress(CStream* compress);


class CArray3DMask : public ReferenceUtiOB
{
public:
    virtual bool getValue(const CIVector3& v) const=0;
    virtual bool isBoxAllowed(const CIntBox& box)
    {
        ENUMBOX_BEGIN(box, pos)
              if (getValue(pos) == false)return false;

                ENUMBOX_END
                        return true;
    }
};

template<class T> class Array3DYContainer : public ReferenceUtiOB
{
public:
  uti_refptr<CMemoryStream> m_stream;
#ifndef NDEBUG
  int m_usagecount;

#endif
  inline void addUsageCount() {
#ifndef NDEBUG
  m_usagecount++;
#endif
};
inline void decUsageCount() {
#ifndef NDEBUG
  m_usagecount--;
#endif
};
inline int getUsageCount()
{
#ifndef NDEBUG
   return m_usagecount;
#else
   return 0;
#endif
}
    explicit Array3DYContainer() {
      m_origin.set(0,0,0);
#ifndef NDEBUG
    m_usagecount = 0;
#endif

  };
  explicit Array3DYContainer(const CIVector3& size)
{
    m_origin.set(0,0,0);
    m_array.setSize(size);

#ifndef NDEBUG
    m_usagecount = 0;
#endif
};

explicit Array3DYContainer(const CIntBox& box)
{
  m_origin = box.m_min;
  m_array.setSize(box.size());

#ifndef NDEBUG
  m_usagecount = 0;
#endif
};
    CIVector3 m_origin;
  Array3DY<T> m_array;

     void setOrigin(const CIVector3& data)
    {
    m_origin = data;


    }

  inline T& getData(const CIVector3& v)
  {
      return m_array(v-m_origin);
  }
  inline bool contains(const CIVector3& v)
  {
    CIntBox b;
    b.m_min = m_origin;
    b.m_max = b.m_min+m_array.size();
    return b.contains(v);
  }
  inline const T& getData(const CIVector3& v) const
  {
      return m_array(v-m_origin);
  }
  inline void setData(const CIVector3& v, const T& v2)
  {
      m_array(v-m_origin)  = v2;
  }
  void setBlock4(const CIVector3& voxelposition, const T& data)
    {

        CIVector3 v;
        v.x = (voxelposition.x-m_origin.x) >> 2;
        v.y = (voxelposition.y-m_origin.y) >> 2;
        v.z = (voxelposition.z-m_origin.z) >> 2;
        assert(v*4+m_origin == voxelposition);
        CIVector3 pos;
        for (pos.z = voxelposition.z-m_origin.z; pos.z < voxelposition.z+4-m_origin.z; pos.z++) {
            for (pos.x = voxelposition.x-m_origin.x; pos.x < voxelposition.x+4-m_origin.x; pos.x++) {

                for (pos.y = voxelposition.y-m_origin.y; pos.y < voxelposition.y+4-m_origin.y; pos.y++) {
                   m_array(pos) = data;
                }

            }
        }
    }
    inline void packFast(){}
  inline void pack() {}

  void drawBox(const CIntBox& box, const T& v1)
  {
     ENUMBOX_BEGIN(box, pos)
             setData(pos, v1);

             ENUMBOX_END
  }

  void drawBoxMask(const CIntBox& box, const T& v1, CArray3DMask* mask)
  {
      CIntBox bnew;
      CIntBox b;
      b.m_min = m_origin;
      b.m_max = b.m_min+m_array.size();
      bnew = b.intersection(box);
     ENUMBOX_BEGIN(bnew, pos)
               if (/*contains(pos) &&*/ mask->getValue(pos) == true) {
             setData(pos, v1);

     }
             ENUMBOX_END
  }

  void drawBoxClipped(const CIntBox& box, const T& v1)
  {
      CIntBox bnew;
      CIntBox b;
      b.m_min = m_origin;
      b.m_max = b.m_min+m_array.size();
      bnew = b.intersection(box);
     ENUMBOX_BEGIN(bnew, pos)
           //  if (contains(pos)) {
             setData(pos, v1);
    // }

             ENUMBOX_END
  }

  void drawBlock(const CIVector3& box, const T& v1)
  {
      CIntBox bnew;
      CIntBox b;
      b.m_min = m_origin;
      b.m_max = b.m_min+m_array.size();
     if (b.contains(box)) {
         setData(box, v1);
     }
  }


  CIntBox getBox() const
  {
      CIntBox b;
      b.m_min = m_origin;
      b.m_max = b.m_min+m_array.size();
      return b;

  }
  void setBox(const CIntBox& b)
  {
      m_origin = b.m_min;
      m_array.setSize(b.size());
  }
  bool isEqual(const CIntBox& box, const T& v) const
  {
      ENUMBOX_BEGIN(box, pos)
           if (!(getData(pos) == v)) {
       return false;
      }
              ENUMBOX_END
      return true;
  }
  template<class UnaryPredicate> bool isEqualCompare(const CIntBox& box, const UnaryPredicate& v) const
  {
	  ENUMBOX_BEGIN(box, pos)
		  if (!v(getData(pos))) {
			  return false;
		  }
	  ENUMBOX_END
		  return true;
  }

  void copyFrom(const Array3DYContainer<T>& source)
  {
      CIntBox b = getBox();
      ENUMBOX_BEGIN(b, position)
              setData(position, source.getData(position));
      ENUMBOX_END;
  }

  inline bool isCompressed()
  {
      return m_stream.get() != NULL;
  }

  void compress()
  {
    utiassert(m_stream.get() == NULL);
    utiassert(m_array.width() != 0);
    m_stream=new CMemoryStream(true);
    uti_refptr<CStream> zipstream;
    zipstream = (zip_streamCompress(m_stream.get()));
    zipstream->writeInt32(m_origin.x);
    zipstream->writeInt32(m_origin.y);
    zipstream->writeInt32(m_origin.z);
    zipstream->writeInt32(m_array.width());
    zipstream->writeInt32(m_array.height());
    zipstream->writeInt32(m_array.depth());
    zipstream->WriteData(m_array.array, m_array.getDataSize());
    zip_streamEndCompress(zipstream.get());
    m_array.setSize(0,0,0);
  }
  void uncompress()
  {
    utiassert(m_stream.get() != NULL);
     utiassert(m_array.width() == 0);
    m_stream->setPosition(0, CStream::START);
    m_stream->setAllowWrite(false);
    uti_refptr<CStream> zipstream;
    zipstream = (zip_streamUnCompress(m_stream.get()));
    m_origin.x = zipstream->getInt32();
    m_origin.y = zipstream->getInt32();
    m_origin.z = zipstream->getInt32();
    int w = zipstream->getInt32();
    int h = zipstream->getInt32();
    int d = zipstream->getInt32();
    m_array.setSize(w,h,d);
    zipstream->ReadData(m_array.array, m_array.getDataSize());
    zip_streamEndUnCompress(zipstream.get());
    m_stream = NULL;
  }

};


class Array3DYBin
{
   public:
      unsigned char* array;
      int dsize;
      int array_width;
      int array_height;
      int array_depth;
      int array_pagesize;
      int array_newline;

      inline int width() const
      {
         return array_width;
      }
      inline int height() const
      {
         return array_height;
      }
      inline int depth() const
      {
	return array_depth;
      };
      inline CIVector3 size() const {
	return CIVector3(array_width, array_height, array_depth);
      };
      inline int datasize()
      {
	return ((array_pagesize*array_depth) >> 3) +1;
      };
      Array3DYBin(int w, int h, int d)
      {

         array_width = w;
         array_height = h;
	 array_depth = d;
         array_newline = h;
	 array_pagesize = array_newline*array_width;
         dsize = datasize();
	 array = new unsigned char[dsize];

      }

      Array3DYBin()
      {
         array_width = 0;
         array_height = 0;
	 array_depth = 0;
         array_newline = 0;
	 array_pagesize = 0;
	 dsize =0;

         array = NULL;
      };

      ~Array3DYBin()
      {
         if (array)
            delete [] array;
      }
      void setSize(const CIVector3& v) {
	setSize(v.x, v.y, v.z);
      };
      void setSize(int w, int h, int d) {
          if (array_width == w && array_height == h && array_depth == d) {return;};
	 if (array)
            delete [] array;

         array_width = w;
         array_height = h;
	 array_depth = d;
         array_newline = h;
	 array_pagesize = array_newline*array_width;
           dsize = datasize();
	 array = new unsigned char[dsize];
      };
      /*void ReferenzTo(const Array2D &a)
      {
         dynamic = false;
         array_size = a.array_size;
         array = a.array;
      }*/
      inline void set(unsigned int x, unsigned int y, unsigned int z, bool b)
      {
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
	 utiassert(z >= 0 && z < (unsigned int)array_depth);
	 int pos = x*array_newline+y+z*array_pagesize;
	 int bin = pos& 7;
	 pos = pos >> 3;
	 utiassert(pos >= 0 && pos < dsize);
	 if (b) {
	   array[pos] |= 1 << bin;
	 } else {
	   array[pos] &= ~(1 << bin);
	 };

      }
      inline bool operator()(unsigned int x, unsigned int y, unsigned int z) const
      {
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
	 utiassert(z >= 0 && z < (unsigned int)array_depth);
	int pos = x*array_newline+y+z*array_pagesize;
	 int bin = pos& 7;
	 pos = pos >> 3;
	 utiassert(pos >= 0 && pos < dsize);
         return ((array[pos] >> bin) & 1) != 0;

      }
      inline void set(const CIVector3& v, bool b)
      {
         utiassert(array != NULL);
         utiassert(v.x >= 0 && v.x < (int)array_width);
         utiassert(v.y >= 0 && v.y < (int)array_height);
	 utiassert(v.z >= 0 && v.z < (int)array_depth);
         int pos = v.x*array_newline+v.y+v.z*array_pagesize;
	 int bin = pos& 7;
	 pos = pos >> 3;
	 utiassert(pos >= 0 && pos < dsize);
	 if (b) {
	   array[pos] |= 1 << bin;
	 } else {
	   array[pos] &= ~(1 << bin);
	 };

      }

      inline bool getAndSet(const CIVector3& v, bool b)
      {
         utiassert(array != NULL);
         utiassert(v.x >= 0 && v.x < (int)array_width);
         utiassert(v.y >= 0 && v.y < (int)array_height);
     utiassert(v.z >= 0 && v.z < (int)array_depth);
         int pos = v.x*array_newline+v.y+v.z*array_pagesize;
     int bin = pos& 7;
     pos = pos >> 3;
     utiassert(pos >= 0 && pos < dsize);
     bool res =((array[pos] >> bin) & 1) != 0;
     if (b) {
       array[pos] |= 1 << bin;
     } else {
       array[pos] &= ~(1 << bin);
     };
     return res;

      }
      inline bool operator()(const CIVector3& v) const
      {
         utiassert(array != NULL);
         utiassert(v.x >= 0 && v.x < (int)array_width);
         utiassert(v.y >= 0 && v.y < (int)array_height);
	 utiassert(v.z >= 0 && v.z < (int)array_depth);
         int pos = v.x*array_newline+v.y+v.z*array_pagesize;
	 int bin = pos& 7;
	 pos = pos >> 3;
	 utiassert(pos >= 0 && pos < dsize);
         return ((array[pos] >> bin) & 1) != 0;

      }
      void fill(const bool& value)
      {

	  unsigned char v = value ? 255 : 0;
          for (int i = 0; i < dsize; i++) array[i] = v;
       }
       void clear()
       {
	 fill(false);
       };
};

template<class X, bool active=true> class uti_usageptr {
public:
	typedef X element_type;
	inline explicit uti_usageptr(X* p = NULL, bool increment=true) throw() : ptr(p) { if (increment && ptr && active) ptr->addUsage();};
	template<class Y> inline uti_usageptr(const uti_usageptr<Y>& a) throw() : ptr(a.get())  { if(ptr&& active) ptr->addUsage();};
	inline uti_usageptr(const uti_usageptr& a) throw() : ptr(a.ptr)  { if(ptr&& active) ptr->addUsage();};
	inline ~uti_usageptr() throw() {/*if (ptr) ptr->decRef();*/reset(NULL);};
	template<class Y> inline uti_usageptr& operator=(const uti_usageptr<Y>& a) throw()
	{
        reset(a.get());
        return *this;
	};
	inline bool operator==(const X* other) const throw()
	{
       return ptr == other;
	}
	inline uti_usageptr& operator=(const uti_usageptr& a) throw()
	{
        reset(a.get());
        return *this;
	};
	inline bool operator==(const uti_usageptr<X>& other) const throw()
	{
       return ptr == other.get();
	}
	/*inline uti_usageptr<X>& operator=(const uti_usageptr<X>& a) throw()
	{
        reset(a.get());
        return *this;
	}*/
    inline uti_usageptr& operator=(X* a) throw()
	{
        reset(a);
        return *this;
	};
	inline bool operator<(const uti_usageptr& a) const throw()
	{
	  return (ptr-a.ptr) < 0;
	}
	inline bool operator>(const uti_usageptr& a)const throw()
	{
	  return (ptr-a.ptr) > 0;
	}
	inline X& operator*() const throw() {return *get();};
	inline X* operator->() const throw() {return get();};
	inline X* get() const throw() {return ptr;};
	inline void reset(X* p = NULL) throw() {if (get() != p) { X* a = ptr; ptr = p;if(!active) return; if (a) a->decUsage();  if (ptr) ptr->addUsage();};};
	template<class Y> operator uti_usageptr<Y>() throw()
	{ return uti_usageptr<Y>(get());};
	private:
		X* ptr;
};

class CResourceInterface : public ReferenceUtiOB
{
public:
  CResourceInterface(){};
  virtual bool isReady(ReferenceUtiOB* resource)=0;
  virtual bool isReadyUnload(ReferenceUtiOB* resource)=0;
  virtual CBaseJob* loadResource(ReferenceUtiOB* resource)=0;
  virtual ReferenceUtiOB* getResource(ReferenceUtiOB* resource)=0; //Liefert immer brauchbares Objectk
  virtual CBaseJob* destroyResource(ReferenceUtiOB* resource)=0;
  virtual void resetValid(ReferenceUtiOB* resource)=0;
  virtual bool isValid(ReferenceUtiOB* resource)=0;
  virtual void checkType(ReferenceUtiOB* resource)=0;
};

class C2DResourceInterface : public CResourceInterface
{
public:
  C2DResourceInterface(){};
  virtual CIntRect getGridSize()=0;
  virtual ReferenceUtiOB* getResourceDescription(const CIVector2& v)=0;
};

class CResourceUpdateListener : public ReferenceUtiOB
{
public:
  CResourceUpdateListener(){};
  virtual void updatedResource(ReferenceUtiOB* res)=0;
};

enum TResJobMode {rjmLoad, rjmDestroy};

struct SResCurrentJobs
{
  uti_refptr<ReferenceUtiOB> m_resource;
  uti_refptr<CBaseJob> m_job;
  TResJobMode m_mode;
};

enum TResourceState {resNotLoaded, resInterested, resLoading, resNormal, resNotInterested, resFreeing, resDoNotCare};
CUtiString resourceStateToString(TResourceState res);

struct SResourceData
{
  TResourceState m_state;
  uti_refptr<CBaseJob> m_job;
  int m_priority;
  int m_interest;
};

class CGeneralResourceManager;
class C2DResourceInterface;

struct SResourceDataInfo
{
  TResourceState m_state;
  int m_priority;
  int m_interest;
  CUtiString toString();
};

class C2DResourceController : public ReferenceUtiOB
{
protected:
  uti_refptr<CGeneralResourceManager> m_resmanager;
  uti_refptr<C2DResourceInterface> m_interface;
public:

    void modifyInterest(const CIntRect& rect, int p);
    int getInterest(const CIVector2& node);
    C2DResourceInterface* getInterface();
    ReferenceUtiOB* getResource(const CIVector2& v);
    SResourceDataInfo getResourceInfo(const CIVector2& v);
    C2DResourceController(CGeneralResourceManager* resmanager, C2DResourceInterface* p_interface);
};


//TODO: Low Res Resource, wo ist die Eigentliche Resource
class CGeneralResourceManager : public ReferenceUtiOB
{
protected:
  uti_refptr<CResourceInterface> m_resourceinterface;
  CReferenceList<CBaseJobQueue> m_jobqueue;
  std::set<ReferenceUtiOB*> m_interested;
  std::set<ReferenceUtiOB*> m_loading;
  std::set<ReferenceUtiOB*> m_freeing;
  std::set<ReferenceUtiOB*> m_normal;

  std::set<ReferenceUtiOB*> m_notinterested;

  std::map<ReferenceUtiOB*, int> m_interestedcount;

  std::map<ReferenceUtiOB*, SResourceData> m_objectstates;
  void setState(ReferenceUtiOB* obj, TResourceState state, int priority, int interest, CBaseJob* job, bool deleteJob);
  void updatedResource(ReferenceUtiOB*obj);
  void dealNotInterested(ReferenceUtiOB* ref);
  void dealInterested(ReferenceUtiOB* ref);
public:
  CGeneralResourceManager(CResourceInterface* resint);
  virtual ~CGeneralResourceManager();
  void addJobQueue(CBaseJobQueue* jobqueue);
  TResourceState getState(ReferenceUtiOB* obj);
  void waitForFinish();
  bool isFinished();
  inline int getInterestedCount() const {return m_interested.size();};
  SResourceDataInfo getResourceInfo(ReferenceUtiOB* res);
  void increaseInterest(ReferenceUtiOB* res, int priority=1);
  void decreaseInterest(ReferenceUtiOB* res, int priority=1);
  int getInterest(ReferenceUtiOB* res);
  void setPriority(ReferenceUtiOB* res, int priority);
  void think(TIME time);
  ReferenceUtiOB* getResource(ReferenceUtiOB* res);
  void invalidResource(ReferenceUtiOB* res);
  void notifyDestroy(ReferenceUtiOB* res);
};



template<class T> struct CRefResource
{
  uti_refptr<T> m_data;
  bool m_available;
  bool m_valid;
  CRefResource() {
    m_available = false;
    m_valid = true;
  };
  T* getData()
  {
    if (!m_available) return NULL;
    return m_data.get();
  };
  void clear()
  {
    m_available = false;
    m_data = NULL;
  };
  bool isValid()
  {
    return m_valid;
  };
  void resetValid()
  {
    m_valid = true;
  };
  void makeInvalid()
  {
    m_valid = false;
  };
};


CUtiString getRandomString(const std::vector<CUtiString>& str);

template<class T> struct uti_uniqueptr
{
private:
    T* m_ptr;
public:
    typedef T element_type;
    explicit uti_uniqueptr(T* p = 0) throw() : m_ptr(p) {}
    uti_uniqueptr(uti_uniqueptr<T>& a) throw() : m_ptr(a.release()) {}
    ~uti_uniqueptr() {if (m_ptr)delete m_ptr;}
    uti_uniqueptr<T>& operator=(uti_uniqueptr<T>& a) throw()
    {
        reset(a.release());
        return *this;
    }

    T* get() const throw() {return m_ptr;}
    T* release() throw()
    {
        T* temp = m_ptr;
        m_ptr = NULL;
        return temp;
    }
    T& operator*() const throw()
    {
        utiassert(m_ptr != NULL);
        return *m_ptr;
    }
    T* operator->() const throw()
    {
        utiassert(m_ptr != NULL);
        return m_ptr;
    }
    void reset(T* p = NULL) throw()
    {
        if (p != m_ptr) {
            if (m_ptr) {delete m_ptr;};
            m_ptr = p;
        }
    }

};

#endif
