#ifndef __UTIHASH__
#define __UTIHASH__
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Uti3DCore.h"
#include "utilinkedlist.h"
#include "utiarray.h"

#include<set>
#include<stdlib.h>
#include <map>
#include <functional>
#include <algorithm>

template <class T> struct CCacheHash_ConvertToInt
{
  public:
  inline int hashValue(const T& value)
  {
    return abs((int)((UTI_INT64) value));
  }
};

template <class T> struct CCacheHash_Method
{
  public:
  inline int hashValue(const T& value)
  {
    return value->hashCode();
  }
};
#ifndef NDEBUG
//Corrupts the heap
//#define UTI_SETDEBUG
#endif

template <class T, bool removefront, bool fastgrow=false> struct UtiArraySimpleSet
{
  protected:
    Array<T>& m_array;
    int m_position;

#ifdef UTI_SETDEBUG
    std::set<T> m_debug;
#endif
  public:
    explicit UtiArraySimpleSet(Array<T>& a) : m_array(a)
    {
      m_position =0;

      utiassert(0 < m_array.size());
    };

    int add(const T &el)
    {
        if (m_array.size() <= m_position) {
            if (fastgrow) {
                  m_array.setSizeCopyHighSpeed(m_array.size()*2);
            }else {
      m_array.setSizeCopy(m_array.size()*2);
            }
    };
    m_array[m_position] = el;
#ifdef UTI_SETDEBUG
     assert(m_debug.find(el) == m_debug.end());
    m_debug.insert(el);
#endif
    m_position++;
    return m_position-1;
    };
    T removeElement()
    {
        assert(m_position != 0);
        T res;
        if (removefront) {
            res = m_array[0];
            m_array[0] = m_array[m_position-1];
        } else {
            res = m_array[m_position-1];

        }
        m_position -= 1;
        if (m_position < m_array.size()/4 && m_array.size() >= 64) {
            assert(m_position <= m_array.size()/2);
            if (fastgrow) {
                  m_array.setSizeCopyHighSpeed(m_array.size()/2);
            }else {
      m_array.setSizeCopy(m_array.size()/2);
            }

        }
#ifdef UTI_SETDEBUG
        assert(m_debug.find(res) != m_debug.end());
        m_debug.erase(m_debug.find(res));
#endif


        return res;
    }

    /*virtual int grow()
    {
        if (m_array.size() <= m_position) {
            if (fastgrow) {
                  m_array.setSizeCopyHighSpeed(m_array.size()*2);
            }else {
      m_array.setSizeCopy(m_array.size()*2);
            }
    }
//	m_array[m_position] = el;
    m_position++;
    return m_position-1;
    };*/
    virtual int size() const {
#ifdef UTI_SETDEBUG
        assert(m_debug.size() == m_position);
#endif
        return m_position;
    }
    virtual void reset(int maxsize) {
        m_position = 0;
#ifdef UTI_SETDEBUG
        m_debug.clear();
#endif
        if (m_array.size() > maxsize) {
           // m_array.setSizeCopy(maxsize);
            if (fastgrow) {
                  m_array.setSizeCopyHighSpeed(maxsize);
            }else {
                 m_array.setSizeCopy(maxsize);
            }
        }
    }
};

template<class T> struct CUtiHeap_DefaultCompare
{
    static inline bool isGreater(const T& el1, const T& el2)
    {
        return el1 < el2;
    }
};

#define USE_CUtiHeap_STL

template<class T, bool fastgrow, class Compare=CUtiHeap_DefaultCompare<T> > class CUtiHeap {
protected:
  Compare m_compare;
  inline void getChildren(int element, int&left, int&right)
  {
       element = element*2;
       left = element+1;
       right = element+2;
   }
   inline int getLeftChild(int element)
   {
      return element*2+1;
   };
   inline int getRightChild(int element)
   {
       return element*2+2;
   };
   inline int getParent(int k)
{
    if (k == 0)
        return -1;
    if (k && 1 == 0)
    {
        return (k-2) >> 1;
    }
    else
    {
        return (k-1) >> 1;
    };
    return -1;

}

inline void Move(int i, int j)
{
  T temp;
 //  utiassert(m_heap[i]->m_internheapinfo.m_heapposition == i);
    //    utiassert(m_heap[j]->m_internheapinfo.m_heapposition == j);
            temp = m_heap[i];
            m_heap[i] = m_heap[j];
            m_heap[j] = temp;
      //  m_heap[i]->m_internheapinfo.m_heapposition = i;
       // m_heap[j]->m_internheapinfo.m_heapposition = j;
}

void downheap(int item)
{
   // int itemweight = GetItemWeighti(item);
    T itemobj = m_heap[item];
   // utiassert(itemobj->m_internheapinfo.m_heapposition==item);
    int childleft;
    int childright;
    bool check;
    T chwl, chwr;
    while (true)
    {
        childleft = item*2+1;
        childright = item*2+2;
        if (childleft >= m_count)
            return;
        if (childright >= m_count)
        {
           // chwl = GetItemWeighti(childleft);
            chwl = m_heap[childleft];
       //  utiassert(chwl->m_internheapinfo.m_heapposition==childleft);
            if (m_compare.isGreater(chwl, itemobj))
            {
                Move(childleft, item);
                return;
            };
            return;
        };
        chwl = m_heap[childleft];
        chwr = m_heap[childright];
    check = m_compare.isGreater(chwl, itemobj) || m_compare.isGreater(chwr, itemobj);
        if (!(check)
      //chwl <= itemweight && chwr <= itemweight
      )
      {
            return;
        }
        //if (chwl > itemweight || chwr > itemweight)
        if (check)
        {
            if (m_compare.isGreater(chwl , chwr))
            {
                Move(childleft, item);
                item = childleft;

            }
            else
            {
                Move(childright, item);
                item = childright;
            };

        };
    };
}

void upheap(int item)
{
    int parent = getParent(item);
    T itemobj = m_heap[item];
    //int itemweight = GetItemWeighti(item);
    while (parent != -1)
    {
        //int pw =GetItemWeighti(parent);
        T pw = m_heap[parent];
        if (m_compare.isGreater(itemobj, pw)
      //pw < itemweight
      )
        {
            Move(parent, item);

            item = parent;
            parent = getParent(item);
        }
        else
        {
            //Ok
            return;
        };
    };
}
public:
  Array<T> m_heap;
  int m_count;
#ifdef UTI_SETDEBUG
    std::set<T> m_debug;
#endif

  CUtiHeap()
  {
    m_heap.setSize(32);

     m_count = 0;
   };
   inline void clear()
   {
     m_count = 0;
     m_heap.setSize(32);
#ifdef UTI_SETDEBUG
        m_debug.clear();
#endif
   };
  inline int size() const
  {
      #ifdef UTI_SETDEBUG
      assert(m_count == m_debug.size());
      #endif
     return m_count;
  };
  const T& front() const
  {
      assert(m_count != 0);
#ifdef UTI_SETDEBUG
       assert(m_count == m_debug.size());
   /*  for (int i = 1; i < m_count; i++) {

       assert(m_compare.isGreater(m_heap[0], m_heap[i])== true);
     };*/
     T value = *m_debug.begin();

     assert(value== m_heap[0]);

#endif

     return m_heap[0];
  }
  /* void setSize(int s)
  {
    m_count = 0;
    if (m_heap.size() == s) return;
     m_heap.setSize(s);
    m_heap.fill(NULL);

  };*/
 /* inline bool isInHeap(T*h)
  {
      return h->m_internheapinfo.m_heapposition != -1;
   }*/
  inline void push_heap(const T& item)
{
      if (m_heap.size() <= m_count) {
          if (fastgrow) {
                m_heap.setSizeCopyHighSpeed(m_heap.size()*2);
          }else {
    m_heap.setSizeCopy(m_heap.size()*2);
          }
  };
    m_heap[m_count] = item;
#ifdef UTI_SETDEBUG
        m_debug.insert(item);
#endif
   // item->m_internheapinfo.m_heapposition = m_count;
    m_count++;
    utiassert(m_count <= m_heap.size());
    #ifdef USE_CUtiHeap_STL
    std::push_heap(m_heap.array, m_heap.array+m_count, std::greater<int>());
#else
    upheap(m_count-1);
#endif
#ifdef UTI_SETDEBUG
       assert(m_count == m_debug.size());
   /*  for (int i = 1; i < m_count; i++) {

       assert(m_compare.isGreater(m_heap[0], m_heap[i])== true);
     };*/
     T value = *m_debug.begin();

     assert(value== m_heap[0]);

#endif
}
inline void pop_front()
{
#ifdef UTI_SETDEBUG
assert(m_count == m_debug.size());

assert(*m_debug.begin()== m_heap[0]);

m_debug.erase(m_debug.begin());
#endif
    if (m_count == 1)
    {
        m_count = 0;

        return;
    }
   // CUtiOB* t = m_heap[0];
    m_heap[0] = m_heap[m_count-1];


    #ifdef UTI_SETDEBUG
	assert(m_heap.size() >= m_count);
    std::pop_heap(m_heap.array, m_heap.array+m_count,  std::greater<int>());
    m_count = m_count-1;
#else
     m_count = m_count-1;
    downheap(0);
#endif
    if (m_count < m_heap.size()/4 && m_heap.size() >= 64) {
        assert(m_count <= m_heap.size()/2);
        if (fastgrow) {
              m_heap.setSizeCopyHighSpeed(m_heap.size()/2);
        }else {
  m_heap.setSizeCopy(m_heap.size()/2);
        }

    }

#ifdef UTI_SETDEBUG
       assert(m_count == m_debug.size());
   /*  for (int i = 1; i < m_count; i++) {

       assert(m_compare.isGreater(m_heap[0], m_heap[i])== true);
     };*/
       if (m_count){
     T value = *m_debug.begin();

     assert(value== m_heap[0]);
       }

#endif


};/*
inline void update(int j)
{

  //downheap(j);
  //upheap(j);
  if (j == 0) {
   downheap(0);
  } else if (j == m_count-1) {
   upheap(m_count-1);
  } else {
    if (m_compare.isGreater(m_heap[j], m_heap[getParent(j)])) {
      upheap(j);
    } else {
      downheap(j);
    };
  };
}*/
};

template <class T, typename CalcHash = CCacheHash_ConvertToInt<T> > struct CCacheHash
{
  protected:
     struct ArrayElement
     {
       T m_value;
       int m_id;
     };
    Array<ArrayElement> m_hasharray;
    CalcHash m_calchash;
    int m_id;
  public:
    CCacheHash() :  m_hasharray(29)
    {

      m_id = 1;
      clear();
    };
    CCacheHash(int size) : m_hasharray(size)
    {
       m_id = 1;
      clear();
    }
    ~CCacheHash()
    {

    }
    void clear()
    {
      ArrayElement element;
      element.m_value = T();
      element.m_id = 0;
      m_hasharray.fill(element);
    }
    inline void insert(const T& value)
    {
      int index = m_calchash.hashValue(value) % m_hasharray.size();
      m_hasharray[index].m_value = value;
      m_hasharray[index].m_id = m_id;
    }
    inline bool hasElement(const T& value)
    {
       int index = m_calchash.hashValue(value) % m_hasharray.size();
       if (m_hasharray[index].m_id == m_id && m_hasharray[index].m_value == value) return true;
       return false;
    }
    inline T* hasElementReturn(const T& value)
    {
       int index = m_calchash.hashValue(value) % m_hasharray.size();
       if (m_hasharray[index].m_id == m_id && m_hasharray[index].m_value == value) return &m_hasharray[index].m_value ;
       return NULL;
    }
    inline void incID()
    {
        m_id++;
    };

};


template <class T, typename CalcHash = CCacheHash_ConvertToInt<T> > struct CCheckHash
{
 protected:
     struct ArrayElement
     {
       T m_value;
       int m_id;
     };
    Array<ArrayElement> m_hasharray;
    std::set<T> m_myset;
    CalcHash m_calchash;
    int m_id;
 public:
    CCheckHash() :  m_hasharray(29)
    {
      m_id = 1;
      clear();
    };
    CCheckHash(int size) : m_hasharray(size)
    {
      m_id = 1;
      clear();
    }

    ~CCheckHash()
    {

    }
    void clear()
    {
     ArrayElement element;
      element.m_value = T();
      element.m_id = 0;
      m_hasharray.fill(element);
    }
    inline int multiplicationHash(int index)
    {
       int alpha = 375; //max 1024
       int size = m_hasharray.size();

       int m = (size * ((index * alpha) & 1023) >> 10);
       utiassert(m >= 0);
       utiassert(m < size);
       return m;
    }
    inline void insert(const T& value)
    {
      int index = m_calchash.hashValue(value);
      int division = index % m_hasharray.size();
      if (m_hasharray[division].m_id != m_id) {
         m_hasharray[division].m_value = value;
         m_hasharray[division].m_id = m_id;
      } else {
        int mul = multiplicationHash(index);
        if (m_hasharray[mul].m_id != m_id) {
          m_hasharray[mul].m_value = value;
          m_hasharray[division].m_id = m_id;
        } else {
          m_myset.insert(value);
        };
      };
    }
    inline bool hasElement(const T& value)
    {
        int index = m_calchash.hashValue(value);
        int division = index % m_hasharray.size();
         if (m_hasharray[division].m_id != m_id) {
             return false;
         };
         if (m_hasharray[division].m_value == value) {
           return true;
         };
         int mul = multiplicationHash(index);
         if (m_hasharray[mul].m_id != m_id) {
              return false;
         };
         if (m_hasharray[mul].m_value == value) {
           return true;
         };
         return m_myset.find(value) != m_myset.end();
    }
     inline void incID()
    {
        m_id++;
        m_myset.clear();
    };
};
#ifndef NDEBUG
#define HASHDEBUG
#endif

template<class T, class T2, bool multi, typename CalcHash = CCacheHash_ConvertToInt<T> > struct CListHash
{
protected:
    struct OneEntry : public CUtiOB
    {
        CLinkedList_Controller clinkedlist_controller[1];
        std::pair<T, T2> m_value;
    };

    struct ArrayElement
    {
      COwnLinkedList<OneEntry, 0> m_list;

    };
   Array<ArrayElement> m_hasharray;
   CalcHash m_calchash;
   int m_totalcount;
#ifdef HASHDEBUG
   std::multimap<T, T2> m_debugmap;
#endif

public:
   typedef OneEntry* iterator;
   CListHash() :  m_hasharray(29)
   {

     m_totalcount = 0;
     clear();
   };
   CListHash(int size) : m_hasharray(size)
   {

     clear();
   }

   ~CListHash()
   {

   }
   void setSize(int s)
   {
       m_hasharray.setSize(s); clear();
   }
   int size() const
   {
	   return m_totalcount;
   }
   void clear()
   {
     for (int i = 0; i < m_hasharray.size(); i++) {
         m_hasharray[i].m_list.clear();

     }
     m_totalcount = 0;
#ifdef HASHDEBUG
   m_debugmap.clear();
#endif
   }
public:
   iterator insert(const std::pair<T, T2>& value)
   {
       if (!multi) {
           utiassert(hasElement(value.first) == false);
       }
     int index = m_calchash.hashValue(value.first);
     int division = index % m_hasharray.size();
     ArrayElement& el = m_hasharray[division];
     OneEntry* entry = new OneEntry();
     entry->m_value = value;
     el.m_list.push_back(entry);
     m_totalcount++;
#ifdef HASHDEBUG
   m_debugmap.insert(value);
#endif
     return entry;
   }
   bool hasElement(const T& value)
   {

       int index = m_calchash.hashValue(value);
       utiassert(index >= 0);
       int division = index % m_hasharray.size();
       ArrayElement& el = m_hasharray[division];
       for (typename COwnLinkedList<OneEntry, 0>::iterator it = el.m_list.begin(); it != el.m_list.end(); it++)
       {
           OneEntry& entry= *it;
           if (entry.m_value.first == value) {
#ifdef HASHDEBUG
               assert(m_debugmap.find(value) != m_debugmap.end());
#endif
               return true;
           }
       }
#ifdef HASHDEBUG
               assert(m_debugmap.find(value) == m_debugmap.end());
#endif
       return false;
   }
   void erase(iterator iter)
   {
       int index = m_calchash.hashValue(iter->m_value.first);
       utiassert(index >= 0);
       int division = index % m_hasharray.size();
       ArrayElement& el = m_hasharray[division];


#ifdef HASHDEBUG
       std::pair<typename std::multimap<T, T2>::iterator,typename  std::multimap<T, T2>::iterator> iterpair = m_debugmap.equal_range(iter->m_value.first);
       assert(iterpair.first != iterpair.second);
       bool found = false;
       for (typename std::multimap<T, T2>::iterator it = iterpair.first; it != iterpair.second; it++)
       {
           assert(iter->m_value.first == it->first);
           if (iter->m_value.second == it->second) {
               m_debugmap.erase(it);
               found = true;
               break;
           }
       }
       assert(found == true);
       //    m_debugmap.erase(value);
#endif
       el.m_list.remove(iter);
       m_totalcount--;
   }

   void erase(const T& value)
   {
        int index = m_calchash.hashValue(value);
        utiassert(index >= 0);
        int division = index % m_hasharray.size();
        ArrayElement& el = m_hasharray[division];

        typename COwnLinkedList<OneEntry, 0>::iterator it, next;
        it = el.m_list.begin();
#ifdef HASHDEBUG
           m_debugmap.erase(value);
#endif
        while (it != el.m_list.end()) {
            next = it;
            next++;
            OneEntry& entry= *it;
            if (entry.m_value.first == value) {
                el.m_list.erase(it);
                m_totalcount--;
                if (!multi) {
                    return;
                }
            }

            it = next;
        }

   }
   template<class ResList> void find(const T& value, ResList& list)
   {
       int index = m_calchash.hashValue(value);
       utiassert(index >= 0);
       int division = index % m_hasharray.size();
       ArrayElement& el = m_hasharray[division];
#ifdef HASHDEBUG
       int basecount = list.size();
       std::set<T2> added;
#endif
       for (typename COwnLinkedList<OneEntry, 0>::iterator it = el.m_list.begin(); it != el.m_list.end(); it++)
       {
           OneEntry& entry= *it;
           if (entry.m_value.first == value) {
               list.push_back(entry.m_value.second);
#ifdef HASHDEBUG

       added.insert(entry.m_value.second);
#endif
           }
       }
#ifdef HASHDEBUG
       basecount = list.size()-basecount;
       assert(basecount == added.size());
       std::pair<typename std::multimap<T, T2>::iterator,typename std::multimap<T, T2>::iterator> iterpair = m_debugmap.equal_range(value);
       assert(iterpair.first != iterpair.second || basecount == 0);
       for (typename std::multimap<T, T2>::iterator it = iterpair.first; it != iterpair.second; it++)
       {
           assert(value == it->first);

           assert(added.find(it->second) != added.end());
          basecount--;
       }
       assert(basecount == 0);
#endif
   }

   bool findOne(const T& value, T2& list)
   {
	   int index = m_calchash.hashValue(value);
	   utiassert(index >= 0);
	   int division = index % m_hasharray.size();
	   ArrayElement& el = m_hasharray[division];

	   for (typename COwnLinkedList<OneEntry, 0>::iterator it = el.m_list.begin(); it != el.m_list.end(); it++)
	   {
		   OneEntry& entry = *it;
		   if (entry.m_value.first == value) {
			 //  list.push_back(entry.m_value.second);
			   list = entry.m_value.second;
			   return true;
		   }
	   }
	   return false;
   }

   inline void erase(const std::pair<T, T2>& value)
   {
        int index = m_calchash.hashValue(value.first);
        utiassert(index >= 0);
        int division = index % m_hasharray.size();
        ArrayElement& el = m_hasharray[division];

        typename COwnLinkedList<OneEntry, 0>::iterator it, next;
        it = el.m_list.begin();
        bool found = false;
        while (it != el.m_list.end()) {
            next = it;
            next++;
            OneEntry& entry= *it;
            if (entry.m_value == value) {
                el.m_list.erase(it);
                m_totalcount--;
                found = true;
               break;
            }

            it = next;
        }
        assert(found == true);
#ifdef HASHDEBUG
       std::pair<typename std::multimap<T, T2>::iterator,typename std::multimap<T, T2>::iterator> iterpair = m_debugmap.equal_range(value.first);
       assert(iterpair.first != iterpair.second);
       for (typename std::multimap<T, T2>::iterator it = iterpair.first; it != iterpair.second; it++)
       {
           assert(value.first == it->first);
           if (value.second == it->second) {
               m_debugmap.erase(it);
               return;
           }
       }
       assert(false);
       //    m_debugmap.erase(value);
#endif

   }

};

template <class T, class T2, typename CalcHash = CCacheHash_ConvertToInt<T> , bool fastgrow=false> struct CNormalHash
{
 protected:
     struct ArrayElement
     {
       unsigned int m_valueindirect;
       unsigned short m_id;
     };
    Array<ArrayElement> m_hasharray;
    struct GrowArrayElement {
        T m_value;
        T2 m_second;
    };
    inline int createGrow(const T& value, const T2& value2)
    {
        GrowArrayElement g;
        g.m_value = value;
        g.m_second = value2;
		return m_firstgrow.fast_add(g);
    }

    Array<GrowArrayElement> m_firstindirection;
    UtiArrayAccessGrow<GrowArrayElement, fastgrow> m_firstgrow;
    std::map<T, T2> m_myset;
    CalcHash m_calchash;
    int m_id;
 public:
    CNormalHash() :  m_hasharray(29), m_firstindirection(1024), m_firstgrow(m_firstindirection)
    {
      m_id = 1;
      clear();
    }
    CNormalHash(int size) : m_hasharray(size), m_firstindirection(1024),m_firstgrow(m_firstindirection)
    {
      m_id = 1;
      clear();
    }

    ~CNormalHash()
    {

    }
    void setSize(int s)
    {
        m_hasharray.setSize(s); clear();
    }
    protected: void clear()
    {
     ArrayElement element;
      element.m_valueindirect = 0;
      element.m_id = 0;
      m_hasharray.fill(element);
      m_firstgrow.reset(m_hasharray.size());
    }
public:
    inline int multiplicationHash(int index)
    {
       int alpha = 375; //max 1024
       int size = m_hasharray.size();

       int m = ((size * ((index * alpha) & 1023) >> 10) & ~(1 << 31)) % size;
       utiassert(m >= 0);
       utiassert(m < size);
       return m;
    }
    inline void insert(const T& value, const T2& value2)
    {
      int index = m_calchash.hashValue(value);
      int division = index % m_hasharray.size();
      if (m_hasharray[division].m_id != m_id) {
          m_hasharray[division].m_valueindirect = createGrow(value, value2);
     //    m_hasharray[division].m_second = value2;
         m_hasharray[division].m_id = m_id;
      } else {
        int mul = multiplicationHash(index);
        if (m_hasharray[mul].m_id != m_id) {
          m_hasharray[mul].m_valueindirect = createGrow(value, value2);
        //   m_hasharray[mul].m_second = value2;
          m_hasharray[mul].m_id = m_id;
        } else {
            assert(m_myset.find(value) == m_myset.end());
          m_myset.insert(std::make_pair(value, value2));
        };
      };
    }
    inline bool hasElement(const T& value, T2& resvalue)
    {
        int index = m_calchash.hashValue(value);
        utiassert(index >= 0);
        int division = index % m_hasharray.size();
         if (m_hasharray[division].m_id != m_id) {
             return false;
         };
         GrowArrayElement& el1 = m_firstindirection[m_hasharray[division].m_valueindirect];
         if (el1.m_value == value) {
           resvalue = el1.m_second;
           return true;
         };
         int mul = multiplicationHash(index);
         if (m_hasharray[mul].m_id != m_id) {
              return false;
         };
         GrowArrayElement& el2 =m_firstindirection[m_hasharray[mul].m_valueindirect];
         if (el2.m_value == value) {
           resvalue = el2.m_second;
           return true;
         };
         typename std::map<T, T2>::iterator it = m_myset.find(value);
         if (it == m_myset.end()) return false;
         resvalue = it->second;
         return true;
    }
    inline size_t getCacheSize()
    {
      return m_myset.size();
    }
     inline void incID()
    {
        m_id++;
        if (m_id ==65000) {
            m_id = 0;
            clear();
        }
        if ((m_id & 100) == 0) {
             m_firstgrow.reset(64);
        } else {
          m_firstgrow.reset(m_hasharray.size());
        }
        m_myset.clear();
    };
};

template <class T, class T2, typename CalcHash = CCacheHash_ConvertToInt<T> , bool fastgrow=false, int checkamount =4> struct CLinearHash
{
 protected:
     struct ArrayElement
     {
       unsigned int m_valueindirect;
       unsigned short m_id;
     };
    Array<ArrayElement> m_hasharray;
    struct GrowArrayElement {
        T m_value;
        T2 m_second;
    };
    inline int createGrow(const T& value, const T2& value2)
    {
        GrowArrayElement g;
        g.m_value = value;
        g.m_second = value2;
		return m_firstgrow.fast_add(g);
    }

    Array<GrowArrayElement> m_firstindirection;
    UtiArrayAccessGrow<GrowArrayElement, fastgrow> m_firstgrow;
    std::map<T, T2> m_myset;
    CalcHash m_calchash;
    int m_id;
    int m_elementsinhash;
 public:
    CLinearHash() :  m_hasharray(29), m_firstindirection(1024), m_firstgrow(m_firstindirection)
    {
      m_id = 1;m_elementsinhash = 0;
      clear();
    }
    CLinearHash(int size) : m_hasharray(size), m_firstindirection(1024),m_firstgrow(m_firstindirection)
    {
      m_id = 1;
      clear();
    }

    ~CLinearHash()
    {

    }

    inline int size() {
        return m_elementsinhash;
    }

    void setSize(int s)
    {
        m_hasharray.setSize(s); clear();
    }
    protected: void clear()
    {
     ArrayElement element;
      element.m_valueindirect = 0;
      element.m_id = 0; assert(m_id >= 1);
      m_hasharray.fill(element);
      m_firstgrow.reset(m_hasharray.size());
      m_elementsinhash = 0;
    }
public:

    inline void insert(const T& value, const T2& value2)
    {
      int index = m_calchash.hashValue(value);
      int size = m_hasharray.size();
      int division = index % size;
      bool inserted = false;
      for (int i = 0; i < checkamount; i++) {
          ArrayElement& ael = m_hasharray[division];
         if (ael.m_id != m_id) {
            ael.m_valueindirect = createGrow(value, value2);
     //    m_hasharray[division].m_second = value2;
            ael.m_id = m_id;
            inserted = true;
            m_elementsinhash++;
            break;
         }
         division = (division + 1) %size;
      }
      if (!inserted) {
          assert(m_myset.find(value) == m_myset.end());
        m_myset.insert(std::make_pair(value, value2));
        m_elementsinhash++;
      }
      /*else {
        int mul = multiplicationHash(index);
        if (m_hasharray[mul].m_id != m_id) {
          m_hasharray[mul].m_valueindirect = createGrow(value, value2);
        //   m_hasharray[mul].m_second = value2;
          m_hasharray[mul].m_id = m_id;
        } else {
            assert(m_myset.find(value) == m_myset.end());
          m_myset.insert(std::make_pair(value, value2));
        };
      };*/
    }
    inline bool hasElement(const T& value, T2& resvalue)
    {
        int index = m_calchash.hashValue(value);
        int size = m_hasharray.size();
        utiassert(index >= 0);
        int division = index % size;

         for (int i = 0; i < checkamount; i++) {
             if (m_hasharray[division].m_id != m_id) {
                 return false;
             };
             GrowArrayElement& el1 = m_firstindirection[m_hasharray[division].m_valueindirect];
             if (el1.m_value == value) {
               resvalue = el1.m_second;
               return true;
             };
              division = (division + 1) %size;
         }
         typename std::map<T, T2>::iterator it = m_myset.find(value);
         if (it == m_myset.end()) return false;
         resvalue = it->second;
         return true;
    }
    inline size_t getCacheSize()
    {
      return m_myset.size();
    }
     inline void incID()
    {
        m_id++;
        m_elementsinhash = 0;
        if (m_id ==65000) {
            m_id = 1;
            clear();
        }
        if ((m_id & 100) == 0) {
             m_firstgrow.reset(64);
        } else {
          m_firstgrow.reset(m_hasharray.size());
        }
        m_myset.clear();
    };
};


#endif
