/***************************************************************************
 *   Copyright (C) 2006 by Michael Staud   *
 *   michael.staud@uni-ulm.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifdef _WIN32
#include <objbase.h>
//#define swprintf snwprintf
#else

#endif
#include "UtiString.h"
#include "UtiArchiv.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <assert.h>

#include <string.h>
#include <stdio.h>
#include <vector>
#include <algorithm>
#include <sstream>
#include "uti64coder.h"

#include "utisystem.h"
#include "utiarray.h"
#include "lispnames.h"
#include <string>
#include "sha1/sha1.hpp"

using namespace std;
using namespace UtiFramework;

#ifdef _WIN32

#define swprintf _snwprintf
#endif

int utistrlen(const wchar_t* t)
{
    int r = 0;

    while (t[r] != 0) {
        r++;
    }

    return r;

}



void gen_random(wchar_t *s, const int len) {
    static const char alphanum[] =
        "0123456789"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz";

    for (int i = 0; i < len; ++i) {
        s[i] = alphanum[rand() % (sizeof(alphanum) - 1)];
    }

    //s[len] = 0;
}
CUtiString createRandomString(int length)
{
    CUtiString str;
    str.setLength(length);
    gen_random(const_cast<wchar_t*>(str.c_str()), length);
    return str;
}
void ReadString(CStream & Stream, CUtiString &s, int version)
{
  int count;

  Stream.ReadInteger(count);
  utiassert(count >= 0);
  if (version <= 11)
  {
    char *t = new char[count+1];
    Stream.ReadData(t, count*sizeof(char));
    t[count] = 0;
    s.LoadAnsi(t);
    delete []t;
  }
  else
  {
    wchar_t *t = new wchar_t[count+1];
    if (sizeof(wchar_t) == 2)
    {
      Stream.ReadData(t, count*sizeof(wchar_t));
    }
    else
    {
      for (int i = 0; i < count; i++)
      {
        unsigned short s=0;
        Stream.ReadData(&s, 2);
        t[i] = s;
      }
      //Stream.ReadData(t, count*sizeof(char_t));
    }
    t[count] = 0;
    s = t;
    delete []t;

  }

}

void WriteString(CStream & Stream, const CUtiString &s, int version)
{
  int count = s.size();
  Stream.WriteInteger(count);
  if (sizeof(wchar_t) == 2)
  {
    Stream.WriteData(s.getString(), count*sizeof(wchar_t));
  }
  else
  {
    for (int i = 0; i < count; i++)
    {
      unsigned short s2=s[i];
      Stream.WriteData(&s2, 2);
    }

  }
}

std::wostream & operator<<(std::wostream & os, const CUtiString & str)
{
   os << str.c_str();
   return os;
};

std::ostream & operator<<(std::ostream & os, const CUtiString & str)
{
  os << converttomulti(str.c_str());
   return os;
}

CUtiStringContainer::CUtiStringContainer(const U_CHAR* std) : m_string(const_cast<U_CHAR*>(std), utistrlen(std)+1)
{
 // m_buffersize = utistrlen(std);
  m_length = m_string.size()-1;
  //m_string = const_cast<U_CHAR*>(std);
 // m_owns = true;
}

CUtiStringContainer::CUtiStringContainer(int size) : m_string(size)
{
 // m_buffersize = size;
  m_length = 0;
#ifdef DEBUGSTRING
  m_string.fill(0);
#endif
 // m_string = new U_CHAR[size];
 // m_owns = true;
}

CUtiStringContainer::~CUtiStringContainer()
{
   /*if (m_owns) {
      delete [] m_string;
   }*/
}


bool CUtiStringContainer::allowResize()
{

  return m_references <= 1 && m_string.isDynamic();
}

void CUtiStringContainer::checkSize()
{
  //TODO: Hier testen, ob man es verkleinern muss.
  if (m_string.size()/ 2 > m_length+16) {
    resize(m_string.size() / 2);
  };
}


void CUtiStringContainer::resize(int size)
{
  utiassert(allowResize());
  utiassert(size > 0);
  m_string.setSizeCopy(size);
  //U_CHAR* newdata = new U_CHAR[size];
  int newlength = umin(size, m_length);
  //memcpy(newdata, m_string, newlength*sizeof(U_CHAR));
  //m_buffersize = size;
  m_length = newlength;
  //delete [] m_string;
  //m_string = newdata;
}



CUtiStringContainer* CUtiStringContainer::createNew(int size)
{
  CUtiStringContainer* newcontainer = new CUtiStringContainer(size);
  newcontainer->m_string.fill(0);
  newcontainer->m_length = umin(size-1, m_length);
  utiassert(newcontainer->m_length < newcontainer->m_string.size());
  m_string.copyToArray(0, newcontainer->m_length+1, newcontainer->m_string, 0);
  //memcpy(newcontainer->m_string, m_string, (newcontainer->m_length+1)*sizeof(U_CHAR));
  return newcontainer;
}

CUtiString CUtiString::getFileExt() const
    {
    CUtiString res = *this; res.ExtractFileExt();return res;
    }
    CUtiString CUtiString::getFileExtP() const
    {
    CUtiString res = *this; res.ExtractFileExt();return UTISTR(".")+ res;
    }
    CUtiString CUtiString::getFilePath() const
    {
    CUtiString res = *this; res.ExtractFilePath();return res;
    }
    CUtiString CUtiString::getFileName() const
    {
    CUtiString res = *this; res.ExtractFileName();return res;
    }
    CUtiString CUtiString::getFileNameWithoutExt() const
    {
    CUtiString res = *this; res.ExtractFileNameWithoutExt();return res;
    }
    CUtiString CUtiString::getRealFileName() const
    {
    CUtiString res = *this; res.ExtractRealFileName();return res;
    }

bool CUtiString::checkContainer() const
{CHECKEXISTANCE;
#ifndef DEBUGSTRING
     return true;
#endif
  utiassert(m_container.get() != NULL);
  CHECKEXISTANCEOF(m_container.get());
  if (m_container->m_length == 0)return true;
  utiassert(m_container->m_string.size() != 0);
  utiassert(m_container->m_string.getCArray() != NULL);
  //if (utistrlen(m_container->m_string.getCArray()) != m_container->m_length){
  int l =  (int)utistrlen(m_container->m_string.getCArray());
  utiassert(l
                ==
                m_container->m_length);
  //}
  utiassert(m_container->m_length < m_container->m_string.size());

  return true;
}

int CUtiString::getBufferSize(int length)
{
   return length+32;
}

int CUtiString::debugReferences()
{
     return m_container->getReferenceCount();
}

int CUtiString::debugContainerLength()
{
  return m_container->m_length;
}

int CUtiString::debugBufferSize()
{
  return m_container->m_string.size();
}

bool CUtiString::debugOwns()
{
  return m_container->m_string.isDynamic();
}

void CUtiString::ownContainer()
{
  if (m_container.get() == NULL){
     m_container = new CUtiStringContainer(16);
     return;
  };
  if (m_container->allowResize()) {

  } else {
     m_container = m_container->createNew(getBufferSize(m_container->m_length+1));
  };
}

void CUtiString::ownContainer(int length)
{
  utiassert(length >= 0);

  if (m_container.get() == NULL){
     m_container = new CUtiStringContainer(getBufferSize(length+1));
     return;
  };
  utiassert(m_container->m_length < m_container->m_string.size());
  if (m_container->allowResize() && m_container->m_string.size() >= length+1) {

  } else {
     m_container = m_container->createNew(getBufferSize(length+1));
     utiassert(m_container->m_length < m_container->m_string.size());
  };
};

void CUtiString::setLength(int l)
{
    ownContainer(l);
    m_container->m_length = l;
    m_container->m_string[l] = 0;
    for (int i = 0; i < l; i++) {
      m_container->m_string[i] = L' ';
    };
#ifdef DEBUGSTRING
    utiassert(checkContainer());
#endif
}

const CUtiString RandomString(TCHAR* start, int l, int h, TCHAR* ende)
{
    CUtiString s;
    s = start;
    int r = l+urandom(h-l+1);
    s += r;
    s += ende;
    return s;
}

const CUtiString CalcNumberedFileName(const CUtiString& start, const CUtiString& ende, int startindex)
{
    CUtiString res;
    int k=startindex;
    do
    {
        res = start;
        res += k;
        res += ende;
        k++;
    }
    while (FileExists(res.getString()));
    return res;
}

CUtiString CUtiString::operator()(unsigned int p, int l) const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    //utiassert(p+l <= m_container->m_length);
    if (p+l > m_container->m_length) {
        l = m_container->m_length-p;

    }
    if (l < 0) {
        return EMPTYSTRING;
    }
    utiassert(p >= 0);
    utiassert(l >= 0);
    CUtiString tmp;
    tmp.insert(0, l, m_container->m_string.getCArray() +p);
    return(tmp);
}
/*
bool IsInteger(U_CHAR *str, int length)
{
    for (int i = 0; i < length; i++)
    {
        if (IsCharacter(str[i]))
        {
            if (!IsNumber(str[i]))
            {
                return(false);

            };
        };
    };
    return(true);
};
*/
int SearchChar(U_CHAR *str, U_CHAR c, int length)
{
    int res = -1;
    for (int i = 0; i < (int)length; i++)
    {
        if (str[i] == c)
        {
            res = i;
            return(res);
        };
    };
    return(res);
}

int SearchCharBackward(U_CHAR *str, U_CHAR c, int length)
{
    int res = -1;
    for (int i = length-1; i >= 0; i--)
    {
        if (str[i] == c)
        {
            res = i;
            return(res);
        };
    };
    return(res);
}


bool IsEqual(U_CHAR* str, U_CHAR* item, int length)
{
    for (int i = 0; i < length; i++)
    {
        if (toupper(str[i]) != toupper(item[i]))
        {
            return(false);
        };
    };
    return(true);
};

/*
bool ValidBez(U_CHAR* str, int i, int length, int searchlength)
{
    if (i != 0)
    {
        if (IsBezeichner(str[i-1]))
        {
            return(false);
        };
    };
    if (i+searchlength < length)
    {
        if (IsBezeichner(str[i+searchlength]))
        {
            return(false);
        };
    };
    return(true);
};*/
/*
bool ValidVergleich(U_CHAR* str, int i, int length, int searchlength)
{
    if (i != 0)
    {
        if (IsVergleich(str[i-1]))
        {
            return(false);
        };
    };
    if (i+searchlength < length)
    {
        if (IsVergleich(str[i+searchlength]))
        {
            return(false);
        };
    };
    return(true);
};
*/
/*
int SearchString(U_CHAR *str, U_CHAR* c, int length, int searchlength)
{
    int res = -1;
    int i = 0;
    for (i = 0; i < (int)length-(int)searchlength; i++)
    {
        //if (!ValidBez(str, i, length, searchlength)) continue;
        if (IsEqual(&str[i], c, searchlength))
        {
            res = i;
            return(res);
        };
    };
    return(res);
}

#define IsBez(a) if ((i+(int)utistrlen(a)-1 < (int)length) && IsEqual(&str[i], a, utistrlen(a)) && ValidBez(str, i, length, utistrlen(a)))
*//*
void SearchBeginEnd(U_CHAR*str, int length, int &beginpos, int &endpos, bool last)
{
    beginpos = -1;
    endpos = -1;
    int i = 0;
    int level = 0;
    int searchlength = 3;
    for (i = 0; i < (int)length; i++)
    {
        IsBez(UTEXT("begin"))
        {
            if (level == 0)
                beginpos = i;
            level++;
        }
        IsBez(UTEXT("record"))
        {
            if (level == 0)
                beginpos = -1;
            level++;
        };
        IsBez(UTEXT("class"))
        {
            if (level == 0)
                beginpos = -1;
            level++;
        };
        IsBez(UTEXT("case"))
        {
            if (level == 0)
                beginpos = -1;
            level++;
        };
        IsBez(UTEXT("object"))
        {
            if (level == 0)
                beginpos = -1;
            level++;
        };
        IsBez(UTEXT("end"))
        {
            level--;
            if ((level == 0) && (beginpos != -1))
            {
                endpos = i;
                if (!last)
                    return;
            };
        };

    };
};*/

int SearchStringBackward(U_CHAR *str, U_CHAR* c, int length, int searchlength)
{
    int res = -1;
    int i = 0;

    for (i = 0; i < (int)(length-searchlength); i++)
    {
        //if (!ValidBez(str, i, length, searchlength)) continue;
        if (IsEqual(&str[i], c, searchlength))
        {
            res = i;
        };
    };
    return(res);
}



/*

void SearchStrings(U_CHAR* str, U_CHAR*c, int length, int searchlength, int* result, int &rescount)
{
    int i = 0;
    rescount = 0;
    int level = 0;
    for (i = 0; i < (int)length-searchlength; i++)
    {
        if (str[i] == UTEXT('('))
        {
            level = level + 1;
            continue;
        };
        if (str[i] == UTEXT(')'))
        {
            level = level - 1;
            //if (level < 0) DEBUG(UTEXT("SearchStrings: Zu viele geschlossene Klammern."));
            continue;
        };
        if (level != 0)
            continue;
        if (IsBezeichner(c[0])) //nur wenn der Suchstring selbst ein Bezeichner ist
            if (!ValidBez(str, i, length, searchlength))
                continue;
        if (IsVergleich(c[0]))
        {
            if (!ValidVergleich(str, i, length, searchlength))
                continue;
        };
        if (IsEqual(&str[i], c, searchlength))
        {
            if (rescount < MAX_ITEMS)
            {
                result[rescount] = i;
                rescount++;

            }
            else
            {
                //DEBUG(UTEXT("SearchStrings: Zu Viele Items"));
            };
        };
    };
}*/
/*
int SearchStringBez(U_CHAR *str, U_CHAR* c, int length, int searchlength)
{
    int res = -1;
    int i = 0;
    int j = 0;

    if (length < searchlength)
        return(-1);
    for (i = 0; i < (int)(length-searchlength+2); i++)
    {
        //if (!ValidBez(str, i, length, searchlength)) continue;
        //if (IsEqual(&str[i], c, searchlength)) {
        //  res = i;
        //  return(res);
        //};
        IsBez(c)
        {
            res = i;
            return(res);
        };
    };
    return(res);
}

int SearchStringBackwardBez(U_CHAR *str, U_CHAR* c, int length, int searchlength)
{
    int res = -1;
    int i = 0;
    int j = 0;

    for (i = 0; i < (int)(length-searchlength); i++)
    {
        if (!ValidBez(str, i, length, searchlength))
            continue;
        if (IsEqual(&str[i], c, searchlength))
        {
            res = i;
        };
    };
    return(res);
}

int SearchChar(U_CHAR *str, U_CHAR c, int length, int charpos)
{
    int res = -1;
    for (int i = 0; i < (int)length; i++)
    {
        if (str[i] == c)
        {
            charpos--;
            if (charpos == 0)
            {
                res = i;
                return(res);
            };
        };
    };
    if (charpos == 1)
    { //Bei vorletztem, l�ge setzen
        res = length;
    };
    return(res);
}
*/


void CUtiString::replace(const U_CHAR *s)
{


    int newlength = (int)utistrlen(s);

    ownContainer(newlength);
    for (int i = 0; i < newlength; i++)
    {
      m_container->m_string[i] = s[i];
    };
    //memcpy(m_container->m_string.getCArray(), s, (newlength+1)*sizeof(U_CHAR));
    m_container->m_string[newlength] = 0;
    m_container->m_length = newlength;
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

CUtiString::CUtiString() : CUtiOB()
{
   m_container = new CUtiStringContainer(UTEXT(""));
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

CUtiString::CUtiString(const U_CHAR *c, bool is_static) : CUtiOB()
{

#ifdef EDITOR

    parent= NULL;
#endif
   if (is_static) {
     m_container = new CUtiStringContainer(c);
   } else {
   replace(c);
   };
}

void CUtiString::ReadProp(const TCHAR* name, CUtiString* result, const TCHAR* defaultitem)
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    CUtiString str;

    int c = getLineCount();
    int l = utistrlen(name);
    int j;
    for (int i=0; i < c; i++)
    {
        getLine(i, str);
        int p = str.Pos(UTEXT(":"));
        if (p != -1)
        {
            if (l != p)
                continue;
            bool b = true;
            for (j = 0; j < p; j++)
            {
                if (str[j] != name[j])
                {
                    b = false;
                    break;
                };

            };
            //str.Copy(&Name, 0, i);

            //if (wcscmp(Name.getString(),name) == 0) {
            if (b)
            {
                str.Copy(*result, p+1, str.size()-p-1);
                return;
            };
            //};
        };
    };
    *result = defaultitem;
};


CUtiString::CUtiString(const U_CHAR c) : CUtiOB()
{

    U_CHAR s[2];
    s[0] = c;
    s[1] = 0;
    replace(s);
#ifdef DEBUGSTRING
            utiassert(checkContainer());
       #endif
#ifdef EDITOR

    parent= NULL;
#endif
    //	typinfo = UTI_STRING;
}

CUtiString::CUtiString(const CUtiString &s) : CUtiOB()
{
    m_container = const_cast<CUtiStringContainer*>(s.m_container.get());
#ifdef DEBUGSTRING
 utiassert(checkContainer());
#endif

}

CUtiString::~CUtiString()
{

}

CUtiString operator+(const CUtiString &s1, const CUtiString &s2)
	{
		CUtiString temp = const_cast<CUtiString&>(s1);
		 utiassert(temp.checkContainer());
		temp.insert(temp.size(), s2.size(), s2.c_str());
		 utiassert(temp.checkContainer());
		return(temp);
	}
CUtiString operator+(const U_CHAR *s, const CUtiString &s2)
	{

		CUtiString temp = s;
#ifdef DEBUGSTRING
     utiassert(temp.checkContainer());
#endif
		temp.insert(temp.size(), s2.size(), s2.c_str());
#ifdef DEBUGSTRING
     utiassert(temp.checkContainer());
#endif
		return(temp);
	}
CUtiString operator+(const U_CHAR c, const CUtiString &s2)
	{
		CUtiString temp = c;
#ifdef DEBUGSTRING
     utiassert(temp.checkContainer());
#endif
		temp.insert(temp.size(), s2.size(), s2.c_str());
#ifdef DEBUGSTRING
     utiassert(temp.checkContainer());
#endif
		return(temp);
	}

void CUtiString::ReplaceChar(TCHAR oldchar,TCHAR newchar)
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
   utiassert(newchar != 0);
    ownContainer();
    for (int i = 0; i < m_container->m_length; i++)
    {
        TCHAR k=m_container->m_string[i];
        if (k == oldchar)
            m_container->m_string[i] = newchar;

    }
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

CUtiString CUtiString::getLine(int line) const
{
 CUtiString str;
 getLine(line, str);
 return str;
}

void CUtiString::getLine(int line, CUtiString &str) const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    //*str = UTEXT("");
    Array<U_CHAR>& string  = m_container->m_string;
    int i=0;
    while (line != 0 && string[i] != 0)
    {
        if (/*string[i] == 13 ||*/ string[i] == 10)
        {
            line--;
            i++;
            if (string[i] == 10)
                i++;
        }
        else
        {
            i++;
        };

    }
    if (string[i] == 0)
    {
		str = L"";
        return;
    };
    int start = i;
    while (string[i] != 13 && string[i] != 10 && string[i] != 0)
    {
        //  *str += string[i];
        i++;
    }
    ;
    int ende = i;
    Copy(str, start, ende-start);
}

int CUtiString::getLineCount() const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    int i=0;
    int line = 1;
     Array<U_CHAR>& string  = m_container->m_string;
    if (*this == UTEXT(""))
        return(0);
    int s = size();
    while (line != 0 && i < s)
    {
        if (string[i] == 10)
        {
            line++;

        };
        i++;
    }
    //line++;
    return(line);

}

int CUtiString::Pos(const CUtiString& substr)const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    int l = substr.size();
    int length = m_container->m_length;
     Array<U_CHAR>& string  = m_container->m_string;
     if (substr.size() == 0) {
         assert(false);
         return -1;
     }
    if (l == 1)
    {
        for (int i = 0; i < (int)length; i++)
        {
            if (string[i] == substr[0])
            {
                return(i);
            };
        }
    }
    else
    {
        for (int i = 0; i < (int)length-l+1; i++)
        {
            if (wcsncmp(&string[i], substr.c_str(), l)==0)
            {
                return(i);
            };
        }
    };
    return(-1);
}

int CUtiString::Pos(int start, const CUtiString& substr)const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    int l = substr.size();
    int length = m_container->m_length;
     Array<U_CHAR>& string  = m_container->m_string;
    if (l == 1)
    {
        for (int i = start; i < (int)length; i++)
        {
            if (string[i] == substr[0])
            {
                return(i);
            };
        }
    }
    else
    {
        for (int i = start; i < (int)length-l+1; i++)
        {
            if (wcsncmp(&string[i], substr.c_str(), l)==0)
            {
                return(i);
            };
        }
    }
    return(-1);
}

int CUtiString::LastDelimiter(const U_CHAR* substr) const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    int l = utistrlen(substr);
    int length = m_container->m_length;
    int lastpos = length;
     Array<U_CHAR>& string  = m_container->m_string;
    for (int i = 0; i < length; i++)
    {
        for (int j = 0; j < l; j++)
        {
            if (string[i] == substr[j])
            {
                lastpos = i;
            }
        };
    };
    return lastpos;
}

int CUtiString::NextDelimiter(const U_CHAR* substr) const
{
   int l = utistrlen(substr);
   int length = m_container->m_length;
//   int lastpos = length;
    Array<U_CHAR>& string  = m_container->m_string;
   for (int i = 0; i < length; i++)
   {
      for (int j = 0; j < l; j++)
      {
         if (string[i] == substr[j])
         {
            return i;
         }
      };
   };
   return -1;
}


int CUtiString::NextDelimiter(int startpos, const U_CHAR* substr) const
{
   int l = utistrlen(substr);
   int length = m_container->m_length;
 //  int lastpos = length;
    Array<U_CHAR>& string  = m_container->m_string;
   for (int i = startpos; i < length; i++)
   {
      for (int j = 0; j < l; j++)
      {
         if (string[i] == substr[j])
         {
            return i;
         }
      };
   };
   return -1;
}

void CUtiString::Copy(CUtiString& target, int start, int l) const
{
    //*target = UTEXT("");
    utiassert(l >= 0);
    int end = start+l;
    int length = m_container->m_length;
    if (end >= length)
        end = length;

    length = end-start;
    if (length < 0) length = 0;
    target.setLength(length);

    target.ownContainer();
    m_container->m_string.copyToArray(start, length, target.m_container->m_string, 0);
    target.m_container->m_string[length] = 0;
    //memcpy(target.m_container->m_string, m_container->m_string, length*sizeof(U_CHAR));
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

CUtiString CUtiString::Trim()
{
    return this->TrimLeft().TrimRight();
}

CUtiString CUtiString::TrimLeft()
{
    int deleteamount = 0;

    for (int i = 0; i < m_container->m_length; i++) {
        if ((*this)[i] != ' ') {
            break;
        }
        deleteamount++;
    }
    if (deleteamount != 0) {
    CUtiString copy(*this);
    copy.Delete(0, deleteamount);
    return copy;
    }
    return *this;
}

CUtiString CUtiString::TrimRight()
{
    int deleteamount = 0;

    for (int i = m_container->m_length-1; i >= 0; i--) {
        if ((*this)[i] != ' ') {
            break;
        }
        deleteamount++;
    }
    if (deleteamount != 0) {
    CUtiString copy(*this);
    copy.Delete(m_container->m_length-deleteamount, deleteamount);
    return copy;
    } else {
        return *this;
    }
}

void CUtiString::Delete(int index, int length)
{

#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
  ownContainer();
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
   if (length == 0)return;
  if (index >= m_container->m_length) return;
  if (index < 0) return;
  if (index+length >= m_container->m_length) {
    length = m_container->m_length - index;
  };
     Array<U_CHAR>& c  = m_container->m_string;
    int l = 0;
    /*for (int i = 0; i < index; i++)
    {
        c[l] = string[i];
        l++;
    };*/
    l += index;
    int l2 = size();
    for (int i = index+length; i < l2; i++)
    {
        c[l] = c[i];
        l++;
    };
    c[l] = 0;
    m_container->m_length -= length;
    utiassert(m_container->m_length == l);
    m_container->checkSize();
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}


HResult CUtiString::Read(CStream & Stream, int Flags, int version)
{
    /*int i;
    Stream.ReadInteger(i);

    char* c = new char[i+1];
    Stream.ReadData(c, (i+1)*sizeof(char));
    //replace(c);
    LoadAnsi(c);
    delete [] c;*/
    utiassert(false);
    return(UTI_OK);
}

HResult CUtiString::Write(CStream & Stream, int Flags, int version)
{
    //Stream.WriteInteger(length);
    //Stream.WriteData(string, length+1);
    utiassert(false);
    return(UTI_OK);
}

CUtiString& CUtiString::operator =(const CUtiString &s)
{
    //if (this == &s)
    //    return(*this);
    m_container = const_cast<CUtiString&>(s).m_container;
    //replace(s.string);
    return(*this);
}

U_CHAR& CUtiString::operator[](unsigned int p)
{
     ownContainer();
    utiassert((int)p>=0 && (int)p<=m_container->m_length);
    return(m_container->m_string[p]);
}

const U_CHAR& CUtiString::operator[](unsigned int p) const
{
    utiassert(p>=0 && p<=m_container->m_length);
    return(m_container->m_string[p]);
}

CUtiString& CUtiString::operator =(const U_CHAR *s)
{
    if (s == NULL)
        return(*this);
    replace(s);
    return(*this);
}

CUtiString& CUtiString::operator =(const U_CHAR c)
{
    UTI_DECL_ARRAY(U_CHAR, s, 2);
    s.fill(0);
    s[0] = c;
    s[1] = 0;
    replace(s.getCArray());
    return(*this);
}

CUtiString& CUtiString::operator =(int i)
{
    UTI_DECL_ARRAY(U_CHAR, s, 40);
    s.fill(0);
    swprintf(s.getCArray(), 40, UTEXT("%i"), i);

    replace(s.getCArray());
    return(*this);
}

CUtiString& CUtiString::operator =(float i)
{
    UTI_DECL_ARRAY(U_CHAR, s, 40);
    s.fill(0);
    swprintf(s.getCArray(), 40,UTEXT("%.6f"), i);

    replace(s.getCArray());
    return(*this);
}


CUtiString& CUtiString::operator =(utifloat i)
{
    UTI_DECL_ARRAY(U_CHAR, s, 40);
    s.fill(0);
    float f = i;
    swprintf(s.getCArray(), 40,UTEXT("%.6f"), f);

    replace(s.getCArray());
    return(*this);
}


CUtiString& CUtiString::operator+=(const CUtiString &s)
{
    insert(m_container->m_length, s.size(), s.c_str());
    return(*this);
}

CUtiString& CUtiString::operator+=(const U_CHAR *s)
{
    insert(m_container->m_length, utistrlen(s), s);
    return(*this);
}

CUtiString& CUtiString::operator+=(const U_CHAR c)
{
    insert(m_container->m_length, 1, &c);
    return(*this);
}

CUtiString& CUtiString::operator+=(const int i)
{
    UTI_DECL_ARRAY(U_CHAR, s, 40);
    s.fill(0);

    swprintf(s.getCArray(), 40,UTEXT("%i"), i);
    insert(m_container->m_length, utistrlen(s.getCArray()), s.getCArray());
    return(*this);
}

CUtiString& CUtiString::operator+=(const float i)
{
    UTI_DECL_ARRAY(U_CHAR, s, 40);
    s.fill(0);
    swprintf(s.getCArray(), 40,UTEXT("%f"), i);
    insert(m_container->m_length, utistrlen(s.getCArray()), s.getCArray());
    return(*this);
}


CUtiString& CUtiString::operator+=(const double i)
{
    UTI_DECL_ARRAY(U_CHAR, s, 40);
    s.fill(0);
    float f= i;
    swprintf(s.getCArray(), 40,UTEXT("%f"), f);
    insert(m_container->m_length, utistrlen(s.getCArray()), s.getCArray());
    return(*this);
}


CUtiString CUtiString::operator+(const U_CHAR *s) const
{
    CUtiString temp = *this;
    temp.insert(m_container->m_length, utistrlen(s), s);
    return(temp);
}

CUtiString CUtiString::operator+(const U_CHAR c) const
{
    if (c == 0) return *this;
    CUtiString temp = *this;
    temp.insert(m_container->m_length, 1, &c);
    return(temp);
}

CUtiString CUtiString::operator+(const int i) const
{
    CUtiString temp = *this;
    UTI_DECL_ARRAY(U_CHAR, s, 40);
    s.fill(0);
    swprintf(s.getCArray(),40, UTEXT("%i"), i);

    temp.insert(m_container->m_length, utistrlen(s.getCArray()), s.getCArray());
    return(temp);
}

void CUtiString::DecLength(int i)
{
    ownContainer();
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    if (m_container->m_length-i >= 0)
    {
        m_container->m_string[m_container->m_length-i] = 0;
        m_container->m_length -= i;
	m_container->checkSize();
    };
};

void CUtiString::ExpandToSize(int s)
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
  if (s == size()) return;
  if (s < size()) {
    return;
  }
  int ex = s-size();
  CUtiString str;
  str.setLength(ex);
  for (int i = 0;i < ex;i++) {
    str[i] = L' ';
  };
  CUtiString s1(*this);
  *this = str + s1;
}


CUtiString CUtiString::operator+(const float i) const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    CUtiString temp = *this;
    UTI_DECL_ARRAY(U_CHAR, s, 40);
    s.fill(0);
    swprintf(s.getCArray(),40, UTEXT("%f"), i);

    temp.insert(m_container->m_length, utistrlen(s.getCArray()), s.getCArray());
    return(temp);
}

CUtiString CUtiString::operator+(const double f) const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
      CUtiString temp = *this;
      UTI_DECL_ARRAY(U_CHAR, s, 40);
      float i = f;
      s.fill(0);
      swprintf(s.getCArray(),40, UTEXT("%f"), i);

      temp.insert(m_container->m_length, utistrlen(s.getCArray()), s.getCArray());
      return(temp);
}

/*int CUtiString::calcbuffersize(int len)
{
    int minlen = len+1;
    int k = 4;
    for (int i = 0; i < 20;i++)
    {
        if (minlen <= k)
            return k;
        k *= 2;

    }
    return len+1+15;
}*/

void CUtiString::makeThreadResistant()
{
  ownContainer();
}


bool CUtiString::startsWith(const CUtiString& other) const
{
   int s = size();
   int o = other.size();
   if (o > s) return false;

   for (int i = 0; i < o; i++) {
       if (other[i] != operator[] (i)) {
           return false;
       }
   }
   return true;
}

bool CUtiString::ReplaceString(const CUtiString& oldone, const CUtiString& newone)
{
    int oldlength = oldone.size();
    int newlength = newone.size();
    int currentpos=0;
    bool found = false;
    while (true) {

        int nextreplace = Pos(currentpos, oldone);
        if (nextreplace == -1) return found;
        Delete(nextreplace, oldlength);
        insert(nextreplace, newone);
        currentpos = nextreplace+newlength;
        found = true;
    }

    return found;
}

void CUtiString::insert(unsigned int pos, const CUtiString& str)
{
    insert(pos, str.size(), str.c_str());
}

void CUtiString::insert(unsigned int pos, unsigned int slen, const TCHAR *s)
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    utiassert(pos <= m_container->m_length);
    int length = m_container->m_length;
    int newlength = length+slen;
    ownContainer(m_container->m_length+slen);
    Array<U_CHAR>& string = m_container->m_string;
            for (int i = length-pos-1; i >= 0; i--)
            {
                string[pos+slen+i] = string[pos+i];
            };
            for (int i = 0; i < slen; i++)
            {

                string[pos+i] = s[i];
            };

            string[newlength] = 0;
            m_container->m_length = newlength;
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif

}

/*void CUtiString::ReadProp(char* name, CUtiString* result)
{
   CUtiString str;
   CUtiString Name;
   CUtiString inhalt;
   int c = getLineCount();
   for (int i=0; i < c; i++) {
	  getLine(i, &str);
	  int i = str.Pos(UTEXT(":"));
	  if (i != -1) {
		 str.Copy(&Name, 0, i);
		 if (wcscmp(Name.getString(),name) == 0) {
			 str.Copy(result, i+1, str.getlength()-i-1);
		 };
	  };
   };
};*/

const U_CHAR* CUtiString::getString() const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
  //CUtiStringContainer* c = const_cast<CUtiStringContainer*>(m_container.get());
  /* if (c->m_string.isDynamic()) {
  c->m_string[c->m_length] = 0;
   }*/
    return(m_container->m_string.getCArray());
}

const U_CHAR* CUtiString::c_str() const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif

  /*CUtiStringContainer* c = const_cast<CUtiStringContainer*>(m_container.get());
  if (c->m_string.isDynamic()) {
  c->m_string[c->m_length] = 0;
  }*/
    return(m_container->m_string.getCArray());
}

void CUtiString::IntToStr(int i, U_CHAR *str)
{
    utiassert(false);
   // swprintf(str, 40,UTEXT("%i"), i);
    //_itoa(i, str, 10);
}


void CUtiString::IntToStr(int i, const CUtiString& format)
{
    utiassert(checkContainer());
    UTI_DECL_ARRAY(U_CHAR, buffer, 80);
   buffer.fill(0);
      swprintf(buffer.getCArray(),80, format.c_str(), i);
      replace(buffer.getCArray());
      utiassert(checkContainer());
}

CUtiString CUtiString::intToStr(int i, const CUtiString& format)
{
   CUtiString str;
   str.IntToStr(i, format);
   return str;
}

void CUtiString::IntToStr(int i)
{
  utiassert(checkContainer());
  UTI_DECL_ARRAY(U_CHAR, buffer, 40);
 buffer.fill(0);
    swprintf(buffer.getCArray(),40, UTEXT("%i"), i);
    replace(buffer.getCArray());
    utiassert(checkContainer());

}


int CUtiString::strToInt(const CUtiString& str)
{
    return str.ToInt();
}

 CUtiString CUtiString::intToStr(int i)
{
    CUtiString res;
    res.IntToStr(i);
    return res;
}

 CUtiString CUtiString::intToStrZero(int i, int zeros)
 {
     CUtiString res;
     res.IntToStr(i);
     while (res.size() < zeros) {
         res = UTISTR("0")+res;
     }
     return res;
 }

void CUtiString::FloatToStr(const float i)
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
  UTI_DECL_ARRAY(U_CHAR, buffer, 40);
 buffer.fill(0);
    swprintf(buffer.getCArray(), 40,UTEXT("%f"), i);
    replace(buffer.getCArray());
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif

}

void CUtiString::FloatToStr(const float i, int width, int precision)
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    UTI_DECL_ARRAY(U_CHAR, buffer, 40);
   buffer.fill(0);
   UTI_DECL_ARRAY(U_CHAR, maskbuffer, 40);
  maskbuffer.fill(0);
    swprintf(maskbuffer.getCArray(), 40, UTEXT("%%%i.%if"), width, precision);
    swprintf(buffer.getCArray(), 40,maskbuffer.getCArray(), i);
    replace(buffer.getCArray());
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif

}


void CUtiString::ExtractRealFileName()
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    ExtractFileName();
    utiassert(checkContainer());
    int i = Pos(UTEXT("."));
    if (i == -1) {return;};
    CUtiString s;
    Copy(s, 0, i);
    *this = s;
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}


void CUtiString::ExtractFileNameWithoutExt()
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    utiassert(checkContainer());
    int i = LastDelimiter(UTEXT("."));
    if (i == -1) {return;};
    CUtiString s;
    Copy(s, 0, i);
    *this = s;
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

void CUtiString::ExtractFileName()
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    int i = LastDelimiter(UTEXT("\\:/"));
    if (Pos(UTEXT("\\")) == -1 && Pos(UTEXT(":")) == -1 && Pos(UTEXT("/")) == -1)
        return;
    CUtiString s;
    Copy(s, i+1, 100000);
    *this = s;
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

void CUtiString::ExtractFileExt()
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    ExtractFileName();
    utiassert(checkContainer());
    int i = LastDelimiter(UTEXT("."));
    CUtiString s;
    Copy(s, i+1, 1000000);
    *this = s;
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

void CUtiString::ExtractFilePath()
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    int i = LastDelimiter(UTEXT("\\:/"));
    utiassert(checkContainer());
    if (i == m_container->m_length)
    {
        *this = UTEXT("");
    }
    else
    {
        CUtiString s;
        Copy(s, 0, i);
        *this = s;
    }
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

CUtiString CUtiString::LowerCase() const
{
   CUtiString s = *this;
   s.ToLowerCase();
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
   return s;
}

CUtiString CUtiString::UpperCase()const
{
   CUtiString s = *this;
   s.ToUpperCase();
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
   return s;
}
void CUtiString::ToLowerCase()
{
  ownContainer();
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
  Array<U_CHAR>& string = m_container->m_string;
   for (int i = 0; i < m_container->m_length; i++) {
      string[i] = towlower(string[i]);
   };
}

void CUtiString::ToUpperCase()
{
  ownContainer();
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
  Array<U_CHAR>& string = m_container->m_string;
   for (int i = 0; i < m_container->m_length; i++) {
      string[i] = towupper(string[i]);
   };
}


SCode CUtiString::convertToCode()
{

    Array<U_CHAR>& string = m_container->m_string;
  std::string res;
  int s = size();
  res.resize(s);

   for (int i = 0; i < s; i++)
   {
       res[i] =string[i];
   }
    SHA1 checksum1;
    checksum1.update(res);
    const std::string hash = checksum1.final();
    SCode res1;
    memcpy(&res1, checksum1.resultdigest, sizeof(res1));

    res1.Data3 &= 0x0FFF;
    res1.Data3 |= 5 << (8+4);
    return res1;

}

void CUtiString::LoadRawString(const wchar_t* c, int l)
{
    setLength(l);
    for (int i = 0; i < l; i++) {
       (*this)[i] = c[i];
    };
}


void CUtiString::MakeFile()
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
  ownContainer();
 Array<U_CHAR>& string  = m_container->m_string;
    int l = size();
    for (int i = 0; i < l; i++)
    {
        if (string[i] == UTEXT('/') || string[i] == UTEXT('\\') || string[i] == UTEXT(':')|| string[i] == UTEXT('*')
                || string[i] == UTEXT('?')|| string[i] == UTEXT('"')|| string[i] == UTEXT('<')
                || string[i] == UTEXT('>')|| string[i] == UTEXT('|'))
        {
            string[i] = UTEXT('_');


        }
    };
    utiassert(checkContainer());
}

void CUtiString::IntToStrRadix(int i, U_CHAR *str,int length, int radix)
{
    swprintf(str, length,  UTEXT("%i"), i);
    //_itoa(i, str, radix);
}

void CUtiString::UIntToHex(unsigned int i, int minlength)
{
    UTI_DECL_ARRAY(U_CHAR, s, 48);
   s.fill(0);
   switch (minlength) {
     case 0: {
        swprintf(s.getCArray(), 40,  UTEXT("%.0X"), i);
     }; break;
     case 1: {
        swprintf(s.getCArray(), 40,  UTEXT("%.1X"), i);
     }; break;
     case 2: {
        swprintf(s.getCArray(), 40,  UTEXT("%.2X"), i);
     }; break;
     case 3: {
        swprintf(s.getCArray(), 40,  UTEXT("%.3X"), i);
     }; break;
     case 4: {
        swprintf(s.getCArray(), 40,  UTEXT("%.4X"), i);
     }; break;
     case 5: {
        swprintf(s.getCArray(), 40,  UTEXT("%.5X"), i);
     }; break;
     case 6: {
        swprintf(s.getCArray(), 40,  UTEXT("%.6X"), i);
     }; break;
     case 7: {
        swprintf(s.getCArray(), 40,  UTEXT("%.7X"), i);
     }; break;
     case 8: {
        swprintf(s.getCArray(), 40,  UTEXT("%.8X"), i);
     }; break;
   };
   *this = s.getCArray();
   /*while (size() < minlength) {
     *this = UTEXT("0")+*this;
   };*/
}

int CUtiString::ToInt() const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    return(wcstol(m_container->m_string.getCArray(), 0, 10));
}

int CUtiString::HexToInt() const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
   return(wcstol(m_container->m_string.getCArray(), 0, 16));
}

double CUtiString::ToFloat() const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    U_CHAR *stopstring;
    return(wcstod( m_container->m_string.getCArray(), &stopstring ));
}

void CUtiString::SaveToFile(const U_CHAR* filename)
{
  utiassert(checkContainer());
    CUtiString namestr(filename);

if (false) {
    char* ansiname = namestr.GetAnsi();
    wofstream s(ansiname);
    delete [] ansiname;
    s << getString();
    s.flush();
    s.close();
} else {
    std::string str = converttomulti(this->c_str());
    char* ansiname = namestr.GetAnsi();
    ofstream s(ansiname);
    delete [] ansiname;
    s << str;
    s.flush();
    s.close();
}


};

void CUtiString::SaveToFile(const CUtiString &filename)
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    SaveToFile(filename.getString());
};

void CUtiString::LoadFromFile(const CUtiString &filename)
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    LoadFromFile(filename.getString());
}

void SkipBOM(std::ifstream & in)
{
    char test[3] = {0};
    in.read(test, 3);
    if ((unsigned char) test[0] == 0xEF &&
            (unsigned char) test[1] == 0xBB &&
            (unsigned char) test[2] == 0xBF)
    {
        return;
    }
    in.seekg(0);
}

void CUtiString::LoadFromFile(const U_CHAR* filename)
{
    CheckFileName(filename);
    if (!UtiFileExists(filename)) {
        wcout << "Could not load the file: " << filename << "" << std::endl;
        return;

     }
    UTI_INT64 size = UtiFileSize(filename);
    if (size < 0) {
       wcout << "Could not load the file: " << filename << "" << std::endl;
       return;

    }
   /* char* c = new char[size+1];
    CStream* str = UtiOpenFile(filename, otRead);
    str->ReadData(c, size);
    c[size] = 0;*/
      CUtiString fn2 = filename;

      wchar_t ch = fn2[0];
      if (ch == L'§' || ch == 167) {
          CUtiString substr;
          fn2.Copy(substr, 1, fn2.size() - 1);
         fn2 = substr;
         //Besser!
      }
    std::string fn = UtiFramework::converttomulti(fn2.c_str());

     std::ifstream t(fn.c_str());

     if (!t.good()) {
         std::cout << "Could not load the file: " << fn << std::endl;
     }
    std::string str;


    t.seekg(0, std::ios::end);

    if (!t.fail()) {

    str.reserve(umax(0,int(t.tellg())));


     t.seekg(0, std::ios::beg);
    SkipBOM(t);





    str.assign((std::istreambuf_iterator<char>(t)),
            std::istreambuf_iterator<char>());
    }



   //cout << c;
    this->LoadAnsi(str.c_str());

    //delete str;
    //delete [] c;
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif

    /*	ifstream s(filename);
        bool first = true;
    	while (!s.eof()) {
    		if (first) {
    			first = false;
    		} else {
    			U_CHAR c=13;
    			*this += c;
    			c =10;
    			*this += c;
    		};
    		U_CHAR c[1024];
    		s.getline(c, 1024);
    		*this += c;
    	};*/
};

/*void ConvertAnsiStringToWide( WCHAR* wstrDestination, const CHAR* strSource, size_t cchDestChar )
{
    if( wstrDestination==NULL || strSource==NULL )
        return;

    if( cchDestChar == -1 )
        cchDestChar = strlen(strSource)+1;

    MultiByteToWideChar( CP_ACP, 0, strSource, -1,
                         wstrDestination, cchDestChar-1 );

    wstrDestination[cchDestChar-1] = 0;
}
*/
void CUtiString::LoadAnsi(const char* c)
{
    /*size_t size = strlen(c);
    TCHAR* cw = new TCHAR[size+3];
    ConvertAnsiStringToWide(cw, c, size+2);
    *this = cw;
    delete [] cw;*/
    *this = converttowide(c);
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
}

char* CUtiString::GetAnsi() const
{
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
    std::string r = converttomulti(this->c_str());

    char* c = new char[r.length()+1];
    //	int l=length+1;
    //WideCharToMultiByte(CP_UTF8, 0, string,length, c, l,NULL, NULL);
    //	c[length] = 0;
    strcpy(c, r.c_str());
    return c;
};

bool isNumeric(const CUtiString& str)
{
  for (int i = 0; i < str.size(); i++) {
    if (!isNumeric(str[i])) return false;
  };
  return true;
}

bool isBezeichner(const CUtiString& name)
{
	if (name.size() == 0) return false;
	if (!isUnderLine(name[0]) && !isAlpha(name[0])) {
		return(false);
	};
    for (int i = 0; i < name.size(); i++) {
		if (!isBezeichner(name[i])) return false;
    };
	return(true);
};


bool isMultiBezeichner(const CUtiString& name)
{
int start = 0;
CUtiString temp;
    for (int i = 0; i < name.size(); i++) {
	if (name[i] == '.') {
	  name.Copy(temp, start, i);
	  if (!isBezeichner(temp)) return false;
	  start = i;
	};
    };
    if (start == name.size()-1) return false;
	return(true);
};


void CUtiString::AddPascalString(U_CHAR* str, int length)
{
    if (length < 50)
    {
        U_CHAR c[52];
        for (int i =0; i < length; i++)
        {
            c[i] = str[i];
        };
        c[length]=0;
        *this += c;

    }
    else if (length < 256)
    {
        U_CHAR c[258];
        for (int i =0; i < length; i++)
        {
            c[i] = str[i];
        };
        c[length]=0;
        *this += c;

    }
    else
    {
        U_CHAR* c = new U_CHAR[length+1];

        for (int i =0; i < length; i++)
        {
            c[i] = str[i];
        };
        c[length]=0;
        *this += c;
        delete [] c;
    };
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
};


float CUtiString::strToFloat(const CUtiString& str)
{
float value = wcstof(str.c_str(), NULL);
return value;
}

CUtiString CUtiString::floatToStr(const float i, int width, int precision)
{
    CUtiString str;
    str.FloatToStr(i, width, precision);
    return str;
}

void CUtiString::assignSimpleAnsi(const char* data, int size)
{
    setLength(size);
    wchar_t* t = m_container->m_string.getCArray();
    for (int i = 0; i < size; i++) {
        t[i] = data[i];
    }
}

char* CUtiString::createSimpleAnsi()
{

    int s = size();
    char* data = new char[s+1];
    wchar_t* t = m_container->m_string.getCArray();
    for (int i = 0; i < s; i++) {
        data[i] = t[i];
    }
    data[s] = 0;
    return data;

}

void CUtiString::SetFloat(double f, int count)
{
    /*U_CHAR *buf;
    int dec, s;
    double r = f;
    if (r < 0) r = -r;
    if(r >= 1) count++;if (r >= 10) count++;if (r >= 100) count++;if (r >= 100000000) count++;if (r >= 1000) count++;
    if (r >= 10000) count++;if (r >= 100000) count++; if (r >= 1000000) count++;if (r >= 10000000) count++;

    buf  = _ecvt(f, count, &dec, &s);
    replace(buf);
    U_CHAR c[2];
    c[0] = 46;
    c[1] = 0;

    insert(dec, 1, c);
    U_CHAR *c1;
    c1 = UTEXT("-");
    if (s) {
     insert(0, 1, c1);
    };*/
    UTI_DECL_ARRAY(U_CHAR, buffer, 40);
   buffer.fill(0);
    //float f1=f;
    swprintf(buffer.getCArray(), 40, UTEXT("%f"), f);
    replace(buffer.getCArray());

}


void ReadStringWide(CStream & Stream, CUtiString &s)
{
    int count;
    Stream.ReadInteger(count);
    //UTI_DECL_ARRAY(wchar_t, t, count+3);
	Array<wchar_t> t(count+3);
    t.fill(0);
    //wchar_t *t = new wchar_t[count+1];
    if (sizeof(wchar_t) == 2) {
        Stream.ReadData(t.getCArray(), count*sizeof(wchar_t));
    } else {
       for (int i = 0; i < count; i++) {
          t[i] = 0;
          wchar_t tres=0;
          Stream.ReadData(&tres, 2);
          t[i] = tres;
       };
    };
    t[count] = 0;
#ifdef _UNICODE

    s = t.getCArray();
#else
#ifdef WIN32
    char* c = new char[count+1];
    WideCharToMultiByte(CP_UTF8, 0, t, count, c, count,NULL, NULL);
    s = c;
    delete [] c;
#endif
#endif

   // delete []t;
}

void WriteStringWide(CStream & Stream, const CUtiString &s)
{
    int count = s.size();
    Stream.WriteInteger(count);
#ifdef _UNICODE
    if (sizeof(wchar_t) == 2) {
        Stream.WriteData(s.getString(), count*2);
    } else {
       const wchar_t* k = s.getString();
       for (int i = 0;i < count; i++) {
          Stream.WriteData(&k[i], 2);
       };

    };
#else
#ifdef WIN32
    wchar_t *t = new wchar_t[count+1];
    MultiByteToWideChar(CP_UTF8, 0, s.getString(), count, t, count+1);
    Stream.WriteData(t, count*2);
#endif
#endif

}

void ReadStringMulti(CStream & Stream, std::string &s)
{
    int count;
    Stream.ReadInteger(count);
    //UTI_DECL_ARRAY(wchar_t, t, count+3);
    Array<char> t(count+3);
    t.fill(0);

        Stream.ReadData(t.getCArray(), count*sizeof(char));

    t[count] = 0;

    s = t.getCArray();

}

void WriteStringMulti(CStream & Stream, const std::string &s)
{
    int count = s.size();
    Stream.WriteInteger(count);
    if (sizeof(wchar_t) == 2) {
        Stream.WriteData(s.c_str(), count);
    }

}


const CUtiString CodeToFileNameString(const SCode s2)
{
    CUtiString s;
    s += UTEXT("_X_");
    CUtiString s1;
    s1.UIntToHex(s2.Data1, 8);
    s += s1;
    s += UTEXT("_");
    s1.UIntToHex(s2.Data2 & 0xFFFF, 4);
    s += s1;
    s += UTEXT("_");
    s1.UIntToHex(s2.Data3 & 0xFFFF, 4);
    s += s1;
    s += UTEXT("_");
    s1.UIntToHex((s2.Data4[0]), 2);
    s += s1;
    s1.UIntToHex((s2.Data4[1]), 2);
    s += s1;
    s += UTEXT("_");


    s1.UIntToHex((s2.Data4[2]), 2);
    s += s1;
    s1.UIntToHex((s2.Data4[3]), 2);
    s += s1;

    s1.UIntToHex((s2.Data4[4]), 2);
    s += s1;
    s1.UIntToHex((s2.Data4[5]), 2);
    s += s1;


    s1.UIntToHex((s2.Data4[6]), 2);
    s += s1;
    s1.UIntToHex((s2.Data4[7]), 2);
    s += s1;

    s += UTEXT("_X_");


      return s;
}

const CUtiString CodeToString(const SCode s2)
{
  CUtiString s;
  s += UTEXT("{");
  CUtiString s1;
  s1.UIntToHex(s2.Data1, 8);
  s += s1;
  s += UTEXT("-");
  s1.UIntToHex(s2.Data2 & 0xFFFF, 4);
  s += s1;
  s += UTEXT("-");
  s1.UIntToHex(s2.Data3 & 0xFFFF, 4);
  s += s1;
  s += UTEXT("-");
  s1.UIntToHex((s2.Data4[0]), 2);
  s += s1;
  s1.UIntToHex((s2.Data4[1]), 2);
  s += s1;
  s += UTEXT("-");
  
  
  s1.UIntToHex((s2.Data4[2]), 2);
  s += s1;
  s1.UIntToHex((s2.Data4[3]), 2);
  s += s1;
  
  s1.UIntToHex((s2.Data4[4]), 2);
  s += s1;
  s1.UIntToHex((s2.Data4[5]), 2);
  s += s1;
  
  
  s1.UIntToHex((s2.Data4[6]), 2);
  s += s1;
  s1.UIntToHex((s2.Data4[7]), 2);
  s += s1;
  
  s += UTEXT("}");
   /* wchar_t t[100];
    #ifdef _WIN32
    StringFromGUID2(s2, t, 100);
    #else
    CUtiString z;
    return z;
    #endif
    int count = 95;
    CUtiString s;

#ifdef _UNICODE

    s = t;
#else
#ifdef WIN32
    char* c = new char[count+1];
    WideCharToMultiByte(CP_UTF8, 0, t, count, c, count,NULL, NULL);
    s = c;
    delete [] c;
#else

#endif
#endif*/

    return s;
}

bool utimatch(const wchar_t* pattern, const wchar_t* name)
{
   //const wchar_t* pattern = pattern_;
   //const wchar_t* name = name_;

   while (*pattern != 0 && *name != 0) {
      if (*pattern == L'*') {
         while (pattern[1] == L'*') pattern++;
         if (name[0] == pattern[1]) {
            const wchar_t* patt2 = pattern;
            //const wchar_t* name2 = name;
            patt2++;
            patt2++;
            name++;
            if (utimatch(patt2, name)) return true;
         };
      } else if (*pattern == L'?') {
         pattern++;

      } else {
         if (*pattern != *name) return false;
         pattern++;
      };
      name++;
   };
   if (*pattern == 0 && *name == 0) return true;
   return false;

};

/*int LastDelimiter(wstring mask, wstring str)
{
   int pos = -1;
   for (size_t i = 0; i < str.size(); i++) {
       for (size_t j = 0; j < mask.size(); j++) {
          if (str[i] == mask[j]) {
              pos = i;
              break;
}
}
}
   return pos;
}

wstring ExtractFilePath(wstring path)
{
    int p = LastDelimiter(L"/\\:", path);
    if (p==-1)return path;
    return path.substr(0, p);
};

wstring ExtractFileName(wstring path)
{
    int p = LastDelimiter(L"/\\:", path);
    if (p==-1)return path;
    return path.substr(p+1, path.size()-p-1);
};

wstring ExtractFileExt(wstring path)
{
    int p = LastDelimiter(L".", path);
    if (p==-1)return path;
    return path.substr(p, path.size()-p);

};

bool FileExists(wstring filename)
{

  struct stat buffer;
  int k = stat(converttomulti(filename.c_str()).c_str(), &buffer);
  if (k == ENOENT) return false;
  return true;

}

void CheckError(int k, const wstring &p)
{

   if (k == EROFS) {
      wcout << L"Readonly Device: " << p << endl;

   }
   if (k == ENOSPC) {
      wcout << L"Zu wenig Speicherplatz: " << p << endl;

   }
}

void ForceDir(const wstring &path)
{
 //   wcout << "ForceDir(" << path << L")";
   wstring p=ExtractFilePath(path);

   if (!FileExists(p))
   {
      ForceDir(p);

   }
   if (ExtractFileName(path) != L"") {
#ifdef WIN32
      string p1 = converttomulti(path.c_str());
      int k = _mkdir(p1.c_str());
      CheckError(k, path);
#else
      string p1 = converttomulti(path.c_str());
      int k = mkdir(p1.c_str(),S_IRWXU | S_IRWXG | S_IRWXO);
      CheckError(k, path);
#endif
   }
}




class PathInfo
{
   public:
      vector<wstring> dirs;
};

int finddelim(const wstring p)
{
   int pos = p.find(L"/", 0);
   if (pos != -1) return pos;
   pos = p.find(L"\\", 0);
   if (pos != -1) return pos;
   return p.size();
}

const PathInfo pathtoinfo(const wstring path1)
{
   wstring path = path1;
   int pos;
   PathInfo info;
   while ((pos = finddelim(path)) != 0) {
      info.dirs.push_back(path.substr(0, pos));
      wcout << path.substr(0, pos);
      wcout << endl;
      path = path.erase(0, pos+1);
   };
   return info;
};

const wstring ExtractRelativePath(const wstring BaseName, const wstring DestName)
{
   wcout << L"BaseName: " << BaseName << " DestName: " << DestName << endl;
   PathInfo baseinfo = pathtoinfo(BaseName);
   PathInfo destinfo = pathtoinfo(DestName);
   int k=-1;

   for (int i = 0; i < min(baseinfo.dirs.size(), destinfo.dirs.size()); i++) {


      if (baseinfo.dirs[i] != destinfo.dirs[i]) break;
      k =i;

   }

   int rest = baseinfo.dirs.size()-k-1;

   wstring result=L"";
   for (int i = 0; i < rest; i++) {
      result += L"../";

   }

   int depthcount = destinfo.dirs.size()-k;
   for (int i = k+1; i < min((size_t)depthcount, destinfo.dirs.size()); i++) {
      result += destinfo.dirs[i];
      if (i !=min((size_t)depthcount, destinfo.dirs.size())-1) result += L"/";
   }

   return result;
}
*/


CStrings::CStrings()
{

}


void CStrings::addCutLine(const CUtiString& str)
{
    CStringList l;
    l.setText(str);
    for (int i = 0; i <l.size(); i++)
    {
        add(l.getLine(i));
    }
}

void CStrings::saveToFile(const CUtiString& filename)
{
    CUtiString str = getText();
    str.SaveToFile(filename);
}

void CStrings::loadFromFile(const CUtiString& filename)
{
CUtiString str;
str.LoadFromFile(filename);
setText(str);
}


void CStrings::addStrings(const CStrings& str)
{
    for (int i = 0; i < str.size(); i++) {
        add(str.getLine(i));
    }
}

int CStrings::indexOf(const CUtiString& str)
{
    for (int i = 0; i < size(); i++) {
        if (getLine(i) == str)return i;
    }
    return -1;
}

int CStrings::indexOfName(const CUtiString& str)
{
    for (int i = 0; i < size(); i++) {
        CUtiString str2 = getLine(i);
        int p = str2.Pos(0, UTISTR(":"));
        if (p != -1) {
            if (str2(0, p-1) == str) return i;
        } else {
           if (str2 == str)return i;
        }
    }
    return -1;
}

int CStrings::exchange(int i, int j)
{
    CUtiString str = getLine(i);
    setLine(i, getLine(j));
    setLine(j, str);
	return 0;
}


void CStrings::printTextToStdOut()
{
       for (int i = 0; i < size(); i++) {
           std::wcout << std::setfill(L'0') << std::setw(3)<< i << L": " << getLine(i) << std::endl;
       }
}

CUtiString CStrings::getText()
{
   CUtiString res;
   CUtiString ret(UTISTR("\n"));
   int l = 0;
   int ts = size();
   for (int i = 0; i < ts; i++) {
       CUtiString s = getLine(i);
       l += s.size();
       if (i != ts-1) {
           l += ret.size();
       }
   }
   res.setLength(l);
   l = 0;
   for (int i = 0; i < ts; i++) {
       CUtiString s = getLine(i);
       int linesize = s.size();
       for (int j = 0; j < linesize; j++) {
           res[l] = s[j];
           l++;
       }
       if (i != ts-1) {
           for (int j = 0; j < ret.size(); j++) {
               res[l] = ret[j];
               l++;
           }

       }
   }
   return res;

}

void CStrings::parseStringList(const CUtiString& source, const CUtiString& seperator)
{
    int lastpos = 0;
    while (lastpos < source.size()) {
        int p = source.Pos(lastpos+1, seperator);
        if (p == -1) p = source.size();
        CUtiString str =source(lastpos, p-lastpos);
        lastpos = p+1;
        CUtiString t = str.Trim();
        add(t);
    }
}

void CStrings::setText(const CUtiString& t)
{
    clear();


    int i=0;
    int l = t.size();
    int linestart=0;
    CUtiString line;
    bool had10, had13;
    while (i < l && t[i] != 0)
    {
        if (t[i] == 10 || t[i] == 13)
        {
            t.Copy(line, linestart, i-linestart);
            add(line);
            had10 = false;
            had13 = false;
            while (i < l && (t[i] == 10 || t[i] == 13)) {
                if (t[i] == 10) {
                    if (had10) break;
                    had10 = true;
                }
                if (t[i] == 13) {
                    if (had13) break;
                    had13 = true;
                }
                i++;
            };
            linestart = i;
        }
        else
        {
            i++;
        };

    }
    t.Copy(line, linestart, l-linestart);
    add(line);
}


class CStringListInternData : public ReferenceUtiOB
{
public:
    std::vector<CUtiString> m_strings;
public:
    virtual CUtiString getLine(int l) const
    {
        return m_strings.at(l);
    }

    virtual void setLine(int l, const CUtiString& str)
    {

        assert(l >= 0 && l <m_strings.size());
        m_strings.at(l) = str;

    };
    virtual void deleteLine(int l)
    {

        assert(l >= 0 && l <m_strings.size());

        m_strings.erase(m_strings.begin()+l);
    }

    virtual void insertLine(int l, const CUtiString& str)
    {
        assert(l >= 0 && l <=m_strings.size());
        m_strings.insert(m_strings.begin()+l, str);


    }

    virtual void clear()
    {
        m_strings.clear();
    }
    virtual int size()
    {
        return m_strings.size();
    }

    virtual void sort()
    {

        std::sort(m_strings.begin(), m_strings.end());

    }
};

class CUndoStringList;

class CUndoStringList_Object : public ReferenceUtiOB
{
protected:
    CUndoStringList* m_stringlist;
    int m_line;
public:
    CUndoStringList_Object(CUndoStringList* stringlist, int line){m_stringlist = stringlist, m_line = line;};
    virtual void execute()=0;
    virtual void undo()=0;
};

class CUndoStringList_DeleteLine : public CUndoStringList_Object
{
protected:
    CUtiString m_text;
public:
    CUndoStringList_DeleteLine(CUndoStringList* stringlist, int line);
    virtual void execute();
    virtual void undo();
};


class CUndoStringList_InsertLine : public CUndoStringList_Object
{
protected:
    CUtiString m_text;
public:
    CUndoStringList_InsertLine(CUndoStringList* stringlist, int line, const CUtiString& text);
    virtual void execute();
    virtual void undo();
};

class CUndoStringList_Clear : public CUndoStringList_Object
{
protected:
    std::vector<CUtiString> m_strings;

public:
    CUndoStringList_Clear(CUndoStringList* stringlist);
    virtual void execute();
    virtual void undo();
};

class CUndoStringList_ChangeLine : public CUndoStringList_Object
{
protected:
    CUtiString m_newtext;
    CUtiString m_oldtext;
    int m_changestart;
public:
    CUndoStringList_ChangeLine(CUndoStringList* stringlist, int line, const CUtiString& oldline, const CUtiString& newline);
    virtual void execute();
    virtual void undo();
};

class CUndoStringList : public CStringListInternData
{
public:
    CReferenceList<CUndoStringList_Object> m_undo;
    CReferenceList<CUndoStringList_Object> m_redo;
public:
    void addAndExecute(CUndoStringList_Object* obj)
    {
        obj->execute();
        m_undo.push_back(obj);
        m_redo.clear();
        while (m_undo.size() > 500) {
            m_undo.deleteItem(0);
        }
    }
    void undo()
    {
        if (m_undo.size() == 0)return;
        m_undo[m_undo.size()-1].undo();
        m_redo.push_back(&m_undo[m_undo.size()-1]);
        m_undo.deleteItem(m_undo.size()-1);


        while (m_redo.size() > 500) {
            m_redo.deleteItem(0);
        }
    }
    void redo()
    {

        if (m_redo.size() == 0)return;
        m_redo[m_redo.size()-1].execute();
        m_undo.push_back(&m_redo[m_redo.size()-1]);
        m_redo.deleteItem(m_redo.size()-1);


        while (m_undo.size() > 500) {
            m_undo.deleteItem(0);
        }

    }
    bool hasUndo()
    {
        return m_undo.size() != 0;
    }
    bool hasRedo()
    {
        return m_redo.size() != 0;
    }

    void clearUndoRedo()
    {
         m_undo.clear();
         m_redo.clear();
    }

    virtual CUtiString getLine(int l) const
    {
        return m_strings[l];
    }

    virtual void setLine(int l, const CUtiString& str)
    {
        if (str ==  m_strings.at(l)) {return;}
        CUndoStringList_ChangeLine* changeline = new CUndoStringList_ChangeLine(this, l, m_strings.at(l), str);
        addAndExecute(changeline);

    };
    virtual void deleteLine(int l)
    {

        CUndoStringList_DeleteLine* obj = new CUndoStringList_DeleteLine(this, l);
        addAndExecute(obj);
    }
    virtual void insertLine(int l, const CUtiString& str)
    {
        CUndoStringList_InsertLine* obj = new CUndoStringList_InsertLine(this, l, str);
        addAndExecute(obj);
    }
    virtual void clear()
    {
        CUndoStringList_Clear* obj = new CUndoStringList_Clear(this);
        addAndExecute(obj);
    }

    virtual void sort()
    {

        assert(false);
        std::sort(m_strings.begin(), m_strings.end());

    }
};


CUndoStringList_DeleteLine::CUndoStringList_DeleteLine(CUndoStringList* stringlist, int line) : CUndoStringList_Object(stringlist, line)
    {
        m_text = stringlist->m_strings.at(line);

    };
void CUndoStringList_DeleteLine::execute()
    {
        m_stringlist->m_strings.erase(m_stringlist->m_strings.begin()+m_line);

    }
void CUndoStringList_DeleteLine::undo()
    {
        m_stringlist->m_strings.insert(m_stringlist->m_strings.begin()+m_line, m_text);


    }


CUndoStringList_InsertLine::CUndoStringList_InsertLine(CUndoStringList* stringlist, int line, const CUtiString &text) : CUndoStringList_Object(stringlist, line)
    {
        m_text = text;

    };
void CUndoStringList_InsertLine::execute()
    {
        m_stringlist->m_strings.insert(m_stringlist->m_strings.begin()+m_line, m_text);

    }
void CUndoStringList_InsertLine::undo()
    {

        m_stringlist->m_strings.erase(m_stringlist->m_strings.begin()+m_line);


    }


CUndoStringList_Clear::CUndoStringList_Clear(CUndoStringList* stringlist) : CUndoStringList_Object(stringlist, 0)
    {
        for (int i = 0; i < stringlist->m_strings.size(); i++) {
           m_strings.push_back(stringlist->m_strings.at(i));
        }

    };
void CUndoStringList_Clear::execute()
    {
        m_stringlist->m_strings.clear();

    }
void CUndoStringList_Clear::undo()
    {

    assert(m_stringlist->m_strings.size() == 0);
        for (int i = 0; i < m_strings.size(); i++) {
           m_stringlist->m_strings.push_back(m_strings.at(i));
        }

    }


CUndoStringList_ChangeLine::CUndoStringList_ChangeLine(CUndoStringList* stringlist, int line, const CUtiString& oldline, const CUtiString& newline) : CUndoStringList_Object(stringlist, line)
    {
        int start = 0;
        int end = oldline.size();
        if (end == 0) {
            m_changestart = 0;
            m_newtext = newline;
            m_oldtext = oldline;
        }
        while (start < oldline.size() && oldline[start] == newline[start]) {
            start++;
        }

        while (end > 0 && end > start&& end-oldline.size()+newline.size() > start&& end-oldline.size()+newline.size() > 0 && newline[end-oldline.size()+newline.size()-1 ] == oldline[end-1]) {
            end--;
        }
        m_changestart = start;
        m_oldtext = oldline(start, end-start);
        m_newtext = newline(start, end-oldline.size()+newline.size()-start);


    }
void CUndoStringList_ChangeLine::execute()
    {
        CUtiString oldstring = m_stringlist->m_strings.at(m_line);
        oldstring.Delete(m_changestart, m_oldtext.size());
        oldstring.insert(m_changestart, m_newtext);
        m_stringlist->m_strings.at(m_line) = oldstring;

    }
void CUndoStringList_ChangeLine::undo()
    {
        CUtiString oldstring = m_stringlist->m_strings.at(m_line);
        oldstring.Delete(m_changestart, m_newtext.size());
        oldstring.insert(m_changestart, m_oldtext);
        m_stringlist->m_strings.at(m_line) = oldstring;


    }


CStringList::CStringList(bool undo)
{
    if (!undo) {
    m_data = new CStringListInternData();
    } else {
        m_data = new CUndoStringList();
    }
}

CStringList::~CStringList()
{

}

void CStringList::undo()
{

    check_cast<CUndoStringList*>(m_data.get())->undo();
}

void CStringList::redo()
{

    check_cast<CUndoStringList*>(m_data.get())->redo();
}

void CStringList::clearUndoRedo()
{

    check_cast<CUndoStringList*>(m_data.get())->clearUndoRedo();
}

bool CStringList::hasUndo()
{

    return check_cast<CUndoStringList*>(m_data.get())->hasUndo();

}

bool CStringList::hasRedo()
{
   return check_cast<CUndoStringList*>(m_data.get())->hasRedo();
}

CUtiString CStringList::getLine(int l) const
{
return m_data->getLine(l);
}

void CStringList::setLine(int l, const CUtiString& str)
{
    m_data->setLine(l, str);
}

struct SParse_CSVLine
{
    int m_linepos;
    CUtiString m_line;
    SParse_CSVLine()
    {
        m_linepos = 0;
    }
    void moveToChar()
    {
        while (m_linepos < m_line.size() && m_line[m_linepos] == L' ') {
            m_linepos++;
        }
    }
    wchar_t getNextChar()
    {
        moveToChar();
        if (m_linepos < m_line.size()) {
            return m_line[m_linepos];
        }
        return 0;
    }
    CUtiString getNextString(wchar_t sep, wchar_t up)
    {
        moveToChar();
        int m_charstart = m_linepos;
        wchar_t nextchar = getNextChar();
        if (nextchar == 0) {
            return EMPTYSTRING;

        }
        if (nextchar == up) {
            while (m_linepos < m_line.size() && m_line[m_linepos] != up) {
                m_linepos++;

            }
            m_linepos++;
            CUtiString res = m_line(m_charstart, m_linepos-m_charstart);
            return res;
        } else {
            while (m_linepos < m_line.size() && m_line[m_linepos] != sep) {
                m_linepos++;

            }
            CUtiString res = m_line(m_charstart, m_linepos-m_charstart-1);
            return res;

        }
    }
};


void csv_parse(const CUtiString& str, const CUtiString& seperator, const CUtiString& up, std::vector<CUtiString>& seg)
{
   wchar_t sep = seperator[0];
   wchar_t upvalue = up[0];
   SParse_CSVLine line;
   line.m_line = str;
   seg.push_back(line.getNextString(sep, upvalue));
   while (line.getNextChar() == sep) {

       seg.push_back(line.getNextString(sep, upvalue));

   }
}

/*
CUtiString& CStringList::line(int l)
{
    assert(l >= 0 && l <m_data->m_strings.size());
    return m_data->m_strings.at(l);
}

const CUtiString& CStringList::line(int l) const
{
    assert(l >= 0 && l <m_data->m_strings.size());
    return m_data->m_strings.at(l);
}*/

int CStringList::size() const
{
    return m_data->size();
}
void CStringList::clear()
{
    return m_data->clear();
}

void CStringList::deleteLine(int i)
{
    m_data->deleteLine(i);
}

void CStringList::add(const CUtiString& str)
{

   // m_data->m_strings.push_back(str);
    m_data->insertLine(m_data->size(), str);
}

void CStringList::insert(int i, const CUtiString&str)
{
   m_data->insertLine(i, str);
}

void CStringList::sort()
{
m_data->sort();
}


CUtiString CUtiString::replace(U_CHAR oldchar, U_CHAR newchar) const
{
    CUtiString str = *this;
    str.ReplaceChar(oldchar, newchar);
    return str;
}

CUtiString CUtiString::replace(const CUtiString& oldstr, const CUtiString& newstr) const
{
    CUtiString str = *this;
    str.ReplaceString(oldstr, newstr);
    return str;
}


void CUtiString::split(const CUtiString& str, CStrings& array, bool merge) const
{
     int start = 0;
     array.clear();
     while (true) {
         int k = this->indexOf(str, start);
         if (k == -1) {
             CUtiString s = this->substring(start);
             if (s.size() != 0) {
             array.add(s);
             }
             return;
         }
         CUtiString s = this->substring(start, k);
         if (s.size() != 0) {
         array.add(s);
         }
         start = k+str.size();

     }
}

uti_refptr<CStrings> CUtiString::split(const CUtiString& str, bool merge) const
{
    uti_refptr<CStrings> result(new CStringList());
    split(str, *result.get(), merge);
    return result;
}

bool CUtiString::contains(const CUtiString& str) const
{
    return Pos(str) != -1;
}

int CUtiString::indexOf(const CUtiString& str, int start) const
{
    return Pos(start, str);
}
int CUtiString::lastIndexOf(const CUtiString& str, int end) const
{
    if (end == -1) {
        end = size()-1;
    }
#ifdef DEBUGSTRING
     utiassert(checkContainer());
#endif
     int l = str.size();
     int length = m_container->m_length;
      Array<U_CHAR>& string  = m_container->m_string;
      end = umin(end, length-l);
      if (end < 0) {
          return -1;
      }
     if (l == 1)
     {
         for (int i = end; i >=0; i--)
         {
             if (string[i] == str[0])
             {
                 return(i);
             };
         }
     }
     else
     {
         for (int i = end; i >=0; i--)
         {
             if (wcsncmp(&string[i], str.c_str(), l)==0)
             {
                 return(i);
             };
         }
     }
     return(-1);
}

bool CUtiString::endsWith(const CUtiString& other) const
{
    if (this->size() < other.size())return false;
    int s = this->size();
    int s2 = other.size();
    for (int i = 0; i < s2; i++) {
        if (operator[](i+s-s2) != other[i]) {
            return false;
        }
    }

    return true;
}

CUtiString CUtiString::substring(int start, int end) const
{
    if (end == -1) {
        end = size();
    }
    return operator()(start, end-start);
}

CUtiString CUtiString::join(CUtiString delimiter, const CStrings& str)
{
    CUtiString result;

    for (int i = 0; i< str.size();i++) {
        result += str.getLine(i);
        if (i != str.size()-1) {
            result += delimiter;
        }
    }

    return result;
}

void CUtiString::insert(int pos, U_CHAR str)
{
    CUtiString r;
    r = UTISTR(" ");
    r[0] = str;
    insert(pos, r);
}

void CUtiString::insert(int pos, float str)
{
    CUtiString r;
    r += str;
    insert(pos, r);
}
void CUtiString::insert(int pos, double str)
{
    CUtiString r;
    r += (float)str;
    insert(pos, r);
}
void CUtiString::insert(int pos, bool str)
{
    CUtiString r;
    r += str ? 1 : 0;
    insert(pos, r);
}
void CUtiString::insert(int pos, int str)
{
    CUtiString r;
    r += str;
    insert(pos, r);
}


void CUtiString::append(const CUtiString& str)
{
   *this += str;
}

void CUtiString::append(U_CHAR str)
{
    CUtiString r;
    r = UTISTR(" ");
    r[0] = str;
    *this+= r;
}

void CUtiString::append(float str)
{
*this += str;
}

void CUtiString::append(double str)
{

    *this +=(float)str;
}

void CUtiString::append(bool str)
{

    *this += str ? 1 : 0;
}

void CUtiString::append(int str)
{
*this += str;
}


bool CUtiString::isNumber() const
{
    int s = size();
   for (int i = 0; i < s; i++) {
wchar_t ch = (*this)[i];
if (ch == L' ') continue;
if (!isDigit(ch))return false;
   }
   return true;
}

CUtiString CUtiString::valueOf(bool v)
{
    CUtiString str;
    str += v ? 1 : 0;
    return str;
}

CUtiString CUtiString::valueOf(char v)
{
    CUtiString str;
    str = UTISTR(" ");
    str[0] = v;
    return str;
}
CUtiString CUtiString::valueOf(double v)
{
CUtiString str;
str += (float)v;
return str;
}
CUtiString CUtiString::valueOf(float v)
{
    CUtiString str;
    str += v;
    return str;
}
CUtiString CUtiString::valueOf(int v)
{
    CUtiString str;
    str += v;
    return str;
}
bool CUtiString::parseBoolean(const CUtiString& str)
{
    return CUtiString::strToInt(str) == 1;
}

int CUtiString::parseInt(const CUtiString& str)
{
    return CUtiString::strToInt(str);
}

int CUtiString::parseInt(const CUtiString& str, int radix)
{
    assert(radix > 1 && radix <= 36);
   return wcstol(str.m_container->m_string.getCArray(), 0, radix);
}

float CUtiString::parseFloat(const CUtiString& str)
{

    return CUtiString::strToFloat(str);
}

double CUtiString::parseDouble(const CUtiString& str)
{

    return CUtiString::strToFloat(str);
}

CUtiString CUtiString::toHexString(int i)
{

    return toString(i, 16);

}

CUtiString CUtiString::toOctalString(int i)
{

    return toString(i, 8);

}
CUtiString CUtiString::toBinaryString(int i)
{

   /* std::wstringstream str;

    str<< std::binary << i;
     return str.str().c_str();*/
    return toString(i, 2);
}

//From Stackoverflow: User: Dmitri
wchar_t *toBaseWhatever(int num, int rad)
{
    wchar_t digits[] = L"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
    int i;
    wchar_t buf[66];   /* enough space for any 64-bit in base 2 */

    /* bounds check for radix */
    if (rad < 2 || rad > 62)
        return NULL;
    /* if num is zero */
    if (!num)
        return wcsdup(L"0");

    /* null terminate buf, and set i at end */
    buf[65] = '\0';
    i = 65;

    if (num > 0) {  /* if positive... */
        while (num) { /* until num is 0... */
            /* go left 1 digit, divide by radix, and set digit to remainder */
            buf[--i] = digits[num % rad];
            num /= rad;
        }
    } else {    /* same for negative, but negate the modulus and prefix a '-' */
        while (num) {
            buf[--i] = digits[-(num % rad)];
            num /= rad;
        }
        buf[--i] = '-';
    }
    /* return a duplicate of the used portion of buf */
    return wcsdup(buf + i);
}

CUtiString CUtiString::toString(int i, int radix)
{

    wchar_t* answer =toBaseWhatever(i, radix);
    assert(answer != NULL);
    CUtiString res = answer;
            free(answer);
    return res;
}
void CUtiString::parseBase64Binary(const CUtiString& str, Array<unsigned char>& data)
{
    int l_length = str.size();
    int originallength = CUti64Coder::reconstructDataLength(l_length);
    data.setSize(originallength);

    char* c = new char[l_length+1];
    for (int i = 0; i < l_length;i++) {
        c[i] = str[i];
    }
    c[l_length] = 0;
    CUti64Coder::readData(c, l_length, data.array, originallength);

    delete[]c;
}

CUtiString CUtiString::printBase64Binary(const Array<unsigned char>& data)
{
    int size = CUti64Coder::getDataLength(data.size());
    char* c = new char[size+1];

    CUti64Coder::writeData(data.array, data.size(), c, size);
    c[size] = 0;
    CUtiString res;
    res.setLength(size);
    for (int i = 0; i < size; i++) {
        res[i] = c[i];
    }
    delete[]c;
    return res;
}

void CUtiString::parseHexBinary(const CUtiString& str, Array<unsigned char>& data)
{

    assert((str.size() &2) == 0);
    data.setSize(str.size()/2);

    int l = data.size();
    for (int i = 0;i < l; i++) {
        data[i] = CUtiString::digit( str[i*2+0], 16) | (CUtiString::digit( str[i*2+1], 16) << 4);

    }
}
CUtiString CUtiString::printHexBinary(const Array<unsigned char>& data)
{
     CUtiString str;
     str.setLength(data.size());

     int l = data.size();
     for (int i = 0;i < l; i++) {
         str[i*2+0] =forDigit(data[i] & 0xF, 16);
         str[i*2+1] =forDigit((data[i] >> 4) & 0xF, 16);
     }

     return str;
}
bool CUtiString::isDigit(U_CHAR ch) {return ch >= '0' && ch <= '9';}
bool CUtiString::isLetter(U_CHAR ch) {return (ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z');}
bool CUtiString::isWhitespace(U_CHAR ch) {return ch == ' ' || ch == 10 || ch == 13 || ch == 7;}
bool CUtiString::isLetterOrDigit(U_CHAR ch) {return isDigit(ch) || isLetter(ch);}
bool CUtiString::isLowerCase(U_CHAR ch){return towlower(ch) == ch;}
bool CUtiString::isUpperCase(U_CHAR ch) {return towupper(ch) == ch;}
U_CHAR CUtiString::toLowerCase(U_CHAR ch) {return towlower(ch);}
U_CHAR CUtiString::toUpperCase(U_CHAR ch) {return towupper(ch);}
int CUtiString::digit(wchar_t ch, int radix)
{

    wchar_t digits[] = L"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

    int res = -1;
    if (ch >= L'0' && ch < L'9') {
        res = ch-L'0';
    } else if (ch >= L'A' && ch <= L'Z') {
        res = 10+(ch-L'A');

    } else if (ch >= L'a' && ch <= L'z') {
        res = 35+(ch-L'a');
    }
    assert(res < radix);
    assert(res >= 0 && res < 60);
    assert(digits[res] == ch);
    return res;
}

wchar_t CUtiString::forDigit(int digit, int radix)
{
    wchar_t digits[] = L"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

    assert(radix <= 62);
assert(digit < radix);
return digits[digit];
}

CUtiString CUtiString::format(const CUtiString& format, const CStrings& str)
{
    uti_refptr<CStrings> t1 = format.split(UTISTR("%s"));
    CUtiString res;
    res = t1->getLine(0);
    for (int i =1; i < t1->size(); i++) {

        int a = i-1;
        assert(a < str.size());
        if (a < str.size()) {
            res += str.getLine(a);

        }
        res += t1->getLine(i);
    }
    return res;
}

int CUtiString::hashCode() const
{
    int hashvalue = 5381;
    if (!m_container.get()) return hashvalue;
   Array<U_CHAR>& data = this->m_container->m_string;
   int c = this->m_container->m_length;
   for (int i = 0; i < c; i++) {
        hashvalue = ((hashvalue << 5) + hashvalue) + data[i];
   }

   return hashvalue;
}



CStringJoiner::CStringJoiner(const CUtiString& part, const CUtiString&front, const CUtiString&back)
{
    m_hasEmptyValue = false;
    m_part = part;
    m_front = front;
    m_back = back;
}

int CStringJoiner::length()
{
    return m_content.size();
}

CStringJoiner& CStringJoiner::add(const CUtiString& str)
{
    m_content.push_back(str);
    return *this;
}

CStringJoiner& CStringJoiner::add(const CStrings& str)
{
    for (int i = 0; i < str.size(); i++) {
        m_content.push_back(str.getLine(i));
    }
    return *this;
}

CStringJoiner& CStringJoiner::setEmptyValue(const CUtiString& val)
{
    m_emptyvalue = val;
    m_hasEmptyValue = true;
    return *this;
}

CUtiString CStringJoiner::toString()
{
    if (m_hasEmptyValue && m_content.size() == 0) {
        return m_emptyvalue;
    }
    CUtiString res=m_front;

    for (int i = 0;  i< m_content.size(); i++) {
        res += m_content.at(i);
        if (i != m_content.size()-1) {
            res += m_part;
        }
    }
    res += m_back;
    return res;
}


int CGameEngineNameStore::s_emptystring=0;

CGameEngineNameStore::CGameEngineNameStore()
{
    m_nextindex = 0;
    m_emptystring = convertNameToInt(EMPTYSTRING);
    s_emptystring = m_emptystring;
    assert(m_emptystring == EMPTYSTRING_ID);
    m_nullstring = convertNameToInt(UTISTR("NULL"));
    m_zerostring = convertNameToInt(UTISTR("ZERO"));
#define CREATE_LISP_NAME(a, b) \
    if (LISP_ ## a !=  convertNameToInt(UTISTR(b))) {\
    std::wcout << "Enum: " << LISP_ ## a << " und " << convertNameToInt(UTISTR(b)) <<std::endl;\
    assert(false);\
} //3

#define CREATE_LISP_NAME_P0(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P1(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P2(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P3(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P0_REST(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P1_REST(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P2_REST(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P3_REST(a, b) CREATE_LISP_NAME(a,b)


#define CREATE_LISP_NAME_P0_KEY(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P1_KEY(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P2_KEY(a, b) CREATE_LISP_NAME(a,b)
#define CREATE_LISP_NAME_P3_KEY(a, b) CREATE_LISP_NAME(a,b)

    CREATE_LISP_NAME_P0_REST(list, "list")
    CREATE_LISP_NAME_P0_REST(append, "append")
    CREATE_LISP_NAME_P2_KEY(adjoin, "adjoin")
    CREATE_LISP_NAME_P1(reverse, "reverse")
    CREATE_LISP_NAME_P1(eval, "eval")
    CREATE_LISP_NAME_P2_KEY(sort, "sort")
    CREATE_LISP_NAME_P3_KEY(substitute, "substitute")
    CREATE_LISP_NAME_P1(length, "length")
    CREATE_LISP_NAME_P1(evenp, "evenp")
    CREATE_LISP_NAME_P1(oddp, "oddp")
    CREATE_LISP_NAME_P2_KEY(count, "count")
    CREATE_LISP_NAME_P2_KEY(find, "find")
    CREATE_LISP_NAME_P2_KEY(position, "position")
    CREATE_LISP_NAME_P1(first, "first")
    CREATE_LISP_NAME_P1(second, "second")
    CREATE_LISP_NAME_P1(third, "third")
    CREATE_LISP_NAME_P1(last, "last")
    CREATE_LISP_NAME_P1(butlast, "butlast")
    CREATE_LISP_NAME_P2(nth, "nth")
    CREATE_LISP_NAME_P2(nthcdr, "nthcdr")
    CREATE_LISP_NAME_P2_REST(subseq, "subseq")
    CREATE_LISP_NAME_P2_KEY(remove, "remove")
    CREATE_LISP_NAME_P2(push, "push")
    CREATE_LISP_NAME_P1(pop, "pop")
    CREATE_LISP_NAME(globals, "globals")
    CREATE_LISP_NAME_P2_KEY(member, "member")
    CREATE_LISP_NAME_P2(pushnew, "pushnew")
    CREATE_LISP_NAME_P2_KEY(l_union, "union")
    CREATE_LISP_NAME_P2_KEY(intersection, "intersection")
    CREATE_LISP_NAME_P2_KEY(set_difference, "set-difference")
    CREATE_LISP_NAME_P3(l_if, "if")
    CREATE_LISP_NAME_P1_REST(when, "when")
    CREATE_LISP_NAME_P0_REST(cond, "cond")
    CREATE_LISP_NAME(l_case, "case")
    CREATE_LISP_NAME_P2(cons, "cons")
    CREATE_LISP_NAME(rest, "rest")
    CREATE_LISP_NAME_P1(car, "car")
    CREATE_LISP_NAME_P1(cdr, "cdr")
    CREATE_LISP_NAME(mapc, "mapc")
    CREATE_LISP_NAME(otherwise, "otherwise")
    CREATE_LISP_NAME(print, "print")
    CREATE_LISP_NAME_P0(t, "t")
    CREATE_LISP_NAME(l_do, "do")
    CREATE_LISP_NAME(l_return, "return")
    CREATE_LISP_NAME(quote, "quote")
    CREATE_LISP_NAME_P0(nil, "nil")
    CREATE_LISP_NAME_P1(l_not, "not")
    CREATE_LISP_NAME_P1(inc_one, "1+")
    CREATE_LISP_NAME_P1(dec_one, "1-")
    CREATE_LISP_NAME_P2(less, "<")
    CREATE_LISP_NAME_P2(eless, "<=")
    CREATE_LISP_NAME_P2(greater, ">")
    CREATE_LISP_NAME_P2(egreater, ">=")
    CREATE_LISP_NAME_P2(equal_numeric, "=")
    CREATE_LISP_NAME_P2(dotimes, "dotimes")
  CREATE_LISP_NAME(assoc, "assoc")
    CREATE_LISP_NAME(dolist, "dolist")
    CREATE_LISP_NAME(ash, "ash")
    CREATE_LISP_NAME_P0_REST(progn, "progn")
    CREATE_LISP_NAME_P0_REST(block, "block")
    CREATE_LISP_NAME(head, "head")
    CREATE_LISP_NAME_P2(setf, "setf")
    CREATE_LISP_NAME_P2(equal, "equal")
    CREATE_LISP_NAME_P0_REST(let, "let")
    CREATE_LISP_NAME_P2(eql, "eql")
    CREATE_LISP_NAME_P2(let_star, "let*")
    CREATE_LISP_NAME_P1_REST(plus, "+")
    CREATE_LISP_NAME_P1_REST(minus, "-")
    CREATE_LISP_NAME_P1_REST(mul, "*")
    CREATE_LISP_NAME_P1_REST(div, "/")
    CREATE_LISP_NAME_P2_REST(defparameter, "defparameter")
    CREATE_LISP_NAME_P2_REST(defconstant, "defconstant")
    CREATE_LISP_NAME_P2_REST(defun, "defun")
    CREATE_LISP_NAME_P1_REST(lambda, "lambda")
    CREATE_LISP_NAME_P1_REST(apply, "apply")
    CREATE_LISP_NAME_P1_REST(funcall, "funcall")
    CREATE_LISP_NAME(mapcar, "mapcar")
    CREATE_LISP_NAME(maplist, "maplist")
    CREATE_LISP_NAME(values, "values")
    CREATE_LISP_NAME_P0_REST(mutliple_value_bind, "multiple-value-bind")
    CREATE_LISP_NAME_P0_REST(mutliple_value_call, "multiple-value-call")
    CREATE_LISP_NAME_P2(incf, "incf")
    CREATE_LISP_NAME_P2(decf, "decf")
    CREATE_LISP_NAME_P1(abs, "abs")
    CREATE_LISP_NAME_P2(mod, "mod")
    CREATE_LISP_NAME_P1_REST(max, "max")
    CREATE_LISP_NAME_P1_REST(min, "min")
    CREATE_LISP_NAME_P2(expt, "expt")
    CREATE_LISP_NAME_P1(float, "float")
    CREATE_LISP_NAME_P1(numerator, "numerator")
    CREATE_LISP_NAME_P1(denominator, "denominator")
    CREATE_LISP_NAME_P1(round, "round")
    CREATE_LISP_NAME_P1(floor, "floor")
    CREATE_LISP_NAME_P1(ceiling, "ceiling")
    CREATE_LISP_NAME_P1(truncate, "truncate")
    CREATE_LISP_NAME_P1(random, "random")
    CREATE_LISP_NAME_P1(sin, "sin")
    CREATE_LISP_NAME_P1(cos, "cos")
    CREATE_LISP_NAME_P1(tan, "tan")
    CREATE_LISP_NAME_P1(asin, "asin")
    CREATE_LISP_NAME_P1(acos, "acos")
    CREATE_LISP_NAME_P1(atan, "atan")
    CREATE_LISP_NAME_P1(sinh, "sinh")
    CREATE_LISP_NAME_P1(cosh, "cosh")
    CREATE_LISP_NAME_P1(tanh, "tanh")
    CREATE_LISP_NAME_P1(asinh, "asinh")
    CREATE_LISP_NAME_P1(acosh, "acosh")
    CREATE_LISP_NAME_P1(atanh, "atanh")
    CREATE_LISP_NAME_P2(typep, "typep")
    CREATE_LISP_NAME_P2_REST(every, "every")
    CREATE_LISP_NAME_P2_REST(some, "some")



    CREATE_LISP_NAME_P1(null, "null")
    CREATE_LISP_NAME_P1(listp, "listp")
    CREATE_LISP_NAME_P1(zerop, "zerop")
    CREATE_LISP_NAME_P1(plusp, "plusp")
    CREATE_LISP_NAME_P1(minusp, "minusp")
    CREATE_LISP_NAME_P1(upper_case_p, "upper-case-p")
    CREATE_LISP_NAME_P1(lower_case_, "lower-case-p")
    CREATE_LISP_NAME_P1(numberp, "numberp")
    CREATE_LISP_NAME_P1(floatp, "floatp")
    CREATE_LISP_NAME_P1(integerp, "integerp")
    CREATE_LISP_NAME_P1(graphic_char_p, "graphic-char-p")
    CREATE_LISP_NAME_P1(alphanumericp, "alphanumericp")
    CREATE_LISP_NAME_P1(alpha_char_p, "alpha-char-p")
    CREATE_LISP_NAME(make_array, "make-array")
    CREATE_LISP_NAME(aref, "aref")
    CREATE_LISP_NAME(svref, "svref")
    CREATE_LISP_NAME(vector, "vector")
    CREATE_LISP_NAME(defstruct, "defstruct")
    CREATE_LISP_NAME(format, "format")
    CREATE_LISP_NAME_P1_REST(concatenate, "concatenate")
    CREATE_LISP_NAME(char, "char")
    CREATE_LISP_NAME(string_capitalize, "string-capitalize")
    CREATE_LISP_NAME(string_upcase, "string-upcase")
    CREATE_LISP_NAME(string_downcase, "string-downcase")
    CREATE_LISP_NAME(remove_duplicates, "remove-duplicates")
    CREATE_LISP_NAME(read, "read")
    CREATE_LISP_NAME(read_line, "read-line")
    CREATE_LISP_NAME(read_from_string, "read-from-string")
    CREATE_LISP_NAME(defmacro, "defmacro")
    CREATE_LISP_NAME(macroexpand, "macroexpand")
    CREATE_LISP_NAME(nthmost, "nthmost")
    CREATE_LISP_NAME(starts_with, "starts-with")
    CREATE_LISP_NAME(rotate_left, "rotate-left")
    CREATE_LISP_NAME(rotate_right, "rotate-right")
    CREATE_LISP_NAME(insert_between, "insert-between")
    CREATE_LISP_NAME(permutations, "permutations")
    CREATE_LISP_NAME(terpri, "terpri")
    CREATE_LISP_NAME_P1_REST(l_and, "and")
    CREATE_LISP_NAME_P1_REST(l_or, "or")
    CREATE_LISP_NAME_P2(defvar, "defvar")
    CREATE_LISP_NAME_P1_REST(unless, "unless")
    CREATE_LISP_NAME(l_string, "string")
    CREATE_LISP_NAME_P2_KEY(string_equal, "string=")
    CREATE_LISP_NAME_P2_KEY(string_less, "string<")
    CREATE_LISP_NAME_P2_KEY(string_greater, "string>")
    CREATE_LISP_NAME_P2_KEY(string_eless, "string<=")
    CREATE_LISP_NAME_P2_KEY(string_egreater, "string>=")
CREATE_LISP_NAME_P1(caaar, "caaar")
CREATE_LISP_NAME_P1(caadr, "caadr")
CREATE_LISP_NAME_P1(caar, "caar")
CREATE_LISP_NAME_P1(cadar, "cadar")
CREATE_LISP_NAME_P1(caddr, "caddr")
CREATE_LISP_NAME_P1(cadr, "cadr")
CREATE_LISP_NAME_P1(cdaar, "cdaar")
CREATE_LISP_NAME_P1(cdadr, "cdadr")
CREATE_LISP_NAME_P1(cdar, "cdar")
CREATE_LISP_NAME_P1(cddar, "cddar")
CREATE_LISP_NAME_P1(cdddr, "cdddr")
CREATE_LISP_NAME_P1(cddr, "cddr")
    CREATE_LISP_NAME(d_key, ":key")
    CREATE_LISP_NAME(d_test, ":test")
    CREATE_LISP_NAME(d_test_not, ":test-not")
    CREATE_LISP_NAME(d_from_end, ":from-end")
    CREATE_LISP_NAME(d_start, ":start")
    CREATE_LISP_NAME(d_end, ":end")
    CREATE_LISP_NAME(d_count, ":count")
            CREATE_LISP_NAME(symbol, "symbol")
            CREATE_LISP_NAME(integer, "integer")
            CREATE_LISP_NAME(function, "function")
            CREATE_LISP_NAME(number, "number")
            CREATE_LISP_NAME_P1(type_of, "type-of")
            CREATE_LISP_NAME(contains, "contains")
            CREATE_LISP_NAME(not_contains, "not-contains")
            CREATE_LISP_NAME(add, "add")
            CREATE_LISP_NAME(agent, "agent")
            CREATE_LISP_NAME_P2_KEY(l_delete, "delete")
            CREATE_LISP_NAME_P2_REST(map, "map")




#undef CREATE_LISP_NAME

#undef CREATE_LISP_NAME_P0
#undef CREATE_LISP_NAME_P1
#undef CREATE_LISP_NAME_P2
#undef CREATE_LISP_NAME_P3
#undef CREATE_LISP_NAME_P0_REST
#undef CREATE_LISP_NAME_P1_REST
#undef CREATE_LISP_NAME_P2_REST
#undef CREATE_LISP_NAME_P3_REST


#undef CREATE_LISP_NAME_P0_KEY
#undef CREATE_LISP_NAME_P1_KEY
#undef CREATE_LISP_NAME_P2_KEY
#undef CREATE_LISP_NAME_P3_KEY
}

CGameEngineNameStore::~CGameEngineNameStore()
{

}


int CGameEngineNameStore::convertNameToInt(const CUtiString& id)
{
    CHECKEXISTANCE;
    CCSControlSpin spin(spinlock);
  std::map<CUtiString, int>::iterator it = m_name_to_int.find(id);
  if (it == m_name_to_int.end()) {
      int res = m_nextindex;
      m_nextindex++;
      m_name_to_int.insert(std::make_pair(id, res));
      //m_int_to_name.insert(std::make_pair(res, id));
      assert(m_int_to_name_list.size() == res);
      m_int_to_name_list.push_back(id);
      return res;
  }
  return it->second;
}

MAKE_SINGLETON_IMPL_FAST(CGameEngineNameStore)

const CUtiString &CGameEngineNameStore::convertIntToName(int i)
{
    CCSControlSpin spin(spinlock);
    assert(i < m_int_to_name_list.size());
    return m_int_to_name_list[i];

/*
    std::map<int, CUtiString>::const_iterator it = m_int_to_name.find(i);
    assert(it != m_int_to_name.end());
    if (it == m_int_to_name.end()) {
        return EMPTYSTRING;
    }
    return it->second;*/
}


MAKE_SINGLETON_IMPL_FAST(CLogSystem)

CLogSystem::CLogSystem()
{
  m_starttime = UtiFramework::getCurrentTime();
  m_logstate = LOG_STATEMACHINE;
#ifndef NDEBUG
  m_loglevel = 3;
   m_logstate =/* LOG_STATEMACHINE |*/ LOG_LUA | LOG_PDDL | LOG_ENTITY /*| LOG_ANIMATION*/;
#else
   m_loglevel = 0;
   m_logstate = 0;
#endif
   outputFile = new std::wofstream("game.log");
 //  outputFile.open("game.log");

}


std::wostream& logout()
{
    return CLogSystem::getInstance()->lout();
}

std::wostream& CLogSystem::lout()
{
   return *outputFile;
}

void CLogSystem::doLog(int s, int level, int indent, CUtiOB* object, const CUtiString& str)
{

    CUtiString s2;
    s2.ExpandToSize(indent*3);
    s2+= str;
    TIME time = UtiFramework::getCurrentTime()-m_starttime;
   std::wcout << "Log(time = "<< time << ", level = " << level << ", state = " << s << ", object = " <<object << ", \"" << s2 << "\")" << std::endl;
   (*outputFile) <<"Log(time = "<< time << ", level = " << level << ", state = " << s << ", object = " <<object << ", \"" << s2 << "\")" << std::endl;
//assert(level>1);
}
