#pragma once
/***************************************************************************
 *   Copyright (C) 2006 by Michael Staud   *
 *   michael.staud@uni-ulm.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UtiString__
#define __UtiString__


//#include "StdAfx.h"
#include "Uti3DCore.h"


#include "utichar.h"
#include <wchar.h>
#include <iostream>
#include <vector>
#include <map>

//#define wcscmp _tcscmp
/*#define wcslen _tcslen
#define remove _tremove
#define swprintf _stprintf
#define swscanf _stscanf
#define wcscpy _tcscpy*/

class CUtiString;
static const unsigned int INCBUFFER = 15;

//#define IsNumber(a) (a >= UTEXT('0') && a <= UTEXT('9'))
//#define IsLetter(a) ((a >= UTEXT('A') && a <= UTEXT('Z')) || (a >= UTEXT('a') && a <= UTEXT('z')))
//#define IsUnderLine(a) (a == UTEXT('_'))
//#define IsCharacter(a) ((a != UTEXT(' ')) && (a > 30))
//#define IsBezeichner(a) (IsNumber(a) || IsLetter(a) || IsUnderLine(a))
//#define IsVergleich(a) (a == UTEXT('>') || a == UTEXT('<') || a == UTEXT('='))
const int MAX_ITEMS = 256;
#define UTISTR(text) CUtiString(UTEXT(text), true)
#define EMPTYSTRING UTISTR("")

bool IsInteger(U_CHAR *str, int length);
int SearchChar(U_CHAR *str, U_CHAR c, int length);
int SearchCharBackward(U_CHAR *str, U_CHAR c, int length);
int SearchChar(U_CHAR *str, U_CHAR c, int length, int charpos);
int SearchStringBackward(U_CHAR *str, U_CHAR* c, int length, int searchlength);
int SearchString(U_CHAR *str, U_CHAR* c, int length, int searchlength);
int SearchStringBackwardBez(U_CHAR *str, U_CHAR* c, int length, int searchlength);
int SearchStringBez(U_CHAR *str, U_CHAR* c, int length, int searchlength);
bool ValidBez(U_CHAR* str, int i, int length, int searchlength);
void SearchBeginEnd(U_CHAR*str, int length, int &beginpos, int &endpos, bool last);
void SearchStrings(U_CHAR* str, U_CHAR*c, int length, int searchlength, int* result, int &rescount);
const CUtiString CalcNumberedFileName(const CUtiString& start, const CUtiString& ende, int startindex=0);
CUtiString createRandomString(int length);
class CStrings;

//Idee: Diese Objekte sollte in einem MemoryPool hinterlegt werden. Diese sollte dabei Threadlocal arbeiten?
//Nebenläufigkeit??
class CUtiStringContainer : public ReferenceUtiOB
{
  public:
    //U_CHAR* m_string;
    Array<U_CHAR> m_string;
    int m_length;
    //int m_buffersize;
    //bool m_owns;
    CUtiStringContainer(const U_CHAR* std);
    CUtiStringContainer(int size);
    ~CUtiStringContainer();
    bool allowResize();
    void resize(int size);
    void checkSize();
    CUtiStringContainer* createNew(int size);
};

class CUtiString : public CUtiOB
{
private:
	void replace(const U_CHAR *s);
	uti_refptr<CUtiStringContainer> m_container;
	int getBufferSize(int length);
	void ownContainer();
        void ownContainer(int length);
	bool checkContainer() const;
  public:
    int debugReferences();
    int debugContainerLength();
    int debugBufferSize();
    bool debugOwns();
public:
    bool isOk() const {checkContainer();return true;};
        void setLength(int l);
	int Pos(const CUtiString& substr) const;
        int Pos(int start, const CUtiString& substr) const;
        void makeThreadResistant();
	void Copy(CUtiString& target, int start, int length) const;
	int size()const {return(m_container->m_length);};
    int length()const {return(m_container->m_length);};
    bool empty() const{return(m_container->m_length) == 0;};

    void assignSimpleAnsi(const char* data, int size);
    char* createSimpleAnsi();
	void SetFloat(double f, int count);
        int ToInt() const;
	int HexToInt() const;
        double ToFloat() const;
	int calcbuffersize(int len);
	static void IntToStrRadix(int i, U_CHAR *str, int length, int radix);
	void IntToStr(int i);
    static int strToInt(const CUtiString& str);
    static CUtiString intToStr(int i);

    static CUtiString intToStrZero(int i, int zeros);
    void IntToStr(int i, const CUtiString& format);
    static CUtiString intToStr(int i, const CUtiString& format);
	void UIntToHex(unsigned int i,  int minlength);
	void Delete(int index, int length);
    CUtiString Trim();
    CUtiString TrimLeft();
    CUtiString TrimRight();
    bool startsWith(const CUtiString& other) const;
    bool ReplaceString(const CUtiString& oldone, const CUtiString& newone);
	void FloatToStr(const float i);
    void FloatToStr(const float i, int width, int precision);
    static CUtiString floatToStr(const float i, int width, int precision);
    static float strToFloat(const CUtiString& str);
    void getLine(int line, CUtiString &str) const;
    CUtiString getLine(int line) const;
    int getLineCount() const;
	static void IntToStr(int i, U_CHAR *str);
	const U_CHAR* getString() const;
	const U_CHAR* c_str() const;
	void ReadProp(const TCHAR* name, CUtiString* result, const TCHAR* defaultitem);
	void AddPascalString(U_CHAR* str, int length);

    void insert(unsigned int pos, const CUtiString& str);
    void insert(unsigned int pos, unsigned int slen, const TCHAR *s);
	class EStringAllocateFailure{};
	class EOutOfRangeError{};
	CUtiString(const CUtiString &s);
	CUtiString(const U_CHAR c);
	CUtiString(const U_CHAR *c, bool is_static=false);
	CUtiString();
	~CUtiString();
        CUtiString &operator =(const CUtiString &s);
	CUtiString &operator =(const U_CHAR *s);
	CUtiString &operator =(const U_CHAR c);
	CUtiString &operator =(int i);
	CUtiString &operator =(float i);
    CUtiString &operator =(utifloat i);
	CUtiString &operator+=(const CUtiString &s);
        CUtiString &operator+=(const U_CHAR *s);
	CUtiString &operator+=(const U_CHAR c);
        CUtiString &operator+=(const int i);
	CUtiString &operator+=(const float i);
    CUtiString &operator+=(const utifloat i);
	friend CUtiString operator+(const CUtiString &s1, const CUtiString &s2);
	friend CUtiString operator+(const U_CHAR *s, const CUtiString &s2);
	friend CUtiString operator+(const U_CHAR c, const CUtiString &s2);
	CUtiString operator+(const U_CHAR *s) const;
	CUtiString operator+(const U_CHAR c) const;
	CUtiString operator+(const int i) const;
	CUtiString operator+(const float f) const;
    CUtiString operator+(const double f) const;
	U_CHAR &operator[](unsigned int p);
	const U_CHAR &operator[](unsigned int p) const;

#ifndef _MSC_VER
#ifdef WIN32
bool less_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(_wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) < 0); }
	bool lessEqual_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(_wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) <= 0); }
	bool equal_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(_wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) == 0); }
	bool notEqual_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(_wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) != 0); }
	bool greaterEqual_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(_wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) >= 0); }
	bool greater_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(_wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) > 0); }

#else
    bool less_lowerCase(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscasecmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) < 0);}
    bool lessEqual_lowerCase(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscasecmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) <= 0);}
    bool equal_lowerCase(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscasecmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) == 0);}
    bool notEqual_lowerCase(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscasecmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) != 0);}
    bool greaterEqual_lowerCase(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscasecmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) >= 0);}
    bool greater_lowerCase(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscasecmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) > 0);}
#endif
#else

	bool less_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) < 0); }
	bool lessEqual_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) <= 0); }
	bool equal_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) == 0); }
	bool notEqual_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) != 0); }
	bool greaterEqual_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) >= 0); }
	bool greater_lowerCase(const CUtiString &s) const { if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false); return(wcsicmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) > 0); }
#endif


	bool operator<(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) < 0);}
	bool operator<=(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) <= 0);}
	bool operator==(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) == 0);}
	bool operator!=(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) != 0);}
	bool operator>=(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) >= 0);}
	bool operator>(const CUtiString &s) const {if ((!m_container->m_string.getCArray()) || (!s.m_container->m_string.getCArray())) return (false);return(wcscmp(m_container->m_string.getCArray(), s.m_container->m_string.getCArray()) > 0);}
	bool operator==(const TCHAR *s) const {if ((!m_container->m_string.getCArray()) || (!s)) return (false);return(wcscmp(m_container->m_string.getCArray(), s) == 0);}
	bool operator<(const TCHAR *s) const {if ((!m_container->m_string.getCArray()) || (!s)) return (false);return(wcscmp(m_container->m_string.getCArray(), s) < 0);}
	bool operator<=(const TCHAR *s) const {if ((!m_container->m_string.getCArray()) || (!s)) return (false);return(wcscmp(m_container->m_string.getCArray(), s) <= 0);}
	bool operator!=(const TCHAR *s) const {if ((!m_container->m_string.getCArray()) || (!s)) return (false);return(wcscmp(m_container->m_string.getCArray(), s) != 0);}
	bool operator>=(const TCHAR *s) const {if ((!m_container->m_string.getCArray()) || (!s)) return (false);return(wcscmp(m_container->m_string.getCArray(), s) >= 0);}
	bool operator>(const TCHAR *s) const {if ((!m_container->m_string.getCArray()) || (!s)) return (false);return(wcscmp(m_container->m_string.getCArray(), s) > 0);}

    CUtiString operator()(unsigned int p, int l) const;
    bool IsEmpty() const {return(m_container->m_length != 0);};
	HResult Read(CStream & Stream, int Flags, int version);
	HResult Write(CStream & Stream, int Flags, int version);
	void LoadFromFile(const U_CHAR* filename);
	void SaveToFile(const U_CHAR* filename);
	void LoadFromFile(const CUtiString &filename);
	void SaveToFile(const CUtiString &filename);
	void DecLength(int i);
	void ExpandToSize(int s);
    void ExtractFileNameWithoutExt();
	void ExtractFileName();
	void ExtractRealFileName();
	void ExtractFilePath();
	void ExtractFileExt();

    CUtiString getFileExtP() const;
    CUtiString getFileExt() const;
    CUtiString getFilePath() const;
    CUtiString getFileName() const;
    CUtiString getFileNameWithoutExt() const;
    CUtiString getRealFileName() const;
        CUtiString LowerCase() const;
        CUtiString UpperCase() const;
        void ToLowerCase();
        void ToUpperCase();
	void LoadRawString(const wchar_t* c, int l);


    SCode convertToCode();
	void MakeFile();
	int LastDelimiter(const U_CHAR* substr) const;
        int NextDelimiter(const U_CHAR* substr) const;
        int NextDelimiter(int startpos, const U_CHAR* substr) const;
	void ReplaceChar(TCHAR oldchar, TCHAR newchar);

	//void ReadProp(char* name, CUtiString* result);
	void LoadAnsi(const char* c);
	char* GetAnsi() const;
    CUtiString replace(U_CHAR oldchar, U_CHAR newchar) const;
    CUtiString replace(const CUtiString& oldstr, const CUtiString& newstr) const;
    //CUtiString replaceAll(const CUtiString& regex, const CUtiString& replacement);
    //CUtiString replaceFirst(const CUtiString& regex, const CUtiString& replacement);
    void split(const CUtiString& str, CStrings& array, bool merge=false) const;
    uti_refptr<CStrings> split(const CUtiString& str, bool merge=false) const;
    bool contains(const CUtiString& str) const;
    int indexOf(const CUtiString& str, int start=0) const;
    int lastIndexOf(const CUtiString& str, int end=-1) const;
    bool endsWith(const CUtiString& other) const;
   // bool regionMatches(??);
    CUtiString substring(int start, int end=-1) const;
    static CUtiString join(CUtiString delimiter, const CStrings& str);
    void insert(int pos, U_CHAR str);
    void insert(int pos, float str);
    void insert(int pos, double str);
    void insert(int pos, bool str);
    void insert(int pos, int str);


    void append(const CUtiString& str);
    void append(U_CHAR str);
        void append(float str);
        void append(double str);
        void append(bool str);
        void append(int str);
        bool isNumber() const;

    static CUtiString valueOf(bool v);
    static CUtiString valueOf(char v);
    static CUtiString valueOf(double v);
    static CUtiString valueOf(float v);
    static CUtiString valueOf(int v);
    static bool parseBoolean(const CUtiString& str);
    static int parseInt(const CUtiString& str);
    static int parseInt(const CUtiString& str, int radix);
    static float parseFloat(const CUtiString& str);
    static double parseDouble(const CUtiString& str);
    static CUtiString toHexString(int i);
    static CUtiString toOctalString(int i);
    static CUtiString toBinaryString(int i);
    static CUtiString toString(int i, int radix);
    static void parseBase64Binary(const CUtiString& str, Array<unsigned char>& data);
    static CUtiString printBase64Binary(const Array<unsigned char>& data);
    static void parseHexBinary(const CUtiString& str, Array<unsigned char>& data);
    static CUtiString printHexBinary(const Array<unsigned char>& data);
    static bool isDigit(U_CHAR ch);
    static bool isLetter(U_CHAR ch);
    static bool isWhitespace(U_CHAR ch);
    static bool isLetterOrDigit(U_CHAR ch);
    static bool isLowerCase(U_CHAR ch);
    static bool isUpperCase(U_CHAR ch);
    static U_CHAR toLowerCase(U_CHAR ch);
    static U_CHAR toUpperCase(U_CHAR ch);
    static int digit(wchar_t ch, int radix);
    static wchar_t forDigit(int digit, int radix);
    static CUtiString format(const CUtiString &format, const CStrings& str);

    int hashCode() const;
  //Better format?
};

class CStringJoiner : public ReferenceUtiOB
{
protected:
    std::vector<CUtiString> m_content;
    CUtiString m_part,m_front, m_back, m_emptyvalue;
    bool m_hasEmptyValue;

public:
    CStringJoiner(const CUtiString& part, const CUtiString&front=UTISTR(""), const CUtiString&back=UTISTR(""));
    int length();
    CStringJoiner& add(const CUtiString& str);
    CStringJoiner& add(const CStrings& str);
   //CStringJoiner& merge(const CStringJoiner& other);
    CStringJoiner& setEmptyValue(const CUtiString& val);
    virtual CUtiString toString();
};


class CStrings : public ReferenceADR
{
public:
    CStrings();
    virtual CUtiString getLine(int l) const=0;
    //inline CUtiString& operator[](int l) {return line(l);};
    //virtual const CUtiString& line(int l) const=0;
    //inline const CUtiString& operator[](int l) const {return line(l);};
    virtual void setLine(int l, const CUtiString& str)=0;
    virtual int size() const=0;
    virtual void clear()=0;
    virtual void deleteLine(int i)=0;
    virtual void add(const CUtiString& str)=0;
    virtual void addCutLine(const CUtiString& str);
    virtual void saveToFile(const CUtiString& filename);
    virtual void loadFromFile(const CUtiString& filename);

    virtual void addStrings(const CStrings& str);
    virtual int indexOf(const CUtiString& str);
    virtual int indexOfName(const CUtiString& str);
    virtual int exchange(int i, int j);
    virtual void insert(int i, const CUtiString&str)=0;
    virtual void sort()=0;
    virtual CUtiString getText();
    virtual void setText(const CUtiString& t);
    virtual void parseStringList(const CUtiString& source, const CUtiString& seperator);
    virtual void printTextToStdOut();

};

class CStringListInternData;

class CStringList : public CStrings
{
protected:
    uti_refptr<CStringListInternData> m_data;
public:
    CStringList(bool undo=false);
    ~CStringList();
    virtual CUtiString getLine(int l) const;
    virtual void setLine(int l, const CUtiString& str);
    virtual int size() const;
    virtual void clear();
    virtual void deleteLine(int i);
    virtual void add(const CUtiString& str);
    virtual void insert(int i, const CUtiString&str);
    virtual void sort();
    virtual void undo();
    virtual void redo();
    virtual void clearUndoRedo();
    virtual bool hasUndo();
    virtual bool hasRedo();
};

void csv_parse(const CUtiString& str, const CUtiString& seperator, const CUtiString& up, std::vector<CUtiString>& seg);



void ReadString(CStream & Stream, CUtiString &s, int version);
void WriteString(CStream & Stream, const  CUtiString &s, int version);

const CUtiString CodeToString(const SCode s2);
const CUtiString CodeToFileNameString(const SCode s2);

const CUtiString RandomString(TCHAR* start, int l, int h, TCHAR* ende);
void ReadStringWide(CStream & Stream, CUtiString &s);
void WriteStringWide(CStream & Stream, const CUtiString &s);
void ReadStringMulti(CStream & Stream, std::string &s);
void WriteStringMulti(CStream & Stream, const std::string &s);

std::wostream & operator<<(std::wostream & os, const CUtiString & str);
std::ostream & operator<<(std::ostream & os, const CUtiString & str);

bool utimatch(const wchar_t* pattern, const wchar_t* name);

inline bool isWhite(wchar_t t)
{
  return t == ' ' || t == 13 || t == 10 || t == 7;
};

inline bool isInteger(wchar_t t)
{
  return t >= '0' && t <= '9';
};

inline bool isAlpha(wchar_t t)
{
  return t == '_' || (t >= 'a' && t <= 'z') ||  (t >= 'A' && t <= 'Z');
};

inline bool isAlphaNumeric(wchar_t t)
{
  return t == '_' || (t >= 'a' && t <= 'z') ||  (t >= 'A' && t <= 'Z') || (t >= '0' && t <= '9');
};

inline bool isNumeric(wchar_t t)
{
  return (t >= '0' && t <= '9');
};

inline bool isUnderLine(wchar_t a)
{
  return (a == UTEXT('_'));
};

inline bool isBezeichner(wchar_t a)
{
  return (isNumeric(a) || isAlpha(a) || isUnderLine(a));
};

bool isNumeric(const CUtiString& str);
bool isBezeichner(const CUtiString& name);
bool isMultiBezeichner(const CUtiString& name);
int utistrlen(const wchar_t* t);



class CGameEngineNameStore : public CUtiOB
{

    MAKE_SINGLETON_FAST(CGameEngineNameStore);
    std::map<CUtiString, int> m_name_to_int;
   // std::map<int, CUtiString> m_int_to_name;
    std::vector<CUtiString> m_int_to_name_list;
    int m_nextindex;
    int m_emptystring;
    int m_nullstring;
    int m_zerostring;
public:
    CSpinLock spinlock;
    CGameEngineNameStore();
    ~CGameEngineNameStore();
    static int s_emptystring;

    void registerName(int id, const CUtiString& value);
    int convertNameToInt(const CUtiString& id);
    const CUtiString &convertIntToName(int i);
    inline int getEmptyString()const {return m_emptystring;};
    inline int getNullString() const {return m_nullstring;};
    inline int getZeroString() const {return m_zerostring;};
};

#define IDSTR(a) CUtiIDString(UTISTR(a))

const int EMPTYSTRING_ID = 0;
//Problem im PDDLSystem: erlaubt kein SubStrings!
#ifdef NDEBUG
#define NDEBUGSTRING
#endif
struct CUtiIDString
{
protected:
    int m_id;
#ifndef NDEBUGSTRING
    const wchar_t* m_realstring;
#endif
public:
    inline CUtiIDString() {
        m_id = EMPTYSTRING_ID;
#ifndef NDEBUGSTRING
        m_realstring =CGameEngineNameStore::getInstance()->convertIntToName(EMPTYSTRING_ID).c_str();
       #endif
    }
    explicit inline CUtiIDString(const CUtiString& str) {
        m_id = CGameEngineNameStore::getInstance()->convertNameToInt(str);
#ifndef NDEBUGSTRING
        m_realstring =  CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str();
        assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());

       #endif
    }
    inline static CUtiIDString fromID(int id, const CUtiString& str)
    {
        CUtiIDString res;
        res.m_id = id;
#ifndef NDEBUGSTRING
         res.m_realstring =  CGameEngineNameStore::getInstance()->convertIntToName(res.m_id).c_str();
         assert(res.m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(res.m_id).c_str());

#endif
        return res;
    }

    inline bool empty() const {
#ifndef NDEBUGSTRING
        assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());
#endif
        return m_id == EMPTYSTRING_ID;
    };
    inline void load(const CUtiString& str)
    {
        m_id = CGameEngineNameStore::getInstance()->convertNameToInt(str);

#ifndef NDEBUGSTRING
 m_realstring = CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str();
 assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());

#endif
    }

    inline CUtiIDString& operator=(const CUtiIDString& str)
    {
        m_id = str.m_id;

#ifndef NDEBUGSTRING
 m_realstring = str.m_realstring;
 assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());
#endif
 return *this;
    }/*
    inline bool operator==(const CUtiString& str) const
    {

#ifndef NDEBUG
 assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id));

#endif
       return m_id == CGameEngineNameStore::getInstance()->convertNameToInt(str);
    }*/
    inline bool operator==(const CUtiIDString& id) const
    {

#ifndef NDEBUGSTRING
 assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());
 assert(id.m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(id.m_id).c_str());

#endif
 return m_id ==id.m_id;
    }
    inline bool operator!=(const CUtiIDString& id) const
    {

#ifndef NDEBUGSTRING
 assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());
 assert(id.m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(id.m_id).c_str());

#endif
 return m_id !=id.m_id;
    }
    inline bool operator<(const CUtiIDString& id) const
    {

#ifndef NDEBUGSTRING
 assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());
 assert(id.m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(id.m_id).c_str());

#endif
 return m_id < id.m_id;

    }
    inline bool operator>(const CUtiIDString& id) const
    {

#ifndef NDEBUGSTRING
 assert(m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());
 assert(id.m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(id.m_id).c_str());

#endif
 return m_id > id.m_id;

    }
    inline int getID() const
    {
        return m_id;
    }
    inline void setID(int id)
    {
        m_id = id;
#ifndef NDEBUGSTRING
               m_realstring = CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str();

               #endif
    }

    inline CUtiString str() const
    {
#ifndef NDEBUGSTRING
        assert(m_id == 0 || m_realstring == CGameEngineNameStore::getInstance()->convertIntToName(m_id).c_str());
#endif
        return CGameEngineNameStore::getInstance()->convertIntToName(m_id);
   }
    inline bool isLispVariable() const
    {
        CUtiString s = str();
        if (s.size() < 2) return false;
        if (s[0] == L'?') {
            return true;
        }
        return false;

    }
    inline bool isExternalVariable() const
    {
        CUtiString s = str();
        if (s.size() < 2) return false;
        if (s[0] == L'#') {
            return true;
        }
        return false;
    }
};

inline CUtiString operator+(const CUtiString& str, const CUtiIDString& str2)
{
    return str+str2.str();
}

const int LOG_ANIMATION = 1;
const int LOG_ENTITY = 2;
const int LOG_SOUND = 4;
const int LOG_STATEMACHINE = 8;
const int LOG_FILE = 8;
const int LOG_RENDERING = 16;
const int LOG_ENTITYMODUL = 32;
const int LOG_PDDL = 64;
const int LOG_LUA = 128;

class CLogSystem
{
public:
   int m_logstate;
   int m_loglevel;
   TIME m_starttime;
   std::wostream* outputFile;
   MAKE_SINGLETON_FAST(CLogSystem);
   CLogSystem();
   void doLog(int s, int level, int indent, CUtiOB* object, const CUtiString& str);
   std::wostream& lout();
};

std::wostream& logout();

#define LOG_MSG(logstate, loglevel, object, message) \
    if (CLogSystem::Instance != NULL && ((CLogSystem::Instance->m_logstate & (logstate)) != 0) && (CLogSystem::Instance->m_loglevel >= (loglevel))) {\
        CLogSystem::Instance->doLog(logstate, loglevel, 0, object, message);\
    }

#define LOG_MSG_INDENT(logstate, loglevel,indent, object, message) \
    if (CLogSystem::Instance != NULL && ((CLogSystem::Instance->m_logstate & (logstate)) != 0) && (CLogSystem::Instance->m_loglevel >= (loglevel))) {\
        CLogSystem::Instance->doLog(logstate, loglevel, indent, object, message);\
    }

#define OnLogReturn(a, b){if (!(a)) {LOG_MSG(LOG_ENTITY, 2, NULL, UTISTR("Condition: ")+ UTISTR(#a)+ UTISTR(" in ")+ UTISTR(  __FILE__)+UTISTR(": ")+ __LINE__+ UTISTR(" date: ")+ UTISTR( __DATE__)+UTISTR(" Message: ")+ (b)); return;};}
#define OnLogReturnNULL(a, b){if (!(a)) {LOG_MSG(LOG_ENTITY, 2, NULL, UTISTR("Condition: ")+ UTISTR(#a)+ UTISTR(" in ")+ UTISTR(  __FILE__)+UTISTR(": ")+ __LINE__+ UTISTR(" date: ")+ UTISTR( __DATE__)+UTISTR(" Message: ")+ (b)); return NULL;};}
#define OnLogDisplayMessage(a, b){if (!(a)) {LOG_MSG(LOG_ENTITY, 2, NULL, UTISTR("Condition: ")+ UTISTR(#a)+ UTISTR(" in ")+ UTISTR(  __FILE__)+UTISTR(": ")+ __LINE__+ UTISTR(" date: ")+ UTISTR( __DATE__)+UTISTR(" Message: ")+ (b)); };}

#endif
