/***************************************************************************
 *   Copyright (C) 2006 by Michael Staud   *
 *   michael.staud@uni-ulm.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiType.h"
#include "UtiArchiv.h"
#include "utimathbox.h"
#include "UtiTypeHandlers.h"
#include <assert.h>
#include <string.h>

CRTTIManager rttimanager;

CRTTIManager* CRTTIManager::m_manager = NULL;

CRTTIManager* CRTTIManager::getInstance()
{
    if (m_manager == NULL) {
      m_manager = new CRTTIManager();
    };
    return m_manager;
}

CUtiException::CUtiException()
{
    m_n = UTEXT("");
};


CUtiException::CUtiException(const CUtiString& s)
{
   m_n = s;
}

const CUtiString CClassRTTI::getName()
{
    return UTEXT("");
};

const CUtiString CCallClass::getName()
{
    return UTEXT("");
};

void CXWriteInterface::Perform(CUtiOB* obj, CXNode* n, CWriteReadInterface* wri)
{
    CXWriteInterface* i = new CXWriteInterface(n, wri);
    obj->TypeInfo(i);
    delete i;
}

CXWriteInterface::CXWriteInterface(CXNode* n, CWriteReadInterface* wri)
{
    this->m_n = n;
    this->m_wri = wri;
}

void CXWriteInterface::RInt(int &i, int std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropInt(Name, i);
    }
}

void CXWriteInterface::RByte(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropInt(Name, i);
    }
}

void CXWriteInterface::REnum(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropInt(Name, i);
    }
}
void CXWriteInterface::RFloat(float &i, float std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropFloat(Name, i);
    }
}
void CXWriteInterface::RDouble(double &i, double std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropDouble(Name, i);
    }
}
void CXWriteInterface::RVector(CVector &i, const CVector &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropVector(Name, i);
    }
}
void CXWriteInterface::RVector4(CVector4 &i, const CVector4 &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropVector4(Name, i);
    }
}
void CXWriteInterface::RCode(SCode &i, SCode std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropCode(Name, i);
    }
}
void CXWriteInterface::RReference(CUtiOB* &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {

       SCode c1;
       if (i) {
       c1 = i->GetCode();
       } else {
          c1 = NULLCODE;
       };
        m_n->setPropCode(Name, c1);
    }
}
void CXWriteInterface::RObject(CUtiOB &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        CXNode* node = m_n->addNode(Name);
        //t = XReadClass(node, wri, this);
        XWriteClass(node, m_wri, &i);
        //   i.xread(node, wri);
    }
}

void CXWriteInterface::RObjectOwns(CUtiOB* &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
   if (flags & TYPE_STREAM)
   {
      CXNode* node = m_n->addNode(Name);

      XWriteClass(node, m_wri, i);

   }
}
void CXWriteInterface::RBool(bool &i, bool std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropBool(Name, i);
    }
}
void CXWriteInterface::RString(CUtiString &i, const CUtiString&  std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        m_n->setPropString(Name, i.getString());
    }
}
void CXWriteInterface::RObjectList(CCacheList &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        CXNode* node = m_n->addNode(Name);
        node->setPropInt(UTISTR("utireserved_count"), i.count);
        for (int l = 0; l < i.count; l++)
        {
            CXNode* subnode = node->addNode(L"class");
            CUtiOB* ob = (CUtiOB*)i[l];
            XWriteClass(subnode, m_wri, ob);
        };
    };
}

void CXWriteInterface::RObjectListB(CBaseListReferenceAdapter *i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    uti_refptr<CBaseListReferenceAdapter> ad(i);
    if (flags & TYPE_STREAM)
    {
        CXNode* node = m_n->addNode(Name);
        node->setPropInt(UTISTR("utireserved_count"), ad->size());
        for (int l = 0; l < ad->size(); l++)
        {
            CXNode* subnode = node->addNode(L"class");
            CUtiOB* ob = (CUtiOB*)ad->get(l);
            XWriteClass(subnode, m_wri, ob);
        };
    };
}


void CXReadInterface::Perform(CUtiOB* obj, CXNode* n, CWriteReadInterface* wri)
{
    CXReadInterface* i = new CXReadInterface(n, wri);
    obj->TypeInfo(i);
    delete i;
}

CXReadInterface::CXReadInterface(CXNode* n, CWriteReadInterface* wri)
{
    this->m_n = n;
    this->m_wri = wri;
}

void CXReadInterface::RInt(int &i, int std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropInt(Name, std);
    }
}

void CXReadInterface::RByte(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropInt(Name, std);
    }
}

void CXReadInterface::REnum(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropInt(Name, std);
    }
}
void CXReadInterface::RFloat(float &i, float std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropFloat(Name, std);
    }
}
void CXReadInterface::RDouble(double &i, double std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropDouble(Name, std);
    }
}
void CXReadInterface::RVector(CVector &i, const CVector &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropVector(Name, std);
    }
}
void CXReadInterface::RVector4(CVector4 &i, const CVector4 &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropVector4(Name, std);
    }
}
void CXReadInterface::RCode(SCode &i, SCode std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropCode(Name, std);
    }
}
void CXReadInterface::RReference(CUtiOB* &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        SCode code = m_n->getPropCode(Name, NULLCODE);
        if (compareCode(code, NULLCODE))
        {
            i = NULL;

        }
        else
        {
            dynamic_cast<CReadInfo*>(m_wri)->addLink(&i, code);

        };
    }
}
void CXReadInterface::RObject(CUtiOB &i,  int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        CXNode* node = m_n->getNodeByName(Name);
        if (node)
        {
            //t = XReadClass(node, wri, this);
            i.xread(node, m_wri);
            dynamic_cast<CReadInfo*>(m_wri)->addObject(&i, i.GetCode());

        };
    }
}

void CXReadInterface::RObjectOwns(CUtiOB* &i,  int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
   if (flags & TYPE_STREAM)
   {
      CXNode* node = m_n->getNodeByName(Name);
      if (node)
      {
         if (i) delete i;
         i = NULL;
          i = XReadClass(node, m_wri, this);
         //i.xread(node, wri);
          if (i)
         dynamic_cast<CReadInfo*>(m_wri)->addObject(i, i->GetCode());

      };
   }
}



void CXReadInterface::RBool(bool &i, bool std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropBool(Name, std);
    }
}
void CXReadInterface::RString(CUtiString &i, const CUtiString&  std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i = m_n->getPropString(Name, std);
    }
}
void CXReadInterface::RObjectList(CCacheList &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    if (flags & TYPE_STREAM)
    {
        i.Clear();
        CXNode* node = m_n->getNodeByName(Name);
        if (node)
        {
            int c = node->getPropInt(UTISTR("utireserved_count"), 0);
            for (int l = 0; l < c; l++)
            {
                CXNode* subnode = node->getChild(l);
                CUtiOB* ob = XReadClass(subnode, m_wri, this);
                dynamic_cast<CReadInfo*>(m_wri)->addObject(ob, ob->GetCode());
                i.Add(ob);
            };
        };
    };
}

void CXReadInterface::RObjectListB(CBaseListReferenceAdapter *i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a)
{
    uti_refptr<CBaseListReferenceAdapter> ad(i);
    if (flags & TYPE_STREAM)
    {
        ad->clear();
        CXNode* node = m_n->getNodeByName(Name);
        if (node)
        {
            int c = node->getPropInt(UTISTR("utireserved_count"), 0);
            for (int l = 0; l < c; l++)
            {
                CXNode* subnode = node->getChild(l);
                CUtiOB* ob = XReadClass(subnode, m_wri, this);
                dynamic_cast<CReadInfo*>(m_wri)->addObject(ob, ob->GetCode());
                ad->push_back(check_cast<ReferenceUtiOB*>(ob));
            };
        };
    };
}



void TypedReferenceUtiOB::xread(CXNode* n, CWriteReadInterface* wri)
{
   CXReadInterface::Perform(this, n, wri);
}

void TypedReferenceUtiOB::xwrite(CXNode* n, CWriteReadInterface* wri)
{
    CXWriteInterface::Perform(this, n, wri);

}

void CLinkInfo::correct(CUtiOB*obj)
{

}

void CStdLinkInfo::correct(CUtiOB* obj)
{
   *m_obj = obj;
}

CReferenceLinkInfo::~CReferenceLinkInfo()
{
 // delete m_adapter;
}

void CReferenceLinkInfo::correct(CUtiOB* obj)
{
if (obj != NULL) {
  utiassert(dynamic_cast<ReferenceUtiOB*>(obj) != NULL);
    m_adapter->set(static_cast<ReferenceUtiOB*>(obj));
 } else {
    m_adapter->set(NULL);
 }
}


CListReferenceLinkInfo::~CListReferenceLinkInfo()
{
//  delete m_adapter;
}
void CListReferenceLinkInfo::correct(CUtiOB* obj)
{
  if (obj != NULL) {
  utiassert(dynamic_cast<ReferenceUtiOB*>(obj) != NULL);
  m_adapter->set(m_index, static_cast<ReferenceUtiOB*>(obj));
  } else {
    m_adapter->set(m_index, NULL);
  };
}


CReadInfo::CReadInfo()
{
    m_links.deleteItems = true;
  //  ObjList.deleteItems = true;
}

void CReadInfo::link()
{
    for (int i = 0; i < m_links.count; i++)
    {
        CLinkInfo* info = (CLinkInfo*)m_links[i];
       /* std::map<SCode, CUtiOB*>::iterator iter = m_objmap.find(info->m_code);
        if (iter == m_objmap.end()) {
           info->correct( NULL);
           std::wcout << L"Link nicht gefunden" << std::endl;
	  utiassert(false);
        } else {*/
           info->correct(findObject(info->m_code));
        //};
    };
    m_links.Clear();
}

CUtiOB* CReadInfo::findObject(SCode code)
{
  std::map<SCode, CUtiOB*>::iterator iter = m_objmap.find(code);
        if (iter == m_objmap.end()) {
	  std::wcout << L"Link nicht gefunden" << std::endl;
	  utiassert(false);
	  return NULL;
	}
   return iter->second;
}

CInstantReadInfo::CInstantReadInfo()
{
  
}
void CInstantReadInfo::addLink(CUtiOB** obj, SCode code)
{
    CReadInfo::addLink(obj, code);
    link();
}

void CInstantReadInfo::addLink(CBaseReferenceAdapter* obj, SCode code)
{
    CReadInfo::addLink(obj, code);link();
}

void CInstantReadInfo::addLink(CBaseListReferenceAdapter* obj, int index, SCode code)
{
    CReadInfo::addLink(obj, index, code);link();
}

     

CReadInfo::~CReadInfo()
{
}

void CReadInfo::addLink(CUtiOB** obj, SCode code)
{
 // utiassert(m_objmap.find(code) != m_objmap.end());
    CStdLinkInfo* obj1 = new CStdLinkInfo();
    obj1->m_obj = obj;
    obj1->m_code = code;
    m_links.Add(obj1);

}

void CReadInfo::addLink(CBaseReferenceAdapter* obj, SCode code)
{
  CReferenceLinkInfo* obj1 = new CReferenceLinkInfo();
    obj1->m_adapter = obj;
    obj1->m_code = code;
    m_links.Add(obj1);
}

void CReadInfo::addLink(CBaseListReferenceAdapter* obj, int index, SCode code)
{
  CListReferenceLinkInfo* obj1 = new CListReferenceLinkInfo();
    obj1->m_adapter = obj;
    obj1->m_index = index;
    obj1->m_code = code;
    m_links.Add(obj1);
}

void CReadInfo::addObject(CUtiOB* obj, SCode code)
{
  /*  CObjInfo* obj1 = new CObjInfo();
    obj1->obj = obj;
    obj1->code = code;
   ObjList.Add(obj1);*/
  if (m_objmap.find(code) == m_objmap.end())
    m_objmap.insert(std::make_pair(code,obj));
}

CXNode::CXNode(CXNode* parent, CXDocument* doc)
{
    this->m_parent = parent;
    this->m_doc = doc;
}


void CXNode::copyTo(CXNode* node)
{
   node->clearNodes();
   node->clearProperties();
   copyPropertiesTo(node);
   int count = getChildCount();
   for (int i = 0; i < count; i++) {
      CXNode* n = getChild(i);
      CXNode* newnode =  node->addNode(n->getName());
      n->copyTo(newnode);
   }
}

CXNode* findSubElement(CXNode* node, const CUtiString& name)
{
 if (node == NULL) return NULL;
 return node->getNodeByName(name);
}

int getElementCount(CXNode* node,const CUtiString& name )
{
   int r = 0;
   for (int i = 0; i < node->getChildCount(); i++) {
       CXNode* n = node->getChild(i);
       if (n->getName() == name) {
           r++;

       }
   }

   return r;
}

   void getElementsByName(CXNode* node, const CUtiString& name, CNormalList<CXNode>& res)
   {

      for (int i = 0; i < node->getChildCount(); i++) {
          CXNode* n = node->getChild(i);
          if (n->getName() == name) {
              res.push_back(n);

          }
      }


   }


CXNode* findSubElementByProperty(CXNode* node, const CUtiString& nodename, const CUtiString& propertyname, const CUtiString& propertyvalue)
{
  if (node == NULL) return NULL;
  int c = node->getChildCount();
  for (int i = 0; i < c; i++) {
    CXNode* subnode = node->getChild(i);
    if (subnode->getName() != nodename) continue;
    if (subnode->getPropString(propertyname, UTISTR("")) == propertyvalue) {
      return subnode;
    };
  };
  return NULL;
}

CXNode* CXNode::getNodeOrCreate(const CUtiString& name)
{
  CXNode* n = getNodeByName(name);
  if (n != NULL) return n;
  return addNode(name);
}

void CXNode::setPropData(const CUtiString& name, const Array<unsigned char>& data)
{
  setPropData(name, data.array, data.size()); 
}

bool CXNode::getPropData(const CUtiString& name, Array<unsigned char>&data) const
{
  errorassert(data.size() == getPropDataSize(name), UTISTR("The size of the array must be equal to the data size."));
  size_t s = data.size();
  bool ret=  getPropData(name, data.array, s); 
  
  return ret;
}


CXNode* CXNode::getNodeByPath(const CUtiString& str)
{
    if (str.size() == 0) {
        return this;
    }
    {
   // originalname = str;
   // level.setSize(Path_MaxLevels);
      int i = 0;
      int l = str.size();
      int start=0;
      int ende=-1;
     // levels = -1;
      CXNode* ret = this;
      CUtiString nodename;
      while (true) {
          if (str[i] == '/' && i >= l) {
              ende = i;

              str.Copy(nodename, start, ende-start);

              ret = ret->getNodeByName(nodename);
              if (!ret) {
                  return NULL;
              }
              start = i+1;
              if (i >= l) break;
          };
          i++;

      };
      return ret;
    }
}

int CXNode::getNodeCount(const CUtiString& name)
{
  int c = 0;
  int realc = getChildCount();
  for (int i = 0; i < realc; i++) {
    CXNode* n = getChild(i);
    if (n->getName() == name){ c++;}
  };
  return c;
}

CDataProp::CDataProp() : CBaseProp()
{
    m_data = NULL;
    m_size = 0;
}

CDataProp::~CDataProp()
{
    delete [] m_data;
}

CXUtiNode::CXUtiNode(CXNode* parent, CXDocument* doc) : CXNode(parent, doc)
{
    //m_nodes.deleteItems = true;
    m_properties.deleteItems = true;
    m_name = -1;
}
void CXUtiNode::setName(const CUtiString& name)
{
    ((CXUtiDocument*)m_doc)->removeIndex(this->m_name);
    this->m_name = ((CXUtiDocument*)m_doc)->addString(name);
}
CUtiString CXUtiNode::getName() const
{
    return ((CXUtiDocument*)m_doc)->getString(m_name);
}


void CXUtiNode::copyPropertiesTo(CXNode* node)
{
   for (int i = 0; i < m_properties.count; i++) {
       CBaseProp* prop = check_cast<CBaseProp*>(m_properties[i]);
       CUtiString propname = ((CXUtiDocument*)m_doc)->getString(prop->m_id);
       if (dynamic_cast<CStringProp*>(prop) != NULL) {
          CStringProp* pdata = check_cast<CStringProp*>(prop);
          node->setPropString(propname, pdata->m_b);
       } else if (dynamic_cast<CBoolProp*>(prop) != NULL) {
          CBoolProp* pdata = check_cast<CBoolProp*>(prop);
          node->setPropBool(propname, pdata->m_b);
       } else if (dynamic_cast<CIntProp*>(prop) != NULL) {
          CIntProp* pdata = check_cast<CIntProp*>(prop);
          node->setPropInt(propname, pdata->m_b);
       } else if (dynamic_cast<CFloatProp*>(prop) != NULL) {
          CFloatProp* pdata = check_cast<CFloatProp*>(prop);
          node->setPropFloat(propname, pdata->m_b);
       } else if (dynamic_cast<CDoubleProp*>(prop) != NULL) {
          CDoubleProp* pdata = check_cast<CDoubleProp*>(prop);
          node->setPropDouble(propname, pdata->m_b);
       } else if (dynamic_cast<CCodeProp*>(prop) != NULL) {
          CCodeProp* pdata = check_cast<CCodeProp*>(prop);
          node->setPropCode(propname, pdata->m_b);
       } else if (dynamic_cast<CVectorProp*>(prop) != NULL) {
          CVectorProp* pdata = check_cast<CVectorProp*>(prop);
          node->setPropVector(propname, pdata->m_b);
       } else if (dynamic_cast<CVector4Prop*>(prop) != NULL) {
          CVector4Prop* pdata = check_cast<CVector4Prop*>(prop);
          node->setPropVector4(propname, pdata->m_b);
       } else if (dynamic_cast<CDataProp*>(prop) != NULL) {
          CDataProp* pdata = check_cast<CDataProp*>(prop);
          node->setPropData(propname, pdata->m_data, pdata->m_size);
       } else {
       assert(false);
       }

   }
}

//#define XUTITYPE(
/*var
   index: integer;
   prop: TBaseProp;
begin
   index := TXUtiDocument(doc).findString(name);
   if index = -1 then begin
      Result := std;
   end else begin
      if m_prophash.Find(Pointer(index), Pointer(prop)) then begin
         if not (prop is TBoolProp) then begin
            raise Exception.Create(name + UTEXT(' ist keine Bool Eigenschaft.'));
         end else begin
            Result := TBoolProp(prop).value;
         end;
      end else begin

 Result := std;
      end;
   end; */

CBaseProp* CXUtiNode::findPropByIndex(int x) const
{
    if (this->m_prophash.find(x) != this->m_prophash.end())
    {
        return this->m_prophash.find(x)->second ;

    }
    else
    {
        /*for (int i = 0; i < m_properties.count; i++) {
        	CBaseProp* prop = (CBaseProp*)m_properties[i];
        	if (prop->id == x) return prop;

        };*/
        return NULL;
    }
}



void CXUtiNode::setPropBool(const CUtiString& name, const bool b)
{
#define PROP CBoolProp
    int index;
    //PROP* prop;
    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_b = b;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp* prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
           if (!p)
               throw CUtiTypeException();
            p->m_b = b;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_b = b;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
bool CXUtiNode::getPropBool(const CUtiString& name, const bool std) const
{
#define PROP CBoolProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return std;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                return std;
            return p->m_b;
        }
        else
        {
            return std;
        };
    };
#undef PROP
}
void CXUtiNode::setPropInt(const CUtiString& name, const int b)
{
#define PROP CIntProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_b = b;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                throw CUtiTypeException();
            p->m_b = b;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_b = b;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
int CXUtiNode::getPropInt(const CUtiString& name, const int std) const
{
#define PROP CIntProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return std;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                return std;
            return p->m_b;
        }
        else
        {
            return std;
        };
    };
#undef PROP
}
void CXUtiNode::setPropFloat(const CUtiString& name, const float b)
{
#define PROP CFloatProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_b = b;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                throw CUtiTypeException();
            p->m_b = b;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_b = b;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
float CXUtiNode::getPropFloat(const CUtiString& name, const float std) const
{
#define PROP CFloatProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return std;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                return std;
            return p->m_b;
        }
        else
        {
            return std;
        };
    };
#undef PROP
}
void CXUtiNode::setPropDouble(const CUtiString& name, const double b)
{
#define PROP CDoubleProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_b = b;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                throw CUtiTypeException();
            p->m_b = b;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_b = b;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
double CXUtiNode::getPropDouble(const CUtiString& name, const double std) const
{
#define PROP CDoubleProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return std;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p) {
                if (dynamic_cast<CFloatProp*>(prop) != NULL) {
                    return check_cast<CFloatProp*>(prop)->m_b;
                }
                return std;
            }
            return p->m_b;
        }
        else
        {
            return std;
        };
    };
#undef PROP
}
void CXUtiNode::setPropCode(const CUtiString& name, const SCode &b)
{
#define PROP CCodeProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_b = b;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                throw CUtiTypeException();
            p->m_b = b;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_b = b;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
SCode CXUtiNode::getPropCode(const CUtiString& name, const SCode &std) const
{
#define PROP CCodeProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return std;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                return std;
            return p->m_b;
        }
        else
        {
            return std;
        };
    };
#undef PROP
}
void CXUtiNode::setPropVector(const CUtiString& name, const CVector &b)
{
#define PROP CVectorProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_b = b;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                throw CUtiTypeException();
            p->m_b = b;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_b = b;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
CVector CXUtiNode::getPropVector(const CUtiString& name, const CVector &std) const
{
#define PROP CVectorProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return std;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                return std;
            return p->m_b;
        }
        else
        {
            return std;
        };
    };
#undef PROP
}
void CXUtiNode::setPropVector4(const CUtiString& name, const CVector4 &b)
{
#define PROP CVector4Prop
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_b = b;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                throw CUtiTypeException();
            p->m_b = b;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_b = b;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
CVector4 CXUtiNode::getPropVector4(const CUtiString& name, const CVector4 &std) const
{
#define PROP CVector4Prop
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return std;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                return std;
            return p->m_b;
        }
        else
        {
            return std;
        };
    };
#undef PROP
}
void CXUtiNode::setPropString(const CUtiString& name, const CUtiString& b)
{
#define PROP CStringProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_b = b;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                throw CUtiTypeException();
            p->m_b = b;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_b = b;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
CUtiString CXUtiNode::getPropString(const CUtiString& name, const CUtiString& std) const
{
#define PROP CStringProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return std;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
               return std;
            return p->m_b;
        }
        else
        {
            return std;
        };
    };
#undef PROP
}

int CXUtiNode::getPropDataSize(const CUtiString& name) const
{
    #define PROP CDataProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return 0;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                return 0;
            //return p->b;
            return p->m_size;
        }
        else
        {
            return 0;
        };
    };
#undef PROP

}


void CXUtiNode::setPropData(const CUtiString& name, const unsigned char* Data, size_t size)
{
#define PROP CDataProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        PROP* p = new PROP();
        index = ((CXUtiDocument*)m_doc)->addString(name);
        p->m_id = index;
        p->m_data = new unsigned char[size];
        memcpy(p->m_data, Data, size);
        p->m_size = (int)size;
        m_prophash[p->m_id] = p;
        m_properties.Add(p);
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                throw CUtiTypeException();
            if (p->m_data)
                delete [] p->m_data;
            p->m_data = new unsigned char[size];
            memcpy(p->m_data, Data, size);
            p->m_size = (int)size;
        }
        else
        {
            PROP* p = new PROP();
            p->m_id = index;
            p->m_data = new unsigned char[size];
            memcpy(p->m_data, Data, size);
            p->m_size = (int)size;
            m_prophash[p->m_id] = p;
            m_properties.Add(p);
        };
    };
#undef PROP
}
bool CXUtiNode::getPropData(const CUtiString& name, unsigned char* &Data, size_t &size) const
{
#define PROP CDataProp
    int index;

    index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return false;
    }
    else
    {
        CBaseProp*prop = findPropByIndex(index);
        if (prop)
        {
            PROP* p = dynamic_cast<PROP*>(prop);
            if (!p)
                return false;
            //return p->b;
            if (Data == NULL){
            Data = new unsigned char[p->m_size];
            size = p->m_size;
            } else {
                utiassert(size >= (size_t)p->m_size);
            }
            memcpy(Data, p->m_data, size);
            return true;
        }
        else
        {
            return false;
        };
    };
#undef PROP
}
bool CXUtiNode::propExists(const CUtiString& name) const
{
    int index = ((CXUtiDocument*)m_doc)->findString(name);
    if (index == -1)
    {
        return false;
    }
    else
    {
        CBaseProp* prop = findPropByIndex(index);
        if (prop)
            return true;
    };
    return false;
}
HResult CXUtiNode::Read(CStream & Stream, int Flags, int version)
{
    int  c, id, typ;
    CUtiString s;

    clearNodes();
    clearProperties();
    Stream.ReadInteger(m_name);
    ((CXUtiDocument*)m_doc)->incIndex(m_name);
    Stream.ReadInteger(c);
    for (int i = 0; i < c; i++)
    {
        Stream.ReadInteger(id);
        typ = id >> 24;
        id = id & 0xFFFFFF;
        ((CXUtiDocument*)m_doc)->incIndex(id);
        switch (typ) {
           case X_BOOL:

        {
            CBoolProp* b = new CBoolProp();
            b->m_id = id;
            Stream.ReadBoolean(b->m_b);
            m_prophash[b->m_id] = b;
            m_properties.Add(b);

        } ;break;
           case X_CODE:
        {
            CCodeProp* b = new CCodeProp();
            b->m_id = id;
            Stream.ReadData(&b->m_b, sizeof(b->m_b));
            m_prophash[b->m_id] = b;
            m_properties.Add(b);
        };break;
           case X_VECTOR4:
        {
            CVector4Prop* b = new CVector4Prop();
            b->m_id = id;
            CVector4f rd;
            Stream.ReadData(&rd, sizeof(rd));
            b->m_b = toDouble(rd);
            m_prophash[b->m_id] = b;
            m_properties.Add(b);
        };break;
           case X_DATA:
        {
            CDataProp* b = new CDataProp();
            b->m_id = id;
            Stream.ReadData(&b->m_size, sizeof(b->m_size));
            b->m_data = new unsigned char [b->m_size];
            Stream.ReadData(b->m_data, b->m_size);
            m_prophash[b->m_id] = b;
            m_properties.Add(b);
        };break;
           case X_DOUBLE:
        {
            CDoubleProp* b = new CDoubleProp();
            b->m_id = id;
            Stream.ReadData(&b->m_b, sizeof(b->m_b));
            m_prophash[b->m_id] = b;
            m_properties.Add(b);
        };break;
           case X_SINGLE:
        {
            CFloatProp* b = new CFloatProp();
            b->m_id = id;
            Stream.ReadData(&b->m_b, sizeof(b->m_b));
            m_prophash[b->m_id] = b;
            m_properties.Add(b);
        };break;
           case X_INT:
        {
            CIntProp* b = new CIntProp();
            b->m_id = id;
            Stream.ReadData(&b->m_b, sizeof(b->m_b));
            m_prophash[b->m_id] = b;
            m_properties.Add(b);
        };break;
           case X_STRING:
        {
            CStringProp* b = new CStringProp();
            b->m_id = id;
            //Stream.ReadData(&b->b, sizeof(b->b));
            ReadStringWide(Stream, b->m_b);
            m_prophash[b->m_id] = b;
            m_properties.Add(b);
        };break;
           case X_VECTOR:
        {
            CVectorProp* b = new CVectorProp();
            b->m_id = id;
            ReadVector(&b->m_b, Stream);
           // b->b.Read(Stream);
            m_prophash[b->m_id] = b;
            m_properties.Add(b);
        };break;
        };
    };


    Stream.ReadInteger(c);
    for (int i = 0; i < c; i++)
    {
        CXUtiNode* n = new CXUtiNode(this, m_doc);
        n->Read(Stream, Flags, version);
        m_nodes.push_back(n);
        m_nodehash[n->m_name] = n;

    };
    return 0;
}
HResult CXUtiNode::Write(CStream & Stream, int Flags, int version)
{
    int  c, id;
    CUtiString s;


    Stream.WriteInteger(m_name);
    c = m_properties.count;
    Stream.WriteInteger(c);
    for (int i = 0; i < c; i++)
    {
        CBaseProp* prop = (CBaseProp*)m_properties[i];
        if (dynamic_cast<CBoolProp*>(prop))
        {
            CBoolProp* b = (CBoolProp*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_BOOL << 24);
            Stream.WriteInteger(id);
            Stream.WriteBoolean(b->m_b);
        } else
        if (dynamic_cast<CCodeProp*>(prop))
        {
            CCodeProp* b = (CCodeProp*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_CODE << 24);
            Stream.WriteInteger(id);
            Stream.WriteData(&b->m_b, sizeof(b->m_b));
        } else
        if (dynamic_cast<CVector4Prop*>(prop))
        {
            CVector4Prop* b = (CVector4Prop*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_VECTOR4 << 24);
            Stream.WriteInteger(id);
            CVector4f data = toFloat(b->m_b);
            Stream.WriteData(&data, sizeof(data));
        } else
        if (dynamic_cast<CDataProp*>(prop))
        {
            CDataProp* b = (CDataProp*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_DATA << 24);
            Stream.WriteInteger(id);
            //Stream.WriteData(&b->b, sizeof(b->b));
            Stream.WriteInteger(b->m_size);
            Stream.WriteData(b->m_data, b->m_size);
        } else
        if (dynamic_cast<CDoubleProp*>(prop))
        {
            CDoubleProp* b = (CDoubleProp*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_DOUBLE << 24);
            Stream.WriteInteger(id);
            Stream.WriteData(&b->m_b, sizeof(b->m_b));
        } else
        if (dynamic_cast<CFloatProp*>(prop))
        {
            CFloatProp* b = (CFloatProp*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_SINGLE << 24);
            Stream.WriteInteger(id);
            Stream.WriteData(&b->m_b, sizeof(b->m_b));
        } else
        if (dynamic_cast<CIntProp*>(prop))
        {
            CIntProp* b = (CIntProp*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_INT << 24);
            Stream.WriteInteger(id);
            Stream.WriteData(&b->m_b, sizeof(b->m_b));
        } else

        if (dynamic_cast<CStringProp*>(prop))
        {
            CStringProp* b = (CStringProp*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_STRING << 24);
            Stream.WriteInteger(id);
            //Stream.WriteData(&b->b, sizeof(b->b));
            WriteStringWide(Stream, b->m_b);
        } else
        if (dynamic_cast<CVectorProp*>(prop))
        {
            CVectorProp* b = (CVectorProp*)(prop);
            id = (b->m_id & 0xFFFFFF) | (X_VECTOR << 24);
            Stream.WriteInteger(id);
            CVector3f data = toFloat(b->m_b);
            Stream.WriteData(&data, sizeof(data));
        };

    };

    c = m_nodes.size();
    Stream.WriteInteger(c);
    for (int i = 0; i < c; i++)
    {
        CXUtiNode* n = &m_nodes[i];
        n->Write(Stream, Flags, version);

    };

    return 0;
}
CXNode* CXUtiNode::addNode(const CUtiString& name)
{
    CHECKEXISTANCE;
    CXUtiNode* n = new CXUtiNode(this, m_doc);
    n->setName(name);
//    utiassert(m_nodehash.find(n->m_name) == m_nodehash.end());
    m_nodehash[n->m_name] = n;
    m_nodes.push_back(n);
    return n;
}
CXNode* CXUtiNode::getNodeByName(const CUtiString& name)
{
   int nm = static_cast<CXUtiDocument*>(m_doc)->findString(name);
   if (nm == -1) return NULL;
   std::map <int, CXUtiNode*>::iterator iter = m_nodehash.find(nm);
   if (iter == m_nodehash.end()) {
      return NULL;
   } else {
      return iter->second;
   };
  /*  for (int i = 0; i < Nodes.count; i++)
    {
        CXUtiNode* n = (CXUtiNode*)Nodes[i];
        if (n->Name == nm)
        {
            return n;
        };

};*/
    return NULL;
}
void CXUtiNode::deleteNode(const CUtiString& name)
{
   int nm = static_cast<CXUtiDocument*>(m_doc)->findString(name);
   if (nm == -1) return;
   std::map <int, CXUtiNode*>::iterator iter = m_nodehash.find(nm);
   if (iter == m_nodehash.end()) {
      return;
   }
   m_nodehash.erase(iter);
    for (int i = 0; i < m_nodes.size(); i++)
    {
        CXUtiNode* n = &m_nodes[i];
        if (n->m_name ==  nm)
        {
            m_nodes.deleteItem(i);
            return;

        };

    };
}
void CXUtiNode::deleteItem(int i)
{
   std::map <int, CXUtiNode*>::iterator iter = m_nodehash.find( (&m_nodes[i])->m_name);
   if (iter == m_nodehash.end()) {
   //   return;
   } else {
   m_nodehash.erase(iter);
   }
    m_nodes.deleteItem(i);
}
void CXUtiNode::remove(CXNode* n)
{
   std::map <int, CXUtiNode*>::iterator iter = m_nodehash.find( ((CXUtiNode*)n)->m_name);
   if (iter == m_nodehash.end()) {
      return;
   }
   m_nodehash.erase(iter);
    m_nodes.remove(check_cast<CXUtiNode*>( n));
}

void CXUtiNode::removeProperty(const CUtiString& name)
{
    int nm = static_cast<CXUtiDocument*>(m_doc)->findString(name);
    if (nm == -1) return;
    std::map<int, CBaseProp*>::iterator it = m_prophash.find(nm);
    if (it == m_prophash.end())
    {
        return;
    }
    m_properties.Remove(it->second);
    m_prophash.erase(it);

}

int CXUtiNode::getChildCount() const
{
    return m_nodes.size();
}
CXNode* CXUtiNode::getChild(int i) const
{

    return const_cast<CXUtiNode*>( &m_nodes[i]);
}
void CXUtiNode::clearNodes()
{
    m_nodes.clear();
    m_nodehash.clear();
}
void CXUtiNode::clearProperties()
{
    m_properties.Clear();
    m_prophash.clear();
}

void CXUtiNode::xread(CXNode* n, CWriteReadInterface* wri)
{
assert(m_doc->getRoot() == this);
Array<unsigned char> data;
   if (n->getPropData(UTISTR("nodedata"), data)) {
       uti_refptr<CMemoryStream> str(new CMemoryStream(data.getCArray(), data.size(), false));
       m_doc->loadFromStream(*str.get());
   } else {
       clearNodes();
       clearProperties();
   }
}

void CXUtiNode::xwrite(CXNode* n, CWriteReadInterface* wri)
{
    assert(m_doc->getRoot() == this);
    uti_refptr<CMemoryStream> str(new CMemoryStream());
    m_doc->saveToStream(*str.get());
    n->setPropData(UTISTR("nodedata"), str->getMemory(), str->getMemorySize());

}

CXDocument::CXDocument() 
{
  m_root = NULL;
}

CXDocument::~CXDocument()
{
}
CXNode* CXDocument::getRoot()
{
    return m_root.get();

}

CXUtiDocument::CXUtiDocument(): CXDocument()
{
    m_strlist.deleteItems = true;
    m_root = new CXUtiNode(NULL, this);
    m_root->setName(L"root");
}
CXUtiDocument::~CXUtiDocument()
{
}
int CXUtiDocument::addString(const CUtiString&s)
{
    int index = findString(s);
    if (index == -1)
    {
        CXName* n = new CXName();
        n->m_name = s;
        n->m_ref = 1;
        m_listindex[s] = m_strlist.count;
        m_strlist.Add(n);
        return m_strlist.count-1;


    }
    else
    {
        incIndex(index);
        return index;
    }
}
void CXUtiDocument::removeString(const CUtiString&id)
{
    int index = findString(id);
    const CUtiString& c = ((CXName*)m_strlist[index])->m_name.getString();

    m_listindex.erase(c);

    removeIndex(index);
}
void CXUtiDocument::removeIndex(int index)
{
    if (index == -1)
        return;
    if (index >= m_strlist.count) return;
    CXName* n = (CXName*)m_strlist[index];
    n->m_ref--;
    if (n->m_ref == 0)
    {
        m_strlist.Delete(index);
    }
}
void CXUtiDocument::incIndex(int index)
{
    CXName* n = (CXName*)m_strlist[index];
    n->m_ref++;
}
int CXUtiDocument::findString(const CUtiString& s) const
{

    if (m_listindex.find(s) != m_listindex.end())
    {
        int index = m_listindex.find(s)->second;
        return index;
    }
    return -1;
}
const CUtiString CXUtiDocument::getString(int id)
{
    CXName* n = (CXName*)m_strlist[id];
    return n->m_name;
}
int CXUtiDocument::saveToStream(CStream &Stream)
{
    char code[4];
    int c;
    code[0] = 'U';
    code[1] =('t');
    code[2] = ('i');
    code[3] = 'X';
    Stream.WriteData(code, 4);
    int flags = 0;
    Stream.WriteData(&flags, sizeof(flags));
    int version = 12;
    Stream.WriteData(&version, sizeof(version));

    // node.read(str, 0, currentversion);
    c = m_strlist.count;
    Stream.WriteData(&c, sizeof(c));
    for (int i = 0; i < c; i++)
    {
        WriteStringWide(Stream, static_cast<CXName*>(m_strlist[i])->m_name);
    };
    m_root->Write(Stream, 0, version);
    return UTI_OK;
}
int CXUtiDocument::loadFromStream(CStream &Stream)
{
    int c;
    CUtiString s;
    int flags, version;
    char code[4];
    Stream.ReadData(code, 4);
    if (code[0] != 'U' || code[1] != ('t') || code[2] != ('i') || code[3] != ('X')) {
        return UTI_WRONGFILEFORMAT;
    }
    //  flags := 0;
    Stream.ReadInteger(flags);
    if (flags != 0) {
       return UTI_WRONGFILEFORMAT;
    }
    // version := currentversion;
    Stream.ReadInteger(version);
    if (version != 12) {
       return UTI_WRONGFILEFORMAT;
    }
    // node.read(str, 0, currentversion);
    //c := hash.Count;
    Stream.ReadInteger(c);
    if (c <= 0) {
       return UTI_FILEFORMATERROR;
    }
    m_root->clearProperties();
    m_root->clearNodes();
    m_strlist.Clear();
    for (int i = 0; i < c; i++)
    {
        ReadStringWide(Stream, s);
        //cout << s.c_str() << endl;
        CXName* n = new CXName();
        n->m_ref = 0;
        n->m_name = s;
        m_listindex[s.getString()] = m_strlist.count;
        m_strlist.Add(n);

    };
    m_root->Read(Stream, 0, version);
    return UTI_OK;
}

void CXUtiDocument::xread(CXNode* n, CWriteReadInterface* wri)
{
//TODO: Implement me
   getRoot()->copyTo(n);
}

void CXUtiDocument::xwrite(CXNode* n, CWriteReadInterface* wri)
{
n->copyTo(getRoot());
}


int CXUtiDocument::saveToString(CUtiString& res)
{
    uti_refptr<CMemoryStream> str(new CMemoryStream(true));
    int r = saveToStream(*str.get());
    if (r != 0) {
        return r;
    }
   /* int size = CUti64Coder::getDataLength(str->getMemorySize());
    char* c = new char[size+1];

    CUti64Coder::writeData(str->getMemory(), str->getMemorySize(), c, size);
    c[size] = 0;
    CUtiString res;
    res.setLength(size);
    for (int i = 0; i < size; i++) {
        res[i] = c[i];
    }
    delete[]c;*/
    res = CUtiString::printBase64Binary(Array<unsigned char>(str->getMemory(), str->getMemorySize()));
    return 0;
}

int CXUtiDocument::loadFromString(const CUtiString& res)
{
   Array<unsigned char> data;
   CUtiString::parseBase64Binary(res, data);
   uti_refptr<CMemoryStream> str(new CMemoryStream(data, false));
   int r = loadFromStream(*str.get());
return r;
}

int CXUtiDocument::saveToFile(const CUtiString& name)
{
    CFileStream* str = new CFileStream(name, S_WRITE);
    if (str == NULL || str->fail())
    {
        if (str) delete str;
        return UTI_WRITEERROR;
    }
    saveToStream(*str);
    delete str;
    return UTI_OK;
}
int CXUtiDocument::loadFromFile(const CUtiString& name)
{
	if (FileExists(name.c_str()) == false) {
        return UTI_READERROR;
	};
    uti_refptr<CFileStream> str(new CFileStream(name, S_READ));
    if (str.get() == NULL || str->fail())
    {

        return UTI_READERROR;
    }
    UTI_INT64 filesize = FileSize(name);
    uti_refptr<CBufferedReadStream> bfrstr(new CBufferedReadStream(str.get(), 16384,filesize));
    loadFromStream(*bfrstr.get());
   // delete str;
    return UTI_OK;

}

CRTTIManager::CRTTIManager()
{
    this->m_classes.deleteItems = true;
}

CRTTIManager::~CRTTIManager()
{
}


SCode CRTTIManager::classNameToCode(const CUtiString& name)
{
   auto it = m_name_to_code.find(name);
   if (it == m_name_to_code.end()) {
       return NULLCODE;
   }
   return it->second;
}

void CRTTIManager::addRTTI(CClassRTTI* cl, LPVOID Data)
{
    m_classes.Add(cl);
    utiassert(m_clmap.find(cl->GetCode()) == m_clmap.end());
    m_clmap.insert(std::make_pair(cl->GetCode(), cl));
    m_name_to_code.insert(std::make_pair(cl->getName(), cl->GetCode()));
}

CUtiOB* CRTTIManager::createClass(SCode code, CUtiOB* parent)
{
   /* for (int i = 0; i < classes.count; i++)
    {
        CClassRTTI* cl = (CClassRTTI*)classes[i];
        if (compareCode(cl->GetCode(), code))
        {
            return cl->CreateClass(parent);
        };

};*/
    std::map<SCode, CClassRTTI*>::iterator it = m_clmap.find(code);
    if (it == m_clmap.end()) return NULL;
    CClassRTTI* cl = it->second;

    return cl->createClass(parent);
}

CUtiString CRTTIManager::getName(SCode code)
{
  std::map<SCode, CClassRTTI*>::iterator it = m_clmap.find(code);
    if (it == m_clmap.end()) return UTISTR("Unknown Object");
    CClassRTTI* cl = it->second;
    return cl->getName();
};

CWriteInfo::CWriteInfo()
{

}

CWriteInfo::~CWriteInfo()
{

}

bool CWriteInfo::needToBeSaved(CUtiOB* obj)
{
  std::map<CUtiOB*, SCode>::iterator it = m_objmap.find(obj);
  return it == m_objmap.end();
}

void CWriteInfo::addObject(CUtiOB* obj, const  SCode &code)
{
  utiassert(m_objmap.find(obj) == m_objmap.end());
  m_objmap.insert(std::make_pair(obj, code));
}
SCode CWriteInfo::getCode(CUtiOB* obj)
{
  std::map<CUtiOB*, SCode>::iterator it = m_objmap.find(obj);
  utiassert(it != m_objmap.end());
  return it->second;
}

CDuplicateWriteInfo::CDuplicateWriteInfo(CClassCollector* tosaveobjs, CClassCollector* nottosave)
{
    m_tosaveobjs = tosaveobjs;
   
  m_nottosave = nottosave;
}
bool CDuplicateWriteInfo::needToBeSaved(CUtiOB* obj)
{
  if (obj == NULL) return CWriteInfo::needToBeSaved(obj);
  if (m_tosaveobjs->hasObject(obj)) {
   return CWriteInfo::needToBeSaved(obj); 
  }
  m_nottosave->addObject(obj);
  return false;
}

void CDuplicateWriteInfo::addObject(CUtiOB* obj, const  SCode &code)
{
  utiassert(obj->GetCode() == code);
  CWriteInfo::addObject(obj, code);
}

SCode CDuplicateWriteInfo::getCode(CUtiOB* obj)
{
  if (m_nottosave->hasObject(obj)) return obj->GetCode();
  return CWriteInfo::getCode(obj);
}


CDuplicateReadInfo::CDuplicateReadInfo(CClassCollector* tosaveobjs, CClassCollector* nottosave)
{
  m_tosaveobjs = tosaveobjs;
  m_nottosave = nottosave;
}

CUtiOB* CDuplicateReadInfo::findObject(SCode code)
{
    if (m_nottosave->hasObject(code)) {
      return m_nottosave->getObject(code);
      
    }
    return CReadInfo::findObject(code);
}

void XWriteClass(CXNode* n, LPVOID wri, CUtiOB* cl)
{
    CHECKEXISTANCEOF(n)
     CHECKEXISTANCEOF(cl)
   if (cl) {
      SCode code= cl->GetTypeCode();
      n->setPropCode(UTISTR("utireserved_typecode"), code);
      n->setPropCode(UTISTR("utireserved_objectcode"), cl->GetCode());
      cl->xwrite(n,(CWriteReadInterface*) wri);
   } else {
      n->setPropCode(UTISTR("utireserved_typecode"), NULLCODE);
      n->setPropCode(UTISTR("utireserved_objectcode"), NULLCODE);
   };
}


void XReadClassNoCreate(CXNode* n, LPVOID wri, CUtiOB* cl)
{
    SCode code = n->getPropCode(UTISTR("utireserved_typecode"), NULLCODE);
    if (code == NULLCODE){
      return;
    }
    assert(code == cl->GetTypeCode());
    CUtiOB* result = cl;//CRTTIManager::getInstance()->createClass(code, parent);
    if (result == NULL) {
     std::wcout << "Did not found the class code: " << CodeToString(code) << std::endl;
       return;
    };

    result->xread(n, (CWriteReadInterface*) wri);
    //return result;
};

CUtiOB* XReadClass(CXNode* n, LPVOID wri, CUtiOB* parent)
{
    SCode code = n->getPropCode(UTISTR("utireserved_typecode"), NULLCODE);
    if (code == NULLCODE){
      return NULL;
    }
    CUtiOB* result = CRTTIManager::getInstance()->createClass(code, parent);
    if (result == NULL) {
     std::wcout << "Did not found the class code: " << CodeToString(code) << std::endl;
       return NULL;
    };
  
    result->xread(n, (CWriteReadInterface*) wri);
    return result;
};

CUtiOB* XCopyClass(CUtiOB* obj, CUtiOB* parent, CWriteInfo* writeinfo, CReadInfo* readinfo)
{
   CXDocument* xml = new CXUtiDocument();
    uti_refptr<CWriteInfo> writeinfo2;
   if (!writeinfo) {
       writeinfo2 = (new CWriteInfo());
       writeinfo = writeinfo2.get();


   }
   uti_refptr<CReadInfo> readinfo2;
   if (!readinfo) {
     readinfo2 = (new CReadInfo());
     readinfo = readinfo2.get();
   }
    writeinfo->addObject(obj, obj->GetCode());
    XWriteClass(xml->getRoot(), writeinfo, obj);
    CUtiOB* result;
    
    result = XReadClass(xml->getRoot(), readinfo, parent);
    SCode c = xml->getRoot()->getPropCode(UTISTR("utireserved_objectcode"), NULLCODE);
    delete xml;
    utiassert(result != NULL);
    readinfo->addObject(result, c);
    readinfo->link();
    return result;
}


int XSaveToFile(const CUtiString&  string, CUtiOB* cl)
{
      uti_refptr<CXDocument> xml(new CXUtiDocument());
    uti_refptr<CWriteInfo> writeinfo(new CWriteInfo());
     writeinfo->addObject(cl, cl->GetCode());
    XWriteClass(xml->getRoot(), writeinfo.get(), cl);
    int errorcode = xml->saveToFile(string);
 
  
    return errorcode;
};

int XLoadFromFile(const CUtiString&  string, CUtiOB* parent, CUtiOB* &result)
{

#ifndef NDEBUG
   // std::wcout <<"Loading XML:" << string << std::endl;
#endif
    result = NULL;
    uti_refptr<CXDocument> xml(new CXUtiDocument());
    int errorcode = xml->loadFromFile(string);
    if (errorcode != UTI_OK) {
      
		result = NULL;
        return errorcode;
    }
    uti_refptr<CReadInfo> r(new CReadInfo());
    result = XReadClass(xml->getRoot(), r.get(), parent);
    SCode c = xml->getRoot()->getPropCode(UTISTR("utireserved_objectcode"), NULLCODE);
    xml = NULL;
    if (result == NULL) {
    
        return UTI_CLASSNOTFOUND;
    }
    r->addObject(result, c);
    r->link();

    return UTI_OK;
}

int XSaveToNode(CXNode* stream, CUtiOB* cl)
{
    CHECKEXISTANCEOF(stream)
    CHECKEXISTANCEOF(cl)
            // uti_refptr<CXDocument> xml(new CXUtiDocument());
    uti_refptr<CWriteInfo> writeinfo(new CWriteInfo());
     writeinfo->addObject(cl, cl->GetCode());
    XWriteClass(stream, writeinfo.get(), cl);
  //  int errorcode = xml->saveToStream(*stream);


    return UTI_OK;
};
int XLoadFromNode(CXNode* stream, CUtiOB* parent, CUtiOB* &result)
{
    result = NULL;

    uti_refptr<CReadInfo> r(new CReadInfo());
    result = XReadClass(stream, r.get(), parent);
    SCode c = stream->getPropCode(UTISTR("utireserved_objectcode"), NULLCODE);

    if (result == NULL) {

        return UTI_CLASSNOTFOUND;
    }
    r->addObject(result, c);
    r->link();

    return UTI_OK;
}


int XSaveToStream(CStream* stream, CUtiOB* cl)
{
      uti_refptr<CXDocument> xml(new CXUtiDocument());
    uti_refptr<CWriteInfo> writeinfo(new CWriteInfo());
     writeinfo->addObject(cl, cl->GetCode());
    XWriteClass(xml->getRoot(), writeinfo.get(), cl);
    int errorcode = xml->saveToStream(*stream);
 
  
    return errorcode;
};

int XLoadFromStream(CStream* stream, CUtiOB* parent, CUtiOB* &result)
{
    result = NULL;
    uti_refptr<CXDocument> xml(new CXUtiDocument());
    int errorcode = xml->loadFromStream(*stream);
    if (errorcode != UTI_OK) {
      
		result = NULL;
        return errorcode;
    }
    uti_refptr<CReadInfo> r(new CReadInfo());
    result = XReadClass(xml->getRoot(), r.get(), parent);
    SCode c = xml->getRoot()->getPropCode(UTISTR("utireserved_objectcode"), NULLCODE);
    xml = NULL;
    if (result == NULL) {
    
        return UTI_CLASSNOTFOUND;
    }
    r->addObject(result, c);
    r->link();

    return UTI_OK;
}

CClassCollector::CClassCollector()
{
  m_tag = 0;
}

CClassCollector::~CClassCollector()
{
  
}

void CClassCollector::addObject(CUtiOB* obj)
{
  if (hasObject(obj) == false) {
    m_objtocode.insert(std::make_pair(obj, obj->GetCode()));
    m_codetoobj.insert(std::make_pair(obj->GetCode(), obj));
  };
}

CUtiOB* CClassCollector::getObject(const SCode& code)
{
  std::map<SCode, CUtiOB*>::iterator it = m_codetoobj.find(code);
  if (it == m_codetoobj.end()) {
    utiassert(false);
    return NULL;
  };
  return it->second;
}

bool CClassCollector::hasObject(CUtiOB* obj)
{
  return m_objtocode.find(obj) != m_objtocode.end();
}

bool CClassCollector::hasObject(const SCode& code)
{
  return m_codetoobj.find(code) != m_codetoobj.end();
}

void CClassCollector::setTag(int tag)
{
  m_tag = tag;
}

int CClassCollector::getTag()
{
  return m_tag;
}


InputOutputHook::InputOutputHook()
{

}

void InputOutputHook::printInt(int i)
{
    CUtiString s;
    s = i;
    printString(s);
}
void InputOutputHook::printDouble(double d)
{
    CUtiString s;
    s = (float)d;
    printString(s);
}
void InputOutputHook::printBool(bool d)
{
    CUtiString s;
    if (d) {
        s = UTISTR("1");
    } else {
        s = UTISTR("0");
    }
    printString(s);
}
void InputOutputHook::printChar(wchar_t ch)
{
    CUtiString s(UTISTR(" "));
    s[0] = ch;
    printString(s);
}

CUtiString convertToString(const CMatrixf& b)
{
    CUtiString str;
    str = UTISTR("[");
    str += b(0, 0);
    str += UTISTR(", ");
    str += b(0, 1);
    str += UTISTR(", ");
    str += b(0, 2);
    str += UTISTR(", ");
    str += b(0, 3);
    str += UTISTR("\n");
    str += b(1, 0);
    str += UTISTR(", ");
    str += b(1, 1);
    str += UTISTR(", ");
    str += b(1, 2);
    str += UTISTR(", ");
    str += b(1, 3);
    str += UTISTR("\n");
    str += b(2, 0);
    str += UTISTR(", ");
    str += b(2, 1);
    str += UTISTR(", ");
    str += b(2, 2);
    str += UTISTR(", ");
    str += b(2, 3);
    str += UTISTR("\n");
    str += b(3, 0);
    str += UTISTR(", ");
    str += b(3, 1);
    str += UTISTR(", ");
    str += b(3, 2);
    str += UTISTR(", ");
    str += b(3, 3);
    str += UTISTR("]");
    return str;
}

CUtiString convertToString(const CMatrix& b)
{
    CUtiString str;
    str = UTISTR("[");
    str += b(0, 0);
    str += UTISTR(", ");
    str += b(0, 1);
    str += UTISTR(", ");
    str += b(0, 2);
    str += UTISTR(", ");
    str += b(0, 3);
    str += UTISTR("\n");
    str += b(1, 0);
    str += UTISTR(", ");
    str += b(1, 1);
    str += UTISTR(", ");
    str += b(1, 2);
    str += UTISTR(", ");
    str += b(1, 3);
    str += UTISTR("\n");
    str += b(2, 0);
    str += UTISTR(", ");
    str += b(2, 1);
    str += UTISTR(", ");
    str += b(2, 2);
    str += UTISTR(", ");
    str += b(2, 3);
    str += UTISTR("\n");
    str += b(3, 0);
    str += UTISTR(", ");
    str += b(3, 1);
    str += UTISTR(", ");
    str += b(3, 2);
    str += UTISTR(", ");
    str += b(3, 3);
    str += UTISTR("]");
    return str;
}

CUtiString convertToString(const CBox& b)
{
    CUtiString str;
    str = UTISTR("(");
    str += b.m_min.x;
    str += UTISTR(", ");
    str += b.m_min.y;
    str += UTISTR(", ");
    str += b.m_min.z;
    str += UTISTR(") - (");
    str += b.m_max.x;
    str += UTISTR(", ");
    str += b.m_max.y;
    str += UTISTR(", ");
    str += b.m_max.z;
    str += UTISTR(")");
    return str;
}


CUtiString convertToString(const CRect& b)
{
    CUtiString str;
    str = UTISTR("(");
    str += b.m_min.x;
    str += UTISTR(", ");
    str += b.m_min.y;
    str += UTISTR(") - (");
    str += b.m_max.x;
    str += UTISTR(", ");
    str += b.m_max.y;
    str += UTISTR(")");
    return str;
}


CUtiString convertToString(const CBoxf& b)
{
    CUtiString str;
    str = UTISTR("(");
    str += b.m_min.x;
    str += UTISTR(", ");
    str += b.m_min.y;
    str += UTISTR(", ");
    str += b.m_min.z;
    str += UTISTR(") - (");
    str += b.m_max.x;
    str += UTISTR(", ");
    str += b.m_max.y;
    str += UTISTR(", ");
    str += b.m_max.z;
    str += UTISTR(")");
    return str;
}


CUtiString convertToString(const CRectf& b)
{
    CUtiString str;
    str = UTISTR("(");
    str += b.m_min.x;
    str += UTISTR(", ");
    str += b.m_min.y;
    str += UTISTR(") - (");
    str += b.m_max.x;
    str += UTISTR(", ");
    str += b.m_max.y;
    str += UTISTR(")");
    return str;
}

CUtiString convertToString(const CIntBox& b)
{
    CUtiString str;
    str = UTISTR("(");
    str += b.m_min.x;
    str += UTISTR(", ");
    str += b.m_min.y;
    str += UTISTR(", ");
    str += b.m_min.z;
    str += UTISTR(") - (");
    str += b.m_max.x;
    str += UTISTR(", ");
    str += b.m_max.y;
    str += UTISTR(", ");
    str += b.m_max.z;
    str += UTISTR(")");
    return str;
}


CUtiString convertToString(const CIntRect& b)
{
    CUtiString str;
    str = UTISTR("(");
    str += b.m_min.x;
    str += UTISTR(", ");
    str += b.m_min.y;
    str += UTISTR(") - (");
    str += b.m_max.x;
    str += UTISTR(", ");
    str += b.m_max.y;
    str += UTISTR(")");
    return str;
}


CUtiString convertToString(const CVector4f& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(", ");
        str += v.z;
        str += UTISTR(", ");
        str += v.w;
        str += UTISTR(")");

        return str;
}

CUtiString convertToString(const CVector3f& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(", ");
        str += v.z;
        str += UTISTR(")");

        return str;
}


CUtiString convertToString(const CVector4& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(", ");
        str += v.z;
        str += UTISTR(", ");
        str += v.w;
        str += UTISTR(")");

        return str;
}

CUtiString convertToString(const CVector3& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(", ");
        str += v.z;
        str += UTISTR(")");

        return str;
}


CUtiString convertToString(bool i)
{
    CUtiString s;
    if (i) {
        return UTISTR("1");
    } else {
        return UTISTR("0");

    }

}

CUtiString convertToString(int i)
{
    CUtiString str;
        str += i;

        return str;
}

CUtiString convertToString(float i)
{
    CUtiString str;
        str += i;

        return str;
}

CUtiString convertToString(double i)
{
    CUtiString str;
        str += float(i);

        return str;
}

CVector2 parseVector2(const CUtiString& str)
{
    CVector2f result;

    int v = swscanf(str.c_str(), L"(%f , %f)", &result.x, &result.y);

    if (v < 2) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    return toDouble(result);
}
CVector3 parseVector3(const CUtiString& str)
{
    CVector3f result;

    int v = swscanf(str.c_str(), L"(%f , %f , %f)", &result.x, &result.y, &result.z);

    if (v < 3) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    return toDouble(result);
}
CVector4 parseVector4(const CUtiString& str)
{
    CVector4f result;

    int v = swscanf(str.c_str(), L"(%f , %f , %f, %f)", &result.x, &result.y, &result.z, &result.w);

    if (v < 4) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    return toDouble(result);
}

CRect parseRect(const CUtiString& str)
{CRectf result;

    int v = swscanf(str.c_str(), L"(%f , %f) - ( %f, %f)", &result.m_min.x, &result.m_min.y, &result.m_max.x, &result.m_max.y);

    if (v < 4) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    return toDouble(result);
}
CBox parseBox(const CUtiString& str)
{
    CBoxf result;

        int v = swscanf(str.c_str(), L"(%f, %f ,%f) - (%f, %f, %f)", &result.m_min.x, &result.m_min.y,&result.m_min.z, &result.m_max.x, &result.m_max.y, &result.m_max.z);

        if (v < 6) {

           CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
           error += str + UTEXT("' can not be converted to a CVector value.");

           throw new CUtiException(error);

        }
        return toDouble(result);
}


CIVector2 parseIVector2(const CUtiString& str)
{
    CIVector2 result;

    int v = swscanf(str.c_str(), L"(%i , %i)", &result.x, &result.y);

    if (v < 2) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    return result;
}
CIVector3 parseIVector3(const CUtiString& str)
{
    CIVector3 result;

    int v = swscanf(str.c_str(), L"(%i , %i , %i)", &result.x, &result.y, &result.z);

    if (v < 3) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    return result;
}
CIVector4 parseIVector4(const CUtiString& str)
{
    CIVector4 result;

    int v = swscanf(str.c_str(), L"(%i , %i , %i, %i)", &result.x, &result.y, &result.z, &result.w);

    if (v < 4) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    return result;
}

CIntRect parseIntRect(const CUtiString& str)
{CIntRect result;

    int v = swscanf(str.c_str(), L"(%i , %i) - (%i, %i)", &result.m_min.x, &result.m_min.y, &result.m_max.x, &result.m_max.y);

    if ( v < 4) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    return result;
}
CIntBox parseIntBox(const CUtiString& str)
{
    CIntBox result;

        int v = swscanf(str.c_str(), L"(%i , %i ,%i) - (%i, %i, %i)", &result.m_min.x, &result.m_min.y,&result.m_min.z, &result.m_max.x, &result.m_max.y, &result.m_max.z);

        if (v < 6) {

           CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
           error += str + UTEXT("' can not be converted to a CVector value.");

           throw new CUtiException(error);

        }
        return result;
}

CUtiString convertToString(const CUtiString& str)
{
return str;
}

CUtiString convertToString(const CVector2& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(")");

        return str;
}

CUtiString convertToString(const CVector2f& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(")");

        return str;
}

CUtiString convertToString(const SCode& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += int(v.Data1);
        str += UTISTR(", ");
        str += v.Data2;
        str += UTISTR(", ");
        str += v.Data3;
        str += UTISTR(", ");
        str += v.Data4[0];
        str += UTISTR(", ");
        str += v.Data4[1];
        str += UTISTR(", ");
        str += v.Data4[2];
        str += UTISTR(", ");
        str += v.Data4[3];
        str += UTISTR(", ");
        str += v.Data4[4];
        str += UTISTR(", ");
        str += v.Data4[5];
        str += UTISTR(", ");
        str += v.Data4[6];
        str += UTISTR(", ");
        str += v.Data4[7];
        str += UTISTR(")");

        return str;
}

SCode parseSCode(const CUtiString& str)
{
    SCode result;
    CIVector4 res1, res2, res3;

    int v = swscanf(str.c_str(), L"(%i , %i , %i, %i, %i, %i, %i, %i, %i, %i, %i)", &res1.x, &res1.y, &res1.z, &res1.w, &res2.x, &res2.y, &res2.z, &res2.w, &res3.x, &res3.y, &res3.z);

    if (v < 11) {

       CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
       error += str + UTEXT("' can not be converted to a CVector value.");

       throw new CUtiException(error);

    }
    result.Data1 = res1.x;
    result.Data2 = res1.y;
    result.Data3 = res1.z;
    result.Data4[0] = res1.w;
    result.Data4[1] = res2.x;
    result.Data4[2] = res2.y;
    result.Data4[3] = res2.z;
    result.Data4[4] = res2.w;
    result.Data4[5] = res3.x;
    result.Data4[6] = res3.y;
    result.Data4[7] = res3.z;

    return result;
}

CUtiString convertToString(const CIVector4& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(", ");
        str += v.z;
        str += UTISTR(", ");
        str += v.w;
        str += UTISTR(")");

        return str;
}
CUtiString convertToString(const CIVector3& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(", ");
        str += v.z;
        str += UTISTR(")");

        return str;
}
CUtiString convertToString(const CIVector2& v)
{
    CUtiString str;
        str = UTISTR("(");
        str += v.x;
        str += UTISTR(", ");
        str += v.y;
        str += UTISTR(")");

        return str;
}

CUtiString convertToCpp(const CUtiString&v)
{
  return UTISTR("UTISTR(\"") + v + UTISTR("\")");
}
CUtiString convertToCpp(bool b)
{
  if (b) {
   return UTISTR("true"); 
  } else {
    return UTISTR("false"); 
  }
}
CUtiString convertToCpp(int i)
{
  CUtiString s;
  s += i;
  return s;
}
CUtiString convertToCpp(float b)
{
   CUtiString s;
  s += b;
  return s;
}


bool CBaseCoroutine::execute(ReferenceUtiOB* context)
{
return true;
}

//$cppoutput::begin
CBaseCoroutine::CBaseCoroutine()
{
   m_coroutine_state = -1;
   m_coroutine_subroutine = NULL;
   ;
   ;
}
CBaseCoroutine::~CBaseCoroutine()
{
   ;
}
CUtiString CBaseCoroutine::getClassName() {
  CHECKEXISTANCE;
return UTISTR("CBaseCoroutine");
}
int CBaseCoroutine::getcoroutine_state() const
{
CHECKEXISTANCE;   return m_coroutine_state;
}
void CBaseCoroutine::setcoroutine_state(const int& value0)
{
  CHECKEXISTANCE;
   m_coroutine_state = value0;
}
CBaseCoroutine* CBaseCoroutine::getcoroutine_subroutine()
{
 CHECKEXISTANCE;   return m_coroutine_subroutine.get();
}
const CBaseCoroutine* CBaseCoroutine::getcoroutine_subroutine() const
{
 CHECKEXISTANCE;   return m_coroutine_subroutine.get();
}
void CBaseCoroutine::setcoroutine_subroutine(CBaseCoroutine* value0)
{
  CHECKEXISTANCE;
   m_coroutine_subroutine = value0;
}
const SCode CBaseCoroutine::TYPECODE =
 {
    8643258, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};
const SCode CBaseCoroutine::GetTypeCode()
{
    return CBaseCoroutine::TYPECODE;
}
IMPLEMENT_CLASS_CREATORTYPENOPARENT(CBaseCoroutine);
void CBaseCoroutine::xread(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xread(n, wri);
   uti_readValue(m_coroutine_state, n, wri, UTISTR("coroutine_state"), -1);
   uti_readValue(new CReferenceAdapter<CBaseCoroutine>(m_coroutine_subroutine), n, wri, UTISTR("coroutine_subroutine"), NULL);
   ;
}
void CBaseCoroutine::xwrite(CXNode* n, CWriteReadInterface* wri)
{ CHECKEXISTANCE; 
   ReferenceUtiOB::xwrite(n, wri);
   uti_writeValue(m_coroutine_state, n, wri, UTISTR("coroutine_state"));
   uti_writeValue(new CReferenceAdapter<CBaseCoroutine>(m_coroutine_subroutine), n, wri, UTISTR("coroutine_subroutine"));
}
//$cppoutput::end
