#pragma once
/***************************************************************************
 *   Copyright (C) 2006 by Michael Staud   *
 *   michael.staud@uni-ulm.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UTITYPE__
#define __UTITYPE__

#include <string>
#include <vector>
#include <map>

//#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "utilists.h"


const int TYPE_STREAM = 1;
const int TYPE_EXPOSE = 2;


struct THotKey
{
    int m_first_key;
    int m_alternate_key;
    inline THotKey(int k, int a) {
        m_first_key = k;
        m_alternate_key = a;

    }
    inline THotKey(int k) {
        m_first_key = k;
        m_alternate_key = 0;
    }
    inline THotKey() {
        m_first_key = 0;
        m_alternate_key = 0;
    }
    bool operator==(const THotKey& v) const
    {
        return m_first_key == v.m_first_key && m_alternate_key == v.m_alternate_key;
    }
    bool operator!=(const THotKey& v) const
    {
        return !operator==(v);
    }
};

class CXDocument;
class CXNode;

class CWriteInfo;
class CReadInfo;

class CUtiException
{
public:
  CUtiString m_n;
  CUtiException();
  CUtiException(const CUtiString& s);

};

class CClassRTTI : public CUtiOB
{
public:
  virtual const SCode GetCode(){return NULLCODE;};
  virtual CUtiOB* createClass(CUtiOB* Parent){return NULL;};
  virtual const CUtiString getName();
};



#define IMPLEMENT_CLASS_CREATOR(classname, id) class CCreateClass ## _ ## classname : public CClassRTTI\
{\
   public:\
      CCreateClass ## _ ## classname() : CClassRTTI()\
      {\
         \
      };\
      virtual const SCode GetCode(){SCode CODE \
{\
   id, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }\
};return CODE;};\
      virtual CUtiOB* createClass(CUtiOB* Parent){return new classname(Parent);};\
      virtual const CUtiString getName()\
{\
   return UTEXT(#classname);\
};\
};

#define IMPLEMENT_CLASS_CREATORTYPE(classname) class CCreateClass ## _ ## classname : public CClassRTTI\
{\
   public:\
      CCreateClass ## _ ## classname() : CClassRTTI()\
      {\
         CRTTIManager::getInstance()->addRTTI(this, NULL);\
      };\
      virtual const SCode GetCode(){ return classname::TYPECODE;};\
      virtual CUtiOB* createClass(CUtiOB* Parent){return new classname(Parent);};\
      virtual const CUtiString getName()\
{\
   return UTEXT(#classname);\
};\
};\
CCreateClass ## _ ## classname csclass ## _ ## classname;


#define IMPLEMENT_CLASS_CREATORTYPEPARENT(classname, par) class CCreateClass ## _ ## classname : public CClassRTTI\
{\
   public:\
      CCreateClass ## _ ## classname() : CClassRTTI()\
      {\
         CRTTIManager::getInstance()->addRTTI(this, NULL);\
      };\
      virtual const SCode GetCode(){ return classname::TYPECODE;};\
      virtual CUtiOB* createClass(CUtiOB* Parent){classname* ret = new classname(); ret->set ## par(Parent); return ret;};\
      virtual const CUtiString getName()\
{\
   return UTEXT(#classname);\
};\
};\
CCreateClass ## _ ## classname csclass ## _ ## classname;

#define IMPLEMENT_CLASS_CREATORTYPEENTITYMODUL(classname) class CCreateClass ## _ ## classname : public CClassRTTI\
{\
   public:\
      CCreateClass ## _ ## classname() : CClassRTTI()\
      {\
         CRTTIManager::getInstance()->addRTTI(this, NULL);\
      };\
      virtual const SCode GetCode(){ return classname::TYPECODE;};\
      virtual CUtiOB* createClass(CUtiOB* Parent){classname* ret = new classname(check_cast_or_zero<EntityClass*>(Parent)); return ret;};\
      virtual const CUtiString getName()\
{\
   return UTEXT(#classname);\
};\
};\
CCreateClass ## _ ## classname csclass ## _ ## classname;

#define IMPLEMENT_CLASS_CREATORTYPEENTITYMODUL_HUD(classname) class CCreateClass ## _ ## classname : public CClassRTTI\
{\
   public:\
      CCreateClass ## _ ## classname() : CClassRTTI()\
      {\
         CRTTIManager::getInstance()->addRTTI(this, NULL);\
      };\
      virtual const SCode GetCode(){ return classname::TYPECODE;};\
      virtual CUtiOB* createClass(CUtiOB* Parent){classname* ret = new classname(check_cast_or_zero<EntityClass*>(Parent), dynamic_cast<BaseHUD*>(Parent)); return ret;};\
      virtual const CUtiString getName()\
{\
   return UTEXT(#classname);\
};\
};\
CCreateClass ## _ ## classname csclass ## _ ## classname;

#define IMPLEMENT_CLASS_CREATORTYPEPARENTDAG(classname, par) class CCreateClass ## _ ## classname : public CClassRTTI\
{\
   public:\
      CCreateClass ## _ ## classname() : CClassRTTI()\
      {\
         CRTTIManager::getInstance()->addRTTI(this, NULL);classname::getStaticClassDescription();\
      };\
      virtual const SCode GetCode(){ return classname::TYPECODE;};\
      virtual CUtiOB* createClass(CUtiOB* Parent){classname* ret = dag_create<classname>(); ret->set ## par(Parent); return ret;};\
      virtual const CUtiString getName()\
{\
   return UTEXT(#classname);\
};\
};\
CCreateClass ## _ ## classname csclass ## _ ## classname;
#define IMPLEMENT_CLASS_CREATORTYPENOPARENT(classname) class CCreateClass ## _ ## classname : public CClassRTTI\
{\
   public:\
      CCreateClass ## _ ## classname() : CClassRTTI()\
      {\
         CRTTIManager::getInstance()->addRTTI(this, NULL);\
      };\
      virtual const SCode GetCode(){ return classname::TYPECODE;};\
      virtual CUtiOB* createClass(CUtiOB* Parent){CUtiOB* obj = new classname(); obj->setCreateParent(Parent); return obj;};\
      virtual const CUtiString getName()\
{\
   return UTEXT(#classname);\
};\
};\
CCreateClass ## _ ## classname csclass ## _ ## classname;

#define IMPLEMENT_CLASS_CREATORTYPENOPARENTDAG(classname) class CCreateClass ## _ ## classname : public CClassRTTI\
{\
   public:\
      CCreateClass ## _ ## classname() : CClassRTTI()\
      {\
         CRTTIManager::getInstance()->addRTTI(this, NULL);classname::getStaticClassDescription();\
      };\
      virtual const SCode GetCode(){ return classname::TYPECODE;};\
      virtual CUtiOB* createClass(CUtiOB* Parent){CUtiOB* obj = dag_create<classname>(); obj->setCreateParent(Parent); return obj;};\
      virtual const CUtiString getName()\
{\
   return UTEXT(#classname);\
};\
};\
CCreateClass ## _ ## classname csclass ## _ ## classname;

class CRTTIManager : public CUtiOB
{
protected:
   static CRTTIManager* m_manager;
   
  
public:
  static CRTTIManager* getInstance();
  CCacheList m_classes;
  std::map<SCode, CClassRTTI*> m_clmap;
  std::map<CUtiString, SCode> m_name_to_code;
  CRTTIManager();
  ~CRTTIManager();
  void addRTTI(CClassRTTI* cl, LPVOID Data);
  CUtiOB* createClass(SCode code, CUtiOB* parent);
  SCode classNameToCode(const CUtiString& name);
  CUtiString getName(SCode code);

};

class CCallClass : public CClassRTTI
{
public:
  CUtiOB* (UTICDECL *m_create )(CUtiOB* Parent);
  SCode m_code;
  CCallClass(CUtiOB* (UTICDECL *create)(CUtiOB* Parent), SCode cd) : CClassRTTI()
  {
    m_code = cd;
    m_create = create;
  };
  virtual const SCode GetCode(){return m_code;};
  virtual CUtiOB* createClass(CUtiOB* Parent){return m_create(Parent);};
  virtual const CUtiString getName();
};

class CBaseListReferenceAdapter;


class CTypeInterface : public CUtiOB
{
public:
  virtual void RInt(int &i, int std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RByte(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void REnum(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RFloat(float &i, float std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RDouble(double &i, double std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RVector(CVector &i, const CVector &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){}
  virtual void RVector4(CVector4 &i, const CVector4 &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RCode(SCode &i, SCode std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RReference(CUtiOB* &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RObject(CUtiOB &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  //	virtual void RObject(CUtiOB* &i, int std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RBool(bool &i, bool std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RString(CUtiString &i, const CUtiString&  std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RObjectList(CCacheList &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RObjectListB(CBaseListReferenceAdapter *i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};
  virtual void RObjectOwns(CUtiOB* &i,  int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a){};

};

class CXWriteInterface : public CTypeInterface
{
public:
  CXNode* m_n;
  CWriteReadInterface* m_wri;
  static void Perform(CUtiOB* obj, CXNode* n, CWriteReadInterface* wri);
  CXWriteInterface(CXNode* n, CWriteReadInterface* wri);
  void RInt(int &i, int std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RByte(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void REnum(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RFloat(float &i, float std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RDouble(double &i, double std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RVector(CVector &i, const CVector &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RVector4(CVector4 &i, const CVector4 &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RCode(SCode &i, SCode std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RReference(CUtiOB* &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RObject(CUtiOB &i,  int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RBool(bool &i, bool std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RString(CUtiString &i, const CUtiString&  std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RObjectList(CCacheList &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RObjectOwns(CUtiOB* &i,  int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  virtual void RObjectListB(CBaseListReferenceAdapter *i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
};



class CXReadInterface : public CTypeInterface
{
public:
  CXNode* m_n;
  CWriteReadInterface* m_wri;
  static void Perform(CUtiOB* obj, CXNode* n, CWriteReadInterface* wri);
  CXReadInterface(CXNode* n, CWriteReadInterface* wri);
  void RInt(int &i, int std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RByte(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void REnum(unsigned char &i, unsigned char std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RFloat(float &i, float std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RDouble(double &i, double std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RVector(CVector &i, const CVector &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RVector4(CVector4 &i, const CVector4 &std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RCode(SCode &i, SCode std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RReference(CUtiOB* &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RObject(CUtiOB &i,int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RBool(bool &i, bool std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RString(CUtiString &i, const CUtiString&  std, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RObjectList(CCacheList &i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  void RObjectOwns(CUtiOB* &i,  int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
  virtual void RObjectListB(CBaseListReferenceAdapter *i, int flags, const CUtiString&  Name, const CUtiString&  Beschreibung, CUtiOB* a);
};

class CWriteReadInterface : public ReferenceUtiOB
{
public:
  CWriteReadInterface(): ReferenceUtiOB(){};
 
};



class CClassCollector : public ReferenceUtiOB
{
protected:
  std::map<CUtiOB*, SCode> m_objtocode;
  std::map<SCode, CUtiOB*> m_codetoobj;
  int m_tag;
public:
  CClassCollector();
  ~CClassCollector();
  void addObject(CUtiOB* obj);
  bool hasObject(CUtiOB* obj);
  bool hasObject(const SCode& code);
  CUtiOB* getObject(const SCode& code);
  void setTag(int tag);
  int getTag();
};

void XWriteClass(CXNode* n, LPVOID wri, CUtiOB* cl);
CUtiOB* XReadClass(CXNode* n, LPVOID wri, CUtiOB* parent);
void XReadClassNoCreate(CXNode* n, LPVOID wri, CUtiOB* parent);
CUtiOB* XCopyClass(CUtiOB* obj, CUtiOB* parent, CWriteInfo* writeinfo, CReadInfo* readinfo);
int XSaveToFile(const CUtiString&  string, CUtiOB* cl);
int XLoadFromFile(const CUtiString&  string, CUtiOB* parent, CUtiOB* &result);
int XSaveToStream(CStream* stream, CUtiOB* cl);
int XLoadFromStream(CStream* stream, CUtiOB* parent, CUtiOB* &result);

int XSaveToNode(CXNode* stream, CUtiOB* cl);
int XLoadFromNode(CXNode* stream, CUtiOB* parent, CUtiOB* &result);



/*
class CObjInfo : public CUtiOB
{
public:
  CUtiOB* obj;
  SCode code;
};
*/

class CBaseReferenceAdapter : public ReferenceUtiOB
{
  public:
    
    virtual ~CBaseReferenceAdapter() {};
    CBaseReferenceAdapter() {};
    virtual ReferenceUtiOB* get() const=0;
    virtual void set(ReferenceUtiOB* ob)=0;
};

template<class T> class CReferenceAdapter : public CBaseReferenceAdapter
{
  protected:
    uti_refptr<T>& m_ref;
  public:
    CReferenceAdapter(uti_refptr<T>& ref) : m_ref(ref)
    {
      
    };
    
     virtual ReferenceUtiOB* get() const
     {
         return m_ref.get(); 
     }
    virtual void set(ReferenceUtiOB* ob)
    {
        m_ref = (T*)ob;
    };
};

class CUndoVectorListener;

class CBaseListReferenceAdapter : public ReferenceUtiOB
{
  public:
    
    virtual ~CBaseListReferenceAdapter() {};
    CBaseListReferenceAdapter() {};
    virtual ReferenceUtiOB* get(int i) const=0;
    virtual void set(int i, ReferenceUtiOB* ob)=0;
    virtual void deleteItem(int index)=0;
    virtual void deleteObject(ReferenceUtiOB* ob)=0;
    virtual int size()=0;
    virtual void push_back(ReferenceUtiOB* ob)=0;
    virtual void insert(int index, ReferenceUtiOB* ob)=0;
    virtual void move(int i, int j)=0;
    virtual int indexOf(ReferenceUtiOB* ob)=0;
    virtual void clear()=0;
    virtual void addChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata){};
    virtual void removeChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata){};
};

template<class T> class CListReferenceAdapter : public CBaseListReferenceAdapter
{
  protected:
    CReferenceList<T>& m_ref;
  public:
    CListReferenceAdapter(CReferenceList<T>& ref) : m_ref(ref)
    {
      
    };
    
     virtual ReferenceUtiOB* get(int i) const
     {
         return &m_ref[i]; 
     }
    virtual void set(int i, ReferenceUtiOB* ob)
    {
         m_ref.exchange(i,(T*)ob);
    };
    virtual int size()
    {
        return m_ref.size(); 
    }
    virtual void deleteItem(int index) 
    {
        m_ref.deleteItem(index); 
    }
    virtual void deleteObject(ReferenceUtiOB* ob) 
    {
        m_ref.remove((T*)ob); 
    }
    virtual void push_back(ReferenceUtiOB* ob)
    {
        return m_ref.push_back((T*)ob); 
    }
    virtual void insert(int index, ReferenceUtiOB* ob)
    {
      m_ref.insert(index, (T*)ob);
    }
    virtual void move(int i, int j)
    {
        m_ref.move(i,j);
    }
    virtual int indexOf(ReferenceUtiOB* ob)
    {
     return m_ref.indexOf((T*)ob); 
    }
    virtual void clear()
    {
        return m_ref.clear(); 
    }
};

class CLinkInfo : public CUtiOB
{
public:
  CLinkInfo(){};
  SCode m_code;
  virtual void correct(CUtiOB*obj);
};

class CStdLinkInfo : public CLinkInfo
{
public:
  CStdLinkInfo(){};
  CUtiOB** m_obj;
  
  virtual void correct(CUtiOB* obj);
};

class CReferenceLinkInfo : public CLinkInfo
{
  public:
  uti_refptr<CBaseReferenceAdapter> m_adapter;
  CReferenceLinkInfo(){};
  ~CReferenceLinkInfo();
  void correct(CUtiOB* obj);
};

class CListReferenceLinkInfo : public CLinkInfo
{
    public:
  uti_refptr<CBaseListReferenceAdapter> m_adapter;
  int m_index;
  CListReferenceLinkInfo(){};
  ~CListReferenceLinkInfo();
  void correct(CUtiOB* obj);
};

class CReadInfo : public CWriteReadInterface
{
public:
  CCacheList m_links;
  //CCacheList ObjList;
  std::map<SCode, CUtiOB*> m_objmap;
  
  CReadInfo();
  ~CReadInfo();
  virtual void addLink(CUtiOB** obj, SCode code);
  virtual void addLink(CBaseReferenceAdapter* obj, SCode code);
  virtual void addLink(CBaseListReferenceAdapter* obj, int index, SCode code);
  virtual void addObject(CUtiOB* obj, SCode code);
  virtual void link();
  virtual CUtiOB* findObject(SCode code);
};

class CInstantReadInfo : public CReadInfo
{
public: 
    CInstantReadInfo();
     virtual void addLink(CUtiOB** obj, SCode code);
  virtual void addLink(CBaseReferenceAdapter* obj, SCode code);
  virtual void addLink(CBaseListReferenceAdapter* obj, int index, SCode code);
};

class CWriteInfo : public CWriteReadInterface
{
  public:
   std::map<CUtiOB*, SCode> m_objmap;
   CWriteInfo();
   ~CWriteInfo();
   virtual bool needToBeSaved(CUtiOB* obj);
   virtual void addObject(CUtiOB* obj, const  SCode &code);
   virtual SCode getCode(CUtiOB* obj);
};



class CDuplicateWriteInfo : public CWriteInfo
{
protected:
  uti_refptr<CClassCollector> m_tosaveobjs;
  uti_refptr<CClassCollector> m_nottosave;
public:
  CDuplicateWriteInfo(CClassCollector* tosaveobjs, CClassCollector* nottosave);
  bool needToBeSaved(CUtiOB* obj);
  void addObject(CUtiOB* obj, const  SCode &code);
  SCode getCode(CUtiOB* obj);
};

class CDuplicateReadInfo : public CReadInfo
{
protected:
  uti_refptr<CClassCollector> m_tosaveobjs;
  uti_refptr<CClassCollector> m_nottosave;
public:
  CDuplicateReadInfo(CClassCollector* tosaveobjs, CClassCollector* nottosave);
   CUtiOB* findObject(SCode code);
};

class CXNode : public ReferenceUtiOB
{
protected:
  CXDocument* m_doc;
  CXNode* m_parent;
public:
  CXNode(CXNode* parent, CXDocument* doc);
  CXNode* getParent(){return m_parent;}
  virtual void setName(const CUtiString& name)=0;
  virtual CUtiString getName() const=0;
  virtual void setPropBool(const CUtiString& name, bool b)=0;
  virtual bool getPropBool(const CUtiString& name, bool std) const=0;
  virtual void setPropInt(const CUtiString& name, int b)=0;
  virtual int getPropInt(const CUtiString& name, int std) const=0;
  virtual void setPropFloat(const CUtiString& name, float b)=0;
  virtual float getPropFloat(const CUtiString& name, float std) const=0;
  virtual void setPropDouble(const CUtiString& name, double b)=0;
  virtual double getPropDouble(const CUtiString& name, double std) const=0;
  virtual void setPropCode(const CUtiString& name, const SCode &b)=0;
  virtual SCode getPropCode(const CUtiString& name, const SCode &std) const=0;
  virtual void setPropVector(const CUtiString& name, const CVector &b)=0;
  virtual CVector getPropVector(const CUtiString& name, const CVector &std) const=0;
  virtual void setPropVector4(const CUtiString& name, const CVector4 &b)=0;
  virtual CVector4 getPropVector4(const CUtiString& name, const CVector4 &std) const=0;
  virtual void setPropString(const CUtiString& name, const CUtiString& b)=0;
  virtual CUtiString getPropString(const CUtiString& name, const CUtiString& std) const=0;
  virtual int getPropDataSize(const CUtiString& name) const=0;
  virtual void setPropData(const CUtiString& name, const unsigned char* Data, size_t size)=0;
  virtual bool getPropData(const CUtiString& name, unsigned char* &Data, size_t &size) const=0;
  virtual void setPropData(const CUtiString& name, const Array<unsigned char>& data);
  virtual bool getPropData(const CUtiString& name, Array<unsigned char>&data) const;
  virtual bool propExists(const CUtiString& name) const=0;
  //virtual HResult Read(CStream & Stream, int Flags, int version);
  //virtual HResult Write(CStream & Stream, int Flags, int version);
  virtual CXNode* getNodeOrCreate(const CUtiString& name);
  virtual CXNode* addNode(const CUtiString& name)=0;
  virtual CXNode* getNodeByName(const CUtiString& name)=0;
  virtual CXNode* getNodeByPath(const CUtiString& str);
  virtual int getNodeCount(const CUtiString& name);
  virtual void deleteNode(const CUtiString& name)=0;
  virtual void deleteItem(int i)=0;
  virtual void remove(CXNode* n)=0;
  virtual void removeProperty(const CUtiString& name)=0;
  virtual int getChildCount() const=0;
  virtual CXNode* getChild(int i) const=0;
  virtual void clearNodes()=0;
  virtual void clearProperties()=0;
  virtual void copyPropertiesTo(CXNode* node)=0;
  virtual void copyTo(CXNode* node);
};

class CBaseProp : public CUtiOB
{
public:
  int m_id;
};

class CStringProp : public CBaseProp
{
public:
  CUtiString m_b;
};

class CBoolProp : public CBaseProp
{
public:
  bool m_b;
};

class CIntProp : public CBaseProp
{
public:
  int m_b;
};

class CFloatProp : public CBaseProp
{
public:
  float m_b;
};

class CDoubleProp : public CBaseProp
{
public:
  double m_b;
};

class CCodeProp : public CBaseProp
{
public:
  SCode m_b;
};

class CVectorProp : public CBaseProp
{
public:
  CVector m_b;
};

class CVector4Prop : public CBaseProp
{
public:
  CVector4 m_b;
};

class CDataProp : public CBaseProp
{
public:
  int m_size;
  unsigned char* m_data;
  CDataProp();
  ~CDataProp();
};

class CUtiTypeException : public CUtiException
  {}
;


class CUtiReadErrorException : public CUtiException
{
  public:
    CUtiReadErrorException(const CUtiString& name)
    {
       m_n = name; 
    }
};
const int X_STRING = 1;
const int X_DATA = 2;
const int X_BOOL = 3;
const int X_CODE = 4;
const int X_VECTOR4 = 5;
const int X_INT = 6;
const int X_DOUBLE = 7;
const int X_SINGLE = 8;
const int X_VECTOR = 9;

class CXUtiNode : public CXNode
{
public:
  CReferenceList<CXUtiNode> m_nodes;
  CCacheList m_properties;
  std::map <int, CBaseProp*> m_prophash;
  std::map <int, CXUtiNode*> m_nodehash;
  int m_name;
  CBaseProp* findPropByIndex(int x) const;
  CXUtiNode(CXNode* parent=NULL, CXDocument* doc=NULL);
  virtual void setName(const CUtiString& name);
  virtual CUtiString getName() const;
  virtual void setPropBool(const CUtiString& name, bool b);
  virtual bool getPropBool(const CUtiString& name, bool std) const;
  virtual void setPropInt(const CUtiString& name, int b);
  virtual int getPropInt(const CUtiString& name, int std) const ;
  virtual void setPropFloat(const CUtiString& name, float b);
  virtual float getPropFloat(const CUtiString& name, float std) const;
  virtual void setPropDouble(const CUtiString& name, double b);
  virtual double getPropDouble(const CUtiString& name, double std) const;
  virtual void setPropCode(const CUtiString& name, const SCode &b);
  virtual SCode getPropCode(const CUtiString& name, const SCode &std) const;
  virtual void setPropVector(const CUtiString& name, const CVector &b);
  virtual CVector getPropVector(const CUtiString& name, const CVector &std) const;
  virtual void setPropVector4(const CUtiString& name, const CVector4 &b);
  virtual CVector4 getPropVector4(const CUtiString& name, const CVector4 &std) const;
  virtual void setPropString(const CUtiString& name, const CUtiString& b);
  virtual CUtiString getPropString(const CUtiString& name, const CUtiString& std) const;
  virtual int getPropDataSize(const CUtiString& name) const;
  virtual void setPropData(const CUtiString& name, const unsigned char* Data, size_t size);
  virtual bool getPropData(const CUtiString& name, unsigned char* &Data, size_t &size) const;
  virtual bool propExists(const CUtiString& name) const;
  virtual HResult Read(CStream & Stream, int Flags, int version);
  virtual HResult Write(CStream & Stream, int Flags, int version);
  virtual CXNode* addNode(const CUtiString& name);
  virtual CXNode* getNodeByName(const CUtiString& name);
  virtual void deleteNode(const CUtiString& name);
  virtual void deleteItem(int i);
  virtual void remove(CXNode* n);
  virtual void removeProperty(const CUtiString& name);
  virtual int getChildCount() const;
  virtual CXNode* getChild(int i) const;
  virtual void clearNodes();
  virtual void clearProperties();
  virtual void xread(CXNode* n, CWriteReadInterface* wri);
  virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
  virtual void copyPropertiesTo(CXNode* node);
};

class CXDocument : public ReferenceADR
{
protected:
    uti_refptr<CXNode> m_root;
public:
  virtual CXNode* getRoot();
  CXDocument();
  ~CXDocument();
  virtual int saveToStream(CStream &Stream){return 0;};
  virtual int loadFromStream(CStream &Stream){return 0;};
  virtual int saveToFile(const CUtiString& name){return 0;};
  virtual int loadFromFile(const CUtiString& name){return 0;};
};

CXNode* findSubElement(CXNode* node, const CUtiString& name);
CXNode* findSubElementByProperty(CXNode* node, const CUtiString& nodename, const CUtiString& propertyname, const CUtiString& propertyvalue);
int getElementCount(CXNode* node,const CUtiString& name );
void getElementsByName(CXNode* node, const CUtiString &name, CNormalList<CXNode>& res);

class CXName : public CUtiOB
{
public:
  int m_ref;
  CUtiString m_name;
};

/* Tolle Idee: Speichere Strings und Datenfelder wenn sie größer als ein Schwellenwert sind am Ende der Datei. 2 Modi:
   Komplettes Laden: zuerst normalen Baum und dann der Rest, damit der Lesekopf nicht springen muss.
   Bei Bedarf: Lade die einzelnen Dateneinträge nur, falls sie wirklich angefordert werden. => spezielle Lesefunktion für Daten!*/

class CXUtiDocument : public CXDocument
{
public:
  CCacheList m_strlist;
  std::map <CUtiString, int> m_listindex; //Globaler Name Index?
  CXUtiDocument();
  ~CXUtiDocument();
  int addString(const CUtiString&s);
  void removeString(const CUtiString&id);
  void removeIndex(int index);
  void incIndex(int index);
  int findString(const CUtiString& s) const;
  const CUtiString getString(int id);
  virtual int saveToStream(CStream &Stream);
  virtual int loadFromStream(CStream &Stream);
  virtual int saveToFile(const CUtiString& name);
  virtual int loadFromFile(const CUtiString& name);

  virtual int saveToString(CUtiString& res);
  virtual int loadFromString(const CUtiString& res);


    virtual void xread(CXNode* n, CWriteReadInterface* wri);
    virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
};

class InputOutputHook : public ReferenceUtiOB
{
public:
    InputOutputHook();
    virtual CUtiString readLine()=0;
    virtual void printInt(int i);
    virtual void printDouble(double d);
    virtual void printBool(bool d);
    virtual void printChar(wchar_t ch);
    virtual void printString(const CUtiString &str)=0;
};

class CCommandLineSystemInterface : public ReferenceADR
{
public:
    CCommandLineSystemInterface(){};
    virtual CUtiString executeCommand(const CUtiString& str){return EMPTYSTRING;};
    virtual void setInputOutputHook(InputOutputHook* h){};
    virtual bool hasCommands()=0;
    virtual CUtiString performCommands()=0;
};

struct CBoxf;
struct CRectf;

CUtiString convertToString(float i);
CUtiString convertToString(double i);
CUtiString convertToString(int i);
CUtiString convertToString(bool i);
CUtiString convertToString(const CBox& b);
CUtiString convertToString(const CMatrix& b);
CUtiString convertToString(const CRect& b);
CUtiString convertToString(const CVector4& v);
CUtiString convertToString(const CVector3& v);
CUtiString convertToString(const CVector2& v);



CUtiString convertToString(const CBoxf& b);
CUtiString convertToString(const CMatrixf& b);
CUtiString convertToString(const CRectf& b);
CUtiString convertToString(const CVector4f& v);
CUtiString convertToString(const CVector3f& v);
CUtiString convertToString(const CVector2f& v);

CUtiString convertToString(const CUtiString& str);
CUtiString convertToString(const SCode& v);

CVector2 parseVector2(const CUtiString& str);
CVector3 parseVector3(const CUtiString& str);
CVector4 parseVector4(const CUtiString& str);
SCode parseSCode(const CUtiString& str);

CRect parseRect(const CUtiString& str);
CBox parseBox(const CUtiString& str);

CUtiString convertToCpp(const CUtiString&v);
CUtiString convertToCpp(bool b);
CUtiString convertToCpp(int i);
CUtiString convertToCpp(float b);

class CBaseCoroutine : public ReferenceUtiOB
{
    //$classbody::begin CBaseCoroutine(ancestor="ReferenceUtiOB" stdconde serialize="8643258")
        //$property coroutine_state(type="int" mode="rw" init="-1")
        //$property coroutine_subroutine(type="CBaseCoroutine" mode="rw_ref" init="NULL")
    //$classbody::middle
public: CBaseCoroutine();
public: ~CBaseCoroutine();
public: CUtiString getClassName();
protected: int m_coroutine_state;
public: int getcoroutine_state() const;
public: void setcoroutine_state(const int& value0);
protected: uti_refptr<CBaseCoroutine> m_coroutine_subroutine;
public: CBaseCoroutine* getcoroutine_subroutine();
public: const CBaseCoroutine* getcoroutine_subroutine() const;
public: void setcoroutine_subroutine(CBaseCoroutine* value0);
public: virtual const SCode GetTypeCode();
public: static const SCode TYPECODE;
public: virtual void xread(CXNode* n, CWriteReadInterface* wri);
public: virtual void xwrite(CXNode* n, CWriteReadInterface* wri);
    //$classbody::end
public:
    bool finished() {return m_coroutine_state == -1;};
    virtual bool execute(ReferenceUtiOB *context);
};

#define coroutine_begin(name) \
     bool name::execute(ReferenceUtiOB* context) {\
     switch (m_coroutine_state) {\
     case -1: {assert(false);return true;};\
     case 0:

#define coroutine_interruptl {m_coroutine_state = __LINE__;return false; case __LINE__:;};

#define coroutine_interruptc(code) {m_coroutine_state = (code);return false; case (code):;};

#define coroutine_end };m_coroutine_state=-1;return true;}
#define coroutine_return(a) {m_coroutine_return = a;m_coroutine_state=-1;return true;};

#define coroutine_call2l(function, param1, param2) {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1, param2);case __LINE__:;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = __LINE__; return false;};m_coroutine_subroutine= NULL;};

#define coroutine_call2l_ret(ret, function, param1, param2)  {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1, param2);case __LINE__:;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = __LINE__; return false;};ret = check_cast<function*>(m_coroutine_subroutine.get())->getReturnValue();  m_coroutine_subroutine= NULL;};

#define coroutine_call2c(code, function, param1, param2) {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1, param2);case (code):;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = (code); return false;};m_coroutine_subroutine= NULL;};

#define coroutine_call2c_ret(code, ret, function, param1, param2)  {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1, param2);case (code):;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = (code); return false;};ret = check_cast<function*>(m_coroutine_subroutine.get())->getReturnValue();  m_coroutine_subroutine= NULL;}

#define coroutine_call1l(function, param1) {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1);case __LINE__:;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = __LINE__; return false;};m_coroutine_subroutine= NULL;};

#define coroutine_call1l_ret(ret, function, param1)  {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1);case __LINE__:;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = __LINE__; return false;};ret = check_cast<function*>(m_coroutine_subroutine.get())->getReturnValue();  m_coroutine_subroutine= NULL;};

#define coroutine_call1c(code, function, param1) {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1);case (code):;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = (code); return false;};m_coroutine_subroutine= NULL;};

#define coroutine_call1c_ret(code, ret, function, param1)  {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1);case (code):;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = (code); return false;};ret = check_cast<function*>(m_coroutine_subroutine.get())->getReturnValue();  m_coroutine_subroutine= NULL;}


#define coroutine_call0l(function) {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call();case __LINE__:;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = __LINE__; return false;};m_coroutine_subroutine= NULL;};

#define coroutine_call0l_ret(ret, function)  {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call();case __LINE__:;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = __LINE__; return false;};ret = check_cast<function*>(m_coroutine_subroutine.get())->getReturnValue();  m_coroutine_subroutine= NULL;};

#define coroutine_call0c(code, function) {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call(param1, param2);case (code):;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = (code); return false;};m_coroutine_subroutine= NULL;};

#define coroutine_call0c_ret(code, ret, function)  {assert(m_coroutine_subroutine.get() == NULL);m_coroutine_subroutine = new function();check_cast<function*>(m_coroutine_subroutine.get())->call();case (code):;m_coroutine_subroutine->execute(context);if (!m_coroutine_subroutine->finished()) {m_coroutine_state = (code); return false;};ret = check_cast<function*>(m_coroutine_subroutine.get())->getReturnValue();  m_coroutine_subroutine= NULL;}

#endif
