/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __UTITYPEHANDLERS__
#define __UTITYPEHANDLERS__

#include "UtiType.h"

template<class T> inline T uti_dagclamp(const T&value, const T&min, const T&max)
{
  if (value < min)return min;
  if (value > max) return max;
  return value;
};

inline CVector2 uti_dagclamp(const CVector2& v, utifloat p_min, utifloat p_max)
{
  return CVector2(umin(p_max, umax(p_min, v.x)),umin(p_max, umax(p_min, v.y)));
};

inline CVector4 uti_dagclamp(const CVector4& v, utifloat p_min, utifloat p_max)
{
  return CVector4(umin(p_max, umax(p_min, v.x)),umin(p_max, umax(p_min, v.y)), umin(p_max, umax(p_min, v.z)), umin(p_max, umax(p_min, v.w)));
};


void uti_readValue_ReferenceToNode(CBaseReferenceAdapter *obj, CXNode* newnode, CWriteReadInterface* wri);
void uti_writeValue_ReferenceToNode(CUtiOB* refobj, CXNode* newnode, CWriteReadInterface* wri);

void uti_readValue(CBaseReferenceAdapter *obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, ReferenceUtiOB* init=NULL);
void uti_writeValue(CBaseReferenceAdapter *obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name);


void uti_readValue(CBaseListReferenceAdapter *obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name);
void uti_writeValue(CBaseListReferenceAdapter *obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

bool uti_readValue(CUtiOB* *obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, CUtiOB* init=NULL);
void uti_writeValue(CUtiOB* *obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

//For Instance who won't be changed
void uti_readValue(CUtiOB *obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name);
void uti_writeValue(CUtiOB *obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(CUtiString &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CUtiString& std=EMPTYSTRING);
void uti_writeValue(CUtiString &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(float &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, float std=0.0f);
void uti_writeValue(float &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(double &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, double std=0.0f);
void uti_writeValue(double &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);


void uti_readValue(UTI_UINT64 &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, UTI_UINT64 std=0);
void uti_writeValue(UTI_UINT64 &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(int &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, int std=0);
void uti_writeValue(int &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(SCode &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const SCode& std=NULLCODE);
void uti_writeValue(SCode &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);


void uti_readValue(CVector2 &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CVector2& std=CVector2(0.0f, 0.0f));
void uti_writeValue(CVector2 &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(CVector3 &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CVector3& std=CVector3(0.0f, 0.0f, 0.0f));
void uti_writeValue(CVector3 &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);



void uti_readValue(CVector4 &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CVector4& std=CVector4(0.0f, 0.0f, 0.0f, 1.0f));
void uti_writeValue(CVector4 &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);



void uti_readValue(CMatrix &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CMatrix& std=CMatrix::getIdentity());
void uti_writeValue(CMatrix &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(bool &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, bool std=false);
void uti_writeValue(bool &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(CRect &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CRect& std);
void uti_writeValue(CRect &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

#define DECLARE_READ_WRITE_ENUM(type) \
  void uti_readValue(type &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, type std=(type)0);\
void uti_writeValue(type &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

#define IMPLEMENT_READ_WRITE_ENUM(type) \
  void uti_readValue(type &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, type std)\
  {\
      int x = (int)obj;\
      uti_readValue(x, node, wri, name, (int)std);\
      obj = (type)x;\
  }\
void uti_writeValue(type &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name)\
  {\
      int x = (int)obj;\
      uti_writeValue(x, node, wri, name);\
      obj = (type)x;\
  }

#endif
