/***************************************************************************
 *   Copyright (C) 2008 by Staud   *
 *   staud@linux-yr92   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cdataconsole.h"
#include <assert.h>
#include "utisystem.h"


CDataConsoleStream* s_uti_console_stream = NULL;

CDataConsoleStream& outconsole()
{
 return * s_uti_console_stream;
}


CFixedConsole::CFixedConsole(int w, int h)
{
  m_console.setSize(w,h);
  m_consoleproperty.setSize(w,h);
}

CFixedConsole::~CFixedConsole()
{
  
}

void CFixedConsole::print(int x, int y, const CUtiString& text)
{
  if (y >= 0 && y < m_console.height()) {
    int c = umin(m_console.width()-x, text.size());
    for (int i = 0; i < c; i++) {
       m_console(i+x, y) = text[i];
       m_consoleproperty(i+x, y) = m_property;
    };
  };
}

void CFixedConsole::print(int x, int y, wchar_t ch, int s)
{
  if (y >= 0 && y < m_console.height()) {
    int c = umin(m_console.width()-x, s);
    for (int i = 0; i < c; i++) {
       m_console(i+x, y) = ch;
       m_consoleproperty(i+x, y) = m_property;
    };
  };
}

void CFixedConsole::setCurserPosition(const CIVector2& pos)
{
  m_position = pos;
};

CCommandLine::CCommandLine(int maxlength)
{
   command.setSize(maxlength);
   linelength = 0;
   curserpos = 0;
}

CCommandLine::~CCommandLine()
{
   
}

void CCommandLine::setData(const CUtiString& str)
{
    assert(str.size() < command.size()-1);

    linelength = str.size();
    for (int i = 0; i < linelength; i++) {
        command[i] = str[i];
    }
 //  command = str;
}

wchar_t* CCommandLine::getData()
{
   command[linelength] = 0;
   return command.getCArray();
}

int CCommandLine::getDataLength()
{
   return linelength;
}

void CCommandLine::clearLine()
{
   linelength = 0;
   curserpos = 0;
}

int CCommandLine::getCurserPos()
{
   return curserpos;
}

void CCommandLine::character(wchar_t w, bool insert)
{
   if (insert) {
      if (linelength >= command.size()-1) return;
      for (int i = linelength-1; i >= curserpos; i--) {
         command[i+1] = command[i];
      };
      command[curserpos] = w;
      curserpos++;
      linelength++;
   } else {
      if (curserpos == linelength) {
         if (linelength >= command.size()-1) return;
         command[curserpos] = w;
         curserpos++;
         linelength++;
      } else {
         command[curserpos] = w;
         curserpos++;
      };
   };
}

void CCommandLine::leftButton()
{
   if (curserpos >= 0) curserpos--;
}

void CCommandLine::rightButton()
{
   if (curserpos < linelength) curserpos++;
}

void CCommandLine::deleteButton()
{
   if (curserpos < linelength) {
      deleteCharacter(curserpos);
   };
}

void CCommandLine::backspaceButton()
{
   if (curserpos == 0) return;
   deleteCharacter(curserpos-1);
   curserpos--;
}

void CCommandLine::deleteCharacter(int pos)
{
   utiassert(linelength >= 0);
   if (linelength <= 0) return;
   
   utiassert(pos >= 0 && pos < linelength);
   for (int i = pos; i < linelength-1; i++) {
      command[i] = command[i+1];
   };
   linelength--;
}

void CCommandLine::homeButton()
{
   curserpos = 0;
}
void CCommandLine::endButton()
{
   curserpos = linelength;
}

CDataConsole::CDataConsole(int characters, int linecount): CUtiOB()
{
   characterbuffer.setSize(characters);
   lines.setSize(linecount);
   lastline = -1;
   fullfilled = false;
   w_charpos = 0;
   w_lastposition = 0;
   w_charcount = 0;
}


void CDataConsole::clear()
{
   lastline = -1;
   fullfilled = false;
}

CDataConsole::~CDataConsole()
{
}

void CDataConsole::backward(CLineCallBack* callback, int maxlines, int id, bool developerMode)
{
   if (lastline == -1) return;
   int currentpos = lastline;
   wchar_t* bfr = characterbuffer.getCArray();
   for (int i = 0; i < maxlines; i++) {
      DataLine& l = lines[currentpos];
      if (l.position == pfEngineInfo && developerMode == false) {} else {
      if (l.linestart <= l.lineend) {
         callback->drawLine(bfr + l.linestart, l.lineend - l.linestart, NULL, 0, id,l.position);
      } else {
         callback->drawLine(bfr + l.linestart, characterbuffer.size() - l.linestart, bfr, l.lineend, id,l.position);
      }
      }
      currentpos--;
      if (currentpos < 0) {
         if (fullfilled == false) return;
         currentpos = lines.size()-1;
      };
   }
};

void CDataConsole::forward(CLineCallBack* callback, int maxlines, int id, TIME t, bool developerMode)
{
   if (lastline == -1) return;
   utiassert(maxlines < 128);
   int currentline = 0;
   UTI_DECL_ARRAY(DataLine*, paintlines, 128);
   

   int currentpos = lastline;
   wchar_t* bfr = characterbuffer.getCArray();
   for (int i = 0; i < maxlines; i++) {
      DataLine& l = lines[currentpos];
      if (l.expiretime > t) {
         paintlines[currentline] = &l;
         currentline++;
	 if (l.position == pfEngineInfo) l.expiretime = 0;
      };
      
      currentpos--;
      if (currentpos < 0) {
         if (fullfilled == false) break;
         currentpos = lines.size()-1;
      };
   }
   for (int i = currentline-1; i >= 0; i--) {
      DataLine& l = *paintlines[i];
      if (l.position == pfEngineInfo && developerMode == false) continue;
      if (l.linestart <= l.lineend) {
         callback->drawLine(bfr + l.linestart, l.lineend - l.linestart, NULL, 0, id, l.position);
      } else {
         callback->drawLine(bfr + l.linestart, characterbuffer.size() - l.linestart, bfr, l.lineend, id,l.position);
      }
   };
  /* currentpos = lastline;
   for (int i = 0; i < maxlines; i++) {
      DataLine& l = lines[currentpos];
      if (l.position == pfEngineInfo) {
	l.expiretime = 0;
      };
      currentpos--;
      if (currentpos < 0) {
         currentpos = lines.size()-1;
      };
   }*/
}

int CDataConsole::writeCharacters(int position, const Array<wchar_t> & line, bool addendline)
{
   int c = position;
   utiassert(line.size() < characterbuffer.size()-16);
   utiassert(c >= 0 && c < characterbuffer.size());
   for (int i = 0; i < line.size(); i++) {
      characterbuffer[c] = line[i];
      c++;
      c = c % characterbuffer.size();
   };
   if (addendline) {
      characterbuffer[c] =10;
      c++;
      c = c % characterbuffer.size();
   };
   return c;
}

int CDataConsole::getCharacterPos()
{
   if (lastline == -1) return 0;
   return lines[lastline].lineend;
}

DataLine& CDataConsole::getNewDataLine()
{
   lastline++;
   if (lastline < lines.size()) {
      return lines[lastline];
   } else {
      
      lastline = 0;
      fullfilled = true;
      return lines[lastline];
   };
};

void CDataConsole::writeData(const U_CHAR* data, int length, TIME stayTime, TPositionFlag pos)
{
   Array<wchar_t> refarray;
   Array<wchar_t> source(const_cast<U_CHAR*>(data), length);
   int currentpos = 0;
   int nextpos = 0;
   int end;
   bool hasreturn;
   while (true) {
      end = length;
      nextpos = length;
      hasreturn = false;
      for (int i = currentpos; i < length; i++) {
         if (data[i] == 10) {
            end = i;
            hasreturn = true;
            nextpos = i+1;
         };
      };
      refarray.ReferenzTo(source, currentpos, end-currentpos);
      if (hasreturn) {
         printLine(refarray, UtiFramework::getCurrentTime()+stayTime, pos);
      } else {
         if (w_charcount + refarray.size() < characterbuffer.size()) {
            w_charpos = writeCharacters(w_charpos, refarray, false);
            w_charcount += refarray.size();
         };
      };
      currentpos = nextpos;
      if (nextpos >= length) break;
   };
};

void CDataConsole::printLine(const Array<wchar_t> & line, TIME expiretime, TPositionFlag pos)
{
   
   int p = w_charpos;//getCharacterPos();
   int end = w_charpos;
   if (w_charcount +line.size() < characterbuffer.size()) {
       end = writeCharacters(p, line, true);
   };
   DataLine& l = getNewDataLine();
   l.linestart = w_lastposition;
   l.lineend = end;
   l.expiretime = expiretime;
   l.position = pos;
   w_lastposition = end;
   w_charpos = end;
   w_charcount = 0;
}

void CDataConsole::setLastLine(const Array<wchar_t> & line, TIME expiretime, TPositionFlag pos)
{
   utiassert(lastline != -1);
   DataLine& l = lines[lastline];
   int end = writeCharacters(l.linestart, line, true);
   l.lineend = end;
   l.expiretime = expiretime;
   l.position = pos;
};

CDataConsoleStream::CDataConsoleStream(CDataConsole* _data)
{
   data = _data;
   stayTime = 5.0f;
   currentposition = pfMiddle;
}
CDataConsoleStream::~CDataConsoleStream()
{
   
}

void CDataConsoleStream::setStayTime(TIME t)
{
   stayTime = t;
}
TIME CDataConsoleStream::getStayTime()
{
   return stayTime;
}

void CDataConsoleStream::setPosition(TPositionFlag p)
{
   currentposition = p;
}

TPositionFlag CDataConsoleStream::getPosition()
{
   return currentposition;
}

void CDataConsoleStream::writeData(const U_CHAR* _data, int length)
{
   utiassert(data != NULL);
   data->writeData(_data, length, stayTime, currentposition);
}
