/***************************************************************************
 *   Copyright (C) 2008 by Staud   *
 *   staud@linux-yr92   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef CDATACONSOLE_H
#define CDATACONSOLE_H

#include "Uti3DCore.h"
#include "ctextstream.h"
#include "utiarray.h"
class CDataConsole;
class CCommandLine : public CUtiOB
{
protected:
   Array<wchar_t> command;
   int linelength;
   int curserpos;
   void deleteCharacter(int pos);
public:
   CCommandLine(int maxlength);
   ~CCommandLine();
   void setData(const CUtiString& str);
   wchar_t* getData();
   int getDataLength();
   int getCurserPos();
   void clearLine();
   void character(wchar_t w, bool insert);
   void leftButton();
   void rightButton();
   void deleteButton();
   void homeButton();
   void endButton();
   void backspaceButton();
};

enum TPositionFlag {pfNone, pfTopLeft, pfTopRight, pfMiddle, pfEngineInfo};

struct DataLine
{
   int linestart;
   int lineend;
   TIME expiretime;
   TPositionFlag position;
};

class CDataConsoleStream : public CTextStream
{
   protected:
      CDataConsole* data;
      TIME stayTime;
      TPositionFlag currentposition;
   public:
      CDataConsoleStream(CDataConsole* _data);
      ~CDataConsoleStream();
      void setStayTime(TIME t);
      TIME getStayTime();
      void setPosition(TPositionFlag p);
      TPositionFlag getPosition();
      virtual void writeData(const U_CHAR* _data, int length);
};

extern CDataConsoleStream* s_uti_console_stream;

CDataConsoleStream& outconsole();

class CLineCallBack
{
   public:
      virtual void drawLine(wchar_t* line, int length, wchar_t* line2, int length2, int id, TPositionFlag pos)=0;
};
/**
	@author Staud <staud@linux-yr92>
*/
class CDataConsole : public CUtiOB
{
protected:
   Array<wchar_t> characterbuffer;
   Array<DataLine> lines;
   int w_lastposition;
   int w_charpos;
   int w_charcount;
   int lastline;
   bool fullfilled;
   int getCharacterPos();
   int writeCharacters(int position, const Array<wchar_t> & line, bool addendline);
   DataLine& getNewDataLine();
public:
    CDataConsole(int characters, int linecount);
    ~CDataConsole();
    void backward(CLineCallBack* callback, int maxlines, int id, bool developerMode);
    void forward(CLineCallBack* callback, int maxlines, int id, TIME t, bool developerMode);
    void clear();
    void printLine(const Array<wchar_t> & line, TIME expiretime, TPositionFlag pos);
    void setLastLine(const Array<wchar_t> & line, TIME expiretime, TPositionFlag pos);
    void writeData(const U_CHAR* data, int length, TIME stayTime, TPositionFlag pos);
};

enum TConsoleColor {clWhite, clBlack};

struct SConsoleProperty 
{
  TConsoleColor m_bgcolor;
  TConsoleColor m_fgcolor;
  SConsoleProperty()
  {
    m_bgcolor = clBlack;
    m_fgcolor = clWhite;
  };
};

class CFixedConsole : public ReferenceUtiOB
{
protected:
  Array2D<wchar_t> m_console;
  Array2D<SConsoleProperty> m_consoleproperty;
  SConsoleProperty m_property;
  CIVector2 m_position;
public:
    explicit CFixedConsole(int w, int h);
    virtual ~CFixedConsole();
    virtual void print(int x, int y, const CUtiString& text);
    virtual void print(int x, int y, wchar_t ch, int s);
    virtual void setCurserPosition(const CIVector2& pos);
    
};
#endif
