#ifndef __FASTCOSSINUS__
#define __FASTCOSSINUS__

#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <conio.h>
#include <time.h>

#define MAX_CIRCLE_ANGLE      512
#define HALF_MAX_CIRCLE_ANGLE (MAX_CIRCLE_ANGLE/2)
#define QUARTER_MAX_CIRCLE_ANGLE (MAX_CIRCLE_ANGLE/4)
#define MASK_MAX_CIRCLE_ANGLE (MAX_CIRCLE_ANGLE - 1)
#define PI 3.14159265358979323846f

float fast_cossin_table[MAX_CIRCLE_ANGLE];           // Declare table of fast cosinus and sinus

// Copied from NVidia web site
inline void FloatToInt(int *int_pointer, float f)
{
	__asm  fld  f
	__asm  mov  edx, int_pointer
	__asm  FRNDINT
	__asm  fistp dword ptr[edx];

}

inline float fastcos(float n)
{
	float f = n * HALF_MAX_CIRCLE_ANGLE / PI;
	int i;
	FloatToInt(&i, f);
	if (i < 0)
	{
		return fast_cossin_table[((-i) + QUARTER_MAX_CIRCLE_ANGLE)&MASK_MAX_CIRCLE_ANGLE];
	}
	else
	{
		return fast_cossin_table[(i + QUARTER_MAX_CIRCLE_ANGLE)&MASK_MAX_CIRCLE_ANGLE];
	}

	assert(0);
}

inline float fastsin(float n)
{
	float f = n * HALF_MAX_CIRCLE_ANGLE / PI;
	int i;
	FloatToInt(&i, f);
	if (i < 0)
	{
		return fast_cossin_table[(-((-i)&MASK_MAX_CIRCLE_ANGLE)) + MAX_CIRCLE_ANGLE];
	}
	else
	{
		return fast_cossin_table[i&MASK_MAX_CIRCLE_ANGLE];
	}

	assert(0);
}

void fast_init_sincos()
{
for (int i = 0; i < MAX_CIRCLE_ANGLE; i++)
	{
		fast_cossin_table[i] = (float)sin((double)i * PI / HALF_MAX_CIRCLE_ANGLE);
	}
}

inline float fast_log2 (float val)
{
   int * const    exp_ptr = reinterpret_cast <int * (&val);
   int            x = *exp_ptr;
   const int      log_2 = ((x  23) & 255) - 128;
   x &= ~(255 << 23);
   x += 127 << 23;
   *exp_ptr = x;

   val = ((-1.0f/3) * val + 2) * val - 2.0f/3;   // (1)

   return (val + log_2);
} 

inline float fast_log (const float &val)
{
   return (fast_log2 (val) * 0.69314718f);
} 


UINT64 getCycleCount()
{

   uint32 timehi, timelo;

   // Use the assembly instruction rdtsc, which gets the current
   // cycle count (since the process started) and puts it in edx:eax.
   __asm
   {
      rdtsc
      mov timehi, edx;
      mov timelo, eax;
   }

   return ((UINT64)timehi << 32) + (UINT64)timelo;
} 

#endif