/***************************************************************************
 *   Copyright (C) 2007 by Staud   *
 *   mstaud@linux-fd2d   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef UTIGRAPHGRAPH_H
#define UTIGRAPHGRAPH_H

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "utilists.h"

#include<map>



class Node;
class Edge;
   
/**
	@author Staud <mstaud@linux-fd2d>
*/
class Graph : public CUtiOB
{
public:
    Graph();
    ~Graph();
    virtual Node* getNode(const CVector& pos)=0;
    virtual Node* getWalkToNode(const CVector& pos)=0;
};

class Node : public CUtiOB
{
public:
   Node();
   ~Node();
   virtual bool inNode(const CVector& pos)=0;
   virtual CVector & getGoal(const CVector& mypos)=0;
   virtual int getEdgeCount()=0;
   virtual Edge* getEdge(int i);
         
};

class Edge : public CUtiOB
{
public:   
   Edge();
   ~Edge();
   virtual Node* getFrom()=0;
   virtual Node* getTo()=0;
   virtual bool isOpened()=0;
};

class CostCalcSystem {
public:
   virtual float getCost(Edge* edge)=0;  
};

class GoalDecomposer {
public:
   virtual CVector getSubGoal(const CVector& target, const CVector& myposition)=0;
};

//TODO: Verwende KDStructure aus 3DEngine mit Raycast
//Mit Raycasting + Volumecasting
//Auch dann in 3D Engine verwenden
//+Daten für Volumeninfo
//+Dynamische Structur
//+VolumeInfo
class VolumeStructure : public ReferenceUtiOB { 
public:
   virtual void addObject(CUtiOB* ob, const CBox& bounds)=0;
   virtual void removeObject(CUtiOB* ob)=0;
   virtual void getObjectsAtPosition(const CVector& p, CCacheList& list)=0;
   virtual void getObjectsInBox(const CBox& bounds, CCacheList& list)=0;
};

class DefaultVolumeStructure : public VolumeStructure {
protected:
   std::map<CUtiOB*, CBox> objects;
public:
   DefaultVolumeStructure();
   ~DefaultVolumeStructure();
   virtual void addObject(CUtiOB* ob, const CBox& bounds);
   virtual void removeObject(CUtiOB* ob);
   virtual void getObjectsAtPosition(const CVector& p, CCacheList& list);
   virtual void getObjectsInBox(const CBox& bounds, CCacheList& list);
   
};



#endif
