/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "invalidcontrol.h"
#include <typeinfo>

CInvalidObject::CInvalidObject(const CUtiString& source, ReferenceUtiOB* obj)
{
  m_abort = false;
  m_allowinvalidate = true;
  m_donotdeleteuninterested = false;
  CUtiString str;
 
  if (obj) {
  str.LoadAnsi(typeid(*obj).name());
  } else {
    str = UTISTR("NULL: ");
  };
  m_sourceobj = str+UTISTR(": ")+source;
  m_obj = obj;
  //utiassert(CThread::isMainThread());
  
}

CInvalidObject::~CInvalidObject()
{
  //utiassert(CThread::isMainThread());
  for (int i = 0; i < m_sourceinterest.size(); i++) {utiassert(CThread::isMainThread());
     m_sourceinterest[i].decreaseInterest(this);
  };
  m_sourceinterest.clear();
}

void CInvalidObject::setObj(ReferenceUtiOB* obj)
{
   m_obj = obj; 
}

void CInvalidObject::decreaseInterest(CInvalidObject* obj)
{
  CHECKEXISTANCE;
assert(CThread::isMainThread());
assert(m_targetinterest.find(obj) != m_targetinterest.end());
  m_targetinterest.erase(obj);
  if (m_targetinterest.size()== 0 && m_donotdeleteuninterested == false) doAbort();
}

void CInvalidObject::doAbort()
{
  CHECKEXISTANCE;
  utiassert(CThread::isMainThread());
  utiassert(m_allowinvalidate == true);
  m_abort = true;
  for (int i = 0; i < m_sourceinterest.size(); i++) {
     m_sourceinterest[i].decreaseInterest(this);
    // m_sourceinterest[i].addRef();
  };
  m_sourceinterest.clear();
  std::set<CInvalidObject*>::iterator it;
  int lastcount=0;
  while (m_targetinterest.begin() != m_targetinterest.end()) {
    CInvalidObject* obj = *m_targetinterest.begin();
    lastcount = m_targetinterest.size();
    CHECKEXISTANCE;
    obj->doAbort();
    if (lastcount <= 1) break;
  };
}

void CInvalidObject::clear()
{
  for (int i = 0; i < m_sourceinterest.size(); i++) {
     m_sourceinterest[i].decreaseInterest(this);
  };
  m_sourceinterest.clear();
 /* std::set<CInvalidObject*>::iterator it;
  while (m_targetinterest.begin() != m_targetinterest.end()) {
    CInvalidObject* obj = *m_targetinterest.begin();
    obj->removeMe(this);
  };*/
};

void CInvalidObject::resetAbort()
{
   m_abort = false; 
}

void CInvalidObject::increaseInterest(CInvalidObject* obj)
{
  CHECKEXISTANCE;
assert(CThread::isMainThread());
assert(m_targetinterest.find(obj) == m_targetinterest.end());
    m_targetinterest.insert(obj);
}

void CInvalidObject::addDepend(CInvalidObject* depend)
{
  CHECKEXISTANCE;
  utiassert(CThread::isMainThread());
  if (m_sourceinterest.indexOf(depend) != UTI_INVALIDINDEX) return;
  m_sourceinterest.push_back(depend);
  //TODO: Checke for Zyklus
  depend->increaseInterest(this);
}
int CInvalidObject::getDependCount()
{
return m_sourceinterest.size();
}

int CInvalidObject::getInterestCount()
{
   return m_targetinterest.size(); 
}
