/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __QUADTREE__
#define __QUADTREE__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include <set>
#include <vector>

typedef void (* UTI_ENUMNODES) (void* controller, void* node, void* data);

template <class BaseType, class ContainerObject,  class ObjectAccessor> class CQuadTree;

template <class BaseType, class ContainerObject,  class ObjectAccessor> class CQuadNode : public ReferenceUtiOB
{
  protected:
    //typedef CQuadNode<BaseType, LeafObject, NodeObject> Tree;
    StaticArray<uti_refptr<CQuadNode>, 4> m_children;
    //Tree* m_tree;
    CRect m_rect;
  public:
    ContainerObject m_container;
    //NodeObject m_nodeobject;
    ObjectAccessor m_accessor;
    CQuadNode()
    {
       
    }
    ~CQuadNode()
    {
      
    }
    void subdivide()
    {
      utiassert(m_children[0] == NULL);
      for (int i = 0; i < 4; i++) {
	m_children[i] = new CQuadNode();
      };
      CVector2 center = m_rect.center();
      m_children[0]->m_rect = CRect(m_rect.m_min.x, m_rect.m_min.y, center.x, center.y);
      m_children[1]->m_rect = CRect(center.x, m_rect.m_min.y, m_rect.m_max.x, center.y);
      m_children[2]->m_rect = CRect(m_rect.m_min.x, center.y, center.x, m_rect.m_max.y);
      m_children[3]->m_rect = CRect(center.x, center.y, m_rect.m_max.x, m_rect.m_max.y);
      
      float area = m_rect.getArea();
      typename ContainerObject::iterator it = m_container.begin();
      while (it != m_container.end()) {
	BaseType* t = *it;
	CRect rect = m_accessor.getRect(t);
	if (rect.getArea()/ area < 0.75f) {
	  
	  for (int i = 0; i < 4; i++) {
	     if (m_children[i]->m_rect.intersect(m_rect)) {
	       
	        m_children[i]->insert(t);
	     };
	  };
	  it = m_container.erase(it);
	} else {
	  it++;
	};
      }
     /* for (int i = 0; i < 4; i++) {
	 m_nodeobject->subdivide(m_children[0], m_children[1], m_children[2], m_children[3]);
	 
       };*/
    }
    void merge()
    {
      typename ContainerObject::iterator it;
      std::set<BaseType*> m_elements;
     /* for (int i = 0; i < 4; i++) {
	 m_nodeobject->merge(m_children[0], m_children[1], m_children[2], m_children[3]);
	 
       };*/
      for (int i = 0; i < 4; i++) {
        // if (m_children[i]->m_rect.intersect(m_rect)) {
	     for (it = m_children[i]->m_container.begin(); it != m_children[i]->m_container.end(); it++) {
	       if (m_elements.find(*it) == m_elements.end()) {
		 utiassert(m_container->indexOf(*it) == UTI_INVALIDINDEX);
	          m_container.insert(*it);
		  m_elements.insert(*it);
	       };
	     };
             m_children[i] = NULL;
         //};
      };
    }
    void insert(BaseType* t)
    {
      float area = m_rect.getArea();
      CRect rect = m_accessor.getRect(t);
	if (rect.getArea()/ area < 0.75f) {
	  for (int i = 0; i < 4; i++) {
	     if (m_children[i]->m_rect.intersect(m_rect)) {
	      
	        m_children[i]->insert(t);
	     };   
	  };
	  
	} else {
          utiassert(m_container.find(t) == m_container.end());
	  m_container.insert(t);
	};
    }
    void remove(BaseType* type)
    {
      float area = m_rect.getArea();
      CRect rect = m_accessor.getRect(type);
	if (rect.getArea()/ area < 0.75f) {
	  for (int i = 0; i < 4; i++) {
	     if (m_children[i]->m_rect.intersect(m_rect)) {
	       
                m_children[i]->remove(type);
	     };
	  };
	  
	} else {
          utiassert(m_container->indexOf(type) != UTI_INVALIDINDEX);
          m_container.removeItem(type);
	};
    }
    CRect getRect()
    {
      return m_rect;
    }
    void setRect(const CRect& rect)
    {
      m_rect = rect;
    };
    CQuadNode* getChild(int i)
    {
        utiassert(i >= 0 && i < 4);
	return m_children[i].get();
    }    
    bool hasChildren()
    {
       return m_children[0].get() != NULL;
    };

    template<class List> void enumNodes2D(const CRect& rect, List& list)
    {
        if (hasChildren()) {
            for (int i = 0; i < 4; i++) {
               if (m_children[i]->m_rect.intersect(rect)) {

                  m_children[i]->enumNodes2D(rect, list);
               };
            };
        } else {
            for (auto el : m_container) {
               CRect r2 = m_accessor.getRect(el);
               if (r2.intersect(rect)) {
                   list.push_back(el);
               }
            }

        }
    }
};

#ifndef NDEBUG
#define QUADTREEDEBUG
#endif

template <class BaseType, class ContainerObject,  class ObjectAccessor> class CQuadTree : public ReferenceUtiOB
{
  public:
    typedef CQuadNode<BaseType, ContainerObject, ObjectAccessor> Node;
  protected:
    CRect m_rect;
    uti_refptr<Node> m_root;
    ObjectAccessor m_accessor;
#ifdef QUADTREEDEBUG
    std::set<BaseType*> m_debugdata;
#endif
  public:
    CQuadTree()
    {
      
    }    
    void insert(BaseType* t)
    {
      m_root->insert(t);
#ifdef QUADTREEDEBUG
      assert(m_debugdata.find(t) == m_debugdata.end());
    m_debugdata.insert(t);
#endif
    }
    void remove(BaseType* t)
    {
      m_root->remove(t);
#ifdef QUADTREEDEBUG
              assert(m_debugdata.find(t) != m_debugdata.end());
            m_debugdata.erase(t);
        #endif
    }
    Node* getRoot()
    {
       return m_root.get();
    };
    CRect getRect()
    {
        return m_rect;	
    };
    void setRect(const CRect& rect)
    {
       utiassert(m_root.get() == 0 || m_root->hasChildren() == false);
       m_rect = rect;
       if (m_root.get() != NULL) {
           m_root->setRect(rect);
       };

    };
    template<class List> void enumNodes2D(const CRect& rect, List& list)
    {
        m_root->enumNodes2D(rect, list);

#ifdef QUADTREEDEBUG

        std::set<BaseType*> detectedset;
        std::copy(list.begin(), list.end(), std::inserter(detectedset, detectedset.begin()));
        std::set<BaseType*> foundset;
        for (auto el : m_debugdata)
        {
            CRect r2 = m_accessor.getRect(el);
            if (r2.intersect(rect)) {
                foundset.insert(el);
            }
        }
        assert(foundset == detectedset);
#endif
    }
   /* void enumNodes3D(void* controller, const CBox& box, UTI_ENUMNODES callback,void* data)
    {
       m_root->enumNodes3D(controller, box, callback, data);														   
    }
    void enumNodesPolyeder(void* controller, const std::vector<CVector4> &planes, UTI_ENUMNODES callback,void* data)
    {
	m_root->enumNodesPolyeder(controller, planes, callback, data);													   
    }  */
};



#endif
