/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "undosystem.h"
#include <assert.h>
#include "utisystem.h"


CUndoChangeListener::CUndoChangeListener()
{

}

CUndoVectorListener::CUndoVectorListener()
{

}

CUndoListener::CUndoListener()
{

}

/*
struct CUndoDocumentLocker {
  CUndoDocument* m_doc;
  CUndoDocumentLocker(CUndoDocument* doc);
 ~CUndoDocumentLocker();
};*/

CUndoDocumentLocker::CUndoDocumentLocker(CUndoDocument* doc)
{
   m_doc = doc;
 // utiassert(CUndoControl::getInstance()->getDocument() == NULL);
   if (m_doc)
  CUndoControl::getInstance()->lockDocument(m_doc);
}

CUndoDocumentLocker::~CUndoDocumentLocker()
{
   if (m_doc)
   CUndoControl::getInstance()->unlockDocument(m_doc);
}

static TIME lastusage = -10000000.0f;

CUndoCommandLocker::CUndoCommandLocker()
{
 
   CUndoControl::getInstance()->beginCommand();
}

CUndoCommandLocker::~CUndoCommandLocker()
{
   CUndoControl::getInstance()->endCommand();
   lastusage = -10000000.0f;
}


CUndoCommandValueLocker::CUndoCommandValueLocker()
{

  TIME currenttime = UtiFramework::getCurrentTime();
   m_locked = false;
   if (CUndoControl::getInstance()->getDocument() == NULL) return;
  if (currenttime > lastusage+0.5f) {
   m_locked = true;
   CUndoControl::getInstance()->beginCommand();
  } else {
     m_locked = false;
  
  };
  lastusage = currenttime;
}

CUndoCommandValueLocker::~CUndoCommandValueLocker()
{
  if (m_locked)
   CUndoControl::getInstance()->endCommand();
}


CUndoControl::CUndoControl()
{
  m_hashMap.setSize(UNDO_HASMAPSIZE);
  m_hashMap.fill(0);
  m_doclockcount = 0;
  m_commandlockcount=0;
}

CUndoControl::~CUndoControl()
{

}

void CUndoControl::lockDocument(CUndoDocument* doc)
{
  if (m_doclockcount == 0) {
    utiassert(m_currentdoc.get() == NULL);
    m_currentdoc = doc;
    m_doclockcount++;
  } else {
    utiassert(m_currentdoc.get() == doc);
    m_doclockcount++;
    
  };
}

void CUndoControl::unlockDocument(CUndoDocument*doc)
{
  utiassert(m_doclockcount > 0);
  utiassert(m_currentdoc.get() == doc);
  if (m_doclockcount == 1) {
    m_currentdoc = NULL;
    m_doclockcount--;
  } else {
    m_doclockcount--;
  };
}

void CUndoControl::beginCommand()
{
  utiassert(m_doclockcount != 0);
  if (m_commandlockcount == 0) {
    snapshot();
  };
  m_commandlockcount++;
}

void CUndoControl::endCommand()
{
  utiassert(m_doclockcount != 0);
  utiassert(m_commandlockcount > 0);
  if (m_commandlockcount == 1) {
   
    m_commandlockcount--;
     if (m_currentdoc.get() == NULL) return;
    m_currentdoc->clearIfSnapShotIsEmpty();
  } else {
    m_commandlockcount--;
  };
}


void CUndoControl::snapshot()
{
   if (m_currentdoc.get() == NULL) return;
   m_currentdoc->snapshot();
   std::wcout << "CUndoControl::snapshot()" << std::endl;
}

void CUndoControl::undo()
{
   if (m_currentdoc.get() == NULL) return;
   m_currentdoc->undo();
    std::wcout << "CUndoControl::undo()" << std::endl;
}

void CUndoControl::redo()
{
   if (m_currentdoc.get() == NULL) return;
   m_currentdoc->redo();
    std::wcout << "CUndoControl::redo()" << std::endl;
}


void CUndoControl::setDocument(CUndoDocument* document)
{
   m_currentdoc = document;
}

CUndoDocument* CUndoControl::getDocument()
{
   return m_currentdoc.get();
}


void CUndoControl::rollback()
{
    if (m_currentdoc.get() == NULL) return;
   m_currentdoc->rollback();
}

void CUndoControl::addUndoCommand(CUndoCommand* command)
{
//  utiassert(m_currentdoc.get() != NULL);
  
   if (m_currentdoc.get() == NULL) {delete command;return;}
   m_currentdoc->addCommand(command);
   
}

bool CUndoControl::alreadyInThisSnapShot(void* data)
{
 //  utiassert(m_currentdoc.get() != NULL);
   if (m_currentdoc.get() == NULL) {
   //  std::wcout << "Ohne Document UndoCommand" << std::endl;
   };
    if (m_currentdoc.get() == NULL) return true;
   return m_currentdoc->alreadyInThisSnapShot(data);
}

THREADLOCAL CUndoControl* CUndoControl::s_instance = NULL;

CUndoControl* CUndoControl::getInstance()
{
  utiassert(CThread::isMainThread());
    if (s_instance == NULL) {
       s_instance = new CUndoControl();
    }
    return s_instance;
}

void CUndoControl::addToHashMap(void* th)
{
   unsigned int value = ((UTI_INT64)th) % UNDO_HASMAPSIZE;
    m_hashMap[value]++;
}

void CUndoControl::removeFromHashMap(void* th)
{
   unsigned int value = ((UTI_INT64)th) % UNDO_HASMAPSIZE;
    m_hashMap[value]--;
}

void CUndoControl::vec_addvalue(void* th, int position)
{
  CUndoPositionInfo pinfo;
    pinfo.m_position = position;
#define ENUMLIST(th)\
std::map<void*, CUndoChangeList>::iterator it = m_changelist.find(th);\
         if (it == m_changelist.end()) {\
            return;\
         }\
    CUndoChangeList& change = it->second;\
    \
    for (std::map<CUndoListener*, CChangeListElement >::iterator it = change.m_updateobjects.begin();\
         it != change.m_updateobjects.end(); it++) 
         
     ENUMLIST(th)   
    {
         check_cast<CUndoVectorListener*>(it->first)->undo_addvalue(th, &pinfo, it->second.m_element.get());
    }
}

void CUndoControl::vec_clearvalue(void* th)
{
 ENUMLIST(th)   
    {
         check_cast<CUndoVectorListener*>(it->first)->undo_clearvalue(th, it->second.m_element.get());
    }
}

void CUndoControl::vec_exchangevalue(void* th, int position)
{
 CUndoPositionInfo pinfo;
    pinfo.m_position = position;
     ENUMLIST(th)   
    {
         check_cast<CUndoVectorListener*>(it->first)->undo_exchangevalue(th, &pinfo, it->second.m_element.get());
    }
}

void CUndoControl::vec_movevalue(void* th, int m1, int m2)
{
CUndoPositionInfo pinfo;
    pinfo.m_position = m1;
    pinfo.m_position2 = m2;
     ENUMLIST(th)   
    {
         check_cast<CUndoVectorListener*>(it->first)->undo_movevalue(th, &pinfo, it->second.m_element.get());
    }
}

void CUndoControl::vec_removevalue(void* th, int position)
{
CUndoPositionInfo pinfo;
    pinfo.m_position = position;
     ENUMLIST(th)   
    {
         check_cast<CUndoVectorListener*>(it->first)->undo_removevalue(th, &pinfo, it->second.m_element.get());
    }
}


void CUndoControl::preChange(void* th)
{
     ENUMLIST(th)   
    {
         utiassert(dynamic_cast<CUndoChangeListener*>(it->first) != NULL);
         ((CUndoChangeListener*)it->first)->undo_prechange(th, it->second.m_element.get());
    }
}

void CUndoControl::postChange(void* th)
{
     ENUMLIST(th)   
    {
         utiassert(dynamic_cast<CUndoChangeListener*>(it->first) != NULL);
         ((CUndoChangeListener*)it->first)->undo_postchange(th, it->second.m_element.get());
    }
}

void CUndoControl::deleteWatch(void* th)
{
     ENUMLIST(th)   
    {
         it->first->undo_delete(th, it->second.m_element.get());
    }
      std::map<void*, CUndoChangeList>::iterator it2 = m_changelist.find(th);
      assert(it2!= m_changelist.end());
      assert(it2->second.m_updateobjects.size() == 0);
    m_changelist.erase(th);
    removeFromHashMap(th);
}


void CUndoControl::addChangeListener(void* th, CUndoListener* listener, ReferenceUtiOB* changedata)
{
     std::map<void*, CUndoChangeList>::iterator it = m_changelist.find(th);
         if (it == m_changelist.end()) {
            it = m_changelist.insert(std::make_pair(th, CUndoChangeList())).first;
            addToHashMap(th);
	    //wcout << "Listeners(a): " << m_changelist.size() << endl;
         }
         utiassert(it->second.m_updateobjects.find(listener) == it->second.m_updateobjects.end());
         it->second.m_updateobjects.insert(std::make_pair(listener, CChangeListElement(changedata)));
}

void CUndoControl::removeChangeListener(void* th, CUndoListener* listener, ReferenceUtiOB* changedata)
{
    std::map<void*, CUndoChangeList>::iterator it = m_changelist.find(th);
    utiassert(it != m_changelist.end());
       
    utiassert(it->second.m_updateobjects.find(listener) != it->second.m_updateobjects.end());
    it->second.m_updateobjects.erase(listener);
    if (it->second.m_updateobjects.size() == 0) {
       m_changelist.erase(it);
       removeFromHashMap(th);
     //  wcout << "Listeners(r): " << m_changelist.size() << endl;
    };
}
/*
void CUndoControl::removeChangeListenerDeletedOk(void* th, CUndoListener* listener, ReferenceUtiOB* changedata)
{
    std::map<void*, CUndoChangeList>::iterator it = m_changelist.find(th);
    if (it == m_changelist.end()) return;
       
    utiassert(it->second.m_updateobjects.find(listener) != it->second.m_updateobjects.end());
    it->second.m_updateobjects.erase(listener);
    if (it->second.m_updateobjects.size() == 0) {
       m_changelist.erase(it);
       removeFromHashMap(th);
    };
}*/

CUndoCommand::CUndoCommand()
{

}

CUndoCommand::~CUndoCommand()
{

}
void CUndoCommand::undo()
{

}

void CUndoCommand::redo()
{

}


CUndoSnapShot::CUndoSnapShot()
{

}

CUndoSnapShot::~CUndoSnapShot()
{

}

bool CUndoSnapShot::alreadyInThisSnapShot(void* data)
{
  return m_commandset.find(data) != m_commandset.end();
}


int CUndoSnapShot::commandCount()
{CHECKEXISTANCE;
   return m_commands.size();
}


void CUndoSnapShot::addCommand(CUndoCommand* command)
{CHECKEXISTANCE;
   if (m_commandset.find(command->getOwner()) != m_commandset.end()) {
     m_commandset.insert(command->getOwner());
   }
   m_commands.push_back(command);
}
void CUndoSnapShot::undo()
{CHECKEXISTANCE;
   for (int i = m_commands.size()-1; i >= 0; i--) {
      m_commands[i].undo();
   };
}

void CUndoSnapShot::redo()
{CHECKEXISTANCE;
   for (int i = 0; i < m_commands.size(); i++) {
       m_commands[i].redo();
   };
}

void CUndoSnapShot::rollback()
{CHECKEXISTANCE;
   undo();
   m_commands.clear();
   m_commandset.clear();
}

CUndoDocumentListener::CUndoDocumentListener()
{
  
}

CUndoDocument::CUndoDocument()
{
   m_undosnapshot.push_back(new CUndoSnapShot());
   m_document = NULL;
   m_listener = NULL;
}
CUndoDocument::~CUndoDocument()
{

}

void CUndoDocument::setDocument(ReferenceUtiOB* obj)
{
m_document = obj;
}
ReferenceUtiOB* CUndoDocument::getDocument()
{
return m_document;
}

CUndoDocumentListener* CUndoDocument::getListener()
{
  return m_listener;
}

void CUndoDocument::setListener(CUndoDocumentListener* doclistener)
{CHECKEXISTANCE;
  m_listener = doclistener;
}

void CUndoDocument::clearIfSnapShotIsEmpty()
{CHECKEXISTANCE;
  utiassert(m_undosnapshot.size() != 0);
  if (m_undosnapshot[m_undosnapshot.size()-1].commandCount() == 0) {
    m_undosnapshot.deleteItem(m_undosnapshot.size()-1);
    std::wcout << "CUndoDocument::clearIfSnapShotIsEmpty() was successful" << std::endl;
  };
}

void CUndoDocument::snapshot()
{CHECKEXISTANCE;
   m_undosnapshot.push_back(new CUndoSnapShot());
}

void CUndoDocument::rollback()
{CHECKEXISTANCE;
   m_undosnapshot[m_undosnapshot.size()-1].rollback();
}

void CUndoDocument::undo()
{CHECKEXISTANCE;
    if (m_undosnapshot.size() == 0) return;
    CUndoSnapShot* sh =& m_undosnapshot[m_undosnapshot.size()-1];
    sh->undo();
    m_redosnapshot.push_back(sh);
    m_undosnapshot.pop_back();
}
void CUndoDocument::redo()
{CHECKEXISTANCE;
  if (m_redosnapshot.size() == 0) return;
   CUndoSnapShot* sh =& m_redosnapshot[m_redosnapshot.size()-1];
    sh->redo();
    m_undosnapshot.push_back(sh);
    m_redosnapshot.pop_back();
}

void CUndoDocument::addCommand(CUndoCommand* command)
{CHECKEXISTANCE;
  if (m_undosnapshot.size() == 0) snapshot();
   m_undosnapshot[m_undosnapshot.size()-1].addCommand(command);
  if (m_listener) m_listener->docChanged(this);
}

bool CUndoDocument::alreadyInThisSnapShot(void* data)
{CHECKEXISTANCE;
  CUndoSnapShot& sh = m_undosnapshot[m_undosnapshot.size()-1];
  return sh.alreadyInThisSnapShot(data);
}
