/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UNDOSYSTEMLIST__
#define __UNDOSYSTEMLIST__

#include "undosystem.h"

template<class List> class CWatchedListClearUndo : public CUndoCommand
{
protected:
   void* m_list;
   List m_savelist;
public:
   CWatchedListClearUndo(void* list)
   {
       m_list = list;
       readList();
   }
   void readList();
   virtual void undo();
   virtual void redo();
};

template<class List> class CWatchedListPushBackUndo : public CUndoCommand
{
protected:
   void* m_list;
   uti_refptr<typename List::value_type> m_value;
public:
   CWatchedListPushBackUndo(void* list, typename List::value_type* v)
   {
       m_list = list;
       m_value = v;
   }
   virtual void undo();
   virtual void redo();
};


template<class List> class CWatchedListInsertUndo : public CUndoCommand
{
protected:
   void* m_list;
   size_t m_pos;
   uti_refptr<typename List::value_type> m_value;
public:
   CWatchedListInsertUndo(void* list, size_t p, typename List::value_type* v)
   {
       m_list = list;
       m_value = v;
       m_pos = p;
   }
   virtual void undo();
   virtual void redo();
};

template<class List> class CWatchedListPopBackUndo : public CUndoCommand
{
protected:
   void* m_list;
   uti_refptr<typename List::value_type> m_value;
public:
   CWatchedListPopBackUndo(void* list, typename List::value_type* v)
   {
       m_list = list;
       m_value = v;
   }
   virtual void undo();
   virtual void redo();
};

template<class List> class CWatchedListDeleteItemUndo : public CUndoCommand
{
protected:
   void* m_list;
   uti_refptr<typename List::value_type> m_value;
   int m_index;
public:
   CWatchedListDeleteItemUndo(void* list, int index, typename List::value_type* v)
   {
       m_list = list;
       m_value = v;
       m_index = index;
   }
   virtual void undo();
   virtual void redo();
};

template<class List> class CWatchedListMoveUndo : public CUndoCommand
{
protected:
   void* m_list;
   int m_index1;
   int m_index2;
public:
   CWatchedListMoveUndo(void* list, int index1, int index2)
   {
       m_list = list;
       
       m_index1 = index1;
       m_index2 = index2;
   }
   virtual void undo();
   virtual void redo();
};

template<class List> class CWatchedListExchangeUndo : public CUndoCommand
{
protected:
   void* m_list;
   uti_refptr<typename List::value_type> m_value;
   uti_refptr<typename List::value_type> m_newvalue;
   int m_index;
public:
   CWatchedListExchangeUndo(void* list, int index, typename List::value_type* oldv, typename List::value_type* newv)
   {
       m_list = list;
       m_value = oldv;
       m_newvalue = newv;
       m_index = index;
   }
   virtual void undo();
   virtual void redo();
};

template <class List, bool t_doundo=UNDO_FLAG> struct CListFacade
{
public:
   List m_list;
   int m_listeners;
public:
   typedef typename List::value_type value_type;
   typedef typename List::size_type size_type;
   typedef typename List::difference_type difference_type;
   typedef typename List::pointer pointer;
   typedef typename List::const_pointer const_pointer;
   typedef typename List::reference reference;
   typedef typename List::const_reference const_reference;
   typedef typename List::iterator iterator;
   //inline iterator begin() { return m_list.begin();};
  //inline  iterator end() {return m_list.end();};
  inline List& getInternData() {
      return m_list;
   };
   explicit CListFacade() {m_listeners = 0;}
   ~CListFacade()
   {
     #ifdef UNDO_USECHANGE
       removeListeners();
      #endif
   }
   void removeListeners()
    {
     #ifdef UNDO_USECHANGE
       CUndoControl* c = CUndoControl::getInstance();
       if (c->hasListener(this)) {
          c->deleteWatch(this);
       }
       m_listeners = 0;
       #endif
    }
    void addChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       #ifdef UNDO_USECHANGE
        CUndoControl* c = CUndoControl::getInstance();
        c->addChangeListener(this, listener, changedata);
	m_listeners++;
       #endif
        
    }
    void removeChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       #ifdef UNDO_USECHANGE
        CUndoControl* c = CUndoControl::getInstance();
        c->removeChangeListener(this, listener, changedata);
	m_listeners--;
	utiassert(m_listeners >= 0);
      #endif
    }
   inline reference front() { return m_list.front();};
   inline reference back() { return m_list.back();};
   inline const_reference front() const { return m_list.front();};
   inline const_reference back() const { return m_list.front();};
   void clear()
   {
      if (t_doundo) {
	   CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedListClearUndo<List>(this));
       }
       if (m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_clearvalue(this); 
       };
       m_list.clear();
   }
   void push_back(pointer typ)
   {   
       m_list.push_back(typ);
       if (t_doundo) {
	   CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedListPushBackUndo<List>(this, typ));
       }
       if (m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_addvalue(this, m_list.size()-1); 
       };
       
   }
   void pop_back()
   {
     utiassert(m_list.size() != 0);
       if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedListPopBackUndo<List>(this, &m_list.back()));
       }
       if (m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_removevalue(this, m_list.size()-1); 
       };
       m_list.pop_back();
   }
   inline size_type size() const { return m_list.size(); }
   inline size_type capacity() const { return m_list.capacity();}
   inline bool empty() const {return m_list.empty();};
   inline size_type max_size() const {return m_list.max_size();};
   inline reference operator[](size_type p)  { return m_list[p]; }
    inline const_reference operator[](size_type p) const { return m_list[p]; }
    inline void deleteItem(size_t i)
    {
      utiassert(i >= 0 && i< m_list.size());
        if (t_doundo) {
	   CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedListDeleteItemUndo<List>(this, i, &m_list[i]));
       }
        m_list.deleteItem(i);
	if (m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_removevalue(this, i); 
       };
    }
    inline iterator erase(const iterator& iter)
    {
        deleteItem(iter.position);
        return iter;
    }
    inline void insert(size_type p, pointer typ)
   {
       m_list.insert(p, typ);
       if (t_doundo) {
	   CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedListInsertUndo<List>(this, p, typ));
       }
       if (m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_addvalue(this,p); 
       };
       
   }
    inline void remove(const_pointer t)
    {
        off_t in = indexOf(t);
        utiassert(in != UTI_INVALIDINDEX);
        deleteItem(in);
    }
    inline off_t indexOf(const_pointer t) const  { return m_list.indexOf(t); }
    inline void move(size_t pos, size_t pos2)
    {
        if (t_doundo) {
	   CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedListMoveUndo<List>(this, pos, pos2));
       }
       if (m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_movevalue(this, pos, pos2); 
       };
        m_list.move(pos, pos2);
    }
    inline void exchange(size_t index, pointer t)
    {
      uti_refptr<value_type> old;
      old = &m_list[index];
      m_list.exchange(index, t);
       if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedListExchangeUndo<List>(this, index, old.get(), t));
       }
        if (m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_exchangevalue(this,index); 
       };
      
    }
    /*void extract(size_t index)
    {
         if (t_doundo) {
          c->addUndoCommand(new CWatchedListExtractUndo<List>(this, i));
       }
         m_list.extract(index);
    }
    void extract(const_pointer t)
    {
        off_t in = indexOf(t);
        utiassert(in != UTI_INVALIDINDEX);
        extract(in);
    }*/
};



template<class List> void CWatchedListClearUndo<List>::readList()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
     for (size_t i = 0; i < l->m_list.size(); i++) {
           m_savelist.push_back(l->m_list[i]);
       }
}

template<class List> void CWatchedListClearUndo<List>::undo()
   {
     CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       for (size_t i = 0; i < m_savelist.size(); i++) {
           l->m_list.push_back((m_savelist)[i]);
	     if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_addvalue(l,l->m_list.size()-1); 
       };
       }
   }
   
template<class List> void CWatchedListClearUndo<List>::redo()
   {
     CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_clearvalue(l); 
       };
       l->m_list.clear();
   }



template<class List> void CWatchedListPushBackUndo<List>::undo()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
    if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_removevalue(l,l->m_list.size()-1); 
       };
       l->m_list.pop_back();
}

template<class List> void CWatchedListPushBackUndo<List>::redo()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       l->m_list.push_back(m_value.get());
        if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_addvalue(l,l->m_list.size()-1); 
       };
}


template<class List> void CWatchedListInsertUndo<List>::undo()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
   if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_removevalue(l,m_pos); 
       };
       l->m_list.deleteItem(m_pos);
}

template<class List> void CWatchedListInsertUndo<List>::redo()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       l->m_list.insert(m_pos, m_value.get());
        if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_addvalue(l,m_pos); 
       };
}

template<class List> void CWatchedListPopBackUndo<List>::undo()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       l->m_list.push_back(m_value.get());
       if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_addvalue(l,l->m_list.size()-1); 
       };
}

template<class List> void CWatchedListPopBackUndo<List>::redo()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
  if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_removevalue(l,l->m_list.size()-1); 
       };
       l->m_list.pop_back();
}


template<class List> void CWatchedListDeleteItemUndo<List>::undo()
{
 CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       l->m_list.insert(m_index, m_value.get());
       if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_addvalue(l,m_index); 
       };
}

template<class List> void CWatchedListDeleteItemUndo<List>::redo()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
  if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_removevalue(l,m_index); 
       };
       l->m_list.deleteItem(m_index);
}



template<class List> void CWatchedListMoveUndo<List>::undo()
{
 CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       l->m_list.move(m_index1, m_index2);
        if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_movevalue(l,m_index1, m_index2); 
       };
}

template<class List> void CWatchedListMoveUndo<List>::redo()
{
  CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       l->m_list.move(m_index1, m_index2);
       if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_movevalue(l,m_index1, m_index2); 
       };
}


template<class List> void CWatchedListExchangeUndo<List>::undo()
{
 CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       l->m_list.exchange(m_index, m_value.get());
       if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_exchangevalue(l,m_index); 
       };
}

template<class List> void CWatchedListExchangeUndo<List>::redo()
{
CListFacade<List, true>* l = (CListFacade<List, true>*)m_list;
       l->m_list.exchange(m_index, m_newvalue.get());
        if (l->m_listeners != 0) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->vec_exchangevalue(l,m_index); 
       };
}


template<class T> class CWatchedListReferenceAdapter : public CBaseListReferenceAdapter
{
  protected:
    CListFacade< CReferenceList<T> >& m_ref;
  public:
    CWatchedListReferenceAdapter(CListFacade< CReferenceList<T> >& ref) : m_ref(ref)
    {
      
    };
    
     virtual ReferenceUtiOB* get(int i) const
     {
         return &m_ref[i]; 
     }
    virtual void set(int i, ReferenceUtiOB* ob)
    {
         m_ref.exchange(i,(T*)ob);
    };
    virtual int size()
    {
        return m_ref.size(); 
    }
    virtual void deleteItem(int index) 
    {
        m_ref.deleteItem(index); 
    }
    virtual void deleteObject(ReferenceUtiOB* ob) 
    {
        m_ref.remove((T*)ob); 
    }
    virtual void push_back(ReferenceUtiOB* ob)
    {
        return m_ref.push_back((T*)ob); 
    }
    virtual void insert(int index, ReferenceUtiOB* ob)
    {
      m_ref.insert(index, (T*)ob);
    }
    virtual void move(int i, int j)
    {
      m_ref.move(i,j);
    }

    virtual int indexOf(ReferenceUtiOB* ob)
    {
     return m_ref.indexOf((T*)ob); 
    }
    virtual void clear()
    {
        return m_ref.clear(); 
    }
    virtual void addChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
      m_ref.addChangeListener(listener, changedata);
    };
    virtual void removeChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       m_ref.removeChangeListener(listener, changedata);
    };
};

#endif
