/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UNDOSYSTEMSETMAP__
#define __UNDOSYSTEMSETMAP__

#include "undosystem.h"

template<class T> class CWatchedSetClearUndo : public CUndoCommand
{
protected:
    void* m_setfacade;
    std::vector<T> m_objects;
public:
    CWatchedSetClearUndo(void* p)
    {
       m_setfacade = p;
       readObjects();
       
    }
    ~CWatchedSetClearUndo()
    {
    
    }
    void readObjects();
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_setfacade; 
    }
};

template<class T> class CWatchedSetEraseUndo : public CUndoCommand
{
protected:
    T m_item;
    void* m_setfacade;
public:
    CWatchedSetEraseUndo(void* p, const T& value)
    {
       m_setfacade = p;
       m_item = value;
    }
    ~CWatchedSetEraseUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_setfacade; 
    }
};

template<class T> class CWatchedSetInsertUndo : public CUndoCommand
{
protected:
    T m_item;
    void* m_setfacade;
public:
    CWatchedSetInsertUndo(void* p, const T& value)
    {
       m_setfacade = p;
       m_item = value;
    }
    ~CWatchedSetInsertUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_setfacade; 
    }
};


template <class T, bool t_doundo=UNDO_FLAG> struct CSetFacade
{
public:
   typedef std::set<T> List;
   List m_list;
public:
   typedef typename List::value_type value_type;
   typedef typename List::size_type size_type;
   typedef typename List::difference_type difference_type;
   typedef typename List::pointer pointer;
   typedef typename List::const_pointer const_pointer;
   typedef typename List::reference reference;
   typedef typename List::const_reference const_reference;
   typedef typename List::const_iterator iterator;
   inline iterator begin() { return m_list.begin();};
   inline iterator end() {return m_list.end();};
   
   inline std::set<T>& getInternData() {
      return m_list;
   };
   
   explicit CSetFacade() {}
   ~CSetFacade()
   {
    #ifdef UNDO_USECHANGE
       removeListeners();
      #endif
   }
   void removeListeners()
    {
     #ifdef UNDO_USECHANGE
       CUndoControl* c = CUndoControl::getInstance();
       if (c->hasListener(this)) {
          c->deleteWatch(this);
       }
       #endif
    }
    void addChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       #ifdef UNDO_USECHANGE
        CUndoControl* c = CUndoControl::getInstance();
        c->addChangeListener(this, listener, changedata);
       #endif
        
    }
    void removeChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       #ifdef UNDO_USECHANGE
        CUndoControl* c = CUndoControl::getInstance();
        c->removeChangeListener(this, listener, changedata);
	#endif
    }

   inline reference front() { return m_list.front();};
   inline reference back() { return m_list.back();};
   inline const_reference front() const { return m_list.front();};
   inline const_reference back() const { return m_list.front();};
   void clear()
   { 
       if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedSetClearUndo<T>(this));
       }
       
       m_list.clear();
   }
   inline size_type size() const { return m_list.size(); }
   inline size_type capacity() const { return m_list.capacity();}
   inline bool empty() const {return m_list.empty();};
   inline size_type max_size() const {return m_list.max_size();};
   //inline reference operator[](size_type p)  { return m_list[p]; }
    inline iterator find(const T& t)
    {
        
        return m_list.find(t);
    }
    size_type erase(const T& t)
    {
        utiassert(m_list.find(t) != m_list.end());
        if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedSetEraseUndo<T>(this, t));
       }
       // m_list.erase(t);
      
        return m_list.erase(t);
    }
    iterator erase(iterator iter)
    {
      utiassert(m_list.end() != iter);
        if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedSetEraseUndo<T>(this, *iter));
       }
        m_list.erase(iter);
        return iter;
    }
    std::pair<iterator, bool> insert(const T& t)
    {
        utiassert(m_list.find(t) == m_list.end());
       if (t_doundo) {
	 CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedSetInsertUndo<T>(this, t));
       }
       return m_list.insert(t);
    }
    /*inline iterator insert(iterator p, const T& t)
    {
       m_list.insert(p, t);
    }*/
};


template<class T> void CWatchedSetClearUndo<T>::undo()
{
   CSetFacade<T, true>* watched = (CSetFacade<T, true>*)m_setfacade;
   utiassert(watched->m_list.size() == 0);
   for (typename std::vector<T>::iterator it = m_objects.begin(); it != m_objects.end(); it++) {
      
      watched->m_list.insert(*it);
   };
}

template<class T> void CWatchedSetClearUndo<T>::redo()
{
    CSetFacade<T, true>* watched = (CSetFacade<T, true>*)m_setfacade;
   watched->m_list.clear();
}

template<class T> void CWatchedSetClearUndo<T>::readObjects()
{
    CSetFacade<T, true>* watched = (CSetFacade<T, true>*)m_setfacade;
   for (typename std::set<T>::iterator it = watched->m_list.begin(); it != watched->m_list.end(); it++) {
      m_objects.push_back(*it);
   };
}

template<class T> void CWatchedSetEraseUndo<T>::undo()
{
   CSetFacade<T, true>* watched = (CSetFacade<T, true>*)m_setfacade;
   watched->m_list.insert(m_item);
   
}

template<class T> void CWatchedSetEraseUndo<T>::redo()
{
    CSetFacade<T, true>* watched = (CSetFacade<T, true>*)m_setfacade;
   watched->m_list.erase(m_item);
}

template<class T> void CWatchedSetInsertUndo<T>::undo()
{
   CSetFacade<T, true>* watched = (CSetFacade<T, true>*)m_setfacade;
   watched->m_list.erase(m_item);
   
}

template<class T> void CWatchedSetInsertUndo<T>::redo()
{
    CSetFacade<T, true>* watched = (CSetFacade<T, true>*)m_setfacade;
    watched->m_list.insert(m_item);
}


template<class T, class U> class CWatchedMapClearUndo : public CUndoCommand
{
protected:
    void* m_mapfacade;
    std::vector<std::pair<T, U> > m_objects;
public:
    CWatchedMapClearUndo(void* p)
    {
       m_mapfacade = p;
       readObjects();
       
    }
    ~CWatchedMapClearUndo()
    {
    
    }
    void readObjects();
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_mapfacade; 
    }
};

template<class T, class U> class CWatchedMapEraseUndo : public CUndoCommand
{
protected:
    std::pair<T, U> m_item;
    void* m_mapfacade;
public:
    CWatchedMapEraseUndo(void* p, const T& value)
    {
       m_mapfacade = p;
       m_item.first = value;
       readOtherValue();
    }
    ~CWatchedMapEraseUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void readOtherValue();
    void* getOwner()
    {
       return m_mapfacade; 
    }
};

template<class T, class U> class CWatchedMapInsertUndo : public CUndoCommand
{
protected:
    std::pair<T, U> m_item;
    void* m_mapfacade;
public:
    CWatchedMapInsertUndo(void* p, const std::pair<T, U>& value)
    {
       m_mapfacade = p;
       m_item = value;
    }
    ~CWatchedMapInsertUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_mapfacade; 
    }
};

template <class T, class U, bool t_doundo=UNDO_FLAG> struct CMapFacade
{
public:
   typedef std::map<T, U> List;
   List m_list;
public:
   typedef typename List::value_type value_type;
   typedef typename List::size_type size_type;
   typedef typename List::difference_type difference_type;
   typedef typename List::pointer pointer;
   typedef typename List::const_pointer const_pointer;
   typedef typename List::reference reference;
   typedef typename List::const_reference const_reference;
   typedef typename List::iterator iterator;
   typedef typename List::const_iterator const_iterator;
   inline iterator begin() { return m_list.begin();};
   inline iterator end() {return m_list.end();};
   inline const_iterator begin() const { return m_list.begin();};
   inline const_iterator end() const {return m_list.end();};
   inline std::map<T, U>& getInternData() {
      return m_list;
   };
   
   explicit CMapFacade() {}
   ~CMapFacade()
   {
    #ifdef UNDO_USECHANGE
       removeListeners();
      #endif
   }
    
    void removeListeners()
    {
     #ifdef UNDO_USECHANGE
       CUndoControl* c = CUndoControl::getInstance();
       if (c->hasListener(this)) {
          c->deleteWatch(this);
       }
       #endif
    }
    void addChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       #ifdef UNDO_USECHANGE
        CUndoControl* c = CUndoControl::getInstance();
        c->addChangeListener(this, listener, changedata);
       #endif
        
    }
    void removeChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       #ifdef UNDO_USECHANGE
        CUndoControl* c = CUndoControl::getInstance();
        c->removeChangeListener(this, listener, changedata);
	#endif
    }

   inline reference front() { return m_list.front();};
   inline reference back() { return m_list.back();};
   inline const_reference front() const { return m_list.front();};
   inline const_reference back() const { return m_list.front();};
   void clear()
   {
       if (t_doundo) {
	 CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedMapClearUndo<T, U>(this));
       }
       m_list.clear();
   }
   inline size_type size() const { return m_list.size(); }
   inline size_type capacity() const { return m_list.capacity();}
   inline bool empty() const {return m_list.empty();};
   inline size_type max_size() const {return m_list.max_size();};
   //inline reference operator[](size_type p)  { return m_list[p]; }
    inline const_iterator find(const T& t) const
    {
        
        return m_list.find(t);
    }
    inline iterator find(const T& t) 
    {
        
        return m_list.find(t);
    }
    inline size_type erase(const T& t)
    {
      utiassert(m_list.find(t) != m_list.end());
       if (t_doundo) {
	 CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedMapEraseUndo<T, U>(this, t));
       }
       return m_list.erase(t);
     
    }
    inline iterator erase(iterator iter)
    {
        utiassert(iter != m_list.end());
        if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedMapEraseUndo<T, U>(this, iter->first));
       }
        m_list.erase(iter);
        return iter;
    }
    inline iterator insert(const value_type& t)
    {
      utiassert(m_list.find(t.first) == m_list.end());
       if (t_doundo) {
	 CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedMapInsertUndo<T, U>(this, t));
       }
       m_list.insert(t);
	   return m_list.find(t.first);
    }
    
};

template<class T, class U> void CWatchedMapClearUndo<T, U>::undo()
{
   CMapFacade<T, U, true>* watched = (CMapFacade<T, U, true>*)m_mapfacade;
   utiassert(watched->m_list.size() == 0);
   for (typename std::vector<std::pair<T, U> >::iterator it = m_objects.begin(); it != m_objects.end(); it++) {
      
      watched->m_list.insert(*it);
   };
}

template<class T, class U> void CWatchedMapClearUndo<T, U>::redo()
{
    CMapFacade<T, U, true>* watched = (CMapFacade<T, U, true>*)m_mapfacade;
   watched->m_list.clear();
}

template<class T, class U> void CWatchedMapClearUndo<T, U>::readObjects()
{
    CMapFacade<T, U, true>* watched = (CMapFacade<T, U, true>*)m_mapfacade;
   for (typename std::map<T, U>::iterator it = watched->m_list.begin(); it != watched->m_list.end(); it++) {
      m_objects.push_back(*it);
   };
}

template<class T, class U> void CWatchedMapEraseUndo<T, U>::undo()
{
   CMapFacade<T, U, true>* watched = (CMapFacade<T, U, true>*)m_mapfacade;
   watched->m_list.insert(m_item);
   
}

template<class T, class U> void CWatchedMapEraseUndo<T, U>::redo()
{
    CMapFacade<T, U, true>* watched = (CMapFacade<T, U, true>*)m_mapfacade;
   watched->m_list.erase(m_item.first);
}

template<class T, class U> void CWatchedMapEraseUndo<T, U>::readOtherValue()
{
  CMapFacade<T, U, true>* watched = (CMapFacade<T, U, true>*)m_mapfacade;
  
  typename std::map<T, U>::iterator it = watched->m_list.find(m_item.first);
  utiassert(it != watched->m_list.end());
  m_item.second = it->second;
}

template<class T, class U> void CWatchedMapInsertUndo<T, U>::undo()
{
   CMapFacade<T, U, true>* watched = (CMapFacade<T, U, true>*)m_mapfacade;
   watched->m_list.erase(m_item.first);
   
}

template<class T, class U> void CWatchedMapInsertUndo<T, U>::redo()
{
    CMapFacade<T, U, true>* watched = (CMapFacade<T, U, true>*)m_mapfacade;
    watched->m_list.insert(m_item);
}


#endif
