/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __UNDOSYSTEMVECTOR__
#define __UNDOSYSTEMVECTOR__

#include "undosystem.h"

template<class T> class CWatchedVectorClearUndo : public CUndoCommand
{
protected:
    void* m_vectorfacade;
    std::vector<T> m_objects;
public:
    CWatchedVectorClearUndo(void* p)
    {
       m_vectorfacade = p;
       readObjects();
       
    }
    ~CWatchedVectorClearUndo()
    {
    
    }
    void readObjects();
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_vectorfacade; 
    }
};

template<class T> class CWatchedVectorPushBackUndo : public CUndoCommand
{
protected:
    void* m_vectorfacade;
    T m_value;
public:
    CWatchedVectorPushBackUndo(void* p, const T& value)
    {
       m_vectorfacade = p;
       m_value = value;
       
    }
    ~CWatchedVectorPushBackUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_vectorfacade; 
    }
};


template<class T> class CWatchedVectorPopBackUndo : public CUndoCommand
{
protected:
    void* m_vectorfacade;
    T m_value;
public:
    CWatchedVectorPopBackUndo(void* p, const T& value)
    {
       m_vectorfacade = p;
       m_value = value;
       
    }
    ~CWatchedVectorPopBackUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_vectorfacade; 
    }
};

template<class T> class CWatchedVectorEraseUndo : public CUndoCommand
{
protected:
    void* m_vectorfacade;
    T m_value;
    size_t m_pos;
public:
    CWatchedVectorEraseUndo(void* p, size_t s, const T& value)
    {
       m_vectorfacade = p;
       m_value = value;
       m_pos = s;
       
    }
    ~CWatchedVectorEraseUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_vectorfacade; 
    }
};

template<class T> class CWatchedVectorInsertUndo : public CUndoCommand
{
protected:
    void* m_vectorfacade;
    T m_value;
    size_t m_pos;
public:
    CWatchedVectorInsertUndo(void* p, size_t s, const T& value)
    {
       m_vectorfacade = p;
       m_value = value;
       m_pos = s;
       
    }
    ~CWatchedVectorInsertUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_vectorfacade; 
    }
};


template<class T> class CWatchedVectorSetUndo : public CUndoCommand
{
protected:
    void* m_vectorfacade;
    T m_value;
    T m_oldvalue;
    size_t m_pos;
public:
    CWatchedVectorSetUndo(void* p, size_t s, const T& oldvalue, const T& value)
    {
       m_vectorfacade = p;
       m_oldvalue = oldvalue;
       m_value = value;
       m_pos = s;
       
    }
    ~CWatchedVectorSetUndo()
    {
    
    }
    void readObjects();
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_vectorfacade; 
    }
};


template <class T, bool t_doundo=UNDO_FLAG> struct CVectorFacade
{
public:
   typedef std::vector<T> List;
   List m_list;
public:
   typedef typename List::value_type value_type;
   typedef typename List::size_type size_type;
   typedef typename List::difference_type difference_type;
   typedef typename List::pointer pointer;
   typedef typename List::const_pointer const_pointer;
   typedef typename List::reference reference;
   typedef typename List::const_reference const_reference;
   typedef typename List::const_iterator iterator;
   inline iterator begin() { return m_list.begin();};
   inline iterator end() {return m_list.end();};
   inline std::vector<T>& getInternData() {
      return m_list;
   };
   
   explicit CVectorFacade() {}
   ~CVectorFacade()
   {
    #ifdef UNDO_USECHANGE
       removeListeners();
      #endif
   }
   void removeListeners()
    {
     #ifdef UNDO_USECHANGE
       CUndoControl* c = CUndoControl::getInstance();
       if (c->hasListener(this)) {
          c->deleteWatch(this);
       }
       #endif
    }
    void addChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       #ifdef UNDO_USECHANGE
        CUndoControl* c = CUndoControl::getInstance();
        c->addChangeListener(this, listener, changedata);
       #endif
        
    }
    void removeChangeListener(CUndoVectorListener* listener, ReferenceUtiOB* changedata)
    {
       #ifdef UNDO_USECHANGE
        CUndoControl* c = CUndoControl::getInstance();
        c->removeChangeListener(this, listener, changedata);
      #endif
    }
   inline reference front() { return m_list.front();};
   inline reference back() { return m_list.back();};
   inline const_reference front() const { return m_list.front();};
   inline const_reference back() const { return m_list.front();};
   void clear()
   {
       if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorClearUndo<T>(this));
       }
       m_list.clear();
   }
   void push_back(const T & typ)
   {
      if (t_doundo) {
	 CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorPushBackUndo<T>(this, typ));
       }
       m_list.push_back(typ);
   }
   void pop_back()
   {
       utiassert(m_list.size() != 0);
       if (t_doundo) {
	 
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorPopBackUndo<T>(this, m_list.back()));
       }
       m_list.pop_back();
   }
   inline size_type size() const { return m_list.size(); }
   inline size_type capacity() const { return m_list.capacity();}
   inline bool empty() const {return m_list.empty();};
   inline size_type max_size() const {return m_list.max_size();};
   //inline reference operator[](size_type p)  { return m_list[p]; }
    inline const_reference operator[](size_type p) const {assert(p >= 0); utiassert(p< m_list.size());return m_list[p]; }
     inline const_reference at(size_type p) const {assert(p >= 0); utiassert(p< m_list.size());return m_list.at(p);};
     inline void set(size_type p, const T& v) 
     {
       utiassert(p >= 0); utiassert(p< m_list.size());
         if (t_doundo) {
	   CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorSetUndo<T>(this, p, m_list[p], v));
       }
         m_list[p] = v;
    };
   
    inline void erase(iterator iter)
    {
         utiassert(iter != m_list.end());
	  utiassert(m_list.size() != 0);
        if (t_doundo) {
	  
	   CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorEraseUndo<T>(this, iter-m_list.begin(), *iter));
       }
       int p = iter-m_list.begin();
        m_list.erase(m_list.begin()+p);
    }
   /* inline iterator erase(iterator first, iterator last)
    {
       utiassert(first != m_list.end());
       utiassert(m_list.size() != 0);
        if (t_doundo) {
	   CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorEraseRangeUndo<T>(this, first-m_list.begin(), last-m_list.begin()));
       }
       // m_list.erase(first, last);
        return m_list.erase(first, last);
    }*/
    inline void deleteItem(size_type index)
    {
       erase(m_list.begin() + index);
    }
    inline off_t indexOf(const T& t)
    {
       for (size_type i = 0;i < m_list.size(); i++) {
          if (m_list[i] == t) return i;
       };
       return UTI_INVALIDINDEX;
    }
    inline void insert(iterator p, const T& t)
    {
      utiassert(p != m_list.end());
       if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorInsertUndo<T>(this, p-m_list.begin(), t));
       }
       int p2 = p-m_list.begin();
       m_list.insert(m_list.begin()+p2, t);
    }
     inline void insert(int p, const T& t)
    {
      
       insert(m_list.begin()+ p, t);
    }
   /* inline iterator insert(iterator p, size_type n, const T& x)
    {
         utiassert(p != m_list.end());
       if (t_doundo) {
	  CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorInsertRangeUndo<T>(this, p-m_list.begin(), n, t));
       }
       return m_list.insert(p, n, x);
    }*/
   /* inline void assign(size_type c, const T& t)
    {
      utiassert(c >= 0);
      utiassert(c < m_list.size());
    if (t_doundo) {
         CUndoControl* c = CUndoControl::getInstance();
          c->addUndoCommand(new CWatchedVectorAssignUndo<T>(this, c, m_list[c], t));
       }
       m_list.assign(c, t);
    }*/
    
};

template<class T> void CWatchedVectorClearUndo<T>::undo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   utiassert(watched->m_list.size() == 0);
   for (typename std::vector<T>::iterator it = m_objects.begin(); it != m_objects.end(); it++) {
      
      watched->m_list.push_back(*it);
   };
}

template<class T> void CWatchedVectorClearUndo<T>::redo()
{
    CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   watched->m_list.clear();
}

template<class T> void CWatchedVectorClearUndo<T>::readObjects()
{
    CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   for (typename std::vector<T>::iterator it = watched->m_list.begin(); it != watched->m_list.end(); it++) {
      m_objects.push_back(*it);
   };
}

template<class T> void CWatchedVectorPushBackUndo<T>::undo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   utiassert(watched->m_list.size() != 0);
   watched->m_list.pop_back();
}

template<class T> void CWatchedVectorPushBackUndo<T>::redo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   watched->m_list.push_back(m_value);
}

template<class T> void CWatchedVectorPopBackUndo<T>::undo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   utiassert(watched->m_list.size() != 0);
   watched->m_list.push_back(m_value);
}

template<class T> void CWatchedVectorPopBackUndo<T>::redo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   watched->m_list.pop_back();
}

template<class T> void CWatchedVectorEraseUndo<T>::undo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   
   watched->m_list.insert(watched->m_list.begin()+m_pos, m_value);
}

template<class T> void CWatchedVectorEraseUndo<T>::redo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   watched->m_list.erase(watched->m_list.begin()+m_pos);
}

template<class T> void CWatchedVectorInsertUndo<T>::undo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   watched->m_list.erase(watched->m_list.begin()+m_pos);
}

template<class T> void CWatchedVectorInsertUndo<T>::redo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
  watched->m_list.insert(watched->m_list.begin()+m_pos, m_value);
}

template<class T> void CWatchedVectorSetUndo<T>::undo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
   utiassert(watched->m_list.size() != 0);
   watched->m_list.at(m_pos) = m_oldvalue;
}

template<class T> void CWatchedVectorSetUndo<T>::redo()
{
   CVectorFacade<T, true>* watched = (CVectorFacade<T, true>*)m_vectorfacade;
    watched->m_list.at(m_pos) = m_value;
}

#endif
