/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "uti64coder.h"

#include <assert.h>
#include "UtiString.h"

const char UTI_NUMBERTOBASE64[64] =
   {
       'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
       'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
       'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
       'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
       'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
       'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
       '8', '9', '+', '/'
   };

static int UTI_BASE64TONUMBER[256];
static bool uti_base64init = false;

void uti_initbase64()
{
    if (uti_base64init) return;
    for (int i = 0; i < 256;i++) UTI_BASE64TONUMBER[i] = -1;
    for (int i = 0; i < 64; i++) UTI_BASE64TONUMBER[(unsigned char)UTI_NUMBERTOBASE64[i]] = i;

    uti_base64init = true;
}

CUti64Coder::CUti64Coder()
{
    //ctor
}

CUti64Coder::~CUti64Coder()
{
    //dtor
}


int CUti64Coder::getDataLength(int datasize)
{
   //3 Bytes => 4 Zeichen
   int l_chars = (datasize / 3)*4;
   int l_rest = datasize % 3;
   if (l_rest == 0) return l_chars;
   if (l_rest == 1) return l_chars+2;
   if (l_rest == 2) return l_chars+3;
      utiassert(false);
   return 0;
}

int CUti64Coder::reconstructDataLength(int sourcesize)
{
    int l_blocks = sourcesize / 4;
    int l_rest = sourcesize % 4;
    utiassert(l_rest != 1);
    if (l_rest == 0) return l_blocks*3;
    if (l_rest == 2) return l_blocks*3+1;
    if (l_rest == 3) return l_blocks*3+2;
    throw new CUti64CodeSize();
    return 0;
}


int CUti64Coder::writeData(const void* data, int datasize, char* targetdata, int targetsize)
{
    uassert(getDataLength(datasize)<= targetsize, UTEXT("CUti64Coder::writeData: not enough space free in the target buffer"));
    int l_blocks = (datasize / 3);
    int l_rest = datasize % 3;
    const unsigned char* datapos = (const unsigned char*)data;
    char* targetpos = targetdata;
    unsigned char index;
    for (int i = 0; i < l_blocks; i++) {
        utiassert(datapos - ((unsigned char*)data) < datasize);
        utiassert(targetpos - targetdata < targetsize);
        index = datapos[0] >> 2;
        utiassert(index < 64);
        targetpos[0] = UTI_NUMBERTOBASE64[index];

        index = (datapos[0] & 3) << 4;
        index |= datapos[1] >> 4;
        targetpos[1] = UTI_NUMBERTOBASE64[index];


        index = (datapos[1] & 15) << 2;
        index |= datapos[2] >> 6;
        targetpos[2] = UTI_NUMBERTOBASE64[index];

        index = (datapos[2] & 63) ;

        targetpos[3] = UTI_NUMBERTOBASE64[index];
        datapos += 3;
        targetpos += 4;

    }
    if (l_rest == 1)
    {
       utiassert(targetpos - targetdata+1 < targetsize);
       index = datapos[0] >> 2;
        utiassert(index < 64);
        targetpos[0] = UTI_NUMBERTOBASE64[index];

        index = (datapos[0] & 3) << 4;
        targetpos[1] = UTI_NUMBERTOBASE64[index];
        targetpos += 2;
    } else if (l_rest == 2) {
       utiassert(targetpos - targetdata+2 < targetsize);

       index = datapos[0] >> 2;
        utiassert(index < 64);
        targetpos[0] = UTI_NUMBERTOBASE64[index];

        index = (datapos[0] & 3) << 4;
        index |= datapos[1] >> 4;
        targetpos[1] = UTI_NUMBERTOBASE64[index];


        index = (datapos[1] & 15) << 2;
        targetpos[2] = UTI_NUMBERTOBASE64[index];
        targetpos += 3;
    }
    return targetpos - targetdata;
}

void CUti64Coder::readData(const char* sourcedata, int sourcesize, void* targetdata, int targetdatasize)
{
    uti_initbase64();
    uassert(getDataLength(targetdatasize)<= sourcesize, UTEXT("CUti64Coder::readData: the source buffer is not big enough"));
    int l_blocks = (targetdatasize / 3);
    int l_rest = targetdatasize % 3;
    unsigned char* datapos = (unsigned char*)targetdata;
    const char* sourcepos = sourcedata;
    unsigned char index, index2;

    for (int i = 0; i < l_blocks; i++) {
        utiassert(datapos - ((unsigned char*)targetdata) < targetdatasize);
        utiassert(sourcepos - sourcedata < sourcesize);
        index = UTI_BASE64TONUMBER[(unsigned char)sourcepos[0]];
        index2 = UTI_BASE64TONUMBER[(unsigned char)sourcepos[1]];
        if (index == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata);
        if (index2 == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata+1);
        datapos[0] = (index << 2) | (index2 >> 4);

        index = index2;
        index2 = UTI_BASE64TONUMBER[(unsigned char)sourcepos[2]];
         if (index2 == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata+2);
        datapos[1] = ((index & (1 | 2 | 4 | 8)) << 4) | (index2 >> 2);

        index = index2;
        index2 = UTI_BASE64TONUMBER[(unsigned char)sourcepos[3]];
         if (index2 == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata+3);
        datapos[2] = ((index & (1 | 2)) << 6) | (index2);

        datapos += 3;
        sourcepos += 4;

    }
    if (l_rest == 1)
    {

       index = UTI_BASE64TONUMBER[(unsigned char)sourcepos[0]];
        index2 = UTI_BASE64TONUMBER[(unsigned char)sourcepos[1]];
        if (index == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata+0);
        if (index2 == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata+1);
        datapos[0] = (index << 2) | (index2 >> 4);

    } else if (l_rest == 2) {


       index = UTI_BASE64TONUMBER[(unsigned char)sourcepos[0]];
        index2 = UTI_BASE64TONUMBER[(unsigned char)sourcepos[1]];
        if (index == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata+0);
        if (index2 == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata+1);
        datapos[0] = (index << 2) | (index2 >> 4);

        index = index2;
        index2 = UTI_BASE64TONUMBER[(unsigned char)sourcepos[2]];
         if (index2 == (unsigned char)-1) throw new CUti64Code(sourcepos - sourcedata+2);
        datapos[1] = ((index & (1 | 2 | 4 | 8)) << 4) | (index2 >> 2);

    }
}
