#ifndef UTIADVARRAY_H
#define UTIADVARRAY_H

#include "Uti3DCore.h"
#include "ziparchiv.h"
#include "utiintvector.h"
#include "UtiCacheStructures.h"

template<class T, bool fastgrow=false> class DynArray1D {
public:
    Array<T> m_data;
    UtiArrayAccessGrow<T, fastgrow> m_datagrow;
    int m_maxsize;

public:
    typedef T value_type;
    typedef T& reference;
    typedef const T& const_reference;
    typedef T* iterator;
    typedef const T* const_iterator;
    inline DynArray1D(): m_data(8), m_datagrow(m_data)
    {

        assert(&m_data == &m_datagrow.m_array);
        m_maxsize = 32;
    }
    inline DynArray1D(const DynArray1D& other)
    {
        assert(false);
        //No copy operator
    }

    inline T* getCArray()
    {
       return m_data.getCArray();
    }
	inline Array<T>& getAsArray()
	{
		return Array<T>(m_data, 0, m_datagrow.fast_size());
	};

	inline const Array<T>& getAsArray() const
	{
		return Array<T>(m_data, 0, m_datagrow.fast_size());
	};

    inline void setMaxSize(int m)
    {
       m_maxsize = m;
    }
    inline void resize(int s)
    {
        m_data.setSizeCopy(s+16);
        m_maxsize = s;
        m_datagrow.m_position = s;
    }

    inline int getMaxSize()
    {
        return m_maxsize;
    }

    inline bool empty()
    {
        return size() == 0;
    }

    inline int push_back(const T &el)
    {
assert(&m_data == &m_datagrow.m_array);
            return m_datagrow.fast_add(el);
    }
    inline int push_back(const DynArray1D<T> &el)
    {
        int res =-1;
        for (int i = 0; i < el.size(); i++)   {

            res = m_datagrow.fast_add(el[i]);
        }
        return res;
    }
    inline void setAndCreateElement(int i, const T& element)
    {
        while (i >= size()) {
            push_back(T());
        }
        m_data[i] = element;
    }
    inline T getElement(int i)
    {
        if (size() <= i) return NULL;
        return m_data[i];
    }

    inline void pop_back()
    {
            m_datagrow.reduceSize(1);
    }
    inline int grow()
    {

            return m_datagrow.fast_grow();
    }
    inline size_t size() const
    {
        return m_datagrow.fast_size();
    }
    inline size_t capacity() const
    {
        return m_data.size();
    }
    inline void clear() {
        m_datagrow.reset(m_maxsize);
    }
    inline iterator begin()
    {
        return m_data.begin();
    }
    inline iterator end()
    {
        return m_data.begin()+m_datagrow.size();
    }
    inline const_iterator begin() const
    {
        return m_data.begin();
    }
    inline const_iterator end()const
    {
        return m_data.begin()+m_datagrow.size();
    }


    inline iterator insertItem(int position, const value_type& val)
    {
        return insert(begin()+position, val);
    }
    inline iterator insert(iterator position, const value_type& val)
    {
         return insert(position, 1, val);
    }
    inline iterator insert(iterator position, size_t n, const value_type& val)
    {
        int index = position - begin();

        int oldsize = m_datagrow.size();
        for (int i = 0; i < n; i++) {
            m_datagrow.fast_grow();
        }
        for (int i = oldsize-1; i >= index; i--) {
            m_data[i+n] = m_data[i];
        }
        for (int i = 0; i < n; i++)
        {
            m_data[index+i] = val;
        }

        return position;
    }
    template<class InputIterator> inline void insert(iterator position, InputIterator first, InputIterator last)
    {
        int n = last-first;
        int index = position - begin();


        for (int i = 0; i < n; i++) {
            m_datagrow.fast_grow();
        }
        for (int i = m_datagrow.size()-1; i >= index; i--) {

            m_data[i+n] = m_data[i];
        }
        InputIterator it = first;
        for (int i = 0; i < n; i++)
        {
            m_data[index+i] = it;
            it++;
        }


    }
    inline iterator erase(iterator position)
    {
		
        erase(position, position+1);
		return position;

    }
    inline void removeItem(T el)
    {
        int i = indexOf(el);
        if (i == -1) {
            return;
        }
        deleteItem(i);
    }
    inline void deleteItem(int i)
    {

        erase(begin()+i);
    }
    inline void deleteItemFast(int i)
    {
        int s = size();
        if (s -1!= i)
        {
            T temp = m_data[i];
            m_data[i] = m_data[s-1];
            m_data[s-1] = temp;
        }
        erase(begin()+(s-1));
    }

    inline void erase(iterator first, iterator last)
    {
        int starterase = first-begin();
        int enderase = last-begin();
        assert(starterase >= 0);
        assert(enderase >= 0);
        assert(starterase <= m_datagrow.size());
        assert(enderase <= m_datagrow.size());
        assert(enderase -starterase >= 0);
        //int c = enderase -starterase;
        int shift =enderase-starterase;
        for (int i = enderase; i < m_datagrow.size(); i++) {
            m_data[i-shift] = m_data[i];
        }
        m_datagrow.reduceSize(shift);
    }
    inline reference operator[] (size_t n)
    {
        assert(n >= 0 && n < m_datagrow.size());
        return m_data[n];
    }

    inline const_reference operator[] (size_t n) const
    {
        assert(n >= 0 && n < m_datagrow.size());
        return m_data[n];
    }

    inline reference at(size_t n)
    {
       assert(n >= 0 && n < m_datagrow.size());
       return m_data[n];
    }

    inline const_reference at(size_t n) const
    {
        assert(n >= 0 && n < m_datagrow.size());
        return m_data[n];
    }

    inline reference getAndExpand(size_t i)
    {
        while (m_datagrow.size() <= i)
        {
            push_back(T());
        }
        return at(i);

    }

    inline reference front()
    {

        assert(m_datagrow.size() != 0);
        return m_data[0];
    }

    inline const_reference front() const
    {
        assert(m_datagrow.size() != 0);
        return m_data[0];
    }

    inline reference back()
    {

        assert(m_datagrow.size() != 0);
        return m_data[m_datagrow.fast_size()-1];
    }

    inline const_reference back() const
    {

        assert(m_datagrow.size() != 0);
        return m_data[m_datagrow.fast_size()-1];
    }
    inline void swap(DynArray1D& other)
    {
      m_data.swap(other.m_data);
      m_datagrow.swap(other.m_datagrow);
    }
    void move(int d1, int d2)
    {

        assert(d1 >= 0 &&d1 < m_datagrow.size());
        assert(d2 >= 0 &&d2 < m_datagrow.size());
        T d = m_data[d1];
        m_data[d1] =m_data[d2];
        m_data[d2] = d;
    }
    int indexOf(const T & data)
    {
        for (int i = 0; i < m_datagrow.fast_size(); i++) {
            if (m_data[i] == data) return i;

        }
        return -1;
    }
    bool contains(const T& data)
    {
        return indexOf(data) != -1;
    }

    inline bool operator==(const DynArray1D& other) const
    {
        for (int i = 0; i< size();i++){
        if (!(at(i) == other.at(i))) return false;
        }
        return true;
    }
    DynArray1D& operator=(const DynArray1D& other)
    {

        clear();
        for (int i = 0; i < other.size(); i++) {
            push_back(other[i]);
        }
        return *this;
    }
};

#ifndef NDEBUG
//#define RLEDEBUG
#endif

template<class T, class CountType> struct RLEElement
{
    T m_value;
    CountType m_start;
    CountType m_count;
};

template<class T, class CountType=int, class Container=DynArray1D<RLEElement<T, CountType> >, class DebugContainer= DynArray1D<T> > class RLEArray1D {
public:
    Container m_rledata;
#ifdef RLEDEBUG
    DebugContainer m_debugdata;
#endif
    int m_count;

public:
    typedef T value_type;
    typedef T& reference;
    typedef const T& const_reference;
    typedef RLEElement<T, CountType> LocalRLEElement;
    inline LocalRLEElement createElement(const T& v, CountType s, CountType c)
    {
        LocalRLEElement res;
        res.m_value = v;
        res.m_count = c;
        res.m_start = s;
        return res;
    }

    inline Container& getContainer()
    {
        return m_rledata;
    }
    inline RLEArray1D& operator=(const RLEArray1D& it)
    {
        m_count = it.m_count;
        m_rledata = it.m_rledata;
#ifdef RLEDEBUG
       m_debugdata = it.m_debugdata;
#endif
        return *this;
    }
    inline bool operator==(const RLEArray1D& it) {
        return m_count == it.m_count && m_rledata == it.m_rledata;
    }

    void updateListDataAfterManualUpdate()
    {
        m_count = 0;
        if (m_rledata.size() != 0) {
           LocalRLEElement & l = m_rledata.back();
           m_count = l.m_start+l.m_count;
        }
#ifdef RLEDEBUG
        m_debugdata.clear();
        for (int i = 0; i < m_rledata.size(); i++) {
            LocalRLEElement & l = m_rledata[i];
            m_debugdata.insert(m_debugdata.begin()+ m_debugdata.size(), l.m_count, l.m_value);

        }
        checkData();
#endif
    }

    struct iterator
    {
        RLEArray1D* m_parent;
        int m_position;
        int m_rleelement;
        int m_inrleposition;
        inline iterator()
        {
            m_parent = NULL;
            m_position = 0;
            m_rleelement = 0;
            m_inrleposition=0;
        }
        inline iterator(const iterator& it)
        {
            m_parent = it.m_parent;
            m_position = it.m_position;
            m_rleelement = it.m_rleelement;
            m_inrleposition = it.m_inrleposition;
        }

        inline ~iterator()
        {

        }
        inline iterator& operator=(const iterator& it)
        {
            m_parent = it.m_parent;
            m_position = it.m_position;
            m_rleelement = it.m_rleelement;
            m_inrleposition = it.m_inrleposition;
            return *this;
        }
        inline bool operator==(const iterator& it)
        {
            return (m_parent == it.m_parent) && (m_position == it.m_position);
        }
        inline bool operator!=(const iterator& it)
        {
            return !((m_parent == it.m_parent) && (m_position == it.m_position));
        }
        inline int operator-(const iterator& it)
        {
            assert(m_parent == it.m_parent);
            return m_position - it.m_position;
        }

        inline iterator& operator++()
        {
            m_inrleposition++;
            if (m_rleelement < m_parent->m_rledata.size() &&m_inrleposition >= m_parent->m_rledata[m_rleelement].m_count) {
                m_rleelement++;

                m_inrleposition = 0;

            }


            m_position++;
            assert( (m_rleelement < m_parent->m_rledata.size() && m_inrleposition <m_parent->m_rledata[m_rleelement].m_count) ||
                    (m_rleelement == m_parent->m_rledata.size() && m_position == m_parent->size()));

            assert((m_rleelement >= m_parent->m_rledata.size() ||m_parent->m_rledata[m_rleelement].m_start + m_inrleposition == m_position));
            return this;

        }

        inline iterator operator++(int)
        {
            m_inrleposition++;
            if (m_rleelement < m_parent->m_rledata.size() &&m_inrleposition >= m_parent->m_rledata[m_rleelement].m_count) {
                m_rleelement++;

                m_inrleposition = 0;

            }


            m_position++;
            assert( (m_rleelement < m_parent->m_rledata.size() && m_inrleposition <m_parent->m_rledata[m_rleelement].m_count) ||
                    (m_rleelement == m_parent->m_rledata.size() && m_position == m_parent->size()));

            assert((m_rleelement >= m_parent->m_rledata.size() ||m_parent->m_rledata[m_rleelement].m_start + m_inrleposition == m_position));
            return *this;
        }
        inline iterator& operator--()
        {
            m_inrleposition--;
            if (m_inrleposition < 0) {
                m_rleelement--;

                m_inrleposition = m_parent->m_rledata[m_rleelement].m_count-1;
                assert(m_inrleposition >= 0);

            }
            assert(m_inrleposition >= 0);
            m_position--;

            assert( (m_rleelement < m_parent->m_rledata.size() && m_inrleposition <m_parent->m_rledata[m_rleelement].m_count) ||
                    (m_rleelement == m_parent->m_rledata.size() && m_position == m_parent->size()));

            assert((m_rleelement >= m_parent->m_rledata.size() ||m_parent->m_rledata[m_rleelement].m_start + m_inrleposition == m_position));
            return *this;

        }

        inline iterator& operator+=(int c)
        {


            m_inrleposition+=c;
            while (m_rleelement < m_parent->m_rledata.size() && m_inrleposition >= m_parent->m_rledata[m_rleelement].m_count) {
                 m_inrleposition -= m_parent->m_rledata[m_rleelement].m_count;
                m_rleelement++;



            }
            m_position+=c;
            assert( (m_rleelement < m_parent->m_rledata.size() && m_inrleposition <m_parent->m_rledata[m_rleelement].m_count) ||
                    (m_rleelement == m_parent->m_rledata.size() && m_position == m_parent->size()));

            assert((m_rleelement >= m_parent->m_rledata.size() ||m_parent->m_rledata[m_rleelement].m_start + m_inrleposition == m_position));
            return *this;
        }
        inline iterator operator+(int c)
        {

            iterator it = *this;
            it += c;
            return it;
        }
        inline iterator& operator-=(int c)
        {

            m_inrleposition-= c;
            while (m_inrleposition < 0) {
                 m_inrleposition += m_parent->m_rledata[m_rleelement-1].m_count;
                m_rleelement++;



            }
            assert(m_inrleposition >= 0);
              m_position+=c;
              assert( (m_rleelement < m_parent->m_rledata.size() && m_inrleposition <m_parent->m_rledata[m_rleelement].m_count) ||
                      (m_rleelement == m_parent->m_rledata.size() && m_position == m_parent->size()));

            assert(m_parent->m_rledata[m_rleelement].m_start + m_inrleposition == m_position);
            return *this;
        }
        inline iterator& operator--(int)
        {
            m_inrleposition--;
            if (m_inrleposition < 0) {
                m_rleelement--;

                m_inrleposition = m_parent->m_rledata[m_rleelement].m_count-1;
                assert(m_inrleposition >= 0);

            }
            assert(m_inrleposition >= 0);
              m_position--;
              assert( (m_rleelement < m_parent->m_rledata.size() && m_inrleposition <m_parent->m_rledata[m_rleelement].m_count) ||
                      (m_rleelement == m_parent->m_rledata.size() && m_position == m_parent->size()));
            assert(m_parent->m_rledata[m_rleelement].m_start + m_inrleposition == m_position);
            return *this;


        }

        inline reference operator*() const
        {
            assert(m_parent->m_rledata[m_rleelement].m_value == m_parent->at(m_position));
            return m_parent->m_rledata[m_rleelement].m_value;
        }
    };


    inline RLEArray1D()
    {
        m_count = 0;
    }



#ifdef RLEDEBUG
     void checkData()
     {
         int currentindex = 0;
         for (int i = 0; i < m_rledata.size(); i++) {
             LocalRLEElement& d = m_rledata[i];
             LocalRLEElement* d2 = NULL;
             if (i != 0) {
                 d2 = &m_rledata[i-1];
                 assert(d2->m_count+d2->m_start == currentindex);
                 assert(d2->m_count+d2->m_start  == d.m_start);
                 assert(!(d2->m_value == d.m_value));

             }
             assert(d.m_start == currentindex);
             assert(d.m_count > 0);
             if (i != 0) {
                 LocalRLEElement& d2 = m_rledata[i-1];
                 assert(d.m_value != d2.m_value);
             }
             for (int j = 0; j < d.m_count; j++) {
                 assert(m_debugdata.at(currentindex) == d.m_value);
                 currentindex++;
             }
         }
         assert(currentindex == m_debugdata.size());
         assert(m_count == m_debugdata.size());
     }

#endif

    inline int push_back(const T &el)
    {
        if (m_rledata.size() == 0 || m_rledata.back().m_value != el)
        {
            LocalRLEElement value;
            value.m_count = 1;
            value.m_value = el;
            value.m_start = m_count;
            m_rledata.push_back(value);
        } else {
            m_rledata.back().m_count++;
        }

        m_count++;

#ifdef RLEDEBUG
      m_debugdata.push_back(el);
      //checkData();
#endif

            return m_count-1;
    }
    inline void pop_back()
    {


        if (m_rledata.last().m_count == 1)
        {

            m_rledata.pop_back();
        } else {
            m_rledata.last().m_count--;
        }
        m_count--;
#ifdef RLEDEBUG
       m_debugdata.pop_back();
      checkData();
#endif

    }
    inline size_t size() const
    {
        return m_count;
    }
    inline size_t capacity() const
    {
        return m_count;
    }
    inline void clear() {
        m_rledata.clear();
        m_count = 0;
#ifdef RLEDEBUG
       m_debugdata.clear();
      checkData();
#endif

    }
    inline iterator begin()
    {
        iterator iter;
        iter.m_parent = this;iter.m_inrleposition = 0; iter.m_position = 0; iter.m_rleelement = 0;
        return iter;
    }
    inline iterator end()
    {
        iterator iter;
        iter.m_parent = this;iter.m_inrleposition = 0; iter.m_position = m_count; iter.m_rleelement = m_rledata.size();
        return iter;
        //return m_data.begin()+m_datagrow.size();
    }

    inline iterator insertItem(int position, const value_type& val)
    {
        return insert(begin()+position, val);
    }
    inline iterator insert(iterator position, const value_type& val)
    {
        return insert(position, (size_t)1, val);
    }
    void expandRLEElement(int rleelement, int addsize)
    {
        m_rledata[rleelement].m_count+= addsize;

        for (int i = rleelement+1; i < m_rledata.size(); i++) {
            m_rledata[i].m_start += addsize;
       }

        m_count+= addsize;
    }

    iterator insert(iterator position, size_t n, const value_type& val)
    {
       // LocalRLEElement createElement(const T& v, int s, int c)
         if (n == 0) return position;
         assert((begin()+position.m_position) == position);
        std::pair<int, int> res;
        res.first = position.m_rleelement;
        res.second = position.m_inrleposition;
        if (res.first == m_rledata.size()) {
            if (m_rledata.size() == 0 || m_rledata[res.first-1].m_value != val) {
                m_rledata.push_back(createElement(val, m_count, n));
            } else {
                m_rledata[res.first-1].m_count+= n;
            }
            m_count+= n;
        } else {
            expandRLEElement(res.first, n);
            LocalRLEElement& el = m_rledata[res.first];
            if (el.m_value == val) {
                // done
            } else {
                if (el.m_start == position.m_position) {
                    // 2 segments, insert at front
                    //change first because it will invalid the reference
                    //int oldstart = el.m_start;
                    el.m_start += n;
                    el.m_count -= n;
                    if (res.first > 0 && m_rledata[res.first-1].m_value == val) {
                        m_rledata[res.first-1].m_count += n;
                        position.m_rleelement--;
                        assert(res.first-1 == position.m_rleelement);
                        position.m_inrleposition = position.m_position-m_rledata[position.m_rleelement].m_start;
                    } else {
                    m_rledata.insert(m_rledata.begin()+res.first, createElement(val,  position.m_position, n));

                    }
                } else if (el.m_start+el.m_count == position.m_position+n){
                    //2 segments, insert at end
                    el.m_count -= n;
                    if (res.first +1 < m_rledata.size() && m_rledata[res.first-1].m_value == val) {
                       m_rledata[res.first+1].m_count += n;
                       m_rledata[res.first+1].m_start -= n;
                    } else {
                       m_rledata.insert(m_rledata.begin()+(res.first+1), createElement(val,  position.m_position, n));

                    }
                    position.m_rleelement++;
                    position.m_inrleposition = position.m_position-m_rledata[position.m_rleelement].m_start;
                } else {
                    //3 segments
                    int oldendposition = el.m_start+el.m_count;
                    //int oldstartposition = el.m_start;
                    el.m_count = position.m_position-el.m_start;

                    T oldvalue = el.m_value;

                    m_rledata.insert(m_rledata.begin()+(res.first+1), 2, createElement(val,  position.m_position, n));

                    LocalRLEElement& afterel = m_rledata[res.first+2];
                    afterel.m_start = position.m_position+n;
                    afterel.m_count = oldendposition-afterel.m_start;
                    afterel.m_value = oldvalue;
                    position.m_rleelement++;
                    position.m_inrleposition = 0;
                }
            }


        }
#ifdef RLEDEBUG
       m_debugdata.insert(m_debugdata.begin()+position.m_position, n, val);
      checkData();
#endif
      assert((begin()+position.m_position) == position);
      return position;

    }

    template<class InputIterator> void insert(iterator position, InputIterator first, InputIterator last)
    {
        if (first == last) return;
        T fvalue = *first;
        int currentcount=1;
        iterator pos = position;
      //  int localpos = 0;
        for (InputIterator it = first+1; it != last; it++) {
            T nvalue = *it;
            if (nvalue == fvalue) {
                currentcount++;
            } else {
               // assert((position+localpos) == pos);
                pos = insert(pos, currentcount, fvalue);
                pos += currentcount;
                currentcount = 1;
                fvalue = nvalue;
            }
        }
        pos = insert(pos, currentcount, fvalue);

    }


    inline iterator erase(iterator position)
    {
        erase(position, position+1);
        return position;
    }
    inline void deleteItem(int i)
    {

        erase(begin()+i);
    }

    inline void setData(int i, const value_type& val)
    {
    deleteItem(i);
    insertItem(i, val);
    }

    inline void erase(iterator first, iterator last)
    {
        if (first == last)return;

        if (first.m_rleelement == last.m_rleelement) {
            expandRLEElement(first.m_rleelement, -(last.m_position-first.m_position));
            if (m_rledata[first.m_rleelement].m_count == 0) {
                m_rledata.erase(m_rledata.begin()+first.m_rleelement);

            }

        } else if (last.m_rleelement == m_rledata.size()){
            LocalRLEElement& firstel = m_rledata[first.m_rleelement];

            int shift = 1;
            if (first.m_inrleposition == 0) {
                shift = 0;
            } else {
                firstel.m_count = first.m_inrleposition;
            };

            m_rledata.erase(m_rledata.begin()+(first.m_rleelement+shift), m_rledata.end());


            m_count -= (last.m_position-first.m_position);
        } else {

            LocalRLEElement& firstel = m_rledata[first.m_rleelement];
            LocalRLEElement& lastel = m_rledata[last.m_rleelement];

            if (firstel.m_value == lastel.m_value)
            {
                 int lastpos = lastel.m_start +lastel.m_count;
                 m_rledata.erase(m_rledata.begin()+(first.m_rleelement+1), m_rledata.begin()+(last.m_rleelement+1));


                 LocalRLEElement& firstel2 = m_rledata[first.m_rleelement];
                 firstel2.m_count = lastpos-firstel2.m_start;

                 expandRLEElement(first.m_rleelement, -(last.m_position-first.m_position));

                 if (m_rledata[first.m_rleelement].m_count == 0) {
                     m_rledata.erase(m_rledata.begin()+first.m_rleelement);

                 }
            } else {
                m_rledata.erase(m_rledata.begin()+(first.m_rleelement+1), m_rledata.begin()+(last.m_rleelement));
                 LocalRLEElement& firstel2 = m_rledata[first.m_rleelement];
                 LocalRLEElement& lastel2 = m_rledata[first.m_rleelement+1];
                 int lastend = lastel2.m_start+lastel2.m_count;
                 lastel2.m_start = last.m_position;
                 lastel2.m_count = lastend-lastel2.m_start;
                 firstel2.m_count = lastel2.m_start-firstel2.m_start;

                 expandRLEElement(first.m_rleelement, -(last.m_position-first.m_position));


                 if (m_rledata[first.m_rleelement].m_count == 0) {
                     m_rledata.erase(m_rledata.begin()+first.m_rleelement);

                 }
            }


        }
        if (m_rledata.size() > first.m_rleelement && first.m_rleelement > 0
                && m_rledata[first.m_rleelement].m_value == m_rledata[first.m_rleelement-1].m_value) {
           m_rledata[first.m_rleelement-1].m_count += m_rledata[first.m_rleelement].m_count;
            m_rledata.erase(m_rledata.begin()+first.m_rleelement);

        }


#ifdef RLEDEBUG
       m_debugdata.erase(m_debugdata.begin()+first.m_position, m_debugdata.begin()+last.m_position);
      checkData();
#endif
    }

    inline std::pair<int, int> positionToRLE(int position) const
    {

        assert(position >= 0 && position < m_count);
        int minelement = 0;
        int maxelement = m_rledata.size();

        while (minelement < maxelement-1) {

            if (minelement+2 == maxelement) {

                const LocalRLEElement& minv = m_rledata[minelement];
                //LocalRLEElement& maxv = m_rledata[maxelement-1];
                if (minv.m_start+minv.m_count > position) {
                    break;
                } else {
                    minelement = maxelement-1;
                }
                break;
            }
            int middle = (minelement+maxelement-1) >> 1;

            const LocalRLEElement& mid = m_rledata[middle];
            if (mid.m_start<= position) {
                minelement = middle;
            }else {
                maxelement= middle+1;
                assert(maxelement <= m_count);
            }


        }
        assert(m_rledata[minelement].m_start <= position);
        assert(position < m_rledata[minelement].m_start+m_rledata[minelement].m_count);
        return std::make_pair(minelement, position-m_rledata[minelement].m_start);
    }

    /*inline reference operator[] (size_t n)
    {
        assert(n >= 0 && n < m_datagrow.size());
        return m_data[n];
    }*/

    inline const_reference operator[] (size_t n) const
    {
        std::pair<int, int> res = positionToRLE(n);

        return m_rledata[res.first].m_value;
    }
/*
    inline reference at(size_t n)
    {
       assert(n >= 0 && n < m_datagrow.size());
       return m_data[n];
    }
*/
    inline const_reference at(size_t n) const
    {
        std::pair<int, int> res = positionToRLE(n);

        return m_rledata[res.first].m_value;

    }

    inline reference front()
    {

        assert(m_rledata.size() != 0);
        return m_rledata[0].m_value;
    }

    inline const_reference front() const
    {
        assert(m_rledata.size() != 0);
        return m_rledata[0].m_value;
    }

    inline reference back()
    {

        assert(m_rledata.size() != 0);
        return m_rledata[m_rledata.size()-1].m_value;
    }

    inline const_reference back() const
    {

        assert(m_rledata.size() != 0);
        return m_rledata[m_rledata.size()-1].m_value;
    }
   /* inline void swap(DynArray1D& other)
    {
      m_rledata.swap(other.m_rledata);


#ifdef RLEDEBUG
      m_debugdata.swap(other.m_debugdata);
#endif
      std::swap(m_count, other.m_count);
    }*/
};






template<class T> struct SRLEBaseSlice
        {
            RLEArray1D<T, short > m_slice;
        };

template<class T> struct SSliceDataContainer : public ReferenceUtiOB
{
    CIntBox m_box;


    Array2D< SRLEBaseSlice<T> > m_data;
    SSliceDataContainer()
    {
        m_box = CIntBox(0,0,0,0,0,0);

    }

    SSliceDataContainer(const CIVector3& v)
    {
        m_box = CIntBox(0,0,0,0,0,0);
        setSize(CIntBox(CIVector3(0,0,0), v));

    }
    SSliceDataContainer(Array3DYContainer<T>* input, const CIntBox& box)
    {
        m_box = CIntBox(0,0,0,0,0,0);
        setSize(m_box);
        int h = m_box.height();
        for (int i = 0; i < m_data.width(); i++) {
                        for (int j = 0; j < m_data.height(); j++) {
                SRLEBaseSlice<T>& s = m_data(i,j);
                s.m_slice.clear();
                for (int t = 0; t< h; t++) {
                    s.m_slice.push_back(input->getData(CIVector3(i, t, j)+m_box.m_min));

                }
            }
        }
    }

    void setSize(const CIntBox& b)
    {
        m_box = b;
        m_data.setSize(m_box.width(), m_box.depth());
        for (int i = 0; i < m_data.width(); i++) {
			for (int j = 0; j < m_data.height(); j++) {
                SRLEBaseSlice<T>& s = m_data(i,j);
                s.m_slice.insert(s.m_slice.begin(), m_box.height(), T());
            }
        }
    }
    void changeOriginTo(const CIVector3& v)
    {
        CIVector3 move = v-m_box.m_min;
        m_box.move(move);
    }

    inline const CIntBox &box() const
    {
        return m_box;
    }
    inline int width() const
    {
        return m_box.width();
    }
    inline int height() const
    {
        return m_box.height();
    }

    inline int depth() const
    {
        return m_box.depth();
    }
    inline SRLEBaseSlice<T>& slice(int worldx, int worldz)
    {
        return m_data(worldx-m_box.m_min.x, worldz-m_box.m_min.z);
    }
    inline const T& get(const CIVector3& v) const
    {
        CIVector3 t = v-m_box.m_min;
        return m_data(t.x, t.z).m_slice.at(t.y);
    }
    inline void set(const CIVector3&v, const T& value)
    {
        CIVector3 t = v-m_box.m_min;

        SRLEBaseSlice<T>& s = m_data(t.x, t.z);
        s.m_slice.erase(s.m_slice.begin()+t.y);
        s.m_slice.insert(s.m_slice.begin()+t.y, value);
    }
   void fill(const T& v)
    {
        for (int i = 0; i < m_data.width(); i++) {
			for (int j = 0; j < m_data.height(); j++) {
                SRLEBaseSlice<T>& s = m_data(i,j);
                s.m_slice.clear();
                s.m_slice.insert(s.m_slice.begin(), m_box.height(), v);
            }
        }
    }
   SSliceDataContainer<T>* createCopy()
   {
	   SSliceDataContainer<T>* t = new SSliceDataContainer<T>();
	   t->setSize(m_box);
	   for (int i = 0; i < m_data.width(); i++) {
		   for (int j = 0; j < m_data.height(); j++) {
			   SRLEBaseSlice<T>& s = m_data(i, j);
                           SRLEBaseSlice<T>& target = t->m_data(i, j);
			   s.m_slice.clear();
			   target.m_slice.insert(target.m_slice.begin(), s.m_slice.begin(), s.m_slice.end());
		   }
	   }
   }
   Array3DYContainer<T>* toArray3D()
   {
	   Array3DYContainer<T>* td = new Array3DYContainer<T>(m_box);

	   for (int i = 0; i < m_data.width(); i++) {
		   for (int j = 0; j < m_data.height(); j++) {
			   SRLEBaseSlice<T>& s = m_data(i, j);
			   //s.m_slice.clear();
			   //s.m_slice.insert(s.m_slice.begin(), m_box.height(), v);
			   DynArray1D<RLEElement<T, short> >& c = s.m_slice.getContainer();

			   for (int ty = 0; ty < c.size(); ty++){
				   RLEElement<T, short>&el = c[ty];
				   int m = el.m_start + el.m_count;
				   for (int yvalue = el.m_start; yvalue < m; yvalue++) {
					   td->m_array(i, yvalue, j) = el.m_value;
				   }
			   }
		   }
	   }
	   return td;
   }
};

template<class T, int baseversion> class SCachedElement : public CUtiOB
{
public:
    typedef T value_type;
    enum {version = baseversion};
    struct StaticContainer
    {
        CIndexList<SCachedElement, 0> m_currentactive;
        int m_limit;
        StaticContainer()
        {
            m_limit = 1000;
        }
        void peek(SCachedElement* el)
		{
                        assert(CThread::isMainThread());
            if (el->clist_indexcontroller[0].inList()) {
                m_currentactive.remove(el);
                m_currentactive.push_back(el);
            }else {

                m_currentactive.push_back(el);
            }
            if (m_currentactive.size() >= m_limit)
            {
                SCachedElement* f = m_currentactive.get(0);
                assert(f != el);
                f->unload();
assert(m_currentactive.size() <= m_limit);
            }
        }
        void deleted(SCachedElement* el)
        {
                        assert(CThread::isMainThread());
            if (el->clist_indexcontroller[0].inList()) {
                m_currentactive.remove(el);
            };
        }
    };
    static StaticContainer m_allelements;

    CList_IndexController clist_indexcontroller[1];
    uti_refptr<CMemoryStream> m_zipdata;
    uti_refptr<T> m_data;
    bool m_needwrite;
    UTI_SYNC_DATA
    SCachedElement()
    {
        m_data = NULL;
        m_needwrite = true;

    }
    ~SCachedElement()
    {
		
        m_allelements.deleted(this);
    }

    bool isLoaded()
    {
        if (m_data.get() == NULL)
        {
            return false;
        }
        return true;
    }
    bool hasData()
    {
        return m_data.get()!= NULL || m_zipdata.get() != NULL;
    }

    T* getData()
    {
         CHECKEXISTANCE;UTI_SYNC_CALL
        if (m_data.get() == NULL) {

            m_allelements.peek(this);
            if (m_zipdata.get() == NULL) {
                m_data = new T();
            } else {
                load();
            }

        }
        m_needwrite = true;
        return m_data.get();

    }
    T* getReadData()
    { CHECKEXISTANCE;UTI_SYNC_CALL
        if (m_data.get() == NULL) {

            m_allelements.peek(this);
            if (m_zipdata.get() == NULL) {
                m_data = new T();
            } else {
                load();
            }

        }
        return m_data.get();

    }
	T* getDataManual()
        {UTI_SYNC_CALL
		if (m_data.get() == NULL) {

			//m_allelements.peek(this);
			if (m_zipdata.get() == NULL) {
				m_data = new T();
			}
			else {
				load();
			}

		}
		m_needwrite = true;
		return m_data.get();

	}
	T* getReadDataManual()
        {UTI_SYNC_CALL
		if (m_data.get() == NULL) {

			//m_allelements.peek(this);
			if (m_zipdata.get() == NULL) {
				m_data = new T();
			}
			else {
				load();
			}

		}
		return m_data.get();

	}
    void load()
    { CHECKEXISTANCE;
        m_needwrite = false;
        m_data = new T();
        assert(m_zipdata.get() != NULL);
        m_zipdata->setAllowWrite(false);
        m_zipdata->setPosition(0, CStream::START);
        uti_refptr<CZipMemoryStream> compressed(new CZipMemoryStream(m_zipdata.get()));
//HResult Read(CStream & Stream, int Flags, int version);
//            HResult Write(CStream & Stream, int Flags, int version);
        compressed->beginRead();
        int v = compressed->getInt32();
        m_data->Read(*compressed.get(), 0, v);
        compressed->endRead();
    }

    void unload()
    { CHECKEXISTANCE;
        assert(m_data.get() != NULL);

        if (m_needwrite) {

            m_zipdata = new CMemoryStream(true);
            uti_refptr<CZipMemoryStream> compressed(new CZipMemoryStream(m_zipdata.get()));
//HResult Read(CStream & Stream, int Flags, int version);
//            HResult Write(CStream & Stream, int Flags, int version);
            compressed->beginWrite();
            compressed->writeInt32(baseversion);
            m_data->Write(*compressed.get(), 0, baseversion);
            compressed->endWrite();

        }
        //assert(m_zipdata.get() != NULL);
        m_data = NULL;

        m_allelements.deleted(this);
        m_needwrite = false;

    }
	void unloadManual ()
        { CHECKEXISTANCE;
            UTI_SYNC_CALL
		assert(m_data.get() != NULL);

		if (m_needwrite) {

			m_zipdata = new CMemoryStream(true);
			uti_refptr<CZipMemoryStream> compressed(new CZipMemoryStream(m_zipdata.get()));
			//HResult Read(CStream & Stream, int Flags, int version);
			//            HResult Write(CStream & Stream, int Flags, int version);
			compressed->beginWrite();
			compressed->writeInt32(baseversion);
			m_data->Write(*compressed.get(), 0, baseversion);
			compressed->endWrite();

		}
		assert(m_zipdata.get() != NULL);
		m_data = NULL;

                //m_allelements.deleted(this);
		m_needwrite = false;

	}

    HResult Read(CStream & Stream, int Flags, int version)
    { CHECKEXISTANCE;
UTI_SYNC_CALL
        m_needwrite = false;
        if (m_data.get() != NULL) {
        unload();
        }
        m_zipdata = new CMemoryStream(true);
        m_zipdata->ReadFromStreamSize(&Stream);
        m_needwrite = false;


    return 0;
    }
    HResult Write(CStream & Stream, int Flags, int version)
    {
        CHECKEXISTANCE;
UTI_SYNC_CALL
        if (m_data.get() != NULL) {


        unload();
        }
        if (m_zipdata.get() == NULL) {
            getData();
            unload();
        }
        m_zipdata->WriteToStreamSize(&Stream);
    return 0;
    }
};

template <class T, int baseversion> typename SCachedElement<T, baseversion>::StaticContainer SCachedElement<T, baseversion>::m_allelements;

template<class T> struct SCachedElementLock
{
	bool m_locked;
    T* m_cachedarray;
    SCachedElementLock(T* ziparray=NULL)
    {
        m_cachedarray = ziparray;
        m_locked = false;
    }
    void setCachedData(T* ziparray)
    {
        assert(m_locked == false);
        m_cachedarray = ziparray;
    }

    typename T::value_type* lockRead()
    {
        m_locked = true;
       return  m_cachedarray->getReadDataManual();
    }

    typename T::value_type* lockWrite()
    {
        m_locked = true;

        return  m_cachedarray->getDataManual();
    }

    void unlock()
    {
        assert(m_locked == true);
        m_locked = false;
        m_cachedarray->unloadManual();
    }

    ~SCachedElementLock()
    {
        if (m_locked) {
            assert(false);
            unlock();
        }
    }

};

template<int multiplier=1, int divider=1> struct uti_ratio
{
    enum {MULTIPLIER=multiplier, DIVIDER=divider};
    static float toFloat()
    {
        return float(multiplier)/float(divider);
    }
    static double toDouble()
    {
        return double(multiplier)/double(divider);
    }
    static utifloat toFloatingPoint()
    {
        return utifloat(multiplier)/utifloat(divider);
    }
    static utifloat toFloatingPointInv()
    {
        return utifloat(divider)/utifloat(multiplier);
    }

};

template<int xsegsize=16, int ysegsize=16, class floatxsize=uti_ratio<1,1>, class floatysize=uti_ratio<1,1> > struct SSegmentCalculator
{
    inline SSegmentCalculator()
    {

    }

    inline static CIVector2 pixelToSegment(const CIVector2& input)
    {
        CIVector2 res, mod;
        uti_moddiv(input, CIVector2(xsegsize, ysegsize), mod, res);
        return res;
    }
    inline static CIntRect pixelToSegmentRect(const CIntRect& r)
    {

        CIntRect res;
        res.m_min =pixelToSegment(r.m_min);
        CIVector2 endmod;
        pixelToModAndSegment(r.m_max, endmod, res.m_max);

        if (endmod.x != 0) res.m_max.x++;
        if (endmod.y != 0) res.m_max.y++;
        return res;
    }

    inline static CIVector2 pixelToModSegment(const CIVector2& input)

    {
        CIVector2 res, mod;
        uti_moddiv(input, CIVector2(xsegsize, ysegsize), mod, res);
        return mod;
    }
    static inline void pixelToModAndSegment(const CIVector2& input, CIVector2 &mod, CIVector2 &res)
    {
        uti_moddiv(input, CIVector2(xsegsize, ysegsize), mod, res);
    }
    static inline CIntRect segmentToPixelRect(const CIVector2& segment)
    {
        return CIntRect(segment.x*xsegsize, segment.y*ysegsize, (segment.x+1)*xsegsize, (segment.y+1)*ysegsize);
    }
    static inline CIntRect segmentRectToPixelRect(const CIntRect& segment)
    {
        return CIntRect(segment.m_min.x*xsegsize, segment.m_min.y*ysegsize, (segment.m_max.x)*xsegsize, (segment.m_max.y+1)*ysegsize);
    }
    static inline CRect pixelToFloatRect(const CIVector2& pixel)
    {
        return CRect(pixel.x * floatxsize::toFloatingPoint(), pixel.y * floatysize::toFloatingPoint(),
                     (pixel.x+1) * floatxsize::toFloatingPoint(), (pixel.y+1) * floatysize::toFloatingPoint());
    }
    static inline CRect pixelRectToFloatRect(const CIntRect& pixel)
    {
        return CRect(pixel.m_min.x * floatxsize::toFloatingPoint(), pixel.m_min.y * floatysize::toFloatingPoint(),
                     (pixel.m_max.x) * floatxsize::toFloatingPoint(), (pixel.m_max.y) * floatysize::toFloatingPoint());
    }
    static inline CRect segmentToFloatRect(const CIVector2& seg)
    {
        return CRect(seg.x*xsegsize * floatxsize::toFloatingPoint(), seg.y*ysegsize * floatysize::toFloatingPoint(),
                     (seg.x+1)*xsegsize * floatxsize::toFloatingPoint(), (seg.y+1)*ysegsize * floatysize::toFloatingPoint());
    }


    static inline CRect segmentRectToFloatRect(const CIntRect& seg)
    {
        return CRect(seg.m_min.x*xsegsize * floatxsize::toFloatingPoint(), seg.m_min.y*ysegsize * floatysize::toFloatingPoint(),
                     (seg.m_max.x)*xsegsize * floatxsize::toFloatingPoint(), (seg.m_max.y)*ysegsize * floatysize::toFloatingPoint());
    }

    static inline CIVector2 floatToPixel(const CVector2& position)
    {
        CIVector2 res;

        res.x = fasterfloor(position.x*floatxsize::toFloatingPointInv());
        res.y = fasterfloor(position.y*floatysize::toFloatingPointInv());
        return res;
    }
    static inline CIVector2 floatToSegment(const CVector2& position)
    {
        CIVector2 res;

        res.x = fasterfloor(position.x*floatxsize::toFloatingPointInv());
        res.y = fasterfloor(position.y*floatysize::toFloatingPointInv());
        return pixelToSegment(res);
    }
    static inline CIntRect floatRectToPixelRect(const CRect& position)
    {
        CIntRect res;
        res.m_min.x = fasterfloor(position.m_min.x*floatxsize::toFloatingPointInv());
        res.m_min.y = fasterfloor(position.m_min.y*floatysize::toFloatingPointInv());
        res.m_max.x = fasterfloor(position.m_max.x*floatxsize::toFloatingPointInv()-0.00001f)+1;
        res.m_max.y = fasterfloor(position.m_max.y*floatysize::toFloatingPointInv()-0.00001f)+1;
        return res;
    }
};


template<int xsegsize = 16, int ysegsize = 16, int zsegsize = 16, class floatxsize = uti_ratio<1, 1>, class floatysize = uti_ratio<1, 1>, class floatzsize = uti_ratio<1, 1> > struct SSegmentCalculator3D
{
	inline SSegmentCalculator3D()
	{

	}

	inline static CIVector3 pixelToSegment(const CIVector3& input)
	{
		CIVector3 res, mod;
		uti_moddiv(input, CIVector3(xsegsize, ysegsize, zsegsize), mod, res);
		return res;
	}
	inline static CIntBox pixelToSegmentRect(const CIntBox& r)
	{

		CIntBox res;
		res.m_min = pixelToSegment(r.m_min);
		CIVector3 endmod;
		pixelToModAndSegment(r.m_max, endmod, res.m_max);

                if (endmod.x != 0) res.m_max.x++;
                if (endmod.y != 0) res.m_max.y++;
                if (endmod.z != 0) res.m_max.z++;
		return res;
	}

	inline static CIVector3 pixelToModSegment(const CIVector3& input)

	{
		CIVector3 res, mod;
		uti_moddiv(input, CIVector3(xsegsize, ysegsize, zsegsize), mod, res);
		return mod;
	}
	static inline void pixelToModAndSegment(const CIVector3& input, CIVector3 &mod, CIVector3 &res)
	{
		uti_moddiv(input, CIVector3(xsegsize, ysegsize, zsegsize), mod, res);
	}
	static inline CIntBox segmentToPixelRect(const CIVector3& segment)
	{
		return CIntBox(segment.x*xsegsize, segment.y*ysegsize, segment.z*zsegsize, (segment.x + 1)*xsegsize, (segment.y + 1)*ysegsize, (segment.z + 1)*zsegsize);
	}
	static inline CIntBox segmentRectToPixelRect(const CIntBox& segment)
	{
		return CIntBox(segment.m_min.x*xsegsize, segment.m_min.y*ysegsize, segment.m_min.z*zsegsize, (segment.m_max.x)*xsegsize, (segment.m_max.y + 1)*ysegsize, (segment.m_max.z + 1)*zsegsize);
	}
	static inline CBox pixelToFloatRect(const CIVector3& pixel)
	{
		return CBox(pixel.x * floatxsize::toFloatingPoint(), pixel.y * floatysize::toFloatingPoint(), pixel.z * floatzsize::toFloatingPoint(),
			(pixel.x + 1) * floatxsize::toFloatingPoint(), (pixel.y + 1) * floatysize::toFloatingPoint(), (pixel.z + 1) * floatzsize::toFloatingPoint());
	}
	static inline CBox pixelRectToFloatRect(const CIntBox& pixel)
	{
		return CBox(pixel.m_min.x * floatxsize::toFloatingPoint(), pixel.m_min.y * floatysize::toFloatingPoint(), pixel.m_min.z * floatzsize::toFloatingPoint(),
			(pixel.m_max.x) * floatxsize::toFloatingPoint(), (pixel.m_max.y) * floatysize::toFloatingPoint(), (pixel.m_max.z) * floatzsize::toFloatingPoint());
	}
	static inline CBox segmentToFloatRect(const CIVector3& seg)
	{
		return CBox(seg.x*xsegsize * floatxsize::toFloatingPoint(), seg.y*ysegsize * floatysize::toFloatingPoint(), seg.z*zsegsize * floatzsize::toFloatingPoint(),
			(seg.x + 1)*xsegsize * floatxsize::toFloatingPoint(), (seg.y + 1)*ysegsize * floatysize::toFloatingPoint(), (seg.z + 1)*zsegsize * floatzsize::toFloatingPoint());
	}


	static inline CBox segmentRectToFloatRect(const CIntBox& seg)
	{
		return CBox(seg.m_min.x*xsegsize * floatxsize::toFloatingPoint(), seg.m_min.y*ysegsize * floatysize::toFloatingPoint(), seg.m_min.z*zsegsize * floatzsize::toFloatingPoint(),
			(seg.m_max.x)*xsegsize * floatxsize::toFloatingPoint(), (seg.m_max.y)*ysegsize * floatysize::toFloatingPoint(), (seg.m_max.z)*zsegsize * floatzsize::toFloatingPoint());
	}

	static inline CIVector3 floatToPixel(const CVector3& position)
	{
		CIVector3 res;

		res.x = fasterfloor(position.x*floatxsize::toFloatingPointInv());
		res.y = fasterfloor(position.y*floatysize::toFloatingPointInv());
		res.z = fasterfloor(position.z*floatysize::toFloatingPointInv());
		return res;
	}
	static inline CIVector3 floatToSegment(const CVector3& position)
	{
		CIVector3 res;

		res.x = fasterfloor(position.x*floatxsize::toFloatingPointInv());
		res.y = fasterfloor(position.y*floatysize::toFloatingPointInv());
		res.z = fasterfloor(position.z*floatysize::toFloatingPointInv());
		return pixelToSegment(res);
	}
	static inline CIntBox floatRectToPixelRect(const CBox& position)
	{
		CIntBox res;
		res.m_min.x = fasterfloor(position.m_min.x*floatxsize::toFloatingPointInv());
		res.m_min.y = fasterfloor(position.m_min.y*floatysize::toFloatingPointInv());
		res.m_min.z = fasterfloor(position.m_min.z*floatysize::toFloatingPointInv());
		res.m_max.x = fasterfloor(position.m_max.x*floatxsize::toFloatingPointInv() - 0.00001f) + 1;
		res.m_max.y = fasterfloor(position.m_max.y*floatysize::toFloatingPointInv() - 0.00001f) + 1;
		res.m_max.z = fasterfloor(position.m_max.z*floatysize::toFloatingPointInv() - 0.00001f) + 1;
		return res;
	}
};

#endif // UTIADVARRAY_H
