#ifndef UTIARRAY_H
#define UTIARRAY_H
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Uti3DCore.h"
#include "utiintvector.h"
#include "utimathbox.h"


#ifdef USEMEMDUMP
#define ITERATORDEBUG
#endif

template <class T>
class Array3D
{
   protected:
      bool dynamic;
   public:
      class Array3DIterator {
      protected:
          T* m_array;

          int m_newline;
          int m_pagesize;
#ifdef ITERATORDEBUG
          CIVector3 m_currentposition;
          CIVector3 m_size;
#endif
      public:
          inline Array3DIterator(const Array3DIterator& v)
          {
              m_array = v.m_array;
              m_newline = v.m_newline;
              m_pagesize = v.m_pagesize;
#ifdef ITERATORDEBUG
              m_currentposition = v.m_currentposition;
              m_size = v.m_size;
#endif
          }

          inline explicit Array3DIterator(T* a,  int newline, int pagesize, const CIVector3& size,const CIVector3& start)
          {
              m_array = a;
              m_newline = newline;
              m_pagesize = pagesize;
#ifdef ITERATORDEBUG
              m_currentposition = start;
              m_size = size;
#endif
          }
          inline Array3DIterator()
          {
              m_array = NULL;
              m_newline = 0;
              m_pagesize = 0;
#ifdef ITERATORDEBUG
              m_currentposition = CIVector3(0,0,0);
              m_size = CIVector3(0,0,0);
#endif
          }
          inline Array3DIterator& operator=(const Array3DIterator& v)
          {
              m_array = v.m_array;
              m_newline = v.m_newline;
              m_pagesize = v.m_pagesize;
#ifdef ITERATORDEBUG
              m_currentposition = v.m_currentposition;
              m_size = v.m_size;
#endif
              return *this;
          }

          inline const T& get() const  {
              #ifdef ITERATORDEBUG
              assert(m_currentposition.x >= 0 && m_currentposition.x < m_size.x);
              assert(m_currentposition.y >= 0 && m_currentposition.y < m_size.y);
              assert(m_currentposition.z >= 0 && m_currentposition.z < m_size.z);
              #endif
              return *m_array;
          }
          inline T& getPointer()   {
#ifdef ITERATORDEBUG
assert(m_currentposition.x >= 0 && m_currentposition.x < m_size.x);
assert(m_currentposition.y >= 0 && m_currentposition.y < m_size.y);
assert(m_currentposition.z >= 0 && m_currentposition.z < m_size.z);
#endif
              return *m_array;
          }
          inline void set(const T& v)
          {
#ifdef ITERATORDEBUG
assert(m_currentposition.x >= 0 && m_currentposition.x < m_size.x);
assert(m_currentposition.y >= 0 && m_currentposition.y < m_size.y);
assert(m_currentposition.z >= 0 && m_currentposition.z < m_size.z);
#endif
              *m_array = v;
          }
          inline void incX(int amount = 1) {
              m_array += amount;
#ifdef ITERATORDEBUG
              m_currentposition.x += amount;
            //  assert(m_currentposition.x >= 0 && m_currentposition.x < m_size.x);
#endif
          }
          inline void incY(int amount = 1) {
              m_array += amount*m_newline;
#ifdef ITERATORDEBUG
              m_currentposition.y += amount;
             // assert(m_currentposition.y >= 0 && m_currentposition.y < m_size.y);
#endif
          }
          inline void incZ(int amount = 1) {
              m_array += amount*m_pagesize;
#ifdef ITERATORDEBUG
              m_currentposition.z+= amount;
              //assert(m_currentposition.z >= 0 && m_currentposition.z < m_size.z);
#endif
          }
          inline void move(const CIVector3& v)
          {
              m_array += v.x+v.y*m_newline+ v.z*m_pagesize;
#ifdef ITERATORDEBUG
              m_currentposition+= v;

#endif

          }
#ifdef ITERATORDEBUG
          bool checkDebugPosition(const CIVector3& v)
          {

              return v == m_currentposition;
          }
#else
          bool checkDebugPosition(const CIVector3& v)
          {

              return true;
          }
#endif
      };
      typedef Array3DIterator iterator;

      T* array;
      int array_width;
      int array_height;
      int array_depth;
      int array_pagesize;
      int array_newline;

      inline Array3DIterator getIterator()
      {
          return Array3DIterator(array, array_newline, array_pagesize, CIVector3(array_width, array_height, array_depth), CIVector3(0,0,0));
      }
      inline Array3DIterator getConst_Iterator() const
      {
          return Array3DIterator(const_cast<T*>(array), array_newline, array_pagesize, CIVector3(array_width, array_height, array_depth), CIVector3(0,0,0));
      }

      inline Array3DIterator getIterator(const CIVector3& pos)
      {
          return Array3DIterator(&(*this)(pos), array_newline, array_pagesize, CIVector3(array_width, array_height, array_depth), CIVector3(0,0,0));
      }
      inline Array3DIterator getConst_Iterator(const CIVector3& pos) const
      {
          return Array3DIterator(const_cast<T*>(&(*this)(pos)), array_newline, array_pagesize, CIVector3(array_width, array_height, array_depth), CIVector3(0,0,0));
      }


      inline bool isDynamic() const
      {
         return dynamic;
      }
      inline int width() const
      {
         return array_width;
      }
      inline int height() const
      {
         return array_height;
      }
      inline int depth() const
      {
    return array_depth;
      };
	  inline CIVector3 size() const
	  {
		  return CIVector3(array_width, array_height, array_depth);
	  }
          inline CIntBox getBox() const
          {
              return CIntBox(0,0,0, array_width, array_height, array_depth);
          }
      inline Array3D(int w, int h, int d)
      {
         dynamic = true;
         array_width = w;
         array_height = h;
     array_depth = d;
         array_newline = w;
     array_pagesize = array_newline*array_height;
         array = new T[array_pagesize*array_depth];
      }

          inline Array3D(const CIVector3& size)
	  {
		  dynamic = true;
		  array_width = size.x;
		  array_height = size.y;
		  array_depth = size.z;
                  array_newline = size.x;
		  array_pagesize = array_newline*array_height;
		  array = new T[array_pagesize*array_depth];
	  }
      inline Array3D(T* ar, int w, int h, int d)
      {
         dynamic = false;
         array_width = w;
         array_height = h;
     array_depth = d;
         array_newline = w;
     array_pagesize = array_newline*array_height;
         array = ar;
      }

      inline Array3D()
      {
         array_width = 0;
         array_height = 0;
     array_depth = 0;
         array_newline = 0;
     array_pagesize = 0;

         array = NULL;
     dynamic = false;
      };
      Array3D(const Array3D &narray);
      inline ~Array3D()
      {
         if (dynamic)
            delete [] array;
      }
      inline void setSize(int w, int h, int d) {
         utiassert(array == NULL);
      //CHECKEXISTANCE;
         dynamic = true;
         array_width = w;
         array_height = h;
     array_depth = d;
         array_newline = w;
     array_pagesize = array_newline*array_height;
         array = new T[array_pagesize*array_depth];
      };
      inline void setSize(const CIVector3& v)
      {
          setSize(v.x, v.y, v.z);
      }


      inline void setSizeCopy(int w, int h, int d) {
         if (array_width == w && array_height == h && array_depth == d) {
             return;
         }
      //CHECKEXISTANCE;
         int oldw = array_width;
         int oldh = array_height;
         int oldd = array_depth;
         int oldnewline = array_newline;
         int oldpagesize = array_pagesize;
         T* oldarray = array;
         dynamic = true;
         array_width = w;
         array_height = h;
     array_depth = d;
         array_newline = w;
     array_pagesize = array_newline*array_height;
         array = new T[array_pagesize*array_depth];
         if (oldarray != NULL) {
             int xsize = umin(array_width, oldw);
             int ysize = umin(array_height, oldh);
             int zsize = umin(array_depth, oldd);

             for (int z = 0; z < zsize; z++) {
                 for (int y = 0; y < ysize; y++) {
                     for (int x = 0; x < xsize; x++) {
                         array[x+y*array_newline+z*array_pagesize] = oldarray[x+y*oldnewline+z*oldpagesize];
                     }
                 }
             }
             delete []oldarray;
         }
      };
      inline void setSizeCopy(const CIVector3& v)
      {
          setSizeCopy(v.x, v.y, v.z);
      }

      /*void ReferenzTo(const Array2D &a)
      {
         dynamic = false;
         array_size = a.array_size;
         array = a.array;
      }*/
      inline T &operator()(unsigned int x, unsigned int y, unsigned int z)
      {
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
     utiassert(z >= 0 && z < (unsigned int)array_depth);
         return array[x+y*array_newline+z*array_pagesize];

      }
      inline const T &operator()(unsigned int x, unsigned int y, unsigned int z) const
      {
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
     utiassert(z >= 0 && z < (unsigned int)array_depth);
         return array[x+y*array_newline+z*array_pagesize];

      }
      inline T &operator()(const CIVector3& pos)
      {
         utiassert(array != NULL);
         utiassert(pos.x >= 0 && pos.x < (unsigned int)array_width);
         utiassert(pos.y >= 0 && pos.y < (unsigned int)array_height);
     utiassert(pos.z >= 0 && pos.z < (unsigned int)array_depth);
         return array[pos.x+pos.y*array_newline+pos.z*array_pagesize];

      }
      inline const T &operator()(const CIVector3& pos) const
      {
         utiassert(array != NULL);
         utiassert(pos.x >= 0 && pos.x < (unsigned int)array_width);
         utiassert(pos.y >= 0 && pos.y < (unsigned int)array_height);
     utiassert(pos.z >= 0 && pos.z < (unsigned int)array_depth);
         return array[pos.x+pos.y*array_newline+pos.z*array_pagesize];

      }


      bool isEqual(const CIVector3& minvalue,const CIVector3& maxvalue, const T& v) const
      {
          for (int z = minvalue.z; z < maxvalue.z; z++) {
              for (int y = minvalue.y; y < maxvalue.y; y++) {
                  for (int x = minvalue.x; x < maxvalue.x; x++) {
                      if (!(operator()(x, y, z) == v)) {
                  return false;
                 }
                  }
              }
          }
          return true;
      }

      bool isEqual(const CIntBox& b, const T& v) const
      {
          return isEqual(b.m_min, b.m_max, v);
      }
      void setBox(const CIntBox& b, const T&v)
      {
          for (int z = b.m_min.z; z < b.m_max.z; z++) {
              for (int y = b.m_min.y; y < b.m_max.y; y++) {
                  for (int x = b.m_min.x; x < b.m_max.x; x++) {
                      operator()(x, y, z) = v;


                  }
              }
          }

      }
      void fill(const T& value)
      {
      int s = array_pagesize*array_depth;
          for (int i = 0; i < s; i++) array[i] = value;
       }
      void copyTo(Array3D<T>& d)
      {
          d.setSize(size());
          int s = array_pagesize*array_depth;
          for (int i = 0; i < s; i++) d.array[i] = array[i];

      }
};


template <class T>
      Array3D<T>::Array3D(const Array3D &narray)
{
   dynamic = true;
   array_width = narray.width();
   array_height = narray.height();
   array_depth = narray.depth();
   array_newline = array_width;
   array_pagesize = array_newline*array_height;
   array = new T[array_depth*array_pagesize];
//   wcout << "Warnung: Array wird kopiert"<<endl;
   for (int k = 0; k < array_depth; k++)
   {
   for (int j = 0; j < array_height; j++)
   {
      for (int i = 0; i < array_width; i++)
      {
          (*this)(i, j, k) = narray(i, j, k);
      }
   }
   }

}

template <class T>
class Array2D
{
   protected:
      bool dynamic;
   public:
      T* array;
      int array_width;
      int array_height;
      int array_newline;
      bool isDynamic() const
      {
         return dynamic;
      }
      inline int width() const
      {
         return array_width;
      }
      inline int height() const
      {
         return array_height;
      }
      Array2D(int w, int h)
      {
         dynamic = true;
         array_width = w;
         array_height = h;
         array_newline = w;
         array = new T[array_newline*array_height];
      }
      Array2D(T* ar, int w, int h)
      {
         dynamic = false;
         array_width = w;
         array_height = h;
         array_newline = w;
         array = ar;
      }
      Array2D(const CIVector2& size)
      {
         dynamic = true;
         array_width = size.x;
         array_height = size.y;
         array_newline = size.x;
         array = new T[array_newline*array_height];
      }
      Array2D(T* ar, int w, int newline, int h)
      {
         dynamic = false;
         array_width = w;
         array_height = h;
         array_newline = newline;
         utiassert(array_newline >= array_width);
         array = ar;
      }
      Array2D()
      {
         array_width = 0;
         array_height = 0;
         array_newline = 0;
         dynamic = false;
         array = NULL;
      };
      Array2D(const Array2D &narray);
      ~Array2D()
      {
         if (dynamic)
            delete [] array;
      }
      void setDataStack(T* ar, int w, int newline, int h)
      {assert(array == NULL);
    dynamic = false;
         array_width = w;
         array_height = h;
         array_newline = newline;
         utiassert(array_newline >= array_width);
         array = ar;
      }
      inline CIntRect getRect() const
      {
          return CIntRect(0,0, array_width, array_height);
      }
      inline CIVector2 size() const
      {
          return CIVector2(array_width, array_height);
      }
      inline void setSize(const CIVector2& s)
      {
          setSize(s.x, s.y);
      }

      void setSize(int w, int h) {
     //CHECKEXISTANCE;

         if (array != NULL){
       if (w == array_width && h == array_height) return;
       if (dynamic)
            delete [] array;
     };
         dynamic = true;
         array_width = w;
         array_height = h;
         array_newline = w;
         array = new T[array_newline*array_height];
      };
      void setSizeCopy(int w, int h, int basex=0, int basey=0) {
   //  CHECKEXISTANCE;
     if (w == array_width && h == array_height && basex == 0 && basey == 0) return;
    T* newarray = new T[w*h];
    utiassert(basex >= 0);
    utiassert(basey >= 0);
    utiassert(w >= 0);
    utiassert(h >= 0);
    int mw = umin(w, array_width-basex);
    int mh = umin(h, array_height-basey);
    for( int j = 0; j < mh; j++) {
      int jbasey = j+basey;
       for (int i = 0; i < mw; i++) {
          newarray[i+j*w] =  array[(i+basex)+(jbasey)*array_newline];
       };
    };

         if (array != NULL){
       if (dynamic)
            delete [] array;
     };
         dynamic = true;
         array_width = w;
         array_height = h;
         array_newline = w;
         array = newarray;
      };
      //Only for Flat Objects
      void copyToFast(Array2D& obj) const
      {
    obj.setSize(array_width, array_height);
    memcpy(obj.array, array, sizeof(T)*array_width*array_height);

      }
      void copyToArray(Array2D& obj) const
      {
    obj.setSize(array_width, array_height);
    int c = array_width*array_height;
    for (int i = 0; i < c; i++) {
       obj.array[i] = array[i];
    }
      }
      void copyToArrayPos(const CIVector2& pos, Array2D& obj) const
      {

          for (int j = 0; j < array_height; j++) {
              for (int i = 0; i < array_width; i++) {
                  obj(pos.x+i, pos.y+j) = (*this)(i, j);

              }

          }
      }
      /*void ReferenzTo(const Array2D &a)
      {
         dynamic = false;
         array_size = a.array_size;
         array = a.array;
      }*/
      inline T &operator()(unsigned int x, unsigned int y)
      {
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
         return array[x+y*array_newline];

      }
      inline T* getPointer(unsigned int x, unsigned int y)
      {
    utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
         return &array[x+y*array_newline];
      }
      inline const T &operator()(unsigned int x, unsigned int y) const
      {
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
         return array[x+y*array_newline];

      }
      inline T &operator()(const CIVector2& p)
      {
         utiassert(array != NULL);
         utiassert(p.x >= 0 && p.x < array_width);
         utiassert(p.y >= 0 && p.y < array_height);
         return array[p.x+p.y*array_newline];

      }
      inline const T &operator()(const CIVector2& p) const
      {
         utiassert(array != NULL);
         utiassert(p.x >= 0 && p.x < array_width);
         utiassert(p.y >= 0 && p.y < array_height);
         return array[p.x+p.y*array_newline];

      }
      inline void set(const CIVector2& p, const T& v)
          {
             utiassert(array != NULL);
             utiassert(p.x >= 0 && p.x < array_width);
             utiassert(p.y >= 0 && p.y < array_height);
             array[p.x+p.y*array_newline] = v;

          }
          inline T get(const CIVector2& p) const
          {
             utiassert(array != NULL);
             utiassert(p.x >= 0 && p.x < array_width);
             utiassert(p.y >= 0 && p.y < array_height);
             return array[p.x+p.y*array_newline];

          }

      void fill(const T& value)
      {
      int s = array_newline*array_height;
          for (int i = 0; i < s; i++) array[i] = value;
       }
      void fillRectangle(const CIntRect& rect, const T& value)
      {
          for (int j = rect.m_min.y; j < rect.m_max.y; j++) {
              for (int i = rect.m_min.x; i < rect.m_max.x; i++) {
                  *getPointer(i,j) = value;

              }
          }
       }
};

template <class T>
      Array2D<T>::Array2D(const Array2D &narray)
{
   dynamic = true;
   array_width = narray.width();
   array_height = narray.height();
   array_newline = array_width;
   array = new T[array_height*array_newline];
//   wcout << "Warnung: Array wird kopiert"<<endl;
   for (int j = 0; j < array_height; j++)
   {
      for (int i = 0; i < array_width; i++)
      {
          (*this)(i, j) = narray(i, j);
      }
   }

}




template<class T> inline int umaxindex(const Array<T>& a)
{
    int res = UTI_INVALIDINDEX;
    if (a.size() != 0) res = 0;
    for (int i = 1; i <a.size(); i++){
        if (a[res] < a[i]){
            res = i;
        }
    }
    return res;
}

template<class T> inline int uminindex(const Array<T>& a)
{
    int res = UTI_INVALIDINDEX;
    if (a.size() != 0) res = 0;
    for (int i = 1; i <a.size(); i++){
        if (a[res] > a[i]){
            res = i;
        }
    }
    return res;
}

template <class T> struct ArrayAccess
{
  public:
    ArrayAccess()
    {

    }
    virtual ~ArrayAccess(){}
    virtual int getRestSize() const {return INT_MAX;};
    inline int push_back(const T& el) {return add(el);};
    virtual int add(const T &el) {return 0;};
    virtual int size() const {return 0;};
    virtual void flush() {};
};

template <class T> struct UtiArrayAccess : public ArrayAccess<T>
{
  protected:
    Array<T>& m_array;
    int m_startposition;
    int m_position;
    int m_size;
  public:
    explicit UtiArrayAccess(Array<T>& a,int start=0) : m_array(a)
    {
      m_size = a.size();
      m_startposition=start;
      m_position =start;
    };

    explicit UtiArrayAccess(Array<T>& a,int start, int size) : m_array(a)
    {
      utiassert(start+size <= m_array.size());
      m_size = size;
      m_startposition=start;
      m_position =start;
    };
    virtual int getRestSize()const
    {
       return m_size-(m_position-m_startposition);
    }
    virtual void add(const T &el)
    {
        utiassert(getRestSize() > 0);
    m_array[m_position] = el;
    m_position++;
    };
    virtual int size() const {return (m_position-m_startposition);};


};


template <class T, bool fastgrow=false, class Ar=Array<T> > struct UtiArrayAccessGrow : public ArrayAccess<T>
{
  protected:
  public:
          Ar& m_array;
   int m_position;
    explicit UtiArrayAccessGrow(Ar& a,int start=0) : m_array(a)
    {
      m_position =start;
      utiassert(start < m_array.size());
    };

    inline Ar& getArray()
    {
        return m_array;
    }

    inline void reduceSize(int r)
    {
        m_position -= r;
        assert(m_position >= 0);
    }

    virtual int add(const T &el)
    {
        if (m_array.size() <= m_position) {
            if (fastgrow) {
                  m_array.setSizeCopyHighSpeed(m_array.size()*2);
            }else {
      m_array.setSizeCopy(m_array.size()*2);
            }
    };
    m_array[m_position] = el;
    m_position++;
    return m_position-1;
    };
	inline int fast_add(const T &el)
	{
		if (m_array.size() <= m_position) {
			if (fastgrow) {
				m_array.setSizeCopyHighSpeed(m_array.size() * 2);
			}
			else {
				m_array.setSizeCopy(m_array.size() * 2);
			}
		};
		m_array[m_position] = el;
		m_position++;
		return m_position - 1;
	};
    virtual int grow()
    {
        if (m_array.size() <= m_position) {
            if (fastgrow) {
                  m_array.setSizeCopyHighSpeed(m_array.size()*2);
            }else {
      m_array.setSizeCopy(m_array.size()*2);
            }
    };
//	m_array[m_position] = el;
    m_position++;
    return m_position-1;
    };
	inline void pop_back()
	{
		m_position--;
		assert(m_position >= 0);
	}
	inline void growToSize(int newsize)
	{
		assert(m_position == 0);
		int s = m_array.size();
		assert(s != 0);
        if (s >= newsize){
            m_position = newsize;
            return;
        }
		while (s < newsize) {
			s = s * 2;
		}
		if (fastgrow) {
			m_array.setSizeCopyHighSpeed(s);
		}
		else {
			m_array.setSizeCopy(s);
		}
		m_position = newsize;
	}
	inline int fast_grow()
	{
		if (m_array.size() <= m_position) {
			if (fastgrow) {
				m_array.setSizeCopyHighSpeed(m_array.size() * 2);
			}
			else {
				m_array.setSizeCopy(m_array.size() * 2);
			}
		};
		//	m_array[m_position] = el;
		m_position++;
		return m_position - 1;
	};
        inline int fast_size() const { return m_position; }
    virtual int size() const {return m_position;}
    virtual void reset(int maxsize, int start=0) {
        m_position = start;
        if (m_array.size() > maxsize) {
           // m_array.setSizeCopy(maxsize);
            if (fastgrow) {
                  m_array.setSizeCopyHighSpeed(maxsize);
            }else {
                 m_array.setSizeCopy(maxsize);
            }
        }
    }
};

/*
template <class T>
      Array<T>::Array(const Array<T> &narray) : CUtiOB()
{

    if (narray.dynamic) {
   dynamic = true;
   array_size = narray.size();
   array = new T[array_size];
   for (int i = 0; i < array_size; i++)
   {
      array[i] = narray[i];

   }
    } else {
        utiassert(narray.stackvar == false);

       dynamic = false;
       array_size = narray.array_size;
       array = narray.array;
    }

}*/

typedef Array<unsigned char> ArrayUC;

template<class T> struct UtiDynamicArray : ArrayAccess<T>
{

          Array<T> m_array;
          UtiArrayAccessGrow<T> m_arraygrow;
     UtiDynamicArray() : m_array(16), m_arraygrow(m_array)
     {

     }
     virtual int add(const T &el) {return m_arraygrow.push_back(el);};
     virtual int size() const {return m_arraygrow.size();};
     virtual void reset(int maxsize, int start=0) {
         m_arraygrow.reset(maxsize, start);
     }

};



      template<class T, class Parent> struct arrayiterator {
          int m_index;
          Parent* m_array;
          typedef std::random_access_iterator_tag iterator_category;
          typedef arrayiterator self_type;
          typedef T value_type;

           typedef T& reference;
           typedef const T& const_reference;
           typedef T* pointer;
           typedef size_t size_type;
           typedef std::ptrdiff_t difference_type;
          inline arrayiterator(Parent* owner, int index)
          {
              m_array = owner;
              m_index = index;
          }
          inline arrayiterator& operator=(const arrayiterator& iter)
          {
              m_index = iter.m_index;
              m_array = iter.m_array;
              return *this;
          }
          inline bool operator<(const arrayiterator& iter) const
          {
              assert(m_array == iter.m_array);
              return (m_index < iter.m_index);
          }
          inline bool operator>(const arrayiterator& iter) const
          {
              assert(m_array == iter.m_array);
              return (m_index > iter.m_index);
          }
          inline bool operator<=(const arrayiterator& iter) const
          {
              assert(m_array == iter.m_array);
              return (m_index <= iter.m_index);
          }
          inline bool operator>=(const arrayiterator& iter) const
          {
              assert(m_array == iter.m_array);
              return (m_index >= iter.m_index);
          }
          inline bool operator==(const arrayiterator& iter) const
          {
              assert(m_array == iter.m_array);
              return (m_index == iter.m_index) && (m_array == iter.m_array);
          }

          inline bool operator!=(const arrayiterator& iter) const
          {
              assert(m_array == iter.m_array);
              return !((m_index == iter.m_index) && (m_array == iter.m_array));
          }

          inline arrayiterator& operator+=(ptrdiff_t movement){m_index += movement;return (*this);}
          inline arrayiterator& operator-=(ptrdiff_t movement){m_index -= movement;return (*this);}
          inline arrayiterator& operator++(){++m_index;return (*this);}
          inline arrayiterator& operator--(){--m_index;return (*this);}

          inline arrayiterator operator+(ptrdiff_t movement){arrayiterator n1 = *this;n1.m_index+=movement;return n1;}
          inline arrayiterator operator-(ptrdiff_t movement){arrayiterator n1 = *this;n1.m_index-=movement;return n1;}

          ptrdiff_t operator-(const arrayiterator& iter){ assert(m_array == iter.m_array);return m_index-iter.m_index;}

          inline T&   operator*(){return (*m_array)[m_index];}
          inline const T&  operator*()const{return (*m_array)[m_index];}
          inline T* operator->(){return &(*m_array)[m_index];}
          inline T* getPtr()const{return &(*m_array)[m_index];}
          inline const T*  getConstPtr()const{return &(*m_array)[m_index];}
      };

      template<class T, class Parent> struct reverse_arrayiterator {
          int m_index;
          Parent* m_array;
          inline reverse_arrayiterator(Parent* owner, int index)
          {
              m_array = owner;
              m_index = index;
          }
          inline reverse_arrayiterator& operator=(const reverse_arrayiterator& iter)
          {
              m_index = iter.m_index;
              m_array = iter.m_array;
          }
          inline bool operator==(const reverse_arrayiterator& iter) const
          {
              assert(m_array == iter.m_array);
              return (m_index == iter.m_index) && (m_array == iter.m_array);
          }

          inline bool operator!=(const reverse_arrayiterator& iter) const
          {
              assert(m_array == iter.m_array);
              return !((m_index == iter.m_index) && (m_array == iter.m_array));
          }

          inline reverse_arrayiterator& operator+=(ptrdiff_t movement){m_index -= movement;return (*this);}
          inline reverse_arrayiterator& operator-=(ptrdiff_t movement){m_index += movement;return (*this);}
          inline reverse_arrayiterator& operator++(){--m_index;return (*this);}
          inline reverse_arrayiterator& operator--(){++m_index;return (*this);}

          inline reverse_arrayiterator operator+(ptrdiff_t movement){reverse_arrayiterator n1 = *this;n1.m_index-=movement;return n1;}
          inline reverse_arrayiterator operator-(ptrdiff_t movement){reverse_arrayiterator n1 = *this;n1.m_index+=movement;return n1;}

          ptrdiff_t operator-(const reverse_arrayiterator& iter){ assert(m_array == iter.m_array);return -(m_index-iter.m_index);}

          inline T&   operator*(){return (*m_array)[m_index];}
          inline const T&  operator*()const{return (*m_array)[m_index];}
          inline T* operator->(){return &(*m_array)[m_index];}
          inline T* getPtr()const{return &(*m_array)[m_index];}
          inline const T*  getConstPtr()const{return &(*m_array)[m_index];}
      };


#define UTI_DECL_ARRAY(type, name, size) type _intern_ ## name ## _array_var [size]; Array<type> name(_intern_ ## name ## _array_var, size, true);

#define UTI_DECL_DYNAMIC_ARRAY(type, name, basesize, size) type _intern_ ## name ## _array_var [basesize]; Array<type> name; if (basesize >= size) {name.init(_intern_ ## name ## _array_var, size, true);} else {name.setSize(size);};

template <class T, int array_size> struct StaticArray
      {
      protected:
          T array[array_size];
      public:
           typedef T value_type;
           typedef T& reference;
           typedef const T& const_reference;
           typedef T* pointer;
           typedef size_t size_type;


          inline StaticArray()
          {

          }
          inline T* getData()
          {
              return array;
		  }
		  inline const T* getData() const
		  {
			  return array;
		  }

                  typedef arrayiterator<T, StaticArray<T, array_size> > iterator;
                  typedef arrayiterator<const T, StaticArray<T, array_size> > const_iterator;

                  typedef reverse_arrayiterator<T, StaticArray<T, array_size> > reverse_iterator;
                  typedef reverse_arrayiterator<const T, StaticArray<T, array_size> > const_reverse_iterator;

                  iterator  begin(){return iterator(this, 0);}
                  iterator  end(){return iterator(this, size());}
                  const_iterator  cbegin() const{return const_iterator(this, 0);}
                  const_iterator  cend() const {return const_iterator(this, size());}


                  reverse_iterator  rbegin(){return reverse_iterator(this, 0);}
                  reverse_iterator  rend(){return reverse_iterator(this, size());}
                  const_reverse_iterator  crbegin() const{return const_reverse_iterator(this, 0);}
                  const_reverse_iterator  crend() const {return const_reverse_iterator(this, size());}
          inline T &operator[](unsigned int p)
          {
             assert((int)p >= 0 && (int)p < array_size);
             assert(array != NULL);
             return array[p];

          }
          inline const T &operator[](unsigned int p) const
          {
             assert((int)p >= 0 && (int)p < array_size);
             assert(array != NULL);
             return array[p];

          }
          inline const T &get(unsigned int p) const
          {
             assert((int)p >= 0 && (int)p < array_size);
             assert(array != NULL);
             return array[p];

          }
          inline void set(unsigned int p, const T& v)
          {
             assert((int)p >= 0 && (int)p < array_size);
             assert(array != NULL);
             array[p] = v;

          }
          void writeTo(T* ar, int size) const
          {
              int v = size;
              if (v > array_size) v = array_size;
              for (int i = 0; i < v; i++)
              {
                  ar[i] = array[i];

              }

          }
          void fill(const T& value)
          {
         //CHECKEXISTANCE;
              for (int i = 0; i < array_size; i++) array[i] = value;
           }
         void copyTo(int start, int count, Array<T>& target, int targetstart) const
         {
            // CHECKEXISTANCE;
             utiassert(start >= 0);
             utiassert(start+count <= array_size);
         utiassert(targetstart >= 0);
         utiassert(targetstart+count <= target.size());
         if (count == 0) return;
         for (int i = 0; i < count; i++) {
             target.array[targetstart+i] = array[start+i];
         };
         };
         inline int size() const
         {
             return array_size;
         }
         void deleteItems(int start, int end)
         {
             assert(start <= end);
             assert(start >= 0);
             assert(end <= array_size);
             for (int i = end; i < array_size; i++) {
                 array[i-end+start] = array[i];
             }
             //setSizeCopy(array_size-(end-start));

         }
         void insert(int index, const T& value)
         {
             int a = array_size;
             //setSizeCopy(array_size+1);
             for (int i = a-1; i >= index; i--) {
                 array[i+1] = array[i];
             }
             operator[](index) = value;
         }
		 int indexOf(const T& value, int c = -1)
		 {
			 if (c == -1) {
				 c = array_size;
			 }
			 for (int i = 0; i < c; i++) {
				 if (value == array[i]) {
					 return i;
				 }
			 }
			 return -1;
		 }
           StaticArray& operator=(const StaticArray& t)
           {
               for (int i = 0; i < array_size; i++) {
                   array[i] = t.array[i];
               }
			   return *this;
           }

           bool operator==(const StaticArray& t)const
           {
               for (int i = 0; i < array_size; i++) {
                   if (array[i] != t.array[i]) return false;
               }
               return true;
           }

      };


      template<class T, int basesize=6> struct StaticAndDynamicArray
      {
          int m_size;
          StaticArray<T, 6> m_data;
          Array<T>* m_additionalsymbols;
          typedef arrayiterator<T, StaticAndDynamicArray<T> > iterator;
          typedef arrayiterator<const T, StaticAndDynamicArray<T> > const_iterator;

          typedef reverse_arrayiterator<T, StaticAndDynamicArray<T> > reverse_iterator;
          typedef reverse_arrayiterator<const T, StaticAndDynamicArray<T> > const_reverse_iterator;

          iterator  begin(){return iterator(this, 0);}
          iterator  end(){return iterator(this, size());}
          const_iterator  cbegin() const{return const_iterator(this, 0);}
          const_iterator  cend() const {return const_iterator(this, size());}


          reverse_iterator  rbegin(){return reverse_iterator(this, 0);}
          reverse_iterator  rend(){return reverse_iterator(this, size());}
          const_reverse_iterator  crbegin() const{return const_reverse_iterator(this, 0);}
          const_reverse_iterator  crend() const {return const_reverse_iterator(this, size());}

          StaticAndDynamicArray()
          {
              m_size = 0;m_additionalsymbols=NULL;
          }
          ~StaticAndDynamicArray()
          {
              if (m_additionalsymbols)
              delete m_additionalsymbols;
          }



          inline void setSize(int t)
          {

              if (m_additionalsymbols)
              delete m_additionalsymbols;
              m_size = t;
              if (m_size > basesize) {
                  m_additionalsymbols->setSize(m_size-basesize);
              }
          }

          inline T &operator[](unsigned int p)
          {
             assert((int)p >= 0 && (int)p < m_size);
             if (p < basesize) {
             return m_data[p];
             } else {
                 return (*m_additionalsymbols)[p-basesize];

             }


          }
          inline const T &operator[](unsigned int p) const
          {
             assert((int)p >= 0 && (int)p < m_size);
             if (p < basesize) {
             return m_data[p];
             } else {
                 return m_additionalsymbols->get(p-basesize);

             }

          }
          inline const T &get(unsigned int p) const
          {
             assert((int)p >= 0 && (int)p < m_size);
             if (p < basesize) {
             return m_data[p];
             } else {
                 return m_additionalsymbols->get(p-basesize);

             }

          }
          inline void set(unsigned int p, const T& v)
          {
             assert((int)p >= 0 && (int)p < m_size);
             if (p < basesize) {
             m_data[p] = v;
             } else {
                 m_additionalsymbols->set(p-basesize, v);

             }
          }

          void fill(const T& value)
          {
              for (int i = 0; i < m_size; i++) set(i, value);
           }

         inline int size() const
         {
             return m_size;
         }

      };


      template <class T, int array_size> T umax(const StaticArray<T, array_size>& s)
      {
          assert(s.size() != 0);
          T res = s[0];
          for (int i = 1; i < s.size(); i++) {
              if (res < s[i]) {
                  res = s[i];
              }
          }
          return res;
      }
      template <class T, int array_size> T umin(const StaticArray<T, array_size>& s)
      {
          assert(s.size() != 0);
          T res = s[0];
          for (int i = 1; i < s.size(); i++) {
              if (res > s[i]) {
                  res = s[i];
              }
          }
          return res;
      }

#define FOREACH_ARRAY(array, xvar, operation) \
         for (int xvar = 0; xvar < array.size(); xvar++) {\
              operation;\
          }

#define FOREACH_STATICARRAY2D(array, xvar, yvar, operation) \
      for (int yvar = 0; yvar < array.getHeight(); yvar++) {\
          for (int xvar = 0; xvar < array.getWidth(); xvar++) {\
              operation;\
          }\
      }
      template <class T, int array_sizex, int array_sizey> struct StaticArray2D
            {
            protected:
                T array[array_sizex*array_sizey];
            public:
                inline StaticArray2D()
                {

                }
                inline T* getData()
                {
                    return array;
                        }
                        inline const T* getData() const
                        {
                                return array;
                        }


                        inline T &operator()(const CIVector2& pos)
                        {
                           assert((int)pos.x >= 0 && (int)pos.x < array_sizex);
                           assert((int)pos.y >= 0 && (int)pos.y < array_sizey);
                           assert(array != NULL);
                           assert(pos.y*array_sizex+pos.x >= 0 && pos.y*array_sizex+pos.x < array_sizex*array_sizey);
                           return array[pos.y*array_sizex+pos.x];

                        }
                        inline const T &operator()(const CIVector2& pos) const
                        {
                            assert((int)pos.x >= 0 && (int)pos.x < array_sizex);
                            assert((int)pos.y >= 0 && (int)pos.y < array_sizey);
                            assert(array != NULL);
                            assert(pos.y*array_sizex+pos.x >= 0 && pos.y*array_sizex+pos.x < array_sizex*array_sizey);
                            return array[pos.y*array_sizex+pos.x];

                        }

                inline T &operator()(unsigned int x, unsigned int y)
                {
                   assert((int)x >= 0 && (int)x < array_sizex);
                   assert((int)y >= 0 && (int)y < array_sizey);
                   assert(array != NULL);
                   assert(y*array_sizex+x >= 0 && y*array_sizex+x < array_sizex*array_sizey);
                   return array[y*array_sizex+x];

                }
                inline const T &operator()(unsigned int x, unsigned y) const
                {
                    assert((int)x >= 0 && (int)x < array_sizex);
                    assert((int)y >= 0 && (int)y < array_sizey);
                    assert(array != NULL);
                    assert(y*array_sizex+x >= 0 && y*array_sizex+x < array_sizex*array_sizey);
                    return array[y*array_sizex+x];

                }
                inline const T &get(unsigned int x, unsigned int y) const
                {
                    assert((int)x >= 0 && (int)x < array_sizex);
                    assert((int)y >= 0 && (int)y < array_sizey);
                    assert(array != NULL);
                    assert(y*array_sizex+x >= 0 && y*array_sizex+x < array_sizex*array_sizey);
                    return array[y*array_sizex+x];

                }
                inline void set(unsigned int x, unsigned int y, const T& v)
                {
                    assert((int)x >= 0 && (int)x < array_sizex);
                    assert((int)y >= 0 && (int)y < array_sizey);
                    assert(array != NULL);
                    assert(y*array_sizex+x >= 0 && y*array_sizex+x < array_sizex*array_sizey);
                    array[y*array_sizex+x] =  v;



                }
                void fill(const T& value)
                {
               //CHECKEXISTANCE;
                    for (int i = 0; i < array_sizex*array_sizey; i++) array[i] = value;
                 }
               inline CIVector2 size() const
               {
                   return CIVector2(array_sizex, array_sizey);
               }
               inline int getWidth() const
               {
                   return array_sizex;
               }
               inline int getHeight() const
               {
                   return array_sizey;
               }


               inline CIntRect getRect() const
               {
                   return CIntRect(0,0, array_sizex, array_sizey);
               }
               StaticArray2D& operator=(const StaticArray2D& t)
               {
                   for (int i = 0; i < array_sizex*array_sizey; i++) {
                       array[i] = t.array[i];
                   
				   }
				   return *this;
               }
               bool operator==(const StaticArray2D& t) const
               {
                   for (int i = 0; i < array_sizex*array_sizey; i++) {
                       if (array[i] != t.array[i]) return false;
                   }
                   return true;
               }


            };
      template <class T, int array_sizex, int array_sizey> T umax(const StaticArray2D<T, array_sizex, array_sizey>& s)
      {
          assert(s.getWidth() != 0 && s.getHeight() != 0);
          T res = s(0,0);
          for (int j = 0; j < array_sizey; j++) {
          for (int i = 0; i < array_sizex; i++) {
              if (res < s(i,j)) {
                  res = s(i,j);
              }
          }
          }
          return res;
      }
      template <class T, int array_sizex, int array_sizey> T umin(const StaticArray2D<T, array_sizex, array_sizey>& s)
      {
          assert(s.getWidth() != 0 && s.getHeight() != 0);
          T res = s(0,0);
          for (int j = 0; j < array_sizey; j++) {
          for (int i = 0; i < array_sizex; i++) {
              if (res > s(i,j)) {
                  res = s(i,j);
              }
          }
          }
          return res;
      }

#endif // UTIARRAY_H
