#ifndef __COMPRESS__
#define __COMPRESS__
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Uti3DCore.h"
#include "utilinkedlist.h"

class CUtiOB;
class CCompressor;

struct SCompressorHuffmannNode;
struct SCompressorLZOutput
{
  unsigned char m_lengthorbyte;
  unsigned short m_delta;
};

struct SCompressorHuffmannNode
{
 unsigned short m_left;
 unsigned short m_right;
};

struct SCompressorHeapElements
{
  int m_amount;
  int m_noderef;
};



struct SCompressedHashItem : public CUtiOB
{
  int m_exactposition;
  CLinkedList_Controller clinkedlist_controller[1];
};

struct SCompressedSymbol
{
  int m_bits;
  int m_mask;
  CNormalLinkedList<SCompressedHashItem, 0> m_hashelements;
};


struct SCompressedHashGroup
{
  CNormalLinkedList<SCompressedHashItem, 0> m_hashelements;
};

/*template<int hashcompare=2> */struct SCompressedHashItems
{
  Array<SCompressedHashGroup> m_hash;
  CCompressor* m_com;
  SCompressedHashItems(int count, CCompressor* com)
  {
    m_hash.setSize(count);
    m_com = com;
  };
  
};

typedef unsigned char TComChar; 

enum TCompressorState {csNone, csDeflate, csInflate};

class CCompressor : public ReferenceUtiOB
{
protected:
  int m_blocksize;
  Array<SCompressorLZOutput> m_lzoutput;
  Array<TComChar> m_slidingwindow;
  int m_slidingwriteposition;
  int m_slidinglength;
  int m_slidingstreamposition;
  int m_currentwriteposition;
  inline TComChar slidingData(int index)
  {
    utiassert(m_currentwriteposition+index >= m_slidingstreamposition);
    return m_slidingwindow[((m_slidingwriteposition-1-index)+m_slidingwindow.size()) % m_slidingwindow.size()];
    
  };
  inline void writeSliding(TComChar sliding)
  {
    m_slidingwindow[m_slidingwriteposition] = sliding;
    if (m_slidinglength == m_blocksize) {
      //Delete aus HashMap
      m_slidingstreamposition++;
    } else {
      m_slidinglength++;
    };
    m_slidingwriteposition++;
    if (m_slidingwriteposition >= m_blocksize) {
      m_slidingwriteposition = 0; 
    }
    //Add to hashmap Level 1
    utiassert(m_unusedhashelements.size() > 4);
    SCompressedHashItem* item = &m_unusedhashelements.front();
    m_unusedhashelements.pop_front();
    item->m_exactposition = m_currentwriteposition;
    m_symbols[sliding].m_hashelements.push_back(item);
    if (m_currentwriteposition>= 1) {
      TComChar c[2];
      c[0] = slidingData(-1);
      c[1] = slidingData(0);
      item = &m_unusedhashelements.front();
      m_unusedhashelements.pop_front();
      item->m_exactposition = m_currentwriteposition-1;
      // m_hash2.insert(c);
    }
    
    m_currentwriteposition++;
   
  }
  //TODO: Finde längste Übereinstimmtung
  CNormalLinkedList<SCompressedHashItem, 0> m_unusedhashelements;
  void findLastPosition(unsigned char* ch, int start, int length);
protected:
  Array<SCompressorHeapElements> m_priorityheap;
  SCompressorHeapElements getHeapElement();
  void pushHeapElement(const SCompressorHeapElements& element);
  Array<SCompressorHuffmannNode> m_huffmannnodes;
  int m_rootnode;
  Array<SCompressedSymbol> m_symbols;
  SCompressedHashItems m_hash2;
  SCompressedHashItems m_hash4;
  SCompressedHashItems m_hash8;
public:
  CStream* m_output;
  TCompressorState m_state;
public:
  CCompressor();
  ~CCompressor();
  void beginDeflate(CStream* output);
  void endDeflate();
  void beginInflate(CStream* output);
  void endInflate();
  void write(unsigned char* data, int datasize);

};



#endif
