#ifndef UTIDEFRAGMENTATION_H
#define UTIDEFRAGMENTATION_H
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Uti3DCore.h"
#include "utiutility.h"
#include "utiarray.h"
#include "utimathbox.h"


const int UTI_LEVEL_SIZE = 32768;
#ifndef NDEBUG
//#define DEBUGDEFRAG
//#define DEBUGDEFRAGFULL
#endif

template <class T, int count> struct BaseArray_Level1
    #ifdef DEBUGDEFRAG
            : public CUtiOB
     #endif

    {
        enum {POINTER_COUNT = count};
    #ifdef DEBUGDEFRAG
         Array<T*> m_array;
         BaseArray_Level1()
         {
             m_array.setSize(POINTER_COUNT);
             m_array.fill(NULL);
         }

    #else
         T* m_array[POINTER_COUNT];
     #endif


    };


    template <class T, int count> struct BaseArray_Level
            #ifdef DEBUGDEFRAG
                    : public CUtiOB
             #endif
    {
        enum {ELEMENT_COUNT = count};
    #ifdef DEBUGDEFRAG
         Array<T> m_level0_array;
         BaseArray_Level()
         {
             m_level0_array.setSize(ELEMENT_COUNT);
         }

    #else
        T m_level0_array[ELEMENT_COUNT];
     #endif
       // unsigned char m_restfill[count-ELEMENT_COUNT*sizeof(T)];
    };

template <class T, int level1_pow, int level0_pow> class Defrag_SegmentArray : public CUtiOB
{
protected:

   protected:
          enum {LEVEL1_COUNT = intpower<2, level1_pow>::m_value,
                LEVEL0_COUNT = intpower<2, level0_pow>::m_value,
                SUM_SHIFT = level1_pow+level0_pow};

 typedef BaseArray_Level<T, LEVEL0_COUNT> LEVEL0;
          typedef BaseArray_Level1<LEVEL0, LEVEL1_COUNT > LEVEL1;


          int array_size;
          Array<LEVEL1* > m_basearray_level2;
          int m_level2size;
          int m_level1chunks;
#ifdef DEBUGDEFRAGFULL
        Array<T> m_debugarray;
#endif
   public:


      inline int size() const
      {
         return array_size;
      }
      Defrag_SegmentArray(int size) : CUtiOB()
      {

         init();
         setSize(size);


      }
      Defrag_SegmentArray() : CUtiOB()
      {
        init();

      }
      void init()
      {
          array_size = 0;
          m_basearray_level2.setSize(1);
          LEVEL1* l = new LEVEL1();
          m_basearray_level2[0] = l;
          LEVEL0* li = new LEVEL0();
          l->m_array[0]=li;
          m_level2size = 1;
          m_level1chunks = 1;
      }

      void toArrayHighSpeed(Array<T>& t)
      {
          t.setSize(size());
          toArrayHighSpeed(t.array, size());

      }

      void toArrayHighSpeed(T* t_array, int count)
      {
          int visitedchunks = 0;
          int writeposition=0;

         // t.setSize(size());
          assert(count <= size());
          int s = count;
          int locallevel2size = calcLevel2Size(s);
          int locallevel1chunks = calcLevel1Chunks(s);
#ifdef DEBUGDEFRAG
      int writecount = 0;
#endif


          for (int i = 0; i < locallevel2size;i++) {
               LEVEL1* l = m_basearray_level2[i];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(l);
 #endif
               for (int j = 0; j < LEVEL1_COUNT; j++) {
                   if (visitedchunks >= locallevel1chunks) {
                      assert(i == locallevel2size-1);
                       break;
                   };
                   LEVEL0*l1 = l->m_array[j];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(l1);
 #endif
                   if (writeposition+LEVEL0_COUNT <= s) {
#ifdef DEBUGDEFRAG
                       assert(l1->m_level0_array.size() == LEVEL0_COUNT);
                       memcpy(&t_array[writeposition], l1->m_level0_array.array, LEVEL0_COUNT*sizeof(T));
                       writecount += LEVEL0_COUNT;
#else
                       memcpy(&t_array[writeposition], l1->m_level0_array, LEVEL0_COUNT*sizeof(T));
#endif
                       //memcpy(&t.array[writeposition], l1->m_level0_array.array, LEVEL0_COUNT*sizeof(T));
                       writeposition += LEVEL0_COUNT;

                   } else {
                       if (writeposition >= s) {
                          assert(i == locallevel2size-1);
                           break;
                       };
                       int rest = s-writeposition;
                       assert(rest <= LEVEL0_COUNT);
                      // memcpy(&t.array[writeposition], l1->m_level0_array.array, rest*sizeof(T));

#ifdef DEBUGDEFRAG
 memcpy(&t_array[writeposition], l1->m_level0_array.array, rest*sizeof(T));
 writecount += rest;
#else
 memcpy(&t_array[writeposition], l1->m_level0_array, rest*sizeof(T));
 #endif
                       assert(i == locallevel2size-1);
                       break;

                   }

               }
          }
          #ifdef DEBUGDEFRAG
          assert(writecount == count);
           #endif
      }

      ~Defrag_SegmentArray()
      {
          int l2;
          int l1index;
          for (int i = 0; i < m_level1chunks;i++) {
              l2 = i >> level1_pow;
              l1index = i & (LEVEL1_COUNT-1);
               LEVEL1* l = m_basearray_level2[l2];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(l);
 #endif
                   LEVEL0* ldel = l->m_array[l1index];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(ldel);
 #endif

                   l->m_array[l1index] = NULL;
                   delete ldel;
          }
          for (int i = 0; i < m_level2size;i++) {
              LEVEL1* l = m_basearray_level2[i];
#ifdef DEBUGDEFRAG
     CHECKEXISTANCE(l);
#endif
              delete l;
              m_basearray_level2[i] = NULL;
          }
      }
      inline int calcLevel1Chunks(int size)
      {
          return (size >> level0_pow)+1;
      }
      inline int calcLevel2Size(int size)
      {
          return (size >> SUM_SHIFT)+1;
      }

      void adjustLevel2Size(int i)
      {
          while (m_basearray_level2.size() < i) {
              m_basearray_level2.setSizeCopyHighSpeed(m_basearray_level2.size()*2);
          }
          while (m_basearray_level2.size() >= i*4 && m_basearray_level2.size() >= 4) {
              m_basearray_level2.setSizeCopyHighSpeed(m_basearray_level2.size()/2);
          }

          m_level2size = i;
      }

      void setSize(int size) {
         //utiassert(array == NULL || dynamic == true);
          CHECKEXISTANCE;
          int newlevel1chucks =calcLevel1Chunks(size);
          int newlevel2size = calcLevel2Size(size);
          int l2, l1index;
          if (m_level1chunks ==newlevel1chucks)  {
#ifdef DEBUGDEFRAGFULL
        m_debugarray.setSizeCopy(size);
#endif
              array_size = size;
              return;
          }
          if (array_size > size) {
              for (int i = newlevel1chucks; i < m_level1chunks;i++) {
                  l2 = i >> level1_pow;
                  l1index = i & (LEVEL1_COUNT-1);
                   LEVEL1* l = m_basearray_level2[l2];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(l);
 #endif
                   LEVEL0* ldel = l->m_array[l1index];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(ldel);
 #endif

                   l->m_array[l1index] = NULL;
                   delete ldel;
              }



              for (int i = newlevel2size; i < m_level2size;i++) {
                  LEVEL1* l = m_basearray_level2[i];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(l);
 #endif
                  delete l;
                  m_basearray_level2[i] = NULL;
              }
              adjustLevel2Size(newlevel2size);
              m_level2size = newlevel2size;
              m_level1chunks = newlevel1chucks;

          } else {
              adjustLevel2Size(newlevel2size);
              for (int i = m_level2size; i < newlevel2size;i++) {
#ifdef DEBUGDEFRAG
     assert(m_basearray_level2[i] == NULL);
 #endif
                  m_basearray_level2[i] = new LEVEL1();
              }
              for (int i = m_level1chunks; i < newlevel1chucks;i++) {
                   l2 = i >> level1_pow;
                   l1index = i & (LEVEL1_COUNT-1);
                   LEVEL1* l = m_basearray_level2[l2];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(l);
 #endif
#ifdef DEBUGDEFRAG
     assert(l->m_array[l1index] == NULL);
 #endif
     l->m_array[l1index] = new LEVEL0();
              }
              m_level2size = newlevel2size;
              m_level1chunks = newlevel1chucks;

          }
          array_size = size;
#ifdef DEBUGDEFRAGFULL
        m_debugarray.setSizeCopy(size);
#endif
      }
      void setSizeCopy(int size) {
          setSize(size);
      }
      void setSizeCopyHighSpeed(int size) {
          setSize(size);
      }
      inline T &operator[](unsigned int p)
      {
         assert((int)p >= 0 && (int)p < array_size);


         LEVEL1* level1 = m_basearray_level2[p >> SUM_SHIFT];

#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(level1);
 #endif
         unsigned int l1pos = (p >> level0_pow) & (LEVEL1_COUNT-1);
         assert(l1pos < LEVEL1_COUNT);
         LEVEL0* level0 =  level1->m_array[l1pos];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(level0);
 #endif

         unsigned int l0pos = (p) & (LEVEL0_COUNT-1);
         assert(l0pos < LEVEL0_COUNT);

#ifdef DEBUGDEFRAGFULL
         assert(m_debugarray[p] == level0->m_level0_array[l0pos]);
#endif
         return level0->m_level0_array[l0pos];

      }
      inline const T &operator[](unsigned int p) const
      {
         assert((int)p >= 0 && (int)p < array_size);

         LEVEL1* level1 = m_basearray_level2[p >> SUM_SHIFT];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(level1);
 #endif
         unsigned int l1pos = (p >> level0_pow) & (LEVEL1_COUNT-1);
         assert(l1pos < LEVEL1_COUNT);
         LEVEL0* level0 = level1->m_array[l1pos];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(level0);
 #endif


         unsigned int l0pos = (p) & (LEVEL0_COUNT-1);
         assert(l0pos < LEVEL0_COUNT);
#ifdef DEBUGDEFRAGFULL
         assert(m_debugarray[p] == level0->m_level0_array[l0pos]);
#endif
         return level0->m_level0_array[l0pos];

      }
      inline const T &get(unsigned int p) const
      {
         assert((int)p >= 0 && (int)p < array_size);

         LEVEL1* level1 = m_basearray_level2[p >> SUM_SHIFT];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(level1);
 #endif
         unsigned int l1pos = (p >> level0_pow) & (LEVEL1_COUNT-1);
         assert(l1pos < LEVEL1_COUNT);
         LEVEL0* level0 = level1->m_array[l1pos];
#ifdef DEBUGDEFRAG
         CHECKEXISTANCE(level0);
 #endif
         unsigned int l0pos = (p) & (LEVEL0_COUNT-1);
         assert(l0pos < LEVEL0_COUNT);
#ifdef DEBUGDEFRAGFULL
         assert(m_debugarray[p] == level0->m_level0_array[l0pos]);
#endif
         return level0->m_level0_array[l0pos];

      }

};

template <class T, class Ar=Array<T> > struct UtiArrayAccessDefragGrow : public ArrayAccess<T>
{
  protected:
   Ar& m_array;
    int m_position;
  public:
    explicit UtiArrayAccessDefragGrow(Ar& a,int start=0) : m_array(a)
    {
      m_position =start;
      utiassert(start < m_array.size());
    };

    virtual int add(const T &el)
    {
        if (m_array.size() <= m_position) {

      m_array.setSizeCopy(m_array.size()+1024);
            }

       m_array[m_position] = el;
      m_position++;
      return m_position-1;
    };
    virtual int grow()
    {
        if (m_array.size() <= m_position) {

              m_array.setSizeCopy(m_array.size()+1024);
         }

//	m_array[m_position] = el;
       m_position++;
       return m_position-1;
    };
	inline int fast_add(const T &el)
	{
		if (m_array.size() <= m_position) {

			m_array.setSizeCopy(m_array.size() + 1024);
		}

		m_array[m_position] = el;
		m_position++;
		return m_position - 1;
	};
	inline int fast_grow()
	{
		if (m_array.size() <= m_position) {

			m_array.setSizeCopy(m_array.size() + 1024);
		}

		//	m_array[m_position] = el;
		m_position++;
		return m_position - 1;
	};
    virtual int size() const {return m_position;}
    virtual void reset(int maxsize, int start=0) {
        m_position = start;
        if (m_array.size() > maxsize) {
           // m_array.setSizeCopy(maxsize);

                 m_array.setSizeCopy(maxsize);

        }
    }
};

#endif // UTIDEFRAGMENTATION_H
