#ifndef UTIINTVECTOR_H
#define UTIINTVECTOR_H
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Uti3DMath.h"
#include "UtiType.h"


struct CIVector2 {
     typedef int ValueType;
   int x;
   int y;
   inline CIVector2(){};
   inline void set(int _x, int _y){x = _x; y = _y;};
   inline CIVector2(int _x, int _y){x = _x; y = _y;};
   inline bool operator==(const CIVector2& v) const { return x == v.x&& y == v.y;};
   inline bool operator!=(const CIVector2& v) const { return x != v.x|| y != v.y;};
   inline bool operator<(const CIVector2&v) const {
     if (x < v.x) return true;
     if (x == v.x){
       return y< v.y;
     } else{
       return false;
     };
     return false;
   };
   inline CIVector2 operator+(const CIVector2& v) const
   {
     CIVector2 r;
     r.x = x + v.x;
     r.y = y + v.y;
     return r;
   };
   inline CIVector2 operator*(int v) const
   {
     CIVector2 r;
     r.x = x *v ;
     r.y = y *v;
     return r;
   };
   inline CIVector2 operator-(const CIVector2& v) const
   {
     CIVector2 r;
     r.x = x - v.x;
     r.y = y - v.y;
     return r;
   };
   inline CIVector2& operator+=(const CIVector2& v)
   {

     x = x + v.x;
     y = y + v.y;
     return *this;
   };
   inline CIVector2& operator*=(int t)
   {

	   x = x*t;
	   y = y *t;
	   return *this;
   };
   inline CIVector2& operator-=(const CIVector2& v)
   {

     x = x - v.x;
     y = y - v.y;
     return *this;
   };
   inline CIVector2 operator*(const CIVector2& v) const
   {
     CIVector2 r;
     r.x = x*v.x;
     r.y = y*v.y;
     return r;
   };
   inline CIVector2 shift(int d)
   {
     CIVector2 r;
     r.x = x >> d;
     r.y = y >> d;
     return r;
   }
   inline int absDistanceTo(const CIVector2& v) const
   {
       return umax(abs(v.x-x), abs(v.y-y));
   }
   inline int getMaxAxis()
   {
       return abs(x) > abs(y) ? 0 : 1;
   }

   inline int getMinAxis()
   {
       return abs(x) < abs(y) ? 0 : 1;
   }

   inline int &operator[](unsigned int p)
   {
   utiassert(/*p >= 0 &&*/ p < 2);
   return (&x)[p];
   }
   inline const int &operator[](unsigned int p) const
   {
   utiassert(/*p >= 0 &&*/ p < 2);
       return (&x)[p];
   }
   HResult Read(CStream & Stream, int Flags, int version);
   HResult Write(CStream & Stream, int Flags, int version);
};

struct CIVector3 {
    typedef int ValueType;
   int x;
   int y;
   int z;
   inline CIVector3(){};
   inline CIVector3(const CVector3& v) {
       x = round(v.x);
       y = round(v.y);
       z = round(v.z);
   };

   inline CIVector3(int _x, int _y, int _z){x = _x; y = _y; z = _z;};
   inline bool operator==(const CIVector3& v) const { return x == v.x &&  y == v.y && z == v.z;};
    inline bool operator!=(const CIVector3& v) const { return !operator==(v);};
   inline void set(int _x, int _y, int _z){x = _x; y = _y; z = _z;};
   inline CIVector2 xy() const {
       return CIVector2(x,y);
   }
   inline CIVector2 xz() const {
       return CIVector2(x,z);
   }
   inline CIVector2 yz() const {
       return CIVector2(y,z);
   }

   inline CIVector3 operator+(const CIVector3& v)const
   {
     CIVector3 r;
     r.x = x + v.x;
     r.y = y + v.y;
     r.z = z + v.z;
     return r;
   };
   inline const CIVector3 &operator+=(const CIVector3& v)
   {

     x += v.x;
     y += v.y;
     z += v.z;
     return *this;
   };
   inline const CIVector3 &operator*=(int v)
   {

	   x *= v;
	   y *= v;
	   z *= v;
	   return *this;
   };

   inline const CIVector3 &operator-=(const CIVector3& v)
   {

     x -= v.x;
     y -= v.y;
     z -= v.z;
     return *this;
   };
   inline CIVector3 operator-(const CIVector3& v) const
   {
     CIVector3 r;
     r.x = x - v.x;
     r.y = y - v.y;
     r.z = z - v.z;
     return r;
   };
   inline CIVector3 operator*(int v) const
   {
     CIVector3 r;
     r.x = x*v;
     r.y = y*v;
     r.z = z*v;
     return r;
   };
   inline CIVector3 operator*(const CIVector3& v) const
   {
     CIVector3 r;
     r.x = x*v.x;
     r.y = y*v.y;
     r.z = z*v.z;
     return r;
   };
   inline CIVector3 operator/(int v) const
   {
	   CIVector3 r;
	   r.x = x/v;
	   r.y = y/v;
	   r.z = z/v;
	   return r;
   };
   inline const CIVector3 &operator/=(int v)
   {

	   x /= v;
	   y /= v;
	   z /= v;
	   return *this;
   };
   inline bool operator<(const CIVector3& v)const
   {
     if (x < v.x) return true;
     if (x != v.x) return false;

     if (y < v.y) return true;
     if (y != v.y) return false;
     if (z < v.z) return true;

     return false;
   };
   inline int distanceTo(const CIVector3& v) const
   {
     int d1 = abs(x -v.x);
     int d2 = abs(y -v.y);
     int d3 = abs(z - v.z);
     d1 = umax(d1, umax(d2, d3));
     return d1;
   }
   inline CIVector3 getShift(int d)const
   {
     CIVector3 r;
     r.x = x >> d;
     r.y = y >> d;
     r.z = z >> d;
     return r;
   }
   inline int getMaxAxis() const
   {
       if (abs(x) > abs(y)) {
           if (abs(z) > abs(x)) {
               return 2;
           }else {
               return 0;
           }
       } else {
           if (abs(z) > abs(y)) {
               return 2;
           }else {
               return 1;
           }
       }

   }

   inline int getMinAxis()
   {
       if (abs(x) < abs(y)) {
           if (abs(z) < abs(x)) {
               return 2;
           }else {
               return 0;
           }
       } else {
           if (abs(z) < abs(y)) {
               return 2;
           }else {
               return 1;
           }
       }
   }
   HResult Read(CStream & Stream, int Flags, int version);
   HResult Write(CStream & Stream, int Flags, int version) const;
   inline int &operator[](unsigned int p)
   {
   utiassert(/*p >= 0 &&*/ p < 3);
   return (&x)[p];
   }
   inline const int &operator[](unsigned int p) const
   {
   utiassert(/*p >= 0 &&*/ p < 3);
       return (&x)[p];
   }
};

inline CIVector2 bintrunc(const CVector2& f)
{
    return CIVector2(bintrunc(f.x), bintrunc(f.y));
}

inline CIVector3 bintrunc(const CVector3& f)
{
    return CIVector3(bintrunc(f.x), bintrunc(f.y), bintrunc(f.z));
}




inline int umax(const CIVector3& a)
{
    int res = a.x;
    if (a.y > res) res = a.y;
    if (a.z > res) res = a.z;
   return res;
}


inline int umin(const CIVector3& a)
{
    int res = a.x;
    if (a.y < res) res = a.y;
    if (a.z < res) res = a.z;
   return res;
}

inline int umax(const CIVector2& a)
{
    int res = a.x;
    if (a.y > res) res = a.y;
   return res;
}


inline int umin(const CIVector2& a)
{
    int res = a.x;
    if (a.y < res) res = a.y;
   return res;
}


inline void uti_moddiv(const CIVector2& value, const CIVector2& divider, CIVector2& mod, CIVector2& res)
{
        uti_moddiv(value.x, divider.x, mod.x, res.x);
        uti_moddiv(value.y, divider.y, mod.y, res.y);
}


inline void uti_moddiv(const CIVector3& value, const CIVector3& divider, CIVector3& mod, CIVector3& res)
{
	uti_moddiv(value.x, divider.x, mod.x, res.x);
	uti_moddiv(value.y, divider.y, mod.y, res.y);
	uti_moddiv(value.z, divider.z, mod.z, res.z);
}


template<class T> inline T umax(T a, T b, T c)
{
    T res = a;
    if (b > res) res = b;
    if (c > res) res = c;
   return res;
}

template<class T> inline T umax(T a, T b, T c, T d)
{
    T res = a;
    if (b > res) res = b;
    if (c > res) res = c;
    if (d > res) res = d;
   return res;
}

template<class T> inline T umax(T a, T b, T c, T d, T e)
{
    T res = a;
    if (b > res) res = b;
    if (c > res) res = c;
    if (d > res) res = d;
    if (e > res) res = e;
   return res;
}

template<class T> inline T umin(T a, T b, T c)
{
    T res = a;
    if (b > res) res = b;
    if (c > res) res = c;
   return res;
}

template<class T> inline T umin(T a, T b, T c, T d)
{
    T res = a;
    if (b > res) res = b;
    if (c > res) res = c;
    if (d > res) res = d;
   return res;
}

template<class T> inline T umin(T a, T b, T c, T d, T e)
{
    T res = a;
    if (b > res) res = b;
    if (c > res) res = c;
    if (d > res) res = d;
    if (e > res) res = e;
   return res;
}

struct CIVector4 {
     typedef int ValueType;
   int x;
   int y;
   int z;
   int w;
   inline CIVector4(){};
   inline CIVector4(int _x, int _y, int _z, int _w){x = _x; y = _y; z = _z; w = _w;};
   inline bool operator==(const CIVector4& v) const {  return x == v.x &&  y == v.y && z == v.z && w == v.w;};
   inline CIVector2 xy()const{return CIVector2(x, y);};
   inline CIVector3 xyz()const {return CIVector3(x,y,z);};
   inline int &operator[](unsigned int p)
   {
   utiassert(/*p >= 0 &&*/ p < 4);
   return (&x)[p];
   }
   inline const int &operator[](unsigned int p) const
   {
   utiassert(/*p >= 0 &&*/ p < 4);
       return (&x)[p];
   }
   HResult Read(CStream & Stream, int Flags, int version);
   HResult Write(CStream & Stream, int Flags, int version);
};

inline CIVector4 bintrunc(const CVector4& f)
{
    return CIVector4(bintrunc(f.x), bintrunc(f.y), bintrunc(f.z), bintrunc(f.w));
}

inline int umax(const CIVector4& a)
{
    int res = a.x;
    if (a.y > res) res = a.y;
    if (a.z > res) res = a.z;
    if (a.w > res) res = a.w;
   return res;
}


inline int umin(const CIVector4& a)
{
    int res = a.x;
    if (a.y < res) res = a.y;
    if (a.z < res) res = a.z;
    if (a.w < res) res = a.w;
   return res;
}


typedef CIVector2 SPoint;

inline CVector3 intVecToFloat(const CIVector3& v)
{
  CVector3 r;
  r.x = (float)(v.x);
  r.y = (float)(v.y);
  r.z = (float)(v.z);

  return r;
};

inline CVector2 intVecToFloat(const CIVector2& v)
{
  CVector2 r;
  r.x = (float)(v.x);
  r.y = (float)(v.y);

  return r;
};



inline CIVector3 floatVecToInt(const CVector3& v)
{
  CIVector3 r;
  r.x = (int)round(v.x);
  r.y = (int)round(v.y);
  r.z = (int)round(v.z);

  return r;
};

inline CIVector2 floatVecToInt(const CVector2& v)
{
  CIVector2 r;
  r.x = (int)round(v.x);
  r.y = (int)round(v.y);

  return r;
};

typedef struct SHitResult
{
    S3DPoint Vector;
    S3DPoint CheckVector;
    CUtiOB** Ignore;
    int IgnoreCount;
    unsigned int ObjectMask;
    CUtiOB* object;
   CUtiOB* Parent;
    S3DPoint point;
    CIVector3 voxelposition;
    utifloat distance;
    bool HitVoxel;
    bool HitRTS;
    bool nobsp;
    bool NoJoints;
    bool SimpleTest;
    CUtiPHMaterial* materialid;
    ReferenceUtiOB* m_currentblockdata;
    ReferenceUtiOB* m_syntheticpainterpainter;
        uti_refptr<ReferenceUtiOB> m_hitdata;

        Array<void*>* simplehashmap;;
    SHitResult()
    {
        Ignore = NULL;
        materialid = NULL;
        IgnoreCount = 0;
        ObjectMask = 0;
        object = NULL;
        Parent = NULL;
        distance = 0;
        nobsp = false;
        NoJoints = false;
        SimpleTest = false;
        voxelposition.set(0,0,0);
        HitVoxel = false;
        HitRTS = false;
        m_currentblockdata = NULL;
        m_syntheticpainterpainter = NULL;
        simplehashmap = NULL;
        //point.setPosition(0,0,0);

    }
} *PHitResult;



struct CIVector8_2 {
   unsigned char x;
   unsigned char y;
   inline CIVector8_2(){};
   inline CIVector8_2(unsigned char _x, unsigned char _y){x = _x; y = _y;};
};

struct CIVector8_3 {
   unsigned char x;
   unsigned char y;
   unsigned char z;
   inline CIVector8_3(){};
   inline CIVector8_3(const CIVector3& v)
   {
       assert(v.x>= 0 && v.x < 256);
       assert(v.y>= 0 && v.y < 256);
       assert(v.z>= 0 && v.z < 256);
       x = v.x;
       y = v.y;
       z = v.z;
   }

   inline CIVector8_3(unsigned char _x, unsigned char _y, unsigned char _z){x = _x; y = _y; z = _z;};
   inline CIVector8_3& operator=(const CIVector3& v)
   {
       assert(v.x>= 0 && v.x < 256);
       assert(v.y>= 0 && v.y < 256);
       assert(v.z>= 0 && v.z < 256);
       x = v.x;
       y = v.y;
       z = v.z;
       return*this;
   }
   inline CIVector3 toVector()const
   {
       return CIVector3(x, y, z);
   }
};

struct CIVector8_4 {
   unsigned char x;
   unsigned char y;
   unsigned char z;
   unsigned char w;
   inline CIVector8_4(){};
   inline CIVector8_4(unsigned char _x, unsigned char _y, unsigned char _z, unsigned char _w){x = _x; y = _y; z = _z; w = _w;};
   inline CIVector8_4& operator=(const CIVector4 &v)
   {
       assert(v.x>= 0 && v.x < 256);
       assert(v.y>= 0 && v.y < 256);
       assert(v.z>= 0 && v.z < 256);
       assert(v.w>= 0 && v.w < 256);
       x = v.x;
       y = v.y;
       z = v.z;
       w = v.w;
       return *this;
   }
   inline CIVector4 toVector()const
   {
       return CIVector4(x, y, z,w );
   }
};

struct CIVector16_2 {
   short x;
   short y;
   inline CIVector16_2(){};
   inline CIVector16_2(short _x, short _y){x = _x; y = _y;};
};

struct CIVector16_3 {
   short x;
   short y;
   short z;
   inline CIVector16_3(){};
   inline CIVector16_3(short _x, short _y, short _z){x = _x; y = _y; z = _z;};
   inline CIVector16_3& operator=(const CIVector3& v)
   {
   assert(v.x > -16384 && v.x < 16384);
   assert(v.y > -16384 && v.y < 16384);
   assert(v.z > -16384 && v.z < 16384);

       x = v.x;
       y = v.y;
       z = v.z;
       return *this;
   }
   inline CIVector3 toVector()const
   {
       return CIVector3(x, y, z);
   }
};

struct CIVector16_4 {
   short x;
   short y;
   short z;
   short w;
   inline CIVector16_4(){};
   inline CIVector16_4(short _x, short _y, short _z, short _w){x = _x; y = _y; z = _z; w = _w;};
   inline CIVector16_4& operator=(const CIVector4 &v)
   {

       x = v.x;
       y = v.y;
       z = v.z;
       w = v.w;
       return *this;
   }
   inline CIVector4 toVector()const
   {
       return CIVector4(x, y, z,w );
   }
};


float intern_inoise(const CIVector2& position);
float intern_inoise(const CIVector3& position);

inline CIVector3 bindivide(const CIVector3& v, int a)
{
    return CIVector3(bindivide(v.x, a),bindivide(v.y,a), bindivide(v.z,a));
}

CUtiString convertToString(const CIVector4& v);
CUtiString convertToString(const CIVector3& v);
CUtiString convertToString(const CIVector2& v);

CIVector2 parseIVector2(const CUtiString& str);
CIVector3 parseIVector3(const CUtiString& str);
CIVector4 parseIVector4(const CUtiString& str);


inline CIVector2 umin(const CIVector2& a, const CIVector2& b)
{
   return CIVector2(a.x < b.x ? a.x : b.x,
                    a.y < b.y ? a.y : b.y);
}

inline CIVector2 umax(const CIVector2& a, const CIVector2& b)
{
    return CIVector2(a.x > b.x ? a.x : b.x,
                     a.y > b.y ? a.y : b.y);
}

inline CIVector3 umin(const CIVector3& a, const CIVector3& b)
{
   return CIVector3(a.x < b.x ? a.x : b.x,
                    a.y < b.y ? a.y : b.y,
                    a.z < b.z ? a.z : b.z);
}

inline CIVector3 umax(const CIVector3& a, const CIVector3& b)
{
    return CIVector3(a.x > b.x ? a.x : b.x,
                     a.y > b.y ? a.y : b.y,
                     a.z > b.z ? a.z : b.z);
}

inline CIVector4 umin(const CIVector4& a, const CIVector4& b)
{
   return CIVector4(a.x < b.x ? a.x : b.x,
                    a.y < b.y ? a.y : b.y,
                    a.z < b.z ? a.z : b.z,
                    a.w < b.w ? a.w : b.w);
}

inline CIVector4 umax(const CIVector4& a, const CIVector4& b)
{
    return CIVector4(a.x > b.x ? a.x : b.x,
                     a.y > b.y ? a.y : b.y,
                     a.z > b.z ? a.z : b.z,
                     a.w > b.w ? a.w : b.w);
}

void uti_readValue(CIVector3 &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CIVector3& std=CIVector3(0, 0, 0));
void uti_writeValue(CIVector3 &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);
void uti_readValue(CIVector2 &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CIVector2& std=CIVector2(0, 0));
void uti_writeValue(CIVector2 &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

void uti_readValue(THotKey &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const THotKey& std);
void uti_writeValue(THotKey &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

#endif // UTIINTVECTOR_H
