/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "utikeycodes.h"
#include "Uti3DCore.h"
#include "UtiString.h"
#include <map>
#include <assert.h>

class KeyCodeTable
{
MAKE_SINGLETON(KeyCodeTable);
public:
   std::map<CUtiString, int> m_stringtokey;
   std::map<int, CUtiString> m_keytostring;
};

MAKE_SINGLETON_IMPL(KeyCodeTable);

int UTI_KEY_stringToKey(const wchar_t* str)
{
   KeyCodeTable* t = KeyCodeTable::getInstance();
   std::map<CUtiString, int>::iterator it = t->m_stringtokey.find(str);
  if (it == t->m_stringtokey.end()) return -1;
  return it->second;
}
const wchar_t* UTI_KEY_keyToString(int i)
{
KeyCodeTable* t = KeyCodeTable::getInstance();
   std::map<int, CUtiString>::iterator it = t->m_keytostring.find(i);
  if (it == t->m_keytostring.end()) return L"NOCHAR";
  return it->second.c_str();
}

void UTI_KEY_registerKey(const wchar_t* name, int keycode)
{
KeyCodeTable* t = KeyCodeTable::getInstance();
   utiassert(t->m_stringtokey.find(name) == t->m_stringtokey.end());
   utiassert(t->m_keytostring.find(keycode) == t->m_keytostring.end());
   t->m_stringtokey.insert(std::make_pair(name, keycode));
   t->m_keytostring.insert(std::make_pair(keycode, name));
}
