/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "utilexer.h"
#include "utiarray.h"

UtiLexer::UtiLexer()
{

    m_pos = 0;
    m_line = 0;
    m_size = 0;
    m_lispsymbols = false;

    m_parser_line = 1;
    m_parser_lastlinepos = 0;
    m_parser_index = 0;
     m_pascalcomment= true;
}


bool UtiLexer::isInternLetter(wchar_t a)
{
    if (m_lispsymbols) {
    return IsLetter(a) || (a == L'?') || (a==L'-') || (a==L'+') || (a == L'*') || (a == L'_')|| (a == L'#') || (a == L'/') || (a == L'=') || (a == L'<') || (a == L'>') || (a == L':');
    } else {
        return IsLetter(a);
    }
}

UtiLexer::UtiLexer(const UtiLexer& lexer)
{
    m_stream = lexer.m_stream;
    m_pos = lexer.m_pos;
    m_size = lexer.m_size;
    m_line = lexer.m_line;
    m_lispsymbols = false;
    m_parser_line = 1;
    m_parser_lastlinepos = 0;
    m_parser_index = 0;
}


UtiLexer::UtiLexer(const CUtiString& text)
{
    m_stream = text;
    m_pos = 0;
    m_size = m_stream.size();
    m_line = 0;
    m_lispsymbols = false;
    m_parser_line = 1;
    m_parser_lastlinepos = 0;
    m_parser_index = 0;
}

UtiLexer& UtiLexer::operator=(const UtiLexer& lexer)
{
    m_stream = lexer.m_stream;
    m_pos = lexer.m_pos;
    m_size = lexer.m_size;
    m_line = lexer.m_line;
    m_lispsymbols = lexer.m_lispsymbols;
    return *this;
}

bool UtiLexer::IsInteger(U_CHAR *str, int length)
{
  if (length == 0) return false;
  int start = 0;
  if (str[0] == '-') start = 1;
  for (int i = start; i < length; i++)
  {
    if (IsCharacter(str[i]))
    {
      if (!IsNumber(str[i]))
      {
        return(false);

      };
    };
  };
  return(true);
};

bool UtiLexer::IsNumber(wchar_t a)
{
  return (a >= UTEXT('0') && a <= UTEXT('9'));
};
bool UtiLexer::IsLetter(wchar_t a)
{
  return ((a >= UTEXT('A') && a <= UTEXT('Z')) || (a >= UTEXT('a') && a <= UTEXT('z')));
};
bool UtiLexer::IsUnderLine(wchar_t a)
{
  return (a == UTEXT('_'));
};
bool UtiLexer::IsCharacter(wchar_t a)
{
  return ((a != UTEXT(' ')) && (a > 30));
};
bool UtiLexer::isInternBezeichner(wchar_t a)
{
  return (IsNumber(a) || isInternLetter(a) || IsUnderLine(a));
};
bool UtiLexer::IsBezeichner(wchar_t a)
{
  return (IsNumber(a) || IsLetter(a) || IsUnderLine(a));
};
bool UtiLexer::IsVergleich(wchar_t a)
{
  return (a == UTEXT('>') || a == UTEXT('<') || a == UTEXT('='));
};


bool UtiLexer::hasBezeichner(const CUtiString& text, const CUtiString& bez, std::set<std::tuple<CUtiString, int, int> > &positions)
{
    bool inbezeichner = false;
    int start = 0;
    int end = 0;
    bool found = false;
    for (int i = 0; i < text.size(); i++) {
        if (inbezeichner) {

            if (!isBezeichner(text[i])) {
                end = i;

                if (text(start, end-start) == bez) {
                    found = true;
                    positions.insert(std::make_tuple(bez, start, end));

                }

                inbezeichner = false;
            }
        } else {
            if (IsLetter(text[i])) {
                start = i;
                inbezeichner = true;
            }
        }

    }

    if (inbezeichner) {


            end = text.size();

            if (text(start, end-start) == bez){
                found = true;
                positions.insert(std::make_tuple(bez, start, end));

            }


    }

    return found;
}


void UtiLexer::setText(const CUtiString& text)
{
    m_stream = text;
    m_pos = 0;
    m_size = text.size();
    m_line = 0;
}

CUtiString UtiLexer::getText() const
{
    return m_stream;
}

void UtiLexer::setPos(int p)
{
   m_pos = p;
}

int UtiLexer::getPos() const
{
    return m_pos;
}
void UtiLexer::setLine(int l)
{
    m_line = l;
}
int UtiLexer::getLine() const
{
    return m_line;
}

bool UtiLexer::GetNextNumber(CUtiString& c)
{
        int i = SearchNextChar();
        if (i == -1) return false;
        int v = 0;
        if (m_stream[i] == UTEXT('$')) {
       c+= UTEXT('$');
           int v = 1;

           MoveToNextChar();
       for (i = m_pos+1; i < m_size ; i++) {
          if (!(IsNumber(m_stream[i]) || (toupper(m_stream[i]) == UTEXT('a'))
                                                      || (toupper(m_stream[i]) == UTEXT('b'))
                                                                                  || (toupper(m_stream[i]) == UTEXT('c'))
                                                                                  || (toupper(m_stream[i]) == UTEXT('d'))
                                                                                  || (toupper(m_stream[i]) == UTEXT('e'))
                                                                                  || (toupper(m_stream[i]) == UTEXT('f')))) break;
                if (!(v < 255)) break;
                c+= m_stream[i];
                v++;
           };
           m_pos += v;
           c[v] = 0;
       return(true);
        };
        bool minus = false;
        if (m_stream[i] == UTEXT('-')) {
                c+= UTEXT('-');
                v = 1;
                i++;minus = true;

        }
        if (!IsNumber(m_stream[i])) {
       return(false);
        };
        MoveToNextChar();
        if (minus)  m_pos++;
    for (i = m_pos; i < m_size ; i++) {
       if (!(IsNumber(m_stream[i]) || (m_stream[i] == UTEXT('.')))) break;
                if (!(v < 255)) break;
                c+= m_stream[i];
           v++;
    };
        m_pos += v;
        if (minus) m_pos--;
        return(true);
};

int UtiLexer::GetIntConstant()
{
CUtiString value;
   bool b = GetNextIntNumber(value);
   assert(b != true);
   return GetIntValue(value);
}

int UtiLexer::GetIntValue(const CUtiString& name)
{
    return name.ToInt();
}

bool UtiLexer::GetNextIntNumber(CUtiString& c)
{
        int i = SearchNextChar();
        if (i == -1) return false;
        int v = 0;
        if (m_stream[i] == UTEXT('$')) {
       c+= UTEXT('$');
           int v = 1;

           MoveToNextChar();
       for (i = m_pos+1; i < m_size ; i++) {
          if (!(IsNumber(m_stream[i]) || (toupper(m_stream[i]) == UTEXT('a'))
                                                      || (toupper(m_stream[i]) == UTEXT('b'))
                                                                                  || (toupper(m_stream[i]) == UTEXT('c'))
                                                                                  || (toupper(m_stream[i]) == UTEXT('d'))
                                                                                  || (toupper(m_stream[i]) == UTEXT('e'))
                                                                                  || (toupper(m_stream[i]) == UTEXT('f')))) break;
                if (!(v < 255)) break;
                c+= m_stream[i];
                v++;
           };
           m_pos += v;

       return(true);
        };
        bool minus = false;
        if (m_stream[i] == UTEXT('-')) {
                c+= UTEXT('-');
                v = 1;
                i++;
                minus = true;

        }
        if (i >= m_stream.size() || !IsNumber(m_stream[i])) {
       return(false);
        };
        if (minus)  m_pos++;
        MoveToNextChar();

    for (i = m_pos; i < m_size ; i++) {
       if (!(IsNumber(m_stream[i]))) break;
                if (!(v < 255)) break;
                c+= m_stream[i];
         m_pos = i+1;
    };
        //m_pos += v;
        return(true);
};

bool UtiLexer::GetNextBezeichner(CUtiString &id)
{
        int i = SearchNextChar();
        if (i == -1) return false;
        if (!IsUnderLine(m_stream[i]) && !isInternLetter(m_stream[i])) {
                return(false);
        };
        StaticArray<wchar_t,255> c;/// wchar_t c[255];
    MoveToNextChar();
        int v = 0;
    for (i = m_pos; i < m_size; i++) {
                if (!isInternBezeichner(m_stream[i])) break;
                if (!(v < 255)) break;
       c[v] = m_stream[i];
           v++;
    };
        m_pos += v;
        c[v] = 0;
        id = c.getData();
        return(true);
};

bool UtiLexer::GetNextTyp(CUtiString &id)
{
        int i = SearchNextChar();
         StaticArray<wchar_t,255> c;
        if (!IsUnderLine(m_stream[i]) && !isInternLetter(m_stream[i]) && !(m_stream[i] == UTEXT('^'))) {
                return(false);
        };
    MoveToNextChar();
        int v = 0;

        while (IsThisNextCharacter(UTEXT('^'))) {
           c[v]= m_stream[m_pos];
           v++;
           m_pos++;
        };
        int start = 0;
    for (i = m_pos; i < m_size; i++) {
                if (!IsCharacter(m_stream[i])
                        || (m_stream[i] == UTEXT(';'))
                        || (m_stream[i] == UTEXT('='))
                        || (m_stream[i] == UTEXT(':'))
                        || (m_stream[i] == UTEXT('/'))) break;
                if (!(v < 255)) break;
       c[v] = m_stream[i];
           v++;
           start++;
    };

        m_pos += start;
        c[v] = 0;
        id = c.getData();
        return(true);
};


bool UtiLexer::GetBezeichner(CUtiString& id)
{
        int i = SearchNextChar();
        if (i == -1) return false;
        if (!IsUnderLine(m_stream[i]) && !isInternLetter(m_stream[i])) {
                return(false);
        };
    MoveToNextChar();
         StaticArray<wchar_t,255> c;
        int v = 0;
    for (i = m_pos; i < m_size; i++) {
       if (!isInternBezeichner(m_stream[i])) break;
                if (!(v < 255)) break;
                c[v] = m_stream[i];
           v++;
    };
        c[v] = 0;
        id = c.getData();
        return(true);
};


void UtiLexer::printError(const CUtiString& data)
{
    int line = 1;
    int lastlinepos = 0;
    for (int i = 0; i < m_pos;i++) {
        if(m_stream[i] == 10){
            lastlinepos = i;
            line++;
        }
    }
    std::wcout << (UTISTR("Line: ")+ line+UTISTR("(")+(m_pos-lastlinepos)+UTISTR("): ")+data) << std::endl;
}


CIVector2 UtiLexer::getLexerPosition()
{

    while (m_parser_index < m_pos) {
        if(m_stream[m_parser_index] == 10){
            m_parser_lastlinepos = m_parser_index;
            m_parser_line++;
        }
        m_parser_index++;
    }
    return CIVector2(m_parser_line,(m_pos-m_parser_lastlinepos) );
}

int UtiLexer::SearchNextChar(int pos)
{
   int oldpos = m_pos;
   m_pos = pos;
   int res = SearchNextChar();
   m_pos = oldpos;
   return(res);
};

int UtiLexer::SearchNextChar()
{
        int result = -1;
        for (int i = m_pos; i < m_size; i++) {
                if (IsCharacter(m_stream[i])) {
           result = i;
                   break;
                };
        };
        if (result == -1) return -1;
        int i = result;
        if (m_pascalcomment) {
        if (m_stream[i] == UTEXT('{') || (i+1 < m_size && m_stream[i] == UTEXT('(') && m_stream[i+1] == UTEXT('*'))) {
                if (m_stream[i] == UTEXT('{')) {
                        i++;
                } else {
                        i += 2;
                };
                int i2 = i;
                bool found = false;

                for (int i = i2; i < m_size; i++) {
                   if (m_stream[i] == UTEXT('}') || (m_stream[i] == UTEXT(')') && m_stream[i-1] == UTEXT('*'))) {
              result = i+1;
                          found = true;
                      break;
                   };
                };
                if (found)
                result = SearchNextChar(result);
        };
        };
        if (i+1 < m_size && m_stream[i] == UTEXT('/') && m_stream[i+1] == UTEXT('/')) {
           i += 2;
           int i2 = i;
           bool found = false;

                for (int i = i2; i < m_size; i++) {
                   if (m_stream[i] == 10) {
              result = i+1;
                          found = true;

                      break;
                   };
                };
                if (found)
                result = SearchNextChar(result);

        };

        return(result);
};

void UtiLexer::MoveToNextChar()
{
   int i = SearchNextChar();
   if (i == -1) {
       m_pos = m_stream.size();
   } else {
      m_pos = i;
   };
};

U_CHAR UtiLexer::GetNextChar()
{
   MoveToNextChar();
   if (m_pos >= m_stream.size()) return L' ';
   U_CHAR res = m_stream[m_pos];
   m_pos++;
   return(res);
};

U_CHAR UtiLexer::GetNextCharNoMove()
{MoveToNextChar();
    if (m_pos >= m_stream.size()) return L' ';

   U_CHAR res = m_stream[m_pos];

   return(res);
};

void UtiLexer::MovePos(int i)
{
   m_pos += i;
};

bool UtiLexer::IsThisNextCharacterNeg(U_CHAR c)
{

    if (m_pos >= m_stream.size())return false;
  if (c == '-') {//Das hier besser mit einem richtigen Parser
    if (m_pos+1 < m_size) {
      return(m_stream[m_pos] == c && (!IsNumber(m_stream[m_pos+1])));
    } else {
      return(m_stream[m_pos] == c);
    };
  } else {
    return(m_stream[m_pos] == c);
  };
};


bool UtiLexer::IsThisNextCharacterSequence(const CUtiString& c)
{

    MoveToNextChar();

    if (m_pos+c.size()-1 >= m_stream.size())return false;
    for (int i = 0; i < c.size(); i++) {
        if (m_stream[m_pos+i] != c[i]) {
            return false;
        }
    }
    return true;
}

bool UtiLexer::GetNextCharSequence(const CUtiString& c)
{

    MoveToNextChar();

    if (m_pos+c.size()-1 >= m_stream.size())return false;
    for (int i = 0; i < c.size(); i++) {
        if (m_stream[m_pos+i] != c[i]) {
            assert(false);
            return false;
        }
    }
    m_pos += c.size();
    return true;
}


bool UtiLexer::IsThisNextCharacter(U_CHAR c)
{

    MoveToNextChar();
    if (m_pos >= m_stream.size())return false;
    wchar_t value = m_stream[m_pos];
    return(value == c);

};

bool UtiLexer::IsThisNextBezeichner(const CUtiString& id)
{
       MoveToNextChar();
       if (m_pos >= m_stream.size())return false;
        CUtiString str = id;
        const U_CHAR* c = str.c_str();
    int l = wcslen(c);
        int rest = m_size - m_pos;
    if (rest < l) return(false);
        for (int i = 0; i < l ; i++) {
                if (toupper(m_stream[m_pos+i]) != toupper(c[i])) {
           return(false);
                };
        };
        if (rest > l) {
                if (isInternBezeichner(c[0])) {
                   if (isInternBezeichner(m_stream[m_pos+l])) {
              return(false);
                   };
                };
                if ((c[0] == UTEXT('<') || c[0] == UTEXT('>')) && l == 1) {
           if (m_stream[m_pos+l] == UTEXT('>') ||
                           m_stream[m_pos+l] == UTEXT('=') ||
                           m_stream[m_pos+l] == UTEXT('<'))
                           return(false);
                };
        };
        return(true);

};
