/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef UTILEXER_H
#define UTILEXER_H

#include "Uti3DCore.h"
#include "UtiString.h"
#include "Uti3DMath.h"
#include "utiintvector.h"
#include <tuple>
#include <set>

class UtiLexer
{
protected:
    CUtiString m_stream;
    int m_pos;
    int m_size;
    int m_line;
    bool m_lispsymbols;
    int m_parser_line;
    int m_parser_lastlinepos ;
    int m_parser_index;
public:
    static bool IsInteger(U_CHAR *str, int length);
    static bool IsNumber(wchar_t a);
    static bool IsLetter(wchar_t a);
    static bool IsUnderLine(wchar_t a);
    static bool IsCharacter(wchar_t a);
    static bool IsBezeichner(wchar_t a);
    static bool IsVergleich(wchar_t a);

    bool m_pascalcomment;
    inline void setLispSymbols(bool l) {m_lispsymbols = l;}
    bool isEOF() {return m_pos >= m_size;}
    UtiLexer();
    UtiLexer(const CUtiString& text);
    UtiLexer(const UtiLexer& lexer);
    UtiLexer& operator=(const UtiLexer& lexer);
    bool isInternLetter(wchar_t a);
    bool isInternBezeichner(wchar_t a);
    void setText(const CUtiString& text);
    CUtiString getText() const;
    int getPos() const;
    void setPos(int p);
    void setLine(int l);
    int getLine() const;

    int GetIntConstant();
    int GetIntValue(const CUtiString& name);
    void MovePos(int i);
    bool GetBezeichner(CUtiString &id); //Gibt den n�chsten Bezeichner zur�ck ohne den Curser zu bewegen
    bool GetNextBezeichner(CUtiString &id);  //Gibt den n�chsten Bezeichner zur�ck
    bool GetNextTyp(CUtiString& id);
    U_CHAR GetNextChar();
    U_CHAR GetNextCharNoMove();
    bool GetNextNumber(CUtiString& c);
    bool GetNextIntNumber(CUtiString& c);
    int SearchNextChar();
    void MoveToNextChar();

    bool GetNextCharSequence(const CUtiString& c);
    bool IsThisNextCharacterSequence(const CUtiString& c);
    bool IsThisNextCharacter(U_CHAR c);  //Test ob das aktuelle Zeichen dem gegebenen entspricht
    bool IsThisNextCharacterNeg(U_CHAR c);
    bool IsThisNextBezeichner(const CUtiString& id);
    int Match(int &i, const CUtiString& c);
    int MatchChar(U_CHAR c);
    int SearchNextChar(int pos);
    void printError(const CUtiString& data);
    CIVector2 getLexerPosition();
    static bool hasBezeichner(const CUtiString& text, const CUtiString& bez, std::set< std::tuple< CUtiString, int, int> >& positions);
};

#endif // UTILEXER_H
