/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef UTILINKEDLIST_H
#define UTILINKEDLIST_H


#include "Uti3DCore.h"


struct CLinkedList_Controller
{
   void* m_next;
   void* m_prev;
//#ifdef _DEBUG
   bool m_inlist;
//#endif
   CLinkedList_Controller() : m_next(NULL), m_prev(NULL)

       , m_inlist(false)


   {};
   inline bool inList() {return m_inlist;};
};

template <class T, int c_index> class CLinkedList_Iterator
{
    public:
    typedef T value_type;
   typedef size_t size_type;
   typedef ptrdiff_t difference_type;
   typedef T* pointer;
   typedef const T* const_pointer;
   typedef T& reference;
   typedef const T& const_reference;
  // typedef random_access_iterator iterator_category;
   //typedef CList_Iterator<T, MemoryContainer, MemoryController> selftype;
    public:
       T* m_element;
       inline CLinkedList_Iterator() {
          m_element = NULL;
       }
       inline CLinkedList_Iterator(const CLinkedList_Iterator& iter)
       {
           m_element = iter.m_element;
       }
       inline explicit CLinkedList_Iterator(T * element) {
           m_element = element;
       }
       inline T & operator*() {return *m_element;};
       inline T * operator->() {return m_element;};
       inline CLinkedList_Iterator &operator++() {
     utiassert(m_element != NULL);
     m_element = (T*)m_element->clinkedlist_controller[c_index].m_next;
     return *this;

      };
       inline CLinkedList_Iterator operator++(int) {CLinkedList_Iterator tmp(*this); ++(*this); return tmp;};
       //inline CLinkedList_Iterator &operator--() {uassert(m_element != NULL); m_element = (T*)m_element->clinkedlist_controller[c_index]->m_prev;uassert(m_element != NULL); return *this;};
       //inline CLinkedList_Iterator operator--(int) {CLinkedList_Iterator tmp(*this); --(*this); return tmp;};

       inline bool operator==(const CLinkedList_Iterator &other) const
       {
           return (m_element == other.m_element);
       }
       inline bool operator!=(const CLinkedList_Iterator &other) const
       {
           return (m_element != other.m_element);
       }
       inline CLinkedList_Iterator& operator=(const CLinkedList_Iterator& iter)
       {
           m_element = iter.m_element;
           return *this;
       }

};

template <class T> struct CLinkedList_NoFree_MemoryController
{
    inline void insert(T* element) {

    }
    inline void delete_all(T* element) {

    };
    inline void extract_all(T* element) {

    };

};

template <class T> struct CLinkedList_Reference_MemoryController
{
    inline void insert(T* element) {
       element->addRef();
    }
    inline void delete_all(T* element) {
      element->decRef();
    };
    inline void extract_all(T* element) {
      element->decRef();
    };

};

template <class T> struct CLinkedList_Free_MemoryController
{
    inline void insert(T* element) {

    }
    inline void delete_all(T* element) {
      delete element;
    };
    inline void extract_all(T* element) {

    };

};

template <class T, int c_index, class MemoryController> struct CLinkedList
{
protected:
    MemoryController controller;
    T* m_first;
    T* m_last;

public:
   typedef T value_type;
   typedef size_t size_type;
   typedef ptrdiff_t difference_type;
   typedef T* pointer;
   typedef const T* const_pointer;
   typedef T& reference;
   typedef const T& const_reference;

   typedef CLinkedList_Iterator<T, c_index> iterator;
   iterator begin() { return iterator(m_first);};
   iterator end() {return iterator(NULL);};
protected:
    size_t m_size;
public:
   explicit CLinkedList() {
       m_first = NULL;
       m_last = NULL;
       m_size = 0;

   }
   ~CLinkedList()
   {
       clear();
   }
   reference front()
   {
       utiassert(m_first != NULL);
       return* m_first;
   }
   reference back()
   {
       utiassert(m_last != NULL);
       return *m_last;
   }
   const_reference front() const
   {
       utiassert(m_first != NULL);
       return *m_first;
   }
   const_reference back() const
   {
       utiassert(m_last != NULL);
       return *m_last;
   }
   T& getElementByIndex(int index) {
       for (iterator it = begin(); it != end(); it++) {
           if (index == 0) {
               return *it;
           }
           index--;
       }
       assert(false);
       return back();
   }
   void clear()
   {
       T* element = m_first;
       while (element)
       {
       CHECKEXISTANCEOF(element);
           T* next = (T*)element->clinkedlist_controller[c_index].m_next;

           element->clinkedlist_controller[c_index].m_prev = NULL;
           element->clinkedlist_controller[c_index].m_next = NULL;
//#ifdef _DEBUG
       utiassert(element->clinkedlist_controller[c_index].m_inlist == true);
       element->clinkedlist_controller[c_index].m_inlist = false;

//#endif
           controller.delete_all(element);
           if (next != NULL) {
          CHECKEXISTANCEOF(next);
           }
           element = next;
       }
       m_first = NULL;
       m_last = NULL;
       m_size = 0;

   }
   void push_back(const uti_refptr<T> &typ)
   {
       push_back(typ.get());
   }
   void push_back(pointer typ)
   {
       utiassert(typ != NULL);
       utiassert(typ->clinkedlist_controller[c_index].m_prev == NULL);
       utiassert(typ->clinkedlist_controller[c_index].m_next == NULL);
         CHECKEXISTANCEOF(typ);
//#ifdef _DEBUG
       utiassert(typ->clinkedlist_controller[c_index].m_inlist == false);
       typ->clinkedlist_controller[c_index].m_inlist = true;

//#endif
       if (m_last == NULL) {
           utiassert(m_first == NULL);
           m_first = typ;
           m_last = typ;
           utiassert(m_last->clinkedlist_controller[c_index].m_next == NULL);
           utiassert(m_first->clinkedlist_controller[c_index].m_prev == NULL);
       } else {
         CHECKEXISTANCEHIGHSPEEDOF(m_last);
           m_last->clinkedlist_controller[c_index].m_next = typ;
           utiassert(typ->clinkedlist_controller[c_index].m_prev == NULL);
                utiassert(typ->clinkedlist_controller[c_index].m_next == NULL);
           typ->clinkedlist_controller[c_index].m_prev = m_last;
                utiassert(typ->clinkedlist_controller[c_index].m_next == NULL);
           m_last = typ;
          utiassert(m_last->clinkedlist_controller[c_index].m_next == NULL);
       };
       m_size++;
       controller.insert(typ);

   }
   void push_back(T& typ)
   {
       push_back(&typ);
   }
   void pop_back()
   {
       utiassert(m_last != NULL);
       T* l = (T*)m_last->clinkedlist_controller[c_index].m_prev;
       m_last->clinkedlist_controller[c_index].m_prev = NULL;
       utiassert(m_last->clinkedlist_controller[c_index].m_next == NULL);
//#ifdef _DEBUG
       utiassert(m_last->clinkedlist_controller[c_index].m_inlist == true);
       m_last->clinkedlist_controller[c_index].m_inlist = false;

//#endif
       CHECKEXISTANCEOF(m_last);
       controller.delete_all(m_last);

       m_last = l;
        if (m_last){ CHECKEXISTANCEOF(m_last);};
       if (m_last != NULL) m_last->clinkedlist_controller[c_index].m_next = NULL;
       if (l == NULL) m_first = NULL;
       utiassert(m_size != 0);
       m_size--;


   }

   void pop_back_extract()
   {
       utiassert(m_last != NULL);
       T* l = m_last->clinkedlist_controller[c_index].m_prev;
       m_last->clinkedlist_controller[c_index].m_prev = NULL;
       utiassert(m_last->clinkedlist_controller[c_index].m_next == NULL);
//#ifdef _DEBUG
       utiassert(m_last->clinkedlist_controller[c_index].m_inlist == true);
       m_last->clinkedlist_controller[c_index].m_inlist = false;

//#endif
       if (l){  CHECKEXISTANCEOF(l);}
           CHECKEXISTANCEOF(m_last);
       controller.extract(m_last);
       m_last = l;
       if (m_last != NULL) m_last->clinkedlist_controller[c_index].m_next = NULL;
       if (l == NULL) m_first = NULL;
       utiassert(m_size != 0);
       m_size--;


   }

   void push_front(const uti_refptr<T> &typ)
   {
       push_front(typ.get());
   }
   void push_front(pointer typ)
   {
       utiassert(typ != NULL);
       utiassert(typ->clinkedlist_controller[c_index].m_prev == NULL);
       utiassert(typ->clinkedlist_controller[c_index].m_next == NULL);
//#ifdef _DEBUG
       utiassert(typ->clinkedlist_controller[c_index].m_inlist == false);
       typ->clinkedlist_controller[c_index].m_inlist = true;

//#endif
       if (m_first == NULL) {
           utiassert(m_last == NULL);
           m_first = typ;
           m_last = typ;
       } else {
           CHECKEXISTANCEOF(m_first);
           m_first->clinkedlist_controller[c_index].m_prev = typ;
           typ->clinkedlist_controller[c_index].m_next = m_first;
           m_first = typ;
           utiassert(m_first->clinkedlist_controller[c_index].m_prev == NULL);
       };
       m_size++;
       controller.insert(typ);

   }
   void push_front(T& typ)
   {
       push_front(&typ);
   }
   void pop_front()
   {
       utiassert(m_first != NULL);
        CHECKEXISTANCEOF(m_first);
       T* l = (T*)m_first->clinkedlist_controller[c_index].m_next;
       m_first->clinkedlist_controller[c_index].m_next = NULL;
       utiassert(m_first->clinkedlist_controller[c_index].m_prev == NULL);
//#ifdef _DEBUG
       utiassert(m_first->clinkedlist_controller[c_index].m_inlist == true);
       m_first->clinkedlist_controller[c_index].m_inlist = false;

//#endif
        CHECKEXISTANCEOF(m_first);
      if (l!= NULL){   CHECKEXISTANCEOF(l);}
       controller.delete_all(m_first);
       m_first = l;
       if (m_first != NULL) m_first->clinkedlist_controller[c_index].m_prev = NULL;
       if (l == NULL) m_last = NULL;
       utiassert(m_size != 0);
       m_size--;


   }
   void pop_front_extract()
   {
       utiassert(m_first != NULL);
        CHECKEXISTANCEOF(m_first);
       T* l = m_first->clinkedlist_controller[c_index].m_next;
       m_first->clinkedlist_controller[c_index].m_next = NULL;
       utiassert(m_first->clinkedlist_controller[c_index].m_prev == NULL);
//#ifdef _DEBUG
       utiassert(m_first->clinkedlist_controller[c_index].m_inlist == true);
       m_first->clinkedlist_controller[c_index].m_inlist = false;

//#endif

       controller.extract(m_first);
       m_first = l;
      if (m_first){  CHECKEXISTANCEOF(m_first);}
       if (m_first != NULL) m_first->clinkedlist_controller[c_index].m_prev = NULL;
       if (l == NULL) m_last = NULL;
       utiassert(m_size != 0);
       m_size--;
   }
   inline size_type size() const {
      return m_size;
   }
   size_type capacity() const {
      return m_size;
   }
   inline bool empty() const {return m_first == NULL;};
   size_type max_size() const {
      return size_type(-1) / sizeof(T);
   }

    void erase(const iterator& iter)
    {
        remove(const_cast<iterator&>(iter).m_element);
       // return iter;
    }

    void remove(pointer t)
    {
        utiassert(m_size != 0);
        utiassert(t != NULL);
        if (m_first == t) {
            pop_front();
         return;
        } else if (m_last == t) {
           pop_back();
           return;
        };
       //  #ifdef _DEBUG
       utiassert(t->clinkedlist_controller[c_index].m_inlist == true);
       t->clinkedlist_controller[c_index].m_inlist = false;
          CHECKEXISTANCEOF(t);
      // #endif
       ((T*)( t->clinkedlist_controller[c_index].m_prev))->clinkedlist_controller[c_index].m_next = t->clinkedlist_controller[c_index].m_next;
    ((T*)( t->clinkedlist_controller[c_index].m_next))->clinkedlist_controller[c_index].m_prev = t->clinkedlist_controller[c_index].m_prev;
        t->clinkedlist_controller[c_index].m_prev = NULL;
        t->clinkedlist_controller[c_index].m_next = NULL;

       controller.delete_all(t);
       m_size--;
    }
    void remove(uti_refptr<T> &typ)
    {
        remove(typ.get());
    }

    void extract(pointer t)
    {
        utiassert(t != NULL);
        if (m_first == t) {
            pop_front_extract();

        } else if (m_last == t) {
           pop_back_extract();
        };
        // #ifdef _DEBUG
       utiassert(t->clinkedlist_controller[c_index].m_inlist == true);
       t->clinkedlist_controller[c_index].m_inlist = false;

      // #endif
        CHECKEXISTANCEOF(t);
         ((T*)( t->clinkedlist_controller[c_index].m_prev))->clinkedlist_controller[c_index].m_next = t->clinkedlist_controller[c_index].m_next;
    ((T*)( t->clinkedlist_controller[c_index].m_next))->clinkedlist_controller[c_index].m_prev = t->clinkedlist_controller[c_index].m_prev;
        t->clinkedlist_controller[c_index].m_prev = NULL;
        t->clinkedlist_controller[c_index].m_next = NULL;
       controller.extract(t);
    }
    void extract(uti_refptr<T> &typ)
    {
        extract(typ.get());
    }
    /*void insert(pointer position, pointer new2)
    {

    };
    void exchange(pointer position, pointer new2)
    {

    }*/

};

template<class T, int c_index> struct CNormalLinkedList : public CLinkedList<T, c_index, CLinkedList_NoFree_MemoryController<T> > {};
template<class T, int c_index> struct COwnLinkedList: public CLinkedList<T, c_index, CLinkedList_Free_MemoryController<T > > {};
template<class T, int c_index> struct CReferenceLinkedList: public CLinkedList<T, c_index, CLinkedList_Reference_MemoryController<T> >{};

#endif // UTILINKEDLIST_H
