/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#pragma once
#ifndef UTILISTS_H
#define UTILISTS_H

#include "Uti3DCore.h"

#include <algorithm>

class CCacheList : public CUtiOB
{
protected:
    virtual CUtiOB** ReserveMem(int count);
    virtual void FreeMem(CUtiOB** vd);
    void ValidNewSize(int newcount);
    virtual void FreeItem(CUtiOB** item);
public:
    CUtiOB* *pointers;
    int count;
    int capacity;
    int inc;
    bool deleteItems;
    CMemController* MemController;
    explicit CCacheList();
    ~CCacheList();
    void Clear();
    inline int getCount() const {return count;};
    //LPVOID &operator[](unsigned int p);
    //const LPVOID &operator[](unsigned int p) const;
    inline CUtiOB* &operator[](unsigned int p)
    {
       utiassert(/*p >= 0 &&*/ p < (unsigned int)count);
       return pointers[p];
    }
    //#ifndef __MINGW32__
    inline const CUtiOB* operator[](unsigned int p) const
    {
       utiassert(/*p >= 0 &&*/ p < (unsigned int)count);
        return pointers[p];
    }
    //#endif
    virtual void Add(CUtiOB* item);

    void AddList(CCacheList* list);
    //void AddList(CUtiList* list);
    virtual void Delete(int index);
    void Remove(CUtiOB* item);
    void Extract(int index);
    void Extract(CUtiOB* item);
    void RemoveExtract(CUtiOB* item);
    int IndexOf(CUtiOB* item) const;
    void Move(int source, int dest);
    void Insert(CUtiOB* item, int index);
    void Pack();
    void SetNewSize(int size);
    void EnumItems(void *data, bool (UTICDECL *enumfunc)(CUtiOB*,void *));
    void Sort(int (UTICDECL *compare )(const void *elem1, const void *elem2 ));
    HResult ReadList(CStream & Stream, int Flags, int version, CUtiOB* (UTICDECL *enumfunc)(CUtiOB* parent, int type), CUtiOB* parent);
    HResult Write(CStream & Stream, int Flags, int version);

};

//typedef CCacheList CList;



template<class T> struct CReferenceListImplementation {
        T** elements;
        size_t count;
        size_t capacity;
        size_t inc;
        CReferenceListImplementation(T** _el=NULL, size_t _c=0, size_t _capacity=0): inc(20)
        {
            elements = _el;
            count = _c;
            capacity = _capacity;
        }
        ~CReferenceListImplementation()
        {
            count = -1;
        }

        void free()
        {
            count = 0; resize(0);
        }
        void resize(size_t size)
        {
            utiassert(count != -1);
            utiassert(size >= count);
            T** newarray = NULL;
            if (size) {
                typedef T* NEW_TYPE;
               newarray  = new NEW_TYPE[size];
               for (size_t i = 0; i < count; i++) {
                  newarray[i] = elements[i];
               }
            }
            if (elements)delete [] elements;
            elements = newarray;
            capacity = size;

        }
        /*void validNewSize(size_t newcount)
        {
           utiassert(newcount >= 0);

           size_t newcap;
           if (newcount > capacity || newcount < capacity-inc*2)
           {
               newcap = newcount+ inc;
           } else {
             count = newcount; return;
           };
           if (newcount < count) count = newcount;
           resize(newcap);
           count = newcount;
         }*/
        void validNewSize(size_t newcount)
        {
           utiassert(newcount >= 0);

           size_t newcap=0;
           if (newcount > capacity)
           {
               newcap = capacity*2;
           } else if (newcount < capacity/3)
           {
               newcap = capacity/2;
           } else {
             count = newcount; return;
           };
           if (newcap < 4) newcap = 4;
           if (newcount < count) count = newcount;
           resize(newcap);
           count = newcount;
         }
         inline void moveData(size_t deleteIndex)
         {
             for (size_t i = deleteIndex+1; i < count; i++)
             {
                elements[i-1] = elements[i];
             }
         }
         inline off_t indexOf(const T* element) const
         {
             for (size_t i = 0; i < count; i++)
             {
                if (elements[i] == element) return (off_t)i;
             }
             return UTI_INVALIDINDEX;
         }

    };

template <class T, class MemoryContainer> struct CList_NoFree_MemoryController
{
    inline void insert(T* element, size_t index) {

    }
    inline void delete_all(T* element) {   };
    inline void deleteItem(MemoryContainer& impl, size_t index) {
       impl.moveData(index);
    };
    inline off_t indexOf(const MemoryContainer& impl, const T* el) const {
        return impl.indexOf(el);
    };
    inline void move(MemoryContainer& impl, size_t el1, size_t el2)
    {
        T* el = impl.elements[el1];
        impl.elements[el1] = impl.elements[el2];
        impl.elements[el2] = el;
    }
    inline void exchange(MemoryContainer& impl, size_t index, T* el)
    {
        impl.elements[index] = el;
    }
    inline void exchangeExtract(MemoryContainer& impl, size_t index, T* el)
    {

        impl.elements[index] = el;
    }
};

template <class T, class MemoryContainer> struct CList_Free_MemoryController
{
    inline void insert(T* element, size_t index) {

    }
    inline void delete_all(T* element) {if (element)delete element;   };
    inline void deleteItem(MemoryContainer& impl, size_t index) {
        delete_all(impl.elements[index]);
       impl.moveData(index);
    };
    inline off_t indexOf(const MemoryContainer& impl, const T* el) const {
        return impl.indexOf(el);
    };
    inline void move(MemoryContainer& impl, size_t el1, size_t el2)
    {
        T* el = impl.elements[el1];
        impl.elements[el1] = impl.elements[el2];
        impl.elements[el2] = el;
    }
    inline void exchange(MemoryContainer& impl, size_t index, T* el)
    {
        if (el != impl.elements[index])  delete_all(impl.elements[index]);
        impl.elements[index] = el;
    }
    inline void exchangeExtract(MemoryContainer& impl, size_t index, T* el)
    {

        impl.elements[index] = el;
    }
};

template <class T, class MemoryContainer> struct CList_Reference_MemoryController
{
    inline void insert(T* element, size_t index) {
        if (element) element->addRef();
    }
    inline void delete_all(T* element) {if (element) element->decRef();   };
    inline void deleteItem(MemoryContainer& impl, size_t index) {
        delete_all(impl.elements[index]);
       impl.moveData(index);
    };
    inline off_t indexOf(const MemoryContainer& impl, const T* el) const {
        return impl.indexOf(el);
    };
    inline void move(MemoryContainer& impl, size_t el1, size_t el2)
    {
        T* el = impl.elements[el1];
        impl.elements[el1] = impl.elements[el2];
        impl.elements[el2] = el;
    }
    inline void exchange(MemoryContainer& impl, size_t index, T* el)
    {
        if (el != impl.elements[index])  delete_all(impl.elements[index]);
        impl.elements[index] = el;
        insert(el, index);
    }
    inline void exchangeExtract(MemoryContainer& impl, size_t index, T* el)
    {
        impl.elements[index] = el;
        insert(el, index);
    }
};


template <class T, class MemoryContainer> struct CList_ReferenceADR_MemoryController
{
    inline void insert(T* element, size_t index) {
        if (element) element->cpp_addRef();
    }
    inline void delete_all(T* element) {if (element) element->cpp_decRef();   };
    inline void deleteItem(MemoryContainer& impl, size_t index) {
        delete_all(impl.elements[index]);
       impl.moveData(index);
    };
    inline off_t indexOf(const MemoryContainer& impl, const T* el) const {
        return impl.indexOf(el);
    };
    inline void move(MemoryContainer& impl, size_t el1, size_t el2)
    {
        T* el = impl.elements[el1];
        impl.elements[el1] = impl.elements[el2];
        impl.elements[el2] = el;
    }
    inline void exchange(MemoryContainer& impl, size_t index, T* el)
    {
        if (el != impl.elements[index])  delete_all(impl.elements[index]);
        impl.elements[index] = el;
        insert(el, index);
    }
    inline void exchangeExtract(MemoryContainer& impl, size_t index, T* el)
    {
        impl.elements[index] = el;
        insert(el, index);
    }
};


struct CList_IndexController
{

public:
   off_t index;
   CList_IndexController() : index(UTI_INVALIDINDEX) {};
   bool inList() {return index != UTI_INVALIDINDEX;};
};

template <class T, class MemoryContainer, int c_index> struct CList_IndexFree_MemoryController
{
    inline void insert(T* element, size_t index) {
        if (!element) return;
        utiassert(element->clist_indexcontroller[c_index].index == UTI_INVALIDINDEX);
        element->clist_indexcontroller[c_index].index = (int)index;
    }
    inline void delete_all(T* element) {
        if (element) {
            utiassert(element->clist_indexcontroller[c_index].index  != UTI_INVALIDINDEX);
            element->clist_indexcontroller[c_index].index  = UTI_INVALIDINDEX;
            delete element;
        };
    };
    inline void deleteItem(MemoryContainer& impl, size_t index) {
        //utiassert(impl.element[index] == NULL || impl.element[index]->clist_indexcontroller.index == index);
        move(impl, index, impl.count-1);
        delete_all(impl.elements[impl.count-1]);
       //impl.moveData(index);
    };
    inline off_t indexOf(const MemoryContainer& impl, const T* el) const {
        utiassert(el != NULL);
        if (el->clist_indexcontroller[c_index].index == UTI_INVALIDINDEX) return UTI_INVALIDINDEX;
        utiassert(el->clist_indexcontroller[c_index].index  != UTI_INVALIDINDEX);
        if (impl.elements[el->clist_indexcontroller[c_index].index] != el) return UTI_INVALIDINDEX;
        return el->clist_indexcontroller[c_index].index;
    };
    inline void move(MemoryContainer& impl, size_t el1, size_t el2)
    {
        if (el1 == el2) return;
        utiassert(impl.elements[el1] == NULL || impl.elements[el1]->clist_indexcontroller[c_index].index == (off_t)el1);
        utiassert(impl.elements[el2] == NULL || impl.elements[el2]->clist_indexcontroller[c_index].index == (off_t)el2);
        T* el = impl.elements[el1];
        impl.elements[el1] = impl.elements[el2];
        impl.elements[el2] = el;
        if (impl.elements[el1]) impl.elements[el1]->clist_indexcontroller[c_index].index = (int)el1;
        if (impl.elements[el2]) impl.elements[el2]->clist_indexcontroller[c_index].index = (int)el2;
    }
    inline void exchange(MemoryContainer& impl, size_t index,T* el)
    {
        if (el != impl.elements[index])  delete_all(impl.elements[index]);
        impl.elements[index] = el;
        insert(el, index);
    }
    inline void exchangeExtract(MemoryContainer& impl, size_t index, T* el)
    {
        T* element = impl.elements[index];
        if (element) {
            utiassert(element->clist_indexcontroller[c_index].index  != UTI_INVALIDINDEX);
            element->clist_indexcontroller[c_index].index  = UTI_INVALIDINDEX;

        };
        impl.elements[index] = el;
        insert(el, index);
    }
};

template <class T, class MemoryContainer, int c_index> struct CList_IndexReference_MemoryController
{
    inline void insert(T* element, size_t index) {
        if (!element) return;
    element->addRef();
        utiassert(element->clist_indexcontroller[c_index].index == UTI_INVALIDINDEX);
        element->clist_indexcontroller[c_index].index = (int)index;
    }
    inline void delete_all(T* element) {
        if (element) {
            utiassert(element->clist_indexcontroller[c_index].index  != UTI_INVALIDINDEX);
            element->clist_indexcontroller[c_index].index  = UTI_INVALIDINDEX;
          element->decRef();
        };
    };
    inline void deleteItem(MemoryContainer& impl, size_t index) {
        //utiassert(impl.element[index] == NULL || impl.element[index]->clist_indexcontroller.index == index);
        move(impl, index, impl.count-1);
        delete_all(impl.elements[impl.count-1]);
       //impl.moveData(index);
    };
    inline off_t indexOf(const MemoryContainer& impl, const T* el) const {
        utiassert(el != NULL);
        if (el->clist_indexcontroller[c_index].index == UTI_INVALIDINDEX) return UTI_INVALIDINDEX;
        utiassert(el->clist_indexcontroller[c_index].index  != UTI_INVALIDINDEX);
        if (impl.elements[el->clist_indexcontroller[c_index].index] != el) return UTI_INVALIDINDEX;
        return el->clist_indexcontroller[c_index].index;
    };
    inline void move(MemoryContainer& impl, size_t el1, size_t el2)
    {
        if (el1 == el2) return;
        utiassert(impl.elements[el1] == NULL || impl.elements[el1]->clist_indexcontroller[c_index].index == (off_t)el1);
        utiassert(impl.elements[el2] == NULL || impl.elements[el2]->clist_indexcontroller[c_index].index == (off_t)el2);
        T* el = impl.elements[el1];
        impl.elements[el1] = impl.elements[el2];
        impl.elements[el2] = el;
        if (impl.elements[el1]) impl.elements[el1]->clist_indexcontroller[c_index].index = (int)el1;
        if (impl.elements[el2]) impl.elements[el2]->clist_indexcontroller[c_index].index = (int)el2;
    }
    inline void exchange(MemoryContainer& impl, size_t index,T* el)
    {
        if (el != impl.elements[index])  delete_all(impl.elements[index]);
        impl.elements[index] = el;
        insert(el, index);
    }
    inline void exchangeExtract(MemoryContainer& impl, size_t index, T* el)
    {
        impl.elements[index] = el;
        insert(el, index);
    }
};

template <class T, class MemoryContainer, int c_index> struct CList_Index_MemoryController
{
    inline void insert(T* element, size_t index) {
        if (!element) return;
        utiassert(element->clist_indexcontroller[c_index].index == UTI_INVALIDINDEX);
        element->clist_indexcontroller[c_index].index = (int)index;
    }
    inline void delete_all(T* element) {
        if (element) {
            utiassert(element->clist_indexcontroller[c_index].index  != UTI_INVALIDINDEX);
            element->clist_indexcontroller[c_index].index  = UTI_INVALIDINDEX;
        };
    };
    inline void deleteItem(MemoryContainer& impl, size_t index) {
        //utiassert(impl.element[index] == NULL || impl.element[index]->clist_indexcontroller.index == index);
        move(impl, index, impl.count-1);
        delete_all(impl.elements[impl.count-1]);
       //impl.moveData(index);
    };
    inline off_t indexOf(const MemoryContainer& impl, const T* el) const {
        utiassert(el != NULL);
        if (el->clist_indexcontroller[c_index].index == UTI_INVALIDINDEX) return UTI_INVALIDINDEX;
        utiassert(el->clist_indexcontroller[c_index].index  != UTI_INVALIDINDEX);
        if (impl.elements[el->clist_indexcontroller[c_index].index] != el) return UTI_INVALIDINDEX;
        return el->clist_indexcontroller[c_index].index;
    };
    inline void move(MemoryContainer& impl, size_t el1, size_t el2)
    {
        if (el1 == el2) return;
        utiassert(impl.elements[el1] == NULL || impl.elements[el1]->clist_indexcontroller[c_index].index == (off_t)el1);
        utiassert(impl.elements[el2] == NULL || impl.elements[el2]->clist_indexcontroller[c_index].index == (off_t)el2);
        T* el = impl.elements[el1];
        impl.elements[el1] = impl.elements[el2];
        impl.elements[el2] = el;
        if (impl.elements[el1]) impl.elements[el1]->clist_indexcontroller[c_index].index = (int)el1;
        if (impl.elements[el2]) impl.elements[el2]->clist_indexcontroller[c_index].index = (int)el2;
    }
    inline void exchange(MemoryContainer& impl, size_t index, T* el)
    {
        if (el != impl.elements[index])  delete_all(impl.element[index]);
        impl.elements[index] = el;
        insert(el, index);
    }
    inline void exchangeExtract(MemoryContainer& impl, size_t index, T* el)
    {
        impl.elements[index] = el;
        insert(el, index);
    }
};

template <class T, class MemoryContainer> class CList_Iterator
{
    public:
       typedef std::random_access_iterator_tag iterator_category;
    typedef T value_type;
   typedef size_t size_type;
   typedef ptrdiff_t difference_type;
   typedef T* pointer;
   typedef const T* const_pointer;
   typedef T& reference;
   typedef const T& const_reference;
  // typedef random_access_iterator iterator_category;
   //typedef CList_Iterator<T, MemoryContainer, MemoryController> selftype;
    public:
       MemoryContainer *impl;
       difference_type position;
      // MemoryController<T, MemoryContainer> controller;
       inline CList_Iterator() {
          position = UTI_INVALIDINDEX;
          impl = NULL;
       }
       inline CList_Iterator(const CList_Iterator& iter)
       {
           impl = iter.impl;
           position = iter.position;
       }
       inline CList_Iterator(MemoryContainer *_impl, int pos) {
           impl = _impl;
           position = pos;
       }
       inline T & operator*() {return *impl->elements[position];};
       inline T * operator->() {return impl->elements[position];};
       inline CList_Iterator &operator++() {++position; return *this;};
       inline CList_Iterator operator++(int) {CList_Iterator tmp(*this); ++(*this); return tmp;};
       inline CList_Iterator &operator--() {--position; return *this;};
       inline CList_Iterator operator--(int) {CList_Iterator tmp(*this); --(*this); return tmp;};
       inline CList_Iterator operator+(difference_type n) {CList_Iterator tmp(*this); tmp.position += n; return tmp;};
       inline CList_Iterator &operator+=(difference_type n) {position += n; return *this;};
       inline CList_Iterator operator-(difference_type n) {CList_Iterator tmp(*this); tmp.position -= n; return tmp;};
       inline CList_Iterator &operator-=(difference_type n) {position -= n; return *this;}
       inline bool operator==(const CList_Iterator &other) const
       {
           return (position == other.position) && (impl == other.impl);
       }
       inline bool operator!=(const CList_Iterator &other) const
       {
           return (position != other.position) || (impl != other.impl);
       }
       inline CList_Iterator& operator=(const CList_Iterator& iter)
       {
           impl = iter.impl;
           position = iter.position;
           return *this;
       }
       inline difference_type operator-(const CList_Iterator& iter) {   utiassert(impl == iter.impl);
           return position - iter.position;   };
       inline bool operator<(const CList_Iterator& iter) const{  utiassert(impl == iter.impl);return position < iter.position;  };
       inline bool operator>(const CList_Iterator& iter) const{  utiassert(impl == iter.impl);return position > iter.position;  };
       inline bool operator<=(const CList_Iterator& iter) const{  utiassert(impl == iter.impl);return position <= iter.position;  };
       inline bool operator>=(const CList_Iterator& iter) const{  utiassert(impl == iter.impl);return position >= iter.position;  };
       inline T& operator[](difference_type n) {return *impl->elements[position+n];};
       inline const T& operator[](difference_type n) const {return *impl->elements[position+n];};

};

template <class T, class MemoryContainer> class CList_Iterator_Const
{
    public:
       typedef std::random_access_iterator_tag iterator_category;
    typedef T value_type;
   typedef size_t size_type;
   typedef ptrdiff_t difference_type;
   typedef T* pointer;
   typedef const T* const_pointer;
   typedef T& reference;
   typedef const T& const_reference;
  // typedef random_access_iterator iterator_category;
   //typedef CList_Iterator<T, MemoryContainer, MemoryController> selftype;
    public:
       const MemoryContainer *impl;
       difference_type position;
      // MemoryController<T, MemoryContainer> controller;
       inline CList_Iterator_Const() {
          position = UTI_INVALIDINDEX;
          impl = NULL;
       }
       inline CList_Iterator_Const(const CList_Iterator_Const& iter)
       {
           impl = iter.impl;
           position = iter.position;
       }
       inline CList_Iterator_Const(const MemoryContainer *_impl, int pos) {
           impl = _impl;
           position = pos;
       }
       inline const T & operator*() {return *impl->elements[position];};
       inline const T * operator->() {return impl->elements[position];};
       inline CList_Iterator_Const &operator++() {++position; return *this;};
       inline CList_Iterator_Const operator++(int) {CList_Iterator_Const tmp(*this); ++(*this); return tmp;};
       inline CList_Iterator_Const &operator--() {--position; return *this;};
       inline CList_Iterator_Const operator--(int) {CList_Iterator_Const tmp(*this); --(*this); return tmp;};
       inline CList_Iterator_Const operator+(difference_type n) {CList_Iterator_Const tmp(*this); tmp.position += n; return tmp;};
       inline CList_Iterator_Const &operator+=(difference_type n) {position += n; return *this;};
       inline CList_Iterator_Const operator-(difference_type n) {CList_Iterator_Const tmp(*this); tmp.position -= n; return tmp;};
       inline CList_Iterator_Const &operator-=(difference_type n) {position -= n; return *this;}
       inline bool operator==(const CList_Iterator_Const &other) const
       {
           return (position == other.position) && (impl == other.impl);
       }
       inline bool operator!=(const CList_Iterator_Const &other) const
       {
           return (position != other.position) || (impl != other.impl);
       }
       inline CList_Iterator_Const& operator=(const CList_Iterator_Const& iter)
       {
           impl = iter.impl;
           position = iter.position;
           return *this;
       }
       inline difference_type operator-(const CList_Iterator_Const& iter) {   utiassert(impl == iter.impl);
           return position - iter.position;   };
       inline bool operator<(const CList_Iterator_Const& iter) const{  utiassert(impl == iter.impl);return position < iter.position;  };
       inline bool operator>(const CList_Iterator_Const& iter) const{  utiassert(impl == iter.impl);return position > iter.position;  };
       inline bool operator<=(const CList_Iterator_Const& iter) const{  utiassert(impl == iter.impl);return position <= iter.position;  };
       inline bool operator>=(const CList_Iterator_Const& iter) const{  utiassert(impl == iter.impl);return position >= iter.position;  };
       inline const T& operator[](difference_type n) const {return *impl->elements[position+n];};

};


template <class T, class MemoryContainer> class CList_Iterator_Pointer
{
    public:
       typedef std::random_access_iterator_tag iterator_category;
    typedef T* value_type;
   typedef size_t size_type;
   typedef ptrdiff_t difference_type;
   typedef value_type* pointer;
   typedef const value_type* const_pointer;
   typedef value_type& reference;
   typedef const value_type& const_reference;
  // typedef random_access_iterator iterator_category;
   //typedef CList_Iterator<T, MemoryContainer, MemoryController> selftype;
    public:
       MemoryContainer *impl;
       difference_type position;
      // MemoryController<T, MemoryContainer> controller;
       inline CList_Iterator_Pointer() {
          position = UTI_INVALIDINDEX;
          impl = NULL;
       }
       inline CList_Iterator_Pointer(const CList_Iterator_Pointer& iter)
       {
           impl = iter.impl;
           position = iter.position;
       }
       inline CList_Iterator_Pointer(MemoryContainer *_impl, int pos) {
           impl = _impl;
           position = pos;
       }
       inline value_type &operator*() {return impl->elements[position];};
       inline T * operator->() {return impl->elements[position];};
       inline CList_Iterator_Pointer &operator++() {++position; return *this;};
       inline CList_Iterator_Pointer operator++(int) {CList_Iterator_Pointer tmp(*this); ++(*this); return tmp;};
       inline CList_Iterator_Pointer &operator--() {--position; return *this;};
       inline CList_Iterator_Pointer operator--(int) {CList_Iterator_Pointer tmp(*this); --(*this); return tmp;};
       inline CList_Iterator_Pointer operator+(difference_type n) {CList_Iterator_Pointer tmp(*this); tmp.position += n; return tmp;};
       inline CList_Iterator_Pointer &operator+=(difference_type n) {position += n; return *this;};
       inline CList_Iterator_Pointer operator-(difference_type n) {CList_Iterator_Pointer tmp(*this); tmp.position -= n; return tmp;};
       inline CList_Iterator_Pointer &operator-=(difference_type n) {position -= n; return *this;}
       inline bool operator==(const CList_Iterator_Pointer &other) const
       {
           return (position == other.position) && (impl == other.impl);
       }
       inline bool operator!=(const CList_Iterator_Pointer &other) const
       {
           return (position != other.position) || (impl != other.impl);
       }
       inline CList_Iterator_Pointer& operator=(const CList_Iterator_Pointer& iter)
       {
           impl = iter.impl;
           position = iter.position;
           return *this;
       }
       inline difference_type operator-(const CList_Iterator_Pointer& iter) {   utiassert(impl == iter.impl);
           return position - iter.position;   };
       inline bool operator<(const CList_Iterator_Pointer& iter) const{  utiassert(impl == iter.impl);return position < iter.position;  };
       inline bool operator>(const CList_Iterator_Pointer& iter) const{  utiassert(impl == iter.impl);return position > iter.position;  };
       inline bool operator<=(const CList_Iterator_Pointer& iter) const{  utiassert(impl == iter.impl);return position <= iter.position;  };
       inline bool operator>=(const CList_Iterator_Pointer& iter) const{  utiassert(impl == iter.impl);return position >= iter.position;  };
       inline T* operator[](difference_type n) {return impl->elements[position+n];};
       inline const T* operator[](difference_type n) const {return impl->elements[position+n];};

};

template <class T, class MemoryContainer, class MemoryController> struct CList
{
protected:
    //virtual void FreeItem(CUtiOB** item);

    MemoryContainer impl;
    MemoryController controller;
public:
   typedef T value_type;
   typedef size_t size_type;
   typedef ptrdiff_t difference_type;
   typedef T* pointer;
   typedef const T* const_pointer;
   typedef T& reference;
   typedef const T& const_reference;
   typedef CList_Iterator<T, MemoryContainer> iterator;
    typedef CList_Iterator_Const<T, MemoryContainer> const_iterator;
    typedef CList_Iterator_Pointer<T, MemoryContainer> iterator_pointer;
   iterator begin() { return iterator(&impl, 0);};
   iterator end() {return iterator(&impl, (int)impl.count);};
   const_iterator begin() const { return const_iterator(&impl, 0);};
   const_iterator end() const {return const_iterator(&impl, (int)impl.count);};
   iterator_pointer begin_pointer() { return iterator_pointer(&impl, 0);};
   iterator_pointer end_pointer() {return iterator_pointer(&impl, (int)impl.count);};
   explicit CList() {

   }
   explicit CList(const CList<T, MemoryContainer, MemoryController>& other)
   {
       int c = other.size();
       for (int i = 0; i < c; i++) {
           push_back(const_cast<T&>(other[i]));
       }
   }

   ~CList()
   {
       clear();
   }
   reference front()
   {
       utiassert(impl.count > 0);
       return* impl.elements[0];
   }
   reference back()
   {
       utiassert(impl.count > 0);
       return *impl.elements[impl.count-1];
   }
   const_reference front() const
   {
       utiassert(impl.count > 0);
       return *impl.elements[0];
   }
   const_reference back() const
   {
       utiassert(impl.count > 0);
       return *impl.elements[impl.count-1];
   }
   void clear()
   {
       for (size_t i = 0; i < impl.count; i++)
       {
           controller.delete_all(impl.elements[i]);//->decRef();
       }
       impl.free();
   }
   void push_back(const uti_refptr<T> &typ)
   {
       push_back(typ.get());
   }
   void push_back(pointer typ)
   {
       controller.insert(typ, impl.count);
       impl.validNewSize(impl.count+1);
       impl.elements[impl.count-1] = typ;

   }
   void insert(int index,pointer typ)
   {
       utiassert(index >= 0 && (size_t)index <= impl.count);
       controller.insert(typ, index);
       impl.validNewSize(impl.count+1);
       impl.elements[impl.count-1] = NULL;
       for (int i = impl.count-1; i > index; i--)
       {
           controller.move(impl, i, i-1);

       }
       utiassert(impl.elements[index] == NULL);
       impl.elements[index] = typ;

   }
   void push_back(T& typ)
   {
       push_back(&typ);
   }
   void push_back(const  T& typ)
   {
       push_back(const_cast<T*>(&typ));
   }

   void push_back(CList<T, MemoryContainer, MemoryController>& other)
   {
       for (int i = 0; i < other.size(); i++) {
           push_back(other[i]);
       }
   }

   void pop_back()
   {
       utiassert(impl.count > 0);
       controller.delete_all(impl.elements[impl.count-1]);
       impl.validNewSize(impl.count-1);
   }
   inline size_type size() const {
      return impl.count;
   }
   size_type capacity() const {
      return impl.capacity;
   }
   inline bool empty() const {return impl.count == 0;};
   size_type max_size() const {
      return size_type(-1) / sizeof(value_type);
   }
   inline reference operator[](size_type p)
    {
       utiassert(/*p >= 0 &&*/ p < impl.count);
       return *impl.elements[p];
    }
   inline pointer get(size_type p)
    {
       utiassert(/*p >= 0 &&*/ p < impl.count);
       return impl.elements[p];
    }
   inline const pointer get(size_type p) const
    {
       utiassert(/*p >= 0 &&*/ p < impl.count);
       return impl.elements[p];
    }

   inline pointer at(size_type p)
    {
       utiassert(/*p >= 0 &&*/ p < impl.count);
       return impl.elements[p];
    }
    //#ifndef __MINGW32__
    inline const_reference operator[](size_type p) const
    {
       utiassert(/*p >= 0 &&*/ p < impl.count);
        return *impl.elements[p];
    }
    template<typename _Compare> void sort(_Compare compare)
    {
        if (impl.count > 1) {
        std::sort(&impl.elements[0], &impl.elements[impl.count], compare);
        };
    }

    void deleteItem(size_t i)
    {utiassert(i >= 0 && i < size());
        controller.deleteItem(impl, i);
        impl.validNewSize(impl.count-1);
    }
    iterator erase(const iterator& iter)
    {
        utiassert(iter.position < impl.count);
        deleteItem(iter.position);
        return iter;
    }
   /* iterator erase(const iterator& q1, const iterator & q2)
    {
        uassert(false);
        return q1;
    }*/
    void remove(const_pointer t)
    {
        off_t in = indexOf(t);
        utiassert(in != UTI_INVALIDINDEX);
        deleteItem(in);
    }
    void remove(const uti_refptr<T> &typ)
    {
        remove(typ.get());
    }
    off_t indexOf(const_pointer t) const
    {
         return controller.indexOf(impl, t);
    }
    off_t indexOf(const uti_refptr<T> &typ) const
    {
         return indexOf(typ.get());
    }
    void move(size_t pos, size_t pos2)
    {utiassert(pos >= 0 && pos < size());
    utiassert(pos2 >= 0 && pos2 < size());
        controller.move(impl, pos, pos2);
    }
    void exchange(size_t index, uti_refptr<T> &typ)
    {utiassert(index >= 0 && index < size());
        exchange(index, typ.get());
    }
    void exchange(size_t index, pointer t)
    {utiassert(index >= 0 && index < size());
       controller.exchange(impl, index, t);
    }
    void extractViaIndex(size_t index)
    {utiassert(index >= 0 && index < size());
         controller.exchangeExtract(impl, index, NULL);
         deleteItem(index);
    }
    void extract(const_pointer t)
    {
        off_t in = indexOf(t);
        utiassert(in != UTI_INVALIDINDEX);
        deleteItem(in);
    }
    CList& operator=(const CList& t)
    {
        clear();
        for (int i = 0;i < t.size(); i++) {
            push_back(t.get(i));
        }


        return *this;
    }
};

template<class T> struct CNormalList : public CList<T, CReferenceListImplementation<T>, CList_NoFree_MemoryController<T, CReferenceListImplementation<T> > > {};
template<class T> struct COwnList: public CList<T, CReferenceListImplementation<T>, CList_Free_MemoryController<T, CReferenceListImplementation<T> > > {};
template<class T> struct CReferenceList: public CList<T, CReferenceListImplementation<T>, CList_Reference_MemoryController<T, CReferenceListImplementation<T> > >{};
template<class T> struct CReferenceListADR: public CList<T, CReferenceListImplementation<T>, CList_ReferenceADR_MemoryController<T, CReferenceListImplementation<T> > >{};

template<class T, int c_index> struct COwnIndexList: public CList<T, CReferenceListImplementation<T>, CList_IndexFree_MemoryController<T, CReferenceListImplementation<T>, c_index > >{};
template<class T, int c_index> struct CIndexList: public CList<T, CReferenceListImplementation<T>, CList_Index_MemoryController<T, CReferenceListImplementation<T>, c_index > >{};
template<class T, int c_index> struct CReferenceIndexList: public CList<T, CReferenceListImplementation<T>, CList_IndexReference_MemoryController<T, CReferenceListImplementation<T>, c_index > >{};

class CHeapList : public CCacheList
{
protected:
    inline int GetItemWeighti(int i);
    inline int getParent(int k);
    void upheap(int item);
    void downheap(int item);
public:

    void Add(CUtiOB* item);
    void Delete(int index);
};

#endif // UTILISTS_H
