
/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "utilogic.h"

UtiLogicObject::UtiLogicObject(UtiLogic *parent)
{
   m_sticky = false;
   m_guisticky = false;
           m_parent = parent;
	   m_guireceiver = NULL;
       m_stickyindex = -1;
       m_guiindex = -1;
}

UtiLogicObject::~UtiLogicObject()
{
   if (m_sticky) {
       m_parent->removeStickyObject(this, m_stickyindex);
       m_sticky = false;
   }
   if (m_guisticky) {
       m_parent->removeGuiStickyObject(this, m_guiindex);
       m_guisticky = false;

   }
   if (clist_indexcontroller[1].inList()) {
       m_parent->m_logicthinkobjects.remove(this);
   }
}

void UtiLogicObject::update()
{
  CHECKEXISTANCE;
}

void UtiLogicObject::notifyGUI()
{
   CHECKEXISTANCE;
  if (m_guireceiver) {
  CHECKEXISTANCEOF(m_guireceiver);
  if (m_guisticky == false) {
      assert(m_guiindex == -1);
   m_guiindex = m_parent->addNotifyGUI(this);
    m_guisticky = true;
  };
 
  }
  
}

void UtiLogicObject::notifyRealGUI()
{
   CHECKEXISTANCE;
  if (m_guireceiver) {
  CHECKEXISTANCEOF(m_guireceiver);
 
  assert(m_guiindex != -1);
  assert(m_guisticky == true);
    bool handled = false;
    m_guisticky = false;
    m_guiindex = -1;

   m_guireceiver->GetMessageUti(this, 0, 0, 0, handled);
  }
  
}

void UtiLogicObject::rewriteSlotData()
{ CHECKEXISTANCE;
  for (int i = 0; i < m_slots.size(); i++) {
     UtiLogicSlot& s = m_slots[i];
     if (s.m_connection && ( (s.m_flags & UTI_LOGIC_OUTPUT_SLOT) != 0)) {
       s.m_connection->m_data = s.m_lastdata;
     }
  }
}

void UtiLogicObject::disconnect()
{ CHECKEXISTANCE;
    for (int i = 0; i < m_slots.size(); i++)
    {
        if (m_slots[i].m_connection != NULL) {
            CHECKEXISTANCEOF(m_slots[i].m_connection);
            int con = getSlotConnectionCountFromLogicConnectionToMe(i);
            m_slots[i].m_connection->removeObject(this, i, con);
        }
    }
}


int UtiLogicObject::getOutputConnectionCount()
{
    int connectioncount = 0;

    for (int i = 0; i < m_slots.size(); i++) {
       if (isOutputSlot(i) == false)continue;
        //UtiLogicSlot& s = m_slots[i];
       connectioncount += umax(getSlotConnectionCountFromOtherObjects(i), 0);
    }
    return connectioncount;
}

bool UtiLogicObject::hasOutputConnection()
{
    return getOutputConnectionCount() != 0;
}

int UtiLogicObject::getInputConnectionCount()
{

    int connectioncount = 0;

    for (int i = 0; i < m_slots.size(); i++) {
       if (isOutputSlot(i) == true)continue;
        //UtiLogicSlot& s = m_slots[i];
       connectioncount += umax(getSlotConnectionCountFromOtherObjects(i), 0);
    }
    return connectioncount;
}

bool UtiLogicObject::hasInputConnection()
{
    return getInputConnectionCount() != 0;
}

int UtiLogicObject::getSlotConnectionCountFromLogicConnectionToMe(int slot)
{ CHECKEXISTANCE;
    UtiLogicSlot& s = m_slots[slot];
    if (s.m_connection == NULL) {
        return 0;
    }
    UtiLogicOneConnection& con = s.m_connection->m_connectedobjects.at(s.m_objindex);
    assert(con.m_object == this);
    assert(con.m_slot == slot);


    return con.m_connectioncount;
}


int UtiLogicObject::getSlotConnectionCountFromOtherObjects(int slot)
{
    UtiLogicSlot& s = m_slots[slot];
    if (s.m_connection == NULL) {
        return 0;
    }
    bool selfoutput = (s.m_flags & UTI_LOGIC_OUTPUT_SLOT) != 0;
    int res = 0;
    for (int i = 0; i < s.m_connection->m_connectedobjects.size(); i++) {
           UtiLogicOneConnection& con = s.m_connection->m_connectedobjects.at(i);
           if (con.m_object == this) continue;
           UtiLogicSlot& otherslot = con.m_object->m_slots[con.m_slot];
           bool otheroutput = (otherslot.m_flags & UTI_LOGIC_OUTPUT_SLOT) != 0;
           if (selfoutput != otheroutput) {
               res++;
           }

        }
    return res;

}


void UtiLogicObject::deleteObject()
{ CHECKEXISTANCE;
    disconnect();
}

void UtiLogicObject::tick()
{

}


void UtiLogicObject::think(TIME frametime)
{

}


CUtiString UtiLogicObject::toString()
{
  return UTISTR("UnknownLogicObject");
}

int UtiLogicObject::addToSticky()
{ CHECKEXISTANCE;
    assert(m_sticky == false);
    assert(m_stickyindex == -1);
    m_sticky = true;

         m_stickyindex= m_parent->addStickyObject(this);
         return m_stickyindex;
}

UtiLogicNot::UtiLogicNot(UtiLogic* parent) :UtiLogicObject(parent)
{
    m_slots.setSize(2);
    m_slots[1].m_flags = UTI_LOGIC_OUTPUT_SLOT;
}

UtiLogicNot::~UtiLogicNot()
{

}

void UtiLogicNot::update()
{ CHECKEXISTANCE;
    unsigned int input = getSlotData(0);
    input = ~input;
    setSlotData(1, input);
}

CUtiString UtiLogicNot::toString()
{
  return UTISTR("Not");
}

UtiLogicAddID::UtiLogicAddID(UtiLogic* parent) :UtiLogicObject(parent)
{
    m_slots.setSize(2);
    m_slots[1].m_flags = UTI_LOGIC_OUTPUT_SLOT;
    m_objectid = 0;
}

UtiLogicAddID::~UtiLogicAddID()
{

}



void UtiLogicAddID::update()
{ CHECKEXISTANCE;
    unsigned int input = getSlotData(0);
    input = input;
    setSlotData(1, input);
}

CUtiString UtiLogicAddID::toString()
{
  return UTISTR("AddID");
}

void UtiLogicAddID::setObjectID(int objid)
{
    m_objectid = objid;
}

int UtiLogicAddID::getObjectID()
{
return m_objectid;
}

HResult UtiLogicAddID::Read(CStream & Stream, int Flags, int version)
{
   m_objectid = Stream.getInt32();
   return 0;
}

HResult UtiLogicAddID::Write(CStream & Stream, int Flags, int version)
{
   Stream.writeInt32(m_objectid);
   return 0;
}


void UtiLogicAddID::receiveData(int slot, int eventid, int objectid, const Array<unsigned char> & data)
{
sendEvent(1, eventid, m_objectid, data);
}

UtiLogicGatter::UtiLogicGatter(UtiLogic* parent, TLogicGatterType type) : UtiLogicObject(parent)
{
    m_type = type;
    m_slots.setSize(3);
    m_slots[2].m_flags = UTI_LOGIC_OUTPUT_SLOT;

}

UtiLogicGatter::~UtiLogicGatter()
{


}

void UtiLogicGatter::update()
{ CHECKEXISTANCE;
    unsigned int input1 = getSlotData(0);
    unsigned int input2 = getSlotData(1);
    unsigned int output=0;

    switch (m_type) {
       case lgtAnd: {
        output = input1 & input2;
    }; break;
       case lgtOr: {
        output = input1 | input2;
    }; break;
       case lgtNand: {
        output = ~(input1 & input2);
    }; break;
       default: {
        assert(false);
    }; break;
    }


    setSlotData(2, output);

}

CUtiString UtiLogicGatter::toString()
{ CHECKEXISTANCE;
  switch (m_type) {
       case lgtAnd: {
        return UTISTR("And");
    }; break;
       case lgtOr: {
        return UTISTR("Or");
    }; break;
       case lgtNand: {
        return UTISTR("Nand");
    }; break;
       default: {
        assert(false);
    }; break;
    }
  return UTISTR("Unknown");
}

//FlipFlop

UtiFlipFlop::UtiFlipFlop(UtiLogic* parent) : UtiLogicObject(parent)
{

    m_slots.setSize(3);
    m_slots[2].m_flags = UTI_LOGIC_OUTPUT_SLOT;

}

UtiFlipFlop::~UtiFlipFlop()
{


}

void UtiFlipFlop::update()
{ CHECKEXISTANCE;
    unsigned int data = getSlotData(0);
    unsigned int readdata = getSlotData(1);


    if (readdata!= 0) {
        m_state = data;
        setSlotData(2, m_state);
    }





}


HResult UtiFlipFlop::Read(CStream & Stream, int Flags, int version)
{

    /*int localflags = */Stream.getUInt32();
    int newstate = Stream.getUInt32();
    m_state = newstate;
    writeState();
    //setSlotData(0,newstate);
    return 0;
}

HResult UtiFlipFlop:: Write(CStream & Stream, int Flags, int version)
{
    Stream.writeUInt32(0);

    Stream.writeUInt32(m_state);
    return 0;
}



int UtiFlipFlop::getState()
{ CHECKEXISTANCE;
  return m_state;
}

void UtiFlipFlop::setState(int b)
{ CHECKEXISTANCE;
  m_state = b;
  //  setSlotData(0,b ? 1 : 0);  // addToSticky();
  writeState();
}

void UtiFlipFlop::writeState()
{ CHECKEXISTANCE;
  setSlotData(2,m_state);  // addToSticky();
}

CUtiString UtiFlipFlop::toString()
{ CHECKEXISTANCE;

        return UTISTR("FlipFlop");

}


UtiLED::UtiLED(UtiLogic* parent): UtiLogicObject(parent)
{
    m_state = false;
    m_slots.setSize(1);
   // m_slots[2].m_flags = UTI_LOGIC_OUTPUT_SLOT;

}

UtiLED::~UtiLED()
{

}


bool UtiLED::getState()
{
   return m_state;
}

void UtiLED::update()
{ CHECKEXISTANCE;
   m_state = (getSlotData(0) & 1) != 0;
   notifyGUI();
}

CUtiString UtiLED::toString()
{
  return UTISTR("LED");
}

HResult UtiLED::Read(CStream & Stream, int Flags, int version)
{

    /*int localflags = */Stream.getUInt32();
    int newstate = Stream.getUInt32();
    m_state = (newstate != 0);
    return 0;
}

HResult UtiLED:: Write(CStream & Stream, int Flags, int version)
{
    Stream.writeUInt32(0);
    Stream.writeUInt32(m_state ? 1 : 0);
    return 0;
}

UtiSwitch::UtiSwitch(UtiLogic* parent): UtiLogicObject(parent)
{
    m_slots.setSize(1);
    m_slots[0].m_flags = UTI_LOGIC_OUTPUT_SLOT;
    m_state = false;
    
}

UtiSwitch::~UtiSwitch()
{

}

CUtiString UtiSwitch::toString()
{
  return UTISTR("Switch");
}

void UtiSwitch::update()
{ CHECKEXISTANCE;
   assert(m_state == (getSlotData(0) != 0));
}

HResult UtiSwitch::Read(CStream & Stream, int Flags, int version)
{

    /*int localflags =*/ Stream.getUInt32();
    int newstate = Stream.getUInt32();
    m_state = newstate != 0;
    writeState();
    //setSlotData(0,newstate);
    assert(newstate == 0 || newstate == 1);
    return 0;
}

HResult UtiSwitch:: Write(CStream & Stream, int Flags, int version)
{
    Stream.writeUInt32(0);
    
    Stream.writeUInt32(m_state ? 1 : 0);
    return 0;
}

bool UtiSwitch::getState()
{ CHECKEXISTANCE;
  return m_state;
}

void UtiSwitch::setState(bool b)
{ CHECKEXISTANCE;
  m_state = b;
  //  setSlotData(0,b ? 1 : 0);  // addToSticky();
  writeState();
}

void UtiSwitch::writeState()
{ CHECKEXISTANCE;
  setSlotData(0,m_state ? 1 : 0);  // addToSticky();
}

//UtiClock

UtiClock::UtiClock(UtiLogic* parent): UtiLogicObject(parent)
{
    m_slots.setSize(1);
    m_slots[0].m_flags = UTI_LOGIC_OUTPUT_SLOT;
    m_state = false;

    m_pulsewidth = 100;
    m_phase = 0;
    m_dutywidth = 50;

}

UtiClock::~UtiClock()
{

}

CUtiString UtiClock::toString()
{
  return UTISTR("Clock");
}

void UtiClock::update()
{ CHECKEXISTANCE;
   assert(m_state == (getSlotData(0) != 0));
}

void UtiClock::tick()
{
  int tick = m_parent->getCurrentTick()+m_phase;
  int pulsewidth = umax(2, m_pulsewidth);
  tick = tick % pulsewidth;
  if (tick < m_dutywidth) {
      setState(true);
  } else {
      setState(false);

  }
}


void UtiClock::setPulseWidth(int w)
{
    m_pulsewidth = w;
}

void UtiClock::setPhase(int p)
{
    m_phase = p;
}

void UtiClock::setDutyWidth(int d)
{
    m_dutywidth = d;
}


HResult UtiClock::Read(CStream & Stream, int Flags, int version)
{

    /*int localflags =*/ Stream.getUInt32();
    int newstate = Stream.getUInt32();
    m_pulsewidth = Stream.getInt32();
    m_phase = Stream.getInt32();
    m_dutywidth = Stream.getInt32();
    m_state = newstate != 0;
    writeState();
    //setSlotData(0,newstate);
    assert(newstate == 0 || newstate == 1);
    return 0;
}

HResult UtiClock:: Write(CStream & Stream, int Flags, int version)
{
    Stream.writeUInt32(0);

    Stream.writeUInt32(m_state ? 1 : 0);
    Stream.writeInt32(m_pulsewidth);
    Stream.writeInt32(m_phase);
    Stream.writeInt32(m_dutywidth);
    return 0;
}

bool UtiClock::getState()
{ CHECKEXISTANCE;
  return m_state;
}

void UtiClock::setState(bool b)
{ CHECKEXISTANCE;
    if (b == m_state)return;
  m_state = b;
  //  setSlotData(0,b ? 1 : 0);  // addToSticky();
  writeState();
}

void UtiClock::writeState()
{ CHECKEXISTANCE;
  setSlotData(0,m_state ? 1 : 0);  // addToSticky();
}



void UtiLogicObjectEventBuffer::addToQueue(int slot, int eventid, int objectid, const Array<unsigned char> & data)
{
    if (m_eventbuffersize + data.size() > m_eventbuffermaxsize) return;
    UtiLogic_SEventBufferEntry* entry = new UtiLogic_SEventBufferEntry;
    entry->m_eventid = eventid;
    entry->m_objectid = objectid;
    entry->m_slot = slot;
    entry->m_data.setSize(data.size());
    data.copyToArrayHighSpeed(0, data.size(), entry->m_data, 0);
    m_eventbuffersize += data.size()+8;
    m_dataqueue.push(uti_refptr<UtiLogic_SEventBufferEntry> (entry));
}

uti_refptr<UtiLogic_SEventBufferEntry> UtiLogicObjectEventBuffer::getFromQueue()
{

    if (m_dataqueue.size() == 0) return uti_refptr<UtiLogic_SEventBufferEntry>(NULL);
    uti_refptr<UtiLogic_SEventBufferEntry> e(m_dataqueue.front());
    CHECKEXISTANCEOF(e.get());
    m_dataqueue.pop();
    m_eventbuffersize -= e->m_data.size()+8;
    assert(m_eventbuffersize >= 0);
    return e;

}

UtiLogicObjectEventBuffer::UtiLogicObjectEventBuffer(UtiLogic* parent) : UtiLogicObject(parent)
{
   m_eventbuffermaxsize = 6553;
   m_eventbuffersize = 0;
}

UtiLogicObjectEventBuffer::~UtiLogicObjectEventBuffer()
{

}

void UtiLogicObjectEventBuffer::receiveData(int slot, int eventid, int objectid, const Array<unsigned char> & data)
{
addToQueue(slot, eventid, objectid, data);
}

void UtiLogicConnection::makeSticky()
{ CHECKEXISTANCE;
    int c = m_connectedobjects.size();
    for (int i = 0; i < c; i++) {
        UtiLogicOneConnection& lo = m_connectedobjects[i];
        if ((lo.m_flags & UTI_LOGIC_OUTPUT_SLOT) == 0) {
            lo.m_object->makeSticky();
	    //std::wcout << "Made Sticky: " << lo.m_object->toString() << std::endl;
        }
    }
}

void UtiLogicConnection::sendData(int eventid, int objectid, const Array<unsigned char> & data)
{ CHECKEXISTANCE;
    int c = m_connectedobjects.size();
    for (int i = 0; i < c; i++) {
        UtiLogicOneConnection& lo = m_connectedobjects[i];
        if ((lo.m_flags & UTI_LOGIC_OUTPUT_SLOT) == 0) {
            lo.m_object->receiveData(lo.m_slot, eventid, objectid, data);
            //std::wcout << "Made Sticky: " << lo.m_object->toString() << std::endl;
        }
    }
}



bool UtiLogicConnection::hasConnection(UtiLogicObject* obj, int slot)
{ CHECKEXISTANCE;
  UtiLogicSlot& s = obj->getSlot(slot);
  if (s.m_connection == this) {
      return true;
   } else {
      return false;
   }
}


void UtiLogicConnection::rewriteSlotData()
{
  CHECKEXISTANCE;
    int c = m_connectedobjects.size();
    for (int i = 0; i < c; i++) {
        UtiLogicOneConnection& lo = m_connectedobjects[i];
        lo.m_object->rewriteSlotData();
    }
}

void UtiLogicConnection::addObject(UtiLogicObject* obj, int slot, int amount)
{ CHECKEXISTANCE;
   UtiLogicSlot& s = obj->getSlot(slot);
   if (s.m_connection == this) {
       assert(s.m_objindex >= 0 && s.m_objindex < m_connectedobjects.size());
       UtiLogicOneConnection& lo = m_connectedobjects[s.m_objindex];
       assert(lo.m_object == obj);
       assert(lo.m_slot == slot);
       assert(lo.m_connectioncount+amount <=255);
       lo.m_connectioncount+= amount;
       assert(lo.m_connectioncount <= 1);
   } else {
       assert(s.m_connection == NULL);
       UtiLogicOneConnection con;
       con.m_connectioncount = amount;
       assert(amount <= 255);
       con.m_flags = s.m_flags;
       con.m_slot = slot;
       con.m_object = obj;
       s.m_connection = this;
       s.m_objindex = m_connectedobjects.size();
       m_connectedobjects.push_back(con);
   }
}

void UtiLogicConnection::swapIndex(int index1, int index2)
{ CHECKEXISTANCE;
    UtiLogicOneConnection& lo_index1 = m_connectedobjects[index1];
    UtiLogicOneConnection& lo_index2 = m_connectedobjects[index2];
    assert(lo_index1.m_object->getSlot(lo_index1.m_slot).m_connection == this);
    assert(lo_index1.m_object->getSlot(lo_index1.m_slot).m_objindex == index1);

    assert(lo_index2.m_object->getSlot(lo_index2.m_slot).m_connection == this);
    assert(lo_index2.m_object->getSlot(lo_index2.m_slot).m_objindex == index2);

    lo_index1.m_object->getSlot(lo_index1.m_slot).m_objindex = index2;
    lo_index2.m_object->getSlot(lo_index2.m_slot).m_objindex = index1;

    UtiLogicOneConnection temp;
    temp = lo_index1;
    lo_index1 = lo_index2;
    lo_index2 = temp;

}

void UtiLogicConnection::removeObject(UtiLogicObject* obj, int slot, int amount)
{ CHECKEXISTANCE;
    UtiLogicSlot& s = obj->getSlot(slot);
    assert(s.m_connection == this);
    assert(s.m_objindex >= 0 && s.m_objindex < m_connectedobjects.size());
    UtiLogicOneConnection& lo = m_connectedobjects[s.m_objindex];
    assert(lo.m_object == obj);
    assert(lo.m_slot == slot);
    assert(lo.m_connectioncount>= amount);
    lo.m_connectioncount-= amount;
    if (lo.m_connectioncount == 0) {
        int objindex = s.m_objindex;
        //Change index
        if (objindex != m_connectedobjects.size()-1) {
            swapIndex(objindex, m_connectedobjects.size()-1);
        }
        s.m_connection = NULL;
        
	assert(s.m_objindex == m_connectedobjects.size()-1);
        s.m_objindex = -1;
        m_connectedobjects.pop_back();
    }

}


void UtiLogicConnection::print()
{
   for (int i = 0; i < m_connectedobjects.size(); i++) {
       UtiLogicOneConnection& c  = m_connectedobjects.at(i);
       std::wcout << "  <> " << c.m_object->toString() << " slot = " << c.m_slot << " object = " << c.m_object <<" flags = " <<c.m_flags << std::endl;
   }
}

UtiLogicConnection::UtiLogicConnection(UtiLogic *parent)
{
    m_parent = parent;
    m_data = 0;
}

UtiLogicConnection::~UtiLogicConnection()
{
   assert(m_connectedobjects.size() == 0);
}

UtiLogic::UtiLogic() : m_stickyobject(32),m_stickyobject2(32), m_guiobject(32), m_stickyobject_grow(m_stickyobject), m_stickyobject2_grow(m_stickyobject2), m_guiobject_grow(m_guiobject)
{
    m_currenttick = 0;
}

UtiLogic::~UtiLogic()
{
   m_logicconnections.clear();
}

UtiLogicConnection* UtiLogic::createWire()
{ CHECKEXISTANCE;
    UtiLogicConnection* connection = new UtiLogicConnection(this);
    m_logicconnections.push_back(connection);
    return connection;
}

void UtiLogic::destroyWire(UtiLogicConnection* con)
{ CHECKEXISTANCE;
   // UtiLogicConnection* connection = new UtiLogicConnection(this);
    CHECKEXISTANCEOF(con);
    assert(con->m_connectedobjects.size() == 0);
    m_logicconnections.remove(con);
}

void UtiLogic::connectTo(UtiLogicConnection* conn, UtiLogicObject* obj, int slot)
{ CHECKEXISTANCE;
    CHECKEXISTANCEOF(conn);
    CHECKEXISTANCEOF(obj);
   conn->addObject(obj, slot);
   obj->rewriteSlotData();
   conn->makeSticky();
}

void UtiLogic::disconnectFrom(UtiLogicConnection* conn, UtiLogicObject* obj, int slot)
{ CHECKEXISTANCE;
    CHECKEXISTANCEOF(conn);
    CHECKEXISTANCEOF(obj);
   conn->removeObject(obj, slot);
   if (obj->isOutputSlot(slot)) {
     conn->m_data = 0;
     conn->rewriteSlotData();
   };
   conn->makeSticky();
}


bool UtiLogic::hasConnection(UtiLogicConnection* conn, UtiLogicObject* obj, int slot)
{ CHECKEXISTANCE;
  return conn->hasConnection(obj, slot);
}

void UtiLogic::addLogicObject(UtiLogicObject* obj)
{ CHECKEXISTANCE;
  // m_logicobjects.push_back(obj);
   if (obj->needThink()) {
       m_logicthinkobjects.push_back(obj);
   }
}

int UtiLogic::getCurrentTick()
{
    return m_currenttick;
}

void UtiLogic::tick()
{ CHECKEXISTANCE;
    int c = m_stickyobject_grow.size();
    m_stickyobject2_grow.reset(16384,0);
    for (int i = 0; i < c; i++) {
        if (m_stickyobject[i] != NULL) {
            m_stickyobject[i]->resetSticky();
			m_stickyobject2_grow.fast_add(m_stickyobject[i]);
        }
    }
    m_stickyobject_grow.reset(16384,0);
    c = m_stickyobject2_grow.size();
    for (int i = 0; i < c; i++) {
        m_stickyobject2[i]->update();
    }
    m_stickyobject2_grow.reset(16384,0);
  m_currenttick++;
    CIndexList<UtiLogicObject, 1>::iterator it;
    for (it = m_logicthinkobjects.begin(); it != m_logicthinkobjects.end(); it++) {
        it->tick();
    }

}

void UtiLogic::think(TIME frametime)
{ CHECKEXISTANCE;
   CIndexList<UtiLogicObject, 1>::iterator it;
   for (it = m_logicthinkobjects.begin(); it != m_logicthinkobjects.end(); it++) {
       it->think(frametime);
   }
   int c = m_guiobject_grow.size();
    for (int i = 0; i < c; i++) {
        if (m_guiobject[i] != NULL) {
           m_guiobject[i]->notifyRealGUI();
        }
    }
     m_guiobject_grow.reset(16384,0);
}
