/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef UTILOGIC_H
#define UTILOGIC_H

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "utilists.h"
#include "utiarray.h"

#include<vector>
#include<deque>
#include<queue>


class UtiLogicConnection;
class UtiLogic;
struct UtiLogicSlot;
class UtiLogicObject;
const int UTI_LOGIC_OUTPUT_SLOT=1;

/* Speichere komplettes Netzwerk in BigSegment: Vorteil: Kein Synchronisationsproblem
-


*/


class UtiLogic_SEventBufferEntry : public ReferenceUtiOB
{
public:
    int m_eventid;
    int m_objectid;
    int m_slot;
    Array<unsigned char> m_data;
};

struct UtiLogicSlot
{
   UtiLogicConnection* m_connection;
   int m_objindex;
   int m_flags;
   unsigned int m_lastdata;
   UtiLogicSlot()
   {
       m_connection = NULL;
       m_objindex = -1;
       m_flags = 0;
       m_lastdata = 0;
   }
};

struct UtiLogicOneConnection
{
    UtiLogicObject* m_object;
    unsigned short m_slot;
    unsigned char m_connectioncount;
    unsigned char m_flags;
    UtiLogicOneConnection()
    {
        m_object = NULL;
        m_slot = 0;
        m_connectioncount=0;
        m_flags = 0;
    }
};

class UtiLogicConnection : public CUtiOB
{
public:
    UtiLogic* m_parent;
    CList_IndexController clist_indexcontroller[2];

    unsigned int m_data;
    std::vector<UtiLogicOneConnection> m_connectedobjects;
    void swapIndex(int index1, int index2);
    void makeSticky();
    void addObject(UtiLogicObject* obj, int slot, int amount=1);
    void removeObject(UtiLogicObject* obj, int slot, int amount=1);
    bool hasConnection(UtiLogicObject* obj, int slot);
    void rewriteSlotData();
    void sendData(int eventid, int objectid, const Array<unsigned char> & data);
    inline int getConnectionCount() {return m_connectedobjects.size();};
    UtiLogicConnection(UtiLogic* parent);
    ~UtiLogicConnection();
    void print();
};

class UtiLogicObject : public ReferenceUtiOB
{
public:
    CList_IndexController clist_indexcontroller[2];

protected:
    bool m_sticky;
    bool m_guisticky;
    int m_stickyindex;
    int m_guiindex;
    Array<UtiLogicSlot> m_slots;
    UtiLogic* m_parent;
    CUtiOB* m_guireceiver;
public:
    UtiLogicObject(UtiLogic* parent);
    ~UtiLogicObject();
    inline void setGUIReceiver(CUtiOB* r) {m_guireceiver = r;}
    inline CUtiOB* getGUIReceiver() {return m_guireceiver;};
    
    inline void makeSticky()
    {
        if (!m_sticky) {
          addToSticky();
        }
    }
    virtual void receiveData(int slot, int eventid, int objectid, const Array<unsigned char> & data) {}
    int addToSticky();
    inline void resetSticky()
    {
        assert(m_sticky == true);
        assert(m_stickyindex != -1);
        m_sticky = false;
        m_stickyindex = -1;
    }

    inline UtiLogicSlot& getSlot(int i)
    {
        return m_slots[i];
    }
    inline int getSlotCount()
    {
        return m_slots.size();
    }
    int getSlotConnectionCountFromLogicConnectionToMe(int slot);
    int getSlotConnectionCountFromOtherObjects(int slot);
    int getOutputConnectionCount();
    bool hasOutputConnection();
    int getInputConnectionCount();
    bool hasInputConnection();
    inline unsigned int getSlotData(int slot)
    {
        UtiLogicSlot& s = m_slots[slot];
        if (s.m_connection) {
            //assert((s.m_flags & UTI_LOGIC_OUTPUT_SLOT) == 0);
            assert(s.m_connection->m_connectedobjects.at(s.m_objindex).m_object == this);
            return s.m_connection->m_data;
        } else {
	 return s.m_lastdata; 
	}
        return 0;
    }


    inline void setSlotData(int slot, unsigned int data)
    {
        UtiLogicSlot& s = m_slots[slot];
        if (s.m_connection) {
            assert((s.m_flags & UTI_LOGIC_OUTPUT_SLOT) != 0);
            assert(s.m_connection->m_connectedobjects.at(s.m_objindex).m_object == this);
            s.m_connection->m_data = data;
	    s.m_lastdata = data;
            s.m_connection->makeSticky();
        } else {
	  s.m_lastdata = data;
	}
    }
    inline void sendEvent(int slot, unsigned int eventid,int objectid, const Array<unsigned char> & data)
    {
        UtiLogicSlot& s = m_slots[slot];
        if (s.m_connection) {
            if ((s.m_flags & UTI_LOGIC_OUTPUT_SLOT) == 0) return;
            assert(s.m_connection->m_connectedobjects.at(s.m_objindex).m_object == this);
            //s.m_connection->m_data = data;
           // s.m_lastdata = data;
            s.m_connection->sendData(eventid,objectid, data);
        }
    }
    inline bool isOutputSlot(int slot)
    {
      UtiLogicSlot& s = m_slots[slot];
      return (s.m_flags & UTI_LOGIC_OUTPUT_SLOT) != 0;
    }
    virtual void notifyGUI();
    void notifyRealGUI();
    virtual void rewriteSlotData();

    virtual void update();
    virtual void tick();
    virtual void think(TIME frametime);
    virtual bool needThink(){return false;};
    virtual void disconnect();
    virtual void deleteObject();
    virtual CUtiString toString();
    virtual HResult Read(CStream & Stream, int Flags, int version){return 0;};
    virtual HResult Write(CStream & Stream, int Flags, int version){return 0;};
    //virtual bool needSlowThink(){return false;};
};



class UtiLogicObjectEventBuffer : public UtiLogicObject
{
public:

protected:
    int m_eventbuffersize;
    int m_eventbuffermaxsize;
    std::queue< uti_refptr<UtiLogic_SEventBufferEntry> > m_dataqueue;
    void addToQueue(int slot,int eventid, int objectid, const Array<unsigned char> & data);
public:
    UtiLogicObjectEventBuffer(UtiLogic* parent);
    ~UtiLogicObjectEventBuffer();
    uti_refptr<UtiLogic_SEventBufferEntry> getFromQueue();
    virtual void receiveData(int slot, int eventid, int objectid, const Array<unsigned char> & data);
};

class UtiLogicNot : public UtiLogicObject
{
public:
    UtiLogicNot(UtiLogic* parent);
    ~UtiLogicNot();
    virtual void update();
    virtual CUtiString toString();
};

class UtiLogicAddID : public UtiLogicObject
{
protected:
    int m_objectid;
public:
    UtiLogicAddID(UtiLogic* parent);
    ~UtiLogicAddID();
    virtual void setObjectID(int objid);
    virtual int getObjectID();
    virtual void update();
    virtual CUtiString toString();

    virtual HResult Read(CStream & Stream, int Flags, int version);
    virtual HResult Write(CStream & Stream, int Flags, int version);
    virtual void receiveData(int slot, int eventid, int objectid, const Array<unsigned char> & data);
};

enum TLogicGatterType {lgtAnd, lgtOr, lgtNand};

class UtiLogicGatter : public UtiLogicObject
{
protected:
    TLogicGatterType m_type;
public:
    UtiLogicGatter(UtiLogic* parent, TLogicGatterType type);
    ~UtiLogicGatter();
    virtual void update();
    virtual CUtiString toString();
};


class UtiFlipFlop : public UtiLogicObject
{
protected:
   int m_state;
public:
    UtiFlipFlop(UtiLogic* parent);
    ~UtiFlipFlop();
    virtual void update();
    virtual void writeState();
    virtual void setState(int b);
    virtual int getState();
    virtual HResult Read(CStream & Stream, int Flags, int version);
    virtual HResult Write(CStream & Stream, int Flags, int version);
    virtual CUtiString toString();
};



class UtiLED : public UtiLogicObject
{
protected:
    bool m_state;
public:
    UtiLED(UtiLogic* parent);
    ~UtiLED();
    virtual bool getState();
    virtual void update();
    virtual HResult Read(CStream & Stream, int Flags, int version);
    virtual HResult Write(CStream & Stream, int Flags, int version);
    virtual CUtiString toString();
};

class UtiSwitch : public UtiLogicObject
{
protected:
  bool m_state;
public:
    UtiSwitch(UtiLogic* parent);
    ~UtiSwitch();
    virtual void writeState();
    virtual void setState(bool b);
    virtual bool getState();
    virtual void update();
    virtual HResult Read(CStream & Stream, int Flags, int version);
    virtual HResult Write(CStream & Stream, int Flags, int version);
    virtual CUtiString toString();
};

//Clock, FlipFlop!

class UtiClock : public UtiLogicObject
{
protected:
    int m_pulsewidth;
    int m_phase;
    int m_dutywidth;
    bool m_state;
public:
    UtiClock(UtiLogic* parent);
    ~UtiClock();
    virtual void tick();
    virtual void writeState();
    virtual void setPulseWidth(int w);
    virtual void setPhase(int p);
    virtual void setDutyWidth(int d);

    virtual void setState(bool b);

    virtual bool getState();
    virtual bool needThink() {return true;};
    virtual void update();
    virtual HResult Read(CStream & Stream, int Flags, int version);
    virtual HResult Write(CStream & Stream, int Flags, int version);
    virtual CUtiString toString();
};

class UtiLogic : public ReferenceUtiOB
{
protected:
   // COwnIndexList<UtiLogicObject, 0> m_logicobjects;
    int m_currenttick;

    //CIndexList<UtiLogicObject, 2> m_slowthinkobjects;
    COwnIndexList<UtiLogicConnection, 0> m_logicconnections;
    Array<UtiLogicObject*> m_stickyobject;
    Array<UtiLogicObject*> m_stickyobject2;
    Array<UtiLogicObject*> m_guiobject;
    UtiArrayAccessGrow<UtiLogicObject*, true> m_stickyobject_grow;
    UtiArrayAccessGrow<UtiLogicObject*, true> m_stickyobject2_grow;
    UtiArrayAccessGrow<UtiLogicObject*, true> m_guiobject_grow;
public:
    CIndexList<UtiLogicObject, 1> m_logicthinkobjects;
    inline int addStickyObject(UtiLogicObject* obj)
    {
		return m_stickyobject_grow.fast_add(obj);
    }
    inline int addNotifyGUI(UtiLogicObject* obj)
    {
		return m_guiobject_grow.fast_add(obj);
    }
    inline void removeStickyObject(UtiLogicObject* obj, int index)
    {
        assert(index < m_stickyobject_grow.size());
        assert(m_stickyobject[index]  == obj);
        m_stickyobject[index] = NULL;
    }
    inline void removeGuiStickyObject(UtiLogicObject* obj, int index)
    {
        assert(index < m_guiobject_grow.size());
        assert(m_guiobject[index]  == obj);
        m_guiobject[index] = NULL;
    }

public:
    UtiLogic();
    ~UtiLogic();
    UtiLogicConnection* createWire();
    void destroyWire(UtiLogicConnection* con);
    void connectTo(UtiLogicConnection* conn, UtiLogicObject* obj, int slot);
    void disconnectFrom(UtiLogicConnection* conn, UtiLogicObject* obj, int slot);
    bool hasConnection(UtiLogicConnection* conn, UtiLogicObject* obj, int slot);
    void addLogicObject(UtiLogicObject* obj);
    void tick();
    int getCurrentTick();
    void think(TIME frametime);
};

#endif // UTILOGIC_H
