/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef UTIMATHBOX_H
#define UTIMATHBOX_H

#include "Uti3DMath.h"
#include "utiintvector.h"
#include "UtiString.h"

struct CBox
{
    CVector m_min;
    CVector m_max;
    inline utifloat height()const  {return m_max.y - m_min.y;};
    inline CVector size() const {return m_max-m_min;};

    static CBox INIT() {        CBox r; r.init(); return r;};
    static CBox ADVANCEINIT() {        CBox r; r.m_min=CVector3(1000000, 1000000, 1000000); r.m_max=CVector3(-1000000, -1000000, -1000000);return r;};

    inline CBox() {};
    inline CBox(utifloat x1, utifloat y1, utifloat z1, utifloat x2, utifloat y2, utifloat z2) {m_min.x = x1; m_min.y = y1; m_min.z = z1; m_max.x = x2; m_max.y = y2; m_max.z = z2;};
    inline CBox(const CVector3& v1, const CVector3&v2) {m_min = v1; m_max=v2;};
    inline void init() {m_min = CVector3(0.0, 0.0,0.0); m_max = CVector3(0.0,0.0,0.0);};
    inline CVector3 center() const {return (m_max+m_min)/2.0f;};
    inline utifloat getWidth() const {return m_max.x-m_min.x;};
    inline utifloat getHeight()const  {return m_max.y-m_min.y;};
    inline utifloat getDepth()const  {return m_max.z-m_min.z;};
    inline void setWidth(utifloat v) {m_max.x = m_min.x+v;};
    inline void setHeight(utifloat v) {m_max.y = m_min.y+v;};
    inline void setDepth(utifloat v) {m_max.z = m_min.z+v;};
    inline utifloat getLeft()const  {return m_min.x;};
    inline utifloat getTop() const {return m_min.y;};
    inline utifloat getFront() const {return m_min.z;};
    inline utifloat getRight()const  {return m_max.x;};
    inline utifloat getBottom() const {return m_max.y;};
    inline utifloat getBack() const {return m_max.z;};
     inline CVector3 getSize()const  {return m_max-m_min;};
	 CPlane getPlane(int i, bool pout) const;
         CPlanef getPlanef(int i, bool pout) const;
     inline void increaseSize(const CVector3& v) {m_min-=v; m_max+= v;};
    inline void move(const CVector3& v) { m_min+= v; m_max += v;};
    inline void setLeft(utifloat v) {move(CVector3(v - m_min.x, 0,0));};
    inline void setTop(utifloat v) {move(CVector3(0, v - m_min.y,0));};
    inline void setFront(utifloat v) {move(CVector3(0, 0, v - m_min.z));};
    inline void correct() {if (m_min.x > m_max.x) {utiswap(m_min.x, m_max.x);}; if (m_min.y > m_max.y) {utiswap(m_min.y, m_max.y);}; if (m_min.z > m_max.z) {utiswap(m_min.z, m_max.z);};}
    inline bool contains(const CVector3& position) const {
       return (m_min.x <= position.x ) && (m_max.x >= position.x) && (m_min.y <= position.y) && (m_max.y >= position.y)&& (m_min.z <= position.z) && (m_max.z >= position.z);
    };
    inline bool intersectMathCorrect(const CVector3& position) const {
       return (m_min.x <= position.x ) && (m_max.x > position.x) && (m_min.y <= position.y) && (m_max.y > position.y)&& (m_min.z <= position.z) && (m_max.z > position.z);
    };
    inline utifloat volume() const {
        return getWidth()*getHeight()*getDepth();
    }
    inline CVector3 interpolate(const CVector3& v)const
    {
        return CVector3(m_min.x+(m_max.x-m_min.x)*v.x,m_min.y+(m_max.y-m_min.y)*v.y,m_min.z+(m_max.z-m_min.z)*v.z);
    }
	template<class T> inline void expandArray(T &ar, int startindex = 0, int count = 1)
	{
		for (int i = 0; i < count; i++)
		{
			expand(ar[i + startindex]);
		}
	}
    template<class T> inline void getEdgePoints(T &ar, int startindex = 0) const
	{
		ar[startindex+0] = CVector3(m_min.x, m_min.y, m_min.z);
		ar[startindex + 1] = CVector3(m_max.x, m_min.y, m_min.z);
		ar[startindex + 2] = CVector3(m_min.x, m_max.y, m_min.z);
		ar[startindex + 3] = CVector3(m_max.x, m_max.y, m_min.z);
		ar[startindex + 4] = CVector3(m_min.x, m_min.y, m_max.z);
		ar[startindex + 5] = CVector3(m_max.x, m_min.y, m_max.z);
		ar[startindex + 6] = CVector3(m_min.x, m_max.y, m_max.z);
		ar[startindex + 7] = CVector3(m_max.x, m_max.y, m_max.z);

	}
	inline bool contains(const CBox& v) const
	{
		utiassert(m_min.x <= m_max.x);
		utiassert(m_min.y <= m_max.y);
		utiassert(m_min.z <= m_max.z);
		
		if (m_min.x > v.m_min.x) return false;
		if (m_min.y > v.m_min.y) return false;
		if (m_min.z > v.m_min.z) return false;
		if (m_max.x < v.m_max.x) return false;
		if (m_max.y < v.m_max.y) return false;
		if (m_max.z < v.m_max.z) return false;
		return true;
	};
    inline void expand(const CBox& rect) {
      if (m_min.x > rect.m_min.x)m_min.x = rect.m_min.x;

      if (m_max.x < rect.m_max.x) m_max.x = rect.m_max.x;
      if (m_min.y > rect.m_min.y)m_min.y = rect.m_min.y;

      if (m_max.y < rect.m_max.y) m_max.y = rect.m_max.y;
      if (m_min.z > rect.m_min.z)m_min.z = rect.m_min.z;

      if (m_max.z < rect.m_max.z) m_max.z = rect.m_max.z;
    };
    inline void expand(const CVector3& v)
    {
         if (m_min.x > v.x)m_min.x = v.x;
      if (m_max.x < v.x) m_max.x = v.x;
      if (m_min.y > v.y)m_min.y = v.y;
      if (m_max.y < v.y) m_max.y = v.y;
      if (m_min.z > v.z)m_min.z = v.z;
      if (m_max.z < v.z) m_max.z = v.z;

    }
    inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y <= m_max.y && m_min.z <= m_max.z;
    }
    inline CBox intersectionCSG(const CBox& b) const
    {
        CBox res;
        res.m_min.x = umax(m_min.x, b.m_min.x);
        res.m_min.y = umax(m_min.y, b.m_min.y);
        res.m_min.z = umax(m_min.z, b.m_min.z);

        res.m_max.x = umin(m_max.x, b.m_max.x);
        res.m_max.y = umin(m_max.y, b.m_max.y);
        res.m_max.z = umin(m_max.z, b.m_max.z);
        return res;
    }
    inline bool intersect(const CBox& rect) const {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      if (m_min.x>= rect.m_max.x) return false;
      if (m_min.y>= rect.m_max.y) return false;
      if (m_min.z>= rect.m_max.z) return false;

      if (m_max.x<= rect.m_min.x) return false;
      if (m_max.y<= rect.m_min.y) return false;
      if (m_max.z<= rect.m_min.z) return false;

      return true;
    };

    inline CBox moveAndReturn(const CVector3& p) const
  {
          CBox b;
          b.m_min.x = m_min.x+  p.x;
          b.m_min.y = m_min.y + p.y;
          b.m_min.z = m_min.z + p.z;
          b.m_max.x = m_max.x + p.x;
          b.m_max.y = m_max.y + p.y;
          b.m_max.z = m_max.z + p.z;
          return b;
  };
    inline bool intersectMathCorrect(const CBox& rect) const {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      if (m_min.x> rect.m_max.x) return false;
      if (m_min.y> rect.m_max.y) return false;
      if (m_min.z> rect.m_max.z) return false;

      if (m_max.x<= rect.m_min.x) return false;
      if (m_max.y<= rect.m_min.y) return false;
      if (m_max.z<= rect.m_min.z) return false;

      return true;
    };
    inline bool operator==(const CBox& other) const
    {
        return m_max == other.m_max && m_min == other.m_min;
    };
    inline bool operator!=(const CBox& other) const {return !(operator==(other));};
    inline int getMaximumAxis() const
    {
        int res = 0;
    utifloat size = getWidth();
    if (getHeight() > size) {size = getHeight(); res = 1;};
    if (getDepth() > size) {size = getDepth(); res =2;};
    return res;
    };
    inline utifloat getMaxSize() const
    {
      return umax(getWidth(), umax(getHeight(), getDepth()));
    };
    inline CBox operator*(float v) const
    {
        CBox b;
        b.m_min = m_min*v;
        b.m_max = m_max*v;
        return b;
    }
    inline const CBox& operator*=(float v)
    {
        m_min*=v;
        m_max*=v;
        return *this;
    }
    inline const CBox& operator/=(float v)
    {
        m_min/=v;
        m_max/=v;
        return *this;
    }
    inline CBox operator/(float v) const
    {
        CBox b;
        b.m_min = m_min/v;
        b.m_max = m_max/v;
        return b;
    }

    inline CRect xy()const ;
    inline CRect xz()const ;
    inline CRect yz()const ;
	utifloat getMinMaxDistance(const CVector3& p)  const;
	utifloat getMinMaxDistanceXZ(const CVector3& p, float factor)  const;
    void octreeSubdivide(CBox* boxes) const;
    inline void createOctreeBox(int index, CBox& octreebox) const
    {
       utifloat xhalf = (m_max.x-m_min.x) / 2.0f;

       utifloat yhalf = (m_max.y-m_min.y)  / 2.0f;
       utifloat zhalf = (m_max.z-m_min.z) / 2.0f;

       utifloat v = (utifloat) (index & 1);
       octreebox.m_min.x = v*xhalf+m_min.x;
       octreebox.m_max.x = (v+1)*xhalf+m_min.x;
       v = (utifloat)( (index & 2) >> 1);
       octreebox.m_min.y = (v)*yhalf+m_min.y;
       octreebox.m_max.y = (v+1)*yhalf+m_min.y;
       v = (utifloat)( (index & 4) >> 2);
       octreebox.m_min.z = (v)*zhalf+m_min.z;
       octreebox.m_max.z = (v+1)*zhalf+m_min.z;


    }
};

struct COBB
{
    CVector3 m_origin;
    CVector3 m_bu, m_bv, m_bw;

    float m_halfu, m_halfv, m_halfw;
    inline COBB()
    {
       m_origin = CVector3(0.5, 0.5, 0.5);
       m_bu = CVector3(1,0,0);
       m_bv = CVector3(0,1,0);
       m_bw = CVector3(0,0,1);
       m_halfu = 0.5;
       m_halfv = 0.5;
       m_halfw = 0.5;

    }
    inline COBB(const CBox& b)
    {
        m_origin = b.center();
       m_bu = CVector3(1,0,0);
       m_bv = CVector3(0,1,0);
       m_bw = CVector3(0,0,1);
       m_halfu = b.getWidth()/2.0f;
       m_halfv = b.getHeight()/2.0f;
       m_halfw = b.getDepth()/2.0f;

    }
   inline  void move(const CVector3& v)
    {
    m_origin += v;
    }
    inline void transform(const CMatrix & m)
    {
        m_origin = m.TransformP(m_origin);
        m_bu = m.TransformNormal(m_bu);
        m_bv = m.TransformNormal(m_bv);
        m_bw = m.TransformNormal(m_bw);
        m_bu.normalize();
        m_bv.normalize();
        m_bw.normalize();
    }

    inline double getRadius() const
    {
        return umin(m_halfu, umin(m_halfv, m_halfw))*2.0f;
    }
    inline CVector3 getPositionInBox(const CVector3& v)
    {
        return m_origin + m_bu*(v.x-0.5)*2.0*m_halfu+ m_bv*(v.y-0.5)*2.0*m_halfv+ m_bw*(v.z-0.5)*2.0*m_halfw;
    }
    inline CVector3 getRandomPosition()
    {
        return getPositionInBox(CVector3(urandom(), urandom(), urandom()));
    }

    inline CBox getBoundingBox() const
    {
       CBox b;
       b = CBox::ADVANCEINIT();
       b.expand(m_origin-m_bu*m_halfu-m_bv*m_halfv-m_bw*m_halfw);
       b.expand(m_origin+m_bu*m_halfu-m_bv*m_halfv-m_bw*m_halfw);
       b.expand(m_origin-m_bu*m_halfu+m_bv*m_halfv-m_bw*m_halfw);
       b.expand(m_origin+m_bu*m_halfu+m_bv*m_halfv-m_bw*m_halfw);
       b.expand(m_origin-m_bu*m_halfu-m_bv*m_halfv+m_bw*m_halfw);
       b.expand(m_origin+m_bu*m_halfu-m_bv*m_halfv+m_bw*m_halfw);
       b.expand(m_origin-m_bu*m_halfu+m_bv*m_halfv+m_bw*m_halfw);
       b.expand(m_origin+m_bu*m_halfu+m_bv*m_halfv+m_bw*m_halfw);

       return b;
    }
    template<class T> inline void fillPoints(T& points) const
    {
       points[0] =m_origin-m_bu*m_halfu-m_bv*m_halfv-m_bw*m_halfw;
       points[1] =m_origin+m_bu*m_halfu-m_bv*m_halfv-m_bw*m_halfw;
       points[2] =m_origin-m_bu*m_halfu+m_bv*m_halfv-m_bw*m_halfw;
       points[3] =m_origin+m_bu*m_halfu+m_bv*m_halfv-m_bw*m_halfw;
       points[4] =m_origin-m_bu*m_halfu-m_bv*m_halfv+m_bw*m_halfw;
       points[5] =m_origin+m_bu*m_halfu-m_bv*m_halfv+m_bw*m_halfw;
       points[6] =m_origin-m_bu*m_halfu+m_bv*m_halfv+m_bw*m_halfw;
       points[7] =m_origin+m_bu*m_halfu+m_bv*m_halfv+m_bw*m_halfw;
    }
    template<class T> inline void fillPlanes(T& planes) const
    {
       planes[0] = CPlane::createPlane(m_origin+m_bu*m_halfu,m_bu);
       planes[1] = CPlane::createPlane(m_origin-m_bu*m_halfu,-m_bu);
       planes[2] = CPlane::createPlane(m_origin+m_bv*m_halfv,m_bv);
       planes[3] = CPlane::createPlane(m_origin-m_bv*m_halfv,-m_bv);
       planes[4] = CPlane::createPlane(m_origin+m_bw*m_halfw,m_bw);
       planes[5] = CPlane::createPlane(m_origin-m_bw*m_halfw,-m_bw);
    }
    bool intersect(const COBB& ob) const
    {
        CPlane myplanes[6];
        CPlane otherplanes[6];
        CVector3 mypoints[8];
        CVector3 otherpoints[8];
        fillPoints(mypoints);
        fillPlanes(myplanes);
        fillPoints(otherpoints);
        fillPlanes(otherplanes);
        for (int i = 0; i < 6;i++) {
            auto & p = myplanes[i];
            bool allonside = true;
            for (int j = 0; j < 8; j++) {
                 if (p.getPlanePos(otherpoints[j]) == CP_BACK) {
                  allonside = false;
                 }
            }
            if (allonside) return false;
        }
        return true;


    }
};


struct CBoxf
{
    CVector3f m_min;
    CVector3f m_max;
    inline float height()const  {return m_max.y - m_min.y;};
    inline CVector3f size() const {return m_max-m_min;};

    static CBoxf INIT() {        CBoxf r; r.init(); return r;};
    static CBoxf ADVANCEINIT() {        CBoxf r; r.m_min=CVector3f(1000000, 1000000, 1000000); r.m_max=CVector3f(-1000000, -1000000, -1000000);return r;};

    inline CBoxf() {};
    inline CBoxf(float x1, float y1, float z1, float x2, float y2, float z2) {m_min.x = x1; m_min.y = y1; m_min.z = z1; m_max.x = x2; m_max.y = y2; m_max.z = z2;};
    inline CBoxf(const CVector3f& v1, const CVector3f&v2) {m_min = v1; m_max=v2;};
    inline void init() {m_min = CVector3f(0.0, 0.0,0.0); m_max = CVector3f(0.0,0.0,0.0);};
    inline CVector3f center() const {return (m_max+m_min)/2.0f;};
    inline float getWidth() const {return m_max.x-m_min.x;};
    inline float getHeight()const  {return m_max.y-m_min.y;};
    inline float getDepth()const  {return m_max.z-m_min.z;};
    inline void setWidth(float v) {m_max.x = m_min.x+v;};
    inline void setHeight(float v) {m_max.y = m_min.y+v;};
    inline void setDepth(float v) {m_max.z = m_min.z+v;};
    inline float getLeft()const  {return m_min.x;};
    inline float getTop() const {return m_min.y;};
    inline float getFront() const {return m_min.z;};
    inline float getRight()const  {return m_max.x;};
    inline float getBottom() const {return m_max.y;};
    inline float getBack() const {return m_max.z;};
     inline CVector3f getSize()const  {return m_max-m_min;};
         CPlane getPlane(int i, bool pout) const;
         CPlanef getPlanef(int i, bool pout) const;
     inline void increaseSize(const CVector3f& v) {m_min-=v; m_max+= v;};
    inline void move(const CVector3f& v) { m_min+= v; m_max += v;};
    inline void setLeft(float v) {move(CVector3f(v - m_min.x, 0,0));};
    inline void setTop(float v) {move(CVector3f(0, v - m_min.y,0));};
    inline void setFront(float v) {move(CVector3f(0, 0, v - m_min.z));};
    inline void correct() {if (m_min.x > m_max.x) {utiswap(m_min.x, m_max.x);}; if (m_min.y > m_max.y) {utiswap(m_min.y, m_max.y);}; if (m_min.z > m_max.z) {utiswap(m_min.z, m_max.z);};}
    inline bool contains(const CVector3f& position) const {
       return (m_min.x <= position.x ) && (m_max.x >= position.x) && (m_min.y <= position.y) && (m_max.y >= position.y)&& (m_min.z <= position.z) && (m_max.z >= position.z);
    };
    inline bool intersectMathCorrect(const CVector3f& position) const {
       return (m_min.x <= position.x ) && (m_max.x > position.x) && (m_min.y <= position.y) && (m_max.y > position.y)&& (m_min.z <= position.z) && (m_max.z > position.z);
    };
    inline CVector3f interpolate(const CVector3f& v)const
    {
        return CVector3f(m_min.x+(m_max.x-m_min.x)*v.x,m_min.y+(m_max.y-m_min.y)*v.y,m_min.z+(m_max.z-m_min.z)*v.z);
    }
        template<class T> inline void expandArray(T &ar, int startindex = 0, int count = 1)
        {
                for (int i = 0; i < count; i++)
                {
                        expand(ar[i + startindex]);
                }
        }
        template<class T> inline void getEdgePoints(T &ar, int startindex = 0)
        {
                ar[startindex+0] = CVector3f(m_min.x, m_min.y, m_min.z);
                ar[startindex + 1] = CVector3f(m_max.x, m_min.y, m_min.z);
                ar[startindex + 2] = CVector3f(m_min.x, m_max.y, m_min.z);
                ar[startindex + 3] = CVector3f(m_max.x, m_max.y, m_min.z);
                ar[startindex + 4] = CVector3f(m_min.x, m_min.y, m_max.z);
                ar[startindex + 5] = CVector3f(m_max.x, m_min.y, m_max.z);
                ar[startindex + 6] = CVector3f(m_min.x, m_max.y, m_max.z);
                ar[startindex + 7] = CVector3f(m_max.x, m_max.y, m_max.z);

        }
        inline bool contains(const CBoxf& v) const
        {
                utiassert(m_min.x <= m_max.x);
                utiassert(m_min.y <= m_max.y);
                utiassert(m_min.z <= m_max.z);

                if (m_min.x > v.m_min.x) return false;
                if (m_min.y > v.m_min.y) return false;
                if (m_min.z > v.m_min.z) return false;
                if (m_max.x < v.m_max.x) return false;
                if (m_max.y < v.m_max.y) return false;
                if (m_max.z < v.m_max.z) return false;
                return true;
        };
    inline void expand(const CBoxf& rect) {
      if (m_min.x > rect.m_min.x)m_min.x = rect.m_min.x;

      if (m_max.x < rect.m_max.x) m_max.x = rect.m_max.x;
      if (m_min.y > rect.m_min.y)m_min.y = rect.m_min.y;

      if (m_max.y < rect.m_max.y) m_max.y = rect.m_max.y;
      if (m_min.z > rect.m_min.z)m_min.z = rect.m_min.z;

      if (m_max.z < rect.m_max.z) m_max.z = rect.m_max.z;
    };
    inline void expand(const CVector3f& v)
    {
         if (m_min.x > v.x)m_min.x = v.x;
      if (m_max.x < v.x) m_max.x = v.x;
      if (m_min.y > v.y)m_min.y = v.y;
      if (m_max.y < v.y) m_max.y = v.y;
      if (m_min.z > v.z)m_min.z = v.z;
      if (m_max.z < v.z) m_max.z = v.z;

    }
    inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y <= m_max.y && m_min.z <= m_max.z;
    }
    inline bool intersect(const CBoxf& rect) const {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      if (m_min.x>= rect.m_max.x) return false;
      if (m_min.y>= rect.m_max.y) return false;
      if (m_min.z>= rect.m_max.z) return false;

      if (m_max.x<= rect.m_min.x) return false;
      if (m_max.y<= rect.m_min.y) return false;
      if (m_max.z<= rect.m_min.z) return false;

      return true;
    };
    inline bool intersectMathCorrect(const CBoxf& rect) const {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      if (m_min.x> rect.m_max.x) return false;
      if (m_min.y> rect.m_max.y) return false;
      if (m_min.z> rect.m_max.z) return false;

      if (m_max.x<= rect.m_min.x) return false;
      if (m_max.y<= rect.m_min.y) return false;
      if (m_max.z<= rect.m_min.z) return false;

      return true;
    };
    inline bool operator==(const CBoxf& other) const
    {
        return other.m_max == other.m_max && other.m_min == other.m_min;
    };
    inline bool operator!=(const CBoxf& other) const {return !(operator==(other));};
    inline int getMaximumAxis() const
    {
        int res = 0;
    float size = getWidth();
    if (getHeight() > size) {size = getHeight(); res = 1;};
    if (getDepth() > size) {size = getDepth(); res =2;};
    return res;
    };
    inline float getMaxSize() const
    {
      return umax(getWidth(), umax(getHeight(), getDepth()));
    };


    inline void createOctreeBox(int index, CBoxf& octreebox) const
    {
       float xhalf = (m_max.x-m_min.x) / 2.0f;

       float yhalf = (m_max.y-m_min.y)  / 2.0f;
       float zhalf = (m_max.z-m_min.z) / 2.0f;

       float v = (float) (index & 1);
       octreebox.m_min.x = v*xhalf+m_min.x;
       octreebox.m_max.x = (v+1)*xhalf+m_min.x;
       v = (float)( (index & 2) >> 1);
       octreebox.m_min.y = (v)*yhalf+m_min.y;
       octreebox.m_max.y = (v+1)*yhalf+m_min.y;
       v = (float)( (index & 4) >> 2);
       octreebox.m_min.z = (v)*zhalf+m_min.z;
       octreebox.m_max.z = (v+1)*zhalf+m_min.z;


    }
};

CBox transformBox(const CMatrix& mat, const CBox& box);

inline CBox toDouble(const CBoxf& v)
{
   CBox m;
   m.m_min = toDouble(v.m_min);
   m.m_max = toDouble(v.m_max);

   return m;
}

inline CBoxf toFloat(const CBox& v)
{
   CBoxf m;
   m.m_min = toFloat(v.m_min);
   m.m_max = toFloat(v.m_max);

   return m;
}


unsigned int UtiRGBA(const CVector4&v);
unsigned int UtiRGBA(const CVector4f&v);

struct SHSPlane {
   CVector3 positive;
  CVector3 negative;
  //CVector4 plane;
  float plane_pos;
  CVector4 original;
  void load(const CVector4& plane);
};

enum TUtiBoxVis {ubvOFull=0, ubvOHalf, ubvONone};
TUtiBoxVis utiVisBox(const CBox & b, SHSPlane* planes, int planecount);


typedef CBox S3DRect;
struct CRectf;

struct CIntRect {
     typedef CIVector2 VectorClass;
  CIVector2 m_min;
  CIVector2 m_max;
 /* int y1;
  int x2;
  int y2;*/
  inline void move(int x, int y)
   {
       m_min.x += x;
        m_min.y += y;
       m_max.x += x;
       m_max.y += y;
   };
  inline int area() const
  {
      return width()*height();
  }

  inline CIVector2 size() const {return m_max-m_min;};
   inline int width() const {return m_max.x-m_min.x;};
   inline int height() const{return m_max.y-m_min.y;};
   inline void init() {m_min = CIVector2(0, 0); m_max = CIVector2(0,0);};
   inline bool intersect(const SPoint& p) const{ if (p.x < m_min.x || p.x > m_max.x || p.y < m_min.y || p.y> m_max.y) return false; return true;};
   inline bool intersect(const CIntRect& box) const {
                  if (box.m_max.x <= m_min.x || box.m_max.y <= m_min.y) return false;
                  if (box.m_min.x >= m_max.x || box.m_min.y >= m_max.y) return false;
          return true;
  };
   static CIntRect MAXINIT() {CIntRect r; r.m_max=CIVector2(-1000000, -1000000); r.m_min=CIVector2(1000000, 1000000);return r;};
   inline CIntRect()
   {
     m_min.x =m_min.y = m_max.x = m_max.y = 0;
   };
   inline explicit CIntRect(const CRect& r);
   inline explicit CIntRect(const CRectf& r);
   inline CIntRect(const CIVector2& m1, const CIVector2&m2)
   {
     m_min = m1;
     m_max = m2;
   };
   inline CIntRect intersectionCSG(const CIntRect& b) const
   {
       CIntRect res;
       res.m_min.x = umax(m_min.x, b.m_min.x);
       res.m_min.y = umax(m_min.y, b.m_min.y);

       res.m_max.x = umin(m_max.x, b.m_max.x);
       res.m_max.y = umin(m_max.y, b.m_max.y);
       return res;
   }
   inline CIntRect(int tx1, int ty1, int tx2, int ty2)
   {
     m_min.x = tx1;
     m_min.y = ty1;
     m_max.x = tx2;
     m_max.y = ty2;
   };
   inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y <= m_max.y;
    }
     inline bool contains(const CIVector2& v) const
    {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);


      if (m_min.x <= v.x && v.x < m_max.x && m_min.y <= v.y && v.y < m_max.y)return true;
      return false;
    };
     inline void expand(const CIVector2& v)
     {
          if (m_min.x > v.x)m_min.x = v.x;
       if (m_max.x <= v.x) m_max.x = v.x+1;
       if (m_min.y > v.y)m_min.y = v.y;
       if (m_max.y <= v.y) m_max.y = v.y+1;
     }
     inline void expand(const CIntRect& v)
     {
         expand(v.m_min);
         expand(v.m_max-CIVector2(1,1));
     }
     inline void expand(int x, int y)
     {
          if (m_min.x > x)m_min.x = x;
       if (m_max.x <= x) m_max.x = x+1;
       if (m_min.y > y)m_min.y = y;
       if (m_max.y <= y) m_max.y = y+1;
     }
     inline void expandSize(int x, int y)
     {
         m_min.x -= x;
          m_min.y -= y;
         m_max.x += x;
         m_max.y += y;
     };
   inline bool contains(const CIntRect& v) const
    {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(v.valid());

      if (m_min.x > v.m_min.x) return false;
      if (m_min.y > v.m_min.y) return false;
      if (m_max.x < v.m_max.x) return false;
      if (m_max.y < v.m_max.y) return false;
      return true;
    };
   inline CIVector2 center() const
   {
	   return CIVector2((m_min.x + m_max.x) / 2, (m_min.y + m_max.y) / 2);
   }
   inline void createQuadBox(const CIVector2& position, CIntRect& result) const
   {
      int xhalf = (m_max.x-m_min.x) >> 1;
      utiassert(xhalf*2 == width());
      int yhalf = (m_max.y-m_min.y) >> 1;
      utiassert(yhalf*2 == height());

      result.m_min.x = (position.x)*xhalf+m_min.x;
      result.m_max.x = (position.x+1)*xhalf+m_min.x;

      result.m_min.y = (position.y)*yhalf+m_min.y;
      result.m_max.y = (position.y+1)*yhalf+m_min.y;

   }
    inline bool operator==(const CIntRect& other) const
    {
        return m_max == other.m_max && m_min == other.m_min;
    };
    inline CIntRect operator+(const CIVector2& other) const
    {
        CIntRect r;
        r = *this;
        r.m_min += other;
        r.m_max += other;
        return r;
    };
    inline bool operator!=(const CIntRect& other) const
    {
        return !(operator==(other));
    };
   inline CBox getBounds(float zmin=0.0, float zmax=1.0) const {        CBox b; b.m_min = CVector3((float)m_min.x, (float)m_min.y, (float)zmin);b.m_max = CVector3((float)m_max.x, (float)m_max.y, (float)zmax);return b;};
};




struct CIntBox {
    typedef CIVector3 VectorClass;
  CIVector3 m_min;
  CIVector3 m_max;
 /* int y1;
  int x2;
  int y2;*/
  inline int volume() const
  {
      return (m_max.x-m_min.x)*(m_max.y-m_min.y)*(m_max.z-m_min.z);
  }

  inline void move(int x, int y, int z)
   {
       m_min.x += x;
        m_min.y += y;
        m_min.z += z;
       m_max.x += x;
       m_max.y += y;
       m_max.z += z;
   };
  inline void move(const CIVector3& p)
   {
       m_min.x += p.x;
        m_min.y += p.y;
        m_min.z += p.z;
       m_max.x += p.x;
       m_max.y += p.y;
       m_max.z += p.z;
   };
  inline CIntBox moveAndReturn(const CIVector3& p) const
  {
	  CIntBox b;
          b.m_min.x = m_min.x+  p.x;
	  b.m_min.y = m_min.y + p.y;
	  b.m_min.z = m_min.z + p.z;
	  b.m_max.x = m_max.x + p.x;
	  b.m_max.y = m_max.y + p.y;
	  b.m_max.z = m_max.z + p.z;
	  return b;
  };

  inline void setMinViaMove(int axis, int y) {
      CIVector3 m(0,0,0);
      m[axis] = y-m_min[axis];
      move(m);
      assert(m_min[axis] == y);
  }
  inline void setMaxViaMove(int axis, int y) {
      CIVector3 m(0,0,0);
      m[axis] = y-m_max[axis];
      move(m);
      assert(m_max[axis] == y);
  }

  inline void setMinYViaMove(int y) {
      move(0,y-m_min.y, 0);
      assert(m_min.y == y);
  }
  inline void setMaxYViaMove(int y) {
      move(0,y-m_max.y, 0);
      assert(m_max.y == y);
  }

  inline void setMinXViaMove(int x) {
      move(x-m_min.x, 0,0);
      assert(m_min.x == x);
  }
  inline void setMaxXViaMove(int x) {
      move(x-m_max.x,0, 0);
      assert(m_max.x == x);
  }


  inline void setMinZViaMove(int z) {
      move(0,0,z-m_min.z);
      assert(m_min.z == z);
  }
  inline void setMaxZViaMove(int z) {
      move(0,0,z-m_max.z);
      assert(m_max.z == z);
  }


  inline CIntBox ret_min(int axis, int x)const
  {
    CIntBox ret = *this;
    ret.m_min[axis] = x;
    return ret;
  };

  inline CIntBox ret_max(int axis, int x)const
  {
    CIntBox ret = *this;
    ret.m_max[axis] = x;
    return ret;
  };

  inline CIntBox ret_v(int axis, int min, int max) const
  {
    CIntBox ret = *this;
    ret.m_min[axis] = min;
    ret.m_max[axis] = max;
    return ret;
  };
  inline CIntBox ret_v_delta(int axis, int min, int max) const
  {
    CIntBox ret = *this;
    ret.m_min[axis] += min;
    ret.m_max[axis] += max;
    return ret;
  };
   inline CIVector3 center() const
   {
     CIVector3 res;
     res.x = (m_min.x+m_max.x) >> 1;
     res.y = (m_min.y+m_max.y) >> 1;
     res.z = (m_min.z+m_max.z) >> 1;
     return res;
   };
   inline void expand(const CIntBox& rect) {
     if (m_min.x > rect.m_min.x)m_min.x = rect.m_min.x;

     if (m_max.x < rect.m_max.x) m_max.x = rect.m_max.x;
     if (m_min.y > rect.m_min.y)m_min.y = rect.m_min.y;

     if (m_max.y < rect.m_max.y) m_max.y = rect.m_max.y;
     if (m_min.z > rect.m_min.z)m_min.z = rect.m_min.z;

     if (m_max.z < rect.m_max.z) m_max.z = rect.m_max.z;
   };
   inline void expand(const CIVector3& v)
   {
       if (m_min.x > v.x) m_min.x = v.x;
       if (m_min.y > v.y) m_min.y = v.y;
       if (m_min.z > v.z) m_min.z = v.z;

       if (m_max.x < v.x+1) m_max.x = v.x+1;
       if (m_max.y < v.y+1) m_max.y = v.y+1;
       if (m_max.z < v.z+1) m_max.z = v.z+1;
   }
   inline void expandNumeric(const CIVector3& v)
   {
       if (m_min.x > v.x) m_min.x = v.x;
       if (m_min.y > v.y) m_min.y = v.y;
       if (m_min.z > v.z) m_min.z = v.z;

       if (m_max.x < v.x) m_max.x = v.x;
       if (m_max.y < v.y) m_max.y = v.y;
       if (m_max.z < v.z) m_max.z = v.z;
   }

   inline void expandSize(int x, int y, int z)
   {
       m_min.x -= x;
        m_min.y -= y;
        m_min.z -= z;
       m_max.x += x;
       m_max.y += y;
       m_max.z += z;
   };

   inline CIntBox expandSizeAndReturn(int x, int y, int z) const
   {
       CIntBox res = *this;
       res.m_min.x -= x;
        res.m_min.y -= y;
        res.m_min.z -= z;
       res.m_max.x += x;
       res.m_max.y += y;
       res.m_max.z += z;
       return res;
   };
   inline CIntBox getExpandedBox(const CIVector3&v) const
   {
	   CIntBox r;
	   r = *this;
	   r.m_min.x -= v.x;
	   r.m_min.y -= v.y;
	   r.m_min.z -= v.z;
	   r.m_max.x += v.x;
	   r.m_max.y += v.y;
	   r.m_max.z += v.z;
	   return r;
   };
   inline CIVector3 size() const{return m_max-m_min;};
   inline int width() const {return m_max.x-m_min.x;};
   inline int height() const{return m_max.y-m_min.y;};
   inline int depth() const{return m_max.z-m_min.z;};
   inline void init() {m_min = CIVector3(0, 0, 0); m_max = CIVector3(0,0, 0);};
   inline bool intersect(const CIVector3& p) const{ if (p.x < m_min.x || p.x >= m_max.x || p.y < m_min.y || p.y >= m_max.y || p.z < m_min.z || p.z>= m_max.z) return false; return true;};
   inline bool intersect(const CIntBox& box) const {
                  if (box.m_max.x <= m_min.x || box.m_max.y <= m_min.y || box.m_max.z <= m_min.z) return false;
                  if (box.m_min.x >= m_max.x || box.m_min.y >= m_max.y || box.m_min.z >= m_max.z) return false;
          return true;
  };
   inline static CIntBox MinMaxCheck(int x1, int y1, int z1, int x2, int y2, int z2)
   {
       CIntBox res;
       if (x1 <= x2) {res.m_min.x = x1; res.m_max.x = x2;} else {res.m_min.x = x2; res.m_max.x = x1;};
       if (y1 <= y2) {res.m_min.y = y1; res.m_max.y = y2;} else {res.m_min.y = y2; res.m_max.y = y1;};
       if (x1 <= x2) {res.m_min.z = z1; res.m_max.z = z2;} else {res.m_min.z = z2; res.m_max.z = z1;};
       return res;
   }

   static CIntBox MAXINIT() {CIntBox r; r.m_min=CIVector3(1000000, 1000000, 1000000); r.m_max=CIVector3(-1000000, -1000000, -1000000);return r;};
   inline CIntBox()
   {
     //m_min.x =m_min.y = m_min.z = m_max.x = m_max.y = m_max.z= 0;
   };
   inline CIntBox(const CIVector3& m1, const CIVector3& m2)
   {
     m_min = m1;
     m_max = m2;
   };
   inline CIntBox(int tx1, int ty1, int tz1, int tx2, int ty2, int tz2)
   {
     m_min.x = tx1;
     m_min.y = ty1;
     m_min.z = tz1;
     m_max.x = tx2;
     m_max.y = ty2;
     m_max.z = tz2;
   };

   inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y <= m_max.y && m_min.z <= m_max.z;
    }
   inline bool contains(const CIntBox& v) const
    {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      utiassert(v.valid());

      if (m_min.x > v.m_min.x) return false;
      if (m_min.y > v.m_min.y) return false;
      if (m_min.z > v.m_min.z) return false;
      if (m_max.x < v.m_max.x) return false;
      if (m_max.y < v.m_max.y) return false;
      if (m_max.z < v.m_max.z) return false;
      return true;
    };

    inline bool contains(const CIntBox& v, int expandtestboxmin, int expandtestboxmax) const
    {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      utiassert(v.valid());

      if (m_min.x > v.m_min.x-expandtestboxmin) return false;
      if (m_min.y > v.m_min.y-expandtestboxmin) return false;
      if (m_min.z > v.m_min.z-expandtestboxmin) return false;
      if (m_max.x < v.m_max.x+expandtestboxmax) return false;
      if (m_max.y < v.m_max.y+expandtestboxmax) return false;
      if (m_max.z < v.m_max.z+expandtestboxmax) return false;
      return true;
    };
    inline bool contains(const CIVector3& v) const
    {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);


      if (m_min.x > v.x) return false;
      if (m_min.y > v.y) return false;
      if (m_min.z > v.z) return false;
      if (m_max.x <= v.x) return false;
      if (m_max.y <= v.y) return false;
      if (m_max.z <= v.z) return false;
      return true;
    };
    inline bool operator==(const CIntBox& other) const
    {
        return m_max == other.m_max && m_min == other.m_min;
    };
    inline bool operator!=(const CIntBox& other) const
    {
        return !(operator==(other));
    }
    inline const CIntBox& operator+=(const CIVector3& v)
    {
         m_min+=v;
         m_max += v;
            return *this;
     }
    inline const CIntBox& operator-=(const CIVector3& v)
{
     m_min-=v;
     m_max -= v;
        return *this;
 }

    inline int getDistanceToBorder(const CIVector3& v) const
    {
      //assert(contains(v) == true);
      int r = umin(umin(umin(v.x-m_min.x, v.y-m_min.y), umin(v.z-m_min.z, m_max.x-1-v.x)),
          umin(m_max.y-1-v.y, m_max.z-1-v.z));
      utiassert(r >= 0);
      return r;

    }
    inline void createOctreeBox(const CIVector3& position, CIntBox& octreebox) const
    {
       int xhalf = (m_max.x-m_min.x) >> 1;
       utiassert(xhalf*2 == m_max.x-m_min.x);
       int yhalf = (m_max.y-m_min.y) >> 1;
       utiassert(yhalf*2 == m_max.y-m_min.y);
       int zhalf = (m_max.z-m_min.z) >> 1;
       utiassert(zhalf*2 == m_max.z-m_min.z);

       octreebox.m_min.x = (position.x)*xhalf+m_min.x;
       octreebox.m_max.x = (position.x+1)*xhalf+m_min.x;

       octreebox.m_min.y = (position.y)*yhalf+m_min.y;
       octreebox.m_max.y = (position.y+1)*yhalf+m_min.y;

       octreebox.m_min.z = (position.z)*zhalf+m_min.z;
       octreebox.m_max.z = (position.z+1)*zhalf+m_min.z;


    }
    inline void createOctreeBox(int index, CIntBox& octreebox) const
    {
       int xhalf = (m_max.x-m_min.x)>> 1;

       int yhalf = (m_max.y-m_min.y) >> 1;
       int zhalf = (m_max.z-m_min.z)>> 1;
        utiassert(xhalf*2 == width());
    utiassert(yhalf*2 == height());
     utiassert(zhalf*2 == depth());
       int v = (int) (index & 1);
       octreebox.m_min.x = v*xhalf+m_min.x;
       octreebox.m_max.x = (v+1)*xhalf+m_min.x;
       v = (int)( (index & 2) >> 1);
       octreebox.m_min.y = (v)*yhalf+m_min.y;
       octreebox.m_max.y = (v+1)*yhalf+m_min.y;
       v = (int)( (index & 4) >> 2);
       octreebox.m_min.z = (v)*zhalf+m_min.z;
       octreebox.m_max.z = (v+1)*zhalf+m_min.z;


    }
     inline void getOctreePositionOld(const CIVector3&p, CIVector3&position, int &index) const
    {
       index = 0;
    CIntBox octreebox;

       for (position.z = 0; position.z < 2; position.z++) {
     for (position.y = 0; position.y < 2; position.y++) {
       for (position.x = 0; position.x < 2; position.x++) {

       createOctreeBox(position, octreebox);
        if (octreebox.contains(p)) {
          return;
        };
        index++;
       }}};
     utiassert(false);
    }
    inline void getOctreePosition(const CIVector3&p, CIVector3&position, int &index) const
    {
      CIVector3 middle;
      middle.x = (m_min.x+m_max.x) >> 1;
      middle.y = (m_min.y+m_max.y) >> 1;
      middle.z = (m_min.z+m_max.z) >> 1;
      position.x = p.x >= middle.x ? 1 : 0;
      position.y = p.y >= middle.y ? 1 : 0;
      position.z = p.z >= middle.z ? 1 : 0;
      index = position.z << 2 | position.y << 1 | position.x;
      /*CIVector3 temppos;
      int tempindex;
      getOctreePositionOld(p, temppos, tempindex);
      utiassert(temppos == position);
      utiassert(tempindex == index);*/


    }
    inline CIntBox intersection(const CIntBox& b) const
    {
        CIntBox res;
        res.m_min.x = umax(m_min.x, b.m_min.x);
        res.m_min.y = umax(m_min.y, b.m_min.y);
        res.m_min.z = umax(m_min.z, b.m_min.z);

        res.m_max.x = umin(m_max.x, b.m_max.x);
        res.m_max.y = umin(m_max.y, b.m_max.y);
        res.m_max.z = umin(m_max.z, b.m_max.z);
        return res;
    }
    inline CBox getAsBox(float scal) const
    {
      CBox b;
      b.m_min.x = ((float)m_min.x) * scal;
      b.m_min.y = ((float)m_min.y) * scal;
      b.m_min.z = ((float)m_min.z) * scal;


      b.m_max.x = ((float)m_max.x) * scal;
      b.m_max.y = ((float)m_max.y) * scal;
      b.m_max.z = ((float)m_max.z) * scal;

      return b;
    };
    inline CIntRect getAsRect() const
    {
      CIntRect b;
      b.m_min.x = (m_min.x);
      b.m_min.y = (m_min.y);


      b.m_max.x = (m_max.x);
      b.m_max.y = (m_max.y);

      return b;
    };
    inline CIntRect xy() const
    {
      CIntRect b;
      b.m_min.x = (m_min.x);
      b.m_min.y = (m_min.y);


      b.m_max.x = (m_max.x);
      b.m_max.y = (m_max.y);

      return b;
    };

    inline CIntRect xz() const
    {
      CIntRect b;
      b.m_min.x = (m_min.x);
      b.m_min.y = (m_min.z);


      b.m_max.x = (m_max.x);
      b.m_max.y = (m_max.z);

      return b;
    };

    inline CIntRect yz() const
    {
      CIntRect b;
      b.m_min.x = (m_min.y);
      b.m_min.y = (m_min.z);


      b.m_max.x = (m_max.y);
      b.m_max.y = (m_max.z);

      return b;
    };
    HResult Read(CStream & Stream, int Flags, int version);
    HResult Write(CStream & Stream, int Flags, int version);

};

#define FOREACH_VECTOR_BEGIN(size, pos) { CIVector3 pos; for (pos.z = 0; pos.z < size.z; pos.z++) {\
    for (pos.x = 0; pos.x < size.x; pos.x++) {\
    for (pos.y = 0; pos.y < size.y; pos.y++)   {


#define FOREACH_VECTOR_END };};};};

#define ENUMBOX_BEGIN(box, pos) { CIVector3 pos; for (pos.z = box.m_min.z; pos.z < box.m_max.z; pos.z++) {\
    for (pos.x = box.m_min.x; pos.x < box.m_max.x; pos.x++) {\
    for (pos.y = box.m_min.y; pos.y < box.m_max.y; pos.y++)   {

#define ENUMBOX_BEGINEXIT(box, pos, condition) { CIVector3 pos; for (pos.z = box.m_min.z; pos.z < box.m_max.z; pos.z++) { if (condition) break;\
    for (pos.x = box.m_min.x; pos.x < box.m_max.x; pos.x++) { if (condition) break;\
    for (pos.y = box.m_min.y; pos.y < box.m_max.y; pos.y++)   { if (condition) break;

#define ENUMBOX_END };};};};

#define ENUMRECT_BEGIN(box, pos) { CIVector2 pos; \
    for (pos.y = box.m_min.y; pos.y < box.m_max.y; pos.y++) {\
    for (pos.x = box.m_min.x; pos.x < box.m_max.x; pos.x++)   {

#define ENUMRECT_BEGINEXIT(box, pos, condition) { CIVector2 pos;\
    for (pos.y = box.m_min.y; pos.y < box.m_max.y; pos.y++) { if (condition) break;\
    for (pos.x = box.m_min.x; pos.x < box.m_max.x; pos.x++)   { if (condition) break;

#define ENUMRECT_END };};};


#define ENUMBOX_BEGIN_Y(box, pos) { CIVector3 pos; for (pos.y = box.m_min.y; pos.y < box.m_max.y; pos.y++)   {\
      for (pos.z = box.m_min.z; pos.z < box.m_max.z; pos.z++) {\
    for (pos.x = box.m_min.x; pos.x < box.m_max.x; pos.x++) {


#define ENUMBOX_END_Y };};};};

#define ENUMBOX_BEGIN_S(box, pos, s) { CIVector3 pos; \
      for (pos.z = box.m_min.z; pos.z < box.m_max.z; pos.z+=(s)) {\
    for (pos.x = box.m_min.x; pos.x < box.m_max.x; pos.x+=(s)) {\
    for (pos.y = box.m_min.y; pos.y < box.m_max.y; pos.y+=(s))   {\
    assert((pos.x % s) == 0);assert((pos.y % s) == 0);assert((pos.z % s) == 0);


#define ENUMBOX_END_S };};};};


inline CBox intBoxToFloat(const CIntBox& b)
{
    return CBox(intVecToFloat(b.m_min), intVecToFloat(b.m_max));
}


inline CIntBox floatBoxToInt(const CBox& b)
{
    return CIntBox(intVecToFloat(b.m_min), intVecToFloat(b.m_max));
}


int binomial(int n, int k);

struct CRect
{
     typedef CVector2 VectorClass;
    static inline CRect INIT() {CRect r; r.init(); return r;};
    static inline CRect ADVANCEINIT() {CRect r; r.m_min=CVector2(1000000, 1000000); r.m_max=CVector2(-1000000, -1000000);return r;};
    static inline CRect MAXINIT() {CRect r; r.m_max=CVector2(1000000, 1000000); r.m_min=CVector2(-1000000, -1000000);return r;};
    CVector2 m_min;
    CVector2 m_max;
    inline CRect() {};
    inline CRect(utifloat x1, utifloat y1, utifloat x2, utifloat y2) {m_min.x = x1; m_min.y = y1; m_max.x = x2; m_max.y = y2;};
    inline CRect(const CVector2& v1, const CVector2&v2) {m_min = v1; m_max=v2;};
    inline explicit CRect(const CIntRect& rect) {m_min.x = (utifloat)rect.m_min.x; m_min.y =(utifloat) rect.m_min.y; m_max.x =(utifloat) rect.m_max.x; m_max.y = (utifloat)rect.m_max.y;};

    inline CRect(const CVector2& pos, utifloat size) {m_min.x = pos.x-size/2.0f;m_min.y = pos.y-size/2.0f;m_max.x = pos.x+size/2.0f;m_max.y = pos.y+size/2.0f;};
    inline void init() {m_min = CVector2(0.0, 0.0); m_max = CVector2(0.0,0.0);};
    inline CBox getBounds(utifloat zmin=0.0, utifloat zmax=1.0) const {        CBox b; b.m_min = CVector3(m_min, zmin);b.m_max = CVector3(m_max, zmax);return b;};
    inline CVector2 getCoord(const CVector2& pos) const {
      return CVector2(m_min.x+(m_max.x-m_min.x)*pos.x,
              m_min.y+(m_max.y-m_min.y)*pos.y);
    };
    inline CVector2f getCoord(const CVector2f& pos) const {
      return CVector2f(m_min.x+(m_max.x-m_min.x)*pos.x,
              m_min.y+(m_max.y-m_min.y)*pos.y);
    };
    inline CVector2 getInvCoord(const CVector2& pos) const {
      return CVector2((pos.x-m_min.x)/(m_max.x-m_min.x),
              (pos.y-m_min.y)/(m_max.y-m_min.y));
    };
	inline CRect getCoordRect(const CRect& pos) const {
		return CRect(getCoord(pos.m_min), getCoord(pos.m_max));
	}
    inline CVector2 center() const {return (m_max+m_min)/2.0f;};
    void getRectAsArray(Array<CVector2> &marray) const;
    inline utifloat getWidth() const {return m_max.x-m_min.x;};
    inline utifloat getHeight()const  {return m_max.y-m_min.y;};
    inline void setWidth(utifloat v) {m_max.x = m_min.x+v;};
    inline void setHeight(utifloat v) {m_max.y = m_min.y+v;};
    inline CVector2 getSize() {return m_max-m_min;};
    inline utifloat getLeft()const  {return m_min.x;};
    inline utifloat getTop() const {return m_min.y;};
    inline utifloat getRight()const  {return m_max.x;};
    inline utifloat getBottom() const {return m_max.y;};
    inline void move(const CVector2& v) { m_min+= v; m_max += v;};
    inline void setLeft(utifloat v) {move(CVector2(v - m_min.x, 0));};
    inline void setTop(utifloat v) {move(CVector2(0, v - m_min.y));};
    inline void correct() {if (m_min.x > m_max.x) {utiswap(m_min.x, m_max.x);}; if (m_min.y > m_max.y) {utiswap(m_min.y, m_max.y);};}
    inline bool contains(const CVector2& position) const {
       return (m_min.x <= position.x ) && (m_max.x >= position.x) && (m_min.y <= position.y) && (m_max.y >= position.y);
    };
    inline CBox toBoxY(float miny, float maxy) const
    {
        return CBox(m_min.x, miny, m_min.y, m_max.x, maxy, m_max.y);
    }


    inline void expandRect(float w)
    {
      m_min.x -= w;
      m_min.y -= w;
      m_max.x += w;
      m_max.y += w;
    }
    inline void roundToInt()
    {
      m_min.x = (float)((int)m_min.x);
      m_min.y = (float)((int)m_min.y);
      m_max.x = (float)((int)m_max.x);
      m_max.y = (float)((int)m_max.y);
    };
    inline void expandRect(const CVector2& v)
    {
      m_min.x -= v.x;
      m_min.y -= v.y;
      m_max.x += v.x;
      m_max.y += v.y;
    }
    inline void expand(const CRect& rect) {
      if (m_min.x > rect.m_min.x)m_min.x = rect.m_min.x;

      if (m_max.x < rect.m_max.x) m_max.x = rect.m_max.x;
      if (m_min.y > rect.m_min.y)m_min.y = rect.m_min.y;

      if (m_max.y < rect.m_max.y) m_max.y = rect.m_max.y;
    };
    inline void expand(const CVector2& v)
    {
         if (m_min.x > v.x)m_min.x = v.x;
      if (m_max.x < v.x) m_max.x = v.x;
      if (m_min.y > v.y)m_min.y = v.y;
      if (m_max.y < v.y) m_max.y = v.y;

    }

    template <class T> inline void fromList(T begin, T end)
    {
        *this = ADVANCEINIT();
        while (begin != end) {
           expand(*begin);
           begin++;

        }
    }
    inline bool contains(const CRect& v) const
    {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(v.valid());

      if (m_min.x > v.m_min.x) return false;
      if (m_min.y > v.m_min.y) return false;
      if (m_max.x < v.m_max.x) return false;
      if (m_max.y < v.m_max.y) return false;
      return true;
    };

    inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y <= m_max.y;
    }
   inline bool intersect(const CRect& rect) const {
      //utiassert(m_min.x <= m_max.x);
      //utiassert(m_min.y <= m_max.y);
      if (m_min.x> rect.m_max.x) return false;
      if (m_min.y> rect.m_max.y) return false;
      if (m_max.x< rect.m_min.x) return false;
      if (m_max.y< rect.m_min.y) return false;
      return true;
    };
    inline bool operator==(const CRect& other) const
    {
        return m_max == other.m_max && m_min == other.m_min;
    };
    inline bool operator!=(const CRect& other) const {return !(operator==(other));};
    inline utifloat getArea()
    {
      return absf((m_max.x-m_min.x)*(m_max.y-m_min.y));
    };
    inline CRect operator+(const CVector2& v) const
    {
        CRect r(*this);
        r.m_min += v;
        r.m_max += v;
        return r;
    }
    inline CRect operator*(utifloat v) const
    {
        CRect r(*this);
        r.m_min *= v;
        r.m_max *= v;
        return r;
    }
    inline CRect operator/(utifloat v) const
    {
        CRect r(*this);
        r.m_min /= v;
        r.m_max /= v;
        return r;
    }
};

inline CRect CBox::xy()const {CRect r;r.m_min = m_min.xy(); r.m_max = m_max.xy();return r;};
inline CRect CBox::xz() const {CRect r;r.m_min = m_min.xz(); r.m_max = m_max.xz();return r;};
inline CRect CBox::yz()const  {CRect r;r.m_min = m_min.yz(); r.m_max = m_max.yz();return r;};


struct CRectf
{
     typedef CVector2f VectorClass;
    static CRectf INIT() {CRectf r; r.init(); return r;};
    static CRectf ADVANCEINIT() {CRectf r; r.m_min=CVector2f(1000000, 1000000); r.m_max=CVector2f(-1000000, -1000000);return r;};
    static CRectf MAXINIT() {CRectf r; r.m_max=CVector2f(1000000, 1000000); r.m_min=CVector2f(-1000000, -1000000);return r;};
    CVector2f m_min;
    CVector2f m_max;
    inline CRectf() {};
    inline CRectf(float x1, float y1, float x2, float y2) {m_min.x = x1; m_min.y = y1; m_max.x = x2; m_max.y = y2;};
    inline CRectf(const CVector2f& v1, const CVector2f&v2) {m_min = v1; m_max=v2;};
    inline explicit CRectf(const CIntRect& rect) {m_min.x = (float)rect.m_min.x; m_min.y =(float) rect.m_min.y; m_max.x =(float) rect.m_max.x; m_max.y = (float)rect.m_max.y;};

    inline CRectf(const CVector2f& pos, float size) {m_min.x = pos.x-size/2.0f;m_min.y = pos.y-size/2.0f;m_max.x = pos.x+size/2.0f;m_max.y = pos.y+size/2.0f;};
    inline void init() {m_min = CVector2f(0.0, 0.0); m_max = CVector2f(0.0,0.0);};
   // inline CBox getBounds(float zmin=0.0, float zmax=1.0) const {        CBox b; b.m_min = CVector3(m_min, zmin);b.m_max = CVector3(m_max, zmax);return b;};
    inline CVector2f getCoord(const CVector2f& pos) const {
      return CVector2f(m_min.x+(m_max.x-m_min.x)*pos.x,
              m_min.y+(m_max.y-m_min.y)*pos.y);
    };
    inline CVector2f getInvCoord(const CVector2f& pos) const {
      return CVector2f((pos.x-m_min.x)/(m_max.x-m_min.x),
              (pos.y-m_min.y)/(m_max.y-m_min.y));
    };
        inline CRectf getCoordRect(const CRectf& pos) const {
                return CRectf(getCoord(pos.m_min), getCoord(pos.m_max));
        }
    inline CVector2f center() const {return (m_max+m_min)/2.0f;};
    void getRectAsArray(Array<CVector2f> &marray);
    inline float getWidth() const {return m_max.x-m_min.x;};
    inline float getHeight()const  {return m_max.y-m_min.y;};
    inline void setWidth(float v) {m_max.x = m_min.x+v;};
    inline void setHeight(float v) {m_max.y = m_min.y+v;};
    inline CVector2f getSize() {return m_max-m_min;};
    inline float getLeft()const  {return m_min.x;};
    inline float getTop() const {return m_min.y;};
    inline float getRight()const  {return m_max.x;};
    inline float getBottom() const {return m_max.y;};
    inline void move(const CVector2f& v) { m_min+= v; m_max += v;};
    inline void setLeft(float v) {move(CVector2f(v - m_min.x, 0));};
    inline void setTop(float v) {move(CVector2f(0, v - m_min.y));};
    inline void correct() {if (m_min.x > m_max.x) {utiswap(m_min.x, m_max.x);}; if (m_min.y > m_max.y) {utiswap(m_min.y, m_max.y);};}
    inline bool contains(const CVector2f& position) const {
       return (m_min.x <= position.x ) && (m_max.x >= position.x) && (m_min.y <= position.y) && (m_max.y >= position.y);
    };
    inline void expandRect(float w)
    {
      m_min.x -= w;
      m_min.y -= w;
      m_max.x += w;
      m_max.y += w;
    }
    inline void roundToInt()
    {
      m_min.x = (float)((int)m_min.x);
      m_min.y = (float)((int)m_min.y);
      m_max.x = (float)((int)m_max.x);
      m_max.y = (float)((int)m_max.y);
    };
    inline void expandRect(const CVector2f& v)
    {
      m_min.x -= v.x;
      m_min.y -= v.y;
      m_max.x += v.x;
      m_max.y += v.y;
    }
    inline void expand(const CRectf& rect) {
      if (m_min.x > rect.m_min.x)m_min.x = rect.m_min.x;

      if (m_max.x < rect.m_max.x) m_max.x = rect.m_max.x;
      if (m_min.y > rect.m_min.y)m_min.y = rect.m_min.y;

      if (m_max.y < rect.m_max.y) m_max.y = rect.m_max.y;
    };
    inline void expand(const CVector2f& v)
    {
         if (m_min.x > v.x)m_min.x = v.x;
      if (m_max.x < v.x) m_max.x = v.x;
      if (m_min.y > v.y)m_min.y = v.y;
      if (m_max.y < v.y) m_max.y = v.y;

    }


    inline bool contains(const CRectf& v) const
    {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(v.valid());

      if (m_min.x > v.m_min.x) return false;
      if (m_min.y > v.m_min.y) return false;
      if (m_max.x < v.m_max.x) return false;
      if (m_max.y < v.m_max.y) return false;
      return true;
    };

    inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y <= m_max.y;
    }
   inline bool intersect(const CRectf& rect) const {
      //utiassert(m_min.x <= m_max.x);
      //utiassert(m_min.y <= m_max.y);
      if (m_min.x> rect.m_max.x) return false;
      if (m_min.y> rect.m_max.y) return false;
      if (m_max.x< rect.m_min.x) return false;
      if (m_max.y< rect.m_min.y) return false;
      return true;
    };
    inline bool operator==(const CRectf& other) const
    {
        return m_max == other.m_max && m_min == other.m_min;
    };
    inline bool operator!=(const CRectf& other) const {return !(operator==(other));};
    inline float getArea()
    {
      return absf((m_max.x-m_min.x)*(m_max.y-m_min.y));
    };
    inline CRectf operator+(const CVector2f& v) const
    {
        CRectf r(*this);
        r.m_min += v;
        r.m_max += v;
        return r;
    }
    inline CRectf operator*(float v) const
    {
        CRectf r(*this);
        r.m_min *= v;
        r.m_max *= v;
        return r;
    }
    inline CRectf operator/(float v) const
    {
        CRectf r(*this);
        r.m_min /= v;
        r.m_max /= v;
        return r;
    }
};

inline CRect intRectToFloat(const CIntRect& b)
{
    return CRect(intVecToFloat(b.m_min), intVecToFloat(b.m_max));
}


inline CIntRect floatRectToInt(const CRect& b)
{
    return CIntRect(floatVecToInt(b.m_min), floatVecToInt(b.m_max));
}


inline CRect toDouble(const CRectf& f1)
{
    return CRect(toDouble(f1.m_min), toDouble(f1.m_max));
}

inline CRectf toFloat(const CRect& f1)
{
    return CRectf(toFloat(f1.m_min), toFloat(f1.m_max));
}

inline CIntRect::CIntRect(const CRect& r)
   {
     m_min.x = (int) round(r.m_min.x);
     m_min.y = (int)  round(r.m_min.y);
     m_max.x = (int)  round(r.m_max.x);
     m_max.y = (int) round( r.m_max.y);
   };

inline CIntRect::CIntRect(const CRectf& r)
   {
     m_min.x = (int) round(r.m_min.x);
     m_min.y = (int)  round(r.m_min.y);
     m_max.x = (int)  round(r.m_max.x);
     m_max.y = (int) round( r.m_max.y);
   };


inline void AdvanceRect3D(CBox &Bounds, S3DPoint &p)
{
    if (Bounds.m_min.x > p.x) Bounds.m_min.x = p.x;
              if (Bounds.m_max.x < p.x) Bounds.m_max.x = p.x;

              if (Bounds.m_min.y > p.y) Bounds.m_min.y = p.y;
              if (Bounds.m_max.y < p.y) Bounds.m_max.y = p.y;

              if (Bounds.m_min.z > p.z) Bounds.m_min.z = p.z;
              if (Bounds.m_max.z < p.z) Bounds.m_max.z = p.z;
};
TUtiBoxVis utiVisBox(const CBox & b, SHSPlane* planes, int planecount, const CMatrix& invtransform);
CVector4 utiTransformPlane(CVector4&plane, const CMatrix& m);

CUtiString convertToString(const CIntBox& b);
CUtiString convertToString(const CIntRect& b);

CIntRect parseIntRect(const CUtiString& str);
CIntBox parseIntBox(const CUtiString& str);

void uti_readValue(CBox &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CBox& std=CBox(CVector3(0.0f, 0.0f, 0.0f), CVector3(0.0,0.0f, 0.0f)));
void uti_writeValue(CBox &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

CUtiString intBoxToString(const CIntBox& box);

class CUtiLine
{
protected:
    CVector3f m_origin;
    CVector3f m_dir;
    float m_length;
public:
    inline CUtiLine()
    {
        m_origin.setVector(0,0,0);
        m_dir.setVector(1,0,0);
        m_length = 0.0f;
    }

    inline CUtiLine(const CVector3f& start, const CVector3f& end)
    {
        m_origin = start;
        m_dir = end-start;
        m_length = m_dir.normalize();
    }

    inline CUtiLine(const CVector3f& origin, const CVector3f& dir, float l)
    {
        assert(l >= 0);
        m_origin = origin;
        m_dir = dir;
        m_length = l;
    }

    inline void set(const CVector3f& start, const CVector3f& end)
    {
        m_origin = start;
        m_dir = end-start;
        m_length = m_dir.normalize();
    }
    inline void set(const CVector3f& origin, const CVector3f& dir, float l)
    {
        assert(l >= 0);
        m_origin = origin;
        m_dir = dir;
        m_dir.normalize();
        m_length = l;
    }

    inline float getLength() const
    {
        return m_length;
    }
    inline CVector3f get(float pos) const
    {
        return m_origin + m_dir*pos;
    }
    inline float projectOnLine(const CVector3f& pos) const
    {
        return m_dir.dotProduct(pos-m_origin);

    }
    inline float distanceToLine(const CVector3f& pos) const
    {
        float s = projectOnLine(pos);
        return get(s).distanceTo(pos);

    }
    inline float distanceToLineSegment(const CVector3f& pos) const
    {
        float s = projectOnLine(pos);
        if (s <= 0.0f) {
            return m_origin.distanceTo(pos);
        } else if (s >= m_length) {
            return get(m_length).distanceTo(pos);
        } else {
            return get(s).distanceTo(pos);
        }

    }
};


void uti_readValue(CIntBox &obj, CXNode* node, CWriteReadInterface* wri, const CUtiString& name, const CIntBox& std);
void uti_writeValue(CIntBox &obj,  CXNode* node, CWriteReadInterface* wri, const CUtiString& name);

#endif // UTIMATHBOX_H
