/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef UTIMEMMANAGER_H
#define UTIMEMMANAGER_H

#include "Uti3DCore.h"
#include "utiarray.h"
#ifndef NDEBUG
//#define MEMORYMANAGER_DEBUG
#endif
struct SMemoryPool_BaseBehaviour
{
    inline void setData(ReferenceUtiOB* data)
    {

    }
};

struct SThreadLocalStatic_SMemElement
{
    void* m_pool;
    unsigned char* m_data;
    inline SThreadLocalStatic_SMemElement()
    {
        m_pool = NULL;
        m_data = NULL;
    }
    inline void setData(void* p, unsigned char* d)
    {
        m_pool = p;
        m_data = d;
    }
};

struct SThreadLocalStatic
{
    StaticArray<SThreadLocalStatic_SMemElement, 64> m_datastore;
    int m_datacount;
    bool m_inited;
    bool m_locked;
    SThreadLocalStatic* m_next;
    SThreadLocalStatic()
    {
       m_next = NULL;
       m_inited = false;
       m_locked = false;
       m_datacount = 0;
    }


};

template<class T, size_t TYPESIZE = 16, size_t blocksize = 4096, class Behaviour = SMemoryPool_BaseBehaviour> class CMemoryPool : public ReferenceUtiOB
{
protected:
    Behaviour m_behavior;
	//enum { TYPESIZE = datasize };
    enum {RESTDATASIZE= blocksize-sizeof(void*)};
    enum {MAXELEMENTCOUNT = RESTDATASIZE / TYPESIZE};
    enum {SHIFTONBLOCK = 0};
    void setData(ReferenceUtiOB* obj)
    {
        m_behavior.setData(obj);
    }

    struct SBlock
    {
        SBlock* m_next;
        StaticArray<unsigned char,RESTDATASIZE> m_data;
    };
    union SSlot {
        unsigned char m_element[TYPESIZE];
        SSlot* m_next;
    };
    SBlock* m_currentblock;
    SBlock* m_firstblock;
    SBlock* m_lastblock;
    int m_indexinblock;
    int m_blockcount;
    SSlot* m_freeslot;

    size_t m_infreelistcount;
    size_t m_infreeblocks;
    size_t m_reservedcount;
#ifdef MEMORYMANAGER_DEBUG
    std::set<unsigned char*> m_debug_freelist;
    std::set<unsigned char*> m_debug_reserved;
#endif

    SThreadLocalStatic* m_firstthreadblock;
    SThreadLocalStatic* m_lastthreadblock;
    CSpinLock m_accessdata;
public:
    UTI_SYNC_DATA;
    inline size_t getInFreeListCount() const
    {
        return m_infreelistcount;
    }
    inline size_t getInFreeBlock() const
    {
        return m_infreeblocks;
    }
    inline size_t getReservedCount() const
    {
        return m_reservedcount;
    }
    inline size_t getMemCount() const
    {
        return m_reservedcount*TYPESIZE;
    }
    CMemoryPool()
    {
        m_currentblock = NULL;
        m_firstblock = NULL;
        m_indexinblock = MAXELEMENTCOUNT;
        m_lastblock = NULL;
        m_freeslot = NULL;
        m_reservedcount = 0;
        m_infreelistcount = 0;
        m_infreeblocks = 0;m_blockcount = 0;
        m_firstblock = NULL;
        m_lastblock = NULL;
    }
    ~CMemoryPool()
    {
       freeAll();
    }
//-----------------------------------
//Memory LEAK!!!!!!!!!
//Thread local variable gets not freed from heap!!!!!
//-----------------------------------

    static THREADLOCAL  SThreadLocalStatic* m_threadlocal;
    void freeAll()
    {
        CCSControlSpin spin(m_accessdata);
        UTI_SYNC_CALL;

        SThreadLocalStatic* threadblock = m_firstthreadblock;
        while (threadblock != NULL) {
            assert(threadblock->m_locked == false);
            threadblock->m_locked = true;
            threadblock = threadblock->m_next;
        }
        threadblock = m_firstthreadblock;

        while (threadblock != NULL) {
            for (int i = 0; i < threadblock->m_datacount; i++) {
                deallocate_internal(threadblock->m_datastore[i].m_data);
                assert(threadblock->m_datastore[i].m_pool == this);
            }
            threadblock->m_datacount = 0;
            threadblock = threadblock->m_next;
        }

        threadblock = m_firstthreadblock;
        while (threadblock != NULL) {
            assert(threadblock->m_locked == true);
            threadblock->m_locked = false;
            threadblock = threadblock->m_next;
        }



      //  assert(m_reservedcount == 0);
        while (m_firstblock) {
            SBlock* next = m_firstblock->m_next;
            delete m_firstblock;
            m_firstblock = next;
        }
        m_freeslot = NULL;
        m_currentblock = 0;
        m_lastblock = NULL;
        m_indexinblock = MAXELEMENTCOUNT;
        m_firstblock = NULL;
        m_reservedcount = 0;
        m_infreelistcount = 0;
        m_infreeblocks = 0;m_blockcount = 0;
    }
    void allocateBlock()
    {UTI_SYNC_CALL;
        SBlock* newblock = new SBlock();
        if (m_lastblock)  {
            m_lastblock->m_next = newblock;

        }m_blockcount++;
        m_infreeblocks += MAXELEMENTCOUNT;
        m_lastblock = newblock;
        if (m_currentblock == NULL) {
            m_currentblock = m_lastblock;
            m_indexinblock = 0;
        }
    }

    inline unsigned char* allocate()
    {

        if (m_threadlocal == NULL) {
            CCSControlSpin spin(m_accessdata);
            m_threadlocal = new SThreadLocalStatic();
            if (m_firstthreadblock == NULL) {
                m_firstthreadblock = m_threadlocal;
                m_lastthreadblock = m_threadlocal;
            } else {
                m_lastthreadblock->m_next = m_threadlocal;
                m_lastthreadblock = m_threadlocal;
            }
        }
        assert(m_threadlocal->m_datacount >= 0);
        assert(m_threadlocal->m_locked == false);
        if (m_threadlocal->m_datacount == 0) {
           CCSControlSpin spin(m_accessdata);
           int hcount = m_threadlocal->m_datastore.size()/2;
           for (int i = 0; i < hcount; i++){
               m_threadlocal->m_datastore[m_threadlocal->m_datacount].setData(this, allocate_internal());
               m_threadlocal->m_datacount++;
           }
        }
        unsigned char* res=m_threadlocal->m_datastore[m_threadlocal->m_datacount-1].m_data;
        assert(m_threadlocal->m_datastore[m_threadlocal->m_datacount-1].m_pool == this);

        m_threadlocal->m_datacount--;
        assert(m_threadlocal->m_datacount >= 0);
        assert(m_threadlocal->m_locked == false);
        return res;

    }
    inline void deallocate(unsigned char* data)
    {

        if (m_threadlocal == NULL) {
            CCSControlSpin spin(m_accessdata);
            m_threadlocal = new SThreadLocalStatic();
            if (m_firstthreadblock == NULL) {
                m_firstthreadblock = m_threadlocal;
                m_lastthreadblock = m_threadlocal;
            } else {
                m_lastthreadblock->m_next = m_threadlocal;
                m_lastthreadblock = m_threadlocal;
            }
        }
       assert(m_threadlocal->m_datacount >= 0);
       assert(m_threadlocal->m_locked == false);
       if (m_threadlocal->m_datacount >= m_threadlocal->m_datastore.size()) {
          CCSControlSpin spin(m_accessdata);
          int hcount = m_threadlocal->m_datastore.size()/2;
          for (int i = 0; i < hcount; i++){
             deallocate_internal(m_threadlocal->m_datastore[m_threadlocal->m_datacount-1].m_data);
             assert(m_threadlocal->m_datastore[m_threadlocal->m_datacount-1].m_pool == this);
             m_threadlocal->m_datastore[m_threadlocal->m_datacount-1].setData(NULL, NULL);

             m_threadlocal->m_datacount--;
          }
       }
       m_threadlocal->m_datastore[m_threadlocal->m_datacount].setData(this, data);
       m_threadlocal->m_datacount++;
       assert(m_threadlocal->m_datacount <=  m_threadlocal->m_datastore.size());
       assert(m_threadlocal->m_locked == false);
    }

    inline unsigned char* allocate_internal()
    {
       // CCSControlSpin spin(m_accessdata);
        UTI_SYNC_CALL;
        if (m_freeslot != NULL) {
            unsigned char* res = m_freeslot->m_element;
            m_freeslot = m_freeslot->m_next;
            m_infreelistcount--;
            assert(m_infreelistcount >= 0);
            m_reservedcount++;
#ifdef MEMORYMANAGER_DEBUG
            assert(m_debug_reserved.find(res) == m_debug_reserved.end());
            m_debug_reserved.insert(res);
            assert(m_debug_freelist.find(res) != m_debug_freelist.end());
            m_debug_freelist.erase(res);
#endif
            return res;
        }else {
            if (m_indexinblock >= MAXELEMENTCOUNT) {
                if (m_currentblock == m_lastblock) {assert(m_infreeblocks == 0);
                    allocateBlock();
                    m_currentblock = m_lastblock;

                    m_indexinblock = 0;
                } else {
                    assert(m_currentblock->m_next != NULL);
                m_currentblock = m_currentblock->m_next;

                m_indexinblock = 0;
                }
                assert(m_currentblock!= NULL);
            }
            assert(m_indexinblock < MAXELEMENTCOUNT);
            unsigned char* res = m_currentblock->m_data.getData()+TYPESIZE*m_indexinblock;
            m_indexinblock++;
            m_reservedcount++;
            m_infreeblocks--;
            assert(m_infreeblocks >= 0);
#ifdef MEMORYMANAGER_DEBUG
            assert(m_debug_reserved.find(res) == m_debug_reserved.end());
            m_debug_reserved.insert(res);
            assert(m_debug_reserved.size() == m_reservedcount);
#endif
            return res;
        }
    }
    inline void deallocate_internal(unsigned char* data)
    {
       // CCSControlSpin spin(m_accessdata);
        UTI_SYNC_CALL;
        SSlot* s = reinterpret_cast<SSlot*>(data);
        s->m_next = m_freeslot;
        m_freeslot = s;
        m_infreelistcount++;
        m_reservedcount--;
#ifdef MEMORYMANAGER_DEBUG
            assert(m_debug_reserved.find(s->m_element) != m_debug_reserved.end());
            m_debug_reserved.erase(s->m_element);
            assert(m_debug_freelist.find(s->m_element) == m_debug_freelist.end());
            m_debug_freelist.insert(s->m_element);
            assert(m_debug_reserved.size() == m_reservedcount);
#endif
    }
};

template<class T, size_t TYPESIZE, size_t blocksize, class Behaviour> THREADLOCAL SThreadLocalStatic* CMemoryPool<T, TYPESIZE, blocksize, Behaviour>::m_threadlocal;

#define UTI_MEMORY_POOL_DECL(ClassName, BlockSize, Behaviour)\
    static inline void* operator new(std::size_t sz);\
    static inline void operator delete(void*ptr, std::size_t sz);

#define UTI_MEMORY_POOL_DECL_EXTERN(ClassName, BlockSize, Behaviour)\
extern CMemoryPool<ClassName, sizeof(ClassName), BlockSize, Behaviour> ClassName ## _m_uti_mem_pool; \
    inline void* ClassName::operator new(std::size_t sz) {\
    assert(sz == sizeof(ClassName));\
         return  ClassName ## _m_uti_mem_pool.allocate();\
}\
    inline void ClassName::operator delete(void*ptr, std::size_t sz) {\
    assert(sz == sizeof(ClassName));\
         ClassName ## _m_uti_mem_pool.deallocate((unsigned char*)ptr);\
}

#define UTI_MEMORY_POOL_IMPL(ClassName, BlockSize, Behaviour)\
    CMemoryPool<ClassName, sizeof(ClassName), BlockSize, Behaviour> ClassName ## _m_uti_mem_pool;

#endif // UTIMEMMANAGER_H
