/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "utistream.h"

#include <cmath>

#include <cstdio>

#include <fstream>
#include <sstream>
#include <set>
#include <typeinfo>
#include <iostream>
#include <cstdlib>
#include <vector>
#include <deque>
#include <locale>

#include <algorithm>
#include "utisystem.h"
#include "UtiArchiv.h"
#include "utiarray.h"
#include "easyprofiler.h"

using namespace std;

using namespace UtiFramework;



void CFileStream::ReadFromStreamSize(CStream* stream)
{
  //  Position = 0;
   // m_allowwrite=false;

    assert(iswrite == true);
    int datasize = stream->getUInt32();
    utiassert(datasize < 2000000000);
//    Memory.setSize(datasize);
    StaticArray<char, 65536> data;

    while(datasize > 0) {
        int c = umin(datasize, data.size());


        stream->ReadData(data.getData(), c);
        WriteData(data.getData(), c);
        datasize -= c;
        assert(datasize >= 0);

    }
}

void CFileStream::WriteToStreamSize(CStream* stream)
{
  setPosition(0, END);
  assert(iswrite == false);
  int datasize = getReadPosition();


  StaticArray<char, 65536> data;
  stream->writeUInt32(datasize);
  int writtenbytes=0;
  setPosition(0, START);
  while (true) {
      int r = ReadData(data.getData(), data.size());
      stream->WriteData(data.getData(), r);
      writtenbytes += r;
      if (r < data.size()) {
          break;
      }
  }
  assert(writtenbytes == datasize);
}

size_t CFileStream::WriteData(const void *data, size_t size)
{
assert(iswrite);
    fstream* str = (fstream*) strm;
    if (docount)
    {
        count = count +size;
    }
    else
    {
        int p = getWritePosition();
        str->write((U_SCHAR* )data, size);
        return getWritePosition()-p;
    };
    return UTI_OK;
}

size_t CFileStream::ReadData(void *data, size_t size)
{

    utiassert(!iswrite);
    std::fstream* str = (std::fstream*) strm;
    int p = getReadPosition();
str->read((U_SCHAR* )data, size);
bool setback=false;
ios_base::iostate oldstate;
if (str->eof()) {
    if (str->fail()) {
      setback = true;
      oldstate = str->rdstate();
        str->clear(/*str->rdstate() & (~ios_base::failbit)*/);
    };

};
int pres = getReadPosition();
/*if (pres < 0) {
  //if (str->fail()) {
        str->clear();
//	};
   pres = getReadPosition();
};*/
if (setback) {
str->clear(oldstate & ios_base::eofbit);
}
assert(pres-p >= 0);
    return pres-p;
}

CFileStream::CFileStream(const CUtiString& name, int i, bool trunc) : CStream()
{
    filename = new CUtiString();
   loadFile(name.c_str(), i, trunc);
}

CUtiString& CFileStream::getFileName()
{
    return *filename;
}

bool CFileStream::loadFile(const TCHAR *name, int i, bool trunc)
{
    *filename = name;
   fstream* str = new fstream();
    strm = str;
    if (i == S_READ)
        CheckFileName(name);

    docount = false;
    count = 0;
    iswrite = false;
    CUtiString namestr(name);
    char* ansiname = namestr.GetAnsi();
    if (i == S_READ) {
        str->open(ansiname, ios::binary | ios::in);
        if (str->good())return false;
      //  utiassert(str->good());
    }
    else
    {
        iswrite = true;
        /*if (!trunc)
        r |= ios::trunc;
        else
        r|= ios::ate;*/
        if (trunc)
        {
            str->open(ansiname, ios::binary | ios::out | ios::trunc);
        }
        else
        {
            str->open(ansiname, ios::binary | ios::out | ios::ate);
        }
    };
    delete [] ansiname;
    return true;
}

CFileStream::CFileStream(const U_CHAR *name, int i, bool trunc) : CStream()
{
    filename = new CUtiString();
   loadFile(name, i, trunc);


}


CFileStream::~CFileStream()
{

    delete filename;
    std::fstream* str = (std::fstream*) strm;
    str->clear();// - K�nte ntzlich sein
    str->flush();
    str->close();
    delete str;

}

bool CFileStream::fail()
{
    fstream* str = (fstream*) strm;
    return(str->fail()!=0);
};
bool CFileStream::eof()
{
    std::fstream* str = (std::fstream*) strm;
    return(str->eof()!=0);
};

void CFileStream::setPosition(off_t i, TPositionType postype)
{
    fstream* str = (fstream*) strm;
    ios_base::seekdir d;
    switch (postype) {
       case START: {d = ios_base::beg;}; break;
       case END: {d = ios_base::end;}; break;
       case CURRENT: {d=ios_base::cur;}; break;
       default: utiassert(false);break;
    };

    if (iswrite)
    {
        str->seekp(i, d);
        //strm.seekg(i);
    }
    else
    {
        str->clear();
        str->seekg(i, d);
    }

    /* CUtiString k;
    k = (int)i;
    MessageBox(0, k.getString(), UTEXT("klk"), 0);*/
}

size_t CFileStream::getReadPosition()
{
    fstream* str = (fstream*) strm;
    return(str->tellg());
}

size_t CFileStream::getWritePosition()
{
    fstream* str = (fstream*) strm;
    return(str->tellp());
}

CMemoryStream::~CMemoryStream()
{
 //   if (m_deletemem && (Memory != NULL))
  //      delete[] (U_SCHAR*)Memory;
}


CMemoryStream::CMemoryStream() : CStream()
{
    //Memory=NULL;
    Position = 0;
    failed = false;
   // m_deletemem = true;
    m_allowwrite = false;
    datasize = 0;
    docount = false;
    //  FileName[0] = 0;
}

CMemoryStream::CMemoryStream(bool allowwrite, size_t size) : Memory(size)
{
 // Memory = mem;
  Position = 0;
//  Size = size;
  datasize = 0;
  failed = false;
 // m_deletemem = false;
  m_allowwrite = allowwrite;
   docount = false;
}

void CMemoryStream::setAllowWrite(bool b)
{
   m_allowwrite = b;
}

CMemoryStream::CMemoryStream(LPVOID mem, size_t size, bool allowwrite) : Memory((unsigned char*)mem,size)
{
 // Memory = mem;
  Position = 0;
//  Size = size;
  datasize = size;
  failed = false;
 // m_deletemem = false;
  m_allowwrite = allowwrite;
   docount = false;
}


CMemoryStream::CMemoryStream(Array<unsigned char>& data, bool allowwrite) : Memory(data)
{
    Position = 0;
  //  Size = size;
    datasize = data.size();
    failed = false;
   // m_deletemem = false;
    m_allowwrite = allowwrite;
     docount = false;
}

unsigned char* CMemoryStream::getMemory()
{
   return Memory.array;
}
size_t CMemoryStream::getMemorySize()
{
  return datasize;
}

void CMemoryStream::setExactMemorySize()
{
  Memory.setSizeCopy(datasize);
};

int CMemoryStream::LoadFromFile(const CUtiString& c)
{

    CheckFileName(c.c_str());
    UTI_INT64 Size = UtiFileSize(c);
    CStream* f = UtiOpenFile(c, otRead);
    if (f == NULL) return UTI_READERROR;
   // m_deletemem = true;
    m_allowwrite = false;
    //Memory = new char[Size];
    Memory.setSize(Size);
    datasize = Size;
    f->ReadData(Memory.array, Size);
    Position = 0;
    delete f;

    /* #ifdef WIN32
    HANDLE res;
    res = CreateFile(c, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
    if (res == INVALID_HANDLE_VALUE) return;

    if (Memory) delete Memory;
    Size = GetFileSize(res, NULL);
    Memory = new char[Size];
    unsigned long Read=0;
    ReadFile(res, Memory, Size, &Read, NULL);

    Position = 0;
    CloseHandle(res);
    #endif*/
    //	FileName = c;
    /* CFileStream* f = new CFileStream(c, S_READ);
     f->strm.


     delete f;*/
     return UTI_OK;
};

int CMemoryStream::LoadFromFileNormal(const CUtiString& filename)
{
 // CheckFileName(c);

    EASY_FUNCTION(profiler::colors::Blue);
 bool b = FileExists(filename.getString());

   if (!b)
   {
     return UTI_READERROR;

   }

    UTI_INT64 Size = FileSize(filename.c_str());
    CStream* f = new CFileStream(filename, S_READ);
    if (f == NULL) return UTI_READERROR;
   // m_deletemem = true;
    m_allowwrite = false;
    //Memory = new char[Size];
    Memory.setSize(Size);
    datasize = Size;
    f->ReadData(Memory.array, Size);
    Position = 0;
    delete f;

    /* #ifdef WIN32
    HANDLE res;
    res = CreateFile(c, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
    if (res == INVALID_HANDLE_VALUE) return;

    if (Memory) delete Memory;
    Size = GetFileSize(res, NULL);
    Memory = new char[Size];
    unsigned long Read=0;
    ReadFile(res, Memory, Size, &Read, NULL);

    Position = 0;
    CloseHandle(res);
    #endif*/
    //	FileName = c;
    /* CFileStream* f = new CFileStream(c, S_READ);
     f->strm.


     delete f;*/
     return UTI_OK;
}


int CMemoryStream::SaveToFileSecure(const CUtiString& filename)
{

    EASY_FUNCTION(profiler::colors::Blue);
    if (FileExists(filename)) {
       CUtiString newfilename = filename+UTISTR("new");
      CUtiString old1filename = filename+UTISTR("old1");
      CUtiString old2filename = filename+UTISTR("old2");
       SaveToFile(newfilename);

       if (FileExists(old2filename)) {
           UtiFramework::deleteFile(old2filename);
       }
       if (FileExists(old1filename)) {
           UtiFramework::renameFile(old1filename, old2filename);

       }
       UtiFramework::renameFile(filename, old1filename);
       //UtiFramework::deleteFile(filename);
       UtiFramework::renameFile(newfilename, filename);
    } else {
        SaveToFile(filename);
    }
      return UTI_OK;
}

int CMemoryStream::SaveToFile(const CUtiString& filename)
{
  CFileStream* str = new CFileStream(filename, S_WRITE);
    if (str == NULL || str->fail())
    {
        if (str) delete str;
        return UTI_WRITEERROR;
    }
  str->WriteData(Memory.array, datasize);
    delete str;
    return UTI_OK;

}

void CMemoryStream::ReadFromStreamSize(CStream* stream)
{
    Position = 0;
    m_allowwrite=false;
    datasize = stream->getUInt32();
    utiassert(datasize < 2000000000);
    Memory.setSize(datasize);
    stream->ReadData(Memory.array, datasize);
}

void CMemoryStream::WriteToStreamSize(CStream* stream)
{
  utiassert(datasize < 2000000000);
  stream->writeUInt32(datasize);
  stream->WriteData(Memory.array, datasize);
}

size_t CMemoryStream::WriteData(const void *data, size_t size)
{

    if (docount)
    {
        count = count +size;
        return size;
    }
    else
    {
        // strm.write((U_SCHAR* )data, size);
       utiassert(m_allowwrite == true);
       utiassert(Position+size <= 2000000000);
       if (Position + size > datasize)
       {
       if (Memory.isDynamic()) {
         int newsize = Position+size;
         if (Memory.size() < 32) {
           Memory.setSizeCopy(32);
         };
         while (Memory.size() < newsize) {
           Memory.setSizeCopyHighSpeed(Memory.size()*2);
         };
         datasize = newsize;
       } else {
           failed = true;
       utiassert(false);
           return 0;
       };
       };
       memcpy((LPVOID) (Memory.array+Position), data, size);
        Position += size;
    return size;
    };
    return 0;
}

size_t CMemoryStream::ReadData(void *data, size_t size)
{
  utiassert(m_allowwrite == false);
    if (Position + size > datasize)
    {

         if (Position >= datasize) {
        failed = true;


        return 0;
     } else {
        size = datasize-Position;
//		utiassert(false);
     };
    };
    memcpy(data,(LPVOID) ((unsigned char*)Memory.array+Position), size);
    Position += size;assert(Position <= datasize);
    return size;
}


void CMemoryStream::setPosition(off_t i, TPositionType postype)
{
    switch (postype) {
        case START: {Position = i;}; break;
        case CURRENT: {Position += i;}; break;
        case END: {Position = datasize-i;}; break;
    };
    if (Position <= 0) Position = 0;
    if (Position >= datasize) Position = datasize;
}

size_t CMemoryStream::getReadPosition()
{
    return(Position);
}

size_t CMemoryStream::getWritePosition()
{
    return(Position);
}

//CCompareStream
CCompareStream::~CCompareStream()
{
 //   if (m_deletemem && (Memory != NULL))
  //      delete[] (U_SCHAR*)Memory;
}


CCompareStream::CCompareStream(CStream *other) : CStream()
{
    //Memory=NULL;
    m_otherstream = other;
    Position = 0;
    failed = false;
   // m_deletemem = true;
    //m_allowwrite = false;
   // datasize = 0;
    docount = false;
    //  FileName[0] = 0;
}

size_t CCompareStream::WriteData(const void *data, size_t size)
{
    if (docount)
    {
        count = count +size;
        return size;
    }
    else
    {
     Array<unsigned char> datacom;
     datacom.setSize(size);
     m_otherstream->setPosition(Position, START);
     m_otherstream->ReadData(datacom.array, size);
     unsigned char* ch = (unsigned char*)data;

     for (int i = 0; i < size; i++) {
         unsigned char newdata = ch[i];
         unsigned char olddata = datacom[i];
         if (newdata != olddata) {
         if (abs((int)newdata - (int)olddata) <2) {
             std::wcout << "Small Missmatch " << Position+i << " '" << newdata << "' old: '" << olddata << "'" << std::endl;
         }else
          {
             std::wcout << "Found Missmatch at Position " << Position+i << " '" << newdata << "' old: '" << olddata << "'" << std::endl;
         }
         }
     }


        Position += size;
    return size;
    }

}

size_t CCompareStream::ReadData(void *data, size_t size)
{
 assert(false);
    return 0;
}


void CCompareStream::setPosition(off_t i, TPositionType postype)
{
    switch (postype) {
        case START: {Position = i;}; break;
        case CURRENT: {Position += i;}; break;
        case END: {assert(false);}; break;
    };
    if (Position <= 0) Position = 0;
    //if (Position >= datasize) Position = datasize;
}

size_t CCompareStream::getReadPosition()
{
    return(Position);
}

size_t CCompareStream::getWritePosition()
{
    return(Position);
}

CBufferedReadStream::CBufferedReadStream(CStream* stream, int buffersize, size_t totaldatasize)
{
//#define DEBUGMYSTREAM
   m_otherstream = stream;
   m_totaldatasize = totaldatasize;
   m_hasbuffer =false;
   m_bufferstart=0;
   m_bufferend=0;
   m_currentposition=stream->getReadPosition();
   m_oldstreampos = m_currentposition;

   m_databuffer.setSize(buffersize);
}

CBufferedReadStream::~CBufferedReadStream()
{

}

bool CBufferedReadStream::fail()
{
   return eof();
}

bool CBufferedReadStream::eof()
{
  /*  if (!m_hasbuffer) {
        loadBuffer();
    }
    utiassert(m_currentposition <= m_bufferend);
    if (m_currentposition < m_bufferend || m_bufferstart == m_bufferend) {
        return false;
    } else {
        return m_otherstream->eof();
    }*/
    return m_currentposition >= m_totaldatasize;
}

bool CBufferedReadStream::loadBuffer()
{
#ifdef DEBUGMYSTREAM
m_otherstream->setPosition(m_oldstreampos, CStream::START);
#endif

 //  utiassert(m_oldstreampos == m_otherstream->getReadPosition());
    int readdata = umin(m_databuffer.size(), m_totaldatasize-m_oldstreampos);
    size_t count = m_otherstream->ReadData(m_databuffer.array, readdata);
    utiassert(count <= readdata);
    m_bufferstart = m_oldstreampos;
    m_oldstreampos+= count;
    m_bufferend = m_oldstreampos;
    m_hasbuffer = count != 0;

    utiassert(m_bufferend-m_bufferstart <= m_databuffer.size());
   // m_foundeof = count != m_databuffer.size();
    return count != 0;
}

size_t CBufferedReadStream::ReadData(void* data, size_t size)
{

    #ifdef DEBUGMYSTREAM
   size_t oldcurrentpos = m_currentposition;
#define STREAMCHECKDATA(datasize) \
    {\
        Array<unsigned char> tempdata(size);\
        m_otherstream->setPosition(oldcurrentpos, CStream::START);\
        size_t c = m_otherstream->ReadData(tempdata.array, size);\
        assert(c == datasize);\
        for (int i = 0; i < c; i++) {\
            assert(targetbuffer[i] == tempdata[i]);\
        }\
\
    }

#else
#define STREAMCHECKDATA(datasize)
#endif

    unsigned char* targetbuffer = (unsigned char*)data;

    if (m_hasbuffer == false) {
        if (!loadBuffer()) {
            return 0;
        }
    };
    utiassert(m_bufferend-m_bufferstart <= m_databuffer.size());
    utiassert(m_currentposition <= m_bufferend  && m_bufferstart<= m_currentposition);
    if (m_currentposition+size <= m_bufferend) {

        for (int i = 0; i < size;i++) {
            targetbuffer[i] = m_databuffer[m_currentposition-m_bufferstart+i];
        }
        m_currentposition += size;
        STREAMCHECKDATA(size);
        return size;
    }

   utiassert(m_currentposition <= m_bufferend  && m_bufferstart<= m_currentposition);
   size_t writepos = 0;
   while(writepos < size) {
       if (m_currentposition == m_bufferend) {
           if (!loadBuffer()) {
               STREAMCHECKDATA(writepos);
               return writepos;
           }
       }
       utiassert(m_currentposition < m_bufferend  && m_bufferstart<= m_currentposition);
       int readcount = std::min(m_bufferend-m_currentposition, size-writepos); assert(writepos+readcount <= size);
       for (int i = 0; i < readcount;i++) {
           targetbuffer[i+writepos] = m_databuffer[m_currentposition-m_bufferstart+i];
       }
       m_currentposition += readcount;
       writepos += readcount;
   }
   STREAMCHECKDATA(writepos);
   return writepos;
}

size_t CBufferedReadStream::WriteData(const void* data, size_t size)
{
   utiassert(false);
   return 0;
}

void CBufferedReadStream::setPosition(off_t i, TPositionType postype)
{
    m_hasbuffer = false;
	if (postype != CURRENT) {
		m_otherstream->setPosition(i, postype);
	}
	else {
		m_otherstream->setPosition(m_currentposition+i, START);
		assert(m_otherstream->getReadPosition() == m_currentposition + i);
	}
    m_currentposition = m_otherstream->getReadPosition();
    m_oldstreampos = m_currentposition;
}

size_t CBufferedReadStream::getReadPosition()
{
   return m_currentposition;
}

size_t CBufferedReadStream::getWritePosition()
{
 assert(false);
 return getReadPosition();
}
