/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef UTISTREAM_H
#define UTISTREAM_H

#include "Uti3DCore.h"



class CFileStream : public CStream
{
  protected:
      bool loadFile(const TCHAR *name, int i, bool trunc);
      CUtiString* filename;
public:
    ~CFileStream();
    explicit CFileStream(const U_CHAR *name, int i, bool trunc=true);
    explicit CFileStream(const CUtiString& name, int i, bool trunc=true);

    void* strm;

    bool iswrite;
    bool fail();
    bool eof();
    CUtiString& getFileName();
    size_t ReadData(void* data, size_t size);
    size_t WriteData(const void* data, size_t size);


    void ReadFromStreamSize(CStream* stream);
    void WriteToStreamSize(CStream* stream);
    void setPosition(off_t i, TPositionType postype);
    size_t getReadPosition();
    size_t getWritePosition();
    size_t getPosition(){return iswrite ? getWritePosition() : getReadPosition();};
};

class CMemoryStream : public CStream
{
protected:
    Array<unsigned char> Memory;
    size_t datasize;
    bool failed;
    off_t Position;
    //bool m_deletemem;
    bool m_allowwrite;
public:

    explicit CMemoryStream();
    CMemoryStream(bool allowwrite, size_t size=16);
    CMemoryStream(LPVOID mem, size_t size, bool allowwrite);
    CMemoryStream(Array<unsigned char>& data, bool allowwrite);
    ~CMemoryStream();
    void setAllowWrite(bool b);
    bool fail()
    {
        return(failed);
    };
    bool eof()
    {
        return(Position == (off_t)datasize);
    };
    unsigned char* getMemory();
    size_t getMemorySize();
    void setExactMemorySize();
    size_t ReadData(void* data, size_t size);
    size_t WriteData(const void* data, size_t size);
    void setPosition(off_t i, TPositionType postype);
    size_t getReadPosition();
    size_t getWritePosition();
    int LoadFromFile(const CUtiString& filename);
    int LoadFromFileNormal(const CUtiString& filename); //Thread Save
    int SaveToFile(const CUtiString& filename);//Thread Save
    int SaveToFileSecure(const CUtiString& filename);//Thread Save
    size_t getDataSize() {return datasize;};
    void ReadFromStreamSize(CStream* stream);
    void WriteToStreamSize(CStream* stream);

    inline void writeFloatFast(float i)
    {

        utiassert(docount == false);
        utiassert(datasize <= Memory.size());
        utiassert(m_allowwrite == true);
        utiassert(Position+4 <= Memory.size());
        utiassert(Position == datasize);
		*((float*)((unsigned char*)Memory.array + Position)) = i;
          Position += 4;
          datasize+= 4;
    }

    inline void writeUInt32Fast(U_UINT32 i)
    {

        utiassert(docount == false);
        utiassert(datasize <= Memory.size());
        utiassert(m_allowwrite == true);
        utiassert(Position+4 <= Memory.size());
        utiassert(Position == datasize);
          *((unsigned int*) ((unsigned char*)Memory.array+Position)) = i;
          Position += 4;
          datasize+= 4;
    }

    inline void writeInt32Fast(int i)
    {

        utiassert(docount == false);
        utiassert(datasize <= Memory.size());
        utiassert(m_allowwrite == true);
        utiassert(Position+4 <= Memory.size());
        utiassert(Position == datasize);
          *(( int*) ((unsigned char*)Memory.array+Position)) = i;
          Position += 4;
          datasize+= 4;
    }
    inline void writeUInt16Fast(unsigned short i)
    {

utiassert(docount == false);
        utiassert(datasize <= Memory.size());
        utiassert(m_allowwrite == true);
        utiassert(Position+2 <= Memory.size());
        utiassert(Position == datasize);
        assert(sizeof(unsigned short) == 2);
          *((unsigned short*) ((unsigned char*)Memory.array+Position)) = i;
          Position += 2;
          datasize+= 2;
    }

    inline void writeUInt8Fast(unsigned char i)
    {
utiassert(docount == false);
        utiassert(datasize <= Memory.size());
        utiassert(m_allowwrite == true);
        utiassert(Position+1 <= Memory.size());
        utiassert(Position == datasize);
        assert(sizeof(unsigned char) == 1);
          *((unsigned char*) ((unsigned char*)Memory.array+Position)) = i;
          Position += 1;
          datasize+= 1;

          utiassert(datasize <= Memory.size());
    }

    inline unsigned short getUInt16Fast()
    {

utiassert(docount == false);
        utiassert(datasize <= Memory.size());
          utiassert(m_allowwrite == false);
          utiassert(Position+2 <= Memory.size());
          utiassert(Position+2 <= datasize);
            unsigned short res = *((unsigned short*) ((unsigned char*)Memory.array+Position));
            Position += 2;

            utiassert(datasize <= Memory.size());
            return res;


    }
    inline unsigned char getUInt8Fast()
    {

utiassert(docount == false);
        utiassert(datasize <= Memory.size());
          utiassert(m_allowwrite == false);
          utiassert(Position+1 <= Memory.size());
          utiassert(Position+1 <= datasize);
            unsigned char res = *((unsigned char*) ((unsigned char*)Memory.array+Position));
            Position += 1;

            utiassert(datasize <= Memory.size());
            return res;

    }
    inline unsigned int getUInt32Fast()
    {

utiassert(docount == false);
        utiassert(datasize <= Memory.size());
          utiassert(m_allowwrite == false);
          utiassert(Position+4 <= Memory.size());
          utiassert(Position+4 <= datasize);
            unsigned int res = *((unsigned int*) ((unsigned char*)Memory.array+Position));
            Position += 4;

            utiassert(datasize <= Memory.size());
            return res;

    }
    inline int getInt32Fast()
    {

utiassert(docount == false);
        utiassert(datasize <= Memory.size());
          utiassert(m_allowwrite == false);
          utiassert(Position+4 <= Memory.size());
          utiassert(Position+4 <= datasize);
            int res = *((int*) ((unsigned char*)Memory.array+Position));
            Position += 4;
            return res;

    }
    inline float getFloatFast()
    {
utiassert(docount == false);
        utiassert(datasize <= Memory.size());
          utiassert(m_allowwrite == false);
          utiassert(Position+4 <= Memory.size());
          assert(sizeof(float) == 4);
          utiassert(Position+4 <= datasize);
            float res = *((float*) ((unsigned char*)Memory.array+Position));
            Position += 4;
            return res;

    }
};


class CCompareStream : public CStream
{
protected:
    bool failed;
    off_t Position;
    uti_refptr<CStream> m_otherstream;
public:

    explicit CCompareStream(CStream* other);
    ~CCompareStream();
    bool fail()
    {
        return(failed);
    };
    bool eof()
    {
        return(m_otherstream->eof());
    };
    size_t ReadData(void* data, size_t size);
    size_t WriteData(const void* data, size_t size);
    void setPosition(off_t i, TPositionType postype);
    size_t getReadPosition();
    size_t getWritePosition();
};

class CBufferedReadStream : public CStream
{
protected:
    bool m_hasbuffer;
    size_t m_bufferstart;
    size_t m_bufferend;
    size_t m_currentposition;
    size_t m_oldstreampos;
    size_t m_totaldatasize;
    Array<unsigned char> m_databuffer;
    uti_refptr<CStream> m_otherstream;
    bool loadBuffer();
public:
    explicit CBufferedReadStream(CStream* stream, int buffersize, size_t totaldatasize);
    ~CBufferedReadStream();
    bool fail();
    bool eof();
    size_t ReadData(void* data, size_t size);
    size_t WriteData(const void* data, size_t size);
    void setPosition(off_t i, TPositionType postype);
    size_t getReadPosition();
    size_t getWritePosition();
};

#endif // UTISTREAM_H
