/***************************************************************************
 *   Copyright (C) 2006 by Michael Staud   *
 *   michael.staud@uni-ulm.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _UTISYSTEM_
#define _UTISYSTEM_

// body file: utisystem.cpp

#include <string>
#ifndef UFRAMEWORK
#define UFRAMEWORK
#endif

#ifdef UFRAMEWORK
#include "UtiString.h"
#else
typedef std::wstring CUtiString;
#endif

#define UTI_API extern
#define UTI_APIIMPORT

//#define CUtiString wstring

#include "gdef.h"

namespace UtiFramework {


struct SystemInformation
{
    UTI_INT64 m_virtualmemsize;
    UTI_INT64 m_virtualmemused;
    UTI_INT64 m_physicalmemsize;
    UTI_INT64 m_physicalmemused;
    UTI_INT64 m_virtualapplicationusage;
    UTI_INT64 m_physicalapplicationusage;
    UTI_INT64 m_heapsize;
    UTI_INT64 m_allocatedmem;
    UTI_INT64 m_maxmemory;
};

void getSystemInformation(SystemInformation& system);


std::string intToString(int i);


enum TFileDateType {dtAccess, dtModification, dtStatusChange};

double getFileDate(CUtiString path, TFileDateType type);
void initTime();
double getCurrentTime();
const CUtiString getCurrentDir();
void ChangeDir(CUtiString str);
struct Dir_FileInfo {
   LPVOID Dir;
   LPVOID entity;
   CUtiString name;
   UTI_INT64 size;
   bool dir;
};

bool FindFirst(CUtiString path, Dir_FileInfo *fi);
bool FindNext(Dir_FileInfo*fi);
void FindClose(Dir_FileInfo*fi);
const CUtiString converttowide(const wchar_t*c);
const CUtiString converttowide(const char* c);
const CUtiString convertutf8towide(const char* c);
const std::string converttomulti(const wchar_t * c);
const std::string converttomulti(const CUtiString &c);
void utisleep(int ms);
LPVOID UtiLoadLibrary(wchar_t* name);
void UtiFreeLibrary(LPVOID module);
LPVOID UtiGetProcAddress(LPVOID module, char* name);
int getFileSize(const CUtiString path);
bool deleteFile(const CUtiString& path);
bool renameFile(const CUtiString& oldname, const CUtiString& newname);
void setDPIAware();
std::string exec(const char* cmd) ;

CUtiString getAppDir();
CUtiString getDataDir();
CUtiString getUserConfigDir();
CUtiString getDocumentsDir();
bool LaunchDefaultBrowser(const CUtiString& url);

};


#endif
