/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "utitimemeasure.h"
#include "utisystem.h"
#include <iostream>

#ifdef DOTIMEMEASURE
STimeMeasure::STimeMeasure(const CUtiString& str, int level)
{
  m_name = str; 
  m_starttime = UtiFramework::getCurrentTime();
  m_level = level;
}

STimeMeasure::~STimeMeasure()
{
    if (m_level >= 1) {
  TIME rest = UtiFramework::getCurrentTime() - m_starttime;
  std::wcout << m_name.c_str() << ": " << rest << std::endl;
    }
}
#endif

SReleaseTimeMeasure::SReleaseTimeMeasure(const CUtiString& str, int level)
{
  m_name = str;
  m_starttime = UtiFramework::getCurrentTime();
  m_level = level;
}

SReleaseTimeMeasure::~SReleaseTimeMeasure()
{
    if (m_level >= 1) {
  TIME rest = UtiFramework::getCurrentTime() - m_starttime;
  std::wcout << m_name.c_str() << ": " << rest << std::endl;
    }
}


void SStatisticTimeMeasure::writeValue(TIME t)
{
	m_timemeasures[m_writeposition] = t;
	m_writeposition = (m_writeposition + 1) % m_timemeasures.size();
	m_timemeasurecount = umax(m_timemeasurecount, m_writeposition+1);
}
bool SStatisticTimeMeasure::isFull() const
{
	return m_timemeasurecount == m_timemeasures.size();
}
SStatisticTimeMeasure::SStatisticTimeMeasure(const CUtiString& str, int timesteps)
{
	m_name = str;
	assert(timesteps > 0);
        m_timemeasures.setSize(timesteps);m_timemeasures.fill(0);
	m_starttime = 0.0;
	m_timemeasurecount = 0;
	m_writeposition = 0;
}
SStatisticTimeMeasure::~SStatisticTimeMeasure()
{

}
void SStatisticTimeMeasure::beginMeasure()
{
	m_starttime = UtiFramework::getCurrentTime();
}
void SStatisticTimeMeasure::endMeasure()
{

	TIME rest = UtiFramework::getCurrentTime() - m_starttime;
	writeValue(rest);
	if (isFull()) {
		TIME e = 0.0f;
		int n = m_timemeasures.size();
		for (int i = 0; i < n; i++) {
			e += m_timemeasures[i];
		}
		e /= TIME(m_timemeasures.size());

		TIME var = 0.0f;
		for (int i = 0; i < n; i++) {
			var += pow(m_timemeasures[i]-e, 2);
		}
		var = var / TIME(n);
		//float quantil = 1.644850; // 5%
		float quantil =  2.326350*2.5f; // 1%
		float minvalue = e - quantil*sqrt(var) / sqrt(n);
		float maxvalue = e + quantil*sqrt(var) / sqrt(n);
		if (rest < minvalue || rest > maxvalue) {
	//		std::wcout << m_name << " (StErr): " << rest << " e = " << e << " var = " << var << std::endl;
		}
	}
}

MAKE_SINGLETON_IMPL(CCentralTimeMeasurement);


CCentralTimeMeasurement::CCentralTimeMeasurement()
{
m_model = 0.0f;
m_texture = 0.0f;
m_sound = 0.0f;
m_domeasure = false;
}

CCentralTimeMeasurement::~CCentralTimeMeasurement()
{

}

void CCentralTimeMeasurement::printData()
{
   std::wcout << "Total Time: " << std::endl;
   std::wcout << "   Model Loading: " << m_model << std::endl;
   std::wcout << "   Texture Loading: "<< m_texture << std::endl;
   std::wcout << "   Sound Loading: "<< m_sound << std::endl;
}

SCentralTimeMeasure::SCentralTimeMeasure(int index)
{
    m_index = index;
    m_starttime = UtiFramework::getCurrentTime();
    assert(CCentralTimeMeasurement::getInstance()->m_domeasure == false);
    CCentralTimeMeasurement::getInstance()->m_domeasure = true;
}

SCentralTimeMeasure::~SCentralTimeMeasure()
{
    assert(CCentralTimeMeasurement::getInstance()->m_domeasure == true);
    CCentralTimeMeasurement::getInstance()->m_domeasure = false;
    TIME rest = UtiFramework::getCurrentTime() - m_starttime;
    switch (m_index) {
      case 0: {
        CCentralTimeMeasurement::getInstance()->m_texture += rest;

    }; break;
    case 1: {
      CCentralTimeMeasurement::getInstance()->m_model += rest;

  }; break;
    case 2: {
      CCentralTimeMeasurement::getInstance()->m_sound += rest;

  }; break;
    }
}
