/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UTITIMEMEASURE__
#define __UTITIMEMEASURE__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"

//#ifndef NDEBUG
//#define DOTIMEMEASURE
//#endif


#ifdef DOTIMEMEASURE

struct STimeMeasure
{
  TIME m_starttime;
  CUtiString m_name;
  int m_level;
  STimeMeasure(const CUtiString& str, int level=0);
  ~STimeMeasure();
};

#else
struct STimeMeasure
{
  inline STimeMeasure(const CUtiString& str, int level=0){};
  inline ~STimeMeasure(){};
};

#endif

struct SReleaseTimeMeasure
{
  TIME m_starttime;
  CUtiString m_name;
  int m_level;
  SReleaseTimeMeasure(const CUtiString& str, int level=0);
  ~SReleaseTimeMeasure();
};

struct SStatisticTimeMeasure
{
	TIME m_starttime;
	CUtiString m_name;
	Array<TIME> m_timemeasures;
	int m_timemeasurecount;
	int m_writeposition;
	void writeValue(TIME t);
	bool isFull() const;
	SStatisticTimeMeasure(const CUtiString& str, int timesteps);
	~SStatisticTimeMeasure();
	void beginMeasure();
	void endMeasure();
};

struct SStatisticTimeCall
{
	SStatisticTimeMeasure& m_time;
	SStatisticTimeCall(SStatisticTimeMeasure& time) : m_time(time)
	{
		m_time.beginMeasure();
	}
	~SStatisticTimeCall()
	{
		m_time.endMeasure();
	}
};

class CCentralTimeMeasurement : public CUtiOB
{
protected:
    MAKE_SINGLETON(CCentralTimeMeasurement);

public:
    TIME m_model;
    TIME m_texture;
    TIME m_sound;
    bool m_domeasure;
    CCentralTimeMeasurement();
    ~CCentralTimeMeasurement();
    void printData();
};

struct SCentralTimeMeasure
{
  TIME m_starttime;
  int m_index;
  SCentralTimeMeasure(int index);
  ~SCentralTimeMeasure();
};

#define UTI_MEASURE_TEXTURE_BEGIN {
//SCentralTimeMeasure no_var_talk(0);

#define UTI_MEASURE_TEXTURE_END }

#define UTI_MEASURE_MODEL_BEGIN {
//SCentralTimeMeasure no_var_talk(1);

#define UTI_MEASURE_MODEL_END }


#define UTI_MEASURE_SOUND_BEGIN {
//SCentralTimeMeasure no_var_talk(2);

#define UTI_MEASURE_SOUND_END }

#endif
