/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __UTIUTILITY__
#define __UTIUTILITY__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "utimathbox.h"
#include "utiintvector.h"
#include "utiarray.h"
#include <set>



struct CIVector2_Control
{
	typedef float DistanceType;
	enum TData { MAX_CHILD_COUNT = 8 };
	uti_refptr<ReferenceUtiOB> m_data;


	void setData(ReferenceUtiOB* obj)
	{
		m_data = obj;
	}
	inline void addToNode(Array<CIVector2>& ar, int & count, const CIVector2& pos) const
	{
		bool contains = CIntRect(0, 0, 100, 100).contains(pos) && !CIntRect(30, 30, 70, 70).contains(pos);
		if (contains) {
			ar[count] = pos;
			count++;
		}
	}

	inline void expandNode(const CIVector2& original, Array<CIVector2>& ar, int & count) const
	{
		count = 0;
		addToNode(ar, count, CIVector2(original.x - 1, original.y));
		addToNode(ar, count, CIVector2(original.x + 1, original.y));
		addToNode(ar, count, CIVector2(original.x, original.y - 1));
		addToNode(ar, count, CIVector2(original.x, original.y + 1));
		addToNode(ar, count, CIVector2(original.x - 1, original.y + 1));
		addToNode(ar, count, CIVector2(original.x + 1, original.y + 1));
		addToNode(ar, count, CIVector2(original.x - 1, original.y - 1));
		addToNode(ar, count, CIVector2(original.x + 1, original.y - 1));

	}
	/*inline float step(const CVector2& prev, const CIVector2& v1, const CIVector2& v2) const
	{
	CVector2 p1(v1.x,v1.y);
	CVector2 p2(v2.x, v2.y);

	float dis =  p1.distanceTo(p2);
	//dis = dis*dis;
	if (dis < 2.0f) {
	//  dis = dis* ( 1.0f+0.5f*umax((1.0f- p1.distanceTo(CVector2(20,20)) /20.0f),0.0f));
	} else {

	}

	return dis;// sqrt((v1.x-v2.x)*(v1.x-v2.x)+(v1.y-v2.y)*(v1.y-v2.y));
	}*/
	inline float distance(const CIVector2& v1, const CIVector2& v2) const
	{
		CVector2 p1(v1.x, v1.y);
		CVector2 p2(v2.x, v2.y);

		float dis = p1.distanceTo(p2);
		//dis = dis*dis;
		if (dis < 2.0f) {
			//  dis = dis* ( 1.0f+0.5f*umax((1.0f- p1.distanceTo(CVector2(20,20)) /20.0f),0.0f));
		}
		else {

		}

		return dis;// sqrt((v1.x-v2.x)*(v1.x-v2.x)+(v1.y-v2.y)*(v1.y-v2.y));
	}
	inline float distanceToLine(const CIVector2 &s, const CIVector2 &t, const CIVector2 &point) const
	{
		CVector2 origin(s.x, s.y);
		CVector2 dir(t.x - s.x, t.y - s.y);
		CVector2 p(point.x, point.y);
		dir.normalize();
		float dis = dir.dotProduct(p - origin);
		CVector2 ponline = origin + dir*dis;
		return ponline.distanceTo(p);
	}
};

template <class T, class TControl>
class CAStarAlgorithm : public ReferenceUtiOB
{
protected:

	T m_start;
	T m_end;
	struct Node
	{
		typename TControl::DistanceType m_distance;
		typename TControl::DistanceType m_approxdistance;
		T m_element;
		int m_prevelement;
        int m_prevelement_edgeindex;
		bool m_openlist, m_closedlist;
		inline bool operator==(const Node& n) const
		{
			return m_element = n.m_element;
		}
		inline bool operator <(const Node&n) const
		{
			return m_element < n.m_element;
		}
		inline void init()
		{
			m_distance = 0;
			m_approxdistance = 0;
			m_prevelement = -1;
            m_prevelement_edgeindex =-1;
			m_element = T();
			m_openlist = false;
			m_closedlist = false;
		}
	};
	int m_goalnode;
	Array<Node> m_nodes;
	UtiArrayAccessGrow<Node, true> m_nodesgrow;
	std::map<T, int> m_objecttonode;
	void clearNodes()
	{
		m_objecttonode.clear();
		m_nodesgrow.reset(100000);
	}
public:
	inline Node& n(int index)
	{
		assert(index >= 0 && index < m_nodesgrow.size());
		return m_nodes[index];
	}
protected:
	inline int getNode(const T& el)
	{
		typename std::map<T, int>::iterator it = m_objecttonode.find(el);
		if (it != m_objecttonode.end()) {
			assert(it->second < m_nodesgrow.size());
			return it->second;
		}
		else {
			Node n;
			n.init();
			n.m_element = el;
			int index = m_nodesgrow.fast_add(n);
			m_objecttonode.insert(std::make_pair(el, index));
			return index;
		}
	}


	struct OpenListElement
	{
		typename TControl::DistanceType m_approxdistance;
		int m_node;
		inline OpenListElement() {

		}
		inline OpenListElement(typename TControl::DistanceType distance, int node)
		{
			m_approxdistance = distance;
			m_node = node;
		}

		inline bool operator==(const OpenListElement& n) const
		{
			return m_approxdistance = n.m_approxdistance && m_node == n.m_node;
		}
		inline bool operator <(const OpenListElement&n) const
		{
			if (m_approxdistance == n.m_approxdistance)
			{
				return m_node < n.m_node;
			}
			return m_approxdistance < n.m_approxdistance;
		}
	};


	bool m_haspath;
	bool m_reachedgoal;
	Array<T> m_expandednodes;
public:
        TControl m_control;
protected:
        std::set<OpenListElement> m_openlist;
    struct SPathElement {
        T m_element;
        int m_edgeindex;
    };

    Array<SPathElement> m_pathelement;
	//std::set<int> m_closedlist;
public:
	CAStarAlgorithm() : m_nodes(16), m_nodesgrow(m_nodes)
	{
		m_haspath = false;
		m_reachedgoal = false;
		m_goalnode = 0;
		m_expandednodes.setSize(TControl::MAX_CHILD_COUNT);
	}

	~CAStarAlgorithm()
	{

		// assert(m_haspath == false);
	}
	inline void setData(ReferenceUtiOB* obj)
	{
		m_control.setData(obj);
	}

	inline void start(const T& start, const T& end)
	{
		assert(m_haspath == false);
                assert(m_openlist.size() == 0);
		m_haspath = true;
		m_start = start;
		m_end = end;
        m_goalnode = getNode(end);
		m_reachedgoal = false;
        int startnode = getNode(start);
        Node& en = n(startnode);
		en.m_distance = 0;
        en.m_approxdistance = (typename TControl::DistanceType)(0) + m_control.distance(start, end);
        addToOpenList(startnode);

	}
	inline bool hasReachedGoal() const
	{
		return m_reachedgoal;
	}
	inline void removeFromOpenList(int i)
	{
		Node& en = n(i);
		assert(en.m_openlist == true);
		typename std::set<OpenListElement>::iterator minit = m_openlist.find(OpenListElement(en.m_approxdistance, i));
		assert(minit != m_openlist.end());
		en.m_openlist = false;
		m_openlist.erase(minit);
	}

	inline void addToOpenList(int i)
	{
		Node& en = n(i);
		assert(en.m_openlist == false);
                en.m_openlist = true;
                m_openlist.insert(OpenListElement(en.m_approxdistance, i));
	}

	inline int extractMinElement()
        {
                typename std::set<OpenListElement>::iterator minit = m_openlist.begin();
                OpenListElement minel = *minit;
                m_openlist.erase(minit);
		Node& node = n(minel.m_node);
		assert(minel.m_approxdistance == node.m_approxdistance);
		assert(node.m_openlist == true);
		node.m_openlist = false;
		return minel.m_node;
	}

    bool steps(int count)
	{
                assert(m_openlist.empty() == false);
		assert(m_reachedgoal == false);
		assert(m_haspath == true);
		for (int currentstep = 0; currentstep < count; currentstep++)
		{
			int currentnode = extractMinElement();
			if (currentnode == m_goalnode) {
				m_reachedgoal = true;
                createPath();
				return false;
			}

			//m_closedlist.insert(minel.m_node);
			n(currentnode).m_closedlist = true;
			expandNode(currentnode);
			if (m_openlist.empty()) {
				return false;
			}
		}
		return true;
	}
	inline void expandNode(int node)
	{
		int count = 0;
		Node& nodeptrold = n(node);
		m_control.expandNode(nodeptrold.m_element, m_expandednodes, count);
		//Node successor;
		//successor.init();

		// successor.m_prevelement = node;
		for (int i = 0; i < count; i++)
		{
			int successor = getNode(m_expandednodes[i]);
			Node& nodeptr = n(node);
			Node& successorptr = n(successor);
			if (successorptr.m_closedlist) {
				continue;
			}
			float enddistance = m_control.distance(nodeptr.m_element, successorptr.m_element);
			float newdistance = nodeptr.m_distance + enddistance;

			if (successorptr.m_openlist && successorptr.m_distance < newdistance)
			{
				continue;
			}
			if (successorptr.m_openlist) {
				removeFromOpenList(successor);
			}
			assert(successorptr.m_openlist == false);
			successorptr.m_prevelement = node;
            successorptr.m_prevelement_edgeindex = i;
			successorptr.m_distance = newdistance;
			successorptr.m_approxdistance = successorptr.m_distance + m_control.distance(successorptr.m_element, m_start);
			addToOpenList(successor);
			//std::map<TControl::DistanceType, T>::iterator  it = m_openlist.find(successors);

		}
	}

	inline void endSearch()
	{
                assert(m_haspath == true);
                m_openlist.clear();
		m_haspath = false;
		m_reachedgoal = false;
		clearNodes();
	}

    void createPath()
    {
        assert(m_reachedgoal != false);
        int nodecount=0;
        int currentnode = m_goalnode;
        while (currentnode != -1) {
            nodecount++;
            currentnode = n(currentnode).m_prevelement;
        }
        m_pathelement.setSize(nodecount);
        int currentindex = nodecount-1;
        int prevchildvalue=-1;


        currentnode = m_goalnode;

        while (currentindex >= 0) {
            assert(currentnode != -1);
            SPathElement& p = m_pathelement[currentindex];
            Node& node = n(currentnode);
            p.m_element = node.m_element;
            p.m_edgeindex = prevchildvalue;
            currentnode = node.m_prevelement;
            prevchildvalue = node.m_prevelement_edgeindex;
            currentindex--;
        }



        //SPathElement
    }

	struct iterator
	{
		CAStarAlgorithm* m_staralgorithm;
        int m_nodeindex;
		inline iterator()
		{
			m_staralgorithm = NULL;
            m_nodeindex = -1;
		}

        inline iterator(CAStarAlgorithm* staralgorithm, int nodeindex)
		{
			m_staralgorithm = staralgorithm;
            m_nodeindex = nodeindex;

		}
		inline iterator(const iterator& iter)
		{
			m_staralgorithm = iter.m_staralgorithm;
            m_nodeindex = iter.m_nodeindex;
		}
		inline iterator& operator=(const iterator& iter)
		{
			m_staralgorithm = iter.m_staralgorithm;
            m_nodeindex = iter.m_nodeindex;
			return *this;
		}
		inline bool operator==(const iterator& iter) const
		{
            return m_staralgorithm == iter.m_staralgorithm && m_nodeindex == iter.m_nodeindex;
		}
		inline bool operator!=(const iterator& iter) const
		{
            return m_staralgorithm != iter.m_staralgorithm || m_nodeindex != iter.m_nodeindex;
		}
		inline const T& operator*() const
		{
            assert(m_nodeindex != -1 && m_nodeindex < m_staralgorithm->m_pathelement.size());
            return m_staralgorithm->m_pathelement[m_nodeindex].m_element;
		} 
        inline const T& getElement() const
        {
            assert(m_nodeindex != -1 && m_nodeindex < m_staralgorithm->m_pathelement.size());
            return m_staralgorithm->m_pathelement[m_nodeindex].m_element;
        }
        inline int getEdgeIndex() const
        {
            assert(m_nodeindex != -1 && m_nodeindex < m_staralgorithm->m_pathelement.size());
            return m_staralgorithm->m_pathelement[m_nodeindex].m_edgeindex;
        }
        inline iterator operator++(int)
                                {
                                    return operator++();
                                }
		inline iterator& operator++()
        {
            assert(m_nodeindex != -1 && m_nodeindex < m_staralgorithm->m_pathelement.size());
			/*if (m_node == m_staralgorithm->m_goalnode) {
			m_node = -1;
			} else {*/
            m_nodeindex++;// = m_staralgorithm->n(m_node).m_prevelement;
			//}
			return *this;
		}
	};
	inline iterator begin()
	{
        assert(m_reachedgoal);
        return iterator(this, 0);
	}
	inline iterator end()
	{
        assert(m_reachedgoal);
        return iterator(this, m_pathelement.size());
	}

};

template<class T, class TControl, class Source> class CCreatePolygonFromVertices : public ReferenceUtiOB
{
protected:
	float m_maxlength;
	float m_maxerror;
	int m_maxpoints;
	TControl m_control;
public:
	inline CCreatePolygonFromVertices()
	{
		m_maxlength = 1.0f;
		m_maxerror = 0.1f;
		m_maxpoints = 5;
	}
        inline void setData(ReferenceUtiOB* obj)
        {
                m_control.setData(obj);
        }
	inline void setMaxLength(float ml)
	{
		m_maxlength = ml;
	}
	inline void setMaxError(float me)
	{
		m_maxerror = me;
	}
	inline void setMaxPoints(int i)
	{
		m_maxpoints = i;
	}
	inline float getMaxLength() const
	{
		return m_maxlength;
	}
	inline float getMaxError() const
	{
		return m_maxerror;
	}
	inline int getMaxPoints() const
	{
		return m_maxpoints;
	}
	struct iterator {
		CCreatePolygonFromVertices* m_parent;
		Source* m_source;
		typename Source::iterator m_currentposition;
		typename Source::iterator m_end;
		inline iterator()
		{
			m_parent = NULL;
			m_source = NULL;
		}

		inline iterator(CCreatePolygonFromVertices* parent, Source* source)
		{
			m_parent = parent;
			m_source = source;
			if (source) {
				m_currentposition = m_source->begin();
				m_end = m_source->end();
			}
		}
		inline void setData(ReferenceUtiOB* obj)
		{
            m_parent->m_control.setData(obj);
		}


		inline iterator(const iterator& iter)
		{
			m_parent = iter.m_parent;
			m_source = iter.m_source;
			m_currentposition = iter.m_currentposition;
			m_end = iter.m_end;
		}
		inline iterator& operator=(const iterator& iter)
		{
			m_parent = iter.m_parent;
			m_source = iter.m_source;
			m_currentposition = iter.m_currentposition;
			m_end = iter.m_end;
			return *this;
		}
		inline bool operator==(const iterator& iter) const
		{
			if (iter.m_source == NULL && m_source == NULL)
			{
				return m_parent == iter.m_parent;
			}
			return m_parent == iter.m_parent && m_source == iter.m_source && m_currentposition == iter.m_currentposition && m_end == iter.m_end;
		}
		inline bool operator!=(const iterator& iter) const
		{
			return !operator ==(iter);
		}
		inline const T& operator*() const
		{
			assert(m_source != NULL);
			assert(m_currentposition != m_end);
			return *m_currentposition;
		}
                inline iterator operator++(int)
                {
                    return operator++();
                }

		inline iterator& operator++()
		{
			assert(m_source != NULL);
			typename Source::iterator next = m_currentposition;
                        next++;
			if (next == m_end) {
				m_currentposition = next;
				m_source = NULL;
				return *this;
			}
			T startpos = *m_currentposition;

                        m_currentposition++;
			typename Source::iterator starttocheck = m_currentposition;
			int pointcount = 1;
                        next++;
			while (true) {
				if (next == m_end) {
					break;
				}
                                if (pointcount >= m_parent->m_maxpoints) {
					break;
				}
				T endpos = *next;
				float l = m_parent->m_control.distance(startpos, endpos);
				if (l > m_parent->m_maxlength) {
					break;
				}
				float error = 0.0f;
                                for (typename Source::iterator it = starttocheck; it != next; it++)
				{
					float linedistance = m_parent->m_control.distanceToLine(startpos, endpos, *it);
					error = umax(error, linedistance);
				}
				if (error > m_parent->m_maxerror)
				{
					break;
				}

				m_currentposition = next;
                                next++;
			}

			return *this;
		}
	};
	inline iterator begin(Source& source)
	{
		return iterator(this, &source);
	}
	inline  iterator end()
	{
		return iterator(this, NULL);
	}

};

class BlockSaver : public ReferenceUtiOB
{
protected:
	unsigned short m_currentindex;
	CIntRect m_rect;
	Array2D<unsigned short> m_data;
public:
	inline BlockSaver()
	{
		m_data.setSize(128, 128);
		m_data.fill(0);
		m_currentindex = 1;
		m_rect = CIntRect(0, 0, 0, 0);
	}
	inline bool getBlock(const CIVector2& s) const
	{
		if (!m_rect.contains(s))
		{
			return false;
		}
		int index = m_data(s - m_rect.m_min);
		return index == m_currentindex;
	}
	void reset()
	{
		m_currentindex++;
		if (m_currentindex > 65000) {
			m_data.fill(0);
			m_currentindex = 1;
		}
	}
	CIntRect rect() const
	{
		return m_rect;
	}

        template<class iter> void parseResult(iter begin, iter end)
	{

		reset();
		CIntRect r = CIntRect::MAXINIT();
		for (iter it = begin; it != end; it++) {
			r.expand(*it);
		}
		if (!r.valid()) {
			m_rect = CIntRect(0, 0, 0, 0);
			return;
		}
		assert(begin != end);
		r.m_max += CIVector2(1, 1);
		int w = r.width();
		int h = r.height();
		int neww = m_data.width();
		int newh = m_data.height();
		bool newsize = false;
		while (neww < w) {

			newsize = true;
			neww *= 2;
		}
		while (newh < h) {

			newsize = true;
			newh *= 2;
		}
		if (newsize) {
			m_data.setSize(neww, newh);
			m_data.fill(0);
			m_currentindex = 1;
		}
		m_rect = r;
		CIVector2 last = *begin;
		for (iter it = begin; it != end; it++) {
			CIVector2 current = *it;
			CIntRect r = CIntRect::MAXINIT();
			r.expand(last);
			r.expand(current);
			for (int j = r.m_min.y; j <= r.m_max.y; j++) {
				for (int i = r.m_min.x; i <= r.m_max.x; i++) {
					assert(m_rect.contains(CIVector2(i, j)));
					m_data(CIVector2(i, j) - m_rect.m_min) = m_currentindex;
				}
			}
			last = current;

		}
	}
};

#endif 
