/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __UTIVECTOR__
#define __UTIVECTOR__
#include <math.h>

#define UTI_VECTOR_DECLARE_OPERATORS(Vector, Ftype) \
    inline Vector add(const Vector& v) const {return *this+v;}\
    inline Vector sub(const Vector& v) const {return *this-v;}\
    inline Vector mul(const Vector& v) const {return *this*v;}\
    inline Vector mul(Ftype v) const {return *this*v;}\
    inline Vector div(Ftype v) const {return *this/v;}\
    inline bool equal(const Vector& v) const {return *this==v;}\
   inline void addto(const Vector& v) {return *this+=v;}\
    inline void subto(const Vector& v){return *this-=v;}


inline float sqrf(float f) {return f*f;};
inline double sqr(double f) {return f*f;};
struct CVector2 {
    UTI_VECTOR_DECLARE_OPERATORS(CVector2, double);
double x;
double y;
inline CVector2(){/*x=0.0f;y=0.0f;*/};
inline CVector2(const double f) {
x = f;
y = f;
};
inline CVector2(const double arg0, const double arg1){
x = arg0;y = arg1;
}
inline double &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 2);
return (&x)[p];
}
inline CVector2 returnNormal() const
{
    CVector2 res =* this;
    res.normalize();
    return res;
}
inline const double &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 2);
    return (&x)[p];
}
inline CVector2 xx() const
{
return CVector2(x, x);
}
inline CVector2 yx() const
{
return CVector2(y, x);
}
inline CVector2 xy() const
{
return CVector2(x, y);
}
inline CVector2 yy() const
{
return CVector2(y, y);
}
inline void operator += (const CVector2 &v)
{
x+=v.x;
y+=v.y;
}
inline void operator -= (const CVector2 &v)
{
x-=v.x;
y-=v.y;
}
inline void operator *= (const double v)
{
x*=v;
y*=v;
}
inline void operator *= (const CVector2 &v)
{
x*=v.x;
y*=v.y;
}
inline void operator /= (const double v)
{
x/=v;
y/=v;
}
inline void operator /= (const CVector2 &v)
{
x/=v.x;
y/=v.y;
}
inline void fromTo(const CVector2 & start, const CVector2 & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
};
void setVector(double arg0, double arg1)
{
x = arg0;
y = arg1;
}
inline void addVector(double arg0, double arg1)
{
x += arg0;
y += arg1;
}
inline void subVector(double arg0, double arg1)
{
x -= arg0;
y -= arg1;
}
inline void neg()
{
x = -x;
y = -y;
}
inline double length() const
{
return sqrt(x*x+y*y);
}
inline double normalize()
{
double r = sqrt(x*x+y*y);
   if (r < 0.00001) return r;
x /= r;
y /= r;
  return r;
}
inline double dotProduct(const CVector2 &b) const
{
return (x* b.x+y* b.y);
}
inline int getMajorAxis() const
{
double current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
return index;
}
inline double distanceTo(const CVector2 &v) const;
inline CVector2 operator +(const CVector2 &v2)const
{
CVector2 r;
r.x = x + v2.x;
r.y = y + v2.y;
return r;
}
inline CVector2 operator +(const double v2)const
{
CVector2 r;
r.x = x + v2;
r.y = y + v2;
return r;
}
inline CVector2 operator -(const CVector2 &v2)const
{
CVector2 r;
r.x = x - v2.x;
r.y = y - v2.y;
return r;
}
inline CVector2 operator -(const double v2)const
{
CVector2 r;
r.x = x - v2;
r.y = y - v2;
return r;
}
inline CVector2 operator -()const
{
CVector2 r;
r.x = -x;
r.y = -y;
return r;
}
inline CVector2 operator *(const CVector2 &v2)const
{
CVector2 r;
r.x = x * v2.x;
r.y = y * v2.y;
return r;
}
inline CVector2 operator *(const double v2)const
{
CVector2 r;
r.x = x * v2;
r.y = y * v2;
return r;
}

inline CVector2 operator /(const CVector2 &v2)const
{
CVector2 r;
r.x = x / v2.x;
r.y = y / v2.y;
return r;
}
inline CVector2 operator /(const double v2)const
{
CVector2 r;
r.x = x / v2;
r.y = y / v2;
return r;
}
inline bool operator ==(const CVector2 &v2)const
{
bool r;
r =  x == v2.x
 &&
 y == v2.y
;
return r;
}
inline bool operator !=(const CVector2 &v2)const
{
bool r;
r =  x != v2.x
 ||
 y != v2.y
;
return r;
}
};

inline CVector2 operator *(const double v1, const CVector2 &v2)
{
CVector2 r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
return r;
}

inline CVector2 operator /(const double v1, const CVector2 &v2)
{
CVector2 r;
r.x = v1 / v2.x;
r.y = v1 / v2.y;
return r;
}

inline double CVector2::distanceTo(const CVector2 &v) const
{
return (v-*this).length();
}
inline double dot(const CVector2 &a, const CVector2 &b)
{
return (a.x* b.x+a.y* b.y);
}
inline CVector2 umin(const CVector2 &a, const CVector2 &b)
{
CVector2 r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
return r;
}
inline CVector2 umax(const CVector2 &a, const CVector2 &b)
{
CVector2 r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
return r;
}
inline CVector2 normalize(const CVector2 &a)
{
CVector2 rvec;
double r = sqrt(a.x* a.x+a.y* a.y);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
  return rvec;
}
inline double length(const CVector2 &a)
{
CVector2 rvec;
double r = sqrt(a.x*a.x+a.y*a.y);
   return r;
}
inline CVector2 sign(const CVector2 &a)
{
CVector2 rvec;
rvec.x = a.x > 0.0f ? 1.0f : -1.0f;
rvec.y = a.y > 0.0f ? 1.0f : -1.0f;
  return rvec;
}
inline double distance(const CVector2 &a, const CVector2 &b)
{
return sqrtf(sqr(a.x-b.x)+sqr(a.y-b.y));
}
inline CVector2 linear(const CVector2 &a, const CVector2 &b, const double v)
{
CVector2 rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
return rvec;
}
struct CVector3 {
double x;
double y;
double z;
UTI_VECTOR_DECLARE_OPERATORS(CVector3, double);
inline CVector3(){/*x=0.0f;y=0.0f;z=0.0f;*/};
explicit inline CVector3(const double f) {
x = f;
y = f;
z = f;
};
inline CVector3(const double arg0, const double arg1, const double arg2){
x = arg0;y = arg1;z = arg2;
}
inline CVector3(const double arg0, const CVector2 & arg1){
x = arg0;y = arg1.x;z = arg1.y;
}
inline CVector3(const CVector2 & arg0, const double arg1){
x = arg0.x;y = arg0.y;z = arg1;
}
inline double &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 3);
return (&x)[p];
}
inline const double &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 3);
    return (&x)[p];
}
inline CVector3 returnNormal() const
{
    CVector3 res = *this;
    res.normalize();
    return res;
}
inline CVector2 xx() const
{
return CVector2(x, x);
}
inline CVector2 yx() const
{
return CVector2(y, x);
}
inline CVector2 zx() const
{
return CVector2(z, x);
}
inline CVector2 xy() const
{
return CVector2(x, y);
}
inline CVector2 yy() const
{
return CVector2(y, y);
}
inline CVector2 zy() const
{
return CVector2(z, y);
}
inline CVector2 xz() const
{
return CVector2(x, z);
}
inline CVector2 yz() const
{
return CVector2(y, z);
}
inline CVector2 zz() const
{
return CVector2(z, z);
}
inline CVector3 xxx() const
{
return CVector3(x, x, x);
}
inline CVector3 yxx() const
{
return CVector3(y, x, x);
}
inline CVector3 zxx() const
{
return CVector3(z, x, x);
}
inline CVector3 xyx() const
{
return CVector3(x, y, x);
}
inline CVector3 yyx() const
{
return CVector3(y, y, x);
}
inline CVector3 zyx() const
{
return CVector3(z, y, x);
}
inline CVector3 xzx() const
{
return CVector3(x, z, x);
}
inline CVector3 yzx() const
{
return CVector3(y, z, x);
}
inline CVector3 zzx() const
{
return CVector3(z, z, x);
}
inline CVector3 xxy() const
{
return CVector3(x, x, y);
}
inline CVector3 yxy() const
{
return CVector3(y, x, y);
}
inline CVector3 zxy() const
{
return CVector3(z, x, y);
}
inline CVector3 xyy() const
{
return CVector3(x, y, y);
}
inline CVector3 yyy() const
{
return CVector3(y, y, y);
}
inline CVector3 zyy() const
{
return CVector3(z, y, y);
}
inline CVector3 xzy() const
{
return CVector3(x, z, y);
}
inline CVector3 yzy() const
{
return CVector3(y, z, y);
}
inline CVector3 zzy() const
{
return CVector3(z, z, y);
}
inline CVector3 xxz() const
{
return CVector3(x, x, z);
}
inline CVector3 yxz() const
{
return CVector3(y, x, z);
}
inline CVector3 zxz() const
{
return CVector3(z, x, z);
}
inline CVector3 xyz() const
{
return CVector3(x, y, z);
}
inline CVector3 yyz() const
{
return CVector3(y, y, z);
}
inline CVector3 zyz() const
{
return CVector3(z, y, z);
}
inline CVector3 xzz() const
{
return CVector3(x, z, z);
}
inline CVector3 yzz() const
{
return CVector3(y, z, z);
}
inline CVector3 zzz() const
{
return CVector3(z, z, z);
}
inline void operator += (const CVector3 &v)
{
x+=v.x;
y+=v.y;
z+=v.z;
}
inline void operator -= (const CVector3 &v)
{
x-=v.x;
y-=v.y;
z-=v.z;
}
inline void operator *= (const double v)
{
x*=v;
y*=v;
z*=v;
}
inline void operator *= (const CVector3 &v)
{
x*=v.x;
y*=v.y;
z*=v.z;
}
inline void operator /= (const double v)
{
x/=v;
y/=v;
z/=v;
}
inline void operator /= (const CVector3 &v)
{
x/=v.x;
y/=v.y;
z/=v.z;
}
inline void fromTo(const CVector3 & start, const CVector3 & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
z = ende.z - start.z;
};
void setVector(double arg0, double arg1, double arg2)
{
x = arg0;
y = arg1;
z = arg2;
}
inline void addVector(double arg0, double arg1, double arg2)
{
x += arg0;
y += arg1;
z += arg2;
}
inline void subVector(double arg0, double arg1, double arg2)
{
x -= arg0;
y -= arg1;
z -= arg2;
}
inline void neg()
{
x = -x;
y = -y;
z = -z;
}
inline double length() const
{
return sqrt(x*x+y*y+z*z);
}
inline double normalize()
{
double r = sqrt(x*x+y*y+z*z);
   if (r < 0.00001) return r;
x /= r;
y /= r;
z /= r;
  return r;
}
inline double dotProduct(const CVector3 &b) const
{
return (x* b.x+y* b.y+z* b.z);
}
inline int getMajorAxis() const
{
double current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
if (current < fabs(z)) {
index = 2;
current = fabs(z);
};
return index;
}
inline double distanceTo(const CVector3 &v) const;
inline CVector3 operator +(const CVector3 &v2) const
{
CVector3 r;
r.x = x + v2.x;
r.y = y + v2.y;
r.z = z + v2.z;
return r;
}
inline CVector3 operator +(const double v2) const
{
CVector3 r;
r.x = x + v2;
r.y = y + v2;
r.z = z + v2;
return r;
}
inline CVector3 operator -(const CVector3 &v2) const
{
CVector3 r;
r.x = x - v2.x;
r.y = y - v2.y;
r.z = z - v2.z;
return r;
}
inline CVector3 operator -(const double v2) const
{
CVector3 r;
r.x = x - v2;
r.y = y - v2;
r.z = z - v2;
return r;
}
inline CVector3 operator -() const
{
CVector3 r;
r.x = -x;
r.y = -y;
r.z = -z;
return r;
}
inline CVector3 operator *(const CVector3 &v2) const
{
CVector3 r;
r.x = x * v2.x;
r.y = y * v2.y;
r.z = z * v2.z;
return r;
}
inline CVector3 operator *(const double v2) const
{
CVector3 r;
r.x = x * v2;
r.y = y * v2;
r.z = z * v2;
return r;
}

inline CVector3 operator /(const CVector3 &v2) const
{
CVector3 r;
r.x = x / v2.x;
r.y = y / v2.y;
r.z = z / v2.z;
return r;
}
inline CVector3 operator /(const double v2) const
{
CVector3 r;
r.x = x / v2;
r.y = y / v2;
r.z = z / v2;
return r;
}
inline bool operator ==(const CVector3 &v2) const
{
bool r;
r =  x == v2.x
 &&
 y == v2.y
 &&
 z == v2.z
;
return r;
}
inline bool operator !=(const CVector3 &v2) const
{
bool r;
r =  x != v2.x
 ||
 y != v2.y
 ||
 z != v2.z
;
return r;
}
};

inline CVector3 operator *(const double v1, const CVector3 &v2)
{
CVector3 r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
r.z = v1 * v2.z;
return r;
}

inline CVector3 operator /(const double v1, const CVector3 &v2)
{
CVector3 r;
r.x = v1 / v2.x;
r.y = v1 / v2.y;
r.z = v1 / v2.z;
return r;
}

inline double CVector3::distanceTo(const CVector3 &v) const
{
return (v-*this).length();
}
inline double dot(const CVector3 &a, const CVector3 &b)
{
return (a.x* b.x+a.y* b.y+a.z* b.z);
}
inline CVector3 umin(const CVector3 &a, const CVector3 &b)
{
CVector3 r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
r.z = umin(a.z, b.z);
return r;
}
inline CVector3 umax(const CVector3 &a, const CVector3 &b)
{
CVector3 r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
r.z = umax(a.z, b.z);
return r;
}
inline CVector3 normalize(const CVector3 &a)
{
CVector3 rvec;
double r = sqrt(a.x* a.x+a.y* a.y+a.z* a.z);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
rvec.z = a.z / r;
  return rvec;
}
inline double length(const CVector3 &a)
{
CVector3 rvec;
double r = sqrt(a.x*a.x+a.y*a.y+a.z*a.z);
   return r;
}
inline CVector3 sign(const CVector3 &a)
{
CVector3 rvec;
rvec.x = a.x > 0.0f ? 1.0f : -1.0f;
rvec.y = a.y > 0.0f ? 1.0f : -1.0f;
rvec.z = a.z > 0.0f ? 1.0f : -1.0f;
  return rvec;
}
inline double distance(const CVector3 &a, const CVector3 &b)
{
return sqrt(sqr(a.x-b.x)+sqr(a.y-b.y)+sqr(a.z-b.z));
}
inline CVector3 linear(const CVector3 &a, const CVector3 &b, const double v)
{
CVector3 rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
rvec.z = a.z+(b.z - a.z)*v;
return rvec;
}
struct CVector4 {
double x;
double y;
double z;
double w;
UTI_VECTOR_DECLARE_OPERATORS(CVector4, double);
inline CVector4(){/*x=0.0f;y=0.0f;z=0.0f;w=0.0f;*/};
inline CVector4(const double f) {
x = f;
y = f;
z = f;
w = f;
};
inline CVector4 returnNormal() const
{
    CVector4 res = *this;
    res.normalize();
    return res;
}
inline CVector4(const double arg0, const double arg1, const double arg2, const double arg3){
x = arg0;y = arg1;z = arg2;w = arg3;
}
inline CVector4(const double arg0, const double arg1, const CVector2 & arg2){
x = arg0;y = arg1;z = arg2.x;w = arg2.y;
}
inline CVector4(const double arg0, const CVector2 & arg1, const double arg2){
x = arg0;y = arg1.x;z = arg1.y;w = arg2;
}
inline CVector4(const double arg0, const CVector3 & arg1){
x = arg0;y = arg1.x;z = arg1.y;w = arg1.z;
}
inline CVector4(const CVector2 & arg0, const double arg1, const double arg2){
x = arg0.x;y = arg0.y;z = arg1;w = arg2;
}
inline CVector4(const CVector2 & arg0, const CVector2 & arg1){
x = arg0.x;y = arg0.y;z = arg1.x;w = arg1.y;
}
inline CVector4(const CVector3 & arg0, const double arg1){
x = arg0.x;y = arg0.y;z = arg0.z;w = arg1;
}
inline double &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 4);
return (&x)[p];
}
inline const double &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 4);
    return (&x)[p];
}
inline CVector2 xx() const
{
return CVector2(x, x);
}
inline CVector2 yx() const
{
return CVector2(y, x);
}
inline CVector2 zx() const
{
return CVector2(z, x);
}
inline CVector2 wx() const
{
return CVector2(w, x);
}
inline CVector2 xy() const
{
return CVector2(x, y);
}
inline CVector2 yy() const
{
return CVector2(y, y);
}
inline CVector2 zy() const
{
return CVector2(z, y);
}
inline CVector2 wy() const
{
return CVector2(w, y);
}
inline CVector2 xz() const
{
return CVector2(x, z);
}
inline CVector2 yz() const
{
return CVector2(y, z);
}
inline CVector2 zz() const
{
return CVector2(z, z);
}
inline CVector2 wz() const
{
return CVector2(w, z);
}
inline CVector2 xw() const
{
return CVector2(x, w);
}
inline CVector2 yw() const
{
return CVector2(y, w);
}
inline CVector2 zw() const
{
return CVector2(z, w);
}
inline CVector2 ww() const
{
return CVector2(w, w);
}
inline CVector3 xxx() const
{
return CVector3(x, x, x);
}
inline CVector3 yxx() const
{
return CVector3(y, x, x);
}
inline CVector3 zxx() const
{
return CVector3(z, x, x);
}
inline CVector3 wxx() const
{
return CVector3(w, x, x);
}
inline CVector3 xyx() const
{
return CVector3(x, y, x);
}
inline CVector3 yyx() const
{
return CVector3(y, y, x);
}
inline CVector3 zyx() const
{
return CVector3(z, y, x);
}
inline CVector3 wyx() const
{
return CVector3(w, y, x);
}
inline CVector3 xzx() const
{
return CVector3(x, z, x);
}
inline CVector3 yzx() const
{
return CVector3(y, z, x);
}
inline CVector3 zzx() const
{
return CVector3(z, z, x);
}
inline CVector3 wzx() const
{
return CVector3(w, z, x);
}
inline CVector3 xwx() const
{
return CVector3(x, w, x);
}
inline CVector3 ywx() const
{
return CVector3(y, w, x);
}
inline CVector3 zwx() const
{
return CVector3(z, w, x);
}
inline CVector3 wwx() const
{
return CVector3(w, w, x);
}
inline CVector3 xxy() const
{
return CVector3(x, x, y);
}
inline CVector3 yxy() const
{
return CVector3(y, x, y);
}
inline CVector3 zxy() const
{
return CVector3(z, x, y);
}
inline CVector3 wxy() const
{
return CVector3(w, x, y);
}
inline CVector3 xyy() const
{
return CVector3(x, y, y);
}
inline CVector3 yyy() const
{
return CVector3(y, y, y);
}
inline CVector3 zyy() const
{
return CVector3(z, y, y);
}
inline CVector3 wyy() const
{
return CVector3(w, y, y);
}
inline CVector3 xzy() const
{
return CVector3(x, z, y);
}
inline CVector3 yzy() const
{
return CVector3(y, z, y);
}
inline CVector3 zzy() const
{
return CVector3(z, z, y);
}
inline CVector3 wzy() const
{
return CVector3(w, z, y);
}
inline CVector3 xwy() const
{
return CVector3(x, w, y);
}
inline CVector3 ywy() const
{
return CVector3(y, w, y);
}
inline CVector3 zwy() const
{
return CVector3(z, w, y);
}
inline CVector3 wwy() const
{
return CVector3(w, w, y);
}
inline CVector3 xxz() const
{
return CVector3(x, x, z);
}
inline CVector3 yxz() const
{
return CVector3(y, x, z);
}
inline CVector3 zxz() const
{
return CVector3(z, x, z);
}
inline CVector3 wxz() const
{
return CVector3(w, x, z);
}
inline CVector3 xyz() const
{
return CVector3(x, y, z);
}
inline CVector3 yyz() const
{
return CVector3(y, y, z);
}
inline CVector3 zyz() const
{
return CVector3(z, y, z);
}
inline CVector3 wyz() const
{
return CVector3(w, y, z);
}
inline CVector3 xzz() const
{
return CVector3(x, z, z);
}
inline CVector3 yzz() const
{
return CVector3(y, z, z);
}
inline CVector3 zzz() const
{
return CVector3(z, z, z);
}
inline CVector3 wzz() const
{
return CVector3(w, z, z);
}
inline CVector3 xwz() const
{
return CVector3(x, w, z);
}
inline CVector3 ywz() const
{
return CVector3(y, w, z);
}
inline CVector3 zwz() const
{
return CVector3(z, w, z);
}
inline CVector3 wwz() const
{
return CVector3(w, w, z);
}
inline CVector3 xxw() const
{
return CVector3(x, x, w);
}
inline CVector3 yxw() const
{
return CVector3(y, x, w);
}
inline CVector3 zxw() const
{
return CVector3(z, x, w);
}
inline CVector3 wxw() const
{
return CVector3(w, x, w);
}
inline CVector3 xyw() const
{
return CVector3(x, y, w);
}
inline CVector3 yyw() const
{
return CVector3(y, y, w);
}
inline CVector3 zyw() const
{
return CVector3(z, y, w);
}
inline CVector3 wyw() const
{
return CVector3(w, y, w);
}
inline CVector3 xzw() const
{
return CVector3(x, z, w);
}
inline CVector3 yzw() const
{
return CVector3(y, z, w);
}
inline CVector3 zzw() const
{
return CVector3(z, z, w);
}
inline CVector3 wzw() const
{
return CVector3(w, z, w);
}
inline CVector3 xww() const
{
return CVector3(x, w, w);
}
inline CVector3 yww() const
{
return CVector3(y, w, w);
}
inline CVector3 zww() const
{
return CVector3(z, w, w);
}
inline CVector3 www() const
{
return CVector3(w, w, w);
}
inline CVector4 xxxx() const
{
return CVector4(x, x, x, x);
}
inline CVector4 yxxx() const
{
return CVector4(y, x, x, x);
}
inline CVector4 zxxx() const
{
return CVector4(z, x, x, x);
}
inline CVector4 wxxx() const
{
return CVector4(w, x, x, x);
}
inline CVector4 xyxx() const
{
return CVector4(x, y, x, x);
}
inline CVector4 yyxx() const
{
return CVector4(y, y, x, x);
}
inline CVector4 zyxx() const
{
return CVector4(z, y, x, x);
}
inline CVector4 wyxx() const
{
return CVector4(w, y, x, x);
}
inline CVector4 xzxx() const
{
return CVector4(x, z, x, x);
}
inline CVector4 yzxx() const
{
return CVector4(y, z, x, x);
}
inline CVector4 zzxx() const
{
return CVector4(z, z, x, x);
}
inline CVector4 wzxx() const
{
return CVector4(w, z, x, x);
}
inline CVector4 xwxx() const
{
return CVector4(x, w, x, x);
}
inline CVector4 ywxx() const
{
return CVector4(y, w, x, x);
}
inline CVector4 zwxx() const
{
return CVector4(z, w, x, x);
}
inline CVector4 wwxx() const
{
return CVector4(w, w, x, x);
}
inline CVector4 xxyx() const
{
return CVector4(x, x, y, x);
}
inline CVector4 yxyx() const
{
return CVector4(y, x, y, x);
}
inline CVector4 zxyx() const
{
return CVector4(z, x, y, x);
}
inline CVector4 wxyx() const
{
return CVector4(w, x, y, x);
}
inline CVector4 xyyx() const
{
return CVector4(x, y, y, x);
}
inline CVector4 yyyx() const
{
return CVector4(y, y, y, x);
}
inline CVector4 zyyx() const
{
return CVector4(z, y, y, x);
}
inline CVector4 wyyx() const
{
return CVector4(w, y, y, x);
}
inline CVector4 xzyx() const
{
return CVector4(x, z, y, x);
}
inline CVector4 yzyx() const
{
return CVector4(y, z, y, x);
}
inline CVector4 zzyx() const
{
return CVector4(z, z, y, x);
}
inline CVector4 wzyx() const
{
return CVector4(w, z, y, x);
}
inline CVector4 xwyx() const
{
return CVector4(x, w, y, x);
}
inline CVector4 ywyx() const
{
return CVector4(y, w, y, x);
}
inline CVector4 zwyx() const
{
return CVector4(z, w, y, x);
}
inline CVector4 wwyx() const
{
return CVector4(w, w, y, x);
}
inline CVector4 xxzx() const
{
return CVector4(x, x, z, x);
}
inline CVector4 yxzx() const
{
return CVector4(y, x, z, x);
}
inline CVector4 zxzx() const
{
return CVector4(z, x, z, x);
}
inline CVector4 wxzx() const
{
return CVector4(w, x, z, x);
}
inline CVector4 xyzx() const
{
return CVector4(x, y, z, x);
}
inline CVector4 yyzx() const
{
return CVector4(y, y, z, x);
}
inline CVector4 zyzx() const
{
return CVector4(z, y, z, x);
}
inline CVector4 wyzx() const
{
return CVector4(w, y, z, x);
}
inline CVector4 xzzx() const
{
return CVector4(x, z, z, x);
}
inline CVector4 yzzx() const
{
return CVector4(y, z, z, x);
}
inline CVector4 zzzx() const
{
return CVector4(z, z, z, x);
}
inline CVector4 wzzx() const
{
return CVector4(w, z, z, x);
}
inline CVector4 xwzx() const
{
return CVector4(x, w, z, x);
}
inline CVector4 ywzx() const
{
return CVector4(y, w, z, x);
}
inline CVector4 zwzx() const
{
return CVector4(z, w, z, x);
}
inline CVector4 wwzx() const
{
return CVector4(w, w, z, x);
}
inline CVector4 xxwx() const
{
return CVector4(x, x, w, x);
}
inline CVector4 yxwx() const
{
return CVector4(y, x, w, x);
}
inline CVector4 zxwx() const
{
return CVector4(z, x, w, x);
}
inline CVector4 wxwx() const
{
return CVector4(w, x, w, x);
}
inline CVector4 xywx() const
{
return CVector4(x, y, w, x);
}
inline CVector4 yywx() const
{
return CVector4(y, y, w, x);
}
inline CVector4 zywx() const
{
return CVector4(z, y, w, x);
}
inline CVector4 wywx() const
{
return CVector4(w, y, w, x);
}
inline CVector4 xzwx() const
{
return CVector4(x, z, w, x);
}
inline CVector4 yzwx() const
{
return CVector4(y, z, w, x);
}
inline CVector4 zzwx() const
{
return CVector4(z, z, w, x);
}
inline CVector4 wzwx() const
{
return CVector4(w, z, w, x);
}
inline CVector4 xwwx() const
{
return CVector4(x, w, w, x);
}
inline CVector4 ywwx() const
{
return CVector4(y, w, w, x);
}
inline CVector4 zwwx() const
{
return CVector4(z, w, w, x);
}
inline CVector4 wwwx() const
{
return CVector4(w, w, w, x);
}
inline CVector4 xxxy() const
{
return CVector4(x, x, x, y);
}
inline CVector4 yxxy() const
{
return CVector4(y, x, x, y);
}
inline CVector4 zxxy() const
{
return CVector4(z, x, x, y);
}
inline CVector4 wxxy() const
{
return CVector4(w, x, x, y);
}
inline CVector4 xyxy() const
{
return CVector4(x, y, x, y);
}
inline CVector4 yyxy() const
{
return CVector4(y, y, x, y);
}
inline CVector4 zyxy() const
{
return CVector4(z, y, x, y);
}
inline CVector4 wyxy() const
{
return CVector4(w, y, x, y);
}
inline CVector4 xzxy() const
{
return CVector4(x, z, x, y);
}
inline CVector4 yzxy() const
{
return CVector4(y, z, x, y);
}
inline CVector4 zzxy() const
{
return CVector4(z, z, x, y);
}
inline CVector4 wzxy() const
{
return CVector4(w, z, x, y);
}
inline CVector4 xwxy() const
{
return CVector4(x, w, x, y);
}
inline CVector4 ywxy() const
{
return CVector4(y, w, x, y);
}
inline CVector4 zwxy() const
{
return CVector4(z, w, x, y);
}
inline CVector4 wwxy() const
{
return CVector4(w, w, x, y);
}
inline CVector4 xxyy() const
{
return CVector4(x, x, y, y);
}
inline CVector4 yxyy() const
{
return CVector4(y, x, y, y);
}
inline CVector4 zxyy() const
{
return CVector4(z, x, y, y);
}
inline CVector4 wxyy() const
{
return CVector4(w, x, y, y);
}
inline CVector4 xyyy() const
{
return CVector4(x, y, y, y);
}
inline CVector4 yyyy() const
{
return CVector4(y, y, y, y);
}
inline CVector4 zyyy() const
{
return CVector4(z, y, y, y);
}
inline CVector4 wyyy() const
{
return CVector4(w, y, y, y);
}
inline CVector4 xzyy() const
{
return CVector4(x, z, y, y);
}
inline CVector4 yzyy() const
{
return CVector4(y, z, y, y);
}
inline CVector4 zzyy() const
{
return CVector4(z, z, y, y);
}
inline CVector4 wzyy() const
{
return CVector4(w, z, y, y);
}
inline CVector4 xwyy() const
{
return CVector4(x, w, y, y);
}
inline CVector4 ywyy() const
{
return CVector4(y, w, y, y);
}
inline CVector4 zwyy() const
{
return CVector4(z, w, y, y);
}
inline CVector4 wwyy() const
{
return CVector4(w, w, y, y);
}
inline CVector4 xxzy() const
{
return CVector4(x, x, z, y);
}
inline CVector4 yxzy() const
{
return CVector4(y, x, z, y);
}
inline CVector4 zxzy() const
{
return CVector4(z, x, z, y);
}
inline CVector4 wxzy() const
{
return CVector4(w, x, z, y);
}
inline CVector4 xyzy() const
{
return CVector4(x, y, z, y);
}
inline CVector4 yyzy() const
{
return CVector4(y, y, z, y);
}
inline CVector4 zyzy() const
{
return CVector4(z, y, z, y);
}
inline CVector4 wyzy() const
{
return CVector4(w, y, z, y);
}
inline CVector4 xzzy() const
{
return CVector4(x, z, z, y);
}
inline CVector4 yzzy() const
{
return CVector4(y, z, z, y);
}
inline CVector4 zzzy() const
{
return CVector4(z, z, z, y);
}
inline CVector4 wzzy() const
{
return CVector4(w, z, z, y);
}
inline CVector4 xwzy() const
{
return CVector4(x, w, z, y);
}
inline CVector4 ywzy() const
{
return CVector4(y, w, z, y);
}
inline CVector4 zwzy() const
{
return CVector4(z, w, z, y);
}
inline CVector4 wwzy() const
{
return CVector4(w, w, z, y);
}
inline CVector4 xxwy() const
{
return CVector4(x, x, w, y);
}
inline CVector4 yxwy() const
{
return CVector4(y, x, w, y);
}
inline CVector4 zxwy() const
{
return CVector4(z, x, w, y);
}
inline CVector4 wxwy() const
{
return CVector4(w, x, w, y);
}
inline CVector4 xywy() const
{
return CVector4(x, y, w, y);
}
inline CVector4 yywy() const
{
return CVector4(y, y, w, y);
}
inline CVector4 zywy() const
{
return CVector4(z, y, w, y);
}
inline CVector4 wywy() const
{
return CVector4(w, y, w, y);
}
inline CVector4 xzwy() const
{
return CVector4(x, z, w, y);
}
inline CVector4 yzwy() const
{
return CVector4(y, z, w, y);
}
inline CVector4 zzwy() const
{
return CVector4(z, z, w, y);
}
inline CVector4 wzwy() const
{
return CVector4(w, z, w, y);
}
inline CVector4 xwwy() const
{
return CVector4(x, w, w, y);
}
inline CVector4 ywwy() const
{
return CVector4(y, w, w, y);
}
inline CVector4 zwwy() const
{
return CVector4(z, w, w, y);
}
inline CVector4 wwwy() const
{
return CVector4(w, w, w, y);
}
inline CVector4 xxxz() const
{
return CVector4(x, x, x, z);
}
inline CVector4 yxxz() const
{
return CVector4(y, x, x, z);
}
inline CVector4 zxxz() const
{
return CVector4(z, x, x, z);
}
inline CVector4 wxxz() const
{
return CVector4(w, x, x, z);
}
inline CVector4 xyxz() const
{
return CVector4(x, y, x, z);
}
inline CVector4 yyxz() const
{
return CVector4(y, y, x, z);
}
inline CVector4 zyxz() const
{
return CVector4(z, y, x, z);
}
inline CVector4 wyxz() const
{
return CVector4(w, y, x, z);
}
inline CVector4 xzxz() const
{
return CVector4(x, z, x, z);
}
inline CVector4 yzxz() const
{
return CVector4(y, z, x, z);
}
inline CVector4 zzxz() const
{
return CVector4(z, z, x, z);
}
inline CVector4 wzxz() const
{
return CVector4(w, z, x, z);
}
inline CVector4 xwxz() const
{
return CVector4(x, w, x, z);
}
inline CVector4 ywxz() const
{
return CVector4(y, w, x, z);
}
inline CVector4 zwxz() const
{
return CVector4(z, w, x, z);
}
inline CVector4 wwxz() const
{
return CVector4(w, w, x, z);
}
inline CVector4 xxyz() const
{
return CVector4(x, x, y, z);
}
inline CVector4 yxyz() const
{
return CVector4(y, x, y, z);
}
inline CVector4 zxyz() const
{
return CVector4(z, x, y, z);
}
inline CVector4 wxyz() const
{
return CVector4(w, x, y, z);
}
inline CVector4 xyyz() const
{
return CVector4(x, y, y, z);
}
inline CVector4 yyyz() const
{
return CVector4(y, y, y, z);
}
inline CVector4 zyyz() const
{
return CVector4(z, y, y, z);
}
inline CVector4 wyyz() const
{
return CVector4(w, y, y, z);
}
inline CVector4 xzyz() const
{
return CVector4(x, z, y, z);
}
inline CVector4 yzyz() const
{
return CVector4(y, z, y, z);
}
inline CVector4 zzyz() const
{
return CVector4(z, z, y, z);
}
inline CVector4 wzyz() const
{
return CVector4(w, z, y, z);
}
inline CVector4 xwyz() const
{
return CVector4(x, w, y, z);
}
inline CVector4 ywyz() const
{
return CVector4(y, w, y, z);
}
inline CVector4 zwyz() const
{
return CVector4(z, w, y, z);
}
inline CVector4 wwyz() const
{
return CVector4(w, w, y, z);
}
inline CVector4 xxzz() const
{
return CVector4(x, x, z, z);
}
inline CVector4 yxzz() const
{
return CVector4(y, x, z, z);
}
inline CVector4 zxzz() const
{
return CVector4(z, x, z, z);
}
inline CVector4 wxzz() const
{
return CVector4(w, x, z, z);
}
inline CVector4 xyzz() const
{
return CVector4(x, y, z, z);
}
inline CVector4 yyzz() const
{
return CVector4(y, y, z, z);
}
inline CVector4 zyzz() const
{
return CVector4(z, y, z, z);
}
inline CVector4 wyzz() const
{
return CVector4(w, y, z, z);
}
inline CVector4 xzzz() const
{
return CVector4(x, z, z, z);
}
inline CVector4 yzzz() const
{
return CVector4(y, z, z, z);
}
inline CVector4 zzzz() const
{
return CVector4(z, z, z, z);
}
inline CVector4 wzzz() const
{
return CVector4(w, z, z, z);
}
inline CVector4 xwzz() const
{
return CVector4(x, w, z, z);
}
inline CVector4 ywzz() const
{
return CVector4(y, w, z, z);
}
inline CVector4 zwzz() const
{
return CVector4(z, w, z, z);
}
inline CVector4 wwzz() const
{
return CVector4(w, w, z, z);
}
inline CVector4 xxwz() const
{
return CVector4(x, x, w, z);
}
inline CVector4 yxwz() const
{
return CVector4(y, x, w, z);
}
inline CVector4 zxwz() const
{
return CVector4(z, x, w, z);
}
inline CVector4 wxwz() const
{
return CVector4(w, x, w, z);
}
inline CVector4 xywz() const
{
return CVector4(x, y, w, z);
}
inline CVector4 yywz() const
{
return CVector4(y, y, w, z);
}
inline CVector4 zywz() const
{
return CVector4(z, y, w, z);
}
inline CVector4 wywz() const
{
return CVector4(w, y, w, z);
}
inline CVector4 xzwz() const
{
return CVector4(x, z, w, z);
}
inline CVector4 yzwz() const
{
return CVector4(y, z, w, z);
}
inline CVector4 zzwz() const
{
return CVector4(z, z, w, z);
}
inline CVector4 wzwz() const
{
return CVector4(w, z, w, z);
}
inline CVector4 xwwz() const
{
return CVector4(x, w, w, z);
}
inline CVector4 ywwz() const
{
return CVector4(y, w, w, z);
}
inline CVector4 zwwz() const
{
return CVector4(z, w, w, z);
}
inline CVector4 wwwz() const
{
return CVector4(w, w, w, z);
}
inline CVector4 xxxw() const
{
return CVector4(x, x, x, w);
}
inline CVector4 yxxw() const
{
return CVector4(y, x, x, w);
}
inline CVector4 zxxw() const
{
return CVector4(z, x, x, w);
}
inline CVector4 wxxw() const
{
return CVector4(w, x, x, w);
}
inline CVector4 xyxw() const
{
return CVector4(x, y, x, w);
}
inline CVector4 yyxw() const
{
return CVector4(y, y, x, w);
}
inline CVector4 zyxw() const
{
return CVector4(z, y, x, w);
}
inline CVector4 wyxw() const
{
return CVector4(w, y, x, w);
}
inline CVector4 xzxw() const
{
return CVector4(x, z, x, w);
}
inline CVector4 yzxw() const
{
return CVector4(y, z, x, w);
}
inline CVector4 zzxw() const
{
return CVector4(z, z, x, w);
}
inline CVector4 wzxw() const
{
return CVector4(w, z, x, w);
}
inline CVector4 xwxw() const
{
return CVector4(x, w, x, w);
}
inline CVector4 ywxw() const
{
return CVector4(y, w, x, w);
}
inline CVector4 zwxw() const
{
return CVector4(z, w, x, w);
}
inline CVector4 wwxw() const
{
return CVector4(w, w, x, w);
}
inline CVector4 xxyw() const
{
return CVector4(x, x, y, w);
}
inline CVector4 yxyw() const
{
return CVector4(y, x, y, w);
}
inline CVector4 zxyw() const
{
return CVector4(z, x, y, w);
}
inline CVector4 wxyw() const
{
return CVector4(w, x, y, w);
}
inline CVector4 xyyw() const
{
return CVector4(x, y, y, w);
}
inline CVector4 yyyw() const
{
return CVector4(y, y, y, w);
}
inline CVector4 zyyw() const
{
return CVector4(z, y, y, w);
}
inline CVector4 wyyw() const
{
return CVector4(w, y, y, w);
}
inline CVector4 xzyw() const
{
return CVector4(x, z, y, w);
}
inline CVector4 yzyw() const
{
return CVector4(y, z, y, w);
}
inline CVector4 zzyw() const
{
return CVector4(z, z, y, w);
}
inline CVector4 wzyw() const
{
return CVector4(w, z, y, w);
}
inline CVector4 xwyw() const
{
return CVector4(x, w, y, w);
}
inline CVector4 ywyw() const
{
return CVector4(y, w, y, w);
}
inline CVector4 zwyw() const
{
return CVector4(z, w, y, w);
}
inline CVector4 wwyw() const
{
return CVector4(w, w, y, w);
}
inline CVector4 xxzw() const
{
return CVector4(x, x, z, w);
}
inline CVector4 yxzw() const
{
return CVector4(y, x, z, w);
}
inline CVector4 zxzw() const
{
return CVector4(z, x, z, w);
}
inline CVector4 wxzw() const
{
return CVector4(w, x, z, w);
}
inline CVector4 xyzw() const
{
return CVector4(x, y, z, w);
}
inline CVector4 yyzw() const
{
return CVector4(y, y, z, w);
}
inline CVector4 zyzw() const
{
return CVector4(z, y, z, w);
}
inline CVector4 wyzw() const
{
return CVector4(w, y, z, w);
}
inline CVector4 xzzw() const
{
return CVector4(x, z, z, w);
}
inline CVector4 yzzw() const
{
return CVector4(y, z, z, w);
}
inline CVector4 zzzw() const
{
return CVector4(z, z, z, w);
}
inline CVector4 wzzw() const
{
return CVector4(w, z, z, w);
}
inline CVector4 xwzw() const
{
return CVector4(x, w, z, w);
}
inline CVector4 ywzw() const
{
return CVector4(y, w, z, w);
}
inline CVector4 zwzw() const
{
return CVector4(z, w, z, w);
}
inline CVector4 wwzw() const
{
return CVector4(w, w, z, w);
}
inline CVector4 xxww() const
{
return CVector4(x, x, w, w);
}
inline CVector4 yxww() const
{
return CVector4(y, x, w, w);
}
inline CVector4 zxww() const
{
return CVector4(z, x, w, w);
}
inline CVector4 wxww() const
{
return CVector4(w, x, w, w);
}
inline CVector4 xyww() const
{
return CVector4(x, y, w, w);
}
inline CVector4 yyww() const
{
return CVector4(y, y, w, w);
}
inline CVector4 zyww() const
{
return CVector4(z, y, w, w);
}
inline CVector4 wyww() const
{
return CVector4(w, y, w, w);
}
inline CVector4 xzww() const
{
return CVector4(x, z, w, w);
}
inline CVector4 yzww() const
{
return CVector4(y, z, w, w);
}
inline CVector4 zzww() const
{
return CVector4(z, z, w, w);
}
inline CVector4 wzww() const
{
return CVector4(w, z, w, w);
}
inline CVector4 xwww() const
{
return CVector4(x, w, w, w);
}
inline CVector4 ywww() const
{
return CVector4(y, w, w, w);
}
inline CVector4 zwww() const
{
return CVector4(z, w, w, w);
}
inline CVector4 wwww() const
{
return CVector4(w, w, w, w);
}
inline void operator += (const CVector4 &v)
{
x+=v.x;
y+=v.y;
z+=v.z;
w+=v.w;
}
inline void operator -= (const CVector4 &v)
{
x-=v.x;
y-=v.y;
z-=v.z;
w-=v.w;
}
inline void operator *= (const double v)
{
x*=v;
y*=v;
z*=v;
w*=v;
}
inline void operator *= (const CVector4 &v)
{
x*=v.x;
y*=v.y;
z*=v.z;
w*=v.w;
}
inline void operator /= (const double v)
{
x/=v;
y/=v;
z/=v;
w/=v;
}
inline void operator /= (const CVector4 &v)
{
x/=v.x;
y/=v.y;
z/=v.z;
w/=v.w;
}
inline void fromTo(const CVector4 & start, const CVector4 & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
z = ende.z - start.z;
w = ende.w - start.w;
};
void setVector(double arg0, double arg1, double arg2, double arg3)
{
x = arg0;
y = arg1;
z = arg2;
w = arg3;
}
inline void addVector(double arg0, double arg1, double arg2, double arg3)
{
x += arg0;
y += arg1;
z += arg2;
w += arg3;
}
inline void subVector(double arg0, double arg1, double arg2, double arg3)
{
x -= arg0;
y -= arg1;
z -= arg2;
w -= arg3;
}
inline void neg()
{
x = -x;
y = -y;
z = -z;
w = -w;
}
inline double length() const
{
return sqrt(x*x+y*y+z*z+w*w);
}
inline double normalize()
{
double r = sqrt(x*x+y*y+z*z+w*w);
   if (r < 0.00001) return r;
x /= r;
y /= r;
z /= r;
w /= r;
  return r;
}
inline double dotProduct(const CVector4 &b) const
{
return (x* b.x+y* b.y+z* b.z+w* b.w);
}
inline int getMajorAxis() const
{
double current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
if (current < fabs(z)) {
index = 2;
current = fabs(z);
};
if (current < fabs(w)) {
index = 3;
current = fabs(w);
};
return index;
}
inline double distanceTo(const CVector4 &v) const;
inline CVector4 operator +(const CVector4 &v2)const
{
CVector4 r;
r.x = x + v2.x;
r.y = y + v2.y;
r.z = z + v2.z;
r.w = w + v2.w;
return r;
}
inline CVector4 operator +( const double v2)const
{
CVector4 r;
r.x = x + v2;
r.y = y + v2;
r.z = z + v2;
r.w = w + v2;
return r;
}
inline CVector4 operator -(const CVector4 &v2)const
{
CVector4 r;
r.x = x - v2.x;
r.y = y - v2.y;
r.z = z - v2.z;
r.w = w - v2.w;
return r;
}
inline CVector4 operator -(const double v2)const
{
CVector4 r;
r.x = x - v2;
r.y = y - v2;
r.z = z - v2;
r.w = w - v2;
return r;
}
inline CVector4 operator -()const
{
CVector4 r;
r.x = -x;
r.y = -y;
r.z = -z;
r.w = -w;
return r;
}
inline CVector4 operator *(const CVector4 &v2)const
{
CVector4 r;
r.x = x * v2.x;
r.y = y * v2.y;
r.z = z * v2.z;
r.w = w * v2.w;
return r;
}
inline CVector4 operator *(const double v2)const
{
CVector4 r;
r.x = x * v2;
r.y = y * v2;
r.z = z * v2;
r.w = w * v2;
return r;
}

inline CVector4 operator /(const CVector4 &v2)const
{
CVector4 r;
r.x = x / v2.x;
r.y = y / v2.y;
r.z = z / v2.z;
r.w = w / v2.w;
return r;
}
inline CVector4 operator /(const double v2)const
{
CVector4 r;
r.x = x / v2;
r.y = y / v2;
r.z = z / v2;
r.w = w / v2;
return r;
}
inline bool operator ==(const CVector4 &v2)const
{
bool r;
r =  x == v2.x
 &&
 y == v2.y
 &&
 z == v2.z
 &&
 w == v2.w
;
return r;
}
inline bool operator !=(const CVector4 &v2) const
{
bool r;
r =  x != v2.x
 ||
 y != v2.y
 ||
 z != v2.z
 ||
 w != v2.w
;
return r;
}
};

inline CVector4 operator *(const double v1, const CVector4 &v2)
{
CVector4 r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
r.z = v1 * v2.z;
r.w = v1 * v2.w;
return r;
}

inline double CVector4::distanceTo(const CVector4 &v) const
{
return (v-*this).length();
}
inline double dot(const CVector4 &a, const CVector4 &b)
{
return (a.x* b.x+a.y* b.y+a.z* b.z+a.w* b.w);
}
inline CVector4 umin(const CVector4 &a, const CVector4 &b)
{
CVector4 r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
r.z = umin(a.z, b.z);
r.w = umin(a.w, b.w);
return r;
}
inline CVector4 umax(const CVector4 &a, const CVector4 &b)
{
CVector4 r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
r.z = umax(a.z, b.z);
r.w = umax(a.w, b.w);
return r;
}
inline CVector4 normalize(const CVector4 &a)
{
CVector4 rvec;
double r = sqrt(a.x* a.x+a.y* a.y+a.z* a.z+a.w* a.w);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
rvec.z = a.z / r;
rvec.w = a.w / r;
  return rvec;
}
inline double length(const CVector4 &a)
{
CVector4 rvec;
double r = sqrt(a.x*a.x+a.y*a.y+a.z*a.z+a.w*a.w);
   return r;
}
inline CVector4 sign(const CVector4 &a)
{
CVector4 rvec;
rvec.x = a.x > 0.0f ? 1.0f : -1.0f;
rvec.y = a.y > 0.0f ? 1.0f : -1.0f;
rvec.z = a.z > 0.0f ? 1.0f : -1.0f;
rvec.w = a.w > 0.0f ? 1.0f : -1.0f;
  return rvec;
}
inline double distance(const CVector4 &a, const CVector4 &b)
{
return sqrt(sqr(a.x-b.x)+sqr(a.y-b.y)+sqr(a.z-b.z)+sqr(a.w-b.w));
}
inline CVector4 linear(const CVector4 &a, const CVector4 &b, const double v)
{
CVector4 rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
rvec.z = a.z+(b.z - a.z)*v;
rvec.w = a.w+(b.w - a.w)*v;
return rvec;
}

//float

struct CVector2f {
float x;
float y;
//UTI_VECTOR_DECLARE_OPERATORS(CVector2f, float);
inline CVector2f(){/*x=0.0f;y=0.0f;*/};
inline CVector2f(const float f) {
x = f;
y = f;
};
inline CVector2f(const float arg0, const float arg1){
x = arg0;y = arg1;
}
inline CVector2f returnNormal() const
{
    CVector2f res = *this;
    res.normalize();
    return res;
}
inline float &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 2);
return (&x)[p];
}
inline const float &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 2);
    return (&x)[p];
}
inline CVector2f xx() const
{
return CVector2f(x, x);
}
inline CVector2f yx() const
{
return CVector2f(y, x);
}
inline CVector2f xy() const
{
return CVector2f(x, y);
}
inline CVector2f yy() const
{
return CVector2f(y, y);
}
inline void operator += (const CVector2f &v)
{
x+=v.x;
y+=v.y;
}
inline void operator -= (const CVector2f &v)
{
x-=v.x;
y-=v.y;
}
inline void operator *= (const float v)
{
x*=v;
y*=v;
}
inline void operator *= (const CVector2f &v)
{
x*=v.x;
y*=v.y;
}
inline void operator /= (const float v)
{
x/=v;
y/=v;
}
inline void operator /= (const CVector2f &v)
{
x/=v.x;
y/=v.y;
}
inline void fromTo(const CVector2f & start, const CVector2f & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
};
void setVector(float arg0, float arg1)
{
x = arg0;
y = arg1;
}
inline void addVector(float arg0, float arg1)
{
x += arg0;
y += arg1;
}
inline void subVector(float arg0, float arg1)
{
x -= arg0;
y -= arg1;
}
inline void neg()
{
x = -x;
y = -y;
}
inline float length() const
{
return sqrtf(x*x+y*y);
}
inline float normalize()
{
float r = sqrtf(x*x+y*y);
   if (r < 0.00001) return r;
x /= r;
y /= r;
  return r;
}
inline float dotProduct(const CVector2f &b) const
{
return (x* b.x+y* b.y);
}
inline int getMajorAxis() const
{
float current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
return index;
}
inline float distanceTo(const CVector2f &v) const;
};
inline CVector2f operator +(const CVector2f &v1, const CVector2f &v2)
{
CVector2f r;
r.x = v1.x + v2.x;
r.y = v1.y + v2.y;
return r;
}
inline CVector2f operator +(const CVector2f &v1, const float v2)
{
CVector2f r;
r.x = v1.x + v2;
r.y = v1.y + v2;
return r;
}
inline CVector2f operator -(const CVector2f &v1, const CVector2f &v2)
{
CVector2f r;
r.x = v1.x - v2.x;
r.y = v1.y - v2.y;
return r;
}
inline CVector2f operator -(const CVector2f &v1, const float v2)
{
CVector2f r;
r.x = v1.x - v2;
r.y = v1.y - v2;
return r;
}
inline CVector2f operator -(const CVector2f &v1)
{
CVector2f r;
r.x = -v1.x;
r.y = -v1.y;
return r;
}
inline CVector2f operator *(const CVector2f &v1, const CVector2f &v2)
{
CVector2f r;
r.x = v1.x * v2.x;
r.y = v1.y * v2.y;
return r;
}
inline CVector2f operator *(const CVector2f &v1, const float v2)
{
CVector2f r;
r.x = v1.x * v2;
r.y = v1.y * v2;
return r;
}
inline CVector2f operator *(const float v1, const CVector2f &v2)
{
CVector2f r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
return r;
}
inline CVector2f operator /(const CVector2f &v1, const CVector2f &v2)
{
CVector2f r;
r.x = v1.x / v2.x;
r.y = v1.y / v2.y;
return r;
}
inline CVector2f operator /(const CVector2f &v1, const float v2)
{
CVector2f r;
r.x = v1.x / v2;
r.y = v1.y / v2;
return r;
}
inline bool operator ==(const CVector2f &v1, const CVector2f &v2)
{
bool r;
r =  v1.x == v2.x
 &&
 v1.y == v2.y
;
return r;
}
inline bool operator !=(const CVector2f &v1, const CVector2f &v2)
{
bool r;
r =  v1.x != v2.x
 ||
 v1.y != v2.y
;
return r;
}
inline float CVector2f::distanceTo(const CVector2f &v) const
{
return (v-*this).length();
}
inline float dot(const CVector2f &a, const CVector2f &b)
{
return (a.x* b.x+a.y* b.y);
}
inline CVector2f umin(const CVector2f &a, const CVector2f &b)
{
CVector2f r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
return r;
}
inline CVector2f umax(const CVector2f &a, const CVector2f &b)
{
CVector2f r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
return r;
}
inline CVector2f normalize(const CVector2f &a)
{
CVector2f rvec;
float r = sqrtf(a.x* a.x+a.y* a.y);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
  return rvec;
}
inline float length(const CVector2f &a)
{
CVector2f rvec;
float r = sqrtf(a.x*a.x+a.y*a.y);
   return r;
}
inline CVector2f sign(const CVector2f &a)
{
CVector2f rvec;
rvec.x = a.x > 0.0f ? 1.0f : -1.0f;
rvec.y = a.y > 0.0f ? 1.0f : -1.0f;
  return rvec;
}
inline float distance(const CVector2f &a, const CVector2f &b)
{
return sqrtf(sqrf(a.x-b.x)+sqrf(a.y-b.y));
}
inline CVector2f linear(const CVector2f &a, const CVector2f &b, const float v)
{
CVector2f rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
return rvec;
}
struct CVector3f {
float x;
float y;
float z;
//UTI_VECTOR_DECLARE_OPERATORS(CVector3f, float);
inline CVector3f(){/*x=0.0f;y=0.0f;z=0.0f;*/};
explicit inline CVector3f(const float f) {
x = f;
y = f;
z = f;
};
inline CVector3f returnNormal() const
{
    CVector3f res = *this;
    res.normalize();
    return res;
}
inline CVector3f(const float arg0, const float arg1, const float arg2){
x = arg0;y = arg1;z = arg2;
}
inline CVector3f(const float arg0, const CVector2f & arg1){
x = arg0;y = arg1.x;z = arg1.y;
}
inline CVector3f(const CVector2f & arg0, const float arg1){
x = arg0.x;y = arg0.y;z = arg1;
}
inline float &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 3);
return (&x)[p];
}
inline const float &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 3);
    return (&x)[p];
}
inline CVector2f xx() const
{
return CVector2f(x, x);
}
inline CVector2f yx() const
{
return CVector2f(y, x);
}
inline CVector2f zx() const
{
return CVector2f(z, x);
}
inline CVector2f xy() const
{
return CVector2f(x, y);
}
inline CVector2f yy() const
{
return CVector2f(y, y);
}
inline CVector2f zy() const
{
return CVector2f(z, y);
}
inline CVector2f xz() const
{
return CVector2f(x, z);
}
inline CVector2f yz() const
{
return CVector2f(y, z);
}
inline CVector2f zz() const
{
return CVector2f(z, z);
}
inline CVector3f xxx() const
{
return CVector3f(x, x, x);
}
inline CVector3f yxx() const
{
return CVector3f(y, x, x);
}
inline CVector3f zxx() const
{
return CVector3f(z, x, x);
}
inline CVector3f xyx() const
{
return CVector3f(x, y, x);
}
inline CVector3f yyx() const
{
return CVector3f(y, y, x);
}
inline CVector3f zyx() const
{
return CVector3f(z, y, x);
}
inline CVector3f xzx() const
{
return CVector3f(x, z, x);
}
inline CVector3f yzx() const
{
return CVector3f(y, z, x);
}
inline CVector3f zzx() const
{
return CVector3f(z, z, x);
}
inline CVector3f xxy() const
{
return CVector3f(x, x, y);
}
inline CVector3f yxy() const
{
return CVector3f(y, x, y);
}
inline CVector3f zxy() const
{
return CVector3f(z, x, y);
}
inline CVector3f xyy() const
{
return CVector3f(x, y, y);
}
inline CVector3f yyy() const
{
return CVector3f(y, y, y);
}
inline CVector3f zyy() const
{
return CVector3f(z, y, y);
}
inline CVector3f xzy() const
{
return CVector3f(x, z, y);
}
inline CVector3f yzy() const
{
return CVector3f(y, z, y);
}
inline CVector3f zzy() const
{
return CVector3f(z, z, y);
}
inline CVector3f xxz() const
{
return CVector3f(x, x, z);
}
inline CVector3f yxz() const
{
return CVector3f(y, x, z);
}
inline CVector3f zxz() const
{
return CVector3f(z, x, z);
}
inline CVector3f xyz() const
{
return CVector3f(x, y, z);
}
inline CVector3f yyz() const
{
return CVector3f(y, y, z);
}
inline CVector3f zyz() const
{
return CVector3f(z, y, z);
}
inline CVector3f xzz() const
{
return CVector3f(x, z, z);
}
inline CVector3f yzz() const
{
return CVector3f(y, z, z);
}
inline CVector3f zzz() const
{
return CVector3f(z, z, z);
}
inline void operator += (const CVector3f &v)
{
x+=v.x;
y+=v.y;
z+=v.z;
}
inline void operator -= (const CVector3f &v)
{
x-=v.x;
y-=v.y;
z-=v.z;
}
inline void operator *= (const float v)
{
x*=v;
y*=v;
z*=v;
}
inline void operator *= (const CVector3f &v)
{
x*=v.x;
y*=v.y;
z*=v.z;
}
inline void operator /= (const float v)
{
x/=v;
y/=v;
z/=v;
}
inline void operator /= (const CVector3f &v)
{
x/=v.x;
y/=v.y;
z/=v.z;
}
inline void fromTo(const CVector3f & start, const CVector3f & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
z = ende.z - start.z;
};
void setVector(float arg0, float arg1, float arg2)
{
x = arg0;
y = arg1;
z = arg2;
}
inline void addVector(float arg0, float arg1, float arg2)
{
x += arg0;
y += arg1;
z += arg2;
}
inline void subVector(float arg0, float arg1, float arg2)
{
x -= arg0;
y -= arg1;
z -= arg2;
}
inline void neg()
{
x = -x;
y = -y;
z = -z;
}
inline float length() const
{
return sqrtf(x*x+y*y+z*z);
}
inline float normalize()
{
float r = sqrtf(x*x+y*y+z*z);
   if (r < 0.00001) return r;
x /= r;
y /= r;
z /= r;
  return r;
}
inline float dotProduct(const CVector3f &b) const
{
return (x* b.x+y* b.y+z* b.z);
}
inline int getMajorAxis() const
{
float current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
if (current < fabs(z)) {
index = 2;
current = fabs(z);
};
return index;
}
inline float distanceTo(const CVector3f &v) const;
};
inline CVector3f operator +(const CVector3f &v1, const CVector3f &v2)
{
CVector3f r;
r.x = v1.x + v2.x;
r.y = v1.y + v2.y;
r.z = v1.z + v2.z;
return r;
}
inline CVector3f operator +(const CVector3f &v1, const float v2)
{
CVector3f r;
r.x = v1.x + v2;
r.y = v1.y + v2;
r.z = v1.z + v2;
return r;
}
inline CVector3f operator -(const CVector3f &v1, const CVector3f &v2)
{
CVector3f r;
r.x = v1.x - v2.x;
r.y = v1.y - v2.y;
r.z = v1.z - v2.z;
return r;
}
inline CVector3f operator -(const CVector3f &v1, const float v2)
{
CVector3f r;
r.x = v1.x - v2;
r.y = v1.y - v2;
r.z = v1.z - v2;
return r;
}
inline CVector3f operator -(const CVector3f &v1)
{
CVector3f r;
r.x = -v1.x;
r.y = -v1.y;
r.z = -v1.z;
return r;
}
inline CVector3f operator *(const CVector3f &v1, const CVector3f &v2)
{
CVector3f r;
r.x = v1.x * v2.x;
r.y = v1.y * v2.y;
r.z = v1.z * v2.z;
return r;
}
inline CVector3f operator *(const CVector3f &v1, const float v2)
{
CVector3f r;
r.x = v1.x * v2;
r.y = v1.y * v2;
r.z = v1.z * v2;
return r;
}
inline CVector3f operator *(const float v1, const CVector3f &v2)
{
CVector3f r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
r.z = v1 * v2.z;
return r;
}
inline CVector3f operator /(const CVector3f &v1, const CVector3f &v2)
{
CVector3f r;
r.x = v1.x / v2.x;
r.y = v1.y / v2.y;
r.z = v1.z / v2.z;
return r;
}
inline CVector3f operator /(const CVector3f &v1, const float v2)
{
CVector3f r;
r.x = v1.x / v2;
r.y = v1.y / v2;
r.z = v1.z / v2;
return r;
}
inline bool operator ==(const CVector3f &v1, const CVector3f &v2)
{
bool r;
r =  v1.x == v2.x
 &&
 v1.y == v2.y
 &&
 v1.z == v2.z
;
return r;
}
inline bool operator !=(const CVector3f &v1, const CVector3f &v2)
{
bool r;
r =  v1.x != v2.x
 ||
 v1.y != v2.y
 ||
 v1.z != v2.z
;
return r;
}
inline float CVector3f::distanceTo(const CVector3f &v) const
{
return (v-*this).length();
}
inline float dot(const CVector3f &a, const CVector3f &b)
{
return (a.x* b.x+a.y* b.y+a.z* b.z);
}
inline CVector3f umin(const CVector3f &a, const CVector3f &b)
{
CVector3f r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
r.z = umin(a.z, b.z);
return r;
}
inline CVector3f umax(const CVector3f &a, const CVector3f &b)
{
CVector3f r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
r.z = umax(a.z, b.z);
return r;
}
inline CVector3f normalize(const CVector3f &a)
{
CVector3f rvec;
float r = sqrtf(a.x* a.x+a.y* a.y+a.z* a.z);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
rvec.z = a.z / r;
  return rvec;
}
inline float length(const CVector3f &a)
{
CVector3f rvec;
float r = sqrtf(a.x*a.x+a.y*a.y+a.z*a.z);
   return r;
}
inline CVector3f sign(const CVector3f &a)
{
CVector3f rvec;
rvec.x = a.x > 0.0f ? 1.0f : -1.0f;
rvec.y = a.y > 0.0f ? 1.0f : -1.0f;
rvec.z = a.z > 0.0f ? 1.0f : -1.0f;
  return rvec;
}
inline float distance(const CVector3f &a, const CVector3f &b)
{
return sqrtf(sqrf(a.x-b.x)+sqrf(a.y-b.y)+sqrf(a.z-b.z));
}
inline CVector3f linear(const CVector3f &a, const CVector3f &b, const float v)
{
CVector3f rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
rvec.z = a.z+(b.z - a.z)*v;
return rvec;
}
struct CVector4f {
float x;
float y;
float z;
float w;
//UTI_VECTOR_DECLARE_OPERATORS(CVector4f, float);
inline CVector4f(){/*x=0.0f;y=0.0f;z=0.0f;w=0.0f;*/};
inline CVector4f(const float f) {
x = f;
y = f;
z = f;
w = f;
};
inline CVector4f returnNormal() const
{
    CVector4f res = *this;
    res.normalize();
    return res;
}
inline CVector4f(const float arg0, const float arg1, const float arg2, const float arg3){
x = arg0;y = arg1;z = arg2;w = arg3;
}
inline CVector4f(const float arg0, const float arg1, const CVector2f & arg2){
x = arg0;y = arg1;z = arg2.x;w = arg2.y;
}
inline CVector4f(const float arg0, const CVector2f & arg1, const float arg2){
x = arg0;y = arg1.x;z = arg1.y;w = arg2;
}
inline CVector4f(const float arg0, const CVector3f & arg1){
x = arg0;y = arg1.x;z = arg1.y;w = arg1.z;
}
inline CVector4f(const CVector2f & arg0, const float arg1, const float arg2){
x = arg0.x;y = arg0.y;z = arg1;w = arg2;
}
inline CVector4f(const CVector2f & arg0, const CVector2f & arg1){
x = arg0.x;y = arg0.y;z = arg1.x;w = arg1.y;
}
inline CVector4f(const CVector3f & arg0, const float arg1){
x = arg0.x;y = arg0.y;z = arg0.z;w = arg1;
}
inline float &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 4);
return (&x)[p];
}
inline const float &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 4);
    return (&x)[p];
}
inline CVector2f xx() const
{
return CVector2f(x, x);
}
inline CVector2f yx() const
{
return CVector2f(y, x);
}
inline CVector2f zx() const
{
return CVector2f(z, x);
}
inline CVector2f wx() const
{
return CVector2f(w, x);
}
inline CVector2f xy() const
{
return CVector2f(x, y);
}
inline CVector2f yy() const
{
return CVector2f(y, y);
}
inline CVector2f zy() const
{
return CVector2f(z, y);
}
inline CVector2f wy() const
{
return CVector2f(w, y);
}
inline CVector2f xz() const
{
return CVector2f(x, z);
}
inline CVector2f yz() const
{
return CVector2f(y, z);
}
inline CVector2f zz() const
{
return CVector2f(z, z);
}
inline CVector2f wz() const
{
return CVector2f(w, z);
}
inline CVector2f xw() const
{
return CVector2f(x, w);
}
inline CVector2f yw() const
{
return CVector2f(y, w);
}
inline CVector2f zw() const
{
return CVector2f(z, w);
}
inline CVector2f ww() const
{
return CVector2f(w, w);
}
inline CVector3f xxx() const
{
return CVector3f(x, x, x);
}
inline CVector3f yxx() const
{
return CVector3f(y, x, x);
}
inline CVector3f zxx() const
{
return CVector3f(z, x, x);
}
inline CVector3f wxx() const
{
return CVector3f(w, x, x);
}
inline CVector3f xyx() const
{
return CVector3f(x, y, x);
}
inline CVector3f yyx() const
{
return CVector3f(y, y, x);
}
inline CVector3f zyx() const
{
return CVector3f(z, y, x);
}
inline CVector3f wyx() const
{
return CVector3f(w, y, x);
}
inline CVector3f xzx() const
{
return CVector3f(x, z, x);
}
inline CVector3f yzx() const
{
return CVector3f(y, z, x);
}
inline CVector3f zzx() const
{
return CVector3f(z, z, x);
}
inline CVector3f wzx() const
{
return CVector3f(w, z, x);
}
inline CVector3f xwx() const
{
return CVector3f(x, w, x);
}
inline CVector3f ywx() const
{
return CVector3f(y, w, x);
}
inline CVector3f zwx() const
{
return CVector3f(z, w, x);
}
inline CVector3f wwx() const
{
return CVector3f(w, w, x);
}
inline CVector3f xxy() const
{
return CVector3f(x, x, y);
}
inline CVector3f yxy() const
{
return CVector3f(y, x, y);
}
inline CVector3f zxy() const
{
return CVector3f(z, x, y);
}
inline CVector3f wxy() const
{
return CVector3f(w, x, y);
}
inline CVector3f xyy() const
{
return CVector3f(x, y, y);
}
inline CVector3f yyy() const
{
return CVector3f(y, y, y);
}
inline CVector3f zyy() const
{
return CVector3f(z, y, y);
}
inline CVector3f wyy() const
{
return CVector3f(w, y, y);
}
inline CVector3f xzy() const
{
return CVector3f(x, z, y);
}
inline CVector3f yzy() const
{
return CVector3f(y, z, y);
}
inline CVector3f zzy() const
{
return CVector3f(z, z, y);
}
inline CVector3f wzy() const
{
return CVector3f(w, z, y);
}
inline CVector3f xwy() const
{
return CVector3f(x, w, y);
}
inline CVector3f ywy() const
{
return CVector3f(y, w, y);
}
inline CVector3f zwy() const
{
return CVector3f(z, w, y);
}
inline CVector3f wwy() const
{
return CVector3f(w, w, y);
}
inline CVector3f xxz() const
{
return CVector3f(x, x, z);
}
inline CVector3f yxz() const
{
return CVector3f(y, x, z);
}
inline CVector3f zxz() const
{
return CVector3f(z, x, z);
}
inline CVector3f wxz() const
{
return CVector3f(w, x, z);
}
inline CVector3f xyz() const
{
return CVector3f(x, y, z);
}
inline CVector3f yyz() const
{
return CVector3f(y, y, z);
}
inline CVector3f zyz() const
{
return CVector3f(z, y, z);
}
inline CVector3f wyz() const
{
return CVector3f(w, y, z);
}
inline CVector3f xzz() const
{
return CVector3f(x, z, z);
}
inline CVector3f yzz() const
{
return CVector3f(y, z, z);
}
inline CVector3f zzz() const
{
return CVector3f(z, z, z);
}
inline CVector3f wzz() const
{
return CVector3f(w, z, z);
}
inline CVector3f xwz() const
{
return CVector3f(x, w, z);
}
inline CVector3f ywz() const
{
return CVector3f(y, w, z);
}
inline CVector3f zwz() const
{
return CVector3f(z, w, z);
}
inline CVector3f wwz() const
{
return CVector3f(w, w, z);
}
inline CVector3f xxw() const
{
return CVector3f(x, x, w);
}
inline CVector3f yxw() const
{
return CVector3f(y, x, w);
}
inline CVector3f zxw() const
{
return CVector3f(z, x, w);
}
inline CVector3f wxw() const
{
return CVector3f(w, x, w);
}
inline CVector3f xyw() const
{
return CVector3f(x, y, w);
}
inline CVector3f yyw() const
{
return CVector3f(y, y, w);
}
inline CVector3f zyw() const
{
return CVector3f(z, y, w);
}
inline CVector3f wyw() const
{
return CVector3f(w, y, w);
}
inline CVector3f xzw() const
{
return CVector3f(x, z, w);
}
inline CVector3f yzw() const
{
return CVector3f(y, z, w);
}
inline CVector3f zzw() const
{
return CVector3f(z, z, w);
}
inline CVector3f wzw() const
{
return CVector3f(w, z, w);
}
inline CVector3f xww() const
{
return CVector3f(x, w, w);
}
inline CVector3f yww() const
{
return CVector3f(y, w, w);
}
inline CVector3f zww() const
{
return CVector3f(z, w, w);
}
inline CVector3f www() const
{
return CVector3f(w, w, w);
}
inline CVector4f xxxx() const
{
return CVector4f(x, x, x, x);
}
inline CVector4f yxxx() const
{
return CVector4f(y, x, x, x);
}
inline CVector4f zxxx() const
{
return CVector4f(z, x, x, x);
}
inline CVector4f wxxx() const
{
return CVector4f(w, x, x, x);
}
inline CVector4f xyxx() const
{
return CVector4f(x, y, x, x);
}
inline CVector4f yyxx() const
{
return CVector4f(y, y, x, x);
}
inline CVector4f zyxx() const
{
return CVector4f(z, y, x, x);
}
inline CVector4f wyxx() const
{
return CVector4f(w, y, x, x);
}
inline CVector4f xzxx() const
{
return CVector4f(x, z, x, x);
}
inline CVector4f yzxx() const
{
return CVector4f(y, z, x, x);
}
inline CVector4f zzxx() const
{
return CVector4f(z, z, x, x);
}
inline CVector4f wzxx() const
{
return CVector4f(w, z, x, x);
}
inline CVector4f xwxx() const
{
return CVector4f(x, w, x, x);
}
inline CVector4f ywxx() const
{
return CVector4f(y, w, x, x);
}
inline CVector4f zwxx() const
{
return CVector4f(z, w, x, x);
}
inline CVector4f wwxx() const
{
return CVector4f(w, w, x, x);
}
inline CVector4f xxyx() const
{
return CVector4f(x, x, y, x);
}
inline CVector4f yxyx() const
{
return CVector4f(y, x, y, x);
}
inline CVector4f zxyx() const
{
return CVector4f(z, x, y, x);
}
inline CVector4f wxyx() const
{
return CVector4f(w, x, y, x);
}
inline CVector4f xyyx() const
{
return CVector4f(x, y, y, x);
}
inline CVector4f yyyx() const
{
return CVector4f(y, y, y, x);
}
inline CVector4f zyyx() const
{
return CVector4f(z, y, y, x);
}
inline CVector4f wyyx() const
{
return CVector4f(w, y, y, x);
}
inline CVector4f xzyx() const
{
return CVector4f(x, z, y, x);
}
inline CVector4f yzyx() const
{
return CVector4f(y, z, y, x);
}
inline CVector4f zzyx() const
{
return CVector4f(z, z, y, x);
}
inline CVector4f wzyx() const
{
return CVector4f(w, z, y, x);
}
inline CVector4f xwyx() const
{
return CVector4f(x, w, y, x);
}
inline CVector4f ywyx() const
{
return CVector4f(y, w, y, x);
}
inline CVector4f zwyx() const
{
return CVector4f(z, w, y, x);
}
inline CVector4f wwyx() const
{
return CVector4f(w, w, y, x);
}
inline CVector4f xxzx() const
{
return CVector4f(x, x, z, x);
}
inline CVector4f yxzx() const
{
return CVector4f(y, x, z, x);
}
inline CVector4f zxzx() const
{
return CVector4f(z, x, z, x);
}
inline CVector4f wxzx() const
{
return CVector4f(w, x, z, x);
}
inline CVector4f xyzx() const
{
return CVector4f(x, y, z, x);
}
inline CVector4f yyzx() const
{
return CVector4f(y, y, z, x);
}
inline CVector4f zyzx() const
{
return CVector4f(z, y, z, x);
}
inline CVector4f wyzx() const
{
return CVector4f(w, y, z, x);
}
inline CVector4f xzzx() const
{
return CVector4f(x, z, z, x);
}
inline CVector4f yzzx() const
{
return CVector4f(y, z, z, x);
}
inline CVector4f zzzx() const
{
return CVector4f(z, z, z, x);
}
inline CVector4f wzzx() const
{
return CVector4f(w, z, z, x);
}
inline CVector4f xwzx() const
{
return CVector4f(x, w, z, x);
}
inline CVector4f ywzx() const
{
return CVector4f(y, w, z, x);
}
inline CVector4f zwzx() const
{
return CVector4f(z, w, z, x);
}
inline CVector4f wwzx() const
{
return CVector4f(w, w, z, x);
}
inline CVector4f xxwx() const
{
return CVector4f(x, x, w, x);
}
inline CVector4f yxwx() const
{
return CVector4f(y, x, w, x);
}
inline CVector4f zxwx() const
{
return CVector4f(z, x, w, x);
}
inline CVector4f wxwx() const
{
return CVector4f(w, x, w, x);
}
inline CVector4f xywx() const
{
return CVector4f(x, y, w, x);
}
inline CVector4f yywx() const
{
return CVector4f(y, y, w, x);
}
inline CVector4f zywx() const
{
return CVector4f(z, y, w, x);
}
inline CVector4f wywx() const
{
return CVector4f(w, y, w, x);
}
inline CVector4f xzwx() const
{
return CVector4f(x, z, w, x);
}
inline CVector4f yzwx() const
{
return CVector4f(y, z, w, x);
}
inline CVector4f zzwx() const
{
return CVector4f(z, z, w, x);
}
inline CVector4f wzwx() const
{
return CVector4f(w, z, w, x);
}
inline CVector4f xwwx() const
{
return CVector4f(x, w, w, x);
}
inline CVector4f ywwx() const
{
return CVector4f(y, w, w, x);
}
inline CVector4f zwwx() const
{
return CVector4f(z, w, w, x);
}
inline CVector4f wwwx() const
{
return CVector4f(w, w, w, x);
}
inline CVector4f xxxy() const
{
return CVector4f(x, x, x, y);
}
inline CVector4f yxxy() const
{
return CVector4f(y, x, x, y);
}
inline CVector4f zxxy() const
{
return CVector4f(z, x, x, y);
}
inline CVector4f wxxy() const
{
return CVector4f(w, x, x, y);
}
inline CVector4f xyxy() const
{
return CVector4f(x, y, x, y);
}
inline CVector4f yyxy() const
{
return CVector4f(y, y, x, y);
}
inline CVector4f zyxy() const
{
return CVector4f(z, y, x, y);
}
inline CVector4f wyxy() const
{
return CVector4f(w, y, x, y);
}
inline CVector4f xzxy() const
{
return CVector4f(x, z, x, y);
}
inline CVector4f yzxy() const
{
return CVector4f(y, z, x, y);
}
inline CVector4f zzxy() const
{
return CVector4f(z, z, x, y);
}
inline CVector4f wzxy() const
{
return CVector4f(w, z, x, y);
}
inline CVector4f xwxy() const
{
return CVector4f(x, w, x, y);
}
inline CVector4f ywxy() const
{
return CVector4f(y, w, x, y);
}
inline CVector4f zwxy() const
{
return CVector4f(z, w, x, y);
}
inline CVector4f wwxy() const
{
return CVector4f(w, w, x, y);
}
inline CVector4f xxyy() const
{
return CVector4f(x, x, y, y);
}
inline CVector4f yxyy() const
{
return CVector4f(y, x, y, y);
}
inline CVector4f zxyy() const
{
return CVector4f(z, x, y, y);
}
inline CVector4f wxyy() const
{
return CVector4f(w, x, y, y);
}
inline CVector4f xyyy() const
{
return CVector4f(x, y, y, y);
}
inline CVector4f yyyy() const
{
return CVector4f(y, y, y, y);
}
inline CVector4f zyyy() const
{
return CVector4f(z, y, y, y);
}
inline CVector4f wyyy() const
{
return CVector4f(w, y, y, y);
}
inline CVector4f xzyy() const
{
return CVector4f(x, z, y, y);
}
inline CVector4f yzyy() const
{
return CVector4f(y, z, y, y);
}
inline CVector4f zzyy() const
{
return CVector4f(z, z, y, y);
}
inline CVector4f wzyy() const
{
return CVector4f(w, z, y, y);
}
inline CVector4f xwyy() const
{
return CVector4f(x, w, y, y);
}
inline CVector4f ywyy() const
{
return CVector4f(y, w, y, y);
}
inline CVector4f zwyy() const
{
return CVector4f(z, w, y, y);
}
inline CVector4f wwyy() const
{
return CVector4f(w, w, y, y);
}
inline CVector4f xxzy() const
{
return CVector4f(x, x, z, y);
}
inline CVector4f yxzy() const
{
return CVector4f(y, x, z, y);
}
inline CVector4f zxzy() const
{
return CVector4f(z, x, z, y);
}
inline CVector4f wxzy() const
{
return CVector4f(w, x, z, y);
}
inline CVector4f xyzy() const
{
return CVector4f(x, y, z, y);
}
inline CVector4f yyzy() const
{
return CVector4f(y, y, z, y);
}
inline CVector4f zyzy() const
{
return CVector4f(z, y, z, y);
}
inline CVector4f wyzy() const
{
return CVector4f(w, y, z, y);
}
inline CVector4f xzzy() const
{
return CVector4f(x, z, z, y);
}
inline CVector4f yzzy() const
{
return CVector4f(y, z, z, y);
}
inline CVector4f zzzy() const
{
return CVector4f(z, z, z, y);
}
inline CVector4f wzzy() const
{
return CVector4f(w, z, z, y);
}
inline CVector4f xwzy() const
{
return CVector4f(x, w, z, y);
}
inline CVector4f ywzy() const
{
return CVector4f(y, w, z, y);
}
inline CVector4f zwzy() const
{
return CVector4f(z, w, z, y);
}
inline CVector4f wwzy() const
{
return CVector4f(w, w, z, y);
}
inline CVector4f xxwy() const
{
return CVector4f(x, x, w, y);
}
inline CVector4f yxwy() const
{
return CVector4f(y, x, w, y);
}
inline CVector4f zxwy() const
{
return CVector4f(z, x, w, y);
}
inline CVector4f wxwy() const
{
return CVector4f(w, x, w, y);
}
inline CVector4f xywy() const
{
return CVector4f(x, y, w, y);
}
inline CVector4f yywy() const
{
return CVector4f(y, y, w, y);
}
inline CVector4f zywy() const
{
return CVector4f(z, y, w, y);
}
inline CVector4f wywy() const
{
return CVector4f(w, y, w, y);
}
inline CVector4f xzwy() const
{
return CVector4f(x, z, w, y);
}
inline CVector4f yzwy() const
{
return CVector4f(y, z, w, y);
}
inline CVector4f zzwy() const
{
return CVector4f(z, z, w, y);
}
inline CVector4f wzwy() const
{
return CVector4f(w, z, w, y);
}
inline CVector4f xwwy() const
{
return CVector4f(x, w, w, y);
}
inline CVector4f ywwy() const
{
return CVector4f(y, w, w, y);
}
inline CVector4f zwwy() const
{
return CVector4f(z, w, w, y);
}
inline CVector4f wwwy() const
{
return CVector4f(w, w, w, y);
}
inline CVector4f xxxz() const
{
return CVector4f(x, x, x, z);
}
inline CVector4f yxxz() const
{
return CVector4f(y, x, x, z);
}
inline CVector4f zxxz() const
{
return CVector4f(z, x, x, z);
}
inline CVector4f wxxz() const
{
return CVector4f(w, x, x, z);
}
inline CVector4f xyxz() const
{
return CVector4f(x, y, x, z);
}
inline CVector4f yyxz() const
{
return CVector4f(y, y, x, z);
}
inline CVector4f zyxz() const
{
return CVector4f(z, y, x, z);
}
inline CVector4f wyxz() const
{
return CVector4f(w, y, x, z);
}
inline CVector4f xzxz() const
{
return CVector4f(x, z, x, z);
}
inline CVector4f yzxz() const
{
return CVector4f(y, z, x, z);
}
inline CVector4f zzxz() const
{
return CVector4f(z, z, x, z);
}
inline CVector4f wzxz() const
{
return CVector4f(w, z, x, z);
}
inline CVector4f xwxz() const
{
return CVector4f(x, w, x, z);
}
inline CVector4f ywxz() const
{
return CVector4f(y, w, x, z);
}
inline CVector4f zwxz() const
{
return CVector4f(z, w, x, z);
}
inline CVector4f wwxz() const
{
return CVector4f(w, w, x, z);
}
inline CVector4f xxyz() const
{
return CVector4f(x, x, y, z);
}
inline CVector4f yxyz() const
{
return CVector4f(y, x, y, z);
}
inline CVector4f zxyz() const
{
return CVector4f(z, x, y, z);
}
inline CVector4f wxyz() const
{
return CVector4f(w, x, y, z);
}
inline CVector4f xyyz() const
{
return CVector4f(x, y, y, z);
}
inline CVector4f yyyz() const
{
return CVector4f(y, y, y, z);
}
inline CVector4f zyyz() const
{
return CVector4f(z, y, y, z);
}
inline CVector4f wyyz() const
{
return CVector4f(w, y, y, z);
}
inline CVector4f xzyz() const
{
return CVector4f(x, z, y, z);
}
inline CVector4f yzyz() const
{
return CVector4f(y, z, y, z);
}
inline CVector4f zzyz() const
{
return CVector4f(z, z, y, z);
}
inline CVector4f wzyz() const
{
return CVector4f(w, z, y, z);
}
inline CVector4f xwyz() const
{
return CVector4f(x, w, y, z);
}
inline CVector4f ywyz() const
{
return CVector4f(y, w, y, z);
}
inline CVector4f zwyz() const
{
return CVector4f(z, w, y, z);
}
inline CVector4f wwyz() const
{
return CVector4f(w, w, y, z);
}
inline CVector4f xxzz() const
{
return CVector4f(x, x, z, z);
}
inline CVector4f yxzz() const
{
return CVector4f(y, x, z, z);
}
inline CVector4f zxzz() const
{
return CVector4f(z, x, z, z);
}
inline CVector4f wxzz() const
{
return CVector4f(w, x, z, z);
}
inline CVector4f xyzz() const
{
return CVector4f(x, y, z, z);
}
inline CVector4f yyzz() const
{
return CVector4f(y, y, z, z);
}
inline CVector4f zyzz() const
{
return CVector4f(z, y, z, z);
}
inline CVector4f wyzz() const
{
return CVector4f(w, y, z, z);
}
inline CVector4f xzzz() const
{
return CVector4f(x, z, z, z);
}
inline CVector4f yzzz() const
{
return CVector4f(y, z, z, z);
}
inline CVector4f zzzz() const
{
return CVector4f(z, z, z, z);
}
inline CVector4f wzzz() const
{
return CVector4f(w, z, z, z);
}
inline CVector4f xwzz() const
{
return CVector4f(x, w, z, z);
}
inline CVector4f ywzz() const
{
return CVector4f(y, w, z, z);
}
inline CVector4f zwzz() const
{
return CVector4f(z, w, z, z);
}
inline CVector4f wwzz() const
{
return CVector4f(w, w, z, z);
}
inline CVector4f xxwz() const
{
return CVector4f(x, x, w, z);
}
inline CVector4f yxwz() const
{
return CVector4f(y, x, w, z);
}
inline CVector4f zxwz() const
{
return CVector4f(z, x, w, z);
}
inline CVector4f wxwz() const
{
return CVector4f(w, x, w, z);
}
inline CVector4f xywz() const
{
return CVector4f(x, y, w, z);
}
inline CVector4f yywz() const
{
return CVector4f(y, y, w, z);
}
inline CVector4f zywz() const
{
return CVector4f(z, y, w, z);
}
inline CVector4f wywz() const
{
return CVector4f(w, y, w, z);
}
inline CVector4f xzwz() const
{
return CVector4f(x, z, w, z);
}
inline CVector4f yzwz() const
{
return CVector4f(y, z, w, z);
}
inline CVector4f zzwz() const
{
return CVector4f(z, z, w, z);
}
inline CVector4f wzwz() const
{
return CVector4f(w, z, w, z);
}
inline CVector4f xwwz() const
{
return CVector4f(x, w, w, z);
}
inline CVector4f ywwz() const
{
return CVector4f(y, w, w, z);
}
inline CVector4f zwwz() const
{
return CVector4f(z, w, w, z);
}
inline CVector4f wwwz() const
{
return CVector4f(w, w, w, z);
}
inline CVector4f xxxw() const
{
return CVector4f(x, x, x, w);
}
inline CVector4f yxxw() const
{
return CVector4f(y, x, x, w);
}
inline CVector4f zxxw() const
{
return CVector4f(z, x, x, w);
}
inline CVector4f wxxw() const
{
return CVector4f(w, x, x, w);
}
inline CVector4f xyxw() const
{
return CVector4f(x, y, x, w);
}
inline CVector4f yyxw() const
{
return CVector4f(y, y, x, w);
}
inline CVector4f zyxw() const
{
return CVector4f(z, y, x, w);
}
inline CVector4f wyxw() const
{
return CVector4f(w, y, x, w);
}
inline CVector4f xzxw() const
{
return CVector4f(x, z, x, w);
}
inline CVector4f yzxw() const
{
return CVector4f(y, z, x, w);
}
inline CVector4f zzxw() const
{
return CVector4f(z, z, x, w);
}
inline CVector4f wzxw() const
{
return CVector4f(w, z, x, w);
}
inline CVector4f xwxw() const
{
return CVector4f(x, w, x, w);
}
inline CVector4f ywxw() const
{
return CVector4f(y, w, x, w);
}
inline CVector4f zwxw() const
{
return CVector4f(z, w, x, w);
}
inline CVector4f wwxw() const
{
return CVector4f(w, w, x, w);
}
inline CVector4f xxyw() const
{
return CVector4f(x, x, y, w);
}
inline CVector4f yxyw() const
{
return CVector4f(y, x, y, w);
}
inline CVector4f zxyw() const
{
return CVector4f(z, x, y, w);
}
inline CVector4f wxyw() const
{
return CVector4f(w, x, y, w);
}
inline CVector4f xyyw() const
{
return CVector4f(x, y, y, w);
}
inline CVector4f yyyw() const
{
return CVector4f(y, y, y, w);
}
inline CVector4f zyyw() const
{
return CVector4f(z, y, y, w);
}
inline CVector4f wyyw() const
{
return CVector4f(w, y, y, w);
}
inline CVector4f xzyw() const
{
return CVector4f(x, z, y, w);
}
inline CVector4f yzyw() const
{
return CVector4f(y, z, y, w);
}
inline CVector4f zzyw() const
{
return CVector4f(z, z, y, w);
}
inline CVector4f wzyw() const
{
return CVector4f(w, z, y, w);
}
inline CVector4f xwyw() const
{
return CVector4f(x, w, y, w);
}
inline CVector4f ywyw() const
{
return CVector4f(y, w, y, w);
}
inline CVector4f zwyw() const
{
return CVector4f(z, w, y, w);
}
inline CVector4f wwyw() const
{
return CVector4f(w, w, y, w);
}
inline CVector4f xxzw() const
{
return CVector4f(x, x, z, w);
}
inline CVector4f yxzw() const
{
return CVector4f(y, x, z, w);
}
inline CVector4f zxzw() const
{
return CVector4f(z, x, z, w);
}
inline CVector4f wxzw() const
{
return CVector4f(w, x, z, w);
}
inline CVector4f xyzw() const
{
return CVector4f(x, y, z, w);
}
inline CVector4f yyzw() const
{
return CVector4f(y, y, z, w);
}
inline CVector4f zyzw() const
{
return CVector4f(z, y, z, w);
}
inline CVector4f wyzw() const
{
return CVector4f(w, y, z, w);
}
inline CVector4f xzzw() const
{
return CVector4f(x, z, z, w);
}
inline CVector4f yzzw() const
{
return CVector4f(y, z, z, w);
}
inline CVector4f zzzw() const
{
return CVector4f(z, z, z, w);
}
inline CVector4f wzzw() const
{
return CVector4f(w, z, z, w);
}
inline CVector4f xwzw() const
{
return CVector4f(x, w, z, w);
}
inline CVector4f ywzw() const
{
return CVector4f(y, w, z, w);
}
inline CVector4f zwzw() const
{
return CVector4f(z, w, z, w);
}
inline CVector4f wwzw() const
{
return CVector4f(w, w, z, w);
}
inline CVector4f xxww() const
{
return CVector4f(x, x, w, w);
}
inline CVector4f yxww() const
{
return CVector4f(y, x, w, w);
}
inline CVector4f zxww() const
{
return CVector4f(z, x, w, w);
}
inline CVector4f wxww() const
{
return CVector4f(w, x, w, w);
}
inline CVector4f xyww() const
{
return CVector4f(x, y, w, w);
}
inline CVector4f yyww() const
{
return CVector4f(y, y, w, w);
}
inline CVector4f zyww() const
{
return CVector4f(z, y, w, w);
}
inline CVector4f wyww() const
{
return CVector4f(w, y, w, w);
}
inline CVector4f xzww() const
{
return CVector4f(x, z, w, w);
}
inline CVector4f yzww() const
{
return CVector4f(y, z, w, w);
}
inline CVector4f zzww() const
{
return CVector4f(z, z, w, w);
}
inline CVector4f wzww() const
{
return CVector4f(w, z, w, w);
}
inline CVector4f xwww() const
{
return CVector4f(x, w, w, w);
}
inline CVector4f ywww() const
{
return CVector4f(y, w, w, w);
}
inline CVector4f zwww() const
{
return CVector4f(z, w, w, w);
}
inline CVector4f wwww() const
{
return CVector4f(w, w, w, w);
}
inline void operator += (const CVector4f &v)
{
x+=v.x;
y+=v.y;
z+=v.z;
w+=v.w;
}
inline void operator -= (const CVector4f &v)
{
x-=v.x;
y-=v.y;
z-=v.z;
w-=v.w;
}
inline void operator *= (const float v)
{
x*=v;
y*=v;
z*=v;
w*=v;
}
inline void operator *= (const CVector4f &v)
{
x*=v.x;
y*=v.y;
z*=v.z;
w*=v.w;
}
inline void operator /= (const float v)
{
x/=v;
y/=v;
z/=v;
w/=v;
}
inline void operator /= (const CVector4f &v)
{
x/=v.x;
y/=v.y;
z/=v.z;
w/=v.w;
}
inline void fromTo(const CVector4f & start, const CVector4f & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
z = ende.z - start.z;
w = ende.w - start.w;
};
void setVector(float arg0, float arg1, float arg2, float arg3)
{
x = arg0;
y = arg1;
z = arg2;
w = arg3;
}
inline void addVector(float arg0, float arg1, float arg2, float arg3)
{
x += arg0;
y += arg1;
z += arg2;
w += arg3;
}
inline void subVector(float arg0, float arg1, float arg2, float arg3)
{
x -= arg0;
y -= arg1;
z -= arg2;
w -= arg3;
}
inline void neg()
{
x = -x;
y = -y;
z = -z;
w = -w;
}
inline float length() const
{
return sqrtf(x*x+y*y+z*z+w*w);
}
inline float normalize()
{
float r = sqrtf(x*x+y*y+z*z+w*w);
   if (r < 0.00001) return r;
x /= r;
y /= r;
z /= r;
w /= r;
  return r;
}
inline float dotProduct(const CVector4f &b) const
{
return (x* b.x+y* b.y+z* b.z+w* b.w);
}
inline int getMajorAxis() const
{
float current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
if (current < fabs(z)) {
index = 2;
current = fabs(z);
};
if (current < fabs(w)) {
index = 3;
current = fabs(w);
};
return index;
}
inline float distanceTo(const CVector4f &v) const;
};
inline CVector4f operator +(const CVector4f &v1, const CVector4f &v2)
{
CVector4f r;
r.x = v1.x + v2.x;
r.y = v1.y + v2.y;
r.z = v1.z + v2.z;
r.w = v1.w + v2.w;
return r;
}
inline CVector4f operator +(const CVector4f &v1, const float v2)
{
CVector4f r;
r.x = v1.x + v2;
r.y = v1.y + v2;
r.z = v1.z + v2;
r.w = v1.w + v2;
return r;
}
inline CVector4f operator -(const CVector4f &v1, const CVector4f &v2)
{
CVector4f r;
r.x = v1.x - v2.x;
r.y = v1.y - v2.y;
r.z = v1.z - v2.z;
r.w = v1.w - v2.w;
return r;
}
inline CVector4f operator -(const CVector4f &v1, const float v2)
{
CVector4f r;
r.x = v1.x - v2;
r.y = v1.y - v2;
r.z = v1.z - v2;
r.w = v1.w - v2;
return r;
}
inline CVector4f operator -(const CVector4f &v1)
{
CVector4f r;
r.x = -v1.x;
r.y = -v1.y;
r.z = -v1.z;
r.w = -v1.w;
return r;
}
inline CVector4f operator *(const CVector4f &v1, const CVector4f &v2)
{
CVector4f r;
r.x = v1.x * v2.x;
r.y = v1.y * v2.y;
r.z = v1.z * v2.z;
r.w = v1.w * v2.w;
return r;
}
inline CVector4f operator *(const CVector4f &v1, const float v2)
{
CVector4f r;
r.x = v1.x * v2;
r.y = v1.y * v2;
r.z = v1.z * v2;
r.w = v1.w * v2;
return r;
}
inline CVector4f operator *(const float v1, const CVector4f &v2)
{
CVector4f r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
r.z = v1 * v2.z;
r.w = v1 * v2.w;
return r;
}
inline CVector4f operator /(const CVector4f &v1, const CVector4f &v2)
{
CVector4f r;
r.x = v1.x / v2.x;
r.y = v1.y / v2.y;
r.z = v1.z / v2.z;
r.w = v1.w / v2.w;
return r;
}
inline CVector4f operator /(const CVector4f &v1, const float v2)
{
CVector4f r;
r.x = v1.x / v2;
r.y = v1.y / v2;
r.z = v1.z / v2;
r.w = v1.w / v2;
return r;
}
inline bool operator ==(const CVector4f &v1, const CVector4f &v2)
{
bool r;
r =  v1.x == v2.x
 &&
 v1.y == v2.y
 &&
 v1.z == v2.z
 &&
 v1.w == v2.w
;
return r;
}
inline bool operator !=(const CVector4f &v1, const CVector4f &v2)
{
bool r;
r =  v1.x != v2.x
 ||
 v1.y != v2.y
 ||
 v1.z != v2.z
 ||
 v1.w != v2.w
;
return r;
}
inline CVector4 operator /(const double v1, const CVector4 &v2)
{
CVector4 r;
r.x = v1 / v2.x;
r.y = v1 / v2.y;
r.z = v1 / v2.z;
r.w = v1 / v2.w;
return r;
}

inline float CVector4f::distanceTo(const CVector4f &v) const
{
return (v-*this).length();
}
inline float dot(const CVector4f &a, const CVector4f &b)
{
return (a.x* b.x+a.y* b.y+a.z* b.z+a.w* b.w);
}
inline CVector4f umin(const CVector4f &a, const CVector4f &b)
{
CVector4f r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
r.z = umin(a.z, b.z);
r.w = umin(a.w, b.w);
return r;
}
inline CVector4f umax(const CVector4f &a, const CVector4f &b)
{
CVector4f r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
r.z = umax(a.z, b.z);
r.w = umax(a.w, b.w);
return r;
}
inline CVector4f normalize(const CVector4f &a)
{
CVector4f rvec;
float r = sqrtf(a.x* a.x+a.y* a.y+a.z* a.z+a.w* a.w);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
rvec.z = a.z / r;
rvec.w = a.w / r;
  return rvec;
}
inline float length(const CVector4f &a)
{
CVector4f rvec;
float r = sqrtf(a.x*a.x+a.y*a.y+a.z*a.z+a.w*a.w);
   return r;
}
inline CVector4f sign(const CVector4f &a)
{
CVector4f rvec;
rvec.x = a.x > 0.0f ? 1.0f : -1.0f;
rvec.y = a.y > 0.0f ? 1.0f : -1.0f;
rvec.z = a.z > 0.0f ? 1.0f : -1.0f;
rvec.w = a.w > 0.0f ? 1.0f : -1.0f;
  return rvec;
}
inline float distance(const CVector4f &a, const CVector4f &b)
{
return sqrtf(sqrf(a.x-b.x)+sqrf(a.y-b.y)+sqrf(a.z-b.z)+sqrf(a.w-b.w));
}
inline CVector4f linear(const CVector4f &a, const CVector4f &b, const float v)
{
CVector4f rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
rvec.z = a.z+(b.z - a.z)*v;
rvec.w = a.w+(b.w - a.w)*v;
return rvec;
}


#endif

