/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "utixml.h"
#include "locale"
#include "utisystem.h"

#include "uti64coder.h"
#include "UtiArchiv.h"

#include "easyprofiler.h"
#ifdef _MSC_VER
#define snprintf _snprintf
#endif

using namespace UtiFramework;

TiXmlDocument* uti_createDocument()
{
    TiXmlDocument* l_document = new TiXmlDocument();
    TiXmlDeclaration * decl = new TiXmlDeclaration( "1.0", "", "" );
	TiXmlElement * element = new TiXmlElement( "root" );

	l_document->LinkEndChild( decl );
	l_document->LinkEndChild( element );
	return l_document;
}

CXMLDocument::CXMLDocument()
{
    m_document = uti_createDocument();

    m_root = new CXMLNode(NULL, this, m_document->RootElement());
}

CXMLDocument::~CXMLDocument()
{
    //dtor
    delete m_document;
    //delete m_root;
}

int CXMLDocument::saveToStream(CStream &Stream)
{
   return 0;
}

int CXMLDocument::loadFromStream(CStream &Stream)
{

return 0;
}





int CXMLDocument::saveToFile(const CUtiString& name)
{
   /*CFileStream* str = new CFileStream(name, S_WRITE);
    if (str == NULL || str->fail())
    {
        if (str) delete str;
        return UTI_WRITEERROR;
    }
    SaveToStream(*str);
    delete str;*/

    std::string fname = converttomulti(name);
    if (m_document->SaveFile(fname)) {
    return UTI_OK;
    }
    return UTI_WRITEERROR;
}


int CXMLDocument::loadFromFile(const CUtiString& name)
{
    CUtiString realname = UtiGetResolvedFileName(name);
    std::string fname = converttomulti(realname);
    if (!FileExists(realname)) {
        return UTI_READERROR;
    }
    if (m_document->LoadFile(fname)) {
        //delete m_root;
        if (m_document->RootElement() == NULL){
            TiXmlElement * element = new TiXmlElement( "root" );
            m_document->LinkEndChild( element );
        }
        m_root = new CXMLNode(NULL, this, m_document->RootElement());
    return UTI_OK;
	}
	else {
		OnErrorDisplayMessage(false, UTISTR("The XML file '") + name + UTISTR("' could not be loaded!"));
	}
    return UTI_READERROR;
}


CXMLNode::CXMLNode(CXNode* parent, CXDocument* doc, TiXmlElement* element) : CXNode(parent, doc)
{
    m_element = element;
    m_name = converttowide(m_element->Value());
    TiXmlElement* el = m_element->FirstChildElement();
    while (el)
    {
        m_children.push_back(new CXMLNode(this, m_doc, el));
        el = el->NextSiblingElement();
    }
}

CUtiString CXMLNode::getText()
{
  CUtiString res;
  const char* str = m_element->GetText();
  return converttowide(str);
};

void CXMLNode::setName(const CUtiString& name)
{
    assert(isBezeichner(name));
    std::string l_name = converttomulti(name.c_str());
    m_element->SetValue(l_name);
    m_name = name;
}

CUtiString CXMLNode::getName() const
{
   return m_name;//converttowide(m_element->Value());

}
void CXMLNode::setPropBool(const CUtiString& name, bool b)
{
    assert(isBezeichner(name));
   std::string l_name = converttomulti(name.c_str());
   if (b) {
      m_element->SetAttribute(l_name, "1");
   } else {
       m_element->SetAttribute(l_name, "0");
   }

}
bool CXMLNode::getPropBool(const CUtiString& name, bool std) const
{
    assert(isBezeichner(name));
   std::string l_name = converttomulti(name.c_str());
   const char* l_value = m_element->Attribute(l_name.c_str());
   if (!l_value) return std;
   if (strcmp(l_value, "0")==0) return false;
   if (strcmp(l_value, "1")==0) return true;
   if (strcmp(l_value, "true")==0) return true;
   if (strcmp(l_value, "false")==0) return false;
   if (strcmp(l_value, "Y")==0) return true;
   if (strcmp(l_value, "N")==0) return false;
   if (strcmp(l_value, "y")==0) return true;
   if (strcmp(l_value, "n")==0) return false;
   {

      CUtiString error(UTEXT("CXMLNode::getPropBool: the value '"));
      error += converttowide(l_value) + UTEXT("' can not be converted to a bool value. Use 0, 1, true, false, Y, N, y, n instead.");

      throw new CUtiXMLException(error);
   }
   return false;

}

void CXMLNode::setPropInt(const CUtiString& name, int b)
{
    assert(isBezeichner(name));
   std::string l_name = converttomulti(name.c_str());
   char intvalue[42];

   snprintf(intvalue, 40, "%i", b);

   m_element->SetAttribute(l_name, intvalue);
}
int CXMLNode::getPropInt(const CUtiString& name, int std) const
{
    assert(isBezeichner(name));
    std::string l_name = converttomulti(name.c_str());
   const char* l_value = m_element->Attribute(l_name.c_str());
   if (!l_value) return std;
   char* l_tail;
   long int value = strtol(l_value, &l_tail, 10);
   if (l_value == l_tail) {

      CUtiString error(UTEXT("CXMLNode::getPropInt: the value '"));
      error += converttowide(l_value) + UTEXT("' can not be converted to a int value.");

      throw new CUtiXMLException(error);

   }
   return value;


}
void CXMLNode::setPropFloat(const CUtiString& name, float b)
{assert(isBezeichner(name));
       setPropDouble(name, b);

}
float CXMLNode::getPropFloat(const CUtiString& name, float std) const
{assert(isBezeichner(name));
   return getPropDouble(name, std);

}
void CXMLNode::setPropDouble(const CUtiString& name, double b)
{
assert(isBezeichner(name));
    std::string l_name = converttomulti(name.c_str());
   char floatvalue[42];
   snprintf(floatvalue, 40, "%f", b);
   m_element->SetAttribute(l_name, floatvalue);

}
double CXMLNode::getPropDouble(const CUtiString& name, double std) const
{assert(isBezeichner(name));
    std::string l_name = converttomulti(name.c_str());
   const char* l_value = m_element->Attribute(l_name.c_str());
   if (!l_value) return std;
   char* l_tail;
   double value = strtod(l_value, &l_tail);
   if (l_value == l_tail) {

      CUtiString error(UTEXT("CXMLNode::getPropDouble: the value '"));
      error += converttowide(l_value) + UTEXT("' can not be converted to a int value.");
      throw new CUtiXMLException(error);

   }
   return value;

}
void CXMLNode::setPropCode(const CUtiString& name, const SCode &b)
{
assert(isBezeichner(name));
   int size = CUti64Coder::getDataLength(sizeof(SCode));
   char c[64];
   utiassert(size < 64);

   CUti64Coder::writeData(&b, sizeof(SCode), c, size);
   c[size] = 0;
   std::string l_name = converttomulti(name.c_str());
    m_element->SetAttribute(l_name, c);


}
SCode CXMLNode::getPropCode(const CUtiString& name, const SCode &std) const
{assert(isBezeichner(name));

     std::string l_name = converttomulti(name.c_str());
   const char* l_value = m_element->Attribute(l_name.c_str());
   if (!l_value) return std;
   int l_length = strlen(l_value);
   if (l_length != CUti64Coder::getDataLength(sizeof(SCode)))
   {
       throw new CUtiXMLException(UTEXT("CXMLNode::getPropCode: Error in Code"));
   }
   SCode result;
   CUti64Coder::readData(l_value, l_length, &result, sizeof(SCode));


   return result;


}
CUtiXMLException::CUtiXMLException(const CUtiString &name)
 {m_n = name;std::wcout << m_n << std::endl;};

void CXMLNode::setPropVector(const CUtiString& name, const CVector &b)
{
assert(isBezeichner(name));
   std::string l_name = converttomulti(name.c_str());
   char floatvalue[128];
   snprintf(floatvalue, 124, "%f, %f, %f", (float)b.x, (float)b.y, (float)b.z);
   m_element->SetAttribute(l_name, floatvalue);
}
CVector CXMLNode::getPropVector(const CUtiString& name, const CVector &std) const
{assert(isBezeichner(name));
   std::string l_name = converttomulti(name.c_str());
   const char* l_value = m_element->Attribute(l_name.c_str());
   if (!l_value) return std;
   //char* l_tail;
   //double value = strtod(l_value, &l_tail, 10);
   CVector3f result;
  // wcout << "Locale is: "<< setlocale(LC_ALL,NULL) << endl;
   int v = sscanf(l_value, "%f , %f , %f", &result.x, &result.y, &result.z);

   if (v == EOF || v < 3) {

      CUtiString error(UTEXT("CXMLNode::getPropVector: the value '"));
      error += converttowide(l_value) + UTEXT("' can not be converted to a CVector value.");

      throw new CUtiXMLException(error);

   }
   return toDouble(result);


}
void CXMLNode::setPropVector4(const CUtiString& name, const CVector4 &b)
{assert(isBezeichner(name));
    std::string l_name = converttomulti(name.c_str());
   char floatvalue[128+40];
   snprintf(floatvalue, 124+40, "%f, %f, %f, %f", (float)b.x, (float)b.y, (float)b.z, (float)b.w);
   m_element->SetAttribute(l_name, floatvalue);

}
CVector4 CXMLNode::getPropVector4(const CUtiString& name, const CVector4 &std) const
{assert(isBezeichner(name));
   std::string l_name = converttomulti(name.c_str());
   const char* l_value = m_element->Attribute(l_name.c_str());
   if (!l_value) return std;
   //char* l_tail;
   //double value = strtod(l_value, &l_tail, 10);
   CVector4f result;

   int v = sscanf(l_value, "%f , %f , %f , %f", &result.x, &result.y, &result.z, &result.w);

   if (v == EOF || v < 4) {

      CUtiString error(UTEXT("CXMLNode::getPropVector4: the value '"));
      error += converttowide(l_value) + UTEXT("' can not be converted to a CVector4 value.");

      throw new CUtiXMLException(error);

   }
   return toDouble(result);

}
void CXMLNode::setPropString(const CUtiString& name, const CUtiString& b)
{assert(isBezeichner(name));
    std::string l_name = converttomulti(name.c_str());
    std::string l_value = converttomulti(b.c_str());

   m_element->SetAttribute(l_name, l_value);

}
CUtiString CXMLNode::getPropString(const CUtiString& name, const CUtiString& std) const
{assert(isBezeichner(name));
    std::string l_name = converttomulti(name.c_str());
    const char* l_value = m_element->Attribute(l_name.c_str());
    if (!l_value) return std;
    return converttowide(l_value);
}

int CXMLNode::getPropDataSize(const CUtiString& name) const
{assert(isBezeichner(name));
   std::string l_name = converttomulti(name.c_str());
    const char* l_value = m_element->Attribute(l_name.c_str());
    if (!l_value) return 0;
    int l_length = strlen(l_value);
    int datasize = CUti64Coder::reconstructDataLength(l_length);
    return datasize;
}

void CXMLNode::setPropData(const CUtiString& name, const unsigned char* Data, size_t size)
{assert(isBezeichner(name));
   int l_size = CUti64Coder::getDataLength(size);
   char* c = new char[l_size+1];

   CUti64Coder::writeData(Data, size, c, l_size);
   c[l_size] = 0;
   std::string l_name = converttomulti(name.c_str());
    m_element->SetAttribute(l_name, c);
    delete [] c;

}
bool CXMLNode::getPropData(const CUtiString& name, unsigned char* &Data, size_t &size) const
{assert(isBezeichner(name));
    std::string l_name = converttomulti(name.c_str());
    const char* l_value = m_element->Attribute(l_name.c_str());
    if (!l_value) return false;
    int l_length = strlen(l_value);
    int datasize = CUti64Coder::reconstructDataLength(l_length);
    if (Data == NULL) {
        Data = new unsigned char[datasize];
        size = datasize;
    } else {
        uassert(datasize <= size, UTEXT("CXMLNode::getPropData: The memory buffer data is not big enough"));
    }
    CUti64Coder::readData(l_value,l_length, Data, datasize);
    return true;

}
bool CXMLNode::propExists(const CUtiString& name) const
{assert(isBezeichner(name));
     std::string l_name = converttomulti(name.c_str());
    const char* l_value = m_element->Attribute(l_name.c_str());
    if (!l_value) return false;
    return true;

}
CXNode* CXMLNode::addNode(const CUtiString& name)
{assert(isBezeichner(name));
      std::string l_name = converttomulti(name.c_str());
      TiXmlElement* element = new TiXmlElement(l_name);
      m_element->LinkEndChild(element);
      CXMLNode* n;
      m_children.push_back(n = new CXMLNode(this, m_doc, element));
       return n;
}
CXNode* CXMLNode::getNodeByName(const CUtiString& name)
{assert(isBezeichner(name));
     for (COwnList<CXMLNode>::iterator it = m_children.begin(); it != m_children.end(); it++)
     {
         if (it->getName() == name)
         {
             return &(*it);
         }
     }
     return NULL;

}
void CXMLNode::deleteNode(const CUtiString& name)
{assert(isBezeichner(name));
    for (COwnList<CXMLNode>::iterator it = m_children.begin(); it != m_children.end(); it++)
     {
         if (it->getName() == name)
         {
             int index = it-m_children.begin();
             deleteItem(index);

         }
     }


}
void CXMLNode::deleteItem(int i)
{
    bool b = m_element->RemoveChild(m_children[i].m_element);
    utiassert(b == true);
    m_children.deleteItem(i);

}
void CXMLNode::remove(CXNode* n)
{
    for (COwnList<CXMLNode>::iterator it = m_children.begin(); it != m_children.end(); it++)
     {
         if (&(*it) == n)
         {
             int index = it-m_children.begin();
            deleteItem(index);
             return;
         }
     }
     return;

}
int CXMLNode::getChildCount() const
{
    return m_children.size();

}
CXNode* CXMLNode::getChild(int i) const
{
    return (CXNode*)&m_children[i];

}

void CXMLNode::removeProperty(const CUtiString& name)
{
    std::string l_name = converttomulti(name.c_str());
     m_element->RemoveAttribute(l_name);
}

void CXMLNode::clearNodes()
{
    //TODO: ALle löschen?
   // doc->beginNoDelete();
     m_children.clear();
     m_element->Clear();
    // doc->endNoDelete();

}

void CXMLNode::copyPropertiesTo(CXNode* node)
{

    assert(false);
  /* for (int i = 0; i < m_element->; i++) {
       CBaseProp* prop = check_cast<CBaseProp*>(m_properties[i]);
       CUtiString propname = ((CXUtiDocument*)m_doc)->getString(prop->m_id);
       if (dynamic_cast<CStringProp*>(prop) != NULL) {
          CStringProp* pdata = check_cast<CStringProp*>(prop);
          node->setPropString(propname, pdata->m_b);
       } else if (dynamic_cast<CBoolProp*>(prop) != NULL) {
          CBoolProp* pdata = check_cast<CBoolProp*>(prop);
          node->setPropBool(propname, pdata->m_b);
       } else if (dynamic_cast<CIntProp*>(prop) != NULL) {
          CIntProp* pdata = check_cast<CIntProp*>(prop);
          node->setPropInt(propname, pdata->m_b);
       } else if (dynamic_cast<CFloatProp*>(prop) != NULL) {
          CFloatProp* pdata = check_cast<CFloatProp*>(prop);
          node->setPropFloat(propname, pdata->m_b);
       } else if (dynamic_cast<CDoubleProp*>(prop) != NULL) {
          CDoubleProp* pdata = check_cast<CDoubleProp*>(prop);
          node->setPropDouble(propname, pdata->m_b);
       } else if (dynamic_cast<CCodeProp*>(prop) != NULL) {
          CCodeProp* pdata = check_cast<CCodeProp*>(prop);
          node->setPropCode(propname, pdata->m_b);
       } else if (dynamic_cast<CVectorProp*>(prop) != NULL) {
          CVectorProp* pdata = check_cast<CVectorProp*>(prop);
          node->setPropVector(propname, pdata->m_b);
       } else if (dynamic_cast<CVector4Prop*>(prop) != NULL) {
          CVector4Prop* pdata = check_cast<CVector4Prop*>(prop);
          node->setPropVector4(propname, pdata->m_b);
       } else if (dynamic_cast<CDataProp*>(prop) != NULL) {
          CDataProp* pdata = check_cast<CDataProp*>(prop);
          node->setPropData(propname, pdata->m_data, pdata->m_size);
       } else {
       assert(false);
       }

   }*/
}

void CXMLNode::clearProperties()
{
     TiXmlAttribute* attribute;
    while ((attribute = m_element->FirstAttribute()) != NULL)
    {
        m_element->RemoveAttribute(attribute->Name());
    }
}

int XMLSaveToFile(const CUtiString&  string, CUtiOB* cl)
{
    CXDocument* xml = new CXMLDocument();
    CWriteInfo* writeinfo = new CWriteInfo();
    writeinfo->addObject(cl, cl->GetCode());
    XWriteClass(xml->getRoot(), writeinfo, cl);
    int errorcode = xml->saveToFile(string);
    delete writeinfo;
    delete xml;
    return errorcode;
};


int XMLLoadFromFileNoCreate(const CUtiString&  string, CUtiOB* obj)
{
    CXDocument* xml = new CXMLDocument();
    int errorcode = xml->loadFromFile(string);
    if (errorcode != UTI_OK) {
        delete xml;
        return errorcode;
    }
    CReadInfo * r = new CReadInfo();
    XReadClassNoCreate(xml->getRoot(), r, obj);
    SCode c = xml->getRoot()->getPropCode(L"utireserved_objectcode", NULLCODE);
    delete xml;

    r->addObject(obj, c);
    r->link();
    delete r;
    return UTI_OK;
}

int XMLLoadFromFile(const CUtiString&  string, CUtiOB* parent, CUtiOB* &result)
{
    result = NULL;
    CXDocument* xml = new CXMLDocument();
    int errorcode = xml->loadFromFile(string);
    if (errorcode != UTI_OK) {
        delete xml;
		result = NULL;
        return errorcode;
    }
    CReadInfo * r = new CReadInfo();
    result = XReadClass(xml->getRoot(), r, parent);
    SCode c = xml->getRoot()->getPropCode(L"utireserved_objectcode", NULLCODE);
    delete xml;
    if (result == NULL) {
        delete r;
        return UTI_CLASSNOTFOUND;
    }
    r->addObject(result, c);
    r->link();
    delete r;
    return UTI_OK;
}


int XMLSaveToNode(CXNode* node, CUtiOB* cl)
{

    EASY_FUNCTION(profiler::colors::Blue);
	//CXDocument* xml = new CXMLDocument();
	CWriteInfo* writeinfo = new CWriteInfo();
	writeinfo->addObject(cl, cl->GetCode());
	XWriteClass(node, writeinfo, cl);
	//int errorcode = xml->saveToFile(string);
	delete writeinfo;
//	delete xml;
	return 0;
}

int XMLLoadFromNode(CXNode* node, CUtiOB* parent, CUtiOB* &result)
{

    EASY_FUNCTION(profiler::colors::Blue);
	result = NULL;
//	CXDocument* xml = new CXMLDocument();
	//int errorcode = xml->loadFromFile(string);
	//if (errorcode != UTI_OK) {
	//	delete xml;
	//	result = NULL;
	//	return errorcode;
	//}
	CReadInfo * r = new CReadInfo();
	result = XReadClass(node, r, parent);
	SCode c = node->getPropCode(L"utireserved_objectcode", NULLCODE);
//	delete xml;
	if (result == NULL) {
		delete r;
		return UTI_CLASSNOTFOUND;
	}
	r->addObject(result, c);
	r->link();
	delete r;
	return UTI_OK;
}
