/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef UTIXML_H
#define UTIXML_H

#include "UtiType.h"
#include "tinyxml/tinyxml.h"
#include <assert.h>


class CUtiXMLException : public CUtiException
{
public:
  CUtiXMLException(const CUtiString& name);

};

class CXMLNode : public CXNode
{
protected:

   TiXmlElement* m_element;
   CReferenceList<CXMLNode> m_children;
   CUtiString m_name;
public:

  CXMLNode(CXNode* parent, CXDocument* doc, TiXmlElement* element);
  virtual void setName(const CUtiString& name);
  virtual CUtiString getName() const;
  virtual void setPropBool(const CUtiString& name, bool b);
  virtual bool getPropBool(const CUtiString& name, bool std) const;
  virtual void setPropInt(const CUtiString& name, int b);
  virtual int getPropInt(const CUtiString& name, int std) const ;
  virtual void setPropFloat(const CUtiString& name, float b);
  virtual float getPropFloat(const CUtiString& name, float std) const;
  virtual void setPropDouble(const CUtiString& name, double b);
  virtual double getPropDouble(const CUtiString& name, double std) const;
  virtual void setPropCode(const CUtiString& name, const SCode &b);
  virtual SCode getPropCode(const CUtiString& name, const SCode &std) const;
  virtual void setPropVector(const CUtiString& name, const CVector &b);
  virtual CVector getPropVector(const CUtiString& name, const CVector &std) const;
  virtual void setPropVector4(const CUtiString& name, const CVector4 &b);
  virtual CVector4 getPropVector4(const CUtiString& name, const CVector4 &std) const;
  virtual void setPropString(const CUtiString& name, const CUtiString& b);
  virtual CUtiString getPropString(const CUtiString& name, const CUtiString& std) const;
  virtual int getPropDataSize(const CUtiString& name) const;
  virtual void setPropData(const CUtiString& name, const unsigned char* Data, size_t size);
  virtual bool getPropData(const CUtiString& name, unsigned char* &Data, size_t &size) const;
  virtual bool propExists(const CUtiString& name) const;

  virtual CXNode* addNode(const CUtiString& name);
  virtual CXNode* getNodeByName(const CUtiString& name);
  virtual void deleteNode(const CUtiString& name);
  virtual void deleteItem(int i);
  virtual void remove(CXNode* n);
  virtual void removeProperty(const CUtiString& name);
  virtual int getChildCount() const;
  virtual CXNode* getChild(int i) const;
  virtual void clearNodes();
  virtual void clearProperties();
  virtual CUtiString getText();

  virtual void copyPropertiesTo(CXNode* node);
};

class CXMLDocument : public CXDocument
{
    protected:
        TiXmlDocument* m_document;
    public:
        /** Default constructor */
        CXMLDocument();
        /** Default destructor */
        virtual ~CXMLDocument();
        virtual int saveToStream(CStream &Stream);
        virtual int loadFromStream(CStream &Stream);
        virtual int saveToFile(const CUtiString& name);
        virtual int loadFromFile(const CUtiString& name);

};

int XMLSaveToFile(const CUtiString&  string, CUtiOB* cl);
int XMLLoadFromFile(const CUtiString&  string, CUtiOB* parent, CUtiOB* &result);
int XMLSaveToNode(CXNode* node, CUtiOB* cl);
int XMLLoadFromNode(CXNode* node, CUtiOB* parent, CUtiOB* &result);
int XMLLoadFromFileNoCreate(const CUtiString&  string, CUtiOB* obj);

#endif // UTIXML_H
