/***************************************************************************
 *   Copyright (C) 2007 by Staud   *
 *   mstaud@linux-fd2d   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "ziparchiv.h"
#include "utisystem.h"
#include <assert.h>
#ifdef HAVE_ZZIP
#include <zzip/zzip.h>
#endif
#include "objectfactory.h"
#include "CDirectoryTree.h"
#include "CDirFile.h"
#include "CDirDirectory.h"
#include "zlib.h"
using namespace UtiFramework;

#ifdef HAVE_ZZIP
CRegisterClassSystem<CFileSystem, CZipArchiv> registerZipArchiv(UTEXT("zip"));
#endif

size_t uti_compressBound(size_t s)
{
return compressBound(s);
};

int uti_compress(unsigned char* desttarget, UTI_UINT64*destsize, unsigned char* source, UTI_UINT64 sourcesize)
{
  //utiassert(sizeof(UINT64) == sizeof(uLong));
 // utiassert(sizeof(UINT64) == sizeof(uLongf));

    uLongf localdest=*destsize;

    int res = compress(desttarget, &localdest, source, (uLong)sourcesize);

    *destsize = localdest;
    return res;
}

int uti_uncompress(unsigned char* desttarget, UTI_UINT64*destsize, unsigned char* source, UTI_UINT64 sourcesize)
{
    //utiassert(sizeof(UINT64) == sizeof(uLong));
   // utiassert(sizeof(UINT64) == sizeof(uLongf));

      uLongf localdest=*destsize;

      int res = uncompress(desttarget, &localdest, source, (uLong)sourcesize);
      *destsize = localdest;
      return res;
}
 
#ifdef HAVE_ZZIP
CZipArchiv::CZipArchiv(): CBaseArchiv()
{
   zipdir = NULL;
   tree = NULL;
}


CZipArchiv::~CZipArchiv()
{

   ZZIP_DIR * d = (ZZIP_DIR*) zipdir;
   if (d) {
   zzip_dir_close(d);
   delete tree;
   };
}
void CZipArchiv::init(const CUtiString& _name)
{
  CBaseArchiv::init(_name);
  utiassert(zipdir == NULL);
  errorassert(FileExists(archivname.c_str()), CUtiString(UTEXT("Archiv not found: ")) + archivname);
  std::string str = converttomulti(archivname.c_str());
  //assert(zzip_dir_open != NULL);
  ZZIP_DIR * d = zzip_dir_open(str.c_str(), 0);
  errorassert(d != NULL, CUtiString(L"Archiv could not be opened: ") + archivname);
  zipdir = d;

  tree = new CDirectoryTree();

  ZZIP_DIRENT dirent;
  memset(&dirent, 0, sizeof(dirent));
  CUtiString name;
  while (zzip_dir_read(d,&dirent)) {
        /* show info for first file */
     /*struct zzip_dirent
     {
     int	 	d_compr;	 compression method
     int         d_csize;         compressed size
     int	 	st_size;	 file size / decompressed size
     char * 	d_name;		 file name / strdupped name
  };*/
     //Zerteile Name und Füge in Baum ein
     name.LoadAnsi(dirent.d_name);
     //wcout << name << endl;
     CDirFile* f = tree->getRootNode()->getFile(name, true);
     if (f) {
        f->setSize(dirent.st_size);
     };
     //   print("%s %i/%i", dirent.d_name, dirent.d_csize, dirent.st_size);
   }
  zzip_dir_close(d);

}

bool CZipArchiv::fileExists(const CUtiString& name)
{
   if (name.size() == 0) return false;
   CDirFile* f = tree->getRootNode()->getFile(name, false);
   return f != NULL;
}


class CZipStream : public CStream
{
   protected:
      ZZIP_DIR* dir;
      ZZIP_FILE* file;
      UTI_INT64 size;
   public:
      ~CZipStream()
      {
         zzip_file_close(file);
      }
      CZipStream(ZZIP_DIR* _dir,
      ZZIP_FILE* _file,
      UTI_INT64 _size)
      {
         dir = _dir;
         file = _file;
         size = _size;
      }

      bool fail()
      {
         return eof();
      }
      bool eof()
      {
         utiassert(zzip_tell(file) >= size);
         return zzip_tell(file) >= size;
      }

      size_t ReadData(void* data, size_t size)
      {
         zzip_ssize_t len = zzip_file_read(file, data, size);
         return len;
      }

      size_t WriteData(const void* data, size_t size)
      {
         utiassert(false);
		 return 0;
      }

      void setPosition(off_t i, TPositionType postype)
      {
         switch (postype) {
            case START: {
               zzip_seek(file, i, SEEK_SET);
            }; break;
            case CURRENT: {
               zzip_seek(file, i, SEEK_CUR);
            }; break;
            case END: {
               zzip_seek(file, i, SEEK_END);
            }; break;
         };

      }
      size_t getReadPosition()
      {
         return zzip_tell(file);
      }

      size_t getWritePosition()
      {
         utiassert(false);
		 return 0;
      }
      size_t getPosition()
      {
         return getReadPosition();
      };
};

CStream* CZipArchiv::openFile(const CUtiString& name, TOpenType ot)
{
   errorassert(ot == otRead, CUtiString(L"File in a zip archiv can only be opened"));
   utiassert(zipdir != NULL);
   ZZIP_DIR * d = (ZZIP_DIR*) zipdir;
   if (!d) return NULL;
   std::string str = converttomulti(name.c_str());
    ZZIP_FILE* file = zzip_file_open(d, str.c_str(), ZZIP_ONLYZIP | ZZIP_CASELESS);
    if (!file) return NULL;
    ZZIP_STAT s;
    zzip_dir_stat(d, str.c_str(), &s, ZZIP_CASEINSENSITIVE);
    return new CZipStream(d, file, s.st_size);
}


UTI_INT64 CZipArchiv::fileSize(const CUtiString& name)
{
   CDirFile* f = tree->getRootNode()->getFile(name, false);
   if (f != NULL) {
      return f->getSize();
   } else {
      return 0;
   };
}

void CZipArchiv::deleteFile(const CUtiString& str)
{

}

void CZipArchiv::getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set< CUtiString >& result, bool relativ, bool searchdirs)
{
    if (!relativ)return;
   CDirDirectory* d = tree->getRootNode()->getDirectory(dir, false);
   if (d == NULL) return;
   if (searchdirs) {
       for (int i = 0; i < d->getChildrenCount(); i++) {
          CDirectoryItem* item = d->getChildren(i);
          if (dynamic_cast<CDirDirectory*>(item) != NULL)  {
          CUtiString k;
          k = item->getName();
          k.ExtractFileExt();
          k= UTEXT('.')+k;
          if (k == type || type.size() == 0)
          {
             k = item->getName();
             if (result.find(k) != result.end()) {
                result.insert(k);
             };
          }
          }
       };
   } else {
   for (int i = 0; i < d->getChildrenCount(); i++) {
      CDirectoryItem* item = d->getChildren(i);
      if (dynamic_cast<CDirFile*>(item) != NULL)  {
      CUtiString k;
      k = item->getName();
      k.ExtractFileExt();
      k= UTEXT('.')+k;
      if (k == type || type.size() == 0)
      {
         k = item->getName();
         if (result.find(k) != result.end()) {
            result.insert(k);
         };
      }
      }
   };
};
}
#endif
class CZipInternData : public ReferenceUtiOB
{
public:
    explicit CZipInternData() {m_strm.zalloc = Z_NULL;
    m_strm.zfree = Z_NULL;
    m_strm.opaque = Z_NULL;};
    z_stream m_strm;
};



CZipMemoryStream::~CZipMemoryStream()
{
 //   if (m_deletemem && (Memory != NULL))
  //      delete[] (U_SCHAR*)Memory;
  
}


CZipMemoryStream::CZipMemoryStream(CStream* basestream) : CStream()
{
    //Memory=NULL;
   /* Position = 0;
    failed = false;
   // m_deletemem = true;
    m_allowwrite = false;
    datasize = 0;*/
    m_basestream = basestream;
    docount = false;
    m_data = new CZipInternData();
 
    m_state = zsNone;
   
    //  FileName[0] = 0;
}

void CZipMemoryStream::beginRead()
{
  utiassert(m_state == zsNone);
 CHECKEXISTANCEOF(m_basestream.get());
 
 m_state = zsRead;
 m_data->m_strm.next_in = inputbfr;
 m_data->m_strm.avail_in = 0;
 
 m_data->m_strm.avail_out = ZIP_BUFFERSIZE;
 m_data->m_strm.next_out = outputbfr;
 m_outputsize = 0;
 m_outputposition = 0;
 m_outputlast = false;
 int ret = inflateInit(&m_data->m_strm);
 utiassert(ret == Z_OK);
 m_basestream->setPosition(0, START);
  
  FillOutputRead();
}

SZipResult CZipMemoryStream::endRead()
{
  utiassert(m_state  == zsRead);
 // EmptyInputBfrWrite();
  SZipResult res;
  res.m_compressedsize = m_data->m_strm.total_in;
  res.m_uncompressedsize = m_data->m_strm.total_out;
  inflateEnd(&m_data->m_strm);
  m_state = zsNone;
  m_outputsize = 0;
 m_outputposition = 0;
 m_outputlast = false;
 return res;
}

void CZipMemoryStream::FillOutputRead() 
{assert(m_state  == zsRead);
CZipInternData* d = m_data.get();

if (m_outputsize != m_outputposition) return;
  int ret;
    //int have;
   
   if (d->m_strm.avail_in == 0) {
     if (m_basestream->eof()) return;
     d->m_strm.avail_in = m_basestream->ReadData(&inputbfr,ZIP_BUFFERSIZE);
      m_data->m_strm.next_in = inputbfr;
      
   };
   if (d->m_strm.avail_in == 0) {
    
     return;
   };
   d->m_strm.avail_out = ZIP_BUFFERSIZE;
   d->m_strm.next_out = outputbfr;
   
   
   ret = inflate(&d->m_strm, Z_NO_FLUSH);
            utiassert(ret != Z_STREAM_ERROR); 
            switch (ret) {
            case Z_NEED_DICT:
                ret = Z_DATA_ERROR;   utiassert(false);
            case Z_DATA_ERROR:
            case Z_MEM_ERROR:
                (void)inflateEnd(&d->m_strm);
		utiassert(false);
                return;
            }
    if (ret == Z_STREAM_END) {
      m_outputlast = true;
    };
    m_outputsize = ZIP_BUFFERSIZE - d->m_strm.avail_out;
    m_outputposition = 0;

   
}


void CZipMemoryStream::beginWrite()
{
   utiassert(m_state == zsNone);
  int ret = deflateInit(&m_data->m_strm, -1);
 utiassert(ret == Z_OK);
 
 m_state = zsWrite;
 m_data->m_strm.next_in = inputbfr;
 m_data->m_strm.avail_in = 0;
 
 m_data->m_strm.avail_out = ZIP_BUFFERSIZE;
 m_data->m_strm.next_out = outputbfr;
 m_basestream->setPosition(0, START);
 
}

SZipResult CZipMemoryStream::endWrite()
{
  utiassert(m_state  == zsWrite);
  EmptyInputBfrWrite(true);
  SZipResult res;
  res.m_uncompressedsize = m_data->m_strm.total_in;
  res.m_compressedsize = m_data->m_strm.total_out;
  deflateEnd(&m_data->m_strm);
  m_state = zsNone;
  return res;
}

void CZipMemoryStream::EmptyInputBfrWrite(bool finish) 
{assert(m_state  == zsWrite);
  int ret;
    int have;
   CZipInternData* d = m_data.get();
   do {
	// while (d->m_strm.avail_in > 0) {
	    d->m_strm.avail_out = ZIP_BUFFERSIZE;
            d->m_strm.next_out = outputbfr;
	    //wcout <<"BVla:"<< finish << endl;
	    ret = deflate(&d->m_strm, finish ? Z_FINISH : Z_NO_FLUSH);    /* no bad return value */
            utiassert(ret != Z_STREAM_ERROR);  /* state not clobbered */
	    have = ZIP_BUFFERSIZE - d->m_strm.avail_out;
	    m_basestream->WriteData(outputbfr, have);
	 } while (d->m_strm.avail_out == 0);
	 utiassert(d->m_strm.avail_in == 0);
      m_data->m_strm.next_in = inputbfr;
      m_data->m_strm.avail_in = 0;
};

size_t CZipMemoryStream::WriteData(const void *data, size_t size)
{
   
       utiassert(m_state  == zsWrite);
       unsigned char* datamove = (unsigned char*)data;
       int restamount;
       size_t totalamount = size;
       CZipInternData* d = m_data.get();
       while (size != 0) {
	 utiassert(size >= 0);
	 if (d->m_strm.avail_in +size <= ZIP_BUFFERSIZE) {
	   memcpy(inputbfr+d->m_strm.avail_in, datamove, size);
	   d->m_strm.avail_in += size;
	   size = 0;
	   return totalamount;
	 } else {
	   restamount = ZIP_BUFFERSIZE-d->m_strm.avail_in;
	   utiassert(restamount <= size);
	   memcpy(inputbfr+d->m_strm.avail_in, datamove, restamount);
	   size -= restamount;
	   d->m_strm.avail_in = ZIP_BUFFERSIZE;
	   datamove += restamount;
	 };
	 EmptyInputBfrWrite(false);
       };
	return totalamount;
}

size_t CZipMemoryStream::ReadData(void *data, size_t size)
{assert(m_state  == zsRead);
 size_t totalamount = size;
 unsigned char* datamove = (unsigned char*)data;
  int restamount;
   while (size != 0) {
    if (m_outputposition + size <= m_outputsize) {
      memcpy(datamove, outputbfr+m_outputposition, size);
      m_outputposition+= size;
         size = 0;
      return totalamount;
    } else {
      restamount = m_outputsize-m_outputposition;
      utiassert(restamount <= size);
      memcpy(datamove, outputbfr+m_outputposition, restamount);
      m_outputposition = m_outputsize;
      size -= restamount;
         datamove += restamount;
	 if (m_outputlast) {
	   return totalamount-size;
	 };
    };
     FillOutputRead();
     if (m_outputsize== 0) {
       return totalamount-size;
     };
   };
   
	return totalamount;
}


void CZipMemoryStream::setPosition(off_t i, TPositionType postype)
{
  utiassert(false);
	/*switch (postype) {
		case START: {Position = i;}; break;
		case CURRENT: {Position += i;}; break;
		case END: {Position = datasize-i;}; break;
	};
	if (Position <= 0) Position = 0;
	if (Position >= datasize) Position = datasize;*/
}

size_t CZipMemoryStream::getReadPosition()
{
  utiassert(false);
    return(0);
}

size_t CZipMemoryStream::getWritePosition()
{ utiassert(false);
    return(0);
}
