/***************************************************************************
 *   Copyright (C) 2007 by Staud   *
 *   mstaud@linux-fd2d   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ZIPARCHIV_H
#define ZIPARCHIV_H

#include "UtiArchiv.h"
#include "utiarray.h"

class CDirectoryTree;

/**
	@author Staud <mstaud@linux-fd2d>
*/

#ifdef HAVE_ZZIP
class CZipArchiv : public CBaseArchiv
{
protected:
   void* zipdir;
   CDirectoryTree* tree;
public:
    CZipArchiv();

    ~CZipArchiv();

    virtual bool fileExists(const CUtiString& name);
    virtual CStream* openFile(const CUtiString& name, TOpenType ot);
    virtual UTI_INT64 fileSize(const CUtiString& name);
    virtual void deleteFile(const CUtiString& str);
    virtual void getFilesInDir(const CUtiString& dir, const CUtiString& type, std::set< CUtiString >& result, bool relativ, bool searchdirs);
    virtual void init(const CUtiString& _name);

};
#endif

class CZipInternData;

struct SZipResult
{
  int m_uncompressedsize;
  int m_compressedsize;
};

enum TZipState {zsNone, zsWrite, zsRead};
const int ZIP_BUFFERSIZE = 2048*4;
class CZipMemoryStream : public CStream
{
protected:
   /* Array<unsigned char> Memory;
    size_t datasize;
    bool failed;
    off_t Position;*/
 //  bool eof;
    unsigned char inputbfr[ZIP_BUFFERSIZE];
    unsigned char outputbfr[ZIP_BUFFERSIZE];
    //bool m_deletemem;
    uti_refptr<CStream> m_basestream;
    TZipState m_state;
    uti_refptr<CZipInternData> m_data;
    int m_outputsize;
    bool m_outputlast;
 int m_outputposition;
    void EmptyInputBfrWrite(bool finish);
    void FillOutputRead();
public:
    explicit CZipMemoryStream(CStream* basestream);
    ~CZipMemoryStream();
    void beginRead();
    SZipResult endRead();
    void beginWrite();
    SZipResult endWrite();
    bool fail()
    {
        return(eof());
    };
    bool eof()
    {
      switch (m_state) {
	case zsNone:
	case zsWrite: 
	{
	  return false;
	}; break;
	case zsRead: {
	  return m_outputlast && m_outputsize <= m_outputposition;
	};
      };
        return(false);
    };
    size_t ReadData(void* data, size_t size);
    size_t WriteData(const void* data, size_t size);
    void setPosition(off_t i, TPositionType postype);
    size_t getReadPosition();
    size_t getWritePosition();
   
};
size_t uti_compressBound(size_t s);
int uti_compress(unsigned char* desttarget, UTI_UINT64*destsize, unsigned char* source, UTI_UINT64 sourcesize);
int uti_uncompress(unsigned char* desttarget, UTI_UINT64*destsize, unsigned char* source, UTI_UINT64 sourcesize);



enum TZipArrayState {zasNormal=0, zasWrite, zasRead, zasCompressed};



template <class T>
class ZipArray2D : public CUtiOB
{
protected:
  
   public:
      Array<unsigned char> m_compresseddata;
      TZipArrayState m_state;
      T* array;
      int array_width;
      int array_height;
      int array_newline;
      inline TZipArrayState getState()
      {
          return m_state;
      }

      inline int width() const
      {
         return array_width;
      }
      inline int height() const
      {
         return array_height;
      }
      ZipArray2D(int w, int h) : CUtiOB()
      {
         m_state = zasNormal;
         array_width = w;
         array_height = h;
         array_newline = w;
         array = new T[array_newline*array_height];
      }
     
      ZipArray2D() : CUtiOB()
      {
	 m_state = zasNormal;
         array_width = 0;
         array_height = 0;
         array_newline = 0;

         array = NULL;
      };
      ~ZipArray2D()
      {
         if (array)
            delete [] array;
      }
      void data_copyTo(ZipArray2D& other)
      {
          other.m_state = m_state;
          other.array = array;
          other.array_width = array_width;
          other.array_height = array_height;
          other.array_newline = array_newline;
          other.m_compresseddata.setSize(m_compresseddata.size());
          m_compresseddata.copyToArrayHighSpeed(0, m_compresseddata.size(), other.m_compresseddata, 0);
      }

      Array<unsigned char> &getCompressedRaw()
      {
	return m_compresseddata;
      };
      bool iscompressed()
      {

	return m_compresseddata.size() != 0;
      };
      bool isReadAble()
      {
	return m_state ==zasNormal || m_state == zasRead || m_state == zasWrite;
      };
      bool isWriteAble()
      {
	return m_state ==zasNormal ||  m_state == zasWrite;
      };
      void endWrite()
      {
	utiassert(m_state == zasWrite);
	utiassert(array != NULL);
	UTI_UINT64 s = array_height*array_newline*sizeof(T);
	UTI_UINT64 cp = uti_compressBound(array_height*array_newline*sizeof(T));
	m_compresseddata.setSize(cp);
	uti_compress(m_compresseddata.array, &cp, (unsigned char*)array,s);
	m_compresseddata.setSizeCopy(cp);
	 delete [] array;
	 array = NULL;
	 m_state = zasCompressed;
      };
      void beginWrite()
      {
	if (m_state == zasNormal){
	  m_state = zasWrite;
	  return;
	};
	utiassert(m_state == zasCompressed);
	utiassert(array == NULL && m_compresseddata.size() != 0);
	array = new T[array_newline*array_height];
	UTI_UINT64 s = array_height*array_newline*sizeof(T);
	uti_uncompress((unsigned char*)array, &s, m_compresseddata.array, m_compresseddata.size());
	utiassert(s == array_height*array_newline*sizeof(T));
	m_compresseddata.setSize(0);
	m_state = zasWrite;
      };
      
      void endRead()
      {
	utiassert(m_state == zasRead);
	utiassert(array != NULL);
	 delete [] array;
	 array = NULL;
	 m_state = zasCompressed;
      };
      void beginRead()
      {
	if (m_state == zasNormal){
	  m_state = zasRead;
	  return;
	};
	utiassert(m_state == zasCompressed);
	utiassert(array == NULL && m_compresseddata.size() != 0);
	array = new T[array_newline*array_height];
	UTI_UINT64 s = array_height*array_newline*sizeof(T);
	uti_uncompress((unsigned char*)array, &s, m_compresseddata.array, m_compresseddata.size());
	utiassert(s == array_height*array_newline*sizeof(T));
	m_state = zasRead;
      };
      void decompressTo(Array2D<T>& data)

      {
          utiassert(m_state == zasCompressed);
          utiassert(m_compresseddata.size() != 0);
          data.setSize(array_width, array_height);
          assert(data.array_newline == array_newline);
          UTI_UINT64 s = array_height*array_newline*sizeof(T);
          uti_uncompress((unsigned char*)data.array, &s, m_compresseddata.array, m_compresseddata.size());
          utiassert(s == array_height*array_newline*sizeof(T));
      }

      void setSize(int w, int h) {
	 CHECKEXISTANCE; 

	 utiassert(m_state != zasCompressed);
	 utiassert(m_state != zasRead);
	 utiassert(array != NULL || array_width == 0);
         if (array != NULL){
	  
            delete [] array;
	 };
       
         array_width = w;
         array_height = h;
         array_newline = w;
         array = new T[array_newline*array_height];
      };
      void setSizeCopy(int w, int h, int basex=0, int basey=0) {
	 CHECKEXISTANCE; 
	
	 utiassert(m_state != zasCompressed);
	 utiassert(m_state != zasRead);
	T* newarray = new T[w*h];
	utiassert(array != NULL || array_width != 0);
	utiassert(basex >= 0);
	utiassert(basey >= 0);
	utiassert(w >= 0);
	utiassert(h >= 0);
	int mw = umin(w, array_width-basex);
	int mh = umin(h, array_height-basey);
	for( int j = 0; j < mh; j++) {
	  int jbasey = j+basey;
	   for (int i = 0; i < mw; i++) {
	      newarray[i+j*w] =  array[(i+basex)+(jbasey)*array_newline];
	   };
	};
	
         if (array != NULL){
            delete [] array;
	 };
         array_width = w;
         array_height = h;
         array_newline = w;
         array = newarray;
      };
      //Only for Flat Objects
      void copyToFast(Array2D<T>& obj) const
      {
	utiassert(m_state !=zasCompressed);
	utiassert(array != NULL);
	
	memcpy(obj.array, array, sizeof(T)*array_width*array_height);
	
      }
      void copyToArray(Array2D<T>& obj) const
      {
	
	utiassert(array != NULL);
	obj.setSize(array_width, array_height);
	int c = array_width*array_height;
	for (int i = 0; i < c; i++) {
	   obj.array[i] = array[i];
	}
      }
      /*void ReferenzTo(const Array2D &a)
      {
         dynamic = false;
         array_size = a.array_size;
         array = a.array;
      }*/
      inline T &operator()(unsigned int x, unsigned int y)
      {
	utiassert(m_state ==zasNormal || m_state == zasWrite);
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
         return array[x+y*array_newline];

      }
      inline T* getPointer(unsigned int x, unsigned int y)
      {assert(m_state ==zasNormal || m_state == zasWrite);
	utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
         return &array[x+y*array_newline];
      }
      inline const T &operator()(unsigned int x, unsigned int y) const
      {assert(m_state ==zasNormal || m_state == zasRead || m_state == zasWrite);
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
         return array[x+y*array_newline];

      }
      inline const T &get(unsigned int x, unsigned int y) const
      {assert(m_state ==zasNormal || m_state == zasRead || m_state == zasWrite);
         utiassert(array != NULL);
         utiassert(x >= 0 && x < (unsigned int)array_width);
         utiassert(y >= 0 && y < (unsigned int)array_height);
         return array[x+y*array_newline];

      }

      inline const T &get(const CIVector2& p) const
      {assert(m_state ==zasNormal || m_state == zasRead || m_state == zasWrite);
         utiassert(array != NULL);
         utiassert(p.x >= 0 && p.x < (unsigned int)array_width);
         utiassert(p.y >= 0 && p.y < (unsigned int)array_height);
         return array[p.x+p.y*array_newline];

      }
      inline T &operator()(const CIVector2& p)
      {assert(m_state ==zasNormal || m_state == zasWrite);
         utiassert(array != NULL);
         utiassert(p.x >= 0 && p.x < (unsigned int)array_width);
         utiassert(p.y >= 0 && p.y < (unsigned int)array_height);
         return array[p.x+p.y*array_newline];

      }
      inline const T &operator()(const CIVector2& p) const
      {assert(m_state ==zasNormal || m_state == zasRead || m_state == zasWrite);
         utiassert(array != NULL);
         utiassert(p.x >= 0 && p.x < (unsigned int)array_width);
         utiassert(p.y >= 0 && p.y < (unsigned int)array_height);
         return array[p.x+p.y*array_newline];

      }
      void fill(const T& value)
      {assert(m_state ==zasNormal || m_state == zasWrite);
	utiassert(array != NULL);
	  int s = array_newline*array_height;
          for (int i = 0; i < s; i++) array[i] = value;
       }
};


template<class T> struct SZipArray2DLock
{
    bool m_locked;
    bool m_write;
    bool m_callendwrite;
    bool m_callendread;
    bool m_callendwritetoread;
    TZipArrayState m_oldstate;
    ZipArray2D<T>* m_ziparray;
    SZipArray2DLock(ZipArray2D<T>* ziparray=NULL)
    {
        m_ziparray = ziparray;
        m_locked = false;
        m_write = false;
        m_callendwrite = false;
        m_callendread = false;
        m_oldstate = zasCompressed;
    }
    void setZipArray(ZipArray2D<T>* ziparray)
    {
        assert(m_locked == false);
        m_ziparray = ziparray;
    }

    void lockRead()
    {
        assert(m_locked == false);
        m_locked = true;
        m_write = false;
        m_callendwrite = false;
        m_callendread = false;
        m_callendwritetoread = false;
        m_oldstate = m_ziparray->getState();
        switch (m_oldstate) {
        case zasNormal: {

        }; break;
        case zasCompressed: {
            m_ziparray->beginRead();
            m_callendread = true;

        }; break;
        case zasWrite:{

        }; break;
        case zasRead: {

        }; break;
        default: {
            assert(false);
        }; break;
        }
    }

    void lockWrite()
    {
        assert(m_locked == false);
        m_locked = true;
        m_write = true;
        m_callendwrite = false;
        m_callendread = false;
        m_callendwritetoread = false;
        m_oldstate = m_ziparray->getState();
        switch (m_oldstate) {
        case zasNormal: {

        }; break;
        case zasCompressed: {
            m_ziparray->beginWrite();
            m_callendwrite = true;

        }; break;
        case zasWrite:{

        }; break;
        case zasRead: {
            m_ziparray->endRead();
            m_ziparray->beginWrite();
            m_callendwritetoread = true;

        }; break;
        default: {
            assert(false);
        }; break;
        }
    }

    void unlock()
    {
        assert(m_locked = true);
        m_locked = false;
        if (m_callendread) {
            m_ziparray->endRead();
        }
        if (m_callendwrite) {
            m_ziparray->endWrite();
        }
        if (m_callendwritetoread) {
            m_ziparray->endWrite();
            m_ziparray->beginRead();
        }
    }
    void decompressTo(Array2D<T>& target) {
        assert(m_ziparray != NULL);
        if (m_ziparray->getState() == zasCompressed) {
            m_ziparray->decompressTo(target);
        } else {
            m_ziparray->copyToArray(target);
        }
    }

    ~SZipArray2DLock()
    {
        if (m_locked) {
            assert(false);
            unlock();
        }
    }

};

template <class T>
      class ZipArray : public CUtiOB
{
   public:
     Array<unsigned char> m_compresseddata;
      TZipArrayState m_state;
      T* array;
      int array_size;
     
      inline int size() const
      {
         return array_size;
      }
      inline TZipArrayState getState()
      {
          return m_state;
      }

	  T* getCArray()
	  {
          return array;
	  };
      ZipArray(int size) : CUtiOB()
      {
        m_state = zasNormal;
         array_size = size;
         array = new T[umax(1,array_size)];
	
      }
      
      ZipArray() : CUtiOB()
      {
	m_state = zasNormal;
         array_size = 0;
         array = NULL;
        
      };
      
      ~ZipArray()
      {
         if (array)
            delete [] array;

      }
      inline bool isCompressed()
      {
          return m_state == zasCompressed;
      }

      Array<unsigned char> &getCompressedRaw()
      {
	return m_compresseddata;
      };
      void writeTo(T* ar, int size) const
      {
	utiassert(array != NULL);
          int v = size;
          if (v > array_size) v = array_size;
          for (int i = 0; i < v; i++)
          {
              ar[i] = array[i];

          }

      }
      void endWrite()
      {
	utiassert(m_state == zasWrite);
	utiassert(array != NULL);
	UTI_UINT64 s = array_size*sizeof(T);
	UTI_UINT64 cp = uti_compressBound(array_size*sizeof(T));
	m_compresseddata.setSize(cp);
	uti_compress(m_compresseddata.array, &cp, (unsigned char*)array,s);
	m_compresseddata.setSizeCopy(cp);
	 delete [] array;
	 array = NULL;
	 m_state = zasCompressed;
      };
      void beginWrite()
      {
	if (m_state == zasNormal){
	  m_state = zasWrite;
	  return;
	};
	utiassert(m_state == zasCompressed);
	utiassert(array == NULL && m_compresseddata.size() != 0);
	array = new T[umax(array_size, 1)];
	UTI_UINT64 s = array_size*sizeof(T);
	uti_uncompress((unsigned char*)array, &s, m_compresseddata.array, m_compresseddata.size());
	m_compresseddata.setSize(0);
	m_state = zasWrite;
      };
      
      void endRead()
      {
	utiassert(m_state == zasRead);
	utiassert(array != NULL);
	 delete [] array;
	 array = NULL;
	 m_state = zasCompressed;
      };
      void beginRead()
      {
	if (m_state == zasNormal){
	  m_state = zasRead;
	  return;
	};
	utiassert(m_state == zasCompressed);
	utiassert(array == NULL && m_compresseddata.size() != 0);
	array = new T[umax(1,array_size)];
	UTI_UINT64 s = array_size*sizeof(T);
	uti_uncompress((unsigned char*)array, &s, m_compresseddata.array, m_compresseddata.size());
	m_state = zasRead;
      };

      void decompressTo(Array<T>& target)
      {

    utiassert(m_state == zasCompressed || m_state == zasRead);
    utiassert(m_compresseddata.size() != 0);
    target.setSize(array_size);

    UTI_UINT64 s = array_size*sizeof(T);
    uti_uncompress((unsigned char*)target.array, &s, m_compresseddata.array, m_compresseddata.size());

      };
      void setSize(int size) {
         //utiassert(array == NULL || dynamic == true);
         utiassert(array != NULL || array_size == 0);
	 utiassert(m_state != zasCompressed);
	 utiassert(m_state != zasRead);
          CHECKEXISTANCE; 
         if (array) delete [] array;
         array_size = size;
	 if (array_size == 0) {
	   array = new T[1];
	 } else {
         array = new T[array_size];
	// dynamic = true;
	 };
      };
      void setSizeCopy(int size) {
         //utiassert(array == NULL || dynamic == true);
          CHECKEXISTANCE; 
	  utiassert(array != NULL || array_size == 0);
	 utiassert(m_state != zasCompressed);
	 utiassert(m_state != zasRead);
         T* newarray = new T[umax(1,size)];

         if (array) {
            int mins = umin(array_size, size);
            for (int i = 0; i < mins; i++) {
               newarray[i] = array[i];
            };

            delete [] array;
         }
         array_size = size;
         array = newarray;
      };

      void setSizeCopyHighSpeed(int size) {
         //utiassert(array == NULL || dynamic == true);
          CHECKEXISTANCE;
      utiassert(array != NULL || array_size == 0);
     utiassert(m_state != zasCompressed);
     utiassert(m_state != zasRead);
         T* newarray = new T[umax(1,size)];

         if (array) {
            int mins = umin(array_size, size);
             memcpy(newarray, array, mins*sizeof(T));
            /*for (int i = 0; i < mins; i++) {
               newarray[i] = array[i];
            };*/

            delete [] array;
         }
         array_size = size;
         array = newarray;
      };
     
      inline T &operator[](unsigned int p)
      {
assert(m_state ==zasNormal || m_state == zasWrite);
         utiassert((int)p >= 0 && (int)p < array_size);
         utiassert(array != NULL);
         return array[p];

      }
      inline const T & get(unsigned int p)
      {
assert(m_state ==zasNormal || m_state == zasWrite || m_state == zasRead);
         utiassert((int)p >= 0 && (int)p < array_size);
         utiassert(array != NULL);
         return array[p];

      }
      inline const T &operator[](unsigned int p) const
      {
         utiassert((int)p >= 0 && (int)p < array_size);
         utiassert(array != NULL);
         return array[p];

      }
      void fill(const T& value)
      {assert(m_state ==zasNormal || m_state == zasWrite);
	utiassert(array != NULL);
	 CHECKEXISTANCE; 
          for (int i = 0; i < array_size; i++) array[i] = value;
       }
     void copyToArray(int start, int count, Array<T>& target, int targetstart)
     {
       utiassert(array != NULL);
         utiassert(start >= 0);
		 utiassert(start+count <= array_size);
	 utiassert(targetstart >= 0);
	 utiassert(targetstart+count <= target.size());
	 if (count == 0) return;
	 for (int i = 0; i < count; i++) {
	     target.array[targetstart+i] = array[start+i];
	 };
     }

    // Array& operator=(const Array & str);


}; 

#endif
