/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __ADVMATH__
#define __ADVMATH__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "utiintvector.h"
#include <cmath>



inline CVector2f vRotateVector(const CVector2f& v, float rotate)
{
  CVector2f res;
  rotate = rotate*pi180;
  res.x = v.x*cos(rotate)-v.y*sin(rotate);
  res.y = v.x*sin(rotate)+v.y*cos (rotate);

  return res;
};

inline CVector2 vRotateVector(const CVector2& v, double rotate)
{
  CVector2 res;
  rotate = rotate*pi180;
  res.x = v.x*cos(rotate)-v.y*sin(rotate);
  res.y = v.x*sin(rotate)+v.y*cos (rotate);

  return res;
};
inline float vfrac(float v)
{
   float res;
   res = v-(float)fasterfloor(v);
   return res;
}


inline double vfrac(double v)
{
   double res;
   res = v-(double)fasterfloor(v);
   return res;
}


inline CVector2 vfrac(const CVector2& v)
{
   CVector2 res;
   res.x = v.x-(double)fasterfloor(v.x);
   res.y = v.y-(double)fasterfloor(v.y);
   return res;
}

inline CVector3 vfrac(const CVector3& v)
{
   CVector3 res;
   res.x =v.x-(double)fasterfloor(v.x);
   res.y = v.y-(double)fasterfloor(v.y);
   res.z = v.z-(double)fasterfloor(v.z);
   return res;
}

inline CVector4 vfrac(const CVector4& v)
{
   CVector4 res;
   res.x = v.x-(double)fasterfloor(v.x);
   res.y =v.y- (double)fasterfloor(v.y);
   res.z =v.z- (double)fasterfloor(v.z);
   res.w =v.w- (double)fasterfloor(v.w);
   return res;
}

inline CVector2 vfloor(const CVector2& v)
{
   CVector2 res;
   res.x = (double)fasterfloor(v.x);
   res.y = (double)fasterfloor(v.y);
   return res;
}
inline CVector3 vfloor(const CVector3& v)
{
   CVector3 res;
   res.x =(double)fasterfloor(v.x);
   res.y = (double)fasterfloor(v.y);
   res.z = (double)fasterfloor(v.z);
   return res;
}
inline CVector4 vfloor(const CVector4& v)
{
   CVector4 res;
   res.x = (double)fasterfloor(v.x);
   res.y =(double) fasterfloor(v.y);
   res.z = (double)fasterfloor(v.z);
   res.w = (double)fasterfloor(v.w);
   return res;
}

inline double vfloori(const double& v)
{
   double res;
   res = (double)fasterfloor(v);

   return res;
}


inline float vfloori(const float& v)
{
   float res;
   res = (float)fasterfloor(v);

   return res;
}

inline CIVector2 vfloori(const CVector2& v)
{
   CIVector2 res;
   res.x = fasterfloor(v.x);
   res.y = fasterfloor(v.y);
   return res;
}
inline CIVector3 vfloori(const CVector3& v)
{
   CIVector3 res;
   res.x = fasterfloor(v.x);
   res.y = fasterfloor(v.y);
   res.z = fasterfloor(v.z);
   return res;
}
inline CIVector4 vfloori(const CVector4& v)
{
   CIVector4 res;
   res.x = fasterfloor(v.x);
   res.y = fasterfloor(v.y);
   res.z = fasterfloor(v.z);
   res.w = fasterfloor(v.w);
   return res;
}


inline CIVector2 vfloori(const CVector2f& v)
{
   CIVector2 res;
   res.x = fasterfloor(v.x);
   res.y = fasterfloor(v.y);
   return res;
}
inline CIVector3 vfloori(const CVector3f& v)
{
   CIVector3 res;
   res.x = fasterfloor(v.x);
   res.y = fasterfloor(v.y);
   res.z = fasterfloor(v.z);
   return res;
}
inline CIVector4 vfloori(const CVector4f& v)
{
   CIVector4 res;
   res.x = fasterfloor(v.x);
   res.y = fasterfloor(v.y);
   res.z = fasterfloor(v.z);
   res.w = fasterfloor(v.w);
   return res;
}


inline CVector2f vfrac(const CVector2f& v)
{
   CVector2f res;
   res.x = v.x-(float)fasterfloor(v.x);
   res.y = v.y-(float)fasterfloor(v.y);
   return res;
}

inline CVector3f vfrac(const CVector3f& v)
{
   CVector3f res;
   res.x =v.x-(float)fasterfloor(v.x);
   res.y = v.y-(float)fasterfloor(v.y);
   res.z = v.z-(float)fasterfloor(v.z);
   return res;
}

inline CVector4f vfrac(const CVector4f& v)
{
   CVector4f res;
   res.x = v.x-(float)fasterfloor(v.x);
   res.y =v.y- (float)fasterfloor(v.y);
   res.z =v.z- (float)fasterfloor(v.z);
   res.w =v.w- (float)fasterfloor(v.w);
   return res;
}

inline CVector2f vfloor(const CVector2f& v)
{
   CVector2f res;
   res.x = (float)fasterfloor(v.x);
   res.y = (float)fasterfloor(v.y);
   return res;
}
inline CVector3f vfloor(const CVector3f& v)
{
   CVector3f res;
   res.x =(float)fasterfloor(v.x);
   res.y = (float)fasterfloor(v.y);
   res.z = (float)fasterfloor(v.z);
   return res;
}
inline CVector4f vfloor(const CVector4f& v)
{
   CVector4f res;
   res.x = (float)fasterfloor(v.x);
   res.y =(float) fasterfloor(v.y);
   res.z = (float)fasterfloor(v.z);
   res.w = (float)fasterfloor(v.w);
   return res;
}

/*
template<class T> inline T vinterpolate(const T& v1, const T& v2,  float p)
{
  T res;
  res = v1*(1-p)+v2*p;
  return res;
};

template<class T> inline T vinterpolateCosine(const T& v1, const T& v2, float p)
{
  float ft = p* M_PI;
  ft = (1-std::cos(ft))*0.5;
  T res;
  res = v1*(1-p)+v2*p;
  return res;
};*/

struct SInterpolate
{
  template<class T> inline static T interpolate(const T& v1, const T& v2, float p)
  {
        assert(p >= -0.01 && p < 1.01);
   T res;
  res = v1*(1-p)+v2*p;
  return res;
  };
};

struct SInterpolateHermite
{
  template<class T> inline static T interpolate(const T &v1, const T &v2, float p)
  {
        assert(p >= -0.01 && p < 1.01);
      p = p*p*(3.0-2.0*p);
   T res;
  res = v1*(1-p)+v2*p;
  return res;
  };
};

struct SInterpolateQuinticFunc
{
  template<class T> inline static T interpolate(const T& v1, const T &v2, float p)
  {
        assert(p >= -0.01 && p < 1.01);
      p = p*p*p*(p*(p*6-15)+10);
   T res;
  res = v1*(1-p)+v2*p;
  return res;
  };
};

inline float cubicLerp(float a, float b, float c, float d, float t)
{
    assert(t >= -0.01 && t < 1.01);
   float p = (d-c)-(a-b);
   return t*t*t*p+t*t*((a-b)-p)+t*(c-a)+b;
}
struct SInterpolateCosine
{

  template<class T> inline static T interpolate(T v1, T v2, float p)
  {
   /*float ft = p* M_PI;
  ft = (1-std::cos(ft))*0.5;
  T res;
  res = v1*(1-ft)+v2*ft;
  return res;*/
      assert(p >= -0.01 && p <= 1.01f);
      if (p < 0.5f) {
           float ft = p* M_PI;

           ft = (1-cos_taylor8(ft))*0.5;
           T res;
           res = v1*(1-ft)+v2*ft;
           return res;
           } else {
           float ft = (1.0f-p)* M_PI;

           ft = (1+cos_taylor8(ft))*0.5;
           T res;
           res = v1*(1-ft)+v2*ft;
           return res;
                  }
  };
};

template <class T> struct SDistanceEuklid
{
  inline static float distance(const T& v1, const T& v2)
  {
    return v1.distanceTo(v2);
  };
};

struct SDistanceCombine
{
  inline static float distance(const CVector2f& v1, const CVector2f& v2)
  {
    return 0.5f*v1.distanceTo(v2) + 0.5f*umin(absf(v1.x-v2.x),absf(v1.y-v2.y));
  };
};

struct SDistanceManhatten
{
  inline static float distance(const CVector2f& v1, const CVector2f& v2)
  {
    return absf(v1.x-v2.x)+absf(v1.y-v2.y);
  };
};

struct SDistanceNatural
{
  inline static float distance(const CVector2f& v1, const CVector2f& v2)
  {

      float dx = v1.x-v2.x;
      float dy = v1.y-v2.y;
    return absf(dx)+absf(dy)+dx*dx+dy*dy;
  };
};

struct SDistanceAbs
{
  inline static float distance(const CVector2f& v1, const CVector2f& v2)
  {
    return umax(absf(v1.x-v2.x),absf(v1.y-v2.y));
  };
};

struct SInterpolationClass
    {
      float m_start;
      float m_current;
      float m_end;
      float m_onestep;

      SInterpolationClass(float start, float end, float steps)
      {
	m_start = start;
	m_end = end;
	m_current = m_start;
	m_onestep = (m_end-m_start) /steps;

      };
      void reset()
      {
	m_current = m_start;

      };
      void step(float amount=1.0f)
      {
	m_current += m_onestep* amount;
      }
    };

#endif
