/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __UTI_ALGEBRA__
#define __UTI_ALGEBRA__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "matrices.h"
#include <ginac/ginac.h>


namespace Algebra {

	
class Field;
class Ring;
class Ideal;
class Group;
class Monoid;
class NumberField;

typedef long Integer;

class NumberData : public ReferenceUtiOB
{
public:
	uti_refptr<Ring> m_container;
	uti_refptr<Ring> m_basis_container;
    NumberData();
    void copyTo(CUtiOB* obj) const
    {
        NumberData* d = check_cast<NumberData*>(obj);
        d->m_container = m_container;
        d->m_basis_container = m_basis_container;

    }
    virtual void modDiv(NumberData* divvalue, NumberData* avalue, NumberData* rvalue)=0;
    virtual double toDouble() const=0;
    virtual Integer tryToConvertToInt() const=0;
    virtual NumberData* createNumberCopy() const=0;
    virtual void add(NumberData* data) =0;
    virtual void neg()=0;
    virtual void mul(NumberData* data) =0;
    virtual void divide(NumberData* data) =0;
    virtual void inverse()=0;
	virtual bool isEqualNumber(NumberData* data) const = 0;
	virtual bool isLesser(NumberData* data) const = 0;
};

class GiNacNumberData : public NumberData
{
public:
    GiNaC::ex m_value;
GiNacNumberData()
{
  m_value = 0;
};

GiNacNumberData(Integer iv)
{
  m_value = GiNaC::ex(iv);
};


GiNacNumberData(int iv)
{
  m_value = GiNaC::ex(iv);
};
GiNacNumberData(double iv)
{
  m_value = GiNaC::ex(iv);
};
    void copyTo(CUtiOB* obj) const
    {
        NumberData::copyTo(obj);
        GiNacNumberData* d = check_cast<GiNacNumberData*>(obj);
        d->m_value = m_value;
    }

    virtual void modDiv(NumberData* divvalue, NumberData* avalue, NumberData* rvalue)
    {
        //???? richtig, oder macht der processor das falsch? ????
    //    check_cast<IntegerNumberData*>(avalue)->m_value = m_value/ check_cast<IntegerNumberData*>(divvalue)->m_value;
        //check_cast<IntegerNumberData*>(avalue)->m_value = m__value/ check_cast<IntegerNumberData*>(divvalue)->m_value;
    }

    GiNaC::ex convertValue(NumberData* data) const { return check_cast<GiNacNumberData*>(data)->m_value; };
    virtual double toDouble() const{return GiNaC::ex_to<GiNaC::numeric>(m_value.evalf()).to_double();};
    virtual Integer tryToConvertToInt() const {return GiNaC::ex_to<GiNaC::numeric>(m_value.evalf()).to_long();};
    virtual NumberData* createNumberCopy() const {GiNacNumberData* res = new GiNacNumberData();copyTo(res); return res;};
    virtual void add(NumberData* data)  {m_value += convertValue(data);};
    virtual void neg() { m_value = -m_value;};
    virtual void mul(NumberData* data)  {m_value *= convertValue(data);};
    virtual void inverse() {assert(m_value == -1 || m_value == 1); };
    virtual void divide(NumberData* data){m_value /= convertValue(data);};
    //virtual bool isEqualNumber(NumberData* data) const {return m_value == convertValue(data);};
    virtual bool isEqualNumber(NumberData* data) const {

        return m_value.compare(check_cast<GiNacNumberData*>(data)->m_value)==0;// convertValue(data);
    };
    virtual bool isLesser(NumberData* data) const {

        return m_value.compare(check_cast<GiNacNumberData*>(data)->m_value) < 0;// convertValue(data);
    };
};

/*class IntegerNumberData : public NumberData
{
public:
    Integer m_value;
IntegerNumberData()
{
  m_value = 0;
};

IntegerNumberData(Integer iv)
{
  m_value = iv;
};
    void copyTo(CUtiOB* obj) const
    {
        NumberData::copyTo(obj);
        IntegerNumberData* d = check_cast<IntegerNumberData*>(obj);
        d->m_value = m_value;
    }

    virtual void modDiv(NumberData* divvalue, NumberData* avalue, NumberData* rvalue)
    {
        //???? richtig, oder macht der processor das falsch? ????
        check_cast<IntegerNumberData*>(avalue)->m_value = m_value/ check_cast<IntegerNumberData*>(divvalue)->m_value;
        //check_cast<IntegerNumberData*>(avalue)->m_value = m__value/ check_cast<IntegerNumberData*>(divvalue)->m_value;
    }

    Integer convertValue(NumberData* data) const { return check_cast<IntegerNumberData*>(data)->m_value; };
    virtual double toDouble() const{return m_value;};
    virtual Integer tryToConvertToInt() const {return m_value;};
    virtual NumberData* createNumberCopy() const {IntegerNumberData* res = new IntegerNumberData();copyTo(res); return res;};
    virtual void add(NumberData* data)  {m_value += convertValue(data);};
    virtual void neg() { m_value = -m_value;};
    virtual void mul(NumberData* data)  {m_value *= convertValue(data);};
    virtual void inverse() {assert(m_value == -1 || m_value == 1); };
    virtual void divide(NumberData* data){m_value /= convertValue(data);};
    //virtual bool isEqualNumber(NumberData* data) const {return m_value == convertValue(data);};
	virtual bool isEqualNumber(NumberData* data) const {
		if (!dynamic_cast<IntegerNumberData*>(data)) {
			return true;
		}
		return m_value == check_cast<IntegerNumberData*>(data)->m_value;// convertValue(data);
	};
	virtual bool isLesser(NumberData* data) const {
		if (!dynamic_cast<IntegerNumberData*>(data)) {
			return true;
		}
		return m_value < check_cast<IntegerNumberData*>(data)->m_value;// convertValue(data);
	};
};*/
/*
class FloatNumberData : public NumberData
{
public:
	double m_value;

	FloatNumberData()
	{
		m_value = 0.0;
	};

	FloatNumberData(double iv)
	{
		m_value = iv;
	};
	void copyTo(CUtiOB* obj) const
	{
		NumberData::copyTo(obj);
		FloatNumberData* d = check_cast<FloatNumberData*>(obj);
		d->m_value = m_value;
	}
	double convertValue(NumberData* data) const { return check_cast<FloatNumberData*>(data)->m_value; };
	virtual double toDouble() const{ return m_value; };
	virtual Integer tryToConvertToInt() const { return m_value; };
//	virtual NumberData* createNumberCopy() const { FloatNumberData* res = new FloatNumberData(); copyTo(res); return res; };
	virtual void add(NumberData* data)  { m_value += convertValue(data); };
	virtual void neg() { m_value = -m_value; };
	virtual void mul(NumberData* data)  { m_value *= convertValue(data); };
	virtual void inverse() { assert(m_value != 0); m_value = 1.0 / m_value; };
	virtual void divide(NumberData* data){ m_value /= convertValue(data); };

	virtual bool isEqualNumber(NumberData* data) const {
		if (!dynamic_cast<FloatNumberData*>(data)) {
			return true;
		}
		return m_value == check_cast<FloatNumberData*>(data)->m_value;// convertValue(data);
	};

	virtual bool isLesser(NumberData* data) const {
		if (!dynamic_cast<FloatNumberData*>(data)) {
			return true;
		}
		return m_value < check_cast<FloatNumberData*>(data)->m_value;// convertValue(data);
	};
};*/

class Number;

UMat<Number> getKernel(const UMat<Number>& basis);
void valueToBasis(Number value, const std::vector<Number> &simplebase, UMat<Number>integralmatrix, int line);
bool isIntegralAndGetMinimalPolynomial(Number integralelement, const std::vector<Number> &simplebase, int d, Polynom<Number>& res);

Number ggt(const Number& a, const Number& b);


class Number : public CUtiOB
{
public:
    uti_refptr<NumberData> m_data;
	Polynom<Number> getMinimalPolynomial();
	bool isIrreduzible();
    bool isComplex();
	bool isPrime();
	int getDegree();
    void modDiv(Number divvalue, Number& avalue, Number & rvalue)
    {
        m_data->modDiv(divvalue.m_data.get(), avalue.m_data.get(), rvalue.m_data.get());
    }

    inline Ring* getContainer() const {return m_data->m_container.get();};
    inline Integer tryToConvertToInt() const
    {
        return m_data->tryToConvertToInt();
    }
    inline operator double() const
	{
        return m_data->toDouble();
	};
    inline void singleElement()
    {
        if (m_data->getReferenceCount() != 1) {
            m_data = m_data->createNumberCopy();
        }
    }
    inline Number()
    {
        m_data = NULL;
    }
    explicit inline Number(Integer d)
    {
        m_data = new GiNacNumberData(d);
    }

    explicit inline Number(int d)
    {
        m_data = new GiNacNumberData(d);
    }

    explicit inline Number(double d)
    {
        m_data = new GiNacNumberData(d);
    }

    inline Number(const Number& n)
    {
        m_data = n.m_data;
    }
    inline Number& operator=(const Number& n)
    {
        m_data = n.m_data;
    }


    inline bool operator==(const Number& n)
    {
        return m_data->isEqualNumber(n.m_data.get());
    }

    inline Number operator+(const Number& n) const
    {
        Number res;
        res.m_data = m_data->createNumberCopy();
        res.m_data->add(n.m_data.get());
        return res;
    }
    inline Number operator-(const Number& n) const
    {
        Number res;
        res.m_data = m_data->createNumberCopy();
        Number b2;
        b2.m_data = n.m_data->createNumberCopy();
        b2.m_data->neg();

        res.m_data->add(b2.m_data.get());
        return res;
    }
    inline Number operator-() const
    {

        Number b2;
        b2.m_data = m_data->createNumberCopy();
        b2.m_data->neg();


        return b2;
    }
    inline Number operator*(const Number& n) const
    {
        Number res;
        res.m_data = m_data->createNumberCopy();
        res.m_data->mul(n.m_data.get());
        return res;
    }
    inline Number operator/(const Number& n) const
        {
            Number res;
            res.m_data = m_data->createNumberCopy();
            res.m_data->divide(n.m_data.get());
            return res;
        }

    inline Number &operator+=(const Number& n)
    {
        singleElement();
        m_data->add(n.m_data.get());
        return *this;
    }

    inline Number &operator-=(const Number& n)
    {
        singleElement();
        Number b2;
        b2.m_data = n.m_data->createNumberCopy();
        b2.m_data->neg();
        m_data->add(b2.m_data.get());
        return *this;
    }
    inline Number &operator*=(const Number& n)
    {
        singleElement();
        m_data->mul(n.m_data.get());
        return *this;
    }
    inline Number &operator/=(const Number& n)
    {
        singleElement();
        m_data->divide(n.m_data.get());
        return *this;
    }


};

Number fabs(const Number& number);

/*class RationalNumberData : public NumberData
{
public:
    Number m_numerator;
    Number m_denominator;
RationalNumberData()
{
  m_numerator = Number(0);
  m_denominator = Number(1);
};
RationalNumberData(Integer iv)
{
  m_numerator = Number(iv);
  m_denominator = Number(1);
};

RationalNumberData(Integer iv, Integer divvalue)
{
  m_numerator = Number(iv);
  m_denominator = Number(divvalue);
  assert(m_denominator.isZero() == false);
  assert(m_denominator.isPositiv() == false);
};
    void copyTo(CUtiOB* obj) const
    {
        NumberData::copyTo(obj);
        RationalNumberData* d = check_cast<RationalNumberData*>(obj);
        d->m_numerator = m_numerator;
        d->m_denominator = m_denominator;
    }

    void reduce()
    {
        assert(m_denominator >= 0);
        Number value = ggT(m_numerator, m_denominator);
        if (!value.isOne()) {
            m_numerator /= value;
            m_denominator = value;
        }

    }

   // Integer convertValue(NumberData* data) const {return data->tryToConvertToInt();};
    virtual double toDouble() const{return double(m_numerator) / double(m_denominator);};
    virtual Integer tryToConvertToInt() const {return m_numerator / m_denominator;};
    virtual NumberData* createNumberCopy() const {RationalNumberData* res = new RationalNumberData();copyTo(res); return res;};
    virtual void add(NumberData* data)  {m_value += convertValue(data);};
    virtual void neg() { m_value = -m_value;};
    virtual void mul(NumberData* data)  {m_value *= convertValue(data);};
    virtual void inverse() {assert(m_value == -1 || m_value == 1); };
    virtual bool isEqualNumber(NumberData* data) const {return m_value == convertValue(data);};
};
*/

class IdealData : public ReferenceUtiOB
{
public:

	uti_refptr<Ring> m_owner;
    std::vector<Number> m_generators;
    void copyTo(CUtiOB* obj) const
    {
        IdealData* d = check_cast<IdealData*>(obj);
        d->m_owner = m_owner;
        //d->m_basis_container = m_basis_container;

    }
    virtual void modDiv(IdealData* divvalue, IdealData* avalue, IdealData* rvalue)=0;
    virtual double toDouble() const=0;
    virtual Integer tryToConvertToInt() const=0;
    virtual IdealData* createNumberCopy() const=0;
    virtual void add(IdealData* data) =0;
    virtual void neg()=0;
    virtual void mul(IdealData* data) =0;
    virtual void divide(IdealData* data) =0;
    virtual void inverse()=0;
    virtual bool isEqualNumber(IdealData* data) const = 0;
    virtual bool isLesser(IdealData* data) const = 0;
	
};




class Ideal : public CUtiOB
{
protected:
    //factoring general ideals
    Number getGcdOfNm_Si() const;
    void getRationalPrimesOfGcdOfNm_Si(std::vector<Number>& primes) const;

public:
	uti_refptr<IdealData> m_data;
    bool factor(std::vector<Ideal> m_ideal, std::vector<Number> &powers, std::vector<Number> &norms) const ;
	bool isPrime();
    bool isPrincipal(Number mynorm) const;
    bool contains(Number value);
    bool doesDivide(Ideal otherideal);
    Number getNorm() const;
    int generatorCount() const;
    Ring* getOwner() const {return m_data.get()->m_owner.get();};
    inline void singleElement()
    {
        if (m_data->getReferenceCount() != 1) {
            m_data = m_data->createNumberCopy();
        }
    }
    Number getGenerator(int i) const;
    inline Ideal()
    {
        m_data = NULL;
    }
    inline Ideal(const Ideal& n)
    {
        m_data = n.m_data;
    }
    inline Ideal& operator=(const Ideal& n)
    {
        m_data = n.m_data;
    }


    inline bool operator==(const Ideal& n)
    {
        return m_data->isEqualNumber(n.m_data.get());
    }

    inline Ideal operator+(const Ideal& n) const
    {
        Ideal res;
        res.m_data = m_data->createNumberCopy();
        res.m_data->add(n.m_data.get());
        return res;
    }
    inline Ideal operator-(const Ideal& n) const
    {
        Ideal res;
        res.m_data = m_data->createNumberCopy();
        Ideal b2;
        b2.m_data = n.m_data->createNumberCopy();
        b2.m_data->neg();

        res.m_data->add(b2.m_data.get());
        return res;
    }
    inline Ideal operator-() const
    {

        Ideal b2;
        b2.m_data = m_data->createNumberCopy();
        b2.m_data->neg();


        return b2;
    }
    inline Ideal operator*(const Ideal& n) const
    {
        Ideal res;
        res.m_data = m_data->createNumberCopy();
        res.m_data->mul(n.m_data.get());
        return res;
    }
    inline Ideal operator/(const Ideal& n) const
        {
            Ideal res;
            res.m_data = m_data->createNumberCopy();
            res.m_data->divide(n.m_data.get());
            return res;
        }

    inline Ideal &operator+=(const Ideal& n)
    {
        singleElement();
        m_data->add(n.m_data.get());
        return *this;
    }

    inline Ideal &operator-=(const Ideal& n)
    {
        singleElement();
        Ideal b2;
        b2.m_data = n.m_data->createNumberCopy();
        b2.m_data->neg();
        m_data->add(b2.m_data.get());
        return *this;
    }
    inline Ideal &operator*=(const Ideal& n)
    {
        singleElement();
        m_data->mul(n.m_data.get());
        return *this;
    }
    inline Ideal &operator/=(const Ideal& n)
    {
        singleElement();
        m_data->divide(n.m_data.get());
        return *this;
    }
};

bool isPrime(Integer value);
void getPrimesTillN(Integer n, std::vector<Integer>& values);

	class Ring : public ReferenceUtiOB
	{
	public:
		virtual uti_refptr<Field> getFieldOfFractions();
		virtual bool isUID();
		virtual bool isPID();//Principal Ideal Domain
		virtual bool isEuclideanDomain();
		virtual bool isDedekind();
		virtual bool isNoetherian();
		virtual Number getOne();
		virtual Number getZero();
		virtual Number getFromInteger(Integer value);
		virtual Number getStdBasis(Integer value);
        //virtual  createIdeal(Number number);
		virtual bool isDomain();
		virtual bool isFactorial();
		virtual uti_refptr<Ring> factorRing(Ideal id);  //Erzeuge damit Number field
		virtual uti_refptr<Ring> expandRing(Number number);
		virtual uti_refptr<Group> getAdditiveGroup();
		virtual uti_refptr<Monoid> getMultiplicativeGroup();
		virtual Number lift(Number n) const;
        bool contains(Ideal id) const;
        bool contains(Number number) const;

        Ideal createIdealFromGenerators(std::vector<Number>& basis) const;
        Ideal createIdealFromGenerators(Number number) const;
        virtual void getIntegralBasis(std::vector<Number>& integralbasis) const;
        virtual Polynom<Number> getMinimalPolynom() const;
        virtual Number getNorm_of_k_minus_alpha(Number k) const;

	};



	


	class Field : public Ring
	{
	public:
		virtual Polynom<Number> getMinimalPolynomial(Ring* subfieldorring);
		virtual Polynom<Number> getMinimalPolynomialOfNumber(Number number);
        virtual bool isIrreduzible(Polynom<Number> data);

	};

	class RealField : public Field
	{
	public:
	};

	class ComplexField : public Field
	{
	public:
	};

	class FinitField : public Field
	{
	protected:
		Integer m_prime;
		Integer m_power;
	public:
		FinitField(int p, int n);
		Integer getElementCount();
		Polynom<Number> getMinimalPolynom();
		Number getPrimitiveElement();
		Integer getDegree() const { return m_power; };
		Integer getPrime() const { return m_prime; };
	};

	class RingOfPolynomials : public Ring
	{
	public:
		RingOfPolynomials(Ring* coefficientring);
		uti_refptr<Ring> m_coefficientring;
		bool isSeparabel(Polynom<Number> poly);
		Polynom<Number> lift(Polynom<Number> poly);


	};

	class NumberRing : public Ring
	{ //Extension of Z, not real Number Ring because it need not be an order
    public:
        uti_refptr<NumberField> m_isringofintegersin;
	public:
		static uti_refptr<NumberRing> createNumberRingFromGenerator(Number number);
		static uti_refptr<NumberRing> createNumberRingFromGenerators(const std::vector<Number>& numbers);
		bool isOrderIn(NumberField* field) const;
		Number getDiscriminant();
		virtual Polynom<Number> getMinimalPolynomial();
		virtual void getIntegralBasis(std::vector<Number>& integralbasis) const;
        Ideal createIdealFromGenerators(std::vector<Number>& basis) const;
        uti_refptr<RingOfPolynomials> getRingOfPolynomials() const;
	};

	void factorIntoRationalPrimes(Number num, std::vector<Number>& factors, std::vector < Number>& powers);
	void factorPolynom(Polynom<Number> & poly, std::vector< Polynom<Number> > & g, std::vector<Integer> & e_power);
    Number traceANT(Number num);
    Number normANT(Number num);

    Number normIdeal(Ideal ideal);

    UMat<Number> M_alpha(Number alpha, const std::vector<Number>& basis);
	Number discriminantOfBasis(std::vector<Number>& basis);

	class NumberField : public Field
	{
	protected:
		int m_degree;
		Number m_integralelement;
		std::vector<Number> m_originalgenerators;
		Polynom<Number> m_minimalpolynomial;
	public:
		static uti_refptr<NumberField> createNumberFieldFromGenerators(const std::vector<Number>& numbers);
        virtual uti_refptr<Ring> determineRingOfIntegers() const;
		virtual Number getIntegralElement() const;
		virtual void getConjugates(std::vector<Number>& conjugates) const;
        virtual void getEmbeddings(std::vector<Polynom<Number> >& embeddings) const;
		virtual void getIntegralBasis(std::vector<Number>& integralbasis) const;
		virtual Number getDiscriminantOfRingOfIntegers() const;

        virtual Number getNorm_of_k_minus_alpha(Number k) const;
        virtual void getEmbeddingsData(int& real, int & complex) const;
        virtual int getNumberFieldDegree() const;
        virtual double getMinkowskiConstant() const;
        //Todo: General Ideal Class for Dedekind ring and not only for ring of integers
        virtual void getGenerators_of_Ideal_Class_Group(std::vector<Ideal>& generators) const;
        virtual void getGenerators_of_Unit_Group(Number& firstgroup, std::vector<Number>& generators) const;
	};

    Integer factorial(Integer value)
    {
        Integer res = 1;
        for (Integer k = 1; k <= value; k++) {
            res *= k;
        }
        return res;

    }

	class Monoid : public ReferenceUtiOB
	{
	public:
		Monoid();
		bool abelian() const;
	};

	class Group : public ReferenceUtiOB
	{
	protected:
		uti_refptr<Group> m_parent;
	public:
		Group();
		Integer indexTo(Group* group);
		uti_refptr<Group> createSubGroup(Number generator);
	};

	uti_refptr<Group> getGaloisGroup(Field* L, Field* K);

	
};
#endif
   
