#include "curves.h"
#include <math.h>

void calculateBezierCurve(const CVector2& p1, const CVector2& p2, const CVector2& p3, const CVector2& p4, ArrayAccess<CVector2>& result)
{
  int points = result.getRestSize();
  for (int i = 0; i < points; i++) {
    result.add(calculateBezier4(p1, p2, p3, p4, (float)i/(float)(points-1)));
  };
  
}


void calculateBezierCurveAdaptiv(const CVector2& p1, const CVector2& p2, const CVector2& p3, const CVector2& p4, float pixelsize, float intolerance, ArrayAccess<CVector2>& result)
{
  
  int maxsteps =65536;
  utiassert(result.getRestSize() > maxsteps);
  float onestep = 1.0f/ (float)(maxsteps);
  int currentsteps = maxsteps / 8;
  int courselevel = currentsteps;
  CVector2 lastpoint = p1;
  result.add(lastpoint);
  int currentposition = 0;
  while (currentposition+currentsteps <= maxsteps)
  {
    float lastu = (float)currentposition*onestep;
    float nextu = (float)(currentposition+currentsteps)*onestep;
    CVector2 nextpoint = calculateBezier4(p1, p2, p3, p4, nextu);
    CVector2 midpoint = calculateBezier4(p1, p2, p3, p4, (lastu+nextu)/2.0f);
    CVector2 t1 = nextpoint-lastpoint;
    CVector2 mid = (nextpoint+lastpoint)/2.0f;;
    float pixels1 = t1.normalize() /pixelsize;
    float pixelsmid = mid.distanceTo(midpoint);
    if (pixels1 < 1.8) {
      result.add(nextpoint);
      lastpoint = nextpoint;
      currentposition += currentsteps;
      if (currentsteps < courselevel) {
	 currentsteps *= 2;
      };
      continue;
    };
    if (pixelsmid < intolerance) {
      result.add(nextpoint);
      lastpoint = nextpoint;
      currentposition += currentsteps;
      if (currentsteps < courselevel) {
	 currentsteps *= 2;
      };
      continue;
    } else {
      if (currentsteps > 1) {
	currentsteps /= 2;
	continue;
      } else {
	result.add(nextpoint);
	lastpoint = nextpoint;
      currentposition += currentsteps;
      };
    };
  };
//  result.add(p4);
  
}