#ifndef __ERRORNUMBER__
#define __ERRORNUMBER__

#include <cmath>
#include <assert.h>
#include <iostream>

//#define ERRORDEBUG

template<int bits> struct CFixedULP
{
  double operator()(double d) const {
    return pow(2, -bits)*2.0;
  };
};


struct CFloatError
{
  double operator()(double d) const {
    return pow(2.0, 1-23)*d;
  };
};

struct CDoubleError
{
  double operator()(double d) const {
    return pow(2.0, 1-50)*d;
  };
};


template<typename T, typename ULP>
struct CErrorNumber
{
  protected:

    T m_minvalue;
    T m_maxvalue;
    T m_error;
    ULP my_ulp;
  public:
    typedef CErrorNumber<T, ULP> NumberType;
    inline CErrorNumber() {};
    inline CErrorNumber(const CErrorNumber& f)
    {
      m_minvalue = f.m_minvalue;
      m_maxvalue = f.m_maxvalue;
      m_error = f.m_error;
      utiassert(m_minvalue <= m_maxvalue);

    };
    inline CErrorNumber(int f)
    {
      m_minvalue = f;
      m_maxvalue = f;
      m_error = 0;
    };
    inline CErrorNumber(long f)
    {
      m_minvalue = f;
      m_maxvalue = f;
      m_error = 0;
    };
    inline void set(T p_min, T p_max, T p_error)
    {
      m_minvalue = p_min;
      m_maxvalue = p_max;
      m_error = p_error;
    }
    inline CErrorNumber(T p_min, T p_max, T p_error)
    {
      m_minvalue = p_min;
      m_maxvalue = p_max;
      m_error = p_error;
    }
    inline CErrorNumber(double f, T error=0)
    {
      m_minvalue = f;
      m_maxvalue = f;
      m_error = error;
    };
    inline CErrorNumber(float f, T error=0)
    {
      m_minvalue = f;
      m_maxvalue = f;
      m_error = error;
    };
    inline T minvalue() const {
      return m_minvalue;

    };
    inline T maxvalue() const {
      return m_maxvalue;
    };
    inline T error() const {
       return m_error;
    }

    inline NumberType& operator=(const NumberType & t)
    {
      m_minvalue = t.m_minvalue;
      m_maxvalue = t.m_maxvalue;
      m_error = t.m_error;
      return *this;
    };
    //inline bool operator<(const NumberType& t) const {return m_value < t.m_value;};
    inline bool operator==(const NumberType& t) const {
      return m_minvalue == t.m_minvalue && m_maxvalue == t.m_maxvalue  && m_error = t.m_error;

    };
    //inline bool operator>(const NumberType& t) const {return m_value > t.m_value;};
    //inline bool operator<=(const NumberType& t) const {return m_value <= t.m_value;};
    //inline bool operator>=(const NumberType& t) const {return m_value >= t.m_value;};
    inline bool operator!=(const NumberType& t) const {return !(*this == t);};
    //inline operator bool() const { return (bool) m_value;};
    //inline bool operator!() const {return m_value == 0;};
    inline NumberType& operator-() const
    {
      NumberType result;
      result.m_maxvalue = -m_minvalue;
      result.m_minvalue = -m_maxvalue;
      result.m_error = m_error;
      utiassert(result.m_minvalue <= result.m_maxvalue);
      #ifdef ERRORDEBUG
      std::cout << "result"; result.print(); std::cout << " = op"; this->print();  std::cout << std::endl;
#endif
      return result;

    };
    inline NumberType& operator++() {*this = *this+ 1; return *this;};
    inline NumberType& operator--() {*this = *this- 1;  return *this;};
    inline NumberType operator++(int) const {NumberType result; result = *this; ++result; return result;};
    inline NumberType operator--(int) const {NumberType result; result = *this; --result; return result;};
    inline NumberType& operator+=(const NumberType& t) {*this = *this+t; return *this;};
    inline NumberType& operator-=(const NumberType& t) {*this = *this-t; return *this;};
    inline NumberType operator+(const NumberType& t) const
    {

      NumberType result(*this);
      T myerror = my_ulp(fmax(std::abs(m_minvalue+t.m_minvalue),std::abs(m_maxvalue+t.m_maxvalue)));

      result.m_minvalue += t.m_minvalue-myerror;
      result.m_maxvalue += t.m_maxvalue+myerror;
      utiassert(result.m_minvalue <= result.m_maxvalue);
      result.m_error += t.m_error + myerror;
#ifdef ERRORDEBUG
      std::cout << "result"; result.print(); std::cout << " = op1"; this->print(); std::cout << " + op2";t.print(); std::cout << std::endl;
#endif
      return result;

    };
    inline NumberType operator-(const NumberType& t) const
    {
      NumberType result(*this);
      T myerror = my_ulp(fmax(std::abs(m_minvalue-t.m_maxvalue),std::abs(m_maxvalue-t.m_minvalue)));

      result.m_minvalue -= t.m_maxvalue-myerror;
      result.m_maxvalue -= t.m_minvalue+myerror;
      if (result.m_minvalue >= result.m_maxvalue) {
        std::swap(result.m_minvalue, result.m_maxvalue);
      }
      utiassert(result.m_minvalue <= result.m_maxvalue);
      result.m_error += t.m_error + myerror;
#ifdef ERRORDEBUG
      std::cout << "result"; result.print(); std::cout << " = op1"; this->print(); std::cout << " - op2";t.print(); std::cout << std::endl;
#endif
      return result;

    };
    inline NumberType& operator*=(const NumberType& t)
    {

      *this = *this *t;
      return *this;
    };

    inline NumberType& operator/=(const NumberType& t)
    {
      *this = *this/t;
      return *this;
    };

    inline NumberType operator*(const NumberType& t) const
    {
      NumberType result(*this);
      //result *= t;
      T rmin =fmin(m_minvalue*t.m_minvalue, fmin(m_minvalue*t.m_maxvalue,fmin(m_maxvalue*t.m_minvalue, m_maxvalue*t.m_maxvalue)));
      T rmax = fmax(m_minvalue*t.m_minvalue,
			fmax(m_minvalue*t.m_maxvalue,
				 fmax(m_maxvalue*t.m_minvalue, m_maxvalue*t.m_maxvalue)));
      T myerror = my_ulp(fmax(std::abs(rmin),std::abs(rmax)));
      rmin -= myerror;
      rmax += myerror;
      result.m_minvalue = rmin;
      result.m_maxvalue = rmax;
      result.m_error =fmax(std::abs(m_minvalue),std::abs(m_maxvalue))*t.m_error+
         fmax(std::abs(t.m_minvalue),std::abs(t.m_maxvalue))*m_error+m_error*t.m_error+myerror;
#ifdef ERRORDEBUG

      std::cout << "result"; result.print(); std::cout << " = op1"; this->print(); std::cout << " * op2";t.print(); std::cout << std::endl;
#endif
      return result;

    };
    inline NumberType operator/(const NumberType& t) const
    {
      NumberType result(*this);
      //result *= t;
      //TODO: Was ist mit der 0?
      //TODO: In Ausarbeitung Fehler?
      T rmin = fmin(m_minvalue/t.m_minvalue, fmin(m_minvalue/t.m_maxvalue, fmin(m_maxvalue/t.m_minvalue, m_maxvalue/t.m_maxvalue)));
      T rmax = fmax(m_minvalue/t.m_minvalue,
			fmax(m_minvalue/t.m_maxvalue,
				 fmax(m_maxvalue/t.m_minvalue, m_maxvalue/t.m_maxvalue)));
      T myerror = my_ulp(fmax(std::abs(rmin),std::abs(rmax)));
      rmin -= myerror;
      rmax += myerror;
      result.m_minvalue = rmin;
      result.m_maxvalue = rmax;
      T thismax = fmax(std::abs(m_minvalue),std::abs(m_maxvalue));
      T tmax = fmax(std::abs(t.m_minvalue),std::abs(t.m_maxvalue));
      result.m_error = fmax(0, (tmax*m_error-thismax*t.m_error)/(tmax*tmax))+myerror;
      #ifdef ERRORDEBUG
      std::cout << "result"; result.print(); std::cout << " = op1"; this->print(); std::cout << " / op2";t.print(); std::cout << std::endl;
#endif
      return result;
    };

    inline NumberType& operator*=(int t)
    {
      *this *= NumberType(t); return *this;
    };

    inline NumberType& operator/=(int t)
    {
      *this /= NumberType(t);
      return *this;
    };

    inline NumberType operator*(int t) const {NumberType result(*this);  result *= t;return result;};
    inline NumberType operator/(int t) const {NumberType result(*this); result /= t; return result;};

    inline NumberType& operator*=(long t)
    {
      *this *= NumberType(t); return *this;
    };

    inline NumberType& operator/=(long t)
    {
      *this /= NumberType(t);
      return *this;
    };

    inline NumberType operator*(long t) const {NumberType result(*this);  result *= t;return result;};
    inline NumberType operator/(long t) const {NumberType result(*this); result /= t; return result;};
    //inline NumberType& operator>>=(size_t shift) {m_value >>= shift; return *this;};
    //inline NumberType& operator<<=(size_t shift) {m_value <<= shift; return *this;};
    //inline NumberType operator>>(size_t shift) const {NumberType result(*this);  result >>= shift;return result;};
    //inline NumberType operator<<(size_t shift) const {NumberType result(*this); result <<=shift; return result;};
    void print() const
    {
      std::cout << "(min = " << m_minvalue << ", max = " << m_maxvalue << ", error = " << m_error << ")";
    };

};
typedef CErrorNumber<double, CFloatError> CDefEN;

#endif
