/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef  __LU__
#define  __LU__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "UtiType.h"

#include "matrices.h"
#include <cmath>
#include <iostream>

class CUtiSingularException : public CUtiException
  {}
;

template<class T> bool gaussAlgorithmNoChange(const UMat<T>& a, UMat<T>& b)
{
  UMat<T> a2 = a;
  return gaussAlgorithm(a2, b);
};


template<class T> UMat<T> invert(const UMat<T>& a)
{
  utiassert(a.n() == a.m());
  
  int n = a.n();

  UMat<T> a2 = a;
  UMat<T> b = UMat<T>::identity(n,n);
  if (!gaussAlgorithm(a2, b)) {
    throw CUtiSingularException();
  };
  return b;
};

template<class T> bool gaussAlgorithm(UMat<T>& a, UMat<T>& b)
{
  utiassert(a.n() == a.m());
  utiassert(b.n() == a.n());
  int n = a.n();
  int m = b.m();
  Array<int> xindex(n);
  Array<int> yindex(n);
  Array<int> solutionarray(n);
  Array<int> temparray(n);
  for (int i = 0; i < n; i++) {
     xindex[i] = i;
     yindex[i] = i;
  };
  T maximum = a(yindex[0], xindex[0]);
  T testvalue;
  CIVector2 maxvalue(0,0);
  for (int column=0; column < n-1; column++) {
    //Eliminate one column after another
    
    //Search Pivot;
    maxvalue = CIVector2(column, column);
    maximum = absf(a(yindex[maxvalue.y], xindex[maxvalue.x]));
    for (int yvalue = column; yvalue < n; yvalue++) {
       for (int xvalue = column; xvalue < n; xvalue++) {
	 testvalue = absf(a(yindex[yvalue], xindex[xvalue]));
	 if (testvalue > maximum) {
	   maximum = testvalue;
	   maxvalue = CIVector2(xvalue, yvalue);
	 };
       };
    };
    if (maximum < 0.0001f) return false;
  //  std::wcout << "Pivot: " << yindex[maxvalue.y] << ", " << xindex[maxvalue.x] << std::endl;
    //Exchange Elements
    if (maxvalue.x != column) {
      utiswap(xindex[maxvalue.x], xindex[column]);
    };
    
    if (maxvalue.y != column) {
      utiswap(yindex[maxvalue.y], yindex[column]);
    };
    //Pivot now in (column, column)
    
    //Calc
    T pivot = a(yindex[column], xindex[column]);
    T pivotinverse = T(1)/pivot;
    for (int y_row = column+1; y_row < n; y_row++) {
      T eliminate = a(yindex[y_row], xindex[column]);
      T scal = -pivotinverse*eliminate;
      
#define MOVECOLUMN(scal, targety, sourcey) for (int x_column =0; x_column < n; x_column++) {\
	a.set(yindex[targety], xindex[x_column], a(yindex[targety], xindex[x_column])+scal*a(yindex[sourcey], xindex[x_column]));\
      }; \
      for (int x_column = 0; x_column < m; x_column++) {\
	b.set(yindex[targety], x_column, b(yindex[targety], x_column)+scal*b(yindex[sourcey], x_column));\
      };
#define MULCOLUMN(scal, targety) for (int x_column =0; x_column < n; x_column++) {\
	a.set(yindex[targety], xindex[x_column], scal*a(yindex[targety], xindex[x_column]));\
      }; \
      for (int x_column = 0; x_column < m; x_column++) {\
	b.set(yindex[targety], x_column, scal*b(yindex[targety], x_column));\
      };
#define SWAPCOLUMNNOINDEX( t1, t2) \
      for (int x_column = 0; x_column < m; x_column++) {\
	temp = b(t1, x_column);\
	b.set(t1, x_column,  b(t2, x_column));\
	b.set(t2, x_column, temp);\
      };
      MOVECOLUMN(scal, y_row, column);
     /* for (int x_column =0; x_column < n; x_column++) {
	a.set(yindex[y_row], xindex[x_column], a(yindex[y_row], xindex[x_column])+scal*a(yindex[column], xindex[x_column]));
      }; 
      for (int x_column = 0; x_column < m; x_column++) {
	b.set(yindex[y_row], x_column, b(yindex[y_row], x_column)+scal*b(yindex[column], x_column));
      };*/
      
      
      a.set(yindex[y_row], xindex[column], 0/*scal*/);
   
    };   
   // std::wcout <<a.toString() << std::endl;
  // std::wcout <<b.toString() << std::endl;
  };
  
   
  //Fertig machen
 // std::wcout << "------" << std::endl;
  for (int row=n-1; row >= 0; row--) {
    for (int column = n-1; column > row; column--) {
      T pivot = a(yindex[column], xindex[column]);
      T eliminate = a(yindex[row], xindex[column]);
      if (absf(pivot) < 0.0001f) return false;
      T scal = -T(1)/pivot*eliminate;
      MOVECOLUMN(scal, row, column);
      a.set(yindex[row], xindex[column], 0.0f);
    };
   // std::wcout <<a.toString() << std::endl;
   // std::wcout <<b.toString() << std::endl;
    {
      T pivot = a(yindex[row], xindex[row]);
      T scal = T(1)/pivot;
      MULCOLUMN(scal, row);
      solutionarray[xindex[row]] =yindex[row] ;
    };
   // std::wcout <<a.toString() << std::endl;
   // std::wcout <<b.toString() << std::endl;
  }
  /*for (int i = 0; i < n; i++) {
    temparray[i] = xindex[yindex[i]];
  };
  for (int i = 0; i < n; i++) {
    xindex[i] = temparray[i];
  };*/
  for (int i = 0; i < n; i++) {
    temparray[solutionarray[i]] = i;
  };
  T temp;
  for (int row = 0; row < n; row++) {
    if (solutionarray[row] != row) {
      int pos = solutionarray[row];
      utiassert(temparray[pos] == row);
      utiassert(pos > row);
      SWAPCOLUMNNOINDEX(row, pos);
      utiswap(temparray[pos], temparray[row]);
      utiassert(temparray[row] == row);
      solutionarray[row] = row;
     solutionarray[temparray[pos]] = pos;
    };
    
  };
  
  
  return true;
};

enum TSimplexReturn {srFoundSolution, srUnlimited, srNoSolution, srDegenerate};

template<class T> TSimplexReturn slowSimplex(const UMat<T>& A, const UMat<T>& b, const UMat<T>& c, UMat<T>& result)
{
  int varcount = A.m();
  int conditions = A.n();
  utiassert(conditions < varcount);
  utiassert(b.n() == conditions);
  utiassert(b.m() == 1);
  utiassert(c.n() == varcount);
  utiassert(c.m() == 1);
  
  UMat<int> xsorted(1, A.m());
  //The first condition vars are the current vars
  for (int i = 0; i < varcount; i++) {
     if (i < conditions) {
        xsorted.set(i, varcount-conditions+i); 
     } else {
        xsorted.set(i, i-conditions);
     };
  }
  UMAT_SETROW5(xsorted,0,0,   0,3,4,1,2);
  std::wcout << "xsorted = " << std::endl << xsorted.toString() << std::endl;
  int steps = 0;
  UMat<T> xcurrent;
  do {
    steps++;
    //Find x?
    std::wcout << "xsorted = " << std::endl << xsorted.toString() << std::endl;
    UMat<T> BN;
    BN = A(UALL, xsorted(UALL, UMat<int>::steps(0, conditions)));
    xcurrent = b;
    gaussAlgorithmNoChange(BN, xcurrent);
    std::wcout << "x = " << std::endl << xcurrent.toString() << std::endl;
    
    UMat<T> ymat;
    ymat = A(UALL, xsorted(UALL, UMat<int>::steps(conditions, varcount-conditions)));
     std::wcout << "arest = " << std::endl << ymat.toString() << std::endl;
    gaussAlgorithmNoChange(BN, ymat);
    std::wcout << "y = " << std::endl << ymat.toString() << std::endl;
    UMat<T> localc =  c(xsorted(UALL, UMat<int>::steps(0, conditions)), UALL);
    std::wcout << "localc = " << std::endl << localc.toString() << std::endl;
    UMat<T> z(1, ymat.m());
    
    for (int currentz = 0; currentz < ymat.m(); currentz++) {
      z.set(0,currentz, ymat(UALL, UMat<int>(currentz)).transpose().multiply(localc).getSingleValue());
    };
    std::wcout << "z = " << std::endl << z.toString() << std::endl;
    UMat<T> otherc =  c(xsorted(UALL, UMat<int>::steps(conditions, varcount-conditions)), UALL).transpose();
    bool isfinalsolution = true;
    
    bool foundminvalue = false;
    int minj = 0;
    int mini  = 0;
    T minvalue = 0;
    
    for (int j = 0; j < otherc.m();j++) {
       T zminc = z(0,j)-otherc(0,j);
       if (zminc >= 0) {
	 continue;
       };
       isfinalsolution = false;
       bool isinfinite = true;
       T abortminvalue  = 10000000.0f;
       for (int i = 0; i < ymat.n(); i++) {
	 if (ymat(i,j) <= 0) {
	  continue; 
	 }
	 isinfinite = false;
	 T xdiv = xcurrent(i,0)/ymat(i,j);
	 utiassert(xdiv >= 0.0f);//??? ist das mathematisch immer erfüllt.
	 abortminvalue = umin(xdiv, abortminvalue);
	 T checkminvalue = xdiv/**zminc*/;
	 std::wcout << "(i,j) = (" << i <<", " << j+conditions << ") = " << checkminvalue << std::endl;
	 if (foundminvalue == false || checkminvalue < minvalue) {
	   minj = j+conditions;
	   mini = i;
	   minvalue = checkminvalue;
	   foundminvalue = true;
	 };
       };
       if (absf(abortminvalue) < 0.0001) {
	 return srDegenerate;
       }
       if (isinfinite) {
	 return srUnlimited;
       }
    };
    if (isfinalsolution) {
      //extract x values
      result = zeros<T>(varcount, 1);
      for (int i = 0; i < conditions; i++) {
	result.set(xsorted(0,i), 0, xcurrent(i, 0));
      };
      return srFoundSolution;
    };
    if (foundminvalue == false) {
      return srDegenerate;
    };
    std::wcout << "Swap: " << minj << " <=> " << mini << std::endl;
    xsorted.swap(0,minj, 0, mini);
    
  } while(steps < 1000);
  
  return srNoSolution;
};

#endif
