/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __MATRICES__
#define __MATRICES__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "utimathbox.h"
#include <vector>
#include "utivectordouble.h"
#include "utiarray.h"

#define MATH_SINGLE 0
#define MATH_RANGE 1
#define MATH_RANGESSE 2

#define MATH_MODE MATH_RANGE

template<class T> struct CUtiSSEVector
{
protected:
  T m_data[4];
public:
  explicit inline CUtiSSEVector()
  {

  }
 inline CUtiSSEVector(const CUtiSSEVector<T>& v)
  {
   m_data[0] = v.m_data[0];
   m_data[1] = v.m_data[1];
   m_data[2] = v.m_data[2];
   m_data[3] = v.m_data[3];
  }
  explicit inline CUtiSSEVector(const T& v)
  {
   m_data[0] = v;
   m_data[1] = v;
   m_data[2] = v;
   m_data[3] = v;
  }
  inline ~CUtiSSEVector()
  {


  }
  inline CUtiSSEVector<T>& operator=(const CUtiSSEVector<T>& v)
  {
    m_data[0] = v.m_data[0];m_data[1] = v.m_data[1]; m_data[2] = v.m_data[2]; m_data[3] = v.m_data[3];
   return *this;
  }
  inline CUtiSSEVector<T>& operator+=(const CUtiSSEVector<T>& v)
  {
    m_data[0] += v.m_data[0];m_data[1] += v.m_data[1]; m_data[2] += v.m_data[2]; m_data[3] += v.m_data[3];
   return *this;
  }
  inline CUtiSSEVector<T>& operator-=(const CUtiSSEVector<T>& v)
  {
    m_data[0] -= v.m_data[0];m_data[1] -= v.m_data[1]; m_data[2] -= v.m_data[2]; m_data[3] -= v.m_data[3];
   return *this;
  }
  inline CUtiSSEVector<T>& operator*=(const CUtiSSEVector<T>& v)
  {
    m_data[0] *= v.m_data[0];m_data[1] *= v.m_data[1]; m_data[2] *= v.m_data[2]; m_data[3] *= v.m_data[3];
   return *this;
  }
  inline CUtiSSEVector<T>& operator/=(const CUtiSSEVector<T>& v)
  {
    m_data[0] /= v.m_data[0];m_data[1] /= v.m_data[1]; m_data[2] /= v.m_data[2]; m_data[3] /= v.m_data[3];
   return *this;
  }
  inline CUtiSSEVector<T>& operator=(const T& v)
  {
     m_data[0] = v; m_data[1] = v; m_data[2] = v;m_data[3] = v;
  }
  inline void set(const T& v1,const T& v2,const T& v3,const T& v4)
  {
     m_data[0] = v1; m_data[1] = v2; m_data[2] = v3;m_data[3] = v4;
  }
  inline void set(const T& v1)
  {
     m_data[0] = v1; m_data[1] = v1; m_data[2] = v1;m_data[3] = v1;
  }
  inline void loadAligned(T* data)
  {
    m_data[0] = data[0]; m_data[1] = data[1]; m_data[2] = data[2];m_data[3] = data[3];
  };
  inline void loadUnAligned(T* data)
  {
    m_data[0] = data[0]; m_data[1] = data[1]; m_data[2] = data[2];m_data[3] = data[3];
  };
  inline void storeAligned(T* data)
  {
    data[0] = m_data[0]; data[1] = m_data[1]; data[2] = m_data[2];data[3] = m_data[3];
  };
  inline void storeUnAligned(T* data)
  {
    data[0] = m_data[0]; data[1] = m_data[1]; data[2] = m_data[2];data[3] = m_data[3];
  };
  inline T operator[](int x) {
    utiassert(x >= 0 && x < 3);
    return m_data[x];
  };
};

//#ifdef MATH_MODE == MATH_RANGESSE
//#include <xmmintrin.h>
//#endif

template<class T> class UMat;

typedef UMat<int> UIntMat;

template<class T> inline T conj(const T& v)
{
  return v;
};

template<class T> inline bool needConj()
{
  return false;
};

inline void calcSSE(int v, int& sse, int & normal)
{
  sse = v >> 2;
  normal = v & 3;
};

inline void calcSSE(int v, int& sse, int & normal, int & ssefull)
{
  sse = v >> 2;
  normal = v & 3;
  ssefull = normal == 0 ? sse : sse+1;
};

template<class T> class BaseMatInternData : public ReferenceUtiOB
{
public:
   BaseMatInternData() : ReferenceUtiOB() {

   }
   ~BaseMatInternData(){}
   virtual int sizex()
   {
       return 0;
   }
   virtual int sizey()
   {
       return 0;
   }
   virtual BaseMatInternData<T>* makeUnique();
   virtual T getData(int y, int x) const
   {
       return T();
   }
   virtual void setData(int y, int x, const T& value)
   {

   }

   virtual BaseMatInternData<T>* getData(const UIntMat&x, const UIntMat&y)const;
   virtual void setData(const UIntMat&access, BaseMatInternData* data)
   {

   }
   virtual void readRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {
     utiassert(source.m_min.x >= 0 && source.m_max.x < sizex());
     utiassert(source.m_min.y >= 0 && source.m_max.y < sizey());
     utiassert(source.m_min.x < source.m_max.x);
     utiassert(source.m_min.y < source.m_max.y);
     int pos=0;

    if (dataxoriented) {
       for (int j = source.m_min.y; j<source.m_max.y; j++) {
          for (int i = source.m_min.x; i < source.m_max.x; i++) {
           d[pos] = getData(j,i);
	  pos++;
	  }
       };
       } else {

          for (int i = source.m_min.x; i < source.m_max.x; i++) {
	    for (int j = source.m_min.y; j<source.m_max.y; j++) {
           d[pos] = getData(j,i);
	  pos++;
	  }
       };
      };

   };
   virtual void writeRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {
     utiassert(source.m_min.x >= 0 && source.m_max.x < sizex());
     utiassert(source.m_min.y >= 0 && source.m_max.y < sizey());
     utiassert(source.m_min.x < source.m_max.x);
     utiassert(source.m_min.y < source.m_max.y);
     int pos=0;
     if (dataxoriented) {
     for (int j = source.m_min.y; j<source.m_max.y; j++) {
        for (int i = source.m_min.x; i < source.m_max.x; i++) {
           setData(j,i, d[pos]);
	  pos++;
	}
     };
     } else {

         for (int i = source.m_min.x; i < source.m_max.x; i++) {
	   for (int j = source.m_min.y; j<source.m_max.y; j++) {
            setData(j,i, d[pos]);
	  pos++;
	 }
       };
     };
   };
   /*
   virtual void readRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     utiassert(source.m_min.x >= 0 && source.m_max.x < sizex());
     utiassert(source.m_min.y >= 0 && source.m_max.y < sizey());
     utiassert(source.m_min.x < source.m_max.x);
     utiassert(source.m_min.y < source.m_max.y);
     int pos=0;
     int datapos=0;
     int sse;
     int normal;


       if (dataxoriented) {
	 calcSSE(source.m_max.x - source.m_min.x, sse, normal);
         for (int j = source.m_min.y; j<source.m_max.y; j++) {
	   datapos = source.m_min.x;
	   for (int i = 0; i < sse; i++) {
	      d[pos].set(getData(j,datapos), getData(j, datapos+1), getData(j, datapos+2), getData(j, datapos+3));
	      pos++;
	      datapos+= 4;
	   };
	   switch (normal){
	     case 0: { }; break;
	     case 1: {d[pos].set(getData(j, datapos),0,0,0);pos++; }; break;
	     case 2: {d[pos].set(getData(j, datapos),getData(j, datapos+1),0,0);pos++; }; break;
	     case 3: {d[pos].set(getData(j, datapos),getData(j, datapos+1),getData(j, datapos+2),0);pos++; }; break;
	   };
        };
       } else {
	  calcSSE(source.m_max.y - source.m_min.y, sse, normal);
	 for (int i = source.m_min.x; i < source.m_max.x; i++) {
	    datapos = source.m_min.y;
	   for (int j = 0; j < sse; j++) {
	      d[pos].set(getData(datapos, i), getData(datapos+1, i), getData(datapos+2,i), getData(datapos+3,i));
	      pos++;
	      datapos+= 4;
	   };
	   switch (normal){
	     case 0: { }; break;
	     case 1: {d[pos].set(getData(datapos,i),0,0,0);pos++; }; break;
	     case 2: {d[pos].set(getData(datapos,i),getData(datapos+1,i),0,0);pos++; }; break;
	     case 3: {d[pos].set(getData(datapos, i),getData(datapos+1,i),getData(datapos+2,i),0);pos++; }; break;
	   };

         };
       }

   };
   virtual void writeRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     utiassert(source.m_min.x >= 0 && source.m_max.x < sizex());
     utiassert(source.m_min.y >= 0 && source.m_max.y < sizey());
     utiassert(source.m_min.x < source.m_max.x);
     utiassert(source.m_min.y < source.m_max.y);
     int pos=0;
     int datapos=0;
     int sse;
     int normal;


       if (dataxoriented) {
	 calcSSE(source.m_max.x - source.m_min.x, sse, normal);
         for (int j = source.m_min.y; j<source.m_max.y; j++) {
	   datapos = source.m_min.x;
	   for (int i = 0; i < sse; i++) {
	      CUtiSSEVector<T>& v1 = d[pos];
	      setData(j, datapos, v1[0]);setData(j, datapos+1, v1[1]);setData(j, datapos+2, v1[2]);setData(j, datapos+3, v1[3]);
	      pos++;
	      datapos+= 4;
	   };
	   switch (normal){
	     case 0: { }; break;
	     case 1: {CUtiSSEVector<T>& v1 = d[pos]; setData(j, datapos, v1[0]);pos++; }; break;
	     case 2: {CUtiSSEVector<T>& v1 = d[pos]; setData(j, datapos, v1[0]);setData(j, datapos+1, v1[1]);pos++; }; break;
	     case 3: {CUtiSSEVector<T>& v1 = d[pos]; setData(j, datapos, v1[0]);setData(j, datapos+1, v1[1]);setData(j, datapos+2, v1[2]);pos++; }; break;
	   };
        };
       } else {
	  calcSSE(source.m_max.y - source.m_min.y, sse, normal);
	 for (int i = source.m_min.x; i < source.m_max.x; i++) {
	    datapos = source.m_min.y;
	   for (int j = 0; j < sse; j++) {
	      CUtiSSEVector<T>& v1 = d[pos];
	      setData(datapos, i, v1[0]);setData(datapos+1, i, v1[1]);setData(datapos+2, i, v1[2]);setData(datapos+3, i, v1[3]);
	      pos++;
	      datapos+= 4;
	   };
	   switch (normal){
	     case 0: { }; break;
	     case 1: {CUtiSSEVector<T>& v1 = d[pos];setData(datapos, i, v1[0]);pos++; }; break;
	     case 2: {CUtiSSEVector<T>& v1 = d[pos];setData(datapos, i, v1[0]);setData(datapos+1, i, v1[1]);pos++; }; break;
	     case 3: {CUtiSSEVector<T>& v1 = d[pos];setData(datapos, i, v1[0]);setData(datapos+1, i, v1[1]);setData(datapos+2, i, v1[2]);pos++; }; break;
	   };

         };
       }
   };*/

};




template<class T> class Mat_SimpleData : public BaseMatInternData<T>
{
protected:
    Array2D<T> m_data;
public:
    Mat_SimpleData(float v)
    {
        m_data.setSize(1,1);
        m_data(0,0)=v;
    }
    Mat_SimpleData(int ysize, int xsize)
    {
        m_data.setSize(xsize, ysize);
    }
    Mat_SimpleData(const Mat_SimpleData& data)
    {
        m_data.copyFrom(data.m_data);

    }
    virtual BaseMatInternData<T>* makeUnique()
   {
       if (this->m_references <= 1)return NULL;
       return BaseMatInternData<T>::makeUnique();
   }
    virtual int sizex()
   {
       return m_data.width();
   }
   virtual int sizey()
   {
       return m_data.height();
   }
   T getData(int y, int x) const
   {
       return m_data(x,y);
   }
   virtual void setData(int y, int x, const T& value)
   {
       m_data(x,y) = value;
   }
   virtual void readRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {
     utiassert(source.m_min.x >= 0 && source.m_max.x < sizex());
     utiassert(source.m_min.y >= 0 && source.m_max.y < sizey());
     utiassert(source.m_min.x < source.m_max.x);
     utiassert(source.m_min.y < source.m_max.y);
     int pos=0;

    if (dataxoriented) {
      utiassert(source.m_max.x <= m_data.width());
      int c = source.m_max.x-source.m_min.x;
      for (int j = source.m_min.y; j<source.m_max.y; j++) {
	  T* src = m_data.getPointer(source.m_min.x,j);

          for (int i = 0; i < c; i++) {
           d[pos] = src[i];
	  pos++;
	  }
       };
       } else {
	  int w = source.width();
	  int c = source.m_max.y-source.m_min.y;
          for (int i = source.m_min.x; i < source.m_max.x; i++) {
	    T* src = m_data.getPointer(i, source.m_min.y);
	    for (int i = 0; i < c; i++) {
               d[pos] = *src;
	       src += w;
	       pos++;
	     }

          };
      };

   };
   virtual void writeRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {
     utiassert(source.m_min.x >= 0 && source.m_max.x < sizex());
     utiassert(source.m_min.y >= 0 && source.m_max.y < sizey());
     utiassert(source.m_min.x < source.m_max.x);
     utiassert(source.m_min.y < source.m_max.y);
     int pos=0;
     if (dataxoriented) {
      utiassert(source.m_max.x <= m_data.width());
      int c = source.m_max.x-source.m_min.x;
      for (int j = source.m_min.y; j<source.m_max.y; j++) {
	  T* src = m_data.getPointer(source.m_min.x,j);

          for (int i = 0; i < c; i++) {
          src[i] = d[pos];
	  pos++;
	  }
       };
       } else {
	  int w = source.width();
	  int c = source.m_max.y-source.m_min.y;
          for (int i = source.m_min.x; i < source.m_max.x; i++) {
	    T* src = m_data.getPointer(i, source.m_min.y);
	    for (int i = 0; i < c; i++) {
               *src = d[pos];
	       src += w;
	       pos++;
	     }

          };
      };
   };

  /* virtual void readRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     utiassert(source.m_min.x >= 0 && source.m_max.x < sizex());
     utiassert(source.m_min.y >= 0 && source.m_max.y < sizey());
     utiassert(source.m_min.x < source.m_max.x);
     utiassert(source.m_min.y < source.m_max.y);
     int pos=0;
     int datapos=0;
     int sse;
     int normal;


       if (dataxoriented) {
	 calcSSE(source.m_max.x - source.m_min.x, sse, normal);
         for (int j = source.m_min.y; j<source.m_max.y; j++) {
	   datapos = source.m_min.x;
	   T* src = m_data.getPointer(source.m_min.x,j);
	   for (int i = 0; i < sse; i++) {
	      d[pos].loadAligned(src);
	      pos++;
	      src+= 4;
	   };
	   switch (normal){
	     case 0: { }; break;
	     case 1: {d[pos].set(src[0],0,0,0);pos++; }; break;
	     case 2: {d[pos].set(src[0],src[1],0,0);pos++; }; break;
	     case 3: {d[pos].set(src[0],src[1],src[2],0);pos++; }; break;
	   };
        };
       } else {
	  calcSSE(source.m_max.y - source.m_min.y, sse, normal);
	  int w = source.width();
	 for (int i = source.m_min.x; i < source.m_max.x; i++) {
	    datapos = source.m_min.y;
	    T* src = m_data.getPointer(i, source.m_min.y);

	   for (int j = 0; j < sse; j++) {
	      d[pos].set(src[0],src[w], src[w*2], src[w*3]);
	      pos++;
	      src+= 4*w;
	   };
	   switch (normal){
	     case 0: { }; break;
	     case 1: {d[pos].set(src[0],0,0,0);pos++; }; break;
	     case 2: {d[pos].set(src[0],src[w],0,0);pos++; }; break;
	     case 3: {d[pos].set(src[0],src[w],src[w*2],0);pos++; }; break;
	   };

         };
       }

   };
   virtual void writeRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     utiassert(source.m_min.x >= 0 && source.m_max.x < sizex());
     utiassert(source.m_min.y >= 0 && source.m_max.y < sizey());
     utiassert(source.m_min.x < source.m_max.x);
     utiassert(source.m_min.y < source.m_max.y);
     int pos=0;
     int datapos=0;
     int sse;
     int normal;


       if (dataxoriented) {
	 calcSSE(source.m_max.x - source.m_min.x, sse, normal);
         for (int j = source.m_min.y; j<source.m_max.y; j++) {
	   datapos = source.m_min.x;
	   T* src = m_data.getPointer(source.m_min.x,j);
	   for (int i = 0; i < sse; i++) {
	      d[pos].storeAligned(src);
	      pos++;
	      src+= 4;
	   };
	    switch (normal){
	     case 0: { }; break;
	     case 1: {CUtiSSEVector<T>& v1 = d[pos]; src[0] = v1[0];pos++; }; break;
	     case 2: {CUtiSSEVector<T>& v1 = d[pos]; src[0] = v1[0];src[1] = v1[1];pos++; }; break;
	     case 3: {CUtiSSEVector<T>& v1 = d[pos]; src[0] = v1[0];src[1] = v1[1];src[2] = v1[2];pos++; }; break;
	   };
        };
       } else {
	  calcSSE(source.m_max.y - source.m_min.y, sse, normal);
	  int w = source.width();
	 for (int i = source.m_min.x; i < source.m_max.x; i++) {
	    datapos = source.m_min.y;
	    T* src = m_data.getPointer(i, source.m_min.y);

	   for (int j = 0; j < sse; j++) {
	     CUtiSSEVector<T>& v1 = d[pos];
	     src[0] = v1[0];
	     src[w] = v1[1];
	     src[w*2] = v1[2];
	     src[w*3] = v1[3];
	      pos++;
	      src+= 4*w;
	   };
	   switch (normal){
	     case 0: { }; break;
	     case 1: {CUtiSSEVector<T>& v1 = d[pos]; src[0] = v1[0];pos++; }; breakd
	     case 2: {CUtiSSEVector<T>& v1 = d[pos]; src[0] = v1[0];src[w] = v1[1];pos++; }; break;
	     case 3: {CUtiSSEVector<T>& v1 = d[pos]; src[0] = v1[0];src[w] = v1[1];src[w*2] = v1[2];pos++; }; break;
	   };


         };
       }
   };*/


};



template<class T> class Mat_Transpose : public BaseMatInternData<T>
{
protected:
    uti_refptr<BaseMatInternData<T> > m_data;
public:
    Mat_Transpose(BaseMatInternData<T>* d)
    {
        m_data = d;
    }

    virtual int sizex()
   {
       return m_data->sizey();
   }
   virtual int sizey()
   {
       return m_data->sizex();
   }

   T getData(int y, int x) const
   {
       return conj(m_data->getData(x,y));
   }
   virtual void setData(int y, int x, const T& value)
   {
       utiassert(false);
   }
   virtual void readRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {
     utiassert(false);
     m_data->readRawData(CIntRect(source.m_min.y, source.m_min.x, source.m_max.y, source.m_max.x), d, !dataxoriented);
     if (needConj<T>()) {
     for (int i = 0; i < d.size(); i++) {
       d[i] = conj(d[i]);
     };
     };
   };
   virtual void writeRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {
    utiassert(false);
   };

  /* virtual void readRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     utiassert(false);
     m_data->readRawDataSSE(CIntRect(source.m_min.y, source.m_min.x, source.m_max.y, source.m_max.x), d, !dataxoriented);
     if (needConj<T>()) {
     for (int i = 0; i < d.size(); i++) {
       d[i] = conj(d[i]);
     };
     };
   };
   virtual void writeRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     utiassert(false);
   };*/

};

template<class T> class Mat_Steps : public BaseMatInternData<T>
{
protected:
    T m_start;
    T m_steps;
    int m_c;
public:
    Mat_Steps(const T& s, const T& steps, int c)
    {
        m_start = s;
        m_steps = steps;
        m_c = c;
    }

    virtual int sizex()
   {
       return m_c;
   }
   virtual int sizey()
   {
       return 1;
   }



    T getData(int y, int x) const
   {
       utiassert(y == 0);
       utiassert(x >= 0 && x < m_c);
       return m_start + m_steps*x;
   }
   virtual void readRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {
     utiassert(source.m_min.y == 0 && source.m_max.y == 1);
     utiassert(source.m_min.x >= 0 && source.m_max.x <= m_c);
     for (int i = 0; i < d.size(); i++) {
       d[i] = m_start + m_steps*(i+source.m_min.x);
     };
   };
   virtual void writeRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {
    utiassert(false);
   };

   virtual void readRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     CUtiSSEVector<T> s(m_start);
     CUtiSSEVector<T> step(m_steps);
     CUtiSSEVector<T> base(source.m_min.x);
     CUtiSSEVector<T> index;
     int b=0;
     if (dataxoriented) {
     for (int i = 0; i < d.size(); i++) {
       b = i << 2;
       index.set(b,b+1,b+2,b+3);
       index += base;
       index *= step;
       index += s;
       d[i] = index;
     };
     } else {
       for (int i = 0; i < d.size(); i++) {
       index.set(m_start + m_steps*(i+source.m_min.x),0,0,0);
       d[i] = index;
     };
     };
   };
   virtual void writeRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     utiassert(false);
   };

};



template<class T> class Mat_Identity : public BaseMatInternData<T>
{
protected:
    int m_x;
    int m_y;
public:
    Mat_Identity(int y, int x)
    {
       m_x = x;
        m_y = y;
    }

    virtual int sizex()
   {
       return m_x;
   }
   virtual int sizey()
   {
       return m_y;
   }



    T getData(int y, int x) const
   {
     utiassert(x >= 0 && x < m_x);
      utiassert(y >= 0 && y< m_y);
       if (x==y) {
	return T(1);
       }
       return T(0);
   }
   virtual void readRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {


     utiassert(source.m_min.x >= 0 && source.m_max.x <= m_x);
     utiassert(source.m_min.y >= 0 && source.m_max.y <= m_y);
     int w = source.m_min.x >= 0 && source.m_max.x;
     T* data = d.array;
     int s = d.size();
     for (int i = 0; i < s; i++) {
       data[s] = T(0);
     };
     int j;

     if (dataxoriented) {
     for (int y = source.m_min.y; y < source.m_max.y; y++) {

       if (y >= source.m_min.x && y < source.m_max.x) {
	  j = y-source.m_min.y;
	  d[j*w+j-source.m_min.x] = T(1);
       };
     };
     } else {
       int i;
       for (int x = source.m_min.x; x < source.m_max.x; x++) {

         if (x >= source.m_min.y && x < source.m_max.y) {
	    i = x-source.m_min.x;
	    d[i+(i-source.m_min.x)*w] = T(1);
         };
       };
     };

   };

   /*virtual void readRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     CUtiSSEVector<T> s(m_start);
     CUtiSSEVector<T> step(m_steps);
     CUtiSSEVector<T> base(source.m_min.x);
     CUtiSSEVector<T> index;
     int b=0;
     for (int i = 0; i < d.size(); i++) {
       b = i << 2;
       index.set(b,b+1,b+2,b+3);
       index += base;
       index *= step;
       index += s;
       d[i].set(index);
     };
   };*/

};

template<class T> class Mat_Value : public BaseMatInternData<T>
{
protected:
    int m_x;
    int m_y;
    T m_value;
public:
    Mat_Value(int y, int x, const T& v1)
    {
       m_x = x;
        m_y = y;
	m_value = v1;
    }

    virtual int sizex()
   {
       return m_x;
   }
   virtual int sizey()
   {
       return m_y;
   }



    T getData(int y, int x) const
   {
     utiassert(x >= 0 && x < m_x);
      utiassert(y >= 0 && y< m_y);

       return m_value;
   }
   virtual void readRawData(const CIntRect& source, Array<T>& d, bool dataxoriented)
   {


     utiassert(source.m_min.x >= 0 && source.m_max.x <= m_x);
     utiassert(source.m_min.y >= 0 && source.m_max.y <= m_y);
   //  int w = source.m_min.x >= 0 && source.m_max.x;
     T* data = d.array;
     int s = d.size();
     for (int i = 0; i < s; i++) {
       data[s] = m_value;
     };

   };

   /*virtual void readRawDataSSE(const CIntRect& source, Array< CUtiSSEVector<T> >& d, bool dataxoriented)
   {
     CUtiSSEVector<T> s(m_start);
     CUtiSSEVector<T> step(m_steps);
     CUtiSSEVector<T> base(source.m_min.x);
     CUtiSSEVector<T> index;
     int b=0;
     for (int i = 0; i < d.size(); i++) {
       b = i << 2;
       index.set(b,b+1,b+2,b+3);
       index += base;
       index *= step;
       index += s;
       d[i].set(index);
     };
   };*/

};

CUtiString convertToString(int i, int s);
CUtiString convertToString(float i, int s);
CUtiString convertToString(double i, int s);


template <class T> inline T getValueMat(const T& v1, const T& reference)
{
  return v1;
};

template<class T> inline UMat<T> getValueMat(const T& v1, const UMat<T>& reference)
{
  return UMat<T>(v1, reference.size());
};

template<class T> struct UMatSetter
{
  UMat<T> & m_data;
  int m_x;
  int m_y;
  inline UMatSetter(UMat<T>& data, int x, int y): m_data(data), m_x(x), m_y(y){};
  inline const T& operator=(const T& value);
};

const int UTI_USEDYNAMIC = 16;
template<class T> class UMat : public ReferenceUtiOB
{
    protected:
        uti_refptr<BaseMatInternData<T> >m_data;
    public:
        UMat() {


        }
        UMat(const T& value)
        {
            //m_data = new Mat_SimpleData<T>(1,1);
            //m_data->setData(0,0, value);
	    m_data = new Mat_Value<T>(1, 1, value);

        }
        UMat(const T& value, int sy, int sx)
        {
            m_data = new Mat_Value<T>(sy, sx, value);
        }
        UMat(const T& value, const CIVector2& v)
        {
            m_data = new Mat_Value<T>(v.y, v.x, value);
        }

        UMat(int sy, int sx)
        {
            m_data = new Mat_SimpleData<T>(sy,sx);
        }
        UMat(const CIVector2& v)
        {
            m_data = new Mat_SimpleData<T>(v.y,v.x);
        }
        UMat(const UMat& mat)
        {
            m_data = mat.m_data;
        }
        UMat(BaseMatInternData<T>* d)
        {
            m_data = d;
        }
        UMat<T>& operator=(const UMat<T>& v)
	{
	  m_data = v.m_data;
	  return *this;
	};
	inline T getSingleValue()
	{
	  utiassert(m() == 1);
	  utiassert(n() == 1);
	  return operator()(0,0);
	};
        inline void makeUnique()
        {
            BaseMatInternData<T>* u = m_data->makeUnique();
            if (u) {
                m_data = u;
            }
        }
        inline CIVector2 size() const
       {

         return  CIVector2(m_data->sizex(), m_data->sizey());
       }
        inline int sizex() const
        {
            CHECKEXISTANCE;
            return m_data->sizex();

        }
         inline int sizey() const
        { CHECKEXISTANCE;
            return m_data->sizey();

        }
        inline int n() const
        { CHECKEXISTANCE;
	  return m_data->sizey();
	};
	inline int m() const
	{ CHECKEXISTANCE;
	  return m_data->sizex();
	};
        inline int length() const
        { CHECKEXISTANCE;
	  utiassert(m_data->sizey());
	  return m_data->sizex();
	}

	UMat<T> operator()(const UMat<int>& y, const UMat<int>& x) const;
       inline T operator()(int y, int x) const
       { CHECKEXISTANCE;
           return m_data->getData(y,x);
       }
       inline T operator()(int x) const
       { CHECKEXISTANCE;
           return m_data->getData(0, x);
       }
       inline T operator[](int x) const
       { CHECKEXISTANCE;
           return m_data->getData(0, x);
       }
       inline T get(int y, int x) const
       { CHECKEXISTANCE;
	   return m_data->getData(y,x);
       }
       inline void set(int y, int x, const T& v)
       { CHECKEXISTANCE;
	 makeUnique();
	  m_data->setData(y,x,v);
       }
       inline void set(int x,  const T& v)
       { CHECKEXISTANCE;
	 makeUnique();
	  m_data->setData(0,x,v);
       }
       inline void swap(int y1, int x1, int y2, int x2)
       { CHECKEXISTANCE;
	  T temp = operator()(y1, x1);
	  set(y1, x1, operator()(y2, x2));
	  set(y2, x2, temp);
       }
      /* inline T& operator(int x, int y)
       {
           makeUnique();
           return m_data->getData(x,y);
       }*/
       inline UMat<T> transpose()
       { CHECKEXISTANCE;
           UMat<T> m(new Mat_Transpose<T>(m_data.get()));
           return m;
       }
      /* inline UMat<T> inverse()
       {

       }*/

       inline UMat<T> multiply(const UMat<T>& other) const
       { CHECKEXISTANCE;
	   utiassert(other.sizey() == sizex());
	   int steps = sizex();
	   int newxsize = other.sizex();
	   int newysize = sizey();
           UMat<T> result(newysize, newxsize);
	   T current;
//#ifdef MATH_MODE == MATH_SINGLE
	   for (int i = 0; i < newxsize; i++) {
	      for (int j = 0; j < newysize; j++) {
		current = T(0);
		for (int k = 0; k < steps; k++) {
		  current += this->get(j,k)*other.get(k,i);
		}
		result.set(j,i,current);
	      };
	   };
/*#else
#if MATH_MODE == MATH_RANGE

	   T tempdatastack1[UTI_USEDYNAMIC];
	   T tempdatastack2[UTI_USEDYNAMIC];
	   Array<T> tempdata1;
	   Array<T> tempdata2;
	   if (steps <= UTI_USEDYNAMIC) {
	     tempdata1.setData(tempdatastack1, steps);
	     tempdata2.setData(tempdatastack2, steps);
	   } else {
	     tempdata1.setSize(steps);
	     tempdata2.setSize(steps);
	   };
	   for (int i = 0; i < newxsize; i++) {
	      for (int j = 0; j < newysize; j++) {
		//virtual void readRawData(const CRect& source, Array<T>& d, bool dataxoriented)
		m_data->readRawData(CIntRect(j,0, j+1,steps), tempdata1, true);
		other->m_data->readRawData(CIntRect(0,k, steps, k+1), tempdata2, false);
		current = T(0);
		T* d1 = tempdata1.array;
		T* d2 = tempdata2.array;
		for (int k = 0; k < steps; k++) {
		  current +=  d1[k]*d2[k];
		}
		result.set(j,i,current);
	      };
	   };
#else
	   utiassert(UTI_USEDYNAMIC % 4 == 0);
	   int sse;
	   int ssefull
	   int normal;
	   calcSSE(steps, sse, normal, ssefull);
	   CUtiSSEVector<T> tempdatastack1[UTI_USEDYNAMIC/4];
	   CUtiSSEVector<T> tempdatastack2[UTI_USEDYNAMIC/4];
	   Array<CUtiSSEVector<T> > tempdata1;
	   Array<CUtiSSEVector<T> > tempdata2;
	   if (ssefull <= UTI_USEDYNAMIC) {
	     tempdata1.setData(tempdatastack1, ssefull);
	     tempdata2.setData(tempdatastack2, ssefull);
	   } else {
	     tempdata1.setSize(ssefull);
	     tempdata2.setSize(ssefull);
	   };
	   CUtiSSEVector<T> ssecurrent;
	   ssecurrent.loadZero();
	   for (int i = 0; i < newxsize; i++) {
	      for (int j = 0; j < newysize; j++) {
		//virtual void readRawData(const CRect& source, Array<T>& d, bool dataxoriented)
		m_data->readRawDataSSE(CIntRect(j,0, j+1,steps), tempdata1, true);
		other->m_data->readRawDataSSE(CIntRect(0,k, steps, k+1), tempdata2, false);
		ssecurrent.loadZero();
		T* d1 = tempdata1.array;
		T* d2 = tempdata2.array;
		for (int k = 0; k < ssefull; k++) {
		  d1[k] *= d2[k];
		  ssecurrent += d1[k];

		}
		current =  ssecurrent.sum();
		result.set(j,i,current);
	      };
	   };
#endif
#endif*/

	   return result;
       }
       static UMat<T> identity(int s)
       {
           UMat<T> m(new Mat_Identity<T>(s,s));
           return m;
       }
       static UMat<T> steps(const T& start, const T& delta, int amount)
       {
          UMat<T> m(new Mat_Steps<T>(start, delta, amount));
           return m;
       }
       static UMat<T> steps(const T& start, int amount)
       {
          UMat<T> m(new Mat_Steps<T>(start, T(1), amount));
           return m;
       }
       CUtiString toString()
       {
	 CUtiString str;
	 int maxsize=8;
	 int sx = sizex();
	 int sy = sizey();

	 for (int j = 0; j < sy; j++) {
           if (j == 0) {str += UTISTR("[");} else {str += UTISTR(" ");};
	   for (int i = 0; i < sx; i++) {
	      str += UTISTR(" ")+convertToString(get(j,i), maxsize);

            }
            if (j == sy-1) {str += UTISTR(" ]");} else {str += UTISTR(";\n");};
	 }
	 return str;
       }
       void addVector(const UMat<T>& other, const T& add, const T&scal, int mypositiony=0, int mypositionx=0, int otherpositiony=0,int otherpositionx=0, int otherlength=-1) {
	 utiassert(mypositionx >= 0);
	 utiassert(mypositionx < sizex());
	 utiassert(otherpositionx >= 0);
	 utiassert(otherpositionx < other.sizex());
	 if (otherlength < 0) {
	   otherlength = umin(other.sizex()-otherpositionx, sizex()-mypositionx);
	 }
	 utiassert(otherpositionx+otherlength <= other.sizex());
	 utiassert(mypositionx+otherlength <= sizex());
	 for (int i = 0; i < otherlength; i++) {
	   set(mypositiony, i+mypositionx, add+scal*other(otherpositiony, i+otherpositionx));
	 }
       };

};

extern UMat<int> UALL;

template<class T> const T& UMatSetter<T>::operator=(const T& value)
{

  m_data.set(m_y,m_x,value);
  return value;
}

template<class T> T sum(const UMat<T>& v)
{
  int n = v.n();
  int m = v.m();
  T sum(0);
  for (int i = 0; i < n; i++) {
     for (int j = 0; j < m; j++) {
       sum += v(i,j);
     };
  };
  return sum;
};

#define UTI_MATHAPPLYSINGLEFUNCTION(name, funcname)\
template<class T> UMat<T> name(const UMat<T> &v)\
{\
 UMat<T> r;\
  int n = v.n();\
  int m = v.m();\
 \
  for (int i = 0; i < n; i++) {\
     for (int j = 0; j < m; j++) {\
       r(i,j) = funcname(v(i,j));\
     };\
  };\
  return r;\
}

UTI_MATHAPPLYSINGLEFUNCTION(sin, sin)
UTI_MATHAPPLYSINGLEFUNCTION(cos, cos);
UTI_MATHAPPLYSINGLEFUNCTION(log, log);
UTI_MATHAPPLYSINGLEFUNCTION(sqrt, sqrt);
UTI_MATHAPPLYSINGLEFUNCTION(acos, acos);
UTI_MATHAPPLYSINGLEFUNCTION(asin, asin);

template<class T> UMat<T> pow(const UMat<T> &v, const T&value)
{

  int n = v.n();
  int m = v.m();
  UMat<T> r(n,m);
  for (int i = 0; i < n; i++) {
     for (int j = 0; j < m; j++) {
       r.set(i,j, pow(v(i,j), value));
     };
  };
  return r;
}

template<class T> UMat<T> pow(const UMat<T> &v, const UMat<T> &value)
{
 UMat<T> r;
  int n = v.n();
  int m = v.m();

  for (int i = 0; i < n; i++) {
     for (int j = 0; j < m; j++) {
       r(i,j) = pow(v(i,j), value(i,j));
     };
  };
  return r;
}

template<class T> UMat<T> pow(const T& v, const UMat<T> &value)
{
 UMat<T> r;
  int n = v.n();
  int m = v.m();

  for (int i = 0; i < n; i++) {
     for (int j = 0; j < m; j++) {
       r(i,j) = pow(v, value(i,j));
     };
  };
  return r;
}


template<class T> UMat<T> linearspace(const T& start, const T& ende, int count)
{
  return UMat<T>::steps(start, (ende-start)/T(count), count);
};

template<class T> UMat<T> zeros(int y, int x)
{
 return UMat<T>(T(0), y, x);
}

template<class T> UMat<T> ones(int y, int x)
{
 return UMat<T>(T(1), y, x);
}

template<class T> UMat<T> operator*(const UMat<T>& value, const T& t)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {

            st.set(i,j, value(i,j)*t);
        }
    }
    return st;
}

template<class T> UMat<T> operator/(const UMat<T>& value, const T& t)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, value(i,j)/t);
        }
    }
    return st;
}

template<class T> UMat<T> operator/=(UMat<T>& value, const T& t)
{
    int sx = value.sizex();
    int sy = value.sizey();

    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            value.set(i,j, value(i,j)/t);
        }
    }
    return value;
}

template<class T> UMat<T> operator*=(UMat<T>& value, const T& t)
{
    int sx = value.sizex();
    int sy = value.sizey();

    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            value.set(i,j, value(i,j)*t);
        }
    }
    return value;
}


template<class T> UMat<T> operator*(const T& t, const UMat<T>& value)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, t* value(i,j));
        }
    }
    return st;

}

template<class T> UMat<T> operator/(const T& t, const UMat<T>& value)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, t/ value(i,j));
        }
    }
    return st;

}

template<class T> UMat<T> operator+(const UMat<T>& value, const UMat<T>& value1)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, value(i,j)+value1(i,j));
        }
    }
    return st;
}

template<class T> UMat<T> operator+=(UMat<T>& value, const UMat<T>& value1)
{
    int sx = value.sizex();
    int sy = value.sizey();
   // UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            value.set(i,j, value(i,j)+value1(i,j));
        }
    }
    return value;
}


template<class T> UMat<T> operator+(const UMat<T>& value, const T& value1)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, value(i,j)+value1);
        }
    }
    return st;
}

template<class T> UMat<T> operator+(const T& value1, const UMat<T>& value)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, value1+value(i,j));
        }
    }
    return st;
}

template<class T> UMat<T> operator-(const UMat<T>& value, const T& value1)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, value(i,j)-value1);
        }
    }
    return st;
}

template<class T> UMat<T> operator-(const T& value1, const UMat<T>& value)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, value1-value(i,j));
        }
    }
    return st;
}

template<class T> UMat<T> operator-(const UMat<T>& value, const UMat<T>& value2)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j,  value(i,j)-value2(i,j));
        }
    }
    return st;
}

template<class T> UMat<T> operator*(const UMat<T>& value, const UMat<T>& value2)
{
    int sx = value.sizex();
    int sy = value.sizey();
    UMat<T> st(sy, sx);
    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            st.set(i,j, value(i,j)*value2(i,j));
        }
    }
    return st;
}

typedef UMat<double> UFMat;

//Implementation


template<class T> BaseMatInternData<T>* BaseMatInternData<T>::makeUnique()
   {
      // if (getReferenceCount() == 0)return NULL;
        int ysize = sizey();
       int xsize = sizex();
       Mat_SimpleData<T>* sdata= new Mat_SimpleData<T>(ysize, xsize);
       for (int j = 0; j < ysize; j++) {
           for (int i = 0; i < xsize; i++) {
               sdata->setData(j,i, getData(j,i));
           }
       }
       return sdata;
   }

template<class T> BaseMatInternData<T>* BaseMatInternData<T>::getData(const UIntMat&x, const UIntMat&y)const
   {
       int ysize = x.sizex();
       int xsize = y.sizex();
       Mat_SimpleData<T>* sdata= new Mat_SimpleData<T>(ysize, xsize);
       for (int j = 0; j < ysize; j++) {
           for (int i = 0; i < xsize; i++) {
               sdata->setData(j,i, getData(y(j),x(i)));
           }
       }
       return NULL;

   }

template<class T> UMat<T> getFromColumns(const UMat<T>& source, const UMat<int>& columns)
{
  utiassert(columns.n() == 1);
  int newm = columns.m();
  int newn = source.n();
  UMat<T> res(newn, newm);

  for (int x = 0; x < newm; x++) {
    int currentcolumn = columns(0, x);
    for (int y = 0; y < newn; y++) {
      res.set(y, x,  source(y, currentcolumn));
    };
  };

  return res;
};

template<class T> UMat<T> getFromRows(const UMat<T>& source, const UMat<int>& rows)
{
  utiassert(rows.n() == 1);
  int newn = rows.m();
  int newm = source.m();
  UMat<T> res(newn, newm);

  for (int y = 0; y < newn; y++) {
    int currentrow = rows(0, y);
    for (int x = 0; x < newm; x++) {
      res.set(y, x,  source(currentrow, x));
    };
  };

  return res;
};

template<class T> UMat<T> getFromRowColumns(const UMat<T>& source, const UMat<int>& rows, const UMat<int>& columns)
{
  utiassert(columns.n() == 1);
  int newm = columns.m();
  int newn = source.n();
  UMat<T> res(newn, newm);

  for (int x = 0; x < newm; x++) {
    int currentcolumn = columns(0, x);
    for (int y = 0; y < newn; y++) {
      res.set(y, x,  source(rows(0, y), currentcolumn));
    };
  };

  return res;
}

template<class T>  UMat<T> UMat<T>::operator()(const UMat<int>& y, const UMat<int>& x) const
{
  if (&y == &UALL) {
    return getFromColumns(*this, x);
  } else if ( &x== &UALL) {
    return getFromRows(*this, y);
  } else {
 return getFromRowColumns(*this, y, x);
  }
}

#define UMAT_SETROW2(mat, y1, x1, v1, v2) (mat).set((y1),(x1), v1);\
 (mat).set((y1),(x1)+1, v2);

#define UMAT_SETROW3(mat, y1, x1, v1, v2, v3) (mat).set((y1),(x1), v1);\
 (mat).set((y1),(x1)+1, v2);(mat).set((y1),(x1)+2, v3);

 #define UMAT_SETROW4(mat, y1, x1, v1, v2, v3, v4) (mat).set((y1),(x1), v1);\
 (mat).set((y1),(x1)+1, v2);(mat).set((y1),(x1)+2, v3);(mat).set((y1),(x1)+3, v4);

 #define UMAT_SETROW5(mat, y1, x1, v1, v2, v3, v4, v5) (mat).set((y1),(x1), v1);\
 (mat).set((y1),(x1)+1, v2);(mat).set((y1),(x1)+2, v3);(mat).set((y1),(x1)+3, v4);(mat).set((y1),(x1)+4, v5);

 #define UMAT_SETROW6(mat, y1, x1, v1, v2, v3, v4, v5, v6) (mat).set((y1),(x1), v1);\
 (mat).set((y1),(x1)+1, v2);(mat).set((y1),(x1)+2, v3);(mat).set((y1),(x1)+3, v4);(mat).set((y1),(x1)+4, v5);\
 (mat).set((y1),(x1)+5, v6);

 #define UMAT_SETROW7(mat, y1, x1, v1, v2, v3, v4, v5, v6, v7) (mat).set((y1),(x1), v1);\
 (mat).set((y1),(x1)+1, v2);(mat).set((y1),(x1)+2, v3);(mat).set((y1),(x1)+3, v4);(mat).set((y1),(x1)+4, v5);\
 (mat).set((y1),(x1)+5, v6);(mat).set((y1),(x1)+6, v7);
 #define UMAT_SETROW8(mat, y1, x1, v1, v2, v3, v4, v5, v6, v7) (mat).set((y1),(x1), v1);\
 (mat).set((y1),(x1)+1, v2);(mat).set((y1),(x1)+2, v3);(mat).set((y1),(x1)+3, v4);(mat).set((y1),(x1)+4, v5);\
 (mat).set((y1),(x1)+5, v6);(mat).set((y1),(x1)+6, v7); (mat).set((y1),(x1)+7, v8);


#define UMAT_SETCOLUMN2(mat, y1, x1, v1, v2) (mat).set((y1),(x1), v1);\
 (mat).set((y1)+1,(x1), v2);

#define UMAT_SETCOLUMN3(mat, y1, x1, v1, v2, v3) (mat).set((y1),(x1), v1);\
 (mat).set((y1)+1,(x1), v2);(mat).set((y1)+2,(x1), v3);

 #define UMAT_SETCOLUMN4(mat, y1, x1, v1, v2, v3, v4) (mat).set((y1),(x1), v1);\
 (mat).set((y1)+1,(x1), v2);(mat).set((y1)+2,(x1), v3);(mat).set((y1)+3,(x1), v4);

 #define UMAT_SETCOLUMN5(mat, y1, x1, v1, v2, v3, v4, v5) (mat).set((y1),(x1), v1);\
 (mat).set((y1)+1,(x1), v2);(mat).set((y1)+2,(x1), v3);(mat).set((y1)+3,(x1), v4);(mat).set((y1)+4,(x1), v5);

 #define UMAT_SETCOLUMN6(mat, y1, x1, v1, v2, v3, v4, v5, v6) (mat).set((y1),(x1), v1);\
 (mat).set((y1)+1,(x1), v2);(mat).set((y1)+2,(x1), v3);(mat).set((y1)+3,(x1), v4);(mat).set((y1)+4,(x1), v5);\
 (mat).set((y1)+5,(x1), v6);

 #define UMAT_SETCOLUMN7(mat, y1, x1, v1, v2, v3, v4, v5, v6, v7) (mat).set((y1),(x1), v1);\
 (mat).set((y1)+1,(x1), v2);(mat).set((y1)+2,(x1), v3);(mat).set((y1)+3,(x1), v4);(mat).set((y1)+4,(x1), v5);\
 (mat).set((y1)+5,(x1), v6);(mat).set((y1)+6,(x1), v7);
 #define UMAT_SETCOLUMN8(mat, y1, x1, v1, v2, v3, v4, v5, v6, v7) (mat).set((y1),(x1), v1);\
 (mat).set((y1)+1,(x1), v2);(mat).set((y1)+2,(x1), v3);(mat).set((y1)+3,(x1), v4);(mat).set((y1)+4,(x1), v5);\
 (mat).set((y1)+5,(x1), v6);(mat).set((y1)+6,(x1), v7); (mat).set((y1)+7,(x1), v8);

 #define ROTATE(a,i,j,k,l) g = a(i,j); h = a(k,l); a.set(i,j, g-s*(h+g*tau)); a.set(k,l, h+s*(g-h*tau));

template<class T> void eigenvalues_sym_jacobi(UMat<T>&a, UMat<T>& d, UMat<T>&v, int nrot)
{

   T tresh, theta, tau, t, sm, s, h, g,c;
   int n = a.sizex();
   utiassert(s.sizex() == s.sizey());
   UMat<T> b(1,n);
   UMat<T> z(1,n);
   v = UMat<T>::identity(n);
   for (int ip = 0; ip < n; ip++) {
    b.set(ip, a(ip, ip));
    d.set(ip, a(ip, ip));
    z.set(ip, 0);
   }
   nrot = 0;
   for (int i = 0; i < 50; i++) {
    sm =0.0;
    for (int ip =0; ip < n-1; ip++) {
        for (int iq=ip+1; iq < n; iq++) {
            sm += fabs(a(ip, iq));
        }
    }
    if (sm == 0.0) {
        return;
    }
    if (i < 3) {
        tresh = 0.2*sm/(n*n);
    } else {
        tresh = 0.0;
    }
    for (int ip=0; ip < n-1;ip++) {
        for (int iq=ip+1; iq < n; iq++) {
            g = 100.0*fabs(a(ip, iq));
            if (i >= 4 && (fabs(d(ip)+g) == fabs(d[ip])) && (fabs(d(iq)+g) == fabs(d[iq]))) {
                a.set(ip, iq, 0);
            } else if (fabs(a(ip, iq)) > tresh) {
                h = d[iq]-d[ip];
                if (fabs(h)+g == fabs(h)) {
                    t = (a(ip, iq)) / h;
                } else {
                   theta = 0.5*h/(a(ip, iq));
                   t = 1.0/(fabs(theta)+sqrt(1.0+theta*theta));
                   if (theta < 0.0) t= -t;
                }
                c = 1.0/sqrt(1+t*t);
                s = t*c;
                tau = s/ (1.0+c);
                h = t*a(ip, iq);
                z.set(ip, z(ip)-h);
                z.set(iq, z(iq)+h);

                d.set(ip, d(ip)-h);
                d.set(iq, d(iq)+h);
                a(ip, iq, 0.0);
                for (int j = 0; j <= ip-1; j++) {
                    ROTATE(a, j, ip, j, iq);
                }
                for (int j = ip+1; j <= iq-1; j++) {
                    ROTATE(a, ip, j, j, iq);
                }
                for (int j = iq+1; j < n; j++) {
                    ROTATE(a, ip, j, iq, j);
                }
                for (int j = 1; j < n; j++) {
                    ROTATE(a, j, ip, j, iq);
                }
                nrot++;

            }
        }
    }
    for (int ip=0; ip< n; ip++) {
        b.set(ip, b(ip)+z(ip));
        d.set(ip, b(ip));
        z.set(ip, 0.0);
    }

   }
}

#undef ROTATE

template<class T>  UMat<T> polynomAdd(const UMat<T>& x, const UMat<T>& y)
{
    int degree1 = x.m();
    int degree2 = y.m();
    int d = umax(degree1, degree2);
    UMat<T> result(1, d);
    for (int i = 0; i < d; i++) {
        T value=T();
        if (i < degree1) {
            value += x.get(0, i);
        }
        if (i < degree2) {
            value += y.get(0, i);
        }
        result.set(i, value);
    }
    return result;
}

template<class T>  UMat<T> polynomSub(const UMat<T>& x, const UMat<T>& y)
{
    int degree1 = x.m();
    int degree2 = y.m();
    int d = umax(degree1, degree2);
    UMat<T> result(1, d);
    for (int i = 0; i < d; i++) {
        T value=T(0.0);
        if (i < degree1) {
            value += x.get(0, i);
        }
        if (i < degree2) {
            value -= y.get(0, i);
        }
        result.set(i, value);
    }
    return result;
}


template<class T>  UMat<T> polynomMultiply(const UMat<T>& x, const UMat<T>& y)
{
    int degree1 = x.m()-1;
    int degree2 = y.m()-1;
    int d = degree1+degree2;
  UMat<T> result(1, d+1);

  for (int s = 0; s <= d; s++) {
      T value = T(0);
      for (int r = 0; r <= s; r++) {
          int r2 = s-r;
          if (r2 <= degree1 && r <= degree2) {
                T v1= x(r2);
                T v2 = y(r);
          value += v1*v2;
          }
      }
      result.set(s, value);
   }
   return result;
}

template<class T> void polynomDivision(const UMat<T>& x, const UMat<T>& y, UMat<T>& divres, UMat<T>& res)
{
  //  #define POLYNOMDEBUG
    #ifdef POLYNOMDEBUG
         std::wcout << "Divide " << x.toString() << " throught " << y.toString() << std::endl;
    #endif
    int degree1 = x.m()-1;
    int degree2 = y.m()-1;
    //int d = degree1+degree2;
    //UMat<T> result(1, d);
    utiassert(degree1 >= degree2);
    UMat<T> currentdata(1, degree1+1);
    int ressize = degree1-degree2;

    UMat<T> transformed(1, degree2+1);
    T divvalue = T(1.0) / y(degree2);
    #ifdef POLYNOMDEBUG
     std::wcout << "divvalue= " << divvalue << std::endl;
    #endif
    for (int i = 0; i <= degree2; i++) {
            transformed.set(i, y(i)*divvalue);
    }
    #ifdef POLYNOMDEBUG
         std::wcout << "Transformed= " << transformed.toString() << std::endl;
    #endif

    UMat<T> result(1, ressize+1);
    for (int i = 0; i <= degree1; i++) {
            currentdata.set(i, x(i));
    }
    #ifdef POLYNOMDEBUG
     std::wcout << "currentdata= " << currentdata.toString() << std::endl;
     #endif
   // int restsize =???;
    UMat<T> rest(1, degree2);
    /*for (int i = 0; i < degree2; i++) {
            currentdata(degree1-i-1) = x(degree1-i-1);
    }*/
    int currentpos = degree1;
    int positioninres = ressize;
    T pvalue, scal;
    pvalue = currentdata(currentpos);
     //bleibt immer so

     while (positioninres >= 0) {
          pvalue = currentdata(currentpos);
            result.set(positioninres, pvalue*divvalue);
            #ifdef POLYNOMDEBUG
             std::wcout << "pvalue= " << pvalue << std::endl;
     std::wcout << "result= " << currentdata.toString() << std::endl;
     #endif
         for (int i = 0; i <= degree2; i++) {
            currentdata.set(currentpos-i, currentdata[currentpos-i] - pvalue*transformed(degree2-i));
         }
         #ifdef POLYNOMDEBUG
     std::wcout << "currentdata= " << currentdata.toString() << std::endl;
     #endif
         positioninres--;
         currentpos--;


     }
     divres = result;

     for (int i = 0; i < degree2; i++) {
          rest.set(i, currentdata(i));
     }
     res = reducePolynom(rest);

}

template<class T>  UMat<T> reducePolynom(const UMat<T>& x)
{
   int maxdegree = x.m();
   int realdegree=maxdegree;
   while (realdegree > 0 && x(realdegree-1) == 0) {
        realdegree--;
   }
   realdegree = umax(0, realdegree);
   if (maxdegree == realdegree) {
    return x;
   }
   UMat<T> result(1, realdegree);
   for (int i = 0; i < realdegree; i++) {
    result.set(i, x(i));
   }


   return result;
}

template<class T>  bool isReducedPolynom(const UMat<T>& x)
{
   int maxdegree = x.m();
   int realdegree=maxdegree;
   while (realdegree > 0 && x(realdegree-1) == 0) {
        realdegree--;
   }
   realdegree = umax(0, realdegree);
   if (maxdegree == realdegree) {
    return true;
   }


   return false;
}

template<class T>  UMat<T> derivatePolynom(const UMat<T>& x)
{
   int maxdegree = x.m();
   UMat<T> result(1, umax(maxdegree-1, 0));
   for (int i = 0; i < maxdegree-1; i++) {
    result.set(i, x(i+1)*T(i+1));
   }


   return result;
}

template<class T> UMat<T> powPolynom(const UMat<T>& x, int n)
{
    UMat<T> res(1,1);
    res.set(0, 1);

    for (int i = 0; i < n; i++) {
        res = polynomMultiply(res, x);
    }

    return res;
}

template<class T> UMat<T> x_pow_n(int n)
{
    UMat<T> res(1,n+1);
    for (int i = 0; i < n; i++) {
        res.set(i, 0);
    }
    res.set(n, 1);
    return res;
}


template<class T> UMat<T> one_minus_x_pow_n(int n)
{
    UMat<T> res(1,n+1);
    res.set(0, 1);
    for (int i = 1; i < n; i++) {
        res.set(i,0);
    }
    res.set(n, -1);
    return res;
}

template<class T> T polynomEval(const UMat<T>& x, const T& value)
{

   int degree1 = x.m()-1;
   int currenti=degree1;
   T res = T(0);
   /*T xvalue = value;
   res = x(0);
   for (int s = 1; s <= degree1; s++) {
       res += x(s)*xvalue;
        xvalue = xvalue * value;
   }*/
   //Horner Schema
   if (degree1 < 0) return res;
   res = x(currenti);
   while (currenti > 0) {
      res = res*value+x(currenti-1);
      currenti--;
   }
   return res;

}



template<class T> class Polynom : public ReferenceUtiOB
{
    protected:
        UMat<T> m_mat;
		uti_refptr<ReferenceUtiOB> m_polynomring;
    public:
        inline Polynom()
        {
            m_mat = UMat<T>(1,1);
            m_mat.set(0, T(0));
        }
        inline Polynom(const Polynom<T>& p)
        {CHECKEXISTANCE;
            m_mat = const_cast<UMat<T>& >(p.m_mat);

        }
		inline void setPolynomRing(ReferenceUtiOB* obj) const
		{
			m_polynomring = obj;
		}
		inline ReferenceUtiOB* getPolynomRing() const
		{
			return m_polynomring.get();
		}
        explicit inline Polynom(int t)
        {CHECKEXISTANCE;
            m_mat = UMat<T>(1,t+1);
            for (int i = 0; i < t; i++) {
            m_mat.set(i, T(0));
            }
        }
        explicit inline Polynom(const UMat<T>& mat)
        {
            CHECKEXISTANCEOF(&mat);
            m_mat = mat;
        }
        explicit inline Polynom(const T &value)
        {CHECKEXISTANCE;
            m_mat = UMat<T>(value);
            CHECKEXISTANCEOF(m_mat);
        }
        inline UMat<T> getMat() const
        {CHECKEXISTANCE;
            return m_mat;

        }
        inline UMat<T>& getMatRef()
        {CHECKEXISTANCE;
            return m_mat;

        }
        inline const UMat<T>& getMatRef() const
        {CHECKEXISTANCE;
            return m_mat;

        }
        inline int getSize() const
        {CHECKEXISTANCE;
           return m_mat.m();
        }
         inline int getDegree() const
        {CHECKEXISTANCE;
	  return m_mat.m()-1;
	};


       inline T operator()(int x) const
       {CHECKEXISTANCE;
           return m_mat(x);
       }
       inline T operator[](int x) const
       {CHECKEXISTANCE;
           return m_mat[x];
       }

       inline void set(int x,  const T& v)
       {CHECKEXISTANCE;
           m_mat.set(x, v);
       }
       inline void reduce()
       {
           m_mat = reducePolynom(m_mat);
            utiassert(isReduced());

       }
       inline Polynom<T> derivate() const
       {CHECKEXISTANCE;
           return Polynom<T>(derivatePolynom(m_mat));
       }
       inline T eval(const T& value) const
       {   CHECKEXISTANCE;
           return polynomEval(m_mat, value);

       }
       inline void division(const Polynom<T> & d, Polynom<T> &q, Polynom<T> &r) const
       {CHECKEXISTANCE;
           UMat<T> qres, rres;
           polynomDivision(m_mat, d.m_mat, qres, rres);
           q = Polynom<T>(qres);
           r = Polynom<T>(rres);
       }
       inline bool identical(const Polynom<T>& other, const T& tolerance) const
       {CHECKEXISTANCE;
           int otherd = other.getDegree();
           int d = getDegree();
           if (d != otherd) {
            return false;
           }
           for (int i = 0; i <= d; i++) {
              if (fabs(operator[](i)-other[i]) > tolerance) {
                  return false;
              }
           }

           return true;


       }

     Polynom<T>& operator=(const Polynom<T>& v)
	{
	  m_mat = v.m_mat;
	  CHECKEXISTANCEOF(&v);
	  CHECKEXISTANCEOF(&m_mat);
	  return *this;
	};
	bool isReduced() const
	{
	    return isReducedPolynom(m_mat);
	}
	CUtiString toString()
       {
	     CUtiString str;
	     int maxdegree = m_mat.m()-1;
	     for (int i = maxdegree; i >= 0; i--) {
             str += float(m_mat(i));
             if (i != 0) {
                str += UTISTR(" x^");
                str += i;
                str += UTISTR(" + ");
             }
	     }
	     return str;
       }

};

template<class T> Polynom<T> createRandomPolynom(int n)
{
    T mul = urandom(2) == 0 ? 1 : -1;
    Polynom<T> bt(n);
    bt.set(n, urandom()*9.8+0.2);

    for (int i = 0; i < n; i++) {
        bt.set(i, urandom()*10);
    }
    return bt;
}

template<class T> Polynom<T> bernsteinPolynom(int i, int n)
{
    int b = binomial(n, i);

    UMat<T> t = x_pow_n<T>(i);
    //std::wcout << t.toString() << std::endl;
    UMat<T> oneminust = one_minus_x_pow_n<T>(1);
    //std::wcout << oneminust.toString() << std::endl;
    oneminust = powPolynom(oneminust, n-i);
    //std::wcout << oneminust.toString() << std::endl;

    t = polynomMultiply(t, oneminust);
    t = t*T(b);
    return Polynom<T>(t);

}

template<class T> Polynom<T> operator+(const Polynom<T>& value, const T& value1)
{
    Polynom<T> result = value;
    //result(0) += value1;
     if (result.getDegree()< 0) {
        result = Polynom<T>(0);
    }
    result.set(0, result(0)+value1);
    return result;
}

template<class T> Polynom<T> operator+(const T& value1, const Polynom<T>& value)
{
    Polynom<T> result = value;
 if (result.getDegree()< 0) {
        result = Polynom<T>(0);
    }
    result.set(0, value1+result(0));
    return result;
}

template<class T> Polynom<T> operator-(const Polynom<T>& value, const T& value1)
{
    Polynom<T> result = value;
    //result(0) -= value1;
     if (result.getDegree()< 0) {
        result = Polynom<T>(0);
    }
    result.set(0, result(0)-value1);
    return result;
}

template<class T> Polynom<T> operator-(const T& value1, const Polynom<T>& value)
{
    Polynom<T> result = value;
    if (result.getDegree()< 0) {
        result = Polynom<T>(0);
    }
    result.set(0, value1-result(0));
    return result;
}

template<class T> Polynom<T> operator-(const Polynom<T>& value, const Polynom<T>& value2)
{

    return Polynom<T>(reducePolynom(polynomSub(value.getMatRef(), value2.getMatRef())));
}

template<class T> Polynom<T> operator+(const Polynom<T>& value, const Polynom<T>& value2)
{

    return Polynom<T>(reducePolynom(polynomAdd(value.getMatRef(), value2.getMatRef())));
}

template<class T> Polynom<T> operator*(const Polynom<T>& value, const Polynom<T>& value2)
{

    return Polynom<T>(polynomMultiply(value.getMatRef(), value2.getMatRef()));
}

template<class T> Polynom<T> operator*(const Polynom<T>& value, const T& value1)
{
    Polynom<T> result = value;
    int s = result.getSize();
    for (int i = 0;i< s; i++) {
        result.set(i, result(i)*value1);
    }
    if (value1 == 0.0) {
        result.reduce();
    }

    return result;
}

template<class T> Polynom<T> operator/(const Polynom<T>& value, const T& value1)
{
    Polynom<T> result = value;
    int s = result.getSize();
    for (int i = 0;i< s; i++) {
        result.set(i, result(i)/value1);
    }
    if (value1 == 0.0) {
        result.reduce();
    }

    return result;
}

template<class T> Polynom<T> operator*(const T& value1, const Polynom<T>& value)
{
    Polynom<T> result = value;
    int s = result.getSize();
    for (int i = 0;i< s; i++) {
        result.set(i, value1*result(i));
    }

if (value1 == 0.0) {
        result.reduce();
    }
    return result;
}

template<class T> int signChanges(const std::vector<Polynom<T> >& chain, const T& a)
{
    utiassert(chain.size() >= 1);
    T value = chain[0].eval(a);
    bool positiv = value >= 0;
    int signchanges = 0;
    for (int i = 1; i < chain.size(); i++) {
            value = chain[i].eval(a);

        bool currentsignpositiv = value >= 0;
        if (positiv != currentsignpositiv) {
            signchanges++;
            positiv = currentsignpositiv;
        }
    }

    return signchanges;
}

const double CONVERGENCELIMIT = 0.000001;

unsigned int ggT(unsigned int a, unsigned int b);

long ggTLong(long a, long b);

inline unsigned int kgV(unsigned int a, unsigned int b)
{
    return (a*b)/ggT(a,b);
}

struct Fraction
{
    int m_zaehler;
    int m_nenner;
    inline Fraction(int zaehler = 0, int nenner = 1)
    {
        m_zaehler = zaehler;
        m_nenner = nenner;

    }
    inline void reduce()
    {
        int tmp = ggT(m_zaehler, m_nenner);
        m_zaehler /= tmp;
        m_nenner /= tmp;
    }
};

template<class T> Polynom<T> ggT(const Polynom<T>& v1, const Polynom<T>& v2)
{
    Polynom<T> var1;
    Polynom<T> var2;
    if (v1.getDegree()< v2.getDegree()) {
            var1 = v2;
            var2 = v1;
    } else {
    var1 = v1;
            var2 = v2;

    }
    Polynom<T> r, q;

    //std::wcout << var1.toString() << std::endl;
    //std::wcout << var2.toString() << std::endl;

    while (true) {
        var1.division(var2, q, r);
        //std::wcout << r.toString() << std::endl;
        if ((r.getDegree() == -1) || ((r.getDegree() == 0) && (fabs(r(0)) < 0.0001))) {
            return var2;
        }
        var1 = var2;
        var2 = r;

    }
    return r;
}

template<class T> bool findRootNewtonInInterval(const Polynom<T>& poly, const Polynom<T>& derv, const T& a, const T& b, T&res)
{
    T xi = (a+b) / 2.0;
    T xi1;
    T dx;
    int steps = 0;
    while (true) {
        if (xi < a || xi > b) {
            return false;
        }
        if (steps >= 20)return false;
        steps++;
        dx = poly.eval(xi) / derv.eval(xi);
        xi1 = xi - dx;
        if (fabs(dx) < CONVERGENCELIMIT) {
            res = xi1;
            return true;
        }
        xi = xi1;
    }
    return false;
}

template<class T> bool findRootBisectionInIntervalRecursive(const Polynom<T>& poly, const T& a, const T& b, const T& valuea, const T& valueb, T&res, int depth, int maxdepth)
{
    if (valuea*valueb >= 0) return false;
    T middle = (a+b)/2.0;
    utiassert(a <= b);
    if (depth == maxdepth || fabs(a-b) < CONVERGENCELIMIT) {
        res = middle;
        return true;
    }
    T middlevalue = poly.eval(middle);
    if (findRootBisectionInIntervalRecursive(poly, a, middle, valuea, middlevalue, res, depth+1, maxdepth)) {
        return true;
    }
    if (findRootBisectionInIntervalRecursive(poly, middle, b,  middlevalue,valueb, res, depth+1, maxdepth)) {
        return true;
    }
    return false;

}

template<class T> bool findRootBisectionInInterval(const Polynom<T>& poly, const T& a, const T& b, T&res)
{
   T valuea = poly.eval(a);
   T valueb = poly.eval(b);
   return findRootBisectionInIntervalRecursive(poly, a, b, valuea, valueb, res, 0, 20);

}


template<class T> void findZeroSetRecursive(const std::vector<Polynom<T> >& chain,
                                            const T& a, const T& b,
                                            int signa, int signb,int depth,
                                            std::vector<T>& resvalues)
{
  //  std::wcout << "findZeroSetRecursive(" << a << ", " << b << ", " << signa << ", " << signb << ")" << std::endl;
     utiassert(a <= b);
     int zeros = signa - signb;
     if (zeros == 0) {
        return;
     }
     if (zeros == 1 || fabs(a-b) < 0.00000001 || depth > 32) {
        T res;
        if (findRootNewtonInInterval(chain[0], chain[1], a, b, res)) {
           resvalues.push_back(res);
           return;
        }
        findRootBisectionInInterval(chain[0], a, b, res);
        resvalues.push_back(res);
        return;
     }
     //divide interval
     T middle = (a+b)/ 2;
     int signmiddle = signChanges(chain, middle);
     findZeroSetRecursive(chain, a, middle, signa, signmiddle,depth+1, resvalues);
     findZeroSetRecursive(chain, middle, b, signmiddle, signb, depth+1, resvalues);
}

template<class T> void findCompleteZeroSet(const Polynom<T>& polynom2, const T& a, const T& b, std::vector<T>& resvalues)
{
Polynom<T> polynom = polynom2;
polynom.reduce();

  utiassert(polynom.isReduced());

  Polynom<T> derivative = polynom.derivate();
  Polynom<T> g = ggT(polynom, derivative);
  Polynom<T> newpolynom, r;
  T topvalue = g(g.getDegree());
  g = g/topvalue;
  std::wcout << "Erase multiroots: " << g.toString() << std::endl;
  polynom.division(g, newpolynom,r);
std::wcout << "Newpolynom: " << newpolynom.toString() << std::endl;
findCompleteZeroSetNoMultiRoots(newpolynom, a,b, resvalues);
}

template<class T> void findCompleteZeroSetNoMultiRoots(const Polynom<T>& polynom2, const T& a, const T& b, std::vector<T>& resvalues)
{
//Keine Mehrfachen Nullstellen!
Polynom<T> polynom = polynom2;
polynom.reduce();

  utiassert(polynom.isReduced());
  int degreepolynom = polynom.getDegree();
  if (degreepolynom == 0) {
        return;
  }
  #ifdef MULTIROOTS
  Polynom<T> derivative = polynom.derivate();
  Polynom<T> g = ggT(polynom, derivative);
  Polynom<T> newpolynom, r;
  T topvalue = g(g.getDegree());
  g = g/topvalue;
  std::wcout << "Erase multiroots: " << g.toString() << std::endl;
  polynom.division(g, newpolynom,r);
std::wcout << "Newpolynom: " << newpolynom.toString() << std::endl;
#else
Polynom<T> newpolynom = polynom;
#endif

  std::vector<Polynom<T> > sturmchain;
  sturmchain.push_back(newpolynom);
  sturmchain.push_back(newpolynom.derivate());

  Polynom<T> pn2;
  Polynom<T> q;
  while (sturmchain.back().getDegree() >= 1) {
        Polynom<T> pn = sturmchain.at(sturmchain.size()-2);
        Polynom<T> pn1 = sturmchain.at(sturmchain.size()-1);
        pn.division(pn1, q, pn2);
        pn2 = pn2*-1.0;
        sturmchain.push_back(pn2);

  }
 /* std::wcout << "Storm chain:" << std::endl;
  for (int i = 0; i < sturmchain.size(); i++){
      std::wcout << sturmchain[i].toString() << std::endl;
  }*/
  int signa = signChanges(sturmchain, a);
  int signb = signChanges(sturmchain, b);
  findZeroSetRecursive(sturmchain, a, b, signa, signb, 0, resvalues);


}

template<class T> void findZeroSetAnalytical(const Polynom<T>& polynom2,std::vector<T>& resvalues)
{
    Polynom<T> polynom = polynom2;
    polynom.reduce();
    int degree = polynom.getDegree();
    switch (degree) {
        case -1: {
            resvalues.push_back(0);
        }; break;
       case 0: {
            T b = polynom[0];
            if (b == 0) {
           resvalues.push_back(0);
            }
           return;
       };break;
       case 1: {
           T a = polynom[1];
           T b = polynom[0];
           if (a != 0) {
            resvalues.push_back(-b/a);
           }

       };break;
       case 2: {
           T a = polynom[2];
           T b = polynom[1];
           T c = polynom[0];
           T D = b*b-4.0*a*c;
           T bsign = b >= 0 ? 1 : -1;
           if (D < -0.000001) {
            return;
           } else if (D > 0.000001) {
               T q = -0.5*(b+bsign*sqrt(D));
               resvalues.push_back(q/a);
               resvalues.push_back(c/q);
           } else {
               resvalues.push_back(-b/(2*a));
           }

       }; break;
       case 3: {
           T scal = polynom[3];
           T a = polynom[2]/scal;
           T b = polynom[1]/scal;
           T c = polynom[0]/scal;
           T Q = (a*a-3.0*b)/9.0;
           T R = (2.0*a*a*a-9.0*a*b+27.0*c)/54.0;
           T R2 =R*R;
           T Q3 = Q*Q*Q;
           if (R2 < Q3) {
            T theta = arccos(R/sqrt(Q3));
            resvalues.push_back(-2.0*sqrt(Q)*cos(theta / 3.0)-a/3.0);
            resvalues.push_back(-2.0*sqrt(Q)*cos((theta+2*M_PI) / 3.0)-a/3.0);
            resvalues.push_back(-2.0*sqrt(Q)*cos((theta-2*M_PI) / 3.0)-a/3.0);
           } else {
              /*T Atemp = sqrt(R2-Q3);
              if (R*Atemp < 0) {
                Atemp = -Atemp;
              }
              T A = pow(Atemp, 1.0/3.0);*/
              T Rsign = R >= 0 ? 1 : -1;
              T A = -Rsign*pow(fabs(R)+sqrt(R2-Q3), 1.0/3.0);
              T B = 0;
              if (A != 0) {
                B = Q / A;
              }
              resvalues.push_back((A+B)-a/3.0);
           }

       }; break;
       default: {
           utiassert(false);
       };break;
    }

}



template<class T> class RationalPolynom : public ReferenceUtiOB
{
    protected:
        Polynom<T> m_top;
        Polynom<T> m_bottom;
    public:
        inline RationalPolynom()
        {
            m_top = UMat<T>(1,1);
            m_top(0)  = T(0);
             m_bottom = Polynom<T>(T(1.0));
        }
        inline RationalPolynom(const UMat<T>& mat)
        {
            m_top =  Polynom<T>(mat);
            m_bottom = Polynom<T>(T(1.0));
        }
        inline RationalPolynom(const Polynom<T>& mat)
        {
            m_top = mat;

          m_bottom = Polynom<T>(T(1.0));
        }

        inline RationalPolynom(const UMat<T>& mat, const UMat<T>& mat2)
        {
            m_top =  Polynom<T>(mat);
            m_bottom = Polynom<T>(mat2);
        }
        inline RationalPolynom(const Polynom<T>& mat,const Polynom<T>& mat2)
        {
            m_top = mat;
          m_bottom = mat2;
        }
        inline RationalPolynom(const T &value)
        {
            m_top = Polynom<T>(value);
            m_bottom = Polynom<T>(T(1.0));
        }
        inline RationalPolynom<T> getTop() const
        {
            return m_top;
        }
        inline RationalPolynom<T> getBottom() const
        {
            return m_top;
        }
        inline T eval(const T& value) const
        {
            return m_top.eval(value) / m_bottom.eval(value);
        }
        inline RationalPolynom<T> derivate() const
        {
            return RationalPolynom<T>(m_top.derivate()*m_bottom - m_top*m_bottom.derivate(), m_bottom*m_bottom);
        }
};

template<class T> RationalPolynom<T> operator+(const RationalPolynom<T>& value, const T& value1)
{

    return value+RationalPolynom<T>(value1);
}

template<class T> RationalPolynom<T> operator+(const T& value1, const RationalPolynom<T>& value)
{
    return value+RationalPolynom<T>(value1);
}

template<class T> RationalPolynom<T> operator-(const RationalPolynom<T>& value, const T& value1)
{
    return value-RationalPolynom<T>(value1);
}

template<class T> RationalPolynom<T> operator-(const T& value1, const RationalPolynom<T>& value)
{
    return RationalPolynom<T>(value1)-value;
}

template<class T> RationalPolynom<T> operator-(const RationalPolynom<T>& value, const RationalPolynom<T>& value2)
{

    return RationalPolynom<T>(value.getTop()*value2.getBottom()- value2.getTop()*value.getBottom(), value.getBottom()*value2.getBottom());
}

template<class T> RationalPolynom<T> operator+(const RationalPolynom<T>& value, const RationalPolynom<T>& value2)
{

    return RationalPolynom<T>(value.getTop()*value2.getBottom()+ value2.getTop()*value.getBottom(), value.getBottom()*value2.getBottom());
}

template<class T> RationalPolynom<T> operator*(const RationalPolynom<T>& value, const RationalPolynom<T>& value2)
{

    return RationalPolynom<T>(value.getTop()*value2.getTop(), value.getBottom()*value2.getBottom());
}

template<class T> RationalPolynom<T> operator*(const RationalPolynom<T>& value, const T& value1)
{
    RationalPolynom<T> result(value.getTop()*value1, value.getBottom());

    return result;
}

template<class T> RationalPolynom<T> operator*(const T& value1, const RationalPolynom<T>& value)
{
    RationalPolynom<T> result(value.getTop()*value1, value.getBottom());

    return result;
}

typedef Polynom<double> PolynomDouble;
typedef UMat<int> UPermutation;

template <class T> bool isIdentical(const UMat<T> &a, const UMat<T>& b, T precision = 0.0001)
{
    if (a.n() != b.n()) return false;
    if (a.m() != b.m()) return false;

    int sx = a.sizex();
    int sy = a.sizey();

    for (int j = 0; j < sx; j++) {
       for (int i = 0; i < sy; i++) {
            T value1 = a(i,j);
            T value2 = b(i,j);
            if (fabs(value1-value2) > precision) return false;
        }
    }

    return true;
}

template<class T> class CULUDecomposition : public ReferenceUtiOB
{
public:
    UMat<T> m_original;
    UMat<T> m_a;
    UPermutation m_permutation;
    T m_signum;
    explicit CULUDecomposition(const UMat<T>& input)
    {
        m_original = input;
        UMat<T> a = input;

        //std::wcout << "original:\n" << m_original.toString() << std::endl;
        int imax;
        T big, dum,sum, temp;
        int n = input.n();
        utiassert(input.n() == input.m());
        UMat<T>  vv(1, n);
        m_signum = T(1);
        m_permutation = UPermutation(1,n);

        for (int i = 0; i < n; i++) {
            big = T(0);
            for (int j = 0; j < n; j++) {
                temp = fabs(a(i,j));
                if (temp > big) {
                    big = temp;
                }
            }
            if (big == T(0)) {
                    utiassert(false);
                    return;
            }
            vv.set(i, T(1)/big);
        }
        for (int j = 0; j < n; j++) {
            for (int i = 0; i < j; i++) {
                sum = a(i,j);
                for (int k = 0; k < i; k++) {
                    sum -= a(i,k)*a(k,j);
                }
                a.set(i,j,sum);
            }
            big = T(0);
            for (int i = j; i < n; i++) {
                sum = a(i,j);
                for (int k = 0; k < j; k++) {
                    sum -= a(i,k)*a(k,j);
                }
                a.set(i,j, sum);
                dum = vv[i]*fabs(sum);
                if (dum >= big) {
                    big = dum;
                    imax = i;
                }
            }
            if (j != imax) {
                for (int k = 0; k < n; k++) {
                    dum = a(imax, k);
                    a.set(imax, k, a(j,k));
                    a.set(j,k, dum);
                }
                m_signum = -m_signum;
                vv.set(imax, vv(j));
            }
            m_permutation.set(j, imax);
            if (a(j,j) == T(0))
            {
                a.set(j,j, T(0.00000001));
            }
            if (j != n) {
                dum = T(1)/a(j,j);
                for (int i = j+1; i <n; i++) {
                    a.set(i,j, a(i,j)*dum);
                }
            }
        }

        //std::wcout << "a:\n" << a.toString() << std::endl;


        m_a = a;
    }
    UMat<T> getInverse() const
    {
         int n = m_a.n();
         UMat<T> res(n,n);
         UMat<T> d(n);

         for (int j = 0; j < n; j++) {
            for (int i = 0; i < n; i++) {
                d.set(i, 0.0);
            }
            d.set(j, 1.0);
            solveSystemInPlace(d);
            for (int i = 0; i < n; i++) {
                res.set(i,j, d(i));
            }
         }
         #ifndef NDEBUG
         UMat<T> shouldbeidentity = res.multiply(m_original);
         utiassert(isIdentical(shouldbeidentity, UMat<T>::identity(n)));
         #endif
         return res;

    }
     void solveSystemInPlace(UMat<T>& b) const
    {
         utiassert(m_original.sizex()==b.sizey());
         #ifndef NDEBUG
         UMat<T> copy = b;
         #endif
        int  ii,ip;
         int n = m_a.n();
         ii = -1;
         T sum;
         //std::wcout << "asolve:\n" << m_a.toString() << std::endl;
         for (int i = 0; i < n; i++) {
            ip = m_permutation(i);
            sum = b(ip, 0);
            b.set(ip, 0, b(i, 0));
            if (ii != -1) {
                for (int j = ii; j <= i-1; j++) {
                    sum -= m_a(i,j)*b(j, 0);
                }
            } else if (sum != T(0)) {
               ii = i;
            }
            b.set(i, 0, sum);

         }
         //std::wcout << "b:\n" << b.toString() << std::endl;
         for (int i = n-1; i >= 0; i--) {
            sum = b(i,0);
            for (int j = i+1; j < n; j++) {
                sum -= m_a(i,j)*b(j,0);
            }
            T a = m_a(i,i);
            b.set(i,0, sum / a);
         }
//std::wcout << "b:\n" << b.toString() << std::endl;
          #ifndef NDEBUG
          //std::wcout << "copy:\n" << copy.toString() << std::endl;
         UMat<T> newcopy = m_original.multiply(b);
         //std::wcout << "newcopy:\n" << newcopy.toString() << std::endl;
         utiassert(isIdentical(copy, newcopy));
         #endif

    }
    UMat<T> solveSystem(const UMat<T>& input) const
    {
         UMat<T> b = input;
       solveSystemInPlace(b);
         return b;
    }
    CVector4Double solveSystem(const CVector4Double& bdl) const
    {
        UMat<T> b(4,1);
       b.set(0,0, bdl.x);
       b.set(1,0, bdl.y);
       b.set(2,0, bdl.z);
       b.set(3,0, bdl.w);
       solveSystemInPlace(b);
       CVector4Double res;
       res.x = b(0,0);
       res.y = b(1,0);
       res.z = b(2,0);
       res.w = b(3,0);

       return res;
    }
    CVector3Double solveSystem(const CVector3Double& bdl) const
    {
      UMat<T> b(3, 1);
       b.set(0, 0,bdl.x);
       b.set(1, 0,bdl.y);
       b.set(2,0, bdl.z);
       solveSystemInPlace(b);
       CVector3Double res;
       res.x = b(0,0);
       res.y = b(1,0);
       res.z = b(2,0);

       return res;
    }
    CVector2Double solveSystem(const CVector2Double& bdl) const
    {
       UMat<T> b(2,1);
       b.set(0,0, bdl.x);
       b.set(1,0, bdl.y);
       solveSystemInPlace(b);
       CVector2Double res;
       res.x = b(0,0);
       res.y = b(1,0);

       return res;
    }
    T getDeterminant()
    {
        T d = m_signum;
        int n = m_a.n();
        for (int j = 0; j < n; j++) {
                d *= m_a(j,j);
        }
        return d;
    }

};

template<class T> T determinant(const UMat<T>& d)
{
        CULUDecomposition<T> decompose(d);
	return decompose.getDeterminant();
}

template<class T> UMat<T> solveSystem(const UMat<T>& M, const UMat<T>&b)
{
    CULUDecomposition<T> decompose(M);
    UMat<T> res = decompose.solveSystem(b);

    return res;
}

template<class T> T trace(const UMat<T>& d)
{
        T ret(0);
        assert(d.sizex() == d.sizey());
        int s = d.sizex();
        for (int i = 0; i < s; i++) {
            ret += d.get(i,i);
        }
return ret;
}
typedef UMat<double> UMatDouble;

#endif
