#include "numeric.h"
#include "odelibrary.h"

bool collideBoxBoxAdvanced(const COBB& box1, const COBB& box2, Array<CollisionContactPoint>& rescontactpoints, int& rescount, double&maxpenetration)
{
    dxBox o1;
    dxBox o2;
    {
        dxBox &o = o1;
        const COBB & box = box1;
    o.final_posr.pos[0] = box.m_origin.x;
    o.final_posr.pos[1] = box.m_origin.y;
    o.final_posr.pos[2] = box.m_origin.z;
    o.gflags = 0;
    o.type = dBoxClass;
    o.side[0] = box.m_halfu*2.0f;
    o.side[1] = box.m_halfv*2.0f;
    o.side[2] = box.m_halfw*2.0f;

    /*Just from the definition (dReal[12]) it is not clear whether it is in
    row-major or column-major format. I'm not actually sure which is which - but
    I'd guess that the example "[col*4+row] in general" indicates it is in
    column-major format. (Though I'd be happy if somebody cleared that up for me
    once and for all.)

    > dMatrix3 is defined as dReal[12].  So to access column 3, row 1, you would
    > ask for [3*4+1], or [col*4+row] in general.  Why the strangeness?  It's a
    > 3x3 matrix with SIMD padding, so it needs to have rows of length 4.*/


    o.final_posr.R[0*4+0] = box.m_bu.x;
    o.final_posr.R[0*4+1] = box.m_bu.y;
    o.final_posr.R[0*4+2] = box.m_bu.z;
    o.final_posr.R[0*4+3] = 0;


    o.final_posr.R[1*4+0] = box.m_bv.x;
    o.final_posr.R[1*4+1] = box.m_bv.y;
    o.final_posr.R[1*4+2] = box.m_bv.z;
    o.final_posr.R[1*4+3] = 0;


    o.final_posr.R[2*4+0] = box.m_bw.x;
    o.final_posr.R[2*4+1] = box.m_bw.y;
    o.final_posr.R[2*4+2] = box.m_bw.z;
    o.final_posr.R[2*4+3] = 0;
    }

    {
        dxBox &o = o2;
        const COBB & box = box2;
    o.final_posr.pos[0] = box.m_origin.x;
    o.final_posr.pos[1] = box.m_origin.y;
    o.final_posr.pos[2] = box.m_origin.z;
    o.gflags = 0;
    o.type = dBoxClass;
    o.side[0] = box.m_halfu*2.0f;
    o.side[1] = box.m_halfv*2.0f;
    o.side[2] = box.m_halfw*2.0f;

    o.final_posr.R[0*4+0] = box.m_bu.x;
    o.final_posr.R[0*4+1] = box.m_bu.y;
    o.final_posr.R[0*4+2] = box.m_bu.z;
    o.final_posr.R[0*4+3] = 0;


    o.final_posr.R[1*4+0] = box.m_bv.x;
    o.final_posr.R[1*4+1] = box.m_bv.y;
    o.final_posr.R[1*4+2] = box.m_bv.z;
    o.final_posr.R[1*4+3] = 0;


    o.final_posr.R[2*4+0] = box.m_bw.x;
    o.final_posr.R[2*4+1] = box.m_bw.y;
    o.final_posr.R[2*4+2] = box.m_bw.z;
    o.final_posr.R[2*4+3] = 0;
    }

    int flags=rescontactpoints.size();
    StaticArray<dContactGeom, 64>  contact;
    int skip = sizeof(dContactGeom);

    int res = dCollideBoxBox (&o1, &o2, flags,
                      contact.getData(), skip);

    rescount = umin(res, rescontactpoints.size());

    maxpenetration = 0.0;
    for (int i = 0; i < rescount; i++) {
        CollisionContactPoint& respoint = rescontactpoints[i];
        const dContactGeom& cgeom = contact[i];
        maxpenetration = umax(maxpenetration, cgeom.depth);
        assert (cgeom.g1 == &o1);
        assert(cgeom.g2 == &o2);
        respoint.m_worldposition.x = cgeom.pos[0];
        respoint.m_worldposition.y = cgeom.pos[1];
        respoint.m_worldposition.z = cgeom.pos[2];
        respoint.m_penetrationnormal.x = cgeom.normal[0];
        respoint.m_penetrationnormal.y = cgeom.normal[0];
        respoint.m_penetrationnormal.z = cgeom.normal[0];
        respoint.m_penetrationdepth = cgeom.depth;

    }
    return res != 0;
}
void SRasterScreen::setup(int w, int h, float openangle)
{

    m_width = w;
    m_height = h;
    m_widthi = w;
    m_heighti = h;

    m_halfwidth = m_width /2.0f;
    m_halfheight = m_height / 2.0f;
    m_halfwidthi = m_widthi / 2;
    m_halfheighti = m_heighti / 2;

    m_focal = m_halfwidth / tan(openangle*pi180);
}

void SRasterTriangle::createSpawnCollection(const SRasterScreen& screen, SSpanCollection& s1, SSpanCollection &s2)
{
    s1.m_valid = false;
    s2.m_valid = false;
    //culling
    if (m_vertices[0].dotProduct(m_plane.cvVector) < 0.0f) return;


    SRaster3DLine x1_x2(m_vertices[0], m_vertices[1]);
    SRaster3DLine x2_x3(m_vertices[1], m_vertices[2]);
    SRaster3DLine x3_x1(m_vertices[2], m_vertices[0]);

    SRasterInterval i_x1_x2 = x1_x2.getPaintInterval(screen);
    SRasterInterval i_x2_x3 = x2_x3.getPaintInterval(screen);
    SRasterInterval i_x3_x1 = x3_x1.getPaintInterval(screen);

    SRasterInterval i_full = union_interval(i_x1_x2, i_x2_x3);

    //TODO: Intelligente fallunterscheidung
}



void SRasterTriangle::calculatePlane()
{

   m_plane.getPlane(m_vertices[0], m_vertices[1], m_vertices[2]);

}
