/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef NURBS_H
#define NURBS_H

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "matrices.h"
#include "utilists.h"


#include "utivectordouble.h"
#include "advarray.h"

#include <vector>
#include <algorithm>


typedef double UTIFLOAT;

int binomial(int n, int k);
void initBinomialArray(Array2D<int> &b, int s);
template<class T> Polynom<T> bernsteinPolynomCache(int i, int n)
{
    utiassert(false);
}

template<> Polynom<double> bernsteinPolynomCache(int i, int n);


inline int bin(int n, int k)
{
    utiassert(k <= n);
    static Array2D<int> b;
    if (b.width() == 0)
    {
        initBinomialArray(b, 4);
    }
    if (n <= b.width())
    {
        return b(n,k);
    }
    else
    {
        return binomial(n,k);
    }
}


struct SKnotVectorResult
{
    int m_index;
    float m_a;
    int m_a_count;
    float m_b;
    int m_b_count;
    SKnotVectorResult()
    {

        m_a = -1;
        m_b = -2;
        m_a_count=-1;
        m_b_count=-1;
        m_index = -1;
    }
};

class KnotVector : public ReferenceADR
{
protected:
    Array<int> m_compile_count;
    Array<UTIFLOAT> m_compile_knotvectorpos;
    Array<int> m_compile_index;
    Array<UTIFLOAT> m_knotvector;

//static Array<UTIFLOAT> m_left;
    // static Array<UTIFLOAT> m_right;
    //static Array<UTIFLOAT> m_N;
    //static Array2D<UTIFLOAT> m_ndu;
    //static Array2D<UTIFLOAT> m_a;
    //static Array<UTIFLOAT> m_ND;
    bool m_compiled;
    bool m_isbezier;
    SKnotVectorResult m_lastresult;
    // UTIFLOAT getBasisFunctionSimpleIntern(int i, int p, int degree, UTIFLOAT u);
public:
    KnotVector(int p);
    KnotVector(const Array<UTIFLOAT>& pos);
    static KnotVector* createBezierKnotVector(int p);
    void setEntry(int i, UTIFLOAT value);
    inline UTIFLOAT getEntry(int i) const
    {
        return m_knotvector[i];
    }
    void setBezier(bool b);
    bool getBezier() const;

    void compile();
    bool isCompiled() const;
    void findKnotVectorPosition(UTIFLOAT v, SKnotVectorResult& result) const;
    SKnotVectorResult findKnotVectorPositionSimple(UTIFLOAT v) const;
    int size() const;
    UTIFLOAT getMinElement() const;
    UTIFLOAT getMaxElement() const;
    int size_single() const;
    UTIFLOAT getSingle(int i) const;
    int getSingleMultiplicy(int i) const;
    UTIFLOAT getBasisFunctionSimple(int i, int p, UTIFLOAT u) const;

    UTIFLOAT dersBasisFunctionSimple(int i, int p, UTIFLOAT u, int n) const;
    void dersBasisFunctionSimple(int i, int p, UTIFLOAT u, Array<UTIFLOAT>& ders, int derv) const;
    //void dersBasisFunctionSimple(int i, UTIFLOAT u, int p, Array<UTIFLOAT>& ders);
    UTIFLOAT getBasisFunction(int i, int p, UTIFLOAT u) const;
    void dersBasisFunction(int i, int p, UTIFLOAT u, Array<UTIFLOAT>& ders, int n) const;
    void calcBasisFunctions(int i, UTIFLOAT u, Array<UTIFLOAT>& N, int p) const;
    void dersBasisFunctions(int i, UTIFLOAT u, Array2D<UTIFLOAT> &ders, int n, int p) const;
    void copyToKnot(int ownstart, int targetstart, int count, KnotVector* target) const;
    void print() const;
};

class SubdivisionNode : public ReferenceUtiOB
{
public:
    uti_refptr<ReferenceUtiOB> m_nurbs;
    uti_refptr<ReferenceUtiOB> m_box;
    double m_tolerance;
    uti_refptr<SubdivisionNode> m_left;
    uti_refptr<SubdivisionNode> m_right;
    SubdivisionNode() {
       m_tolerance = 0.0;
    }
};

class SubdivisionContainer : public ReferenceUtiOB
{
public:
    CReferenceList<SubdivisionNode> m_root;

};

template <class T> void curveKnotInsertTemplate(T& accessor, UTIFLOAT newknotvalue, int amount)
{
    int k;
    // if (before_positioninknotvector == -2) {
    KnotVector* m_knotvector = accessor.getKnotVector();
    int m_degree = accessor.getDegree();
    int m_verticessize = accessor.getVertexCount();
    SKnotVectorResult result;
    m_knotvector->findKnotVectorPosition(newknotvalue, result);
    k = result.m_index;
    utiassert(k <= m_knotvector->size());
    int s = 0;
    if (fabs(result.m_a-newknotvalue) < 0.0001)
    {
        newknotvalue = result.m_a;
        s =result.m_a_count;
    }

    // k -= s;
    // k += 1;

    utiassert(amount+s <= m_degree);
    int olds = s;
    s = 0;

    // }
    //  Array<UTIFLOAT> newknotarray;
    int np = m_verticessize-1;
    int mp = np+m_degree+1;
    utiassert(mp == m_knotvector->size()-1);
    //int nq = np+amount;

    // newknotarray.setSize(m_knotvector->size()+amount);
    KnotVector* vnew = new KnotVector(m_knotvector->size()+amount);

    //copyTo(0, 0, before_positioninknotvector+1);

    for (int i = 0; i <= k; i++)
    {
        vnew->setEntry(i, m_knotvector->getEntry(i));
    }
    for (int i = 1; i <= amount; i++)
    {
        vnew->setEntry(k+i, newknotvalue);
    }
    //copyTo(before_positioninknotvector+1, i+amount, mp-(before_positioninknotvector+1)+1);
    for (int i = k+1; i <= mp; i++)
    {
        vnew->setEntry(i+amount, m_knotvector->getEntry(i));
    }


    vnew->compile();
    accessor.createCurve(vnew, k+1-olds); //Check dies auch beim erweitern, muss ich -s machen?
    //Nurbs1D* nurbs = new Nurbs1D(vnew, m_degree);
    /*copyTo(0,0,before_positioninknotvector-degree+1, nurbs);
    copyTo(before_positioninknotvector-s, before_positioninknotvector-s+amount, np-(before_positioninknotvector-s)+1, nurbs);
    */
    typename T::SOneVertex tempvertex, tempvertex2;
    for (int i = 0; i <= k-m_degree; i++)
    {

        accessor.getVertex(i, tempvertex);
        accessor.setVertex(i, tempvertex);
    }
    for (int i = k-s; i <= np; i++)
    {
        accessor.getVertex(i, tempvertex);
        accessor.setVertex(i+amount, tempvertex);
    }
    Array<typename T::SOneVertex> Rw(m_degree+1);
    for (int i = 0; i <= m_degree-s; i++)
    {
        // Rw[i] = m_vertices[k-m_degree+i];

        accessor.getVertexWeight(k-m_degree+i, Rw[i]);
    }
    int L = k-m_degree;
    for (int j = 1; j <= amount; j++)
    {
        L = k-m_degree+j;
        for (int i = 0; i <= m_degree-j-s; i++)
        {
            typename T::Scalar alpha = (newknotvalue-m_knotvector->getEntry(L+i))/(m_knotvector->getEntry(i+k+1)-m_knotvector->getEntry(L+i));
            accessor.scal(tempvertex, Rw[i+1], alpha);
            accessor.scal(tempvertex2, Rw[i], (1.0-alpha));
            accessor.add(Rw[i], tempvertex, tempvertex2);

            // Rw[i] = (Rw[i+1])*alpha + (Rw[i])* (1.0-alpha);
        }
        accessor.setVertexProject(L, Rw[0]);
        accessor.setVertexProject(k+amount-j+s, Rw[m_degree-j-s]);
        // nurbs->setVertex(L, Rw[0]);
        //nurbs->setVertex(k+amount-j+s, Rw[m_degree-j-s]);
    }
    for (int i = L+1; i < k-s; i++)
    {
        accessor.setVertexProject(i, Rw[i-L]);
    }

}

template <class T> void degreeElevateCurve(T& accessor, int t)
{
    int p = accessor.getDegree();
    int n = accessor.getVertexCount()-1;
    int m = n+p+1;
    int ph = p+t;
    int ph2 = p/2;
    Array2D<typename T::Scalar> bezahlfs(p+t+1,p+1);
    for (int i = 1; i <= ph2; i++)
    {
        typename T::Scalar inv = 1.0/bin(ph,i);
        int mpi = umin(p, i);
        for (int j = umax(0,i-t); j <= mpi; j++)
        {
            bezahlfs[i][j] = inv*bin(p,j)*bin(t,i-j);
        }
    }
    for (int i = ph2+1; i <= ph-1; i++)
    {
        int mpi = umin(p,i);
        for (int j = umax(0,i-t); j <= mpi; j++)
        {
            bezahlfs[i][j] = bezahlfs[ph-i][p-j];
        }
    }
    int mh = ph;
    int kind = ph+1;
    int r = -1;
    int a = p;
    int b = p+1;
    int cind = 1;
    KnotVector* U = accessor.getKnotVector();
    UTIFLOAT ua = U->getEntry(0);
    Array<UTIFLOAT> Uh;
    Array<typename T::SOneVertex> Qw;
    Array<typename T::SOneVertex> bpts(p+1);
    Array<typename T::SOneVertex> alfs(p+1);
    Array<typename T::SOneVertex> ebpts(p+t+1);
    Array<typename T::SOneVertex> Nextbpts(p-1);
    Array<UTIFLOAT> alphas(p-1);
    accessor->getVertex(0, Qw[0]);
    for (int i = 0; i < ph; i++)
    {
        Uh[i] = ua;
    }
    for (int i=0; i <= p; i++)
    {
        accessor->getVertex(i, bpts[i]);
    }

    typename T::SOneVertex tempvertex;

    while (b < m)
    {
        int i = b;
        while (b < m && U->getEntry(b) == U->getEntry(b+1))
        {
            b = b+1;
        }
        int mul = b-i+1;
        mh = mh+mul+t;
        UTIFLOAT ub = U->getEntry(b);
        int oldr = r;
        r = p-mul;
        int lbz=1;
        int rbz = ph;
        if (oldr > 0)
        {
            lbz = (oldr+2)/2;
        }
        else
        {
            lbz = 1;
        }
        if (r > 0)
        {
            rbz = ph-(r+1)/2;
        }
        else
        {
            rbz = ph;
        }
        if (r > 0)
        {
            int numer = ub-ua;
            for (int k = p; k > mul; k--)
            {
                alfs[k-mul-1] = numer/(U->getEntry(a+k)-ua);
            }
            for (int j = 1; j <= r; j++)
            {
                int save = r-j;
                int s= mul+j;
                for (int k = p; k >= s; j--)
                {
                    bpts[k] = alfs[k-s]*bpts[k]+ (1.0-alfs[k-s])*bpts[k-1];
                    //accessor.scal(tempvertex, bpts[k])
                }
                Nextbpts[save] = bpts[p];
            }
        }//End of "insert knot"
        //Degree evevate Bezier
        for (int i = lbz; i <= ph; i++)
        {
            accessor.initZero(ebpts[i]);
            int mpi = umin(p,i);
            for (int j=umax(0, i-t); j <= mpi; j++)
            {
                ebpts[i] = ebpts[i] + bezahlfs[i][j]*bpts[j];
            }

        } //End of degree evelating bezier
        if (oldr > 1)
        {
            int first = kind-2;
            int last = kind;
            UTIFLOAT den = ub - ua;
            UTIFLOAT bet = (ub-Uh[kind-1])/den;

            for (int tr = 1; tr < oldr; tr++)
            {
                //Knot removal
                int i = first;
                int j = last;
                int kj = j-kind+1;
                while (j-i > tr)
                {
                    if (i < cind)
                    {
                        UTIFLOAT alf = (ub-Uh[i]) / (ua-Uh[i]);
                        Qw[i] = alf*Qw[i] + (1.0-alf)*Qw[i-1];

                    }
                    if (j >= lbz)
                    {
                        if (j-tr <= kind-ph-oldr)
                        {
                            float gam = (ub-Uh[j-tr])/den;
                            ebpts[kj] = gam* ebpts[kj]+(1.0-gam)*ebpts[kj+1];
                        }
                        else
                        {
                            ebpts[kj] = bet*ebpts[kj]+(1.0-bet)*ebpts[kj+1];
                        }
                    }
                    i = i+1;
                    j = j-1;
                    kj = kj-1;

                }
                first = first-1;
                last = last+1;
            }

        } /*End of removing knot, u=U[a]*/
        if (a != p)
        {
            for (int i = 0; i < ph-oldr; i++)
            {
                Uh[kind] = ua;
                kind = kind + 1;
            }
        }
        for (int j = lbz; j<= rbz; j++)
        {
            Qw[cind] = ebpts[j];
            cind = cind+1;
        }
        if (b < m)
        {
            for (int j = 0; j < r; j++)
            {
                bpts[j] = Nextbpts[j];
            }
            for (int j = r; j <= p; j++)
            {
                //bpts[j] = Pw[b-p+j];
                accessor->getVertex(b-p+j, bpts[j]);
            }
            a=b;
            b = b+1;
            ua = ub;
        }
        else
        {
            for (int i = 0; i <= ph; i++)
            {
                Uh[kind+i] = ub;
            }
        }


    }

    int nh = mh-ph-1;
    //m_degree = degree;
    //m_vertices.setSize(u->size()-degree-1);

    int knotsize = nh+p+1;
    KnotVector* vnew = new KnotVector(knotsize);

    //copyTo(0, 0, before_positioninknotvector+1);

    for (int i = 0; i < knotsize; i++)
    {
        vnew->setEntry(i, Uh[i]);
    }

    vnew->compile();
    accessor.createCurve(vnew, nh);
    for (int i = 0; i< nh; i++)
    {
        accessor.setVertex(i, Qw[i]);
    }
}

template<class T> void decomposeCurveTemplate(T& accessor)
{
    //typedef SAccessNurbs1DToBezier T;
    KnotVector* U = accessor.getKnotVector();

    int p = accessor.getDegree();
    int n = accessor.getVertexCount()-1;
    int m = n+p+1;
    int a = p;
    int b = p+1;
    int nb = 0;
    Array<typename T::Scalar> alphas(p+1);
    typename T::Scalar numer;
    //  accessor.addBezierSegment();
    typename T::SOneVertex tempvertex, tempvertex2;
    for (int i = 0; i <= p; i++)
    {
        accessor.getVertexWeight(i, tempvertex);
        accessor.setBezierVertex(nb, i, tempvertex);
    }
    while (b < m)
    {
        int i = b;
        while (b < m && U->getEntry(b+1) == U->getEntry(b))
        {
            b++;
        }
        int mult = b-i+1;
        if (mult < p)
        {
            numer = U->getEntry(b)-U->getEntry(a);
            for (int j = p; j > mult; j--)
            {
                alphas[j-mult-1] = numer / (U->getEntry(a+j)-U->getEntry(a));
            }
            int r = p - mult;
            for (int j = 1; j <= r; j++)
            {
                int save = r-j;
                int s = mult + j;
                for (int k = p; k >= s; k--)
                {
                    UTIFLOAT alpha = alphas[k-s];
                    accessor.getBezierVertex(nb, k, tempvertex);
                    accessor.scal(tempvertex, tempvertex, alpha);

                    accessor.getBezierVertex(nb, k-1, tempvertex2);
                    accessor.scal(tempvertex2, tempvertex2, 1.0-alpha);

                    accessor.add(tempvertex, tempvertex, tempvertex2);
                    accessor.setBezierVertex(nb, k, tempvertex);
                }
                if (b < m)
                {
                    accessor.getBezierVertex(nb, p, tempvertex);
                    accessor.setBezierVertex(nb+1, save, tempvertex);
                }
            }
        }
        nb = nb + 1;
        if (b < m)
        {
            for (int i = p-mult; i <= p; i++)
            {
                accessor.getVertexWeight(b-p+i, tempvertex);
                accessor.setBezierVertex(nb, i, tempvertex);
            }
            a = b;
            b = b + 1;
        }
    }
}

template<class T> typename T::Scalar getToleranceTemplate(T& accessor, bool* p_allonbaseline=NULL)
    {
        utiassert(accessor.getKnotVector()->getBezier());
        int n = accessor.getVertexCount()-1;
        typename T::SOneVertex temp, avertex, bvertex;
        accessor.getVertex(0, avertex);
        typename T::Vector a = avertex.m_pos;
        accessor.getVertex(n, bvertex);
        typename T::Vector b = bvertex.m_pos;
        typename T::Vector dir = b-a;
        typename T::Scalar distance;
        typename T::Scalar length = accessor.m_controldata.normalize(dir);
        if (length < 0.0000001)
        {
            return 0.0;
        }
        bool allonbaseline = true;

        typename T::Scalar maxdistancetobaseline=0;
        typename T::Vector r, lineposition;
        typename T::Scalar pos, maxweight, sk;
        maxweight = 0.0;

        for (int i = 1; i < n; i++)
        {
            typename T::SOneVertex temp;
        accessor.getVertex(i, temp);
            const typename T::Vector & v = temp.m_pos;
            r = v-a;
            maxweight = umax(temp.m_weight, maxweight);
            pos = accessor.m_controldata.dot(dir, r);
            if (pos < 0 || pos > length)
            {
                allonbaseline = false;
                break;
            }
            lineposition = a+dir*pos;

            distance = accessor.m_controldata.distanceTo(lineposition, v);
            maxdistancetobaseline = umax(distance, maxdistancetobaseline);
        }

        if (p_allonbaseline){
            *p_allonbaseline = allonbaseline;
        }
        if (allonbaseline)
        {
            typename T::Scalar value1 = umax(1.0/avertex.m_weight, 1.0/bvertex.m_weight);
            value1 *= maxweight;
            value1 *= (pow(T::Scalar(2), T::Scalar(n))-1);
            value1 += 1;
            value1 = 1.0/value1;
            value1 = 1.0-value1;

            const typename T::Scalar LINETOLERANCE= 0.01;
            return maxdistancetobaseline*value1;

        }
        return 1E+10;

    }

template<class Vector> struct SBounds
{
    Vector m_min;
    Vector m_max;

};

template<class Vector> struct SOrientedBounds
{
    SBounds<Vector> m_oriented;
    UMatDouble m_A;
    //uti_refptr<CLUDecomposition> m_workspace;
};



template <class ControlData> typename ControlData::Scalar getMinDistanceToBounds(ControlData& m_controldata, const
        SBounds<typename ControlData::Vector> &bounds, const typename ControlData::Vector& q)
{
    //return typename ControlData::Scalar(0);
    typename ControlData::Scalar distance = 0;
    int d = m_controldata.getDimensions();
    typename ControlData::Scalar min_value, max_value, q_value;
    // bool contains=true;
    for (int i = 0; i < d; i++)
    {
        min_value = m_controldata.getVertexIndex(bounds.m_min, i);
        max_value = m_controldata.getVertexIndex(bounds.m_max, i);
        q_value = m_controldata.getVertexIndex(q, i);
        if (q_value < min_value)
        {
            // contains = false;
            distance += sqr(min_value-q_value);
        }
        else if (q_value > max_value)
        {
            // contains = false;
            distance += sqr(q_value-max_value);
        }
        else
        {
            //intern
        }
    }
    return sqrt(distance);
    /* if (contains) {
         return typename ControlData::Scalar(0);
     }
     assert(d <= 30);
     int amount = intpow(2, d);
     typename ControlData::Vector checkpoint;

     m_controldata.init(checkpoint);

     for (int i = 0; i < amount; i++) {
         int bit = 1;

         for (int dim = 0; dim < d; dim++) {
             if ((i & bit) != 0) {
                 q_value = m_controldata.getVertexIndex(bounds.m_min, dim);
             } else {
                 q_value = m_controldata.getVertexIndex(bounds.m_max, dim);

             }
              m_controldata.setVertexIndex(checkpoint, dim, q_value);
         }
         q_value = m_controldata.distanceTo(q, checkpoint);
         mindistance = std::min(mindistance, q_value);

     }
    */
    //  return mindistance;
}

template<class ControlData> class CBaseBoundsContainer : public ReferenceUtiOB
{
public:
    virtual typename ControlData::Scalar getMinimumDistance(const ControlData& m_controldata, const typename ControlData::Vector& vector)=0;
};

template<class ControlData> class CBoundsContainer : public CBaseBoundsContainer<ControlData>
{
public:
    SBounds<typename ControlData::Vector> m_bounds;
    typename ControlData::Scalar getMinimumDistance(const ControlData& m_controldata, const typename ControlData::Vector& vector)
    {
        return getMinDistanceToBounds(m_controldata, m_bounds, vector);

    }

};
template<class ControlData> class COrientedBoundsContainer : public CBaseBoundsContainer<ControlData>
{
public:
    SOrientedBounds<typename ControlData::Vector> m_orientedbounds;
    typename ControlData::Scalar getMinimumDistance(const ControlData& m_controldata, const typename ControlData::Vector& vector)
    {
        typename ControlData::Vector temp = m_orientedbounds.m_workspace->solveSystem(vector);
        return getMinDistanceToBounds(m_controldata, m_orientedbounds.m_oriented, temp);

    }
};

template <class ControlData> typename ControlData::Scalar getMinDistanceToLine(ControlData& m_controldata, const
        typename ControlData::Vector& a, const typename ControlData::Vector & b, const typename ControlData::Vector& q)
{
    typename ControlData::Vector dir, d1;
    dir = b-a;
    typename ControlData::Scalar length = m_controldata.normalize(dir);
    if (length < 0.0000001)
    {
        return 0;
    }

    d1 = q-a;
    typename ControlData::Scalar pos = m_controldata.dot(dir, d1)/length;
    if (pos < 0.0) pos = 0.0;
    if (pos > 1.0) pos = 1.0;
    return pos;
}


template <class ControlData> bool IntersectLines(ControlData& m_controldata,
        const typename ControlData::Vector& P1,
        const typename ControlData::Vector& T1,
        const typename ControlData::Vector& P2,
        const typename ControlData::Vector& T2,
        typename ControlData::Scalar& final_u1res,
        typename ControlData::Scalar& final_u2res,
        typename ControlData::Vector& Pres)
{
    typename ControlData::Scalar u1res;
    typename ControlData::Scalar u2res;
    typename ControlData::Scalar  A, B, C, D, E, F, p1i, t1i, p2i, t2i, deter;
    int dim = m_controldata.getDimensions();
    A=0;
    B=0;
    C=0;
    D=0;
    E=0;
    F=0;
    for (int i = 0; i < dim; i++)
    {
        p1i = m_controldata.getVertexIndex(P1,i);
        t1i = m_controldata.getVertexIndex(T1,i);
        p2i = m_controldata.getVertexIndex(P2,i);
        t2i = m_controldata.getVertexIndex(T2,i);
        A += t1i*t1i;
        E += -t2i*t2i;
        B += -t2i*t1i;
        D += t2i*t1i;

        C += p1i*t1i-p2i*t1i;
        F += p1i*t2i-p2i*t2i;
    }
    utiassert(fabs(D+B)<=0.0001);
    deter = A*E+B*B;
    if (fabs(deter)< 0.0001)
    {
        u1res = 0;
        u2res = 0;//?
        final_u1res = u1res;
        final_u2res = u2res;
        return false;
    }
    else
    {
        deter = 1.0/deter;
        u1res = deter*(E*-C+B*F);
        u2res = deter*(D*C-A*F);
    }
    m_controldata.init(Pres);
    Pres = P1+T1*u1res;



    final_u1res = u1res;
    final_u2res = u2res;
    return true;

}

template<class Scalar> bool sortNURBS(const std::pair<int, Scalar>& d1, const std::pair<int, Scalar>& d2)
{
    return d1.second < d2.second;
}



template <class ControlData> class Nurbs1D : public ReferenceUtiOB
{
public:
    typedef typename ControlData::Vector Vector;
    typedef typename ControlData::Scalar Scalar;
    typedef ControlData CurveControlData;
    ControlData m_controldata;
    struct SNurbsVertex
    {
        Vector m_pos;
        Scalar m_weight;
        SNurbsVertex operator*(Scalar v) const
        {
            SNurbsVertex r;
            r.m_pos = m_pos*v;
            r.m_weight = m_weight*v;
            return r;

        }

        SNurbsVertex operator+(const SNurbsVertex& v) const
        {
            SNurbsVertex r;
            r.m_pos = m_pos+v.m_pos;
            r.m_weight = m_weight+v.m_weight;
            return r;

        }
    };

    struct SAccessNurbs1D
    {
    public:
        const Nurbs1D* m_oldcurve;
        Nurbs1D* m_newcurve;
        ControlData m_controldata;

        typedef typename ControlData::Vector Vector;
        typedef typename ControlData::Scalar Scalar;
        typedef SNurbsVertex SOneVertex;
        inline void add(SOneVertex& res, const SOneVertex &a, const SOneVertex &b)const
        {
            res = a + b;
        }

        inline void scal(SOneVertex& res, const SOneVertex &a, Scalar weight)const
        {
            res = a* weight;
        }
        inline void getVertex(int index, SOneVertex& v)const
        {
            v = m_oldcurve->getVertex(index);
        }
		inline void getVertexWeight(int index, SOneVertex& v)const
        {
            v = m_oldcurve->getVertex(index);
			v.m_pos *= v.m_weight;
        }
        inline int getVertexCount()const
        {
            return m_oldcurve->getVertexCount();
        }

        inline int getDegree()const
        {
            return m_oldcurve->getDegree();
        }

        inline KnotVector* getKnotVector()const
        {
            return m_oldcurve->getKnotVector();
        }

        inline void setVertex(int index, const SOneVertex& v)const
        {
            m_newcurve->setVertex(index, v);
        }
		inline void setVertexProject(int index, const SOneVertex& v)const
        {
		    SOneVertex b = v;
			b.m_pos /= b.m_weight;
            m_newcurve->setVertex(index, v);
        }

        inline void createCurve(KnotVector* newknot, int insertpoint)
        {
            m_newcurve = new Nurbs1D(newknot, getDegree(), m_controldata);
        }
    };
    struct SAccessNurbs1DToBezier : public SAccessNurbs1D
    {
        CReferenceList<Nurbs1D> m_bezierlist;
        //uti_refptr<KnotVector> m_sharedknotvector;
        void addBezierSegment()
        {
            int p = getDegree();
            KnotVector* oldkv = getKnotVector();
            //if (m_sharedknotvector.get() == NULL)
            //{
            double start = oldkv->getSingle(m_bezierlist.size());
            double ende = oldkv->getSingle(m_bezierlist.size()+1);
            KnotVector* newvector= new KnotVector((p+1)*2);
            for (int i = 0; i <= p; i++)
            {
                newvector->setEntry(i, start);
                newvector->setEntry(i+(p+1), ende);
            }
            newvector->setBezier(true);
            newvector->compile();
            //}

            Nurbs1D* n = new Nurbs1D(newvector, p, this->m_oldcurve->getControlData());
            m_bezierlist.push_back(n);

        }

        void finalizeBezier()
        {
            for (int index = 0; index < m_bezierlist.size(); index++) {
                Nurbs1D* n = &m_bezierlist[index];
                for (int i = 0; i < n->getVertexCount(); i++) {
                    n->m_vertices[i].m_pos /= n->m_vertices[i].m_weight;

                }
            }
        }


        void setBezierVertex(int segment, int index, const typename SAccessNurbs1D::SOneVertex& v)
        {
            while (segment >= m_bezierlist.size())
            {
                addBezierSegment();
            };
            m_bezierlist[segment].setVertex(index, v);

        }

        void getBezierVertex(int segment, int index, typename SAccessNurbs1D::SOneVertex& v)const
        {
            v = m_bezierlist[segment].getVertex(index);

        }/*
		void setBezierVertexProject(int segment, int index, const typename SAccessNurbs1D::SOneVertex& v)
        {
            while (segment >= m_bezierlist.size())
            {
                addBezierSegment();
            };
			 typename SAccessNurbs1D::SOneVertex temp;
			 temp = v;
			 temp.m_pos /= temp.m_weight;
            m_bezierlist[segment].setVertex(index, v);

        }

        void getBezierVertexWeight(int segment, int index, typename SAccessNurbs1D::SOneVertex& v)const
        {
            v = m_bezierlist[segment].getVertex(index);
			v.m_pos *= v.m_weight;

        }*/
    };


    struct SAccessNurbs1DSplit
    {
        ControlData m_controldata;
        const Nurbs1D* m_oldcurve;
        Nurbs1D* m_newcurve1;
        Nurbs1D* m_newcurve2;
        int m_newcurve1end;
        int m_newcurve2start;

        typedef typename ControlData::Vector Vector;
        typedef typename ControlData::Scalar Scalar;
        typedef SNurbsVertex SOneVertex;
        inline void add(SOneVertex& res, const SOneVertex &a, const SOneVertex &b)const
        {
            res = a + b;
        }

        inline void scal(SOneVertex& res, const SOneVertex &a, Scalar weight)const
        {
            res = a* weight;
        }
        inline void getVertex(int index, SOneVertex& v)const
        {
            v = m_oldcurve->getVertex(index);
        }
        inline void getVertexWeight(int index, SOneVertex& v)const
        {
            v = m_oldcurve->getVertex(index);
            v.m_pos *= v.m_weight;
        }
        inline int getVertexCount()const
        {
            return m_oldcurve->getVertexCount();
        }

        inline int getDegree()const
        {
            return m_oldcurve->getDegree();
        }

        inline KnotVector* getKnotVector()const
        {
            return m_oldcurve->getKnotVector();
        }

        inline void setVertex(int index, const SOneVertex& v)const
        {
            if (index < m_newcurve1end)
            {
                m_newcurve1->setVertex(index, v);
            }
            if (m_newcurve2start <= index)
            {
                m_newcurve2->setVertex(index-m_newcurve2start, v);
            }
        }
        inline void setVertexProject(int index, const SOneVertex& v)const
        {
            SOneVertex temp = v;
            temp.m_pos /= temp.m_weight;
            if (index < m_newcurve1end)
            {
                m_newcurve1->setVertex(index, temp);
            }
            if (m_newcurve2start <= index)
            {
                m_newcurve2->setVertex(index-m_newcurve2start, temp);
            }
        }


        inline void createCurve(KnotVector* newknot, int insertpoint)
        {
//            newknot->print();
            m_newcurve1end = insertpoint;
            m_newcurve2start = insertpoint-1;
            int degree = getDegree();
            UTIFLOAT value = newknot->getEntry(insertpoint); //Splitpoint

            KnotVector* newknot1 = new KnotVector(insertpoint+degree+1);
            for (int i = 0; i < insertpoint; i++)
            {
                newknot1->setEntry(i, newknot->getEntry(i));
            }
            for (int i = insertpoint; i <= insertpoint+degree; i++)
            {
                newknot1->setEntry(i, value);
            }
//            newknot1->print();
            newknot1->setBezier(m_oldcurve->getKnotVector()->getBezier());
            newknot1->compile();

            KnotVector* newknot2 = new KnotVector(newknot->size()- insertpoint+1);
            for (int i = 0; i <= degree; i++)
            {
                newknot2->setEntry(i, value);
            }
            for (int i = insertpoint+degree; i < newknot->size(); i++)
            {
                newknot2->setEntry(i-( insertpoint)+1, newknot->getEntry(i));
            }

            // newknot2->print();

            newknot2->setBezier(m_oldcurve->getKnotVector()->getBezier());
            newknot2->compile();

            m_newcurve1 = new Nurbs1D(newknot1, degree, m_controldata);
            m_newcurve2 = new Nurbs1D(newknot2, degree, m_controldata);

        }
    };

    uti_refptr<KnotVector> m_knotvector;
    Array<SNurbsVertex> m_vertices;
    //Array<UTIFLOAT> m_basisfunctions;
    int m_degree;
    uti_refptr<SubdivisionContainer> m_subdivisioncontainer;

public:
    struct SSplitResult
    {
        Nurbs1D* first;
        Nurbs1D* second;
    };
    bool m_useinfluence;
    Nurbs1D(KnotVector* u, int degree, const ControlData& data=ControlData())
    {
        m_controldata = data;
        m_knotvector = u;
        utiassert(m_knotvector->isCompiled() == true);
        m_degree = degree;
        m_useinfluence = true;
        m_vertices.setSize(u->size()-degree-1);
        for (int i = 0; i < m_vertices.size(); i++)
        {
            m_controldata.init(m_vertices[i].m_pos);
        }

    }
   /* Vector getMeanOfAllVectors()
{
    Vector v;
    m_controldata.initZero(v);
    for (int i = 0; i < m_vertices.width(); i++) {
        for (int j = 0; j < m_vertices.height(); j++) {
            v += m_vertices(i,j).m_pos;
        }
    }
    v = v * (1.0/ (m_vertices.width()*m_vertices.height()));
    return v;
}
UMat<Scalar> getCovarianceMatrix()
{
    Vector mean = getMeanOfAllVectors();
    int d = m_controldata.getDimensions();
    Scalar N = m_vertices.width()*m_vertices.height();
    UMat<Scalar> res(d, d);
    for (int k = 0; k < d; k++) {
        for (int l = 0; l < d; l++) {
                Scalar currentvalue=0.0;
                Scalar mean1 = m_controldata.getVertexIndex(mean, k);
                Scalar mean2 = m_controldata.getVertexIndex(mean, l);
                for (int i = 0; i < m_vertices.width(); i++) {
                   for (int j = 0; j < m_vertices.height(); j++) {
                        Vector & x = m_vertices(i,j).m_pos;
                        currentvalue += (m_controldata.getVertexIndex(x, k)-mean1)*(m_controldata.getVertexIndex(x, l)-mean2);
                   }
                }
                currentvalue /= N;
                res.set(k,l, currentvalue);
        }
    }
    return res;

}*/
    inline const ControlData& getControlData()const
    {
        CHECKEXISTANCE;
        return m_controldata;
    }
     inline bool hasContainer()
    {
        return m_subdivisioncontainer.get() != NULL;
    }
    inline SubdivisionContainer* getSubdivisionContainer()
    {
        return m_subdivisioncontainer.get();
    }
    inline void setVertex(int index, const Vector& pos, const Scalar& weight)
    {
        CHECKEXISTANCE;
        m_vertices[index].m_pos = pos;
        m_vertices[index].m_weight = weight;
    }
    inline void setVertex(int index, const SNurbsVertex& v)
    {
        m_vertices[index] = v;
    }
    inline SNurbsVertex getVertex(int index)const
    {
        CHECKEXISTANCE;
        return m_vertices[index];
    }

    inline int getVertexCount()const
    {
        return m_vertices.size();
    }

    inline int getDegree()const
    {
        return m_degree;
    };
    inline KnotVector* getKnotVector()const
    {
        CHECKEXISTANCE;
        return m_knotvector.get();
    }

    inline Vector getPosition(int index)const
    {
        CHECKEXISTANCE;
        return m_vertices[index].m_pos;
    }

    inline Scalar getWeight(int index)const
    {
        CHECKEXISTANCE;
        return m_vertices[index].m_weight;
    }

    Vector calcReferenceValue(Scalar u)const
    {
        CHECKEXISTANCE;
        utiassert(false);
        Vector res;
        m_controldata.initZero(res);
        int n = m_vertices.size();
        Scalar wf = 0.0;

        for (int i = 0; i < n; i++)
        {
            Scalar v = m_knotvector->getBasisFunctionSimple(i, m_degree, u);
            res += v*m_vertices[i].m_weight*m_vertices[i].m_pos;
            wf += v*m_vertices[i].m_weight;
        }
        res /= wf;
        return res;
    }

    Vector calcValue(Scalar u)const
    {
        CHECKEXISTANCE;
        SKnotVectorResult result;

        m_knotvector->findKnotVectorPosition(u, result);
        int span = result.m_index;

        static Array<UTIFLOAT> m_basisfunctions;
        if (m_basisfunctions.size() < m_degree+1)
        {
            m_basisfunctions.setSize(m_degree+1);
        }


        m_knotvector->calcBasisFunctions(span, u, m_basisfunctions, m_degree);

        Vector res;
        m_controldata.initZero(res);

        Scalar wf(0.0);
        for (int j = 0; j <= m_degree; j++)
        {
            Scalar v = m_basisfunctions[j];
            int index = span-m_degree+j;
            res += v*m_vertices[index].m_weight*m_vertices[index].m_pos;
            wf += v*m_vertices[index].m_weight;
        }
        res /= wf;
        /*#ifndef NDEBUG
                Vector r = calcReferenceValue(u);
                utiassert(m_controldata.distanceTo(res, r) < 0.01);
        #endif*/
        return res;
    }
    void calcDervValueNoWeight(Scalar u, Array<SNurbsVertex>& CK)const
    {
        CHECKEXISTANCE;
        int d = CK.size()-1;
        int du = umin(m_degree, d);
        for (int i = m_degree+1; i <= d; i++)
        {
            m_controldata.initZero(CK[i].m_pos);
            CK[i].m_weight = 1.0;
        }
        SKnotVectorResult result;

        m_knotvector->findKnotVectorPosition(u, result);
        int span = result.m_index;
        //m_basisfunctions2D;
        static Array2D<UTIFLOAT> m_basisfunctions2D;
        if (m_basisfunctions2D.width() < m_degree+1 || m_basisfunctions2D.height() < m_degree+1)
        {
            m_basisfunctions2D.setSize(m_degree+1, m_degree+1);
        }

        m_knotvector->dersBasisFunctions(span, u, m_basisfunctions2D, m_degree, m_degree);

        for (int k = 0; k <= du; k++)
        {
            m_controldata.initZero(CK[k].m_pos);
            CK[k].m_weight = 0.0;
            for (int j = 0; j <= m_degree; j++)
            {
                UTIFLOAT f = m_basisfunctions2D(k,j);
                int index = span-m_degree+j;
                CK[k].m_pos += f*m_vertices[index].m_pos*m_vertices[index].m_weight;//Correct?
                CK[k].m_weight += f*m_vertices[index].m_weight;

            }
        }
        //dersBasisFunction(int i, int p, u, Array<UTIFLOAT>& ders);

    }
    void calcDervValue(Scalar u, Array<Vector>& CK)const
    {
        CHECKEXISTANCE;
        int d = CK.size()-1;
        static Array<SNurbsVertex> Aderv;
        if (Aderv.size() < CK.size())
        {
            Aderv.setSize(CK.size());
        }
        calcDervValueNoWeight(u, Aderv);
        Vector v;
        m_controldata.initZero(v);
        for (int k = 0; k <= d; k++)
        {
            v = Aderv[k].m_pos;
            for (int i = 1; i <= k; i++)
            {
                v = v - bin(k,i)*Aderv[i].m_weight*CK[k-i];
            }
            CK[k] = v / Aderv[0].m_weight;
        }
    }


    Nurbs1D* curveKnotInsert(UTIFLOAT newknotvalue, int amount)const
    {
        CHECKEXISTANCE;
        SAccessNurbs1D nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;

        curveKnotInsertTemplate(nurbsaccess, newknotvalue, amount);

        return nurbsaccess.m_newcurve;
    }
    SSplitResult splitCurve(UTIFLOAT splitposition)const
    {
        CHECKEXISTANCE;
        SKnotVectorResult result;
        m_knotvector->findKnotVectorPosition(splitposition, result);

        int s = 0;
        if (fabs(result.m_a-splitposition) < 0.0001)
        {
            splitposition = result.m_a;
            s =result.m_a_count;
        }
        utiassert(s <= getDegree());
        SAccessNurbs1DSplit nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;

        curveKnotInsertTemplate(nurbsaccess, splitposition, getDegree()-s);
        Nurbs1D::SSplitResult res;
        res.first = nurbsaccess.m_newcurve1;
        res.second = nurbsaccess.m_newcurve2;

        return res;
    }
    void decomposeCurve(CReferenceList<Nurbs1D>& result)const
    {
        CHECKEXISTANCE;

        if (m_knotvector->getBezier())
        {
            result.push_back(const_cast<Nurbs1D*>(this));
            return;
        }
        SAccessNurbs1DToBezier accessor;
        accessor.m_oldcurve = this;
        accessor.m_controldata = m_controldata;
        decomposeCurveTemplate(accessor);
        accessor.finalizeBezier();
        for (int i = 0; i < accessor.m_bezierlist.size(); i++)
        {
            result.push_back(&accessor.m_bezierlist[i]);
        }
    }
    void createSubdivisionStructureNode(SubdivisionNode* node, UTIFLOAT minvalue, UTIFLOAT maxvalue, int depth,int maxdepth, float tolerance, bool axisaligned )
    {
        UTIFLOAT splitposition = (minvalue+maxvalue) / 2.0f;
        Nurbs1D* n = check_cast<Nurbs1D*>(node->m_nurbs.get());
        /*CBoundsContainer<ControlData>* c = new CBoundsContainer<ControlData>();
        c->m_bounds = n->getBounds();
        node->m_box = c;*/

        if (axisaligned) {
        CBoundsContainer<ControlData>* c = new CBoundsContainer<ControlData>();
        c->m_bounds = n->getBounds();
        node->m_box = c;
        } else {
        COrientedBoundsContainer<ControlData>* c = new COrientedBoundsContainer<ControlData>();
        n->getOrientedBounds(&c->m_orientedbounds);
        node->m_box = c;
        }

        Scalar tol = n->getTolerance();
        node->m_tolerance = tol;
        if (depth >= maxdepth || tol < tolerance)
        {
            return;
        }

        SSplitResult res = n->splitCurve(splitposition);
        SubdivisionNode* newnode = new SubdivisionNode();
        newnode->m_nurbs = res.first;
        node->m_left =newnode;
        createSubdivisionStructureNode(newnode, minvalue, splitposition, depth+1, maxdepth, tolerance, axisaligned);

        newnode = new SubdivisionNode();
        newnode->m_nurbs = res.second;
        node->m_right =newnode;
        createSubdivisionStructureNode(newnode, splitposition, maxvalue, depth+1, maxdepth, tolerance, axisaligned);


    }
    void createSubdivisionStructure(int m_maxdepth, float tolerance, bool axisaligned = false)
    {
        m_subdivisioncontainer = new SubdivisionContainer();
        CReferenceList<Nurbs1D> result;
        decomposeCurve(result);
        for (int i = 0; i < result.size(); i++)
        {
            SubdivisionNode* node = new SubdivisionNode();
            node->m_nurbs = &result[i];
            Scalar minvalue = result[i].m_knotvector->getMinElement();
            Scalar maxvalue = result[i].m_knotvector->getMaxElement();
            createSubdivisionStructureNode(node, minvalue, maxvalue, 0, m_maxdepth, tolerance, axisaligned);
            m_subdivisioncontainer->m_root.push_back(node);
        }
    }
    Vector getMeanOfAllVectors()
    {
        Vector v;
        m_controldata.initZero(v);
        for (int i = 0; i < m_vertices.size(); i++) {

                v += m_vertices[i].m_pos;
        }
        v = v * (1.0/ m_vertices.size());
        return v;
    }
    UMat<Scalar> getCovarianceMatrix()
    {
        Vector mean = getMeanOfAllVectors();
        int d = m_controldata.getDimensions();
        Scalar N = m_vertices.size();
        UMat<Scalar> res(d, d);
        for (int k = 0; k < d; k++) {
            for (int l = 0; l < d; l++) {
                    Scalar currentvalue=0.0;
                    Scalar mean1 = m_controldata.getVertexIndex(mean, k);
                    Scalar mean2 = m_controldata.getVertexIndex(mean, l);
                    for (int i = 0; i < m_vertices.size(); i++) {

                            Vector & x = m_vertices[i].m_pos;
                            currentvalue += (m_controldata.getVertexIndex(x, k)-mean1)*(m_controldata.getVertexIndex(x, l)-mean2);

                    }
                    currentvalue /= N;
                    res.set(k,l, currentvalue);
            }
        }
        return res;

    }
    void getOrientedBounds(SOrientedBounds<Vector>* bounds)
    {
      /*  UMat<Scalar> covariance(getCovarianceMatrix());
        int m = covariance.m();
        uti_refptr<CEigenSymmWorkspace> eigensymm(new CEigenSymmWorkspace(m));
        eigensymm->calculateEvalAndEvec(covariance);
        eigensymm->sortEvalAndEvec(estAbsDesc);
        UMat<Scalar> newsystem(eigensymm->getEvec());

        uti_refptr<CLUDecomposition> decomposition;
        decomposition = new CLUDecomposition(newsystem);
        bounds->m_A = newsystem;
        bounds->m_workspace = decomposition;

        SBounds<Vector> b;
        m_controldata.init(b.m_min);
        m_controldata.init(b.m_max);
        int d = m_controldata.getDimensions();
        Vector temp = decomposition->solveSystem(m_vertices[0].m_pos);
        b.m_min = temp;
        b.m_max = temp;
        for (int i = 1; i < m_vertices.size(); i++)
        {
                Vector v = m_vertices[i].m_pos;
                v = decomposition->solveSystem(v);
                for (int k = 0; k < d; k++)
                {
                    m_controldata.setVertexIndex(b.m_min, k, umin(m_controldata.getVertexIndex(b.m_min, k), m_controldata.getVertexIndex(v, k)));
                    m_controldata.setVertexIndex(b.m_max, k, umax(m_controldata.getVertexIndex(b.m_max, k), m_controldata.getVertexIndex(v, k)));
                }

        }
        bounds->m_oriented = b;*/


    }
    Polynom<Scalar> getDistancePolynomRational(const Vector& v) const
    {
        //(p-C(u))*C'(u)
        utiassert(m_knotvector->getBezier());
        int n = getDegree();
        Polynom<Scalar> res;
        Polynom<Scalar> r, rdu;

        for (int i = 0; i <= n; i++)
        {
                Scalar controlpoint = m_vertices[i].m_weight;
                r = r+bernsteinPolynomCache<Scalar>(i, n)*controlpoint;
        };
        rdu = r.derivate();

        int d = m_controldata.getDimensions();
        for (int cd = 0; cd < d; cd++)
        {
            Scalar px = m_controldata.getVertexIndex(v,cd);
            Polynom<Scalar> cu;

            for (int i = 0; i <= n; i++)
            {
                Scalar controlpoint = m_controldata.getVertexIndex(m_vertices[i].m_pos,cd)*m_vertices[i].m_weight;
                cu = cu+bernsteinPolynomCache<Scalar>(i, n)*controlpoint;
            };

            Polynom<Scalar> cdu;

            cdu = cu.derivate();
            cdu = cdu*r-cu*rdu;
            cu = cu-(px*r);
            cu = cu*cdu;
            res = res + cu;
        }
        return res;
    }
    Polynom<Scalar> getDistancePolynomNonRational(const Vector& v) const
    {
        //(p-C(u))*C'(u)
        utiassert(m_knotvector->getBezier());
        int n = getDegree();
        Polynom<Scalar> res;
        int d = m_controldata.getDimensions();
        for (int cd = 0; cd < d; cd++)
        {
            Scalar px = m_controldata.getVertexIndex(v,cd);
            Polynom<Scalar> cu;

            for (int i = 0; i <= n; i++)
            {
                Scalar controlpoint = m_controldata.getVertexIndex(m_vertices[i].m_pos,cd);
                cu = cu+bernsteinPolynomCache<Scalar>(i, n)*controlpoint;
            };

            Polynom<Scalar> cdu;


            cdu = cu.derivate();

            // std::wcout <<"cu = " << cu.toString() << std::endl;
            //std::wcout <<"cdu = " << cdu.toString() << std::endl;

            cu = cu-px;
            // std::wcout <<"px-cu = " << cu.toString() << std::endl;
            cu = cu*cdu;
            res = res + cu;
        }
        return res;
    }
    struct SAnalyticSolution
    {
        Scalar m_basesolution;
        std::vector<Scalar> m_allsolutions;
        bool m_infinity;
        Scalar m_mindistance;
        SAnalyticSolution()
        {
            m_basesolution = 1E+10;
            m_infinity = false;
            m_mindistance = 1E+10;

        }
    };
    Scalar findNearestPointSturm(const Vector& v, SAnalyticSolution* sol=NULL)const
    {
        utiassert(m_knotvector->getBezier()==true);
      int n = m_vertices.size()-1;
      Polynom<Scalar> p;
      bool isrational = false;
        for (int i = 0; i <= n; i++) {
            if (fabs(m_vertices[i].m_weight - 1.0) > 0.0001) {
                isrational = true;
            }
        }
        if (isrational) {
            p = getDistancePolynomRational(v);
        } else {
            p = getDistancePolynomNonRational(v);
        }
        bool infinity = true;
        for (int i = 0; i < p.getSize(); i++) {
            if (absf(p(i)) > 0.0001) {
                infinity = false;
            }
        }
        //std::wcout << p.toString() << std::endl;
        std::vector<Scalar> resvalues;

        Vector a = m_vertices[0].m_pos;
        Vector b = m_vertices[n].m_pos;
        UTIFLOAT minpos = m_knotvector->getMinElement();
        UTIFLOAT maxpos = m_knotvector->getMaxElement();
        findCompleteZeroSetNoMultiRoots(p, 0.0, 1.01, resvalues);
        Scalar distance;
        Scalar currentmin = 1E+20;
        Scalar currentvalue = minpos;
        std::vector<Scalar> distances;
        for (int i = 0; i < resvalues.size(); i++)
        {
            Scalar u = minpos+resvalues[i]*(maxpos-minpos);
            resvalues[i] = u;
            //Scalar u = resvalues[i]*2.0;
            if (u >= minpos && u <= maxpos )
            {
                Vector vp = calcValue(u);
                distance = m_controldata.distanceTo(vp, v);
                distances.push_back(distance);
                if (distance < currentmin)
                {
                    currentvalue = u;
                    currentmin = distance;
                }


            } else {
               distances.push_back(1E+10);
            }
        }

        distance = m_controldata.distanceTo(a, v);
        if (distance < currentmin)
        {
            currentvalue = minpos;
            currentmin = distance;
        }
        Scalar SAME_SOLUTION = 0.0001;
        if (sol) {
                bool foundvalue = false;
                for (int i = 0; i < resvalues.size(); i++) {
                        Scalar u = resvalues[i];
                    if (u >= minpos && u <= maxpos ) {
                    if (fabs(resvalues[i]-minpos) < SAME_SOLUTION) {
                        foundvalue = true;
                    break;
                    }
                    }
                }
                if  (!foundvalue) {
                resvalues.push_back(minpos);
                distances.push_back(distance);
                }
        }
        distance = m_controldata.distanceTo(b, v);
        if (distance < currentmin)
        {
            currentvalue = maxpos;
            currentmin = distance;
        }
        if (sol) {
                 bool foundvalue = false;
                for (int i = 0; i < resvalues.size(); i++) {
                        Scalar u = resvalues[i];
                        if (u >= minpos && u <= maxpos ) {
                    if (fabs(resvalues[i]-maxpos) < SAME_SOLUTION) {
                        foundvalue = true;
                    break;
                    }
                        }
                }
                if  (!foundvalue) {
                resvalues.push_back(maxpos);
                distances.push_back(distance);
                }
        }
        if (sol) {
            bool addsolutions= true;
            if (fabs(sol->m_mindistance-currentmin) < SAME_SOLUTION) {
                sol->m_infinity = infinity;
                addsolutions = true;
                sol->m_basesolution = currentvalue;
            } else if (sol->m_mindistance > currentmin) {
               sol->m_allsolutions.clear();
               sol->m_infinity = sol->m_infinity || infinity;
               addsolutions = true;
            } else {
               addsolutions = false;
            }
            if (addsolutions) {
                int solutioncount = 0;
                assert(resvalues.size() == distances.size());
                for (int i = 0; i < resvalues.size(); i++)
                {
                   Scalar u = resvalues[i];
                   if (u >= minpos && u <= maxpos )
                   {
                       assert(i < distances.size());
                      if (fabs(distances[i]-currentmin) <SAME_SOLUTION) {
                        solutioncount++;
                        sol->m_allsolutions.push_back(u);
                      }
                   }
                }
            }
            assert(sol->m_allsolutions.size() > 0);
        }

        return currentvalue;

    }
    void copyToNurbs(int ownstart, int targetstart, int count, Nurbs1D* target)
    {
        CHECKEXISTANCE;
        for (int i = 0; i < count; i++)
        {
            target->m_vertices[targetstart+i] = m_vertices[ownstart+i];
        }
    }

    void print()
    {
        m_knotvector->print();

        int d = m_controldata.getDimensions();
        for (int i = 0; i < m_vertices.size(); i++)
        {
            std::wcout << "(";
            for (int j = 0; j< d; j++)
            {
                std::wcout << m_controldata.getVertexIndex(m_vertices[i].m_pos, j);
                if (j != d-1)
                {
                    std::wcout << ", ";
                }
            }
            std::wcout <<")" << std::endl;
        }
    }
    void move(const Vector& v)
    {
        CHECKEXISTANCE;
        for (int i = 0; i< m_vertices.size(); i++)
        {
            m_vertices[i].m_pos += v;
        }
    }
    Scalar newtonDelta(Scalar u, const Vector& p) const
    {
        //void calcDervValue(Scalar u, Array<Vector>& CK)const
        Array<Vector> CK(3);
        calcDervValue(u, CK);
        Scalar fx = m_controldata.dot(p-CK[0], CK[1]);
        Scalar fdx = -m_controldata.dot(CK[1], CK[1]) + m_controldata.dot(p-CK[0], CK[2]);
        Scalar dx = -fx / fdx;
        return dx;
    }
    bool improveFoundMinima(Scalar &u, const Vector& p, Scalar tolerance, Scalar aborttolerance) const
    {
        Scalar minvalue = m_knotvector->getMinElement();
        Scalar maxvalue = m_knotvector->getMaxElement();
        Scalar currentvalue = u;
        Scalar originalvalue = u;
        for (int i = 0; i < 10; i++)
        {
            if (currentvalue < minvalue || currentvalue > maxvalue)
            {
                return false;
            }

            Scalar delta = newtonDelta(currentvalue, p);
            currentvalue += delta;
            if (fabs(currentvalue-originalvalue) > tolerance)
            {
                return false;
            }
            if (fabs(delta) < aborttolerance)
            {
                u = currentvalue;
                return true;
            }
        }
        u = currentvalue;
        return true;

    }

    Scalar findNearestPointBruteForce(const Vector& v, int amount)const
    {
        CHECKEXISTANCE;
        UTIFLOAT minpos = m_knotvector->getMinElement();
        UTIFLOAT maxpos = m_knotvector->getMaxElement();
        UTIFLOAT currentdistance = 1E+10;
        UTIFLOAT resultpos = 0;
        Vector res;
        m_controldata.init(res);
        for (int i = 0; i < amount; i++)
        {
            UTIFLOAT s = (UTIFLOAT)i/(UTIFLOAT)(amount-1);
            s = minpos + s*(maxpos-minpos);
            res  = calcValue(s);
            UTIFLOAT newdistance = m_controldata.distanceTo(res, v);
            if (newdistance < currentdistance)
            {
                currentdistance = newdistance;
                resultpos = s;
            }
        }
        return resultpos;
    }

    SBounds<Vector> getBounds()const
    {
        CHECKEXISTANCE;
        SBounds<Vector> b;
        m_controldata.init(b.m_min);
        m_controldata.init(b.m_max);
        int d = m_controldata.getDimensions();
        b.m_min = m_vertices[0].m_pos;
        b.m_max = m_vertices[0].m_pos;
        for (int i = 1; i < m_vertices.size(); i++)
        {
            Vector v = m_vertices[i].m_pos;
            for (int j = 0; j < d; j++)
            {
                m_controldata.setVertexIndex(b.m_min, j, umin(m_controldata.getVertexIndex(b.m_min, j), m_controldata.getVertexIndex(v, j)));
                m_controldata.setVertexIndex(b.m_max, j, umax(m_controldata.getVertexIndex(b.m_max, j), m_controldata.getVertexIndex(v, j)));
            }
        }
        /* for (int i = 0; i < d; i++) {
             m_controldata.setVertexIndex(b.m_min, i, 1E+10);
             m_controldata.setVertexIndex(b.m_max, i, -1E+10);
         }*/

        return b;
    }
    struct SNearestPointResult
    {
        bool m_found;
        UTIFLOAT m_s;
        UTIFLOAT m_distance;
        int m_calls;
        int m_distancechecks;
        int m_splits;
        int m_finalstep;
        UTIFLOAT m_maxtolerance;
        Nurbs1D* m_curve;
        SNearestPointResult()
        {
            m_found =false;
            m_s = 0.0f;
            m_distance = 1E+10;
            m_calls = 0;
            m_distancechecks = 0;
            m_splits = 0;
            m_finalstep= 32;
            m_curve = NULL;
            m_maxtolerance = 0.05;

        }
    };
protected:
    Scalar getTolerance()
    {
        SAccessNurbs1D nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;

        return getToleranceTemplate(nurbsaccess);
        /*utiassert(m_knotvector->getBezier());
        int n = m_vertices.size()-1;
        Vector a = m_vertices[0].m_pos;
        Vector b = m_vertices[n].m_pos;
        Vector dir = b-a;
        Scalar distance;
        Scalar length = m_controldata.normalize(dir);
        if (length < 0.0000001)
        {
            return 0.0;
        }
        bool allonbaseline = true;

        Scalar maxdistancetobaseline=0;
        Vector r, lineposition;
        Scalar pos, maxweight, sk;
        maxweight = 0.0;

        for (int i = 1; i < m_vertices.size()-1; i++)
        {
            const Vector & v = m_vertices[i].m_pos;
            r = v-a;
            maxweight = umax(m_vertices[i].m_weight, maxweight);
            pos = m_controldata.dot(dir, r);
            if (pos < 0 || pos > length)
            {
                allonbaseline = false;
                break;
            }
            lineposition = a+dir*pos;

            distance = m_controldata.distanceTo(lineposition, v);
            maxdistancetobaseline = umax(distance, maxdistancetobaseline);
        }

        if (allonbaseline)
        {
            Scalar value1 = umax(1.0/m_vertices[0].m_weight, 1.0/m_vertices[m_vertices.size()-1].m_weight);
            value1 *= maxweight;
            value1 *= (pow(2, n)-1);
            value1 += 1;
            value1 = 1.0/value1;
            value1 = 1.0-value1;

            const Scalar LINETOLERANCE= 0.01;
            return maxdistancetobaseline*value1;

        }
        return 1E+10;*/

    }

    void findNearestPointBi(SNearestPointResult& result, SubdivisionNode* node, const Vector& q, UTIFLOAT minvalue, UTIFLOAT maxvalue, int depth, int maxdepth)const
    {
        CHECKEXISTANCE;
        //#define PrintStatus(text) std::wcout << "Level: " << depth << " " << text << " ("<<minvalue << " - " << maxvalue <<")" << std::endl;
#define PrintStatus(text)
        UTIFLOAT splitposition = (minvalue+maxvalue) / 2.0f;
        bool inbetween = (minvalue < splitposition) && (splitposition < maxvalue);
        result.m_calls += 1;
        bool abort = false;
        Vector dir, d1, a, b;
        int n = m_vertices.size()-1;
        a = m_vertices[0].m_pos;
        b = m_vertices[n].m_pos;
        dir = b-a;
        Scalar distance;
        Scalar length = m_controldata.normalize(dir);
        if (length < 0.0000001)
        {
            abort = true;
        }
        if (depth < maxdepth && !abort)
        {
            //checke, ob man es mit Linie ersetzen kann.
            bool allonbaseline = true;
            bool allonsameside = true;
            Scalar maxdistancetobaseline=0;
            Vector r, lineposition, influence_rj1, influence_multerm;
            Scalar pos, maxweight, sk;
            maxweight = 0.0;

            for (int i = 1; i < m_vertices.size()-1; i++)
            {
                const Vector & v = m_vertices[i].m_pos;
                r = v-a;
                maxweight = umax(m_vertices[i].m_weight, maxweight);
                pos = m_controldata.dot(dir, r);
                if (pos < 0 || pos > length)
                {
                    allonbaseline = false;
                    allonsameside = false;
                    break;
                }
                lineposition = a+dir*pos;
                if (i == 1)
                {
                    influence_rj1 = lineposition;
                    influence_multerm = v-influence_rj1;
                }
                else if (allonsameside)
                {
                    sk = m_controldata.dot(influence_multerm, v-influence_rj1);
                    if (sk < 0)
                    {
                        allonsameside = false;

                    }

                }
                distance = m_controldata.distanceTo(lineposition, v);
                maxdistancetobaseline = umax(distance, maxdistancetobaseline);
            }

            if (allonbaseline)
            {
                Scalar value1 = umax(1.0/m_vertices[0].m_weight, 1.0/m_vertices[m_vertices.size()-1].m_weight);
                value1 *= maxweight;
                value1 *= (pow(Scalar(2), Scalar(n))-1);
                value1 += 1;
                value1 = 1.0/value1;
                value1 = 1.0-value1;

               // const Scalar LINETOLERANCE= 0.05;
                if (maxdistancetobaseline*value1 < result.m_maxtolerance)
                {
                    PrintStatus(UTISTR("abort: line criteria"));
                    abort = true;

                }

            }
            if (false && !abort && allonsameside) {
                if (sk < 0) {

                r = q-a;
                pos = m_controldata.dot(dir, r);
                if (pos < 0 || pos > length)
                {
                    //influence are criteria
                    distance = m_controldata.distanceTo(a, q);
                    if (distance < result.m_distance)
                    {
                        result.m_distance = distance;
                        result.m_found = true;
                        result.m_s = minvalue;
                        result.m_distancechecks += 1;

                    }
                    distance = m_controldata.distanceTo(b, q);
                    if (distance < result.m_distance)
                    {
                        result.m_distance = distance;
                        result.m_found = true;
                        result.m_s = maxvalue;
                        result.m_distancechecks += 1;

                    }
                    PrintStatus(UTISTR("abort: influence criteria"));
                    return;
                }
                }

            }


        }

        if (depth >= maxdepth || (!inbetween) || abort)
        {
            Scalar f = getMinDistanceToLine(m_controldata, a, b, q);

            if (depth >= maxdepth)
            {
                PrintStatus(UTISTR("abort: depth criteria"));
            }

            f = minvalue+f*(maxvalue-minvalue);
            //TODO: Dieses Ergebnis mit Newton verbessern

            Vector v = result.m_curve->calcValue(f);
            distance = m_controldata.distanceTo(v, q);
            if (distance < result.m_distance)
            {
                result.m_distance = distance;
                result.m_found = true;
                result.m_s = f;
                result.m_distancechecks += 1;

            }

        }
        else
        {
            PrintStatus(UTISTR("subdivide"));

            Scalar firstmindistance;
            Scalar secondmindistance;
            SSplitResult res;
            SubdivisionNode* nodeleft = NULL;
            SubdivisionNode* noderight = NULL;

            if (node && node->m_left.get())
            {
                nodeleft = node->m_left.get();
                noderight = node->m_right.get();
                res.first = check_cast<Nurbs1D<ControlData>* >(node->m_left->m_nurbs.get());
                res.second = check_cast<Nurbs1D<ControlData>* >(node->m_right->m_nurbs.get());
                //firstmindistance = getMinDistanceToBounds(m_controldata, check_cast<CBoundsContainer<ControlData>* >(node->m_left->m_box.get())->m_bounds, q);



                //secondmindistance = getMinDistanceToBounds(m_controldata, check_cast<CBoundsContainer<ControlData>* >(node->m_right->m_box.get())->m_bounds, q);

                firstmindistance = check_cast<CBaseBoundsContainer<ControlData>*>(node->m_left->m_box.get())->getMinimumDistance(m_controldata,  q);
                secondmindistance = check_cast<CBaseBoundsContainer<ControlData>*>(node->m_right->m_box.get())->getMinimumDistance(m_controldata,  q);
                /* SBounds<Vector> firstbounds = res.first->getBounds();
                SBounds<Vector> secondbounds = res.second->getBounds();
                firstmindistance = getMinDistanceToBounds(m_controldata, firstbounds, q);
                secondmindistance = getMinDistanceToBounds(m_controldata, secondbounds, q);*/
            }
            else
            {
                res = splitCurve(splitposition);
                result.m_splits += 1;
                SBounds<Vector> firstbounds = res.first->getBounds();
                SBounds<Vector> secondbounds = res.second->getBounds();
                firstmindistance = getMinDistanceToBounds(m_controldata, firstbounds, q);
                secondmindistance = getMinDistanceToBounds(m_controldata, secondbounds, q);
            };
            if (firstmindistance < secondmindistance)
            {
                if (firstmindistance > result.m_distance)
                {
                    return;
                };
                res.first->findNearestPointBi(result,nodeleft, q, minvalue, splitposition, depth+1, maxdepth);
                if (secondmindistance > result.m_distance)
                {
                    return;
                };
                res.second->findNearestPointBi(result, noderight, q,  splitposition, maxvalue, depth+1, maxdepth);

            }
            else
            {
                if (secondmindistance > result.m_distance)
                {
                    return;
                };
                res.second->findNearestPointBi(result, noderight, q,splitposition, maxvalue,  depth+1, maxdepth);
                if (firstmindistance > result.m_distance)
                {
                    return;
                };
                res.first->findNearestPointBi(result, nodeleft, q, minvalue, splitposition, depth+1, maxdepth);

            }

        }
#undef PrintStatus
    }
public:
    SNearestPointResult findNearestPointBisection(const Vector& q, int maxdepth, UTIFLOAT maxdistance) const
    {
        CHECKEXISTANCE;
        SNearestPointResult nearestpoint;
        nearestpoint.m_s = 0.0f;
        nearestpoint.m_distance = maxdistance;
        nearestpoint.m_found = false;

        findNearestPointBi(nearestpoint, NULL, q, m_knotvector->getMinElement(), m_knotvector->getMaxElement(), 0, maxdepth);
        /*std::cout <<"callcount: " << nearestpoint.m_calls << std::endl;
        std::cout <<"splitcount: " << nearestpoint.m_splits << std::endl;
        std::cout <<"distancecount: " << nearestpoint.m_distancechecks << std::endl;*/
        return nearestpoint;




    }
    SNearestPointResult findNearestPointBisectionBezier(const Vector& q, int maxdepth, UTIFLOAT maxdistance, bool polish, UTIFLOAT maxtolerance)const
    {
        CHECKEXISTANCE;

        SNearestPointResult nearestpoint;
        nearestpoint.m_s = 0.0f;
        nearestpoint.m_distance = maxdistance;
        nearestpoint.m_maxtolerance = maxtolerance;
        nearestpoint.m_found = false;
        CReferenceList<Nurbs1D> resultlist;
        std::vector< std::pair<int, Scalar> > distances;
        if (m_subdivisioncontainer.get())
        {
            for (int i = 0; i < m_subdivisioncontainer->m_root.size(); i++)
            {
                SubdivisionNode* n = &m_subdivisioncontainer->m_root[i];
                resultlist.push_back(check_cast<Nurbs1D*>(n->m_nurbs.get()));
                Scalar firstmindistance = check_cast<CBaseBoundsContainer<ControlData>*>(n->m_box.get())->getMinimumDistance(m_controldata,  q);
                distances.push_back(std::make_pair(i, firstmindistance));
            }
        }
        else
        {
            decomposeCurve(resultlist);
            for (int i = 0; i < resultlist.size(); i++)
            {
                SBounds<Vector> firstbounds = resultlist[i].getBounds();
                Scalar firstmindistance = getMinDistanceToBounds(m_controldata, firstbounds, q);
                distances.push_back(std::make_pair(i, firstmindistance));
            }
        }


        std::sort(distances.begin(), distances.end(), sortNURBS<Scalar>);

        for (int i = 0; i < distances.size(); i++)
        {
            std::pair<int, Scalar>& element = distances[i];
            if (element.second > nearestpoint.m_distance)
            {
                break;
            }
            Nurbs1D& n = resultlist[element.first];
            SubdivisionNode* sn = NULL;
            if (m_subdivisioncontainer.get())
            {
                sn = &m_subdivisioncontainer->m_root[element.first];
            }
            nearestpoint.m_curve = &n;
            if (element.second < nearestpoint.m_distance) {
                n.findNearestPointBi(nearestpoint, sn, q, n.getKnotVector()->getMinElement(), n.getKnotVector()->getMaxElement(), 0, maxdepth);
            }
        }

        if (polish) {
        improveFoundMinima(nearestpoint.m_s, q, 0.05, 0.000001);

        UTIFLOAT minpos = m_knotvector->getMinElement();
        UTIFLOAT maxpos = m_knotvector->getMaxElement();
        nearestpoint.m_s = umax(umin(maxpos, nearestpoint.m_s), minpos);

        Vector v = calcValue(nearestpoint.m_s);
        nearestpoint.m_distance = m_controldata.distanceTo(v, q);
        }

        nearestpoint.m_curve = const_cast<Nurbs1D*>(this);

        //void decomposeCurve(CReferenceList<Nurbs1D>& result)
        return nearestpoint;
    }

    static Nurbs1D* makeNurbsCircle(const ControlData &controldata,  const Vector& O, const Vector& X, const Vector& Y, Scalar r, Scalar ths, Scalar the)
    {
        if (the < ths)
        {
            the = 360.0 + the;
        }
        Scalar theta = the - ths;
        int narcs = 1;
        if (theta <= 90.0)
        {
            narcs = 1;
        }
        else if (theta <= 180.0)
        {
            narcs = 2;
        }
        else if (theta <= 270.0)
        {
            narcs = 3;

        }
        else
        {
            narcs = 4;
        }
        Scalar dtheta = theta / (Scalar)narcs;
        Scalar dummy;
        //int n = 2* narcs;
        Scalar w1 = cos(dtheta*pi180/2.0f);
        Vector P0 = O + r*cos(ths*pi180)*X + r*sin(ths*pi180)*Y;
        Vector P1;
        controldata.init(P1);
        Vector T0 = -sin(ths*pi180)*X + cos(ths)*Y;
        Array<Vector> Pw;
        Array<Scalar> weights;
        Pw.setSize((narcs-1)*2+3);
        weights.setSize((narcs-1)*2+3);
        Pw[0] = P0;
        weights[0] = 1.0;

        int index = 0;
        float angle = ths;


        for (int i = 1; i <= narcs; i++)
        {
            angle = angle + dtheta;
            Vector P2 = O + r*cos(angle*pi180)*X + r*sin(angle*pi180)*Y;
            Pw[index+2] = P2;
            weights[index+2] = 1.0;

            Vector T2 = -sin(angle*pi180)*X + cos(angle*pi180)*Y;
            IntersectLines(controldata, P0, T0, P2, T2, dummy, dummy, P1);
            Pw[index+1] = P1;//Eventuelle Scaliere dies nicht, checks NURBS Formel
            weights[index+1] = w1;
            index = index + 2;
            if (i < narcs)
            {
                P0 = P2;
                T0 = T2;
            }
        }
        int j = 2* narcs+1;
        //KnotVector
        Array<UTIFLOAT> U(3+j);
        for (int i = 0; i < 3; i++)
        {
            U[i] = 0.0;
            U[i+j] = 1.0;
        }
        switch (narcs)
        {
        case 1:
            break;
        case 2:
        {
            U[3] = U[4] = 0.5;
        };
        break;
        case 3:
        {
            U[3] = U[4] = 1.0/3.0;
            U[5] = U[6] = 2.0/3.0;
        };
        break;
        case 4:
        {
            U[3] = U[4] = 0.25;
            U[5] = U[6] = 0.5;
            U[7] = U[8] = 0.75;
        };
        break;

        }
        KnotVector* v = new KnotVector(U);
        v->print();
        v->compile();
        Nurbs1D* n2 = new Nurbs1D(v, 2, controldata);
        utiassert(n2->getVertexCount() == Pw.size());
        for (int i = 0; i < Pw.size(); i++)
        {
            n2->setVertex(i, Pw[i], weights[i]);
        }
        return n2;
    }
    static Nurbs1D* leastSquaresApproximation(const ControlData &controldata,
            const Array<Vector> &Q,
            const Array<Scalar> &Wq,
            const Array<Vector> &D,
            const Array<int> &I,
            const Array<Scalar> &Wd,
            int n, int p)
    {
        int r = Q.size()-1;
        utiassert(Wq.size() == r+1);
        utiassert(D.size() == I.size());
        utiassert(D.size() == Wd.size());
        int s = D.size()-1;

        Array<Scalar> ub(r+1);
        //Array<Scalar> N[]

        int ru= -1;
        int rc = -1;
        for (int i = 0; i <= r; i++)
        {
            if (Wq[i] > 0.0)
            {
                ru = ru+1;

            }
            else
            {
                rc = rc+1;
            }
        }
        int su = -1;
        int sc = -1;
        for (int j = 0; j <= s; j++)
        {
            if (Wd[j] > 0.0)
            {
                su = su+1;
            }
            else
            {
                sc = sc+1;
            }
        }
        int mu = ru+su+1;
        int mc = rc+sc+1;
        if (mc >= n || mc+n >= mu+1)
        {
            utiassert(false);
        }

    }


};

template <class ControlData> typename ControlData::Scalar getMinDistanceToTriangleNew(ControlData& m_controldata, const
                                                          typename ControlData::Vector& B, const typename ControlData::Vector & p2, const typename ControlData::Vector & p3, const typename ControlData::Vector& P,
                                                          typename ControlData::Scalar& u,typename ControlData::Scalar& v)
{
    typename ControlData::Vector e0, e1, dvec;
    e0 = p2-B;
    e1 = p3-B;

    dvec = B-P;

    typename ControlData::Scalar a = m_controldata.dot(e0, e0);
    typename ControlData::Scalar b = m_controldata.dot(e0, e1);
    typename ControlData::Scalar c = m_controldata.dot(e1, e1);


 /*   typename ControlData::Scalar a = n1n1;
    typename ControlData::Scalar b_and_c = n1n2;
    typename ControlData::Scalar d = n2n2;*/


    typename ControlData::Scalar d = m_controldata.dot(e0, dvec);
    typename ControlData::Scalar e = m_controldata.dot(e1, dvec);
    typename ControlData::Scalar f = m_controldata.dot(dvec, dvec);

    typename ControlData::Scalar det = a*c-b*b;
    typename ControlData::Scalar invDet;

    if (fabs(det) < 0.00001) {
        u = 0.0;
        v = 0.0;
        return m_controldata.distanceTo(B, P);
    }
    typename ControlData::Scalar s = b*e - c*d;
    typename ControlData::Scalar t = b*d - a*e;
    int region =0;
    if (s+t <= det) {
        if (s < 0) {
            if (t < 0) {
                region = 4;
            } else {
                region = 3;

            }
        }
    } else {
        if (s < 0) {
            region = 2;
        } else if (t < 0) {
            region = 6;
        } else {
            region = 1;
        }
    }
    switch (region) {
        case 0: {
            invDet = 1.0 / det;
            s *= invDet;
            t *= invDet;

        }; break;
        case 1: {
            typename ControlData::Scalar numer = c+d-b-d;
            if (numer <= 0) {
                s = 0;
            } else {
                typename ControlData::Scalar denom = a - 2*b +c;
                s = (numer >= denom ? 1 : numer/denom);
            }
            t = 1-s;
        }; break;
        case 3: {
            s = 0;
            t = (e >= 0 ? 0 : (-e >= c ? 1: -e/c));

        }; break;
        case 5: {
            t = 0;
            s = (d >= 0 ? 0 : (-d >= a ? 1: -d/a));

        }; break;
        case 2: {
            typename ControlData::Scalar tmp0, tmp1, numer, denom;
            tmp0 = b + d;
            tmp1 = c + e;
            if (tmp1 > tmp0) {
                numer = tmp1 - tmp0;
                denom = a - 2*b+c;
                s = (numer>= denom? 1: numer /denom);
                t = 1-s;
            } else {
                s = 0;
                t = (tmp1 <= 0 ? 1 : (e >= 0 ? 0 : -e / c));
            }
        }; break;

        case 6: {
            typename ControlData::Scalar tmp0, tmp1, numer, denom;
            tmp0 = b + e;
            tmp1 = a + d;
            if (tmp1 > tmp0) {
                numer = tmp1 - tmp0;
                denom = c - 2*b+a;
                t = (numer>= denom? 1: numer /denom);
                s = 1-t;
            } else {
                s = 0;
                t = (tmp1 <= 0 ? 1 : (d >= 0 ? 0 : -d / a));
            }
        }; break;
        case 4: {
         s= 0;
         t = 0;
        };

    }
    u = s;
    v = t;
    dvec = B+u*e0+v*e1;
    typename ControlData::Scalar pos = m_controldata.distanceTo(dvec, P);

 //  #define CHECKTRIANGLETEST
    #ifdef CHECKTRIANGLETEST
    {
        typename ControlData::Vector currentpos;
        currentpos = p1+u*n1+v*n2;
        currentpos = q-currentpos;

        typename ControlData::Scalar checkzero = m_controldata.dot(n1, currentpos);
        typename ControlData::Scalar checkzero2 = m_controldata.dot(n2, currentpos);

         utiassert(fabs(checkzero) < 0.01);
          utiassert(fabs(checkzero2) < 0.01);

    }
    #endif



    #ifdef CHECKTRIANGLETEST
    typename ControlData::Vector d2;
    const int c = 10;
    typename ControlData::Scalar resu2 = 0;
    typename ControlData::Scalar resv2 = 0;
    typename ControlData::Scalar currentdistance, distance;
    currentdistance = m_controldata.distanceTo(p1, q);
    for (int i = 0; i <= c; i++) {
        for (int j = 0; j <= c; j++) {
            typename ControlData::Scalar u2 = (typename ControlData::Scalar)i / (typename ControlData::Scalar)c;
            typename ControlData::Scalar v2 = (typename ControlData::Scalar)j / (typename ControlData::Scalar)c;
            if (u2+v2 >= 1.1) continue;

            d2 = p1+u2*n1+v2*n2;
            distance = m_controldata.distanceTo(d2, q);
            if (currentdistance > distance) {
                currentdistance = distance;
                resu2 = u2;
                resv2 = v2;
            }

        }
    }
    utiassert(fabs(u-resu2) < 0.2);
    utiassert(fabs(v-resv2) < 0.2);
    u = resu2;
    v = resv2;
    d1 = p1+u*n1+v*n2;
    pos = m_controldata.distanceTo(d1, q);
    #endif


    return pos;
}


template <class ControlData> typename ControlData::Scalar getMinDistanceToTriangle(ControlData& m_controldata, const
                                                          typename ControlData::Vector& p1, const typename ControlData::Vector & p2, const typename ControlData::Vector & p3, const typename ControlData::Vector& q,
                                                          typename ControlData::Scalar& u,typename ControlData::Scalar& v)
{
    typename ControlData::Vector n1, n2, d1;
    n1 = p2-p1;
    n2 = p3-p1;

    d1 = q-p1;

    typename ControlData::Scalar n1n1 = m_controldata.dot(n1, n1);
    typename ControlData::Scalar n2n2 = m_controldata.dot(n2, n2);
    typename ControlData::Scalar n1n2 = m_controldata.dot(n1, n2);

    typename ControlData::Scalar a = n1n1;
    typename ControlData::Scalar b_and_c = n1n2;
    typename ControlData::Scalar d = n2n2;


    typename ControlData::Scalar tb1 = m_controldata.dot(n1, d1);
    typename ControlData::Scalar tb2 = m_controldata.dot(n2, d1);

    typename ControlData::Scalar det = a*d-b_and_c*b_and_c;

    if (fabs(det) < 0.00001) {
        u = 0.0;
        v = 0.0;
        return m_controldata.distanceTo(p1, q);
    }
    a /= det;
    b_and_c /= det;
    d /= det;

    u = (d*tb1 - b_and_c*tb2);
    v = (-b_and_c*tb1 + a* tb2);
 //  #define CHECKTRIANGLETEST
    #ifdef CHECKTRIANGLETEST
    {
        typename ControlData::Vector currentpos;
        currentpos = p1+u*n1+v*n2;
        currentpos = q-currentpos;

        typename ControlData::Scalar checkzero = m_controldata.dot(n1, currentpos);
        typename ControlData::Scalar checkzero2 = m_controldata.dot(n2, currentpos);

         utiassert(fabs(checkzero) < 0.01);
          utiassert(fabs(checkzero2) < 0.01);

    }
    #endif

    typename ControlData::Scalar u1, v1;
u1 = u;
    v1 = v;
    if (v > 1.0-u) {
    if (v > u-1.0f && v < u+1.0) {
        u1 = 0.5*(-v+u+1);
        v1 = 0.5*(v-u+1);
        u = u1;
        v = v1;

       /* u1 = n1n1-n1n2;
       // if (fabs(u1)> 0.0001) {

        u = (tb1-n1n2)/(u1);
        v = 1.0-u;*/

       // } else {

       // }
    } else if (u > 1.0 && u > v) {
        u = 1.0;
        v = 0.0;
    } else if (v > 1.0) {
        u = 0.0;
        v = 1.0;
    }
    };


    u = std::max(0.0,std::min(1.0, u));
    v = std::max(0.0,std::min(1.0, v));

    d1 = p1+u*n1+v*n2;
    typename ControlData::Scalar pos = m_controldata.distanceTo(d1, q);

    #ifdef CHECKTRIANGLETEST
    typename ControlData::Vector d2;
    const int c = 10;
    typename ControlData::Scalar resu2 = 0;
    typename ControlData::Scalar resv2 = 0;
    typename ControlData::Scalar currentdistance, distance;
    currentdistance = m_controldata.distanceTo(p1, q);
    for (int i = 0; i <= c; i++) {
        for (int j = 0; j <= c; j++) {
            typename ControlData::Scalar u2 = (typename ControlData::Scalar)i / (typename ControlData::Scalar)c;
            typename ControlData::Scalar v2 = (typename ControlData::Scalar)j / (typename ControlData::Scalar)c;
            if (u2+v2 >= 1.1) continue;

            d2 = p1+u2*n1+v2*n2;
            distance = m_controldata.distanceTo(d2, q);
            if (currentdistance > distance) {
                currentdistance = distance;
                resu2 = u2;
                resv2 = v2;
            }

        }
    }
    utiassert(fabs(u-resu2) < 0.2);
    utiassert(fabs(v-resv2) < 0.2);
    u = resu2;
    v = resv2;
    d1 = p1+u*n1+v*n2;
    pos = m_controldata.distanceTo(d1, q);
    #endif


    return pos;
}



template <class ControlData> class Nurbs2D : public ReferenceADR
{
public:
    typedef typename ControlData::Vector Vector;
    typedef typename ControlData::Scalar Scalar;
    typedef ControlData CurveControlData;
    ControlData m_controldata;
    struct SNurbsVertex
    {
        Vector m_pos;
        Scalar m_weight;
        SNurbsVertex operator*(Scalar v) const
        {
            SNurbsVertex r;
            r.m_pos = m_pos*v;
            r.m_weight = m_weight*v;
            return r;

        }

        SNurbsVertex operator+(const SNurbsVertex& v) const
        {
            SNurbsVertex r;
            r.m_pos = m_pos+v.m_pos;
            r.m_weight = m_weight+v.m_weight;
            return r;

        }
    };

    template <bool uvdir> struct SAccessNurbs2DTolerance
    {
    public:
        Nurbs2D* m_oldcurve;
        ControlData m_controldata;
        int m_baseindex;

        typedef typename ControlData::Vector Vector;
        typedef typename ControlData::Scalar Scalar;
        typedef SNurbsVertex SOneVertex;

	SAccessNurbs2DTolerance()
	{
	  m_baseindex = 0;
	}
        inline void getVertex(int index, SOneVertex& v)const
        {
            if (uvdir) {
            v = m_oldcurve->getVertex(index, m_baseindex);
            } else {
            v = m_oldcurve->getVertex(m_baseindex, index);
            }
        }

        inline int getVertexCount()const
        {
            if (uvdir) {
            return m_oldcurve->getSizeX();

            } else {
            return m_oldcurve->getSizeY();
            }
        }

        inline int getDegree()const
        {
             if (uvdir) {
            return m_oldcurve->getDegreeU();
            } else {
              return m_oldcurve->getDegreeV();
            }
        }

        inline KnotVector* getKnotVector()const
        {
             if (uvdir) {
            return m_oldcurve->getKnotVectorU();
             } else {
              return m_oldcurve->getKnotVectorV();
             }
        }


    };

    template <bool uvdir> struct SAccessNurbs2D
    {
    public:
        Nurbs2D* m_oldcurve;
        Nurbs2D* m_newcurve;
        ControlData m_controldata;

        typedef typename ControlData::Vector Vector;
        typedef typename ControlData::Scalar Scalar;
        struct SOneVertex
        {
            Array<SNurbsVertex> m_row_or_column;
        };
        inline void add(SOneVertex& res, const SOneVertex &a, const SOneVertex &b)
        {
            utiassert(a.m_row_or_column.size() == b.m_row_or_column.size());
            if (uvdir)
            {
                utiassert(a.m_row_or_column.size() == this->m_oldcurve->getSizeY());
            }
            else
            {
                utiassert(a.m_row_or_column.size() == this->m_oldcurve->getSizeX());
            }
            if (res.m_row_or_column.size() != a.m_row_or_column.size())
            {
                res.m_row_or_column.setSize(a.m_row_or_column.size());
            }
            for (int i = 0; i < res.m_row_or_column.size(); i++)
            {
                res.m_row_or_column[i] = a.m_row_or_column[i] + b.m_row_or_column[i];
            }
        }

        inline void scal(SOneVertex& res, const SOneVertex &a, Scalar weight)
        {

            if (uvdir)
            {
                utiassert(a.m_row_or_column.size() == this->m_oldcurve->getSizeY());
            }
            else
            {
                utiassert(a.m_row_or_column.size() == this->m_oldcurve->getSizeX());
            }
            if (res.m_row_or_column.size() != a.m_row_or_column.size())
            {
                res.m_row_or_column.setSize(a.m_row_or_column.size());
            }
            for (int i = 0; i < res.m_row_or_column.size(); i++)
            {
                res.m_row_or_column[i] = a.m_row_or_column[i]* weight;
            }
            // res = a* weight;
        }
        inline void getVertexWeight(int index, SOneVertex& v)
        {
            getVertex(index, v);
            for (int i = 0; i < v.m_row_or_column.size(); i++) {
                //temp.m_row_or_column[i] = v.m_row_or_column[i];
                v.m_row_or_column[i].m_pos *= v.m_row_or_column[i].m_weight;
            }
        }
        inline void getVertex(int index, SOneVertex& v)
        {
            if (uvdir)
            {
                v.m_row_or_column.setSize(m_oldcurve->getSizeY());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    v.m_row_or_column[i] = m_oldcurve->getVertex(index, i);
                }
            }
            else
            {
                v.m_row_or_column.setSize(m_oldcurve->getSizeX());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    v.m_row_or_column[i] = m_oldcurve->getVertex(i, index);
                }
            }
            //v = m_oldcurve->getVertex(index);
        }
        inline int getVertexCount()
        {
            if (uvdir)
            {
                return m_oldcurve->getSizeX();
            }
            else
            {
                return m_oldcurve->getSizeY();
            }
        }

        inline int getDegree()
        {
            if (uvdir)
            {
                return m_oldcurve->getDegreeU();
            }
            else
            {
                return m_oldcurve->getDegreeV();
            }
        }

        inline KnotVector* getKnotVector()
        {
            if (uvdir)
            {
                return m_oldcurve->getKnotVectorU();
            }
            else
            {
                return m_oldcurve->getKnotVectorV();
            }
        }

        inline void setVertexProject(int index, const SOneVertex& v)
        {
            SOneVertex temp;
            temp.m_row_or_column.setSize(v.m_row_or_column.size());
            for (int i = 0; i < v.m_row_or_column.size(); i++) {
                temp.m_row_or_column[i] = v.m_row_or_column[i];
                temp.m_row_or_column[i].m_pos /= temp.m_row_or_column[i].m_weight;
            }
            setVertex(index, v);
        }
        inline void setVertex(int index, const SOneVertex& v)
        {

            if (uvdir)
            {
                v.m_row_or_column.setSize(m_oldcurve->getSizeY());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    m_newcurve->setVertex(index, i, v.m_row_or_column[i]);
                }
            }
            else
            {
                v.m_row_or_column.setSize(m_oldcurve->getSizeX());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    m_newcurve->setVertex(i, index, v.m_row_or_column[i]);
                }
            }

        }

        /* inline void setVertex(int index, const SOneVertex& v)
         {
             m_newcurve->setVertex(index, v);
         }*/

        inline void createCurve(KnotVector* newknot, int insertpoint)
        {
            if (uvdir)
            {
                m_newcurve = new Nurbs2D(newknot, m_oldcurve->getKnotVectorV(), m_oldcurve->getDegreeU(), m_oldcurve->getDegreeV(), m_controldata);
            }
            else
            {
                // m_newcurve1 = new Nurbs2D(newknot1, degree, m_controldata);
                //  m_newcurve2 = new Nurbs2D(newknot2, degree, m_controldata);

                m_newcurve = new Nurbs2D(m_oldcurve->getKnotVectorU(), newknot,  m_oldcurve->getDegreeU(),m_oldcurve->getDegreeV(), m_controldata);

            }
            // m_newcurve = new Nurbs1D(newknot, getDegree(), m_controldata);
        }

    };
    template <bool uvdir> struct SAccessNurbs2DToBezier : public SAccessNurbs2D<uvdir>
    {
        CReferenceList<Nurbs2D> m_bezierlist;
        //uti_refptr<KnotVector> m_sharedknotvector;
        void addBezierSegment()
        {

            /*if (m_sharedknotvector.get() == NULL)
            {
                m_sharedknotvector = new KnotVector((p+1)*2);
                for (int i = 0; i <= p; i++)
                {
                    m_sharedknotvector->setEntry(i, 0.0);
                    m_sharedknotvector->setEntry(i+(p+1), 1.0);
                }
                m_sharedknotvector->setBezier(true);
                m_sharedknotvector->compile();
            }
            */
            int p = this->getDegree();
            KnotVector* oldkv = this->getKnotVector();

            //if (m_sharedknotvector.get() == NULL)
            //{
            double start = oldkv->getSingle(m_bezierlist.size());
            double ende = oldkv->getSingle(m_bezierlist.size()+1);
            KnotVector* newvector= new KnotVector((p+1)*2);
            for (int i = 0; i <= p; i++)
            {
                newvector->setEntry(i, start);
                newvector->setEntry(i+(p+1), ende);
            }
            newvector->setBezier(true);
            newvector->compile();

            Nurbs2D* n;
            if (uvdir)
            {
                n = new Nurbs2D(newvector, this->m_oldcurve->getKnotVectorV(), this->m_oldcurve->getDegreeU(), this->m_oldcurve->getDegreeV(), this->m_oldcurve->getControlData());
            }
            else
            {
                n = new Nurbs2D(this->m_oldcurve->getKnotVectorU(), newvector, this->m_oldcurve->getDegreeU(), this->m_oldcurve->getDegreeV(), this->m_oldcurve->getControlData());
            }
            m_bezierlist.push_back(n);

        }
        void finalizeBezier()
        {
            for (int index = 0; index < m_bezierlist.size(); index++) {
                Nurbs2D* n = &m_bezierlist[index];
                for (int i = 0; i < n->getSizeX(); i++) {
                    for (int j = 0; j < n->getSizeY(); j++) {
                        n->m_vertices(i,j).m_pos /= n->m_vertices(i,j).m_weight;
                    }
                }
            }
        }

        /*void setBezierVertexProject(int segment, int index, const typename SAccessNurbs2D<uvdir>::SOneVertex& v)
        {
            typename SAccessNurbs2D<uvdir>::SOneVertex temp;
            temp.m_row_or_column.setSize(v.m_row_or_column.size());
            for (int i = 0; i < v.m_row_or_column.size(); i++) {
                temp.m_row_or_column[i] = v.m_row_or_column[i];
                temp.m_row_or_column[i].m_pos /= temp.m_row_or_column[i].m_weight;
            }
            setBezierVertex(segment, index, v);
        }*/

        void setBezierVertex(int segment, int index, const typename SAccessNurbs2D<uvdir>::SOneVertex& v)
        {
            while (segment >= m_bezierlist.size())
            {
                addBezierSegment();
            };
            // m_bezierlist[segment].setVertex(index, v);
            Nurbs2D& n = m_bezierlist[segment];
            if (uvdir)
            {
                //v.m_row_or_column.setSize(this->m_oldcurve->getSizeY());
                utiassert(v.m_row_or_column.size() == this->m_oldcurve->getSizeY());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    n.setVertex(index, i, v.m_row_or_column[i]);
                }
            }
            else
            {
                //v.m_row_or_column.setSize(this->m_oldcurve->getSizeX());
                utiassert(v.m_row_or_column.size() == this->m_oldcurve->getSizeX());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    n.setVertex(i, index, v.m_row_or_column[i]);
                }
            }


        }

        void getBezierVertex(int segment, int index, typename SAccessNurbs2D<uvdir>::SOneVertex& v)
        {
            //v = m_bezierlist[segment].getVertex(index);
            Nurbs2D& n = m_bezierlist[segment];
            if (uvdir)
            {
                v.m_row_or_column.setSize(this->m_oldcurve->getSizeY());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    v.m_row_or_column[i] = n.getVertex(index, i);
                }
            }
            else
            {
                v.m_row_or_column.setSize(this->m_oldcurve->getSizeX());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    v.m_row_or_column[i] = n.getVertex(i, index);
                }
            }

        }
    };

    template <bool uvdir> struct SAccessNurbs2DSplit
    {
        ControlData m_controldata;
        Nurbs2D* m_oldcurve;
        Nurbs2D* m_newcurve1;
        Nurbs2D* m_newcurve2;
        int m_newcurve1end;
        int m_newcurve2start;

        typedef typename ControlData::Vector Vector;
        typedef typename ControlData::Scalar Scalar;
        struct SOneVertex
        {
            Array<SNurbsVertex> m_row_or_column;
        };
        inline void add(SOneVertex& res, const SOneVertex &a, const SOneVertex &b)
        {
            utiassert(a.m_row_or_column.size() == b.m_row_or_column.size());
            if (uvdir)
            {
                utiassert(a.m_row_or_column.size() == m_oldcurve->getSizeY());
            }
            else
            {
                utiassert(a.m_row_or_column.size() == m_oldcurve->getSizeX());
            }
            if (res.m_row_or_column.size() != a.m_row_or_column.size())
            {
                res.m_row_or_column.setSize(a.m_row_or_column.size());
            }
            for (int i = 0; i < res.m_row_or_column.size(); i++)
            {
                res.m_row_or_column[i] = a.m_row_or_column[i] + b.m_row_or_column[i];
            }
        }

        inline void scal(SOneVertex& res, const SOneVertex &a, Scalar weight)
        {

            if (uvdir)
            {
                utiassert(a.m_row_or_column.size() == m_oldcurve->getSizeY());
            }
            else
            {
                utiassert(a.m_row_or_column.size() == m_oldcurve->getSizeX());
            }
            if (res.m_row_or_column.size() != a.m_row_or_column.size())
            {
                res.m_row_or_column.setSize(a.m_row_or_column.size());
            }
            for (int i = 0; i < res.m_row_or_column.size(); i++)
            {
                res.m_row_or_column[i] = a.m_row_or_column[i]* weight;
            }
            // res = a* weight;
        }
        inline void getVertexWeight(int index, SOneVertex& v)
        {
            getVertex(index, v);
            for (int i = 0; i < v.m_row_or_column.size(); i++) {
                //temp.m_row_or_column[i] = v.m_row_or_column[i];
                v.m_row_or_column[i].m_pos *= v.m_row_or_column[i].m_weight;
            }
        }
        inline void getVertex(int index, SOneVertex& v)
        {
            if (uvdir)
            {
                v.m_row_or_column.setSize(m_oldcurve->getSizeY());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    v.m_row_or_column[i] = m_oldcurve->getVertex(index, i);
                }
            }
            else
            {
                v.m_row_or_column.setSize(m_oldcurve->getSizeX());
                for (int i = 0; i < v.m_row_or_column.size(); i++)
                {
                    v.m_row_or_column[i] = m_oldcurve->getVertex(i, index);
                }
            }
            //v = m_oldcurve->getVertex(index);
        }
        inline int getVertexCount()
        {
            if (uvdir)
            {
                return m_oldcurve->getSizeX();
            }
            else
            {
                return m_oldcurve->getSizeY();
            }
        }

        inline int getDegree()
        {
            if (uvdir)
            {
                return m_oldcurve->getDegreeU();
            }
            else
            {
                return m_oldcurve->getDegreeV();
            }
        }

        inline KnotVector* getKnotVector()
        {
            if (uvdir)
            {
                return m_oldcurve->getKnotVectorU();
            }
            else
            {
                return m_oldcurve->getKnotVectorV();
            }
        }
        inline void setVertexProject(int index, const SOneVertex& v)
        {
            SOneVertex temp;
            temp.m_row_or_column.setSize(v.m_row_or_column.size());
            for (int i = 0; i < v.m_row_or_column.size(); i++) {
                temp.m_row_or_column[i] = v.m_row_or_column[i];
                temp.m_row_or_column[i].m_pos /= temp.m_row_or_column[i].m_weight;
            }
            setVertex(index, temp);
        }

        inline void setVertex(int index, SOneVertex& v)
        {
            if (index < m_newcurve1end)
            {
                //m_newcurve1->setVertex(index, v);

                if (uvdir)
                {
                    v.m_row_or_column.setSize(m_oldcurve->getSizeY());
                    for (int i = 0; i < v.m_row_or_column.size(); i++)
                    {
                        m_newcurve1->setVertex(index, i, v.m_row_or_column[i]);
                    }
                }
                else
                {
                    v.m_row_or_column.setSize(m_oldcurve->getSizeX());
                    for (int i = 0; i < v.m_row_or_column.size(); i++)
                    {
                        m_newcurve1->setVertex(i, index, v.m_row_or_column[i]);
                    }
                }
            }
            if (m_newcurve2start <= index)
            {
                // m_newcurve2->setVertex(index-m_newcurve2start, v);

                if (uvdir)
                {
                    v.m_row_or_column.setSize(m_oldcurve->getSizeY());
                    for (int i = 0; i < v.m_row_or_column.size(); i++)
                    {
                        m_newcurve2->setVertex(index-m_newcurve2start, i, v.m_row_or_column[i]);
                    }
                }
                else
                {
                    v.m_row_or_column.setSize(m_oldcurve->getSizeX());
                    for (int i = 0; i < v.m_row_or_column.size(); i++)
                    {
                        m_newcurve2->setVertex(i, index-m_newcurve2start, v.m_row_or_column[i]);
                    }
                }
            }
        }

        inline void createCurve(KnotVector* newknot, int insertpoint)
        {
//            newknot->print();
            m_newcurve1end = insertpoint;
            m_newcurve2start = insertpoint-1;
            int degree = getDegree();
            UTIFLOAT value = newknot->getEntry(insertpoint); //Splitpoint

            KnotVector* newknot1 = new KnotVector(insertpoint+degree+1);
            for (int i = 0; i < insertpoint; i++)
            {
                newknot1->setEntry(i, newknot->getEntry(i));
            }
            for (int i = insertpoint; i <= insertpoint+degree; i++)
            {
                newknot1->setEntry(i, value);
            }
//            newknot1->print();
            if (uvdir)            {
            newknot1->setBezier(m_oldcurve->getKnotVectorU()->getBezier());
           } else {
                newknot1->setBezier(m_oldcurve->getKnotVectorV()->getBezier());
            }
            newknot1->compile();

            KnotVector* newknot2 = new KnotVector(newknot->size()- insertpoint+1);
            for (int i = 0; i <= degree; i++)
            {
                newknot2->setEntry(i, value);
            }
            for (int i = insertpoint+degree; i < newknot->size(); i++)
            {
                newknot2->setEntry(i-( insertpoint)+1, newknot->getEntry(i));
            }

            // newknot2->print();
           // newknot2->setBezier(m_oldcurve->getKnotVector()->getBezier());
            if (uvdir) {
                newknot2->setBezier(m_oldcurve->getKnotVectorU()->getBezier());
            } else {
                newknot2->setBezier(m_oldcurve->getKnotVectorV()->getBezier());
            }
            newknot2->compile();

            if (uvdir)
            {
                m_newcurve1 = new Nurbs2D(newknot1, m_oldcurve->getKnotVectorV(), degree, m_oldcurve->getDegreeV(), m_controldata);
                m_newcurve2 = new Nurbs2D(newknot2,  m_oldcurve->getKnotVectorV(), degree, m_oldcurve->getDegreeV(),m_controldata);
            }
            else
            {
                // m_newcurve1 = new Nurbs2D(newknot1, degree, m_controldata);
                //  m_newcurve2 = new Nurbs2D(newknot2, degree, m_controldata);

                m_newcurve1 = new Nurbs2D(m_oldcurve->getKnotVectorU(), newknot1,  m_oldcurve->getDegreeU(),degree, m_controldata);
                m_newcurve2 = new Nurbs2D( m_oldcurve->getKnotVectorU(), newknot2, m_oldcurve->getDegreeU(), degree,m_controldata);

            }

        }
    };
protected:
    uti_refptr<KnotVector> m_knotvector_u;
    uti_refptr<KnotVector> m_knotvector_v;
    Array2D<SNurbsVertex> m_vertices;
    int m_degree_u;
    int m_degree_v;
    uti_refptr<SubdivisionContainer> m_subdivisioncontainer;
    //Array<UTIFLOAT> m_basisfunctions_u;
    //Array<UTIFLOAT> m_basisfunctions_v;
    //Array2D<UTIFLOAT> m_basisfunctions_u2D;
    //Array2D<UTIFLOAT> m_basisfunctions_v2D;
public:
    struct SSplitResult
    {
        Nurbs2D* first;
        Nurbs2D* second;
    };
    inline bool hasContainer()
    {
        return m_subdivisioncontainer.get() != NULL;
    }
    inline SubdivisionContainer* getSubdivisionContainer()
    {
        return m_subdivisioncontainer.get();
    }
    Nurbs2D(KnotVector* u, KnotVector* v, int degree_u, int degree_v, const ControlData& data=ControlData())
    {
        m_controldata = data;
        utiassert(u->isCompiled() == true);
        utiassert(v->isCompiled() == true);
        m_knotvector_u = u;
        m_degree_u = degree_u;
        m_knotvector_v = v;
        m_degree_v = degree_v;
        m_vertices.setSize(m_knotvector_u->size()-m_degree_u-1, m_knotvector_v->size()-m_degree_v-1);
        for (int j = 0; j < m_vertices.height(); j++)
        {
            for (int i = 0; i < m_vertices.width(); i++)
            {
                m_controldata.init(m_vertices(i,j).m_pos);

            }
        }

    }
    int getSizeX() const
    {
        CHECKEXISTANCE;
        return m_vertices.width();
    }
    int getSizeY() const
    {
        CHECKEXISTANCE;
        return m_vertices.height();
    }
    KnotVector* getKnotVectorU()
    {
        CHECKEXISTANCE;
        return m_knotvector_u.get();
    }
    KnotVector* getKnotVectorV()
    {
        CHECKEXISTANCE;
        return m_knotvector_v.get();
    }

    int getDegreeU()
    {
        CHECKEXISTANCE;
        return m_degree_u;
    }
    int getDegreeV()
    {
        CHECKEXISTANCE;
        return m_degree_v;
    }
    inline ControlData& getControlData()
    {
        CHECKEXISTANCE;
        return m_controldata;
    }
    void setVertex(int i,int j, const Vector& pos, Scalar weight)
    {
        CHECKEXISTANCE;
        m_vertices(i,j).m_pos = pos;
        m_vertices(i,j).m_weight = weight;
    }
    void setVertex(int i,int j, const SNurbsVertex& pos)
    {
        CHECKEXISTANCE;
        m_vertices(i,j) = pos;
        // m_vertices(i,j).m_weight = weight;
    }
    SNurbsVertex getVertex(int i, int j)
    {
        CHECKEXISTANCE;
        return m_vertices(i,j);
    }
    void createSubdivisionStructureNode(SubdivisionNode* node, UTIFLOAT minvalueu, UTIFLOAT maxvalueu,
                                        UTIFLOAT minvaluev, UTIFLOAT maxvaluev, int depth,int maxdepth, float tolerance, bool udir, bool axisaligned )
    {

        /*void findNearestPointBi(SNearestPointResult& result, const Vector& q, UTIFLOAT minvalueu, UTIFLOAT maxvalueu,
                            UTIFLOAT minvaluev, UTIFLOAT maxvaluev, int depth, int maxdepth, int orientation)
        { CHECKEXISTANCE;*/

        UTIFLOAT splitpositionu = (minvalueu+maxvalueu) / 2.0f;
        //bool inbetweenu = (minvalueu < splitpositionu) && (splitpositionu < maxvalueu);

        UTIFLOAT splitpositionv = (minvaluev+maxvaluev) / 2.0f;
        //bool inbetweenv = (minvaluev < splitpositionv) && (splitpositionv < maxvaluev);


        Nurbs2D* n = check_cast<Nurbs2D*>(node->m_nurbs.get());
        //#define NOORIENTED
        if (axisaligned) {
        CBoundsContainer<ControlData>* c = new CBoundsContainer<ControlData>();
        c->m_bounds = n->getBounds();
        node->m_box = c;
        } else {
        COrientedBoundsContainer<ControlData>* c = new COrientedBoundsContainer<ControlData>();
        n->getOrientedBounds(&c->m_orientedbounds);
        node->m_box = c;
        }
        Scalar tol = n->getTolerance();
        Scalar utolerance  = n->getToleranceU();
        Scalar vtolerance = n->getToleranceV();
        node->m_tolerance = tol;
        if (depth >= maxdepth || tol < tolerance)
        {
            return;
        }

        if (false) {
            udir = utolerance > vtolerance;

        }

        SSplitResult res = n->splitSurfaceDir(udir, udir ? splitpositionu : splitpositionv);
        SubdivisionNode* newnodeleft = new SubdivisionNode();
        newnodeleft->m_nurbs = res.first;
        node->m_left =newnodeleft;


        SubdivisionNode* newnoderight = new SubdivisionNode();
        newnoderight->m_nurbs = res.second;
        node->m_right =newnoderight;

        if (udir)
        {

            createSubdivisionStructureNode(newnodeleft, minvalueu, splitpositionu, minvaluev, maxvaluev, depth+1, maxdepth, tolerance, !udir, axisaligned);
            createSubdivisionStructureNode(newnoderight, splitpositionu, maxvalueu, minvaluev, maxvaluev, depth+1, maxdepth, tolerance, !udir, axisaligned);
        }
        else
        {
            createSubdivisionStructureNode(newnodeleft, minvalueu, maxvalueu,minvaluev, splitpositionv,  depth+1, maxdepth, tolerance, !udir, axisaligned);
            createSubdivisionStructureNode(newnoderight, minvalueu, maxvalueu,splitpositionv, maxvaluev,  depth+1, maxdepth, tolerance, !udir, axisaligned);
        }


    }
    void createSubdivisionStructure(int m_maxdepth, float tolerance, bool axisaligned=false)
    {
        m_subdivisioncontainer = new SubdivisionContainer();
        CReferenceList<Nurbs2D> result;
        decomposeSurfaceFull(result);
        for (int i = 0; i < result.size(); i++)
        {
            SubdivisionNode* node = new SubdivisionNode();
            Nurbs2D* n = &result[i];
            node->m_nurbs = n;
            Scalar minvalueu = n->m_knotvector_u->getMinElement();
            Scalar maxvalueu = n->m_knotvector_u->getMaxElement();

            Scalar minvaluev = n->m_knotvector_v->getMinElement();
            Scalar maxvaluev = n->m_knotvector_v->getMaxElement();
            createSubdivisionStructureNode(node, minvalueu, maxvalueu, minvaluev, maxvaluev, 0, m_maxdepth, tolerance, true, axisaligned);
            m_subdivisioncontainer->m_root.push_back(node);
        }
    }
    Scalar getTolerance()
    {
        Scalar tolerance = 0.0;

        for (int i = 0; i < m_vertices.height(); i++) {
            SAccessNurbs2DTolerance<true> nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;
        nurbsaccess.m_baseindex  = i;


        Scalar sc =  getToleranceTemplate(nurbsaccess);
        tolerance = umax(sc, tolerance);
        }

        for (int i = 0; i < m_vertices.width(); i++) {
            SAccessNurbs2DTolerance<false> nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;
        nurbsaccess.m_baseindex  = i;


        Scalar sc =  getToleranceTemplate(nurbsaccess);
        tolerance = umax(sc, tolerance);
        }
       return tolerance;

    }

    Scalar getToleranceU()
    {
        Scalar tolerance = 0.0;

        for (int i = 0; i < m_vertices.height(); i++) {
            SAccessNurbs2DTolerance<true> nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;
        nurbsaccess.m_baseindex  = i;


        Scalar sc =  getToleranceTemplate(nurbsaccess);
        tolerance = umax(sc, tolerance);
        }


       return tolerance;

    }

    Scalar getToleranceV()
    {
        Scalar tolerance = 0.0;


        for (int i = 0; i < m_vertices.width(); i++) {
            SAccessNurbs2DTolerance<false> nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;
        nurbsaccess.m_baseindex  = i;


        Scalar sc =  getToleranceTemplate(nurbsaccess);
        tolerance = umax(sc, tolerance);
        }
       return tolerance;

    }
    Vector calcReferenceValue(Scalar uv_x, Scalar uv_y)
    {
        CHECKEXISTANCE;
        Vector res;
        m_controldata.initZero(res);
        int n = m_vertices.width();
        int m = m_vertices.height();
        Scalar wf = 0.0;

        for (int i = 0; i < n; i++)
        {
            UTIFLOAT v1 = m_knotvector_u->getBasisFunction(i, m_degree_u, uv_x);
            for (int j = 0; j < m; j++)
            {

                UTIFLOAT v2 =  m_knotvector_v->getBasisFunction(j, m_degree_v, uv_y);
                UTIFLOAT v = v1*v2;
                res += v*m_vertices(i,j).m_weight*m_vertices(i,j).m_pos;
                wf += v*m_vertices(i,j).m_weight;
            };
        }
        res /= wf;
        return res;
    }
    Vector calcValue(Scalar uv_x, Scalar uv_y)
    {
        CHECKEXISTANCE;
        SKnotVectorResult resultu;
        SKnotVectorResult resultv;

        static Array<UTIFLOAT> m_basisfunctions_u;
        static Array<UTIFLOAT> m_basisfunctions_v;
        if (m_basisfunctions_u.size() < m_degree_u+1)
        {
            m_basisfunctions_u.setSize(m_degree_u+1);
        }
        if (m_basisfunctions_v.size() < m_degree_v+1)
        {
            m_basisfunctions_v.setSize(m_degree_v+1);
        }
        m_knotvector_u->findKnotVectorPosition(uv_x, resultu);
        m_knotvector_v->findKnotVectorPosition(uv_y, resultv);
        int span_u = resultu.m_index;
        int span_v = resultv.m_index;
        m_knotvector_u->calcBasisFunctions(span_u, uv_x, m_basisfunctions_u, m_degree_u);
        m_knotvector_v->calcBasisFunctions(span_v, uv_y, m_basisfunctions_v, m_degree_v);

        Vector res;
        m_controldata.initZero(res);

        Scalar wf = 0.0;
        for (int i = 0; i <= m_degree_u; i++)
        {
            UTIFLOAT u = m_basisfunctions_u[i];
            int index_u = span_u-m_degree_u+i;
            for (int i = 0; i <= m_degree_v; i++)
            {
                UTIFLOAT v = m_basisfunctions_v[i]*u;
                int index_v = span_v-m_degree_v+i;

                res += v*m_vertices(index_u, index_v).m_weight*m_vertices(index_u, index_v).m_pos;
                wf += v*m_vertices(index_u, index_v).m_weight;
            };
        }
        res /= wf;
#ifndef NDEBUG
        Vector r = calcReferenceValue(uv_x, uv_y);
        utiassert(m_controldata.distanceTo(r, res) < 0.01);
#endif
        return res;
    }

    void calcValueDervNoWeight(Scalar uv_x, Scalar uv_y, Array2D<SNurbsVertex> &SKL, int d) const
    {
        //CHECKEXISTANCE;
        //int d = SKL.width()-1;
        utiassert(SKL.width() == SKL.height());
        int du = umin(m_degree_u, d);
        for (int k = m_degree_u+1; k <= du; k++)
        {
            for (int l = 0; l <= d-k; l++)
            {
                m_controldata.initZero(SKL(k,l).m_pos);
                SKL(k,l).m_weight = 0.0f;
            }
        }
        int dv = umin(m_degree_v, d);
        for (int l = m_degree_v+1; l <= d; l++)
        {
            for (int k = 0; k <= d-l; k++)
            {
                m_controldata.initZero(SKL(k,l).m_pos);
                SKL(k,l).m_weight = 0.0f;
            }
        }
        SKnotVectorResult result_u;

        m_knotvector_u->findKnotVectorPosition(uv_x, result_u);
        int span_u = result_u.m_index;
        //m_basisfunctions2D;
        static Array2D<UTIFLOAT> m_basisfunctions2D_u;
        static Array2D<UTIFLOAT> m_basisfunctions2D_v;
        if (m_basisfunctions2D_u.width() < m_degree_u+1 || m_basisfunctions2D_u.height() < m_degree_u+1)
        {
            m_basisfunctions2D_u.setSize(m_degree_u+1, m_degree_u+1);
        }

        m_knotvector_u->dersBasisFunctions(span_u, uv_x, m_basisfunctions2D_u, m_degree_u, m_degree_u);


        SKnotVectorResult result_v;

        m_knotvector_v->findKnotVectorPosition(uv_y, result_v);
        int span_v = result_v.m_index;
        //m_basisfunctions2D;
        if (m_basisfunctions2D_v.width() < m_degree_v+1 || m_basisfunctions2D_v.height() < m_degree_v+1)
        {
            m_basisfunctions2D_v.setSize(m_degree_v+1, m_degree_v+1);
        }

        m_knotvector_v->dersBasisFunctions(span_v, uv_y, m_basisfunctions2D_v, m_degree_v, m_degree_v);


        static Array<SNurbsVertex> temp;
        int tempsize = umax(m_degree_v,d)+1;
        if (temp.size() < tempsize)
        {
            temp.setSize(tempsize);
        }
        for (int k = 0; k <= du; k++)
        {
            for (int s = 0; s <= m_degree_v; s++)
            {
                m_controldata.initZero(temp[s].m_pos);
                temp[s].m_weight = 0.0;
                int index_v = span_v-m_degree_v+s;
                for (int r = 0; r <= m_degree_u; r++)
                {
                    int index_u = span_u-m_degree_u+r;
                    double w = m_basisfunctions2D_u(k,r);
                    temp[s].m_pos = temp[s].m_pos + w*m_vertices(index_u, index_v).m_weight*m_vertices(index_u, index_v).m_pos;
                    temp[s].m_weight = temp[s].m_weight + w*m_vertices(index_u, index_v).m_weight;
                }

            }
            int dd = umin(d-k, dv);
            for (int l = 0; l <= dd; l++)
            {
                m_controldata.initZero(SKL(k,l).m_pos);
                SKL(k,l).m_weight = 0.0;
                for (int s = 0; s <= m_degree_v; s++)
                {
                    double w = m_basisfunctions2D_v(l,s);
                    SKL(k,l).m_pos = SKL(k,l).m_pos+w*temp[s].m_pos;
                    SKL(k,l).m_weight = SKL(k,l).m_weight+w*temp[s].m_weight;
                }
            }



        }

        /*Vector res;
        for (int k = 0; k <= du; k++) {

           m_controldata.initZero(res);

           Scalar wf = 0.0;
           for (int i = 0; i <= m_degree_u; i++)
           {
            UTIFLOAT u = m_basisfunctions_u(k,i);
            int index_u = span_u-m_degree_u+i;
            for (int i = 0; i <= m_degree_v; i++)
            {
                UTIFLOAT v = m_basisfunctions_v(k, i)*u;
                int index_v = span_v-m_degree_v+i;

                res += v*m_vertices(index_u, index_v).m_weight*m_vertices(index_u, index_v).m_pos;
                wf += v*m_vertices(index_u, index_v).m_weight;
            };
           }
        // res /= wf;
           CK[k].m_pos = res;
           CK[k].m_weight = wf;
        }*/
    }
    void calcValueDerv(Scalar uv_x, Scalar uv_y, Array2D<Vector> &SKL, int d) const
    {

        //3D Matrix
        // int d = SKL.width()-1;

        static Array2D<SNurbsVertex> Aderv;
        if (Aderv.width() < SKL.width() || Aderv.height() < SKL.height())
        {
            Aderv.setSize(SKL.width(), SKL.height());
        }

        calcValueDervNoWeight(uv_x, uv_y, Aderv, d);
//Ohne Wights;
      /*  for (int i = 0; i < SKL.width(); i++) {
            for (int j = 0; j < SKL.height(); j++) {
                SKL(i,j) = Aderv(i,j).m_pos;
            }
        }
        return;*/

        Vector v, v2;
        for (int k = 0; k <= d; k++)
        {
            for (int l = 0; l <= d-k; l++)
            {
                v = Aderv(k,l).m_pos;
                for (int j = 1; j <= l; j++)
                {
                    v = v - bin(l,j)*Aderv(0,j).m_weight*SKL(k,l-j);
                }
                for (int i = 1; i <= k; i++)
                {
                    v = v - bin(k,i)*Aderv(i,0).m_weight*SKL(k-i, l);
                    m_controldata.initZero(v2);
                    for (int j = 1; j <= l; j++)
                    {
                        v2 = v2 + bin(l,j)*Aderv(i,j).m_weight*SKL(k-i, l-j);
                    }
                    v = v - (bin(k,i)*1.0)*v2;
                }
                SKL(k,l) = v / Aderv(0,0).m_weight;
            }
        }
    }
    void newtonDelta(Scalar u, Scalar v, const Vector& p, Scalar &dx, Scalar& dy) const
    {
        //void calcDervValue(Scalar u, Array<Vector>& CK)const
        static Array2D<Vector> CK(3, 3);

        calcValueDerv(u, v, CK, 2);
        //calcValueDervNoWeight(u, v, CK, 2);
        Scalar fu = m_controldata.dot(p-CK(0,0), CK(1, 0));
        Scalar fv = m_controldata.dot(p-CK(0,0), CK(0, 1));

        Vector same = (p-CK(0, 0))*(-1.0);
        Scalar fda = (m_controldata.dot(CK(1, 0), CK(1, 0)) + m_controldata.dot(same, CK(2,0)));
        Scalar fdb = (m_controldata.dot(CK(1, 0), CK(0, 1)) + m_controldata.dot(same, CK(1,1)));
        Scalar fdc = fdb;
        Scalar fdd = (m_controldata.dot(CK(0, 1), CK(0, 1)) + m_controldata.dot(same, CK(0,2)));
        Scalar ddet = fda*fdd-fdb*fdc;

        if (fabs(ddet) < 0.000001)
        {
            dx = 0;
            dy = 0;
            return;
        }
        fda = fda/ddet;
        fdb = fdb/ddet;
        fdc = fdc/ddet;
        fdd = fdd/ddet;

        Scalar du, dv;
        du = (fdd*fu -fdb*fv);
        dv = (-fdc*fu + fda*fv);
        dx = du;
        dy = dv;
        return;
    }
    bool improveFoundMinima(Scalar &u, Scalar &v, const Vector& p, Scalar tolerance, Scalar aborttolerance) const
    {
        UTIFLOAT minposu = m_knotvector_u->getMinElement();
        UTIFLOAT maxposu = m_knotvector_u->getMaxElement();

        UTIFLOAT minposv = m_knotvector_v->getMinElement();
        UTIFLOAT maxposv = m_knotvector_v->getMaxElement();
        Scalar currentvalueu = u;
        Scalar currentvaluev = v;
        Scalar originalvalueu = u;
        Scalar originalvaluev = v;
        Scalar oldvalueu = u;
        Scalar oldvaluev = v;
        Scalar deltau, deltav;
        for (int i = 0; i < 10; i++)
        {
            newtonDelta(currentvalueu, currentvaluev,p, deltau, deltav);
            oldvalueu = currentvalueu;
            oldvaluev = currentvaluev;
            currentvalueu += deltau;
            currentvaluev += deltav;
            currentvalueu = umin(maxposu, umax(minposu, currentvalueu));
            currentvaluev = umin(maxposv, umax(minposv, currentvaluev));
            deltau = currentvalueu-oldvalueu;
            deltav = currentvaluev-oldvaluev;
            if (sqrt(sqr(currentvalueu-originalvalueu)+sqr(currentvaluev-originalvaluev)) > tolerance)
            {
                return false;
            }

            if (sqrt(sqr(deltau)+sqr(deltav)) < aborttolerance)
            {
                u = currentvalueu;
                v = currentvaluev;
                return true;
            }
        }
        u = currentvalueu;
        v = currentvaluev;
        return true;

    }
    Vector getMeanOfAllVectors()
    {
        Vector v;
        m_controldata.initZero(v);
        for (int i = 0; i < m_vertices.width(); i++) {
            for (int j = 0; j < m_vertices.height(); j++) {
                v += m_vertices(i,j).m_pos;
            }
        }
        v = v * (1.0/ (m_vertices.width()*m_vertices.height()));
        return v;
    }
    UMat<Scalar> getCovarianceMatrix()
    {
        Vector mean = getMeanOfAllVectors();
        int d = m_controldata.getDimensions();
        Scalar N = m_vertices.width()*m_vertices.height();
        UMat<Scalar> res(d, d);
        for (int k = 0; k < d; k++) {
            for (int l = 0; l < d; l++) {
                    Scalar currentvalue=0.0;
                    Scalar mean1 = m_controldata.getVertexIndex(mean, k);
                    Scalar mean2 = m_controldata.getVertexIndex(mean, l);
                    for (int i = 0; i < m_vertices.width(); i++) {
                       for (int j = 0; j < m_vertices.height(); j++) {
                            Vector & x = m_vertices(i,j).m_pos;
                            currentvalue += (m_controldata.getVertexIndex(x, k)-mean1)*(m_controldata.getVertexIndex(x, l)-mean2);
                       }
                    }
                    currentvalue /= N;
                    res.set(k,l, currentvalue);
            }
        }
        return res;

    }


    template<bool udir>  Nurbs2D* curveKnotInsert(UTIFLOAT newknotvalue, int amount)
    {
        CHECKEXISTANCE;


        SAccessNurbs2D<udir> nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;

        curveKnotInsertTemplate(nurbsaccess, newknotvalue, amount);

        return nurbsaccess.m_newcurve;

    }
    SSplitResult splitSurfaceDir(bool udir, UTIFLOAT splitposition)
    {
        if (udir)
        {
            return splitSurface<true>(splitposition);
        }
        else
        {
            return splitSurface<false>(splitposition);
        }
    }
    template<bool udir>  SSplitResult splitSurface(UTIFLOAT splitposition)
    {
        CHECKEXISTANCE;
        SKnotVectorResult result;
        if (udir)
        {
            m_knotvector_u->findKnotVectorPosition(splitposition, result);
        }
        else
        {
            m_knotvector_v->findKnotVectorPosition(splitposition, result);
        }

        int s = 0;
        if (fabs(result.m_a-splitposition) < 0.0001)
        {
            splitposition = result.m_a;
            s =result.m_a_count;
        }
        utiassert(s <= (udir ?  getDegreeU() : getDegreeV()));
        SAccessNurbs2DSplit<udir> nurbsaccess;
        nurbsaccess.m_oldcurve = this;
        nurbsaccess.m_controldata = m_controldata;

        curveKnotInsertTemplate(nurbsaccess, splitposition, (udir?  getDegreeU(): getDegreeV())-s);
        Nurbs2D::SSplitResult res;
        res.first = nurbsaccess.m_newcurve1;
        res.second = nurbsaccess.m_newcurve2;

        return res;
    }
    template<bool udir>  void decomposeSurface(CReferenceList<Nurbs2D>& result)
    {
        CHECKEXISTANCE;
        SAccessNurbs2DToBezier<udir> accessor;
        accessor.m_oldcurve = this;
        accessor.m_controldata = m_controldata;
        decomposeCurveTemplate(accessor);
        accessor.finalizeBezier();
        for (int i = 0; i < accessor.m_bezierlist.size(); i++)
        {
            result.push_back(&accessor.m_bezierlist[i]);
        }
    }
    void decomposeSurfaceFull(CReferenceList<Nurbs2D>& result)
    {
        if (m_knotvector_u->getBezier() && m_knotvector_v->getBezier())
        {
            result.push_back(this);
            return;
        }
        CReferenceList<Nurbs2D> tempres;
        decomposeSurface<true>(tempres);
        for (int i = 0; i < tempres.size(); i++)
        {
            tempres[i].decomposeSurface<false>(result);
        }
    }
public:
    void findNearestPointBruteForce(const Vector& v, int amount, Scalar &u_res, Scalar &v_res)
    {
        CHECKEXISTANCE;
        UTIFLOAT minposu = m_knotvector_u->getMinElement();
        UTIFLOAT maxposu = m_knotvector_u->getMaxElement();
        UTIFLOAT minposv = m_knotvector_v->getMinElement();
        UTIFLOAT maxposv = m_knotvector_v->getMaxElement();
        UTIFLOAT currentdistance = 1E+10;
        //UTIFLOAT resultpos = 0;
        Vector res;
        m_controldata.init(res);
        for (int i = 0; i < amount; i++)
        {
            for (int j = 0; j < amount; j++)
            {
                UTIFLOAT s = (UTIFLOAT)i/(UTIFLOAT)(amount-1);
                s = minposu + s*(maxposu-minposu);
                UTIFLOAT t = (UTIFLOAT)j/(UTIFLOAT)(amount-1);
                t = minposv + t*(maxposv-minposv);
                res  = calcValue(s, t);
                UTIFLOAT newdistance = m_controldata.distanceTo(res, v);
                if (newdistance < currentdistance)
                {
                    currentdistance = newdistance;
                    u_res = s;
                    v_res = t;
                }
            }
        }
        //return resultpos;
    }

    SBounds<Vector> getBounds()
    {
        CHECKEXISTANCE;
        SBounds<Vector> b;
        m_controldata.init(b.m_min);
        m_controldata.init(b.m_max);
        int d = m_controldata.getDimensions();
        b.m_min = m_vertices(0, 0).m_pos;
        b.m_max = m_vertices(0, 0).m_pos;
        for (int i = 0; i < m_vertices.width(); i++)
        {
            for (int j = 0; j < m_vertices.height(); j++)
            {
                Vector v = m_vertices(i,j).m_pos;
                for (int k = 0; k < d; k++)
                {
                    m_controldata.setVertexIndex(b.m_min, k, umin(m_controldata.getVertexIndex(b.m_min, k), m_controldata.getVertexIndex(v, k)));
                    m_controldata.setVertexIndex(b.m_max, k, umax(m_controldata.getVertexIndex(b.m_max, k), m_controldata.getVertexIndex(v, k)));
                }
            }
        }
        /* for (int i = 0; i < d; i++) {
             m_controldata.setVertexIndex(b.m_min, i, 1E+10);
             m_controldata.setVertexIndex(b.m_max, i, -1E+10);
         }*/

        return b;
    }
    void getOrientedBounds(SOrientedBounds<Vector>* bounds)
    {
        /*UMat<Scalar> covariance(getCovarianceMatrix());
        int m = covariance.m();
        uti_refptr<CEigenSymmWorkspace> eigensymm(new CEigenSymmWorkspace(m));
        eigensymm->calculateEvalAndEvec(covariance);
        eigensymm->sortEvalAndEvec(estAbsDesc);
        UMat<Scalar> newsystem(eigensymm->getEvec());

        uti_refptr<CLUDecomposition> decomposition;
        decomposition = new CLUDecomposition(newsystem);
        bounds->m_A = newsystem;
        bounds->m_workspace = decomposition;

        SBounds<Vector> b;
        m_controldata.init(b.m_min);
        m_controldata.init(b.m_max);
        int d = m_controldata.getDimensions();
        Vector temp = decomposition->solveSystem(m_vertices(0, 0).m_pos);
        b.m_min = temp;
        b.m_max = temp;
        for (int i = 0; i < m_vertices.width(); i++)
        {
            for (int j = 0; j < m_vertices.height(); j++)
            {
                Vector v = m_vertices(i,j).m_pos;
                v = decomposition->solveSystem(v);
                for (int k = 0; k < d; k++)
                {
                    m_controldata.setVertexIndex(b.m_min, k, umin(m_controldata.getVertexIndex(b.m_min, k), m_controldata.getVertexIndex(v, k)));
                    m_controldata.setVertexIndex(b.m_max, k, umax(m_controldata.getVertexIndex(b.m_max, k), m_controldata.getVertexIndex(v, k)));
                }
            }
        }
        bounds->m_oriented = b;*/


    }
    struct SNearestPointResult
    {
        bool m_found;
        UTIFLOAT m_s;
        UTIFLOAT m_t;
        UTIFLOAT m_distance;
        int m_calls;
        int m_distancechecks;
        int m_splits;
        int m_finalstep;
        UTIFLOAT m_maxtolerance;
        Nurbs2D* m_surface;
        SNearestPointResult()
        {
            m_found =false;
            m_s = 0.0f;
            m_s = 0.0f;
            m_distance = 1E+10;
            m_calls = 0;
            m_distancechecks = 0;
            m_splits = 0;
            m_finalstep= 32;
            m_surface=NULL;
            m_maxtolerance = 0.05;

        }
    };
protected:


    void findNearestPointBi(SNearestPointResult& result, SubdivisionNode* node, const Vector& q, UTIFLOAT minvalueu, UTIFLOAT maxvalueu,
                            UTIFLOAT minvaluev, UTIFLOAT maxvaluev, int depth, int maxdepth, int orientation)
    {
        CHECKEXISTANCE;

        UTIFLOAT splitpositionu = (minvalueu+maxvalueu) / 2.0f;
        bool inbetweenu = (minvalueu < splitpositionu) && (splitpositionu < maxvalueu);

        UTIFLOAT splitpositionv = (minvaluev+maxvaluev) / 2.0f;
        bool inbetweenv = (minvaluev < splitpositionv) && (splitpositionv < maxvaluev);
        result.m_calls += 1;
        Scalar tolerance = 1E+10;
        if (node) {
            tolerance = node->m_tolerance;

        }
        if (depth >= maxdepth || (!inbetweenu) || (!inbetweenv) || tolerance < result.m_maxtolerance)
        {
            Vector& p00 = m_vertices(0,0).m_pos;
            Vector& p10 = m_vertices((m_vertices.width()-1),0).m_pos;
            Vector& p11 = m_vertices((m_vertices.width()-1),(m_vertices.height()-1)).m_pos;
            Vector& p01 = m_vertices(0,(m_vertices.height()-1)).m_pos;

            Scalar u,v;

            UTIFLOAT distance = getMinDistanceToTriangle(m_controldata, p00, p10, p01, q, u,v);
            if (distance < result.m_distance)
                    {
                        result.m_distance = distance;
                        result.m_found = true;
                        result.m_s = minvalueu+(maxvalueu-minvalueu)*u;
                        result.m_t = minvaluev+(maxvaluev-minvaluev)*v;
                        result.m_distancechecks += 1;
                    }
            distance = getMinDistanceToTriangle(m_controldata, p11, p01, p10, q, u,v);
            if (distance < result.m_distance)
                    {
                        result.m_distance = distance;
                        result.m_found = true;
                        u = 1.0-u;
                        v = 1.0-v;
                        result.m_s = minvalueu+(maxvalueu-minvalueu)*u;
                        result.m_t = minvaluev+(maxvaluev-minvaluev)*v;
                        result.m_distancechecks += 1;
                    }
                 /*   const int c = 40;
            for (int i = 0; i <= c; i++)
            {
                for (int j = 0; j <= c; j++)
                {
                    u = (Scalar)i/(Scalar)c;
                    v = (Scalar)j/(Scalar)c;
                    Vector vres = ((1.0-u)*(1.0-v))*p00+((u)*(1.0-v))*p10+((u)*(v))*p11+((1.0-u)*(v))*p01;
                    UTIFLOAT distance = m_controldata.distanceTo(vres, q);
                    if (distance < result.m_distance)
                    {
                        result.m_distance = distance;
                        result.m_found = true;
                        result.m_s = minvalueu+(maxvalueu-minvalueu)*u;
                        result.m_t = minvaluev+(maxvaluev-minvaluev)*v;
                        result.m_distancechecks += 1;
                    }
                }
            }*/
            /*for (int i = 0; i < 2; i++)
            {
                for (int j = 0; j < 2; j++)
                {
                    Vector v = m_vertices(i*(m_vertices.width()-1) ,j*(m_vertices.height()-1)).m_pos;
                    UTIFLOAT distance = m_controldata.distanceTo(v, q);
                    if (distance < result.m_distance)
                    {
                        result.m_distance = distance;
                        result.m_found = true;
                        result.m_s = (i == 0 ? minvalueu : maxvalueu);
                        result.m_t = (j == 0 ? minvaluev : maxvaluev);
                        result.m_distancechecks += 1;
                    }
                }
            }*/

        }
        else
        {

            SSplitResult res;
            SubdivisionNode* nodeleft = NULL;
            SubdivisionNode* noderight = NULL;
            Scalar firstmindistance;
            Scalar secondmindistance;
            int neworientation = (orientation+1) % 2;

            if (node && node->m_left.get())
            {
                nodeleft = node->m_left.get();
                noderight = node->m_right.get();
                res.first = check_cast<Nurbs2D<ControlData>* >(node->m_left->m_nurbs.get());
                res.second = check_cast<Nurbs2D<ControlData>* >(node->m_right->m_nurbs.get());
                //firstmindistance = getMinDistanceToBounds(m_controldata, check_cast<CBoundsContainer<ControlData>* >(node->m_left->m_box.get())->m_bounds, q);
                //secondmindistance = getMinDistanceToBounds(m_controldata, check_cast<CBoundsContainer<ControlData>* >(node->m_right->m_box.get())->m_bounds, q);

                firstmindistance = check_cast<CBaseBoundsContainer<ControlData>* >(node->m_left->m_box.get())->getMinimumDistance(m_controldata, q);
                // getMinDistanceToBounds(m_controldata, check_cast<CBoundsContainer<ControlData>* >(node->m_left->m_box.get())->m_bounds, q);
                secondmindistance = check_cast<CBaseBoundsContainer<ControlData>* >(node->m_right->m_box.get())->getMinimumDistance(m_controldata, q);
                //getMinDistanceToBounds(m_controldata, check_cast<CBoundsContainer<ControlData>* >(node->m_right->m_box.get())->m_bounds, q);


                /* SBounds<Vector> firstbounds = res.first->getBounds();
                SBounds<Vector> secondbounds = res.second->getBounds();
                firstmindistance = getMinDistanceToBounds(m_controldata, firstbounds, q);
                secondmindistance = getMinDistanceToBounds(m_controldata, secondbounds, q);*/
            }
            else
            {
               if (orientation == 0)
               {

                   res = splitSurface<true>(splitpositionu);
               }
               else
               {
                   res = splitSurface<false>(splitpositionv);
               }


               result.m_splits += 1;
               SBounds<Vector> firstbounds = res.first->getBounds();
               SBounds<Vector> secondbounds = res.second->getBounds();
               firstmindistance = getMinDistanceToBounds(m_controldata, firstbounds, q);
               secondmindistance = getMinDistanceToBounds(m_controldata, secondbounds, q);
            }
            if (firstmindistance < secondmindistance)
            {
                if (firstmindistance > result.m_distance)
                {
                    return;
                };
                if (orientation == 0)
                {
                    res.first->findNearestPointBi(result, nodeleft, q, minvalueu, splitpositionu, minvaluev, maxvaluev, depth+1, maxdepth, neworientation);
                }
                else
                {
                    res.first->findNearestPointBi(result, nodeleft, q, minvalueu, maxvalueu, minvaluev, splitpositionv, depth+1, maxdepth, neworientation);
                }
                if (secondmindistance > result.m_distance)
                {
                    return;
                };
                if (orientation == 0)
                {
                    res.second->findNearestPointBi(result, noderight, q,  splitpositionu, maxvalueu,  minvaluev, maxvaluev, depth+1, maxdepth, neworientation);

                }
                else
                {
                    res.second->findNearestPointBi(result, noderight, q,  minvalueu, maxvalueu, splitpositionv, maxvaluev,  depth+1, maxdepth, neworientation);

                }
            }
            else
            {
                if (secondmindistance > result.m_distance)
                {
                    return;
                };
                if (orientation == 0)
                {
                    res.second->findNearestPointBi(result, noderight, q,splitpositionu, maxvalueu, minvaluev, maxvaluev, depth+1, maxdepth, neworientation);
                }
                else
                {
                    res.second->findNearestPointBi(result, noderight, q,minvalueu, maxvalueu,splitpositionv, maxvaluev,  depth+1, maxdepth, neworientation);

                }
                if (firstmindistance > result.m_distance)
                {
                    return;
                };
                if (orientation == 0)
                {
                    res.first->findNearestPointBi(result, nodeleft, q, minvalueu, splitpositionu, minvaluev, maxvaluev,depth+1, maxdepth, neworientation);
                }
                else
                {
                    res.first->findNearestPointBi(result, nodeleft, q, minvalueu, maxvalueu,minvaluev, splitpositionv, depth+1, maxdepth, neworientation);
                }
            }

        }
    }
public:
    SNearestPointResult findNearestPointBisection(const Vector& q, int maxdepth, UTIFLOAT maxdistance, bool polish, UTIFLOAT maxtolerance)
    {
        CHECKEXISTANCE;
        SNearestPointResult nearestpoint;
        nearestpoint.m_s = 0.0f;
        nearestpoint.m_t = 0.0f;
        nearestpoint.m_distance = maxdistance;
        nearestpoint.m_found = false;
        nearestpoint.m_maxtolerance = maxtolerance;

        if (m_subdivisioncontainer.get())
        {
            CReferenceList<Nurbs2D> resultlist;
            std::vector< std::pair<int, Scalar> > distances;
            for (int i = 0; i < m_subdivisioncontainer->m_root.size(); i++)
            {
                SubdivisionNode* n = &m_subdivisioncontainer->m_root[i];
                resultlist.push_back(check_cast<Nurbs2D*>(n->m_nurbs.get()));
                Scalar firstmindistance = check_cast<CBaseBoundsContainer<ControlData>*>(n->m_box.get())->getMinimumDistance(m_controldata, q);
                distances.push_back(std::make_pair(i, firstmindistance));
            }
            std::sort(distances.begin(), distances.end(), sortNURBS<Scalar>);
            for (int i = 0; i < distances.size(); i++) {
               std::pair<int, Scalar>& element = distances[i];
               if (element.second > nearestpoint.m_distance) {
                   break;
               }
               Nurbs2D& n = resultlist[element.first];
               SubdivisionNode* sn = NULL;
               if (m_subdivisioncontainer.get()) {
                   sn = &m_subdivisioncontainer->m_root[element.first];
               }
                nearestpoint.m_surface = &n;
                if (element.second < nearestpoint.m_distance) {
                    n.findNearestPointBi(nearestpoint, sn, q, n.getKnotVectorU()->getMinElement(), n.getKnotVectorU()->getMaxElement(),n.getKnotVectorV()->getMinElement(), n.getKnotVectorV()->getMaxElement(), 0, maxdepth, 0);
                }
            }
        }
        else
        {
            findNearestPointBi(nearestpoint, NULL, q, m_knotvector_u->getMinElement(), m_knotvector_u->getMaxElement(), m_knotvector_v->getMinElement(), m_knotvector_v->getMaxElement(), 0, maxdepth, 0);

        };




        if (polish) {
                improveFoundMinima(nearestpoint.m_s, nearestpoint.m_t, q, 0.10, 0.000001);
           UTIFLOAT minposu = m_knotvector_u->getMinElement();
           UTIFLOAT maxposu = m_knotvector_u->getMaxElement();

           UTIFLOAT minposv = m_knotvector_v->getMinElement();
           UTIFLOAT maxposv = m_knotvector_v->getMaxElement();
           nearestpoint.m_s = umax(umin(maxposu, nearestpoint.m_s), minposu);
           nearestpoint.m_t = umax(umin(maxposv, nearestpoint.m_t), minposv);

           Vector v = calcValue(nearestpoint.m_s, nearestpoint.m_t);
           nearestpoint.m_distance = m_controldata.distanceTo(v, q);
        }

        nearestpoint.m_surface = const_cast<Nurbs2D*>(this);

        return nearestpoint;
    }
    void move(const Vector& v)
    {
        CHECKEXISTANCE;
        for (int i = 0; i< m_vertices.width(); i++)
        {
            for (int j = 0; j< m_vertices.height(); j++)
            {
                m_vertices(i,j).m_pos += v;
            }
        }
    }
    void print()
    {
        std::wcout << "knotvector_u: " << std::endl;
        m_knotvector_u->print();
        std::wcout << "knotvector_v: " << std::endl;
        m_knotvector_v->print();
        for (int i = 0; i< m_vertices.width(); i++)
        {
            for (int j = 0; j< m_vertices.height(); j++)
            {
                // m_vertices(i,j).m_pos += v;
                std::wcout << "(" << i << ", " << j << "): (" << m_vertices(i,j).m_pos.x << ", " << m_vertices(i,j).m_pos.y << ", " << m_vertices(i,j).m_pos.z << ", " << m_vertices(i,j).m_weight <<")" << std::endl;

            }
        }

        /*for (int i = 0; i < m_vertices.size(); i++)
        {
         }*/
    }
};

template <class V, class S, int dimensions> struct StaticControlData
{
    typedef V Vector;
    typedef S Scalar;
    inline StaticControlData()
    {

    }
    inline void setDimensions(int d)
    {
        utiassert(d == dimensions);
    }
    inline int getDimensions()const
    {
        return dimensions;

    }
    inline Scalar getVertexIndex(const V& v, int index)const
    {
        utiassert(index >= 0 && index < dimensions);
        return v[index];

    }
    inline void setVertexIndex(V& v, int index, Scalar value)const
    {
        utiassert(index >= 0 && index < dimensions);
        v[index] = value;
    }
    inline void init(Vector& v) const{}
    inline void initZero(Vector& v)const {
       for (int i = 0; i < dimensions; i++) {
           v[i] = Scalar(0);
       }
    }
    inline Scalar distanceTo(const Vector& v1, const Vector& v2)const
    {
        return v1.distanceTo(v2);
    }
    inline Scalar normalize(Vector& v1) const
    {
       return v1.normalize();

    }
    inline Scalar length(const Vector& v1) const
    {
       return v1.length();

    }
    inline Scalar dot(const Vector& v1, const Vector& v2)const
    {
        return v1.dotProduct(v2);
    }


};

template <class S> struct DynamicMatControlData
{
    int m_dimensions;
    typedef UMat<S> Vector;
    typedef S Scalar;
    inline DynamicMatControlData()
    {
        m_dimensions = 0;
    }
    inline void setDimensions(int d)
    {
        utiassert(m_dimensions == 0);
        utiassert(d != 0);
        m_dimensions = d;
    }
    inline int getDimensions()const
    {
        return m_dimensions;

    }
    inline Scalar getVertexIndex(const Vector& v, int index)const
    {
        utiassert(m_dimensions != 0);
        utiassert(index >= 0 && index < m_dimensions);
        return v(0,index);

    }
    inline void setVertexIndex(Vector& v, int index, Scalar value)const
    {
        utiassert(m_dimensions != 0);
        utiassert(index >= 0 && index < m_dimensions);
        return v.set(0, index, value);
    }
    inline void init(Vector& v)const{
        utiassert(m_dimensions != 0);
        v = Vector(1, m_dimensions);

     }
    inline void initZero(Vector& v)const {
        utiassert(m_dimensions != 0);
        init(v);
       for (int i = 0; i < m_dimensions; i++) {
           v.set(0, i, Scalar(0));
       }
    }
    inline Scalar distanceTo(const Vector& v1, const Vector& v2)const
    {
        Scalar s= 0.0;
        for (int index = 0; index < m_dimensions; index++) {
            s += sqr(v1(0,index)-v2(0,index));
        }
        return sqrt(s);
    }
    inline Scalar normalize(Vector& v1) const
    {
        Scalar s= 0.0;
        for (int index = 0; index < m_dimensions; index++) {
            s += sqr(v1(0,index));
        }
        s = sqrt(s);
        v1 = v1/s;
        return s;

    }
    inline Scalar length(const Vector& v1) const
    {
        Scalar s= 0.0;
        for (int index = 0; index < m_dimensions; index++) {
            s += sqr(v1(0,index));
        }
        s = sqrt(s);
        return s;

    }
    inline Scalar dot(const Vector& v1, const Vector& v2)const
    {
        Scalar s= 0.0;
        for (int index = 0; index < m_dimensions; index++) {
            s += v1(0,index)*v2(0,index);
        }
        return s;

    }

};

typedef StaticControlData<CVector4, double, 4>  Vector4Data;
typedef StaticControlData<CVector3, double, 3>  Vector3Data;
typedef StaticControlData<CVector2, double, 2>  Vector2Data;
typedef StaticControlData<double, double, 1>  Vector1Data;

typedef StaticControlData<CVector4f, float, 3>  Vector4FloatData;
typedef StaticControlData<CVector3f, float, 3>  Vector3FloatData;
typedef StaticControlData<CVector2f, float, 2>  Vector2FloatData;
typedef Nurbs1D<Vector4FloatData> NurbsCurve4Df;
typedef Nurbs1D<Vector3FloatData> NurbsCurve3Df;
typedef Nurbs1D<Vector2FloatData> NurbsCurve2Df;

typedef Nurbs2D<Vector4FloatData> NurbsSurface4Df;
typedef Nurbs2D<Vector3FloatData> NurbsSurface3Df;
typedef Nurbs2D<Vector2FloatData> NurbsSurface2Df;

typedef Nurbs1D<Vector4Data> NurbsCurve4D;
typedef Nurbs1D<Vector3Data> NurbsCurve3D;
typedef Nurbs1D<Vector2Data> NurbsCurve2D;
typedef Nurbs1D<Vector1Data> NurbsCurve1D;

typedef Nurbs2D<Vector4Data> NurbsSurface4D;
typedef Nurbs2D<Vector3Data> NurbsSurface3D;
typedef Nurbs2D<Vector2Data> NurbsSurface2D;

enum TKnotVectorType {kvtGiven, kvtEqually, kvtChordLength, kvtCentripetal};
enum TTangentType {ttGiven, tt3Point, tt4Point};

template<class T> class NurbsGenerator {
public:
  typedef typename T::CurveControlData::Vector Vector;
        typedef typename T::CurveControlData::Scalar Scalar;
	typedef typename T::CurveControlData CurveControlData;
	CurveControlData m_controldata;
	struct SOnePosition {
	  Vector m_vector;
	  Scalar m_position;
	  Vector m_tangent;
	};
   std::vector<SOnePosition> m_positions;
   Array<Scalar> m_uk;
   ArraySE<Vector> m_Dk;
   Array<Vector> m_Tk;
   ArraySE<Vector> m_Vk;
   ArraySE<Vector> m_qk;
   ArraySE<Vector> m_dk;
   ArraySE<Scalar> m_deltauk;
   ArraySE<Scalar> m_ak;
   int m_degree;
   TKnotVectorType  m_knottype;
   Scalar lengthCrossProduct(const Vector& v1, const Vector& v2)
   {
       if (m_controldata.getDimensions() == 3){
            Scalar a1 = m_controldata.getVertexIndex(v1, 0);
            Scalar a2 = m_controldata.getVertexIndex(v1, 1);
            Scalar a3 = m_controldata.getVertexIndex(v1, 2);

            Scalar b1 = m_controldata.getVertexIndex(v2, 0);
            Scalar b2 = m_controldata.getVertexIndex(v2, 1);
            Scalar b3 = m_controldata.getVertexIndex(v2, 2);

            Scalar r1 = a2*b3-a3*b2;
            Scalar r2 = a3*b1-a1*b3;
            Scalar r3 = a1*b2-a2*b1;

            Scalar length = sqrt(r1*r1+r2*r2+r3*r3);
            return length;
       } else {
       Scalar lv1 = m_controldata.length(v1);
       Scalar lv2 = m_controldata.length(v2);
       Vector bv1 = v1;
       Vector bv2 = v2;
       bv1 = bv1 / lv1;
       bv2 = bv2 / lv2;
       float cosangle = m_controldata.dot(bv1, bv2);
       float sinusangle = sqrt(1.0-cosangle*cosangle);
       return lv1*lv2*sinusangle;
       }

   }
   NurbsGenerator(const CurveControlData& controldata= CurveControlData())
   {
       m_degree = 3;
       m_knottype = kvtChordLength;
       m_controldata = controldata;
   }
   /*void setDegree(int d)
   {
     m_degree = d;
   }
   int getDegree()
   {
     return m_degree;
   };*/
   void addPosition(const Vector& v, const Scalar& p=0.0, const Vector& tangent=Vector())
   {
    SOnePosition pres;
    pres.m_vector = v;
    pres.m_position = p;
    pres.m_tangent = tangent;
    m_positions.push_back(pres);
   }
   KnotVector* createKnotVector(TKnotVectorType knottype)
   {
       m_uk.setSize(m_positions.size());
       switch (knottype) {
           case kvtGiven: {
       for (int i = 0; i < m_positions.size(); i++) {
          m_uk[i] = (m_positions.at(i).m_position);

       };
           }; break;
           case kvtEqually: {
                for (int i = 0; i < m_positions.size(); i++) {
                 m_uk[i] = ((double)i / (double) (m_positions.size()-1));

                 };
           }; break;
           case kvtChordLength: {
                Scalar totallength = 0;
                for (int i = 0; i < m_positions.size()-1; i++) {
                        totallength += m_controldata.distanceTo(m_positions.at(i).m_vector, m_positions.at(i+1).m_vector);
                }
                Scalar currentpos=0;
                if (totallength == 0) totallength = 1;
                for (int i = 0; i < m_positions.size(); i++) {
                   m_uk[i] = (currentpos/totallength);
                    if (i != m_positions.size()-1) {
                        currentpos += m_controldata.distanceTo(m_positions.at(i).m_vector, m_positions.at(i+1).m_vector);
                    }

                };
           }; break;
           case kvtCentripetal: {
Scalar totallength = 0;
                for (int i = 0; i < m_positions.size()-1; i++) {
                        totallength +=sqrt(m_controldata.distanceTo(m_positions.at(i).m_vector, m_positions.at(i+1).m_vector));
                }
                  if (totallength == 0) totallength = 1;
                Scalar currentpos=0;
                for (int i = 0; i < m_positions.size(); i++) {
                 m_uk[i] = (currentpos/totallength);
                 if (i != m_positions.size()-1) {
                    currentpos += sqrt(m_controldata.distanceTo(m_positions.at(i).m_vector, m_positions.at(i+1).m_vector));
                 }

                 };
           }; break;
       }

       int knotcount = m_positions.size()+m_degree+1;
     Scalar first = 0;
     Scalar last = 1.0;
     int p = m_degree;
     int n = m_positions.size()-1;
      KnotVector* knotvector = new KnotVector(knotcount);
      for (int i = 0; i <= m_degree; i++) {
	knotvector->setEntry(i, first);

	knotvector->setEntry(knotcount-1-i, last);
      };
      for (int j = 1; j <= n-p;j++) {
        UTIFLOAT sum = 0.0;
        for (int i = j; i <= j+p-1; i++) {
            sum += m_uk[i];

        }
        sum = sum / p;
        knotvector->setEntry(j+p, sum);
      }



      knotvector->compile();
      knotvector->print();
      return knotvector;

   }
   T* globalInterpolation(int degree, TKnotVectorType knottype)
   {
       m_degree = degree;
       KnotVector* knotvector = createKnotVector(knottype);
       int p = m_degree;
       int n = m_positions.size()-1;
       UMat<Scalar> A(n+1, n+1);
       Array<UTIFLOAT> N(p+1);

       for (int i = 0; i <=n; i++) {
          for (int j = 0; j <=n; j++) {
              A.set(i,j,0.0);
           }
       }

       for (int i = 0; i <=n; i++) {
           SKnotVectorResult result;
           knotvector->findKnotVectorPosition(m_uk[i], result);
           //result.m_index
           knotvector->calcBasisFunctions(result.m_index, m_uk[i], N, p);
           for (int j = 0; j < N.size(); j++) {
                A.set(i, result.m_index-p+j, N[j]);
           }
       }

       CULUDecomposition<Scalar> decomp(A);

       T* result = new T(knotvector, m_degree);
       Array<Vector> newcontrolpoints;
       newcontrolpoints.setSize(n+1);
        for (int i = 0; i <= n; i++) {
          m_controldata.initZero(newcontrolpoints[i]);
       }

       int dim = m_controldata.getDimensions();
       UMat<Scalar> Q(n+1, 1);
       UMat<Scalar> P;
       for (int d = 0; d < dim; d++)
       {
           for (int i = 0; i <= n; i++) {
             Q.set(i, 0, m_controldata.getVertexIndex(m_positions.at(i).m_vector, d));
           };
           P = decomp.solveSystem(Q);
           for (int i = 0; i <= n; i++) {
             m_controldata.setVertexIndex(newcontrolpoints[i], d, P(i, 0));
           };

       }
       for (int i = 0; i <= n; i++) {
          result->setVertex(i, newcontrolpoints[i], 1.0);
       }

       return result;
   }
   void createTangents(TTangentType tangent)
   {
       m_Tk.setSize(m_positions.size());
       if (tangent == ttGiven) {

            for (int i = 0; i < m_positions.size(); i++) {
                m_Tk[i] = m_positions[i].m_position;
            }
            return;
       }

       int n = m_positions.size()-1;
       m_deltauk.setRange(1, n+1);
       m_qk.setRange(-1, n+3);
       m_ak.setRange(0, n+1);
       m_Dk.setRange(0,n+1);
       m_Vk.setRange(0,n+1);
       m_dk.setRange(1, n+1);
       Array<Vector> m_Q(n+1);
       for (int i = 0; i <= n; i++) {
        m_Q[i] = m_positions[i].m_vector;

       }
       for (int i = 1; i <= n; i++) {

        m_deltauk[i] = m_uk[i]-m_uk[i-1];
        m_qk[i] = m_Q[i]-m_Q[i-1];
        m_dk[i] = m_qk[i]/m_deltauk[i];
       }

        m_qk[0] = 2.0*m_qk[1]-m_qk[2];
        m_qk[-1] = 2.0*m_qk[0]-m_qk[1];
        m_qk[n+1] = 2.0*m_qk[n]-m_qk[n-1];
        m_qk[n+2] = 2.0*m_qk[n+1]-m_qk[n];

        //todo set m_ak

        switch (tangent) {
            case tt3Point: {
                 m_ak[0] = 0.0;
                for (int i = 1; i < n; i++) {
                        m_ak[i] = m_deltauk[i]/(m_deltauk[i]+m_deltauk[i+1]);

                }
                m_ak[n] = 1.0;
                for (int i = 1; i < n; i++) {
                   m_Dk[i] = (1-m_ak[i])*m_dk[i]+m_ak[i]*m_dk[i+1];
                 }
                 m_Dk[0] = 2.0*m_dk[1]-m_Dk[1];
                 m_Dk[n] = 2.0*m_dk[n]-m_Dk[n-1];

                 for (int i = 0; i <= n; i++) {
                    m_Tk[i] = m_Dk[i]/m_controldata.length(m_Dk[i]);
                 }
            }; break;
            case tt4Point: {

                for (int i = 0; i <= n; i++) {
                        Scalar l1 = lengthCrossProduct(m_qk[i-1], m_qk[i]);
                        Scalar denominator = l1+lengthCrossProduct(m_qk[i+1], m_qk[i+2]);
                        if (denominator < 0.000001) {
                            m_ak[i] = 0.5;

                        } else {
                           m_ak[i] = l1 / denominator;
                        }

                }


              for (int i = 0; i <= n; i++) {
                 m_Vk[i] = (1-m_ak[i])*m_qk[i]+(m_ak[i])*m_qk[i+1];
              }
              for (int i = 0; i <= n; i++) {
                  m_Tk[i] = m_Vk[i]/m_controldata.length(m_Vk[i]);
               }

            }; break;
        default: {
            assert(false);
        }; break;
        };




   }
   struct BezierSegment{
      Vector P0;
      Vector P1;
      Vector P2;
      Vector P3;
   };

   T* localCubicInterpolation(TKnotVectorType knottype, TTangentType tangent)
   {
       uti_refptr<KnotVector> knotvector(createKnotVector(knottype));
       createTangents(tangent);
       int n = m_positions.size()-1;

       int m = 8+(n-1)*2-1;

       KnotVector* newknotvector = new KnotVector(m+1);
       for (int i = 0;i <4; i++) {
        newknotvector->setEntry(i, 0);
        newknotvector->setEntry(m-i, 1);
       }
       Array<Scalar> uk(n+1);
       uk[0] = 0.0;
       Array<BezierSegment> beziersegments(n);
       for (int i = 0; i < n; i++) {
           BezierSegment& s = beziersegments[i];
           s.P0 = m_positions[i].m_vector;
           s.P3 = m_positions[i+1].m_vector;
           Vector T0 = m_Tk[i];
           Vector T3 = m_Tk[i+1];
           Scalar a = 16 - sqr(m_controldata.length(T0+ T3));
           Scalar b = 12*m_controldata.dot(s.P3-s.P0, T0+T3);
           Scalar c = -36*sqr(m_controldata.length(s.P3- s.P0));
           Scalar alpha = (-b + sqrt(b*b-4*a*c))/ (2.0*a);
           s.P1 = s.P0 + (alpha/3.0)*T0;
           s.P2 = s.P3 - (alpha/3.0)*T3;
           uk[i+1] = uk[i] + 3.0*(m_controldata.length(s.P1-s.P0));
       }
       for (int i = 1; i < n; i++) {
            newknotvector->setEntry(4 +(i-1)*2, uk[i]/uk[n]);
            newknotvector->setEntry(4 +(i-1)*2+1, uk[i]/uk[n]);

       }
       newknotvector->print();
       newknotvector->compile();
       T* newcurve = new T(newknotvector, 3);
       newcurve->setVertex(0, beziersegments[0].P0, 1.0);
       int index = 1;
       for (int i = 0; i < n; i++) {
           newcurve->setVertex(index, beziersegments[i].P1, 1.0);
           index++;
           newcurve->setVertex(index, beziersegments[i].P2, 1.0);
           index++;
       }
       newcurve->setVertex(index, beziersegments[n-1].P3, 1.0);
       index++;
       utiassert(newcurve->getVertexCount() == index);
       newcurve->print();

       return newcurve;
   }

};

#endif // NURBS_H
