/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __POLYNOM__
#define __POLYNOM__

#include "matrices.h"

template<class T, class DataType> T polyval(const UMat<DataType>& poly, const T& xeval)
{
  CIVector2 size = poly.size();
  T res=getValueMat(poly(0, 0), xeval);
  //Horner Schema
  
  for (int i = 1;i < size.x; i++) {
    
    res = getValueMat(poly(0,i), xeval)+xeval*res;
  };
  
  return res;
  
}

const double UTI_POLYLIMIT= 0;

template<class T> UMat<T> addpoly(const UMat<T>& a, const UMat<T>& b)
{
  int al = a.length();
  int bl = b.length();
  //int newlength;
  UMat<T> res;
  if (al > bl)
  {
    res = UMat<T>(1, al);
    int rest = al-bl;
    for (int i = 0; i < rest; i++) {
      res.set(i,a[i]);
    };
    
    for (int i = rest; i < al; i++) {
      res.set(i,a[i]+b[i-rest]);
    };
    
  } else if (bl > al) {
    res = UMat<T>(1, bl);
    int rest = bl-al;
    for (int i = 0; i < rest; i++) {
      res.set(i,a[i]);
    };
    
    for (int i = rest; i < bl; i++) {
      res.set(i,a[i]+b[i-rest]);
    };
  } else {
    utiassert(al == bl);
    int start = al;
    T zero(UTI_POLYLIMIT);
    for (int i = 0; i < al; i++) {
      if (abs(a[i]+b[i]) > zero) {
	start = i;
      }
    };
    res =UMat<T>(1, al-start);
    for (int i = start; i < al; i++) {
      res.set(i,a[i]+b[i-start]);
    };
    
  };
  return res;
};


template<class T> UMat<T> conv(const UMat<T>& a, const UMat<T>& b)
{
  int al = a.length();
  int bl = b.length();
  int newsize = al+bl;
  UMat<T> res = zeros<T>(1, newsize);
  for (int i = 0; i < al; i++) {
    res.addVector(b, T(0), a[i], 0, i, 0, 0,bl);
  };
  return res;
 
}

template<class T> UMat<T> poly(const UMat<T>& pzeros)
{
  int zerocount = pzeros.size();
   UMat<T> multiply(1, 2);
   multiply.set(0,0,T(1));
   multiply.set(0,1,pzeros[0]);
   UMat<T> res;
   res = multiply;
   for (int i = 1; i < zerocount; i++) {
       multiply.set(0,1,pzeros[i]);
       res = conv(res, multiply);
   };
   
   return res;
};

template<class T> UMat<T> derv(const UMat<T>& a)
{
  int al = a.length();
  if (al <= 1) {
    return UMat<T>(1, 0);
  };
  {
    al = al-1; 
    UMat<T> res(1, al);
     
     for (int i = 0; i < al; i++) {
       res[i] = (al-i-1)*a[i];
     };
    return res;
  };
}

template<class T> struct SSimpleFunction
{
  SSimpleFunction()
  {
    
  };
  virtual ~SSimpleFunction()
  {
    
  }
  virtual T operator()(const T& v)
  {
    return T(0);
  };
  virtual UMat<T> operator()(const UMat<T>& v)
  {
    //CIVector2 s = v.size();
    UMat<T> res(v.size());
    for (int i = 0; i < v.y; i++) {
       for (int j = 0; j < v.x; j++) {
	  res.set(i,j,operator()(v(i,j)));
       }
    };
    return res;
  };
};

template<class T> struct SPolynomFunc
{
  UMat<T> m_mat;
  SPolynomFunc(const UMat<T>& polynom)
  {
    m_mat = polynom;
  };
  T operator()(const T& v)
  {
    return polyeval(m_mat, v);
  };
  UMat<T> operator()(const UMat<T>& v)
  {
    return polyeval(m_mat, v);
  };
};



#endif
