/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __RANDOMNUMBER__
#define __RANDOMNUMBER__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include <cmath>
#include "advmath.h"
#include "utimathbox.h"
#include <cstdlib>
extern const float uti_random_number[4096];
extern const float uti_gradient_2d[16];
extern const float uti_gradient_3d[16*3];

extern int noiseseed;
//http://freespace.virgin.net/hugo.elias/models/m_perlin.htm
inline float intNoise(int x)
{
    //x = ipow((x << 13), x);
    x = x+noiseseed;
    x = (x << 13)*x*x;
//    return  ( (x * (x * x * 15731 +789221)+1376312589)&0x7fffffff) / 1073741824.0 / 2.0;
   return  ( (x * (x * x * 15731 +789221)+1376312589)&0x7fffffff) / 1073741824.0 * 0.5;
}

inline float noise(int x)
{
  return intNoise(x);
}
inline float noise(const CIVector2& p)
{
  return intNoise(p.x+p.y*34522);
}
inline float noise(const CIVector3& p)
{
  return intNoise(p.x+p.y*34522+p.z*5683732);
}
inline CVector2f noise2(const CIVector2& p)
{
  int xv = (p.x) & 63;
  int yv = (p.y) & 63;
  int xv1 = (p.x+46) & 63;
  int yv1 = (p.y+14) & 63;
  return CVector2f(uti_random_number[xv+yv*64], uti_random_number[xv1+yv1*64]);

}

inline CVector2f noise2(const CIVector3& p)
{
  int xv = (p.x+p.z*13) & 63;
  int yv = (p.y+p.z*7) & 63;
  int xv1 = (p.x+46+p.z*27) & 63;
  int yv1 = (p.y+14+p.z*5) & 63;
  return CVector2f(uti_random_number[xv+yv*64], uti_random_number[xv1+yv1*64]);

}
inline CVector3f noise3(const CIVector3& p)
{
  int xv = (p.x+p.z*13) & 63;
  int yv = (p.y+p.z*7) & 63;
  int xv1 = (p.x+46+p.z*27) & 63;
  int yv1 = (p.y+14+p.z*5) & 63;
  int xv2 = (p.x+21+p.z*49) & 63;
  int yv2 = (p.y+37+p.z*23) & 63;
  return CVector3f(uti_random_number[xv+yv*64], uti_random_number[xv1+yv1*64], uti_random_number[xv2+yv2*64]);

}
const int X_PRIME = 1619;
const int Y_PRIME = 31337;
const int Z_PRIME = 6971;
const int W_PRIME = 1013;
inline float gradCoord2D(int x, int y, float xd, float yd)
{
    int hash = noiseseed;
    hash^= X_PRIME*x;
    hash^= Y_PRIME*y;
    hash = hash*hash*hash*60493;
    hash = (hash >> 13)^hash;
    const float* f = &uti_gradient_2d[(hash & 7)*2];

    return xd*f[0]+yd*f[1];

}

inline float gradCoord3D(int x, int y, int z,float xd, float yd, float zd)
{
    int hash = noiseseed;
    hash^= X_PRIME*x;
    hash^= Y_PRIME*y;
    hash^= Z_PRIME*z;
    hash = hash*hash*hash*60493;
    hash = (hash >> 13)^hash;
    const float* f = &uti_gradient_3d[(hash & 15)*3];

    return xd*f[0]+yd*f[1]+zd*f[2];

}

inline int hash3d(int x, int y, int z)
{
    int hash = noiseseed;
    hash^= X_PRIME*x;
    hash^= Y_PRIME*y;
    hash^= Z_PRIME*z;
    hash = hash*hash*hash*60493;
    hash = (hash >> 13)^hash;
    return hash;
}

struct SGradCoord3DOptimized
{
    int m_hash;
    const float *m_f;
    inline void setPosition(int x, int y, int z)
    {
        m_hash = noiseseed;
        m_hash^= X_PRIME*x;
        m_hash^= Y_PRIME*y;
        m_hash^= Z_PRIME*z;
        m_hash = m_hash*m_hash*m_hash*60493;
        m_hash = (m_hash >> 13)^m_hash;
        m_f = &uti_gradient_3d[(m_hash & 15)*3];
    }
    inline float noise(float xd, float yd, float zd)
    {

        return xd*m_f[0]+yd*m_f[1]+zd*m_f[2];
    }
};

inline float gradCoord4D(int x, int y, int z, int w, float xd, float yd, float zd, float wd)
{
    int hash = noiseseed;
    hash^= X_PRIME*x;
    hash^= Y_PRIME*y;
    hash^= Z_PRIME*z;
    hash^= W_PRIME*w;
    hash = hash*hash*hash*60493;
    hash = (hash >> 13)^hash;
    const float* f = &uti_gradient_3d[(hash & 15)*3];

    hash &= 31;
    float a =yd;
    float b = zd;
    float c = wd;
    switch (hash >> 3) {
    case 1:  {
        a = wd;
        b = xd;
        c = yd;
    }; break;
    case 2:  {
        a = zd;
        b = wd;
        c = xd;
    }; break;
    case 3:  {
        a = yd;
        b = zd;
        c = wd;
    }; break;
    }
    return ((hash & 4) == 0 ? -a : a)+((hash & 2) == 0 ? -b : b)+((hash & 1) == 0 ? -c : c);

}

template<class T> inline T randomParam(const CIVector3& v, int shifty, const T& p1)
{
    return p1;
}


template<class T> inline T randomParam(const CIVector3& v, int shifty, const T& p1, const T& p2)
{
    int res = noise(v+CIVector3(0,shifty, 0))*1000;
    switch (res % 3)
    {
    case 0: return p1; case 1: return p2; default: {assert(false);};
    }

    return p1;
}


template<class T> inline T randomParam(const CIVector3& v, int shifty, const T& p1, const T& p2, const T& p3)
{
    int res = noise(v+CIVector3(0,shifty, 0))*1000;
    switch (res % 4)
    {
    case 0: return p1; case 1: return p2;case 2: return p3; default: {assert(false);};
    }

    return p1;
}


template<class T> inline T randomParam(const CIVector3& v, int shifty, const T& p1, const T& p2, const T& p3, const T& p4)
{
    int res = noise(v+CIVector3(0,shifty, 0))*1000;
    switch (res % 4)
    {
    case 0: return p1; case 1: return p2;case 2: return p3;case 3: return p4; default: {assert(false);};
    }

    return p1;
}


template<class T> inline T randomParam(const CIVector3& v, int shifty, const T& p1, const T& p2, const T& p3, const T& p4, const T& p5)
{
    int res = noise(v+CIVector3(0,shifty, 0))*1000;
    switch (res % 5)
    {
    case 0: return p1; case 1: return p2;case 2: return p3;case 3: return p4; case 4: return p5; default: {assert(false);};
    }

    return p1;
}


template<class T> inline T randomParam(const CIVector3& v, int shifty, const T& p1, const T& p2, const T& p3, const T& p4, const T& p5, const T& p6)
{
    int res = noise(v+CIVector3(0,shifty, 0))*1000;
    switch (res % 6)
    {
    case 0: return p1; case 1: return p2;case 2: return p3;case 3: return p4; case 4: return p5; case 5: return p6; default: {assert(false);};
    }

    return p1;
}


template<class T> inline T randomParam(const CIVector3& v, int shifty, const T& p1, const T& p2, const T& p3, const T& p4, const T& p5, const T& p6, const T& p7)
{
    int res = noise(v+CIVector3(0,shifty, 0))*1000;
    switch (res % 7)
    {
    case 0: return p1; case 1: return p2;case 2: return p3;case 3: return p4; case 4: return p5; case 5: return p6; case 6: return p7; default: {assert(false);};
    }

    return p1;
}


template<class T> inline T randomParam(const CIVector3& v, int shifty, const T& p1, const T& p2, const T& p3, const T& p4, const T& p5, const T& p6, const T& p7, const T& p8)
{
    int res = noise(v+CIVector3(0,shifty, 0))*1000;
    switch (res % 8)
    {
    case 0: return p1; case 1: return p2;case 2: return p3;case 3: return p4; case 4: return p5; case 5: return p6; case 6: return p7; case 7: return p8; default: {assert(false);};
    }

    return p1;
}

template<class T> inline float noiseDirect(const T& v)
{
  return noise(vfloori(v));
};

template<class T> struct SInterpolateJitter
{
  static T interpolate(const T& v1, const T& v2, float p)
  {
   float ft = p* M_PI;
  ft = (1-std::cos(ft))*0.5;
  float noise = noiseDirect((v1-v2)*400+p*64.0f);
  T res;
  ft += (noise-0.5f)*0.2f;
  res = umin(1.0f, umax(0.0f,v1*(1-ft)+v2*ft));
  return res;
  };
};

template<class T> inline float noiseInterpolate2D(const CVector2f& pos)
{
  CIVector2 base = vfloori(pos);
  CVector2f fraction = vfrac(pos);
  float v1 = T::interpolate(noise(base), noise(CIVector2(base.x+1, base.y)), fraction.x);
  float v2 = T::interpolate(noise(CIVector2(base.x, base.y+1)), noise(CIVector2(base.x+1, base.y+1)), fraction.x);
  return T::interpolate(v1, v2, fraction.y);
};
template<class T> inline float noiseInterpolate2D(const CVector2& pos)
{
  CIVector2 base = vfloori(pos);
  CVector2 fraction = vfrac(pos);
  float v1 = T::interpolate(noise(base), noise(CIVector2(base.x+1, base.y)), fraction.x);
  float v2 = T::interpolate(noise(CIVector2(base.x, base.y+1)), noise(CIVector2(base.x+1, base.y+1)), fraction.x);
  return T::interpolate(v1, v2, fraction.y);
};

inline int utimod(int value, int modul) {
  if (modul <= 0) return value;
  if (value < 0) {
    return (value+(value/modul+1)*modul)% modul;
  } else {
    return value % modul;
  };
};

template<class T> inline float noiseInterpolate2D(const CVector2& pos, const CIVector2& repeat)
{
  CIVector2 base = vfloori(pos);
  CVector2 fraction = vfrac(pos);
  float v1 = T::interpolate(noise(CIVector2(utimod(base.x, repeat.x), utimod(base.y, repeat.y))), noise(CIVector2(utimod(base.x+1, repeat.x), utimod(base.y, repeat.y))), fraction.x);
  float v2 = T::interpolate(noise(CIVector2(utimod(base.x, repeat.x), utimod(base.y+1, repeat.y))), noise(CIVector2(utimod(base.x+1, repeat.x), utimod(base.y+1, repeat.y))), fraction.x);
  return T::interpolate(v1, v2, fraction.y);
};

template<class T> inline float noiseInterpolate2DL(const CVector2& pos, const CIVector2& repeat, int layer)
{
  CIVector2 base = vfloori(pos);
  CVector2 fraction = vfrac(pos);
  float v1 = T::interpolate(noise(CIVector3(utimod(base.x, repeat.x), utimod(base.y, repeat.y), layer)), noise(CIVector3(utimod(base.x+1, repeat.x), utimod(base.y, repeat.y), layer)), fraction.x);
  float v2 = T::interpolate(noise(CIVector3(utimod(base.x, repeat.x), utimod(base.y+1, repeat.y), layer)), noise(CIVector3(utimod(base.x+1, repeat.x), utimod(base.y+1, repeat.y), layer)), fraction.x);
  return T::interpolate(v1, v2, fraction.y);
};


template<class T> inline float noiseInterpolate2D(const CVector2f& pos, const CIVector2& repeat)
{
  CIVector2 base = vfloori(pos);
  CVector2f fraction = vfrac(pos);
  float v1 = T::interpolate(noise(CIVector2(utimod(base.x, repeat.x), utimod(base.y, repeat.y))), noise(CIVector2(utimod(base.x+1, repeat.x), utimod(base.y, repeat.y))), fraction.x);
  float v2 = T::interpolate(noise(CIVector2(utimod(base.x, repeat.x), utimod(base.y+1, repeat.y))), noise(CIVector2(utimod(base.x+1, repeat.x), utimod(base.y+1, repeat.y))), fraction.x);
  return T::interpolate(v1, v2, fraction.y);
};

template<class T> inline float noiseInterpolate2DL(const CVector2f& pos, const CIVector2& repeat, int layer)
{
  CIVector2 base = vfloori(pos);
  CVector2f fraction = vfrac(pos);
  float v1 = T::interpolate(noise(CIVector3(utimod(base.x, repeat.x), utimod(base.y, repeat.y), layer)), noise(CIVector3(utimod(base.x+1, repeat.x), utimod(base.y, repeat.y), layer)), fraction.x);
  float v2 = T::interpolate(noise(CIVector3(utimod(base.x, repeat.x), utimod(base.y+1, repeat.y), layer)), noise(CIVector3(utimod(base.x+1, repeat.x), utimod(base.y+1, repeat.y), layer)), fraction.x);
  return T::interpolate(v1, v2, fraction.y);
};



template<class T> struct SNoiseInterpolate2DOptimized
{
    float m_noise[4];
    CVector2f m_last;
    CIVector2 m_basevalue;
    CIVector2 m_repeat;
    int m_layer;

    float v1;
    float v2;
    inline SNoiseInterpolate2DOptimized()
    {
        m_basevalue.set(-1000000,-1000000);
        m_last.setVector(-1E+10, -1E+10);
        m_repeat.set(-1000000, -1000000);
        m_layer = 0;
    }
    inline void setLayer(int l)
    {
        if (l != m_layer) {
            m_basevalue.set(-1000000,-1000000);
            m_layer = l;
        }
    }

    float calcValue(const CVector2f& pos, const CIVector2& repeat)
    {
        float result;
        CIVector2 base;
        CVector2f fraction;
        fraction.x = fasterfloor(pos.x);
        fraction.y = fasterfloor(pos.y);
      //  fraction.z = fasterfloor(pos.z);

        base.x = fraction.x;
        base.y = fraction.y;
      //  base.z = fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
       // fraction.z = pos.z - fraction.z;


        if (base.x != m_basevalue.x || base.y != m_basevalue.y || repeat.x != m_repeat.x || repeat.y != m_repeat.y) {
            m_basevalue = base;
            m_repeat = repeat;
            m_noise[0] = noise(CIVector3(utimod(base.x, repeat.x), utimod(base.y, repeat.y), m_layer));
            m_noise[1] = noise(CIVector3(utimod(base.x+1, repeat.x), utimod(base.y, repeat.y), m_layer));
            m_noise[2] = noise(CIVector3(utimod(base.x, repeat.x), utimod(base.y+1, repeat.y), m_layer));
            m_noise[3] = noise(CIVector3(utimod(base.x+1, repeat.x), utimod(base.y+1, repeat.y), m_layer));

            m_last.y = 1E+10;
        }
        if (m_last.y != pos.y) {
          v1 = T::interpolate(m_noise[0], m_noise[2], fraction.y);
          v2 = T::interpolate(m_noise[1], m_noise[3], fraction.y);
          m_last.y = pos.y;
        }
        //m_last.y = pos.y //nicht nötig
        result = T::interpolate(v1, v2, fraction.x);
       // assert(absf(result-noiseInterpolate2DL<T>(pos, repeat, m_layer)) < 0.001);
        return result;
    }
};

template<class T> inline float noiseInterpolate3D(const CVector3f& pos)
{
  CIVector3 base = vfloori(pos);
  CVector3f fraction = vfrac(pos);
  float v1_1 = T::interpolate(noise(base), noise(CIVector3(base.x, base.y, base.z+1)), fraction.z);
  float v1_2 = T::interpolate(noise(CIVector3(base.x, base.y+1, base.z)), noise(CIVector3(base.x, base.y+1, base.z+1)), fraction.z);
  float v2_1 = T::interpolate(noise(CIVector3(base.x+1, base.y, base.z)), noise(CIVector3(base.x+1, base.y, base.z+1)), fraction.z);
  float v2_2 = T::interpolate(noise(CIVector3(base.x+1, base.y+1, base.z)), noise(CIVector3(base.x+1, base.y+1, base.z+1)), fraction.z);

  float v1 = T::interpolate(v1_1, v2_1, fraction.x);
  float v2 = T::interpolate(v1_2, v2_2, fraction.x);
  return T::interpolate(v1, v2, fraction.y);
};

template<class T, bool normalized=true> inline float noiseInterpolate3DVector(const CVector3f& pos)
{
  CIVector3 base = vfloori(pos);
  CVector3f fraction = vfrac(pos);
  CVector4f a,b;
  CVector2f c;
  a.x = noise(base);
  a.y = noise(CIVector3(base.x, base.y+1, base.z));
  a.z = noise(CIVector3(base.x+1, base.y, base.z));
  a.w = noise(CIVector3(base.x+1, base.y+1, base.z));

  b.x = noise(CIVector3(base.x, base.y, base.z+1));
  b.y = noise(CIVector3(base.x, base.y+1, base.z+1));
  b.z = noise(CIVector3(base.x+1, base.y, base.z+1));
  b.w = noise(CIVector3(base.x+1, base.y+1, base.z+1));

  a = T::interpolate(a,b,fraction.z);
  c = T::interpolate(a.xy(), a.zw(), fraction.x);

/*
  float v1_1 = T::interpolate(noise(base), noise(CIVector3(base.x, base.y, base.z+1)), fraction.z);
  float v1_2 = T::interpolate(noise(CIVector3(base.x, base.y+1, base.z)), noise(CIVector3(base.x, base.y+1, base.z+1)), fraction.z);
  float v2_1 = T::interpolate(noise(CIVector3(base.x+1, base.y, base.z)), noise(CIVector3(base.x+1, base.y, base.z+1)), fraction.z);
  float v2_2 = T::interpolate(noise(CIVector3(base.x+1, base.y+1, base.z)), noise(CIVector3(base.x+1, base.y+1, base.z+1)), fraction.z);

  float v1 = T::interpolate(v1_1, v2_1, fraction.x);
  float v2 = T::interpolate(v1_2, v2_2, fraction.x);*/
 // return T::interpolate(c.x, c.y, fraction.y);
  if (!normalized) {
     return T::interpolate(c.x, c.y, fraction.y)*2.0f-1.0f;
  } else {

      return T::interpolate(c.x, c.y, fraction.y);
  }
};


template<class T, bool normalized=true> inline float perlinInterpolate3DVector(const CVector3f& pos)
{
  CIVector3 base0 = vfloori(pos);
  CIVector3 base1 = base0 + CIVector3(1,1,1);
  CVector3f fraction0 = vfrac(pos);
  CVector3f fraction1 = fraction0-CVector3f(1,1,1);
  CVector4f a,b;
  CVector2f c;
  a.x = gradCoord3D(base0.x, base0.y, base0.z, fraction0.x, fraction0.y, fraction0.z);
  a.y = gradCoord3D(base0.x, base1.y, base0.z, fraction0.x, fraction1.y, fraction0.z);//noise(CIVector3(base.x, base.y+1, base.z));
  a.z = gradCoord3D(base1.x, base0.y, base0.z, fraction1.x, fraction0.y, fraction0.z);//noise(CIVector3(base.x+1, base.y, base.z));
  a.w = gradCoord3D(base1.x, base1.y, base0.z, fraction1.x, fraction1.y, fraction0.z);//noise(CIVector3(base.x+1, base.y+1, base.z));

  b.x = gradCoord3D(base0.x, base0.y, base1.z, fraction0.x, fraction0.y, fraction1.z);//noise(CIVector3(base.x, base.y, base.z+1));
  b.y = gradCoord3D(base0.x, base1.y, base1.z, fraction0.x, fraction1.y, fraction1.z);//noise(CIVector3(base.x, base.y+1, base.z+1));
  b.z = gradCoord3D(base1.x, base0.y, base1.z, fraction1.x, fraction0.y, fraction1.z);//noise(CIVector3(base.x+1, base.y, base.z+1));
  b.w = gradCoord3D(base1.x, base1.y, base1.z, fraction1.x, fraction1.y, fraction1.z);//noise(CIVector3(base.x+1, base.y+1, base.z+1));

  a = T::interpolate(a,b,fraction0.z);
  c = T::interpolate(a.xy(), a.zw(), fraction0.x);

/*
  float v1_1 = T::interpolate(noise(base), noise(CIVector3(base.x, base.y, base.z+1)), fraction.z);
  float v1_2 = T::interpolate(noise(CIVector3(base.x, base.y+1, base.z)), noise(CIVector3(base.x, base.y+1, base.z+1)), fraction.z);
  float v2_1 = T::interpolate(noise(CIVector3(base.x+1, base.y, base.z)), noise(CIVector3(base.x+1, base.y, base.z+1)), fraction.z);
  float v2_2 = T::interpolate(noise(CIVector3(base.x+1, base.y+1, base.z)), noise(CIVector3(base.x+1, base.y+1, base.z+1)), fraction.z);

  float v1 = T::interpolate(v1_1, v2_1, fraction.x);
  float v2 = T::interpolate(v1_2, v2_2, fraction.x);*/

  if (normalized) {
     return( T::interpolate(c.x, c.y, fraction0.y)+1.0f)/2.0f;
  } else {

      return T::interpolate(c.x, c.y, fraction0.y);
  }
};

template<class T, bool normalized=true> inline float simplexInterpolate3DVector(const CVector3f& pos)
{
   const float F3 = (1.0f/ 3.0f);
   const float G3 = (1.0f/ 6.0f);
   const float G33 = G3*3 -1;

   float t = (pos.x+pos.y+pos.z)*F3;
   int i = fasterfloor(pos.x+t);
   int j = fasterfloor(pos.y+t);
   int k = fasterfloor(pos.z+t);

   t = float(i+j+k)*G3;
   float x0 = pos.x- (i-t);
   float y0 = pos.y- (j-t);
   float z0 = pos.z- (k-t);

   int i1,j1,k1;
   int i2,j2,k2;
   if (x0 >= y0) {
       if (y0 >= z0) {
          i1 = 1;
          j1 = 0;
          k1 = 0;
          i2 = 1;
          j2 = 1;
          k2 = 0;
       } else if (x0 >= z0) {
           i1 = 1;
           j1 = 0;
           k1 = 0;
           i2 = 1;
           j2 = 0;
           k2 = 1;
       } else { // x0 < z0
           i1 = 0;
           j1 = 0;
           k1 = 1;
           i2 = 1;
           j2 = 0;
           k2 = 1;
       }

   } else {//x0 < y0
       if (y0 >= z0) {
          i1 = 0;
          j1 = 0;
          k1 = 1;
          i2 = 0;
          j2 = 1;
          k2 = 1;
       } else if (x0 >= z0) {
           i1 = 0;
           j1 = 1;
           k1 = 0;
           i2 = 0;
           j2 = 1;
           k2 = 1;
       } else { // x0 < z0
           i1 = 0;
           j1 = 1;
           k1 = 0;
           i2 = 1;
           j2 = 1;
           k2 = 0;
       }
   }
   float x1 = x0 -i1 + G3;
   float y1 = y0 -j1 + G3;
   float z1 = z0 -k1 + G3;
   float x2 = x0 -i2 + F3;
   float y2 = y0 -i2 + F3;
   float z2 = z0 -i2 + F3;
   float x3 = x0  + G33;
   float y3 = y0  + G33;
   float z3 = z0  + G33;

   float n0,n1,n2, n3;

   t = (float)0.6-x0*x0-y0*y0-z0*z0;
   if (t < 0) {
       n0=0;
   } else {
       t *= t;
       n0 = t*t*gradCoord3D(i,j,k,x0,y0,z0);
   }

   t = (float)0.6-x1*x1-y1*y1-z1*z1;
   if (t < 0) {
       n1=0;
   } else {
       t *= t;
       n1 = t*t*gradCoord3D(i+i1,j+i1,k+k1,x1,y1,z1);
   }

   t = (float)0.6-x2*x2-y2*y2-z2*z2;
   if (t < 0) {
       n2=0;
   } else {
       t *= t;
       n2 = t*t*gradCoord3D(i+i2,j+i2,k+k2,x2,y2,z2);
   }

   t = (float)0.6-x3*x3-y3*y3-z3*z3;
   if (t < 0) {
       n3=0;
   } else {
       t *= t;
       n3 = t*t*gradCoord3D(i+1,j+1,k+1,x3,y3,z3);
   }

   if (normalized) {

       return (32*(n0+n1+n2+n3)+1.0f)/2.0f;
   } else {
   return 32*(n0+n1+n2+n3);
   }
};

template<class T, bool normalized=true> inline float cubicInterpolate3DVector(const CVector3f& pos)
{
    const float CUBIC_3D = 1.0f/(1.5f*1.5f*1.5f);
    CIVector3 v1 = vfloori(pos);

    CIVector3 v0 = v1 + CIVector3(-1,-1,-1);
    CIVector3 v2 = v1 + CIVector3(1,1,1);
    CIVector3 v3 = v1 + CIVector3(2,2,2);
    CVector3f vs = vfrac(pos);
    float xs = vs.x;
    float ys = vs.y;
    float zs = vs.z;
    float res= cubicLerp(
                cubicLerp(
                    cubicLerp(noise(CIVector3(v0.x, v0.y, v0.z)), noise(CIVector3(v1.x, v0.y, v0.z)),noise(CIVector3(v2.x, v0.y, v0.z)),noise(CIVector3(v3.x, v0.y, v0.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v1.y, v0.z)), noise(CIVector3(v1.x, v1.y, v0.z)),noise(CIVector3(v2.x, v1.y, v0.z)),noise(CIVector3(v3.x, v1.y, v0.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v2.y, v0.z)), noise(CIVector3(v1.x, v2.y, v0.z)),noise(CIVector3(v2.x, v2.y, v0.z)),noise(CIVector3(v3.x, v2.y, v0.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v3.y, v0.z)), noise(CIVector3(v1.x, v3.y, v0.z)),noise(CIVector3(v2.x, v3.y, v0.z)),noise(CIVector3(v3.x, v3.y, v0.z)),xs),
                         ys),
                cubicLerp(
                    cubicLerp(noise(CIVector3(v0.x, v0.y, v1.z)),noise( CIVector3(v1.x, v0.y, v1.z)),noise(CIVector3(v2.x, v0.y, v1.z)),noise(CIVector3(v3.x, v0.y, v1.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v1.y, v1.z)), noise(CIVector3(v1.x, v1.y, v1.z)),noise(CIVector3(v2.x, v1.y, v1.z)),noise(CIVector3(v3.x, v1.y, v1.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v2.y, v1.z)), noise(CIVector3(v1.x, v2.y, v1.z)),noise(CIVector3(v2.x, v2.y, v1.z)),noise(CIVector3(v3.x, v2.y, v1.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v3.y, v1.z)), noise(CIVector3(v1.x, v3.y, v1.z)),noise(CIVector3(v2.x, v3.y, v1.z)),noise(CIVector3(v3.x, v3.y, v1.z)),xs),
                         ys),
                cubicLerp(
                    cubicLerp(noise(CIVector3(v0.x, v0.y, v2.z)), noise(CIVector3(v1.x, v0.y, v2.z)),noise(CIVector3(v2.x, v0.y, v2.z)),noise(CIVector3(v3.x, v0.y, v2.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v1.y, v2.z)),noise( CIVector3(v1.x, v1.y, v2.z)),noise(CIVector3(v2.x, v1.y, v2.z)),noise(CIVector3(v3.x, v1.y, v2.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v2.y, v2.z)), noise(CIVector3(v1.x, v2.y, v2.z)),noise(CIVector3(v2.x, v2.y, v2.z)),noise(CIVector3(v3.x, v2.y, v2.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v3.y, v2.z)), noise(CIVector3(v1.x, v3.y, v2.z)),noise(CIVector3(v2.x, v3.y, v2.z)),noise(CIVector3(v3.x, v3.y, v2.z)),xs),
                         ys),
                cubicLerp(
                    cubicLerp(noise(CIVector3(v0.x, v0.y, v3.z)), noise(CIVector3(v1.x, v0.y, v3.z)),noise(CIVector3(v2.x, v0.y, v3.z)),noise(CIVector3(v3.x, v0.y, v3.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v1.y, v3.z)), noise(CIVector3(v1.x, v1.y, v3.z)),noise(CIVector3(v2.x, v1.y, v3.z)),noise(CIVector3(v3.x, v1.y, v3.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v2.y, v3.z)), noise(CIVector3(v1.x, v2.y, v3.z)),noise(CIVector3(v2.x, v2.y, v3.z)),noise(CIVector3(v3.x, v2.y, v3.z)),xs),
                    cubicLerp(noise(CIVector3(v0.x, v3.y, v3.z)), noise(CIVector3(v1.x, v3.y, v3.z)),noise(CIVector3(v2.x, v3.y, v3.z)),noise(CIVector3(v3.x, v3.y, v3.z)),xs),
                         ys)
                ,zs);
    if (normalized) {
        return res;
    } else {
        return res*2.0f-1.0f;
    }
}

struct RandomNumberGenerator
{
	CIVector3 m_position;
	int m_index;
	inline RandomNumberGenerator()
	{
		init(CIVector3(0,0,0));
	}
		inline void init(const CIVector3& pos)
	{
		m_position = pos;
		m_index = 0;

	}
        inline void initViaSeed(int randseed)
        {
            init(CIVector3((randseed >> 16)&255, (randseed >> 8)&255, randseed & 255));
        }

		inline float getRandomNumber(int index) const {
			return noise(CIVector3(m_position.x + (index % 1024), m_position.y + (index / 1024), m_position.z));
				
		}
		inline float random()
		{
			float r = getRandomNumber(m_index);
			m_index++;
			return r;
		}
		inline int random(int m) {
			assert(m != 0);
			return int(round(random()*m)) % m;
		}

                inline int randomOrZero(int m) {
                        if (m == 0)return 0;
                        return int(round(random()*m)) % m;
                }
};

template<class T> struct SNoiseInterpolate3DOptimized
{
   // float m_noise[8];
    CVector4f m_noise_a, m_noise_b, m_noise_res;// m_noise[8];
    CVector2f m_c;
    CVector3f m_last;
    CIVector3 m_basevalue;
   /* float v1_1;
    float v1_2;
    float v2_1;
    float v2_2;

    float v1;
    float v2;*/
    inline SNoiseInterpolate3DOptimized()
    {
        m_basevalue.set(-1000000,-1000000, -1000000);
        m_last.setVector(-1E+10, -1E+10,-1E+10);
    }

    inline float calcValue(const CVector3f& pos)
    {
        float result;
        CIVector3 base;
        CVector3f fraction;
        fraction.x = (float)fasterfloor(pos.x);
        fraction.y = (float)fasterfloor(pos.y);
        fraction.z = (float)fasterfloor(pos.z);

        base.x = (int)fraction.x;
        base.y = (int)fraction.y;
        base.z =(int) fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
        fraction.z = pos.z - fraction.z;

        if (base.x != m_basevalue.x || base.y != m_basevalue.y || base.z != m_basevalue.z) {
            m_basevalue = base;
            m_noise_a.x = noise(base);
            m_noise_b.x = noise(CIVector3(base.x, base.y, base.z+1));
            m_noise_a.y = noise(CIVector3(base.x, base.y+1, base.z));
            m_noise_b.y = noise(CIVector3(base.x, base.y+1, base.z+1));


            m_noise_a.z = noise(CIVector3(base.x+1, base.y, base.z));
            m_noise_b.z = noise(CIVector3(base.x+1, base.y, base.z+1));
            m_noise_a.w = noise(CIVector3(base.x+1, base.y+1, base.z));
            m_noise_b.w = noise(CIVector3(base.x+1, base.y+1, base.z+1));
            m_last.z = 1E+10;
        }

        if (m_last.z != pos.z) {

            m_noise_res = T::interpolate(m_noise_a, m_noise_b, fraction.z);
       /* v1_1 = T::interpolate(m_noise[0], m_noise[1], fraction.z);
        v1_2 = T::interpolate(m_noise[2], m_noise[3], fraction.z);
        v2_1 = T::interpolate(m_noise[4], m_noise[5], fraction.z);
        v2_2 = T::interpolate(m_noise[6], m_noise[7], fraction.z);*/
            m_c = T::interpolate(m_noise_res.xy(), m_noise_res.zw(), fraction.x);

          m_last.z = pos.z;
          m_last.x = pos.x;
        } else if (m_last.x != pos.x) {

            m_c = T::interpolate(m_noise_res.xy(), m_noise_res.zw(), fraction.x);
            //v1 = T::interpolate(v1_1, v2_1, fraction.x);
           // v2 = T::interpolate(v1_2, v2_2, fraction.x);
            m_last.x = pos.x;
        }
        //m_last.y = pos.y //nicht nötig
        result = T::interpolate(m_c.x, m_c.y, fraction.y);
        assert(absf(result-noiseInterpolate3D<T>(pos)) < 0.001);
        return result;
    }
   /* float calcValueOld(const CVector3f& pos)
    {
        float result;
        CIVector3 base;
        CVector3f fraction;
        fraction.x = (float)fasterfloor(pos.x);
        fraction.y = (float)fasterfloor(pos.y);
        fraction.z = (float)fasterfloor(pos.z);

        base.x = (int)fraction.x;
        base.y = (int)fraction.y;
        base.z =(int) fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
        fraction.z = pos.z - fraction.z;


        //Wurde noch schneller gehen
        if (base.x != m_basevalue.x || base.y != m_basevalue.y || base.z != m_basevalue.z) {
            m_basevalue = base;
            m_noise[0] = noise(base);
            m_noise[1] = noise(CIVector3(base.x, base.y, base.z+1));
            m_noise[2] = noise(CIVector3(base.x, base.y+1, base.z));
            m_noise[3] = noise(CIVector3(base.x, base.y+1, base.z+1));


            m_noise[4] = noise(CIVector3(base.x+1, base.y, base.z));
            m_noise[5] = noise(CIVector3(base.x+1, base.y, base.z+1));
            m_noise[6] = noise(CIVector3(base.x+1, base.y+1, base.z));
            m_noise[7] = noise(CIVector3(base.x+1, base.y+1, base.z+1));
            m_last.z = 1E+10;
        }

        if (m_last.z != pos.z) {

        v1_1 = T::interpolate(m_noise[0], m_noise[1], fraction.z);
        v1_2 = T::interpolate(m_noise[2], m_noise[3], fraction.z);
        v2_1 = T::interpolate(m_noise[4], m_noise[5], fraction.z);
        v2_2 = T::interpolate(m_noise[6], m_noise[7], fraction.z);

        v1 = T::interpolate(v1_1, v2_1, fraction.x);
        v2 = T::interpolate(v1_2, v2_2, fraction.x);
          m_last.z = pos.z;
          m_last.x = pos.x;
        } else if (m_last.x != pos.x) {
            v1 = T::interpolate(v1_1, v2_1, fraction.x);
            v2 = T::interpolate(v1_2, v2_2, fraction.x);
            m_last.x = pos.x;
        }
        //m_last.y = pos.y //nicht nötig
        result = T::interpolate(v1, v2, fraction.y);
       // assert(absf(result-noiseInterpolate3D<T>(pos)) < 0.001);
        return result;
    }
*/
    inline float calcValueAsynchron(const CVector3f& pos)
    {
        return noiseInterpolate3DVector<T>(pos);
       /* float result;
        CIVector3 base;
        CVector3f fraction;
        fraction.x = (float)fasterfloor(pos.x);
        fraction.y = (float)fasterfloor(pos.y);
        fraction.z = (float)fasterfloor(pos.z);

        base.x = (int)fraction.x;
        base.y = (int)fraction.y;
        base.z =(int) fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
        fraction.z = pos.z - fraction.z;



            m_noise[0] = noise(base);
            m_noise[1] = noise(CIVector3(base.x, base.y, base.z+1));
            m_noise[2] = noise(CIVector3(base.x, base.y+1, base.z));
            m_noise[3] = noise(CIVector3(base.x, base.y+1, base.z+1));


            m_noise[4] = noise(CIVector3(base.x+1, base.y, base.z));
            m_noise[5] = noise(CIVector3(base.x+1, base.y, base.z+1));
            m_noise[6] = noise(CIVector3(base.x+1, base.y+1, base.z));
            m_noise[7] = noise(CIVector3(base.x+1, base.y+1, base.z+1));
            m_last.z = 1E+10;



        v1_1 = T::interpolate(m_noise[0], m_noise[1], fraction.z);
        v1_2 = T::interpolate(m_noise[2], m_noise[3], fraction.z);
        v2_1 = T::interpolate(m_noise[4], m_noise[5], fraction.z);
        v2_2 = T::interpolate(m_noise[6], m_noise[7], fraction.z);

        v1 = T::interpolate(v1_1, v2_1, fraction.x);
        v2 = T::interpolate(v1_2, v2_2, fraction.x);
          m_last.z = pos.z;
          m_last.x = pos.x;


        result = T::interpolate(v1, v2, fraction.y);
        return result;*/
    }
};

template<class T> struct SPerlinInterpolate3DOptimized
{
    SGradCoord3DOptimized m_noisesource[8];
    CVector3f m_last;
    CIVector3 m_basevalue;
    CVector4f m_noise_a, m_noise_b, m_noise_res;// m_noise[8];
    CVector2f m_c;

    inline SPerlinInterpolate3DOptimized()
    {
        m_basevalue.set(-1000000,-1000000, -1000000);
        m_last.setVector(-1E+10, -1E+10,-1E+10);
    }

    inline float calcValue(const CVector3f& pos)
    {
        //Trick does not work?
        float result;
        CIVector3 base;
        CVector3f fraction;
        CVector3f fraction1;
        fraction.x = (float)fasterfloor(pos.x);
        fraction.y = (float)fasterfloor(pos.y);
        fraction.z = (float)fasterfloor(pos.z);

        base.x = (int)fraction.x;
        base.y = (int)fraction.y;
        base.z =(int) fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
        fraction.z = pos.z - fraction.z;

        fraction1.x = fraction.x-1.0f;
        fraction1.y = fraction.y-1.0f;
        fraction1.z = fraction.z-1.0f;


        //Wurde noch schneller gehen
        if (base.x != m_basevalue.x || base.y != m_basevalue.y || base.z != m_basevalue.z) {
            m_basevalue = base;
            m_noisesource[0].setPosition(base.x, base.y, base.z);
            m_noisesource[1].setPosition(base.x, base.y, base.z+1);
            m_noisesource[2].setPosition(base.x, base.y+1, base.z);
            m_noisesource[3].setPosition(base.x, base.y+1, base.z+1);


            m_noisesource[4].setPosition(base.x+1, base.y, base.z);
            m_noisesource[5].setPosition(base.x+1, base.y, base.z+1);
            m_noisesource[6].setPosition(base.x+1, base.y+1, base.z);
            m_noisesource[7].setPosition(base.x+1, base.y+1, base.z+1);
            m_last.z = 1E+10;
        }
        m_noise_a.x = m_noisesource[0].noise(fraction.x, fraction.y, fraction.z);
        m_noise_b.x = m_noisesource[1].noise(fraction.x, fraction.y, fraction1.z);//noise(CIVector3(base.x, base.y, base.z+1));
        m_noise_a.y = m_noisesource[2].noise(fraction.x, fraction1.y, fraction.z);//noise(CIVector3(base.x, base.y+1, base.z));
        m_noise_b.y = m_noisesource[3].noise(fraction.x, fraction1.y, fraction1.z);//noise(CIVector3(base.x, base.y+1, base.z+1));


        m_noise_a.z = m_noisesource[4].noise(fraction1.x, fraction.y, fraction.z);//noise(CIVector3(base.x+1, base.y, base.z));
        m_noise_b.z = m_noisesource[5].noise(fraction1.x, fraction.y, fraction1.z);//noise(CIVector3(base.x+1, base.y, base.z+1));
        m_noise_a.w = m_noisesource[6].noise(fraction1.x, fraction1.y, fraction.z);//noise(CIVector3(base.x+1, base.y+1, base.z));
        m_noise_b.w = m_noisesource[7].noise(fraction1.x, fraction1.y, fraction1.z);//noise(CIVector3(base.x+1, base.y+1, base.z+1));


        m_noise_res = T::interpolate(m_noise_a, m_noise_b, fraction.z);

        m_c = T::interpolate(m_noise_res.xy(), m_noise_res.zy(), fraction.x);

          m_last.z = pos.z;
          m_last.x = pos.x;

        result = T::interpolate(m_c.x, m_c.y, fraction.y);
        assert(absf(result-perlinInterpolate3DVector<T>(pos)) < 0.001);
        return result;
    }

    inline float calcValueAsynchron(const CVector3f& pos)
    {

        return perlinInterpolate3DVector<T>(pos);
    }
};


struct SPerlinInfo
{
  float m_amplitude;
  float m_persistence;
  float m_lucunairty;
  float m_scal;
  int m_octaves;
  inline SPerlinInfo() {
    m_amplitude = 0.5f;
    m_persistence=0.5f;
    m_lucunairty = 2.0f;
    m_scal = 1.0f;

    m_octaves= 4;
  };
  inline SPerlinInfo(float a, float p, int o) {
    m_amplitude = a;
    m_persistence=p;
    m_lucunairty = 0.5f;
    m_scal = 1.0f;
    m_octaves= o;
  };
 template<class T, typename CB> inline  float getFBM(const T& pos, CB& n)
  {
       float sum = n(pos);
       float amp = 1.0f;

       T p = pos;
       for (int i = 1; i < m_octaves; i++) {
           p *= m_lucunairty;
           amp *= m_persistence;
           sum += n(p)*amp;
       }
       return sum* m_scal;
  }
  template<class T, typename CB>inline  float getBillow(const T& pos, CB& n)
  {
       float sum = absf(n(pos))*2.0f-1.0f;
       float amp = 1.0f;

       T p = pos;
       for (int i = 1; i < m_octaves; i++) {
           p *= m_lucunairty;
           amp *= m_persistence;
           sum += (absf(n(pos))*2.0f-1.0f)*amp;
       }
       return sum* m_scal;
  }
  template<class T, typename CB> inline float getRigidMulti(const T& pos, CB& n)
  {
       float sum = 1.0f-absf(n(pos));
       float amp = 1.0f;

       T p = pos;
       for (int i = 1; i < m_octaves; i++) {
           p *= m_lucunairty;
           amp *= m_persistence;
           sum -=  (1.0f-absf(n(pos)))*amp;
       }
       return sum* m_scal;
  }
  template<class T, typename CB> inline float getRigidMultiScal2(const T& pos, CB& n)
  {
       float sum = 1.0f-absf(n(pos)*2.0f-1.0f);
       float amp = 1.0f;

       T p = pos;
       for (int i = 1; i < m_octaves; i++) {
           p *= m_lucunairty;
           amp *= m_persistence;
           sum -=  (1.0f-absf(n(pos)*2.0f-1.0f))*amp;
       }
       return sum* m_scal;
  }
};



template<class T> inline void gradientTurbulence(float perturbAmb, float frequency, CVector3f& v3)
{
    float xf = v3.x * frequency;
    float yf = v3.y * frequency;
    float zf = v3.z * frequency;

    int x0 = fasterfloor(xf);
    int y0 = fasterfloor(yf);
    int z0 = fasterfloor(zf);

    int x1 = x0+1;
    int y1 = y0+1;
    int z1 = z0+1;

    float xs = xf-x0, ys=yf-y0, zs=zf-z0;

    const float* vec0 = &uti_random_number[(hash3d(x0,y0,z0) & 255)*3];
    const float* vec1 = &uti_random_number[(hash3d(x1,y0,z0) & 255)*3];
    CVector3f l0x = T::interpolate(CVector3f(vec0[0], vec0[1],vec0[2]), CVector3f(vec1[0], vec1[1],vec1[2]), xs);

    vec0 = &uti_random_number[(hash3d(x0,y1,z0) & 255)*3];
    vec1 = &uti_random_number[(hash3d(x1,y1,z0) & 255)*3];
    CVector3f l1x = T::interpolate(CVector3f(vec0[0], vec0[1],vec0[2]), CVector3f(vec1[0], vec1[1],vec1[2]), xs);

    CVector3f l0y =  T::interpolate(l0x, l1x, ys);

    vec0 = &uti_random_number[(hash3d(x0,y0,z1) & 255)*3];
    vec1 = &uti_random_number[(hash3d(x1,y0,z1) & 255)*3];
     l0x = T::interpolate(CVector3f(vec0[0], vec0[1],vec0[2]), CVector3f(vec1[0], vec1[1],vec1[2]), xs);

    vec0 = &uti_random_number[(hash3d(x0,y1,z1) & 255)*3];
    vec1 = &uti_random_number[(hash3d(x1,y1,z1) & 255)*3];
    l1x = T::interpolate(CVector3f(vec0[0], vec0[1],vec0[2]), CVector3f(vec1[0], vec1[1],vec1[2]), xs);

    CVector3f l1y =  T::interpolate(l0x, l1x, ys);

    CVector3f shift = T::interpolate(l1x, l1y, zs);
    v3 += shift* perturbAmb;
}

template<class T> struct SGradientTurbulence
{
    float m_gain;
    float m_lacunarity;
    float m_gradientPerturbAmp;
    float m_fractalBounding;
    float m_frequency;
    int m_octaves;
    void calcValue(CVector3f& v3)
    {
        float amp = m_gradientPerturbAmp* m_fractalBounding;
        float freq = m_frequency;
        gradientTurbulence<T>(amp, m_frequency, v3);
        for (int i = 1; i < m_octaves; i++) {
            freq *= m_lacunarity;
            amp *= m_gain;
            gradientTurbulence<T>(amp, freq, v3);

        }
    }
};

template<class T> inline float perlin2D(const CVector2f& pos, const SPerlinInfo& perlininfo)
{
  float amplitude = perlininfo.m_amplitude;
  float value = 0.0f;
  CVector2f p = pos;
  for (int i = 0; i < perlininfo.m_octaves; i++) {
     value += noiseInterpolate2D<T>(p)*amplitude;
     p *= 0.5f;
     amplitude *= perlininfo.m_persistence;
  };
  value = umin(1.0f, umax(0.0f,value));
  return value;
};

template<class T> inline float perlin3D(const CVector3f& pos, const SPerlinInfo& perlininfo)
{
  float amplitude = perlininfo.m_amplitude;
  float value = 0.0f;
  CVector3f p = pos;
  for (int i = 0; i < perlininfo.m_octaves; i++) {
     value += noiseInterpolate3D<T>(p)*amplitude;
     p *= 0.5f;
     amplitude *= perlininfo.m_persistence;
  };
  value = umin(1.0f, umax(0.0f,value));
  return value;
};

template<class T> inline float perlin3DInverse(const CVector3f& pos, const SPerlinInfo& perlininfo)
{
  float amplitude = perlininfo.m_amplitude;
  float value = 0.0f;
  CVector3f p = pos;
  for (int i = 0; i < perlininfo.m_octaves; i++) {
     value += noiseInterpolate3D<T>(p)*amplitude;
     p *= 2.0f;
     amplitude *= perlininfo.m_persistence;
  };
  value = umin(1.0f, umax(0.0f,value));
  return value;
};

template<class T, int m_octaves> struct SPerlinOptimized
{
    float m_amplitude;
    float m_persistence;
    SNoiseInterpolate3DOptimized<T> m_levels[m_octaves];

    inline SPerlinOptimized() {
      m_amplitude = 0.5f;
      m_persistence=0.5f;
    };

    inline float inversePerlin(const CVector3f& pos)
    {
        float amplitude = m_amplitude;
        float value = 0.0f;
        CVector3f p = pos;
        for (int i = 0; i < m_octaves; i++) {
           value += m_levels[i].calcValue(p)*amplitude;
           p.x *= 2.0f;
           p.y *= 2.0f;
           p.z *= 2.0f;
           amplitude *= m_persistence;
        };
        if (value < 0.0f) value = 0.0f;
        if (value > 1.0f) value = 1.0f;
        //assert(absf(perlin3DInverse<T>(pos, SPerlinInfo(m_amplitude, m_persistence, m_octaves))-value ) < 0.001);
        return value;

    }
    inline float perlin(const CVector3f& pos)
    {
        float amplitude = m_amplitude;
        float value = 0.0f;
        CVector3f p = pos;
        for (int i = 0; i < m_octaves; i++) {
           value += m_levels[i].calcValue(p)*amplitude;
           p.x *= 0.5f;
           p.y *= 0.5f;
           p.z *= 0.5f;
           amplitude *= m_persistence;
        };
        if (value < 0.0f) value = 0.0f;
        if (value > 1.0f) value = 1.0f;
        //assert(absf(perlin3D<T>(pos, SPerlinInfo(m_amplitude, m_persistence, m_octaves))-value ) < 0.001);
        return value;

    }

    inline float inversePerlinAsynchron(const CVector3f& pos)
    {
        float amplitude = m_amplitude;
        float value = 0.0f;
        CVector3f p = pos;
        for (int i = 0; i < m_octaves; i++) {
           value += m_levels[i].calcValueAsynchron(p)*amplitude;
           p.x *= 2.0f;
           p.y *= 2.0f;
           p.z *= 2.0f;
           amplitude *= m_persistence;
        };
        if (value < 0.0f) value = 0.0f;
        if (value > 1.0f) value = 1.0f;
        //assert(absf(perlin3DInverse<T>(pos, SPerlinInfo(m_amplitude, m_persistence, m_octaves))-value ) < 0.001);
        return value;

    }
    inline float perlinAsynchron(const CVector3f& pos)
    {
        float amplitude = m_amplitude;
        float value = 0.0f;
        CVector3f p = pos;
        for (int i = 0; i < m_octaves; i++) {
           value += m_levels[i].calcValueAsynchron(p)*amplitude;
           p.x *= 0.5f;
           p.y *= 0.5f;
           p.z *= 0.5f;
           amplitude *= m_persistence;
        };
        if (value < 0.0f) value = 0.0f;
        if (value > 1.0f) value = 1.0f;
        //assert(absf(perlin3D<T>(pos, SPerlinInfo(m_amplitude, m_persistence, m_octaves))-value ) < 0.001);
        return value;

    }
};


struct SVoroniInfo
{
  CVector2f m_pointmovement;
  float m_pointmovement_z;
  int m_points;
  CVector2f m_value;
  float m_frequency;
  inline SVoroniInfo() : m_pointmovement(1,1) {
     m_points = 1;
     m_value = CVector2f(0.5f,0.5f);
     m_pointmovement = CVector2f(0.5f, 0.5f);m_pointmovement_z=0.0f;
    m_frequency = 16;
  }
};

struct SVoroni
{
  static float eval(const CIVector3& minv, float distance, float distance2, const SVoroniInfo& info)
  {
    //if (distance < 0.05) return 1;
    return noise(minv);
  };
};

struct SVoroniMesh
{
  static float eval(const CIVector3& minv, float distance, float distance2, const SVoroniInfo& info)
  {
    //if (distance < 0.05) return 1;
    if (distance2 - distance < info.m_value.x) return 1.0f;

    return 0.0f;
  };
};

struct SVoroniMesh2
{
  static inline float eval(const CIVector3& minv, float distance, float distance2, const SVoroniInfo& info)
  {
    //if (distance < 0.05) return 1;
   // float value = (distance2-distance)*();

  //  if (distance2 - distance < info.m_value.x) return 1.0f;
    float size = info.m_value.x;
    if (size < 0.0001f) size = 0.0001f;
    return umin(1.0f,(distance2 - distance)/size);
  };
};

struct SVoroniPoint
{
  static inline float eval(const CIVector3& minv, float distance, float distance2, const SVoroniInfo& info)
  {
    return noise(minv)*umax(0.0f,1.0f-distance/info.m_value.x);
  };
};

struct SVoroniPoint2
{
  static inline  float eval(const CIVector3& minv, float distance, float distance2, const SVoroniInfo& info)
  {
    return umax(0.0f,1.0f-distance/(info.m_value.x*noise(minv)));
  };
};


struct SVoroniWavesVec2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float dis = sqrt(sqr(distance.x /info.m_value.x)+sqr(distance.y /info.m_value.y));

    return umax(0.0f,1.0f-dis/(noise(minv)+0.01f))* sin(dis*info.m_frequency)*0.5f;
  };
};



struct SVoroniPointVec2
{
  static float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float dis = sqrt(sqr(distance.x /info.m_value.x)+sqr(distance.y /info.m_value.y));

    return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};


struct SVoroniPointRotate2
{
  static inline  float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float rotation = noise(CIVector3(minv.x+1, minv.y-1, minv.z))*360;
    CVector2f dir = vRotateVector(distance, rotation);
    float dis = sqrt(sqr(dir.x /info.m_value.x)+sqr(dir.y /info.m_value.y));

    return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};


struct SVoroniPointSVec2
{
  static inline  float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float dis = sqrt(sqr(distance.x /info.m_value.x)+sqr(distance.y /info.m_value.y));

    if (dis/(noise(minv)+0.01f) >= 1.0f) {
      return 0.0f;
    } else {
      return 1.0f;
    }
    //return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};


struct SVoroniPointSRotate2
{
  static inline  float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float rotation = noise(CIVector3(minv.x+1, minv.y-1, minv.z))*360;
    CVector2f dir = vRotateVector(distance, rotation);
    float dis = sqrt(sqr(dir.x /info.m_value.x)+sqr(dir.y /info.m_value.y));

     if (dis/(noise(minv)+0.01f) >= 1.0f) {
      return 0.0f;
    } else {
      return 1.0f;
    }
    //return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};



struct SVoroniRectVec2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float dis = umax(absf(distance.x /info.m_value.x),absf(distance.y /info.m_value.y));

    return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};


struct SVoroniRectRotate2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float rotation = noise(CIVector3(minv.x+1, minv.y-1, minv.z))*360;
    CVector2f dir = vRotateVector(distance, rotation);
    float dis = umax(absf(dir.x /info.m_value.x),absf(dir.y /info.m_value.y));

    return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};


struct SVoroniRectSVec2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float dis = umax(absf(distance.x /info.m_value.x),absf(distance.y /info.m_value.y));

    if (dis/(noise(minv)+0.01f) >= 1.0f) {
      return 0.0f;
    } else {
      return 1.0f;
    }
    //return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};


struct SVoroniRectSRotate2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float rotation = noise(CIVector3(minv.x+1, minv.y-1, minv.z))*360;
    CVector2f dir = vRotateVector(distance, rotation);
    float dis = umax(absf(dir.x /info.m_value.x),absf(dir.y /info.m_value.y));

     if (dis/(noise(minv)+0.01f) >= 1.0f) {
      return 0.0f;
    } else {
      return 1.0f;
    }
    //return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};

struct SVoroniStarVec2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
   // float rotation = noise(CIVector2(minv.x+1, minv.y-1))*360;
    CVector2f dir = distance;//rotateVector(distance, rotation);
    float size = (noise(minv)+0.01f);
    dir.x /= info.m_value.x*size;
    dir.y /= info.m_value.y*size;
    if (dir.x <= -1.0f || dir.x >= 1.0f) return 0.0f;
    if (dir.y <= -1.0f || dir.y >= 1.0f) return 0.0f;
    float rect = 1.0f-umax(absf(dir.x),absf(dir.y));
   //float dis = umax(absf(dir.x /info.m_value.x),absf(dir.y /info.m_value.y));
    return pow(1.0f-umin(absf((dir.x)-dir.y), absf((dir.y)-(-dir.x))), 3)*rect;
    /* if (dis/(noise(minv)+0.01f) >= 1.0f) {
      return 0.0f;
    } else {
      return 1.0f;
    }*/
    //return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};

struct SVoroniStarRotateVec2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float rotation = noise(CIVector3(minv.x+1, minv.y-1, minv.z))*360;
    CVector2f dir = vRotateVector(distance, rotation);
    float size = (noise(minv)+0.01f);
    dir.x /= info.m_value.x*size;
    dir.y /= info.m_value.y*size;
    if (dir.x <= -1.0f || dir.x >= 1.0f) return 0.0f;
    if (dir.y <= -1.0f || dir.y >= 1.0f) return 0.0f;
    float rect = 1.0f-umax(absf(dir.x),absf(dir.y));
    //float dis = umax(absf(dir.x /info.m_value.x),absf(dir.y /info.m_value.y));
    return pow(1.0f-umin(absf((dir.x)-dir.y), absf((dir.y)-(-dir.x))), 3)*rect;
    /* if (dis/(noise(minv)+0.01f) >= 1.0f) {
      return 0.0f;
    } else {
      return 1.0f;
    }*/
    //return umax(0.0f,1.0f-dis/(noise(minv)+0.01f));
  };
};

struct SVoroniStar2Vec2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {
    float angle = atan2(distance.x, distance.y);
    float rotation = noise(CIVector3(minv.x+1, minv.y-1, minv.z))*360;
    CVector2f dir = vRotateVector(distance, rotation);
    float size = (noise(minv)+0.01f);
    dir.x /= info.m_value.x*size;
    dir.y /= info.m_value.y*size;
   // if (dir.x <= -1.0f || dir.x >= 1.0f) return 0.0f;
   // if (dir.y <= -1.0f || dir.y >= 1.0f) return 0.0f;
    CIVector2 repeat;
    repeat.x = repeat.y = 0;
    float mul = noiseInterpolate2D<SInterpolateCosine >(CVector2f(angle*4, (float) (minv.y+minv.x*345)), repeat)*0.5f+0.5f;
    float dis = umin(1.0f,(float)sqrt(sqr(dir.x)+sqr(dir.y))*mul);
    return Interpolatef(1.0f, 0.0f, dis);
  };
};


struct SVoroniPieVec2
{
  static inline float eval(const CIVector3& minv, const CVector2f& distance, float distance2, const SVoroniInfo& info)
  {

    float rotation = noise(CIVector3(minv.x+1, minv.y-1, minv.z))*360;
    CVector2f dir = vRotateVector(distance, rotation);
    float angle = atan2(dir.x, dir.y);
    float size = (noise(minv)+0.01f);
    dir.x /= info.m_value.x*size;
    dir.y /= info.m_value.y*size;
    float dis = sqrt(sqr(dir.x)+sqr(dir.y));
    return 1.0f*(umax(0.0f, 1.0f-absfloat(1.0-angle)))*umax(0.0f, 1.0f-absfloat(0.75f-dis)*4);
   // if (dir.x <= -1.0f || dir.x >= 1.0f) return 0.0f;
   // if (dir.y <= -1.0f || dir.y >= 1.0f) return 0.0f;
   /* CIVector2 repeat;
    repeat.x = repeat.y = 0;
    float mul = noiseInterpolate2D<SInterpolateCosine >(CVector2(angle*4, minv.y+minv.x*345), repeat)*0.5f+0.5f;
    float dis = umin(1.0f,sqrt(sqr(dir.x)+sqr(dir.y))*mul);
    return Interpolatef(1.0f, 0.0f, dis);*/
  };
};

int intcompare (const void *x1, const void *x2);

//Auch Punkte
template <class T, class Color> inline float voroniDiagram(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat, int layer)
{
  CIVector2 base = vfloori(pos);
  CVector2f fraction = vfrac(pos);
  float min = 100.0f;
  float min2 = 80.0f;
  CIVector3 minv(0,0, 0);
  int tp = info.m_points;
// Array<float> distances((tp*2+1)*(tp*2+1));
// int index = 0;
  for (int i = -tp; i <= tp; i++) {
    for (int j = -tp; j <= tp; j++) {
      CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), layer);
      //float value = noise(pi);
      CVector2f pmove = noise2(pi);
      pmove.x *= info.m_pointmovement.x;
      pmove.y *= info.m_pointmovement.y;
      CVector2f point = CVector2f(0.5f+i, 0.5f+j)+pmove;
      float distance = T::distance(point, fraction);
     // distances[index] = distance;
     //index++;
      //float distance = std::sqrt(sqr(point.x-fraction.x)+sqr(point.y-fraction.y));
      //float distance = 0.5f*point.distanceTo(fraction) + 0.5f*umin(absf(point.x-fraction.x),absf(point.y-fraction.y));
      //float distance = umin(absf(point.x-fraction.x),absf(point.y-fraction.y));

      if (distance < min) {
    minv = pi;
	min2 = min;
	min = distance;
      } else if (distance < min2){
	min2 = distance;
      };
      /*if (distance != min) {
	if (distance < min2){
	  min2 = distance;
         };
      };*/
    };
  };
 // std::qsort(distances.array, distances.size(), sizeof(int), intcompare);
 // utiassert(distances[0] == min);
  //assert(distances[1] == min2);
  return Color::eval(minv, min, min2, info);
};

inline float voroniDiagram1DNearestPoint(float pos, const SVoroniInfo& info, const int& repeat, int layer)
{
  int base = vfloori(pos);
  float fraction = vfrac(pos);
  float min = 100000.0f;
  int tp = info.m_points;
  float res(pos);
  for (int i = -tp; i <= tp; i++) {
      CIVector3 pi = CIVector3(utimod(base+i, repeat),0, layer);
      //float value = noise(pi);
      float pmove = noise(pi);
      pmove *= info.m_pointmovement.x;
      float point =(0.5f+float(i))+pmove;
      float distance = fabsf(point- fraction);
      if (distance < min) {
    min = distance;
    res = point+base;
      }

  };
  return res;
};


template <class T> CVector2f voroniDiagram2DNearestPoint(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat, int layer)
{
  CIVector2 base = vfloori(pos);
  CVector2f basef(base.x, base.y);
  CVector2f fraction = vfrac(pos);
  float min = 100000.0f;
  int tp = info.m_points;
  CVector2f res(pos);
  for (int i = -tp; i <= tp; i++) {
    for (int j = -tp; j <= tp; j++) {
      CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), layer);
      //float value = noise(pi);
      CVector2f pmove = noise2(pi);
      pmove.x *= info.m_pointmovement.x;
      pmove.y *= info.m_pointmovement.y;
      CVector2f point = CVector2f(0.5f+i, 0.5f+j)+pmove;
      float distance = T::distance(point, fraction);
      if (distance < min) {
    min = distance;
    res = point+basef;
      }
    };
  };
  return res;
};


template <class T> CVector3f voroniDiagram3DNearestPoint(const CVector3f& pos, const SVoroniInfo& info, const CIVector3& repeat, int layer)
{
  CIVector3 base = vfloori(pos);

  CVector3f basef(base.x, base.y, base.z);
  CVector3f fraction = vfrac(pos);
  float min = 100000.0f;
  int tp = info.m_points;
  CVector3f res(pos);
  for (int i = -tp; i <= tp; i++) {
    for (int j = -tp; j <= tp; j++) {
        for (int k = -tp; k <= tp; k++) {
      CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y),  utimod(base.z+k, repeat.z));
      //float value = noise(pi);
      CVector3f pmove = noise3(pi);
      pmove.x *= info.m_pointmovement.x;
      pmove.y *= info.m_pointmovement.y;
      pmove.z *= info.m_pointmovement_z;
      CVector3f point = CVector3f(0.5f+i, 0.5f+j, 0.5+k)+pmove;
      float distance = T::distance(point, fraction);
      if (distance < min) {
    min = distance;
    res = point+basef;
      }
      }
    };
  };
  return res;
};


void voroniPointInTraingle(Array<CVector2f> &p1, int count, const SVoroniInfo& info, int layer, Array<CVector2f> &res, int &pointcount);

template <class T, class Color> float voroniPoint(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat, int layer)
{
  CIVector2 base = vfloori(pos);
  CVector2f fraction = vfrac(pos);
  //float min = 100.0f;
 // CIVector3 minv(0,0, 0);
  int tp = info.m_points;
  float res = 0.0f;
  for (int i = -tp; i <= tp; i++) {
    for (int j = -tp; j <= tp; j++) {
      //CIVector2 pi = CIVector2(base.x+i, base.y+j);
      CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), layer);
      //float value = noise(pi);
      CVector2f pmove = noise2(pi);//*info.m_pointmovement;
      pmove.x *= info.m_pointmovement.x;
      pmove.y *= info.m_pointmovement.y;
      CVector2f point = CVector2f(0.5f+i, 0.5f+j)+pmove;
      float distance = T::distance(point, fraction);
      //float distance = std::sqrt(sqr(point.x-fraction.x)+sqr(point.y-fraction.y));
      //float distance = 0.5f*point.distanceTo(fraction) + 0.5f*umin(absf(point.x-fraction.x),absf(point.y-fraction.y));
      //float distance = umin(absf(point.x-fraction.x),absf(point.y-fraction.y));
      res += Color::eval(pi, distance, 0,info);

    };
  };
  return umin(1.0f, res);
};

template <class Color> float voroniPoint2(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat, int layer)
{
  CIVector2 base = vfloori(pos);
  CVector2f fraction = vfrac(pos);
  //float min = 100.0f;
  CIVector3 minv(0,0, 0);
  CIVector3 pi;
  CVector2f pmove;
  CVector2f point;
  CVector2f distance;
  int tp = info.m_points;
  float res = 0.0f;
  for (int i = -tp; i <= tp; i++) {
    for (int j = -tp; j <= tp; j++) {
      //CIVector2 pi = CIVector2(base.x+i, base.y+j);
      pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), layer);
      //float value = noise(pi);
      pmove = noise2(pi);//*info.m_pointmovement;
      pmove.x *= info.m_pointmovement.x;
      pmove.y *= info.m_pointmovement.y;
      point = CVector2f(0.5f+i, 0.5f+j)+pmove;
      distance = point- fraction;
      //float distance = std::sqrt(sqr(point.x-fraction.x)+sqr(point.y-fraction.y));
      //float distance = 0.5f*point.distanceTo(fraction) + 0.5f*umin(absf(point.x-fraction.x),absf(point.y-fraction.y));
      //float distance = umin(absf(point.x-fraction.x),absf(point.y-fraction.y));
      res += Color::eval(pi, distance,0, info);

    };
  };
  return umin(1.0f, res);
};

template <class T> CIVector2 voroniNext(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat, int layer)
{
  CIVector2 base = vfloori(pos);
  CVector2f fraction = vfrac(pos);
  float min = 100.0f;
  float min2 = 80.0f;
  CIVector3 minv(0,0, 0);
  int tp = info.m_points;
// Array<float> distances((tp*2+1)*(tp*2+1));
// int index = 0;
  for (int i = -tp; i <= tp; i++) {
    for (int j = -tp; j <= tp; j++) {
      CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), layer);
      CVector2f pmove = noise2(pi);
      pmove.x *= info.m_pointmovement.x;
      pmove.y *= info.m_pointmovement.y;
      CVector2f point = CVector2f(0.5f+i, 0.5f+j)+pmove;
      float distance = T::distance(point, fraction);

      if (distance < min) {
	minv = pi;
	min2 = min;
	min = distance;
      } else if (distance < min2){
	min2 = distance;
      };
    };
  };
 // std::qsort(distances.array, distances.size(), sizeof(int), intcompare);
 // utiassert(distances[0] == min);
  //assert(distances[1] == min2);
  return CIVector2(minv.x, minv.y);
};



template <class T, class Color, bool voronioptimize = true> struct SVoroniOptimized
{
    Array<CVector2f> m_noise;
    Array<CIVector3> m_pi;
   // CVector2 m_last;
    CIVector2 m_basevalue;
    CIVector2 m_repeat;
    CVector2f m_lastposition;
    float m_distance;
    int m_layer;
    int m_points;
    int m_last_min_index;
    CIVector3 m_last_minv;

    float v1;
    float v2;
    inline SVoroniOptimized()
    {
        m_basevalue.set(-1000000,-1000000);
        m_repeat.set(-1000000, -1000000);
        m_layer = 0;
        m_points = -2;
        m_distance = 0;
        //m_lastposition.set(-1000000, -100000);
    }
    void setLayer(int l)
    {
        if (l != m_layer) {
            m_basevalue.set(-1000000,-1000000);
            m_layer = l;
             m_distance = 0;
            // m_lastposition.set(-1000000, -100000);
        }
    }
    void setPoints(int l)
    {
        if (l != m_points) {
            m_points = l;
            m_basevalue.set(-1000000,-1000000);
            m_noise.setSize((m_points*2+1)*(m_points*2+1));
            m_pi.setSize((m_points*2+1)*(m_points*2+1));
             m_distance = 0;
            // m_lastposition.set(-1000000, -100000);

        }
    }

   void calcElementsOfArea(const CRectf& pos, const SVoroniInfo& info, const CIVector2& repeat, std::vector< std::pair<CVector2f, float> >& res )
    {

        assert(m_points == info.m_points);
        CIntRect r;
        r.m_min = CIVector2(pos.m_min.x-m_points, pos.m_min.y-m_points);
        r.m_max = CIVector2(pos.m_max.x+m_points, pos.m_max.y+m_points);

        for (int i = r.m_min.x; i <= r.m_max.x; i++) {
          for (int j = r.m_min.y; j <= r.m_max.y; j++) {
            CIVector3 pi = CIVector3(utimod(i, repeat.x), utimod(j, repeat.y), m_layer);
            CVector2f pmove = noise2(pi);
            pmove.x *= info.m_pointmovement.x;
            pmove.y *= info.m_pointmovement.y;

            CVector2f finalpos = CVector2f(0.5f+i, 0.5f+j)+pmove;
            float finalvalue = Color::eval(pi, 0.0f, 2.0f, info);

            if (pos.contains(finalpos)) {
                res.push_back(std::make_pair(finalpos, finalvalue));
            }
          };
        };

    }

    float calcDiagram(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat)
    {
        float result=0.0f;
        assert(m_points == info.m_points);
        int currentposition;
        CIVector2 base;
        CVector2f fraction;
        float min = 100.0f;
        float min2 = 80.0f;
        CIVector3 minv(0,0, 0);
        fraction.x = fasterfloor(pos.x);
        fraction.y = fasterfloor(pos.y);
      //  fraction.z = fasterfloor(pos.z);

        base.x = fraction.x;
        base.y = fraction.y;
      //  base.z = fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
       // fraction.z = pos.z - fraction.z;


        if (base.x != m_basevalue.x || base.y != m_basevalue.y || repeat.x != m_repeat.x || repeat.y != m_repeat.y) {
            m_basevalue = base;
            m_repeat = repeat;
            currentposition=0;
            for (int i = -m_points; i <= m_points; i++) {
              for (int j = -m_points; j <= m_points; j++) {
                CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), m_layer);
                CVector2f pmove = noise2(pi);
                pmove.x *= info.m_pointmovement.x;
                pmove.y *= info.m_pointmovement.y;

                m_noise[currentposition] = CVector2f(0.5f+i, 0.5f+j)+pmove;
                currentposition++;

              };
            };
             m_distance = 0;
             m_lastposition = CVector2f(-1000000, -100000);
        }
        if (voronioptimize == false || T::distance(m_lastposition, pos) >= m_distance) {
        currentposition = 0;
        m_last_min_index = 0;
        for (int i = -m_points; i <= m_points; i++) {
          for (int j = -m_points; j <= m_points; j++) {

              float distance = T::distance(m_noise[currentposition], fraction);

              if (distance < min) {
                 minv.x = i;
                 minv.y = j;
                 min2 = min;
                 m_last_min_index = currentposition;
                 min = distance;
              } else if (distance < min2){
                 min2 = distance;
              };
              currentposition++;

          };
        };
        m_lastposition = pos;
        m_last_minv = CIVector3(utimod(base.x+minv.x, repeat.x), utimod(base.y+minv.y, repeat.y), m_layer);
           m_distance = (min2-min)/2.0f;
        } else {
            min = T::distance(m_noise[m_last_min_index], fraction);
            min2 = min + 0.1f;
        }
        result = Color::eval(m_last_minv, min, min2, info);
        //assert(absf(result-voroniDiagram<T, Color>(pos, info, repeat, m_layer)) < 0.001);
        return  result;
    }

    CIVector2 calcNext(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat)
    {
        CIVector2 result;
        assert(m_points == info.m_points);
        int currentposition;
        CIVector2 base;
        CVector2f fraction;
        float min = 100.0f;
        float min2 = 80.0f;
        CIVector3 minv(0,0,0);
        fraction.x = fasterfloor(pos.x);
        fraction.y = fasterfloor(pos.y);
      //  fraction.z = fasterfloor(pos.z);

        base.x = fraction.x;
        base.y = fraction.y;
      //  base.z = fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
       // fraction.z = pos.z - fraction.z;


        if (base.x != m_basevalue.x || base.y != m_basevalue.y || repeat.x != m_repeat.x || repeat.y != m_repeat.y) {
            m_basevalue = base;
            m_repeat = repeat;
            currentposition=0;
            for (int i = -m_points; i <= m_points; i++) {
              for (int j = -m_points; j <= m_points; j++) {
                CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), m_layer);
                CVector2f pmove = noise2(pi);
                pmove.x *= info.m_pointmovement.x;
                pmove.y *= info.m_pointmovement.y;

                m_noise[currentposition] = CVector2f(0.5f+i, 0.5f+j)+pmove;
                currentposition++;

              };
            };
        }
        currentposition = 0;
        for (int i = -m_points; i <= m_points; i++) {
          for (int j = -m_points; j <= m_points; j++) {

              float distance = T::distance(m_noise[currentposition], fraction);
              currentposition++;
              if (distance < min) {
                 minv.x = i;
                 minv.y = j;
                 min2 = min;
                 min = distance;
              } else if (distance < min2){
                 min2 = distance;
              };

          };
        };
        result = CIVector2(utimod(base.x+minv.x, repeat.x), utimod(base.y+minv.y, repeat.y));
        assert(result==voroniNext<T>(pos, info, repeat, m_layer));
        return  result;
    }

    float calcPoint(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat)
    {
        float result=0.0f;
        assert(m_points == info.m_points);
        int currentposition;
        CIVector2 base;
        CVector2f fraction;
        //float min = 100.0f;
        //float min2 = 80.0f;
        //CIVector2 minv(0,0);
        fraction.x = fasterfloor(pos.x);
        fraction.y = fasterfloor(pos.y);
      //  fraction.z = fasterfloor(pos.z);

        base.x = fraction.x;
        base.y = fraction.y;
      //  base.z = fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
       // fraction.z = pos.z - fraction.z;


        if (base.x != m_basevalue.x || base.y != m_basevalue.y || repeat.x != m_repeat.x || repeat.y != m_repeat.y) {
            m_basevalue = base;
            m_repeat = repeat;
            currentposition=0;
            for (int i = -m_points; i <= m_points; i++) {
              for (int j = -m_points; j <= m_points; j++) {
                CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), m_layer);
                CVector2f pmove = noise2(pi);
                pmove.x *= info.m_pointmovement.x;
                pmove.y *= info.m_pointmovement.y;

                m_noise[currentposition] = CVector2f(0.5f+i, 0.5f+j)+pmove;
                m_pi[currentposition] = pi;
                currentposition++;

              };
            };
        }
        currentposition = 0;
        for (int i = -m_points; i <= m_points; i++) {
          for (int j = -m_points; j <= m_points; j++) {

              float distance = T::distance(m_noise[currentposition], fraction);

              result += Color::eval(m_pi[currentposition], distance, 0,info);
              currentposition++;

          };
        };
        result =  umin(1.0f, result);
      //  assert(absf(result-voroniPoint<Color>(pos, info, repeat, m_layer)) < 0.001);
        return  result;
    }
    float calcPoint2(const CVector2f& pos, const SVoroniInfo& info, const CIVector2& repeat)
    {
        float result=0.0f;
        assert(m_points == info.m_points);
        int currentposition;
        CIVector2 base;
        CVector2f fraction;
        CVector2f distance;
        //float min = 100.0f;
        //float min2 = 80.0f;
        //CIVector3 minv(0,0,0);
        fraction.x = fasterfloor(pos.x);
        fraction.y = fasterfloor(pos.y);
      //  fraction.z = fasterfloor(pos.z);

        base.x = fraction.x;
        base.y = fraction.y;
      //  base.z = fraction.z;

        fraction.x = pos.x - fraction.x;
        fraction.y = pos.y - fraction.y;
       // fraction.z = pos.z - fraction.z;


        if (base.x != m_basevalue.x || base.y != m_basevalue.y || repeat.x != m_repeat.x || repeat.y != m_repeat.y) {
            m_basevalue = base;
            m_repeat = repeat;
            currentposition=0;
            for (int i = -m_points; i <= m_points; i++) {
              for (int j = -m_points; j <= m_points; j++) {
                CIVector3 pi = CIVector3(utimod(base.x+i, repeat.x), utimod(base.y+j, repeat.y), m_layer);
                CVector2f pmove = noise2(pi);
                pmove.x *= info.m_pointmovement.x;
                pmove.y *= info.m_pointmovement.y;

                m_noise[currentposition] = CVector2f(0.5f+i, 0.5f+j)+pmove;
                 m_pi[currentposition] = pi;
                currentposition++;

              };
            };
        }
        currentposition = 0;
        for (int i = -m_points; i <= m_points; i++) {
          for (int j = -m_points; j <= m_points; j++) {
              distance = m_noise[currentposition]- fraction;
              result += Color::eval(m_pi[currentposition], distance,0, info);
              currentposition++;
          };
        };
        result =  umin(1.0f, result);
      //  assert(absf(result-voroniPoint2<Color>(pos, info, repeat, m_layer)) < 0.001);
        return  result;
    }
};




#endif
