#include "simplematrix.h"


namespace em {
    void ExpandedSum::sort()
    {
        for (int i = 0; i < m_sum.size(); i++) {
            m_sum.at(i).sort();
        }

        std::sort(m_sum.begin(), m_sum.end());

        std::vector<ExpandedProduct> sum;
        for (int i = 0; i < m_sum.size(); i++) {
            if (sum.size() == 0) {
               sum.push_back(m_sum[i]);
            } else if (sum[sum.size()-1].sameVariables(m_sum[i])) {
               sum[sum.size()-1].m_constant += m_sum[i].m_constant;
            } else {
               sum.push_back(m_sum[i]);

            }
        }
        m_sum.clear();
        for (int i = 0; i < sum.size(); i++) {
            m_sum.push_back(sum[i]);

        }
    }
    void ExpandedSum::fillVars(std::set<SExpandedVariable>& m_variables)
    {
       for (int i = 0; i < m_sum.size(); i++) {
            m_sum.at(i).fillVars(m_variables);
        }
    }


    CUtiString ExpandedProduct::toString()
    {
        CUtiString str;
        int varcount = getVarCount();
        if (m_constant != 1.0 || varcount == 0) {
            str += m_constant;
            if (varcount != 0) {
                str+=UTISTR("*");
            }
        }

        for (int i = 0; i < varcount; i++) {
            str += m_variable[i].m_basevariable.str();
            str += UTISTR("[");
            str += m_variable[i].m_variableindex;
            str += UTISTR("]");
        }

        return str;
    }


    CUtiString ExpandedSum::toString()
    {
        CUtiString str;
        for (int i = 0; i < m_sum.size(); i++) {
            str += m_sum[i].toString();
            if (i != m_sum.size()-1) {
                str += UTISTR(" + ");
            }
        }
        return str;
    }
    EquationCompiled_Equation*ExpandedSum::createCompiledEquation(std::map<SExpandedVariable, int> &var_to_index)
    {
        EquationCompiled_Equation* eq = new EquationCompiled_Equation();

        eq->m_sum.setSize(m_sum.size());
        for (int j = 0; j< m_sum.size(); j++) {
            ExpandedProduct&pp = m_sum.at(j);
            eq->m_sum[j].m_factor = pp.m_constant;
            eq->m_sum[j].m_count = pp.getVarCount();
            for (int k = 0; k < eq->m_sum[j].m_count; k++) {
                auto itt = var_to_index.find(pp.m_variable[k]);
                assert(itt != var_to_index.end());
                eq->m_sum[j].m_prodelement[k] = itt->second;
            }

        }
        return eq;
    }

    std::pair<double, double> ExpandedSum::getMinMax01(std::set<SExpandedVariable>& variables) const
    {
        double minvalue = 1000000;
        double maxvalue = -1000000;

        int checkamount = 11;

        ::Array<int> checkindex(variables.size());
        ::Array<SExpandedVariable> varnames(variables.size());
        checkindex.fill(0);
        int index=0;
        for (auto p: variables) {varnames[index] = p; index++;};

        while (true) {
            std::map<SExpandedVariable, double> vars;
            for (int i = 0; i < variables.size(); i++) {
                vars.insert(std::make_pair(varnames[i],double(checkindex[i])/double(checkamount)));
            }

            double res = eval(vars);
            minvalue = umin(minvalue, res);
            maxvalue = umax(maxvalue, res);

        }

        return std::make_pair(minvalue, maxvalue);
    }

    double ExpandedSum::eval(std::map<SExpandedVariable, double>& vars) const
    {
        double res=0.0;

        for (int i = 0; i < m_sum.size(); i++) {
            double r1 = m_sum[i].eval(vars);

            res += r1;
        }
        return res;
    }
    void ExpandedSum::derv(const SExpandedVariable& variable, ExpandedSum& sum)
    {
        sum.m_sum.clear();
        for (int i = 0; i < m_sum.size(); i++) {
            ExpandedProduct p = m_sum[i].derv(variable);
            if (p.m_constant == 0.0) {
                continue;
            }
            sum.m_sum.push_back(p);
        }

    }


    double ExpandedProduct::eval(std::map<SExpandedVariable, double>& vars) const
    {
        double res = m_constant;
        for (int i = 0; i < m_variable.size(); i++) {
            if (m_variable[i].m_basevariable.empty()) continue;
             auto f = vars.find(m_variable[i]);
             assert(f != vars.end());
             res += f->second;
        }

        return res;
    }
    ExpandedProduct ExpandedProduct::derv(const SExpandedVariable& variable)
    {
        int vc = getVarCount();

        if (vc == 0)
        {
            ExpandedProduct p;
            return p;
        }
        int c = 0;
        for (int i = 0; i < m_variable.size(); i++) {
            if (m_variable[i] == variable) {
                c++;
            }
        }
        if (c == 0) {
            ExpandedProduct p;
            return p;
        }

        ExpandedProduct p;
        p.m_constant = m_constant * double(c);

        bool removed = false;
        int writepos = 0;
        for (int i = 0; i < m_variable.size(); i++) {
            if (m_variable[i] == variable && removed == false) {
                removed = true;
                continue;
            }
            p.m_variable[writepos] = m_variable[i];
            writepos++;
        }

        return p;
    }


    SExpandedVariable::operator ExpandedSum () const
    {
         ExpandedSum eq;
         eq.m_sum.push_back(ExpandedProduct(m_basevariable,m_variableindex));
         return eq;
    }


    ExpandedSum SExpandedVariable::sum()const
{
    ExpandedSum eq;
    eq.m_sum.push_back(ExpandedProduct(m_basevariable,m_variableindex));
    return eq;
}
    ExpandedSum operator+(double value, const ExpandedSum& prod)
    {
        ExpandedSum result;
        result += value;
        result += prod;
        return result;
    }

    ExpandedSum operator*(double value, const ExpandedSum& prod)
    {
        ExpandedSum result;
        result += value;

        return result*prod;
    }


    ExpandedSum ExpandedSum::operator+(const ExpandedSum& prod) const
    {
         ExpandedSum result;
         result += *this;
         result += prod;
         return result;
    }
    ExpandedSum ExpandedSum::operator+(double value) const
    {
         ExpandedSum result;
         result += *this;
         result.m_sum.push_back( ExpandedProduct(value));
         return result;
    }

    ExpandedSum ExpandedSum::operator-(const ExpandedSum& prod) const
    {
         ExpandedSum result;
         result += *this;
         result -= prod;
         return result;
    }
    ExpandedSum ExpandedSum::operator-() const
    {
        ExpandedSum result;
        result += (*this)*-1;
        return result;
    }

    ExpandedSum ExpandedSum::operator-(double value) const
    {
         ExpandedSum result;
         result += *this;
         result.m_sum.push_back( ExpandedProduct(-value));
         return result;
    }

    ExpandedSum& ExpandedSum::operator+=(const ExpandedSum& prod)
    {
        for (int i = 0; i < prod.m_sum.size(); i++) {

            m_sum.push_back(prod.m_sum.at(i));
        }
        return *this;
    }
    ExpandedSum& ExpandedSum::operator-=(double value)
    {
        m_sum.push_back(ExpandedProduct(-value));
        return *this;
    }

    ExpandedSum& ExpandedSum::operator-=(const ExpandedSum& prod)
    {
        for (int i = 0; i < prod.m_sum.size(); i++) {

            ExpandedProduct v = prod.m_sum.at(i);
            v.m_constant = -v.m_constant;
            m_sum.push_back(v);
        }
        return *this;
    }
    ExpandedSum& ExpandedSum::operator+=(double value)
    {
        m_sum.push_back(ExpandedProduct(value));
        return *this;
    }

    ExpandedSum& ExpandedSum::operator=(const ExpandedSum& prod)
    {
        m_sum.clear();
        for (int i = 0; i < prod.m_sum.size(); i++) {

            m_sum.push_back(prod.m_sum.at(i));
        }
        return *this;
    }

    ExpandedSum ExpandedSum::operator^(int v) const
    {
        if (v == 0) return c(1);
        ExpandedSum sum = *this;

        for (int i = 1; i < v;i++) {
            sum = sum * *this;
        }
        return sum;
    }

    ExpandedSum ExpandedSum::operator*(const ExpandedSum& prod) const
    {
         ExpandedSum result;

         for (int i = 0; i < m_sum.size(); i++) {
            for (int j = 0; j < prod.m_sum.size(); j++) {
                ExpandedProduct p1 = m_sum.at(i)*prod.m_sum.at(j);
                result.m_sum.push_back(p1);

            }

         }

         return result;
    }

    ExpandedSum ExpandedSum::operator/(double value) const
    {
        return operator*(1.0f/value);
    }

    ExpandedSum ExpandedSum::operator*(double value) const
    {
         ExpandedSum result;
         result = *this;
         for (int i = 0; i < result.m_sum.size(); i++) {

             result.m_sum.at(i).m_constant *= value;
         }

         return result;

    }
    ExpandedSum& ExpandedSum::operator*=(double value)
    {
        for (int i = 0; i < m_sum.size(); i++) {

             m_sum.at(i).m_constant *= value;
         }

        return *this;
    }
    ExpandedEquation ExpandedSum::operator==(const ExpandedSum& prod) const
    {
        ExpandedEquation eq;
        eq.m_condition = 0;

        for (int i = 0; i < m_sum.size(); i++) {

            eq.m_sum1.m_sum.push_back(m_sum.at(i));
        }
        for (int i = 0; i < prod.m_sum.size(); i++) {
            ExpandedProduct p = prod.m_sum.at(i);
            p.m_constant *= -1.0f;
            eq.m_sum1.m_sum.push_back(p);
        }
        return eq;
    }
    ExpandedEquation ExpandedSum::operator<(const ExpandedSum& prod) const
    {
        ExpandedEquation eq;
        eq.m_condition = -1;

        for (int i = 0; i < m_sum.size(); i++) {

            eq.m_sum1.m_sum.push_back(m_sum.at(i));
        }
        for (int i = 0; i < prod.m_sum.size(); i++) {
            ExpandedProduct p = prod.m_sum.at(i);
            p.m_constant *= -1.0f;
            eq.m_sum1.m_sum.push_back(p);
        }
        return eq;
    }
    ExpandedEquation ExpandedSum::operator>(const ExpandedSum& prod) const
    {
        ExpandedEquation eq;
        eq.m_condition = 1;

        for (int i = 0; i < m_sum.size(); i++) {

            eq.m_sum1.m_sum.push_back(m_sum.at(i));
        }
        for (int i = 0; i < prod.m_sum.size(); i++) {
            ExpandedProduct p = prod.m_sum.at(i);
            p.m_constant *= -1.0f;
            eq.m_sum1.m_sum.push_back(p);
        }
        return eq;
    }

    EquationGenerator::EquationGenerator()
    {

    }
    void EquationGenerator::addCondition(const ExpandedEquation& ee)
    {
        m_conditions.push_back(ee);
    }
    void EquationGenerator::addConditionFirst(const ExpandedEquation& ee)
    {
        m_conditions.insert(m_conditions.begin(), ee);
    }
    void EquationGenerator::updateData()
    {
        m_variables.clear();
        m_var_to_index.clear();
        m_index_to_var.clear();

        m_optimisationfunction.sort();
        m_optimisationfunction.fillVars(m_variables);

        for (int i = 0; i < m_conditions.size(); i++) {
              m_conditions[i].sort();
              m_conditions[i].fillVars(m_variables);
        }

        for (auto p: m_variables) {
             m_var_to_index.insert(std::make_pair(p, m_index_to_var.size()));
             m_index_to_var.push_back(p);
        }
    }



    CUtiString ExpandedEquation::toString()
    {
        CUtiString end = UTISTR(" == 0");
        if (m_condition < 0) {
            end = UTISTR(" < 0");
        }
        if (m_condition > 0) {
            end = UTISTR(" > 0");

        }
        return m_sum1.toString()+end;
    }

    void EquationGenerator::print()
    {

        updateData();
        CStringList l;
        toStringList(l);
        for (int i = 0; i < l.size(); i++) {
            std::wcout << l.getLine(i) << std::endl;
        }
        std::wcout << "Variable count: " << getVarCount() << std::endl;

        std::wcout << "Equation count: "  << m_conditions.size() << std::endl;
    }

    void EquationGenerator::toStringList(CStringList& list)
    {
        for (int i = 0; i < m_conditions.size(); i++) {

            list.add(m_conditions[i].toString());
        }
    }

    void EquationGenerator::printDerivate(const SExpandedVariable& var)
    {

        std::wcout << var.toString() << ": ";
        for (int i = 0; i < m_conditions.size(); i++) {

            ExpandedSum sum;
            m_conditions[i].m_sum1.derv(var, sum);
            std::wcout << sum.toString() << ", ";
        }

        std::wcout << std::endl;

    }

    /*double EquationGenerator::evalFunction(const ExpandedSum& sum, const Eigen::MatrixXd &x0)
    {
       double result = 0.0;

       for (int i = 0; i < sum.m_sum.size(); i++) {
           const ExpandedProduct& prod = sum.m_sum[i];
           double currentvar = prod.m_constant;
           for (int j = 0; j< prod.getVarCount(); j++) {
               auto tind = m_var_to_index.find(prod.m_variable[j]);
               assert(tind != m_var_to_index.end());
               currentvar *= x0(tind->second);
           }

           result += currentvar;
       }
       return result;
    }*/

   /* float EquationGenerator::getValue(const ExpandedSum& sum, const SSimpleMatrix &x0)
    {
        assert (sum.m_sum.size() == 1);
       ExpandedProduct p = sum.m_sum.at(0);
       assert(p.getVarCount() == 1);
       auto tind = m_var_to_index.find(p.m_variable[0]);
       assert(tind != m_var_to_index.end());
       return x0(tind->second);
    }*/

    /*float EquationGenerator::getValue(const ExpandedSum& sum, const Eigen::MatrixXd &x0)
    {
       assert (sum.m_sum.size() == 1);
      ExpandedProduct p = sum.m_sum.at(0);
      assert(p.getVarCount() == 1);
      auto tind = m_var_to_index.find(p.m_variable[0]);
      assert(tind != m_var_to_index.end());
      return x0(tind->second);
    }*/


    EquationCompiled::EquationCompiled()
    {

    }


    /*void EquationCompiled::eval(const Eigen::MatrixXd& x, Eigen::MatrixXd& f)
    {
        for (int i = 0; i < m_equation.size(); i++) {
            EquationCompiled_Equation& eq = m_equation[i];

            double sum = 0;

            for (auto &p : eq.m_sum) {
                double prod = p.m_factor;

                for (int j = 0; j < p.m_count; j++) {
                    prod *= x(p.m_prodelement[j]);
                }

                sum += prod;
            }

            f(i) = sum;
        }
    }*/



    /*void SEquationCompiled::operator()(const Eigen::MatrixXd & x, Eigen::MatrixXd& f)
    {
        m_data->eval(x, f);
    }*/


    void EquationGenerator::createCompiledEquation(SEquationCompiled& compiled)
    {
        updateData();
        //int vc = m_variables.size();
        compiled.m_data = new EquationCompiled();
        compiled.m_data->m_equation.clear();
        for (int i = 0; i < m_conditions.size(); i++) {
           ExpandedEquation& e = m_conditions[i];
           if (e.m_condition == 0) {
               EquationCompiled_Equation* eq = new EquationCompiled_Equation();
               compiled.m_data->m_equation.push_back(eq);
               eq->m_sum.setSize(e.m_sum1.m_sum.size());
               for (int j = 0; j< e.m_sum1.m_sum.size(); j++) {
                   ExpandedProduct&pp = e.m_sum1.m_sum.at(j);
                   eq->m_sum[j].m_factor = pp.m_constant;
                   eq->m_sum[j].m_count = pp.getVarCount();
                   for (int k = 0; k < eq->m_sum[j].m_count; k++) {
                       auto itt = m_var_to_index.find(pp.m_variable[k]);
                       assert(itt != m_var_to_index.end());
                       eq->m_sum[j].m_prodelement[k] = itt->second;
                   }

               }
           }
        };

    }

}
