/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef SIMPLEMATRIX_H
#define SIMPLEMATRIX_H


#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "utiadvarray.h"
#include "utiarray.h"


namespace em {

class SSimpleMatrix;
class SSimpleVector;

class ExpandedEquation;
class ExpandedSum;

struct SExpandedVariable
{
    CUtiIDString m_basevariable;
    int m_variableindex;
    SExpandedVariable()
    {
       m_variableindex = 0;
    }
    SExpandedVariable(const CUtiIDString& base, int index)
    {
        m_basevariable = base;
        m_variableindex = index;
    }

    bool operator==(const SExpandedVariable& v) const
    {
         return m_basevariable == v.m_basevariable && m_variableindex == v.m_variableindex;
    }
    bool operator!=(const SExpandedVariable& v) const
    {
         return !(m_basevariable == v.m_basevariable && m_variableindex == v.m_variableindex);
    }
    bool operator>(const SExpandedVariable& v) const
    {
         if (m_basevariable > v.m_basevariable) return true;
         if (m_basevariable != v.m_basevariable) return false;
         return m_variableindex > v.m_variableindex;
    }
    bool operator<(const SExpandedVariable& v) const
    {
         if (m_basevariable < v.m_basevariable) return true;
         if (m_basevariable != v.m_basevariable) return false;
         return m_variableindex < v.m_variableindex;
    }

    operator ExpandedSum () const;
    ExpandedSum sum()const;
    CUtiString toString() const
    {
        return m_basevariable.str() + UTISTR("[")+m_variableindex+UTISTR("]");
    }
};


const int EQUATION_MAX_PRODUCT = 8;

class ExpandedProduct
{
public:
    double m_constant;
    StaticArray<SExpandedVariable, EQUATION_MAX_PRODUCT> m_variable;
    ExpandedProduct()
    {

    m_constant = 0;
    }

    ExpandedProduct(double c)
    {
        m_constant = c;
    }
    ExpandedProduct(const CUtiIDString& varname, int index)
    {

    m_constant = 1;
       m_variable[0].m_basevariable = varname;
       m_variable[0].m_variableindex = index;
    }
    ExpandedProduct(const CUtiIDString& varname1, int index1, const CUtiIDString& varname2, int index2)
    {

    m_constant = 1;
       m_variable[0].m_basevariable = varname1;
       m_variable[0].m_variableindex = index1;
       m_variable[1].m_basevariable = varname2;
       m_variable[1].m_variableindex = index2;
    }

    virtual CUtiString toString();

    ExpandedProduct derv(const SExpandedVariable& variable);
    double eval(std::map<SExpandedVariable, double>& vars) const;

    ExpandedProduct operator*(const ExpandedProduct& p1) const
    {
        ExpandedProduct result;
        result.m_constant = m_constant*p1.m_constant;
        int localindex = 0;
        for (int i = 0; i < m_variable.size(); i++) {
           if (m_variable[i].m_basevariable.empty() == false) {
              result.m_variable[localindex] = m_variable[i];
              localindex++;
           }
        }
        for (int i = 0; i < p1.m_variable.size(); i++) {
           if (p1.m_variable[i].m_basevariable.empty() == false) {
              result.m_variable[localindex] = p1.m_variable[i];
              localindex++;
           }
        }

        return result;
    }
    bool sameVariables(const ExpandedProduct& prod) const
    {
        int vcount = getVarCount();
        if (prod.getVarCount() != getVarCount()) {
           return false;
        }

        for (int i = 0; i < vcount; i++) {
            if (m_variable[i] == prod.m_variable[i]) {

            } else {
            return false;
            }
        }

       return true;
    }
    int getVarCount() const
    {
        for (int i = 0; i < m_variable.size(); i++) {
            if (m_variable[i].m_basevariable.empty()) {
                return i;
            }
        }

        return m_variable.size();
    }

    void fillVars(std::set<SExpandedVariable>& m_variables)
    {
        for (int i = 0; i < m_variable.size(); i++) {
            if (!m_variable[i].m_basevariable.empty()) {
                m_variables.insert(m_variable[i]);
            }
        }
    }

    void sort()
    {

        int vcount = getVarCount();
        std::sort(m_variable.begin(), m_variable.begin()+vcount);

    }
    bool operator<(const ExpandedProduct& prod) const
    {
        for (int i = 0; i < m_variable.size(); i++) {
            /*if (m_variable[i].m_basevariable.empty()) {
                return i;
            }*/
            if (m_variable[i] < prod.m_variable[i]) {

                return true;
            }
            if (m_variable[i] > prod.m_variable[i]) {

                return false;
            }
        }
        return m_constant < prod.m_constant;


    }
};

class EquationCompiled_Equation;

class ExpandedSum
{
public:
    std::vector<ExpandedProduct> m_sum;
    ExpandedSum(){};
    void sort();
    void fillVars(std::set<SExpandedVariable>& m_variables);
    ExpandedSum operator+(const ExpandedSum& prod) const;
    ExpandedSum operator+(double value) const;
    ExpandedSum operator-(const ExpandedSum& prod) const;
    ExpandedSum operator-(double value) const;
    ExpandedSum operator-() const;
    ExpandedSum& operator+=(const ExpandedSum& prod);
    ExpandedSum& operator+=(double value);

    ExpandedSum& operator-=(const ExpandedSum& prod);
    ExpandedSum& operator-=(double value);
    ExpandedSum& operator=(const ExpandedSum& prod);
    ExpandedSum operator*(const ExpandedSum& prod) const;
    ExpandedSum operator^(int v) const;
    ExpandedSum operator*(double value) const;
    ExpandedSum operator/(double value) const;
    ExpandedSum& operator*=(double value);
    ExpandedEquation operator==(const ExpandedSum& prod) const;
    ExpandedEquation operator<(const ExpandedSum& prod) const;
    ExpandedEquation operator>(const ExpandedSum& prod) const;

    virtual CUtiString toString();

    SExpandedVariable asVariable() const
    {
        assert(m_sum.size() == 1);
        assert(m_sum[0].getVarCount() == 1);
        return m_sum[0].m_variable[0];
    }

    void derv(const SExpandedVariable &variable, ExpandedSum& sum);
    double eval(std::map<SExpandedVariable, double>& vars) const;
    std::pair<double, double> getMinMax01AllVars() const
    {
        std::set<SExpandedVariable> s;
        return getMinMax01(s);
    }
    std::pair<double, double> getMinMax01(std::set<SExpandedVariable>& variables) const;

    EquationCompiled_Equation*createCompiledEquation(std::map<SExpandedVariable, int> &var_to_index);
};

ExpandedSum operator+(double value, const ExpandedSum& prod);
ExpandedSum operator*(double value, const ExpandedSum& prod);

class ExpandedEquation
{
public:
   ExpandedSum m_sum1;
   int m_condition;
   ExpandedEquation()
   {
      m_condition = 0;
   }
   //sum1 < 0 (condition < 0)
   //sum1 == 0 (condition == 0)
   //sum1 > 0 (condition > 0)
   void sort()
   {
      m_sum1.sort();
   }
   void fillVars(std::set<SExpandedVariable>& m_variables)
   {
      m_sum1.fillVars(m_variables);
   }

   virtual CUtiString toString();
};
inline ExpandedSum c(double value)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(value));
   return eq;
}

inline ExpandedSum x(int index)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(IDSTR("x"),index));
   return eq;
}

inline ExpandedSum y(int index)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(IDSTR("y"),index));
   return eq;
}

inline ExpandedSum z(int index)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(IDSTR("z"),index));
   return eq;
}

inline ExpandedSum a(int index)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(IDSTR("a"),index));
   return eq;
}

inline ExpandedSum b(int index)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(IDSTR("b"),index));
   return eq;
}
/*
inline ExpandedSum c(int index)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(IDSTR("c"),index));
   return eq;
}*/

inline ExpandedSum var(const CUtiIDString& str)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(str,0));
   return eq;
}

inline ExpandedSum var(const CUtiIDString& str, int index)
{
   ExpandedSum eq;
   eq.m_sum.push_back(ExpandedProduct(str,index));
   return eq;
}

struct SEquationCompiled;


class EquationGenerator : public ReferenceUtiOB
{
public:
    ExpandedSum m_optimisationfunction;

    std::vector<ExpandedEquation> m_conditions;
    std::set<SExpandedVariable> m_variables;
    std::map<SExpandedVariable, int> m_var_to_index;
    std::vector<SExpandedVariable> m_index_to_var;
    EquationGenerator();
    int getVarCount() const {return m_variables.size();}
    void addConditionFirst(const ExpandedEquation& ee);
    void addCondition(const ExpandedEquation& ee);
    void updateData();

    void createCompiledEquation(SEquationCompiled& compiled);
  //  float getValue(const ExpandedSum& sum, const Eigen::MatrixXd &x0);
  //  float getValue(const ExpandedSum& sum, const SSimpleMatrix &x0);
   template<class Matrix>  float getValue(const ExpandedSum& sum, const Matrix &x0)
    {
        assert (sum.m_sum.size() == 1);
       ExpandedProduct p = sum.m_sum.at(0);
       assert(p.getVarCount() == 1);
       auto tind = m_var_to_index.find(p.m_variable[0]);
       if (tind == m_var_to_index.end())  {

           return 0;//x0(p.m_variable[0].m_variableindex);
       }
       return x0(tind->second);
    }

   // double evalFunction(const ExpandedSum& sum, const Eigen::MatrixXd &x0);
    void print();
    void toStringList(CStringList& list);

    void printDerivate(const SExpandedVariable& var);


};

struct TEC_SumElement
{
    double m_factor;
    int m_count;
    StaticArray<int, EQUATION_MAX_PRODUCT> m_prodelement;
};

class EquationCompiled_Equation : public ReferenceUtiOB
{
public:
    Array<TEC_SumElement> m_sum;
};

class EquationCompiled : public ReferenceUtiOB
{
public:
    CReferenceList< EquationCompiled_Equation > m_equation;
    EquationCompiled();
    template<class Matrix> void eval(const Matrix& x, Matrix& f)
   // void EquationCompiled::eval(const SSimpleMatrix& x, SSimpleMatrix& f)
    {
        for (int i = 0; i < m_equation.size(); i++) {
            EquationCompiled_Equation& eq = m_equation[i];

            double sum = 0;

            for (auto &p : eq.m_sum) {
                double prod = p.m_factor;

                for (int j = 0; j < p.m_count; j++) {
                    prod *= x(p.m_prodelement[j]);
                }

                sum += prod;
            }

            f(i) = sum;
        }
    }

    //void eval(const Eigen::MatrixXd& x, Eigen::MatrixXd& f);
};

struct SEquationCompiled
{
    uti_refptr<EquationCompiled> m_data;
   // void operator()(const SSimpleMatrix & x, SSimpleMatrix& f);
    template<class Matrix> void operator()(const Matrix & x, Matrix& f)
    {
        m_data->eval(x, f);
    }
    //void operator()(const Eigen::MatrixXd & x, Eigen::MatrixXd& f);
};

class EquationCompiledJacobi: public ReferenceUtiOB
{
public:

    CReferenceList<EquationCompiled> m_data;
};
struct SEquationCompiledJacobi
{
    uti_refptr<EquationCompiledJacobi> m_data;
   // void operator()(const SSimpleMatrix & x, SSimpleMatrix& f);
    template<class Matrix> void operator()(const Matrix & x, Matrix& f)
    {
        int index = 0;
        for (auto& p: m_data->m_data) {

            Matrix d = f.block(0,index,f.rows(), 1);
           p.eval(x,d);
           f.block(0,index,f.rows(), 1) = d;
           index++;
        }
    }
    //void operator()(const Eigen::MatrixXd & x, Eigen::MatrixXd& f);
};


template<class ArrayType> inline double veclength(const ArrayType&ar)
{
    int n = ar.size();
    double res = 0.0;
    for (int i = 0; i < n; i++) {
        res = ar(i)*ar(i);
    }
    return sqrt(res);
}

template<class ArrayType> inline void vecscal(ArrayType&ar, double f)
{
    int n = ar.size();
    for (int i = 0; i < n; i++) {
        ar(i) *= f;
    }

}

template<class ArrayType> inline void veclineeval(const ArrayType&origin, const ArrayType& direction, double f, ArrayType& result)
{
    int n = origin.size();
    for (int i = 0; i < n; i++) {
        result(i) = origin(i)+direction(i)*f;
    }

}

const double ALF = 1.0e-4;
const double TOLX = 1.0e-7;
const double TINY = 1.0e-20;

template<class Matrix> void ludcmp(Matrix& a, int n, Array<int>& indx, double& d)
{
    double big, dum, sum, temp;
    int imax;
    Matrix vv(n,1);
    d = 1.0;
    for (int i = 0; i < n; i++) {
        big = 0.0;
        for (int j = 0; j < n; j++) {
            temp = fabs(a(i,j));
            if (temp > big) {
                big = temp;
            }

        }
        if (big == 0.0) {
            std::wcout << "Singular matrix in routine ludcmp when collecting column " << i  << std::endl;
        }
        vv(i) = 1.0 / big;
    }
    for (int j = 0; j < n; j++) {
        for (int i = 0; i < j; i++) {
            sum = a(i,j);
            for (int k = 0; k < i; k++) {
                sum -= a(i,k)*a(k,j);
            }
            a(i,j) = sum;
        }
        big = 0.0;
        for (int i = j; i < n; i++) {
            sum = a(i,j);
            for (int k = 0; k < j; k++) {
                sum -= a(i,k)*a(k,j);
            }
            a(i,j) = sum;
            dum = vv(i)*fabs(sum);
            if (dum >= big) {
                big = dum;
                imax = i;
            }
        }
        if (j != imax) {
            for (int k = 0; k < n; k++) {
                dum = a(imax, k);
                a(imax, k) = a(j,k);
                a(j,k) = dum;
            }
            d = -d;
            vv(imax) = vv(j);
        }
        indx[j]=imax;
        if (a(j,j) == 0.0) {
            a(j,j) = TINY;
        }
        if (j != n) {
            dum = 1.0/a(j,j);
            for (int i = j+1; i < n;i++) {
                a(i,j) *= dum;
            }
        }
    }
}

template<class Matrix> void ludksb(Matrix& a, int n, Array<int>& indx, Matrix& b)
{
    int ip, ii= -1;
    double sum;
    for (int i = 0; i < n; i++) {
        ip = indx[i];
        sum = b(ip);
        b(ip) = b(i);
        if (ii != -1) {
            for (int j = ii; j <=i-1; j++) {
                sum -= a(i,j)*b(j);
            }
        } else if (sum) {
            ii = i;
        }
        b(i) = sum;
    }
    for (int i = n-1; i >= 0; i--) {
        sum = b(i);
        for (int j = i+1; j < n; j++) {
            sum -= a(i,j)*b(j);
        }
        b(i) = sum/a(i,i);
    }
}


enum TNumericResultType {rtContinue=0, rtConvergence};

template<class ArrayType, class FunctionCall> void linesearch(ArrayType& xold, double fold, ArrayType& g, ArrayType& p, ArrayType& x, double &f, double stpmax, int &check, FunctionCall & call)
{

    double tmplam, rhs1, rhs2, a, b, disc, alam2, f2;
    int n = xold.size();

    check = 0;
    double sum = veclength(p);
    if (sum > stpmax) {vecscal(p, stpmax/sum);}

    double slope = 0.0;
    double test=0.0;
    for (int i = 0; i < n; i++) {
        slope += g(i)*p(i);
        double temp = fabs(p(i))/fmax(fabs(xold(i)), 1.0);
        if (temp > test) test = temp;
    }
    if (slope >= 0.0)
    {
        std::wcout << "Round off problem in linesearch" << std::endl;
    }
    double alamin = TOLX/test;
    double alam = 1.0;
    while (true) {
        veclineeval(xold, p, alam, x);
        f = call(x);
        if (alam < alamin) {
            x = xold; check = 1; return;
        } else if (f <= fold+ALF*alam*slope) {
            return;
        } else {
            if (alam == 1.0) {
                tmplam = -slope/(2.0*(f-fold-slope));

            } else {
                rhs1 = f-fold-alam*slope;
                rhs2 = f2 - fold -alam2*slope;
                a = (rhs1 / (alam*alam)-rhs2/(alam2*alam2))/(alam-alam2);
                b= (-alam2*rhs1/(alam*alam)+alam*rhs2/(alam2*alam2))/(alam-alam2);
                if (a == 0.0) {
                    tmplam = -slope/(2.0*b);
                } else {
                    disc = b*b-3.0*a*slope;
                    if (disc < 0.0) {
                        tmplam = 0.5*alam;

                    } else if (b <= 0.0) {
                        tmplam = (-b+sqrt(disc))/(3.0*a);
                    } else {
                        tmplam= -slope/(b+sqrt(disc));
                    }
                    if (tmplam > 0.5*alam) {
                        tmplam= 0.5*alam;
                    }
                }
            }
            alam2 = alam;
            f2 = f;
            alam = fmax(tmplam, 0.1*alam);
        }
    }
}

const double MAXITS = 200;
const double TOLF = 1.0e-4;
const double TOLMIN = 1.0e-6;
//const double TOLX = 1.0e-7;
const double STPMAX = 100.0;
const double EPS = 1.0e-5;

template<class Matrix, class VectorFunctionCall> void fdjac(int n, Matrix& x, Matrix& fvec, Matrix& df, VectorFunctionCall& vecfunc)
{
    Matrix f(x.rows(), x.cols());
    for (int j = 0; j < n; j++) {
        double temp = x(j);
        double h = EPS*fabs(temp);
        if (h == 0.0)h = EPS;
        x(j) = temp+h;
        h = x(j)-temp;
        vecfunc(x, f);
        x(j) = temp;
        for (int i = 0; i < n; i++) {
            df(i,j) = (f(i)-fvec(i))/h;
        }
    }
}
/*
template<class Matrix> void simplx(const Matrix& A, int m, int n, int m1, int m2, int m3, int &icase, Array<int>& izrow, Array<int>& iposv)
{

}*/

/*
template<class Matrix, class VectorFunctionCall> void newton_global(Matrix& x, int& check, VectorFunctionCall& vecfunc)
{

    int n = x.rows();
    Array<int> indx(n);
    Matrix fjac(n,n);
    Matrix g(n, 1);
    Matrix p(n, 1);
    Matrix xold(n, 1);
    Matrix fvec(n, 1);
    struct fminfunc
    {
        VectorFunctionCall* m_vecfunc;
        Matrix* m_fvec;
        int m_n;
        double operator()(const Matrix& x)
        {
            int i = 0;
            double sum = 0.0f;
            (*m_vecfunc)(x, *m_fvec);
            for (int i = 0; i < m_n; i++) {
                sum = (*m_fvec)(i)*(*m_fvec)(i);
            }
            return 0.5*sum;
        }
    };
    fminfunc fmin;
    fmin.m_vecfunc = &vecfunc;
    fmin.m_n = n;
    fmin.m_fvec = &fvec;

    double f = fmin(x);
    double fold, den;
    double test = 0.0;

    for (int i = 0; i< n; i++) {
        double v = fabs(fvec(i));
        if (v > test) test = v;

    }
    if (test < 0.01*TOLF) {
        check = 0;
        return;
    }

    double sum = veclength(x);

    double stpmax = STPMAX* fmax(sum, (double)n);
    for (int its = 1; its <= MAXITS; its++) {
        fdjac(n, x, fvec, fjac, vecfunc);

        for (int i = 0; i < n; i++) {
            sum = 0.0;
            for (int j = 0; j < n; j++) {
                sum += fjac(j,i)*fvec(j);
                g(i)= sum;
            }
        }
        xold = x;
        fold = f;
       // p = -fvec;
        for (int tk = 0; tk < n; tk++) {
            p(tk) = -fvec(tk);
        }
        double d;
       // ludcmp(fjac, n, indx, d);//LU Decomposition
       // ludksb(fjac, n, indx, p);

        //Eigen::BiCGSTAB<Matrix > solver;
      //  solver.compute(fjac);
       //7 solver.setMaxIterations(5000);
      //  Eigen::MatrixXd sol = solver.solve(p);

      //  p =sol;
       // std::wcout << "#iterations:     " << solver.iterations() << std::endl;
       // std::wcout << "estimated error: " << solver.error()      << std::endl;

        Eigen::MatrixXd sol = fjac.fullPivLu().solve(p);
        p =sol;
        //Test 11
        linesearch(xold, fold, g, p, x, f, stpmax, check, fmin);
        test = 0.0;
        for (int i = 0; i < n; i++) {
            double v = fabs(fvec(i));
            if (v > test) test = v;
        }
        if (test < TOLF) {
            check = 0;
            return;
        }
        if (check) {
            test = 0.0;
            den = fmax(f, 0.5*double(n));
            for (int i = 0; i < n; i++) {
                double temp = fabs(g(i))*fmax(fabs(x(i)), 1.0)/den;
                if (temp > test) {
                    test = temp;
                }
            }
            check = (test < TOLMIN ? 1 : 0);
            return;
        }
        test = 0.0;
        for (int i = 0; i < n; i++)
        {
            double temp =(fabs(x(i)-xold(i)))/fmax(fabs(x(i)), 1.0);
            if (temp > test) test = temp;
        }
        if (test < TOLX) return;

    }
    std::wcout << "Too many steps!" << std::endl;
}*/

template<class Matrix> void rsolv(Matrix& a, int n, Matrix& d, Matrix& b)
{
    int i,j;
    double sum;
    b(n-1) /= d(n-1);
    for (int i = n-2; i >= 0; i--) {
        sum = 0.0;
        for (int j = i+1; j < n; j++) {
            sum += a(i,j)*b(j);
            b(i) = (b(i)-sum)/d(i);
        }
    }
}

template<class Matrix> void qrsolv(Matrix& a, int n, Matrix& c, Matrix& d, Matrix& b)
{
    double sum, tau;
    for (int j = 0; j < n-1; j++) {
        sum = 0.0;
        for (int i = j; i < n; i++) {
            sum += a(i,j)*b(i);
        }
        tau = sum / c(j);
        for (int i = j; i < n; i++) {
            b(i) -= tau*a(i,j);
        }
    }
    rsolv(a,n, d, b);
}
template<class Matrix> void qrdcmp(Matrix& a, int n, Matrix& c, Matrix& d, int &sing)
{

    double scale, sigma, sum, tau;
    sing = 0;
    for (int k = 0; k < n-1; k++) {
        scale = 0.0;
        for (int i = k; i < n; i++) {
            scale = std::fmax(scale, fabs(a(i,k)));
        }
        if (scale == 0.0) {
            sing = 1;
            c(k) = d(k) = 0.0;
        } else {
            for (int i = k; i < n; i++) {
                a(i,k) /= scale;
            }
            sum = 0.0;
            for (int i = k; i < n; i++) {
                sum += a(i,k)*a(i,k);
            }
            sigma = std::copysign(sqrt(sum), a(k,k));
            a(k,k) += sigma;
            c(k) = sigma*a(k,k);
            d(k) = -scale * sigma;
            for (int j = k+1;j < n; j++) {
                sum = 0.0;
                for (int i = k; i < n; i++) {
                    sum += a(i,k)*a(i,j);

                }
                tau = sum /c(k);
                for (int i = k; i < n; i++) {
                    a(i,j) -= tau*a(i,k);
                }
            }
        }
    }
    d(n-1) = a(n-1,n-1);
    if (d(n-1) == 0.0) sing = 1;
}

template<class Matrix> void rotate(Matrix& r, Matrix& qt, int n, int i, double a, double b)
{

    double c, fact, s, w, y;
    if (a == 0.0) {
        c = 0.0;
        s = (b >= 0.0 ? 1.0 : -1.0);
    } else if (fabs(a) > fabs(b)) {
        fact = b/a;
        c = std::copysign(1.0/sqrt(1.0+(fact*fact)), a);
        s = fact*c;
    } else {
        fact = a/b;
        s = std::copysign(1.0/sqrt(1.0+(fact*fact)), b);
        c = fact*s;
    }
    for (int j = i; j < n; j++) {
        y = r(i,j);
        w= r(i+1, j);
        r(i,j)=c*y-s*w;
        r(i+1,j)=s*y+c*w;
    }
    for (int j = 0; j < n; j++) {
        y = qt(i,j);
        w = qt(i+1,j);
        qt(i,j) = c*y-s*w;
        qt(i+1, j) = s*y+c*w;
    }
}

template<class Matrix> void qrupdt(Matrix& r, Matrix& qt, int n, Matrix& u, Matrix& v)
{
    int k = 0;
    for (int k1 = n-1; k1 >= 0; k1--)
    {
        k = k1;
        if (u(k) != 0.0) break;
    }

    for (int i = k-1; i >= 0; i++) {
        rotate(r, qt, n, i, u(i), -u(i+1));
        if (u(i) == 0.0) {
            u(i) = fabs(u(i+1));
        } else if (fabs(u(i)) > fabs(u(i+1))) {
            u(i) = fabs(u(i))*sqrt(1.0+sqr(u(i+1)/u(i)));
        } else {

            u(i) = fabs(u(i+1))*sqrt(1.0+sqr(u(i)/u(i+1)));
        }
    }
    for (int j = 0; j < n; j++) {
        r(0,j) += u(0)*v(j);
    }
    for (int i = 0; i < k; i++) {
        rotate(r, qt, n, i, r(i,i), -r(i+1, 1));
    }
}

template<class Matrix, class VectorFunctionCall> void brody_global(Matrix& x, int& check, VectorFunctionCall& vecfunc)
{
    int sing;

    double fold, f, den;
    int n = x.rows();
    Matrix c(n,1);
   Matrix d(n,1);
    Matrix fvcold(n, 1);
    Matrix g(n, 1);
    Matrix p(n, 1);
    Matrix qt(n, n);
    Matrix r(n, n);
    Matrix s(n, n);
    Matrix t(n, n);
    Matrix w(n, n);
    Matrix xold(n, 1);
    Matrix fvec(n, 1);
    struct fminfunc
    {
        VectorFunctionCall* m_vecfunc;
        Matrix* m_fvec;
        int m_n;
        double operator()(const Matrix& x)
        {
            int i = 0;
            double sum = 0.0f;
            (*m_vecfunc)(x, *m_fvec);
            for (int i = 0; i < m_n; i++) {
                sum = (*m_fvec)(i)*(*m_fvec)(i);
            }
            return 0.5*sum;
        }
    };
    fminfunc fmin;
    fmin.m_vecfunc = &vecfunc;
    fmin.m_n = n;
    fmin.m_fvec = &fvec;

    double test = 0.0;
    for (int i = 0; i< n; i++) {
        double v = fabs(fvec(i));
        if (v > test) test = v;

    }
    if (test < 0.01*TOLF) {
        check = 0;
        return;
    }

    double sum = veclength(x);

    double stpmax = STPMAX* fmax(sum, (double)n);
    int restrt = 1;
    for (int its = 0; its < MAXITS; its++) {
        if (restrt) {
            fdjac(n, x, fvec, r, vecfunc);
            qrdcmp(r,n, c, d, sing);
            if (sing) {
                std::wcout << "singular Jacobian in broydn" << std::endl;
            }
            for (int i = 0; i < n; i++) {
                for (int j = 0; j < n; j++) {
                    qt(i,j) = 0.0;

                }
                qt(i,i) = 1.0;
            }
            for (int k = 0; k < n; k++) {
                if (c(k)) {
                    for (int j = 0; j < n; j++) {
                        double sum = 0;
                        for (int i = k; i < n; i++) {
                            sum += r(i,k)*qt(i,j);
                        }
                        sum /= c(k);
                        for (int i = k; i < n; i++) {
                            qt(i,j) -= sum*r(i,k);
                        }
                    }
                }
            }
            for (int i = 0; i < n; i++) {
                r(i,i) = d(i);
                for (int j = 0; j < i; j++) {
                     r(i,j)= 0.0;
                }
            }
        } else {
            for (int i = 0; i < n; i++) {
                s(i) = x(i) - xold(i);
            }

            for (int i = 0; i < n; i++) {
                double sum = 0.0;
                for (int j = i; j < n; j++) {
                    sum += r(i,j)*s(j);
                    t(i) = sum;
                }
            }
            int skip = 1;
            for (int i = 0; i < n; i++) {
                double sum = 0.0;

                for (int j = 0; j < n; j++) {
                    sum += qt(j,i)*t(j);
                }
                w(i) = fvec(i)-fvcold(i)-sum;
                if (fabs(w(i)) >= EPS* (fabs(fvec(i))+fabs(fvcold(i)))) {
                    skip = 0;
                }else {
                    w(i) = 0.0;
                }
            }
            if (!skip) {
                for (int i = 0; i<n; i++) {
                    double sum = 0.0;
                    for (int j = 0; j < n; j++) {
                        sum += qt(i,j)*w(j);
                        t(i) = sum;
                    }
                }
                double den = 0.0;
                for (int i = 0; i < n; i++) {
                    den += s(i)*s(i);
                }
                for (int i = 0; i < n; i++) {
                    s(i) /= den;
                }
                qrupdt(r, qt, n, t, s);
                for (int i = 0; i < n; i++) {
                    if (r(i,i) == 0.0) {
                        std::wcout << "r singular in broydn" << std::endl;
                        d(i) = r(i,i);
                    }
                }
            }
            for (int i = 0; i < n; i++) {
                double sum = 0.0;
                for (int j = 0; j < n; j++) {
                    sum += qt(i,j)*fvec(j);
                }
                g(i) = sum;
            }
            for (int i = 0; i < n; i++) {
                double sum = 0.0;
                for (int j = 0; j < n; j++) {
                    sum += r(i,j)*g(j);
                }
                g(i) = sum;
            }
            for (int i = 0; i < n; i++) {
                xold(i) = x(i);
                fvcold(i) = fvec(i);
            }
            fold = f;
            for (int i = 0; i < n; i++) {
                double sum = 0.0;
                for (int j = 0; j < n; j++) {
                    sum += qt(i,j)*fvec(j);
                }
                p(i) = -sum;
            }
            rsolv(r, n, d, p);
            linesearch(xold, fold, g, p, x, f, stpmax, check, fmin);
            test = 0.0;
            for (int i = 0; i < n; i++) {
                double v = fabs(fvec(i));
                if (v > test) test = v;
            }
            if (test < TOLF) {
                check = 0;
                return;
            }
            if (check) {
                if (restrt) {return;}
                else {
                test = 0.0;
                den = fmax(f, 0.5*double(n));
                for (int i = 0; i < n; i++) {
                    double temp = fabs(g(i))*fmax(fabs(x(i)), 1.0)/den;
                    if (temp > test) {
                        test = temp;
                    }
                }
                if (test < TOLMIN){ return;} else{restrt = 1;}

                 }
            } else {
                restrt = 0;
               test = 0.0;
               for (int i = 0; i < n; i++)
               {
                  double temp =(fabs(x(i)-xold(i)))/fmax(fabs(x(i)), 1.0);
                   if (temp > test) test = temp;
               }
               if (test < TOLX) return;
            }

        }

    }
}

class SSimpleVector : public ReferenceUtiOB
{
public:
    Array<double> m_data;
    inline SSimpleVector(int n) {
        m_data.setSize(n);
    }
    inline int rows() const
    {
        return m_data.size();
    }
    inline int cols() const
    {
        return 1;

    }
    inline int size() const
    {
        return m_data.size();
    }
    inline double& operator()(int index)
    {
        return m_data[index];
    }
    inline const double& operator()(int index) const
    {
        return m_data[index];
    }
};

class SSimpleMatrix : public ReferenceUtiOB
{
public:
    int m_icount, m_jcount;
    Array<double> m_data;
    inline SSimpleMatrix(int icount, int jcount) {
        m_data.setSize(icount*jcount);
        m_icount = icount;
        m_jcount = jcount;
    }
    inline SSimpleMatrix() {

        m_icount = 0;
        m_jcount = 0;
    }
    inline int rows() const
    {
        return m_icount;
    }
    inline int cols() const
    {
        return m_jcount;

    }
    inline int size() const
    {
        return m_icount*m_jcount;
    }

    inline double& operator()(int i, int j)
    {
        return m_data[i*m_jcount + j];
    }
    inline const double& operator()(int i, int j) const
    {
        return m_data[i*m_jcount + j];
    }
    inline double& operator()(int index)
    {
        return m_data[index];
    }
    inline const double& operator()(int index) const
    {
        return m_data[index];
    }
    inline SSimpleMatrix& operator=(const SSimpleMatrix& mat)
    {
        assert(m_icount == mat.m_icount);
        assert(m_jcount == mat.m_jcount);
        int c = m_icount* m_jcount;
        for (int i = 0; i < c; i++) {
            m_data[i] = mat.m_data[i];
        }
        return *this;
    }
};

template<class Matrix> void matrixmul(const Matrix& a, const Matrix& b, Matrix& res)
{
    assert(a.m_icount == res.m_icount);
    assert(b.m_jcount == res.m_jcount);
    assert(a.m_jcount == b.m_jcount);

    for (int i = 0; i < res.m_icount; i++) {
        for (int j = 0; j < res.m_jcount; j++) {
            double sum = 0.0;

            for (int k = 0; k < res.m_jcount; k++) {
                sum += a(i, k) * b(k, j);
            }

            res(i,j) = sum;
        }
    }
}



struct SSparseMatrix :public ReferenceUtiOB
{
    int m_n;
    Array<double> m_sa;

    Array<unsigned long> m_ija;
    inline SSparseMatrix(int n) {
       // m_data.setSize(n);
        m_ija.setSize(n+1);
        m_sa.setSize(n+1);
    }
};

template<class Matrix> bool sprsin(const Matrix& a, int n, double thresh, SSparseMatrix* result)
{
    unsigned long k;
    for (int j = 0; j < n; j++) {
        result->m_sa[j] = a(j,j);
    }
    result->m_ija[0] = n+2;
    k = n+1;
    for (int i = 0; i < n; i++) {
        for (int j = 0; j < n; j++) {
            if (fabs(a(i,j)) >= thresh && i != j) {
                k++;
                if (result->m_sa.size() <= k) {
                    result->m_sa.setSizeCopyHighSpeed(result->m_sa.size()*2);
                    result->m_ija.setSizeCopyHighSpeed(result->m_ija.size()*2);
                }
                result->m_sa[k] = a(i,j);
                result->m_ija[k]= j+1;
            }

        }
        result->m_ija[i+1+1]=k+1;
    }
    return true;
}



template<class Matrix> bool sprsax(SSparseMatrix* matrix,const Matrix& x, const Matrix& b)
{
   int n = matrix->m_n;
   assert(matrix->m_ija[1] == n+2);

   for (int i = 0; i < n; i++) {
       b(i) = matrix->m_sa[i]*x(i);
       for (int k = matrix->m_ija[i+1]; k <= matrix->m_ija[i+1+1]-1; k++) {
           b(i) += matrix->m_sa[k]*x(matrix->m_ija[k]-1);
       }
   }
}

template<class Matrix> bool sprstx(SSparseMatrix* matrix,const Matrix& x, const Matrix& b)
{
   int n = matrix->m_n;
   assert(matrix->m_ija[1] == n+2);

   for (int i = 0; i < n; i++) {
       b(i) = matrix->m_sa[i]*x(i);
   }
   for (int i = 0; i < n; i++) {
       for (int k = matrix->m_ija[i+1]; k <= matrix->m_ija[i+1+1]-1; k++) {
           int j = matrix->m_ija[k]-1;
           b(j) += matrix->m_sa[k]*x(i);
       }
   }
}

template<class Matrix> double snrm(unsigned long n, Matrix& sx, int itol)
{
    unsigned long i, isamax;
    double ans;

    if (itol <= 3) {
        ans = 0.0;
        for (int i = 0; i < n; i++) {
            ans += sx(i)*sx(i);
        }
        return sqrt(ans);
    } else {
        isamax=0;
        for (int i = 0; i < n; i++) {
            if (fabs(sx(i))>fabs(sx(isamax))) isamax = i;
        }
        return fabs(sx(isamax));
    }
}



template<class Matrix> void linbcg(SSparseMatrix* matrix, unsigned int n, Matrix& b, Matrix& x, int itol, double tol, int itmax, int &iter, double &err)
{
    struct SATimesType {
         SSparseMatrix* m_matrix;
         void operator()(unsigned long n, Matrix& x, Matrix& r, int itrnsp)
         {
             if (itrnsp) {
                 sprstx(m_matrix, x, r);
             } else {
                 sprsax(m_matrix, x, r);
             }
         }
    };
    struct SASolveType {
        SSparseMatrix* m_matrix;
        void operator()(unsigned long n, Matrix& b, Matrix&x, int itrnsp)
        {
            for (int i = 0; i < n; i++) {
                x(i) = (m_matrix->m_sa[i] != 0.0 ? b(i)/m_matrix->m_sa[i] : b(i));
            }
        }
   };

    SATimesType atimes;
    atimes.m_matrix = matrix;
    SASolveType asolve;
    asolve.m_matrix = matrix;

    double ak, akden, bk, bkden, bknum, bnrm, dxnrm, xnrm, zminrm, znrm;
    Matrix p(n,1);
    Matrix pp(n,1);
    Matrix r(n,1);
    Matrix rr(n,1);
    Matrix z(n,1);
    Matrix zz(n,1);
    iter  = 0;
    atimes(n,x, r, 0);
    for (int j = 0; j < n; j++) {
        r(j) = b(j)-r(j);
        rr(j) = r(j);
    }
    // atimes(n,r,rr, 0);
    if (itol == 1) {
        bnrm = snrm(n, b, itol);
        asolve(n,r,z,0);
    } else if (itol == 2) {
        asolve(n,b,z,0);
        bnrm = snrm(n, z, itol);
        asolve(n,r,z,0);
    } else if (itol == 3 || itol == 4) {
        asolve(n,b,z,0);
        bnrm = snrm(n, z, itol);
        asolve(n,r,z,0);

        znrm = snrm(n, z, itol);
    } else {
        assert(false);
    }

    while (iter < itmax) {
        iter++;
        asolve(n, rr, zz, 1);
        bknum = 0.0;
        for (int j = 0; j < n; j++) {
            bknum += z(j)*rr(j);
        }
        if (iter == 1) {
            for (int j = 0; j < n; j++) {
                p(j) = z(j);
                pp(j)= zz(j);
            }
        } else {
            bk = bknum/bkden;
            for (int j = 0; j < n; j++) {
                p(j) = bk*p(j)+z(j);
                pp(j) = bk*pp(j)+zz(j);
            }
        }
        bkden = bknum;
        atimes(n,p, z, 0);
        akden = 0.0;
        for (int j = 0; j < n; j++) {
            akden += z(j)*pp(j);
        }
        ak = bknum/akden;
        atimes(n, pp, zz, 1);
        for (int j = 0; j < n; j++) {

            x(j) += ak*p(j);
            r(j) -= ak*z(j);
            rr(j) -= ak*zz(j);
        }
        asolve(n, r, z, 0);
        if (itol == 1) {
            err = snrm(n, r, itol)/bnrm;

        }else if (itol == 2) {
            err = snrm(n, z, itol)/bnrm;

        } else if (itol == 3 || itol == 4) {

            zminrm = znrm;
            znrm = snrm(n,z,itol);
            if (fabs(zminrm-znrm) > EPS*znrm) {
                dxnrm = fabs(ak)*snrm(n,p, itol);
                err = znrm/fabs(zminrm-znrm)*dxnrm;
            } else {
                err = znrm /bnrm;
                continue;
            }
            xnrm = snrm(n, x, itol);
            if (err <= 0.5*xnrm) {
                err /= xnrm;
            } else {
                err = znrm/bnrm;
                continue;
            }
        }
        std::wcout << "iter = " << iter << ", err = " << err << std::endl;
        if (err <= tol) break;
    }
}

const double MACH4_EPS = 1e-7;
const double MACH_EPS = 1e-7;

template<class Matrix> int ueber_system(Matrix& a, int m, int n, Matrix&x)
{
    Matrix d(n, 1);

    double radiant, alpha, ak, factor, sum;
    if (m < n) return 1;
    if (n <= 0) return 2;

    for (int i = 0; i < n; i++) {
        radiant = sqr(a(i,i));
        for (int k = i+1; k < m; k++) {
            radiant =sqr(a(k,i));
        }
        if (radiant < MACH4_EPS) return 4;
        alpha = sqrt(radiant) * std::copysign(a(i,i))*1.0;
        ak = 1.0/(radiant + alpha * a(i,i));
        a(i,i) += alpha;

        d(i) = -alpha;
        for (int k = i+1; k < n+1; k++) {
            factor = 0.0;
            for (int j = i; j < m; j++) {
                factor += a(j,k)*a(j,i);
            }
            factor *= ak;
            for (int j = i; j < m; j++) {
                a(j,k) -= factor * a(j,i);
            }
        }

    }
    //Rueckwartssubsitution
    for (int i = n-1; i>= 0; i--) {
        sum = 0.0;
        for (int k = i+1; k < n; k++) {
            sum += a(i,k)*x(k);
        }
        x(i) = (a(i,n+1) - sum) /d(i);
    }
    return 0;
}
template<class Matrix> int seidel(int ITERMAX, int krit, int n, Matrix & matrix, Matrix& b, double omega, Matrix&x, Matrix& residu, int& iter)
{
    int res = 0;
    double temp;

    iter = 0;

    if (n < 2 || omega <= 0.0 || omega >= 2.0) return 1;

    for (int i = 0; i<n; i++) {
        if (matrix(i,i) == 0.0) return 3;
        temp = 1.0 / matrix(i,i);
        for (int j = 0; j < n; j++) {
            matrix(i,j) *= temp;
        }
        b(i) *= temp;
    }

    switch (krit) {
       case 1:
        for (int i = 0; i < n; i++) {
            temp = 0.0;
            for (int j = 0; j < n; j++) {
                temp += fabs(matrix(i,j));
            }
            if (temp >= 2.0) return 11;
        }
         break;
       case 2:
        for (int j = 0; j < n; j++) {
            temp = 0.0;
            for (int i = 0; i < n; i++) {
                temp += fabs(matrix(i,j));
            }
            if (temp >= 2.0) return 12;
        }
         break;
       case 3:
        temp = 0.0;
        for (int i = 0; i < n; i++) {

            for (int j = 0; j < n; j++) {
                temp += sqr(matrix(i,j));
            }

        }
        temp = sqrt(temp-1.0);
        if (temp >= 1.0) return 13;
         break;
    default: break;
    }
    for (int i = 0; i < n; i++) residu(i) =x(i);

    while (iter <= ITERMAX) {
        iter++;
        for (int i = 0; i < n; i++) {
            temp = b(i);
            for (int j = 0; j < n; j++) {
                temp -= matrix(i,j)*residu(j);
            }
            residu(i) += omega* temp;
        }
        for (int i = 0; i < n; i++) {
            temp = fabs(x(i)-residu(i));
            if (temp <= MACH_EPS* 16.0)
            {
                x(i) = residu(i);
                res = 0;
            } else {
                for (int j = 0; j < n; j++) {
                    x(j) = residu(j);

                }
                res = 4;
                break;
            }
        }
        if (res == 0) break;

        for (int i = 0; i< n; i++) {
            temp = b(i);
            for (int j = 0; j < n; j++) {
                temp -= matrix(i,j)*x(j);
            }
            residu(i) = temp;
        }
    }
    return res;

}
/*
template<class Matrix> int gauss(int cas, int n, Matrix& matrix, Matrix& lumat, Matrix& perm, Matrix& b, Matrix& x, int &signdet)
{
    if (n < 2)
    {
        return 1;
    }
    switch(cas) {
    case 0: {
        int res = gaudec(n, matrix, lumat, perm, signdet);
        if (res == 0) {
            return gausol(n, lumat, perm, b, x);
        } else {
            return res;
        }
    }; break;
    case 1: {
        return gaudec(n, matrix, lumat, perm, signdet);
    }; break;
    case 2: {
        return gausol(n, lumat, perm, b, x);
    }; break;
    }
    return 5;
}

template<class Matrix> int gaudec(int n, Matrix& matrix, Matrix& lumat,  Matrix& perm, int &signdet)
{

    int j0;
    double piv, temp, zmax, abslu;

}*/

}

#endif // SIMPLEMATRIX_H
