/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __STRUCTURES__
#define __STRUCTURES__
#include "Uti3DCore.h"
#include "UtiString.h"



typedef int STRUCTURENAME;

enum TLogicType {ltNone, ltPredicate, ltNamedPredicate, ltVariable, ltInt, ltFloat, ltString, ltBool, ltReferenceUtiOB};

const STRUCTURENAME UTI_LOGIC_VAR=-1;

class UtiPredicateLogic : public CUtiOB
{
  public:
    TLogicType m_type;
   uti_refptr<ReferenceUtiOB> m_object;
  public:
    UtiPredicateLogic();
    //bool operator==(UtiPredicateLogic& logic);
    bool unify(UtiPredicateLogic logic); //==
    void mustUnify(UtiPredicateLogic logic);//=
    bool canUnify(UtiPredicateLogic logic);
    bool isInt();
    bool isFloat();
    bool isBool();
    bool isString();
    bool isUnBoundVar();
    bool isBoundVar();
    bool isPredicate();
    bool isReferenceUtiOB();
    TLogicType getType();
    CUtiString toString();
    STRUCTURENAME getName();
    UtiPredicateLogic getPredicate(int i);
};

UtiPredicateLogic ulint(int value);
UtiPredicateLogic ulfloat(double value);
UtiPredicateLogic ulstring(const CUtiString& str);
UtiPredicateLogic ulbool(bool b);
UtiPredicateLogic ulvar();
UtiPredicateLogic ulrefobj(ReferenceUtiOB* obj);
UtiPredicateLogic ulpredicate(int name);
UtiPredicateLogic ulpredicate(int name, UtiPredicateLogic param1);
UtiPredicateLogic ulpredicate(int name, UtiPredicateLogic param1, UtiPredicateLogic param2);
UtiPredicateLogic ulpredicate(int name, UtiPredicateLogic param1, UtiPredicateLogic param2, UtiPredicateLogic param3);
UtiPredicateLogic ulpredicate(int name, UtiPredicateLogic param1, UtiPredicateLogic param2, UtiPredicateLogic param3, UtiPredicateLogic param4);



class UtiPredicateIntern : public ReferenceUtiOB
{
  public:
  STRUCTURENAME m_name;
    Array<UtiPredicateLogic> m_children;
};

class UtiVariableIntern : public ReferenceUtiOB
{
  public:
    UtiPredicateLogic m_target;
};


class UtiIntIntern : public ReferenceUtiOB
{
   public:
     int m_value;
};

class UtiFloatIntern : public ReferenceUtiOB
{
   public:
     double m_value;
};

class UtiStringIntern : public ReferenceUtiOB
{
   public:
     CUtiString m_value;
};

class UtiBoolIntern : public ReferenceUtiOB
{
   public:
     bool m_value;
};


#endif
