/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UTIVECTORDOUBLE__
#define __UTIVECTORDOUBLE__
#include <math.h>
#include "utiintvector.h"
template<class Scalar> struct CVector2Template {
Scalar x;
Scalar y;
inline CVector2Template(){x=0;y=0;};
inline CVector2Template(const Scalar f) {
x = f;
y = f;
};
inline CVector2Template(const Scalar arg0, const Scalar arg1){
x = arg0;y = arg1;
}
inline Scalar &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 2);
return (&x)[p];
}
inline const Scalar &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 2);
    return (&x)[p];
}
inline CVector2Template<Scalar> xx() const
{
return CVector2Template<Scalar>(x, x);
}
inline CVector2Template<Scalar> yx() const
{
return CVector2Template<Scalar>(y, x);
}
inline CVector2Template<Scalar> xy() const
{
return CVector2Template<Scalar>(x, y);
}
inline CVector2Template<Scalar> yy() const
{
return CVector2Template<Scalar>(y, y);
}
inline void operator += (const CVector2Template<Scalar> &v)
{
x+=v.x;
y+=v.y;
}
inline void operator -= (const CVector2Template<Scalar> &v)
{
x-=v.x;
y-=v.y;
}
inline void operator *= (const Scalar v)
{
x*=v;
y*=v;
}
inline void operator *= (const CVector2Template<Scalar> &v)
{
x*=v.x;
y*=v.y;
}
inline void operator /= (const Scalar v)
{
x/=v;
y/=v;
}
inline void operator /= (const CVector2Template<Scalar> &v)
{
x/=v.x;
y/=v.y;
}
inline void fromTo(const CVector2Template<Scalar> & start, const CVector2Template<Scalar> & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
};
void setVector(Scalar arg0, Scalar arg1)
{
x = arg0;
y = arg1;
}
inline void addVector(Scalar arg0, Scalar arg1)
{
x += arg0;
y += arg1;
}
inline void subVector(Scalar arg0, Scalar arg1)
{
x -= arg0;
y -= arg1;
}
inline void neg()
{
x = -x;
y = -y;
}
inline Scalar length() const
{
return sqrt(x*x+y*y);
}
inline Scalar normalize()
{
Scalar r = sqrt(x*x+y*y);
   if (r < 0.00001) return r;
x /= r;
y /= r;
  return r;
}
inline Scalar dotProduct(const CVector2Template<Scalar> &b) const
{
return (x* b.x+y* b.y);
}
inline int getMajorAxis() const
{
Scalar current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
return index;
}
inline Scalar distanceTo(const CVector2Template<Scalar> &v) const;
};
template<class Scalar> inline CVector2Template<Scalar> operator +(const CVector2Template<Scalar> &v1, const CVector2Template<Scalar> &v2)
{
CVector2Template<Scalar> r;
r.x = v1.x + v2.x;
r.y = v1.y + v2.y;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator +(const CVector2Template<Scalar> &v1, const Scalar v2)
{
CVector2Template<Scalar> r;
r.x = v1.x + v2;
r.y = v1.y + v2;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator -(const CVector2Template<Scalar> &v1, const CVector2Template<Scalar> &v2)
{
CVector2Template<Scalar> r;
r.x = v1.x - v2.x;
r.y = v1.y - v2.y;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator -(const CVector2Template<Scalar> &v1, const Scalar v2)
{
CVector2Template<Scalar> r;
r.x = v1.x - v2;
r.y = v1.y - v2;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator -(const CVector2Template<Scalar> &v1)
{
CVector2Template<Scalar> r;
r.x = -v1.x;
r.y = -v1.y;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator *(const CVector2Template<Scalar> &v1, const CVector2Template<Scalar> &v2)
{
CVector2Template<Scalar> r;
r.x = v1.x * v2.x;
r.y = v1.y * v2.y;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator *(const CVector2Template<Scalar> &v1, const Scalar v2)
{
CVector2Template<Scalar> r;
r.x = v1.x * v2;
r.y = v1.y * v2;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator *(const Scalar v1, const CVector2Template<Scalar> &v2)
{
CVector2Template<Scalar> r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator /(const CVector2Template<Scalar> &v1, const CVector2Template<Scalar> &v2)
{
CVector2Template<Scalar> r;
r.x = v1.x / v2.x;
r.y = v1.y / v2.y;
return r;
}
template<class Scalar> inline CVector2Template<Scalar> operator /(const CVector2Template<Scalar> &v1, const Scalar v2)
{
CVector2Template<Scalar> r;
r.x = v1.x / v2;
r.y = v1.y / v2;
return r;
}
template<class Scalar> inline bool operator ==(const CVector2Template<Scalar> &v1, const CVector2Template<Scalar> &v2)
{
bool r;
r =  v1.x == v2.x
 &&
 v1.y == v2.y
;
return r;
}
template<class Scalar> inline bool operator !=(const CVector2Template<Scalar> &v1, const CVector2Template<Scalar> &v2)
{
bool r;
r =  v1.x != v2.x
 ||
 v1.y != v2.y
;
return r;
}
template<class Scalar> inline Scalar CVector2Template<Scalar>::distanceTo(const CVector2Template<Scalar> &v) const
{
return (v-*this).length();
}
template<class Scalar> inline Scalar dot(const CVector2Template<Scalar> &a, const CVector2Template<Scalar> &b)
{
return (a.x* b.x+a.y* b.y);
}
template<class Scalar> inline CVector2Template<Scalar> umin(const CVector2Template<Scalar> &a, const CVector2Template<Scalar> &b)
{
CVector2Template<Scalar> r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
return r;
}
template<class Scalar> inline CVector2Template<Scalar> umax(const CVector2Template<Scalar> &a, const CVector2Template<Scalar> &b)
{
CVector2Template<Scalar> r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
return r;
}
template<class Scalar> inline CVector2Template<Scalar> normalize(const CVector2Template<Scalar> &a)
{
CVector2Template<Scalar> rvec;
Scalar r = sqrt(a.x* a.x+a.y* a.y);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
  return rvec;
}
template<class Scalar> inline Scalar length(const CVector2Template<Scalar> &a)
{
CVector2Template<Scalar> rvec;
Scalar r = sqrt(a.x*a.x+a.y*a.y);
   return r;
}
template<class Scalar> inline CVector2Template<Scalar> sign(const CVector2Template<Scalar> &a)
{
CVector2Template<Scalar> rvec;
rvec.x = a.x > 0 ? 1 : -1;
rvec.y = a.y > 0 ? 1 : -1;
  return rvec;
}
template<class Scalar> inline Scalar distance(const CVector2Template<Scalar> &a, const CVector2Template<Scalar> &b)
{
return sqrt(sqr(a.x-b.x)+sqr(a.y-b.y));
}
template<class Scalar> inline CVector2Template<Scalar> linear(const CVector2Template<Scalar> &a, const CVector2Template<Scalar> &b, const Scalar v)
{
CVector2Template<Scalar> rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
return rvec;
}
template<class Scalar> struct CVector3Template {
Scalar x;
Scalar y;
Scalar z;
inline CVector3Template(){x=0;y=0;z=0;};
inline CVector3Template(const Scalar f) {
x = f;
y = f;
z = f;
};
inline CVector3Template(const Scalar arg0, const Scalar arg1, const Scalar arg2){
x = arg0;y = arg1;z = arg2;
}
inline CVector3Template(const Scalar arg0, const CVector2Template<Scalar> & arg1){
x = arg0;y = arg1.x;z = arg1.y;
}
inline CVector3Template(const CVector2Template<Scalar> & arg0, const Scalar arg1){
x = arg0.x;y = arg0.y;z = arg1;
}
inline Scalar &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 3);
return (&x)[p];
}
inline const Scalar &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 3);
    return (&x)[p];
}
inline CVector2Template<Scalar> xx() const
{
return CVector2Template<Scalar>(x, x);
}
inline CVector2Template<Scalar> yx() const
{
return CVector2Template<Scalar>(y, x);
}
inline CVector2Template<Scalar> zx() const
{
return CVector2Template<Scalar>(z, x);
}
inline CVector2Template<Scalar> xy() const
{
return CVector2Template<Scalar>(x, y);
}
inline CVector2Template<Scalar> yy() const
{
return CVector2Template<Scalar>(y, y);
}
inline CVector2Template<Scalar> zy() const
{
return CVector2Template<Scalar>(z, y);
}
inline CVector2Template<Scalar> xz() const
{
return CVector2Template<Scalar>(x, z);
}
inline CVector2Template<Scalar> yz() const
{
return CVector2Template<Scalar>(y, z);
}
inline CVector2Template<Scalar> zz() const
{
return CVector2Template<Scalar>(z, z);
}
inline CVector3Template<Scalar> xxx() const
{
return CVector3Template<Scalar>(x, x, x);
}
inline CVector3Template<Scalar> yxx() const
{
return CVector3Template<Scalar>(y, x, x);
}
inline CVector3Template<Scalar> zxx() const
{
return CVector3Template<Scalar>(z, x, x);
}
inline CVector3Template<Scalar> xyx() const
{
return CVector3Template<Scalar>(x, y, x);
}
inline CVector3Template<Scalar> yyx() const
{
return CVector3Template<Scalar>(y, y, x);
}
inline CVector3Template<Scalar> zyx() const
{
return CVector3Template<Scalar>(z, y, x);
}
inline CVector3Template<Scalar> xzx() const
{
return CVector3Template<Scalar>(x, z, x);
}
inline CVector3Template<Scalar> yzx() const
{
return CVector3Template<Scalar>(y, z, x);
}
inline CVector3Template<Scalar> zzx() const
{
return CVector3Template<Scalar>(z, z, x);
}
inline CVector3Template<Scalar> xxy() const
{
return CVector3Template<Scalar>(x, x, y);
}
inline CVector3Template<Scalar> yxy() const
{
return CVector3Template<Scalar>(y, x, y);
}
inline CVector3Template<Scalar> zxy() const
{
return CVector3Template<Scalar>(z, x, y);
}
inline CVector3Template<Scalar> xyy() const
{
return CVector3Template<Scalar>(x, y, y);
}
inline CVector3Template<Scalar> yyy() const
{
return CVector3Template<Scalar>(y, y, y);
}
inline CVector3Template<Scalar> zyy() const
{
return CVector3Template<Scalar>(z, y, y);
}
inline CVector3Template<Scalar> xzy() const
{
return CVector3Template<Scalar>(x, z, y);
}
inline CVector3Template<Scalar> yzy() const
{
return CVector3Template<Scalar>(y, z, y);
}
inline CVector3Template<Scalar> zzy() const
{
return CVector3Template<Scalar>(z, z, y);
}
inline CVector3Template<Scalar> xxz() const
{
return CVector3Template<Scalar>(x, x, z);
}
inline CVector3Template<Scalar> yxz() const
{
return CVector3Template<Scalar>(y, x, z);
}
inline CVector3Template<Scalar> zxz() const
{
return CVector3Template<Scalar>(z, x, z);
}
inline CVector3Template<Scalar> xyz() const
{
return CVector3Template<Scalar>(x, y, z);
}
inline CVector3Template<Scalar> yyz() const
{
return CVector3Template<Scalar>(y, y, z);
}
inline CVector3Template<Scalar> zyz() const
{
return CVector3Template<Scalar>(z, y, z);
}
inline CVector3Template<Scalar> xzz() const
{
return CVector3Template<Scalar>(x, z, z);
}
inline CVector3Template<Scalar> yzz() const
{
return CVector3Template<Scalar>(y, z, z);
}
inline CVector3Template<Scalar> zzz() const
{
return CVector3Template<Scalar>(z, z, z);
}
inline void operator += (const CVector3Template<Scalar> &v)
{
x+=v.x;
y+=v.y;
z+=v.z;
}
inline void operator -= (const CVector3Template<Scalar> &v)
{
x-=v.x;
y-=v.y;
z-=v.z;
}
inline void operator *= (const Scalar v)
{
x*=v;
y*=v;
z*=v;
}
inline void operator *= (const CVector3Template<Scalar> &v)
{
x*=v.x;
y*=v.y;
z*=v.z;
}
inline void operator /= (const Scalar v)
{
x/=v;
y/=v;
z/=v;
}
inline void operator /= (const CVector3Template<Scalar> &v)
{
x/=v.x;
y/=v.y;
z/=v.z;
}
inline void fromTo(const CVector3Template<Scalar> & start, const CVector3Template<Scalar> & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
z = ende.z - start.z;
};
void setVector(Scalar arg0, Scalar arg1, Scalar arg2)
{
x = arg0;
y = arg1;
z = arg2;
}
inline void addVector(Scalar arg0, Scalar arg1, Scalar arg2)
{
x += arg0;
y += arg1;
z += arg2;
}
inline void subVector(Scalar arg0, Scalar arg1, Scalar arg2)
{
x -= arg0;
y -= arg1;
z -= arg2;
}
inline void neg()
{
x = -x;
y = -y;
z = -z;
}
inline Scalar length() const
{
return sqrt(x*x+y*y+z*z);
}
inline Scalar normalize()
{
Scalar r = sqrt(x*x+y*y+z*z);
   if (r < 0.00001) return r;
x /= r;
y /= r;
z /= r;
  return r;
}
inline Scalar dotProduct(const CVector3Template<Scalar> &b) const
{
return (x* b.x+y* b.y+z* b.z);
}
inline int getMajorAxis() const
{
Scalar current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
if (current < fabs(z)) {
index = 2;
current = fabs(z);
};
return index;
}
inline Scalar distanceTo(const CVector3Template<Scalar> &v) const;
};
template<class Scalar> inline CVector3Template<Scalar> operator +(const CVector3Template<Scalar> &v1, const CVector3Template<Scalar> &v2)
{
CVector3Template<Scalar> r;
r.x = v1.x + v2.x;
r.y = v1.y + v2.y;
r.z = v1.z + v2.z;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator +(const CVector3Template<Scalar> &v1, const Scalar v2)
{
CVector3Template<Scalar> r;
r.x = v1.x + v2;
r.y = v1.y + v2;
r.z = v1.z + v2;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator -(const CVector3Template<Scalar> &v1, const CVector3Template<Scalar> &v2)
{
CVector3Template<Scalar> r;
r.x = v1.x - v2.x;
r.y = v1.y - v2.y;
r.z = v1.z - v2.z;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator -(const CVector3Template<Scalar> &v1, const Scalar v2)
{
CVector3Template<Scalar> r;
r.x = v1.x - v2;
r.y = v1.y - v2;
r.z = v1.z - v2;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator -(const CVector3Template<Scalar> &v1)
{
CVector3Template<Scalar> r;
r.x = -v1.x;
r.y = -v1.y;
r.z = -v1.z;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator *(const CVector3Template<Scalar> &v1, const CVector3Template<Scalar> &v2)
{
CVector3Template<Scalar> r;
r.x = v1.x * v2.x;
r.y = v1.y * v2.y;
r.z = v1.z * v2.z;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator *(const CVector3Template<Scalar> &v1, const Scalar v2)
{
CVector3Template<Scalar> r;
r.x = v1.x * v2;
r.y = v1.y * v2;
r.z = v1.z * v2;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator *(const Scalar v1, const CVector3Template<Scalar> &v2)
{
CVector3Template<Scalar> r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
r.z = v1 * v2.z;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator /(const CVector3Template<Scalar> &v1, const CVector3Template<Scalar> &v2)
{
CVector3Template<Scalar> r;
r.x = v1.x / v2.x;
r.y = v1.y / v2.y;
r.z = v1.z / v2.z;
return r;
}
template<class Scalar> inline CVector3Template<Scalar> operator /(const CVector3Template<Scalar> &v1, const Scalar v2)
{
CVector3Template<Scalar> r;
r.x = v1.x / v2;
r.y = v1.y / v2;
r.z = v1.z / v2;
return r;
}
template<class Scalar> inline bool operator ==(const CVector3Template<Scalar> &v1, const CVector3Template<Scalar> &v2)
{
bool r;
r =  v1.x == v2.x
 &&
 v1.y == v2.y
 &&
 v1.z == v2.z
;
return r;
}
template<class Scalar> inline bool operator !=(const CVector3Template<Scalar> &v1, const CVector3Template<Scalar> &v2)
{
bool r;
r =  v1.x != v2.x
 ||
 v1.y != v2.y
 ||
 v1.z != v2.z
;
return r;
}
template<class Scalar> inline Scalar CVector3Template<Scalar>::distanceTo(const CVector3Template<Scalar> &v) const
{
return (v-*this).length();
}
template<class Scalar> inline Scalar dot(const CVector3Template<Scalar> &a, const CVector3Template<Scalar> &b)
{
return (a.x* b.x+a.y* b.y+a.z* b.z);
}
template<class Scalar> inline CVector3Template<Scalar> umin(const CVector3Template<Scalar> &a, const CVector3Template<Scalar> &b)
{
CVector3Template<Scalar> r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
r.z = umin(a.z, b.z);
return r;
}
template<class Scalar> inline CVector3Template<Scalar> umax(const CVector3Template<Scalar> &a, const CVector3Template<Scalar> &b)
{
CVector3Template<Scalar> r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
r.z = umax(a.z, b.z);
return r;
}
template<class Scalar> inline CVector3Template<Scalar> normalize(const CVector3Template<Scalar> &a)
{
CVector3Template<Scalar> rvec;
Scalar r = sqrt(a.x* a.x+a.y* a.y+a.z* a.z);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
rvec.z = a.z / r;
  return rvec;
}
template<class Scalar> inline Scalar length(const CVector3Template<Scalar> &a)
{
CVector3Template<Scalar> rvec;
Scalar r = sqrt(a.x*a.x+a.y*a.y+a.z*a.z);
   return r;
}
template<class Scalar> inline CVector3Template<Scalar> sign(const CVector3Template<Scalar> &a)
{
CVector3Template<Scalar> rvec;
rvec.x = a.x > 0 ? 1 : -1;
rvec.y = a.y > 0 ? 1 : -1;
rvec.z = a.z > 0 ? 1 : -1;
  return rvec;
}
template<class Scalar> inline Scalar distance(const CVector3Template<Scalar> &a, const CVector3Template<Scalar> &b)
{
return sqrt(sqr(a.x-b.x)+sqr(a.y-b.y)+sqr(a.z-b.z));
}
template<class Scalar> inline CVector3Template<Scalar> linear(const CVector3Template<Scalar> &a, const CVector3Template<Scalar> &b, const Scalar v)
{
CVector3Template<Scalar> rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
rvec.z = a.z+(b.z - a.z)*v;
return rvec;
}
template<class Scalar> struct CVector4Template {
Scalar x;
Scalar y;
Scalar z;
Scalar w;
inline CVector4Template(){x=0;y=0;z=0;w=0;};
inline CVector4Template(const Scalar f) {
x = f;
y = f;
z = f;
w = f;
};
inline CVector4Template(const Scalar arg0, const Scalar arg1, const Scalar arg2, const Scalar arg3){
x = arg0;y = arg1;z = arg2;w = arg3;
}
inline CVector4Template(const Scalar arg0, const Scalar arg1, const CVector2Template<Scalar> & arg2){
x = arg0;y = arg1;z = arg2.x;w = arg2.y;
}
inline CVector4Template(const Scalar arg0, const CVector2Template<Scalar> & arg1, const Scalar arg2){
x = arg0;y = arg1.x;z = arg1.y;w = arg2;
}
inline CVector4Template(const Scalar arg0, const CVector3Template<Scalar> & arg1){
x = arg0;y = arg1.x;z = arg1.y;w = arg1.z;
}
inline CVector4Template(const CVector2Template<Scalar> & arg0, const Scalar arg1, const Scalar arg2){
x = arg0.x;y = arg0.y;z = arg1;w = arg2;
}
inline CVector4Template(const CVector2Template<Scalar> & arg0, const CVector2Template<Scalar> & arg1){
x = arg0.x;y = arg0.y;z = arg1.x;w = arg1.y;
}
inline CVector4Template(const CVector3Template<Scalar> & arg0, const Scalar arg1){
x = arg0.x;y = arg0.y;z = arg0.z;w = arg1;
}
inline Scalar &operator[](unsigned int p)
{
utiassert(/*p >= 0 &&*/ p < 4);
return (&x)[p];
}
inline const Scalar &operator[](unsigned int p) const
{
utiassert(/*p >= 0 &&*/ p < 4);
    return (&x)[p];
}
inline CVector2Template<Scalar> xx() const
{
return CVector2Template<Scalar>(x, x);
}
inline CVector2Template<Scalar> yx() const
{
return CVector2Template<Scalar>(y, x);
}
inline CVector2Template<Scalar> zx() const
{
return CVector2Template<Scalar>(z, x);
}
inline CVector2Template<Scalar> wx() const
{
return CVector2Template<Scalar>(w, x);
}
inline CVector2Template<Scalar> xy() const
{
return CVector2Template<Scalar>(x, y);
}
inline CVector2Template<Scalar> yy() const
{
return CVector2Template<Scalar>(y, y);
}
inline CVector2Template<Scalar> zy() const
{
return CVector2Template<Scalar>(z, y);
}
inline CVector2Template<Scalar> wy() const
{
return CVector2Template<Scalar>(w, y);
}
inline CVector2Template<Scalar> xz() const
{
return CVector2Template<Scalar>(x, z);
}
inline CVector2Template<Scalar> yz() const
{
return CVector2Template<Scalar>(y, z);
}
inline CVector2Template<Scalar> zz() const
{
return CVector2Template<Scalar>(z, z);
}
inline CVector2Template<Scalar> wz() const
{
return CVector2Template<Scalar>(w, z);
}
inline CVector2Template<Scalar> xw() const
{
return CVector2Template<Scalar>(x, w);
}
inline CVector2Template<Scalar> yw() const
{
return CVector2Template<Scalar>(y, w);
}
inline CVector2Template<Scalar> zw() const
{
return CVector2Template<Scalar>(z, w);
}
inline CVector2Template<Scalar> ww() const
{
return CVector2Template<Scalar>(w, w);
}
inline CVector3Template<Scalar> xxx() const
{
return CVector3Template<Scalar>(x, x, x);
}
inline CVector3Template<Scalar> yxx() const
{
return CVector3Template<Scalar>(y, x, x);
}
inline CVector3Template<Scalar> zxx() const
{
return CVector3Template<Scalar>(z, x, x);
}
inline CVector3Template<Scalar> wxx() const
{
return CVector3Template<Scalar>(w, x, x);
}
inline CVector3Template<Scalar> xyx() const
{
return CVector3Template<Scalar>(x, y, x);
}
inline CVector3Template<Scalar> yyx() const
{
return CVector3Template<Scalar>(y, y, x);
}
inline CVector3Template<Scalar> zyx() const
{
return CVector3Template<Scalar>(z, y, x);
}
inline CVector3Template<Scalar> wyx() const
{
return CVector3Template<Scalar>(w, y, x);
}
inline CVector3Template<Scalar> xzx() const
{
return CVector3Template<Scalar>(x, z, x);
}
inline CVector3Template<Scalar> yzx() const
{
return CVector3Template<Scalar>(y, z, x);
}
inline CVector3Template<Scalar> zzx() const
{
return CVector3Template<Scalar>(z, z, x);
}
inline CVector3Template<Scalar> wzx() const
{
return CVector3Template<Scalar>(w, z, x);
}
inline CVector3Template<Scalar> xwx() const
{
return CVector3Template<Scalar>(x, w, x);
}
inline CVector3Template<Scalar> ywx() const
{
return CVector3Template<Scalar>(y, w, x);
}
inline CVector3Template<Scalar> zwx() const
{
return CVector3Template<Scalar>(z, w, x);
}
inline CVector3Template<Scalar> wwx() const
{
return CVector3Template<Scalar>(w, w, x);
}
inline CVector3Template<Scalar> xxy() const
{
return CVector3Template<Scalar>(x, x, y);
}
inline CVector3Template<Scalar> yxy() const
{
return CVector3Template<Scalar>(y, x, y);
}
inline CVector3Template<Scalar> zxy() const
{
return CVector3Template<Scalar>(z, x, y);
}
inline CVector3Template<Scalar> wxy() const
{
return CVector3Template<Scalar>(w, x, y);
}
inline CVector3Template<Scalar> xyy() const
{
return CVector3Template<Scalar>(x, y, y);
}
inline CVector3Template<Scalar> yyy() const
{
return CVector3Template<Scalar>(y, y, y);
}
inline CVector3Template<Scalar> zyy() const
{
return CVector3Template<Scalar>(z, y, y);
}
inline CVector3Template<Scalar> wyy() const
{
return CVector3Template<Scalar>(w, y, y);
}
inline CVector3Template<Scalar> xzy() const
{
return CVector3Template<Scalar>(x, z, y);
}
inline CVector3Template<Scalar> yzy() const
{
return CVector3Template<Scalar>(y, z, y);
}
inline CVector3Template<Scalar> zzy() const
{
return CVector3Template<Scalar>(z, z, y);
}
inline CVector3Template<Scalar> wzy() const
{
return CVector3Template<Scalar>(w, z, y);
}
inline CVector3Template<Scalar> xwy() const
{
return CVector3Template<Scalar>(x, w, y);
}
inline CVector3Template<Scalar> ywy() const
{
return CVector3Template<Scalar>(y, w, y);
}
inline CVector3Template<Scalar> zwy() const
{
return CVector3Template<Scalar>(z, w, y);
}
inline CVector3Template<Scalar> wwy() const
{
return CVector3Template<Scalar>(w, w, y);
}
inline CVector3Template<Scalar> xxz() const
{
return CVector3Template<Scalar>(x, x, z);
}
inline CVector3Template<Scalar> yxz() const
{
return CVector3Template<Scalar>(y, x, z);
}
inline CVector3Template<Scalar> zxz() const
{
return CVector3Template<Scalar>(z, x, z);
}
inline CVector3Template<Scalar> wxz() const
{
return CVector3Template<Scalar>(w, x, z);
}
inline CVector3Template<Scalar> xyz() const
{
return CVector3Template<Scalar>(x, y, z);
}
inline CVector3Template<Scalar> yyz() const
{
return CVector3Template<Scalar>(y, y, z);
}
inline CVector3Template<Scalar> zyz() const
{
return CVector3Template<Scalar>(z, y, z);
}
inline CVector3Template<Scalar> wyz() const
{
return CVector3Template<Scalar>(w, y, z);
}
inline CVector3Template<Scalar> xzz() const
{
return CVector3Template<Scalar>(x, z, z);
}
inline CVector3Template<Scalar> yzz() const
{
return CVector3Template<Scalar>(y, z, z);
}
inline CVector3Template<Scalar> zzz() const
{
return CVector3Template<Scalar>(z, z, z);
}
inline CVector3Template<Scalar> wzz() const
{
return CVector3Template<Scalar>(w, z, z);
}
inline CVector3Template<Scalar> xwz() const
{
return CVector3Template<Scalar>(x, w, z);
}
inline CVector3Template<Scalar> ywz() const
{
return CVector3Template<Scalar>(y, w, z);
}
inline CVector3Template<Scalar> zwz() const
{
return CVector3Template<Scalar>(z, w, z);
}
inline CVector3Template<Scalar> wwz() const
{
return CVector3Template<Scalar>(w, w, z);
}
inline CVector3Template<Scalar> xxw() const
{
return CVector3Template<Scalar>(x, x, w);
}
inline CVector3Template<Scalar> yxw() const
{
return CVector3Template<Scalar>(y, x, w);
}
inline CVector3Template<Scalar> zxw() const
{
return CVector3Template<Scalar>(z, x, w);
}
inline CVector3Template<Scalar> wxw() const
{
return CVector3Template<Scalar>(w, x, w);
}
inline CVector3Template<Scalar> xyw() const
{
return CVector3Template<Scalar>(x, y, w);
}
inline CVector3Template<Scalar> yyw() const
{
return CVector3Template<Scalar>(y, y, w);
}
inline CVector3Template<Scalar> zyw() const
{
return CVector3Template<Scalar>(z, y, w);
}
inline CVector3Template<Scalar> wyw() const
{
return CVector3Template<Scalar>(w, y, w);
}
inline CVector3Template<Scalar> xzw() const
{
return CVector3Template<Scalar>(x, z, w);
}
inline CVector3Template<Scalar> yzw() const
{
return CVector3Template<Scalar>(y, z, w);
}
inline CVector3Template<Scalar> zzw() const
{
return CVector3Template<Scalar>(z, z, w);
}
inline CVector3Template<Scalar> wzw() const
{
return CVector3Template<Scalar>(w, z, w);
}
inline CVector3Template<Scalar> xww() const
{
return CVector3Template<Scalar>(x, w, w);
}
inline CVector3Template<Scalar> yww() const
{
return CVector3Template<Scalar>(y, w, w);
}
inline CVector3Template<Scalar> zww() const
{
return CVector3Template<Scalar>(z, w, w);
}
inline CVector3Template<Scalar> www() const
{
return CVector3Template<Scalar>(w, w, w);
}
inline CVector4Template<Scalar> xxxx() const
{
return CVector4Template<Scalar>(x, x, x, x);
}
inline CVector4Template<Scalar> yxxx() const
{
return CVector4Template<Scalar>(y, x, x, x);
}
inline CVector4Template<Scalar> zxxx() const
{
return CVector4Template<Scalar>(z, x, x, x);
}
inline CVector4Template<Scalar> wxxx() const
{
return CVector4Template<Scalar>(w, x, x, x);
}
inline CVector4Template<Scalar> xyxx() const
{
return CVector4Template<Scalar>(x, y, x, x);
}
inline CVector4Template<Scalar> yyxx() const
{
return CVector4Template<Scalar>(y, y, x, x);
}
inline CVector4Template<Scalar> zyxx() const
{
return CVector4Template<Scalar>(z, y, x, x);
}
inline CVector4Template<Scalar> wyxx() const
{
return CVector4Template<Scalar>(w, y, x, x);
}
inline CVector4Template<Scalar> xzxx() const
{
return CVector4Template<Scalar>(x, z, x, x);
}
inline CVector4Template<Scalar> yzxx() const
{
return CVector4Template<Scalar>(y, z, x, x);
}
inline CVector4Template<Scalar> zzxx() const
{
return CVector4Template<Scalar>(z, z, x, x);
}
inline CVector4Template<Scalar> wzxx() const
{
return CVector4Template<Scalar>(w, z, x, x);
}
inline CVector4Template<Scalar> xwxx() const
{
return CVector4Template<Scalar>(x, w, x, x);
}
inline CVector4Template<Scalar> ywxx() const
{
return CVector4Template<Scalar>(y, w, x, x);
}
inline CVector4Template<Scalar> zwxx() const
{
return CVector4Template<Scalar>(z, w, x, x);
}
inline CVector4Template<Scalar> wwxx() const
{
return CVector4Template<Scalar>(w, w, x, x);
}
inline CVector4Template<Scalar> xxyx() const
{
return CVector4Template<Scalar>(x, x, y, x);
}
inline CVector4Template<Scalar> yxyx() const
{
return CVector4Template<Scalar>(y, x, y, x);
}
inline CVector4Template<Scalar> zxyx() const
{
return CVector4Template<Scalar>(z, x, y, x);
}
inline CVector4Template<Scalar> wxyx() const
{
return CVector4Template<Scalar>(w, x, y, x);
}
inline CVector4Template<Scalar> xyyx() const
{
return CVector4Template<Scalar>(x, y, y, x);
}
inline CVector4Template<Scalar> yyyx() const
{
return CVector4Template<Scalar>(y, y, y, x);
}
inline CVector4Template<Scalar> zyyx() const
{
return CVector4Template<Scalar>(z, y, y, x);
}
inline CVector4Template<Scalar> wyyx() const
{
return CVector4Template<Scalar>(w, y, y, x);
}
inline CVector4Template<Scalar> xzyx() const
{
return CVector4Template<Scalar>(x, z, y, x);
}
inline CVector4Template<Scalar> yzyx() const
{
return CVector4Template<Scalar>(y, z, y, x);
}
inline CVector4Template<Scalar> zzyx() const
{
return CVector4Template<Scalar>(z, z, y, x);
}
inline CVector4Template<Scalar> wzyx() const
{
return CVector4Template<Scalar>(w, z, y, x);
}
inline CVector4Template<Scalar> xwyx() const
{
return CVector4Template<Scalar>(x, w, y, x);
}
inline CVector4Template<Scalar> ywyx() const
{
return CVector4Template<Scalar>(y, w, y, x);
}
inline CVector4Template<Scalar> zwyx() const
{
return CVector4Template<Scalar>(z, w, y, x);
}
inline CVector4Template<Scalar> wwyx() const
{
return CVector4Template<Scalar>(w, w, y, x);
}
inline CVector4Template<Scalar> xxzx() const
{
return CVector4Template<Scalar>(x, x, z, x);
}
inline CVector4Template<Scalar> yxzx() const
{
return CVector4Template<Scalar>(y, x, z, x);
}
inline CVector4Template<Scalar> zxzx() const
{
return CVector4Template<Scalar>(z, x, z, x);
}
inline CVector4Template<Scalar> wxzx() const
{
return CVector4Template<Scalar>(w, x, z, x);
}
inline CVector4Template<Scalar> xyzx() const
{
return CVector4Template<Scalar>(x, y, z, x);
}
inline CVector4Template<Scalar> yyzx() const
{
return CVector4Template<Scalar>(y, y, z, x);
}
inline CVector4Template<Scalar> zyzx() const
{
return CVector4Template<Scalar>(z, y, z, x);
}
inline CVector4Template<Scalar> wyzx() const
{
return CVector4Template<Scalar>(w, y, z, x);
}
inline CVector4Template<Scalar> xzzx() const
{
return CVector4Template<Scalar>(x, z, z, x);
}
inline CVector4Template<Scalar> yzzx() const
{
return CVector4Template<Scalar>(y, z, z, x);
}
inline CVector4Template<Scalar> zzzx() const
{
return CVector4Template<Scalar>(z, z, z, x);
}
inline CVector4Template<Scalar> wzzx() const
{
return CVector4Template<Scalar>(w, z, z, x);
}
inline CVector4Template<Scalar> xwzx() const
{
return CVector4Template<Scalar>(x, w, z, x);
}
inline CVector4Template<Scalar> ywzx() const
{
return CVector4Template<Scalar>(y, w, z, x);
}
inline CVector4Template<Scalar> zwzx() const
{
return CVector4Template<Scalar>(z, w, z, x);
}
inline CVector4Template<Scalar> wwzx() const
{
return CVector4Template<Scalar>(w, w, z, x);
}
inline CVector4Template<Scalar> xxwx() const
{
return CVector4Template<Scalar>(x, x, w, x);
}
inline CVector4Template<Scalar> yxwx() const
{
return CVector4Template<Scalar>(y, x, w, x);
}
inline CVector4Template<Scalar> zxwx() const
{
return CVector4Template<Scalar>(z, x, w, x);
}
inline CVector4Template<Scalar> wxwx() const
{
return CVector4Template<Scalar>(w, x, w, x);
}
inline CVector4Template<Scalar> xywx() const
{
return CVector4Template<Scalar>(x, y, w, x);
}
inline CVector4Template<Scalar> yywx() const
{
return CVector4Template<Scalar>(y, y, w, x);
}
inline CVector4Template<Scalar> zywx() const
{
return CVector4Template<Scalar>(z, y, w, x);
}
inline CVector4Template<Scalar> wywx() const
{
return CVector4Template<Scalar>(w, y, w, x);
}
inline CVector4Template<Scalar> xzwx() const
{
return CVector4Template<Scalar>(x, z, w, x);
}
inline CVector4Template<Scalar> yzwx() const
{
return CVector4Template<Scalar>(y, z, w, x);
}
inline CVector4Template<Scalar> zzwx() const
{
return CVector4Template<Scalar>(z, z, w, x);
}
inline CVector4Template<Scalar> wzwx() const
{
return CVector4Template<Scalar>(w, z, w, x);
}
inline CVector4Template<Scalar> xwwx() const
{
return CVector4Template<Scalar>(x, w, w, x);
}
inline CVector4Template<Scalar> ywwx() const
{
return CVector4Template<Scalar>(y, w, w, x);
}
inline CVector4Template<Scalar> zwwx() const
{
return CVector4Template<Scalar>(z, w, w, x);
}
inline CVector4Template<Scalar> wwwx() const
{
return CVector4Template<Scalar>(w, w, w, x);
}
inline CVector4Template<Scalar> xxxy() const
{
return CVector4Template<Scalar>(x, x, x, y);
}
inline CVector4Template<Scalar> yxxy() const
{
return CVector4Template<Scalar>(y, x, x, y);
}
inline CVector4Template<Scalar> zxxy() const
{
return CVector4Template<Scalar>(z, x, x, y);
}
inline CVector4Template<Scalar> wxxy() const
{
return CVector4Template<Scalar>(w, x, x, y);
}
inline CVector4Template<Scalar> xyxy() const
{
return CVector4Template<Scalar>(x, y, x, y);
}
inline CVector4Template<Scalar> yyxy() const
{
return CVector4Template<Scalar>(y, y, x, y);
}
inline CVector4Template<Scalar> zyxy() const
{
return CVector4Template<Scalar>(z, y, x, y);
}
inline CVector4Template<Scalar> wyxy() const
{
return CVector4Template<Scalar>(w, y, x, y);
}
inline CVector4Template<Scalar> xzxy() const
{
return CVector4Template<Scalar>(x, z, x, y);
}
inline CVector4Template<Scalar> yzxy() const
{
return CVector4Template<Scalar>(y, z, x, y);
}
inline CVector4Template<Scalar> zzxy() const
{
return CVector4Template<Scalar>(z, z, x, y);
}
inline CVector4Template<Scalar> wzxy() const
{
return CVector4Template<Scalar>(w, z, x, y);
}
inline CVector4Template<Scalar> xwxy() const
{
return CVector4Template<Scalar>(x, w, x, y);
}
inline CVector4Template<Scalar> ywxy() const
{
return CVector4Template<Scalar>(y, w, x, y);
}
inline CVector4Template<Scalar> zwxy() const
{
return CVector4Template<Scalar>(z, w, x, y);
}
inline CVector4Template<Scalar> wwxy() const
{
return CVector4Template<Scalar>(w, w, x, y);
}
inline CVector4Template<Scalar> xxyy() const
{
return CVector4Template<Scalar>(x, x, y, y);
}
inline CVector4Template<Scalar> yxyy() const
{
return CVector4Template<Scalar>(y, x, y, y);
}
inline CVector4Template<Scalar> zxyy() const
{
return CVector4Template<Scalar>(z, x, y, y);
}
inline CVector4Template<Scalar> wxyy() const
{
return CVector4Template<Scalar>(w, x, y, y);
}
inline CVector4Template<Scalar> xyyy() const
{
return CVector4Template<Scalar>(x, y, y, y);
}
inline CVector4Template<Scalar> yyyy() const
{
return CVector4Template<Scalar>(y, y, y, y);
}
inline CVector4Template<Scalar> zyyy() const
{
return CVector4Template<Scalar>(z, y, y, y);
}
inline CVector4Template<Scalar> wyyy() const
{
return CVector4Template<Scalar>(w, y, y, y);
}
inline CVector4Template<Scalar> xzyy() const
{
return CVector4Template<Scalar>(x, z, y, y);
}
inline CVector4Template<Scalar> yzyy() const
{
return CVector4Template<Scalar>(y, z, y, y);
}
inline CVector4Template<Scalar> zzyy() const
{
return CVector4Template<Scalar>(z, z, y, y);
}
inline CVector4Template<Scalar> wzyy() const
{
return CVector4Template<Scalar>(w, z, y, y);
}
inline CVector4Template<Scalar> xwyy() const
{
return CVector4Template<Scalar>(x, w, y, y);
}
inline CVector4Template<Scalar> ywyy() const
{
return CVector4Template<Scalar>(y, w, y, y);
}
inline CVector4Template<Scalar> zwyy() const
{
return CVector4Template<Scalar>(z, w, y, y);
}
inline CVector4Template<Scalar> wwyy() const
{
return CVector4Template<Scalar>(w, w, y, y);
}
inline CVector4Template<Scalar> xxzy() const
{
return CVector4Template<Scalar>(x, x, z, y);
}
inline CVector4Template<Scalar> yxzy() const
{
return CVector4Template<Scalar>(y, x, z, y);
}
inline CVector4Template<Scalar> zxzy() const
{
return CVector4Template<Scalar>(z, x, z, y);
}
inline CVector4Template<Scalar> wxzy() const
{
return CVector4Template<Scalar>(w, x, z, y);
}
inline CVector4Template<Scalar> xyzy() const
{
return CVector4Template<Scalar>(x, y, z, y);
}
inline CVector4Template<Scalar> yyzy() const
{
return CVector4Template<Scalar>(y, y, z, y);
}
inline CVector4Template<Scalar> zyzy() const
{
return CVector4Template<Scalar>(z, y, z, y);
}
inline CVector4Template<Scalar> wyzy() const
{
return CVector4Template<Scalar>(w, y, z, y);
}
inline CVector4Template<Scalar> xzzy() const
{
return CVector4Template<Scalar>(x, z, z, y);
}
inline CVector4Template<Scalar> yzzy() const
{
return CVector4Template<Scalar>(y, z, z, y);
}
inline CVector4Template<Scalar> zzzy() const
{
return CVector4Template<Scalar>(z, z, z, y);
}
inline CVector4Template<Scalar> wzzy() const
{
return CVector4Template<Scalar>(w, z, z, y);
}
inline CVector4Template<Scalar> xwzy() const
{
return CVector4Template<Scalar>(x, w, z, y);
}
inline CVector4Template<Scalar> ywzy() const
{
return CVector4Template<Scalar>(y, w, z, y);
}
inline CVector4Template<Scalar> zwzy() const
{
return CVector4Template<Scalar>(z, w, z, y);
}
inline CVector4Template<Scalar> wwzy() const
{
return CVector4Template<Scalar>(w, w, z, y);
}
inline CVector4Template<Scalar> xxwy() const
{
return CVector4Template<Scalar>(x, x, w, y);
}
inline CVector4Template<Scalar> yxwy() const
{
return CVector4Template<Scalar>(y, x, w, y);
}
inline CVector4Template<Scalar> zxwy() const
{
return CVector4Template<Scalar>(z, x, w, y);
}
inline CVector4Template<Scalar> wxwy() const
{
return CVector4Template<Scalar>(w, x, w, y);
}
inline CVector4Template<Scalar> xywy() const
{
return CVector4Template<Scalar>(x, y, w, y);
}
inline CVector4Template<Scalar> yywy() const
{
return CVector4Template<Scalar>(y, y, w, y);
}
inline CVector4Template<Scalar> zywy() const
{
return CVector4Template<Scalar>(z, y, w, y);
}
inline CVector4Template<Scalar> wywy() const
{
return CVector4Template<Scalar>(w, y, w, y);
}
inline CVector4Template<Scalar> xzwy() const
{
return CVector4Template<Scalar>(x, z, w, y);
}
inline CVector4Template<Scalar> yzwy() const
{
return CVector4Template<Scalar>(y, z, w, y);
}
inline CVector4Template<Scalar> zzwy() const
{
return CVector4Template<Scalar>(z, z, w, y);
}
inline CVector4Template<Scalar> wzwy() const
{
return CVector4Template<Scalar>(w, z, w, y);
}
inline CVector4Template<Scalar> xwwy() const
{
return CVector4Template<Scalar>(x, w, w, y);
}
inline CVector4Template<Scalar> ywwy() const
{
return CVector4Template<Scalar>(y, w, w, y);
}
inline CVector4Template<Scalar> zwwy() const
{
return CVector4Template<Scalar>(z, w, w, y);
}
inline CVector4Template<Scalar> wwwy() const
{
return CVector4Template<Scalar>(w, w, w, y);
}
inline CVector4Template<Scalar> xxxz() const
{
return CVector4Template<Scalar>(x, x, x, z);
}
inline CVector4Template<Scalar> yxxz() const
{
return CVector4Template<Scalar>(y, x, x, z);
}
inline CVector4Template<Scalar> zxxz() const
{
return CVector4Template<Scalar>(z, x, x, z);
}
inline CVector4Template<Scalar> wxxz() const
{
return CVector4Template<Scalar>(w, x, x, z);
}
inline CVector4Template<Scalar> xyxz() const
{
return CVector4Template<Scalar>(x, y, x, z);
}
inline CVector4Template<Scalar> yyxz() const
{
return CVector4Template<Scalar>(y, y, x, z);
}
inline CVector4Template<Scalar> zyxz() const
{
return CVector4Template<Scalar>(z, y, x, z);
}
inline CVector4Template<Scalar> wyxz() const
{
return CVector4Template<Scalar>(w, y, x, z);
}
inline CVector4Template<Scalar> xzxz() const
{
return CVector4Template<Scalar>(x, z, x, z);
}
inline CVector4Template<Scalar> yzxz() const
{
return CVector4Template<Scalar>(y, z, x, z);
}
inline CVector4Template<Scalar> zzxz() const
{
return CVector4Template<Scalar>(z, z, x, z);
}
inline CVector4Template<Scalar> wzxz() const
{
return CVector4Template<Scalar>(w, z, x, z);
}
inline CVector4Template<Scalar> xwxz() const
{
return CVector4Template<Scalar>(x, w, x, z);
}
inline CVector4Template<Scalar> ywxz() const
{
return CVector4Template<Scalar>(y, w, x, z);
}
inline CVector4Template<Scalar> zwxz() const
{
return CVector4Template<Scalar>(z, w, x, z);
}
inline CVector4Template<Scalar> wwxz() const
{
return CVector4Template<Scalar>(w, w, x, z);
}
inline CVector4Template<Scalar> xxyz() const
{
return CVector4Template<Scalar>(x, x, y, z);
}
inline CVector4Template<Scalar> yxyz() const
{
return CVector4Template<Scalar>(y, x, y, z);
}
inline CVector4Template<Scalar> zxyz() const
{
return CVector4Template<Scalar>(z, x, y, z);
}
inline CVector4Template<Scalar> wxyz() const
{
return CVector4Template<Scalar>(w, x, y, z);
}
inline CVector4Template<Scalar> xyyz() const
{
return CVector4Template<Scalar>(x, y, y, z);
}
inline CVector4Template<Scalar> yyyz() const
{
return CVector4Template<Scalar>(y, y, y, z);
}
inline CVector4Template<Scalar> zyyz() const
{
return CVector4Template<Scalar>(z, y, y, z);
}
inline CVector4Template<Scalar> wyyz() const
{
return CVector4Template<Scalar>(w, y, y, z);
}
inline CVector4Template<Scalar> xzyz() const
{
return CVector4Template<Scalar>(x, z, y, z);
}
inline CVector4Template<Scalar> yzyz() const
{
return CVector4Template<Scalar>(y, z, y, z);
}
inline CVector4Template<Scalar> zzyz() const
{
return CVector4Template<Scalar>(z, z, y, z);
}
inline CVector4Template<Scalar> wzyz() const
{
return CVector4Template<Scalar>(w, z, y, z);
}
inline CVector4Template<Scalar> xwyz() const
{
return CVector4Template<Scalar>(x, w, y, z);
}
inline CVector4Template<Scalar> ywyz() const
{
return CVector4Template<Scalar>(y, w, y, z);
}
inline CVector4Template<Scalar> zwyz() const
{
return CVector4Template<Scalar>(z, w, y, z);
}
inline CVector4Template<Scalar> wwyz() const
{
return CVector4Template<Scalar>(w, w, y, z);
}
inline CVector4Template<Scalar> xxzz() const
{
return CVector4Template<Scalar>(x, x, z, z);
}
inline CVector4Template<Scalar> yxzz() const
{
return CVector4Template<Scalar>(y, x, z, z);
}
inline CVector4Template<Scalar> zxzz() const
{
return CVector4Template<Scalar>(z, x, z, z);
}
inline CVector4Template<Scalar> wxzz() const
{
return CVector4Template<Scalar>(w, x, z, z);
}
inline CVector4Template<Scalar> xyzz() const
{
return CVector4Template<Scalar>(x, y, z, z);
}
inline CVector4Template<Scalar> yyzz() const
{
return CVector4Template<Scalar>(y, y, z, z);
}
inline CVector4Template<Scalar> zyzz() const
{
return CVector4Template<Scalar>(z, y, z, z);
}
inline CVector4Template<Scalar> wyzz() const
{
return CVector4Template<Scalar>(w, y, z, z);
}
inline CVector4Template<Scalar> xzzz() const
{
return CVector4Template<Scalar>(x, z, z, z);
}
inline CVector4Template<Scalar> yzzz() const
{
return CVector4Template<Scalar>(y, z, z, z);
}
inline CVector4Template<Scalar> zzzz() const
{
return CVector4Template<Scalar>(z, z, z, z);
}
inline CVector4Template<Scalar> wzzz() const
{
return CVector4Template<Scalar>(w, z, z, z);
}
inline CVector4Template<Scalar> xwzz() const
{
return CVector4Template<Scalar>(x, w, z, z);
}
inline CVector4Template<Scalar> ywzz() const
{
return CVector4Template<Scalar>(y, w, z, z);
}
inline CVector4Template<Scalar> zwzz() const
{
return CVector4Template<Scalar>(z, w, z, z);
}
inline CVector4Template<Scalar> wwzz() const
{
return CVector4Template<Scalar>(w, w, z, z);
}
inline CVector4Template<Scalar> xxwz() const
{
return CVector4Template<Scalar>(x, x, w, z);
}
inline CVector4Template<Scalar> yxwz() const
{
return CVector4Template<Scalar>(y, x, w, z);
}
inline CVector4Template<Scalar> zxwz() const
{
return CVector4Template<Scalar>(z, x, w, z);
}
inline CVector4Template<Scalar> wxwz() const
{
return CVector4Template<Scalar>(w, x, w, z);
}
inline CVector4Template<Scalar> xywz() const
{
return CVector4Template<Scalar>(x, y, w, z);
}
inline CVector4Template<Scalar> yywz() const
{
return CVector4Template<Scalar>(y, y, w, z);
}
inline CVector4Template<Scalar> zywz() const
{
return CVector4Template<Scalar>(z, y, w, z);
}
inline CVector4Template<Scalar> wywz() const
{
return CVector4Template<Scalar>(w, y, w, z);
}
inline CVector4Template<Scalar> xzwz() const
{
return CVector4Template<Scalar>(x, z, w, z);
}
inline CVector4Template<Scalar> yzwz() const
{
return CVector4Template<Scalar>(y, z, w, z);
}
inline CVector4Template<Scalar> zzwz() const
{
return CVector4Template<Scalar>(z, z, w, z);
}
inline CVector4Template<Scalar> wzwz() const
{
return CVector4Template<Scalar>(w, z, w, z);
}
inline CVector4Template<Scalar> xwwz() const
{
return CVector4Template<Scalar>(x, w, w, z);
}
inline CVector4Template<Scalar> ywwz() const
{
return CVector4Template<Scalar>(y, w, w, z);
}
inline CVector4Template<Scalar> zwwz() const
{
return CVector4Template<Scalar>(z, w, w, z);
}
inline CVector4Template<Scalar> wwwz() const
{
return CVector4Template<Scalar>(w, w, w, z);
}
inline CVector4Template<Scalar> xxxw() const
{
return CVector4Template<Scalar>(x, x, x, w);
}
inline CVector4Template<Scalar> yxxw() const
{
return CVector4Template<Scalar>(y, x, x, w);
}
inline CVector4Template<Scalar> zxxw() const
{
return CVector4Template<Scalar>(z, x, x, w);
}
inline CVector4Template<Scalar> wxxw() const
{
return CVector4Template<Scalar>(w, x, x, w);
}
inline CVector4Template<Scalar> xyxw() const
{
return CVector4Template<Scalar>(x, y, x, w);
}
inline CVector4Template<Scalar> yyxw() const
{
return CVector4Template<Scalar>(y, y, x, w);
}
inline CVector4Template<Scalar> zyxw() const
{
return CVector4Template<Scalar>(z, y, x, w);
}
inline CVector4Template<Scalar> wyxw() const
{
return CVector4Template<Scalar>(w, y, x, w);
}
inline CVector4Template<Scalar> xzxw() const
{
return CVector4Template<Scalar>(x, z, x, w);
}
inline CVector4Template<Scalar> yzxw() const
{
return CVector4Template<Scalar>(y, z, x, w);
}
inline CVector4Template<Scalar> zzxw() const
{
return CVector4Template<Scalar>(z, z, x, w);
}
inline CVector4Template<Scalar> wzxw() const
{
return CVector4Template<Scalar>(w, z, x, w);
}
inline CVector4Template<Scalar> xwxw() const
{
return CVector4Template<Scalar>(x, w, x, w);
}
inline CVector4Template<Scalar> ywxw() const
{
return CVector4Template<Scalar>(y, w, x, w);
}
inline CVector4Template<Scalar> zwxw() const
{
return CVector4Template<Scalar>(z, w, x, w);
}
inline CVector4Template<Scalar> wwxw() const
{
return CVector4Template<Scalar>(w, w, x, w);
}
inline CVector4Template<Scalar> xxyw() const
{
return CVector4Template<Scalar>(x, x, y, w);
}
inline CVector4Template<Scalar> yxyw() const
{
return CVector4Template<Scalar>(y, x, y, w);
}
inline CVector4Template<Scalar> zxyw() const
{
return CVector4Template<Scalar>(z, x, y, w);
}
inline CVector4Template<Scalar> wxyw() const
{
return CVector4Template<Scalar>(w, x, y, w);
}
inline CVector4Template<Scalar> xyyw() const
{
return CVector4Template<Scalar>(x, y, y, w);
}
inline CVector4Template<Scalar> yyyw() const
{
return CVector4Template<Scalar>(y, y, y, w);
}
inline CVector4Template<Scalar> zyyw() const
{
return CVector4Template<Scalar>(z, y, y, w);
}
inline CVector4Template<Scalar> wyyw() const
{
return CVector4Template<Scalar>(w, y, y, w);
}
inline CVector4Template<Scalar> xzyw() const
{
return CVector4Template<Scalar>(x, z, y, w);
}
inline CVector4Template<Scalar> yzyw() const
{
return CVector4Template<Scalar>(y, z, y, w);
}
inline CVector4Template<Scalar> zzyw() const
{
return CVector4Template<Scalar>(z, z, y, w);
}
inline CVector4Template<Scalar> wzyw() const
{
return CVector4Template<Scalar>(w, z, y, w);
}
inline CVector4Template<Scalar> xwyw() const
{
return CVector4Template<Scalar>(x, w, y, w);
}
inline CVector4Template<Scalar> ywyw() const
{
return CVector4Template<Scalar>(y, w, y, w);
}
inline CVector4Template<Scalar> zwyw() const
{
return CVector4Template<Scalar>(z, w, y, w);
}
inline CVector4Template<Scalar> wwyw() const
{
return CVector4Template<Scalar>(w, w, y, w);
}
inline CVector4Template<Scalar> xxzw() const
{
return CVector4Template<Scalar>(x, x, z, w);
}
inline CVector4Template<Scalar> yxzw() const
{
return CVector4Template<Scalar>(y, x, z, w);
}
inline CVector4Template<Scalar> zxzw() const
{
return CVector4Template<Scalar>(z, x, z, w);
}
inline CVector4Template<Scalar> wxzw() const
{
return CVector4Template<Scalar>(w, x, z, w);
}
inline CVector4Template<Scalar> xyzw() const
{
return CVector4Template<Scalar>(x, y, z, w);
}
inline CVector4Template<Scalar> yyzw() const
{
return CVector4Template<Scalar>(y, y, z, w);
}
inline CVector4Template<Scalar> zyzw() const
{
return CVector4Template<Scalar>(z, y, z, w);
}
inline CVector4Template<Scalar> wyzw() const
{
return CVector4Template<Scalar>(w, y, z, w);
}
inline CVector4Template<Scalar> xzzw() const
{
return CVector4Template<Scalar>(x, z, z, w);
}
inline CVector4Template<Scalar> yzzw() const
{
return CVector4Template<Scalar>(y, z, z, w);
}
inline CVector4Template<Scalar> zzzw() const
{
return CVector4Template<Scalar>(z, z, z, w);
}
inline CVector4Template<Scalar> wzzw() const
{
return CVector4Template<Scalar>(w, z, z, w);
}
inline CVector4Template<Scalar> xwzw() const
{
return CVector4Template<Scalar>(x, w, z, w);
}
inline CVector4Template<Scalar> ywzw() const
{
return CVector4Template<Scalar>(y, w, z, w);
}
inline CVector4Template<Scalar> zwzw() const
{
return CVector4Template<Scalar>(z, w, z, w);
}
inline CVector4Template<Scalar> wwzw() const
{
return CVector4Template<Scalar>(w, w, z, w);
}
inline CVector4Template<Scalar> xxww() const
{
return CVector4Template<Scalar>(x, x, w, w);
}
inline CVector4Template<Scalar> yxww() const
{
return CVector4Template<Scalar>(y, x, w, w);
}
inline CVector4Template<Scalar> zxww() const
{
return CVector4Template<Scalar>(z, x, w, w);
}
inline CVector4Template<Scalar> wxww() const
{
return CVector4Template<Scalar>(w, x, w, w);
}
inline CVector4Template<Scalar> xyww() const
{
return CVector4Template<Scalar>(x, y, w, w);
}
inline CVector4Template<Scalar> yyww() const
{
return CVector4Template<Scalar>(y, y, w, w);
}
inline CVector4Template<Scalar> zyww() const
{
return CVector4Template<Scalar>(z, y, w, w);
}
inline CVector4Template<Scalar> wyww() const
{
return CVector4Template<Scalar>(w, y, w, w);
}
inline CVector4Template<Scalar> xzww() const
{
return CVector4Template<Scalar>(x, z, w, w);
}
inline CVector4Template<Scalar> yzww() const
{
return CVector4Template<Scalar>(y, z, w, w);
}
inline CVector4Template<Scalar> zzww() const
{
return CVector4Template<Scalar>(z, z, w, w);
}
inline CVector4Template<Scalar> wzww() const
{
return CVector4Template<Scalar>(w, z, w, w);
}
inline CVector4Template<Scalar> xwww() const
{
return CVector4Template<Scalar>(x, w, w, w);
}
inline CVector4Template<Scalar> ywww() const
{
return CVector4Template<Scalar>(y, w, w, w);
}
inline CVector4Template<Scalar> zwww() const
{
return CVector4Template<Scalar>(z, w, w, w);
}
inline CVector4Template<Scalar> wwww() const
{
return CVector4Template<Scalar>(w, w, w, w);
}
inline void operator += (const CVector4Template<Scalar> &v)
{
x+=v.x;
y+=v.y;
z+=v.z;
w+=v.w;
}
inline void operator -= (const CVector4Template<Scalar> &v)
{
x-=v.x;
y-=v.y;
z-=v.z;
w-=v.w;
}
inline void operator *= (const Scalar v)
{
x*=v;
y*=v;
z*=v;
w*=v;
}
inline void operator *= (const CVector4Template<Scalar> &v)
{
x*=v.x;
y*=v.y;
z*=v.z;
w*=v.w;
}
inline void operator /= (const Scalar v)
{
x/=v;
y/=v;
z/=v;
w/=v;
}
inline void operator /= (const CVector4Template<Scalar> &v)
{
x/=v.x;
y/=v.y;
z/=v.z;
w/=v.w;
}
inline void fromTo(const CVector4Template<Scalar> & start, const CVector4Template<Scalar> & ende)
{
x = ende.x - start.x;
y = ende.y - start.y;
z = ende.z - start.z;
w = ende.w - start.w;
};
void setVector(Scalar arg0, Scalar arg1, Scalar arg2, Scalar arg3)
{
x = arg0;
y = arg1;
z = arg2;
w = arg3;
}
inline void addVector(Scalar arg0, Scalar arg1, Scalar arg2, Scalar arg3)
{
x += arg0;
y += arg1;
z += arg2;
w += arg3;
}
inline void subVector(Scalar arg0, Scalar arg1, Scalar arg2, Scalar arg3)
{
x -= arg0;
y -= arg1;
z -= arg2;
w -= arg3;
}
inline void neg()
{
x = -x;
y = -y;
z = -z;
w = -w;
}
inline Scalar length() const
{
return sqrt(x*x+y*y+z*z+w*w);
}
inline Scalar normalize()
{
Scalar r = sqrt(x*x+y*y+z*z+w*w);
   if (r < 0.00001) return r;
x /= r;
y /= r;
z /= r;
w /= r;
  return r;
}
inline Scalar dotProduct(const CVector4Template<Scalar> &b) const
{
return (x* b.x+y* b.y+z* b.z+w* b.w);
}
inline int getMajorAxis() const
{
Scalar current = fabs(x);
int index = 0;
if (current < fabs(y)) {
index = 1;
current = fabs(y);
};
if (current < fabs(z)) {
index = 2;
current = fabs(z);
};
if (current < fabs(w)) {
index = 3;
current = fabs(w);
};
return index;
}
inline Scalar distanceTo(const CVector4Template<Scalar> &v) const;
};
template<class Scalar> inline CVector4Template<Scalar> operator +(const CVector4Template<Scalar> &v1, const CVector4Template<Scalar> &v2)
{
CVector4Template<Scalar> r;
r.x = v1.x + v2.x;
r.y = v1.y + v2.y;
r.z = v1.z + v2.z;
r.w = v1.w + v2.w;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator +(const CVector4Template<Scalar> &v1, const Scalar v2)
{
CVector4Template<Scalar> r;
r.x = v1.x + v2;
r.y = v1.y + v2;
r.z = v1.z + v2;
r.w = v1.w + v2;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator -(const CVector4Template<Scalar> &v1, const CVector4Template<Scalar> &v2)
{
CVector4Template<Scalar> r;
r.x = v1.x - v2.x;
r.y = v1.y - v2.y;
r.z = v1.z - v2.z;
r.w = v1.w - v2.w;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator -(const CVector4Template<Scalar> &v1, const Scalar v2)
{
CVector4Template<Scalar> r;
r.x = v1.x - v2;
r.y = v1.y - v2;
r.z = v1.z - v2;
r.w = v1.w - v2;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator -(const CVector4Template<Scalar> &v1)
{
CVector4Template<Scalar> r;
r.x = -v1.x;
r.y = -v1.y;
r.z = -v1.z;
r.w = -v1.w;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator *(const CVector4Template<Scalar> &v1, const CVector4Template<Scalar> &v2)
{
CVector4Template<Scalar> r;
r.x = v1.x * v2.x;
r.y = v1.y * v2.y;
r.z = v1.z * v2.z;
r.w = v1.w * v2.w;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator *(const CVector4Template<Scalar> &v1, const Scalar v2)
{
CVector4Template<Scalar> r;
r.x = v1.x * v2;
r.y = v1.y * v2;
r.z = v1.z * v2;
r.w = v1.w * v2;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator *(const Scalar v1, const CVector4Template<Scalar> &v2)
{
CVector4Template<Scalar> r;
r.x = v1 * v2.x;
r.y = v1 * v2.y;
r.z = v1 * v2.z;
r.w = v1 * v2.w;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator /(const CVector4Template<Scalar> &v1, const CVector4Template<Scalar> &v2)
{
CVector4Template<Scalar> r;
r.x = v1.x / v2.x;
r.y = v1.y / v2.y;
r.z = v1.z / v2.z;
r.w = v1.w / v2.w;
return r;
}
template<class Scalar> inline CVector4Template<Scalar> operator /(const CVector4Template<Scalar> &v1, const Scalar v2)
{
CVector4Template<Scalar> r;
r.x = v1.x / v2;
r.y = v1.y / v2;
r.z = v1.z / v2;
r.w = v1.w / v2;
return r;
}
template<class Scalar> inline bool operator ==(const CVector4Template<Scalar> &v1, const CVector4Template<Scalar> &v2)
{
bool r;
r =  v1.x == v2.x
 &&
 v1.y == v2.y
 &&
 v1.z == v2.z
 &&
 v1.w == v2.w
;
return r;
}
template<class Scalar> inline bool operator !=(const CVector4Template<Scalar> &v1, const CVector4Template<Scalar> &v2)
{
bool r;
r =  v1.x != v2.x
 ||
 v1.y != v2.y
 ||
 v1.z != v2.z
 ||
 v1.w != v2.w
;
return r;
}
template<class Scalar> inline Scalar CVector4Template<Scalar>::distanceTo(const CVector4Template<Scalar> &v) const
{
return (v-*this).length();
}
template<class Scalar> inline Scalar dot(const CVector4Template<Scalar> &a, const CVector4Template<Scalar> &b)
{
return (a.x* b.x+a.y* b.y+a.z* b.z+a.w* b.w);
}
template<class Scalar> inline CVector4Template<Scalar> umin(const CVector4Template<Scalar> &a, const CVector4Template<Scalar> &b)
{
CVector4Template<Scalar> r;r.x = umin(a.x, b.x);
r.y = umin(a.y, b.y);
r.z = umin(a.z, b.z);
r.w = umin(a.w, b.w);
return r;
}
template<class Scalar> inline CVector4Template<Scalar> umax(const CVector4Template<Scalar> &a, const CVector4Template<Scalar> &b)
{
CVector4Template<Scalar> r;r.x = umax(a.x, b.x);
r.y = umax(a.y, b.y);
r.z = umax(a.z, b.z);
r.w = umax(a.w, b.w);
return r;
}
template<class Scalar> inline CVector4Template<Scalar> normalize(const CVector4Template<Scalar> &a)
{
CVector4Template<Scalar> rvec;
Scalar r = sqrt(a.x* a.x+a.y* a.y+a.z* a.z+a.w* a.w);
   if (r < 0.00001) return a;
rvec.x = a.x / r;
rvec.y = a.y / r;
rvec.z = a.z / r;
rvec.w = a.w / r;
  return rvec;
}
template<class Scalar> inline Scalar length(const CVector4Template<Scalar> &a)
{
CVector4Template<Scalar> rvec;
Scalar r = sqrt(a.x*a.x+a.y*a.y+a.z*a.z+a.w*a.w);
   return r;
}
template<class Scalar> inline CVector4Template<Scalar> sign(const CVector4Template<Scalar> &a)
{
CVector4Template<Scalar> rvec;
rvec.x = a.x > 0 ? 1 : -1;
rvec.y = a.y > 0 ? 1 : -1;
rvec.z = a.z > 0 ? 1 : -1;
rvec.w = a.w > 0 ? 1 : -1;
  return rvec;
}
template<class Scalar> inline Scalar distance(const CVector4Template<Scalar> &a, const CVector4Template<Scalar> &b)
{
return sqrt(sqr(a.x-b.x)+sqr(a.y-b.y)+sqr(a.z-b.z)+sqr(a.w-b.w));
}
template<class Scalar> inline CVector4Template<Scalar> linear(const CVector4Template<Scalar> &a, const CVector4Template<Scalar> &b, const Scalar v)
{
CVector4Template<Scalar> rvec;
rvec.x = a.x+(b.x - a.x)*v;
rvec.y = a.y+(b.y - a.y)*v;
rvec.z = a.z+(b.z - a.z)*v;
rvec.w = a.w+(b.w - a.w)*v;
return rvec;
}

template<class Scalar> inline CVector3Template<Scalar> crossProduct(const CVector3Template<Scalar>& v1, const CVector3Template<Scalar>& v2)
{
   CVector3Template<Scalar> res;
   res.x = v1.y*v2.z-v1.z*v2.y;
   res.y = v1.z*v2.x-v1.x*v2.z;
   res.z = v1.x*v2.y-v1.y*v2.x;
   return res;
};

typedef CVector4Template<double> CVector4Double;
typedef CVector2Template<double> CVector4d;
typedef CVector3Template<double> CVector3Double;
typedef CVector3Template<double> CVector3d;
typedef CVector2Template<double> CVector2Double;
typedef CVector2Template<double> CVector2d;

template <class Scalar> struct CRectTemplate;

template <class Scalar> struct CBoxTemplate
{
    CVector3Template<Scalar> m_min;
    CVector3Template<Scalar> m_max;
    inline Scalar height()const  {return m_max.y - m_min.y;};
    inline CVector3Template<Scalar> size() const {return m_max-m_min;};

    static CBoxTemplate INIT() {        CBoxTemplate r; r.init(); return r;};
    static CBoxTemplate ADVANCEINIT() {        CBoxTemplate r; r.m_min=CVector3Template<Scalar>(1000000, 1000000, 1000000); r.m_max=CVector3Template<Scalar>(-1000000, -1000000, -1000000);return r;};

    inline CBoxTemplate() {};
    inline CBoxTemplate(Scalar x1, Scalar y1, Scalar z1, Scalar x2, Scalar y2, Scalar z2) {m_min.x = x1; m_min.y = y1; m_min.z = z1; m_max.x = x2; m_max.y = y2; m_max.z = z2;};
    inline CBoxTemplate(const CVector3Template<Scalar>& v1, const CVector3Template<Scalar>&v2) {m_min = v1; m_max=v2;};
    inline void init() {m_min = CVector3Template<Scalar>(0, 0,0); m_max = CVector3Template<Scalar>(0,0,0);};
    inline CVector3Template<Scalar> center() const {return (m_max+m_min)/Scalar(2);};
    inline Scalar getWidth() const {return m_max.x-m_min.x;};
    inline Scalar getHeight()const  {return m_max.y-m_min.y;};
    inline Scalar getDepth()const  {return m_max.z-m_min.z;};
    inline void setWidth(Scalar v) {m_max.x = m_min.x+v;};
    inline void setHeight(Scalar v) {m_max.y = m_min.y+v;};
    inline void setDepth(Scalar v) {m_max.z = m_min.z+v;};
    inline Scalar getLeft()const  {return m_min.x;};
    inline Scalar getTop() const {return m_min.y;};
    inline Scalar getFront() const {return m_min.z;};
    inline Scalar getRight()const  {return m_max.x;};
    inline Scalar getBottom() const {return m_max.y;};
    inline Scalar getBack() const {return m_max.z;};
     inline CVector3Template<Scalar> getSize()const  {return m_max-m_min;};
     inline void increaseSize(const CVector3Template<Scalar>& v) {m_min-=v; m_max+= v;};
    inline void move(const CVector3Template<Scalar>& v) { m_min+= v; m_max += v;};
    inline void setLeft(Scalar v) {move(CVector3Template<Scalar>(v - m_min.x, 0,0));};
    inline void setTop(Scalar v) {move(CVector3Template<Scalar>(0, v - m_min.y,0));};
    inline void setFront(Scalar v) {move(CVector3Template<Scalar>(0, 0, v - m_min.z));};
    inline void correct() {if (m_min.x > m_max.x) {utiswap(m_min.x, m_max.x);}; if (m_min.y > m_max.y) {utiswap(m_min.y, m_max.y);}; if (m_min.z > m_max.z) {utiswap(m_min.z, m_max.z);};}
    inline bool contains(const CVector3Template<Scalar>& position) const {
       return (m_min.x <= position.x ) && (m_max.x >= position.x) && (m_min.y <= position.y) && (m_max.y >= position.y)&& (m_min.z <= position.z) && (m_max.z >= position.z);
    };
    inline bool intersectMathCorrect(const CVector3Template<Scalar>& position) const {
       return (m_min.x <= position.x ) && (m_max.x > position.x) && (m_min.y <= position.y) && (m_max.y > position.y)&& (m_min.z <= position.z) && (m_max.z > position.z);
    };
    inline CVector3Template<Scalar> interpolate(const CVector3Template<Scalar>& v)const
    {
        return CVector3Template<Scalar>(m_min.x+(m_max.x-m_min.x)*v.x,m_min.y+(m_max.y-m_min.y)*v.y,m_min.z+(m_max.z-m_min.z)*v.z);
    }

    inline void expand(const CBoxTemplate& rect) {
      if (m_min.x > rect.m_min.x)m_min.x = rect.m_min.x;

      if (m_max.x < rect.m_max.x) m_max.x = rect.m_max.x;
      if (m_min.y > rect.m_min.y)m_min.y = rect.m_min.y;

      if (m_max.y < rect.m_max.y) m_max.y = rect.m_max.y;
      if (m_min.z > rect.m_min.z)m_min.z = rect.m_min.z;

      if (m_max.z < rect.m_max.z) m_max.z = rect.m_max.z;
    };
    inline void expand(const CVector3& v)
    {
         if (m_min.x > v.x)m_min.x = v.x;
      if (m_max.x < v.x) m_max.x = v.x;
      if (m_min.y > v.y)m_min.y = v.y;
      if (m_max.y < v.y) m_max.y = v.y;
      if (m_min.z > v.z)m_min.z = v.z;
      if (m_max.z < v.z) m_max.z = v.z;

    }
    inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y <= m_max.y && m_min.z <= m_max.z;
    }
    inline bool intersect(const CBoxTemplate& rect) const {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      if (m_min.x>= rect.m_max.x) return false;
      if (m_min.y>= rect.m_max.y) return false;
      if (m_min.z>= rect.m_max.z) return false;

      if (m_max.x<= rect.m_min.x) return false;
      if (m_max.y<= rect.m_min.y) return false;
      if (m_max.z<= rect.m_min.z) return false;

      return true;
    };
    inline bool intersectMathCorrect(const CBoxTemplate& rect) const {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(m_min.z <= m_max.z);
      if (m_min.x> rect.m_max.x) return false;
      if (m_min.y> rect.m_max.y) return false;
      if (m_min.z> rect.m_max.z) return false;

      if (m_max.x<= rect.m_min.x) return false;
      if (m_max.y<= rect.m_min.y) return false;
      if (m_max.z<= rect.m_min.z) return false;

      return true;
    };
    inline bool operator==(const CBoxTemplate& other) const
    {
        return other.m_max == other.m_max && other.m_min == other.m_min;
    };
    inline bool operator!=(const CBoxTemplate& other) const {return !(operator==(other));};
    inline int getMaximumAxis() const
    {
        int res = 0;
    Scalar size = getWidth();
    if (getHeight() > size) {size = getHeight(); res = 1;};
    if (getDepth() > size) {size = getDepth(); res =2;};
    return res;
    };
    inline Scalar getMaxSize() const
    {
        Scalar w = getWidth();
        Scalar t = getHeight();
        if (t > w) w = t;
        t = getDepth();
        if (t > w) w = t;
        return w;
    };
    inline CRectTemplate<Scalar> xy()const ;
    inline CRectTemplate<Scalar> xz()const ;
    inline CRectTemplate<Scalar> yz()const ;
    //Scalar getMinMaxDistance(const CVector3Template<Scalar>& p)  const;
    void octreeSubdivide(CBoxTemplate* boxes) const;
    inline void createOctreeBox(int index, CBoxTemplate& octreebox) const
    {
       Scalar xhalf = (m_max.x-m_min.x) / Scalar(2);

       Scalar yhalf = (m_max.y-m_min.y)  / Scalar(2);
       Scalar zhalf = (m_max.z-m_min.z) / Scalar(2);

       Scalar v = Scalar (index & 1);
       octreebox.m_min.x = v*xhalf+m_min.x;
       octreebox.m_max.x = (v+1)*xhalf+m_min.x;
       v = Scalar((index & 2) >> 1);
       octreebox.m_min.y = (v)*yhalf+m_min.y;
       octreebox.m_max.y = (v+1)*yhalf+m_min.y;
       v = Scalar( (index & 4) >> 2);
       octreebox.m_min.z = (v)*zhalf+m_min.z;
       octreebox.m_max.z = (v+1)*zhalf+m_min.z;


    }
};


template<class Scalar> struct CRectTemplate
{
     typedef CVector2Template<Scalar> VectorClass;
    static CRectTemplate INIT() {CRectTemplate r; r.init(); return r;};
    static CRectTemplate ADVANCEINIT() {CRectTemplate r; r.m_min=CVector2Template<Scalar>(1000000, 1000000); r.m_max=CVector2Template<Scalar>(-1000000, -1000000);return r;};
    static CRectTemplate MAXINIT() {CRectTemplate r; r.m_max=CVector2Template<Scalar>(1000000, 1000000); r.m_min=CVector2Template<Scalar>(-1000000, -1000000);return r;};
    CVector2Template<Scalar> m_min;
    CVector2Template<Scalar> m_max;
    inline CRectTemplate() {};
    inline CRectTemplate(Scalar x1, Scalar y1, Scalar x2, Scalar y2) {m_min.x = x1; m_min.y = y1; m_max.x = x2; m_max.y = y2;};
    inline CRectTemplate(const CVector2Template<Scalar>& v1, const CVector2Template<Scalar>&v2) {m_min = v1; m_max=v2;};
    inline explicit CRectTemplate(const CIntRect& rect) {m_min.x = rect.m_min.x; m_min.y = rect.m_min.y; m_max.x = rect.m_max.x; m_max.y = rect.m_max.y;};
    inline CRectTemplate(const CVector2Template<Scalar>& pos, Scalar size) {m_min.x = pos.x-size/Scalar(2);m_min.y = pos.y-size/Scalar(2);m_max.x = pos.x+size/Scalar(2);m_max.y = pos.y+size/Scalar(2);};
    inline void init() {m_min = CVector2Template<Scalar>(0, 0); m_max = CVector2Template<Scalar>(0,0);};
    inline CBoxTemplate<Scalar> getBounds(Scalar zmin=0, Scalar zmax=1) const {        CBoxTemplate<Scalar> b; b.m_min = CVector3(m_min, zmin);b.m_max = CVector3(m_max, zmax);return b;};
    inline CVector2Template<Scalar> getCoord(const CVector2Template<Scalar>& pos) const {
      return CVector2Template<Scalar>(m_min.x+(m_max.x-m_min.x)*pos.x,
              m_min.y+(m_max.y-m_min.y)*pos.y);
    };
    inline CVector2Template<Scalar> center() const {return (m_max+m_min)/Scalar(2);};
    void getRectAsArray(Array<CVector2> &marray);
    inline Scalar getWidth() const {return m_max.x-m_min.x;};
    inline Scalar getHeight()const  {return m_max.y-m_min.y;};
    inline void setWidth(Scalar v) {m_max.x = m_min.x+v;};
    inline void setHeight(Scalar v) {m_max.y = m_min.y+v;};
    inline CVector2Template<Scalar> getSize() {return m_max-m_min;};
    inline Scalar getLeft()const  {return m_min.x;};
    inline Scalar getTop() const {return m_min.y;};
    inline Scalar getRight()const  {return m_max.x;};
    inline Scalar getBottom() const {return m_max.y;};
    inline void move(const CVector2Template<Scalar>& v) { m_min+= v; m_max += v;};
    inline void setLeft(Scalar v) {move(CVector2Template<Scalar>(v - m_min.x, 0));};
    inline void setTop(Scalar v) {move(CVector2Template<Scalar>(0, v - m_min.y));};
    inline void correct() {if (m_min.x > m_max.x) {utiswap(m_min.x, m_max.x);}; if (m_min.y > m_max.y) {utiswap(m_min.y, m_max.y);};}
    inline bool contains(const CVector2Template<Scalar>& position) const {
       return (m_min.x <= position.x ) && (m_max.x >= position.x) && (m_min.y <= position.y) && (m_max.y >= position.y);
    };
    inline void expandRect(Scalar w)
    {
      m_min.x -= w;
      m_min.y -= w;
      m_max.x += w;
      m_max.y += w;
    }
    inline void roundToInt()
    {
      m_min.x = (int)m_min.x;
      m_min.y = (int)m_min.y;
      m_max.x = (int)m_max.x;
      m_max.y = (int)m_max.y;
    };
    inline void expandRect(const CVector2Template<Scalar>& v)
    {
      m_min.x -= v.x;
      m_min.y -= v.y;
      m_max.x += v.x;
      m_max.y += v.y;
    }
    inline void expand(const CRectTemplate& rect) {
      if (m_min.x > rect.m_min.x)m_min.x = rect.m_min.x;

      if (m_max.x < rect.m_max.x) m_max.x = rect.m_max.x;
      if (m_min.y > rect.m_min.y)m_min.y = rect.m_min.y;

      if (m_max.y < rect.m_max.y) m_max.y = rect.m_max.y;
    };
    inline void expand(const CVector2Template<Scalar>& v)
    {
         if (m_min.x > v.x)m_min.x = v.x;
      if (m_max.x < v.x) m_max.x = v.x;
      if (m_min.y > v.y)m_min.y = v.y;
      if (m_max.y < v.y) m_max.y = v.y;

    }


    inline bool contains(const CRectTemplate& v) const
    {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      utiassert(v.valid());

      if (m_min.x > v.m_min.x) return false;
      if (m_min.y > v.m_min.y) return false;
      if (m_max.x < v.m_max.x) return false;
      if (m_max.y < v.m_max.y) return false;
      return true;
    };

    inline bool valid() const {
       return m_min.x <= m_max.x && m_min.y <= m_max.y;
    }
   inline bool intersect(const CRectTemplate& rect) const {
      utiassert(m_min.x <= m_max.x);
      utiassert(m_min.y <= m_max.y);
      if (m_min.x> rect.m_max.x) return false;
      if (m_min.y> rect.m_max.y) return false;
      if (m_max.x< rect.m_min.x) return false;
      if (m_max.y< rect.m_min.y) return false;
      return true;
    };
    inline bool operator==(const CRectTemplate& other) const
    {
        return m_max == other.m_max && m_min == other.m_min;
    };
    inline bool operator!=(const CRectTemplate& other) const {return !(operator==(other));};
    inline Scalar getArea()
    {
        Scalar res = (m_max.x-m_min.x)*(m_max.y-m_min.y);
        if (res < 0) res = -res;
      return res;
    };
};

template<class Scalar> inline CRectTemplate<Scalar> CBoxTemplate<Scalar>::xy()const {CRectTemplate<Scalar> r;r.m_min = m_min.xy(); r.m_max = m_max.xy();return r;};
template<class Scalar> inline CRectTemplate<Scalar> CBoxTemplate<Scalar>::xz() const {CRectTemplate<Scalar> r;r.m_min = m_min.xz(); r.m_max = m_max.xz();return r;};
template<class Scalar> inline CRectTemplate<Scalar> CBoxTemplate<Scalar>::yz()const  {CRectTemplate<Scalar> r;r.m_min = m_min.yz(); r.m_max = m_max.yz();return r;};



inline CIVector3 floatVecToInt(const CVector3d& v)
{
  CIVector3 r;
  r.x = (int)round(v.x);
  r.y = (int)round(v.y);
  r.z = (int)round(v.z);

  return r;
};

inline CIVector2 floatVecToInt(const CVector2d& v)
{
  CIVector2 r;
  r.x = (int)round(v.x);
  r.y = (int)round(v.y);

  return r;
};

inline CVector3 toFloatVec(const CVector3d& v)
{
    return CVector3(v.x, v.y, v.z);
}

inline CVector3d toFloatVec(const CVector3& v)
{
    return CVector3d(v.x, v.y, v.z);
}

#endif
