/***************************************************************************
 *   Copyright (C) 2018 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "pddllib.h"
#include "syntheticrule.h"

CRPDDLCompiledAction* m_compileaction = NULL;

CUtiString SRuntimeVariable::print()
{
   return PrologSystem::toLisp(*this).print();
}

CUtiString SRuntimeVariable::printPDDL(std::set<CUtiIDString> &usedvars)
{
   SRuntimeVariable var = *this;
   switch (var.m_type) {
    case rvtIntegerValue:
    {

        return CUtiString::intToStr( var.m_integer_value);
    }; break;
    case rctFloatValue:
    {

        return CUtiString::floatToStr(var.m_float_value, 0, 6);

    }; break;
    case rctBoolValue:
    {

        return CUtiString::intToStr(var.m_bool_value ? 1 : 0);
    }; break;
    case rctStringValue:
    {
        return EMPTYSTRING;

    }; break;
    case rctAtomValue:
    {

        return UTISTR("")+var.m_atom_value.str();
    }; break;
    case rvtIntegerPDDL:
    {
        usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
        return UTISTR("?")+var.m_atom_value.str();
    }; break;
    case rvtFloatPDDL:
    {

        usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
        return UTISTR("?")+var.m_atom_value.str();
    }; break;
    case rvtBoolPDDL:
    {

        usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
        return UTISTR("?")+var.m_atom_value.str();
    }; break;
    case rvtAtomPDDL:
    {

        usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
        return UTISTR("?")+var.m_atom_value.str();
    }; break;
    default:  {
 assert(false);
    }; break;
    }
    return EMPTYSTRING;
}

namespace PrologSystem {

TPreconditionType getPLState(FP_term term)
{
    FP_atom resultvalue;
    bool res = FP_get_atom(term, &resultvalue);
    CUtiString st(FP_wstring_from_atom(resultvalue));
    if (st == UTISTR("start")) {
        return ttStartCondition;
    }
    if (st == UTISTR("end")) {
        return ttEndCondition;
    }
    if (st == UTISTR("overall")) {
        return ttOverAllCondition;
    }


    assert(false);
}

SRuntimeVariable fromProlog(FP_term term)
{
    FP_atom name;
    bool res = FP_get_structure(term, &name, NULL);//FP_get_atom(term, &name);
    assert(res == true);
    CUtiString tc = FP_wstring_from_atom(name);
    FP_term value;
     res=  FP_get_arg(term, 0, &value);
    assert(res == true);
    SRuntimeVariable v;

    if (tc == UTISTR("vint")) {
       v.m_type =rvtIntegerValue;
       FP_int resultvalue;
       res = FP_get_integer(value, &resultvalue);
       v.m_integer_value = resultvalue;
    } else if (tc == UTISTR("vfloat")) {
        v.m_type =rctFloatValue;
        FP_real resultvalue;
        res = FP_get_real(value, &resultvalue);
        v.m_float_value = resultvalue;
    } else if (tc == UTISTR("vbool")) {
        v.m_type =rctBoolValue;
        FP_int resultvalue;
        res = FP_get_integer(value, &resultvalue);
        v.m_bool_value = resultvalue != 0;
    } else if (tc == UTISTR("vatom")) {
        v.m_type =rctAtomValue;
        FP_atom resultvalue;
        res = FP_get_atom(value, &resultvalue);
        v.m_atom_value = CUtiIDString(CUtiString(FP_wstring_from_atom(resultvalue)));
    } else if (tc == UTISTR("vstring")) {
        v.m_type =rctStringValue;
        FP_atom resultvalue;
        res = FP_get_atom(value, &resultvalue);
        v.m_atom_value = CUtiIDString(CUtiString(FP_wstring_from_atom(resultvalue)));
    } else if (tc == UTISTR("vvari")) {
        v.m_type =rvtIntegerPDDL;
        FP_atom resultvalue;
        res = FP_get_atom(value, &resultvalue);
        v.m_atom_value = CUtiIDString(CUtiString(FP_wstring_from_atom(resultvalue)));
    } else if (tc == UTISTR("vvarf")) {
        v.m_type =rvtFloatPDDL;
        FP_atom resultvalue;
        res = FP_get_atom(value, &resultvalue);
        v.m_atom_value = CUtiIDString(CUtiString(FP_wstring_from_atom(resultvalue)));
    } else if (tc == UTISTR("vvarb")) {
        v.m_type =rvtBoolPDDL;
        FP_atom resultvalue;
        res = FP_get_atom(value, &resultvalue);
        v.m_atom_value = CUtiIDString(CUtiString(FP_wstring_from_atom(resultvalue)));
    } else if (tc == UTISTR("vvara")) {
        v.m_type =rvtAtomPDDL;
        FP_atom resultvalue;
        res = FP_get_atom(value, &resultvalue);
        v.m_atom_value = CUtiIDString(CUtiString(FP_wstring_from_atom(resultvalue)));
    }else {
        assert(false);
    }

    return v;
}
FP_term toProlog(const SRuntimeVariable& var)
{
   switch (var.m_type) {
   case rvtIntegerValue:
   {

       FP_term term = FP_create_integer(var.m_integer_value);
       return FP_create_structure_1(FP_atom_from_wstring(L"vint"), term);
   }; break;
   case rctFloatValue:
   {

       FP_term term = FP_create_real(var.m_float_value);
       return FP_create_structure_1(FP_atom_from_wstring(L"vfloat"), term);
   }; break;
   case rctBoolValue:
   {

       FP_term term = FP_create_integer(var.m_bool_value ? 1 : 0);
       return FP_create_structure_1(FP_atom_from_wstring(L"vbool"), term);
   }; break;
   case rctStringValue:
   {
       FP_term term = FP_create_atom(FP_atom_from_wstring(var.m_atom_value.str().c_str()));
       return FP_create_structure_1(FP_atom_from_wstring(L"vstring"), term);
   }; break;
   case rctAtomValue:
   {

       FP_term term = FP_create_atom(FP_atom_from_wstring(var.m_atom_value.str().c_str()));
       return FP_create_structure_1(FP_atom_from_wstring(L"vatom"), term);
   }; break;
   case rvtIntegerPDDL:
   {

       FP_term term = FP_create_atom(FP_atom_from_wstring(var.m_atom_value.str().c_str()));
       return FP_create_structure_1(FP_atom_from_wstring(L"vvari"), term);
   }; break;
   case rvtFloatPDDL:
   {

       FP_term term = FP_create_atom(FP_atom_from_wstring(var.m_atom_value.str().c_str()));
       return FP_create_structure_1(FP_atom_from_wstring(L"vvarf"), term);
   }; break;
   case rvtBoolPDDL:
   {

       FP_term term = FP_create_atom(FP_atom_from_wstring(var.m_atom_value.str().c_str()));
       return FP_create_structure_1(FP_atom_from_wstring(L"vvarb"), term);
   }; break;
   case rvtAtomPDDL:
   {

       FP_term term = FP_create_atom(FP_atom_from_wstring(var.m_atom_value.str().c_str()));
       return FP_create_structure_1(FP_atom_from_wstring(L"vvara"), term);
   }; break;
   default:  {
assert(false);
   }; break;
   }
   return 0;

}


Lisp::LObject toLisp(const SRuntimeVariable& var)
{
    switch (var.m_type) {
    case rvtIntegerValue:
    {

        return Lisp::list(IDSTR("vint"), var.m_integer_value);
    }; break;
    case rctFloatValue:
    {

        return Lisp::list(IDSTR("vfloat"), var.m_float_value);

    }; break;
    case rctBoolValue:
    {

        return Lisp::list(IDSTR("vbool"), var.m_bool_value ? 1 : 0);
    }; break;
    case rctStringValue:
    {
        return Lisp::list(IDSTR("vstring"), var.m_atom_value);

    }; break;
    case rctAtomValue:
    {


        return Lisp::list(IDSTR("vatom"), var.m_atom_value);
    }; break;
    case rvtIntegerPDDL:
    {

        return Lisp::list(IDSTR("vvari"), var.m_atom_value);
    }; break;
    case rvtFloatPDDL:
    {

        return Lisp::list(IDSTR("vvarf"), var.m_atom_value);
    }; break;
    case rvtBoolPDDL:
    {

        return Lisp::list(IDSTR("vvarb"), var.m_atom_value);
    }; break;
    case rvtAtomPDDL:
    {

        return Lisp::list(IDSTR("vvara"), var.m_atom_value);
    }; break;
    default:  {
 assert(false);
    }; break;
    }
    return Lisp::LObject();
}

bool pl_foreign_outputPDDLNum(FP_term a, FP_term value)
{

    int lcount = FP_get_list_count(a);
    Array<FP_term> t1(lcount);
    DynArray1D<SRuntimeVariable> runtime;
    for (int i = 0; i < lcount; i++) {
        bool res=  FP_get_list_element(a, i, &t1[i]);
        assert(res == true);
        if (i> 0) {
            runtime.push_back(fromProlog(t1[i]));
        }
    }
    FP_atom name;
    FP_get_atom(t1[0], &name);
    CUtiString tc = FP_wstring_from_atom(name);
    std::wcout << "Found Name Command: " << tc << std::endl;
    SRuntimeVariable res;
    CRPDDLNumericOperation* num = NULL;
    if (tc == UTISTR("add") || tc == UTISTR("sub") || tc == UTISTR("mul") || tc == UTISTR("div")) {

        SRuntimeVariable p1 = runtime[0];
        SRuntimeVariable p2 = runtime[1];

        TRExpressionType et = eAdd;
        if (tc == UTISTR("add")) {
            et = eAdd;
        } else if (tc == UTISTR("sub")) {
            et = eSub;
        } else if (tc == UTISTR("mul")) {
            et = eMul;
        } else if (tc == UTISTR("div")) {
            et = eDiv;
        }

        if (p1.m_type == rctFloatValue && p2.m_type == rctFloatValue) {
           res.m_type = rctFloatValue;
           switch (et) {
           case eAdd: {

               res.m_float_value = p1.m_float_value + p2.m_float_value;
           }; break;
           case eSub: {

               res.m_float_value = p1.m_float_value - p2.m_float_value;
           }; break;
           case eMul: {

               res.m_float_value = p1.m_float_value * p2.m_float_value;
           }; break;
           case eDiv: {

               res.m_float_value = p1.m_float_value / p2.m_float_value;
           }; break;
           }


        }else if (p1.m_type == rvtIntegerValue && p2.m_type == rvtIntegerValue) {
           res.m_type = rvtIntegerValue;
           switch (et) {
           case eAdd: {

               res.m_integer_value = p1.m_integer_value + p2.m_integer_value;
           }; break;
           case eSub: {

               res.m_integer_value = p1.m_integer_value - p2.m_integer_value;
           }; break;
           case eMul: {

               res.m_integer_value = p1.m_integer_value * p2.m_integer_value;
           }; break;
           case eDiv: {

               res.m_integer_value = p1.m_integer_value / p2.m_integer_value;
           }; break;
           }


        }else  if (p1.m_type == rvtFloatPDDL || p2.m_type == rvtFloatPDDL)   {
           CUtiIDString newexistsatom = m_compileaction->getNewPDDLVariable();

            num = new CRPDDLNumericOperation();
            num->m_op1 = p1;
            num->m_op2 = p2;
           num->m_res.m_atom_value = newexistsatom;
           num->m_res.m_type = rvtFloatPDDL;
           num->m_type = et;
           res = num->m_res;

        }else  if (p1.m_type == rvtIntegerPDDL || p2.m_type == rvtIntegerPDDL)   {
           CUtiIDString newexistsatom = m_compileaction->getNewPDDLVariable();

            num = new CRPDDLNumericOperation();
            num->m_op1 = p1;
            num->m_op2 = p2;
           num->m_res.m_atom_value = newexistsatom;
           num->m_res.m_type = rvtIntegerPDDL;
           num->m_type = et;
           res = num->m_res;

        } else {
            assert(false);
        }

    } else if (tc == UTISTR("less") || tc == UTISTR("greater")|| tc == UTISTR("eless")|| tc == UTISTR("egreater")|| tc == UTISTR("equal")) {

        SRuntimeVariable p1 = runtime[0];
        SRuntimeVariable p2 = runtime[1];

        TRExpressionType et = eLess;
        if (tc == UTISTR("less")) {
            et = eLess;
        } else if (tc == UTISTR("greater")) {
            et = eGreater;
        } else if (tc == UTISTR("eless")) {
            et = eELess;
        } else if (tc == UTISTR("egreater")) {
            et = eEGreater;
        } else if (tc == UTISTR("equal")) {
            et = eEqual;
        }

        if (p1.m_type == rctFloatValue && p2.m_type == rctFloatValue) {
           res.m_type = rctBoolValue;
           switch (et) {
           case eLess: {
               res.m_bool_value = p1.m_float_value < p2.m_float_value;
           }; break;

           case eGreater: {
               res.m_bool_value = p1.m_float_value > p2.m_float_value;
           }; break;

           case eELess: {
               res.m_bool_value = p1.m_float_value <= p2.m_float_value;
           }; break;

           case eEGreater: {
               res.m_bool_value = p1.m_float_value >= p2.m_float_value;
           }; break;

           case eEqual: {
               res.m_bool_value = p1.m_float_value == p2.m_float_value;
           }; break;
           }


        }else if (p1.m_type == rvtIntegerValue && p2.m_type == rvtIntegerValue) {
           res.m_type = rctBoolValue;
           switch (et) {
           case eLess: {
               res.m_bool_value = p1.m_integer_value < p2.m_integer_value;

           }; break;

           case eGreater: {
               res.m_bool_value = p1.m_integer_value > p2.m_integer_value;

           }; break;

           case eELess: {
               res.m_bool_value = p1.m_integer_value <= p2.m_integer_value;

           }; break;

           case eEGreater: {
               res.m_bool_value = p1.m_integer_value >= p2.m_integer_value;

           }; break;

           case eEqual: {
               res.m_bool_value = p1.m_integer_value == p2.m_integer_value;

           }; break;
           }

        }else  if (p1.m_type == rvtFloatPDDL || p2.m_type == rvtFloatPDDL)   {
           CUtiIDString newexistsatom = m_compileaction->getNewPDDLVariable();

            num = new CRPDDLNumericOperation();
            num->m_op1 = p1;
            num->m_op2 = p2;
           num->m_res.m_atom_value = newexistsatom;
           num->m_res.m_type = rvtBoolPDDL;
           num->m_type = et;
           res = num->m_res;

        }else  if (p1.m_type == rvtIntegerPDDL || p2.m_type == rvtIntegerPDDL)   {
           CUtiIDString newexistsatom = m_compileaction->getNewPDDLVariable();

            num = new CRPDDLNumericOperation();
            num->m_op1 = p1;
            num->m_op2 = p2;
           num->m_res.m_atom_value = newexistsatom;
           num->m_res.m_type = rvtBoolPDDL;
           num->m_type = et;
           res = num->m_res;

        }else {
            assert(false);
        }


    }else if (tc == UTISTR("logicalor") || tc == UTISTR("logicaland")) {

        SRuntimeVariable p1 = runtime[0];
        SRuntimeVariable p2 = runtime[1];

        TRExpressionType et = eLogicalOr;
        if (tc == UTISTR("logicalor")) {
            et = eLogicalOr;
        } else if (tc == UTISTR("logicaland")) {
            et = eLogicalAnd;
        }
        if (p1.m_type == rctBoolValue && p2.m_type == rctBoolValue) {
           res.m_type = rctBoolValue;
           switch (et) {
           case eLogicalOr: {
               res.m_bool_value = p1.m_bool_value || p2.m_bool_value;
           }; break;

           case eLogicalAnd: {
               res.m_bool_value = p1.m_bool_value && p2.m_bool_value;
           }; break;


           }

        }else  if (p1.m_type == rvtBoolPDDL || p2.m_type == rvtBoolPDDL)   {
           CUtiIDString newexistsatom = m_compileaction->getNewPDDLVariable();

            num = new CRPDDLNumericOperation();
            num->m_op1 = p1;
            num->m_op2 = p2;
           num->m_res.m_atom_value = newexistsatom;
           num->m_res.m_type = rvtBoolPDDL;
           num->m_type = et;
           res = num->m_res;

        }else {
            assert(false);
        }


    }

    FP_term term  = toProlog(res);
    FP_unify(value, term);
    if (num) {
       m_compileaction->m_globalstore.insert(std::make_pair(res.m_atom_value, uti_refptr<CGlobalStoreCommand>(num)));

    }

    return true;
}

bool pl_foreign_outputPDDLBool (FP_term a, FP_term value)
{

    pl_foreign_outputPDDLNum(a, value);
    return true;
}

bool pl_foreign_outputPDDLAssert (FP_term a, FP_term state)
{

    TPreconditionType plstate = getPLState(state);

    SRuntimeVariable var = fromProlog(a);
    if (m_compileaction->getCurrentPath() == m_compileaction->m_block.get()) {
        m_compileaction->m_preconditions.push_back(std::make_pair(var, plstate));
    } else {
        std::wcout << "You cannot add a precondition after an if statement or a loop" << std::endl;
    }


    return true;
}


bool pl_foreign_createPDDLEmptyFluent (FP_term value, FP_term type)
{
    FP_atom name;
    FP_get_atom(type, &name);
    const wchar_t* tc = FP_wstring_from_atom(name);
    std::wcout << "Found Name Command: " << tc << std::endl;
    CUtiString s = tc;


    //createPDDLEmptyFluent(Value, Type)
    CUtiIDString newexistsatom = m_compileaction->getNewPDDLVariable();
    SRuntimeVariable var;
    var.m_atom_value = newexistsatom;
    if (s == UTISTR("float")) {
    var.m_type = rvtFloatPDDL;
    } else if (s == UTISTR("int")) {
    var.m_type = rvtIntegerPDDL;

    }
    FP_term term  = toProlog(var);
    FP_unify(value, term);
    return true;
}

bool pl_foreign_outputPDDLExists (FP_term a, FP_term value)
{
    int lcount = FP_get_list_count(a);
    Array<FP_term> t1(lcount);
    DynArray1D<SRuntimeVariable> runtime;
    for (int i = 0; i < lcount; i++) {
        bool res=  FP_get_list_element(a, i, &t1[i]);
        assert(res == true);
        if (i> 0) {
            runtime.push_back(fromProlog(t1[i]));
        }
    }
    FP_atom name;
    FP_get_atom(t1[0], &name);
    const wchar_t* tc = FP_wstring_from_atom(name);
    std::wcout << "Found Name Command: " << tc << std::endl;


    uti_refptr<CRPDDLEffect> effect = new CRPDDLEffect();
    effect->m_name = CUtiIDString(CUtiString(tc));
    effect->m_add = true;
    effect->m_runtimevars.push_back(runtime);

    CRPDDLExists* exists = new CRPDDLExists();
    exists->m_effect = effect;

//create fluents
    auto it = m_compileaction->m_program->m_nametotable.find(effect->m_name);
    assert(it != m_compileaction->m_program->m_nametotable.end());
    CRStructDeclaration* s = it->second;
    int atomcount = 0;
    {
        DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> > params;
        s->createFlatVariableList(params, EMPTYSTRING);
        assert(params.size() == runtime.size());

        // output.add(d + ef.printPDDL(usedvars));
        {
            for (int fluentindex = 0; fluentindex < params.size(); fluentindex++) {

                if (dynamic_cast<CRClassDeclaration*>(params[fluentindex].second) != NULL) {
                    atomcount++;
                    continue;
                }
                CRPDDLFluentAccess* fluentaccess = new CRPDDLFluentAccess();
                fluentaccess->m_accessindex = fluentindex;
                fluentaccess->m_predicatename =effect->m_name.str();
                s->createFlatVariableList(fluentaccess->m_list, EMPTYSTRING);
                fluentaccess->m_struct = s;
                fluentaccess->m_runtimevariable.push_back(runtime);
                SRuntimeVariable targetvariable = runtime[fluentindex];
                assert(targetvariable.m_type == rvtIntegerPDDL || targetvariable.m_type == rvtFloatPDDL);
                assert(m_compileaction->m_globalstore.find(targetvariable.m_atom_value) == m_compileaction->m_globalstore.end());
                m_compileaction->m_globalstore.insert(std::make_pair(targetvariable.m_atom_value, uti_refptr<CGlobalStoreCommand>(fluentaccess)));
                /*for (int i = 0; i < params.size(); i++) {
                    if (dynamic_cast<CRClassDeclaration*>(params[i].second) != NULL) {
                        paramsres += UTISTR(" ")+ef.m_runtimevars[i].printPDDL(usedvars);
                        c++;
                    }
                }*/
            }
        }
    }


    if (atomcount != 0) {
        CUtiIDString newexistsatom = m_compileaction->getNewPDDLVariable();
        SRuntimeVariable var;
        var.m_atom_value = newexistsatom;
        var.m_type = rvtBoolPDDL;
        FP_term term  = toProlog(var);
        FP_unify(value, term);
        m_compileaction->m_globalstore.insert(std::make_pair(var.m_atom_value, uti_refptr<CGlobalStoreCommand>(exists)));
    } else {
        // CUtiIDString newexistsatom = m_compileaction->getNewPDDLVariable();
        SRuntimeVariable var;
        var.m_bool_value = true;
        var.m_type = rctBoolValue;
        FP_term term  = toProlog(var);
        FP_unify(value, term);
        m_compileaction->m_globalstore.insert(std::make_pair(var.m_atom_value, uti_refptr<CGlobalStoreCommand>(exists)));
    }


    return true;
}


bool pl_foreign_outputEffect (FP_term a, FP_term state)
{

    TPreconditionType plstate = getPLState(state);

     int lcount = FP_get_list_count(a);
    Array<FP_term> t1(lcount);
    DynArray1D<SRuntimeVariable> runtime;
    for (int i = 0; i < lcount; i++) {
        bool res=  FP_get_list_element(a, i, &t1[i]);
        assert(res == true);
        if (i> 0) {
            runtime.push_back(fromProlog(t1[i]));
        }
    }
    FP_atom name;
    FP_get_atom(t1[0], &name);
    const wchar_t* tc = FP_wstring_from_atom(name);

    CRPDDLEffect* effect = new CRPDDLEffect();
    effect->m_name = CUtiIDString(CUtiString(tc));
    effect->m_add = true;
    effect->m_plstate = plstate;
    effect->m_runtimevars.push_back(runtime);

    m_compileaction->getCurrentPath()->m_addremovecommand.push_back(effect);

    /*int lcount = FP_get_list_count(a);
    Array<FP_term> t1(lcount);
    for (int i = 0; i < lcount; i++) {
        bool res=  FP_get_list_element(a, i, &t1[lcount]);
        assert(res == true);
    }
    DynArray1D<Lisp::LObject> neweffect;
    for (int i = 0; i < t1.size(); i++) {
       FP_atom name;
       FP_real r;
       if (FP_get_atom(t1[i], &name)) {
           const wchar_t* tc = FP_wstring_from_atom(name);
          neweffect.push_back(Lisp::LObject(tc));
       }else if (FP_get_real(t1[i], &r)) {
          neweffect.push_back(Lisp::LObject(r));

       } else {
           assert(false);
       }
    }
    Lisp::LObject neweffectobj = Lisp::list(neweffect);
   // m_compileaction->getCurrentPath()->m_addremovecommand.push_back(Lisp::list(IDSTR("not"), neweffectobj));
    m_compileaction->getCurrentPath()->m_addremovecommand.push_back(neweffectobj);*/
    return true;
}


bool pl_foreign_outputNotEffect (FP_term a, FP_term state)
{


    TPreconditionType plstate = getPLState(state);
 int lcount = FP_get_list_count(a);
    Array<FP_term> t1(lcount);
    DynArray1D<SRuntimeVariable> runtime;
    for (int i = 0; i < lcount; i++) {
        bool res=  FP_get_list_element(a, i, &t1[i]);
        assert(res == true);
        if (i> 0) {
            runtime.push_back(fromProlog(t1[i]));
        }
    }
    FP_atom name;
    FP_get_atom(t1[0], &name);
    const wchar_t* tc = FP_wstring_from_atom(name);

    CRPDDLEffect* effect = new CRPDDLEffect();
    effect->m_name = CUtiIDString(CUtiString(tc));
    effect->m_add = false;
    effect->m_plstate = plstate;
    effect->m_runtimevars.push_back(runtime);


    m_compileaction->getCurrentPath()->m_addremovecommand.push_back(effect);

   /* int lcount = FP_get_list_count(a);
    Array<FP_term> t1(lcount);
    for (int i = 0; i < lcount; i++) {
        bool res=  FP_get_list_element(a, i, &t1[lcount]);
        assert(res == true);
    }
    DynArray1D<Lisp::LObject> neweffect;
    for (int i = 0; i < t1.size(); i++) {
       FP_atom name;
       FP_real r;
       if (FP_get_atom(t1[i], &name)) {
           const wchar_t* tc = FP_wstring_from_atom(name);
          neweffect.push_back(Lisp::LObject(tc));
       }else if (FP_get_real(t1[i], &r)) {
          neweffect.push_back(Lisp::LObject(r));

       } else {
           assert(false);
       }
    }
    Lisp::LObject neweffectobj = Lisp::list(neweffect);
   m_compileaction->getCurrentPath()->m_addremovecommand.push_back(Lisp::list(IDSTR("not"), neweffectobj));*/
    //m_compileaction->getCurrentPath()->m_addremovecommand.push_back(neweffectobj);
    return true;
}


bool pl_foreign_pathBegin (FP_term a, FP_term b)
{
    FP_atom name;
    FP_get_atom(a, &name);
    const wchar_t* tc = FP_wstring_from_atom(name);
    uti_refptr<CRPDDLPairBlock> m_pairblock = new CRPDDLPairBlock();
    m_pairblock->m_condition = CUtiIDString(tc);
    m_pairblock->m_true = new CRPDDLExecutionBlock();
    m_pairblock->m_false = new CRPDDLExecutionBlock();
    CRPDDLExecutionBlock* block = m_compileaction->getCurrentPath();
    m_pairblock->m_true->m_parent = block;
    m_pairblock->m_false->m_parent = block;
    block->m_subblocks.push_back(m_pairblock.get());
    m_compileaction->m_stack.push(std::make_pair(m_pairblock.get(), true));
    return true;
}

bool pl_foreign_pathElse (FP_term a)
{
    auto tp = m_compileaction->m_stack.top();
    assert(tp.second == true);
    m_compileaction->m_stack.pop();
    tp.second = false;
    m_compileaction->m_stack.push(tp);
    return true;
}


bool pl_foreign_pathEnd (FP_term a)
{
   m_compileaction->m_stack.pop();
   return true;
}

bool pl_foreign_pathWait(FP_term a)
{
    FP_atom name;
    FP_get_atom(a, &name);
    const wchar_t* tc = FP_wstring_from_atom(name);
    m_compileaction->m_duration.push_back(CUtiIDString(tc));
    return true;
}

}
