/***************************************************************************
 *   Copyright (C) 2018 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "syntheticrule.h"
#include "utilisp.h"
#include "frogprolog.h"
#include "executecpp.h"
#include "pddllib.h"
#include "utisystem.h"
#include <iomanip>

using namespace PrologSystem;


Lisp::LObject CRBoolCell::toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj) {
    if (mode == FLATMODE_CREATETEMPVARS) {
        CRAssembler* ass = check_cast<CRAssembler*>(obj);
        CUtiIDString newvar =ass->getNewTempVariable(UTISTR(""),IDSTR("bool"),false);
        ass->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),newvar,Lisp::list(IDSTR("vbool"), data ? 1 : 0)));
        return Lisp::LObject(newvar);

    }
    if (mode ==FLATMODE_COLLECTYPES) {
        return IDSTR("bool");
    }
    return Lisp::LObject(data);
}

Lisp::LObject CRClassCell::toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj)
{    assert(m_objectname.empty() == false);
    if (mode == FLATMODE_CREATETEMPVARS) {
        CRAssembler* ass = check_cast<CRAssembler*>(obj);
        CUtiIDString newvar =ass->getNewTempVariable(UTISTR(""),m_type->m_name,false);
        ass->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),newvar,Lisp::list(IDSTR("vatom"), m_objectname)));
        return Lisp::LObject(newvar);

    }
    if (mode ==FLATMODE_COLLECTYPES) {
        return m_type->m_name;
    }
    return m_objectname;
}


Lisp::LObject CRFloatCell::toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj) {
    if (mode == FLATMODE_CREATETEMPVARS) {
        CRAssembler* ass = check_cast<CRAssembler*>(obj);
        CUtiIDString newvar =ass->getNewTempVariable(UTISTR(""),IDSTR("float"),false);
        ass->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),newvar,Lisp::list(IDSTR("vfloat"), data)));
        return Lisp::LObject(newvar);

    }
    if (mode ==FLATMODE_COLLECTYPES) {
        return IDSTR("float");
    }
    return Lisp::LObject(data);
}

Lisp::LObject CRIntCell::toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj) {
    if (mode == FLATMODE_CREATETEMPVARS) {
        CRAssembler* ass = check_cast<CRAssembler*>(obj);
        CUtiIDString newvar =ass->getNewTempVariable(UTISTR(""),IDSTR("int"),false);
        ass->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),newvar,Lisp::list(IDSTR("vint"), data)));
        return Lisp::LObject(newvar);

    }
    if (mode ==FLATMODE_COLLECTYPES) {
        return IDSTR("float");
    }
    return Lisp::LObject(data);
}
int CRPDDLExecutionBlock::calculateUsageCount()
{
    m_usagecount = 0;
    m_usagecount += m_addremovecommand.size();
    for (auto & p : m_subblocks) {
        m_usagecount += p.calculateUsageCount();
    }

    return m_usagecount;
}

bool CRPDDLExecutionBlock::tryToMerge()
{
    bool changed = false;
    for (auto &p : m_subblocks) {
        bool c = p.m_true->tryToMerge();
        changed = c || changed;
        c = p.m_false->tryToMerge();
        changed = c || changed;

        bool abort = false;
        int trueindex=0;
        while (trueindex < p.m_true->m_addremovecommand.size()) {
            int falseindex = 0;
            while (falseindex < p.m_false->m_addremovecommand.size()) {
                CRPDDLEffect* te = &p.m_true->m_addremovecommand[trueindex];
                CRPDDLEffect* fe = & p.m_false->m_addremovecommand[falseindex];
                if (te->isEqualToEffect(fe)) {
                    //We can merge it
                    changed = true;
                    m_addremovecommand.push_back(te);
                    p.m_true->m_addremovecommand.deleteItem(trueindex);
                    p.m_false->m_addremovecommand.deleteItem(falseindex);
                    //abort = true;
                    falseindex = 0;
                    if (trueindex >= p.m_true->m_addremovecommand.size()) {
                        break;
                    }
                } else {
                    falseindex++;
                }
            }
            trueindex++;

           // if (abort) {
           //     break;
            //}
        }
    }


    return changed;
}

CRPDDLPairBlock::CRPDDLPairBlock()
{
    m_usagecount = 0;
}
int CRPDDLPairBlock::calculateUsageCount()
{
    m_usagecount = 0;
    m_usagecount += m_true->calculateUsageCount();
    m_usagecount += m_false->calculateUsageCount();
    return m_usagecount;
}
CRAssembler::CRAssembler()
{
    m_labelindex = 1000;
m_tempvarindex = 0;
m_syntheticpddl = false;
    m_global = new CRStackElement();

}
CRAssembler::~CRAssembler()
{

}
int CRAssembler::createLabel()
{
   int res = m_labelindex;
   m_labelindex++;
   return res;
}
Lisp::LObject replaceNthArgument(Lisp::LObject command, int n, Lisp::LObject newobj)
{
int index = 0;

  DynArray1D<Lisp::LObject> obj;
  for (auto p: command) {
     if (index == n) {
     obj.push_back(newobj);
     } else {
     obj.push_back(p);
     }
     index++;
  }
  Lisp::LObject newcommand = Lisp::list(obj);
  return newcommand;
}

//Befehler wie jump erhalten als Ziel den Labelindex
void CRAssembler::add(const Lisp::LObject& command, int mylabel)
{

  Lisp::LObject newcommand = replaceNthArgument(command, 1,Lisp::LObject((int) m_commands.size()));
  std::wcout << newcommand.print() << std::endl;
  m_commands.push_back(std::make_pair(mylabel, newcommand));
}

void CRAssembler::linkLabels()
{
   std::map<int, int> labeltoindex;
   for (int i = 0; i < m_commands.size(); i++) {
       auto& p = m_commands[i];
       if (p.first != -1) {
          assert(labeltoindex.find(p.first) == labeltoindex.end());
          labeltoindex.insert(std::make_pair(p.first, i));
       }
   }
   for (int i = 0; i < m_commands.size(); i++) {
       auto& p = m_commands[i];
       CUtiIDString commandtype = Lisp::nth(p.second, 0).getAtom();
       if (commandtype == IDSTR("asm_jump")) {
          int labelid = Lisp::nth(p.second, 2).toInt();
          assert(labeltoindex.find(labelid) != labeltoindex.end());
          int codepos = labeltoindex[labelid];
          Lisp::LObject newcommand = replaceNthArgument(p.second, 2, codepos);
          p.second = newcommand;
       } else if (commandtype == IDSTR("asm_conditionaljump")) {
          int labelid = Lisp::nth(p.second, 3).toInt();
          assert(labeltoindex.find(labelid) != labeltoindex.end());
          int codepos = labeltoindex[labelid];
          Lisp::LObject newcommand = replaceNthArgument(p.second,3, codepos);
          p.second = newcommand;
       } else if (commandtype == IDSTR("asm_procedure_call")) {
           int labelid = Lisp::nth(p.second, 2).toInt();
           assert(labeltoindex.find(labelid) != labeltoindex.end());
           int codepos = labeltoindex[labelid];
           Lisp::LObject newcommand = replaceNthArgument(p.second,2, codepos);
           p.second = newcommand;
        }
   }


}

void CRAssembler::printCode()
{
    for (int i = 0; i< m_commands.size(); i++) {
        auto& el = m_commands[i];
        std::wcout <<std::setw(5) << std::setfill(L'0') <<i <<": "  << std::setw(5) << std::setfill(L'0') << el.first << ": " << el.second.print() << std::endl;
    }
}


Lisp::LObject CRAssembler::blockToVar(CRDataCell* cell, TRBasicDataType type)
{
    if (type == erBool) {
        if (dynamic_cast<CRPDDLVariable*>(cell) != NULL) {
            CRPDDLVariable* var = check_cast<CRPDDLVariable*>(cell);
            assert(dynamic_cast<CRBaseTypeDeclaration*>(var->m_type) != NULL);
            assert(dynamic_cast<CRBaseTypeDeclaration*>(var->m_type)->m_data == erBool);
            return var->m_name;

        } else {

            CUtiIDString tempvar = getNewTempVariable(UTISTR(""), IDSTR("bool"), false);
            add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),tempvar,Lisp::list(IDSTR("vbool"), check_cast<CRBoolCell*>(cell)->data ? 1 : 0)));


        }
    } else if (type == erFloat) {
            if (dynamic_cast<CRPDDLVariable*>(cell) != NULL) {
                CRPDDLVariable* var = check_cast<CRPDDLVariable*>(cell);
                assert(dynamic_cast<CRBaseTypeDeclaration*>(var->m_type) != NULL);
                assert(dynamic_cast<CRBaseTypeDeclaration*>(var->m_type)->m_data == erFloat);
                return var->m_name;

            } else {

                CUtiIDString tempvar = getNewTempVariable(UTISTR(""), IDSTR("float"), false);
                add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),tempvar,Lisp::list(IDSTR("vfloat"), check_cast<CRFloatCell*>(cell)->data)));


            }

    } else if (type == erInt) {
        if (dynamic_cast<CRPDDLVariable*>(cell) != NULL) {
            CRPDDLVariable* var = check_cast<CRPDDLVariable*>(cell);
            assert(dynamic_cast<CRBaseTypeDeclaration*>(var->m_type) != NULL);
            assert(dynamic_cast<CRBaseTypeDeclaration*>(var->m_type)->m_data == erFloat);
            return var->m_name;

        } else {

            CUtiIDString tempvar = getNewTempVariable(UTISTR(""), IDSTR("int"), false);
            add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),tempvar,Lisp::list(IDSTR("vint"), check_cast<CRIntCell*>(cell)->data)));


        }

} else {
        assert(false);
    }
    return Lisp::LObject();
}

bool CRPDDLExecutionBlock::equals(CUtiOB *other)
{
    CRPDDLExecutionBlock* otherblock = (CRPDDLExecutionBlock*)other;
    if (otherblock->m_addremovecommand.size() != m_addremovecommand.size())return false;
    //if (otherblock->m_condition.size() != m_condition.size())return false;
    if (otherblock->m_subblocks.size() != m_subblocks.size())return false;
    //if (otherblock->m_fullexecutioncondition.size() != m_fullexecutioncondition.size())return false;
    assert(false);

    /*for (int i = 0;i < m_condition.size();i++) {
        if (m_condition[i] != otherblock->m_condition[i]) return false;
    }
    for (int i = 0;i < m_fullexecutioncondition.size();i++) {
        if (m_fullexecutioncondition[i] != otherblock->m_fullexecutioncondition[i]) return false;
    }*/
    return true;
}

CRPDDLDecision::CRPDDLDecision()
{
m_createboolresult = false;
}
/*
CRPDDLDecision::CRPDDLDecision(CRPDDLWhen* when, CRPDDLDecision* parent)
{

m_codeparent = parent;
m_when = when;
}*/

/*)
CRPDDLVariableValue* CRPDDLVariable::getValueByDecision(CRPDDLDecision* ret)
{
   std::set<CRPDDLDecision*> decisions;
   CRPDDLDecision* r = ret;
   while (r) {
      decisions.insert(r);
      r = r->m_codeparent;
   }
   for (int i = 0; i < m_values.size(); i++) {
       CRPDDLVariableValue* v = m_values.at(i);


       std::set<CRPDDLDecision*> d2;
       std::copy(v->m_currentpath->m_decisions.begin_pointer(), v->m_currentpath->m_decisions.end_pointer(), std::inserter(d2, d2.begin()));
       if (d2 == decisions) {
       return v;
       }
   }
   return NULL;
}*/


/*CRPDDLVariableValue::CRPDDLVariableValue()
{

}

CRPDDLVariableValue* CRPDDLVariableValue::createVarCopy()
{
   CRPDDLVariableValue* v = new CRPDDLVariableValue();
   v->m_contenttype = m_contenttype;
   v->m_data = m_data;
   v->m_name = m_name;
   v->m_currentpath = new CRPDDLPath();
   std::copy(m_currentpath->m_decisions.begin(), m_currentpath->m_decisions.end(), std::back_inserter(v->m_currentpath->m_decisions));

   return v;
}
*/
/*void CRPDDLVariable::prepareNewValue(CRPDDLWhen* decision)
{

    CReferenceList<CRPDDLVariableValue> oldvalues;
    std::copy(m_values.begin(), m_values.end(), std::back_inserter(oldvalues));
    m_values.clear();
    std::vector<CRPDDLWhen*> whens;
    CRPDDLWhen* element = decision;
    while (element) {
      whens.push_back(element);
      element = element->m_parent;

    }
    for (auto&varvalue : oldvalues) {
       std::vector<CRPDDLWhen*> missingwhens;
       for (auto p : whens) {
           bool found = std::find_if(varvalue.m_currentpath->m_decisions.begin(), varvalue.m_currentpath->m_decisions.end(),
            [p] (const CRPDDLDecision& decision) -> bool
            {
               return decision.m_when == p;
            }) != varvalue.m_currentpath->m_decisions.end();


           if (!found) {
              missingwhens.push_back(p);
           }
       }
       if (missingwhens.size() == 0 ) {
          m_values.push_back(varvalue);
          continue;
       }
       int count = intpow(2, missingwhens.size());
       for (int i = 0; i < count; i++) {
           CRPDDLVariableValue* newvalue = varvalue.createVarCopy();
           assert(newvalue->m_currentpath->getReferenceCount() == 1);
           for (int j = 0; j < missingwhens.size(); j++) {
               newvalue->m_currentpath->m_decisions.push_back( ((i & (1 << j)) != 0) ? missingwhens[j]->m_truedecision.get() : missingwhens[j]->m_falsedecision.get() );
           }
          m_values.push_back(newvalue);
       }
    }

}

void CRPDDLVariable::setValue_cell(CRPDDLDecision* decision, CRDataCell *value)
{
CRPDDLVariableValue* v = getValueByDecision(decision);
   assert(v != NULL);
   v->m_data = value;
   v->m_contenttype = ctPDDLFixedCell;
}
void CRPDDLVariable::setValue_atom(CRPDDLDecision* decision, const CUtiIDString& atom)
{
CRPDDLVariableValue* v = getValueByDecision(decision);
   assert(v != NULL);
   assert(atom.isLispVariable() == false);
   v->m_name  = atom;
   v->m_contenttype = ctPDDLAtom;
}

void CRPDDLVariable::setValue_var(CRPDDLDecision* decision, const CUtiIDString& variable)
{
CRPDDLVariableValue* v = getValueByDecision(decision);
   assert(v != NULL);
   assert(variable.isLispVariable() == true);
   v->m_name  = variable;
   v->m_contenttype = ctPDDLVariable;
}*/


uti_refptr<CRDataCell> CRValue::compile(CRAssembler* assembler)
{
    uti_refptr<CRDataCell> result(this->m_typedata->createCell());
    if (m_type == erInt) {
        check_cast<CRIntCell*>(result.get())->data = round(m_value);
        return result;
      };
      if (m_type == erFloat) {
          check_cast<CRFloatCell*>(result.get())->data = m_value;
          return result;
        };
      if (m_type == erBool) {
          check_cast<CRBoolCell*>(result.get())->data = m_value > 0.5;
          return result;
        };
      if (m_type == erString) {
          check_cast<CRStringCell*>(result.get())->data =m_atom.str();
          return result;
        };
      if (m_type == erObject) {
          // check_cast<CRStringCell*>(result.get())->data =m_atom.str();
           assert(false);
         };
      if (m_type == erVariable) {
          CUtiIDString variablename;
          variablename = m_atom;
          result = assembler->getVariable(variablename);
          if (result.get() == NULL)
          {
              error(UTISTR("Variable not found: ")+variablename.str());
          }
          return result;
      }

      assert(false);
        return result;
}

uti_refptr<CRDataCell> CRValue::execute(CRNormalExecutionContext* econtext)
{
    uti_refptr<CRDataCell> result(this->m_typedata->createCell());
    if (m_type == erInt) {
        check_cast<CRIntCell*>(result.get())->data = round(m_value);
      };
      if (m_type == erFloat) {
          check_cast<CRFloatCell*>(result.get())->data = m_value;
        };
      if (m_type == erBool) {
          check_cast<CRBoolCell*>(result.get())->data = m_value > 0.5;
        };
      if (m_type == erString) {
          check_cast<CRStringCell*>(result.get())->data =m_atom.str();
        };
      if (m_type == erObject) {
         // check_cast<CRStringCell*>(result.get())->data =m_atom.str();
          assert(false);
        };
      if (m_type == erVariable) {
          CUtiIDString variablename;
          variablename = m_atom;
          result = econtext->getVariable(variablename);

        };


        return result;
}

uti_refptr<CRDataCell> CRFunctionDeclaration::execute(CRNormalExecutionContext* econtext)
{
    econtext->m_stack[econtext->m_stack.size()-1].m_declaration = this;
     m_block->execute(econtext);
     econtext->m_doreturn = false;
     return uti_refptr<CRDataCell>(econtext->m_returnvalue.get());
}

uti_refptr<CRDataCell> CRFunctionDeclaration::compileNonMainFunction(CRAssembler* assembler)
{
    if (!m_block.get()) {
        return uti_refptr<CRDataCell>();
    }
    //Create function block
    //assert(false);
    //:- dynamic(/('asm_procedure_start', 2)).
    //:- dynamic(/('asm_return', 2)).
    // executecode(PC, Env) :- asm_procedure_start(PC, ParamVariables),
    // asm_return(PC, ListReturnValues)
    assert(assembler->m_norootmode == false);
    assembler->m_norootmode = true;
    DynArray1D<Lisp::LObject> l;
    l.push_back(IDSTR("!!!list"));
   // if (result.get() == NULL) {sdaf
    for (int i = 0; i < m_param_type_and_name.size(); i++) {
        uti_refptr<CRDataCell> r = std::get<0>(m_param_type_and_name[i])->createCell();

        r = r->replaceVarWithPDDL(assembler, NULL, std::get<1>(m_param_type_and_name[i]).str(), true);
        r->createFlatPDDLList(l,0, NULL);
        assembler->reserveVariable(std::get<1>(m_param_type_and_name[i]),r.get());

    }

       assembler->add(Lisp::list(IDSTR("asm_procedure_start"), IDSTR("PC"), Lisp::list(l)), getLabelOrCreate(assembler));

   // }

       uti_refptr<CRDataCell> result = m_block->compile(assembler);
    assembler->m_norootmode = false;
 //   if (result.get() == NULL) {
       assembler->add(Lisp::list(IDSTR("asm_return"), IDSTR("PC"), Lisp::list(IDSTR("!!!list"))));
   /* } else {
         DynArray1D<Lisp::LObject> l;
         l.push_back(IDSTR("!!!list"));

         result->createFlatPDDLList(l);
         assembler->add(Lisp::list(IDSTR("asm_return"), IDSTR("PC"), Lisp::list(l)));
    }*/
    return uti_refptr<CRDataCell>();

}

uti_refptr<CRDataCell> CRFunctionDeclaration::compile(CRAssembler* assembler)
{
   // asm_procedure_start(PC, ListParamVariables)
// DynArray1D< std::tuple<uti_refptr<CRBaseDeclaration>, CUtiIDString> > m_param_type_and_name;
    /*for (auto & el : m_param_type_and_name) {
        CRBaseDeclaration* decl = std::get<0>(el).get();
        CUtiIDString name = std::get<1>(el);
//todo rekursiv absteigen und declarieren
        assembler->add(Lisp::list(IDSTR("asm_declparam"), IDSTR("PC"),name));

    }*/
    m_block->compile(assembler);
    return uti_refptr<CRDataCell>();
}

CRDataCell* CRStructDeclaration::createCell() {
    CRRecordCell* cell = new CRRecordCell();
    cell->m_type = this;
    for (int i = 0;  i < m_children_tan.size(); i++) {
        cell->m_subcells.push_back(m_children_tan.at(i)->m_typedata->createCell());
    }
    return cell;
}


CRRecordCell::~CRRecordCell()
{
#ifndef NDEBUG
    std::wcout << "Delete Record" << std::endl;
#endif
}

CRDataCell* CRRecordCell::createRCopy()
{
    CRRecordCell* c  = new CRRecordCell();
    c->m_type = m_type;

    for (int i = 0; i < m_subcells.size(); i++) {
        auto&p  = m_subcells[i];
         c->m_subcells.push_back( p.createRCopy());
    }

    return c;
}


CRDataCell * CRStringCell::replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param)
{return this;};

CRDataCell * CRIntCell::replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param)
{
    CRPDDLVariable* newcell = new CRPDDLVariable();
    newcell->m_name = context->getNewTempVariable(basename, IDSTR("int"), param);
    newcell->m_type = m_type;

    return newcell;
}


CRDataCell * CRFloatCell::replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param)
{
    CRPDDLVariable* newcell = new CRPDDLVariable();
    newcell->m_name = context->getNewTempVariable(basename, IDSTR("float"), param);
    newcell->m_type = m_type;

    return newcell;
}


CRDataCell * CRArrayCell::replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param)
{
    for (int i = 0; i < m_cells.size(); i++) {
       m_cells[i]= m_cells[i]->replaceVarWithPDDL(context, data, basename+UTISTR("_")+i, param);
    }
    return this;
}

void CRArrayCell::replaceVarWithAtom(CRNormalExecutionContext* context, ReferenceUtiOB* data, const CUtiString& basename)
{
    for (int i = 0; i < m_cells.size(); i++) {
       m_cells[i]->replaceVarWithAtom(context, data, basename+UTISTR("_")+i);
    }
}

void CRArrayCell::replaceVarWithAtomAss(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename)
{
    for (int i = 0; i < m_cells.size(); i++) {
       m_cells[i]->replaceVarWithAtomAss(context, data, basename+UTISTR("_")+i);
    }
}

CRDataCell* CRRecordCell::replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param)
{

    CRStructDeclaration* type = check_cast<CRStructDeclaration*>(m_type);
    for (int i = 0; i < m_subcells.size(); i++) {
       m_subcells.exchange(i,  m_subcells[i].replaceVarWithPDDL(context, data, basename+UTISTR(".")+type->m_children_tan[i].m_name.str(), param));
    }
    return this;
}

void CRRecordCell::replaceVarWithAtom(CRNormalExecutionContext* context, ReferenceUtiOB* data, const CUtiString& basename)
{
    CRStructDeclaration* type = check_cast<CRStructDeclaration*>(m_type);
    for (int i = 0; i < m_subcells.size(); i++) {
      m_subcells[i].replaceVarWithAtom(context, data, basename+UTISTR(".")+type->m_children_tan[i].m_name.str());
    }
}

void CRRecordCell::replaceVarWithAtomAss(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename)
{
    CRStructDeclaration* type = check_cast<CRStructDeclaration*>(m_type);
    for (int i = 0; i < m_subcells.size(); i++) {
      m_subcells[i].replaceVarWithAtomAss(context, data, basename+UTISTR(".")+type->m_children_tan[i].m_name.str());
    }
}



CRDataCell *CRClassCell::replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param)
{
    CRPDDLVariable* newcell = new CRPDDLVariable();
    newcell->m_name = context->getNewTempVariable(basename, m_type->m_name, param);

    newcell->m_type = m_type;
    return newcell;
}

CUtiString CRRecordCell::printValue() {
    CUtiString str;
    str += UTISTR("(");
    for (int i = 0; i < m_subcells.size(); i++) {
        str += m_subcells[i].m_type->m_name.str();
        str += UTISTR(": ");
       str += m_subcells[i].printValue();
       if (i != m_subcells.size()-1) {
           str += UTISTR(", ");
       }
    }
    str += UTISTR(")");
    return str;

}

CRDataCell* CRClassCell::createRCopy()
{
    CRClassCell* c  = new CRClassCell();
    c->m_type = m_type;
    c->m_objectname = m_objectname;


    return c;
}
CUtiString CRClassCell::printValue() {
    CUtiString str;
    str += m_type->m_name.str();
    str += UTISTR("(");
    str += (int)((int64_t)this);
    str += UTISTR(")");
    return str;
}

CRStackElement::CRStackElement()
{
 m_declaration = NULL;
}


CRDataCell* CRAssembler::replaceVarWithPDDL(CRDataCell* dcell, const CUtiString& basename, bool param)
{
   return dcell->replaceVarWithPDDL(this, NULL, basename, param);
}

void CRNormalExecutionContext::reserveVariable(const CUtiIDString& str, CRDataCell* cell)
{
   CRStackElement* s = m_global.get();
   if (m_stack.size() != 0) {
       s = &m_stack[m_stack.size()-1];
   }
   //assert(s->m_variables.find(str) == s->m_variables.end());

  // s->m_variables.insert(std::make_pair(str, uti_refptr<CRDataCell>(cell)));
  s->m_variables.push_back(new CRVariableEntry(str,cell));
}


void CRNormalExecutionContext::setVariable(const CUtiIDString& str, CRDataCell* cell)
{
    for (int i = m_stack.size()-1; i >= 0; i--) {
        CRStackElement* st = &m_stack[i];
        auto it = st->findVariable(str);
        if (it!= NULL) {
            it->m_data = cell;
        }
        if (st->m_declaration == NULL) {
            break;
        }
    }
    auto it = m_global->findVariable(str);
    if (it!= NULL) {
        it->m_data = cell;
    }
    assert(false);
}

CRDataCell* CRNormalExecutionContext::getVariable(const CUtiIDString& str)
{
    for (int i = m_stack.size()-1; i >= 0; i--) {
        CRStackElement* st =& m_stack[i];
        auto it = st->findVariable(str);
        if (it!= NULL) {
            return it->m_data.get();
        }
        if (st->m_declaration != NULL) {
            break;
        }
    }
    auto it = m_global->findVariable(str);
    if (it!=  NULL) {
        return it->m_data.get();
    }
    return NULL;
}



void CRAssembler::reserveVariable(const CUtiIDString& str, CRDataCell* cell)
{
   CRStackElement* s = m_global.get();
   if (m_stack.size() != 0) {
       s = &m_stack[m_stack.size()-1];
   }
   //assert(s->m_variables.find(str) == s->m_variables.end());

  // s->m_variables.insert(std::make_pair(str, uti_refptr<CRDataCell>(cell)));
  s->m_variables.push_back(new CRVariableEntry(str,cell));
}


void CRAssembler::setVariable(const CUtiIDString& str, CRDataCell* cell)
{
    for (int i = m_stack.size()-1; i >= 0; i--) {
        CRStackElement* st = &m_stack[i];
        auto it = st->findVariable(str);
        if (it!= NULL) {
            it->m_data = cell;
        }
        if (st->m_declaration == NULL) {
            break;
        }
    }
    auto it = m_global->findVariable(str);
    if (it!= NULL) {
        it->m_data = cell;
    }
    assert(false);
}

CRDataCell* CRAssembler::getVariable(const CUtiIDString& str)
{
    for (int i = m_stack.size()-1; i >= 0; i--) {
        CRStackElement* st =& m_stack[i];
        auto it = st->findVariable(str);
        if (it!= NULL) {
            return it->m_data.get();
        }
        if (st->m_declaration != NULL) {
            break;
        }
    }
    auto it = m_global->findVariable(str);
    if (it!=  NULL) {
        return it->m_data.get();
    }
    return NULL;
}

bool CRTypeDefDeclaration::canConvertTo(CRBaseDeclaration* decl)
{
    CRTypeDefDeclaration* cd = this;

    while (cd != NULL) {
       if (cd == this) {
           return true;
       }
       cd = dynamic_cast<CRTypeDefDeclaration*>(cd->m_parenttype.get());
    }

    return false;
}
CRBaseDeclaration* resolveTypeDef(CRBaseDeclaration* d) {
assert(d != NULL);
    if (dynamic_cast<CRTypeDefDeclaration*>(d) != NULL) {
        return resolveTypeDef(check_cast<CRTypeDefDeclaration*>(d)->m_parenttype.get());
    }
    return d;
}

CRBaseDeclaration* CRTypeDefDeclaration::executeTypeOperation(CRBaseDeclaration* decl, int operatortype)
{
    CRBaseDeclaration* d = resolveTypeDef(decl);
    //Vielleicht sollten wir ueberall typedef auflosen
   return resolveTypeDef(this)->executeTypeOperation(d, operatortype);
}

uti_refptr<CRDataCell> CRTypeDefDeclaration::execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, int operatortype,  CRDataCell* othercell)
{
return resolveTypeDef(this)->execute(econtext, mycell, operatortype, othercell);
}

bool CRBaseTypeDeclaration::canConvertTo(CRBaseDeclaration* decl)
{
   switch (m_data) {
      case erInt: {

      return (dynamic_cast<CRBaseTypeDeclaration*>(decl) != NULL) &&  (erInt == check_cast<CRBaseTypeDeclaration*>(decl)->m_data ||
      erFloat == check_cast<CRBaseTypeDeclaration*>(decl)->m_data);
      }; break;
      case  erFloat: {

      return (dynamic_cast<CRBaseTypeDeclaration*>(decl) != NULL) &&  m_data == check_cast<CRBaseTypeDeclaration*>(decl)->m_data;
      }; break;
      case  erObject: {

assert(false);
      }; break;
      case  erBool: {


      return (dynamic_cast<CRBaseTypeDeclaration*>(decl) != NULL) &&  m_data == check_cast<CRBaseTypeDeclaration*>(decl)->m_data;
      }; break;
      case  erComplex: {

      return (dynamic_cast<CRBaseTypeDeclaration*>(decl) != NULL) &&  m_data == check_cast<CRBaseTypeDeclaration*>(decl)->m_data;
      }; break;
      case  erString: {
      return (dynamic_cast<CRBaseTypeDeclaration*>(decl) != NULL) &&  m_data == check_cast<CRBaseTypeDeclaration*>(decl)->m_data;
      }; break;

   }
   return false;
}


CRBaseTypeDeclaration::CRBaseTypeDeclaration(CRBaseRule* p) : CRBaseDeclaration(p)
{

}


uti_refptr<CRDataCell> CRBaseTypeDeclaration::execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, int operatortype,  CRDataCell* othercell)
{
CRBaseTypeDeclaration* basetype = dynamic_cast<CRBaseTypeDeclaration*>(resolveTypeDef( othercell->m_type));
   if (basetype == NULL) return uti_refptr<CRDataCell>();
   CRProgram* prog = getParentProgram();

   if (prog->m_operator_to_opdata.find((TRExpressionType)operatortype) == prog->m_operator_to_opdata.end()) {
      return uti_refptr<CRDataCell>();
   }
   CRExpressionOperator* expoperator = prog->m_operator_to_opdata[(TRExpressionType)operatortype];
   if (!expoperator->m_binary) return NULL;
   for (int i = 0; i < expoperator->m_type.size(); i++) {
      CRTypedOperator* to = expoperator->m_type.get(i);
      if (to->m_type1 == m_data &&  to->m_type2 ==  basetype->m_data) {
         return to->execute(econtext, mycell, othercell);
      }
   }

   return uti_refptr<CRDataCell>();
}

CRBaseDeclaration* CRBaseTypeDeclaration::executeTypeOperation(CRBaseDeclaration* decl, int operatortype)
{
   CRBaseTypeDeclaration* basetype = dynamic_cast<CRBaseTypeDeclaration*>(resolveTypeDef( decl));
   if (basetype == NULL) return NULL;
   CRProgram* prog = getParentProgram();

   if (prog->m_operator_to_opdata.find((TRExpressionType)operatortype) == prog->m_operator_to_opdata.end()) {
      return NULL;
   }
   CRExpressionOperator* expoperator = prog->m_operator_to_opdata[(TRExpressionType)operatortype];
   if (!expoperator->m_binary) return NULL;
   for (int i = 0; i < expoperator->m_type.size(); i++) {
      CRTypedOperator* to = expoperator->m_type.get(i);
      if (to->m_type1 == m_data &&  to->m_type2 ==  basetype->m_data) {
         return prog->getBaseType(to->m_res);
      }
   }

   return NULL;
}

bool CRArrayDeclaration::canConvertTo(CRBaseDeclaration* decl)
{
    CRArrayDeclaration* a = dynamic_cast<CRArrayDeclaration*>(decl);
    if (a && a->m_basetype.get() == m_basetype.get() && a->m_size == m_size) {
        return true;
    }
    return false;
}

void CRArrayDeclaration::createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB *obj)
{
    int s = 1;
    for (int i = 0; i < m_size.size(); i++) {
        s *= m_size[i];
    }
    for (int i = 0; i < s; i++) {
        m_basetype->createFlatPDDLList(list, index,mode, obj);
    }
}

void CRArrayDeclaration::createFlatVariableList(DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> >& list, const CUtiString& basename)
{
    int s = 1;
    for (int i = 0; i < m_size.size(); i++) {
        s *= m_size[i];
    }
    for (int i = 0; i < s; i++) {
    CUtiString newname = basename + UTISTR("_")+i+UTISTR("_");
    m_basetype->createFlatVariableList(list, newname);
  }
}


CRBaseDeclaration* CRArrayDeclaration::executeTypeOperation(CRBaseDeclaration* decl, int operatortype)
{
  return NULL;
}

bool CRClassDeclaration::canConvertTo(CRBaseDeclaration* decl)
{
    CRClassDeclaration* cd = this;

    while (cd != NULL) {
       if (cd == this) {
           return true;
       }
       cd = cd->m_parenttype.get();
    }

    return false;
}
CRBaseDeclaration* CRClassDeclaration::executeTypeOperation(CRBaseDeclaration* decl, int operatortype)
{
    return NULL;

}

void CRClassDeclaration::createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB *obj)
{
   int paramindex = index;
   index++;
   CUtiIDString str(UTISTR("p")+index);
   list.push_back(Lisp::list(IDSTR("-"), m_name, str));
}


void CRBaseTypeDeclaration::createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB *obj)
{
   int paramindex = index;
   index++;
   CUtiIDString str(UTISTR("f")+index);
   list.push_back(Lisp::list(IDSTR("-"), IDSTR("float"), str));

}

void CRDeclaration::createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB *obj)
{
  m_typedata->createFlatPDDLList(list, index, mode, obj);

}

CRBaseRule::CRBaseRule(CRBaseRule* p)
{
   m_parent = p;
   m_program = getParentProgram();

}

CRBlock* CRBaseRule::getParentBlock()
{

   if (m_parent == NULL) return NULL;
   if (dynamic_cast<CRBlock*>(m_parent)) {
      return check_cast<CRBlock*>(m_parent);
   }
   return m_parent->getParentBlock();
}

CRTypedOperator::CRTypedOperator(TRBasicDataType t1, TRBasicDataType t2, TRBasicDataType r)
{
m_type1 = t1;
m_type2 = t2;
m_res = r;
}


CRTypedOperator* CRExpressionOperator::getRType(CRBaseDeclaration* type1, CRBaseDeclaration* type2)
{
   CRBaseTypeDeclaration* t1 = check_cast<CRBaseTypeDeclaration*>(type1);
   CRBaseTypeDeclaration* t2 = check_cast<CRBaseTypeDeclaration*>(type2);

   for (auto& t : m_type) {
       if (t.m_type1 == t1->m_data && t.m_type2 == t2->m_data) {
           return &t;
       }
   }
   assert(false);
   return NULL;
}

CRExpressionOperator* CRExpressionOperator::addType(CRTypedOperator* t)
{
   m_type.push_back(t);
   return this;
}


bool CRExpressionOperator::isNextThenEat(UtiLexer& parse)
{
   if (m_isIdentitifer) {
       if (parse.IsThisNextBezeichner(m_name)) {
           CUtiString temp;
           parse.GetNextBezeichner(temp);
           return true;
       }
   } else {
       if (m_name.size() == 1) {
          if (parse.IsThisNextCharacter(m_name[0])) {
          parse.GetNextChar();
          return true;
          }
       } else {
          if (parse.IsThisNextCharacterSequence(m_name)) {
          parse.GetNextCharSequence(m_name);
          return true;
          }

       }
   }

   return false;
}

CRExpressionOperator::CRExpressionOperator(const CUtiString &name, TRExpressionType t, bool binary)
{
   m_name = name;
   m_expressiontype = t;
   m_binary = binary;
   m_isIdentitifer = true;
   for (int i = 0; i < m_name.size(); i++) {
       if (!UtiLexer::IsBezeichner(m_name[i])) {
          m_isIdentitifer = false;
       }
   }
}

CRProgram* CRBaseRule::getParentProgram()
{
if (dynamic_cast<CRProgram*>(this)) {
      return check_cast<CRProgram*>(this);
   }
if (m_parent == NULL){
 //assert(false);
 return NULL;
 }
   if (dynamic_cast<CRProgram*>(m_parent)) {
      return check_cast<CRProgram*>(m_parent);
   }
   assert(m_parent != this);
   return m_parent->getParentProgram();
}

void CRBaseRule::changeParent(CRBaseRule* p)
{
   m_parent = p;
   m_program = getParentProgram();
}

void CRBaseRule::error(UtiLexer& parse, const CUtiString& data)
{
   parse.printError(data);
   throw new RuleException(parse.getLexerPosition(), data);

}


void CRBaseRule::error(const CUtiString& data)
{
   //parse.printError(data);
    std::wcout << "Error: " << data << std::endl;
   throw new RuleException(CIVector2(0,0), data);

}


void CRBaseRule::MatchCharacter(UtiLexer& parse, U_CHAR c)
{
   if (parse.IsThisNextCharacter(c)) {
      parse.GetNextChar();
   } else {
      error(parse, UTISTR("Character expected: ")+c);

   }
}

void CRBaseRule::MatchIdentifier(UtiLexer& parse, const CUtiString& str)
{
   if (parse.IsThisNextBezeichner(str)) {
   CUtiString t;
      parse.GetNextBezeichner(t);
   } else {
      error(parse, UTISTR("Identifier expected: ")+str);

   }
}

CRCommand::CRCommand(CRBaseRule* p) :CRBaseRule(p)
{

};



CREval::CREval(CRBaseRule* p) :CRBaseRule(p)
{

}


CRExpression::CRExpression(CRBaseRule* p) :CREval(p)
{

    m_datatype = NULL;
}

uti_refptr<CRDataCell> CRExpression::execute(CRNormalExecutionContext* econtext)
{
   // assert(false);
    // virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell)=0;
    return m_datatype->execute(econtext, m_a->execute(econtext).get(), m_b->execute(econtext).get());
}



uti_refptr<CRDataCell> CRExpression::compile(CRAssembler* assembler)
{

    return m_datatype->compile(assembler, m_a->compile(assembler).get(), m_b->compile(assembler).get());
}

CRValue::CRValue(CRBaseRule* p) : CREval(p)
{
    m_value = 0;
    m_type = erInt;

}



CRVariable::CRVariable(CRBaseRule* p) : CRBaseDeclaration(p)
{
   m_isconstant= false;
   m_isparameter =false;
};

CREvalFunction::CREvalFunction(CRBaseRule* p) : CREval(p)
{

}


uti_refptr<CRDataCell> CREvalFunction::compile(CRAssembler* assembler)
{

    return m_call->compile(assembler);
}

uti_refptr<CRDataCell> CREvalFunction::execute(CRNormalExecutionContext* econtext)
{
   return m_call->execute(econtext);
}

CRExists::CRExists(CRBaseRule* p) : CREval(p)
{
m_typedata = p->getParentProgram()->getBaseType(erBool);
}

void CRExists::doParse(UtiLexer& parse)
{
    MatchIdentifier(parse, UTISTR("exists"));
    MatchCharacter(parse, L'(');
m_value = getParentBlock()->parseExpression(parse, this);
MatchCharacter(parse, L')');
}


/*Lisp::LObject CRCheckExistanceCell::toLisp(CRPDDLDecision* decision)
{
    DynArray1D<Lisp::LObject> params;
    params.push_back(m_type->m_name);
    for (int i = 0; i < m_cells.size(); i++) {
       m_cells[i]->createFlatPDDLList(params);

    }
    return Lisp::list(params);
}*/

uti_refptr<CRDataCell> CRExists::compile(CRAssembler* assembler)
{
   CUtiIDString tempvar = assembler->getNewTempVariable(UTISTR(""), IDSTR("bool"), false);
   assembler->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),tempvar,Lisp::list(IDSTR("vbool"), 0)));
   Lisp::LObject atom;
   uti_refptr<CRDataCell> datacell =m_value->compile(assembler);
   CRRecordCell* rc = check_cast<CRRecordCell*>(datacell.get());
   DynArray1D<Lisp::LObject> params;
   params.push_back(IDSTR("!!!list"));
   params.push_back(rc->m_type->m_name);
   for (int i = 0; i < rc->m_subcells.size(); i++) {
      rc->m_subcells[i].createFlatPDDLList(params, 0, NULL);
      // nec->m_cells[i] = &rc->m_subcells[i];
   }
   atom = Lisp::list(params);


   CRStructDeclaration* s = check_cast<CRStructDeclaration*>(rc->m_type);
   assert(s != NULL);
    DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> > structparams;
        s->createFlatVariableList(structparams, EMPTYSTRING);
        assert(structparams.size() == params.size()-2);


    for (int fluentindex = 0; fluentindex < structparams.size(); fluentindex++) {

                if (dynamic_cast<CRClassDeclaration*>(structparams[fluentindex].second) == NULL) {

                      assembler->add(Lisp::list(IDSTR("asm_emptyfluent"), IDSTR("PC"),params[fluentindex+2], IDSTR("float")));
                }
    }

   assembler->add(Lisp::list(IDSTR("asm_exists"), IDSTR("PC"),tempvar,atom));
   return uti_refptr<CRDataCell>(new CRPDDLVariable(tempvar, erBool));
}

uti_refptr<CRDataCell> CRExists::execute(CRNormalExecutionContext* econtext)
{
    /*if (econtext->m_pddlmode) {
        //Was wenn wir tatsachliche eine Boolsche Variable zu einem string transformieren wollen?
        uti_refptr<CRDataCell> datacell =m_value->execute(econtext);
        CRRecordCell* rc = check_cast<CRRecordCell*>(datacell.get());
        DynArray1D<Lisp::LObject> params;
        params.push_back(rc->m_type->m_name);

        CRCheckExistanceCell* nec = new CRCheckExistanceCell();
        nec->m_type = getParentProgram()->getBaseType(erBool);
        nec->m_cells.setSize(rc->m_subcells.size());
        nec->m_not = false;
        nec->m_type = rc->m_type;
        nec->m_originalcell = rc;
        for (int i = 0; i < rc->m_subcells.size(); i++) {
           //rc->m_subcells[i].createFlatPDDLList(params);
            nec->m_cells[i] = &rc->m_subcells[i];
        }

        return uti_refptr<CRDataCell>(nec);

    } else {*/
    return uti_refptr<CRDataCell>(new CRBoolCell(false));
    //}
}

CRBaseDeclaration::CRBaseDeclaration(CRBaseRule* p) : CRCommand(p)
{

};


void CRStructDeclaration::createPDDLDeclaration(CStringList* list, bool createpredicates)
{
    DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> > res;
    createFlatVariableList(res, UTISTR(""));

    if (createpredicates) {
        CUtiString data;
        data += UTISTR("   (")+m_name.str();

        int atomcount = 0;
        for (int i = 0; i < res.size(); i++) {
            if (dynamic_cast<CRClassDeclaration*>(res[i].second)) {
                atomcount++;
                data += UTISTR(" ?")+res[i].first + UTISTR(" - ")+res[i].second->m_name.str();
            }

        }
        data += UTISTR(")");
        if (atomcount != 0) {
           list->add(data);
        }
    } else {
        for (int j = 0; j < res.size(); j++) {
            if (!dynamic_cast<CRClassDeclaration*>(res[j].second)) {
                CUtiString data;
                data += UTISTR("(")+m_name.str()+UTISTR("_")+res[j].first;

                for (int i = 0; i < res.size(); i++) {
                    if (dynamic_cast<CRClassDeclaration*>(res[i].second)) {
                        data += UTISTR(" ?")+res[i].first + UTISTR(" - ")+res[i].second->m_name.str();
                    }

                }
                data += UTISTR(")");
                list->add(data);
            }
        }
    }
}

CRDeclaration::CRDeclaration(CRBaseRule* p, CRBaseDeclaration* t): CRBaseDeclaration(p)
{
  m_typedata = t;
  m_type = t->m_name;
}

CRDeclaration::CRDeclaration(CRBaseRule* p) : CRBaseDeclaration(p)
{
  m_typedata = NULL;
};


CRArrayDeclaration::CRArrayDeclaration(CRBaseRule* p) : CRBaseDeclaration(p)
{

}

CRStructDeclaration::CRStructDeclaration(CRBaseRule* p) : CRBaseDeclaration(p)
{
};

void CRStructDeclaration::createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB *obj)
{
  for (auto& p: m_children_tan) {
  p.createFlatPDDLList(list, index, mode, obj);
  }
}

void CRStructDeclaration::createFlatVariableList(DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> >& list, const CUtiString& basename)
{
   for (auto& p: m_children_tan) {
       CUtiString newname;
       if (basename.size() == 0) {
            newname = p.m_name.str();
       } else {
           newname = basename + UTISTR(".")+p.m_name.str();
       }

     p.m_typedata->createFlatVariableList(list, newname);
  }
}




Lisp::LObject CRStructDeclaration::toLispDeclaration()
{
   DynArray1D<Lisp::LObject> dect;

   dect.push_back(m_name);
   int index=0;
   createFlatPDDLList(dect, index, 0, NULL);

   return Lisp::list(dect);
}

void CRStructDeclaration::doParse(UtiLexer& parse)
{
    MatchIdentifier(parse, UTISTR("struct"));
    CUtiString name;
    if (!parse.GetNextBezeichner(name)) {
       error(parse, UTISTR("struct name expected"));
    }
    m_name = CUtiIDString(name);

    MatchCharacter(parse, L'{');

   // void parseDeclaration(UtiLexer& parse, CUtiIDString& typestr, uti_refptr<CRBaseDeclaration> & typedata, CUtiIDString& name);
    while (!parse.IsThisNextCharacter(L'}')) {
        CRDeclaration* decl = new CRDeclaration(this);
        decl->doParse(parse);
        m_children_tan.push_back(decl);
        MatchCharacter(parse, L';');
    }


    MatchCharacter(parse, L'}');
}

void CRClassDeclaration::doParse(UtiLexer& parse)
{
    MatchIdentifier(parse, UTISTR("class"));
    CUtiString name;
    if (!parse.GetNextBezeichner(name)) {
       error(parse, UTISTR("struct name expected"));
    }
    m_name = CUtiIDString(name);

    if (parse.IsThisNextCharacter(L':')) {
        MatchCharacter(parse, L':');
        MatchIdentifier(parse, UTISTR("public"));
        CUtiString baseclass;
        if (!parse.GetNextBezeichner(baseclass)) {
            error(parse, UTISTR("Identifier expected!"));
        }
        CRClassDeclaration* cl = dynamic_cast<CRClassDeclaration*>(getParentBlock()->find(CUtiIDString(baseclass)));
        if (cl == NULL) {

            error(parse, UTISTR("Class Type expected!"));
        }
        m_parenttype = cl;
}

    MatchCharacter(parse, L'{');

  /*  while (!parse.IsThisNextCharacter(L'}')) {
        CRDeclaration* decl = new CRDeclaration(this);
        decl->doParse(parse);
        m_children_tan.push_back(decl);
        MatchCharacter(parse, L';');
    }
*/

    MatchCharacter(parse, L'}');
}

int CRStructDeclaration::findIndex(const CUtiIDString& name)
{
   for (int i = 0; i < m_children_tan.size(); i++) {
      CRDeclaration* decl = m_children_tan.get(i);
      if (decl->m_name == name) return i;
   }
    return -1;
}

CRReference::CRReference(CRBaseRule* p) : CREval(p)
{

}



CRVarReference::CRVarReference(CRBaseRule* p) : CRReference(p)
{

}


CRDatabaseCommand::CRDatabaseCommand(CRBaseRule* p, TDCommandType ct) : CRCommand(p)
{
    m_ct = ct;
}

CRAssert::CRAssert(CRBaseRule* p): CRCommand(p)
{
m_myline = -1;
}


CRPDDLVariable::CRPDDLVariable(const CUtiIDString& n, TRBasicDataType type)
{
    m_name = n;
    m_type = new CRBaseTypeDeclaration(type);
    m_owntype = m_type;
}


void CRPDDLVariable::copyFromCompile(CRAssembler* assembler, CRDataCell* cell)
{
    CRPDDLVariable* t = check_cast<CRPDDLVariable*>(cell);
    assert(t->m_type == m_type ||  check_cast<CRBaseTypeDeclaration*>(t->m_type)->m_data == check_cast<CRBaseTypeDeclaration*>(m_type)->m_data);
    if (dynamic_cast<CRClassDeclaration*>(m_type)) {

        assembler->add(Lisp::list(IDSTR("asm_copyobject"), IDSTR("PC"),m_name,t->m_name));
        return;
    }
    switch (check_cast<CRBaseTypeDeclaration*>(m_type)->m_data) {
      case erInt: {
        assembler->add(Lisp::list(IDSTR("asm_copyobject"), IDSTR("PC"),m_name,t->m_name));

    }; break;
       case erFloat: {
        assembler->add(Lisp::list(IDSTR("asm_copyobject"), IDSTR("PC"),m_name,t->m_name));

    }; break;
       case erObject: {
        assembler->add(Lisp::list(IDSTR("asm_copyobject"), IDSTR("PC"),m_name,t->m_name));

    }; break;
       case erBool: {
        assembler->add(Lisp::list(IDSTR("asm_copyobject"), IDSTR("PC"),m_name,t->m_name));

    }; break;
       case erString: {
        assembler->add(Lisp::list(IDSTR("asm_copyobject"), IDSTR("PC"),m_name,t->m_name));

    }; break;
    }
}
void CRPDDLVariable::declPDDLVariable(CRNormalExecutionContext* context)
{
 //   context->m_usedpddlvariables.insert(m_name);
}


CRPrint::CRPrint(CRBaseRule* p): CRCommand(p)
{
 m_newline = false;
}


CRReturn::CRReturn(CRBaseRule* p): CRCommand(p)
{

}


CRIf::CRIf(CRBaseRule* p): CRCommand(p)
{

}


CRFor::CRFor(CRBaseRule* p): CRCommand(p)
{

};


CRReferenceArrayAccess::CRReferenceArrayAccess(CRBaseRule* p) : CRReference(p)
{

}


CRReferenceStruct::CRReferenceStruct(CRBaseRule* p) : CRReference(p)
{
   m_subelementindex = -1;
}



CRSetVariable::CRSetVariable(CRBaseRule* p) : CRCommand(p)
{

}

CRFunctionCall::CRFunctionCall(CRBaseRule* p) : CRCommand(p)
{

   m_declaration = NULL;
   m_structdeclaration = NULL;
}


CRFunctionDeclaration::CRFunctionDeclaration(CRBaseRule* p) : CRBaseDeclaration(p)
{
    m_type =rNormal;
    m_actiontype = ratAction;
}

CRBlock::CRBlock(CRBaseRule* p) : CRCommand(p)
{
    m_blocklevel = 0;
    m_needbraces = true;
    m_singleline = false;


};

void CRBlock::addChild(UtiLexer& parse, CRCommand* rule)
{
   m_command.push_back(rule);
   CRBaseDeclaration* decl = dynamic_cast<CRBaseDeclaration*>(rule);
   if (decl) {
      if (m_block.find(decl->m_name) != m_block.end()) {
          error(parse, UTISTR("Double Identifier found: ")+decl->m_name);
      }
      m_block.insert(std::make_pair(decl->m_name, decl));
   }
}

CRBaseDeclaration* CRBlock::find(const CUtiIDString& str)
{
   auto it = m_block.find(str);
   if (it == m_block.end()) {
      CRBlock* bl = getParentBlock();
      if (!bl) return NULL;
      return bl->find(str);

   }
   return it->second;
}


void CRTypeDefDeclaration::doParse(UtiLexer& parse)
{
    MatchIdentifier(parse, UTISTR("typedef"));

    CUtiIDString typestr;
    getParentBlock()->parseDeclaration(parse, typestr, m_parenttype, m_name);

}

CRTypeDefDeclaration::CRTypeDefDeclaration(CRBaseRule* p) :CRBaseDeclaration(p)
{
   m_parenttype = NULL;
}



uti_refptr<CRDataCell> CRDatabaseCommand::compile(CRAssembler* assembler)
{
    uti_refptr<CRDataCell> datacell(m_index->compile(assembler));
    CRRecordCell* rc = check_cast<CRRecordCell*>(datacell.get());
    DynArray1D<Lisp::LObject> params;


   params.push_back(IDSTR("!!!list"));
    params.push_back(rc->m_type->m_name);

    for (int i = 0; i < rc->m_subcells.size(); i++) {
        if (rc->m_subcells.get(i) == NULL) {

            //uti_refptr<CRDataCell> datacell2(m_index->compile(assembler));
            assert(false);
            continue;
        }
       rc->m_subcells[i].createFlatPDDLList(params, FLATMODE_CREATETEMPVARS, assembler);
    }
    Lisp::LObject as = Lisp::list(params);

    if (m_ct== dcInsert) {
    assembler->add(Lisp::list(IDSTR("asm_pddleffect"), IDSTR("PC"),as));
    } else  if (m_ct == dcRemove){
        assembler->add(Lisp::list(IDSTR("asm_pddlnoteffect"), IDSTR("PC"),as));

    } else if (m_ct == dcGoal) {
        std::wcout << "Error: Goal used in PDDL compilation!" << std::endl;
    }else if (m_ct == dcNotGoal) {
        std::wcout << "Error: Not Goal used in PDDL compilation!" << std::endl;
    }
    return uti_refptr<CRDataCell>();
}

uti_refptr<CRDataCell> CRDatabaseCommand::execute(CRNormalExecutionContext* econtext)
{
    uti_refptr<CRDataCell> datacell(m_index->execute(econtext));

     // uti_refptr<CRActionGenerator> generator = econtext->getCurrentGenerator(datacell.get());

         CRRecordCell* rc = check_cast<CRRecordCell*>(datacell.get());
         DynArray1D<Lisp::LObject> params;
         DynArray1D<Lisp::LObject> paramstype;
         DynArray1D<Lisp::LObject> paramandcelltype;
         params.push_back(rc->m_type->m_name);
         paramstype.push_back(Lisp::LObject(false));

         DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> > paramnameandtypelist;
         rc->m_type->createFlatVariableList(paramnameandtypelist, UTISTR(""));

         bool filterforatoms = econtext->m_syntheticpddl == false;
         for (int i = 0; i < paramnameandtypelist.size(); i++) {
             if (filterforatoms && dynamic_cast<CRClassDeclaration*>(paramnameandtypelist[i].second)== NULL) {
                 continue;
             }
            // paramstype.push_back(paramnameandtypelist[i].second->m_name);
         }
         for (int i = 0; i < rc->m_subcells.size(); i++) {
            rc->m_subcells[i].createFlatPDDLList(params, 0, NULL);
         }
         assert(paramnameandtypelist.size() == params.size()-1);
         for (int i = 0; i < rc->m_subcells.size(); i++) {
            rc->m_subcells[i].createFlatPDDLList(paramandcelltype, FLATMODE_COLLECTYPES, NULL);
         }
         assert(paramnameandtypelist.size() == paramandcelltype.size());
         for (int i = 0; i < paramandcelltype.size(); i++) {
             if (filterforatoms && dynamic_cast<CRClassDeclaration*>(paramnameandtypelist[i].second)== NULL) {
                 continue;
             }
             paramstype.push_back(paramandcelltype[i]);
         }

         DynArray1D<std::pair<CUtiIDString, Lisp::LObject> > fluentproperties;
         if (filterforatoms) {
             DynArray1D<Lisp::LObject> tempparams;
             tempparams.push_back(params[0]);
             for (int i = 0; i < paramnameandtypelist.size(); i++) {
                 if (dynamic_cast<CRClassDeclaration*>(paramnameandtypelist[i].second)== NULL) {
                     fluentproperties.push_back(std::make_pair(CUtiIDString(paramnameandtypelist[i].first), params[i+1]));

                     continue;
                 }
                 tempparams.push_back(params[i+1]);
             }
             params.clear();
             params.push_back(tempparams);

         }
         Lisp::LObject as = Lisp::list(params);
         Lisp::LObject typedata = Lisp::list(paramstype);
         std::wcout << "databaseCommand(" << as.print() << ", " << typedata.print() << ")" << std::endl;
          if (m_ct== dcGoal) {
            if (params.size() != 1)  econtext->m_goal.push_back(std::make_pair(as, typedata));

            //TODO: <, >
            for (auto &p: fluentproperties) {
                Lisp::LObject fluentpredicate = replaceNthArgument(as, 0, CUtiIDString(rc->m_type->m_name.str()+UTISTR("_")+p.first));
                std::wcout << "fluentpredicate = " << fluentpredicate.print() << std::endl;
                econtext->m_initial.push_back(std::make_pair(Lisp::list(IDSTR("="), fluentpredicate, p.second), Lisp::LObject()));
            }
         } else  if (m_ct== dcNotGoal) {
             if (params.size() != 1)  econtext->m_goal.push_back(std::make_pair(Lisp::list(IDSTR("not"), as), typedata));


         } else if (m_ct== dcInsert) {
              if (params.size() != 1) econtext->m_initial.push_back(std::make_pair(as, typedata));
             for (auto &p: fluentproperties) {
                 Lisp::LObject fluentpredicate = replaceNthArgument(as, 0, CUtiIDString(rc->m_type->m_name.str()+UTISTR("_")+p.first));
                 std::wcout << "fluentpredicate = " << fluentpredicate.print() << std::endl;
                 econtext->m_initial.push_back(std::make_pair(Lisp::list(IDSTR("="), fluentpredicate, p.second), Lisp::LObject()));
             }
          } else  if (m_ct== dcRemove) {
              if (params.size() != 1) econtext->m_initial.push_back(std::make_pair(Lisp::list(IDSTR("not"), as), typedata));

          }




    return uti_refptr<CRDataCell>();
}


uti_refptr<CRDataCell> CRAssert::compile(CRAssembler* assembler)
{
    uti_refptr<CRDataCell> res(m_index->compile(assembler));
    if (dynamic_cast<CRBoolCell*>(res.get()) != NULL) {
        CRBoolCell* c = (CRBoolCell*)res.get();
        if (c->toBool()) {

            return uti_refptr<CRDataCell>();
        } else {
            assembler->failedCurrentPath();
           return uti_refptr<CRDataCell>();
        }
    }else {
        CRPDDLVariable* pddlvar = check_cast<CRPDDLVariable*>(res.get());
        assert(dynamic_cast<CRBaseTypeDeclaration*>(pddlvar->m_type)!= NULL);
        assert(dynamic_cast<CRBaseTypeDeclaration*>(pddlvar->m_type)->m_data == erBool);
        assembler->add(Lisp::list(IDSTR("asm_assert"), IDSTR("PC"),pddlvar->m_name));

    }
    return uti_refptr<CRDataCell>();

}

uti_refptr<CRDataCell> CRAssert::execute(CRNormalExecutionContext* econtext)
{
   /* if (econtext->m_pddlmode) {
        uti_refptr<CRDataCell> res(m_index->execute(econtext));
        if (dynamic_cast<CRBoolCell*>(res.get()) != NULL) {
            CRBoolCell* c = (CRBoolCell*)res.get();
            if (c->toBool()) {

                return uti_refptr<CRDataCell>();
            } else {
                econtext->getPDDLTop()->m_failed = true;
               return uti_refptr<CRDataCell>();
            }
        } else {
            m_savedreference = res.get();
            econtext->getPDDLTop()->m_condition.push_back(res->toLisp(NULL));
        }  return uti_refptr<CRDataCell>();
    } else {*/


        bool b = m_index->execute(econtext)->toBool();
        if (!b) {
            std::wcout << "Assert failed: " << m_myline << std::endl;
            econtext->m_doexit = true;
        }
        return uti_refptr<CRDataCell>();
   // }
}



uti_refptr<CRDataCell> CRPrint::execute(CRNormalExecutionContext* econtext)
{
    for (int i = 0; i < m_params.size(); i++) {
        uti_refptr<CRDataCell> cell = m_params[i].execute(econtext);
        std::wcout << cell->printValue();
    }
    if (m_newline) {
        std::wcout << std::endl;
    }
    return uti_refptr<CRDataCell>();
}

uti_refptr<CRDataCell> CRPrint::compile(CRAssembler* assembler)
{
    for (int i = 0; i < m_params.size(); i++) {
        uti_refptr<CRDataCell> cell = m_params[i].compile(assembler);
        cell->compilePrint(assembler);
    }
    if (m_newline) {

        assembler->add(Lisp::list(IDSTR("asm_newline"), IDSTR("PC")));
    }
    return uti_refptr<CRDataCell>();
}

void CRDatabaseCommand::doParse(UtiLexer& parse)
{
    switch (m_ct) {
       case dcInsert: {

        MatchIdentifier(parse, UTISTR("insert"));
    }; break;
    case dcGoal: {

     MatchIdentifier(parse, UTISTR("goal"));
 }; break;
    case dcNotGoal: {

     MatchIdentifier(parse, UTISTR("notgoal"));
 }; break;
    case dcRemove: {

        MatchIdentifier(parse, UTISTR("remove"));
    }; break;
    }
    MatchCharacter(parse, L'(');
    m_index = getParentBlock()->parseExpression(parse, this, 0);
    MatchCharacter(parse, L')');
    if (!dynamic_cast<CRStructDeclaration*>(m_index->m_typedata.get())) {
       error(parse, UTISTR("A struct is expected in the database command!"));

    }
    CRStructDeclaration* str = check_cast<CRStructDeclaration*>(m_index->m_typedata.get());
    if (!getParentProgram()->isTableStruct(str)) {
       error(parse, UTISTR("A struct must be registered as a database command."));

    }
}

void CRAssert::doParse(UtiLexer& parse)
{

    m_myline = parse.getLine();
MatchIdentifier(parse, UTISTR("assert"));
   MatchCharacter(parse, L'(');
   m_index = getParentBlock()->parseExpression(parse, this, 0);
   MatchCharacter(parse, L')');
}


void CRPrint::doParse(UtiLexer& parse)
{
    if (m_newline) {
        MatchIdentifier(parse, UTISTR("println"));

    } else {
MatchIdentifier(parse, UTISTR("print"));
    }
   MatchCharacter(parse, L'(');
   while (!parse.IsThisNextCharacter(L')')) {

       m_params.push_back(getParentBlock()->parseExpression(parse, this, 0));
       if (!parse.IsThisNextCharacter(L')')) {
           MatchCharacter(parse, L',');
       }
   }
   MatchCharacter(parse, L')');
}

CRWait::CRWait(CRBaseRule* p) : CRCommand(p)
{

}

uti_refptr<CRDataCell> CRWait::compile(CRAssembler* assembler)
{
uti_refptr<CRDataCell> result = m_index->compile(assembler);
    assembler->add(Lisp::list(IDSTR("asm_wait"), IDSTR("PC"),assembler->blockToVar(result.get(),erFloat )));
    return uti_refptr<CRDataCell>();
}

uti_refptr<CRDataCell> CRWait::execute(CRNormalExecutionContext* econtext)
{

return uti_refptr<CRDataCell>();
}

void CRWait::doParse(UtiLexer& parse)
{
   MatchIdentifier(parse, UTISTR("wait"));
   MatchCharacter(parse, L'(');
   m_index = getParentBlock()->parseExpression(parse, this, 0);
   MatchCharacter(parse, L')');
}


uti_refptr<CRDataCell> CRReturn::execute(CRNormalExecutionContext* econtext)
{
    econtext->m_doreturn = true;
    if (m_index.get() != NULL) {
        econtext->m_returnvalue = m_index->execute(econtext).get();
    };
return uti_refptr<CRDataCell>();
}


uti_refptr<CRDataCell> CRReturn::compile(CRAssembler* assembler)
{
    DynArray1D<Lisp::LObject> params;


    uti_refptr<CRDataCell> res = m_index->compile(assembler).get();

    params.push_back(IDSTR("!!!list"));
    res->createFlatPDDLList(params,FLATMODE_CREATETEMPVARS, assembler);
  Lisp::LObject as = Lisp::list(params);

    assembler->add(Lisp::list(IDSTR("asm_return"), IDSTR("PC"),as));

    return uti_refptr<CRDataCell>();
}


uti_refptr<CRDataCell> CRIf::compile(CRAssembler* assembler)
{

    uti_refptr<CRDataCell> cell = m_condition->compile(assembler);
    int truelabel = assembler->createLabel();
    int endlabel = assembler->createLabel();

    assembler->add(Lisp::list(IDSTR("asm_conditionaljump"), IDSTR("PC"), assembler->blockToVar(cell.get(), erBool),  truelabel));
   if (m_falseblock.get()) { m_falseblock->compile(assembler);}
    assembler->add(Lisp::list(IDSTR("asm_jump"), IDSTR("PC"), endlabel));
    assembler->add(Lisp::list(IDSTR("asm_noop"), IDSTR("PC")), truelabel);
    uti_refptr<CRDataCell> res = m_trueblock->compile(assembler);

    assembler->add(Lisp::list(IDSTR("asm_noop"), IDSTR("PC")), endlabel);
    return uti_refptr<CRDataCell>();
}

uti_refptr<CRDataCell> CRIf::execute(CRNormalExecutionContext* econtext)
{
    uti_refptr<CRDataCell> cell = m_condition->execute(econtext);
/*if (econtext->m_pddlmode && cell->isBool() == false) {
   Lisp::LObject condition = cell->toLisp(NULL);
    econtext->pushPDDLBlock(condition);
    CUndoControl::getInstance()->snapshot();
       m_trueblock->execute(econtext);
    CUndoControl::getInstance()->rollback();
    econtext->popPDDLBlock();


    if (m_falseblock.get() != NULL) {
    econtext->pushPDDLBlock(Lisp::list(IDSTR("not"), condition));
    CUndoControl::getInstance()->snapshot();
       m_falseblock->execute(econtext);
    CUndoControl::getInstance()->rollback();
    econtext->popPDDLBlock();
    }
}else {*/
   if (cell->toBool()) {
       m_trueblock->execute(econtext);
   } else if (m_falseblock.get() != NULL) {
       m_falseblock->execute(econtext);

   }
 //  }
return uti_refptr<CRDataCell>();
}


uti_refptr<CRDataCell> CRFor::compile(CRAssembler* assembler)
{

    assembler->pushStack();
    m_startcommand->compile(assembler);

    int loopstart = assembler->createLabel();
    int endlabel = assembler->createLabel();
    assembler->add(Lisp::list(IDSTR("asm_noop"), IDSTR("PC")), loopstart);
     uti_refptr<CRDataCell> cell = m_checkcondition->compile(assembler);
     CUtiIDString resultvar = assembler->getNewTempVariable(EMPTYSTRING, IDSTR("bool"), false);
      assembler->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),resultvar,Lisp::list(IDSTR("vbool"), 0)));
     assembler->add(Lisp::list(IDSTR("asm_not"), IDSTR("PC"), resultvar, assembler->blockToVar(cell.get(), erBool)));
     assembler->add(Lisp::list(IDSTR("asm_conditionaljump"), IDSTR("PC"),resultvar, endlabel));



     assembler->add(Lisp::list(IDSTR("asm_jump"), IDSTR("PC"), loopstart));
    assembler->add(Lisp::list(IDSTR("asm_noop"), IDSTR("PC")), endlabel);


    assembler->popStack();
    return uti_refptr<CRDataCell>();
}

uti_refptr<CRDataCell> CRFor::execute(CRNormalExecutionContext* econtext)
{
    econtext->pushStack();
    m_startcommand->execute(econtext);
    while (m_checkcondition->execute(econtext)->toBool()) {

        m_block->execute(econtext);

        m_stepcommand->execute(econtext);
    }

    econtext->popStack();
return uti_refptr<CRDataCell>();
}


void CRReturn::doParse(UtiLexer& parse)
{

   MatchIdentifier(parse, UTISTR("return"));
   bool hasklammer = false;
   if (parse.IsThisNextCharacter(L'(')) {
      hasklammer = true;
   }
   if (hasklammer) {
      MatchCharacter(parse, L'(');
   }
   m_index = getParentBlock()->parseExpression(parse, this, 0);
   if (hasklammer) {
      MatchCharacter(parse, L')');
   }
}


void CRIf::doParse(UtiLexer& parse)
{

      MatchIdentifier(parse, UTISTR("if"));
      MatchCharacter(parse, L'(');
      m_condition = getParentBlock()->parseExpression(parse, this, 0);


      MatchCharacter(parse, L')');

      CRBlock* block = new CRBlock(this);
      m_trueblock = block;
      block->doParse(parse);

      if (parse.IsThisNextBezeichner(UTISTR("else"))) {
          MatchIdentifier(parse, UTISTR("else"));
          block = new CRBlock(this);
          m_falseblock = block;
          block->doParse(parse);
      }
}


void CRFor::doParse(UtiLexer& parse)
{
      MatchIdentifier(parse, UTISTR("if"));
      MatchCharacter(parse, L'(');

      CRSetVariable* as = new CRSetVariable(this);
      as->doParse(parse);
      m_startcommand = as;

      MatchCharacter(parse, L';');

      m_checkcondition = getParentBlock()->parseExpression(parse, this, 0);

      MatchCharacter(parse, L';');

      CRSetVariable* as2 = new CRSetVariable(this);
      as2->doParse(parse);
      m_stepcommand = as2;
      MatchCharacter(parse, L')');
      CRBlock* block = new CRBlock(this);
      m_block = block;
      block->doParse(parse);
}


uti_refptr<CRDataCell> CRFunctionCall::compile(CRAssembler* assembler)
{
    Array< uti_refptr<CRDataCell> > celldata;
    celldata.setSize(m_parameters.size());
    for (int i = 0; i < celldata.size(); i++) {
        uti_refptr<CRDataCell> param =m_parameters[i].compile(assembler);
        celldata[i] = param.get();
        //celldata[i] =param->m_type->createCell();
        //celldata[i]  = celldata[i]->replaceVarWithPDDL(assembler, NULL, EMPTYSTRING);
       // celldata[i]->copyFromCompile(assembler, param.get());
    }
    if (m_declaration != NULL) {
    if (m_declaration->m_callback.get() != NULL) {
   /* assembler->pushStack();
    for (int i = 0; i < celldata.size(); i++) {
        assembler->reserveVariable(std::get<1>(m_declaration->m_param_type_and_name[i]), celldata[i].get());
    }
    uti_refptr<CRDataCell> res = m_declaration->execute(assembler);

    assembler->popStack();

    return res;*/
       // m_declaration->m_callback->compile(assembler, celldata.get());
        assert(false);
    } else {
  //  return m_declaration->m_callback->compile(m_declaration, assembler, celldata);

        //Wandele alle Parameter in List
        DynArray1D<Lisp::LObject> paramlist;
        paramlist.push_back(IDSTR("!!!list"));

        for (int i = 0; i < celldata.size(); i++) {
            celldata[i]->createFlatPDDLList(paramlist, 0, NULL);
        }

        DynArray1D<Lisp::LObject> returnvalues;
        returnvalues.push_back(IDSTR("!!!list"));

        uti_refptr<CRDataCell> cell(NULL);
        if (m_declaration->m_returntypedata.get()) {
            cell = m_declaration->m_returntypedata->createCell();
            cell = assembler->replaceVarWithPDDL(cell.get(),m_declaration->getCallerVarName(), false);
            cell->createFlatPDDLList(returnvalues, 0 /*FLATMODE_CREATETEMPVARS*/, assembler);

        }
        assembler->add(Lisp::list(IDSTR("asm_procedure_call"), IDSTR("PC"),m_declaration->getLabelOrCreate(assembler), Lisp::list(paramlist), Lisp::list(returnvalues)));


        //ReturnTargetValues anlegen als Temp
       // asm_procedure_call(PC, NewPC, ListParamValues, ListReturnTargetVariables);

        //Call function
        return cell;
    }
    } else if (m_structdeclaration) {
       CRRecordCell* cell = check_cast<CRRecordCell*>( m_structdeclaration->createCell());
       for (int i = 0; i < celldata.size(); i++) {
           cell->m_subcells.exchange(i, celldata[i].get());
       }


       uti_refptr<CRDataCell> res(cell);
       return res;
    }
    uti_refptr<CRDataCell> res;
    return res;
}

uti_refptr<CRDataCell> CRFunctionCall::execute(CRNormalExecutionContext* econtext)
{

    Array< uti_refptr<CRDataCell> > celldata;
    celldata.setSize(m_parameters.size());
    for (int i = 0; i < celldata.size(); i++) {
        uti_refptr<CRDataCell> res = m_parameters[i].execute(econtext);
        celldata[i] = res->createRCopy();
    }
    if (m_declaration != NULL) {
    if (m_declaration->m_callback.get() == NULL) {
    econtext->pushStack();
    for (int i = 0; i < celldata.size(); i++) {
        econtext->reserveVariable(std::get<1>(m_declaration->m_param_type_and_name[i]), celldata[i].get());
    }
    uti_refptr<CRDataCell> res = m_declaration->execute(econtext);

    econtext->popStack();

    return res;
    } else {
    return m_declaration->m_callback->execute(m_declaration, econtext, celldata);
    }
    } else if (m_structdeclaration) {
       CRRecordCell* cell = check_cast<CRRecordCell*>( m_structdeclaration->createCell());
       for (int i = 0; i < celldata.size(); i++) {
           cell->m_subcells.exchange(i, celldata[i].get());
       }


       uti_refptr<CRDataCell> res(cell);
       return res;
    }
    uti_refptr<CRDataCell> res;
    return res;
}

void CRFunctionCall::doParse(UtiLexer& parse)
{
    if (m_declaration) {
   MatchIdentifier(parse, m_declaration->m_name.str());

   if (m_declaration->m_template_type_and_name.size() != 0) {

      MatchCharacter(parse, L'<');
      for (int i = 0; i < m_declaration->m_template_type_and_name.size(); i++) {
          CUtiString str;
          if (!parse.GetNextBezeichner(str)) {
              error(parse, UTISTR("Identifier expected!") );
          }
         uti_refptr<CRBaseDeclaration> eval;
         eval = dynamic_cast<CRBaseDeclaration*>(getParentBlock()->find(CUtiIDString(str)));
         if (!eval.get()) {
             error(parse, UTISTR("Type expected!") );
         }
         m_templateparameters.push_back(eval.get());
       //  CRBaseDeclaration* evaltype = eval->getType();
        /* if (!evaltype->canConvertTo(std::get<0>(m_declaration->m_param_type_and_name.at(i)).get())) {

             error(parse, UTISTR("Cannot convert '")+evaltype->m_name.str()+UTISTR("'' to '")+std::get<0>(m_declaration->m_param_type_and_name.at(i))->m_name.str()+UTISTR("'") );

         }*/

         if (i != m_declaration->m_template_type_and_name.size()-1) {

              MatchCharacter(parse, L',');
         }
      }

      MatchCharacter(parse, L'>');
      if (m_declaration->m_name == IDSTR("createTable")) {
          CRBaseDeclaration* dec = m_templateparameters.get(0);
          if (!dynamic_cast<CRStructDeclaration*>(dec)) {
              error(parse, UTISTR("The createTable function expects a struct!"));
          }
          getParentProgram()->m_structs.insert(check_cast<CRStructDeclaration*>(dec));

      }
   }

      MatchCharacter(parse, L'(');
      for (int i = 0; i < m_declaration->m_param_type_and_name.size(); i++) {
         uti_refptr<CREval> eval = getParentBlock()->parseExpression(parse, this, 0);
         m_parameters.push_back(eval);
         CRBaseDeclaration* evaltype = eval->getType();
         if (!evaltype->canConvertTo(std::get<0>(m_declaration->m_param_type_and_name.at(i)).get())) {

             error(parse, UTISTR("Cannot convert '")+evaltype->m_name.str()+UTISTR("'' to '")+std::get<0>(m_declaration->m_param_type_and_name.at(i))->m_name.str()+UTISTR("'") );

         }

         if (i != m_declaration->m_param_type_and_name.size()-1) {

              MatchCharacter(parse, L',');
         }
      }

      MatchCharacter(parse, L')');
    } else {
        MatchIdentifier(parse, m_structdeclaration->m_name.str());
        MatchCharacter(parse, L'(');
        for (int i = 0; i < m_structdeclaration->m_children_tan.size(); i++) {
           uti_refptr<CREval> eval = getParentBlock()->parseExpression(parse, this, 0);
           m_parameters.push_back(eval);
           CRBaseDeclaration* evaltype = eval->getType();
           if (!evaltype->canConvertTo(m_structdeclaration->m_children_tan.at(i)->m_typedata.get())) {

               error(parse, UTISTR("Cannot convert '")+evaltype->m_name.str()+UTISTR("' to '")+m_structdeclaration->m_children_tan.at(i)->m_typedata->m_name.str()+UTISTR("'") );

           }

           if (i != m_structdeclaration->m_children_tan.size()-1) {

                MatchCharacter(parse, L',');
           }
        }

        MatchCharacter(parse, L')');
    }
}

void CRVariable::doParse(UtiLexer& parse)
{
// void parseDeclaration(UtiLexer& parse, CUtiIDString& typestr, uti_refptr<CRBaseDeclaration> & typedata, CUtiIDString& name);
    getParentBlock()->parseDeclaration(parse, m_typename, m_type, m_name);
    if (parse.IsThisNextCharacter(L'=')) {
       MatchCharacter(parse, '=');

       m_initvalue = getParentBlock()->parseExpression(parse, this);
    }
}

void CRDeclaration::doParse(UtiLexer& parse)
{
// void parseDeclaration(UtiLexer& parse, CUtiIDString& typestr, uti_refptr<CRBaseDeclaration> & typedata, CUtiIDString& name);
    getParentBlock()->parseDeclaration(parse, m_type, m_typedata, m_name);
    if (parse.IsThisNextCharacter(L'=')) {
       MatchCharacter(parse, '=');

       m_value = getParentBlock()->parseExpression(parse, this);
    }
}

void CRSetVariable::doParse(UtiLexer& parse)
{
    m_variable = getParentBlock()->parseReference(parse, this);
    MatchCharacter(parse, '=');

    m_value = getParentBlock()->parseExpression(parse, this);
}


uti_refptr<CREval> CRBlock::parseExpression(UtiLexer& parse, CRBaseRule* parent, int layer)
{
   if (layer == m_program->m_operators.height()) {
    //Final layer
       CUtiString output;
       if (parse.GetNextIntNumber(output)) {
          int value = parse.GetIntValue(output);
          CRValue* v = new CRValue(parent);
          v->m_value = value;
          v->m_type = erInt;
          v->m_typedata = getParentBlock()->find(IDSTR("int"));
          assert(v->m_typedata.get() != NULL);
          return v;
       } else  if (parse.GetNextNumber(output)) {
          double value = CUtiString::strToFloat(output);
          CRValue* v = new CRValue(parent);
          v->m_value = value;
          v->m_type = erFloat;
          v->m_typedata = getParentBlock()->find(IDSTR("float"));
          assert(v->m_typedata.get() != NULL);
          return v;
       } else  if (parse.IsThisNextCharacter(34)) {
           CUtiString str;
          // inhalt = str;
           U_CHAR c = parse.GetNextChar();
           int count = 0;
           while ((c = parse.GetNextCharNoMove()) != 34)
           {
               if (c == 0)
               {
                    error(parse, UTEXT("String constant is not closed!"));

               };
               if (count > 1024) {
                    error(parse, UTEXT("String constant is not closed!"));

               }
               count++;
               parse.setPos(parse.getPos()+1);
               str += c;
           };
           parse.setPos(parse.getPos()+1);
           CRValue* v = new CRValue(parent);
           v->m_atom = CUtiIDString(str);
           v->m_type = erString;
           v->m_typedata = getParentBlock()->find(IDSTR("string"));
           assert(v->m_typedata.get() != NULL);
           return v;
        } else  if (parse.GetNextCharNoMove() == L'(') {
           /*double value = CUtiString::strToFloat(output);
           CRValue* v = new CRValue(parent);
           v->m_value = value;
           v->m_type = erFloat;
           v->m_typedata = getParentBlock()->find(IDSTR("float"));
           assert(v->m_typedata.get() != NULL);
           return v;*/
           MatchCharacter(parse, L'(');
           uti_refptr<CREval> res = parseExpression(parse,parent, 0);


           MatchCharacter(parse, L')');
           return res;
        } else if (parse.IsThisNextBezeichner(UTISTR("true"))) {
       parse.GetNextBezeichner(output);
          CRValue* v = new CRValue(parent);
          v->m_value = 1;
          v->m_type = erBool;
          v->m_typedata = getParentBlock()->find(IDSTR("bool"));
          assert(v->m_typedata.get() != NULL);
          return v;
       }else  if (parse.IsThisNextBezeichner(UTISTR("false"))) {
       parse.GetNextBezeichner(output);

          CRValue* v = new CRValue(parent);
          v->m_value = 0;
          v->m_type = erBool;
          v->m_typedata = getParentBlock()->find(IDSTR("bool"));
          assert(v->m_typedata.get() != NULL);
          return v;
       }else  if (parse.GetBezeichner(output)) {
          CUtiIDString id(output);
          CRBaseDeclaration* decl = find(id);

          if (id == IDSTR("exists")) {
              CRExists* func = new CRExists(parent);

             func->doParse(parse);
             return func;
          }


          if (decl == NULL) {

              error(parse, UTISTR("Identifier expected: ")+output);
          }
          if (decl->isConstant()) {
          CRValue* v = new CRValue(parent);
          v->m_atom = id;
          v->m_type = erVariable;
          v->m_typedata = check_cast<CRVariable*>(decl)->m_type;
          parse.GetNextBezeichner(output);
          return v;
          } else if (decl->isType()) {
              //call to constructor
              if (dynamic_cast<CRStructDeclaration*>(decl) == NULL) {

                  error(parse, UTISTR("Struct expected: ")+output);
              }
              CREvalFunction* func = new CREvalFunction(parent);
              uti_refptr<CRFunctionCall> call(new CRFunctionCall(func));
              call->m_structdeclaration = check_cast<CRStructDeclaration*>(decl);
             call->doParse(parse);
             func->m_call = call;
             func->m_typedata = decl;
             return func;
          } else if (decl->isFunction()) {
          //call to function
              CREvalFunction* func = new CREvalFunction(parent);
              uti_refptr<CRFunctionCall> call(new CRFunctionCall(func));
              call->m_declaration = check_cast<CRFunctionDeclaration*>(decl);

             call->doParse(parse);
             func->m_call = call;
             func->m_typedata =  check_cast<CRFunctionDeclaration*>(decl)->m_returntypedata;

             if (!func->m_typedata.get()) {
                 error(parse, UTISTR("A function is here expected!"));
             }


             return func;
          } else if (decl->isVariable()) {
              //Das könnte eine Referenz sein
              uti_refptr<CRReference> res(parseReference(parse, parent));
              assert(res->m_typedata.get() != NULL);
              return uti_refptr<CREval>(res.get());
          } else {

              decl = find(id);
                error(parse, UTISTR("Unknown Identifier: ")+output);
                return NULL;
          }
       } else {
           error(parse, UTISTR("Unknown Symbol: ")+CUtiString(parse.getText()[parse.getPos()]));
           return NULL;
       }
   } else {
      DynArray1D<std::tuple<int, CRTypedOperator*, uti_refptr<CREval> > > elements;
      uti_refptr<CREval> current(parseExpression(parse, parent,layer+1));
      assert(current.get() != nullptr);
      elements.push_back(std::make_tuple(-1,nullptr, current));
      CRBaseDeclaration* currenttype= current->m_typedata.get();
      while (true) {
          int operatorindex = -1;
          CRExpressionOperator* operatordata=NULL;
          for (int i = 0; i< m_program->m_operators.width(); i++) {
              if (m_program->m_operators(i, layer) == NULL) {
                 break;
              }
              CRExpressionOperator* op = m_program->m_operators(i, layer).get();
              if (op->isNextThenEat(parse)) {
                 operatorindex = op->m_expressiontype;
                 operatordata = op;
              }

          }

          if (operatorindex == -1) {
              break;
          }


          uti_refptr<CREval> next(parseExpression(parse, parent,layer+1));

           CRBaseDeclaration* oldtype=currenttype;
          currenttype = currenttype->executeTypeOperation(next->m_typedata.get(), operatorindex);
          elements.push_back(std::make_tuple(operatorindex,operatordata->getRType(oldtype, next->m_typedata.get()), next));
          if (currenttype == NULL) {
              error(parse, UTISTR("Cannnot convert ")+oldtype->m_name.str()+ UTISTR(" to ")+currenttype->m_name.str());
          }
      }
      if (elements.size() == 1) {
      return std::get<2>(elements[0]);
      } else {
          CRExpression* exp = new CRExpression(parent);
          exp->m_type = (TRExpressionType) std::get<0>(elements[1]);
          exp->m_datatype = std::get<1>(elements[1]);
          exp->m_a = std::get<2>(elements[0]);
          exp->m_b = std::get<2>(elements[1]);
          exp->m_a->changeParent(exp);
          exp->m_b->changeParent(exp);
          exp->m_typedata = exp->m_a->m_typedata->executeTypeOperation(exp->m_b->m_typedata.get(), exp->m_type);
          uti_refptr<CREval> result = exp;

          for (int i = 2; i < elements.size(); i++) {
              exp = new CRExpression(parent);
              exp->m_type = (TRExpressionType) std::get<0>(elements[i]);
              exp->m_datatype =  std::get<1>(elements[i]);
              exp->m_a = result;
              exp->m_b = std::get<2>(elements[i]);
              exp->m_a->changeParent(exp);
              exp->m_b->changeParent(exp);
              exp->m_typedata = exp->m_a->m_typedata->executeTypeOperation(exp->m_b->m_typedata.get(), exp->m_type);
              result = exp;


          }
          return result;
      }
   }
   assert(false);
   return uti_refptr<CREval>();
}


uti_refptr<CRDataCell> CRBlock::compile(CRAssembler* assembler)
{
    assembler->pushStack();
    for (int i = 0; i < m_command.size(); i++) {
        m_command[i].compile(assembler);
    }
    assembler->popStack();
    uti_refptr<CRDataCell> res(NULL);
    return res;
}

uti_refptr<CRDataCell> CRBlock::execute(CRNormalExecutionContext* econtext)
{
    //2 mal push, unsauber?

    if ( econtext->m_doreturn || econtext->m_doexit ) return uti_refptr<CRDataCell>();
   econtext->pushStack();
   for (int i = 0; i < m_command.size(); i++) {
       m_command[i].execute(econtext);
       if ( econtext->m_doreturn || econtext->m_doexit  ) return uti_refptr<CRDataCell>();
   }


   econtext->popStack();
   return uti_refptr<CRDataCell>();
}

uti_refptr<CRReference> CRBlock::parseReference(UtiLexer& parse, CRBaseRule* parent)
{
    uti_refptr<CRReference> currentreference;
    CUtiString id;
    if (!parse.GetNextBezeichner(id)) {
        error(parse, UTISTR("Identifier expected!"));
        return currentreference;
    }

    CRBaseDeclaration* typedata = find(CUtiIDString(id));
    if (typedata == NULL) {

        error(parse, UTISTR("The variable '")+id+ UTISTR("' was not found!"));
        return currentreference;
    }

    if (!dynamic_cast<CRVariable*>(typedata)) {
        error(parse, UTISTR("A variable was expected, but '")+id+ UTISTR("' was found!"));
        return currentreference;
    }
    CRVariable* var = check_cast<CRVariable*>(typedata);
    CRVarReference* varref = new CRVarReference(parent);
    varref->m_variable = var;
    varref->m_typedata = var->m_type;
    currentreference = varref;
    CRBaseDeclaration* currenttype = var->m_type.get();

    while (true) {
        if (parse.IsThisNextCharacter(L'.')) {
            if (!dynamic_cast<CRStructDeclaration*>(currenttype)) {

                error(parse, UTISTR("A struct type was expected"));
            }
            CRStructDeclaration* str = check_cast<CRStructDeclaration*>(currenttype);
            MatchCharacter(parse, L'.');
            CUtiString subelement;
            parse.GetNextBezeichner(subelement);
            CUtiIDString s(subelement);
            int sindex = str->findIndex(s);
            if (sindex == -1) {

                error(parse, UTISTR("The struct member does not exists: ")+subelement);
            }
            CRDeclaration* de = str->m_children_tan.get(sindex);
            CRReferenceStruct* structreference = new CRReferenceStruct(parent);
            structreference->m_source = currentreference;
            currentreference->changeParent(structreference);
            structreference->m_subelement = s;
            structreference->m_subelementindex = sindex;
            structreference->m_typedata = de->m_typedata;
            currentreference = structreference;
            currenttype = de->m_typedata.get();


        } else if (parse.IsThisNextCharacter(L'[')) {
            if (!dynamic_cast<CRArrayDeclaration*>(currenttype)) {

                error(parse, UTISTR("An array type was expected!"));
            }
            CRArrayDeclaration* str = check_cast<CRArrayDeclaration*>(currenttype);
            CReferenceList<CREval> subexpressions;
            while (parse.IsThisNextCharacter('[')) {
                MatchCharacter(parse, L'[');
                uti_refptr<CREval> exp = parseExpression(parse, parent);
                exp->doParse(parse);
                subexpressions.push_back(exp);
                MatchCharacter(parse, L']');

            }
            if (subexpressions.size() != str->m_size.size()) {

                error(parse, UTISTR("The dimension of the array is not correct!"));
            }
            uti_refptr<CRReferenceArrayAccess> aaccess(new CRReferenceArrayAccess(parent));
            aaccess->m_source = currentreference;
            for (auto & el: subexpressions) {
                el.changeParent(aaccess.get());
                aaccess->m_index.push_back(el);
            }
            currentreference = aaccess;
            aaccess->m_source->changeParent(aaccess.get());
            currenttype = str->m_basetype->m_typedata.get();
            currentreference->m_typedata = aaccess->m_typedata;

        } else {
            break;
        }
    }

    return currentreference;
}

void CRBlock::parseDeclaration(UtiLexer& parse, CUtiIDString& typestr, uti_refptr<CRBaseDeclaration> &typedata, CUtiIDString& name)
{
    CUtiString id;
    if (!parse.GetNextBezeichner(id)) {
       error(parse, UTISTR("Identifier expected!"));
       return;
    }
    typedata = NULL;
    if (id == UTISTR("struct")) {

       error(parse, UTISTR("Struct declaration is not supported!"));
       return;
    }
    typedata = dynamic_cast<CRBaseDeclaration*>(find(CUtiIDString(id)));
    if (typedata == NULL) {

       error(parse, UTISTR("The type '")+id+ UTISTR("' was not found!"));
       return;
    }
    if (!dynamic_cast<CRBaseDeclaration*>(typedata.get())) {
       error(parse, UTISTR("A type was expected, but '")+id+ UTISTR("' was found!"));
       return;
    }
    CUtiString n;
    if (!parse.GetNextBezeichner(n)) {

       error(parse, UTISTR("The name of the element is here expected."));
       return;
    }
    name = CUtiIDString(n);
    DynArray1D<int> dimensions;
    while (parse.IsThisNextCharacter(L'[')) {
       MatchCharacter(parse, L'[');
       CUtiString value;
       if (!parse.GetNextIntNumber(value)) {

          error(parse, UTISTR("The Array size is here expected!"));
       }
       int v = parse.GetIntValue(value);

       MatchCharacter(parse, L']');
       dimensions.push_back(v);

    }
    if (dimensions.size() != 0) {
       CUtiIDString newname = m_program->getUniqueVarName();
       uti_refptr<CRArrayDeclaration> vararray(new CRArrayDeclaration(this));
       vararray->m_name = newname;
       vararray->m_basetype = new CRDeclaration(vararray.get(), typedata.get());
       vararray->m_size.push_back(dimensions);
       addChild(parse, vararray.get());
       typedata = vararray;
    }
}

bool isFunction(UtiLexer& parse, CRBlock* parentblock)
{
assert(parentblock != NULL);
    int p = parse.getPos();

    CUtiString id;
    if (!parse.GetNextBezeichner(id)) {
      parse.setPos(p);
       return false;
    }
    CUtiIDString m_returntype;
    CRBaseDeclaration* m_returntypedata = NULL;
    if (id == UTISTR("struct")) {

       parse.setPos(p);
       return false;
    }
    m_returntypedata = dynamic_cast<CRBaseDeclaration*>(parentblock->find(CUtiIDString(id)));
    if (m_returntypedata == NULL) {
       parse.setPos(p);
       return false;
    }
    m_returntype = CUtiIDString(id);
    if (!dynamic_cast<CRBaseDeclaration*>(m_returntypedata)) {
       parse.setPos(p);
       return false;
    }

    CUtiString n;
    if (!parse.GetNextBezeichner(n)) {
parse.setPos(p);
       return false;
    }
   // m_type = rNormal;
    if (parse.IsThisNextCharacterSequence(UTISTR("::"))) {
        parse.GetNextCharSequence(UTISTR("::"));
        if (n == UTISTR("Actions")) {
        //   m_type = rAction;
        } else if (n == UTISTR("Events")) {
          // m_type = rEvent;
        } else {
         //  parse.setPos(p);
          // return false;
        }
       if (!parse.GetBezeichner(n)) {

          parse.setPos(p);
          return false;
       }
    }

//Doto, how do we deal with no forward declaration, first step: collect functions?
//First pass: collection function bodies with parse the bodies
//Second pass: when we hit the function body, we parse it completely again

    if (!parse.IsThisNextCharacter(L'(')) {
          parse.setPos(p);
    return false;
    }
   // MatchCharacter(parse, '(');

    parse.setPos(p);
    return true;
}

void CRFunctionDeclaration::doParse(UtiLexer& parse)
{
    CUtiString id;
    if (!parse.GetNextBezeichner(id)) {
       error(parse, UTISTR("Identifier expected!"));
       return;
    }
    m_returntypedata = NULL;
    if (id == UTISTR("struct")) {

       error(parse, UTISTR("Struct declaration is not supported!"));
       return;
    }
    m_returntypedata = dynamic_cast<CRBaseDeclaration*>(getParentBlock()->find(CUtiIDString(id)));
    if (m_returntypedata.get() == NULL) {

       error(parse, UTISTR("The type '")+id+ UTISTR("' was not found!"));
       return;
    }
    m_returntype = CUtiIDString(id);
    if (!dynamic_cast<CRBaseDeclaration*>(m_returntypedata.get())) {
       error(parse, UTISTR("A type was expected, but '")+id+ UTISTR("' was found!"));
       return;
    }

    CUtiString n;
    if (!parse.GetNextBezeichner(n)) {

       error(parse, UTISTR("The name of the procedure is here expected."));
       return;
    }
    m_type = rNormal;
    if (parse.IsThisNextCharacterSequence(UTISTR("::"))) {
        parse.GetNextCharSequence(UTISTR("::"));
        if (n == UTISTR("Actions")) {
           m_type = rAction;
        } else if (n == UTISTR("Events")) {
           m_type = rEvent;
        } else {
           error(parse, UTISTR("Unknown Class: ")+n);
        }
       if (!parse.GetNextBezeichner(n)) {

          error(parse, UTISTR("The name of the procedure is here expected."));
          return;
       }
    }

//Doto, how do we deal with no forward declaration, first step: collect functions?
//First pass: collection function bodies with parse the bodies
//Second pass: when we hit the function body, we parse it completely again

    MatchCharacter(parse, '(');

    m_name = CUtiIDString(n);
    bool first = true;
    while (!parse.IsThisNextCharacter(L')')) {

        if (!first) MatchCharacter(parse, ',');
        first = false;
        CUtiIDString typestr;
        uti_refptr<CRBaseDeclaration> typedata;
        CUtiIDString name;
        getParentBlock()->parseDeclaration(parse, typestr, typedata,name);
        m_param_type_and_name.push_back(std::make_pair(typedata, name));
    }

    MatchCharacter(parse, ')');
    m_block = new CRBlock(this);
    for (int i = 0; i < m_param_type_and_name.size(); i++) {
       CRBaseDeclaration* decl = std::get<0>(m_param_type_and_name[i]).get();
       CUtiIDString name = std::get<1>(m_param_type_and_name[i]);
       CRVariable* var = new CRVariable(m_block.get());
       var->m_name = name;
       var->m_isconstant = false;
       var->m_isparameter = true;
       var->m_type = decl;
       m_block->addChild(parse, var);
    }
    m_block->doParse(parse);

}
/*
Lisp::LObject CRBoolExpression::toLisp(CRPDDLDecision* decision)
{
    return Lisp::list(IDSTR("<"), m_arg1->toLisp(decision), m_arg2->toLisp(decision));

}*/

void CRBlock::doParse(UtiLexer& parse)
{

   if (m_needbraces) {
    MatchCharacter(parse, '{');
    }
    while (parse.IsThisNextCharacter('}') == false && parse.isEOF() == false) {
        if (parse.IsThisNextBezeichner(UTISTR("assert"))) {
            CRAssert* as = new CRAssert(this);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else if (parse.IsThisNextBezeichner(UTISTR("print"))) {
            CRPrint* as = new CRPrint(this);
            as->m_newline = false;
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        }else if (parse.IsThisNextBezeichner(UTISTR("println"))) {
            CRPrint* as = new CRPrint(this);
            as->m_newline = true;
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else  if (parse.IsThisNextBezeichner(UTISTR("insert"))) {
            CRDatabaseCommand* as = new CRDatabaseCommand(this, dcInsert);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else  if (parse.IsThisNextBezeichner(UTISTR("goal"))) {
            CRDatabaseCommand* as = new CRDatabaseCommand(this, dcGoal);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        }  else  if (parse.IsThisNextBezeichner(UTISTR("notgoal"))) {
            CRDatabaseCommand* as = new CRDatabaseCommand(this, dcNotGoal);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else  if (parse.IsThisNextBezeichner(UTISTR("remove"))) {
            CRDatabaseCommand* as = new CRDatabaseCommand(this, dcRemove);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else if (parse.IsThisNextBezeichner(UTISTR("struct"))) {
            CRStructDeclaration* as = new CRStructDeclaration(this);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else if (parse.IsThisNextBezeichner(UTISTR("class"))) {
            CRClassDeclaration* as = new CRClassDeclaration(this);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else if (parse.IsThisNextBezeichner(UTISTR("typedef"))) {
            CRTypeDefDeclaration* as = new CRTypeDefDeclaration(this);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        }else  /*else if (parse.IsThisNextBezeichner(UTISTR("class"))) {
            CRObjectType* as = new CRObjectType(this);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        }else */if (parse.IsThisNextBezeichner(UTISTR("wait"))) {
            CRWait* as = new CRWait(this);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else if (parse.IsThisNextBezeichner(UTISTR("if"))) {
            CRIf* as = new CRIf(this);
            as->doParse(parse);
            addChild(parse, as);
            //MatchCharacter(parse, ';');
        } else if (parse.IsThisNextBezeichner(UTISTR("return"))) {
            CRReturn* as = new CRReturn(this);
            as->doParse(parse);
            addChild(parse, as);
            MatchCharacter(parse, ';');
        } else if (parse.IsThisNextBezeichner(UTISTR("for"))) {
            CRFor* as = new CRFor(this);
            as->doParse(parse);
            addChild(parse, as);
          //  MatchCharacter(parse, ';');
        } else if (parse.IsThisNextBezeichner(UTISTR("const"))) {
            CUtiString tokenstr;
            parse.GetNextBezeichner(tokenstr);
            if (parse.GetBezeichner(tokenstr)) {
                CUtiIDString token(tokenstr);
                CRBaseDeclaration* decl = find(token);
                if (!decl) {
                    error(parse, UTISTR("Unexpected Identifier: ")+token.str());
                    return;
                }
                if (dynamic_cast<CRBaseDeclaration*>(decl) != NULL) {
                //TODO: It could also be function declaration
                    bool isfunction = isFunction(parse, this);


                    if (!isfunction) {
                    CRVariable* as = new CRVariable(this);
                    as->m_isconstant = true;
                    as->doParse(parse);
                    addChild(parse, as);
                    MatchCharacter(parse, ';');
                    }else {

                        error(parse, UTISTR("Expected Variable Deklaration: ")+token.str());
                   // MatchCharacter(parse, ';');
                    }

                }  else {
                    error(parse, UTISTR("Unexpected Identifier: ")+token.str());
                }
            } else {

                    error(parse, UTISTR("Expected Identifier "));
            }
          //  MatchCharacter(parse, ';');
        } else if (parse.IsThisNextCharacter(L'{')) {
            CRBlock* as = new CRBlock(this);

            addChild(parse, as);
            as->doParse(parse);
        } else {
            CUtiString tokenstr;
            if (parse.GetBezeichner(tokenstr)) {
                CUtiIDString token(tokenstr);
                CRBaseDeclaration* decl = find(token);
                if (!decl) {
                    error(parse, UTISTR("Unexpected Identifier: ")+token.str());
                    return;
                }
                if (dynamic_cast<CRFunctionDeclaration*>(decl) != NULL) {
                    CRFunctionCall* as = new CRFunctionCall(this);
                    as->m_declaration = (CRFunctionDeclaration*)decl;
                    as->doParse(parse);
                    addChild(parse, as);
                    MatchCharacter(parse, ';');

                }else if (dynamic_cast<CRVariable*>(decl) != NULL) {
                    CRSetVariable* as = new CRSetVariable(this);
                    as->doParse(parse);
                    addChild(parse, as);
                    MatchCharacter(parse, ';');

                } else if (dynamic_cast<CRBaseDeclaration*>(decl) != NULL) {
                //TODO: It could also be function declaration
                    bool isfunction = isFunction(parse, this);


                    if (!isfunction) {
                    CRVariable* as = new CRVariable(this);
                    as->doParse(parse);
                    addChild(parse, as);
                    MatchCharacter(parse, ';');
                    }else {
                    CRFunctionDeclaration* as = new CRFunctionDeclaration(this);
                    as->doParse(parse);
                    addChild(parse, as);
                   // MatchCharacter(parse, ';');
                    }

                }  else {
                    error(parse, UTISTR("Unexpected Identifier: ")+token.str());
                }
            } else {

                    error(parse, UTISTR("Expected Identifier "));
            }
        }
        if (m_singleline) {
           break;
        }
    }

   if (m_needbraces) {
    MatchCharacter(parse, '}');
    }
}

enum TSpecialFunction {spCreateTable, spRegisterAction, spRegisterEvent,spRegisterProcess, spSqrt, spSqr, spSin, spCos, spTan, spASin, spACos, spATan, spRound, spCeil, spFloor};

class CRSpecialFunction : public CRFunctionDeclaration_Callback
{
public:
    TSpecialFunction m_func;
    CRSpecialFunction(TSpecialFunction f)
    {
    m_func = f;
    }

    virtual uti_refptr<CRDataCell> execute(CRFunctionDeclaration* func, CRNormalExecutionContext* econtext, Array<uti_refptr<CRDataCell> >& parameters)
    {
         switch (m_func) {
         case spCreateTable: {


         return uti_refptr<CRDataCell>();
         }; break;
         case spRegisterAction: {


         return uti_refptr<CRDataCell>();
         }; break;
         case spRegisterProcess: {


             return uti_refptr<CRDataCell>();
             }; break;
         case spRegisterEvent: {


         return uti_refptr<CRDataCell>();
         }; break;
         case spSqrt: {


         return uti_refptr<CRDataCell>(new CRFloatCell(sqrt(check_cast<CRFloatCell*>(parameters[0].get())->data) ));
         }; break;
         case spSqr: {
 return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(parameters[0].get())->data*check_cast<CRFloatCell*>(parameters[0].get())->data ));

         }; break;
         case spSin: {

         return uti_refptr<CRDataCell>(new CRFloatCell(sin(check_cast<CRFloatCell*>(parameters[0].get())->data) ));

         }; break;
         case spCos: {

         return uti_refptr<CRDataCell>(new CRFloatCell(cos(check_cast<CRFloatCell*>(parameters[0].get())->data) ));

         }; break;
         case spTan: {

         return uti_refptr<CRDataCell>(new CRFloatCell(tan(check_cast<CRFloatCell*>(parameters[0].get())->data) ));

         }; break;
         case spASin: {


         return uti_refptr<CRDataCell>(new CRFloatCell(asin(check_cast<CRFloatCell*>(parameters[0].get())->data) ));
         }; break;
         case spACos: {


         return uti_refptr<CRDataCell>(new CRFloatCell(acos(check_cast<CRFloatCell*>(parameters[0].get())->data) ));
         }; break;
         case spATan: {


         return uti_refptr<CRDataCell>(new CRFloatCell(atan(check_cast<CRFloatCell*>(parameters[0].get())->data) ));
         }; break;
         case spRound: {

         return uti_refptr<CRDataCell>(new CRIntCell(binround(check_cast<CRFloatCell*>(parameters[0].get())->data) ));

         }; break;
         case spCeil: {


         return uti_refptr<CRDataCell>(new CRIntCell(std::ceil(check_cast<CRFloatCell*>(parameters[0].get())->data) ));
         }; break;
         case spFloor: {

         return uti_refptr<CRDataCell>(new CRIntCell(bintrunc(check_cast<CRFloatCell*>(parameters[0].get())->data) ));

         }; break;

         default: {
         assert(false);
         }; break;
         }
         return uti_refptr<CRDataCell>();
    }
};


CRProgram::CRProgram() : CRBlock(NULL)
{
m_program = this;
m_name = UTISTR("noname");
   m_operators.setSize(10,5);
   // erInt, erFloat, erObject, erBool, erComplex, erString
   m_operators(0,0) = (new CRExpressionOperator(UTISTR("||"), eLogicalOr, true))->addType(new CRLogicalTypedOperator(eLogicalOr, erBool, erBool, erBool));

   m_operators(0,1) = (new CRExpressionOperator(UTISTR("&&"), eLogicalAnd, true))->addType(new CRLogicalTypedOperator(eLogicalAnd, erBool, erBool, erBool));

    m_operators(0,2) = (new CRExpressionOperator(UTISTR("<"), eLess, true))->addType(new CRLessTypedOperator(eLess, erInt, erInt, erBool))->addType(new CRLessTypedOperator(eLess, erFloat, erInt, erBool))->addType(new CRLessTypedOperator(eLess, erFloat, erFloat, erBool))->addType(new CRLessTypedOperator(eLess, erInt, erFloat, erBool));
    m_operators(1,2) = (new CRExpressionOperator(UTISTR(">"), eGreater, true))->addType(new CRLessTypedOperator(eGreater, erInt, erInt, erBool))->addType(new CRLessTypedOperator(eGreater, erFloat, erInt, erBool))->addType(new CRLessTypedOperator(eGreater, erFloat, erFloat, erBool))->addType(new CRLessTypedOperator(eGreater, erInt, erFloat, erBool));
    m_operators(2,2) = (new CRExpressionOperator(UTISTR(">="), eEGreater, true))->addType(new CRLessTypedOperator(eEGreater, erInt, erInt, erBool))->addType(new CRLessTypedOperator(eEGreater, erFloat, erInt, erBool))->addType(new CRLessTypedOperator(eEGreater, erFloat, erFloat, erBool))->addType(new CRLessTypedOperator(eEGreater, erInt, erFloat, erBool));
    m_operators(3,2) = (new CRExpressionOperator(UTISTR("<="), eELess, true))->addType(new CRLessTypedOperator(eELess, erInt, erInt, erBool))->addType(new CRLessTypedOperator(eELess, erFloat, erInt, erBool))->addType(new CRLessTypedOperator(eELess, erFloat, erFloat, erBool))->addType(new CRLessTypedOperator(eELess, erInt, erFloat, erBool));
    m_operators(4,2) = (new CRExpressionOperator(UTISTR("=="), eEqual, true))->addType(new CRLessTypedOperator(eEqual, erInt, erInt, erBool))->addType(new CRLessTypedOperator(eEqual, erFloat, erInt, erBool))->addType(new CRLessTypedOperator(eEqual, erFloat, erFloat, erBool))->addType(new CRLessTypedOperator(eEqual, erInt, erFloat, erBool));

   m_operators(0,3) = (new CRExpressionOperator(UTISTR("+"), eAdd, true))->addType(new CRPlusTypedOperator(eAdd, erInt, erInt, erInt))->addType(new CRPlusTypedOperator(eAdd, erInt, erFloat, erFloat))->addType(new CRPlusTypedOperator(eAdd, erFloat, erInt, erFloat))->addType(new CRPlusTypedOperator(eAdd, erFloat, erFloat, erFloat))->addType(new CRPlusTypedOperator(eAdd, erString, erString, erString));
   m_operators(1,3) = (new CRExpressionOperator(UTISTR("-"), eSub, true))->addType(new CRPlusTypedOperator(eSub, erInt, erInt, erInt))->addType(new CRPlusTypedOperator(eSub, erInt, erFloat, erFloat))->addType(new CRPlusTypedOperator(eSub,erFloat, erInt, erFloat))->addType(new CRPlusTypedOperator(eSub, erFloat, erFloat, erFloat));
   m_operators(0,4) = (new CRExpressionOperator(UTISTR("*"), eMul, true))->addType(new CRPlusTypedOperator(eMul, erInt, erInt, erInt))->addType(new CRPlusTypedOperator(eMul, erInt, erFloat, erFloat))->addType(new CRPlusTypedOperator(eMul,erFloat, erInt, erFloat))->addType(new CRPlusTypedOperator(eMul, erFloat,erFloat, erFloat));
   m_operators(1,4) = (new CRExpressionOperator(UTISTR("/"), eDiv, true))->addType(new CRPlusTypedOperator(eDiv, erInt, erInt, erFloat))->addType(new CRPlusTypedOperator(eDiv, erInt, erFloat, erFloat))->addType(new CRPlusTypedOperator(eDiv,erFloat, erInt, erFloat))->addType(new CRPlusTypedOperator(eDiv, erFloat, erFloat, erFloat));
   m_needbraces = false;
   m_tempindex = 0;
   initOperators();UtiLexer l;
{
   CRBaseTypeDeclaration* decl_void = new CRBaseTypeDeclaration(this);
   decl_void->m_name = IDSTR("void");
   decl_void->m_data = erVoid;

   addChild(l, decl_void);
   }
   {
   CRBaseTypeDeclaration* decl_void = new CRBaseTypeDeclaration(this);
   decl_void->m_name = IDSTR("int");
   decl_void->m_data = erInt;

   addChild(l, decl_void);
   }
    {
   CRBaseTypeDeclaration* decl_void = new CRBaseTypeDeclaration(this);
   decl_void->m_name = IDSTR("float");
   decl_void->m_data = erFloat;

   addChild(l, decl_void);
   }
    {
   CRBaseTypeDeclaration* decl_void = new CRBaseTypeDeclaration(this);
   decl_void->m_name = IDSTR("bool");
   decl_void->m_data = erBool;

   addChild(l, decl_void);
   }
    {
   CRBaseTypeDeclaration* decl_void = new CRBaseTypeDeclaration(this);
   decl_void->m_name = IDSTR("string");
   decl_void->m_data = erString;

   addChild(l, decl_void);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("createTable");
       rfunc->m_callback = new CRSpecialFunction(spCreateTable);
       rfunc->m_template_type_and_name.push_back(std::make_pair(uti_refptr<CRBaseDeclaration>(NULL),IDSTR("structtype")));
       rfunc->m_returntypedata = getBaseType(erVoid);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("registerAction");
       rfunc->m_callback = new CRSpecialFunction(spRegisterAction);
       rfunc->m_template_type_and_name.push_back(std::make_pair(uti_refptr<CRBaseDeclaration>(NULL),IDSTR("proctype")));
       rfunc->m_returntypedata = getBaseType(erVoid);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("registerEvent");
       rfunc->m_callback = new CRSpecialFunction(spRegisterEvent);
       rfunc->m_template_type_and_name.push_back(std::make_pair(uti_refptr<CRBaseDeclaration>(NULL),IDSTR("proctype")));
       rfunc->m_returntypedata = getBaseType(erVoid);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("registerProcess");
       rfunc->m_callback = new CRSpecialFunction(spRegisterProcess);
       rfunc->m_template_type_and_name.push_back(std::make_pair(uti_refptr<CRBaseDeclaration>(NULL),IDSTR("proctype")));
       rfunc->m_returntypedata = getBaseType(erVoid);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("sqrt");
       rfunc->m_callback = new CRSpecialFunction(spSqrt);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erFloat);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("sqr");
       rfunc->m_callback = new CRSpecialFunction(spSqr);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erFloat);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("round");
       rfunc->m_callback = new CRSpecialFunction(spRound);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erInt);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("ceil");
       rfunc->m_callback = new CRSpecialFunction(spCeil);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erInt);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("floor");
       rfunc->m_callback = new CRSpecialFunction(spFloor);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erInt);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("cos");
       rfunc->m_callback = new CRSpecialFunction(spCos);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erFloat);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("sin");
       rfunc->m_callback = new CRSpecialFunction(spSin);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erFloat);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("tan");
       rfunc->m_callback = new CRSpecialFunction(spTan);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erFloat);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("acos");
       rfunc->m_callback = new CRSpecialFunction(spACos);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erFloat);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("asin");
       rfunc->m_callback = new CRSpecialFunction(spASin);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erFloat);
       addChild(l, rfunc);
   }
   {
       CRFunctionDeclaration* rfunc = new CRFunctionDeclaration(this);
       rfunc->m_name = IDSTR("atan");
       rfunc->m_callback = new CRSpecialFunction(spATan);
       rfunc->m_param_type_and_name.push_back(std::make_pair(getBaseType(erFloat),IDSTR("value")));
       rfunc->m_returntypedata = getBaseType(erFloat);
       addChild(l, rfunc);
   }
}


CRBaseTypeDeclaration *CRProgram::getBaseType(TRBasicDataType t)
{
     for (int i = 0; i < m_command.size(); i++) {
         CRBaseTypeDeclaration* d = dynamic_cast<CRBaseTypeDeclaration*>(m_command.at(i));
         if (d) {
             if (d->m_data == t) return d;
         }
     }
     assert(false);
     return NULL;
}


bool CRProgram::isTableStruct(CRStructDeclaration* decl)
{
   return m_structs.find(decl) != m_structs.end();
}

void CRProgram::collectObjects(CRDataCell* data, std::map<CUtiIDString, CUtiIDString> &objects)
{
    if (dynamic_cast<CRClassCell*>(data) != NULL) {
        CRClassCell* d = (CRClassCell*)data;
        if (d->m_objectname.str() != 0) {
            objects.insert(std::make_pair(d->m_objectname, d->m_type->m_name));
        }
    } else if (dynamic_cast<CRArrayCell*>(data) != NULL) {
        CRArrayCell* d = (CRArrayCell*)data;
        for (int i = 0; i < d->m_cells.size(); i++) {
            collectObjects(d->m_cells[i].get(), objects);
        }
    }else if (dynamic_cast<CRRecordCell*>(data) != NULL) {
        CRRecordCell* d = (CRRecordCell*)data;
        for (int i = 0; i < d->m_subcells.size(); i++) {
            collectObjects(d->m_subcells.get(i), objects);
        }
    }
}

void CRProgram::collectObjects(const Lisp::LObject & params, const Lisp::LObject& types, std::map<CUtiIDString, CUtiIDString>& objects)
{
    if (types.isNil()) return;
    Lisp::LObject extractedparams;
    if (Lisp::car(params).getAtom() == IDSTR("not")) {
      extractedparams = Lisp::cdr(params);
    } else {
       extractedparams = params;
    }
    std::wcout << "collectObjects(" << params.print() << ", " << types.print()<< ")" << std::endl;
   auto it = extractedparams.begin();
   auto ittype = types.begin();
   auto itend =extractedparams.end();
   int index = 0;
   while (it != itend) {

       if (index > 0) {
           objects.insert(std::make_pair((*it).getAtom(), (*ittype).getAtom()));
       }
       ittype++;
       it++;
       index++;
   }
}

void CRProgram::executeProgram(const CUtiString& filename)
{
  uti_refptr<CRNormalExecutionContext> context(new CRNormalExecutionContext());
  std::map<CUtiIDString, CUtiIDString > objects;

  CRFunctionDeclaration* main=NULL;
  for (int i = 0; i < m_command.size(); i++) {
      CRCommand* r = m_command.get(i);
      if (dynamic_cast<CRFunctionDeclaration*>(r) != NULL && check_cast<CRFunctionDeclaration*>(r)->m_param_type_and_name.size() == 0 && check_cast<CRFunctionDeclaration*>(r)->m_name == IDSTR("main")) {
          CRFunctionDeclaration* fd = check_cast<CRFunctionDeclaration*>(r);
          main = fd;
      } else if (dynamic_cast<CRVariable*>(r)) {
          uti_refptr<CRDataCell>data = r->execute(context.get());
          //Nicht notig, sollte in Domain schon erzeugt sein!
          //collectObjects(data.get(), objects);
      }

  }
  if (main != NULL) {
      context->pushStack();
      main->execute(context.get());
      context->popStack();
  }
  uti_refptr<CStringList> output(new CStringList());
  CUtiString problemname = UTISTR("main");
  output->add(UTISTR("(define (problem ")+problemname+UTISTR(")"));
  output->add(UTISTR("(:domain ")+m_name+UTISTR(")"));
  //Objektlsite aus den verwendeten!
  CUtiString objectlist;
  for (auto p: context->m_initial) {
      collectObjects(p.first, p.second, objects);
  }
  for (auto p: context->m_goal) {
      collectObjects(p.first, p.second, objects);
  }

  for (auto it : objects) {
      if (it.second.empty()) {
          objectlist += UTISTR(" ")+it.first.str();
      } else {
          objectlist += UTISTR(" ")+it.first.str()+UTISTR(" - ")+it.second.str();

      }
  }

  output->add(UTISTR("(:objects")+objectlist+UTISTR(")"));

  output->add(UTISTR("(:init "));
  for (auto p: context->m_initial) {
      //TODO: We need to filter this!
      output->add(UTISTR("    ")+p.first.print());

  }

  /*(= (length roomaroomb) 30 )(= (length roombrooma) 30 )(= (total-cost) 0 )*/


  output->add(UTISTR(")"));
  output->add(UTISTR("(:goal "));
  for (auto p: context->m_goal) {
      output->add(UTISTR("    ")+p.first.print());

  }

  output->add(UTISTR(")"));

  output->add(UTISTR(")"));

  /*(define (problem strips-gripper2)
    (:domain gripper-strips)
    (:objects rooma roomb ball1 ball2 left right)
    (:init (room rooma)
           (room roomb)
           (ball ball1)
           (ball ball2)
           (gripper left)
           (gripper right)
           (at-robby rooma)
           (free left)
           (free right)
           (at ball1 rooma)
           (at ball2 rooma))
    (:goal (at ball1 roomb)))
*/
  for (int i = 0;i < output->size(); i++) {
     std::wcout << output->getLine(i) << std::endl;
  }

  output->saveToFile(filename);
}

void CRProgram::initOperators()
{

   for (int i = 0; i < m_operators.width(); i++) {
      for (int j = 0; j < m_operators.height(); j++) {
          CRExpressionOperator*op = m_operators(i,j).get();
          if (op == NULL) continue;
          assert(m_operator_to_opdata.find(op->m_expressiontype) == m_operator_to_opdata.end());
          m_operator_to_opdata.insert(std::make_pair(op->m_expressiontype, op));


      }
    }
}


void CRProgram::compileTo(CRActionContainer* action, const CUtiString& filename)
{
    uti_refptr<CRNormalExecutionContext> context(new CRNormalExecutionContext());
    std::map<CUtiIDString, CUtiIDString> objects;

    CRFunctionDeclaration* main=NULL;
    for (int i = 0; i < m_command.size(); i++) {
        CRCommand* r = m_command.get(i);
        if (dynamic_cast<CRFunctionDeclaration*>(r) != NULL && check_cast<CRFunctionDeclaration*>(r)->m_param_type_and_name.size() == 0 && check_cast<CRFunctionDeclaration*>(r)->m_name == IDSTR("main")) {
            CRFunctionDeclaration* fd = check_cast<CRFunctionDeclaration*>(r);
            main = fd;
        } else if (dynamic_cast<CRVariable*>(r)) {
            uti_refptr<CRDataCell>data = r->execute(context.get());
            collectObjects(data.get(), objects);
        }

    }

    DynArray1D<CRFunctionDeclaration*> compilefunctions;
    DynArray1D<CRStructDeclaration*> tables;
   for (int i = 0; i < m_command.size(); i++) {
       CRCommand* c = m_command.get(i);
       if (dynamic_cast<CRFunctionCall*>(c)) {
           CRFunctionCall* funccall = check_cast<CRFunctionCall*>(c);
           if (funccall->m_declaration != NULL && funccall->m_declaration->m_name == IDSTR("registerAction")) {
                CRBaseDeclaration* decl = funccall->m_templateparameters.get(0);
                if (dynamic_cast<CRFunctionDeclaration*>(decl)) {
                    compilefunctions.push_back(check_cast<CRFunctionDeclaration*>(decl));

                } else {
                    error(UTISTR("registerAction: ")+decl->m_name.str()+UTISTR(" is not a function."));
                }
           } else if (funccall->m_declaration != NULL && funccall->m_declaration->m_name == IDSTR("registerEvent")) {
               CRBaseDeclaration* decl = funccall->m_templateparameters.get(0);
               if (dynamic_cast<CRFunctionDeclaration*>(decl)) {
                   check_cast<CRFunctionDeclaration*>(decl)->m_actiontype = ratEvent;
                   compilefunctions.push_back(check_cast<CRFunctionDeclaration*>(decl));

               } else {
                   error(UTISTR("registerEvent: ")+decl->m_name.str()+UTISTR(" is not a function."));
               }
          } else if (funccall->m_declaration != NULL && funccall->m_declaration->m_name == IDSTR("registerProcess")) {
               CRBaseDeclaration* decl = funccall->m_templateparameters.get(0);
               if (dynamic_cast<CRFunctionDeclaration*>(decl)) {
                   check_cast<CRFunctionDeclaration*>(decl)->m_actiontype = ratProcess;
                   compilefunctions.push_back(check_cast<CRFunctionDeclaration*>(decl));

               } else {
                   error(UTISTR("registerProcess: ")+decl->m_name.str()+UTISTR(" is not a function."));
               }
          } else if (funccall->m_declaration != NULL && funccall->m_declaration->m_name == IDSTR("createTable")) {
               CRBaseDeclaration* decl = funccall->m_templateparameters.get(0);
               if (dynamic_cast<CRStructDeclaration*>(decl)) {
                   CRStructDeclaration *sd = (CRStructDeclaration*)decl;
                   tables.push_back(sd);
                   m_nametotable.insert(std::make_pair(sd->m_name , sd));
                   CRStructDeclaration* r = check_cast<CRStructDeclaration*>(decl);
                   DynArray1D<CUtiIDString> str;
                   for (auto& p: r->m_children_tan) {str.push_back(p.m_name);}
                   if (action)
                   action->addStruct(r->m_name, str);

               } else {
                   error(UTISTR("createTable: ")+decl->m_name.str()+UTISTR(" is not a struct."));
               }
          }
       }

   }
 std::wcout  << std::endl;
      uti_refptr<CStringList> l(new CStringList());
      l->add(UTISTR("(define (domain ")+ m_name+UTISTR(")"));
      l->add(UTISTR("(:requirements :equality :typing :numeric-fluents :negative-preconditions :conditional-effects)"));
 //  DynArray1D<Lisp::LObject> predicates;
  // predicates.push_back(IDSTR(":predicates"));
  l->add(UTISTR("(:types"));
   for (int i = 0; i < m_command.size(); i++) {
       CRCommand* c = m_command.get(i);
       c->declarePDDLType(l.get());
   }



  l->add(UTISTR(")"));

  CUtiString objectlist;
  for (auto it : objects) {
      if (it.second.empty()) {
          objectlist += UTISTR(" ")+it.first.str();
      } else {
          objectlist += UTISTR(" ")+it.first.str()+UTISTR(" - ")+it.second.str();

      }
  }

  if (objectlist.size() != 0) {
  l->add(UTISTR("(:constants")+objectlist+UTISTR(")"));
  }

  l->add(UTISTR("(:predicates"));
   for (auto t: tables) {
       t->createPDDLDeclaration(l.get(), true);
      // predicates.push_back(t->toLispDeclaration());

   }

  l->add(UTISTR(")"));
  l->add(UTISTR("(:functions"));
   for (auto t: tables) {
       t->createPDDLDeclaration(l.get(), false);
      // predicates.push_back(t->toLispDeclaration());
//(amount ?j - jug)(capacity ?j - jug
   }

  l->add(UTISTR(")"));
//      std::wcout  << Lisp::list(predicates).printAligned() << std::endl;


  CReferenceList<CRAssembler> assm;
   for (int i = 0; i < compilefunctions.size(); i++) {
      CRFunctionDeclaration* funcdecl = compilefunctions[i];

    //  CRActionGenerator* actiongen = action.createAction(funcdecl->m_name);

      uti_refptr<CRAssembler> context(new CRAssembler());
      assm.push_back(context.get());
      //uti_refptr<CRPDDLWhen> mainwhen(new CRPDDLWhen(NULL));
      //uti_refptr<CRPDDLDecision> maindecision(new CRPDDLDecision(mainwhen.get(), NULL));
     // context->m_decisionstack.push(maindecision.get());
     // context->m_actiongenerator = actiongen;
      //mainwhen->m_truedecision = maindecision.get();
      //maindecision->m_istruedecision = true;
     // maindecision->m_actiongenerator = actiongen;
   // uti_refptr< CUndoDocument> document(new CUndoDocument());
  //  CUndoControl::getInstance()->setDocument(document.get());
   // CUndoControl::getInstance()->snapshot();

      //Register Normal Vars
      for (int i = 0; i < m_command.size(); i++) {
          CRCommand* r = m_command.get(i);
          if (dynamic_cast<CRVariable*>(r)) {
              CRVariable* var =check_cast<CRVariable*>(r);
              var->compileDecl(context.get());
          }

      }

      context->pushStack();

      context->add(Lisp::list(IDSTR("asm_init"), IDSTR("PC"),Lisp::list(IDSTR("!!!list"))));
      funcdecl->compile(context.get());
      context->add(Lisp::list(IDSTR("asm_end"), IDSTR("PC")));


      context->popStack();

      for (int i = 0; i < m_command.size(); i++) {
          CRCommand* r = m_command.get(i);
          if (dynamic_cast<CRFunctionDeclaration*>(r) != NULL && r != funcdecl) {
              CRFunctionDeclaration* fd = check_cast<CRFunctionDeclaration*>(r);
              //main = fd;
              fd->compileNonMainFunction(context.get());
          }

      }

     context->linkLabels();
      context->printCode();
      uti_refptr<CRPDDLCompiledAction> action(new CRPDDLCompiledAction(context.get(), this));
      m_compileaction = action.get();
      executePrologEngine(context.get());
      action->generatePDDLAction(funcdecl->m_name.str(), funcdecl->m_actiontype, *l.get(), false);

      m_compileaction = NULL;
    // Lisp::LObject result = convertToAction(funcdecl,context.get(), context->m_pddlmainblock.get(), true);
      //std::wcout  << result.printAligned() << std::endl;

    //CUndoControl::getInstance()->setDocument(NULL);

   }


  l->add(UTISTR(")"));
  std::wcout << "------------- Output: " << std::endl;
      for (int i = 0;i < l->size(); i++) {
         std::wcout << l->getLine(i) << std::endl;
      }
  l->saveToFile(filename);
}

CUtiString CRPDDLEffect::printPDDL(std::set<CUtiIDString> &usedvars)
{
   CUtiString res =UTISTR("(")+ m_name.str();

   for (int i = 0; i < m_runtimevars.size(); i++) {
       res += UTISTR(" ")+m_runtimevars[i].printPDDL(usedvars);
   }
   return res+UTISTR(")");
}


CUtiString CRPDDLEffect::print()
{
   CUtiString res = UTISTR("(")+ m_name.str();

   for (int i = 0; i < m_runtimevars.size(); i++) {
       res += UTISTR(" ")+m_runtimevars[i].print();
   }
   return res+UTISTR(")");
}

void CRPDDLCompiledAction::generatePDDLEffect(int shift, CRPDDLExecutionBlock* block, CStringList& output, bool ownpddl, std::set<CUtiIDString> &usedvars)
{
   CUtiString d;
   for (int i = 0; i < shift; i++) {
      d += UTISTR("   ");
   }
   output.add(d+UTISTR("(and "));
   for (int i = 0; i < block->m_addremovecommand.size(); i++) {
       CRPDDLEffect& ef = block->m_addremovecommand[i];
       CUtiIDString effectname = ef.m_name;
       auto it = m_program->m_nametotable.find(effectname);
       assert(it != m_program->m_nametotable.end());
       CRStructDeclaration* decl = it->second;

       DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> > params;
       decl->createFlatVariableList(params, EMPTYSTRING);
       assert(params.size() == ef.m_runtimevars.size());

      // output.add(d + ef.printPDDL(usedvars));
      {
         CUtiString res =UTISTR("(")+ decl->m_name.str();

         CUtiString paramsres;
         int c = 0;
         for (int i = 0; i < params.size(); i++) {
             if (dynamic_cast<CRClassDeclaration*>(params[i].second) != NULL || m_assembler->m_syntheticpddl) {
            paramsres += UTISTR(" ")+ef.m_runtimevars[i].printPDDL(usedvars);
            c++;
            }
         }
         res = res+paramsres;

         res = res+UTISTR(")");
         if (c != 0) {
             if (ef.m_plstate == ttEndCondition) {
                 if (ef.m_add) {
                     output.add(d+UTISTR("(at end ")+ res+UTISTR(")"));
                 } else {
                     output.add(d+UTISTR("(at end ")+ UTISTR("(not ")+res + UTISTR("))"));
                 }
             } else {
                 if (ef.m_add) {
                     output.add(d+res);
                 } else {
                     output.add(d+UTISTR("(not ")+res + UTISTR(")"));
                 }
             }
         }
         if (!ef.m_add){
         continue;
         }
         if (m_assembler->m_syntheticpddl == false) {
             for (int fluentindex = 0; fluentindex < params.size(); fluentindex++) {
                 if (dynamic_cast<CRClassDeclaration*>(params[fluentindex].second) != NULL) {
                     continue;
                 }
                 output.add(d+UTISTR("(assign (")+decl->m_name.str()+UTISTR("_")+params[fluentindex].first+paramsres+UTISTR(") ")+createNumericExpression(ef.m_runtimevars[fluentindex], usedvars)+ UTISTR(")"));

             }
         }


       }
   }
   for (int i = 0; i < block->m_subblocks.size(); i++) {
      CRPDDLPairBlock* pb = block->m_subblocks.get(i);
      if (pb->m_usagecount == 0) continue;
      CUtiString conditions = createBoolExpression(pb->m_condition, usedvars);
      if (pb->m_true->m_usagecount != 0) {
         output.add(d+UTISTR("(when ")+conditions+UTISTR(" "));

          generatePDDLEffect(shift+1, pb->m_true.get(), output, ownpddl, usedvars);

          output.add(d+UTISTR(")"));
       }
      if (pb->m_false->m_usagecount != 0) {

          output.add(d+UTISTR("(when (not (")+conditions+UTISTR(")) "));
          generatePDDLEffect(shift+1, pb->m_false.get(), output, ownpddl, usedvars);
          output.add(d+UTISTR(")"));
      }
   }
   output.add(d+UTISTR(")"));
}


CUtiString CRPDDLCompiledAction::createBoolExpression(const SRuntimeVariable&var,  std::set<CUtiIDString>& usedvars)
{
   switch (var.m_type) {
    case rvtIntegerValue:
    {

       assert(false);

        return CUtiString::intToStr( var.m_integer_value);
    }; break;
    case rctFloatValue:
    {


       assert(false);
        return CUtiString::floatToStr(var.m_float_value, 0, 6);

    }; break;
    case rctBoolValue:
    {

        assert(false);
    }; break;
    case rctStringValue:
    {
        assert(false);
        return EMPTYSTRING;

    }; break;
    case rctAtomValue:
    {

        assert(false);
        return UTISTR("")+var.m_atom_value.str();
    }; break;
    case rvtIntegerPDDL:
    {
       assert(false);
       // usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
       // return UTISTR("?")+var.m_atom_value.str();
      // return  createNumericExpression(var.m_atom_value,usedvars);
    }; break;
    case rvtFloatPDDL:
    {

       assert(false);
       //return  createNumericExpression(var.m_atom_value,usedvars);
    }; break;
    case rvtBoolPDDL:
    {

      //  assert(false);
       // usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
        return UTISTR("?")+var.m_atom_value.str();
    }; break;
    case rvtAtomPDDL:
    {

        assert(false);
       // usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
        return UTISTR("?")+var.m_atom_value.str();
    }; break;
    default:  {
 assert(false);
    }; break;
    }
    return UTISTR("ERROR");
}

CUtiString CRPDDLCompiledAction::createBoolExpression(const CUtiIDString&variable,  std::set<CUtiIDString>& usedvars)
{
    auto it = m_globalstore.find(variable);
    assert(it != m_globalstore.end());
     CGlobalStoreCommand* com = it->second.get();
     if (dynamic_cast<CRPDDLExists*>(com) != NULL) {
        CRPDDLExists* ex = (CRPDDLExists*)com;
        CUtiString res;
        if (ex->m_effect->m_add == false) {
           res += UTISTR("(not (");
        } else {

           res += UTISTR("(");
        }
        res += ex->m_effect->m_name.str();
        auto it = m_program->m_nametotable.find(ex->m_effect->m_name);
        assert(it != m_program->m_nametotable.end());
        CRStructDeclaration* decl = it->second;

        DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> > params;
        decl->createFlatVariableList(params, EMPTYSTRING);
        assert(params.size() == ex->m_effect->m_runtimevars.size());


        int atomcount = 0;
        for (int i = 0; i < ex->m_effect->m_runtimevars.size(); i++) {
            if (dynamic_cast<CRClassDeclaration*>(params[i].second) == NULL) {
                continue;
            }
            atomcount++;
            SRuntimeVariable rv = ex->m_effect->m_runtimevars[i];
            if (rv.m_type == rctAtomValue) {
               res += UTISTR(" ")+rv.m_atom_value.str();
            } else if (rv.m_type == rvtAtomPDDL) {
               res += UTISTR(" ?")+rv.m_atom_value.str();
               usedvars.insert(CUtiIDString(UTISTR("?")+rv.m_atom_value.str()));
            } else {
               res += UTISTR(" ?????? ");
            }
        }
        if (atomcount == 0)
        {
            return EMPTYSTRING;
        }

        if (ex->m_effect->m_add == false) {
           res += UTISTR("))");
        } else {

           res += UTISTR(")");
        }
        return res;
     } else  if (dynamic_cast<CRPDDLNumericOperation*>(com) != NULL) {
        CRPDDLNumericOperation* num = (CRPDDLNumericOperation*)com;
        if (num->m_type == eLogicalAnd) {
           CUtiString res;
           res += UTISTR("(and ");
           res += createBoolExpression(num->m_op1.m_atom_value, usedvars);
           res += UTISTR(" ");
           res += createBoolExpression(num->m_op2.m_atom_value, usedvars);

           res += UTISTR(")");
           return res;
        } else if (num->m_type == eLogicalOr) {
           CUtiString res;
           res += UTISTR("(or ");
           res += createBoolExpression(num->m_op1.m_atom_value, usedvars);
           res += UTISTR(" ");
           res += createBoolExpression(num->m_op2.m_atom_value, usedvars);

           res += UTISTR(")");
           return res;
        } else if (num->m_type == eLess) {
           CUtiString res;
           res += UTISTR("(< ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        } else if (num->m_type == eGreater) {
           CUtiString res;
           res += UTISTR("(> ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        } else if (num->m_type == eELess) {
           CUtiString res;
           res += UTISTR("(<= ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        } else if (num->m_type == eEGreater) {
           CUtiString res;
           res += UTISTR("(>= ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        }  else if (num->m_type == eEqual) {
           CUtiString res;
           res += UTISTR("(= ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        }else {
        assert(false);
        }

     }
     return EMPTYSTRING;
}


CUtiString CRPDDLCompiledAction::createNumericExpression(const SRuntimeVariable&var,  std::set<CUtiIDString>& usedvars)
{
   switch (var.m_type) {
    case rvtIntegerValue:
    {

        return CUtiString::intToStr( var.m_integer_value);
    }; break;
    case rctFloatValue:
    {

        return CUtiString::floatToStr(var.m_float_value, 0, 6);

    }; break;
    case rctBoolValue:
    {

        assert(false);
    }; break;
    case rctStringValue:
    {
        assert(false);
        return EMPTYSTRING;

    }; break;
    case rctAtomValue:
    {

        assert(false);
        return UTISTR("")+var.m_atom_value.str();
    }; break;
    case rvtIntegerPDDL:
    {
       // usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
       // return UTISTR("?")+var.m_atom_value.str();
       if (m_assembler->m_syntheticpddl == false) {
            return  createNumericExpression(var.m_atom_value,usedvars);
       } else {

           return makePDDLToVar(var.m_atom_value);
       }
    }; break;
    case rvtFloatPDDL:
    {

        if (m_assembler->m_syntheticpddl == false) {
            return  createNumericExpression(var.m_atom_value,usedvars);
        } else {

            return makePDDLToVar(var.m_atom_value);
        }
    }; break;
    case rvtBoolPDDL:
    {

        assert(false);
       // usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
        return UTISTR("?")+var.m_atom_value.str();
    }; break;
    case rvtAtomPDDL:
    {

        assert(false);
       // usedvars.insert(CUtiIDString(UTISTR("?")+var.m_atom_value.str()));
        return UTISTR("?")+var.m_atom_value.str();
    }; break;
    default:  {
 assert(false);
    }; break;
    }
    return UTISTR("ERROR");
}
CUtiString CRPDDLCompiledAction::createNumericExpression(const CUtiIDString&variable,  std::set<CUtiIDString>& usedvars)
{
    auto it = m_globalstore.find(variable);
    assert(it != m_globalstore.end());
     CGlobalStoreCommand* com = it->second.get();
      if (dynamic_cast<CRPDDLNumericOperation*>(com) != NULL) {
        CRPDDLNumericOperation* num = (CRPDDLNumericOperation*)com;
        if (num->m_type == eAdd) {
           CUtiString res;
           res += UTISTR("(+ ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        } else if (num->m_type == eSub) {
           CUtiString res;
           res += UTISTR("(- ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        } else if (num->m_type == eMul) {
           CUtiString res;
           res += UTISTR("(* ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        } else if (num->m_type == eDiv) {
           CUtiString res;
           res += UTISTR("(/ ");
           res += createNumericExpression(num->m_op1, usedvars);
           res += UTISTR(" ");
           res += createNumericExpression(num->m_op2, usedvars);

           res += UTISTR(")");
           return res;
        }
     }else if (dynamic_cast<CRPDDLConstOperation*>(com) != NULL){
          return createNumericExpression(check_cast<CRPDDLConstOperation*>(com)->m_variable, usedvars);
       } else if (dynamic_cast<CRPDDLFluentAccess*>(com) != NULL){
           CRPDDLFluentAccess* num = (CRPDDLFluentAccess*)com;
           CUtiString accessstr;
           accessstr += UTISTR("(");
           accessstr += num->m_predicatename + UTISTR("_")+num->m_list[num->m_accessindex].first;


        int c = 0;
        for (int i = 0; i <  num->m_list.size(); i++) {
            if (dynamic_cast<CRClassDeclaration*>(num->m_list[i].second) != NULL) {
           accessstr += UTISTR(" ")+num->m_runtimevariable[i].printPDDL(usedvars);
           c++;
           }
        }



           accessstr += UTISTR(")");
           return accessstr;

        } else {
       assert(false);
       }

     return EMPTYSTRING;
}

CUtiString CRPDDLCompiledAction::numericOpToSyntheticPDDL(CRPDDLNumericOperation* num)
{
    switch (num->m_type) {
        case eLogicalOr: {

        assert(false);
    }; break;
        case eLogicalAnd: {

        assert(false);
    }; break;
        case eAdd: {
        return UTISTR("(= ")+makePDDLToVar(num->m_res)+ UTISTR(" (+ ") +makePDDLToVar(num->m_op1) + UTISTR(" ")+makePDDLToVar(num->m_op2)+UTISTR("))");

    }; break;
        case eSub: {
        return UTISTR("(= ")+makePDDLToVar(num->m_res)+ UTISTR(" (- ") +makePDDLToVar(num->m_op1) + UTISTR(" ")+makePDDLToVar(num->m_op2)+UTISTR("))");

    }; break;
        case eMul: {

        return UTISTR("(= ")+makePDDLToVar(num->m_res)+ UTISTR(" (* ") +makePDDLToVar(num->m_op1) + UTISTR(" ")+makePDDLToVar(num->m_op2)+UTISTR("))");
    }; break;
        case eDiv: {

        return UTISTR("(= ")+makePDDLToVar(num->m_res)+ UTISTR(" (/ ") +makePDDLToVar(num->m_op1) + UTISTR(" ")+makePDDLToVar(num->m_op2)+UTISTR("))");
    }; break;
        case eLess: {

        assert(false);
    }; break;
        case eGreater: {

        assert(false);
    }; break;
        case eEqual: {

        assert(false);
    }; break;
        case eELess: {

        assert(false);
    }; break;
        case eEGreater: {

        assert(false);
    }; break;
        case eModulo: {

    }; break;
    case eIntDiv: {
        return UTISTR("(= ")+makePDDLToVar(num->m_res)+ UTISTR(" (/ ") +makePDDLToVar(num->m_op1) + UTISTR(" ")+makePDDLToVar(num->m_op2)+UTISTR("))");

    }; break;
    }

    return EMPTYSTRING;
}

void CRPDDLCompiledAction::generatePDDLAction(const CUtiString &name, TRActionType m_actiontype, CStringList& output, bool ownpddl)
{
   //eigentlich brauchen wir nur parameter
    while (m_block->tryToMerge()) {

    }
    m_block->calculateUsageCount();
    if (m_actiontype == ratEvent) {
        if (m_duration.size() == 0) {
        output.add(UTISTR("(:event ")+name);
        }

    }else if (m_actiontype ==ratAction) {
        if (m_duration.size() == 0) {
            output.add(UTISTR("(:action ")+name);
        } else {

            output.add(UTISTR("(:durative-action ")+name);
        }
    }else if (m_actiontype == ratProcess) {
        output.add(UTISTR("(:process ")+name);

    }
   CUtiString params;
   std::set<CUtiIDString> usedvariables;
  for (auto p : m_parameters) {
     params += p.first.str() + UTISTR(" - ") + p.second.str();
  }
  if (m_assembler->m_syntheticpddl) {
      //Collect new variables+types
      for (auto it : m_globalstore) {
          CRPDDLNumericOperation* num = dynamic_cast<CRPDDLNumericOperation*>(it.second.get());
          if (!num) continue;
          params +=  makePDDLToVar(num->m_res) + UTISTR(" - float");

      }

  }
  int paramindex = output.size();
   output.add(UTISTR("   :parameters (") + params+ UTISTR(")"));
   output.add(UTISTR("   :precondition (and "));
   if (m_assembler->m_syntheticpddl) {
       //Collect new variables+types
       for (auto it : m_globalstore) {
           CRPDDLNumericOperation* num = dynamic_cast<CRPDDLNumericOperation*>(it.second.get());
           if (!num) continue;
           output.add(UTISTR("       ")+numericOpToSyntheticPDDL(num));
       }

   }

   for (auto p : m_preconditions) {

       if (p.second == ttStartCondition) {
           output.add(UTISTR("       ")+createBoolExpression(p.first.m_atom_value, usedvariables));
       } else if (p.second == ttEndCondition) {
           output.add(UTISTR("       (at end ")+createBoolExpression(p.first.m_atom_value, usedvariables)+UTISTR(")"));
       } else if (p.second == ttOverAllCondition) {
           output.add(UTISTR("       (over all")+createBoolExpression(p.first.m_atom_value, usedvariables)+UTISTR(")"));
       }
   }


   output.add(UTISTR("   )"));

   output.add(UTISTR("   :effect "));

   generatePDDLEffect(2, m_block.get(), output, ownpddl, usedvariables);
   output.add(UTISTR("   "));
   output.add(UTISTR(")"));

 for (auto p : usedvariables) {
     CUtiString n = p.str();
     n.Delete(0,1);
     auto it = m_assembler->m_tempvariables.find(CUtiIDString(n));
     if (it == m_assembler->m_tempvariables.end()) {
     params += p.str() + UTISTR(" ");
     } else {

     params += p.str() + UTISTR(" - ")+it->second + UTISTR(" ");
     }
   //  params += p.first.str() + UTISTR(" - ") + p.second.str();
  }

   output.setLine(paramindex, UTISTR("   :parameters (") + params+ UTISTR(")"));
}

FP_term CRProgram::convertVarContentToProlog(const Lisp::LObject& obj)
{

    if (obj.getType() == Lisp::tctInteger) {
        return FP_create_integer(obj.toInt());
    } else if (obj.isNumber()) {
        return FP_create_real(obj.toFloat());
    } else if (obj.isAtom()) {
        return FP_create_atom(FP_atom_from_wstring(obj.getAtom().str().c_str()));
    } else if (obj.isBool()) {
        return FP_create_integer(obj.toBool() ? 1 : 0);
    }else if (obj.isConList()) {
    int length = Lisp::length(obj);
    CUtiIDString atom = Lisp::nth(obj,0).getAtom();
    if (atom == IDSTR("!!!list")) {
       DynArray1D<FP_term> elements;
       for (int i = 1; i < length; i++) {
          elements.push_back(convertVarContentToProlog(Lisp::nth(obj, i)));
       }
       return FP_create_list_from_array(elements.size(), elements.getCArray());
    }

    switch (length) {
    case 2: {
        return FP_create_structure_1(FP_atom_from_wstring(atom.str().c_str()), convertVarContentToProlog(Lisp::nth(obj, 1)));

    };break;
    case 3: {
        return FP_create_structure_2(FP_atom_from_wstring(atom.str().c_str()), convertVarContentToProlog(Lisp::nth(obj, 1)), convertVarContentToProlog(Lisp::nth(obj,2)));

    };break;
    case 4: {
        return FP_create_structure_3(FP_atom_from_wstring(atom.str().c_str()), convertVarContentToProlog(Lisp::nth(obj, 1)), convertVarContentToProlog(Lisp::nth(obj, 2)), convertVarContentToProlog(Lisp::nth(obj, 3)));

    };break;
    case 5: {
        return FP_create_structure_4(FP_atom_from_wstring(atom.str().c_str()), convertVarContentToProlog(Lisp::nth(obj, 1)), convertVarContentToProlog(Lisp::nth(obj, 2)), convertVarContentToProlog(Lisp::nth(obj, 3)), convertVarContentToProlog(Lisp::nth(obj, 4)));

    };break;

    default: {

    }; break;
    }
    }
    assert(false);
    return 0;
}

void CRProgram::addCommandToProlog(const Lisp::LObject & obj)
{
    CUtiIDString atom = Lisp::nth(obj,0).getAtom();
    int l = Lisp::length(obj);

  // if (atom == IDSTR("asm_declparam")) {
       FP_DATABASE db = FP_dyn_getDatabase(FP_atom_from_wstring(atom.str().c_str()), l-1);
       assert(db != NULL);
       FP_term t = convertVarContentToProlog(obj);
               /*FP_create_structure_2(
             FP_atom_from_wstring(L"asm_declparam"),
              FP_create_integer(Lisp::nth(obj,1).toInt()),
              FP_atom_from_wstring(Lisp::nth(obj,2).getAtom().str().c_str())
           );*/
       FP_dyn_addlast(db, t);
  /* }
   if (atom == IDSTR("asm_setvariable")) {
       FP_DATABASE db = FP_dyn_getDatabase(FP_atom_from_string("asm_setvariable"), 3);
       FP_term t = FP_create_structure_3(
             FP_atom_from_wstring(L"asm_setvariable"),
              FP_create_integer(Lisp::nth(obj,1).toInt()),
              FP_atom_from_wstring(Lisp::nth(obj,2).getAtom().str().c_str()),
              convertVarContentToProlog(Lisp::nth(obj,3))
           );
       FP_dyn_addlast(db, t);
   }*/
}

void CRProgram::executePrologEngine(CRAssembler* assembler)
{

    pl_Initialize();
       pl_initModule_executecpp();
       pl_initengine();
       //Mit Backtracking ueber Loesungen
       for (int i = 0; i < assembler->m_commands.size(); i++)
       {
           Lisp::LObject obj = assembler->m_commands[i].second;
           addCommandToProlog(obj);

       }

      /* HP_term t = HP_create_structure_3(
             HP_atom_from_wstring(L"testterm"),
              HP_create_integer(5),
              HP_create_char(L'a'),
              HP_create_atom(HP_atom_from_wstring(L"atom"))
           );*/
       plr_startprogram();
       FP_dyn_clearAll();
       pl_Free();
}

Lisp::LObject CRProgram::convertToAction( CRFunctionDeclaration* funcdecl, CRNormalExecutionContext* context, CRPDDLExecutionBlock* block, bool maindata)
{
   /* if (maindata) {
        DynArray1D<Lisp::LObject> parameters;
        parameters.push_back(IDSTR(":parameters"));

        for (auto& p: context->m_tempvariables) {
         //  auto&p = funcdecl->m_param_type_and_name[j];
          //CRBaseDeclaration* decl = std::get<0>(p).get();
          //CUtiIDString name = std::get<1>(p);
            if (context->m_usedpddlvariables.find(p.first) == context->m_usedpddlvariables.end()) {
                continue;
            }
            parameters.push_back(Lisp::list(IDSTR("-"), p.first, p.second));
        }

        DynArray1D<Lisp::LObject> preconditions;
        preconditions.push_back(IDSTR(":preconditions"));
        for (auto& d: block->m_condition) {
           preconditions.push_back(d);
        }


        for (auto& d : context->m_numericaloperation) {
            CUtiIDString result =d.second;
            Lisp::LObject op1 = std::get<0>(d.first);
            CUtiIDString type = std::get<1>(d.first);
            Lisp::LObject op2 = std::get<2>(d.first);
            preconditions.push_back(Lisp::list(IDSTR("="), result, Lisp::list(type, op1, op2)));

        }

        DynArray1D<Lisp::LObject> effect;
        effect.push_back(IDSTR(":effect"));
        for (auto &p : block->m_addremovecommand) {
            effect.push_back(p);
        }
        for (auto &sub : block->m_subblocks) {
            effect.push_back(convertToAction(funcdecl, context, &sub, false));
        }
        return Lisp::list(IDSTR(":action"),funcdecl->m_name, Lisp::list(parameters), Lisp::list(preconditions), Lisp::list(effect));
    } else {
        Lisp::LObject condition;
        if (block->m_condition.size() == 1) {
            condition = block->m_condition.at(0);
        } else  {

            DynArray1D<Lisp::LObject> condlist;
            condlist.push_back(IDSTR("and"));
            condlist.push_back(block->m_condition);
            condition = Lisp::list(condition);
        }
        DynArray1D<Lisp::LObject> whenresult;
        whenresult.push_back(IDSTR("when"));
        whenresult.push_back(condition);
        for (auto &p : block->m_addremovecommand) {
            whenresult.push_back(p);
        }
        for (auto &sub : block->m_subblocks) {
            whenresult.push_back(convertToAction(funcdecl, context, &sub, false));
        }


        return Lisp::list(whenresult);

    }*/
    return Lisp::LObject();
}

void CRProgram::loadFromFile(const CUtiString& name)
{
  m_name = name;
  m_name.ExtractRealFileName();
    CUtiString document;
  document.LoadFromFile(name);
  UtiLexer lexer(document);
  lexer.m_pascalcomment = false;
  doParse(lexer);

}

void CRProgram::loadFromString(const CUtiString& filename, const std::string& name)
{
  m_name = filename;
  m_name.ExtractRealFileName();
    CUtiString document;
    document = UtiFramework::converttowide(name.c_str());
  UtiLexer lexer(document);
  lexer.m_pascalcomment = false;
  doParse(lexer);

}

CUtiIDString CRProgram::getUniqueVarName()
{
   CUtiIDString t(UTISTR("§temp")+m_tempindex);
   m_tempindex++;

   return t;
}

/*CRDuplicateEntry::CRDuplicateEntry()
{
    m_entrytype = etParam;
    m_paramtype = ratObject;
    m_conditiontype = ctEqual;
    m_operatortype = otAdd;
    m_positive = true;
}


CRActionGenerator::CRActionGenerator()
{

}

CRDuplicateGenerator::CRDuplicateGenerator()
{

}

void CRDuplicateGenerator::addParam(const CUtiIDString& str, TRType type)
{
    CRDuplicateEntry* entry = new CRDuplicateEntry();
    entry->m_entrytype = etParam;
    entry->m_name_or_target = str;
    entry->m_paramtype = type;
    m_commands.push_back(entry);

}
void CRDuplicateGenerator::addDatabaseCheck(const CUtiIDString& name,bool positive, const DynArray1D<Lisp::LObject>& params)
{
    CRDuplicateEntry* entry = new CRDuplicateEntry();
    entry->m_entrytype = etDatabaseCheck;
    entry->m_name_or_target = name;
    entry->m_positive = positive;
    entry->m_params  = params;
    m_commands.push_back(entry);
}
void CRDuplicateGenerator::addDatabaseChange(const CUtiIDString& name, bool add, const DynArray1D<Lisp::LObject> &params)
{
    CRDuplicateEntry* entry = new CRDuplicateEntry();
    entry->m_entrytype = etDatabaseChange;
    entry->m_name_or_target = name;
    entry->m_positive = add;
    entry->m_params  = params;
    m_commands.push_back(entry);
}
void CRDuplicateGenerator::addFluentCondition(const CUtiIDString& var1, TRConditionType type, const CUtiIDString& var2)
{
    CRDuplicateEntry* entry = new CRDuplicateEntry();
    entry->m_entrytype = etFluentCondition;
    entry->m_var1 = var2;
    entry->m_var2 = var2;
    entry->m_conditiontype = type;
    m_commands.push_back(entry);
}
void CRDuplicateGenerator::setFluentVar(const CUtiIDString& target, const CUtiIDString& var1,  TROperatorType type, const CUtiIDString& var2)
{
    CRDuplicateEntry* entry = new CRDuplicateEntry();
    entry->m_entrytype = etSetFluentVar;
    entry->m_name_or_target = target;
    entry->m_var1 = var1;
    entry->m_var2 = var2;
    entry->m_operatortype = type;
    m_commands.push_back(entry);
}
CRActionGenerator* CRDuplicateGenerator::createOn(CRActionContainer* container, const CUtiIDString& name)
{
   CRActionGenerator* action = container->createAction(name);
   for (int i = 0; i < m_commands.size(); i++) {
       CRDuplicateEntry& entry = m_commands[i];
       switch (entry.m_entrytype) {
          case etParam: {
           action->addParam(entry.m_name_or_target, entry.m_paramtype);
       }; break;
           case etDatabaseCheck: {
           action->addDatabaseCheck(entry.m_name_or_target,entry.m_positive, entry.m_params);

       }; break;
           case etDatabaseChange: {
           action->addDatabaseChange(entry.m_name_or_target,entry.m_positive, entry.m_params);

       }; break;
           case etFluentCondition: {
           action->addFluentCondition(entry.m_var1, entry.m_conditiontype, entry.m_var2);

       }; break;
           case etSetFluentVar: {
           action->setFluentVar(entry.m_name_or_target, entry.m_var1, entry.m_operatortype, entry.m_var2);

       }; break;
          default:  {

       }; break;

       }
   }

   return action;
}
*/
