/***************************************************************************
 *   Copyright (C) 2018 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef SYNTHETICRULE_H
#define SYNTHETICRULE_H

//Wir machen einen allgemeinen C Parser, um dann eventuell den Quellcode nach Lua zu konvertieren

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "UtiString.h"
#include "utilists.h"
#include "utiarray.h"
#include "utiadvarray.h"
#include "utilexer.h"
#include <tuple>
#include <iostream>
#include <stack>
#include "utilisp.h"
#include "frogprolog.h"
#include "pddllib.h"
//#include "undosystem.h"
//#include "undosystemlist.h"

class CRBlock;
class CRProgram;
class CRDataCell;
class CRAssembler;
struct SRuntimeVariable;
enum TRBasicDataType {
    erInt, erFloat, erObject, erBool, erComplex, erString, erVoid, erVariable
};

enum TRExpressionType {eLogicalOr, eLogicalAnd, eAdd, eSub, eMul, eDiv, eLess, eGreater, eEqual, eELess, eEGreater, eModulo, eIntDiv};


enum TRActionType {ratAction, ratEvent, ratProcess};
enum TPreconditionType {ttStartCondition, ttEndCondition, ttOverAllCondition};
class RuleException : public Exception
{
public:
    CIVector2 m_object;
    CUtiString m_data;
    RuleException(const CIVector2& pos, const CUtiString& data)
    {
        m_data = data;
        m_object = pos;
        std::wcout << m_data << " in " << convertToString(pos) << std::endl;

    }
};

class CRBaseDeclaration;
class CRFunctionDeclaration;
class CRActionGenerator;
class CRPDDLDecision;
class CRNormalExecutionContext;

const int FLATMODE_CREATETEMPVARS=1;
const int FLATMODE_COLLECTYPES=2;

class CRDataCell : public ReferenceUtiOB
{
public:
    CRBaseDeclaration* m_type;
    CRDataCell(){}
    virtual CRDataCell* createRCopy(){return NULL;};
    virtual int toInt(){assert(false);return 0;}
    virtual bool toBool(){assert(false);return false;}
    virtual bool isBool(){return false;}
    virtual bool isPDDLVariable(){return false;}
    virtual CUtiString printValue() {return EMPTYSTRING;};
    virtual void copyFrom(CRDataCell* cell) {}
    virtual void copyFromCompile(CRAssembler* assembler, CRDataCell* cell)  {assert(false);}
    virtual Lisp::LObject toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj) {assert(false);};
    virtual void declPDDLVariable(CRNormalExecutionContext* context){}
    virtual CRDataCell * replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param) {assert(false);return this;}
    virtual void replaceVarWithAtom(CRNormalExecutionContext* context, ReferenceUtiOB* data, const CUtiString& basename){};
    virtual void replaceVarWithAtomAss(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename){};
    virtual void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int mode, ReferenceUtiOB* obj) {list.push_back(toLisp(NULL, mode, obj));}
    virtual void compilePrint(CRAssembler* context) {}
};

class CRIntCell : public CRDataCell
{
public:
    int data;
    CRIntCell(){data = 0;}
    CRIntCell(int b){data = b;}
    virtual int toInt(){return data;}
    virtual CRDataCell* createRCopy(){ CRIntCell* c  = new CRIntCell();c->m_type = m_type; c->data = data; return c;}
    virtual CUtiString printValue() {CUtiString str; str += data; return str;}
    virtual void copyFrom(CRDataCell* cell) {data = check_cast<CRIntCell*>(cell)->data;}

    virtual Lisp::LObject toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj);
    virtual CRDataCell * replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param);
};

class CRFloatCell : public CRDataCell
{
public:
    double data;
    CRFloatCell(){data = 0;}
    CRFloatCell(double b){data = b;}
    virtual CRDataCell* createRCopy(){ CRFloatCell* c  = new CRFloatCell();c->m_type = m_type;  c->data = data; return c;}
    virtual CUtiString printValue() {CUtiString str; str += data; return str;}
    virtual void copyFrom(CRDataCell* cell) {data = check_cast<CRFloatCell*>(cell)->data;}
    virtual Lisp::LObject toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj);
    virtual CRDataCell * replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param);
};


class CRBoolCell : public CRDataCell
{
public:
    bool data;
    CRBoolCell(){data = false;}
    CRBoolCell(bool b){data = b;}
    virtual bool toBool(){return data;}
    virtual bool isBool(){return true;}
    virtual CRDataCell* createRCopy(){ CRBoolCell* c  = new CRBoolCell();c->m_type = m_type;  c->data = data; return c;}
    virtual CUtiString printValue() {CUtiString str; str += data; return str;}
    virtual void copyFrom(CRDataCell* cell) {data = check_cast<CRBoolCell*>(cell)->data;}
    virtual CRDataCell * replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param){return this;};
    virtual Lisp::LObject toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj);
};
/*
class CRBoolExpression : public CRDataCell
{
public:
    TRExpressionType m_exptype;
    uti_refptr<CRDataCell> m_arg1;
    uti_refptr<CRDataCell> m_arg2;
    CRBoolExpression(){m_exptype = eLess;};
    CRBoolExpression(TRExpressionType ex, CRDataCell* a1, CRDataCell* a2) {m_exptype = ex;m_arg1 = a1; m_arg2 = a2;}
    virtual CRDataCell* createRCopy()
    { CRBoolExpression* c  = new CRBoolExpression();
        c->m_arg1 = m_arg1;
        c->m_arg2 =  m_arg2;
        return c;
    }
    virtual Lisp::LObject toLisp(CRPDDLDecision* decision);
    virtual CUtiString printValue() {CUtiString str; str += UTISTR("(")+m_arg1->printValue()+UTISTR(", ")+m_arg2->printValue()+UTISTR(")"); return str;}
    virtual void copyFrom(CRDataCell* cell) {
        m_arg1 = check_cast<CRBoolExpression*>(cell)->m_arg1;
        m_arg2= check_cast<CRBoolExpression*>(cell)->m_arg2;
    }
    virtual CRDataCell * replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename){return this;};
};*/

class CRStringCell : public CRDataCell
{
public:
    CUtiString data;
    CRStringCell(){data = EMPTYSTRING;}
    CRStringCell(const CUtiString& b){data = b;}
    virtual CRDataCell* createRCopy(){ CRStringCell* c  = new CRStringCell();c->m_type = m_type;  c->data = data; return c;}
    virtual CUtiString printValue() {CUtiString str; str += data; return str;}
    virtual void copyFrom(CRDataCell* cell) {data = check_cast<CRStringCell*>(cell)->data;}
    virtual CRDataCell * replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param);
};
class CRArrayCell : public CRDataCell
{
public:
    Array<uti_refptr<CRDataCell> >m_cells;
    CRArrayCell(){}
    virtual void copyFrom(CRDataCell* cell) {m_cells = check_cast<CRArrayCell*>(cell)->m_cells;}
    void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int mode, ReferenceUtiOB* obj)
    {
        for (auto &p: m_cells) {
            p->createFlatPDDLList(list, mode, obj);
        }
    }
    virtual void copyFromCompile(CRAssembler* assembler, CRDataCell* cell)
    {
        CRArrayCell* t = check_cast<CRArrayCell*>(cell);
        assert(t->m_type == m_type);
        assert(m_cells.size() == t->m_cells.size());
        for (int i = 0; i < m_cells.size(); i++) {
            m_cells[i]->copyFromCompile(assembler, t->m_cells[i].get());
        }
    }


    virtual CRDataCell* createRCopy()
    {
        CRArrayCell* c  = new CRArrayCell();
        c->m_cells.setSize(m_cells.size());c->m_type = m_type;
        for (int i = 0; i< m_cells.size(); i++) {
            c->m_cells[i] = m_cells[i]->createRCopy();
        }
        return c;
    }
    virtual CUtiString printValue() {
        CUtiString str;
        str += UTISTR("[");
        for (int i = 0; i < m_cells.size(); i++) {
            str += m_cells[i]->printValue();
            if (i != m_cells.size()-1) {
                str += UTISTR(", ");
            }
        }
        str += UTISTR("]");
        return str;
    }
    virtual CRDataCell * replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param);

    virtual void replaceVarWithAtom(CRNormalExecutionContext* context, ReferenceUtiOB* data, const CUtiString& basename);
    virtual void replaceVarWithAtomAss(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename);
};
/*
class CRCheckExistanceCell : public CRDataCell
{
public:
    bool m_not;
    uti_refptr<CRDataCell> m_originalcell;

    Array<uti_refptr<CRDataCell> >m_cells;
    CRCheckExistanceCell(){m_not = false;}
    virtual void copyFrom(CRDataCell* cell) {m_cells = check_cast<CRArrayCell*>(cell)->m_cells;}
    virtual CRDataCell* createRCopy()
    {
        CRCheckExistanceCell* c  = new CRCheckExistanceCell();
        c->m_cells.setSize(m_cells.size());c->m_type = m_type;
        for (int i = 0; i< m_cells.size(); i++) {
            c->m_cells[i] = m_cells[i]->createRCopy();
        }
        return c;
    }
    virtual CUtiString printValue() {
        CUtiString str;
        str += UTISTR("[");
        for (int i = 0; i < m_cells.size(); i++) {
           str += m_cells[i]->printValue();
           if (i != m_cells.size()-1) {
               str += UTISTR(", ");
           }
        }
        str += UTISTR("]");
        return str;
    }

    virtual Lisp::LObject toLisp(CRPDDLDecision* decision) ;
};*/


class CRRecordCell : public CRDataCell
{
public:
    // std::map<CUtiIDString, uti_refptr<CRDataCell> >m_cells;
    CReferenceList<CRDataCell> m_subcells;
    CRRecordCell(){}
    ~CRRecordCell();
    virtual void copyFrom(CRDataCell* cell) {m_subcells = check_cast<CRRecordCell*>(cell)->m_subcells;}
    virtual CRDataCell* createRCopy();
    virtual CUtiString printValue();
    void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int mode, ReferenceUtiOB* obj)
    {
        for (int i = 0; i < m_subcells.size(); i++) {
            CRDataCell& cell = m_subcells[i];
            cell.createFlatPDDLList(list, mode, obj);
        }
    }
    virtual void copyFromCompile(CRAssembler* assembler, CRDataCell* cell)
    {
        CRRecordCell* t = check_cast<CRRecordCell*>(cell);
        assert(t->m_type == m_type);assert(false);
        assert(m_subcells.size() == t->m_subcells.size());
        for (int i = 0; i < m_subcells.size(); i++) {
            m_subcells[i].copyFromCompile(assembler, &t->m_subcells[i]);
        }
    }
    virtual CRDataCell *replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param);
    virtual void replaceVarWithAtom(CRNormalExecutionContext* context, ReferenceUtiOB* data, const CUtiString& basename);
    virtual void replaceVarWithAtomAss(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename);
};

class CRClassCell : public CRDataCell
{
public:
    // std::map<CUtiIDString, uti_refptr<CRDataCell> >m_cells;
    CUtiIDString m_objectname;//Sollte gleich wie Variable sein
    CRClassCell()
    {

    }
    virtual CRDataCell* createRCopy();
    virtual CUtiString printValue();
    virtual Lisp::LObject toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj);

    virtual CRDataCell *replaceVarWithPDDL(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename, bool param);
    virtual void replaceVarWithAtom(CRNormalExecutionContext* context, ReferenceUtiOB* data, const CUtiString& basename)
    {
        m_objectname = CUtiIDString(basename);
    }
    virtual void replaceVarWithAtomAss(CRAssembler* context, ReferenceUtiOB* data, const CUtiString& basename)
    {
        m_objectname = CUtiIDString(basename);
    }
};

/*
class CRPDDLVarCell : public CRDataCell
{
public:
    CUtiIDString m_string;
};
*/

class CRVariableEntry : public ReferenceUtiOB
{
public:
    CUtiIDString m_name;
    uti_refptr<CRDataCell> m_data;
    CRVariableEntry();
    CRVariableEntry(const CUtiIDString& name, CRDataCell* cell)
    {
        m_name = name;
        m_data = cell;
    }
};

class CRStackElement : public ReferenceUtiOB
{
public:
    //  std::map<CUtiIDString, uti_refptr<CRDataCell> > m_variables;
    CReferenceList<CRVariableEntry> m_variables;
    //TempCells?
    //CRDataCell* reserveTempCell(CRBaseDeclaration* decl);
    //void freeTempCell(CRDataCell* cell);
    CRVariableEntry* findVariable(const CUtiIDString& str)
    {
        for (int i = 0; i < m_variables.size(); i++) {
            if (m_variables[i].m_name == str) {
                return &m_variables[i];
            }
        }
        return NULL;
    }

    CRFunctionDeclaration* m_declaration;
    CRStackElement();

};
/*

struct SPDDLExecutionPath
{
    int m_pathswitchid;
    bool m_truevalue;
};

class CPDDLDefinition : public ReferenceUtiOB
{
public:
   DynArray1D<SPDDLExecutionPath> m_dependendblocks;
   uti_refptr<CRDataCell> m_content;
};
*/
/*
class CRExecutionPath : public ReferenceUtiOB
{
public:
    int m_pathid;
    CRBaseDeclaration* m_ifdeclaration;
    uti_refptr<CRActionGenerator> m_actionpositiv;
    uti_refptr<CRActionGenerator> m_actionnegativ;
    CReferenceList<CRExecutionPath> m_whencondition;
    CRExecutionPath()
    {
        m_pathid = -1;
        m_ifdeclaration = NULL;
    }

};*/
/*
class CRPDDLExecutionContext : public ReferenceUtiOB
{
   uti_refptr<CRExecutionPath> m_mainpath;
   std::map<CUtiIDString, uti_refptr<CPDDLVariable> > m_variables;
};
*/

class CRPDDLDecision;
/*
class CRPDDLWhen : public ReferenceUtiOB
{
public:
    CRPDDLWhen* m_parent;
    uti_refptr<CRPDDLDecision> m_truedecision;
    uti_refptr<CRPDDLDecision> m_falsedecision;
    CReferenceList<CRPDDLWhen> m_subconditions;
    CRPDDLWhen(CRPDDLWhen* parent) {m_parent = parent;}
};
*/

class CRPDDLDecision : public ReferenceUtiOB
{
public:
    // uti_refptr<CRActionGenerator> m_actiongenerator;
    //bool m_istruedecision;
    // CRPDDLWhen* m_when;
    // CRPDDLDecision* m_codeparent;
    bool m_createboolresult;
    CRPDDLDecision();
    // CRPDDLDecision(CRPDDLWhen* when, CRPDDLDecision* parent);
};
/*
class CRPDDLPath : public ReferenceUtiOB
{
public:
    CReferenceList<CRPDDLDecision> m_decisions;
};*/


// enum TContentType {ctPDDLVariable, ctPDDLFixedCell, ctPDDLAtom, ctPDDLBoolExpression};
/*class CRPDDLVariableValue : public ReferenceUtiOB
{
public:
    uti_refptr<CRPDDLPath> m_currentpath;
   // uti_refptr<CRPDDLDecision> m_currentdecision;

  //  Lisp::LObject m_boolexpression; //('<' a, 0), (exists name value value) (and a b) (or (a b)
    CRPDDLVariableValue();
    CRPDDLVariableValue* createVarCopy();
};*/

class CRPDDLVariable : public CRDataCell
{
public:
    // CReferenceList<CRPDDLVariableValue> m_values;
    /* CRPDDLVariableValue* getValueByDecision(CRPDDLDecision* ret);
    void prepareNewValue(CRPDDLWhen *decision);
    void setValue_cell(CRPDDLDecision* decision, CRDataCell* value);
    void setValue_atom(CRPDDLDecision* decision, const CUtiIDString& atom);
    void setValue_var(CRPDDLDecision* decision, const CUtiIDString& variable);*/
    uti_refptr<CRBaseDeclaration> m_owntype;
    CRPDDLVariable(){}
    CRPDDLVariable(const CUtiIDString& n) {m_name = n;}
    CRPDDLVariable(const CUtiIDString& n, TRBasicDataType type);
    //   TContentType m_contenttype;
    CUtiIDString m_name; //pddlvariable and atom
    //  CWatchedReference<CRDataCell> m_data;

    virtual bool isPDDLVariable(){return true;}
    virtual Lisp::LObject toLisp(CRPDDLDecision* decision, int mode, ReferenceUtiOB* obj)
    {
        //  CRPDDLVariableValue* value = getValueByDecision(decision);
        // if (m_contenttype == ctPDDLFixedCell) {
        //return m_data->toLisp(decision);
        //}
        assert(m_name.empty() == false);
        return Lisp::LObject(m_name);
    }
    virtual void declPDDLVariable(CRNormalExecutionContext* context);
    virtual void copyFromCompile(CRAssembler *assembler, CRDataCell* cell);

    virtual CRDataCell* createRCopy(){ CRPDDLVariable* c  = new CRPDDLVariable();
                                       // c->m_contenttype = m_contenttype;
                                       // c->m_data = m_data;
                                       c->m_name = m_name;
                                                                      return c;}
};

class CRPDDLExecutionBlock;

class CRPDDLPairBlock : public ReferenceUtiOB
{
public:
    CUtiIDString m_condition;
   uti_refptr< CRPDDLExecutionBlock> m_true;
   uti_refptr< CRPDDLExecutionBlock> m_false;

   int m_usagecount;
   CRPDDLPairBlock();
   int calculateUsageCount();

};

class CRPDDLEffect : public ReferenceUtiOB
{
public:
    CUtiIDString m_name;
    bool m_add;
    TPreconditionType m_plstate;
    DynArray1D<SRuntimeVariable> m_runtimevars;
    CRPDDLEffect()
    {
    m_add = true;
    m_plstate = ttStartCondition;
    }

    bool isEqualToEffect(CRPDDLEffect* effect)
    {
        if (m_name != effect->m_name) {
            return false;
        }
        if (m_plstate != effect->m_plstate) {
            return false;
        }
        if (m_add != effect->m_add) {
            return false;
        }
        assert(m_runtimevars.size() == effect->m_runtimevars.size());

        for (int i = 0; i < m_runtimevars.size(); i++) {
            if (!(m_runtimevars[i] == effect->m_runtimevars[i])) {
                return false;
            }
        }
        return true;
    }
    CUtiString print();
    CUtiString printPDDL(std::set<CUtiIDString> &usedvars);
};

class CRPDDLExecutionBlock : public ReferenceUtiOB
{
public:
    int m_usagecount;
    CRPDDLExecutionBlock* m_parent;
    //Wir sammeln fur jeden Code rekursive alle executionconditions
  //  DynArray1D<Lisp::LObject> m_fullexecutioncondition;
    CReferenceList<CRPDDLEffect> m_addremovecommand;
    CReferenceList<CRPDDLPairBlock > m_subblocks;

    CRPDDLExecutionBlock()
    {
    m_parent = NULL;
    m_usagecount = 0;

    }
    bool tryToMerge();
    int calculateUsageCount();

    virtual bool equals(CUtiOB *other);


};

class CGlobalStoreCommand : public ReferenceUtiOB
{
public:

};

class CRPDDLExists:  public CGlobalStoreCommand
{
public:
   uti_refptr<CRPDDLEffect> m_effect;
};

class CRPDDLNumericOperation : public CGlobalStoreCommand
{
public:
    SRuntimeVariable m_op1;
    TRExpressionType m_type;
    SRuntimeVariable m_op2;
    SRuntimeVariable m_res;
};

class CRPDDLConstOperation : public CGlobalStoreCommand
{
public:
   SRuntimeVariable m_variable;
};

class CRStructDeclaration;

class CRPDDLFluentAccess : public CGlobalStoreCommand
{
public:
    CUtiString m_predicatename;
    int m_accessindex;
    uti_refptr<CRStructDeclaration> m_struct;
    DynArray1D<SRuntimeVariable> m_runtimevariable;//Full Vars
    DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> > m_list;

};

inline CUtiString makePDDLToVar(const CUtiIDString& input)
{
    if (input.isLispVariable()) {
        return input.str();
    }
    CUtiString str = input.str();
    return UTISTR("?")+str;
}

inline CUtiString makePDDLToVar(const SRuntimeVariable& input)
{
    switch (input.m_type) {
    case rvtIntegerValue: {
        return input.m_integer_value;
    }; break;
    case rctFloatValue: {

        return input.m_float_value;
    }; break;
    case rctBoolValue: {
        return input.m_bool_value ? UTISTR("true") : UTISTR("false");
    }; break;
    case rctStringValue: {
        return UTISTR("\"")+ input.m_atom_value.str()+UTISTR("\"");

    }; break;
    case rctAtomValue:
    {
        assert(input.m_atom_value.isLispVariable() == false);
        return input.m_atom_value.str();

    }; break;
    case rvtIntegerPDDL: {
        return makePDDLToVar(input.m_atom_value);

    }; break;
    case rvtFloatPDDL: {
        return makePDDLToVar(input.m_atom_value);

    }; break;
    case rvtBoolPDDL: {
        return makePDDLToVar(input.m_atom_value);

    }; break;
    case rvtAtomPDDL: {
        return makePDDLToVar(input.m_atom_value);

    }; break;

    }
    assert(false);
    return EMPTYSTRING;
}


class CRPDDLCompiledAction : public ReferenceUtiOB
{
public:
   std::map<CUtiIDString, uti_refptr<CGlobalStoreCommand> > m_globalstore;
   uti_refptr<CRPDDLExecutionBlock> m_block;
   uti_refptr<CRPDDLPairBlock> m_pairblock;
   std::stack<std::pair<CRPDDLPairBlock*, bool> > m_stack;
   DynArray1D<std::pair<SRuntimeVariable, TPreconditionType> > m_preconditions;
   std::vector<std::pair<CUtiIDString, CUtiIDString> > m_parameters;
   CRAssembler* m_assembler;
   CRProgram* m_program;
   DynArray1D<CUtiIDString> m_duration;
   CRPDDLCompiledAction(CRAssembler* a, CRProgram* program)
   {
      m_block = new CRPDDLExecutionBlock();
      m_pairblock = new CRPDDLPairBlock();
      m_pairblock->m_true = m_block.get();
      m_program = program;
      m_tempvarindex = 0;
      m_assembler = a;
      m_stack.push(std::make_pair(m_pairblock.get(), true));
   }

   CRPDDLExecutionBlock* getCurrentPath()
   {
      auto & t = m_stack.top();
      if (t.second) {
         return t.first->m_true.get();
      } else {
         return t.first->m_false.get();

      }
   }


   CUtiString numericOpToSyntheticPDDL(CRPDDLNumericOperation* num);
    void  generatePDDLEffect(int shift, CRPDDLExecutionBlock* block, CStringList& output, bool ownpddl,  std::set<CUtiIDString>& usedvars);
   void generatePDDLAction(const CUtiString& name, TRActionType isevent, CStringList &output, bool ownpddl);


   CUtiString createBoolExpression(const SRuntimeVariable&var,  std::set<CUtiIDString>& usedvars);
    CUtiString createBoolExpression(const CUtiIDString&variable, std::set<CUtiIDString> &usedvars);

    CUtiString createNumericExpression(const SRuntimeVariable&var,  std::set<CUtiIDString>& usedvars);
    CUtiString createNumericExpression(const CUtiIDString&variable,  std::set<CUtiIDString>& usedvars);
public:

    int m_tempvarindex;
    CUtiIDString getNewPDDLVariable()
    {
        CUtiIDString str(UTISTR("?ptemp")+m_tempvarindex);
        m_tempvarindex++;
        return str;
    }
};

Lisp::LObject replaceNthArgument(Lisp::LObject command, int n, Lisp::LObject newobj);

class CRAssembler : public ReferenceUtiOB
{
protected:
    int m_labelindex;
    int m_tempvarindex;
public:
    DynArray1D< std::pair<int, Lisp::LObject> > m_commands;
    bool m_norootmode = false;
    bool m_syntheticpddl;
    CRAssembler();
    ~CRAssembler();
    int createLabel();
    //Befehler wie jump erhalten als Ziel den Labelindex
    void add(const Lisp::LObject& command, int mylabel=-1);
    void printCode();

    void linkLabels();
    Lisp::LObject blockToVar(CRDataCell* cell,  TRBasicDataType type);
public:
    uti_refptr<CRDataCell> m_returnvalue;
    CReferenceList<CRStackElement> m_stack;
    uti_refptr<CRStackElement> m_global;
    std::map<CUtiIDString, CUtiIDString> m_tempvariables;
    void reserveVariable(const CUtiIDString& str, CRDataCell* cell);
    void setVariable(const CUtiIDString& str, CRDataCell* cell);
    CRDataCell* getVariable(const CUtiIDString& str);
    void pushStack() {m_stack.push_back(new CRStackElement());}
    void popStack() {m_stack.pop_back();}
    void failedCurrentPath() {assert(false);}
    CRDataCell* replaceVarWithPDDL(CRDataCell* dcell, const CUtiString& basename, bool param);
    CUtiIDString getNewTempVariable(const CUtiString&preferredname, const CUtiIDString& pddltype, bool parameter)
    {
        if (parameter && m_norootmode)
        {

            if (preferredname.size() != 0 && m_tempvariables.find(CUtiIDString(preferredname)) == m_tempvariables.end()) {
                CUtiIDString str(preferredname);
                // m_tempvarindex++;
                m_tempvariables.insert(std::make_pair(str, pddltype));
                if (parameter) {
                   add(Lisp::list(IDSTR("asm_copyobject"), IDSTR("PC"),str, CUtiIDString(preferredname)));
                }
                return str;
            } else {
                std::wcout << "parameter in root mode detected" << std::endl;
            }
        }
        if (preferredname.size() != 0 && m_tempvariables.find(CUtiIDString(preferredname)) == m_tempvariables.end()) {
            CUtiIDString str(preferredname);
            // m_tempvarindex++;
            m_tempvariables.insert(std::make_pair(str, pddltype));
            if (parameter ) {
               add(Lisp::list(IDSTR("asm_declparam"), IDSTR("PC"),str));
               assert( m_norootmode == false);
            }
            return str;
        }


        CUtiIDString str(UTISTR("?temp@")+m_tempvarindex);
        m_tempvarindex++;
        m_tempvariables.insert(std::make_pair(str, pddltype));
        if (parameter) {
           add(Lisp::list(IDSTR("asm_declparam"), IDSTR("PC"),str));
           //assert( m_norootmode == false);
        }
        return str;
    }

};

class CRNormalExecutionContext : public ReferenceUtiOB
{
public:
 //   bool m_pddlmode;
    bool m_doreturn;
    bool m_dobreak;
    bool m_docontinue;
    bool m_doexit;
    bool m_syntheticpddl;
   /* std::map< std::tuple<Lisp::LObject, CUtiIDString, Lisp::LObject>, CUtiIDString> m_numericaloperation;
    std::map<CUtiIDString, CUtiIDString> m_tempvariables;
    std::set<CUtiIDString> m_usedpddlvariables;
    void useVariable(const CUtiIDString& str)
    {
        m_usedpddlvariables.insert(str);
    }

    CUtiIDString getNewTempVariable(const CUtiString&preferredname, const CUtiIDString& pddltype)
    {
        if (preferredname.size() != 0 && m_tempvariables.find(CUtiIDString(preferredname)) == m_tempvariables.end()) {
            CUtiIDString str(preferredname);
            // m_tempvarindex++;
            m_tempvariables.insert(std::make_pair(str, pddltype));
            return str;
        }


        CUtiIDString str(UTISTR("?temp@")+m_tempvarindex);
        m_tempvarindex++;
        m_tempvariables.insert(std::make_pair(str, pddltype));
        return str;
    }

    CUtiIDString executeOperation(const Lisp::LObject& v1, const CUtiIDString& operation, const Lisp::LObject& v2, const CUtiIDString& pddltype=IDSTR("float"))
    {
        std::tuple<Lisp::LObject, CUtiIDString, Lisp::LObject> t;
        t = std::make_tuple(v1, operation, v2);
        auto it = m_numericaloperation.find(t);
        if (it != m_numericaloperation.end()) {
            return it->second;
        }
        CUtiIDString newtempvariable = getNewTempVariable(UTISTR(""), pddltype);
        m_numericaloperation.insert(std::make_pair(t, newtempvariable));
        return newtempvariable;
    }*/

    /* uti_refptr<CRActionGenerator> m_actiongenerator;
    uti_refptr<CRActionGenerator> getCurrentGenerator(CRDataCell* datacell) {
    return m_actiongenerator.get();
    }*/

    //uti_refptr<CRPDDLPath> m_currentpddlpath;
    // std::stack<CRPDDLDecision*> m_decisionstack;
   // uti_refptr<CRPDDLExecutionBlock> m_pddlmainblock;
   /* CReferenceList<CRPDDLExecutionBlock> m_pddlstack;
    void pushPDDLBlock(const Lisp::LObject& condition)
    {
        CRPDDLExecutionBlock* block = m_pddlstack.get(m_pddlstack.size()-1);
        CRPDDLExecutionBlock* newblock = new CRPDDLExecutionBlock();
        newblock->m_condition.push_back( condition);
        block->m_subblocks.push_back(newblock);
        m_pddlstack.push_back(newblock);
    };
    void popPDDLBlock()
    {
        m_pddlstack.pop_back();
    }
    CRPDDLExecutionBlock* getPDDLTop()
    {
        return m_pddlstack.get(m_pddlstack.size()-1);
    }
*/
    DynArray1D<std::pair<Lisp::LObject, Lisp::LObject> > m_initial;
    DynArray1D<std::pair<Lisp::LObject, Lisp::LObject> > m_goal;
    uti_refptr<CRDataCell> m_returnvalue;
    CReferenceList<CRStackElement> m_stack;
    uti_refptr<CRStackElement> m_global;
    void reserveVariable(const CUtiIDString& str, CRDataCell* cell);
    void setVariable(const CUtiIDString& str, CRDataCell* cell);
    CRDataCell* getVariable(const CUtiIDString& str);
    void pushStack() {m_stack.push_back(new CRStackElement());}
    void popStack() {m_stack.pop_back();}
    CRNormalExecutionContext() {//m_pddlmode = true;
                         m_syntheticpddl = false;
                                m_doreturn = false; m_dobreak = false; m_docontinue = false;m_doexit=false;
                               // m_pddlmainblock = new CRPDDLExecutionBlock();
                                       //                 m_pddlstack.push_back(m_pddlmainblock.get());
                                                                               // m_tempvarindex = 0;
                                                                                                        m_global = new CRStackElement();
                               }
};


class CRBaseRule : public ReferenceUtiOB
{
public:
    CRBaseRule* m_parent;
    CRProgram* m_program;
    CRBaseRule(CRBaseRule* p);
    virtual void doParse(UtiLexer& parse){};


    virtual void error(const CUtiString& data);
    virtual void error(UtiLexer& parse, const CUtiString& data);
    virtual void MatchCharacter(UtiLexer& parse, U_CHAR c);

    virtual void MatchIdentifier(UtiLexer& parse, const CUtiString& str);
    CRBlock* getParentBlock();
    CRProgram* getParentProgram();
    void changeParent(CRBaseRule* p);
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext){return uti_refptr<CRDataCell>();};
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler) {return uti_refptr<CRDataCell>();};
};


class CRCommand : public CRBaseRule
{
public:
    CRCommand(CRBaseRule* p);
    virtual void declarePDDLType(CStringList* output){};

};

class CRBaseDeclaration;



class CREval : public CRBaseRule
{
public:
    uti_refptr<CRBaseDeclaration> m_typedata;
    CREval(CRBaseRule* p);
    virtual CRBaseDeclaration* getType() {return m_typedata.get();}
};

class CRExpressionOperator;
class CRTypedOperator;

class CRExpression : public CREval
{
public:

    CRExpression(CRBaseRule* p);
    TRExpressionType m_type;
    CRTypedOperator* m_datatype;
    uti_refptr<CREval> m_a;
    uti_refptr<CREval> m_b;

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

class CRValue : public CREval
{
public:
    double m_value;
    CUtiIDString m_atom;
    TRBasicDataType m_type;
    CRValue(CRBaseRule* p);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

class CRReference;
class CRFunctionCall;
class CRBlock;
class CRFunctionDeclaration;
class CRBaseTypeDeclaration;
class CRBaseDeclaration;


class CREvalFunction : public CREval
{
public:
    uti_refptr<CRFunctionCall> m_call;
    CREvalFunction(CRBaseRule* p);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};


class CRExists : public CREval
{
public:
    uti_refptr<CREval> m_value;
    CRExists(CRBaseRule* p);
    void doParse(UtiLexer& parse);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

class CRBaseDeclaration : public CRCommand
{
public:

    CUtiIDString m_name;
    CRBaseDeclaration(CRBaseRule* p);
    virtual bool isConstant() {return false;};
    virtual bool isType() {return false;};
    virtual bool isFunction() {return false;};
    virtual bool isVariable() {return false;};

    virtual bool canConvertTo(CRBaseDeclaration* decl){return this == decl;};
    virtual CRDataCell* createCell() {return NULL;};
    virtual CRBaseDeclaration* executeTypeOperation(CRBaseDeclaration* decl, int operatortype){return NULL;}

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, int operatortype,  CRDataCell* othercell) {return uti_refptr<CRDataCell>();}
    virtual void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB* obj) {assert(false);}
    virtual void createFlatVariableList(DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> >& list, const CUtiString& basename)
    {
       list.push_back(std::make_pair(basename, this));
    };
    virtual void createPDDLDeclaration(CStringList* list, bool createpredicates){};
};

class CRVariable : public CRBaseDeclaration
{
public:
    uti_refptr<CRBaseDeclaration> m_type;
    CUtiIDString m_typename;
    bool m_isconstant;
    bool m_isparameter;
    uti_refptr<CREval> m_initvalue;


    CRVariable(CRBaseRule* p);
    virtual bool isConstant() {return m_isconstant;};
    virtual bool isVariable() {return !m_isconstant;};
    void doParse(UtiLexer& parse);
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext)
    {
        /* uti_refptr<CRDataCell> result(m_type->createCell());
       if (econtext->m_pddlmode) {
           result = econtext->replaceVarWithPDDL(result.get(),m_name.str());
       }

       econtext->reserveVariable(m_name,result.get());*/
        if (m_initvalue.get()) {
            uti_refptr<CRDataCell> ed = m_initvalue->execute(econtext);

            //result->copyFrom(ed.get());
            //   uti_refptr<CRDataCell> result(m_type->createCell());
            /* if (econtext->m_pddlmode) {
               result = econtext->replaceVarWithPDDL(result.get(),m_name.str());
           }*/

            econtext->reserveVariable(m_name,ed.get());
            return ed;
        }  else {
            uti_refptr<CRDataCell> result(m_type->createCell());
            /*if (econtext->m_pddlmode) {

            }*/
            result->replaceVarWithAtom(econtext,NULL,m_name.str());
            if (m_isparameter) {
             //   TODO: Wie ordnen wir die Vars zu, das durch nur in der primary function geschehen
               // result = result->replaceVarWithPDDL(result.get(),m_name.str());
            }

            econtext->reserveVariable(m_name,result.get());

            return result;
        }
    }

    //Wir wollen die Variablen rekursiv fullen
    //Rekursiv reservieren beim assembler mit Variablen Stamm
    //Init:Value Rekursiv anmelden
    //Parameter Deklarieren
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler)
    {
        if (m_initvalue.get()) {
            uti_refptr<CRDataCell> ed = m_initvalue->compile(assembler);


            assembler->reserveVariable(m_name,ed.get());
            if (m_isparameter) {
              //  TODO: Rekursive durchlaufen
               assert(false);
               // assembler->add(Lisp::list(IDSTR("asm_declparam"), IDSTR("PC"),m_name));

                   //ed->copyFromCompile(assembler,ed.get() );
                 //  assembler->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),m_name, Lisp::list(IDSTR("vvarf"), m_name)));

            } else {
                 //ed->copyFromCompile(assembler,ed.get() );
            }
            return ed;
        } else {
            uti_refptr<CRDataCell> result;
            /*if (econtext->m_pddlmode) {

            }*/

            if (m_isparameter &&assembler->m_norootmode) {
                return result;
            }
            result = m_type->createCell();
            result->replaceVarWithAtom(NULL,NULL,m_name.str());
           result = assembler->replaceVarWithPDDL(result.get(),m_name.str(), m_isparameter);
            assembler->reserveVariable(m_name,result.get());
            /*if (m_isparameter) {
                   assembler->add(Lisp::list(IDSTR("asm_declparam"), IDSTR("PC"),m_name));

            } else {
                // result->copyFromCompile(assembler,ed.get() );
            }*/
            return result;
        }
    }
    virtual uti_refptr<CRDataCell> compileDecl(CRAssembler* assembler)
    {
        if (m_isconstant == false) {

            error(UTISTR("Global variabes are not supported"));
        }
        if (m_initvalue.get()) {
            uti_refptr<CRDataCell> ed = m_initvalue->compile(assembler);


            assembler->reserveVariable(m_name,ed.get());

            return ed;
        } else {
            uti_refptr<CRDataCell> result(m_type->createCell());
            /*if (econtext->m_pddlmode) {

            }*/
            result->replaceVarWithAtomAss(assembler,NULL,m_name.str());
            if (m_isparameter) {
             //   TODO: Wie ordnen wir die Vars zu, das durch nur in der primary function geschehen
               // result = result->replaceVarWithPDDL(result.get(),m_name.str());
            }

            assembler->reserveVariable(m_name,result.get());

            return result;


         //  error(UTISTR("Undefined global variabes are not supported"));
           // return result;
        }
    }

};


class CRDeclaration : public CRBaseDeclaration
{
public:
    CUtiIDString m_type;
    uti_refptr<CRBaseDeclaration> m_typedata;
    uti_refptr<CREval> m_value;
    CRDeclaration(CRBaseRule* p);
    CRDeclaration(CRBaseRule* p, CRBaseDeclaration* t);

    virtual void doParse(UtiLexer& parse);
    void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB* obj);

};

class CRBaseTypeDeclaration : public CRBaseDeclaration
{
public:
    TRBasicDataType m_data;
    CRBaseTypeDeclaration(CRBaseRule* p);
    CRBaseTypeDeclaration(TRBasicDataType t) : CRBaseDeclaration(NULL) {m_data = t;}
    virtual bool isType() {return true;};
    virtual bool canConvertTo(CRBaseDeclaration* decl);
    virtual CRBaseDeclaration* executeTypeOperation(CRBaseDeclaration* decl, int operatortype);

    uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, int operatortype,  CRDataCell* othercell);
    virtual CRDataCell* createCell() {
        switch (m_data) {
        case erInt: {
            CRIntCell* c = new CRIntCell();
            c->m_type = this;
            return c;
        }; break;
        case erFloat: {
            CRFloatCell* c = new CRFloatCell();
            c->m_type = this;
            return c;
        }; break;
        case erBool: {
            CRBoolCell* c = new CRBoolCell();
            c->m_type = this;
            return c;
        }; break;
        case erString: {
            CRStringCell* c = new CRStringCell();
            c->m_type = this;
            return c;
        }; break;
        default: {
            assert(false);
        };
        }
        return NULL;
    };

    void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB* obj);
};


class CRStructDeclaration : public CRBaseDeclaration
{
public:
    CReferenceList< CRDeclaration > m_children_tan; //Type, Array, Name
    CRStructDeclaration(CRBaseRule* p);
    virtual bool isType() {return true;};
    int findIndex(const CUtiIDString& name);
    virtual void doParse(UtiLexer& parse);

    virtual CRDataCell* createCell();
    Lisp::LObject toLispDeclaration();

    void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB* obj);

    void createFlatVariableList(DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> >& list, const CUtiString& basename);
    void createPDDLDeclaration(CStringList* list, bool createpredicates);

};

class CRClassDeclaration : public CRBaseDeclaration
{
public:
    //   CReferenceList< CRDeclaration > m_children_tan; //Type, Array, Name
    // CRStructDeclaration(CRBaseRule* p);
    //   int findIndex(const CUtiIDString& name);
    uti_refptr<CRClassDeclaration> m_parenttype;
    virtual bool isType() {return true;};

    CRClassDeclaration(CRBaseRule* p) : CRBaseDeclaration(p)
    {

    }
    virtual void declarePDDLType(CStringList* output)
    {
       if (m_parenttype.get() == NULL) {
          output->add(UTISTR("   ")+m_name.str() + UTISTR(" - object"));
       } else {

          output->add(UTISTR("   ")+m_name.str() + UTISTR(" - ")+m_parenttype->m_name.str());
       }
    }

    virtual void doParse(UtiLexer& parse);

    virtual CRDataCell* createCell() {
        CRClassCell* cell = new CRClassCell();
        cell->m_type = this;

        return cell;
    }

    virtual bool canConvertTo(CRBaseDeclaration* decl);
    virtual CRBaseDeclaration* executeTypeOperation(CRBaseDeclaration* decl, int operatortype);
    void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB* obj) ;
};

class CRArrayDeclaration : public CRBaseDeclaration
{
public:
    DynArray1D<int> m_size;
    uti_refptr<CRDeclaration> m_basetype;
    CRArrayDeclaration(CRBaseRule* p);
    virtual bool isType() {return true;};

    virtual CRDataCell* createCell() {
        CRArrayCell* cell = new CRArrayCell();
        cell->m_type = this;
        int a = 1;
        for (int i = 0; i< m_size.size(); i++) {
            a*= m_size[i];
        }
        cell->m_cells.setSize(a);
        for (int i = 0;  i < a; i++) {
            cell->m_cells[i] = (m_basetype->createCell());
        }
        return cell;
    }
    virtual bool canConvertTo(CRBaseDeclaration* decl);
    virtual CRBaseDeclaration* executeTypeOperation(CRBaseDeclaration* decl, int operatortype);

    void createFlatPDDLList(DynArray1D<Lisp::LObject>& list, int&index, int mode, ReferenceUtiOB* obj) ;
    void createFlatVariableList(DynArray1D<std::pair<CUtiString, CRBaseDeclaration*> >& list, const CUtiString& basename);
};

class CRTypeDefDeclaration : public CRBaseDeclaration
{
public:
    uti_refptr<CRBaseDeclaration> m_parenttype;
    virtual bool isType() {return true;};
    CRTypeDefDeclaration(CRBaseRule* p);
    virtual void doParse(UtiLexer& parse);

    virtual bool canConvertTo(CRBaseDeclaration* decl);
    virtual CRDataCell* createCell() {
        CRDataCell* cell = m_parenttype->createCell();
        cell->m_type = m_parenttype.get();
        return cell;
    }
    virtual CRBaseDeclaration* executeTypeOperation(CRBaseDeclaration* decl, int operatortype);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, int operatortype,  CRDataCell* othercell);
};
/*
class CRObjectType : public CRBaseDeclaration
{
public:
    CUtiString m_entityname;
    CRObjectType* m_parent;
    CRObjectType(CRBaseRule* p);
};*/
class CRReference : public CREval
{
public:
    CRReference(CRBaseRule* p);


};


class CRVarReference : public CRReference
{
public:
    CRVarReference(CRBaseRule* p);
    CRVariable* m_variable;

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext)
    {
        uti_refptr<CRDataCell> result = econtext->getVariable(m_variable->m_name);
       /* if (econtext->m_pddlmode && dynamic_cast<CRPDDLVariable*>(result.get())) {
            CRPDDLVariable* v = check_cast<CRPDDLVariable*>(result.get());
            econtext->useVariable(v->m_name);
        }*/

        return result;
        //return econtext->getVariable(m_variable->m_name);
    }

    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler)
    {
        uti_refptr<CRDataCell> result = assembler->getVariable(m_variable->m_name);
       /* if (econtext->m_pddlmode && dynamic_cast<CRPDDLVariable*>(result.get())) {
            CRPDDLVariable* v = check_cast<CRPDDLVariable*>(result.get());
            econtext->useVariable(v->m_name);
        }*/

        return result;
        //return econtext->getVariable(m_variable->m_name);
    }

};

class CRAssert : public CRCommand
{
protected:
    int m_myline;
public:
    CRAssert(CRBaseRule* p);

  //  uti_refptr<CRDataCell> m_savedreference;
    uti_refptr<CREval> m_index;

    virtual void doParse(UtiLexer& parse);
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};


class CRPrint : public CRCommand
{
public:
    bool m_newline;
    CRPrint(CRBaseRule* p);

    CReferenceList<CREval> m_params;

    virtual void doParse(UtiLexer& parse);
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

enum TDCommandType {dcInsert, dcRemove, dcGoal, dcNotGoal};

class CRDatabaseCommand : public CRCommand
{
protected:
    TDCommandType m_ct;
public:
    CRDatabaseCommand(CRBaseRule* p, TDCommandType ct);

    //u<CRDataCell> m_savedreference;

    uti_refptr<CREval> m_index;

    virtual void doParse(UtiLexer& parse);
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

class CRWait : public CRCommand
{
public:
    CRWait(CRBaseRule* p);

    uti_refptr<CREval> m_index;

    virtual void doParse(UtiLexer& parse);
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};


class CRReturn : public CRCommand
{
public:
    CRReturn(CRBaseRule* p);

    uti_refptr<CREval> m_index;
    virtual void doParse(UtiLexer& parse);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};


class CRIf : public CRCommand
{
public:
    CRIf(CRBaseRule* p);

    uti_refptr<CREval> m_condition;
    uti_refptr<CRBlock> m_trueblock;
    uti_refptr<CRBlock> m_falseblock;
    virtual void doParse(UtiLexer& parse);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

class CRFor : public CRCommand
{
public:
    uti_refptr<CRCommand> m_startcommand;
    uti_refptr<CREval> m_checkcondition;
    uti_refptr<CRCommand> m_stepcommand;

    uti_refptr<CRBlock> m_block;
    CRFor(CRBaseRule* p);
    virtual void doParse(UtiLexer& parse);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);

};

class CRReferenceArrayAccess : public CRReference
{
public:
    CRReferenceArrayAccess(CRBaseRule* p);
    uti_refptr<CRReference> m_source;
    CReferenceList<CREval> m_index;
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler)
    {

        uti_refptr<CRDataCell> r(m_source->compile(assembler));
        CRArrayCell* ac = check_cast<CRArrayCell*>(r.get());
        int accessindex = 0;
        int sidesize= 1;
        CRArrayDeclaration* decl = check_cast<CRArrayDeclaration*>(m_typedata.get());
        assert(false); // Implement me
        //flate target, flat jedes array element, fur jeden flat entry asm_copy
        //asm_array(PC, target, index, [listvalue])
        return uti_refptr<CRDataCell>(ac->m_cells.get(accessindex).get());
    }
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext)
    {
        uti_refptr<CRDataCell> r(m_source->execute(econtext));
        CRArrayCell* ac = check_cast<CRArrayCell*>(r.get());
        int accessindex = 0;
        int sidesize= 1;
        CRArrayDeclaration* decl = check_cast<CRArrayDeclaration*>(m_typedata.get());
        for (int i = 0; i < decl->m_size.size(); i++) {
            accessindex += sidesize*m_index[i].execute(econtext)->toInt();
            sidesize *= decl->m_size[i];
        }
        return uti_refptr<CRDataCell>(ac->m_cells.get(accessindex).get());


    }

};


class CRReferenceStruct : public CRReference
{
public:
    CRReferenceStruct(CRBaseRule* p);
    uti_refptr<CRReference> m_source;
    CUtiIDString m_subelement;
    int m_subelementindex;
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext)
    {

        uti_refptr<CRDataCell> r(m_source->execute(econtext));
        CRRecordCell* ac = check_cast<CRRecordCell*>(r.get());

        return uti_refptr<CRDataCell>(&ac->m_subcells[m_subelementindex]);
    }
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler)
    {

        uti_refptr<CRDataCell> r(m_source->compile(assembler));
        CRRecordCell* ac = check_cast<CRRecordCell*>(r.get());

        return uti_refptr<CRDataCell>(&ac->m_subcells[m_subelementindex]);


    }
};

class CRSetVariable : public CRCommand
{
public:
    CRSetVariable(CRBaseRule* p);
    uti_refptr<CRReference> m_variable;
    uti_refptr<CREval> m_value;


    virtual void doParse(UtiLexer& parse);
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext)
    {
        uti_refptr<CRDataCell> cell = m_variable->execute(econtext);
        uti_refptr<CRDataCell> resultvalue(m_value->execute(econtext));
        cell->copyFrom(resultvalue.get());
        return uti_refptr<CRDataCell>();
    }
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler)
    {
        uti_refptr<CRDataCell> cell = m_variable->compile(assembler);
        uti_refptr<CRDataCell> resultvalue(m_value->compile(assembler));
        cell->copyFromCompile(assembler, resultvalue.get());
        return uti_refptr<CRDataCell>();
    }
};


class CRFunctionCall : public CRCommand
{
public:
    CRFunctionCall(CRBaseRule* p);
    CReferenceList<CRBaseDeclaration> m_templateparameters;
    CReferenceList< CREval > m_parameters;

    CRFunctionDeclaration* m_declaration;
    CRStructDeclaration* m_structdeclaration;
    virtual void doParse(UtiLexer& parse);
    virtual bool isFunction() {return false;};
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

//Process via function
enum TRFunctionType {rNormal, rAction, rEvent};

class CRFunctionDeclaration_Callback : public ReferenceUtiOB
{
public:

    virtual uti_refptr<CRDataCell> execute(CRFunctionDeclaration* func, CRNormalExecutionContext* econtext, Array<uti_refptr<CRDataCell> >& parameters)=0;
};


class CRFunctionDeclaration : public CRBaseDeclaration
{
public:
    TRActionType m_actiontype;
    CRFunctionDeclaration(CRBaseRule* p);
    TRFunctionType m_type;
    std::map<CRAssembler*, int> m_label;
    bool isFunction(){return true;}
    int getLabelOrCreate(CRAssembler* assembler)
    {
        auto it = m_label.find(assembler);
        if (it != m_label.end()) {
            return it->second;
        }else{
            int newlabel = assembler->createLabel();
            m_label.insert(std::make_pair(assembler, newlabel));
            return newlabel;
        }
    }
    int m_callerid = 0;
    CUtiString getCallerVarName()
    {
        CUtiString res = (m_name.str()+m_callerid);
        m_callerid++;
        return res;

    }
    DynArray1D< std::tuple<uti_refptr<CRBaseDeclaration>, CUtiIDString> > m_param_type_and_name;
    DynArray1D< std::tuple<uti_refptr<CRBaseDeclaration>, CUtiIDString> > m_template_type_and_name;

    CUtiIDString m_returntype;
    uti_refptr<CRBaseDeclaration> m_returntypedata;
    uti_refptr<CRFunctionDeclaration_Callback> m_callback;
    uti_refptr<CRBlock> m_block;
    virtual void doParse(UtiLexer& parse);
    uti_refptr<CRDataCell> compileNonMainFunction(CRAssembler* assembler);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

class CRBlock : public CRCommand
{
public:
    CRBlock(CRBaseRule* p);
    bool m_needbraces;
    bool m_singleline;
    int m_blocklevel;
    CReferenceList<CRCommand> m_command;
    std::map<CUtiIDString, CRBaseDeclaration*> m_block;
    void addChild(UtiLexer& parse, CRCommand* rule);

    CRBaseDeclaration *find(const CUtiIDString& str);

    virtual void doParse(UtiLexer& parse);
    void parseDeclaration(UtiLexer& parse, CUtiIDString& typestr, uti_refptr<CRBaseDeclaration> & typedata, CUtiIDString& name);
    uti_refptr<CRReference> parseReference(UtiLexer& parse, CRBaseRule *parent);
    uti_refptr<CREval> parseExpression(UtiLexer& parse, CRBaseRule* parent, int layer=0);

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* assembler);
};

class CRTypedOperator : public ReferenceUtiOB
{
public:
    TRBasicDataType m_type1;
    TRBasicDataType m_type2;
    TRBasicDataType m_res;
    CRTypedOperator(TRBasicDataType t1, TRBasicDataType t2, TRBasicDataType r);
    virtual uti_refptr<CRDataCell> compile(CRAssembler* econtext, CRDataCell* mycell, CRDataCell* othercell)=0;
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell)=0;


};


class CRLessTypedOperator : public CRTypedOperator
{
protected:
    TRExpressionType m_type;
public:

    CRLessTypedOperator(TRExpressionType expt, TRBasicDataType t1, TRBasicDataType t2, TRBasicDataType r) : CRTypedOperator(t1, t2, r)
    {
        m_type = expt;
    }

    virtual uti_refptr<CRDataCell> compile(CRAssembler* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {
        //asm_add(PC, R, vint(P1), vint(P2))
        DynArray1D<Lisp::LObject> list;
        //enum TRExpressionType {eLogicalOr, eLogicalAnd, eAdd, eSub, eMul, eDiv, eLess, eGreater, eEqual, eELess, eEGreater, eModulo, eIntDiv};

        switch (m_type) {
        case eLess: {
        list.push_back(IDSTR("asm_less"));
        }; break;
        case eGreater: {
            list.push_back(IDSTR("asm_greater"));

        }; break;
        case eEGreater: {
            list.push_back(IDSTR("asm_egreater"));

        }; break;
        case eELess: {
            list.push_back(IDSTR("asm_eless"));

        }; break;
        case eEqual: {
            list.push_back(IDSTR("asm_equal"));

        }; break;

        }
        list.push_back(IDSTR("PC"));

        bool needfloat = false;
        if (m_type1 == erFloat) {
            needfloat = true;
        }
        if (m_type2 == erFloat) {
            needfloat = true;
        }

        CUtiIDString result = econtext->getNewTempVariable(UTISTR(""),IDSTR("bool"), false);
        list.push_back(result);
        if (dynamic_cast<CRPDDLVariable*>(mycell) != NULL) {
            CRPDDLVariable* var = dynamic_cast<CRPDDLVariable*>(mycell);
            if (m_type1 == erInt) {
                if (needfloat) {

                    CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
                    econtext->add(Lisp::list(IDSTR("asm_tofloat"), IDSTR("PC"),param,var->m_name));
                    list.push_back(param);
                } else {
                    list.push_back( var->m_name);
                }
            } else if (m_type1 == erFloat) {
                assert(needfloat == true);
                list.push_back(var->m_name);
            } else {
                assert(false);
            }
        } else {
           if (m_type1 == erInt) {
               if (needfloat) {

                   CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
                   econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vfloat"), check_cast<CRIntCell*>(mycell)->data)));
                   list.push_back( param);
               } else {
                   CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("int"), false);
                   econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vint"), check_cast<CRIntCell*>(mycell)->data)));
                   list.push_back(param);
               }
           } else if (m_type1 == erFloat) {
               assert(needfloat == true);
               CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
               econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vfloat"), check_cast<CRFloatCell*>(mycell)->data)));
               list.push_back(param);
           } else {
               assert(false);

           }
        }
        if (dynamic_cast<CRPDDLVariable*>(othercell) != NULL) {
            CRPDDLVariable* var = dynamic_cast<CRPDDLVariable*>(othercell);
            if (m_type2 == erInt) {
                if (needfloat) {

                    CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
                    econtext->add(Lisp::list(IDSTR("asm_tofloat"), IDSTR("PC"),param,var->m_name));
                    list.push_back( param);
                } else {
                    list.push_back( var->m_name);
                }
            } else if (m_type2 == erFloat) {
                list.push_back( var->m_name);
            } else {
                assert(false);
            }
        } else {
           if (m_type2 == erInt) {
               if (needfloat) {

                   CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
                   econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vfloat"), check_cast<CRIntCell*>(othercell)->data)));
                   list.push_back(param);
               } else {
                  CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("int"), false);
                  econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vint"), check_cast<CRIntCell*>(othercell)->data)));
                  list.push_back(param);
               }
           } else if (m_type2 == erFloat) {
               CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
               econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vfloat"), check_cast<CRFloatCell*>(othercell)->data)));
               list.push_back(param);
           } else {
               assert(false);

           }
        }
      /*  if (m_type1 == erInt && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRPDDLVariable(check_cast<CRIntCell*>(mycell)->data<check_cast<CRIntCell*>(othercell)->data ));

        }

        if (m_type1 == erInt && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRBoolCell(double(check_cast<CRIntCell*>(mycell)->data)<check_cast<CRFloatCell*>(othercell)->data ));

        }
        if (m_type1 == erFloat && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data<double(check_cast<CRIntCell*>(othercell)->data) ));

        }
        if (m_type1 == erFloat && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data<check_cast<CRFloatCell*>(othercell)->data) );

        }*/
          econtext->add(Lisp::list(list));
        return uti_refptr<CRDataCell>(new CRPDDLVariable(result, erBool));
    }

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {
        /*if (mycell->isPDDLVariable() || othercell->isPDDLVariable()) {
            mycell->declPDDLVariable(econtext); othercell->declPDDLVariable(econtext);
            CUtiIDString newvariablename = econtext->executeOperation(mycell->toLisp(NULL), IDSTR("<"), othercell->toLisp(NULL));

            return uti_refptr<CRDataCell>(new CRBoolExpression(m_type, mycell, othercell));
        }*/
        switch (m_type) {
        case eLess: {
            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRIntCell*>(mycell)->data<check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(double(check_cast<CRIntCell*>(mycell)->data)<check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data<double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data<check_cast<CRFloatCell*>(othercell)->data) );

            }
        }; break;
        case eGreater: {
            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRIntCell*>(mycell)->data>check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(double(check_cast<CRIntCell*>(mycell)->data)>check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data>double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data>check_cast<CRFloatCell*>(othercell)->data) );

            }

        }; break;
        case eEGreater: {
            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRIntCell*>(mycell)->data>=check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(double(check_cast<CRIntCell*>(mycell)->data)>=check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data>=double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data>=check_cast<CRFloatCell*>(othercell)->data) );

            }

        }; break;
        case eELess: {
            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRIntCell*>(mycell)->data<=check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(double(check_cast<CRIntCell*>(mycell)->data)<=check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data<=double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data<=check_cast<CRFloatCell*>(othercell)->data) );

            }

        }; break;
        case eEqual: {

            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRIntCell*>(mycell)->data==check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(double(check_cast<CRIntCell*>(mycell)->data)==check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data==double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data==check_cast<CRFloatCell*>(othercell)->data) );

            }

        }; break;

        }

        return uti_refptr<CRDataCell>();
    }
};


class CRLogicalTypedOperator : public CRTypedOperator
{
protected:
    TRExpressionType m_typeop;
public:

    CRLogicalTypedOperator(TRExpressionType expt, TRBasicDataType t1, TRBasicDataType t2, TRBasicDataType r) : CRTypedOperator(t1, t2, r)
    {
        m_typeop = expt;
    }

    virtual uti_refptr<CRDataCell> compile(CRAssembler* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {
        //asm_add(PC, R, vint(P1), vint(P2))
        DynArray1D<Lisp::LObject> list;
        //enum TRExpressionType {eLogicalOr, eLogicalAnd, eAdd, eSub, eMul, eDiv, eLess, eGreater, eEqual, eELess, eEGreater, eModulo, eIntDiv};

        switch (m_typeop) {
        case eLogicalOr: {
        list.push_back(IDSTR("asm_logicalor"));
        }; break;
        case eLogicalAnd: {
            list.push_back(IDSTR("asm_logicaland"));

        }; break;
        }
        list.push_back(IDSTR("PC"));



        CUtiIDString result = econtext->getNewTempVariable(UTISTR(""),IDSTR("bool"), false);
        list.push_back(result);
        if (dynamic_cast<CRPDDLVariable*>(mycell) != NULL) {
            CRPDDLVariable* var = dynamic_cast<CRPDDLVariable*>(mycell);
            if (m_type1 == erBool) {

                list.push_back( var->m_name);

            } else {
                assert(false);
            }
        } else {
           if (m_type1 == erBool) {

                   CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("bool"), false);
                   econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vbool"), check_cast<CRBoolCell*>(mycell)->data ? 1 : 0)));
                   list.push_back(param);

           }else {
               assert(false);

           }
        }
        if (dynamic_cast<CRPDDLVariable*>(othercell) != NULL) {
            CRPDDLVariable* var = dynamic_cast<CRPDDLVariable*>(othercell);
            if (m_type2 == erBool) {

                list.push_back( var->m_name);

            }else {
                assert(false);
            }
        } else {
           if (m_type2 == erInt) {

                  CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("bool"), false);
                  econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vbool"), check_cast<CRIntCell*>(othercell)->data ? 1 : 0)));
                  list.push_back(param);

           }else {
               assert(false);

           }
        }
      /*  if (m_type1 == erInt && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRPDDLVariable(check_cast<CRIntCell*>(mycell)->data<check_cast<CRIntCell*>(othercell)->data ));

        }

        if (m_type1 == erInt && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRBoolCell(double(check_cast<CRIntCell*>(mycell)->data)<check_cast<CRFloatCell*>(othercell)->data ));

        }
        if (m_type1 == erFloat && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data<double(check_cast<CRIntCell*>(othercell)->data) ));

        }
        if (m_type1 == erFloat && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRFloatCell*>(mycell)->data<check_cast<CRFloatCell*>(othercell)->data) );

        }*/
          econtext->add(Lisp::list(list));
        return uti_refptr<CRDataCell>(new CRPDDLVariable(result, erBool));
    }

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {
        /*if (mycell->isPDDLVariable() || othercell->isPDDLVariable()) {
            mycell->declPDDLVariable(econtext); othercell->declPDDLVariable(econtext);
            CUtiIDString newvariablename = econtext->executeOperation(mycell->toLisp(NULL), IDSTR("<"), othercell->toLisp(NULL));

            return uti_refptr<CRDataCell>(new CRBoolExpression(m_type, mycell, othercell));
        }*/

        switch (m_typeop) {
        case eLogicalOr: {
            if (m_type1 == erBool && m_type2 == erBool) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRBoolCell*>(mycell)->data ||check_cast<CRBoolCell*>(othercell)->data ));

            }
        }; break;
        case eLogicalAnd: {
            if (m_type1 == erBool && m_type2 == erBool) {
                return uti_refptr<CRDataCell>(new CRBoolCell(check_cast<CRBoolCell*>(mycell)->data && check_cast<CRBoolCell*>(othercell)->data ));

            }

        }; break;
        }


        return uti_refptr<CRDataCell>();
    }
};


class CRPlusTypedOperator : public CRTypedOperator
{
public:
    TRExpressionType m_type;

    CRPlusTypedOperator(TRExpressionType typeop, TRBasicDataType t1, TRBasicDataType t2, TRBasicDataType r) : CRTypedOperator(t1, t2, r)
    {
        m_type = typeop;

    }
    virtual uti_refptr<CRDataCell> compile(CRAssembler* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {
        //asm_add(PC, R, vint(P1), vint(P2))
        DynArray1D<Lisp::LObject> list;
        //enum TRExpressionType {eLogicalOr, eLogicalAnd, eAdd, eSub, eMul, eDiv, eLess, eGreater, eEqual, eELess, eEGreater, eModulo, eIntDiv};
        switch (m_type) {
           case eAdd:{list.push_back(IDSTR("asm_add"));}; break;
        case eSub:{list.push_back(IDSTR("asm_sub"));}; break;
        case eMul:{list.push_back(IDSTR("asm_mul"));}; break;
        case eDiv:{list.push_back(IDSTR("asm_div"));}; break;
        };
        list.push_back(IDSTR("PC"));

        bool needfloat = false;
        if (m_type1 == erFloat) {
            needfloat = true;
        }
        if (m_type2 == erFloat) {
            needfloat = true;
        }

        CUtiIDString result = econtext->getNewTempVariable(UTISTR(""),IDSTR("bool"), false);
        list.push_back(result);
        if (dynamic_cast<CRPDDLVariable*>(mycell) != NULL) {
            CRPDDLVariable* var = dynamic_cast<CRPDDLVariable*>(mycell);
            if (m_type1 == erInt) {
                if (needfloat) {

                    CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
                    econtext->add(Lisp::list(IDSTR("asm_tofloat"), IDSTR("PC"),param,var->m_name));
                    list.push_back(param);
                } else {
                    list.push_back(var->m_name);
                }
            } else if (m_type1 == erFloat) {
                assert(needfloat == true);
                list.push_back(var->m_name);
            } else {
                assert(false);
            }
        } else {
           if (m_type1 == erInt) {
               if (needfloat) {

                   CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
                   econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vfloat"), check_cast<CRIntCell*>(mycell)->data)));
                   list.push_back(param);
               } else {
                   CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("int"), false);
                   econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vint"), check_cast<CRIntCell*>(mycell)->data)));
                   list.push_back(param);
               }
           } else if (m_type1 == erFloat) {
               assert(needfloat == true);
               CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
               econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vfloat"), check_cast<CRFloatCell*>(mycell)->data)));
               list.push_back(param);
           } else {
               assert(false);

           }
        }
        if (dynamic_cast<CRPDDLVariable*>(othercell) != NULL) {
            CRPDDLVariable* var = dynamic_cast<CRPDDLVariable*>(othercell);
            if (m_type2 == erInt) {
                if (needfloat) {

                    CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
                    econtext->add(Lisp::list(IDSTR("asm_tofloat"), IDSTR("PC"),param,var->m_name));
                    list.push_back(param);
                } else {
                    list.push_back(var->m_name);
                }
            } else if (m_type2 == erFloat) {
                list.push_back(var->m_name);
            } else {
                assert(false);
            }
        } else {
           if (m_type2 == erInt) {
               if (needfloat) {

                   CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
                   econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vfloat"), check_cast<CRIntCell*>(othercell)->data)));
                   list.push_back(param);
               } else {
                  CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("int"), false);
                  econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vint"), check_cast<CRIntCell*>(othercell)->data)));
                  list.push_back(param);
               }
           } else if (m_type2 == erFloat) {
               CUtiIDString param = econtext->getNewTempVariable(UTISTR(""),IDSTR("float"), false);
               econtext->add(Lisp::list(IDSTR("asm_setvariable"), IDSTR("PC"),param,Lisp::list(IDSTR("vfloat"), check_cast<CRFloatCell*>(othercell)->data)));
               list.push_back( param);
           } else {
               assert(false);

           }
        }
        if (needfloat) {
            return uti_refptr<CRDataCell>(new CRPDDLVariable(result, erFloat));

        }
         econtext->add(Lisp::list(list));

        return uti_refptr<CRDataCell>(new CRPDDLVariable(result, erInt));
    }

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {
       /* if (mycell->isPDDLVariable() || othercell->isPDDLVariable()) {
            mycell->declPDDLVariable(econtext); othercell->declPDDLVariable(econtext);
            CUtiIDString newvariablename = econtext->executeOperation(mycell->toLisp(NULL), IDSTR("+"), othercell->toLisp(NULL));

            return uti_refptr<CRDataCell>(new CRPDDLVariable(newvariablename));
        }*/

        switch (m_type) {
        case eAdd:{
            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRIntCell(check_cast<CRIntCell*>(mycell)->data+check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRFloatCell(double(check_cast<CRIntCell*>(mycell)->data)+check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data+double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data+check_cast<CRFloatCell*>(othercell)->data) );

            }if (m_type1 == erString && m_type2 == erString) {
                return uti_refptr<CRDataCell>(new CRStringCell(check_cast<CRStringCell*>(mycell)->data+check_cast<CRStringCell*>(othercell)->data) );

            }
        }; break;
        case eSub:
        {
            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRIntCell(check_cast<CRIntCell*>(mycell)->data-check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRFloatCell(double(check_cast<CRIntCell*>(mycell)->data)-check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data-double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data-check_cast<CRFloatCell*>(othercell)->data) );

            }
        }; break;
        case eMul:{
            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRIntCell(check_cast<CRIntCell*>(mycell)->data*check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRFloatCell(double(check_cast<CRIntCell*>(mycell)->data)*check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data*double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data*check_cast<CRFloatCell*>(othercell)->data) );

            }
        }; break;
        case eDiv:{
            if (m_type1 == erInt && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRIntCell(check_cast<CRIntCell*>(mycell)->data/check_cast<CRIntCell*>(othercell)->data ));

            }

            if (m_type1 == erInt && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRFloatCell(double(check_cast<CRIntCell*>(mycell)->data)/check_cast<CRFloatCell*>(othercell)->data ));

            }
            if (m_type1 == erFloat && m_type2 == erInt) {
                return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data/double(check_cast<CRIntCell*>(othercell)->data) ));

            }
            if (m_type1 == erFloat && m_type2 == erFloat) {
                return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data/check_cast<CRFloatCell*>(othercell)->data) );

            }
        }; break;
        };

        return uti_refptr<CRDataCell>();
    }
};
/*
class CRSubTypedOperator : public CRTypedOperator
{
public:

    CRSubTypedOperator(TRBasicDataType t1, TRBasicDataType t2, TRBasicDataType r) : CRTypedOperator(t1, t2, r)
    {

    }

    virtual uti_refptr<CRDataCell> compile(CRAssembler* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {

        return uti_refptr<CRDataCell>();
    }

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {

        if (m_type1 == erInt && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRIntCell(check_cast<CRIntCell*>(mycell)->data-check_cast<CRIntCell*>(othercell)->data ));

        }

        if (m_type1 == erInt && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRFloatCell(double(check_cast<CRIntCell*>(mycell)->data)-check_cast<CRFloatCell*>(othercell)->data ));

        }
        if (m_type1 == erFloat && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data-double(check_cast<CRIntCell*>(othercell)->data) ));

        }
        if (m_type1 == erFloat && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data-check_cast<CRFloatCell*>(othercell)->data) );

        }
        return uti_refptr<CRDataCell>();
    }
};

class CRMulTypedOperator : public CRTypedOperator
{
public:

    CRMulTypedOperator(TRBasicDataType t1, TRBasicDataType t2, TRBasicDataType r) : CRTypedOperator(t1, t2, r)
    {

    }

    virtual uti_refptr<CRDataCell> compile(CRAssembler* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {

        return uti_refptr<CRDataCell>();
    }
    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {

        if (m_type1 == erInt && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRIntCell(check_cast<CRIntCell*>(mycell)->data*check_cast<CRIntCell*>(othercell)->data ));

        }

        if (m_type1 == erInt && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRFloatCell(double(check_cast<CRIntCell*>(mycell)->data)*check_cast<CRFloatCell*>(othercell)->data ));

        }
        if (m_type1 == erFloat && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data*double(check_cast<CRIntCell*>(othercell)->data) ));

        }
        if (m_type1 == erFloat && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data*check_cast<CRFloatCell*>(othercell)->data) );

        }
        return uti_refptr<CRDataCell>();
    }
};

class CRDivTypedOperator : public CRTypedOperator
{
public:

    CRDivTypedOperator(TRBasicDataType t1, TRBasicDataType t2, TRBasicDataType r) : CRTypedOperator(t1, t2, r)
    {

    }
    virtual uti_refptr<CRDataCell> compile(CRAssembler* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {

        return uti_refptr<CRDataCell>();
    }

    virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell)
    {

        if (m_type1 == erInt && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRFloatCell(double(check_cast<CRIntCell*>(mycell)->data)/double(check_cast<CRIntCell*>(othercell)->data )));

        }

        if (m_type1 == erInt && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRFloatCell(double(check_cast<CRIntCell*>(mycell)->data)/check_cast<CRFloatCell*>(othercell)->data ));

        }
        if (m_type1 == erFloat && m_type2 == erInt) {
            return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data/double(check_cast<CRIntCell*>(othercell)->data) ));

        }
        if (m_type1 == erFloat && m_type2 == erFloat) {
            return uti_refptr<CRDataCell>(new CRFloatCell(check_cast<CRFloatCell*>(mycell)->data/check_cast<CRFloatCell*>(othercell)->data) );

        }
        return uti_refptr<CRDataCell>();
    }
};
*/
class CRExpressionOperator :public ReferenceUtiOB
{
public:
    CUtiString m_name;
    bool m_isIdentitifer;
    bool m_binary;
    CReferenceList<CRTypedOperator> m_type;
    TRExpressionType m_expressiontype;
    CRExpressionOperator(const CUtiString &name, TRExpressionType t, bool binary);
    bool isNextThenEat(UtiLexer& parse);
    CRExpressionOperator* addType(CRTypedOperator* t);
    CRTypedOperator* getRType(CRBaseDeclaration* type1, CRBaseDeclaration* type2);
    //  virtual uti_refptr<CRDataCell> execute(CRNormalExecutionContext* econtext, CRDataCell* mycell, CRDataCell* othercell);
};

class CRActionContainer;
class CRActionGenerator;

class CRProgram : public CRBlock
{
public:
    Array2D<uti_refptr<CRExpressionOperator> > m_operators;
    int m_tempindex;
    std::map<TRExpressionType, CRExpressionOperator*> m_operator_to_opdata;
    std::set<CRStructDeclaration*> m_structs;
    CUtiString m_name;
    std::map<CUtiIDString, CRStructDeclaration*> m_nametotable;
    void collectObjects(const Lisp::LObject & params, const Lisp::LObject& types, std::map<CUtiIDString, CUtiIDString>& objects);

    void collectObjects(CRDataCell* data, std::map<CUtiIDString, CUtiIDString> &objects);
public:
    CRProgram();
    void initOperators();
    CUtiIDString getUniqueVarName();
    CRBaseTypeDeclaration*  getBaseType(TRBasicDataType t);
    void loadFromFile(const CUtiString& name);

    void loadFromString(const CUtiString& filename, const std::string& name);
    void compileTo(CRActionContainer *action, const CUtiString& filename);

    FP_term convertVarContentToProlog(const Lisp::LObject& obj);
    void addCommandToProlog(const Lisp::LObject & obj);
    void executePrologEngine(CRAssembler* assembler);
    void executeProgram(const CUtiString& filename);
    bool isTableStruct(CRStructDeclaration* decl);
    Lisp::LObject convertToAction(CRFunctionDeclaration* funcdecl,CRNormalExecutionContext* context, CRPDDLExecutionBlock* block, bool maindata);

};

class CRActionContainer : public ReferenceUtiOB
{
public:
    CRActionContainer();
    virtual bool supportType()=0;
    virtual void addStruct(const CUtiIDString& name, const DynArray1D<CUtiIDString>& params)=0;
    virtual void addType(const CUtiIDString& name, const CUtiIDString& parent)=0;
    virtual CRActionGenerator* createAction(const CUtiIDString& name)=0;
};

enum TRType {ratObject, ratFloat, ratInteger};
enum TRConditionType {ctEqual, ctNotEqual, ctLess, ctGreater, ctEqualLess, ctEqualGreater};
enum TROperatorType {otAdd, otSub, otMul, otDiv, otRound, otSin, otCos, otTan, otSqrt, otSqr, otDivInteger, otMod};
/*
class CRActionGenerator : public ReferenceUtiOB
{
public:
    CRActionGenerator();
    virtual void addParam(const CUtiIDString& str, TRType type)=0;
    virtual void addDatabaseCheck(const CUtiIDString& name,bool positive, const DynArray1D<Lisp::LObject>& params)=0;
    virtual void addDatabaseChange(const CUtiIDString& name,bool add, const DynArray1D<Lisp::LObject>& params)=0;
    virtual void addFluentCondition(const CUtiIDString& var1, TRConditionType type, const CUtiIDString& var2)=0;
    virtual void setFluentVar(const CUtiIDString& target, const CUtiIDString& var1,  TROperatorType type, const CUtiIDString& var2)=0;
    virtual void addWhen(CRActionGenerator* subgenerator, CRActionGenerator* elsesubgenerator)=0;

};*/
/*
enum TREntryType {etParam, etDatabaseCheck, etDatabaseChange, etFluentCondition, etSetFluentVar};

class CRDuplicateEntry : public ReferenceUtiOB
{
public:
    TREntryType m_entrytype;
    CUtiIDString m_name_or_target;
    TRType m_paramtype;
    TRConditionType m_conditiontype;
    TROperatorType m_operatortype;
    CUtiIDString  m_var1, m_var2;
    bool m_positive;
    DynArray1D<Lisp::LObject> m_params;

    CRDuplicateEntry();
};

class CRDuplicateGenerator : public CRActionGenerator
{
protected:
    CReferenceList<CRDuplicateEntry > m_commands;
public:
    CRDuplicateGenerator();

    virtual void addParam(const CUtiIDString& str, TRType type);
    virtual void addDatabaseCheck(const CUtiIDString& name,bool positive, const DynArray1D<Lisp::LObject>& params);
    virtual void addDatabaseChange(const CUtiIDString& name,bool add, const DynArray1D<Lisp::LObject>& params);
    virtual void addFluentCondition(const CUtiIDString& var1, TRConditionType type, const CUtiIDString& var2);
    virtual void setFluentVar(const CUtiIDString& target, const CUtiIDString& var1,  TROperatorType type, const CUtiIDString& var2);
    virtual void addWhen(CRActionGenerator* subgenerator, CRActionGenerator* elsesubgenerator) {assert(false);}
    virtual CRActionGenerator* createOn(CRActionContainer* container, const CUtiIDString &name);
};
*/
#endif // SYNTHETICRULE_H
