
/***************************************************************************
 *   Copyright (C) 2017 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "utilisp.h"
#include "lispnames.h"
#include "utilexer.h"
#include "utisystem.h"

#include <algorithm>
#include <iterator>

namespace Lisp {

UTI_MEMORY_POOL_IMPL(LCell, 65536, SMemoryPool_BaseBehaviour);

UTI_MEMORY_POOL_IMPL(LInteger, 65536, SMemoryPool_BaseBehaviour);

UTI_MEMORY_POOL_IMPL(LFloat, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LTrue, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LReturn, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LNil, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LAtom, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LProc, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LLambda, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LString, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LList, 65536, SMemoryPool_BaseBehaviour);
UTI_MEMORY_POOL_IMPL(LCons, 65536, SMemoryPool_BaseBehaviour);


LGlobalEnvironment* LGlobalEnvironment::s_globalenvironment = NULL;


LGlobalEnvironment* LGlobalEnvironment::init()
{
   s_globalenvironment = new LGlobalEnvironment();
   return s_globalenvironment;
}

bool LCell::l_equal(LCell* other) const
{
    assert(false);
    return false;
}

bool LCell::l_eql(LCell* other) const
{
   return this == other;
}

int LCell::l_hash() const
{
  assert(false);
  return false;
}

bool LCell::l_smaller(LCell* other) const
{
    assert(false);
    return false;
}

LCell* LCell::l_createCopy() const
{
    assert(false);
    return NULL;
}


LObject LCell::l_eval() const
{
 return this;
}


HResult WriteCell(CStream & Stream, LCell* cell, int Flags, int version)
{
    int celltype = cell->m_type;
    Stream.writeInt32(celltype);
    switch (cell->m_type) {
       case tctNone: {

        assert(false);
    }; break;
     case tctFloat: {
        LFloat* f = check_cast<LFloat*>(cell);
        Stream.WriteDouble(f->m_value);
    }; break;
     case tctInteger: {
        LInteger* f = check_cast<LInteger*>(cell);
        Stream.WriteInteger(f->m_value);

    }; break;
     case tctReturn: {
        LReturn* f = check_cast<LReturn*>(cell);
        WriteString(Stream, f->m_name.str(), version);
        WriteCell(Stream, f->m_value.m_data.get(), Flags, version);

    }; break;
     case tctAtom: {
        LAtom* a = check_cast<LAtom*>(cell);
        WriteString(Stream, a->m_value.str(), version);

    }; break;
     case tctCons: {
        LCons* a = check_cast<LCons*>(cell);

        WriteCell(Stream, a->m_head.m_data.get(), Flags, version);
        WriteCell(Stream, a->m_tail.m_data.get(), Flags, version);

    }; break;
     case tctString: {
        LString* a = check_cast<LString*>(cell);
        WriteString(Stream, a->m_value, version);


    }; break;
     case tctArray: {
        assert(false);

    }; break;
     case tctMultiArray: {
        assert(false);

    }; break;
     case tctDynamicArray: {
        assert(false);

    }; break;
     case tctTrue: {
        LTrue* a = check_cast<LTrue*>(cell);

    }; break;
     case tctLambda: {
        LLambda* l = check_cast<LLambda*>(cell);

        WriteCell(Stream, l->m_args.m_data.get(), Flags, version);
        WriteCell(Stream, l->m_sexp.m_data.get(), Flags, version);

        assert(false);
    }; break;
     case tctProc: {
        LProc* p = check_cast<LProc*>(cell);


        assert(false);
    }; break;
     case tctNil: {
        LNil* a = check_cast<LNil*>(cell);

    }; break;
    default: {
        assert(false);
    }; break;
    }

    return 0;
}
LCell* ReadCell(CStream & Stream, int Flags, int version)
{

    LCell* ret = NULL;
    int celltype = Stream.getInt32();
    switch (celltype) {
       case tctNone: {
        assert(false);
    }; break;
     case tctFloat: {
        LFloat* f = new LFloat();

        Stream.ReadDouble(f->m_value);
        ret = f;
    }; break;
     case tctInteger: {
        LInteger* f = new LInteger();
        Stream.ReadInteger(f->m_value);
        ret = f;

    }; break;
     case tctReturn: {
        LReturn* f = new LReturn();
        CUtiString s;
        ReadString(Stream, s, version);
        f->m_name = CUtiIDString(s);
        f->m_value = ReadCell(Stream, Flags, version);
        ret = f;

    }; break;
     case tctAtom: {
        LAtom* a = new LAtom();
        CUtiString s;
        ReadString(Stream, s, version);
        a->m_value = CUtiIDString(s);

        ret = a;
    }; break;
     case tctCons: {
        LCons* a = new LCons();

        a->m_head = ReadCell(Stream, Flags, version);
        a->m_tail = ReadCell(Stream, Flags, version);

        ret = a;
    }; break;
     case tctString: {
        LString* a = new LString();
        ReadString(Stream, a->m_value, version);

        ret = a;

    }; break;
     case tctArray: {
        assert(false);

    }; break;
     case tctMultiArray: {
        assert(false);

    }; break;
     case tctDynamicArray: {
        assert(false);

    }; break;
     case tctTrue: {
        LTrue* a = new LTrue();

        ret = a;
    }; break;
     case tctLambda: {
        LLambda* l = NULL;//new LLambda();

        l->m_args.m_data = ReadCell(Stream,  Flags, version);
        l->m_sexp.m_data=ReadCell(Stream,  Flags, version);

        assert(false);
        ret = l;
    }; break;
     case tctProc: {
        LProc* p = NULL;//new LProc();


        assert(false);
    }; break;
     case tctNil: {
        LNil* a = new LNil();

        ret = a;
    }; break;
    default: {
        assert(false);
    }; break;
    }


    return ret;
}

LCell* xreadFact(CXNode* node, CWriteReadInterface* wri)
{
    LCell* ret = NULL;
    if (node == NULL) return new LNil();
    int celltype = node->getPropInt(UTISTR("celltype"), tctNone);
    switch (celltype) {
       case tctNone: {
        assert(false);
    }; break;
     case tctFloat: {
        LFloat* f = new LFloat();

        f->m_value = node->getPropDouble(UTISTR("value"), 0);
       // Stream.ReadDouble(f->m_value);
        ret = f;
    }; break;
     case tctInteger: {
        LInteger* f =new LInteger();
        f->m_value = node->getPropInt(UTISTR("value"), 0);
      //  Stream.ReadInteger(f->m_value);
        ret = f;

    }; break;
     case tctReturn: {
        LReturn* f = new LReturn();
      //  CUtiString s;
       // ReadString(Stream, s, version);
        f->m_name = CUtiIDString(node->getPropString(UTISTR("name"), EMPTYSTRING));
        f->m_value = xreadFact(node->getNodeByName(UTISTR("value")), wri);
        ret = f;

    }; break;
     case tctAtom: {
        LAtom* a = new LAtom();

        a->m_value =  CUtiIDString(node->getPropString(UTISTR("value"), EMPTYSTRING));

        ret = a;
    }; break;
     case tctCons: {
        LCons* a = new LCons();

        a->m_head = xreadFact(node->getNodeByName(UTISTR("head")), wri);
        a->m_tail = xreadFact(node->getNodeByName(UTISTR("tail")), wri);

        ret = a;
    }; break;
     case tctString: {
        LString* a = new LString();
     //   ReadString(Stream, a->m_value, version);
        a->m_value = node->getPropString(UTISTR("value"), EMPTYSTRING);

        ret = a;

    }; break;
     case tctArray: {
        assert(false);

    }; break;
     case tctMultiArray: {
        assert(false);

    }; break;
     case tctDynamicArray: {
        assert(false);

    }; break;
     case tctTrue: {
        LTrue* a = new LTrue();

        ret = a;
    }; break;
     case tctLambda: {
        LLambda* l = NULL;// check_cast<LLambda*>(cell);

        l->m_args.m_data = xreadFact(node->getNodeByName(UTISTR("args")), wri);
        l->m_sexp.m_data=xreadFact(node->getNodeByName(UTISTR("sexp")), wri);

        assert(false);
        ret = l;
    }; break;
     case tctProc: {
        LProc* p = NULL;//check_cast<LProc*>(cell);


        assert(false);
    }; break;
     case tctNil: {
        LNil* a = new LNil();

        ret = a;
    }; break;
    default: {
        assert(false);
    }; break;
    }


    if (ret == NULL) return new LNil();
    return ret;
}
void xwriteFact(LCell* cell, CXNode* node, CWriteReadInterface* wri)
{
    int celltype = cell->m_type;
    node->setPropInt(UTISTR("celltype"), celltype);
    switch (cell->m_type) {
       case tctNone: {

        assert(false);
    }; break;
     case tctFloat: {
        LFloat* f = check_cast<LFloat*>(cell);
        node->setPropDouble(UTISTR("value"), f->m_value);
        //Stream.WriteDouble(f->m_value);
    }; break;
     case tctInteger: {
        LInteger* f = check_cast<LInteger*>(cell);
        node->setPropInt(UTISTR("value"), f->m_value);
       // Stream.WriteInteger(f->m_value);

    }; break;
     case tctReturn: {
        LReturn* f = check_cast<LReturn*>(cell);
        node->setPropString(UTISTR("name"), f->m_name.str());
        //WriteString(Stream, f->m_name.str(), version);
        xwriteFact(f->m_value.m_data.get(), node->addNode(UTISTR("value")), wri);

    }; break;
     case tctAtom: {
        LAtom* a = check_cast<LAtom*>(cell);
        node->setPropString(UTISTR("value"), a->m_value.str());
       // WriteString(Stream, a->m_value.str(), version);

    }; break;
     case tctCons: {
        LCons* a = check_cast<LCons*>(cell);

        xwriteFact(a->m_head.m_data.get(), node->addNode(UTISTR("head")), wri);
        xwriteFact(a->m_tail.m_data.get(), node->addNode(UTISTR("tail")), wri);
      //  WriteCell(Stream, a->m_head.m_data.get(), Flags, version);
      //  WriteCell(Stream, a->m_tail.m_data.get(), Flags, version);

    }; break;
     case tctString: {
        LString* a = check_cast<LString*>(cell);
        node->setPropString(UTISTR("value"), a->m_value);
        //WriteString(Stream, a->m_value, version);


    }; break;
     case tctArray: {
        assert(false);

    }; break;
     case tctMultiArray: {
        assert(false);

    }; break;
     case tctDynamicArray: {
        assert(false);

    }; break;
     case tctTrue: {
        LTrue* a = check_cast<LTrue*>(cell);

    }; break;
     case tctLambda: {
        LLambda* l = check_cast<LLambda*>(cell);

        xwriteFact(l->m_args.m_data.get(), node->addNode(UTISTR("args")), wri);
        xwriteFact(l->m_sexp.m_data.get(), node->addNode(UTISTR("sexp")), wri);
       // WriteCell(Stream, l->m_args.m_data.get(), Flags, version);
       // WriteCell(Stream, l->m_sexp.m_data.get(), Flags, version);

        assert(false);
    }; break;
     case tctProc: {
        LProc* p = check_cast<LProc*>(cell);


        assert(false);
    }; break;
     case tctNil: {
        LNil* a = check_cast<LNil*>(cell);

    }; break;
    default: {
        assert(false);
    }; break;
    }

}

bool LObject::iterator::operator==(const iterator& it)
{
    if (m_data->m_type == tctNil && it.m_data->m_type == tctNil) {
        return true;
    };
    return m_data == it.m_data;
}

bool LObject::iterator::operator!=(const iterator& it)
{
    if (m_data->m_type == tctNil && it.m_data->m_type == tctNil) {
        return false;
    };
    return !(m_data == it.m_data);

}


void LObject::toList(DynArray1D<LObject>& subids)
{
    subids.clear();
LObject obj = *this;
for (auto p: obj) {
     subids.push_back(p);
};
}


std::string LObject::createDebugData() const
{
  return UtiFramework::converttomulti(print());
}

LObject::iterator& LObject::iterator::operator=(const iterator& it)
{

    m_data = it.m_data;
    return *this;
}
LObject::iterator& LObject::iterator::operator++(int)
{


m_data = check_cast<LList*>(m_data.get())->l_cdr().m_data;
    return *this;
}

LObject::iterator& LObject::iterator::operator++()
{

    m_data = check_cast<LList*>(m_data.get())->l_cdr().m_data;
        return *this;
}

LObject LObject::iterator::operator*() const
{
   return Lisp::car(LObject(m_data.get()));
}


LObject::iterator LObject::begin() const
{
    iterator iter;

    iter.m_data =const_cast<LCell*>(m_data.get());

    return iter;
}
LObject::iterator LObject::end() const
{
    iterator iter;

    iter.m_data = new LNil();

    return iter;
}

LObject::iterator LObject::begin()
{
    iterator iter;

    iter.m_data = m_data;

    return iter;
}
LObject::iterator LObject::end()
{
    iterator iter;

    iter.m_data = new LNil();;

    return iter;
}


struct l_iterator
{
public:
    LObject m_currentelement;
    inline l_iterator()
    {};
    inline l_iterator(const LObject& el)
    {
        m_currentelement = el;

    }
    inline bool end()
    {
        return m_currentelement.isNil();
    }
    inline LObject getObject() {return car(m_currentelement);}
    inline LObject getList() {return m_currentelement;}
    inline void next()
    {
        assert(m_currentelement.isNil() == false);
        m_currentelement = cdr(m_currentelement);
    }
    inline bool operator==(const l_iterator& iter)
    {
        return m_currentelement == iter.m_currentelement;
    }
};


int length(const LObject& obj)
{
    if (obj.isNil())return 0;
    if (!obj.isConList()) throw LispException(obj,UTISTR("Is not a sequence"));
    int count=1;
    LCons* c = check_cast<LCons*>(obj.m_data.get());
    while (c->m_tail.isConList()) {
        count++;
        c = check_cast<LCons*>(c->m_tail.m_data.get());
    }
    return count;

  //  return check_cast<LList*>(obj.m_data.get())->l_length();
}

int LList::l_length()
{
    int l = 0;
    l_iterator iter(this);
    while(!iter.end()) {
        l++;
        iter.next();

    }
  return l;
}

int LList::l_count(const LObject& obj)
{

    int l = 0;
    l_iterator iter(this);
    while(!iter.end()) {
        if (iter.getObject() == obj) {
        l++;
        }
        iter.next();

    }


    return l;
}

LObject LList::l_nth(int i)
{
    l_iterator iter(this);
    while(!iter.end()) {
        if (i == 0) return iter.getObject();
        i--;
        iter.next();

    }
    return LObject();
}

inline LObject setTailAndReturnNewEnd(LObject lastelement, LObject newsegment)
{
   LCons* cons = check_cast<LCons*>(lastelement.m_data.get());

   cons->m_tail = newsegment;

   return newsegment;
}

struct ListGenerator
{
    LObject list;
    LObject lastelement;
    inline void push_back(const LObject& obj)
    {
        if (list.isNil()) {
            list = cons(obj);
            lastelement = list;
        } else {
            lastelement = setTailAndReturnNewEnd(lastelement, cons(obj));

        }
    }
    inline void push_backlist(const LObject& endelement)
    {
        if (endelement.isNil()) return;
         setTailAndReturnNewEnd(lastelement, endelement);
         lastelement.setNil();
    }

    inline LObject getList()
    {
        return list;
    }
    inline LObject getListAsType(const LObject& type)
    {
        return list;
    }
};

LObject LList::l_subseq(int start, int end)
{

    if (end < start) throw LispException(LObject(this),UTISTR("subseq: it must hold start < end"));

    ListGenerator list;
    l_iterator iter(this);
    int i = 0;
    while(!iter.end()) {
        if (start <= i && i < end) {
            list.push_back(iter.getObject());
        };
        i++;
        if (end == i) {
            break;
        }
        iter.next();

    }
    return list.getList();
}

LObject LList::l_filter(const LObject& obj, LFilterCallback* callback )
{
   //Erstelle neue List bis zum letzten Zielelement

    ListGenerator list;
    l_iterator iter(this);
    l_iterator lastok(this);
    int i = 0;
    while(!iter.end()) {
        LObject obj = iter.getObject();
        if (callback->eval(obj)) {
           // list.push_back(obj);
        } else {
            while (!(iter == lastok)) {
                list.push_back(lastok.getObject());
                lastok.next();

            }
            lastok.next(); //uberspringe aktuelles

        }
        iter.next();

    }
    list.push_backlist(lastok.m_currentelement);
    return list.getList();

}


LObject LList::l_map(const LObject& obj, LMapCallback *callback)
{
    ListGenerator list;
    l_iterator iter(this);
    int i = 0;
    while(!iter.end()) {
        LObject obj = iter.getObject();
            list.push_back(callback->eval(obj));
        iter.next();

    }
    return list.getList();
}


LObject list(const DynArray1D<LObject> & elements)
{

     ListGenerator list;

     for (int i = 0; i< elements.size(); i++) {
         list.push_back(elements[i]);
     }

     return list.getList();
}

LObject list(const std::set<LObject> & elements)
{
    ListGenerator list;

    for (std::set<LObject>::iterator i = elements.begin(); i != elements.end(); i++) {
        list.push_back(*i);
    }

    return list.getList();
}

void listToArray(const LObject& list, DynArray1D<LObject> & elements)
{
    l_iterator iter(list);

    while(!iter.end()) {
        LObject obj = iter.getObject();
        elements.push_back(obj);
        iter.next();

    }
}

void listToArray(const LObject& list, std::set<LObject> & elements)
{
    l_iterator iter(list);

    while(!iter.end()) {
        LObject obj = iter.getObject();
        elements.insert(obj);
        iter.next();
    }
}

LObject lunion(const LObject& l1, const LObject& l2)
{
    DynArray1D<LObject> a1;
    listToArray(l1, a1);
    DynArray1D<LObject> a2;
    listToArray(l1, a2);
    std::sort(a1.begin(), a1.end());
    std::sort(a2.begin(), a2.end());
    DynArray1D<LObject> a3;
    a3.resize(a1.size()+a2.size());

    DynArray1D<LObject>::iterator it = std::set_union(a1.begin(), a1.end(), a2.begin(), a2.end(), a3.begin());
    a3.resize(it - a3.begin());
    return list(a1);

}



LObject lsort(const LObject& l1)
{
    DynArray1D<LObject> a1;
    listToArray(l1, a1);
    std::sort(a1.begin(), a1.end());
    return list(a1);
}


LObject reverse(const LObject& l1)
{
    DynArray1D<LObject> a1;
    listToArray(l1, a1);
    ListGenerator list;

    for (int i = 0; i< a1.size(); i++) {
        list.push_back(a1[a1.size()-i-1]);
    }

    return list.getList();
}

LObject lintersection(const LObject& l1, const LObject& l2)
{
    DynArray1D<LObject> a1;
    listToArray(l1, a1);
    DynArray1D<LObject> a2;
    listToArray(l1, a2);
    std::sort(a1.begin(), a1.end());
    std::sort(a2.begin(), a2.end());
    DynArray1D<LObject> a3;
    a3.resize(a1.size()+a2.size());

    DynArray1D<LObject>::iterator it = std::set_intersection(a1.begin(), a1.end(), a2.begin(), a2.end(), a3.begin());
    a3.resize(it - a3.begin());
    return list(a1);
}

LObject lset_difference(const LObject& l1, const LObject& l2)
{
    DynArray1D<LObject> a1;
    listToArray(l1, a1);
    DynArray1D<LObject> a2;
    listToArray(l1, a2);
    std::sort(a1.begin(), a1.end());
    std::sort(a2.begin(), a2.end());
    DynArray1D<LObject> a3;
    a3.resize(a1.size()+a2.size());

    DynArray1D<LObject>::iterator it = std::set_difference(a1.begin(), a1.end(), a2.begin(), a2.end(), a3.begin());
    a3.resize(it - a3.begin());
    return list(a1);
}


LObject push(const LObject& obj, LObject& list)
{
     list.m_data = cons(obj, list).m_data;
      list.updateDebug();
     return list;
}

LObject pop(LObject& list)
{
   list.m_data = cdr(list).m_data;
   list.updateDebug();
   return list;
}


bool LList::l_equal(LCell* other) const
{

    if (m_type != other->m_type)return false;
#ifndef NDEBUG
    if (other->m_type != m_type && LObject(this).isConList() == LObject(other).isConList()) {
        assert(false);

    }
#endif
    if (other->m_type != m_type) throw LispException(other,UTISTR("List expression expected"));
   l_iterator iter(LObject(this));
   LObject o(other);
   l_iterator iter2(o);

   if (iter.end() != iter2.end()) return false;
   while (!iter.end()) {

       if (iter.getObject() != iter2.getObject()) return false;
       iter.next();
       iter2.next();
       if (iter.end() != iter2.end()) return false;

   }
   return true;
   /* if (m_type < other->m_type)return true; if (m_type > other->m_type)return false;
#ifndef NDEBUG
    if (other->m_type != m_type && LObject(this).isConList() == LObject(other).isConList()) {
        assert(false);

    }
#endif
    if (other->m_type != m_type) throw LispException(other,UTISTR("List expression expected"));
   l_iterator iter(LObject(this));
   LObject o(other);
   l_iterator iter2(o);

   if (iter.end() != iter2.end()) return false;
   while (!iter.end()) {

       if (iter.getObject() != iter2.getObject()) return false;
       iter.next();
       iter2.next();
       if (iter.end() != iter2.end()) return false;

   }
   return true;*/

}
int LList::l_hash() const
{

    int hashvalue = 5381;
 l_iterator iter(LObject(this));

 while (!iter.end()) {

     int hashcode2 = iter.getObject().hashCode();
     hashvalue = ((hashvalue << 5) + hashvalue) + hashcode2;
     iter.next();

 }
 return hashvalue;

}


LObject LAtom::l_eval() const
{
     return LGlobalEnvironment::s_globalenvironment->getVariable(m_value);
}

LObject adjoin(const LObject& item, const LObject& list)
{

     if (member(item, list)) {
       return list;
     } else {
         return cons(item, list);
     };


}

#define PARSE_KEYS1(keys, atom1, obj1) \
LObject obj1;\
{l_iterator it(keys);\
while (!it.end()) {\
    switch (it.getObject().getAtom().getID()) {\
        case atom1: {\
           it.next();\
           obj1 = it.getObject();\
        }; break;\
    default: {throw new LispException(it.getObject(), UTISTR("Unknown key: ")); };\
    };\
    it.next();\
}\
}

#define PARSE_KEYS2(keys, atom1, obj1, atom2, obj2) \
LObject obj1;\
LObject obj2;\
{l_iterator it(keys);\
while (!it.end()) {\
    switch (it.getObject().getAtom().getID()) {\
        case atom1: {\
           it.next();\
           obj1 = it.getObject();\
        }; break;\
       case atom2: {\
           it.next();\
           obj2 = it.getObject();\
       }; break;\
    default: {throw new LispException(it.getObject(), UTISTR("Unknown key: ")); };\
    };\
    it.next();\
}\
}

#define PARSE_KEYS3(keys, atom1, obj1, atom2, obj2, atom3, obj3) \
LObject obj1;\
LObject obj2;\
LObject obj3;\
{l_iterator it(keys);\
while (!it.end()) {\
    switch (it.getObject().getAtom().getID()) {\
        case atom1: {\
           it.next();\
           obj1 = it.getObject();\
        }; break;\
       case atom2: {\
           it.next();\
           obj2 = it.getObject();\
       }; break;\
       case atom3: {\
          it.next();\
          obj3 = it.getObject();\
       }; break;\
    default: {throw new LispException(it.getObject(), UTISTR("Unknown key: ")); };\
    };\
    it.next();\
}\
}

#define PARSE_KEYS7(keys, atom1, obj1, atom2, obj2, atom3, obj3, atom4, obj4, atom5, obj5, atom6, obj6, atom7, obj7) \
LObject obj1;\
LObject obj2;\
LObject obj3;\
    LObject obj4;\
    LObject obj5;\
    LObject obj6;\
    LObject obj7;\
{l_iterator it(keys);\
while (!it.end()) {\
    switch (it.getObject().getAtom().getID()) {\
        case atom1: {\
           it.next();\
           obj1 = it.getObject();\
        }; break;\
       case atom2: {\
           it.next();\
           obj2 = it.getObject();\
       }; break;\
       case atom3: {\
          it.next();\
          obj3 = it.getObject();\
       }; break;\
    case atom4: {\
        it.next();\
       obj4 = it.getObject();\
    }; break;\
    case atom5: {\
       it.next();\
       obj5 = it.getObject();\
    }; break;\
    case atom6: {\
       it.next();\
       obj6 = it.getObject();\
    }; break;\
    case atom7: {\
       it.next();\
       obj7 = it.getObject();\
    }; break;\
    default: {throw new LispException(it.getObject(), UTISTR("Unknown key: ")); };\
    };\
    it.next();\
}\
}



LObject adjoin(const LObject& item, const LObject& list, const LObject& keys)
{
    PARSE_KEYS3(keys, LISP_d_key, keyfn, LISP_d_test, testfn, LISP_d_test_not, testnotfn);

       if (keyfn.isNil() == false) {
       LObject keyitem = funccall(keyfn, quote(item));


       if (member(keyitem, list, keys)) {
         return list;
       } else {
           return cons(item, list);
       };
       } else {
           if (member(item, list, keys)) {
             return list;
           } else {
               return cons(item, list);
           };
       }

}

bool member(const LObject& item, const LObject& list)
{
    l_iterator iter;
    iter.m_currentelement = (LObject(list));
    while (!iter.end()) {
        if (iter.getObject() == item) return true;
    };
    return false;
}

bool member(const LObject& item, const LObject& list, const LObject& keys)
{


    PARSE_KEYS3(keys, LISP_d_key, keyfn, LISP_d_test, testfn, LISP_d_test_not, testnotfn);


    l_iterator iter;
    iter.m_currentelement = (LObject(list));

       while (!iter.end()) {
           LObject obj = iter.getObject();
           if (keyfn.isNil() == false) {
               obj =funccall(keyfn, quote(obj));
           }
           if (testfn.isNil() == false) {
               if (funccall(testfn, quote(iter.getObject()), quote(item)).toBool()) return true;
           } else if (testnotfn.isNil() == false) {
               if (!funccall(testnotfn, quote(iter.getObject()), quote(item)).toBool()) return true;
           } else {
               if (iter.getObject() == item) return true;

           }
       };

    return false;
}

struct LObjectCompare
{
    LObject m_predicate;
    bool operator()(const LObject& v1, const LObject& v2) const
    {
        return funccall(m_predicate, quote(v1), quote(v2)).toBool();
    }
};

struct LObjectCompareKey
{
    LObject m_predicate;
    LObject m_key;
    bool operator()(const LObject& v1, const LObject& v2) const
    {
        return funccall(m_predicate, quote(funccall(m_key, quote(v1))), quote(funccall(m_key, quote(v2)))).toBool();
    }
};


LObject sort(const LObject& l1, const LObject& predicate)
{
    DynArray1D<LObject> a1;
    listToArray(l1, a1);
    LObjectCompare comparecode;
    comparecode.m_predicate = predicate;
    std::sort(a1.begin(), a1.end(), comparecode);
    return list(a1);
}

LObject sort(const LObject& l1, const LObject& predicate, const LObject& keys)
{
    PARSE_KEYS1(keys, LISP_d_key, keyfn);
    if (keyfn.isNil()) {
       return sort(l1, predicate);
    } else {
        DynArray1D<LObject> a1;
        listToArray(l1, a1);
        LObjectCompareKey comparecode;
        comparecode.m_predicate = predicate;
        comparecode.m_key = keys;


        std::sort(a1.begin(), a1.end(), comparecode);

        return list(a1);
    }

}

struct l_keyedoperator
{

    l_iterator iter;
    bool m_fromend;
    LObject m_testfn;
    LObject m_testnotfn;
    int m_start;
    int m_end;
    int m_count;
    LObject m_key;
    bool m_replace;
    int m_currentindex;
    LObject m_searchkey;
    l_keyedoperator()
    {

        m_start = -1;
        m_end = -1;
        m_fromend = false;
        m_currentindex = 0;
        m_replace = false;
        m_count = 1000000000;
    };
    inline void init(const LObject& sequence, const LObject& keys, const LObject& olditem, bool inputisalreadykey)
    {
        PARSE_KEYS7(keys, LISP_d_from_end, fromend, LISP_d_test, testfn, LISP_d_test_not, testnotfn,
                    LISP_d_start, start, LISP_d_end, end, LISP_d_count, count, LISP_d_key, key);



        if (!fromend.isNil()) m_fromend = fromend.toBool();
        m_testfn = testfn;
        m_testnotfn = testnotfn;
        if (!start.isNil()) m_start = start.toInt();
        if (!end.isNil()) m_end = end.toInt();
        if (!count.isNil()) m_count = count.toInt();
        m_key = key;
        iter.m_currentelement = sequence;

        m_searchkey = olditem;
        if (!inputisalreadykey) {
        if (!key.isNil()) {m_searchkey = funccall(key, quote(olditem));};
        }
    }
    inline bool canAbort() const
    {
        return m_count == 0 || m_currentindex >= m_end;
    }
    inline void jumpToNextCandidate()
    {
      if (m_start != -1 && m_currentindex < m_start) {
          while (m_currentindex < m_start) {
              next();
          };

      };

    };
    inline void next()
    {
        iter.next();
        m_currentindex++;
    }
    inline bool end() {return iter.end();};
    std::pair<LObject, bool> getObject()
    {
      LObject obj = iter.getObject();
      bool replace = false;

      if (m_count <= 0 || (m_start != -1 && m_currentindex < m_start) || (m_end != -1 && m_currentindex >= m_end)) {

      } else {
        LObject l = iter.getObject();
        if (!m_key.isNil()) {
           l = funccall(m_key, quote(l));
        }
         if (!m_testfn.isNil()) {
             replace = funccall(m_testfn, quote(m_searchkey), quote(l)).toBool();


         } else if (!m_testnotfn.isNil()) {
             replace = funccall(m_testnotfn, quote(m_searchkey), quote(l)).toBool();

         } else {
                 replace = (l == m_searchkey);
         }
      }
      if (replace) {
        m_count--;
      };

      return std::make_pair(obj, replace);
    };
};

LObject substitute(const LObject& newitem, const LObject& olditem, const LObject& sequence, const LObject& keys)
{
    ListGenerator gen;
    l_keyedoperator iter;
    iter.init(sequence, keys, olditem, false);
    while (!iter.end()) {

        std::pair<LObject, bool> obj = iter.getObject();
        if (obj.second) {
            gen.push_back(newitem);
        }else {

            gen.push_back(obj.first);
        }
        iter.next();
    };
    return gen.getList();
}


int count(const LObject& olditem, const LObject& sequence, const LObject& keys)
{
 int c = 0;
 l_keyedoperator iter;
 iter.init(sequence, keys, olditem, false);
 while (!iter.end()) {

     std::pair<LObject, bool> obj = iter.getObject();
     if (obj.second) {
         c++;
     }else {

     }
     iter.next();
 };
 return c;
}

int positioncpp(const LObject& olditem, const LObject& sequence, const LObject& keys)
{
    l_keyedoperator iter;
    iter.init(sequence, keys, olditem, false);
    int index = 0;
    while (!iter.end()) {

        std::pair<LObject, bool> obj = iter.getObject();
        if (obj.second) {
           return index;
        }else {

        }
        index++;
        iter.next();
    };
    return -1;
}


LObject map(const LObject& resulttype, const LObject& function, const LObject& sequences)
{
    DynArray1D<l_iterator> seq;

    l_iterator totalseq;
    totalseq.m_currentelement = sequences;
    while (!totalseq.end()) {
        l_iterator iter;
        iter.m_currentelement = totalseq.getObject();
        seq.push_back(iter);
        totalseq.next();
    };
    if (seq.size() == 0) return LObject();
    ListGenerator generator;

    while (!seq[0].end()) {
        ListGenerator paramlist;
        for (int i = 0; i < seq.size(); i++) {
            paramlist.push_back(quote(seq[i].getObject()));
        };
        generator.push_back(funcall(function,paramlist.getList()));


        for (int i = 0; i < seq.size(); i++) {
             seq[i].next();
        };
    };

    return generator.getListAsType(resulttype);
};

LObject position(const LObject& olditem, const LObject& sequence, const LObject& keys)
{
    l_keyedoperator iter;
    iter.init(sequence, keys, olditem, false);
    int index = 0;
    while (!iter.end()) {

        std::pair<LObject, bool> obj = iter.getObject();
        if (obj.second) {
           return index;
        }else {

        }
        index++;
        iter.next();
    };
    return LObject();
}

LObject find(const LObject& olditem, const LObject& sequence, const LObject& keys)
{
    l_keyedoperator iter;
    iter.init(sequence, keys, olditem, false);
    while (!iter.end()) {

        std::pair<LObject, bool> obj = iter.getObject();
        if (obj.second) {
           return obj.first;
        }else {

        }
        iter.next();
    };
    return LObject();
}

LObject remove(const LObject& olditem, const LObject& sequence, const LObject& keys)
{
    ListGenerator gen;
    l_keyedoperator iter;
    iter.init(sequence, keys, olditem, false);
    while (!iter.end()) {

        std::pair<LObject, bool> obj = iter.getObject();
        if (obj.second) {
        }else {

            gen.push_back(obj.first);
        }
        iter.next();
    };
    return gen.getList();
}


LObject l_delete(const LObject& item,  const LObject& sequence, const LObject& keys)
{
    LObject seq;
    LObject prev;
    l_keyedoperator iter;
    iter.init(sequence, keys, item, true);
    while (!iter.end()) {

        std::pair<LObject, bool> obj = iter.getObject();
        if (obj.second) {
            if (iter.iter.m_currentelement == seq) {
                assert(prev.isNil());
                iter.next();
              seq = iter.iter.m_currentelement;
            } else {
                assert(cdr(prev) == iter.iter.m_currentelement);
                iter.next();
                check_cast<LList*>(prev.m_data.get())->l_setcdr(iter.iter.m_currentelement);

            }
            continue;
        };
        prev = iter.iter.m_currentelement;
        iter.next();
        if (iter.canAbort()) break;
    };
    return seq;
}

LObject append(const LObject & v2)
{

    ListGenerator gen;
    l_iterator iter;
    iter.m_currentelement = LObject(v2);
    while (!iter.end()) {

        LObject sublist = iter.getObject();
        l_iterator iter2(sublist);
         while (!iter2.end()) {
             gen.push_back(iter2.getObject());

             iter2.next();

         };
         iter.next();
    };
    return gen.getList();
}


LObject LList::l_eval() const
{
    l_iterator iter(LObject(this));
    LObject obj = iter.getObject();
    CUtiIDString id;
    if (obj.isAtom()) {

         id = obj.getAtom();
    };


    iter.next();

#define CH_END if (!iter.end()) { throw new LispException(LObject(this), UTISTR("Too many arguments"));};
#define CH_PAR(r) if (iter.end()) { throw new LispException(LObject(this), UTISTR("Too less arguments"));}; LObject r = iter.getObject(); iter.next();
#define CH_PAR_NO_EVAL(r) if (iter.end()) { throw new LispException(LObject(this), UTISTR("Too less arguments"));}; LObject r = iter.getObject(); iter.next();
#define CH_REST(r)  LObject r = iter.getList(); iter.next();

     switch (id.getID()) {
#define CREATE_LISP_NAME(a, b)// \
   //    if (LISP_ ## a !=  convertNameToInt(UTISTR(b))) return false; //3

#define CREATE_LISP_NAME_P0_M(a, b) case LISP_ ## a: { CH_END;  return a(); }; break;
#define CREATE_LISP_NAME_P1_M(a, b) case LISP_ ## a: { CH_PAR(p1); CH_END; return a(p1); }; break;
#define CREATE_LISP_NAME_P1_CALLOTHER_M(a, b, c) case LISP_ ## a: { CH_PAR_NO_EVAL(p1); CH_END; return c(p1); }; break;
#define CREATE_LISP_NAME_P2_M(a, b) case LISP_ ## a: { CH_PAR(p1); CH_PAR(p2);CH_END; return a(p1, p2); }; break;
#define CREATE_LISP_NAME_P3_M(a, b) case LISP_ ## a: { CH_PAR(p1); CH_PAR(p2);CH_PAR(p3);CH_END; return a(p1, p2, p3); }; break;
#define CREATE_LISP_NAME_P0_REST_M(a, b) case LISP_ ## a: {  CH_REST(rest); return a(rest); }; break;
#define CREATE_LISP_NAME_P1_REST_M(a, b) case LISP_ ## a: {CH_PAR(p1); CH_REST(rest); return a(p1, rest); }; break;
#define CREATE_LISP_NAME_P2_REST_M(a, b) case LISP_ ## a: { CH_PAR(p1); CH_PAR(p2); CH_REST(rest); return a(p1, p2, rest); }; break;
#define CREATE_LISP_NAME_P3_REST_M(a, b) case LISP_ ## a: {  CH_PAR(p1); CH_PAR(p2); CH_PAR(p3); CH_REST(rest); return a( p1, p2, p3, rest); }; break;


#define CREATE_LISP_NAME_P0_KEY_M(a, b) case LISP_ ## a: {CH_REST(rest); return a(rest); }; break;
#define CREATE_LISP_NAME_P1_KEY_M(a, b) case LISP_ ## a: {  CH_PAR(p1); CH_REST(rest); return a(p1, rest); }; break;
#define CREATE_LISP_NAME_P2_KEY_M(a, b) case LISP_ ## a: { CH_PAR(p1); CH_PAR(p2); CH_REST(rest); return a(p1, p2, rest); }; break;
#define CREATE_LISP_NAME_P3_KEY_M(a, b) case LISP_ ## a: {  CH_PAR(p1); CH_PAR(p2); CH_PAR(p3); CH_REST(rest); return a(p1, p2, p3, rest); }; break;


#define CREATE_LISP_NAME_P0(a, b) case LISP_ ## a: { CH_END;  return a(); }; break;
#define CREATE_LISP_NAME_P1(a, b) case LISP_ ## a: { CH_PAR(p1); CH_END; return a(p1.eval()); }; break;
#define CREATE_LISP_NAME_P1_CALLOTHER(a, b, c) case LISP_ ## a: { CH_PAR_NO_EVAL(p1); CH_END; return c(p1.eval()); }; break;
#define CREATE_LISP_NAME_P2(a, b) case LISP_ ## a: { CH_PAR(p1); CH_PAR(p2);CH_END; return a(p1.eval(), p2.eval()); }; break;
#define CREATE_LISP_NAME_P3(a, b) case LISP_ ## a: { CH_PAR(p1); CH_PAR(p2);CH_PAR(p3);CH_END; return a(p1.eval(), p2.eval(), p3.eval()); }; break;
#define CREATE_LISP_NAME_P0_REST(a, b) case LISP_ ## a: { CH_REST(rest); return a(rest); }; break;
#define CREATE_LISP_NAME_P1_REST(a, b) case LISP_ ## a: {  CH_PAR(p1); CH_REST(rest); return a(p1.eval(), rest); }; break;
#define CREATE_LISP_NAME_P2_REST(a, b) case LISP_ ## a: { CH_PAR(p1); CH_PAR(p2); CH_REST(rest); return a(p1.eval(), p2.eval(), rest); }; break;
#define CREATE_LISP_NAME_P3_REST(a, b) case LISP_ ## a: {  CH_PAR(p1); CH_PAR(p2); CH_PAR(p3); CH_REST(rest); return a( p1.eval(), p2.eval(), p3.eval(), rest); }; break;


#define CREATE_LISP_NAME_P0_KEY(a, b) case LISP_ ## a: {  CH_REST(rest); return a(rest); }; break;
#define CREATE_LISP_NAME_P1_KEY(a, b) case LISP_ ## a: { CH_PAR(p1); CH_REST(rest); return a(p1.eval(), rest); }; break;
#define CREATE_LISP_NAME_P2_KEY(a, b) case LISP_ ## a: {  CH_PAR(p1); CH_PAR(p2); CH_REST(rest); return a(p1.eval(), p2.eval(), rest); }; break;
#define CREATE_LISP_NAME_P3_KEY(a, b) case LISP_ ## a: {  CH_PAR(p1); CH_PAR(p2); CH_PAR(p3); CH_REST(rest); return a(p1.eval(), p2.eval(), p3.eval(), rest); }; break;



     CREATE_LISP_NAME_P0_REST(list, "list")
     CREATE_LISP_NAME_P0_REST(append, "append")
     CREATE_LISP_NAME_P2_KEY(adjoin, "adjoin")
     CREATE_LISP_NAME_P1(reverse, "reverse")
     CREATE_LISP_NAME_P1(eval, "eval")
     CREATE_LISP_NAME_P2_KEY(sort, "sort")
     CREATE_LISP_NAME_P3_KEY(substitute, "substitute")
     CREATE_LISP_NAME_P1(length, "length")
     CREATE_LISP_NAME_P1(evenp, "evenp")
     CREATE_LISP_NAME_P1(oddp, "oddp")
     CREATE_LISP_NAME_P2_KEY(count, "count")
     CREATE_LISP_NAME_P2_KEY(find, "find")
     CREATE_LISP_NAME_P2_KEY(position, "position")
     CREATE_LISP_NAME_P1(first, "first")
     CREATE_LISP_NAME_P1(second, "second")
     CREATE_LISP_NAME_P1(third, "third")
     CREATE_LISP_NAME_P1(last, "last")
    // CREATE_LISP_NAME_P1(butlast, "butlast")
     CREATE_LISP_NAME_P2(nth, "nth")
    // CREATE_LISP_NAME_P2(nthcdr, "nthcdr")
     CREATE_LISP_NAME_P2(subseq, "subseq")
     CREATE_LISP_NAME_P2_KEY(remove, "remove")
    // CREATE_LISP_NAME_P2(push, "push")

       case LISP_push: { CH_PAR(p1); CH_PAR(p2);CH_END; return push(p1.eval(), p2); }; break;
     case LISP_pop: { CH_PAR(p1); CH_END; return pop(p1); }; break;


    // CREATE_LISP_NAME_P1(pop, "pop")
     CREATE_LISP_NAME(globals, "globals")
     CREATE_LISP_NAME_P2_KEY(member, "member")
    // CREATE_LISP_NAME_P2(pushnew, "pushnew")
     //CREATE_LISP_NAME_P2_KEY(l_union, "union")
   //  CREATE_LISP_NAME_P2_KEY(intersection, "intersection")
    // CREATE_LISP_NAME_P2_KEY(set_difference, "set-difference")
     CREATE_LISP_NAME_P3_M(l_if, "if")
     CREATE_LISP_NAME_P1_REST_M(when, "when")
     CREATE_LISP_NAME_P0_REST_M(cond, "cond")
     CREATE_LISP_NAME(l_case, "case")
     CREATE_LISP_NAME_P2(cons, "cons")
     CREATE_LISP_NAME(rest, "rest")
     CREATE_LISP_NAME_P1(car, "car")
     CREATE_LISP_NAME_P1(cdr, "cdr")
     CREATE_LISP_NAME(mapc, "mapc")
             CREATE_LISP_NAME_P2_REST(map, "map")
     CREATE_LISP_NAME(otherwise, "otherwise")
     CREATE_LISP_NAME(print, "print")
     CREATE_LISP_NAME_P0(t, "t")
     CREATE_LISP_NAME(l_do, "do")
     CREATE_LISP_NAME(l_return, "return")
     CREATE_LISP_NAME_P1_CALLOTHER_M(quote, "quote", executequote)
     CREATE_LISP_NAME_P0(nil, "nil")
     CREATE_LISP_NAME_P1(l_not, "not")
     CREATE_LISP_NAME_P1(inc_one, "1+")
     CREATE_LISP_NAME_P1(dec_one, "1-")
     CREATE_LISP_NAME_P2(less, "<")
     CREATE_LISP_NAME_P2(eless, "<=")
     CREATE_LISP_NAME_P2(greater, ">")
     CREATE_LISP_NAME_P2(egreater, ">=")
     CREATE_LISP_NAME_P2(equal_numeric, "=")
     CREATE_LISP_NAME_P2(dotimes, "dotimes")
   CREATE_LISP_NAME(assoc, "assoc")
     CREATE_LISP_NAME(dolist, "dolist")
    //CREATE_LISP_NAME(ash, "ash")
    // CREATE_LISP_NAME_P0_REST(progn, "progn")
    // CREATE_LISP_NAME_P0_REST(block, "block")
   //  CREATE_LISP_NAME(head, "head")
     CREATE_LISP_NAME_P2_M(setf, "setf")
     CREATE_LISP_NAME_P2(equal, "equal")
     CREATE_LISP_NAME_P1_REST(let, "let")
     CREATE_LISP_NAME_P2(eql, "eql")
     CREATE_LISP_NAME_P2(let_star, "let*")
     CREATE_LISP_NAME_P1_REST(plus, "+")
     CREATE_LISP_NAME_P1_REST(minus, "-")
     CREATE_LISP_NAME_P1_REST(mul, "*")
     CREATE_LISP_NAME_P1_REST(div, "/")
    // CREATE_LISP_NAME_P2_REST(defparameter, "defparameter")
    // CREATE_LISP_NAME_P2_REST(defconstant, "defconstant")
     CREATE_LISP_NAME_P2_REST_M(defun, "defun")
     CREATE_LISP_NAME_P1_REST_M(lambda, "lambda")
     //CREATE_LISP_NAME_P1_REST(apply, "apply")
     CREATE_LISP_NAME_P1_REST_M(funcall, "funcall")
    // CREATE_LISP_NAME(mapcar, "mapcar")
     //CREATE_LISP_NAME(maplist, "maplist")
    // CREATE_LISP_NAME(values, "values")
    // CREATE_LISP_NAME_P0_REST(mutliple_value_bind, "multiple-value-bind")
    // CREATE_LISP_NAME_P0_REST(mutliple_value_call, "multiple-value-call")
     CREATE_LISP_NAME_P2(incf, "incf")
     CREATE_LISP_NAME_P2(decf, "decf")
     CREATE_LISP_NAME_P1(abs, "abs")
     CREATE_LISP_NAME_P2(mod, "mod")
     CREATE_LISP_NAME_P1_REST(max, "max")
     CREATE_LISP_NAME_P1_REST(min, "min")
    // CREATE_LISP_NAME_P2(expt, "expt")
     //CREATE_LISP_NAME_P1(float, "float")
    // CREATE_LISP_NAME_P1(numerator, "numerator")
    // CREATE_LISP_NAME_P1(denominator, "denominator")
     CREATE_LISP_NAME_P1(round, "round")
    // CREATE_LISP_NAME_P1(floor, "floor")
    // CREATE_LISP_NAME_P1(ceiling, "ceiling")
   //  CREATE_LISP_NAME_P1(truncate, "truncate")
     CREATE_LISP_NAME_P1(random, "random")
     CREATE_LISP_NAME_P1(sin, "sin")
     CREATE_LISP_NAME_P1(cos, "cos")
     CREATE_LISP_NAME_P1(tan, "tan")
     CREATE_LISP_NAME_P1(asin, "asin")
     CREATE_LISP_NAME_P1(acos, "acos")
     CREATE_LISP_NAME_P1(atan, "atan")
     CREATE_LISP_NAME_P1(sinh, "sinh")
     CREATE_LISP_NAME_P1(cosh, "cosh")
     CREATE_LISP_NAME_P1(tanh, "tanh")
     CREATE_LISP_NAME_P1(asinh, "asinh")
     CREATE_LISP_NAME_P1(acosh, "acosh")
     CREATE_LISP_NAME_P1(atanh, "atanh")
    // CREATE_LISP_NAME_P2(typep, "typep")
     //CREATE_LISP_NAME_P2_REST(every, "every")
    // CREATE_LISP_NAME_P2_REST(some, "some")



     CREATE_LISP_NAME_P1(null, "null")
   //  CREATE_LISP_NAME_P1(listp, "listp")
     /*  CREATE_LISP_NAME_P1(zerop, "zerop")
     CREATE_LISP_NAME_P1(plusp, "plusp")
     CREATE_LISP_NAME_P1(minusp, "minusp")
   CREATE_LISP_NAME_P1(upper_case_p, "upper-case-p")
     CREATE_LISP_NAME_P1(lower_case_, "lower-case-p")
     CREATE_LISP_NAME_P1(numberp, "numberp")
     CREATE_LISP_NAME_P1(floatp, "floatp")
     CREATE_LISP_NAME_P1(integerp, "integerp")
     CREATE_LISP_NAME_P1(graphic_char_p, "graphic-char-p")
     CREATE_LISP_NAME_P1(alphanumericp, "alphanumericp")
     CREATE_LISP_NAME_P1(alpha_char_p, "alpha-char-p")*/
     CREATE_LISP_NAME(make_array, "make-array")
     CREATE_LISP_NAME(aref, "aref")
     CREATE_LISP_NAME(svref, "svref")
     CREATE_LISP_NAME(vector, "vector")
     CREATE_LISP_NAME(defstruct, "defstruct")
     CREATE_LISP_NAME(format, "format")
    // CREATE_LISP_NAME_P1_REST(concatenate, "concatenate")
     /*CREATE_LISP_NAME(char, "char")
     CREATE_LISP_NAME(position, "position")
     CREATE_LISP_NAME(string_capitalize, "string-capitalize")
     CREATE_LISP_NAME(string_upcase, "string-upcase")
     CREATE_LISP_NAME(string_downcase, "string-downcase")
     CREATE_LISP_NAME(remove_duplicates, "remove-duplicates")
     CREATE_LISP_NAME(read, "read")
     CREATE_LISP_NAME(read_line, "read-line")
     CREATE_LISP_NAME(read_from_string, "read-from-string")
     CREATE_LISP_NAME(defmacro, "defmacro")
     CREATE_LISP_NAME(macroexpand, "macroexpand")
     CREATE_LISP_NAME(nthmost, "nthmost")
     CREATE_LISP_NAME(starts_with, "starts-with")
     CREATE_LISP_NAME(rotate_left, "rotate-left")
     CREATE_LISP_NAME(rotate_right, "rotate-right")
     CREATE_LISP_NAME(insert_between, "insert-between")
     CREATE_LISP_NAME(permutations, "permutations")
     CREATE_LISP_NAME(terpri, "terpri")*/
     CREATE_LISP_NAME_P1_REST(l_and, "and")
     CREATE_LISP_NAME_P1_REST(l_or, "or")
     CREATE_LISP_NAME_P2_M(defvar, "defvar")
    // CREATE_LISP_NAME_P1_REST(unless, "unless")
    // CREATE_LISP_NAME(l_string, "string")
    // CREATE_LISP_NAME_P2_KEY(string_equal, "string=")
    // CREATE_LISP_NAME_P2_KEY(string_less, "string<")
    // CREATE_LISP_NAME_P2_KEY(string_greater, "string>")
    // CREATE_LISP_NAME_P2_KEY(string_eless, "string<=")
    // CREATE_LISP_NAME_P2_KEY(string_egreater, "string>=")
     CREATE_LISP_NAME_P1(caaar, "caaar")
     CREATE_LISP_NAME_P1(caadr, "caadr")
     CREATE_LISP_NAME_P1(caar, "caar")
     CREATE_LISP_NAME_P1(cadar, "cadar")
     CREATE_LISP_NAME_P1(caddr, "caddr")
     CREATE_LISP_NAME_P1(cadr, "cadr")
     CREATE_LISP_NAME_P1(cdaar, "cdaar")
     CREATE_LISP_NAME_P1(cdadr, "cdadr")
     CREATE_LISP_NAME_P1(cdar, "cdar")
     CREATE_LISP_NAME_P1(cddar, "cddar")
     CREATE_LISP_NAME_P1(cdddr, "cdddr")
     CREATE_LISP_NAME_P1(cddr, "cddr")
             CREATE_LISP_NAME(d_key, ":key")
             CREATE_LISP_NAME(d_test, ":test")
             CREATE_LISP_NAME(d_test_not, ":test-not")
             CREATE_LISP_NAME_P1(type_of, "type-of")
             CREATE_LISP_NAME_P2(typep, "typep")
             CREATE_LISP_NAME_P2_KEY(l_delete, "delete")


#undef CREATE_LISP_NAME

#undef CREATE_LISP_NAME_P0
#undef CREATE_LISP_NAME_P1
#undef CREATE_LISP_NAME_P2
#undef CREATE_LISP_NAME_P3
#undef CREATE_LISP_NAME_P0_REST
#undef CREATE_LISP_NAME_P1_REST
#undef CREATE_LISP_NAME_P2_REST
#undef CREATE_LISP_NAME_P3_REST


#undef CREATE_LISP_NAME_P0_KEY
#undef CREATE_LISP_NAME_P1_KEY
#undef CREATE_LISP_NAME_P2_KEY
#undef CREATE_LISP_NAME_P3_KEY
        default: {
            LObject command = obj.m_data->l_eval();
            switch (command.m_data->m_type) {
            case tctLambda: {
                  LLambda* func = check_cast<LLambda*>(command.m_data.get());
                  DynArray1D<LObject> params;
                  DynArray1D<CUtiIDString> names;
                  l_iterator iter2;
                  iter2.m_currentelement = func->m_args;
                  while (!iter2.end()) {
                      if (iter.end()) {
                       throw LispException(LObject(this), UTISTR("Not enough parameters provided. Call: ")+this->l_print() + UTISTR(" Arguments: ")+func->m_args.print());
                      }
                      params.push_back(iter.getObject().m_data->l_eval());
                      names.push_back(iter2.getObject().getAtom());
                      iter.next();
                      iter2.next();

                  };
                  if (!iter.end()) {
                   throw LispException(LObject(this), UTISTR("Too many parameters provided. Call: ")+this->l_print() + UTISTR(" Arguments: ")+func->m_args.print());
                  }
                  uti_refptr<LEnvironment> env =LGlobalEnvironment::s_globalenvironment->getTopStack();

                  LGlobalEnvironment::s_globalenvironment->setTopStack(func->m_environment.get());

                  LGlobalEnvironment::s_globalenvironment->push();

                  for (int i = 0; i < params.size();i++) {
                      LGlobalEnvironment::s_globalenvironment->createVariable(names[i],params[i]);
                  }

                  LObject res = func->m_sexp.m_data->l_eval();


                  LGlobalEnvironment::s_globalenvironment->pop();

                  LGlobalEnvironment::s_globalenvironment->setTopStack(env.get());
                  return res;

            }; break;
            case tctProc: {
                LProc* func = check_cast<LProc*>(command.m_data.get());
                LObject res = func->m_value->call(iter.m_currentelement);

                return res;

            }; break;
            default: {
                    throw LispException(LObject(this), UTISTR("Lambda function in first argument expected!"));
            }; break;
            };

            //Bla, setupup environment


        }; break;
     }
     return LObject();
}

LObject l_case(const LObject& cond)
{

    l_iterator iter(cond);
    LObject keyform = iter.getObject().eval();
    iter.next();

    while (!iter.end()) {
        LObject r = iter.getObject();
        LObject key = car(r);
        if (key == keyform) {
          return car(cdr(r)).eval();
        };

      iter.next();
    };
    return LObject();

}
LObject cond(const LObject& cond)
{

    l_iterator iter(cond);


    while (!iter.end()) {
        LObject r = iter.getObject();
        if (car(r).eval().toBool()) {
          return car(cdr(r)).eval();
        };

      iter.next();
    };
    return LObject();

}
LObject defun(const LObject& name, const LObject& parameter, const LObject& body)
{
    LLambda* l = new LLambda(parameter, Lisp::car(body));

    l->m_environment = LGlobalEnvironment::s_globalenvironment->getTopStack();

    LObject res;
   res.m_data = l;

   res.updateDebug();
   LGlobalEnvironment::s_globalenvironment->createVariable(name.getAtom(), res);

   return  res;

}
//also for loop
LObject loop(const LObject& sexpressions)
{
    /*l_iterator iter(cond);
    LObject keyform = iter.getObject().eval();
    iter.next();

    while (!iter.end()) {
        LObject r = iter.getObject();
        LObject key = car(r);
        if (key == keyform) {
          return car(cdr(r)).eval();
        };

      iter.next();
    };
    return LObject();*/
LObject res;
     while (true)
     {
         l_iterator iter(sexpressions);
         while (!iter.end()) {
             LObject r = iter.getObject().eval();
             if (r.getType() == tctReturn) {
               res = r;
               break;
             };
           iter.next();
         };
        if (res.isNil()== false) break;

     };
     return check_cast<LReturn*>(res.m_data.get())->m_value;
}

bool l_and(const LObject& a1, const LObject& l) {
    bool b = a1.eval().toBool();
    if (!b) return false;
    l_iterator iter(l);
    while (!iter.end()) {
        b = iter.getObject().eval().toBool();
        if (!b)return false;
    }
    return true;
}
bool l_or(const LObject& a1, const LObject& l) {
    bool b = a1.eval().toBool();
    if (b) return true;
    l_iterator iter(l);
    while (!iter.end()) {
        b = iter.getObject().eval().toBool();
        if (b)return true;
    }
    return false;
}
LObject l_do(const LObject& evalop, const LObject& conditions, const LObject& sexpressions)
{
 //sadlkföjsad
    return LObject();
}
LObject dotimes(const LObject& amount, const LObject& sexpressions)
{

    LGlobalEnvironment::s_globalenvironment->push();
    LObject times = amount;
    CUtiIDString variablename = car(times).getAtom();
    int count = car(cdr(times)).eval().toInt();
    LGlobalEnvironment::s_globalenvironment->createVariable(variablename, LObject(0));

    l_iterator iter(sexpressions);

    LObject loopstart = iter.m_currentelement;
    LObject res, r;
    for (int i = 0; i < count; i++) {
        LGlobalEnvironment::s_globalenvironment->setVariable(variablename, LObject(i));

       iter.m_currentelement = loopstart;
       while (!iter.end()) {
          r= iter.getObject().eval();
          if (r.getType() == tctReturn) {
            res = check_cast<LReturn*>(r.m_data.get())->m_value;
            break;
          };
          iter.next();
       };
       if (!res.isNil()) break;

    };
    if (res.isNil()) {
      res = r;
    };

    LGlobalEnvironment::s_globalenvironment->pop();
    return res;
}

LObject dolist(const LObject& amount, const LObject& sexpressions)
{
    LGlobalEnvironment::s_globalenvironment->push();
    LObject times = amount;
    CUtiIDString variablename = car(times).getAtom();
    LObject liste = car(cdr(times)).eval();
    LGlobalEnvironment::s_globalenvironment->createVariable(variablename, LObject(0));



    l_iterator totallist(liste);
    l_iterator iter(sexpressions);

    LObject loopstart = iter.m_currentelement;
    LObject res, r;
    while (!totallist.end())  {

        LGlobalEnvironment::s_globalenvironment->setVariable(variablename, totallist.getObject());
       iter.m_currentelement = loopstart;
       while (!iter.end()) {
          r= iter.getObject().eval();
          if (r.getType() == tctReturn) {
            res = check_cast<LReturn*>(r.m_data.get())->m_value;
            break;
          };
          iter.next();
       };
       if (!res.isNil()) break;

       totallist.next();
    };
    if (res.isNil()) {
      res = r;
    };

    LGlobalEnvironment::s_globalenvironment->pop();
    return res;
}

LObject lambda(const LObject&parameters, const LObject& body)
{

    LLambda* l = new LLambda(parameters, body);
    l->m_environment = LGlobalEnvironment::s_globalenvironment->getTopStack();

    LObject res;
   res.m_data = l;

   res.updateDebug();

   return res;

}
LObject let(const LObject& variables, const LObject& sexpressions)
{
 //sadlkföjsad
    return LObject();
}
LObject let_star(const LObject& variables, const LObject& sexpressions)
{
 //sadlkföjsad
    return LObject();
}
LObject block(const LObject& name, const LObject& sexpressions)
{
 //sadlkföjsad
    return LObject();
}
LObject l_return(const LObject& valuelist)
{
    LReturn* ret = new LReturn();
    l_iterator iter(valuelist);
    if (!iter.end())
    ret->m_value = iter.getObject();
  LObject res;
  res.m_data = ret;
  res.updateDebug();
  return ret;

};
LObject l_return_from(const LObject& name, const LObject& valuelist)
{
    LReturn* ret = new LReturn();
    l_iterator iter(valuelist);
    if (!iter.end())
    ret->m_value = iter.getObject();

    ret->m_name = name.getAtom();
  LObject res;
  res.m_data = ret;
  res.updateDebug();
  return ret;

};
LObject plus(const LObject& f, const LObject& rest)
{
   LObject res = f;
   l_iterator iter(rest);
   while (!iter.end()) {
       res = single_plus(res, iter.getObject());
     iter.next();
   };

   return res;
}
LObject minus(const LObject& f, const LObject& rest)
{
   LObject res = f;
   l_iterator iter(rest);
   while (!iter.end()) {
       res = single_minus(res, iter.getObject());
     iter.next();
   };

   return res;
}
LObject mul(const LObject& f, const LObject& rest)
{
   LObject res = f;
   l_iterator iter(rest);
   while (!iter.end()) {
       res = single_mul(res, iter.getObject());
     iter.next();
   };

   return res;
}
LObject div(const LObject& f, const LObject& rest)
{
   LObject res = f;
   l_iterator iter(rest);
   while (!iter.end()) {
       res = single_div(res, iter.getObject());
     iter.next();
   };

   return res;
}
LObject mod(const LObject& f, const LObject& rest)
{
   LObject res = f;
   l_iterator iter(rest);
   while (!iter.end()) {
       res = single_mod(res, iter.getObject());
     iter.next();
   };

   return res;
}
LObject max(const LObject& f, const LObject& rest)
{
    LObject res = f;
    l_iterator iter(rest);
    while (!iter.end()) {
        LObject obj = iter.getObject();
        if (less(res, obj)) {
          res = obj;
        };
      iter.next();
    };
    return res;

}
LObject min(const LObject& f, const LObject& rest)
{
    LObject res = f;
    l_iterator iter(rest);
    while (!iter.end()) {
        LObject obj = iter.getObject();
        if (less(obj, res)) {
          res = obj;
        };
      iter.next();
    };
    return res;

}
LObject abs(const LObject& f)
{
    switch (f.m_data->m_type) {
        case tctInteger: {
            return ::abs(check_cast<LInteger*>(f.m_data.get())->m_value);
        }; break;
        case tctFloat: {
        return ::abs(check_cast<LFloat*>(f.m_data.get())->m_value);
        }; break;
        default: {throw LispException(f, UTISTR("Type cannot be used in comparsion"));};
    };
}
LObject round(const LObject& f)
{
    switch (f.m_data->m_type) {
        case tctInteger: {
            return check_cast<LInteger*>(f.m_data.get())->m_value;
        }; break;
        case tctFloat: {
        return int(::round(check_cast<LFloat*>(f.m_data.get())->m_value));
        }; break;
        default: {throw LispException(f, UTISTR("Type cannot be used in comparsion"));};
    };
}
LObject trunc(const LObject& f)
{
    switch (f.m_data->m_type) {
        case tctInteger: {
            return check_cast<LInteger*>(f.m_data.get())->m_value;
        }; break;
        case tctFloat: {
           return int(bintrunc(check_cast<LFloat*>(f.m_data.get())->m_value));
        }; break;
        default: {throw LispException(f, UTISTR("Type cannot be used in comparsion"));};
    };
}


LObject setf(const LObject& varname, const LObject& exp)
{
    LGlobalEnvironment::s_globalenvironment->setVariable(varname.getAtom(), exp.eval());
    return varname;
}


LObject executequote(const LObject& exp)
{
    return exp;
}

LObject quote(const LObject& exp)
{
 return seq(LISP_quote, exp);
}

LObject random(const LObject& f)
{
    switch (f.m_data->m_type) {
        case tctInteger: {
            return urandom(check_cast<LInteger*>(f.m_data.get())->m_value);
        }; break;
        case tctFloat: {
           return urandom()*check_cast<LFloat*>(f.m_data.get())->m_value;
        }; break;
        default: {throw LispException(f, UTISTR("Type cannot be used in comparsion"));};
    };
}

#define CREATEOPERATOR(f1, f2, predicate) switch (f1.m_data->m_type) {\
case tctInteger: {\
switch (f2.m_data->m_type) {\
    case tctInteger: {\
        return check_cast<LInteger*>(f1.m_data.get())->m_value predicate check_cast<LInteger*>(f2.m_data.get())->m_value;\
    }; break;\
    case tctFloat: {\
    return check_cast<LInteger*>(f1.m_data.get())->m_value predicate check_cast<LFloat*>(f2.m_data.get())->m_value;\
    }; break;\
    default: {throw new LispException(f2, UTISTR("Type cannot be used in comparsion"));};\
};\
};break;\
case tctFloat: {\
    switch (f2.m_data->m_type) {\
        case tctInteger: {\
            return check_cast<LFloat*>(f1.m_data.get())->m_value predicate check_cast<LInteger*>(f2.m_data.get())->m_value;\
        }; break;\
        case tctFloat: { \
    return check_cast<LFloat*>(f1.m_data.get())->m_value predicate check_cast<LFloat*>(f2.m_data.get())->m_value;\
        }; break;\
    default: {throw new LispException(f2, UTISTR("Type cannot be used in comparsion"));};\
    }; break;\
}; break;\
 default: {throw new LispException(f1, UTISTR("Type cannot be used in comparsion"));};\
};

LObject single_plus(const LObject& f1, const LObject& f2)
{
CREATEOPERATOR(f1, f2, +)
    return LObject();
}

LObject single_minus(const LObject& f1, const LObject& f2)
{
CREATEOPERATOR(f1, f2, -)
    return LObject();
}

LObject single_mul(const LObject& f1, const LObject& f2)
{
CREATEOPERATOR(f1, f2, *)
    return LObject();
}

LObject single_div(const LObject& f1, const LObject& f2)
{
CREATEOPERATOR(f1, f2, /)
    return LObject();
}

LObject single_mod(const LObject& f1, const LObject& f2)
{

    return f1.toInt() % f2.toInt();
}

bool equal_numeric(const LObject& f1,const LObject& f2)
{
    CREATEOPERATOR(f1, f2, ==);
}

bool less(const LObject& f1,const LObject& f2)
{
CREATEOPERATOR(f1, f2, <);
return false;
}
bool greater(const LObject& f1,const LObject& f2)
{
CREATEOPERATOR(f1, f2, >);
return false;
}
bool eless(const LObject& f1,const LObject& f2)
{
CREATEOPERATOR(f1, f2, <=);
return false;
}
bool egreater(const LObject& f1,const LObject& f2)
{
CREATEOPERATOR(f1, f2, >=);
return false;
}
LObject assoc(const LObject& f1,const LObject& f2)
{
     l_iterator iter(f2);
     while (!iter.end()) {
         LObject obj = iter.getObject();
         if (car(obj) == f1) return obj;

         iter.next();
     }

     return LObject();
}


CUtiString LList::l_printAligned() const
{
    l_iterator iter(LObject(this));;
    if (iter.end()) return UTISTR("(nil)");

bool haslist = false;
DynArray1D<LObject> list;
    while (!iter.end()) {

        list.push_back(iter.getObject());
        if (iter.getObject().isConList()) haslist = true;
        iter.next();

    }
    if (!haslist)return l_print();

    CStringList strlist;
    CStringList templist;
    strlist.add(UTISTR("("));

    for (int i = 0; i < list.size(); i++) {
        templist.setText(list[i].printAligned());
        for (int j = 0; j< templist.size(); j++) {
         strlist.add(UTISTR("   ")+templist.getLine(j));
        }
    };



    strlist.add(UTISTR(")"));

    return strlist.getText();
}

CUtiString LList::l_print() const
{

    l_iterator iter(LObject(this));;
    if (iter.end()) return UTISTR("(nil)");
    CUtiString str = UTISTR("(");
    int c = 0;

    while (!iter.end()) {

        if (c != 0) str += UTISTR(" ");
        str += iter.getObject().print();
        c++;
        iter.next();


    }str += UTISTR(")");
    return str;
}

bool LList::l_smaller(LCell* other) const
{
    if (m_type < other->m_type)return true; if (m_type > other->m_type)return false;
    if (other->m_type != m_type) throw LispException(other,UTISTR("List expression expected"));
   l_iterator iter(LObject(this));
   LObject o(other);
   l_iterator iter2;
   iter2.m_currentelement = o;

   if (iter.end() == true && iter2.end() == false) return true;
   if (iter.end() == false && iter2.end() == true) return false;
   while (!iter.end()) {

       if (iter.getObject() < iter2.getObject()) return true;
       if (iter.getObject() != iter2.getObject()) return false;
       iter.next();
       iter2.next();
       if (iter.end() == true && iter2.end() == false) return true;
       if (iter.end() == false && iter2.end() == true) return false;

   }
   return false;
}

LCell* LCons::l_createCopy() const
{
    ListGenerator l;
    l_iterator iter(LObject(this));

    while (!iter.end()) {
        l.push_back( iter.getObject().m_data->l_createCopy());
        iter.next();

    }
    return l.getList().m_data.get();
}

LObject parseLispEntry(UtiLexer& lexer, LispErrorContainer* errorcontainer)
{
    CUtiString res;
    if (lexer.IsThisNextCharacter(L'(')) {
        //Liste
        lexer.GetNextChar();
        if (lexer.IsThisNextCharacter(L')')) {

            lexer.GetNextChar();

             return LObject(false);

        }
        DynArray1D<LObject> l;
        while (!lexer.IsThisNextCharacter(L')') && lexer.isEOF() == false) {
            wchar_t ch = lexer.GetNextCharNoMove();
            l.push_back(parseLispEntry(lexer, errorcontainer));
            lexer.MoveToNextChar();
        }
        lexer.GetNextChar();

        return errorcontainer->take(list(l), lexer.getLexerPosition());

    }else if (lexer.GetNextNumber(res)) {

        return errorcontainer->take(LObject(res.ToFloat()), lexer.getLexerPosition());
    } else if (lexer.GetNextIntNumber(res)) {
        return errorcontainer->take(LObject(res.ToInt()), lexer.getLexerPosition());

    }  else if (lexer.GetNextBezeichner(res)) {
        if (res == UTISTR("nil")) {
            return LObject(false);
        }
        if (res == UTISTR("t")) {
            return LObject(true);
        }
        return errorcontainer->take(LObject(CUtiIDString(res)), lexer.getLexerPosition());
    } else if (lexer.IsThisNextCharacter(34))
    {
        CUtiString str;
        U_CHAR c = lexer.GetNextChar();
        while ((c = lexer.GetNextChar()) != 34)
        {
            if (c == 0)
            {
                lexer.printError(UTEXT("String constant is not closed!"));
                break;
            };
            str += c;
        };

        return errorcontainer->take(LObject(str), lexer.getLexerPosition());
    } else{

       // lexer.GetNextNumber(res);
        LObject res = errorcontainer->take(LObject(), lexer.getLexerPosition());
        //lexer.printError(UTISTR("Unknown Symbol"));
throw LispException(res, UTISTR("Unknown Symbol"));
        lexer.GetNextChar();
        return res;
    }

    return LObject();
}

LObject parseString(const CUtiString& data, LispErrorContainer* errorcontainer)
{
    UtiLexer lexer;
    lexer.setLispSymbols(true);
    lexer.setText(data);

    uti_refptr<LispErrorContainer> ec;
    if (errorcontainer == NULL) {
        ec = new LispErrorContainer();
        errorcontainer = ec.get();
    }

    return parseLispEntry(lexer, errorcontainer);
}


void parseStringAll(const CUtiString& data, LispErrorContainer* errorcontainer, std::vector<LObject>& resobj)
{
    UtiLexer lexer;
    lexer.setLispSymbols(true);
    lexer.setText(data);

    uti_refptr<LispErrorContainer> ec;
    if (errorcontainer == NULL) {
        ec = new LispErrorContainer();
        errorcontainer = ec.get();
    }

    while (!lexer.isEOF()) {
    LObject res = parseLispEntry(lexer, errorcontainer);
    if (!res.isNil()) {
        resobj.push_back(res);
    }
    }
}


std::map<LCell*, CIVector2>::iterator LispErrorContainer::getErrorRecursive(LCell* cell)
{
    auto it = m_element_to_line.find(cell);
    if (it!= m_element_to_line.end()) {
        return it;
    }else {
        LObject obj(cell);
        if (obj.isConList()) {
            for (auto p : obj) {
                it = getErrorRecursive(p.m_data.get());
                if (it!= m_element_to_line.end()) {
                    return it;
                }
            };
            return m_element_to_line.end();
        }else {
         return m_element_to_line.end();
        }

    }
}

bool LispErrorContainer::hasErrorInfo(LCell* cell)
{
   return getErrorRecursive(cell) != m_element_to_line.end();
}

CUtiString LispErrorContainer::createError(LCell* cell, const CUtiString& name)
{

    auto it = getErrorRecursive(cell);
    if (it!= m_element_to_line.end()) {

        return m_filename+UTISTR("(")+it->second.x+UTISTR(", ")+it->second.y+UTISTR("): ")+ name;
    }else {

    return m_filename + UTISTR(": ")+ name + UTISTR(" in ")+cell->l_print();
    };
}


LObject LispErrorContainer::take(const LObject& obj, const CIVector2& pos)
{
  m_element_to_line.insert(std::make_pair(obj.m_data.get(), pos));
return obj;
}

void LispErrorContainer::replacedBy(const LObject& old, const LObject &newobj)
{
    auto it = m_element_to_line.find(old.m_data.get());
    if (it!= m_element_to_line.end()) {
        m_element_to_line.insert(std::make_pair(newobj.m_data.get(), it->second));
    }else {


    };
}

}


