%skeleton "lalr1.cc"
%define "parser_class_name" "prologxx_parser"
%{

#include "tokentype.h"
#include <string>
#include <cmath>
#include <iostream>

using namespace std;

#include "scanner.h"
#include "enginefactory.h"
#include "prolog_driver.h"
#define DEBUGMODE(a)

#define SETPOS(f) \
   f->pos.beginline = yylloc.begin.line;\
   f->pos.begincolumn = yylloc.begin.column;\
   f->pos.endline = yylloc.end.line;\
   f->pos.endcolumn = yylloc.end.column;\
   f->pos.beginfile = yylloc.begin.filename;\
   f->pos.endfile = yylloc.end.filename;

#define STDTERM(a, b) a=b;
#define TERM_SINGLE(result, term, atom) Structure* structure = EngineFactory::getInstance()->createStructure(); SETPOS(structure);  structure->setName(atom);  structure->addElement(dynamic_cast<Term*>(term)); result = structure;

#define TERM_DOUBLE(result, terma, atom, termb) Structure* structure = EngineFactory::getInstance()->createStructure(); SETPOS(structure);  structure->setName(atom);  structure->addElement(dynamic_cast<Term*>(terma)); structure->addElement(dynamic_cast<Term*>(termb)); result = structure;

%}


%parse-param {prolog_driver& driver}
%lex-param {prolog_driver&driver}
%locations
%initial-action
{
   @$.begin.filename = @$.end.filename = &driver.file;
};

%union {
   double real;
   int integer;
   int id;
   char* string;
   BaseEngine* element;
   ListElement le;
};

%debug
%error-verbose

%token INTNUM
%token FLOATNUM
%token BEGINBODY
%token ATOM
%token NORMALVAR
%token NOTTOKEN
%token STRINGLIST
%token CONSTRAINT_ADD
%token CONSTRAINT_GUARDEND
%token CONSTRAINT_SIMPLIFICATION
%token CONSTRAINT_PROPAGATION
%token CONSTRAINT_ID
%token SELECTED_ONEFACT

      /*
%token XFX1  %token YFX1  %token XFY1  %token XF1  %token YF1  %token FX1  %token FY1
%token XFX2  %token YFX2  %token XFY2  %token XF2  %token YF2  %token FX2  %token FY2
%token XFX3  %token YFX3  %token XFY3  %token XF3  %token YF3  %token FX3  %token FY3
%token XFX4  %token YFX4  %token XFY4  %token XF4  %token YF4  %token FX4  %token FY4
%token XFX5  %token YFX5  %token XFY5  %token XF5  %token YF5  %token FX5  %token FY5
%token XFX6  %token YFX6  %token XFY6  %token XF6  %token YF6  %token FX6  %token FY6
%token XFX7  %token YFX7  %token XFY7  %token XF7  %token YF7  %token FX7  %token FY7
%token XFX8  %token YFX8  %token XFY8  %token XF8  %token YF8  %token FX8  %token FY8
%token XFX9  %token YFX9  %token XFY9  %token XF9  %token YF9  %token FX9  %token FY9
%token XFX10  %token YFX10  %token XFY10  %token XF10  %token YF10  %token FX10  %token FY10
%token XFX11  %token YFX11  %token XFY11  %token XF11  %token YF11  %token FX11  %token FY11
%token XFX12  %token YFX12  %token XFY12  %token XF12  %token YF12  %token FX12  %token FY12
%token XFX13  %token YFX13  %token XFY13  %token XF13  %token YF13  %token FX13  %token FY13
%token XFX14  %token YFX14  %token XFY14  %token XF14  %token YF14  %token FX14  %token FY14
%token XFX15  %token YFX15  %token XFY15  %token XF15  %token YF15  %token FX15  %token FY15
%token XFX16  %token YFX16  %token XFY16  %token XF16  %token YF16  %token FX16  %token FY16
%token XFX17  %token YFX17  %token XFY17  %token XF17  %token YF17  %token FX17  %token FY17
%token XFX18  %token YFX18  %token XFY18  %token XF18  %token YF18  %token FX18  %token FY18
%token XFX19  %token YFX19  %token XFY19  %token XF19  %token YF19  %token FX19  %token FY19
      */
      
%%

program:	
      | program declaration   {DEBUGMODE(wcout << "program -> program declaration\n");}
      | onefact {}
;

onefact: SELECTED_ONEFACT term '.' {
	Term* str = dynamic_cast<Term*>($<element>2);
   assert(str != NULL);
   driver.newstr = str;
   DEBUGMODE(wcout << "SELECTED_ONEFACT term '.'\n");
   
};

declaration: fact '.'		{
  
   Fact* f = dynamic_cast<Fact*>($<element>1);
   assert(f != NULL);
  // addFact(f);
   
   Program::getInstance()->addElement(f);
   DEBUGMODE(wcout << "declaration -> fact\n");}
   | rule '.'	{
	   Rule* f = dynamic_cast<Rule*>($<element>1);
	   Program::getInstance()->addRule(f);
	   
	}
   | error '.'	{/*yyerrok;*/}
;


rule: constraintname constraintrule {
		Rule* f = dynamic_cast<Rule*>($<element>2);
	$<element>$ = f;
}
| constraintrule
{
	$<element>$ = $<element>1;
};
	
constraintname: atom CONSTRAINT_ADD {
	$<id>$ = $<id>1;	
};

constraintrule: head CONSTRAINT_SIMPLIFICATION guard body
{
	Rule* f = dynamic_cast<Rule*>($<element>1);
	SETPOS(f);
	f->setGuard(dynamic_cast<Body*>($<element>3));
	f->setBody(dynamic_cast<Body*>($<element>4));
	f->setType(Rule::Simplification);
	$<element>$ = f;
}
| head CONSTRAINT_PROPAGATION guard body
{
	Rule* f = dynamic_cast<Rule*>($<element>1);
	SETPOS(f);
	f->setGuard(dynamic_cast<Body*>($<element>3));
	f->setBody(dynamic_cast<Body*>($<element>4));
	f->setType(Rule::Propagation);
	$<element>$ = f;
}
|
head CONSTRAINT_SIMPLIFICATION body
{
	Rule* f = dynamic_cast<Rule*>($<element>1);
	SETPOS(f);
	f->setBody(dynamic_cast<Body*>($<element>3));
	f->setType(Rule::Simplification);
	$<element>$ = f;
}
| head CONSTRAINT_PROPAGATION body
{
	Rule* f = dynamic_cast<Rule*>($<element>1);
	SETPOS(f);
	f->setBody(dynamic_cast<Body*>($<element>3));
	f->setType(Rule::Propagation);
	$<element>$ = f;
};

guard: body CONSTRAINT_GUARDEND
{
	$<element>$ = $<element>1;
};

head: constraints;

constraints: predicate{
	Rule* f = EngineFactory::getInstance()->createRule();
	SETPOS(f);
	f->addHead((Structure*)$<element>1);
	$<element>$ = f;
}
| constraints ',' predicate
{
	Rule* f = (Rule*)$<element>1;
	SETPOS(f);
	f->addHead((Structure*)$<element>3);
	$<element>$ = f;
};

fact: 	predicate		
{
   Fact* fact = EngineFactory::getInstance()->createFact();
   SETPOS(fact);
   Structure* str = dynamic_cast<Structure*>($<element>1);
   assert(str != NULL);
   fact->setHead(str);
   $<element>$ = fact;
   DEBUGMODE(wcout << "fact -> predicate\n");
}
| predicate BEGINBODY body 
{
	Fact* fact = EngineFactory::getInstance()->createFact();
	SETPOS(fact);
   Structure* str = dynamic_cast<Structure*>($<element>1);
   assert(str != NULL);
   fact->setHead(str);
   Body* body = dynamic_cast<Body*>($<element>3);
   assert(body != NULL);
   fact->setBody(body);
   $<element>$ = fact;
   DEBUGMODE(wcout << "fact -> predicate :- body\n");
}
| BEGINBODY body
{
	Fact* fact = EngineFactory::getInstance()->createFact();
	SETPOS(fact);
   fact->setHead(NULL);
   Body* body = dynamic_cast<Body*>($<element>2);
   assert(body != NULL);
   fact->setBody(body);
   $<element>$ = fact;
   DEBUGMODE(wcout << "fact -> :- body\n");
};


predicate: functor		
{
	Structure* structure = EngineFactory::getInstance()->createStructure();
	SETPOS(structure);
   structure->setName($<id>1);
   $<element>$ = structure;   
//   addStructure(structure);
   DEBUGMODE(wcout << "predicate -> functor\n");
}
| functor '(' arguments ')'	
{
	Structure* structure = dynamic_cast<Structure*>($<element>3);
   assert(structure != NULL);
   structure->setName($<id>1);
   $<element>$ = structure;   
  // addStructure(structure);
   DEBUGMODE(wcout << "predicate -> functor '(' arguments ')'");
}
;

functor: atom		{$<id>$ = $<id>1;DEBUGMODE(wcout << "functor -> atom\n");}
;

arguments: term		
{
	Structure* structure = EngineFactory::getInstance()->createStructure();
	SETPOS(structure);
   assert(dynamic_cast<Term*>($<element>1) != NULL);
   Term *t = dynamic_cast<Term*>($<element>1);
   assert(t != NULL);
   structure->addElement(t);
   $<element>$ = structure;
   DEBUGMODE(wcout << "arguments -> term\n");
}
| arguments ',' term	
{
   
	Structure* structure = dynamic_cast<Structure*>($<element>1);
    assert(structure != NULL);
    Term *t = dynamic_cast<Term*>($<element>3);
    assert(t != NULL);
    structure->addElement(t);
    $<element>$ = structure;
    DEBUGMODE(wcout << "arguments -> arguments ',' term\n");
}
;

/*term: term0   {$<element>$ = $<element>1;}
;
*/
term:	var		{$<element>$ = $<element>1;DEBUGMODE(wcout << "term -> var\n");}
|  number	{$<element>$ = $<element>1;DEBUGMODE(wcout << "term -> number\n");}
|  predicate	{$<element>$ = $<element>1;DEBUGMODE(wcout << "term -> structure\n");}
|  list		{$<element>$ = $<element>1;DEBUGMODE(wcout << "term -> list\n");}
|  stringlist   {$<element>$ = $<element>1;DEBUGMODE(wcout << "term -> stringlist\n");}
//|  term0   {$<element>$ = $<element>1;}
;

/*
term0: term0 YFX0 term1 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term1 XFY0 term0 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term1 XFX0 term1 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
| FX0 term1 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| FY0 term0 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| term1 XF0 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term0 YF0 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term1 {STDTERM($<element>$, $<element>1);}
;

term1: term1 YFX1 term2 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term2 XFY1 term1 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term2 XFX1 term2 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
| FX1 term2 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| FY1 term1 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| term2 XF1 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term1 YF1 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term2 {STDTERM($<element>$, $<element>1);}
;

term2: term2 YFX2 term3 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term3 XFY2 term2 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term3 XFX2 term3 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
| FX2 term3 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| FY2 term2 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| term3 XF2 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term2 YF2 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term3 {STDTERM($<element>$, $<element>1);}
;

term3: term3 YFX3 term4 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term4 XFY3 term3 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term4 XFX3 term4 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
| FX3 term4 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| FY3 term3 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| term4 XF3 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term3 YF3 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term4 {STDTERM($<element>$, $<element>1);}
;

term4: term4 YFX4 term5 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term5 XFY4 term4 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
|term5 XFX4 term5 {TERM_DOUBLE($<element>$, $<element>1, $<id>2, $<element>3);}
| FX4 term5 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| FY4 term4 {TERM_SINGLE($<element>$, $<element>2, $<id>1);}
| term5 XF4 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term4 YF4 {TERM_SINGLE($<element>$, $<element>1, $<id>2);}
| term5 {STDTERM($<element>$, $<element>1);}
;


term5: termold {$<element>$ = $<element>1;};
*/
var: 	'_'		
{
	Var* var = EngineFactory::getInstance()->createVar();
	SETPOS(var);
   var->setName($<id>1);
   $<element>$ = var;
   DEBUGMODE(wcout << "DONTCARE\n");
}
| NORMALVAR 	
{
	Var* var = EngineFactory::getInstance()->createVar();
	SETPOS(var);
   var->setName($<id>1);
   $<element>$ = var;
   DEBUGMODE(wcout << "VARIABLE\n");
}
;

atom: 	ATOM		{$<id>$ = $<id>1;DEBUGMODE(wcout << "atom: \n");}
| '.'	
{
	$<id>$  = EngineFactory::getInstance()->registerAtom(".");
	DEBUGMODE(wcout << "atom: '.'\n");
}
;

number: INTNUM		{
	IntNumber* integer = EngineFactory::getInstance()->createIntNumber();
	SETPOS(integer);
   integer->setValue($<integer>1);
   $<element>$ = integer;
   DEBUGMODE(wcout << "numberi: \n");
   }
| FLOATNUM	{
	   RealNumber* real = EngineFactory::getInstance()->createRealNumber();
   SETPOS(real);
   real->setValue($<real>1);
   $<element>$ = real;
   DEBUGMODE(wcout << "numberf: \n");
}
;

 // '['']' 	 {DEBUGMODE(wcout <<"list -> []\n");} | 

stringlist:  STRINGLIST
{
	char* c = $<string>$;
   PID dot = EngineFactory::getInstance()->registerAtom(".");
   PID emptylist = EngineFactory::getInstance()->registerAtom("[]");
   Structure* lastelement = NULL;
   int length = strlen(c);
   Structure* r_struct;
   if (length == 0) {
	Structure* structure;
   r_struct= EngineFactory::getInstance()->createStructure();
   SETPOS(r_struct);
       structure = r_struct;
       structure->setName(emptylist);
   } else {
      Structure* structure;
      r_struct = EngineFactory::getInstance()->createStructure();
      SETPOS(r_struct);
      structure = r_struct;
      IntNumber* intvalue = EngineFactory::getInstance()->createIntNumber();
      SETPOS(intvalue);
      intvalue->setValue(c[0]);
      structure->addElement(intvalue);
      
      structure->setName(dot);
      lastelement=r_struct;
      for (int i = 1; i < length; i++) {
	      structure = EngineFactory::getInstance()->createStructure();
      SETPOS(structure);
      intvalue = EngineFactory::getInstance()->createIntNumber();
      SETPOS(intvalue);
          intvalue->setValue(c[i]);
          structure->addElement(intvalue);
          structure->setName(dot);
	  lastelement->addElement(structure);
	  lastelement = structure;
      };
      structure = EngineFactory::getInstance()->createStructure();
      SETPOS(structure);
      structure->setName(emptylist);
      lastelement->addElement(structure);
   };
   //Structure* structure = cf->createStructure();
   //structure->setName();
   //wcout << c << endl;
   $<element>$ = r_struct;
   delete c;
}
;

list:   '[' liste ']'	
{
	$<element>$ = $<le>2.element;
	DEBUGMODE(wcout <<"list -> [liste]\n");
}
|	'{' brace '}'
{
	$<element>$ = $<element>2;
	DEBUGMODE(wcout <<"list -> {bracket}\n");
};
;

liste: 	  term		{
	Structure* structure = EngineFactory::getInstance()->createStructure();
	SETPOS(structure);
	structure->setName(EngineFactory::getInstance()->registerAtom("."));
	assert(dynamic_cast<Term*>($<element>1) != NULL);
	Term *t = dynamic_cast<Term*>($<element>1);
   assert(t != NULL);
   structure->addElement(t);
   Structure* closelist = EngineFactory::getInstance()->createStructure();
   SETPOS(closelist);
   closelist->setName(EngineFactory::getInstance()->registerAtom("[]"));
   structure->addElement(closelist);
   //$$.element = structure;
   $<le>$.element = structure;
   $<le>$.laststruct = structure;
	
	DEBUGMODE(wcout << "liste -> term\n");

}
| liste ',' term	{
	
	Structure* structure = dynamic_cast<Structure*>($<le>1.element);
	Structure* lastelement = $<le>1.laststruct;
	assert(structure != NULL);
	assert(lastelement != NULL);
	
	Structure* newstruct = EngineFactory::getInstance()->createStructure();
	SETPOS(newstruct);
	newstruct->setName(EngineFactory::getInstance()->registerAtom("."));
	
	Term *t = dynamic_cast<Term*>($<element>3);
	assert(t != NULL);
	newstruct->addElement(t);
	newstruct->addElement(lastelement->getElement(1));
	lastelement->setElement(1, newstruct);
	$<le>$.element = structure;
	$<le>$.laststruct = newstruct;
	DEBUGMODE(wcout << "liste -> liste,term\n");

}
 | liste CONSTRAINT_GUARDEND term	{
	Structure* structure = dynamic_cast<Structure*>($<le>1.element);
	Structure* lastelement = $<le>1.laststruct;
	assert(structure != NULL);
	assert(lastelement != NULL);
	Term *t = dynamic_cast<Term*>($<element>3);
	Term* t2 =  lastelement->getElement(1);
	
	lastelement->setElement(1, t);
	delete t2;
	$<le>$.element = structure;
	$<le>$.laststruct = lastelement;
	DEBUGMODE(wcout << "liste -> liste|term\n");

}
;

brace: 	  term		{
	Structure* structure = EngineFactory::getInstance()->createStructure();
	SETPOS(structure);
	structure->setName(EngineFactory::getInstance()->registerAtom("{}"));
	assert(dynamic_cast<Term*>($<element>1) != NULL);
	Term *t = dynamic_cast<Term*>($<element>1);
	assert(t != NULL);
	structure->addElement(t);
	$<element>$ = structure;
	
	DEBUGMODE(wcout << "brace -> term\n");

}
| brace ',' term	{
	
	Structure* structure = dynamic_cast<Structure*>($<element>1);
    assert(structure != NULL);
    Term *t = dynamic_cast<Term*>($<element>3);
    assert(t != NULL);
    structure->addElement(t);
    $<element>$ = structure;
    DEBUGMODE(wcout << "brace -> brace,term\n");

}
;

body: 	  bodypart	
{
   Body* body = EngineFactory::getInstance()->createBody();
   SETPOS(body);
   BodyPart* g = dynamic_cast<BodyPart*>($<element>1);
   assert(g != NULL);
   body->addElement(g);
   $<element>$ = body;
   DEBUGMODE(wcout << "body -> bodypart\n");
}
| body ';' bodypart	
{
   Body* body = dynamic_cast<Body*>($<element>1);
   assert(body != NULL);
   BodyPart* g = dynamic_cast<BodyPart*>($<element>3);
   assert(g != NULL);
   body->addElement(g);
   $<element>$ = body;
   DEBUGMODE(wcout << "body -> body ; bodypart\n");
}
;

bodypart: 	goal	
{
   BodyPart* part = EngineFactory::getInstance()->createBodyPart();
   SETPOS(part);
   Goal* g = dynamic_cast<Goal*>($<element>1);
   assert(g != NULL);
   part->addElement(g);
   $<element>$ = part;
   DEBUGMODE(wcout << "bodypart -> goal\n");
}
| bodypart ',' goal	
{
	BodyPart* part = dynamic_cast<BodyPart*>($<element>1);
   assert(part != NULL);
   Goal* g = dynamic_cast<Goal*>($<element>3);
   assert(g != NULL);
   part->addElement(g);
   $<element>$ = part;
   DEBUGMODE(wcout << "bodypart -> bodypart ',' goal\n");
}
;

goal:	term		
{
	Goal* goal = EngineFactory::getInstance()->createGoal();
	SETPOS(goal);
   goal->setGoal(gTerm);
   Term* t = dynamic_cast<Term*>($<element>1);
   assert(t != NULL);
   goal->setElement(t);
   $<element>$ = goal;
   DEBUGMODE(wcout << "goal -> term\n");

}
| '(' body ')'
{
	Goal* goal = EngineFactory::getInstance()->createGoal();
	SETPOS(goal);
   goal->setGoal(gBody);
   Body* t = dynamic_cast<Body*>($<element>2);
   assert(t != NULL);
   goal->setElement(t);
   $<element>$ = goal;
   DEBUGMODE(wcout << "goal -> '(' body ')'\n");
}
| NOTTOKEN goal	
{
   Goal* goal = EngineFactory::getInstance()->createGoal();
	SETPOS(goal);
   goal->setGoal(gNotBody);
   Goal* g = dynamic_cast<Goal*>($<element>2);
 //  Body* t = dynamic_cast<Body*>($<element>2);
   assert(g != NULL);
   goal->setElement(g);
   $<element>$ = goal;
   DEBUGMODE(wcout << "goal -> \\+ goal \n");
}
;

%%


void yy::prologxx_parser::error(const yy::prologxx_parser::location_type &l,
			      const std::string &m)
{
   driver.error(l,m);
}