%option noyywrap nounput batch debug yylineno
%option c++
%{
#include <cmath>
#include <iostream>
using namespace std;
#include "tokentype.h"
#include "prolog_driver.h"
#include "plparse.tab.h"
#include "enginefactory.h"

#define DEBUGMODE(a)
#define MAX_STR_CONST 1024

char string_buf[MAX_STR_CONST];
char* string_buf_ptr;
int strcount;
%}

DIGIT	[0-9]
ATOM	[a-z][a-zA-Z0-9_]*
VAR	[A-Z][a-zA-Z0-9_]*
ONEATOMCHARACTER	["+""*""/""\""^""<"">""=""~"":""?""@""#""$""&"]|"-"
ATOMCHAR	{ONEATOMCHARACTER}+
SPECIALATOM	"!"|"[]"|"{}"
BLANK	[ \t]

%{
#define YY_USER_ACTION flex_yylloc->columns(yyleng);
%}



%x str
%x atomstr

%%
%{
   flex_yylloc->step();
   
%}
\"	{string_buf_ptr = string_buf; strcount = 0;BEGIN(str);}

\'	{string_buf_ptr = string_buf; strcount = 0;BEGIN(atomstr);}

{BLANK}+	flex_yylloc->step();
[\n]+	flex_yylloc->lines(yyleng); flex_yylloc->step();

{DIGIT}+	{
		//printf("An integer: %s (%d)\n", yytext, atoi(yytext));
	flex_yylval->integer = atoi(yytext);
		DEBUGMODE(wcout << "Int: " << yylval.integer << endl);
		return INTNUM;
		}
"-"{DIGIT}+	{
		//printf("An integer: %s (%d)\n", yytext, atoi(yytext));
		   flex_yylval->integer = atoi(yytext);
		DEBUGMODE(wcout << "Int: " << yylval.integer << endl);
		return INTNUM;
}

{DIGIT}+"."{DIGIT}+	{
			DEBUGMODE(wcout << L"A float: " << yytext << L"(" << atof(yytext) <<L")\n");
		flex_yylval->real = atof(yytext);
		
	return FLOATNUM;
	}
-{DIGIT}+"."{DIGIT}+	{
	   DEBUGMODE(wcout << L"A float: " << yytext << L"(" << atof(yytext) <<L")\n");
	flex_yylval->real = atof(yytext);
		
	return FLOATNUM;
}

":""-"	{ DEBUGMODE(wcout << L"BEGINBODY" << endl);return BEGINBODY;}

"|"	{return CONSTRAINT_GUARDEND;};
	
"§§§!!!_SELECTED_ONEFACT_!!!§§§"	{return SELECTED_ONEFACT;};

{ATOM}|{ATOMCHAR}|{SPECIALATOM}	{
	DEBUGMODE(wcout << L"An ATOM " << YYText() <<  L"\n");
	const char* c = YYText();
	if (strcmp(c, "@")==0) {
	   return CONSTRAINT_ADD;
        };
	/*if (strcmp(c, "|")==0) {
	   return CONSTRAINT_GUARDEND;
        };*/
	if (strcmp(c, "#")==0) {
		return CONSTRAINT_ID;
};
	if (strcmp(c, "<=>")==0) {
	   return CONSTRAINT_SIMPLIFICATION;
        };
	if (strcmp(c, "==>")==0) {
	   return CONSTRAINT_PROPAGATION;
        };
	flex_yylval->id = EngineFactory::getInstance()->registerAtom(YYText());
	return ATOM;
}

"\\""+"   { DEBUGMODE(wcout << L"NOTTOKEN"); return NOTTOKEN;}
{VAR}	{
	DEBUGMODE(wcout << L"An VAR " << YYText() <<  L"\n");
        flex_yylval->id = EngineFactory::getInstance()->registerAtom(YYText());
	return NORMALVAR;
	}
	
	
"("|")"|"["|"]"|","|";"|"|"|"{"|"}"	{
	DEBUGMODE(wcout << L"An Operator " << YYText() <<  L"\n");
	return yytext[0];
	}
	
"."	{ 
	DEBUGMODE(wcout << L"An Dot\n");
	return '.'; 
	}



<str>\"  {
		DEBUGMODE(wcout << L"An ListString " << YYText() <<  L"\n");
	BEGIN(INITIAL);
	*string_buf_ptr = '\0';
	
	//wcout << string_buf;
	int l = strlen(string_buf)+3;
	char* c_str = new char[l];
	flex_yylval->string = c_str;
	strncpy(c_str, string_buf, l);
	return STRINGLIST;
}

<str>\n	{

	flex_driver->error(*flex_yylloc, "Error: Return in String");
}

<str>\\[0-7]{1,3} {
	int result;
	sscanf(YYText() + 1, "%o", &result);
	if (result > 0xff) {

		flex_driver->error(*flex_yylloc, "Error: Octal ASCII Code Out-Of-Bounds");
	}
	
	*string_buf_ptr++ = result; strcount++; assert(strcount < MAX_STR_CONST-2);
}

<str>\\[0-9]+	{
	flex_driver->error(*flex_yylloc, "Error: Falsche Escapesequence");
}

<str>\\n	*string_buf_ptr++ = '\n'; strcount++; assert(strcount < MAX_STR_CONST-2);
<str>\\t	*string_buf_ptr++ = '\t'; strcount++; assert(strcount < MAX_STR_CONST-2);
<str>\\r	*string_buf_ptr++ = '\r'; strcount++; assert(strcount < MAX_STR_CONST-2);
<str>\\b	*string_buf_ptr++ = '\b'; strcount++; assert(strcount < MAX_STR_CONST-2);
<str>\\f	*string_buf_ptr++ = '\f'; strcount++; assert(strcount < MAX_STR_CONST-2);
	
<str>\\(.|\n)	*string_buf_ptr++ = yytext[1]; strcount++; assert(strcount < MAX_STR_CONST-2);

<str>[^\\\n\"]+	{
	const char* yptr = YYText();
	
	while(*yptr) {
		*string_buf_ptr++ = *yptr++; strcount++; assert(strcount < MAX_STR_CONST-2);
	}
	
}

<atomstr>\'  {
	BEGIN(INITIAL);
	*string_buf_ptr = '\0';
	DEBUGMODE(wcout << L"An ATOM " << string_buf <<  L"\n");
	flex_yylval->id =EngineFactory::getInstance()->registerAtom(string_buf);
	return ATOM;
}

<atomstr>\n	{

	flex_driver->error(*flex_yylloc, "Error: Return in String");
}

<atomstr>\\[0-7]{1,3} {
	int result;
	sscanf(yytext + 1, "%o", &result);
	if (result > 0xff) {
		flex_driver->error(*flex_yylloc, "Error: Octal ASCII Code Out-Of-Bounds");
}
	
	*string_buf_ptr++ = result; strcount++; assert(strcount < MAX_STR_CONST-2);
}

<atomstr>\\[0-9]+	{
	flex_driver->error(*flex_yylloc, "Error: Falsche Escapesequence");
}

<atomstr>\\n	*string_buf_ptr++ = '\n'; strcount++; assert(strcount < MAX_STR_CONST-2);
<atomstr>\\t	*string_buf_ptr++ = '\t'; strcount++; assert(strcount < MAX_STR_CONST-2);
<atomstr>\\r	*string_buf_ptr++ = '\r'; strcount++; assert(strcount < MAX_STR_CONST-2);
<atomstr>\\b	*string_buf_ptr++ = '\b'; strcount++; assert(strcount < MAX_STR_CONST-2);
<atomstr>\\f	*string_buf_ptr++ = '\f'; strcount++; assert(strcount < MAX_STR_CONST-2);
	
<atomstr>\\(.|\n)	*string_buf_ptr++ = yytext[1]; strcount++; assert(strcount < MAX_STR_CONST-2);

<atomstr>[^\\\n\']+	{
	char* yptr = yytext;
	
	while(*yptr) {
		*string_buf_ptr++ = *yptr++; strcount++; assert(strcount < MAX_STR_CONST-2);
	}
	
}

	
"/*"[^}]*"*/"	/*comments*/

"%"[^\n]*	flex_yylloc->lines(yyleng); flex_yylloc->step();

<<EOF>>	{return 0;}
\xd	{//nix
		}
.	{
	std::string str3 = "Illegales Zeichen: ";
	str3 += yytext;
	int v = yytext[0];
	char c[30];
	snprintf(c, 30, "(%i)/(0x%x)", v);
	str3 += c;
	flex_driver->error(*flex_yylloc, str3.c_str());

}

%%

void prolog_driver::scan_begin()
{
   //yy_flex_debug = trace_scanning;
   /*if (!(yyin=fopen(file.c_str(), "r"))) {
      error(std::string("cannot open") +file);
   }*/
}

void prolog_driver::scan_end()
{
   //fclose(yyin);
};
