//Uti3DMath.cpp
/***************************************************************************
 *   Copyright (C) 2006 by Michael Staud   *
 *   michael.staud@uni-ulm.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



#include "Uti3DMath.h"

#include <math.h>
#include <stdio.h>
#include <algorithm>
#include <iostream>
#include <assert.h>

#include "utivector.cpp"
#include "utiintvector.h"
#include "utimathbox.h"
#include "utiarray.h"
#include "easyprofiler.h"

using namespace std;

double anglemod(double angle)
{
    if (angle >= 0.0f && angle < 360.0f) return angle;
    double a = angle/360.0f;

    int part = bintrunc(a);
    a = a-part;
    return a*360.0f;
}

CUtiString ftos(double value)
{
    return CUtiString::floatToStr(value, 0, 3);
}

/*double rint(double val)
{
    return binround(val);
}*/

void makevectors(const CVector3& angles, CVector3& forward, CVector3& up, CVector3& right)
{
CMatrix m;
m.LoadIdentity();
m.EulerRotation(angles);
forward = m.getColumn(2).xyz();
up = m.getColumn(1).xyz();
right = m.getColumn(0).xyz();
}

CVector vectoangles(const CVector3& v)
{
return uVectorToAngle(v);
}

float vectoyaw(const CVector3& v)
{

    return uVectorToAngle(v).y;
}

CUtiString vtos(const CVector3& v)
{
   return convertToString(v);
}

void getLocalCoordinateSystemByVector(const CVector& input, CVector &xp, CVector &yp, CVector &zp)
{
   
   CVector3 cameravector = normalize(input);
 //  utiassert(udistance(length(cameravector), 1) < 0.01);
   CVector3 up(0, 1, 0);
   CVector3 a;
   CVector3 b;

   a=crossProduct(up, cameravector);
   if (length(a) < 0.01) {
     b= crossProduct(cameravector, CVector(1, 0, 0));
      b = normalize(b);
      a=crossProduct(b, cameravector);
   } else {
      a = normalize(a);
      b=crossProduct(cameravector, a);
   };
   xp = *((CVector*)&a);
   yp = *((CVector*)&cameravector);
   zp = *((CVector*)&b);

}

int binomial(int n, int k)
{
    int tmp0, tmp1;
    if ((k == 0) || (n==k)) {
        return 1;
    } else {
       tmp0 = binomial(n-1, k);
       tmp1 = binomial(n-1, k-1);
       return tmp0+tmp1;
    }

}
//TODO: Im Falle von Multitasking klappt das mit der Globalen tempmatrix nicht
//Man muss fr jeden Thread eine machen
//CMatrix mMat;



//Summand + Summand = Summe
//Subtrahend - Minuend = Differenz
//Faktor * Faktor = Produkt
//Dividend : Divisor = Quotient

CVector vectorToEuler(const CVector& vector)
{
    return uVectorToAngle(vector);
}

NetRotationInfo vectorToRotation(const CVector& vector)
{
   return NetRotationInfo(uVectorToAngle(vector));
}

CUtiString convertToString(const NetRotationInfo& info)
{
    if (info.m_rotationtype == nrtEuler) {

        return UTISTR("E")+convertToString(info.m_values.xyz());
    } else {
        return UTISTR("Q")+convertToString(info.m_values);
    }
}


CVector3 hexRGBtoVec3(unsigned int color)
{
   unsigned int red = (color >> 16)&255;
    unsigned int green = (color >> 8)&255;
    unsigned int blue = (color >> 0)&255;
    return CVector3((utifloat)red/255.0f, (utifloat)green/255.0f, (utifloat)blue/255.0f);
}


CVector3 rgbToHSV(const CVector3& v)
{
  CVector result;
  utifloat max_value = umax(v.x, umax(v.y, v.z));
  utifloat min_value = umin(v.x, umin(v.y, v.z));
  if (max_value == min_value) {
     result.x = 0;
  } else if (max_value == v.x){
    result.x = 60.0f*(0.0f+(v.y-v.z)/(max_value-min_value));
  } else if (max_value == v.y) {
    result.x = 60.0f*(2.0f+(v.z-v.x)/(max_value-min_value));
  } else if (max_value == v.z) {
    result.x = 60.0f*(4.0f+(v.x-v.y)/(max_value-min_value));
  }
  if (result.x < 0.0f) result.x += 360.0f;
  if (result.x > 360.0f) result.x -= 360.0f;
  if (max_value < 0.0001f) {
      result.y = 0.0f;
  } else {
      result.y = (max_value-min_value)/max_value;
  };

  result.z = max_value;
  return result;
}

CVector3 rgbToHSV(const CVector3& v, const CVector3& oldhsv)
{
  CVector result;
  utifloat max_value = umax(v.x, umax(v.y, v.z));
  utifloat min_value = umin(v.x, umin(v.y, v.z));
  if (max_value == min_value) {
     result.x = oldhsv.x;
  } else if (max_value == v.x){
    result.x = 60.0f*(0.0f+(v.y-v.z)/(max_value-min_value));
  } else if (max_value == v.y) {
    result.x = 60.0f*(2.0f+(v.z-v.x)/(max_value-min_value));
  } else if (max_value == v.z) {
    result.x = 60.0f*(4.0f+(v.x-v.y)/(max_value-min_value));
  }
  if (result.x < 0.0f) result.x += 360.0f;
if (result.x > 360.0f) result.x -= 360.0f;
  if (max_value < 0.0001f) {
      result.y = oldhsv.y;
  } else {
      result.y = (max_value-min_value)/max_value;
  };

  result.z = max_value;
  return result;
}

CVector3 hsvToRGB(const CVector3& v)
{
  CVector result;
  utifloat ft = v.x/360.0f;
  ft = (ft - fasterfloor(ft))*360.0f;
  utifloat angle =  ft;
  
  int hi =(int) (angle / 60.0f);
  utifloat f = (angle / 60.0f - hi);
  utifloat p = v.z* (1.0f - v.y);
  utifloat q = v.z*(1.0f - v.y*f);
  utifloat t = v.z*(1-v.y*(1-f));
  switch (hi) {
    case 0:
    case 6: {
      return CVector(v.z, t, p);
    }; break;
    case 1: {
      return CVector(q, v.z, p);
    }; break;
    case 2: {
      return CVector(p, v.z, t);
    }; break;
    case 3: {
      return CVector(p, q, v.z);
    }; break;
    case 4: {
      return CVector(t, p, v.z);
    }; break;
    case 5: {
      return CVector(v.z, p, q);
    }; break;


  };

  return result;
}



CVector3f rgbToHSV(const CVector3f& v)
{
  CVector3f result;
  float max_value = umax(v.x, umax(v.y, v.z));
  float min_value = umin(v.x, umin(v.y, v.z));
  if (max_value == min_value) {
     result.x = 0;
  } else if (max_value == v.x){
    result.x = 60.0f*(0.0f+(v.y-v.z)/(max_value-min_value));
  } else if (max_value == v.y) {
    result.x = 60.0f*(2.0f+(v.z-v.x)/(max_value-min_value));
  } else if (max_value == v.z) {
    result.x = 60.0f*(4.0f+(v.x-v.y)/(max_value-min_value));
  }
  if (result.x < 0.0f) result.x += 360.0f;
  if (result.x > 360.0f) result.x -= 360.0f;
  if (max_value < 0.0001f) {
      result.y = 0.0f;
  } else {
      result.y = (max_value-min_value)/max_value;
  };

  result.z = max_value;
  return result;
}

CVector3f rgbToHSV(const CVector3f& v, const CVector3f& oldhsv)
{
  CVector3f result;
  float max_value = umax(v.x, umax(v.y, v.z));
  float min_value = umin(v.x, umin(v.y, v.z));
  if (max_value == min_value) {
     result.x = oldhsv.x;
  } else if (max_value == v.x){
    result.x = 60.0f*(0.0f+(v.y-v.z)/(max_value-min_value));
  } else if (max_value == v.y) {
    result.x = 60.0f*(2.0f+(v.z-v.x)/(max_value-min_value));
  } else if (max_value == v.z) {
    result.x = 60.0f*(4.0f+(v.x-v.y)/(max_value-min_value));
  }
  if (result.x < 0.0f) result.x += 360.0f;
if (result.x > 360.0f) result.x -= 360.0f;
  if (max_value < 0.0001f) {
      result.y = oldhsv.y;
  } else {
      result.y = (max_value-min_value)/max_value;
  };

  result.z = max_value;
  return result;
}

CVector3f hsvToRGB(const CVector3f& v)
{
  CVector3f result;
  float ft = v.x/360.0f;
  ft = (ft - fasterfloor(ft))*360.0f;
  float angle =  ft;

  int hi =(int) (angle / 60.0f);
  float f = (angle / 60.0f - hi);
  float p = v.z* (1.0f - v.y);
  float q = v.z*(1.0f - v.y*f);
  float t = v.z*(1-v.y*(1-f));
  switch (hi) {
    case 0:
    case 6: {
      return CVector3f(v.z, t, p);
    }; break;
    case 1: {
      return CVector3f(q, v.z, p);
    }; break;
    case 2: {
      return CVector3f(p, v.z, t);
    }; break;
    case 3: {
      return CVector3f(p, q, v.z);
    }; break;
    case 4: {
      return CVector3f(t, p, v.z);
    }; break;
    case 5: {
      return CVector3f(v.z, p, q);
    }; break;


  };

  return result;
}

CVector2 angleToVector(utifloat angle)
{
    CVector2 res;
    res.x = cos(angle*pi180);
    res.y = sin(angle*pi180);
    return res;
}

CVector3 angleToVector(const CVector3 & v)
{
    CMatrix m;
    m.LoadIdentity();
    m.EulerRotation(v.x, v.y, 0);
    CVector3 v1(0,0,1);
    return m.TransformNormal(v1);
}

CVector3 angleToVector(const CVector3 & v, const CVector3& origin)
{
  CMatrix m;
    m.LoadIdentity();
    m.EulerRotation(v.x, v.y, 0);
    return m.TransformNormal(origin);
}


utifloat positionOnRay(const CVector3& s, const CVector3& normvector, const CVector3& point)
{

    if (normvector.length() < 0.0001)return 0;
    assert(absf(normvector.length()-1.0f) < 0.01f);
    CVector3 p1(point-s);
   // p1.normalize();
    utifloat res = normvector.dotProduct(p1);
    return res;
}

utifloat positionOnRay2(const CVector2& s, const CVector2& normvector, const CVector2& point)
{
    if (normvector.length() < 0.0001)return 0;
    assert(absf(normvector.length()-1.0f) < 0.01f);
    CVector2 p1(point-s);
   // p1.normalize();
    utifloat res = normvector.dotProduct(p1);
    return res;
}

utifloat distanceToLine(const CVector3& s, const CVector3& t, const CVector3& point)
{
    CVector dir(t-s);
    utifloat l = dir.normalize();
    utifloat p = positionOnRay(s, dir, point);
    if (p < 0) {
        return s.distanceTo(point);
    } else if (p > l) {
        return t.distanceTo(point);
    }
    CVector on_line = s+dir*p;
    return on_line.distanceTo(point);

}

utifloat distanceToLine2(const CVector2& s, const CVector2& t, const CVector2& point)
{
    CVector2 dir(t-s);
    utifloat l = dir.normalize();
    utifloat p = positionOnRay2(s, dir, point);
    if (p < 0) {
        return s.distanceTo(point);
    } else if (p > l) {
        return t.distanceTo(point);
    }
    CVector2 on_line = s+dir*p;
    return on_line.distanceTo(point);

}



S2DPoint set2DPoint(U_SINGLE x, U_SINGLE y)
{
  S2DPoint p;
  p.x = x;
  p.y = y;
  return p;
}


const CVector setVector(float x, float y, float z)
{
   CVector r;
   r.setVector(x, y, z);
   return r;
}

CLineDrawingDebugOutputInterface* g_uti_debugoutputinterface=NULL;

void setLineDrawingDebugInterface(CLineDrawingDebugOutputInterface* debug)
{
    if (debug) {
        assert(g_uti_debugoutputinterface == NULL);
        g_uti_debugoutputinterface = debug;
    } else {
        assert(g_uti_debugoutputinterface != NULL);
        g_uti_debugoutputinterface = debug;

    }
}

void debug_Line(const CVector3& start, const CVector3& end, const CVector3& color, float duration)
{
    if (g_uti_debugoutputinterface) {
        g_uti_debugoutputinterface->debug_Line(start, end, color, duration);
    }
}

void debug_Ray(const CVector3& start, const CVector3& dir, const CVector3& color, float duration)
{
    if (g_uti_debugoutputinterface) {
        g_uti_debugoutputinterface->debug_Ray(start, dir, color, duration);
    }
}

void debug_Box(const CBox& box, const CVector3& color, float duration)
{
    if (g_uti_debugoutputinterface) {
        g_uti_debugoutputinterface->debug_Box(box, color, duration);
    }
}
void debug_Point(const CVector3& pos, const CVector3& color, float duration)
{
    if (g_uti_debugoutputinterface) {
        g_uti_debugoutputinterface->debug_Point(pos, color, duration);
    }
}


void debug_String(const CVector3& pos, const CVector3& color, float duration, const CUtiString& html)
{
    if (g_uti_debugoutputinterface) {
        g_uti_debugoutputinterface->debug_String(pos, color, duration, html);
    }
}

void debug_StringXY(const CVector2& pos, const CVector3& color, float duration, const CUtiString& html)
{
    if (g_uti_debugoutputinterface) {
        g_uti_debugoutputinterface->debug_StringXY(pos, color, duration, html);
    }
}


int maxf(utifloat x, utifloat y, utifloat z)
{
   int i = 0;
   float m = absf(x);
   if (absf(y) > m) { m = absf(y); i = 1; };
   if (absf(z) > m) { m = absf(z); i = 2; };
   return(i);
};



void Point_Sub(float *diff, float *sub, float *min, int count)
{
    for (int i = 0; i < count; i++) {
		diff[i] = sub[i] - min[i];
	}
}



void Point_Div(float *quot, float *div, float *divisor, int count)
{
    for (int i = 0; i < count; i++) {
		quot[i] = div[i] / divisor[i];
	}
}

void Point_Add_Value(float *sum, float*sum1, float value, int count)
{
	for (int i = 0; i < count; i++) {
		sum[i] = sum1[i] + value;
	}
}

void Point_Sub_Value(float *diff, float*sub1, float min, int count)
{
	for (int i = 0; i < count; i++) {
		diff[i] = sub1[i] - min;
	}
}

void Value_Sub_Point(float *diff, float sub1, float* min, int count)
{
	for (int i = 0; i < count; i++) {
		diff[i] = sub1 - min[i];
	}
}



void Point_Div_Value(float *quot, float*div, float divisor, int count)
{
	for (int i = 0; i < count; i++) {
		quot[i] = div[i] / divisor;
	}
}

void Value_Div_Point(float *quot, float div, float*divisor, int count)
{
	for (int i = 0; i < count; i++) {
		quot[i] = div / divisor[i];
	}
}


/*CVector::CVector(const S3DPoint & start, const S3DPoint & ende)
{
	x = (float)(ende.x - start.x);
	y = (float)(ende.y - start.y);
   z =(float)(ende.z - start.z);
}

CVector::CVector(const CVector &start, const CVector & ende)
{
x = (float)(ende.x - start.x);
	y = (float)(ende.y - start.y);
   z =(float)(ende.z - start.z);
};

CVector::CVector(TVector &v)
{
	x = v.x;
	y = v.y;
	z = v.z;

}

CVector::CVector(float _x, float _y, float _z)
{
	x = _x;
   y = _y;
	z = _z;
}*/

void ReadVector(CVector* v,CStream & Stream)
{
	Stream.ReadFloat(v->x);
        Stream.ReadFloat(v->y);
        Stream.ReadFloat(v->z);
}

void WriteVector(CVector* v, CStream & Stream)
{
   Stream.WriteFloat(v->x);
   Stream.WriteFloat(v->y);
   Stream.WriteFloat(v->z);
}

/*
void CVector::Neg()
{
	x = -x;
	y = -y;
	z = -z;
}

void CVector::print()
{
  // wcout << UTEXT("Vektor X: ") << x << UTEXT(" Y: ") << y << UTEXT(" Z: ") << z;
}

void CVector::setVector(float _x, float _y, float _z)
{
	x = _x;
	y = _y;
	z = _z;
}

float CVector::length()
{
	return((float)sqrt(x*x+y*y+z*z));
}

void CVector::scalVector(float value)
{
	x = x*value;
	y = y*value;
	z = z*value;
}

float CVector::Normalize()
{
   float length = (float)sqrt(x*x+y*y+z*z);
   if (length == 0) return(0);
   x = x / length;
   y = y / length;
   z = z / length;
   return(length);
}

float CVector::DistanceTo( const CVector*v ) const
{
   float kx = x-v->x;
   float ky = y-v->y;
   float kz = z-v->z;
return (float)sqrt(kx*kx+ky*ky+kz*kz);
}

void CVector::addVector(float _x, float _y, float _z)
{
	x = x + _x;
	y = y + _y;
	z = z + _z;
}

void CVector::subVector(float _x, float _y, float _z)
{
	x = x - _x;
	y = y - _y;
	z = z - _z;
}

CVector CVector::operator+(const CVector & b) const
{
	float sx,sy,sz;
	sx = x + b.x;
	sy = y + b.y;
	sz = z + b.z;
	CVector sum;
	sum.x = sx;
	sum.y = sy;
	sum.z = sz;
	return sum;
}

CVector CVector::operator+(const S3DPoint & b)
{
	float sx,sy,sz;
	sx = x + b.x;
	sy = y + b.y;
	sz = z + b.z;
	CVector sum;
	sum.x = sx;
	sum.y = sy;
	sum.z = sz;
	return sum;
}

CVector CVector::operator-(const CVector & b) const
{
	float sx,sy,sz;
	sx = x - b.x;
	sy = y - b.y;
	sz = z - b.z;
	CVector sum;
	sum.x = sx;
	sum.y = sy;
	sum.z = sz;
	return sum;
}

CVector CVector::operator-(const S3DPoint & b)
{
	float sx,sy,sz;
	sx = x - b.x;
	sy = y - b.y;
	sz = z - b.z;
	CVector sum;
	sum.x = sx;
	sum.y = sy;
	sum.z = sz;
	return sum;
}


CVector CVector::operator-()
{
	float sx,sy,sz;
	sx = -x;
	sy = -y;
	sz = -z;
	CVector sum;
	sum.x = sx;
	sum.y = sy;
	sum.z = sz;
	return sum;
}

CVector& CVector::operator= (const S3DPoint &a)
{

	x = a.x;
	y = a.y;
	z = a.z;
	return(*this);
};*/
/*CVector& CVector::operator= (const SVector &a)
{
	CVector r;
	x = a.x;
	y = a.y;
	z = a.z;
	return(*this);
};
*/
CEntity& CEntity::operator= (const SEntity &a)
{
    CEntity r;
	Group = a.Group;
	Position.x = a.Position.x;
	Position.y = a.Position.y;
	Position.z = a.Position.z;
	Rotation.x = a.Rotation.x;
	Rotation.y = a.Rotation.y;
	Rotation.z = a.Rotation.z;
	valid = a.valid;
	return(*this);
}

/*S3DPoint& CVector::operator= (const CVector &a)
{
	S3DPoint r;
	r.x = a.x;
	r.y = a.y;
	r.z = a.z;
	return(r);
}; */
/*
CVector &CVector::operator+=(const CVector &s)
{
	x += s.x;
	y += s.y;
	z += s.z;
	return(*this);
};
CVector &CVector::operator+=(const S3DPoint &s)
{
	x += s.x;
	y += s.y;
	z += s.z;
	return(*this);
};

CVector &CVector::operator-=(const CVector &s)
{

	x -= s.x;
	y -= s.y;
	z -= s.z;
	return(*this);
};

CVector &CVector::operator-=(const S3DPoint &s)
{
	x -= s.x;
	y -= s.y;
	z -= s.z;
	return(*this);
};


float CVector::operator*(const CVector & b) const
{

   return x*b.x+y*b.y+z*b.z;
}

void CVector::ReflectVector(CVector* Norm)
{
    CVector n1;
	float proj = (*this*-1)*(*Norm);
	n1 = (*Norm)*proj;

	CVector f = 2*n1+ *this;
	*this = f;
}

int CVector::GetMajorAxis()
{
   return(maxf(x, y, z));
}

CVector CVector::GetMajorNormal()
{
	CVector result;
	switch(GetMajorAxis()) {
	case 0: result.setVector(1, 0, 0); break;
	case 1: result.setVector(0, 1, 0); break;
	case 2: result.setVector(0, 0, 1); break;
	};
	return result;
}



void CVector::CrossProduct(CVector* p1, CVector *res)
{

	res->x = y*p1->z-z*p1->y;
   res->y = z*p1->x-x*p1->z;
   res->z = x*p1->y-y*p1->x;

}

void CVector::CalcRightUpVector(CVector* right, CVector* up)
{
	CVector maj(GetMajorNormal());

	int major = GetMajorAxis();
	if (x == 1 || y == 1 || z == 1 || x == -1 || y == -1 || z == -1) {
		if ((major == 0 && x > 0) || major == 1) {
			right->setVector(0, 0, -1);
		} else if (major == 0) {
		  right->setVector(0, 0, 1);
		} else {
		  right->setVector(z, 0, 0);
		};

	} else
	   CrossProduct(&maj, right);


	CrossProduct(right, up);
}



CVector CVector::operator*(float n)  const
{
	float sx,sy,sz;
	sx = x * n;
	sy = y * n;
	sz = z * n;
	CVector sum;
	sum.x = sx;
	sum.y = sy;
	sum.z = sz;
	return sum;
}

CVector CVector::operator/(float n)  const
{
   float sx,sy,sz;
   sx = x / n;
   sy = y / n;
   sz = z / n;
   CVector sum;
   sum.x = sx;
   sum.y = sy;
   sum.z = sz;
   return sum;
}



float CVector::operator*(const S3DPoint &b)
{
   return x*b.x+y*b.y+z*b.z;
};

CVector CVector::operator^(const CVector &b)  const
{
	CVector cvVector;
	cvVector.x = y*b.z-z*b.y;
   cvVector.y = z*b.x-x*b.z;
   cvVector.z = x*b.y-y*b.x;
   return(cvVector);
};*/


const CVector setVector(utifloat x, utifloat y, utifloat z)
{
    return CVector(x, y, z);
}

const CVector3f setVectorf(utifloat x, utifloat y, utifloat z)
{
    return CVector3f(x, y, z);
}


CVector crossProduct(const CVector& v1, const CVector& v2)
{
   CVector res;
   res.x = v1.y*v2.z-v1.z*v2.y;
   res.y = v1.z*v2.x-v1.x*v2.z;
   res.z = v1.x*v2.y-v1.y*v2.x;
   return res;
};

CVector3f crossProduct(const CVector3f& v1, const CVector3f& v2)
{
   CVector3f res;
   res.x = v1.y*v2.z-v1.z*v2.y;
   res.y = v1.z*v2.x-v1.x*v2.z;
   res.z = v1.x*v2.y-v1.y*v2.x;
   return res;
};

CVector3 utireflect(const CVector3 &p, const CVector3 &Norm)
{
    CVector n1;
    double proj = p.dotProduct(Norm);
    n1 = (Norm)*proj;

    CVector f = p-2*n1;
   return f;
}

CVector3f utireflect(const CVector3f &p, const CVector3f &Norm)
{
    CVector3f n1;
    float proj = p.dotProduct(Norm);
    n1 = (Norm)*proj;

    CVector3f f = p-2*n1;
   return f;
}


C3DPoint::C3DPoint()
{
	x = 0;
	y = 0;
	z = 0;
}

void C3DPoint::print()
{
//   wcout << UTEXT("Punkt X: ") << x << UTEXT(" Y: ") << y << UTEXT(" Z: ") << z;
}

C3DPoint::C3DPoint(float _x, float _y, float _z)
{
	x = _x; y = _y; z = _z;
}

C3DPoint::~C3DPoint() {}

/*CMatrix::CMatrix()
{
	LoadIdentity();
}*/

//CMatrix::~CMatrix() {}

/*void CMatrix::LoadIdentity()
{
	m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1;
	m[1][0] = m[2][0] = m[3][0] = 0;
	m[0][1] = m[2][1] = m[3][1] = 0;
	m[0][2] = m[1][2] = m[3][2] = 0;
	m[0][3] = m[1][3] = m[2][3] = 0;
}*/

bool isvectorok(const CVector3 &v) {
 return utiisnan(v.x) == false && utiisfinite(v.x) && utiisnan(v.y) == false && utiisfinite(v.y) && utiisnan(v.z) == false && utiisfinite(v.z);
}

bool isvectorok(const CVector2 &v) {
 return utiisnan(v.x) == false && utiisfinite(v.x) && utiisnan(v.y) == false && utiisfinite(v.y);
}

bool isvectorok(const CVector4 &v) {
 return utiisnan(v.x) == false && utiisfinite(v.x) && utiisnan(v.y) == false && utiisfinite(v.y) && utiisnan(v.z) == false && utiisfinite(v.z) && utiisnan(v.w) == false && utiisfinite(v.w);
}

std::wostream & operator<<(std::wostream & os, const CVector2 & str)
{
  os << L"(" << str.x << L", " << str.y << L")" << std::endl;
  return os;
}

std::ostream & operator<<(std::ostream & os, const CVector2 & str)
{
  os << "(" << str.x << ", " << str.y << ")" << std::endl;
  return os;
}


std::wostream & operator<<(std::wostream & os, const CVector3 & str)
{
  os << L"(" << str.x << L", " << str.y << L", " << str.z << L")" << std::endl;
  return os;
}
std::ostream & operator<<(std::ostream & os, const CVector3 & str)
{
  os << "(" << str.x << ", " << str.y << ", " << str.z << ")" << std::endl;
  return os;
}

std::wostream & operator<<(std::wostream & os, const CVector4 & str)
{
  os << L"(" << str.x << L", " << str.y << L", " << str.z << L", " << str.w << L")" << std::endl;
  return os;
}
std::ostream & operator<<(std::ostream & os, const CVector4 & str)
{
  os << "(" << str.x << ", " << str.y << ", " << str.z << ", " << str.w << ")" << std::endl;
  return os;
}

 CMatrix CMatrix::getIdentity() {
      CMatrix m;
      return m;
    };

void CMatrix::loadOpenGLMatrix(const float* mi)
{
   m[0][0] = mi[0];
   m[1][0] = mi[1];
  m[2][0] = mi[2];
    m[3][0] = mi[3];

   m[0][1] = mi[4];
   m[1][1] = mi[5];
   m[2][1] = mi[6];
  m[3][1] = mi[7];

   m[0][2] = mi[8];
   m[1][2] = mi[9];
   m[2][2] = mi[10];
   m[3][2] = mi[11];

    m[0][3] = mi[12];
    m[1][3] = mi[13];
     m[2][3] = mi[14];
     m[3][3] = mi[15];
}

void CMatrix::storeOpenGLMatrix(float* mi)
{
 mi[0] = m[0][0];
 mi[1] =  m[1][0];
 mi[2] = m[2][0];
 mi[3] =   m[3][0];

 mi[4] =  m[0][1];
 mi[5] =  m[1][1];
 mi[6] =  m[2][1];
 mi[7] = m[3][1];

 mi[8] =  m[0][2];
 mi[9] =  m[1][2];
 mi[10] =  m[2][2];
 mi[11] =  m[3][2];

  mi[12] =  m[0][3];
  mi[13] =  m[1][3];
  mi[14] =   m[2][3];
 mi[15] =    m[3][3];
}
    
   
void CMatrix::SetXRotation(utifloat value)
{assert(!utiisnan(value));
	//LoadIdentity();
	m[0][0] = m[3][3] = 1;
	   m[1][0] = m[2][0] = m[3][0] = 0;
	   m[0][1] =  m[3][1] = 0;
	   m[0][2] =  m[3][2] = 0;
	   m[0][3] = m[1][3] = m[2][3] = 0;
    utifloat c = (utifloat)cos(value*pi180);
    utifloat s = (utifloat)sin(value*pi180);
	m[1][1] = c;
	m[2][1] = s;
	m[1][2] = -s;
	m[2][2] = c;
}

void CMatrix::SetYRotation(utifloat value)
{assert(!utiisnan(value));
	 m[1][1] =  m[3][3] = 1;
	   m[1][0] =  m[3][0] = 0;
	   m[0][1] = m[2][1] = m[3][1] = 0;
	    m[1][2] = m[3][2] = 0;
	   m[0][3] = m[1][3] = m[2][3] = 0;
    utifloat c =(utifloat) cos(value*pi180);
    utifloat s = (utifloat)sin(value*pi180);
	m[0][0] = c;
	m[2][0] = -s;
	m[0][2] = s;
	m[2][2] = c;
}

void CMatrix::SetZRotation(utifloat value)
{
     assert(!utiisnan(value));
	  m[2][2] = m[3][3] = 1;
	   m[2][0] = m[3][0] = 0;
	    m[2][1] = m[3][1] = 0;
	   m[0][2] = m[1][2] = m[3][2] = 0;
	   m[0][3] = m[1][3] = m[2][3] = 0;
    utifloat c = (utifloat)cos(value*pi180);
    utifloat s = (utifloat)sin(value*pi180);
	m[0][0] = c;
	m[1][0] = s;
	m[0][1] = -s;
	m[1][1] = c;
}

void CMatrix::SetTranslate(utifloat _x, utifloat _y, utifloat _z)
{
    assert(!utiisnan(_x));
     assert(!utiisnan(_y));
      assert(!utiisnan(_z));
    m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1;
	   m[1][0] = m[2][0] = m[3][0] = 0;
	   m[0][1] = m[2][1] = m[3][1] = 0;
	   m[0][2] = m[1][2] = m[3][2] = 0;

	m[0][3] = _x;
	m[1][3] = _y;
	m[2][3] = _z;
}

void CMatrix::SetScaling(utifloat _x, utifloat _y, utifloat _z)
{
    assert(!utiisnan(_x));
     assert(!utiisnan(_y));
      assert(!utiisnan(_z));
	 m[3][3] = 1;
	   m[1][0] = m[2][0] = m[3][0] = 0;
	   m[0][1] = m[2][1] = m[3][1] = 0;
	   m[0][2] = m[1][2] = m[3][2] = 0;
	   m[0][3] = m[1][3] = m[2][3] = 0;
	m[0][0] = _x;
	m[1][1] = _y;
	m[2][2] = _z;
}


void CMatrix::Rotate(const NetRotationInfo& info)
{
   if (info.m_rotationtype == nrtEuler) {
       Rotate(info.m_values.x, info.m_values.y, info.m_values.z);
   } else {
       RotateQuaternion(info.m_values);
   }
}

void CMatrix::RotateQuaternion(const CVector4 & quat)
{
   CMatrix mMat;
   mMat.SetRotateQuaternion(quat);
   multipyMatrix(mMat);
}
void CMatrix::SetRotateQuaternion(const CVector4 & quat2)
{
    CVector4 quat = quat2;
    double length = quat.normalize();
    assert(quat.length() > 0.1);
  double quatq1 = 2.0f*quat.x*quat.x;
  double quatq2 = 2.0f*quat.y*quat.y;
  double quatq3 = 2.0f*quat.z*quat.z;
  /*m[0][0] = 1.0f - quatq2 - quatq3;
  m[0][1] = 2.0f*(quat.x*quat.y-quat.w*quat.z);
  m[0][2] = 2.0f*(quat.x*quat.z+quat.w*quat.y);
  m[0][3] = 0.0f;
  m[1][0] = 2.0f*(quat.x*quat.y+quat.w*quat.z);
  m[1][1] = 1.0f - quatq1 - quatq3;
  m[1][2] = 2.0f*(quat.y*quat.z-quat.w*quat.x);
  m[1][3] = 0.0f;
  m[2][0] = 2.0f*(quat.x*quat.z-quat.w*quat.y);
  m[2][1] = 2.0f*(quat.y*quat.z+quat.w*quat.x);
  m[2][2] = 1.0f - quatq1 - quatq2;
  m[2][3] = 0.0f;
  m[3][0] = 0.0f;
  m[3][1] = 0.0f;
  m[3][2] = 0.0f;
  m[3][3] = 1.0f;*/

  double quatx = quat.x;
  double quaty = quat.y;
  double quatz = quat.z;
  double quatw = quat.w;


  m[0][0] = (utifloat)(1.0 - quatq2 - quatq3);
  m[0][1] = (utifloat)(2.0*(quatx*quaty-quatw*quatz));
  m[0][2] = (utifloat)(2.0*(quatx*quatz+quatw*quaty));
  m[0][3] = (utifloat)(0.0);
  m[1][0] = (utifloat)(2.0*(quatx*quaty+quatw*quatz));
  m[1][1] = (utifloat)(1.0 - quatq1 - quatq3);
  m[1][2] = (utifloat)(2.0*(quaty*quatz-quatw*quatx));
  m[1][3] = 0.0;
  m[2][0] = (utifloat)(2.0*(quatx*quatz-quatw*quaty));
  m[2][1] = (utifloat)(2.0*(quaty*quatz+quatw*quatx));
  m[2][2] = (utifloat)(1.0 - quatq1 - quatq2);
  m[2][3] = 0.0;
  m[3][0] = 0.0;
  m[3][1] = 0.0;
  m[3][2] = 0.0;
  m[3][3] = 1.0;

  assert(fabs(getXAxis().length()-1.0) < 0.01);
  assert(fabs(getYAxis().length()-1.0) < 0.01);
  assert(fabs(getZAxis().length()-1.0) < 0.01);
  assert(fabs(getXAxis().dotProduct(getYAxis())) < 0.01);
  assert(fabs(getXAxis().dotProduct(getZAxis())) < 0.01);
  assert(fabs(getYAxis().dotProduct(getZAxis())) < 0.01);

#ifndef NDEBUG
   //CVector4 r = GetRotationAsQuaternion();
 //  assert(isQuaternionEqual(r, quat, 0.01));
#endif

  /*btScalar xs = q.x() * s,   ys = q.y() * s,   zs = q.z() * s;
        btScalar wx = q.w() * xs,  wy = q.w() * ys,  wz = q.w() * zs;
        btScalar xx = q.x() * xs,  xy = q.x() * ys,  xz = q.x() * zs;
        btScalar yy = q.y() * ys,  yz = q.y() * zs,  zz = q.z() * zs;
        setValue(
            btScalar(1.0) - (yy + zz), xy - wz, xz + wy,
            xy + wz, btScalar(1.0) - (xx + zz), yz - wx,
            xz - wy, yz + wx, btScalar(1.0) - (xx + yy));*/
}

CVector4 getQuaternionAxisAngle(const CVector& axis, utifloat angle)
{
    assert(hasNan(axis) == false);
   CVector4 q;
  utifloat l = axis.dotProduct(axis);
  if (l > 0.0) {
    angle *= 0.5;
    q.w = cos (angle);
    l = sin(angle) * (1.0f / sqrt(l));
    q.x = axis.x*l;
    q.y = axis.y*l;
    q.z = axis.z*l;
  }
  else {
    q.w = 1;
    q.x = 0;
    q.y = 0;
    q.z = 0;
  }
  return q;
}

CVector4f quaternionMultiply(const CVector4f& a, const CVector4f& b)
{
   CVector4f r;
   r.w = a.w*b.w - a.x*b.x - a.y*b.y - a.z*b.z;
  r.x = a.w*b.x + a.x*b.w + a.y*b.z - a.z*b.y;
  r.y = a.w*b.y + a.y*b.w + a.z*b.x - a.x*b.z;
  r.z = a.w*b.z + a.z*b.w + a.x*b.y - a.y*b.x;
   return r;
}

CVector4 quaternionMultiply(const CVector4& a, const CVector4& b)
{
   CVector4 r;
   r.w = a.w*b.w - a.x*b.x - a.y*b.y - a.z*b.z;
  r.x = a.w*b.x + a.x*b.w + a.y*b.z - a.z*b.y;
  r.y = a.w*b.y + a.y*b.w + a.z*b.x - a.x*b.z;
  r.z = a.w*b.z + a.z*b.w + a.x*b.y - a.y*b.x;
   return r;
}
CVector4 quaternionMultiplyInvA(const CVector4& a, const CVector4& b)
{
    CVector4 r;
   r.w = a.w*b.w + a.x*b.x + a.y*b.y + a.z*b.z;
    r.x = a.w*b.x - a.x*b.w - a.y*b.z + a.z*b.y;
    r.y = a.w*b.y - a.y*b.w - a.z*b.x + a.x*b.z;
    r.z = a.w*b.z - a.z*b.w - a.x*b.y + a.y*b.x;
   return r;
}
CVector4 quaternionMultiplyInvB(const CVector4& a, const CVector4& b)
{
    CVector4 r;
   r.w =  a.w*b.w + a.x*b.x + a.y*b.y + a.z*b.z;
   r.x = -a.w*b.x + a.x*b.w - a.y*b.z + a.z*b.y;
   r.y = -a.w*b.y + a.y*b.w - a.z*b.x + a.x*b.z;
   r.z = -a.w*b.z + a.z*b.w - a.x*b.y + a.y*b.x;
   return r;
}
CVector4 quaternionMultiplyInvAB(const CVector4& a, const CVector4& b)
{
    CVector4 r;
    r.w =  a.w*b.w - a.x*b.x - a.y*b.y - a.z*b.z;
  r.x = -a.w*b.x - a.x*b.w + a.y*b.z - a.z*b.y;
  r.y = -a.w*b.y - a.y*b.w + a.z*b.x - a.x*b.z;
  r.z = -a.w*b.z - a.z*b.w + a.x*b.y - a.y*b.x;
   return r;
}

CVector4 CMatrix::GetRotationAsQuaternion() const
{
/*	CVector4 q;
   float tr,s;
  tr = m[0][0] + m[1][1] + m[2][2];
  if (tr >= 0) {
    s = sqrt (tr + 1);
    q.w = 0.5 * s;
    s =0.5 * (1.0f/s);
    q.x = (m[2][1] - m[1][2]) * s;
    q.y = (m[0][2] - m[2][0]) * s;
    q.z = (m[1][0] - m[0][1]) * s;
	return q;
  }
  else {
    if (m[1][1] > m[0][0]) {
		if (m[2][2] <= m[1][1]){

         s = sqrt((m[1][1] - (m[2][2] + m[0][0])) + 1);
    q.y = 0.5 * s;
    s =0.5 * (1.0f/s);
    q.z = (m[1][2] + m[2][1]) * s;
    q.x = (m[0][1] + m[1][0]) * s;
    q.w = (m[0][2] - m[2][0]) * s;
		  return q;
		};
    } else	if (m[2][2] <= m[0][0]) {

        s = sqrt((m[0][0] - (m[1][1] + m[2][2])) + 1);
    q.x = 0.5 * s;
    s =0.5 * (1.0f/s);
    q.y = (m[0][1] + m[1][0]) * s;
    q.z = (m[2][0] + m[0][2]) * s;
    q.w = (m[2][1] - m[1][2]) * s;
		return q;
	};
    s = sqrt((m[2][2] - (m[0][0] + m[1][1])) + 1);
    q.z = 0.5 * s;
    s =0.5 * (1.0f/s);
    q.x = (m[2][0] + m[0][2]) * s;
    q.y = (m[1][2] + m[2][1]) * s;
    q.w = (m[1][0] - m[0][1]) * s;
	return q;
  }*/

#define ROT_ACC(i,j) m[i][j]

    CVector4 q;
       utifloat tr,s;
      tr = ROT_ACC(0,0) + ROT_ACC(1,1) + ROT_ACC(2,2);
      if (tr >= 0) {
        s = sqrt (tr + 1);
        q.w = 0.5 * s;
        s =0.5 * (1.0f/s);
        q.x = (ROT_ACC(2,1) - ROT_ACC(1,2)) * s;
        q.y = (ROT_ACC(0,2) - ROT_ACC(2,0)) * s;
        q.z = (ROT_ACC(1,0) - ROT_ACC(0,1)) * s;
        return q;
      }
      else {
        if (ROT_ACC(1,1) > ROT_ACC(0,0)) {
            if (ROT_ACC(2,2) <= ROT_ACC(1,1)){

             s = sqrt((ROT_ACC(1,1) - (ROT_ACC(2,2) + ROT_ACC(0,0))) + 1);
        q.y = 0.5 * s;
        s =0.5 * (1.0f/s);
        q.z = (ROT_ACC(1,2) + ROT_ACC(2,1)) * s;
        q.x = (ROT_ACC(0,1) + ROT_ACC(1,0)) * s;
        q.w = (ROT_ACC(0,2) - ROT_ACC(2,0)) * s;
              return q;
            };
        } else	if (ROT_ACC(2,2) <= ROT_ACC(0,0)) {

            s = sqrt((ROT_ACC(0,0) - (ROT_ACC(1,1) + ROT_ACC(2,2))) + 1);
        q.x = 0.5 * s;
        s =0.5 * (1.0f/s);
        q.y = (ROT_ACC(0,1) + ROT_ACC(1,0)) * s;
        q.z = (ROT_ACC(2,0) + ROT_ACC(0,2)) * s;
        q.w = (ROT_ACC(2,1) - ROT_ACC(1,2)) * s;
            return q;
        };
        s = sqrt((ROT_ACC(2,2) - (ROT_ACC(0,0) + ROT_ACC(1,1))) + 1);
        q.z = 0.5 * s;
        s =0.5 * (1.0f/s);
        q.x = (ROT_ACC(2,0) + ROT_ACC(0,2)) * s;
        q.y = (ROT_ACC(1,2) + ROT_ACC(2,1)) * s;
        q.w = (ROT_ACC(1,0) - ROT_ACC(0,1)) * s;
        return q;
      }
}

CVector4f CMatrix::GetRotationAsQuaternionf() const
{
    EASY_FUNCTION(profiler::colors::Green);
/*	CVector4 q;
   float tr,s;
  tr = m[0][0] + m[1][1] + m[2][2];
  if (tr >= 0) {
    s = sqrt (tr + 1);
    q.w = 0.5 * s;
    s =0.5 * (1.0f/s);
    q.x = (m[2][1] - m[1][2]) * s;
    q.y = (m[0][2] - m[2][0]) * s;
    q.z = (m[1][0] - m[0][1]) * s;
    return q;
  }
  else {
    if (m[1][1] > m[0][0]) {
        if (m[2][2] <= m[1][1]){

         s = sqrt((m[1][1] - (m[2][2] + m[0][0])) + 1);
    q.y = 0.5 * s;
    s =0.5 * (1.0f/s);
    q.z = (m[1][2] + m[2][1]) * s;
    q.x = (m[0][1] + m[1][0]) * s;
    q.w = (m[0][2] - m[2][0]) * s;
          return q;
        };
    } else	if (m[2][2] <= m[0][0]) {

        s = sqrt((m[0][0] - (m[1][1] + m[2][2])) + 1);
    q.x = 0.5 * s;
    s =0.5 * (1.0f/s);
    q.y = (m[0][1] + m[1][0]) * s;
    q.z = (m[2][0] + m[0][2]) * s;
    q.w = (m[2][1] - m[1][2]) * s;
        return q;
    };
    s = sqrt((m[2][2] - (m[0][0] + m[1][1])) + 1);
    q.z = 0.5 * s;
    s =0.5 * (1.0f/s);
    q.x = (m[2][0] + m[0][2]) * s;
    q.y = (m[1][2] + m[2][1]) * s;
    q.w = (m[1][0] - m[0][1]) * s;
    return q;
  }*/

#define ROT_ACC(i,j) m[i][j]

    CVector4f q;
       utifloat tr,s;
      tr = ROT_ACC(0,0) + ROT_ACC(1,1) + ROT_ACC(2,2);
      if (tr >= 0) {
        s = sqrt (tr + 1);
        q.w = 0.5 * s;
        s =0.5 * (1.0f/s);
        q.x = (ROT_ACC(2,1) - ROT_ACC(1,2)) * s;
        q.y = (ROT_ACC(0,2) - ROT_ACC(2,0)) * s;
        q.z = (ROT_ACC(1,0) - ROT_ACC(0,1)) * s;
        q.normalize();
        return q;
      }
      else {
        if (ROT_ACC(1,1) > ROT_ACC(0,0)) {
            if (ROT_ACC(2,2) <= ROT_ACC(1,1)){

             s = sqrt((ROT_ACC(1,1) - (ROT_ACC(2,2) + ROT_ACC(0,0))) + 1);
        q.y = 0.5 * s;
        s =0.5 * (1.0f/s);
        q.z = (ROT_ACC(1,2) + ROT_ACC(2,1)) * s;
        q.x = (ROT_ACC(0,1) + ROT_ACC(1,0)) * s;
        q.w = (ROT_ACC(0,2) - ROT_ACC(2,0)) * s;
        q.normalize();
              return q;
            };
        } else	if (ROT_ACC(2,2) <= ROT_ACC(0,0)) {

            s = sqrt((ROT_ACC(0,0) - (ROT_ACC(1,1) + ROT_ACC(2,2))) + 1);
        q.x = 0.5 * s;
        s =0.5 * (1.0f/s);
        q.y = (ROT_ACC(0,1) + ROT_ACC(1,0)) * s;
        q.z = (ROT_ACC(2,0) + ROT_ACC(0,2)) * s;
        q.w = (ROT_ACC(2,1) - ROT_ACC(1,2)) * s;
        q.normalize();
            return q;
        };
        s = sqrt((ROT_ACC(2,2) - (ROT_ACC(0,0) + ROT_ACC(1,1))) + 1);
        q.z = 0.5 * s;
        s =0.5 * (1.0f/s);
        q.x = (ROT_ACC(2,0) + ROT_ACC(0,2)) * s;
        q.y = (ROT_ACC(1,2) + ROT_ACC(2,1)) * s;
        q.w = (ROT_ACC(1,0) - ROT_ACC(0,1)) * s;
        q.normalize();
        return q;
      }
}


bool isQuaternionEqual(const CVector4& quat1, const CVector4& quat2, float threshold)
{
    float d = umin(quat1.distanceTo(quat2), quat1.distanceTo(-quat2));
    return d < threshold;
}

CVector4 farthestQuaternion(const CVector4& self, const CVector4& other)
{
    CVector4 diff = self-other;
    CVector4 sum = self+other;
    if (diff.dotProduct(diff) > sum.dotProduct(sum)) {
        return other;
    }
    return -other;
}

CVector4 nearestQuaternion(const CVector4& self, const CVector4& other)
{
    CVector4 diff = self-other;
    CVector4 sum = self+other;
    //diff.normalize();
    //sum.normalize();
    if (diff.dotProduct(diff) < sum.dotProduct(sum)) {
        return other;
    }
    return -other
            ;;

}


CVector4 slerpQuaternion(const CVector4& self, const CVector4&q, float t)
{

    float magnitude = self.length()*q.length();
    float product = self.dotProduct(q) /magnitude;
    if (absf(product) != 1.0) {
        float sign = product < 0.0f ? -1.0f : 1.0f;

        float theta = acos(sign* product);
        float s1 = sin(sign*t*theta);
        float d = 1.0f/sin(theta);
        float s0 = sin(1.0-t)* theta;

        return CVector4((self.x*s0 + q.x * s1)*d,
                        (self.y*s0 + q.y * s1)*d,
                        (self.z*s0 + q.z * s1)*d,
                        (self.w*s0 + q.w * s1)*d);
    } else

    {
        return self;
    }

}


CVector3 rotateViaQuaternion(const CVector4& self, const CVector3& p)
{
    /*aiQuaterniont q2(0.f,v.x,v.y,v.z), q = *this, qinv = q;
    q.Conjugate();

    q = q*q2*qinv;
    return aiVector3t<TReal>(q.x,q.y,q.z);*/
    CVector4 q2(p.x,p.y, p.z, 0.0f);
    CVector4 q = self;

    CVector4 qinv = self;
   qinv = inverseQuaternion(qinv);
    q2 = quaternionMultiply(quaternionMultiply(q,q2),qinv);

    return q2.xyz();
}


CVector3f rotateViaQuaternion(const CVector4f& self, const CVector3f& p)
{
    /*aiQuaterniont q2(0.f,v.x,v.y,v.z), q = *this, qinv = q;
    q.Conjugate();

    q = q*q2*qinv;
    return aiVector3t<TReal>(q.x,q.y,q.z);*/
    CVector4f q2(p.x,p.y, p.z, 0.0f);
    CVector4f q = self;

    CVector4f qinv = self;
  //  q.normalize();
  //  qinv.normalize();
   qinv = inverseQuaternion(qinv);
    q2 = quaternionMultiply(quaternionMultiply(q,q2),qinv);

    return q2.xyz();
}

CVector3 rotateViaEuler(const CVector3& euler, const CVector3& p)
{
	/*CMatrix m;
	m.SetEulerRotation(euler.x, euler.y, euler.z);

	return m.TransformP(p);*/
     CVector4 quat = eulerToQuaternion(euler);
     CVector3 res = rotateViaQuaternion(quat, p);
     return res;
}

CVector3 rotateViaEuler2(const CVector3& euler, const CVector3& p)
{
	/*CMatrix m;
	m.SetEulerRotation(euler.x, euler.y, euler.z);

	return m.TransformP(p);*/
	/*CVector4 quat = eulerToQuaternion(euler);
	CVector3 res = rotateViaQuaternion(quat, p);
	return res;*/
	CMatrix m;
	m.Rotate(euler);
	return m.TransformP(p);
}

CVector4f eulerToQuaternion(const CVector3f & a)
{

   /* CVector4 q;
    double t0 = std::cos(a.z * 0.5*pi180);
    double t1 = std::sin(a.z * 0.5*pi180);
    double t2 = std::cos(a.x * 0.5*pi180);
    double t3 = std::sin(a.x * 0.5*pi180);
    double t4 = std::cos(a.y * 0.5*pi180);
    double t5 = std::sin(a.y * 0.5*pi180);

    q.w = t0 * t2 * t4 + t1 * t3 * t5;
    q.x = t0 * t3 * t4 - t1 * t2 * t5;
    q.y = t0 * t2 * t5 + t1 * t3 * t4;
    q.z = t1 * t2 * t4 - t0 * t3 * t5;*/
    CMatrix m;
    m.SetEulerRotation(a.x, a.y, a.z);

    CVector4f q = m.GetRotationAsQuaternionf();

#ifndef NDEBUG

    //assert(q2.distanceTo(q) < 0.01f || q2.distanceTo(-q) < 0.01f);
    CMatrix m2;
    m2.SetRotateQuaternion(toDouble(q));
    assert(m.numericalCompare(m2, 0.01));
#endif

    return q;
}



CVector4 eulerToQuaternion(const CVector3 & a)
{

   /* CVector4 q;
    double t0 = std::cos(a.z * 0.5*pi180);
    double t1 = std::sin(a.z * 0.5*pi180);
    double t2 = std::cos(a.x * 0.5*pi180);
    double t3 = std::sin(a.x * 0.5*pi180);
    double t4 = std::cos(a.y * 0.5*pi180);
    double t5 = std::sin(a.y * 0.5*pi180);

    q.w = t0 * t2 * t4 + t1 * t3 * t5;
    q.x = t0 * t3 * t4 - t1 * t2 * t5;
    q.y = t0 * t2 * t5 + t1 * t3 * t4;
    q.z = t1 * t2 * t4 - t0 * t3 * t5;*/
    CMatrix m;
    m.SetEulerRotation(a.x, a.y, a.z);

    CVector4 q = m.GetRotationAsQuaternion();

#ifndef NDEBUG

    //assert(q2.distanceTo(q) < 0.01f || q2.distanceTo(-q) < 0.01f);
    CMatrix m2;
    m2.SetRotateQuaternion(q);
    assert(m.numericalCompare(m2, 0.01));
#endif

    return q;
}

CVector4 eulerToQuaternionOld(const CVector3 & a)
{
    CMatrix m;
    m.SetEulerRotationOld(a.x, a.y, a.z);
    CVector4 q = m.GetRotationAsQuaternion();
#ifndef NDEBUG
    CMatrix m2;
    m2.SetRotateQuaternion(q);
    assert(m.numericalCompare(m2, 0.01));
#endif

    return q;
}

void CMatrix::SetEulerRotation(utifloat x, utifloat y, utifloat z)
{
   /* utifloat sx,cx,sy,cy,sz,cz;
  sx = sin(x*pi180);
  cx = cos(x*pi180);
  sy = sin(y*pi180);
  cy = cos(y*pi180);
  sz = sin(z*pi180);
  cz = cos(z*pi180);
 m[0][0] = cz*cy;
  m[0][1] = sz*cy;
  m[0][2] =-sy;
  m[0][3] = 0.0f;

  m[1][0] = cz*sy*sx - sz*cx;
  m[1][1] = sz*sy*sx + cz*cx;
  m[1][2] = cy*sx;
  m[1][3] = 0.0f;

  m[2][0] = cz*sy*cx + sz*sx;
  m[2][1] = sz*sy*cx - cz*sx;
  m[2][2] = cy*cx;
  m[2][3] = 0.0f;*/

    utifloat s1,c1,s2,c2,s3,c3;
      s1 = sin(y*pi180);
      c1 = cos(y*pi180);
      s2 = sin(x*pi180);
      c2 = cos(x*pi180);
      s3 = sin(z*pi180);
      c3 = cos(z*pi180);

  m[0][0] = c1*c3+s1*s2*s3;
    m[0][1] = c3*s1*s2-c1*s3;
    m[0][2] = c2*s1;
    m[0][3] = 0.0f;

    m[1][0] = c2*s3;
    m[1][1] = c2*c3;
    m[1][2] = -s2;
    m[1][3] = 0.0f;

    m[2][0] = c1*s2*s3-c3*s1;
    m[2][1] = c1*c3*s2+s1*s3;
    m[2][2] = c1*c2;
    m[2][3] = 0.0f;

  m[3][0] = 0.0f;
  m[3][1] = 0.0f;
  m[3][2] = 0.0f;
  m[3][3] = 1.0f;

 /*  CMatrix temp;
   temp.Rotate(x, y, z);
   assert(numericalCompare(temp, 0.01));
*/
};

void CMatrix::SetEulerRotationOld(utifloat x, utifloat y, utifloat z)
{
    utifloat sx,cx,sy,cy,sz,cz;
  sx = sin(x*pi180);
  cx = cos(x*pi180);
  sy = sin(y*pi180);
  cy = cos(y*pi180);
  sz = sin(z*pi180);
  cz = cos(z*pi180);
 m[0][0] = cz*cy;
  m[0][1] = sz*cy;
  m[0][2] =-sy;
  m[0][3] = 0.0f;

  m[1][0] = cz*sy*sx - sz*cx;
  m[1][1] = sz*sy*sx + cz*cx;
  m[1][2] = cy*sx;
  m[1][3] = 0.0f;

  m[2][0] = cz*sy*cx + sz*sx;
  m[2][1] = sz*sy*cx - cz*sx;
  m[2][2] = cy*cx;
  m[2][3] = 0.0f;

  m[3][0] = 0.0f;
  m[3][1] = 0.0f;
  m[3][2] = 0.0f;
  m[3][3] = 1.0f;

 /*  CMatrix temp;
   temp.Rotate(x, y, z);
   assert(numericalCompare(temp, 0.01));
*/
};


void CMatrix::Rotation(const NetRotationInfo& rot)
{
    if (rot.m_rotationtype == nrtEuler) {
        EulerRotation(rot.m_values.x, rot.m_values.y, rot.m_values.z);
    } else {
        RotateQuaternion(rot.m_values);
    }
}

void CMatrix::EulerRotation(utifloat x, utifloat y, utifloat z)
{
   CMatrix l_mat;
   l_mat.SetEulerRotation(x, y, z);
   multipyMatrix(l_mat);
}

CVector CMatrix::rotateEuler(const CVector& rot, const CVector& point)
{
   CMatrix m;
   m.SetEulerRotation(rot.x, rot.y, rot.z);
   return m.TransformP(point);
}

void CMatrix::EulerRotation(const CVector3& v)
{
   CMatrix l_mat;
   l_mat.SetEulerRotation(v.x, v.y, v.z);
   multipyMatrix(l_mat);
}

void CMatrix::Scal(utifloat x, utifloat y, utifloat z)
{
  CMatrix l_mat;
    l_mat.SetScaling(x, y, z);
    multipyMatrix(l_mat);
}

void CMatrix::Scal(const CVector3& v)
{
    CMatrix l_mat;
    l_mat.SetScaling(v.x, v.y, v.z);
    multipyMatrix(l_mat);
}

void CMatrix::ShearXfromY(utifloat x)
{
  CMatrix m;
  m.LoadIdentity();
  m.m[0][1] = x;
  multipyMatrix(m);
}

void CMatrix::ShearYfromX(utifloat x)
{
  CMatrix m;
  m.LoadIdentity();
  m.m[1][0] = x;
  multipyMatrix(m);
}



void CMatrix::Rotate(utifloat _x, utifloat _y, utifloat _z)
{
    CMatrix mMat;

	if (_y != 0) {
	   mMat.SetYRotation(_y);
	   multipyMatrix(mMat);
	}

	if (_x != 0) {
	   mMat.SetXRotation(_x);
      multipyMatrix(mMat);
	}

	if (_z != 0) {
	   mMat.SetZRotation(_z);
   	multipyMatrix(mMat);
	}
}

void CMatrix::RotateInv(utifloat _x, utifloat _y, utifloat _z)
{
    CMatrix mMat;
if (_z != 0) {
	   mMat.SetZRotation(-_z);
   	multipyMatrix(mMat);
	}

	if (_x != 0) {
	   mMat.SetXRotation(-_x);
      multipyMatrix(mMat);
	}
if (_y != 0) {
	   mMat.SetYRotation(-_y);
	   multipyMatrix(mMat);
	}


}


void CMatrix::SetFrustum(utifloat l, utifloat r, utifloat b, utifloat t, utifloat n, utifloat f)
{
   m[0][0] = (2*n)/(r-l); m[0][1] = 0; m[0][2] = (r+l)/(r-l); m[0][3] = 0;
   m[1][0] = 0; m[1][1] = (2*n)/(t-b); m[1][2] = (t+b)/(t-b); m[1][3] = 0;
   m[2][0] = 0; m[2][1] = 0; m[2][2] = -(f+n)/(f-n); m[2][3] = -(2*n*f)/(f-n);
   m[3][0] = 0; m[3][1] = 0;m[3][2] = -1; m[3][3] = 0;
}

void CMatrix::SetOrtho(utifloat l, utifloat r, utifloat b, utifloat t, utifloat n, utifloat f)
{
   m[0][0] = (2)/(r-l); m[0][1] = 0; m[0][2] = 0; m[0][3] = -(r+l)/(r-l);
   m[1][0] = 0; m[1][1] = (2)/(t-b); m[1][2] = 0; m[1][3] =-(t+b)/(t-b);
   m[2][0] = 0; m[2][1] = 0; m[2][2] = -(2)/(f-n); m[2][3] = -(f+n)/(f-n);
   m[3][0] = 0; m[3][1] = 0; m[3][2] = 0; m[3][3] = 1;
}

void CMatrix::SetPerspective(utifloat fovy, utifloat aspect, utifloat near_plane, utifloat far_plane)
{
   utifloat angle = tan(fovy/180.0f*PI/2.0f)*near_plane;
   SetFrustum(-aspect*angle, aspect*angle, -angle, angle, near_plane, far_plane);
}

void CMatrix::Frustum(utifloat l, utifloat r, utifloat b, utifloat t, utifloat n, utifloat f)
{
   CMatrix mMat;
   mMat.SetFrustum(l, r, b, t, n, f);
   multipyMatrix(mMat);
}

void CMatrix::Ortho(utifloat l, utifloat r, utifloat b, utifloat t, utifloat n, utifloat f)
{
   CMatrix mMat;
   mMat.SetOrtho(l, r, b, t, n, f);
   multipyMatrix(mMat);
}

void CMatrix::Perspective(utifloat fovy, utifloat aspect, utifloat near_plane, utifloat far_plane)
{
   CMatrix mMat;
   mMat.SetPerspective(fovy, aspect, near_plane, far_plane);
   multipyMatrix(mMat);
}


void CMatrix::Translation(utifloat _x, utifloat _y, utifloat _z)
{

   if ((_x != 0) || (_y != 0) || (_z != 0)) {
	   CMatrix mMat;
      mMat.SetTranslate(_x, _y, _z);
		multipyMatrix(mMat);
	}
}

void CMatrix::RotateAroundPoint(const CVector3& point, utifloat angle, int axis)
{ Translation(point.x, point.y, point.z);

  switch (axis) {
    case 0: {
      Rotate(angle, 0, 0);
    }; break;
    case 1: {
      Rotate(0,angle, 0);
    }; break;
    case 2: {
      Rotate(0,0,angle);
    }; break;
  };
   Translation(-point.x, -point.y, -point.z);

}

void CMatrix::ShearAroundPoint(const CVector3& point, utifloat amount, int axis)
{ Translation(point.x, point.y, point.z);

  switch (axis) {
    case 0: {
      ShearXfromY(amount);
    }; break;
    case 1: {
      ShearYfromX(amount);
    }; break;

  };
   Translation(-point.x, -point.y, -point.z);

}

void CMatrix::ScalAroundPoint(const CVector3& point, const CVector3& scal)
{
  Translation(point.x, point.y, point.z);

   Scaling(scal.x, scal.y, scal.z);

   Translation(-point.x, -point.y, -point.z);
}


CMatrix CMatrix::TranslationMatrix(utifloat x, utifloat y, utifloat z)
{
  CMatrix mMat;
      mMat.SetTranslate(x, y, z);
      return mMat;
}


void CMatrix::Scaling(utifloat _x, utifloat _y, utifloat _z)
{

   if ((_x != 1) || (_y != 1) || (_z != 1)) {
	  CMatrix mMat;
      mMat.SetScaling(_x, _y, _z);
		multipyMatrix(mMat);
	}
}


bool CMatrix::numericalCompare(const CMatrix & b, utifloat threshold)
{
    int i, j;

     for (i=0; i<4; i++) {
        for (j=0; j<4; j++) {

            if (absf(m[i][j] - b.m[i][j]) > threshold) return false;

          }
     }
     return true;
}


utifloat CMatrix::numericalDistance(const CMatrix & b)
{
    int i, j;
    utifloat maxdistance = 0.0f;

     for (i=0; i<4; i++) {
        for (j=0; j<4; j++) {

            maxdistance = umax(maxdistance, absf(m[i][j] - b.m[i][j]));

          }
     }
     return maxdistance;
}


bool CMatrix::operator==(const CMatrix & b)
{
  int i, j;

   for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {

          if (m[i][j] != b.m[i][j]) return false;

		}
   }
   return true;
}



CMatrix CMatrix::operator*(const CMatrix& b) const
{
   CMatrix mMat;



   //int i, j, k;

   //utifloat sum;
#define CALC_ONE_MATRIX_ELEMENT(i,j) mMat.m[i][j] = m[i][0] * b.m[0][j] + m[i][1] * b.m[1][j] + m[i][2] * b.m[2][j] + m[i][3] * b.m[3][j];

   /*for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
			for (k=0; k<4; k++) {
            sum += m[i][k] * b.m[k][j];
			}
			//mMat.m[i][j] = sum;
		}
   }*/
   CALC_ONE_MATRIX_ELEMENT(0, 0);
   CALC_ONE_MATRIX_ELEMENT(0, 1);
	CALC_ONE_MATRIX_ELEMENT(0, 2);

	CALC_ONE_MATRIX_ELEMENT(0, 3);
	CALC_ONE_MATRIX_ELEMENT(1, 0);
	CALC_ONE_MATRIX_ELEMENT(1, 1);
	CALC_ONE_MATRIX_ELEMENT(1, 2);

	CALC_ONE_MATRIX_ELEMENT(1, 3);
	CALC_ONE_MATRIX_ELEMENT(2, 0);
	CALC_ONE_MATRIX_ELEMENT(2, 1);
	CALC_ONE_MATRIX_ELEMENT(2, 2);

	CALC_ONE_MATRIX_ELEMENT(2, 3);
	CALC_ONE_MATRIX_ELEMENT(3, 0);
	CALC_ONE_MATRIX_ELEMENT(3, 1);
	CALC_ONE_MATRIX_ELEMENT(3, 2);

	CALC_ONE_MATRIX_ELEMENT(3, 3);
   return mMat;
}

void CMatrix::print() const
{
//	int i, j;

//   cout << UTEXT("\n");
  /* for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
         wcout << m[j][i] << UTEXT("   | ");
		}
		wcout << UTEXT("\n");
	}*/

}

C3DPoint CMatrix::Transform(const C3DPoint & b) const
{
	C3DPoint r;
	r.x = b.x * m[0][0] + b.y * m[1][0] + b.z * m[2][0] + m[0][3];
	r.y = b.x * m[0][1] + b.y * m[1][1] + b.z * m[2][1] + m[1][3];
	r.z = b.x * m[0][2] + b.y * m[1][2] + b.z * m[2][2] + m[2][3];
   return r;
}


CBox CMatrix::TransformP(const CBox& box) const
{
	CBox m;
	m = CBox::ADVANCEINIT();
	m.expand(TransformP(CVector3(box.m_min.x, box.m_min.y, box.m_min.z)));
	m.expand(TransformP(CVector3(box.m_max.x, box.m_min.y, box.m_min.z)));
	m.expand(TransformP(CVector3(box.m_min.x, box.m_max.y, box.m_min.z)));
	m.expand(TransformP(CVector3(box.m_max.x, box.m_max.y, box.m_min.z)));
	m.expand(TransformP(CVector3(box.m_min.x, box.m_min.y, box.m_max.z)));
	m.expand(TransformP(CVector3(box.m_max.x, box.m_min.y, box.m_max.z)));
	m.expand(TransformP(CVector3(box.m_min.x, box.m_max.y, box.m_max.z)));
	m.expand(TransformP(CVector3(box.m_max.x, box.m_max.y, box.m_max.z)));
	return m;
}

void CMatrix::TransformNorm(S3DPoint *p) const
{
   utifloat x, y, z;
   x = p->nx * m[0][0] + p->ny * m[0][1] + p->nz * m[0][2];
   y = p->nx * m[1][0] + p->ny * m[1][1] + p->nz * m[1][2];
   z = p->nx * m[2][0] + p->ny * m[2][1] + p->nz * m[2][2];
   p->nx = x;
   p->ny = y;
   p->nz = z;
}

void CMatrix::TransformAll(S3DPoint *p) const
{
   utifloat x, y, z;
   x = p->x * m[0][0] + p->y * m[0][1] + p->z * m[0][2] + m[0][3];
   y = p->x * m[1][0] + p->y * m[1][1] + p->z * m[1][2] + m[1][3];
   z = p->x * m[2][0] + p->y * m[2][1] + p->z * m[2][2] + m[2][3];
   p->x = x;
   p->y = y;
   p->z = z;
   x = p->nx * m[0][0] + p->ny * m[0][1] + p->nz * m[0][2];
   y = p->nx * m[1][0] + p->ny * m[1][1] + p->nz * m[1][2];
   z = p->nx * m[2][0] + p->ny * m[2][1] + p->nz * m[2][2];
   p->nx = x;
   p->ny = y;
   p->nz = z;
}

CVector4 CMatrix::getRow(int index)const
{
  return CVector4(m[index][0], m[index][1], m[index][2],m[index][3]);
}
void CMatrix::setRow(int index, const CVector4& v)
{
  m[index][0] = v.x;
  m[index][1] = v.y;
  m[index][2] = v.z;
  m[index][3] = v.w;
}


template<class T> void uarcsin2(T v, T&a, T&b)
{
    a = asin(v);
    b = PI-a;
};

template<class T> void uarccos2(T v, T&a, T&b)
{
    a = acos(v);
    b = 2*PI-a;
};

template<class T> bool zero(T a)
{
    if (a < 0) a= -a;
    return a < 0.001;
};

Matrix3 CMatrix::getRotationMatrix()
{  Matrix3 res;
    res.m[0][0]=m[0][0];
    res.m[0][1]=m[0][1];
    res.m[0][2]=m[0][2];
    res.m[1][0]=m[1][0];
    res.m[1][1]=m[1][1];
    res.m[1][2]=m[1][2];
    res.m[2][0]=m[2][0];
    res.m[2][1]=m[2][1];
    res.m[2][2]=m[2][2];
    return res;
}



CVector rotateEuler(const CVector& point,const CVector& rot)
{
    return CMatrix::rotateEuler(rot, point);
}

void utianalysematrix(Matrix3* m, utifloat &a, utifloat &b, utifloat &c)
{
    utifloat avalue[2];
    utifloat acosvalue[2];
    utifloat bvalue[2];
    utifloat cvalue[2];
    bvalue[0] = 0;
    bvalue[1] = 0;
    if (zero(m->m[1][2]-1.0)) {
        //falls sin(a)=1
        a = asin(1.0);
        c=0;
        utifloat i1, i2;
        uarccos2(m->m[0][0], i1, i2);
        if (fabs(sin(i1)-m->m[0][1]) < fabs(sin(i2)-m->m[0][1])) {
            b = i1;
        } else {
            b = i2;
        };
        return;

    };
    if (zero(m->m[1][2]+1.0)) {
        //falls sin(a)=-1
        a = asin(-1.0);
        c=0;
        utifloat i1, i2;
        uarccos2(m->m[0][0], i1, i2);
        if (fabs((-sin(i1))-m->m[0][1]) < fabs((-sin(i2))-m->m[0][1])) {
            b = i1;
        } else {
            b = i2;
        };
        return;

    };
    uarcsin2(m->m[1][2], avalue[0], avalue[1]);

    acosvalue[0] = cos(avalue[0]);
    acosvalue[1] = cos(avalue[1]);
    //c

    {
        utifloat v1 = m->m[1][1]/acosvalue[0];
        utifloat v2 = m->m[1][0]/(-acosvalue[0]);
        utifloat i1, i2;
        uarccos2(v1, i1, i2);
        if (fabs(i1-v2) < fabs(i2-v2)) {
            cvalue[0] = i1;
        } else {
            cvalue[0] = i2;

        };
    };


    {
        utifloat v1 = m->m[1][1]/acosvalue[1];
        utifloat v2 = m->m[1][0]/(-acosvalue[1]);
        utifloat i1, i2;
        uarccos2(v1, i1, i2);
        if (fabs(i1-v2) < fabs(i2-v2)) {
            cvalue[1] = i1;
        } else {
            cvalue[1] = i2;
        };
    }
//b

    {
        utifloat v1 = m->m[2][2]/acosvalue[0];
        utifloat v2 = m->m[0][2]/(-acosvalue[0]);
        utifloat i1, i2;
        uarccos2(v1, i1, i2);
        if (fabs(i1-v2) < fabs(i2-v2)) {
            bvalue[0] = i1;
        } else {
            bvalue[0] = i2;

        };
    };


    {
        utifloat v1 = m->m[2][2]/acosvalue[1];
        utifloat v2 = m->m[0][2]/(-acosvalue[1]);
        utifloat i1, i2;
        uarccos2(v1, i1, i2);
        if (fabs(i1-v2) < fabs(i2-v2)) {
            bvalue[1] = i1;
        } else {
            bvalue[1] = i2;
        };
    };
//   cout << "avalue[0] = " << avalue[0] << endl;
//   cout << "avalue[1] = " << avalue[1] << endl;
//   cout << "bvalue[0] = " << bvalue[0] << endl;
//   cout << "bvalue[1] = " << bvalue[1] << endl;
//   cout << "cvalue[0] = " << cvalue[0] << endl;
//   cout << "cvalue[1] = " << cvalue[1] << endl;
    utifloat v[2];
    for (int i = 0; i < 2; i++) {
        utifloat a, b, c;
        a = avalue[i];
        b = bvalue[i];
        c = cvalue[i];
        v[i] = fabs(m->m[0][0] -(cos(b)*cos(c)-sin(b)*sin(a)*sin(c)))+
               fabs(m->m[0][1] - (cos(b)*sin(c)+sin(b)*sin(a)*cos(c)))+
               fabs(m->m[2][0] - (sin(b)*cos(c)+cos(b)*sin(a)*sin(c)))+
               fabs(m->m[2][1] - (sin(b)*sin(c)-cos(b)*sin(a)*cos(c)));
    };
    if (v[0] < v[1]) {
        a = avalue[0];
        b = bvalue[0];
        c = cvalue[0];
    } else {
        a = avalue[1];
        b = bvalue[1];
        c = cvalue[1];
    }
};

CVector4 CMatrix::getColumn(int index) const
{
  return CVector4(m[0][index], m[1][index], m[2][index],m[3][index]);
}
void CMatrix::setColumn(int index, const CVector4& v)
{
  m[0][index] = v.x;
  m[1][index] = v.y;
  m[2][index] = v.z;
  m[3][index] = v.w;
}

void CMatrix::Transpose(CMatrix* target)
{
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			target->m[j][i] = m[i][j];
		}
	};
}

void Perform_OpenGL_matrix_inverse(float* _in,float* _out)
{
	float scale;
   scale = _in[TBGL[0*4 + 0]] * _in[TBGL[0*4 + 0]] +
            _in[TBGL[0*4 + 1]] * _in[TBGL[0*4 + 1]] +
            _in[TBGL[0*4 + 2]] * _in[TBGL[0*4 + 2]];
   scale = 1/scale;
   _out[TBGL[0 * 4 + 0]] = scale * _in[TBGL[0*4 + 0]];
   _out[TBGL[1 * 4 + 0]] = scale * _in[TBGL[0*4 + 1]];
   _out[TBGL[2 * 4 + 0]] = scale * _in[TBGL[0*4 + 2]];
   _out[TBGL[0 * 4 + 1]] = scale * _in[TBGL[1*4 + 0]];
   _out[TBGL[1 * 4 + 1]] = scale * _in[TBGL[1*4 + 1]];
   _out[TBGL[2 * 4 + 1]] = scale * _in[TBGL[1*4 + 2]];
   _out[TBGL[0 * 4 + 2]] = scale * _in[TBGL[2*4 + 0]];
   _out[TBGL[1 * 4 + 2]] = scale * _in[TBGL[2*4 + 1]];
   _out[TBGL[2 * 4 + 2]] = scale * _in[TBGL[2*4 + 2]];

   _out[TBGL[0*4 + 3]]   = -(_out[TBGL[0*4 + 0]] * _in[TBGL[0*4 + 3]] +
                        _out[TBGL[0*4 + 1]] * _in[TBGL[1*4 + 3]] +
                        _out[TBGL[0*4 + 2]] * _in[TBGL[2*4 + 3]]);
   _out[TBGL[1*4 + 3]]   = -(_out[TBGL[1*4 + 0]] * _in[TBGL[0*4 + 3]] +
                        _out[TBGL[1*4 + 1]] * _in[TBGL[1*4 + 3]] +
                        _out[TBGL[1*4 + 2]] * _in[TBGL[2*4 + 3]]);
   _out[TBGL[2*4 + 3]]   = -(_out[TBGL[2*4 + 0]] * _in[TBGL[0*4 + 3]] +
                        _out[TBGL[2*4 + 1]] * _in[TBGL[1*4 + 3]] +
                        _out[TBGL[2*4 + 2]] * _in[TBGL[2*4 + 3]]);
   _out[TBGL[3*4 + 0]] = 0;
   _out[TBGL[3*4 + 1]] = 0;
   _out[TBGL[3*4 + 2]] = 0;
   _out[TBGL[3*4 + 3]] = 1;
}

void CMatrix::multipyMatrix(const CMatrix & b)
{
 //  CMatrix mMat;
   double km[4][4];
   int i, j;


   for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
         /*sum = 0;
			for (k=0; k<4; k++) {
            sum += m[i][k] * b.m[k][j];
			}
			km[i][j] = sum;*/
          km[i][j] = (double)m[i][0] * (double)b.m[0][j]+
                  (double)m[i][1] * (double)b.m[1][j]+
                  (double)m[i][2] * (double)b.m[2][j]+
                  (double)m[i][3] * (double)b.m[3][j];

		}
   }

	//Kopieren
   //memcpy(m, km, 64);
   for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
         m[i][j] =km[i][j];
		}
	}
}
bool CPlane::IntersectTriangle(S3DPoint* p1, S3DPoint* p2, S3DPoint* p3, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   CPlane p;
   p.GetNormal(p1, p2, p3);
   S3DPoint res;
   utifloat dis;
   if (!IntersectLinePlaneV(p.d, &p.cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
	   /*S3DPoint pl[3];
	   pl[0] = *p1;
	   pl[1] = *p2;
	   pl[2] = *p3;*/
	   S3DPoint pl[3];
	   pl[0].x = p1->x;
	   pl[0].y = p1->y;
	   pl[0].z = p1->z;

	   pl[1].x = p2->x;
	   pl[1].y = p2->y;
	   pl[1].z = p2->z;

	   pl[2].x = p3->x;
	   pl[2].y = p3->y;
	   pl[2].z = p3->z;
       bool r = false;
	   switch (p.cvVector.getMajorAxis()) {
	   case 0: {
		   r = Has2DPointX(pl, 3, &res);
			   }; break;
		case 1: {
		   r = Has2DPointY(pl, 3, &res);
			   }; break;
		case 2: {
		   r = Has2DPointZ(pl, 3, &res);
			   }; break;
		default: utiassert(false);break;
	   };
	   if (r) {
    	  HitRec->distance = dis;
	      HitRec->point.x = res.x;
		   HitRec->point.y = res.y;
		   HitRec->point.z = res.z;
		   HitRec->point.nx = p.cvVector.x;
		   HitRec->point.ny = p.cvVector.y;
		   HitRec->point.nz= p.cvVector.z;
		   HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
		   return(true);
	  };
	  return(false);
   };
   return(false);
};

bool CPlane::IntersectTriangle(const CVector& pos, const CVector& direction, const CVector& p1, const CVector& p2, const CVector& p3, utifloat& distance)
{
  // CPlane p;
   //p.GetNormal(p1, p2, p3);
assert(false);
return false;
}



bool CPlane::IntersectTriangle(const CVector3f* p1, const CVector3f* p2, const CVector3f* p3, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   CPlane p;
   p.GetNormal(p1, p2, p3);

   S3DPoint res;
   utifloat dis;
   if (!IntersectLinePlaneV(p.d, &p.cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
       /*S3DPoint pl[3];
       pl[0] = *p1;
       pl[1] = *p2;
       pl[2] = *p3;*/
       S3DPoint pl[3];
       pl[0].x = p1->x;
       pl[0].y = p1->y;
       pl[0].z = p1->z;

       pl[1].x = p2->x;
       pl[1].y = p2->y;
       pl[1].z = p2->z;

       pl[2].x = p3->x;
       pl[2].y = p3->y;
       pl[2].z = p3->z;
       bool r = false;
       switch (p.cvVector.getMajorAxis()) {
       case 0: {
           r = Has2DPointX(pl, 3, &res);
               }; break;
        case 1: {
           r = Has2DPointY(pl, 3, &res);
               }; break;
        case 2: {
           r = Has2DPointZ(pl, 3, &res);
               }; break;
        default: utiassert(false);break;
       };
       if (r) {
          HitRec->distance = dis;
          HitRec->point.x = res.x;
           HitRec->point.y = res.y;
           HitRec->point.z = res.z;
           HitRec->point.nx = p.cvVector.x;
           HitRec->point.ny = p.cvVector.y;
           HitRec->point.nz= p.cvVector.z;
           HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
           return(true);
      };
      return(false);
   };
   return(false);
};


bool CPlane::IntersectTriangle(const CVector* p1, const CVector* p2, const CVector* p3, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   CPlane p;
   p.GetNormal(p1, p2, p3);

   S3DPoint res;
   utifloat dis;
   if (!IntersectLinePlaneV(p.d, &p.cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
	   /*S3DPoint pl[3];
	   pl[0] = *p1;
	   pl[1] = *p2;
	   pl[2] = *p3;*/
	   S3DPoint pl[3];
	   pl[0].x = p1->x;
	   pl[0].y = p1->y;
	   pl[0].z = p1->z;

	   pl[1].x = p2->x;
	   pl[1].y = p2->y;
	   pl[1].z = p2->z;

	   pl[2].x = p3->x;
	   pl[2].y = p3->y;
	   pl[2].z = p3->z;
       bool r = false;
	   switch (p.cvVector.getMajorAxis()) {
	   case 0: {
		   r = Has2DPointX(pl, 3, &res);
			   }; break;
		case 1: {
		   r = Has2DPointY(pl, 3, &res);
			   }; break;
		case 2: {
		   r = Has2DPointZ(pl, 3, &res);
			   }; break;
		default: utiassert(false);break;
	   };
	   if (r) {
    	  HitRec->distance = dis;
	      HitRec->point.x = res.x;
		   HitRec->point.y = res.y;
		   HitRec->point.z = res.z;
		   HitRec->point.nx = p.cvVector.x;
		   HitRec->point.ny = p.cvVector.y;
		   HitRec->point.nz= p.cvVector.z;
		   HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
		   return(true);
	  };
	  return(false);
   };
   return(false);
};


bool CPlane::PointInTriangle(const CVector* p1, const CVector* p2, const CVector* p3, const CVector3* point)
{
   CPlane p;
   p.GetNormal(p1, p2, p3);

   if (absf(p.PlaneDistance(*point))> 0.05) return false;
   S3DPoint res;
   res.setPosition(*point);
   //utifloat dis;
       /*S3DPoint pl[3];
       pl[0] = *p1;
       pl[1] = *p2;
       pl[2] = *p3;*/
       S3DPoint pl[3];
       pl[0].x = p1->x;
       pl[0].y = p1->y;
       pl[0].z = p1->z;

       pl[1].x = p2->x;
       pl[1].y = p2->y;
       pl[1].z = p2->z;

       pl[2].x = p3->x;
       pl[2].y = p3->y;
       pl[2].z = p3->z;
       bool r = false;
       switch (p.cvVector.getMajorAxis()) {
       case 0: {
           r = Has2DPointX(pl, 3, &res);
               }; break;
        case 1: {
           r = Has2DPointY(pl, 3, &res);
               }; break;
        case 2: {
           r = Has2DPointZ(pl, 3, &res);
               }; break;
        default: utiassert(false);break;
       };
       if (r) {

           return(true);
      };
  return false;
};

bool CPlane::IntersectTriangleWithV(S3DPoint* p1, S3DPoint* p2, S3DPoint* p3, CPlane *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   S3DPoint res;
   utifloat dis;
   if (!IntersectLinePlaneV(p->d, &p->cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
	   /*S3DPoint pl[3];
	   pl[0] = *p1;
	   pl[1] = *p2;
	   pl[2] = *p3;*/
	   S3DPoint pl[3];
	   pl[0].x = p1->x;
	   pl[0].y = p1->y;
	   pl[0].z = p1->z;

	   pl[1].x = p2->x;
	   pl[1].y = p2->y;
	   pl[1].z = p2->z;

	   pl[2].x = p3->x;
	   pl[2].y = p3->y;
	   pl[2].z = p3->z;
       bool r = false;
	   switch (p->cvVector.getMajorAxis()) {
	   case 0: {
		   r = Has2DPointX(pl, 3, &res);
			   }; break;
		case 1: {
		   r = Has2DPointY(pl, 3, &res);
			   }; break;
		case 2: {
		   r = Has2DPointZ(pl, 3, &res);
			   }; break;
		default: utiassert(false);break;
	   };
	   if (r) {
    	  HitRec->distance = dis;
	      HitRec->point.x = res.x;
		   HitRec->point.y = res.y;
		   HitRec->point.z = res.z;
		   HitRec->point.nx = p->cvVector.x;
		   HitRec->point.ny = p->cvVector.y;
		   HitRec->point.nz= p->cvVector.z;
		   HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
		   return(true);
	  };
	  return(false);
   };
   return(false);
};

bool CPlane::IntersectTriangleWithV(CVector* p1, CVector* p2, CVector* p3, CPlane *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   S3DPoint res;
   utifloat dis;
   //p->GetNormal(p1, p2, p3);


   if (!IntersectLinePlaneV(p->d, &p->cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
	   /*S3DPoint pl[3];
	   pl[0] = *p1;
	   pl[1] = *p2;
	   pl[2] = *p3;*/
	   S3DPoint pl[3];
      pl[0].x = p1->x;
	   pl[0].y = p1->y;
	   pl[0].z = p1->z;

	   pl[1].x = p2->x;
	   pl[1].y = p2->y;
	   pl[1].z = p2->z;

	   pl[2].x = p3->x;
	   pl[2].y = p3->y;
	   pl[2].z = p3->z;
      // bool r = Has2DPointX(pl, 3, &res) || Has2DPointY(pl, 3, &res) || Has2DPointZ(pl, 3, &res);
	   bool r = false;
      switch (p->cvVector.getMajorAxis()) {
	   case 0: {
		   r = Has2DPointX(pl, 3, &res);
			   }; break;
		case 1: {
		   r = Has2DPointY(pl, 3, &res);
			   }; break;
		case 2: {
		   r = Has2DPointZ(pl, 3, &res);
			   }; break;
		default: utiassert(false);break;
	   };
	   if (r) {
    	  HitRec->distance = dis;
	      HitRec->point.x = res.x;
		   HitRec->point.y = res.y;
		   HitRec->point.z = res.z;
		   HitRec->point.nx = p->cvVector.x;
		   HitRec->point.ny = p->cvVector.y;
		   HitRec->point.nz= p->cvVector.z;
		   HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
		   return(true);
	  };
	  return(false);
   };
   return(false);
};


bool CPlane::IntersectTriangleWithV(CVector3f* p1, CVector3f* p2, CVector3f* p3, CPlanef *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   S3DPoint res;
   utifloat dis;
   //p->GetNormal(p1, p2, p3);


   if (!IntersectLinePlaneV(p->d, &p->cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
       /*S3DPoint pl[3];
       pl[0] = *p1;
       pl[1] = *p2;
       pl[2] = *p3;*/
       S3DPoint pl[3];
      pl[0].x = p1->x;
       pl[0].y = p1->y;
       pl[0].z = p1->z;

       pl[1].x = p2->x;
       pl[1].y = p2->y;
       pl[1].z = p2->z;

       pl[2].x = p3->x;
       pl[2].y = p3->y;
       pl[2].z = p3->z;
      // bool r = Has2DPointX(pl, 3, &res) || Has2DPointY(pl, 3, &res) || Has2DPointZ(pl, 3, &res);
       bool r = false;
      switch (p->cvVector.getMajorAxis()) {
       case 0: {
           r = Has2DPointX(pl, 3, &res);
               }; break;
        case 1: {
           r = Has2DPointY(pl, 3, &res);
               }; break;
        case 2: {
           r = Has2DPointZ(pl, 3, &res);
               }; break;
        default: utiassert(false);break;
       };
       if (r) {
          HitRec->distance = dis;
          HitRec->point.x = res.x;
           HitRec->point.y = res.y;
           HitRec->point.z = res.z;
           HitRec->point.nx = p->cvVector.x;
           HitRec->point.ny = p->cvVector.y;
           HitRec->point.nz= p->cvVector.z;
           HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
           return(true);
      };
      return(false);
   };
   return(false);
};


bool CPlane::IntersectTriangleWithV(CVector** points, CPlane *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   S3DPoint res;
   utifloat dis;
   //p->GetNormal(p1, p2, p3);


   if (!IntersectLinePlaneV(p->d, &p->cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
	   /*S3DPoint pl[3];
	   pl[0] = *p1;
	   pl[1] = *p2;
	   pl[2] = *p3;*/
	  // S3DPoint pl[3];
      //pl[0].x = p1->x;
	   //pl[0].y = p1->y;
	   //pl[0].z = p1->z;

	  // pl[1].x = p2->x;
	  // pl[1].y = p2->y;
	  // pl[1].z = p2->z;

	  // pl[2].x = p3->x;
	  // pl[2].y = p3->y;
	  // pl[2].z = p3->z;
      // bool r = Has2DPointX(pl, 3, &res) || Has2DPointY(pl, 3, &res) || Has2DPointZ(pl, 3, &res);
	   bool r = false;
      switch (p->cvVector.getMajorAxis()) {
	   case 0: {
		   r = Has2DPointX(points, 3, &res);
			   }; break;
		case 1: {
		   r = Has2DPointY(points, 3, &res);
			   }; break;
		case 2: {
		   r = Has2DPointZ(points, 3, &res);
			   }; break;
		default: utiassert(false);break;
	   };
	   if (r) {
    	  HitRec->distance = dis;
	      HitRec->point.x = res.x;
		   HitRec->point.y = res.y;
		   HitRec->point.z = res.z;
		   HitRec->point.nx = p->cvVector.x;
		   HitRec->point.ny = p->cvVector.y;
		   HitRec->point.nz= p->cvVector.z;
		   HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
		   return(true);
	  };
	  return(false);
   };
   return(false);
};

bool CPlane::IntersectTriangleWithV(CVector3f** points, CPlanef *p, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   S3DPoint res;
   utifloat dis;
   //p->GetNormal(p1, p2, p3);


   if (!IntersectLinePlaneV(p->d, &p->cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
       /*S3DPoint pl[3];
       pl[0] = *p1;
       pl[1] = *p2;
       pl[2] = *p3;*/
      // S3DPoint pl[3];
      //pl[0].x = p1->x;
       //pl[0].y = p1->y;
       //pl[0].z = p1->z;

      // pl[1].x = p2->x;
      // pl[1].y = p2->y;
      // pl[1].z = p2->z;

      // pl[2].x = p3->x;
      // pl[2].y = p3->y;
      // pl[2].z = p3->z;
      // bool r = Has2DPointX(pl, 3, &res) || Has2DPointY(pl, 3, &res) || Has2DPointZ(pl, 3, &res);
       bool r = false;
      switch (p->cvVector.getMajorAxis()) {
       case 0: {
           r = Has2DPointX(points, 3, &res);
               }; break;
        case 1: {
           r = Has2DPointY(points, 3, &res);
               }; break;
        case 2: {
           r = Has2DPointZ(points, 3, &res);
               }; break;
        default: utiassert(false);break;
       };
       if (r) {
          HitRec->distance = dis;
          HitRec->point.x = res.x;
           HitRec->point.y = res.y;
           HitRec->point.z = res.z;
           HitRec->point.nx = p->cvVector.x;
           HitRec->point.ny = p->cvVector.y;
           HitRec->point.nz= p->cvVector.z;
           HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
           return(true);
      };
      return(false);
   };
   return(false);
};

bool CPlane::IntersectTriangle(S3DUNLIT* p1, S3DUNLIT* p2, S3DUNLIT* p3, CUtiOB* obj, CUtiOB* parent, SHitResult* HitRec)
{
   CPlane p;
   p.GetNormal(p1, p2, p3);
   S3DPoint res;
   utifloat dis;
   if (!IntersectLinePlaneV(p.d, &p.cvVector, &HitRec->Vector, &res, &dis)) return(false);
   if ((dis > 0) && (dis < HitRec->distance)) {
	   S3DPoint pl[3];
	   pl[0].x = p1->x;
	   pl[0].y = p1->y;
	   pl[0].z = p1->z;

	   pl[1].x = p2->x;
	   pl[1].y = p2->y;
	   pl[1].z = p2->z;

	   pl[2].x = p3->x;
	   pl[2].y = p3->y;
	   pl[2].z = p3->z;


	   //pl[1] = *p2;
	  // pl[2] = *p3;
	  bool r = false;
	   switch (p.cvVector.getMajorAxis()) {
	   case 0: {
		   r = Has2DPointX(pl, 3, &res);
			   }; break;
		case 1: {
		   r = Has2DPointY(pl, 3, &res);
			   }; break;
		case 2: {
		   r = Has2DPointZ(pl, 3, &res);
			   }; break;
		default: utiassert(false);break;
	   };
	   if (r) {
    	   HitRec->distance = dis;
	      HitRec->point.x = res.x;
		   HitRec->point.y = res.y;
		   HitRec->point.z = res.z;
		   HitRec->point.nx = p.cvVector.x;
		   HitRec->point.ny = p.cvVector.y;
		   HitRec->point.nz= p.cvVector.z;
		   HitRec->object = obj;
           HitRec->materialid = NULL;
         HitRec->Parent = parent;
		   return(true);
	  };
	  return(false);
   };
   return(false);
};

/*CPlane::CPlane()
{
	d = 0;
}

CPlane::~CPlane()
{}*/

bool CPlane::IntersectLine(CVector* Start, CVector* Dir, utifloat &dis)
{
   utifloat N = cvVector.x*Dir->x+cvVector.y*Dir->y+cvVector.z*Dir->z;
   if (N==0) return false;
   dis = (d-cvVector.x*Start->x-cvVector.y*Start->y-cvVector.z*Start->z)/N;
   return true;
}








/*int CPlane::getPlanePos(C3DPoint & p)
{
   float distance = PlaneDistance(p);
   if (distance < -EPSILON) return (CP_BACK);
   if (distance > EPSILON) return (CP_FRONT);
   return (CP_ZERO);
}*/

int CPlane::getPlanePos(const S3DPoint& p)
{
   utifloat distance = PlaneDistance(p);
   if (distance < -EPSILON) return (CP_BACK);
   if (distance > EPSILON) return (CP_FRONT);
   return (CP_ZERO);
}


int CPlane::getPlanePos(const CVector &p)
{
   utifloat distance = PlaneDistance(p);
   if (distance < -EPSILON) return (CP_BACK);
   if (distance > EPSILON) return (CP_FRONT);
   return (CP_ZERO);
}


int CPlane::getPlanePos(const CVector3f &p)
{
   utifloat distance = PlaneDistance(p);
   if (distance < -EPSILON) return (CP_BACK);
   if (distance > EPSILON) return (CP_FRONT);
   return (CP_ZERO);
}


int CPlane::getBoundingPos(S3DPoint *p, int count)
{
	if (count <= 0) return(0);
   bool res = PlaneDistance(p[0])> 0;
   for (int i = 1; i < count; i++) {
	   bool b1 = PlaneDistance(p[i]) > 0;
	   if (b1 != res) return(0);
   };
   if (res) return(1); else return(-1);
};

void clippoint3d(S3DPointf *in, S3DPointf*innext, S3DPointf *out,float scale)
{
    out->x = in->x + (innext->x - in->x) * scale;
    out->y = in->y + (innext->y - in->y) * scale;
    out->z = in->z + (innext->z - in->z) * scale;
    out->nx = in->nx + (innext->nx - in->nx) * scale;
    out->ny = in->ny + (innext->ny - in->ny) * scale;
    out->nz = in->nz + (innext->nz - in->nz) * scale;
    out->u = in->u + (innext->u - in->u) * scale;
    out->v = in->v + (innext->v - in->v) * scale;
    out->r = in->r + (innext->r - in->r) * scale;
}

void clippoint(CVector3f *in, CVector3f *innext, CVector3f *out,float scale)
{
    out->x = in->x + (innext->x - in->x) * scale;
    out->y = in->y + (innext->y - in->y) * scale;
    out->z = in->z + (innext->z - in->z) * scale;
}

void clippoint3d(S3DPoint *in, S3DPoint *innext, S3DPoint *out,double scale)
{
    out->x = in->x + (innext->x - in->x) * scale;
    out->y = in->y + (innext->y - in->y) * scale;
	out->z = in->z + (innext->z - in->z) * scale;
	out->nx = in->nx + (innext->nx - in->nx) * scale;
	out->ny = in->ny + (innext->ny - in->ny) * scale;
    out->nz = in->nz + (innext->nz - in->nz) * scale;
    out->u = in->u + (innext->u - in->u) * scale;
	out->v = in->v + (innext->v - in->v) * scale;
	out->r = in->r + (innext->r - in->r) * scale;
}

void clippoint(CVector *in, CVector *innext, CVector *out,double scale)
{
    out->x = in->x + (innext->x - in->x) * scale;
    out->y = in->y + (innext->y - in->y) * scale;
	out->z = in->z + (innext->z - in->z) * scale;
}

/* int other_side(float a, float b)
{
	if ((a > 0 && b <= 0) || (a < 0 && b >= 0))
		return 1;
	else
		return 0;
}*/

#define other_side(a, b) ((a > 0 && b <= 0) || (a < 0 && b >= 0))

void CPlane::cliptoplane(S3DPoint *in, int countin, S3DPoint *out, int &countout, int maxcount)
{
	int i;
    utifloat scale;
    utifloat indot[60];
    countout = 0;

	for (i = 0; i < countin; i++)
	{
		//indot[i] = cvVector.dotProduct(in[i]);
		indot[i] = cvVector.x*in[i].x+cvVector.y*in[i].y+cvVector.z*in[i].z;
	}


	for (i = 0; i < countin-1; i++)
	{
		if (indot[i] >= d)
			out[countout++] = in[i];
        if (countout >= maxcount) return;
		if (other_side(indot[i] - d, indot[i+1] - d))
		{
			scale = (indot[i]-d) / ((indot[i]-d)-(indot[i+1]-d));
			clippoint3d(&in[i], &in[i+1], &out[countout], scale);
			countout++;
            if (countout >= maxcount) return;
		}
	}

	if (indot[countin-1] >= d)
		out[countout++] = in[countin-1];

    if (countout >= maxcount) return;
	if (other_side(indot[0] - d, indot[countin-1] - d))
	{
		scale = (indot[0]-d) / ((indot[0]-d)-(indot[countin-1]-d));
		clippoint3d(&in[0], &in[countin-1], &out[countout], scale);
		countout++;
        if (countout >= maxcount) return;
	}

}

void CPlane::cliptoplane(CVector *in, int countin, CVector *out, int &countout, int maxcount)
{
	int i;
    utifloat scale;
    utifloat indot[60];
	utiassert(countin < 60);
    countout = 0;

	//for (i = 0; i < countin; i++)
	//{
	//	indot[i] = cvVector.dotProduct(in[i]);
	//}
    for (i = 0; i < countin; i++)
	{
		//indot[i] = cvVector.dotProduct(in[i]);
		indot[i] = cvVector.x*in[i].x+cvVector.y*in[i].y+cvVector.z*in[i].z;
	}

	for (i = 0; i < countin-1; i++)
	{
		if (indot[i] >= d)
			out[countout++] = in[i];

        if (countout >= maxcount) return;
		if (other_side(indot[i] - d, indot[i+1] - d))
		{
			scale = (indot[i]-d) / ((indot[i]-d)-(indot[i+1]-d));
			clippoint(&in[i], &in[i+1], &out[countout], scale);
			countout++;

            if (countout >= maxcount) return;
		}
	}

	if (indot[countin-1] >= d)
		out[countout++] = in[countin-1];

    if (countout >= maxcount) return;

	if (other_side(indot[0] - d, indot[countin-1] - d))
	{
		scale = (indot[0]-d) / ((indot[0]-d)-(indot[countin-1]-d));
		clippoint(&in[0], &in[countin-1], &out[countout], scale);
		countout++;

        if (countout >= maxcount) return;
	}

}


void CPlane::cliptoplane(S3DPointf *in, int countin, S3DPointf *out, int &countout, int maxcount)
{
    int i;
    float scale;
    float indot[60];
    countout = 0;

    for (i = 0; i < countin; i++)
    {
        //indot[i] = cvVector.dotProduct(in[i]);
        indot[i] = cvVector.x*in[i].x+cvVector.y*in[i].y+cvVector.z*in[i].z;
    }


    for (i = 0; i < countin-1; i++)
    {
        if (indot[i] >= d)
            out[countout++] = in[i];

        if (countout >= maxcount) return;
        if (other_side(indot[i] - d, indot[i+1] - d))
        {
            scale = (indot[i]-d) / ((indot[i]-d)-(indot[i+1]-d));
            clippoint3d(&in[i], &in[i+1], &out[countout], scale);
            countout++;

            if (countout >= maxcount) return;
        }
    }

    if (indot[countin-1] >= d)
        out[countout++] = in[countin-1];

    if (countout >= maxcount) return;

    if (other_side(indot[0] - d, indot[countin-1] - d))
    {
        scale = (indot[0]-d) / ((indot[0]-d)-(indot[countin-1]-d));
        clippoint3d(&in[0], &in[countin-1], &out[countout], scale);
        countout++;

        if (countout >= maxcount) return;
    }

}



void CPlanef::cliptoplane(S3DPointf *in, int countin, S3DPointf *out, int &countout, int maxcount)
{
    int i;
    float scale;
    float indot[60];
    countout = 0;

    for (i = 0; i < countin; i++)
    {
        //indot[i] = cvVector.dotProduct(in[i]);
        indot[i] = cvVector.x*in[i].x+cvVector.y*in[i].y+cvVector.z*in[i].z;
    }


    for (i = 0; i < countin-1; i++)
    {
        if (indot[i] >= d)
            out[countout++] = in[i];

        if (countout >= maxcount) return;
        if (other_side(indot[i] - d, indot[i+1] - d))
        {
            scale = (indot[i]-d) / ((indot[i]-d)-(indot[i+1]-d));
            clippoint3d(&in[i], &in[i+1], &out[countout], scale);
            countout++;

            if (countout >= maxcount) return;
        }
    }

    if (indot[countin-1] >= d)
        out[countout++] = in[countin-1];

    if (countout >= maxcount) return;

    if (other_side(indot[0] - d, indot[countin-1] - d))
    {
        scale = (indot[0]-d) / ((indot[0]-d)-(indot[countin-1]-d));
        clippoint3d(&in[0], &in[countin-1], &out[countout], scale);
        countout++;

        if (countout >= maxcount) return;
    }

}

void CPlane::cliptoplane(CVector3f *in, int countin, CVector3f*out, int &countout, int maxcount)
{
    int i;
    float scale;
    float indot[60];
    utiassert(countin < 60);
    countout = 0;

    //for (i = 0; i < countin; i++)
    //{
    //	indot[i] = cvVector.dotProduct(in[i]);
    //}
    for (i = 0; i < countin; i++)
    {
        //indot[i] = cvVector.dotProduct(in[i]);
        indot[i] = cvVector.x*in[i].x+cvVector.y*in[i].y+cvVector.z*in[i].z;
    }

    for (i = 0; i < countin-1; i++)
    {
        if (indot[i] >= d)
            out[countout++] = in[i];

        if (countout >= maxcount) return;
        if (other_side(indot[i] - d, indot[i+1] - d))
        {
            scale = (indot[i]-d) / ((indot[i]-d)-(indot[i+1]-d));
            clippoint(&in[i], &in[i+1], &out[countout], scale);
            countout++;

            if (countout >= maxcount) return;
        }
    }

    if (indot[countin-1] >= d)
        out[countout++] = in[countin-1];

    if (countout >= maxcount) return;

    if (other_side(indot[0] - d, indot[countin-1] - d))
    {
        scale = (indot[0]-d) / ((indot[0]-d)-(indot[countin-1]-d));
        clippoint(&in[0], &in[countin-1], &out[countout], scale);
        countout++;

        if (countout >= maxcount) return;
    }

}


void CPlane::GetNormal(S3DPoint *p)
{
   utifloat ax = p[0].x-p[1].x;
   utifloat ay = p[0].y-p[1].y;
   utifloat az = p[0].z-p[1].z;
   utifloat bx = p[2].x-p[1].x;
   utifloat by = p[2].y-p[1].y;
   utifloat bz = p[2].z-p[1].z;
   cvVector.x = ay*bz-az*by;
   cvVector.y = az*bx-ax*bz;
   cvVector.z = ax*by-ay*bx;
   cvVector.normalize();
   d = 0;
   d = PlaneDistance(p[0]);
}

void CPlane::GetNormal(S3DPoint *p1, S3DPoint *p2, S3DPoint *p3)
{
   utifloat ax = p1->x-p2->x;
   utifloat ay = p1->y-p2->y;
   utifloat az = p1->z-p2->z;
   utifloat bx = p3->x-p2->x;
   utifloat by = p3->y-p2->y;
   utifloat bz = p3->z-p2->z;
   cvVector.x = ay*bz-az*by;
   cvVector.y = az*bx-ax*bz;
   cvVector.z = ax*by-ay*bx;
   cvVector.normalize();
   d = 0;
   d = PlaneDistance(*p1);
}

void CPlanef::GetNormal(const CVector3f *p1, const CVector3f *p2,const CVector3f *p3)
{
    utifloat ax = p1->x-p2->x;
    utifloat ay = p1->y-p2->y;
    utifloat az = p1->z-p2->z;
    utifloat bx = p3->x-p2->x;
    utifloat by = p3->y-p2->y;
    utifloat bz = p3->z-p2->z;
    cvVector.x = ay*bz-az*by;
    cvVector.y = az*bx-ax*bz;
    cvVector.z = ax*by-ay*bx;
    cvVector.normalize();
    d = 0;
    d = PlaneDistance(*p1);
}


void CPlane::GetNormal(const CVector3f *p1, const CVector3f *p2,const CVector3f *p3)
{
    utifloat ax = p1->x-p2->x;
    utifloat ay = p1->y-p2->y;
    utifloat az = p1->z-p2->z;
    utifloat bx = p3->x-p2->x;
    utifloat by = p3->y-p2->y;
    utifloat bz = p3->z-p2->z;
    cvVector.x = ay*bz-az*by;
    cvVector.y = az*bx-ax*bz;
    cvVector.z = ax*by-ay*bx;
    cvVector.normalize();
    d = 0;
    d = PlaneDistance(*p1);
}

void CPlane::GetNormal(const CVector *p1, const CVector *p2, const CVector *p3)
{
   utifloat ax = p1->x-p2->x;
   utifloat ay = p1->y-p2->y;
   utifloat az = p1->z-p2->z;
   utifloat bx = p3->x-p2->x;
   utifloat by = p3->y-p2->y;
   utifloat bz = p3->z-p2->z;
   cvVector.x = ay*bz-az*by;
   cvVector.y = az*bx-ax*bz;
   cvVector.z = ax*by-ay*bx;
   cvVector.normalize();
   d = 0;
   d = PlaneDistance(*p1);
}


CVector3 CPlane::pointOnPlane(const CVector3& v)
{
     CVector3 v1;
     float p = PlaneDistance(v);
     v1 = v-cvVector*p;
     return v1;
}

void CPlane::getPlane(const CVector &p1, const CVector &p2,const CVector &p3)
{
  utifloat ax = p1.x-p2.x;
   utifloat ay = p1.y-p2.y;
   utifloat az = p1.z-p2.z;
   utifloat bx = p3.x-p2.x;
   utifloat by = p3.y-p2.y;
   utifloat bz = p3.z-p2.z;
   cvVector.x = ay*bz-az*by;
   cvVector.y = az*bx-ax*bz;
   cvVector.z = ax*by-ay*bx;
   cvVector.normalize();
   d = 0;
   d = PlaneDistance(p1);
}

void CPlanef::getPlane(const CVector3f &p1, const CVector3f &p2,const CVector3f &p3)
{
  utifloat ax = p1.x-p2.x;
   utifloat ay = p1.y-p2.y;
   utifloat az = p1.z-p2.z;
   utifloat bx = p3.x-p2.x;
   utifloat by = p3.y-p2.y;
   utifloat bz = p3.z-p2.z;
   cvVector.x = ay*bz-az*by;
   cvVector.y = az*bx-ax*bz;
   cvVector.z = ax*by-ay*bx;
   cvVector.normalize();
   d = 0;
   d = PlaneDistance(p1);
}

CVector CPlane::getNormal(const CVector &p1, const CVector &p2,const CVector &p3)
{
  utifloat ax = p1.x-p2.x;
   utifloat ay = p1.y-p2.y;
   utifloat az = p1.z-p2.z;
   utifloat bx = p3.x-p2.x;
   utifloat by = p3.y-p2.y;
   utifloat bz = p3.z-p2.z;
   CVector cvVector;
   cvVector.x = ay*bz-az*by;
   cvVector.y = az*bx-ax*bz;
   cvVector.z = ax*by-ay*bx;
   cvVector.normalize();
   return cvVector;
}

CVector3f CPlane::getNormal(const CVector3f &p1, const CVector3f &p2,const CVector3f &p3)
{
  utifloat ax = p1.x-p2.x;
   utifloat ay = p1.y-p2.y;
   utifloat az = p1.z-p2.z;
   utifloat bx = p3.x-p2.x;
   utifloat by = p3.y-p2.y;
   utifloat bz = p3.z-p2.z;
   CVector3f cvVector;
   cvVector.x = ay*bz-az*by;
   cvVector.y = az*bx-ax*bz;
   cvVector.z = ax*by-ay*bx;
   cvVector.normalize();
   return cvVector;
}

CPlane CPlane::createPlane(const CVector3& position, const CVector3& normal)
{
   CPlane res;
   res.cvVector = normal;
   res.d = 0;
   res.d = res.cvVector.dotProduct( position);

   return res;
}

void CPlane::GetNormal(S3DUNLIT *p1, S3DUNLIT *p2, S3DUNLIT *p3)
{
   utifloat ax = p1->x-p2->x;
   utifloat ay = p1->y-p2->y;
   utifloat az = p1->z-p2->z;
   utifloat bx = p3->x-p2->x;
   utifloat by = p3->y-p2->y;
   utifloat bz = p3->z-p2->z;
   cvVector.x = ay*bz-az*by;
   cvVector.y = az*bx-ax*bz;
   cvVector.z = ax*by-ay*bx;
   cvVector.normalize();
   d = p1->x*cvVector.x+p1->y*cvVector.y+p1->z*cvVector.z;
  // d = PlaneDistance(*p1);
}

const S3DPoint CPlane::GetNormal(S3DPoint p1, S3DPoint p2, S3DPoint p3, S3DPoint *res)
{
   S3DPoint result;
   if (res) {
	   result = *res;
   } else {
	   zeromem(&result,sizeof(S3DPoint));
   }
   utifloat ax = p1.x-p2.x;
   utifloat ay = p1.y-p2.y;
   utifloat az = p1.z-p2.z;
   utifloat bx = p3.x-p2.x;
   utifloat by = p3.y-p2.y;
   utifloat bz = p3.z-p2.z;
   CVector v;
   v.x = ay*bz-az*by;
   v.y = az*bx-ax*bz;
   v.z = ax*by-ay*bx;
   v.normalize();
   result.nx = v.x;
   result.ny = v.y;
result.nz = v.z;
return result;
}

CMatrix CMatrix::getInverse() const
{
    CMatrix r;
    inverse(*this, r);
    return r;
}

void CMatrix::inverse(const CMatrix &in, CMatrix &out)
{
   /*float scale;
   scale = in.m[0][0] * in.m[0][0] +
            in.m[0][1] * in.m[0][1] +
            in.m[0][2] * in.m[0][2];
   scale = 1/scale;
   out.m[0][0] = scale * in.m[0][0];
   out.m[1][0] = scale * in.m[0][1];
   out.m[2][0] = scale * in.m[0][2];
   out.m[0][1] = scale * in.m[1][0];
   out.m[1][1] = scale * in.m[1][1];
   out.m[2][1] = scale * in.m[1][2];
   out.m[0][2] = scale * in.m[2][0];
   out.m[1][2] = scale * in.m[2][1];
   out.m[2][2] = scale * in.m[2][2];

   out.m[0][3]   = -(out.m[0][0] * in.m[0][3] +
                        out.m[0][1] * in.m[1][3] +
                        out.m[0][2] * in.m[2][3]);
   out.m[1][3]   = -(out.m[1][0] * in.m[0][3] +
                        out.m[1][1] * in.m[1][3] +
                        out.m[1][2] * in.m[2][3]);
   out.m[2][3]   = -(out.m[2][0] * in.m[0][3] +
                         out.m[2][1] * in.m[1][3] +
                         out.m[2][2] * in.m[2][3]);
   out.m[3][0] = 0;
   out.m[3][1] = 0;
   out.m[3][2] = 0;
   out.m[3][3] = 1;
   */


   double  d00, d01, d02, d03,
  d10, d11, d12, d13,
  d20, d21, d22, d23,
  d30, d31, d32, d33,
  m00, m01, m02, m03,
  m10, m11, m12, m13,
  m20, m21, m22, m23,
  m30, m31, m32, m33, D;

  m00 = in.m[0][0];  m01 = in.m[0][1];  m02 = in.m[0][2];  m03 = in.m[0][3];
  m10 = in.m[1][0];  m11 = in.m[1][1];  m12 = in.m[1][2];  m13 = in.m[1][3];
  m20 = in.m[2][0];  m21 = in.m[2][1];  m22 = in.m[2][2];  m23 = in.m[2][3];
  m30 = in.m[3][0];  m31 = in.m[3][1];  m32 = in.m[3][2];  m33 = in.m[3][3];

  d00 = m11*m22*m33 + m12*m23*m31 + m13*m21*m32 - m31*m22*m13 - m32*m23*m11 - m33*m21*m12;
  d01 = m10*m22*m33 + m12*m23*m30 + m13*m20*m32 - m30*m22*m13 - m32*m23*m10 - m33*m20*m12;
  d02 = m10*m21*m33 + m11*m23*m30 + m13*m20*m31 - m30*m21*m13 - m31*m23*m10 - m33*m20*m11;
  d03 = m10*m21*m32 + m11*m22*m30 + m12*m20*m31 - m30*m21*m12 - m31*m22*m10 - m32*m20*m11;

  d10 = m01*m22*m33 + m02*m23*m31 + m03*m21*m32 - m31*m22*m03 - m32*m23*m01 - m33*m21*m02;
  d11 = m00*m22*m33 + m02*m23*m30 + m03*m20*m32 - m30*m22*m03 - m32*m23*m00 - m33*m20*m02;
  d12 = m00*m21*m33 + m01*m23*m30 + m03*m20*m31 - m30*m21*m03 - m31*m23*m00 - m33*m20*m01;
  d13 = m00*m21*m32 + m01*m22*m30 + m02*m20*m31 - m30*m21*m02 - m31*m22*m00 - m32*m20*m01;

  d20 = m01*m12*m33 + m02*m13*m31 + m03*m11*m32 - m31*m12*m03 - m32*m13*m01 - m33*m11*m02;
  d21 = m00*m12*m33 + m02*m13*m30 + m03*m10*m32 - m30*m12*m03 - m32*m13*m00 - m33*m10*m02;
  d22 = m00*m11*m33 + m01*m13*m30 + m03*m10*m31 - m30*m11*m03 - m31*m13*m00 - m33*m10*m01;
  d23 = m00*m11*m32 + m01*m12*m30 + m02*m10*m31 - m30*m11*m02 - m31*m12*m00 - m32*m10*m01;

  d30 = m01*m12*m23 + m02*m13*m21 + m03*m11*m22 - m21*m12*m03 - m22*m13*m01 - m23*m11*m02;
  d31 = m00*m12*m23 + m02*m13*m20 + m03*m10*m22 - m20*m12*m03 - m22*m13*m00 - m23*m10*m02;
  d32 = m00*m11*m23 + m01*m13*m20 + m03*m10*m21 - m20*m11*m03 - m21*m13*m00 - m23*m10*m01;
  d33 = m00*m11*m22 + m01*m12*m20 + m02*m10*m21 - m20*m11*m02 - m21*m12*m00 - m22*m10*m01;

  D = m00*d00 - m01*d01 + m02*d02 - m03*d03;



  out.m[0][0] =  d00/D; out.m[0][1] = -d10/D;  out.m[0][2] =  d20/D; out.m[0][3] = -d30/D;
  out.m[1][0] = -d01/D; out.m[1][1] =  d11/D;  out.m[1][2] = -d21/D; out.m[1][3] =  d31/D;
  out.m[2][0] =  d02/D; out.m[2][1] = -d12/D;  out.m[2][2] =  d22/D; out.m[2][3] = -d32/D;
  out.m[3][0] = -d03/D; out.m[3][1] =  d13/D;  out.m[3][2] = -d23/D; out.m[3][3] =  d33/D;
}
/*
inline void M4x4_SSE(float *A, float *B, float *C) {
    __m128 row1 = _mm_load_ps(&B[0]);
    __m128 row2 = _mm_load_ps(&B[4]);
    __m128 row3 = _mm_load_ps(&B[8]);
    __m128 row4 = _mm_load_ps(&B[12]);
    for(int i=0; i<4; i++) {
        __m128 brod1 = _mm_set1_ps(A[4*i + 0]);
        __m128 brod2 = _mm_set1_ps(A[4*i + 1]);
        __m128 brod3 = _mm_set1_ps(A[4*i + 2]);
        __m128 brod4 = _mm_set1_ps(A[4*i + 3]);
        __m128 row = _mm_add_ps(
                    _mm_add_ps(
                        _mm_mul_ps(brod1, row1),
                        _mm_mul_ps(brod2, row2)),
                    _mm_add_ps(
                        _mm_mul_ps(brod3, row3),
                        _mm_mul_ps(brod4, row4)));
        _mm_store_ps(&C[4*i], row);
    }
}

void CMatrixf::multiply(CMatrixf & in, CMatrixf& out)
{
M4x4_SSE(&this->m[0][0],&in->m[0][0], &out->m[0][0]);
}*/

void CMatrixf::inverse(const CMatrixf &in, CMatrixf &out)
{
   /*float scale;
   scale = in.m[0][0] * in.m[0][0] +
            in.m[0][1] * in.m[0][1] +
            in.m[0][2] * in.m[0][2];
   scale = 1/scale;
   out.m[0][0] = scale * in.m[0][0];
   out.m[1][0] = scale * in.m[0][1];
   out.m[2][0] = scale * in.m[0][2];
   out.m[0][1] = scale * in.m[1][0];
   out.m[1][1] = scale * in.m[1][1];
   out.m[2][1] = scale * in.m[1][2];
   out.m[0][2] = scale * in.m[2][0];
   out.m[1][2] = scale * in.m[2][1];
   out.m[2][2] = scale * in.m[2][2];

   out.m[0][3]   = -(out.m[0][0] * in.m[0][3] +
                        out.m[0][1] * in.m[1][3] +
                        out.m[0][2] * in.m[2][3]);
   out.m[1][3]   = -(out.m[1][0] * in.m[0][3] +
                        out.m[1][1] * in.m[1][3] +
                        out.m[1][2] * in.m[2][3]);
   out.m[2][3]   = -(out.m[2][0] * in.m[0][3] +
                         out.m[2][1] * in.m[1][3] +
                         out.m[2][2] * in.m[2][3]);
   out.m[3][0] = 0;
   out.m[3][1] = 0;
   out.m[3][2] = 0;
   out.m[3][3] = 1;
   */


   double  d00, d01, d02, d03,
  d10, d11, d12, d13,
  d20, d21, d22, d23,
  d30, d31, d32, d33,
  m00, m01, m02, m03,
  m10, m11, m12, m13,
  m20, m21, m22, m23,
  m30, m31, m32, m33, D;

  m00 = in.m[0][0];  m01 = in.m[0][1];  m02 = in.m[0][2];  m03 = in.m[0][3];
  m10 = in.m[1][0];  m11 = in.m[1][1];  m12 = in.m[1][2];  m13 = in.m[1][3];
  m20 = in.m[2][0];  m21 = in.m[2][1];  m22 = in.m[2][2];  m23 = in.m[2][3];
  m30 = in.m[3][0];  m31 = in.m[3][1];  m32 = in.m[3][2];  m33 = in.m[3][3];

  d00 = m11*m22*m33 + m12*m23*m31 + m13*m21*m32 - m31*m22*m13 - m32*m23*m11 - m33*m21*m12;
  d01 = m10*m22*m33 + m12*m23*m30 + m13*m20*m32 - m30*m22*m13 - m32*m23*m10 - m33*m20*m12;
  d02 = m10*m21*m33 + m11*m23*m30 + m13*m20*m31 - m30*m21*m13 - m31*m23*m10 - m33*m20*m11;
  d03 = m10*m21*m32 + m11*m22*m30 + m12*m20*m31 - m30*m21*m12 - m31*m22*m10 - m32*m20*m11;

  d10 = m01*m22*m33 + m02*m23*m31 + m03*m21*m32 - m31*m22*m03 - m32*m23*m01 - m33*m21*m02;
  d11 = m00*m22*m33 + m02*m23*m30 + m03*m20*m32 - m30*m22*m03 - m32*m23*m00 - m33*m20*m02;
  d12 = m00*m21*m33 + m01*m23*m30 + m03*m20*m31 - m30*m21*m03 - m31*m23*m00 - m33*m20*m01;
  d13 = m00*m21*m32 + m01*m22*m30 + m02*m20*m31 - m30*m21*m02 - m31*m22*m00 - m32*m20*m01;

  d20 = m01*m12*m33 + m02*m13*m31 + m03*m11*m32 - m31*m12*m03 - m32*m13*m01 - m33*m11*m02;
  d21 = m00*m12*m33 + m02*m13*m30 + m03*m10*m32 - m30*m12*m03 - m32*m13*m00 - m33*m10*m02;
  d22 = m00*m11*m33 + m01*m13*m30 + m03*m10*m31 - m30*m11*m03 - m31*m13*m00 - m33*m10*m01;
  d23 = m00*m11*m32 + m01*m12*m30 + m02*m10*m31 - m30*m11*m02 - m31*m12*m00 - m32*m10*m01;

  d30 = m01*m12*m23 + m02*m13*m21 + m03*m11*m22 - m21*m12*m03 - m22*m13*m01 - m23*m11*m02;
  d31 = m00*m12*m23 + m02*m13*m20 + m03*m10*m22 - m20*m12*m03 - m22*m13*m00 - m23*m10*m02;
  d32 = m00*m11*m23 + m01*m13*m20 + m03*m10*m21 - m20*m11*m03 - m21*m13*m00 - m23*m10*m01;
  d33 = m00*m11*m22 + m01*m12*m20 + m02*m10*m21 - m20*m11*m02 - m21*m12*m00 - m22*m10*m01;

  D = m00*d00 - m01*d01 + m02*d02 - m03*d03;



  out.m[0][0] =  d00/D; out.m[0][1] = -d10/D;  out.m[0][2] =  d20/D; out.m[0][3] = -d30/D;
  out.m[1][0] = -d01/D; out.m[1][1] =  d11/D;  out.m[1][2] = -d21/D; out.m[1][3] =  d31/D;
  out.m[2][0] =  d02/D; out.m[2][1] = -d12/D;  out.m[2][2] =  d22/D; out.m[2][3] = -d32/D;
  out.m[3][0] = -d03/D; out.m[3][1] =  d13/D;  out.m[3][2] = -d23/D; out.m[3][3] =  d33/D;
}

void GetBoundsClipPlanes(CPlane *planes, CBox* rect)
{

   planes[0].cvVector.setVector(0, 1, 0);
   planes[0].d = 0;
   planes[0].d = planes[0].PlaneDistance(rect->m_min);
   planes[1].cvVector.setVector(1, 0, 0);
   planes[1].d = 0;
   planes[1].d = planes[1].PlaneDistance(rect->m_min);
   planes[2].cvVector.setVector(0, 0, 1);
   planes[2].d = 0;
   planes[2].d = planes[2].PlaneDistance(rect->m_min);
   planes[3].cvVector.setVector(0, -1, 0);
   planes[3].d = 0;
   planes[3].d = planes[3].PlaneDistance(rect->m_max);
   planes[4].cvVector.setVector(-1, 0, 0);
   planes[4].d = 0;
   planes[4].d = planes[4].PlaneDistance(rect->m_max);
   planes[5].cvVector.setVector(0, 0, -1);
   planes[5].d = 0;
   planes[5].d = planes[5].PlaneDistance(rect->m_max);

}

void GetRectCenter(const CBox* rect,CVector *p)
{
   p->x = rect->m_min.x + (rect->m_max.x - rect->m_min.x) / 2;
   p->y = rect->m_min.y + (rect->m_max.y - rect->m_min.y) / 2;
   p->z = rect->m_min.z + (rect->m_max.z - rect->m_min.z) / 2;
}

bool PointInRect(const CBox* rect, const S3DPoint *p)
{
   if ((rect->m_min.x <= p->x) && (rect->m_max.x >= p->x) &&
	   (rect->m_min.y <= p->y) && (rect->m_max.y >= p->y) &&
	   (rect->m_min.z <= p->z) && (rect->m_max.z >= p->z)) {

	   return(true);
   };
   return(false);
}

bool PointInRect(const CBox* rect, const CVector *p)
{
   if ((rect->m_min.x <= p->x) && (rect->m_max.x >= p->x) &&
	   (rect->m_min.y <= p->y) && (rect->m_max.y >= p->y) &&
	   (rect->m_min.z <= p->z) && (rect->m_max.z >= p->z)) {

	   return(true);
   };
   return(false);
}

void IntersectQuaderRayMinMax(const S3DPoint& vector, const CBox& b, utifloat &t_min, utifloat &t_max)
{
   //Warnung:
   utifloat t1, t2;
   utifloat minx, maxx, miny, maxy, minz, maxz;

      t1 = (b.m_min.x - vector.x) / vector.nx;
      t2 = (b.m_max.x - vector.x) / vector.nx;
      if (t1 < t2)
      {
         minx = t1;
         maxx = t2;
      } else {
         minx = t2;
         maxx = t1;
      };

      t1 = (b.m_min.y - vector.y) / vector.ny;
      t2 = (b.m_max.y - vector.y) / vector.ny;
      if (t1 < t2)
      {
         miny = t1;
         maxy = t2;
      } else {
         miny = t2;
         maxy = t1;
      };

      t1 = (b.m_min.z - vector.z) / vector.nz;
      t2 = (b.m_max.z - vector.z) / vector.nz;
      if (t1 < t2)
      {
         minz = t1;
         maxz = t2;
      } else {
         minz = t2;
         maxz = t1;
      };

    t_min = umax(minx, umax(miny, minz));
    t_max = umin(maxx, umin(maxy, maxz));
};

#ifndef MAXFLOAT
#define MAXFLOAT 1E+10
#endif

bool IntersectRay(const S3DPoint* vector, const CBox* b, utifloat &dis)
{
	dis = 0;
	bool in = false;
	if (PointInRect(b, vector)) {
		in = true;
		//return(true);
	};
	dis = MAXFLOAT;
    utifloat t1, t2;
    utifloat minx, maxx, miny, maxy, minz, maxz;
	if (fabs(vector->nx) < 0.00001) {
		if (b->m_min.x < vector->x && b->m_max.x > vector->x) {
			minx = -MAXFLOAT;
			maxx = MAXFLOAT;
		} else {
			return(false);
			minx = -MAXFLOAT;
			maxx = MAXFLOAT;
		};
	} else {
		t1 = (b->m_min.x - vector->x) / vector->nx;
		t2 = (b->m_max.x - vector->x) / vector->nx;
		if (t1 < t2)
		{
			minx = t1;
			maxx = t2;
		} else {
			minx = t2;
			maxx = t1;
		};
		//if (maxx < 0) ;
			//return(0);
	}

	if (fabs(vector->ny) < 0.00001) {
		if (b->m_min.y < vector->y && b->m_max.y > vector->y) {
			miny = -MAXFLOAT;
			maxy = MAXFLOAT;
		} else {
			return(false);
			miny = -MAXFLOAT;
			maxy = MAXFLOAT;
		};
	} else {
		t1 = (b->m_min.y - vector->y) / vector->ny;
		t2 = (b->m_max.y - vector->y) / vector->ny;
		if (t1 < t2)
		{
			miny = t1;
			maxy = t2;
		} else {
			miny = t2;
			maxy = t1;
		};
		//if (maxy < 0)	   ;
			//return(false);
	}
	if (fabs(vector->nz) < 0.00001) {
		if (b->m_min.z < vector->z && b->m_max.z > vector->z) {
			minz = -MAXFLOAT;
			maxz = MAXFLOAT;
		} else {
			return(false); //Commented this because of unknowning reasons
			minz = -MAXFLOAT;
			maxz = MAXFLOAT;
		};
	} else {
		t1 = (b->m_min.z - vector->z) / vector->nz;
		t2 = (b->m_max.z - vector->z) / vector->nz;
		if (t1 < t2)
		{
			minz = t1;
			maxz = t2;
		} else {
			minz = t2;
			maxz = t1;
		};
		//if (maxz < 0)  ;
			//return(false);
	}

    utifloat t_min = umax(minx, umax(miny, minz));
    utifloat t_max = umin(maxx, umin(maxy, maxz));
	if (t_min <= t_max) {
		if (in) {
			dis = t_max;
		} else {
		    dis = t_min;
		}
	   return(true);
	};
	return(false);
};

bool IntersectRayB(const S3DPoint* vector, const CBox* b, utifloat &dis, bool &in)
{
	dis = 0;
    in = false;
	if (PointInRect(b, vector)) {
		in = true;
		//return(true);
	};
	dis = MAXFLOAT;
    utifloat t1, t2;
    utifloat minx=-MAXFLOAT, maxx=MAXFLOAT, miny=-MAXFLOAT, maxy=MAXFLOAT, minz=-MAXFLOAT, maxz=MAXFLOAT;
	if (fabs(vector->nx) < 0.0001) {
		if (b->m_min.x < vector->x && b->m_max.x > vector->x) {
			minx = -MAXFLOAT;
			maxx = MAXFLOAT;
		} else {
			return(0);
		};
	} else {
		t1 = (b->m_min.x - vector->x) / vector->nx;
		t2 = (b->m_max.x - vector->x) / vector->nx;
		if (t1 < t2)
		{
			minx = t1;
			maxx = t2;
		} else {
			minx = t2;
			maxx = t1;
		};
		//if (maxx < 0) ;
			//return(0);
	}

	if (fabs(vector->ny) < 0.0001) {
		if (b->m_min.y < vector->y && b->m_max.y > vector->y) {
			miny = -MAXFLOAT;
			maxy = MAXFLOAT;
		} else {
			//return(false);
		};
	} else {
		t1 = (b->m_min.y - vector->y) / vector->ny;
		t2 = (b->m_max.y - vector->y) / vector->ny;
		if (t1 < t2)
		{
			miny = t1;
			maxy = t2;
		} else {
			miny = t2;
			maxy = t1;
		};
		//if (maxy < 0)	   ;
			//return(false);
	}
	if (fabs(vector->nz) < 0.0001) {
		if (b->m_min.z < vector->z && b->m_max.z > vector->z) {
			minz = -MAXFLOAT;
			maxz = MAXFLOAT;
		} else {
			//return(false);
		};
	} else {
		t1 = (b->m_min.z - vector->z) / vector->nz;
		t2 = (b->m_max.z - vector->z) / vector->nz;
		if (t1 < t2)
		{
			minz = t1;
			maxz = t2;
		} else {
			minz = t2;
			maxz = t1;
		};
		//if (maxz < 0)  ;
			//return(false);
	}

    utifloat t_min = umax(minx, umax(miny, minz));
    utifloat t_max = umin(maxx, umin(maxy, maxz));
	if (t_min <= t_max) {
		if (in) {
			dis = t_max;
		} else {
		    dis = t_min;
		}
	   return(true);
	};
	return(false);
};

bool IntersectNormRay(const S3DPoint* vector, const CBox* b, utifloat &dis, CVector* normal)
{
	dis = 0;
    utifloat x, y, z;
	bool in = false;
	if (PointInRect(b, vector)) {
		in = true;
		//return(true);
	};
	dis = MAXFLOAT;
    utifloat t1, t2;
    utifloat minx=-MAXFLOAT, maxx=MAXFLOAT, miny=-MAXFLOAT, maxy=MAXFLOAT, minz=-MAXFLOAT, maxz=MAXFLOAT;
	if (fabs(vector->nx) < 0.0001) {
        if (b->m_min.x <= vector->x && b->m_max.x >= vector->x) {
			minx = -MAXFLOAT;
			maxx = MAXFLOAT;
		} else {
			return(0);
		};
	} else {
		t1 = (b->m_min.x - vector->x) / vector->nx;
		t2 = (b->m_max.x - vector->x) / vector->nx;
		if (t1 < t2)
		{
			minx = t1;
			maxx = t2;
			x = -1;
		} else {
			minx = t2;
			maxx = t1;
			x = 1;
		};
		//if (maxx < 0) ;
			//return(0);
	}

	if (fabs(vector->ny) < 0.0001) {
        if (b->m_min.y <= vector->y && b->m_max.y >= vector->y) {
			miny = -MAXFLOAT;
			maxy = MAXFLOAT;
		} else {
			//return(false);
		};
	} else {
		t1 = (b->m_min.y - vector->y) / vector->ny;
		t2 = (b->m_max.y - vector->y) / vector->ny;
		if (t1 < t2)
		{
			miny = t1;
			maxy = t2;
			y = -1;
		} else {
			miny = t2;
			maxy = t1;
			y = 1;
		};
		//if (maxy < 0)	   ;
			//return(false);
	}
	if (fabs(vector->nz) < 0.0001) {
        if (b->m_min.z <= vector->z && b->m_max.z >= vector->z) {
			minz = -MAXFLOAT;
			maxz = MAXFLOAT;
		} else {
			//return(false);
		};
	} else {
		t1 = (b->m_min.z - vector->z) / vector->nz;
		t2 = (b->m_max.z - vector->z) / vector->nz;
		if (t1 < t2)
		{
			minz = t1;
			maxz = t2;
			z = -1;
		} else {
			minz = t2;
			maxz = t1;
			z = 1;
		};
	//	if (maxz < 0)  ;
			//return(false);
	}

    utifloat t_min = umax(minx, umax(miny, minz));
    utifloat t_max = umin(maxx, umin(maxy, maxz));
	if (t_min <= t_max) {
		if (in) {
			dis = t_max;
			if (t_max == maxx) {
				normal->setVector(x, 0, 0);
				return(true);
			}
			if (t_max == maxy) {
				normal->setVector(0, y, 0);
				return(true);
			}
			if (t_max == maxz) {
				normal->setVector(0, 0, z);
				return(true);
			}
		} else {
		    dis = t_min;
			if (t_min == minx) {
				normal->setVector(-x, 0, 0);
				return(true);
			}
			if (t_min == miny) {
				normal->setVector(0, -y, 0);
				return(true);
			}
			if (t_min == minz) {
				normal->setVector(0, 0, -z);
				return(true);
			}

		}
	   return(true);
	};
	return(false);
};


void MoveRect(CBox* rect, S3DPoint* t)
{
   Point_Add(&rect->m_min.x, &rect->m_min.x, &t->x, 3);
   Point_Add(&rect->m_max.x, &rect->m_max.x, &t->x, 3);
};

void getBoxPoints(const CBox& rectm, CVector* points)
{
    points[0].x = rectm.m_min.x;
    points[0].y = rectm.m_min.y;
   points[0].z = rectm.m_min.z;

    points[1].x = rectm.m_max.x;
    points[1].y = rectm.m_min.y;
    points[1].z = rectm.m_min.z;

    points[2].x = rectm.m_max.x;
    points[2].y = rectm.m_max.y;
    points[2].z = rectm.m_min.z;

    points[3].x = rectm.m_min.x;
    points[3].y = rectm.m_max.y;
    points[3].z = rectm.m_min.z;

    //points[4] = rectm.p2;
   points[4].x = rectm.m_max.x;
   points[4].y = rectm.m_max.y;
   points[4].z = rectm.m_max.z;


    points[5].x = rectm.m_max.x;
    points[5].y = rectm.m_min.y;
    points[5].z = rectm.m_max.z;

    points[6].x = rectm.m_min.x;
    points[6].y = rectm.m_min.y;
    points[6].z = rectm.m_max.z;

    points[7].x = rectm.m_min.x;
    points[7].y = rectm.m_max.y;
    points[7].z = rectm.m_max.z;

}

int GetRectPoints(CBox* rectm, S3DPoint* points, int count, CVector* dir)
{
	if (!dir) {
	points[0].x = rectm->m_min.x;
	points[0].y = rectm->m_min.y;
   points[0].z = rectm->m_min.z;

	points[1].x = rectm->m_max.x;
	points[1].y = rectm->m_min.y;
	points[1].z = rectm->m_min.z;

	points[2].x = rectm->m_max.x;
	points[2].y = rectm->m_max.y;
	points[2].z = rectm->m_min.z;

	points[3].x = rectm->m_min.x;
	points[3].y = rectm->m_max.y;
	points[3].z = rectm->m_min.z;

	//points[4] = rectm->p2;
   points[4].x = rectm->m_max.x;
   points[4].y = rectm->m_max.y;
   points[4].z = rectm->m_max.z;


	points[5].x = rectm->m_max.x;
	points[5].y = rectm->m_min.y;
	points[5].z = rectm->m_max.z;

	points[6].x = rectm->m_min.x;
	points[6].y = rectm->m_min.y;
	points[6].z = rectm->m_max.z;

	points[7].x = rectm->m_min.x;
	points[7].y = rectm->m_max.y;
	points[7].z = rectm->m_max.z;
	return count;
	};
return 0;
};

int ClassifyPoints(S3DPoint* points, int count, CPlane*plane)
{
	if (count == 0) return CP_BACK;
   int side = plane->getPlanePos(points[0]);
   if (side == CP_ZERO) return CP_ZERO;
   int s2;
   for (int i = 1; i < count; i++) {
	   s2 = plane->getPlanePos(points[i]);
	   if (side != s2) {
		   return CP_ZERO;
	   }
   };
   return side;
};



void InitBounds(CBox &Bounds)
{
     Bounds.m_min.x = 10000000000.0f;
		Bounds.m_min.y = 10000000000.0f;
		Bounds.m_min.z = 10000000000.0f;
		Bounds.m_max.x = -10000000000.0f;
		Bounds.m_max.y = -10000000000.0f;
		Bounds.m_max.z = -10000000000.0f;
};

bool IntersectRect3D(const CBox* rect1,  const CBox* rect2)
{

   if (((rect1->m_max.x < rect2->m_min.x) || (rect2->m_max.x < rect1->m_min.x)) ||
	   ((rect1->m_max.y < rect2->m_min.y) || (rect2->m_max.y < rect1->m_min.y)) ||
	   ((rect1->m_max.z < rect2->m_min.z) || (rect2->m_max.z < rect1->m_min.z))) {
	  return(false);
   };
   return(true);

}

bool PtInRect3D(S3DPoint *p,  CBox* rect2)
{

   if (rect2->m_min.x <= p->x && rect2->m_min.y <= p->y && rect2->m_min.z <= p->z &&
	   rect2->m_max.x >= p->x && rect2->m_max.y >= p->y && rect2->m_max.z >= p->z)
   return(true);
   return(false);

}


float Length(float*v)
{
   return(sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]));
}

double Length(double*v)
{
   return(sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]));
}


float normalize(float*v)
{
    float r = Length(v);
    v[0] /= r;
    v[1] /= r;
    v[2] /= r;
    return r;
}


double normalize(double*v)
{
    double r = Length(v);
        v[0] /= r;
        v[1] /= r;
        v[2] /= r;
        return r;
}

void Set3DPoint(float x, float y, float z, S3DPoint*p)
{
    p->x = x;
	p->y = y;
	p->z = z;
}

void Set3DPoint(S3DPoint* p, float x, float y, float z, float nx, float ny, float nz, float tx, float ty, float tz, UINT color)
{
   p->x = x;
   p->y = y;
   p->z = z;
   p->nx = nx;
   p->ny = ny;
   p->nz = nz;
   p->u = tx;
   p->v = ty;
   p->r = tz;
   p->color = color;
   p->flags = 0;
}

/*S3DPoint Set3DPoint(float x, float y, float z, S3DPoint* p1)
{
	S3DPoint p;
	if (p1) {
		p = *p1;

	} else {
		zeromem(p, sizeof p);
	}
   p.x = x;
   p.y = y;
   p.z = z;
   return p;

}*/

bool IntersectLineCircle(const CVector2& origin, const CVector2& dir,
                         const CVector2& position, utifloat radius,
                         utifloat& nearest,  CVector2& res)
{

    CVector3 lres(0,0,0);assert(dir.length() > 0.0001);
    bool ret = IntersectLineSphere(CVector3(origin.x, 0, origin.y), CVector3(dir.x, 0, dir.y), CVector3(position.x, 0, position.y), radius, nearest, lres);
    res.x = lres.x;
    res.y = lres.z;
    return ret;
}

bool IntersectLineSphere(const CVector3& origin, const CVector3& dir,
                         const CVector& position, utifloat radius,
                         utifloat &nearest,  CVector& res)
{
    utifloat b,c,a,  discrm;
    CVector3 pMinusC = origin-position;
    b = 2.0f * dot(dir, pMinusC);
    c = dot(pMinusC, pMinusC) - radius * radius;
    a = dot(dir, dir);

    discrm = b*b-4*c*a;
    nearest = 100000.0f;

    if (discrm > 0) {
        utifloat r1 = (-b+sqrt(discrm))/(2*a);
        utifloat r2 = (-b - sqrt(discrm))/(2*a);
        //assert(r1 <= r2);
        if (r2 < r1) {
            std::swap(r1, r2);

        }
        if (r2 < 0) {
            return false;
        } else if (r1 < 0) {
            nearest = r2;
            assert(std::isfinite(nearest));

        }else {
            nearest = r1;
            assert(std::isfinite(nearest));
        }
    } else if (discrm == 0) {
        utifloat r1 = -b/(2*a);
        nearest = r1;
        assert(std::isfinite(nearest));
        if (r1 < 0.0) {
            return false;
        }
    } else {
        return false;
    }
    assert(std::isfinite(nearest));

    res = origin + dir*nearest;

	return(true);
};

bool IntersectLinePlane(S3DPoint *polypoint, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = line->nx*v->x+
	              line->ny*v->y+
				  line->nz*v->z;
   if (bottom == 0) return(false);
   utifloat top = (polypoint->x-line->x)*v->x+
	           (polypoint->y-line->y)*v->y+
			   (polypoint->z-line->z)*v->z;
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}

bool IntersectLinePlaneV(utifloat d, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = line->nx*v->x+
	              line->ny*v->y+
				  line->nz*v->z;
   if (bottom == 0) return(false);
   utifloat top = d-v->x*line->x-v->y*line->y-v->z*line->z;//(polypoint->x-line->x)*v->x+
	           //(polypoint->y-line->y)*v->y+
			   //(polypoint->z-line->z)*v->z;

   //top := V[3]-V[0]*Line.X-V[1]*Line.Y-V[2]*Line.Z;
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}

bool IntersectLinePlaneV(utifloat d, CVector3f *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = line->nx*v->x+
                  line->ny*v->y+
                  line->nz*v->z;
   if (bottom == 0) return(false);
   utifloat top = d-v->x*line->x-v->y*line->y-v->z*line->z;//(polypoint->x-line->x)*v->x+
               //(polypoint->y-line->y)*v->y+
               //(polypoint->z-line->z)*v->z;

   //top := V[3]-V[0]*Line.X-V[1]*Line.Y-V[2]*Line.Z;
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}


bool IntersectLinePlaneZ(S3DPoint *polypoint, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = line->nz;
   if (bottom == 0) return(false);
   utifloat top = polypoint->z-line->z;
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}

bool IntersectLinePlaneX(S3DPoint *polypoint, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = line->nx;
   if (bottom == 0) return(false);
   utifloat top = (polypoint->x-line->x);
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}

bool IntersectLinePlaneY(S3DPoint *polypoint, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = line->ny;
   if (bottom == 0) return(false);
   utifloat top = (polypoint->y-line->y);
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}

bool IntersectLinePlaneNZ(S3DPoint *polypoint, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = -line->nz;
   if (bottom == 0) return(false);
   utifloat top = -(polypoint->z-line->z);
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}

bool IntersectLinePlaneNX(S3DPoint *polypoint, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = -line->nx;
   if (bottom == 0) return(false);
   utifloat top = -(polypoint->x-line->x);
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}

bool IntersectLinePlaneNY(S3DPoint *polypoint, CVector *v, S3DPoint *line, S3DPoint *result, utifloat *dis)
{
   utifloat bottom = -line->ny;
   if ((bottom > -0.01) && (bottom < 0.01)) return(false);
   utifloat top = -(polypoint->y-line->y);
   utifloat mul = top/bottom;

   result->x = line->x + mul * line->nx;
   result->y = line->y + mul * line->ny;
   result->z = line->z + mul * line->nz;
   *dis = mul;
   return(true);
}

utifloat SearchRoundedAreaWidthZ(S3DPoint *points, int npol)
{
    utifloat xmin = (utifloat)1E+20;
    utifloat ymin = (utifloat)1E+20;
    utifloat xmax = (utifloat)-1E+20;
    utifloat ymax = (utifloat)-1E+20;

	for (int i = 0; i < npol; i++) {
		if (points[i].x > xmax) xmax = points[i].x;
		if (points[i].y > ymax) ymax = points[i].y;
		if (points[i].x < xmin) xmin = points[i].x;
		if (points[i].y < ymin) ymin = points[i].y;
	};
    utifloat w = xmin-xmax;
    utifloat h = ymin-ymax;
	return(w*h);
};

utifloat SearchRoundedAreaWidthY(S3DPoint *points, int npol)
{
    utifloat xmin = (utifloat)1E+20;
    utifloat zmin = (utifloat)1E+20;
    utifloat xmax = (utifloat)-1E+20;
    utifloat zmax = (utifloat)-1E+20;

	for (int i = 0; i < npol; i++) {
		if (points[i].x > xmax) xmax = points[i].x;
		if (points[i].z > zmax) zmax = points[i].z;
		if (points[i].x < xmin) xmin = points[i].x;
		if (points[i].z < zmin) zmin = points[i].z;
	};
    utifloat w = xmin-xmax;
    utifloat h = zmin-zmax;
	return(w*h);
};

utifloat SearchRoundedAreaWidthX(S3DPoint *points, int npol)
{
    utifloat ymin = (utifloat)1E+20;
    utifloat zmin = (utifloat)1E+20;
    utifloat ymax = (utifloat)-1E+20;
    utifloat zmax = (utifloat)-1E+20;

	for (int i = 0; i < npol; i++) {
		if (points[i].y > ymax) ymax = points[i].y;
		if (points[i].z > zmax) zmax = points[i].z;
		if (points[i].y < ymin) ymin = points[i].y;
		if (points[i].z < zmin) zmin = points[i].z;
	};
    utifloat w = ymin-ymax;
    utifloat h = zmin-zmax;
	return(w*h);
};


bool intersectPolygon(const Array<CVector2> &points, const CVector2 &point)
{
    bool r = false;
    int npol = points.size();
	int j = npol-1;
	int i;
	for (i = 0; i < npol; i++) {
		if ((((points[i].y <= point.y) && (point.y < points[j].y)) ||
			((points[j].y <= point.y) && (point.y < points[i].y))) &&
			(point.x < (points[j].x - points[i].x) * (point.y - points[i].y) /
			(points[j].y - points[i].y) + points[i].x)) {
			r = !r;
		};
		j = i;
	}
	return(r);
}


void clipPolygonNormal(const Array<CVector2> & inpoly, Array<CVector2> & outpoly, bool closed, const CVector2& normal, utifloat d)
{
  outpoly.setSize(4);
  UtiArrayAccessGrow<S2DPoint> grow(outpoly, 0);
  int segments = inpoly.size();
  if (segments <= 1) return;



  int j=1;
  for (int i = 0; i < segments; i++) {
    if (i == segments-1) {
      j = 0;
      if (!closed) {
	const CVector2& p1 = inpoly[i];
    utifloat a = normal.dotProduct(p1)-d;
	if (a >= 0) {
		grow.fast_add(p1);
	};
	break;
      };
    } else {
      j = i+1;
    };
    const CVector2& p1 = inpoly[i];
    const CVector2& p2 = inpoly[j];
    utifloat a = (normal.dotProduct(p1)-d);
    utifloat b = (normal.dotProduct(p2)-d);
    if (a >= 0 && b >= 0) {
		grow.fast_add(p1);
    } else if (a >= 0 && b < 0) {
		grow.fast_add(p1);
     //Schnittpunkt hinzufügen
      if (absf(a-b)< 0.001) {
		  grow.fast_add(p2);
      } else {
    utifloat t = a/(a-b);CVector2 newpoint = p1 + (p2-p1)*t;
	grow.fast_add(newpoint);
      };
    } else if (a < 0 && b >= 0) {
     //Schnittpunkt hinzufügen
     if (absf(a-b)< 0.001) {
		 grow.fast_add(p1);
      } else {
    utifloat t = a/(a-b);CVector2 newpoint = p1 + (p2-p1)*t;
	grow.fast_add(newpoint);
      };
    }
  };

   outpoly.setSizeCopy(grow.size());
}


float distanceToRectangle(const CRect& rect, const CVector2& point)
{
    if (rect.m_min.x <= point.x && point.x <= rect.m_max.x) {
        if (rect.m_min.y >= point.y) {
            return absf(rect.m_min.y-point.y);
        } else if (rect.m_max.y <= point.y) {
            return absf(rect.m_max.y-point.y);
        } else {
            return umin(umin(absf(rect.m_min.x-point.x), absf(rect.m_max.x-point.x)), umin(absf(rect.m_min.y-point.y), absf(rect.m_max.y-point.y)));
        }
    }
    if (rect.m_min.y <= point.y && point.y <= rect.m_max.y) {
        if (rect.m_min.x >= point.x) {
            return absf(rect.m_min.x-point.x);
        } else if (rect.m_max.x <= point.x) {
            return absf(rect.m_max.x-point.x);
        } else {
            return umin(umin(absf(rect.m_min.x-point.x), absf(rect.m_max.x-point.x)), umin(absf(rect.m_min.y-point.y), absf(rect.m_max.y-point.y)));
        }
    }
    float d1 = CVector2(rect.m_min.x, rect.m_min.y).distanceTo(point);
    float d2 = CVector2(rect.m_max.x, rect.m_min.y).distanceTo(point);
    float d3 = CVector2(rect.m_max.x, rect.m_max.y).distanceTo(point);
    float d4 = CVector2(rect.m_min.x, rect.m_max.y).distanceTo(point);

    return umin(umin(d1, d2), umin(d3, d4));
}
float distanceToRotatedRectangle(const CRect& rect, CVector2& origin, float rotation, const CVector2& point)
{
    CVector2 p = point-origin;
    p = rotateVector(p, -rotation);
    return distanceToRectangle(rect, p);
}

CVector2f rotateVector(const CVector2f& p, float degree)
{
  CVector2f r;
  degree *= pi180;
  float asin = sin(degree);
  float acos = cos(degree);
  r.x = p.x*acos + p.y*asin;
  r.y = -p.x*asin + p.y*acos;
  return r;
}

CVector2 rotateVector(const CVector2& p, utifloat degree)
{
  CVector2 r;
  degree *= pi180;
  utifloat asin = sin(degree);
  utifloat acos = cos(degree);
  r.x = p.x*acos + p.y*asin;
  r.y = -p.x*asin + p.y*acos;
  return r;
}

void clipPolygonLineSegment(const Array<CVector2> & inpoly, Array<CVector2> & outpoly, bool closed, const CVector2& p1, const CVector2& p2)
{
  CVector2 dir = p2-p1;
  dir.normalize();
  CVector2 normal = rotateVector(dir, 90);
  utifloat d = normal.dotProduct(p1);
  clipPolygonNormal(inpoly, outpoly, closed, dir, d);
}

void clipPolygonConvexPolygon(const Array<CVector2> & inpoly, Array<CVector2> & outpoly, bool closed, const Array<CVector2>& convexmesh)
{
  if (convexmesh.size() <= 3) return;
  Array<CVector2> temp1;
  bool in_temp = true;
  clipPolygonLineSegment(inpoly, temp1, closed, convexmesh[0], convexmesh[1]);
  if (temp1.size() <= 1) {
    outpoly.setSize(0);
    return;
  };

  for (int i = 1; i < convexmesh.size(); i++) {
     int j = (i+1)% convexmesh.size();

     if (in_temp) {
       clipPolygonLineSegment(temp1, outpoly, closed, convexmesh[i], convexmesh[j]);
        if (outpoly.size() <= 1) {
        outpoly.setSize(0);
         return;
         };
     } else {
        clipPolygonLineSegment(outpoly, temp1, closed, convexmesh[i], convexmesh[j]);
	if (temp1.size() <= 1) {
          outpoly.setSize(0);
         return;
       };

    }

     in_temp = !in_temp;

  };
  if (in_temp) {
    outpoly.setSize(temp1.size());
    temp1.copyToArray(0, temp1.size(), outpoly, 0);

  };
}

void clipPolygonRect(const Array<CVector2> & inpoly, Array<CVector2> & outpoly, bool closed, const CRect& rect)
{
  utiassert(rect.valid());
  if (rect.getWidth() <= 0.0001 || rect.getHeight() <= 0.0001 || rect.m_min.x > rect.m_max.x || rect.m_min.y > rect.m_max.y) {
      outpoly.setSize(0);
      return;
  };

  Array<CVector2> temp1;
  clipPolygonNormal(inpoly, temp1, closed, CVector2(1, 0), rect.m_min.x);
  if (temp1.size() <= 1) {
    outpoly.setSize(0);
    return;
  };
  clipPolygonNormal(temp1, outpoly, closed, CVector2(-1, 0), -rect.m_max.x);
  if (outpoly.size() <= 1) {
    outpoly.setSize(0);
    return;
  };

  clipPolygonNormal(outpoly, temp1, closed, CVector2(0, 1), rect.m_min.y);
  if (temp1.size() <= 1) {
    outpoly.setSize(0);
    return;
  };

  clipPolygonNormal(temp1, outpoly, closed, CVector2(0, -1), -rect.m_max.y);
  if (temp1.size() <= 1) {
    outpoly.setSize(0);
    return;
  };
}


bool intersectPolygonRect(const Array<CVector2> & inpoly, bool closed, const CRect& rect)
{
  Array<CVector2> res;
  clipPolygonRect(inpoly, res, closed, rect);
  return res.size() >= 2;
}

bool intersectLineRect(const CVector2& start, const CVector2& dir, const CRect& rect, utifloat& pos)
{
  utiassert(rect.valid());
  if (rect.contains(start)) return true;
  //Warnung:
   utifloat t1, t2;
   utifloat minx, maxx, miny, maxy;
   if (abs(dir.x) < 0.0001)
   {
     if (rect.m_min.x <start.x && rect.m_max.x > start.x) {
       minx = -MAXFLOAT;
	maxx = MAXFLOAT;
     } else {
       return(false);
     }
   } else {
      t1 = (rect.m_min.x - start.x) / dir.x;
      t2 = (rect.m_max.x - start.x) / dir.x;
      if (t1 < t2)
      {
         minx = t1;
         maxx = t2;
      } else {
         minx = t2;
         maxx = t1;
      };
   }
   if (abs(dir.y) < 0.0001)
   {
     if (rect.m_min.y <start.y && rect.m_max.y > start.y) {
       miny = -MAXFLOAT;
	maxy = MAXFLOAT;
     } else {
       return(false);
     }
   } else {
      t1 = (rect.m_min.y - start.y) / dir.y;
      t2 = (rect.m_max.y - start.y) / dir.y;
      if (t1 < t2)
      {
         miny = t1;
         maxy = t2;
      } else {
         miny = t2;
         maxy = t1;
      };
   }

   utifloat t_min = umax(minx, miny);
   utifloat t_max = umin(maxx, maxy);
   if (t_min < t_max) {
    if (t_min < pos && t_min >= 0) {
      pos = t_min;
      return true;
    } else if (t_max < pos && t_max >= 0) {
      pos = t_max;
      return true;
    }

   }
   return false;
}


bool intersectPolylineRect(const Array<CVector2> & inpoly, bool closed, const CRect& rect)
{
    int seg = inpoly.size();
    if (!closed) seg -= 1;

    for (int i = 0; i < seg; i++) {
      int j = (i+1)%inpoly.size();
      const CVector2& p1 = inpoly[i];
      const CVector2& p2 = inpoly[j];
      utifloat pos = 1.0f;
      if (intersectLineRect(p1, p2-p1, rect,pos)) return true;
    };
    return false;
}


bool Has2DPointZ(S3DPoint *points, int npol, S3DPoint *point)
{
    bool r = false;
	int j = npol-1;
	int i;
	for (i = 0; i < npol; i++) {
		if ((((points[i].y <= point->y) && (point->y < points[j].y)) ||
			((points[j].y <= point->y) && (point->y < points[i].y))) &&
			(point->x < (points[j].x - points[i].x) * (point->y - points[i].y) /
			(points[j].y - points[i].y) + points[i].x)) {
			r = !r;
		};
		j = i;
	}
	return(r);
}

bool Has2DPointY(S3DPoint *points, int npol, S3DPoint *point)
{
    bool r = false;
	int j = npol-1;
	int i;
	for (i = 0; i < npol; i++) {
		if ((((points[i].z <= point->z) && (point->z < points[j].z)) ||
			((points[j].z <= point->z) && (point->z < points[i].z))) &&
			(point->x < (points[j].x - points[i].x) * (point->z - points[i].z) /
			(points[j].z - points[i].z) + points[i].x)) {
			r = !r;
		};
		j = i;
	}
	return(r);
}

bool Has2DPointX(S3DPoint *points, int npol, S3DPoint *point)
{
    bool r = false;
	int j = npol-1;
	int i;
	for (i = 0; i < npol; i++) {
		if ((((points[i].z <= point->z) && (point->z < points[j].z)) ||
			((points[j].z <= point->z) && (point->z < points[i].z))) &&
			(point->y < (points[j].y - points[i].y) * (point->z - points[i].z) /
			(points[j].z - points[i].z) + points[i].y)) {
			r = !r;
		};
		j = i;
	}
	return(r);
}

bool Has2DPointZ(CVector* *points, int npol, S3DPoint *point)
{
    bool r = false;
	CVector** xj = points+ (npol-1);
	CVector** end = points + npol;
	CVector** xi = NULL;
	for (xi = points; xi != end; xi++) {
	    CVector* i = *xi;
	    CVector* j = *xj;
		if ((((i->y <= point->y) && (point->y < j->y)) ||
			((j->y <= point->y) && (point->y < i->y))) &&
			(point->x < (j->x - i->x) * (point->y - i->y) /
			(j->y - i->y) + i->x)) {
			r = !r;
		};
		xj = xi;
	}
	return(r);
}

bool Has2DPointZ(CVector3f* *points, int npol, S3DPoint *point)
{
    bool r = false;
    CVector3f** xj = points+ (npol-1);
    CVector3f** end = points + npol;
    CVector3f** xi = NULL;
    for (xi = points; xi != end; xi++) {
        CVector3f* i = *xi;
        CVector3f* j = *xj;
        if ((((i->y <= point->y) && (point->y < j->y)) ||
            ((j->y <= point->y) && (point->y < i->y))) &&
            (point->x < (j->x - i->x) * (point->y - i->y) /
            (j->y - i->y) + i->x)) {
            r = !r;
        };
        xj = xi;
    }
    return(r);
}

bool Has2DPoint2D(const Array<CVector2>& points, int npol, const CVector2 &point)
{
    bool r = false;
    int xj = (npol-1);
    int end =  npol;
    int xi = 0;
    for (xi = 0; xi != end; xi++) {
        const CVector2& i = points[xi];
        const CVector2& j = points[xj];
        if ((((i.y <= point.y) && (point.y < j.y)) ||
            ((j.y <= point.y) && (point.y < i.y))) &&
            (point.x < (j.x - i.x) * (point.y - i.y) /
            (j.y - i.y) + i.x)) {
            r = !r;
        };
        xj = xi;
    }
    return(r);
}

bool Has2DPoint2D(const Array<CVector2f>& points, int npol, const CVector2f &point)
{
    bool r = false;
    int xj = (npol-1);
    int end =  npol;
    int xi = 0;
    for (xi = 0; xi != end; xi++) {
        const CVector2f& i = points[xi];
        const CVector2f& j = points[xj];
        if ((((i.y <= point.y) && (point.y < j.y)) ||
            ((j.y <= point.y) && (point.y < i.y))) &&
            (point.x < (j.x - i.x) * (point.y - i.y) /
            (j.y - i.y) + i.x)) {
            r = !r;
        };
        xj = xi;
    }
    return(r);
}


bool Has2DPointY(CVector **points, int npol, S3DPoint *point)
{
    bool r = false;
	CVector** xj = points+ (npol-1);
	CVector** end = points + npol;
	CVector** xi = NULL;
	for (xi = points; xi != end; xi++) {
	    CVector* i = *xi;
	    CVector* j = *xj;
		if ((((i->z <= point->z) && (point->z < j->z)) ||
			((j->z <= point->z) && (point->z < i->z))) &&
			(point->x < (j->x - i->x) * (point->z - i->z) /
			(j->z - i->z) + i->x)) {
			r = !r;
		};
		xj = xi;
	}
	return(r);
}

bool Has2DPointX(CVector **points, int npol, S3DPoint *point)
{
    bool r = false;
	CVector** xj = points+ (npol-1);
	CVector** end = points + npol;
	CVector** xi = NULL;
	for (xi = points; xi != end; xi++) {
	    CVector* i = *xi;
	    CVector* j = *xj;
		if ((((i->z <= point->z) && (point->z < j->z)) ||
			((j->z <= point->z) && (point->z < i->z))) &&
			(point->y < (j->y - i->y) * (point->z - i->z) /
			(j->z - i->z) + i->y)) {
			r = !r;
		};
		xj = xi;
	}
	return(r);
}


bool Has2DPointY(CVector3f **points, int npol, S3DPoint *point)
{
    bool r = false;
    CVector3f** xj = points+ (npol-1);
    CVector3f** end = points + npol;
    CVector3f** xi = NULL;
    for (xi = points; xi != end; xi++) {
        CVector3f* i = *xi;
        CVector3f* j = *xj;
        if ((((i->z <= point->z) && (point->z < j->z)) ||
            ((j->z <= point->z) && (point->z < i->z))) &&
            (point->x < (j->x - i->x) * (point->z - i->z) /
            (j->z - i->z) + i->x)) {
            r = !r;
        };
        xj = xi;
    }
    return(r);
}

bool Has2DPointX(CVector3f **points, int npol, S3DPoint *point)
{
    bool r = false;
    CVector3f** xj = points+ (npol-1);
    CVector3f** end = points + npol;
    CVector3f** xi = NULL;
    for (xi = points; xi != end; xi++) {
        CVector3f* i = *xi;
        CVector3f* j = *xj;
        if ((((i->z <= point->z) && (point->z < j->z)) ||
            ((j->z <= point->z) && (point->z < i->z))) &&
            (point->y < (j->y - i->y) * (point->z - i->z) /
            (j->z - i->z) + i->y)) {
            r = !r;
        };
        xj = xi;
    }
    return(r);
}

void Add(S3DPoint* result, S3DPoint* sum1, S3DPoint* sum2)
{
   result->x = sum1->x + sum2->x;
   result->y = sum1->y + sum2->y;
   result->z = sum1->z + sum2->z;
}

void Add(S3DPoint* result, S3DPoint* sum1, CVector* sum2)
{
   result->x = sum1->x + sum2->x;
   result->y = sum1->y + sum2->y;
   result->z = sum1->z + sum2->z;
}

void Add(S3DPoint* result, CVector* sum1, S3DPoint* sum2)
{
   result->x = sum1->x + sum2->x;
   result->y = sum1->y + sum2->y;
   result->z = sum1->z + sum2->z;
}

void Add(S3DPoint* result, CVector* sum1, CVector* sum2)
{
   result->x = sum1->x + sum2->x;
   result->y = sum1->y + sum2->y;
   result->z = sum1->z + sum2->z;
}

void AddV(S3DPoint* result, S3DPoint* sum1, S3DPoint* sum2)
{
   result->nx = sum1->x + sum2->x;
   result->ny = sum1->y + sum2->y;
   result->nz = sum1->z + sum2->z;
}

void AddV(S3DPoint* result, S3DPoint* sum1, CVector* sum2)
{
   result->nx = sum1->x + sum2->x;
   result->ny = sum1->y + sum2->y;
   result->nz = sum1->z + sum2->z;
}

void AddV(S3DPoint* result, CVector* sum1, S3DPoint* sum2)
{
   result->nx = sum1->x + sum2->x;
   result->ny = sum1->y + sum2->y;
   result->nz = sum1->z + sum2->z;
}

void AddV(S3DPoint* result, CVector* sum1, CVector* sum2)
{
   result->nx = sum1->x + sum2->x;
   result->ny = sum1->y + sum2->y;
   result->nz = sum1->z + sum2->z;
}

int CPlane::ClassifyPolygon(S3DPoint* list, int count)
{

   int side = getPlanePos(list[0]);
   for (int i = 1; i < count; i++) {
	   if (side != getPlanePos(list[i])) {
          return(BSP_CUT);
	   };
   };
   return(side);
}

int CPlane::ClassifyPolygon(UINT* list, CVertex* v, int count)
{
   int side = getPlanePos(v[list[0]].pos);
   for (int i = 1; i < count; i++) {
	   if (side != getPlanePos(v[list[i]].pos)) {
          return(BSP_CUT);
	   };
   };
   return(side);

}

void operator += (S3DPoint &a, const S3DPoint &b)
{
	a.x += b.x;
	a.y += b.y;
	a.z += b.z;
}

void operator += (S3DPoint &a, const CVector &b)
{
	a.x += b.x;
	a.y += b.y;
	a.z += b.z;
}

void operator *= ( S3DPoint &a, const float b )
{
	a.x *= b;
	a.y *= b;
	a.z *= b;
}

S3DPoint operator * (const S3DPoint &a, const float b)
{
	S3DPoint result;

	result.x = a.x * b;
	result.y = a.y * b;
	result.z = a.z * b;
	return result;
}

S3DPoint operator - ( const S3DPoint &a, const S3DPoint &b )
{
	S3DPoint result;

	result.x = a.x - b.x;
	result.y = a.y - b.y;
	result.z = a.z - b.z;

	return result;
}

S3DPoint operator - ( const S3DPoint &a, const CVector &b )
{
	S3DPoint result;

	result.x = a.x - b.x;
	result.y = a.y - b.y;
	result.z = a.z - b.z;

	return result;
}

S3DPoint	operator + ( const S3DPoint &a, const S3DPoint &b )
{
	S3DPoint result;

	result.x = a.x + b.x;
	result.y = a.y + b.y;
	result.z = a.z + b.z;

	return result;
}

S3DPoint	operator + ( const S3DPoint &a, const CVector &b )
{
	S3DPoint result;

	result.x = a.x + b.x;
	result.y = a.y + b.y;
	result.z = a.z + b.z;

	return result;
}


S3DPoint operator ^ ( const S3DPoint &a, const S3DPoint &b )
{
	S3DPoint result;

	result.x = a.y * b.z - b.y * a.z;
	result.y = a.z * b.x - b.z * a.x;
	result.z = a.x * b.y - b.x * a.y;

	return result;
}

S3DPoint operator ^ ( const S3DPoint &a, const CVector &b )
{
	S3DPoint result;

	result.x = a.y * b.z - b.y * a.z;
	result.y = a.z * b.x - b.z * a.x;
	result.z = a.x * b.y - b.x * a.y;

	return result;
}


float	operator * ( const S3DPoint &a, const S3DPoint &b )
{
	return ( a.x * b.x + a.y * b.y + a.z * b.z );
}

float	operator * ( const S3DPoint &a, const CVector &b )
{
	return ( a.x * b.x + a.y * b.y + a.z * b.z );
}

CVector2 solveLinearEquation2(const CVector2&v1, const CVector2&v2, const CVector2& c, bool* hassolution)
{
  CVector2 res;
  float r = v1.x*v2.y-v2.x*v1.y;
  if (absf(r) < 0.000001) {
  if (hassolution) {
	*hassolution = false;
      };
      return CVector2(0,0);
  }

  CVector2 v1new = CVector2(v2.y/r,
			   -v1.y/r);
  CVector2 v2new = CVector2(-v2.x/r,
			   v1.x/r);

   res.x = c.x*v1new.x+c.y*v2new.x;
   res.y = c.x*v1new.y+c.y*v2new.y;

   //CVector2 ctest = CVector2(res.x*v1.x+res.y*v2.x, res.x*v1.y+res.y*v2.y);
  // utiassert(ctest.distanceTo(c) < 0.01);
  /*if (absf(v1.x) < absf(v1.y)) {

  } else {

    float c2 = c.y;
    float f = v2.y;
    float scal=1.0f;
    if (absf(v1.y) > 0.000001) {
      scal = (v1.x/ v1.y);
      c2 = c2 *scal - c.x;
      f = f *scal - v1.x;
    };
    if (absf(f) < 0.00001) {
      if (hassolution) {
	return;
      };
    };
    res.y = c2 / f;


    c2 = c.x-v2.x*res.y;
    f = v1.x;

    if (absf(f) < 0.00001) {
      if (hassolution) {
	return;
      };
    };


  };*/
  if (hassolution) {
	*hassolution = true;
      };
      return res;
}

CVector2 intersectLines(const CVector2&p1, const CVector2&dir1, const CVector2& p2, const CVector2& dir2, bool* hassolution)
{
   CVector2 res = solveLinearEquation2(-dir1, dir2, p1-p2, hassolution);
   //CVector2 r1 = p1+dir1*res.x;
   //CVector2 r2 = p2+dir2*res.y;
   if (hassolution != NULL && *hassolution) {
   //utiassert(r1.distanceTo(r2) < 0.01);
   }
   return res;
}


utifloat angleBetween(const CVector2&v1, const CVector2&v2)
{
  CVector2 n1 = v1;
  CVector2 n2 = v2;
  n1.normalize();
  n2.normalize();
  utifloat angle = dot(n1, n2);
  angle = acos(angle)/PI*180.0f;
  return angle;
}

void GetRandomPointInRect(CVector* p, CBox* r)
{
   p->x = r->m_min.x + (r->m_max.x-r->m_min.x) * urandom();
   p->y = r->m_min.y + (r->m_max.y-r->m_min.y) * urandom();
   p->z = r->m_min.z + (r->m_max.z-r->m_min.z) * urandom();
}

void GetRandomPointInRect(S3DPoint* p, CBox* r)
{
   p->x = r->m_min.x + (r->m_max.x-r->m_min.x) * urandom();
   p->y = r->m_min.y + (r->m_max.y-r->m_min.y) * urandom();
   p->z = r->m_min.z + (r->m_max.z-r->m_min.z) * urandom();
}



void calcBumpTangent(const CVector3f& P, const CVector3f & Q, const CVector3f& N, float s1, float s2, float t1, float t2, CVector3f & tangent, float &m)
{
    float k = s1*t2 - s2*t1;
    if (absf(k) < 0.0001 || N.length() > 10000.0f || (k != k)) {
        tangent.setVector(0,0,1);
        m = 1;
        return;
    }
    k = 1.0f/k;
    CVector3f T, B, bi2;
	/*T.x = k*(t2*P.x - t1 * Q.x);
	T.y = k*(t2*P.y - t1 * Q.y);
	T.z = k*(t2*P.z - t1 * Q.z);*/
    T = (P*t2-Q*t1)*k;
	/*B.x = k*(-s2*P.x + s1*Q.x);
	B.y = k*(-s2*P.y + s1*Q.y);
	B.z = k*(-s2*P.z + s1*Q.z);*/
    B = (P*(-s2) + Q*s1)*k;
	tangent = T - N*(N*T);
	bi2 = B - N*(N*B) - tangent*(tangent*B);
	tangent.normalize();

   // tangent.neg();

	bi2.normalize();
    m = tangent.x*bi2.y*N.z+tangent.y*bi2.z*N.x+tangent.z*bi2.x*N.y
		- tangent.z*bi2.y*N.x- tangent.x*bi2.z*N.y- tangent.y*bi2.x*N.z;
	m = m < 0 ? -1.0f : 1.0f;

   assert(std::isnan(tangent.x) ==false);
   assert(std::isnan(tangent.y) ==false);
   assert(std::isnan(tangent.z) ==false);
	//m = 1.0f;
	//CVector testbi;
	//N.CrossProduct(&tangent, &testbi);
	//testbi = testbi* m;
//	utiassert((testbi-bi2).length() < 0.01);
}

float intern_inoise(const CIVector2& position)
{
  return intern_inoise(position.x+position.y*55);
}

float intern_inoise(const CIVector3& position)
{
  return intern_inoise(position.x+position.y*55+position.z*794);
}

int expandTo2(int w)
{
   int x = 1;
   int c = 0;
   while (x < w) {
     x = x*2;
     c++;
     if (c > 30) return x;
   };
   return x;
}

int bintrunc(double f)
{

    /*#ifdef _WIN32
#ifdef _MSC_VER
#pragma warning Dies ist nicht sehr gut implementiert
	return f > 0.0 ? (int)(f) : (int)(f-1.0f);
#else
    return trunc(f);
#endif
#else*/
    if (f >= 0) {
    return (int)f;
    } else {
      int res = (int) (f-0.999999);
      return res;
    };
//#endif
};



int bintruncd(double f)
{
    if (f >= 0) {
    return (int)f;
    } else {
      int res = (int) (f-0.99999999);
      return res;
    };
//#endif
};

/*
void IntersectQuaderRayMinMax(const S3DPoint& vector, const SBounds& b, float &t_min, float &t_max)
{
   //Warnung:
   float t1, t2;
   float minx, maxx, miny, maxy, minz, maxz;

   t1 = (b.p1.x - vector.x) / vector.nx;
   t2 = (b.p2.x - vector.x) / vector.nx;
   if (t1 < t2)
   {
      minx = t1;
      maxx = t2;
   } else {
      minx = t2;
      maxx = t1;
   };

   t1 = (b.p1.y - vector.y) / vector.ny;
   t2 = (b.p2.y - vector.y) / vector.ny;
   if (t1 < t2)
   {
      miny = t1;
      maxy = t2;
   } else {
      miny = t2;
      maxy = t1;
   };

   t1 = (b.p1.z - vector.z) / vector.nz;
   t2 = (b.p2.z - vector.z) / vector.nz;
   if (t1 < t2)
   {
      minz = t1;
      maxz = t2;
   } else {
      minz = t2;
      maxz = t1;
   };

   t_min = umax(minx, umax(miny, minz));
   t_max = umin(maxx, umin(maxy, maxz));
};
*/

void CRect::getRectAsArray(Array<CVector2> &marray)const
{
  marray.setSize(4);
  marray[0] = m_min;
  marray[1] = CVector2(m_max.x, m_min.y);
  marray[2] = m_max;
  marray[3] = CVector2(m_min.x, m_max.y);
}

utifloat CBox::getMinMaxDistance(const CVector3& p) const
{
  utifloat distance;
   if (contains(p)) {
     distance = 0;
   } else {
       distance = umax (    umax(umax(0.0, m_min.x-p.x), umax(0.0, p.x-m_max.x)),
               umax(    umax(umax(0.0, m_min.y-p.y), umax(0.0, p.y-m_max.y)),


               umax(umax(0.0, m_min.z-p.z) ,umax(0.0, p.z-m_max.z))    ));


    /*  distance = umax (    umin(absf(p.x-m_min.x), absf(p.x-m_max.x)),
		      umax(    umin(absf(p.y-m_min.y), absf(p.y-m_max.y)),


              umin(absf(p.z-m_min.z) ,absf(p.z-m_max.z))    ));*/

    // float otherdistance = umax(absf(p.x-m_max.x), umax(absf(p.y-m_max.y), absf(p.z-m_max.z)));
      //distance = umax(distance, otherdistance);
   };
   return distance;
}

utifloat CBox::getMinMaxDistanceXZ(const CVector3& p, float factor) const
{
	utifloat distance;
	if (contains(p)) {
		distance = 0;
	}
	else {
		
        distance = umax(umax(umax(0.0, m_min.x - p.x), umax(0.0, p.x - m_max.x)),
            umax(umax(umax(0.0, (m_min.y - p.y)*factor), umax(0.0, (p.y - m_max.y)*factor)),


            umax(umax(0.0, m_min.z - p.z), umax(0.0, p.z - m_max.z))));


		/*  distance = umax (    umin(absf(p.x-m_min.x), absf(p.x-m_max.x)),
		umax(    umin(absf(p.y-m_min.y), absf(p.y-m_max.y)),


		umin(absf(p.z-m_min.z) ,absf(p.z-m_max.z))    ));*/

		// float otherdistance = umax(absf(p.x-m_max.x), umax(absf(p.y-m_max.y), absf(p.z-m_max.z)));
		//distance = umax(distance, otherdistance);
	};
	return distance;
}


void CBox::octreeSubdivide(CBox* boxes) const
{
  CVector middle = center();
  boxes[0] = CBox(m_min.x, m_min.y, m_min.z, middle.x, middle.y,middle.z);
  boxes[1] = CBox(middle.x, m_min.y, m_min.z, m_max.x, middle.y,middle.z);
  boxes[2] = CBox(m_min.x, middle.y, m_min.z, middle.x, m_max.y,middle.z);
  boxes[3] = CBox(middle.x, middle.y, m_min.z, m_max.x, m_max.y,middle.z);
  
  boxes[4] = CBox(m_min.x, m_min.y, middle.z, middle.x, middle.y,m_max.z);
  boxes[5] = CBox(middle.x, m_min.y, middle.z, m_max.x, middle.y,m_max.z);
  boxes[6] = CBox(m_min.x, middle.y, middle.z, middle.x, m_max.y,m_max.z);
  boxes[7] = CBox(middle.x, middle.y, middle.z, m_max.x, m_max.y,m_max.z);
}


CPlane CBox::getPlane(int i, bool outp) const
{
	CPlane ret;

	switch (i) {
	case 0:
	{
		//X+
		ret.cvVector.setVector(1,0,0);
		ret.d = m_max.x;
	}; break;
	case 1:
	{
		//X-
		ret.cvVector.setVector(-1, 0, 0);
		ret.d = -m_min.x;
	}; break;
	case 2:
	{
		//Y+
		ret.cvVector.setVector(0, 1, 0);
		ret.d = m_max.y;
	}; break;
	case 3:
	{
		//Y-
		ret.cvVector.setVector(0, -1, 0);
		ret.d = -m_min.y;
	}; break;
	case 4:
	{
		//Z+
		ret.cvVector.setVector(0, 0, 1);
		ret.d = m_max.z;
	}; break;
	case 5:
	{
		//Z-
		ret.cvVector.setVector(0, 0, -1);
		ret.d = -m_min.z;
	}; break;
	default: {

		assert(false);
	}; break;
	}

	if (!outp) {
		ret.cvVector = -ret.cvVector;
		ret.d = -ret.d;
	}
	return ret;
}


CPlanef CBox::getPlanef(int i, bool outp) const
{
    CPlanef ret;

    switch (i) {
    case 0:
    {
        //X+
        ret.cvVector.setVector(1,0,0);
        ret.d = m_max.x;
    }; break;
    case 1:
    {
        //X-
        ret.cvVector.setVector(-1, 0, 0);
        ret.d = -m_min.x;
    }; break;
    case 2:
    {
        //Y+
        ret.cvVector.setVector(0, 1, 0);
        ret.d = m_max.y;
    }; break;
    case 3:
    {
        //Y-
        ret.cvVector.setVector(0, -1, 0);
        ret.d = -m_min.y;
    }; break;
    case 4:
    {
        //Z+
        ret.cvVector.setVector(0, 0, 1);
        ret.d = m_max.z;
    }; break;
    case 5:
    {
        //Z-
        ret.cvVector.setVector(0, 0, -1);
        ret.d = -m_min.z;
    }; break;
    default: {

        assert(false);
    }; break;
    }

    if (!outp) {
        ret.cvVector = -ret.cvVector;
        ret.d = -ret.d;
    }
    return ret;
}

CBox transformBox(const CMatrix& mat, const CBox& box)
{
    UTI_DECL_ARRAY(CVector3, boxedges, 8);

    box.getEdgePoints(boxedges);
    for (int ed = 0; ed < 8; ed++) {
        boxedges[ed] = mat.TransformP(boxedges[ed]);
    }
    CBox b = CBox::ADVANCEINIT();
    b.expandArray(boxedges, 0, boxedges.size());
    return b;
}


void SHSPlane::load(const CVector4& plane)
  {
     positive.x = std::max(plane.x, 0.0);
     positive.y = std::max(plane.y, 0.0);
     positive.z = std::max(plane.z, 0.0);
     negative.x = std::min(plane.x, 0.0);
     negative.y = std::min(plane.y, 0.0);
     negative.z = std::min(plane.z, 0.0);
	 //el->plane = plane;
	 plane_pos = plane.w;
	 original = plane;
  };
  
void transformTranspose(const CMatrix& m, const CVector4 &v, CVector4 &res)
{
    res.x = m.m[0][0]*v.x+m.m[1][0]*v.y+m.m[2][0]*v.z+m.m[3][0]*v.w;
   res.y = m.m[0][1]*v.x+m.m[1][1]*v.y+m.m[2][1]*v.z+m.m[3][1]*v.w;
   res.z = m.m[0][2]*v.x+m.m[1][2]*v.y+m.m[2][2]*v.z+m.m[3][2]*v.w;
   res.w = -m.m[0][3]*v.x-m.m[1][3]*v.y-m.m[2][3]*v.z+m.m[3][3]*v.w;
};
  
utifloat rangecalc(utifloat value, utifloat min, utifloat max, utifloat bordermin, utifloat bordermax)
{
    if (value < min)return 0.0f;

    if (value > max)return 0.0f;
    if (value > min+bordermin && value < max-bordermax) {
        return 1.0f;
    }
    if (max-min >= bordermin+bordermax) {

       if (value < min+bordermin) {
          return (value -min)/bordermin;
       }
       //if (value > max-border) {
           return 1.0f- (value -(max-bordermax))/bordermax;
       //}
    } else {
        assert(false);

        if (value < min+bordermin) {
           return (value -min)/bordermin;
        }
        //if (value > max-border) {
            return 1.0f- (value -(max-bordermax))/bordermax;
        //}

    }

    return 0.0f;
}

CVector4 utiTransformPlane(CVector4&plane, const CMatrix& m)
{
   CVector4 v;
   transformTranspose(m, plane, v);
   CVector3 normal = v.xyz();
   utifloat l = normal.length();
   if (l > 0.001){
       v = v*(1.0f/l);
   }
  return v;
}
  


bool CMatrix::isValid() const
{
    utifloat lx = getXAxis().length();
    utifloat ly = getYAxis().length();
    utifloat lz = getZAxis().length();
    utifloat l = getPosition().length();
   /* assert(fabs(getXAxis().length()-1.0) < 0.01);
    assert(fabs(getYAxis().length()-1.0) < 0.01);
    assert(fabs(getZAxis().length()-1.0) < 0.01);*/
   /* assert(fabs(lx-ly) <0.1);
    assert(fabs(lx-lz) <0.1);*/
    //assert(fabs(getXAxis().dotProduct(getYAxis())) < 0.01);
   // assert(fabs(getXAxis().dotProduct(getZAxis())) < 0.01);
   // assert(fabs(getYAxis().dotProduct(getZAxis())) < 0.01);
    return (lx > 0.0001) &&(lx < 1000)&& (ly > 0.0001) &&(ly < 1000) && (lz > 0.0001) &&(lz < 1000) && (l < 1000000);
}

TUtiBoxVis utiVisBox(const CBox & b, SHSPlane* planes, int planecount)
{
	TUtiBoxVis res = ubvOFull;
   utifloat bmin, bmax;
   SHSPlane* curr = planes;
   for (int i = 0; i < planecount; i++) { 
	   //bmax = dot(b.m_min, curr->negative)+ dot(b.m_max, curr->positive);
	  // bmin = dot(b.m_min, curr->positive)+ dot(b.m_max, curr->negative);
       assert(std::isnan(curr->original.x) == false);
       assert(std::isnan(curr->original.y) == false);
       assert(std::isnan(curr->original.z) == false);
       bmax = b.m_min.x*curr->negative.x + b.m_min.y*curr->negative.y + b.m_min.z*curr->negative.z;
	   bmax += b.m_max.x*curr->positive.x + b.m_max.y*curr->positive.y + b.m_max.z*curr->positive.z;
       assert(absf(dot(b.m_min, curr->negative) + dot(b.m_max, curr->positive) - bmax) < 0.01);
       if (bmax < curr->plane_pos) return ubvONone;
	   bmin = b.m_min.x*curr->positive.x + b.m_min.y*curr->positive.y + b.m_min.z*curr->positive.z;
	   bmin += b.m_max.x*curr->negative.x + b.m_max.y*curr->negative.y + b.m_max.z*curr->negative.z;

	   
       assert(absf(dot(b.m_min, curr->positive) + dot(b.m_max, curr->negative) - bmin) < 0.01);

	   
	   if (bmin < curr->plane_pos) res = ubvOHalf;

	   curr++;
   };
   return res;
};

TUtiBoxVis utiVisBox(const CBox & b, SHSPlane* planes, int planecount, const CMatrix& invtransform)
{
	TUtiBoxVis res = ubvOFull;
   utifloat bmin, bmax;
   SHSPlane* curr = planes;
   //CVector3 minvalue = invtransform.TransformP(b.m_min);
  // CVector3 maxvalue = invtransform.TransformP(b.m_max);
   CVector4 newplane;
   CVector3 positive(0);
   CVector3 negative(0);
   for (int i = 0; i < planecount; i++) { 
           newplane = utiTransformPlane(curr->original, invtransform);
	   
	   
         positive.x = std::max(newplane.x, 0.0);
     positive.y = std::max(newplane.y, 0.0);
     positive.z = std::max(newplane.z, 0.0);
     negative.x = std::min(newplane.x, 0.0);
     negative.y = std::min(newplane.y, 0.0);
     negative.z = std::min(newplane.z, 0.0);
	   
	   bmax = dot(b.m_min, negative)+ dot(b.m_max, positive);
	   bmin = dot(b.m_min, positive)+ dot(b.m_max, negative);
	   if (bmax < newplane.w) return ubvONone;
	   if (bmin < newplane.w) res = ubvOHalf;

	   curr++;
   };
   return res;
};

CVector uVectorToAngle(const CVector& v2)
{
   /* CVector v = v2;
    CVector result;
    float w, l;
    l = (float)sqrt(v.x*v.x + v.z*v.z);
    if (l != 0) {
        w = (float)asin(v.z/l)/(PI / 180);
        if (v.x < 0) {
            w = 180-w;
        };
        w = 90-w;
        result.y=(w);
    } else {
        result.y=(0);
    };

    CMatrix m;
    m.LoadIdentity();
    m.Rotate(0, result.y, 0);
    m.Transform(&v);

    v.normalize();
    w = (float)acos(v.z)/(PI/180);
    if (v.y < 0) {
        w = 360-w;
    };
    result.x =-w;
    result.z =0;

    //Check it
    m.LoadIdentity();
    m.Rotate(result);
    CVector testdir = m.TransformP(CVector(0,0,1));
    v = v2;
    v.normalize();
    assert(testdir.distanceTo(v) < 0.01);*/

	CVector v = v2; v.normalize();
        CVector result;
        utifloat w, l;
        l = (utifloat)sqrt(v.x*v.x + v.z*v.z);
        if (l != 0) {
            w = (utifloat)asin(v.z/l)/(PI / 180);
            if (v.x < 0) {
                w = 180-w;
            };
            w = 90-w;
            result.y=(-w);
        } else {
            result.y=(0);
        };

        CMatrix m;
        m.LoadIdentity();
        m.Rotate(0, result.y, 0);
        m.Transform(&v);

        v.normalize();
        w = (utifloat)acos(v.z)/ (PI/180);
        if (v.y < 0) {
            w = 360-w;
        };
        result.x =-w;
        result.z =0;



        //result.x = 90-result.x;
        result.y = -result.y;

        //Check it
        m.LoadIdentity();
        m.Rotate(result);
        CVector testdir = m.TransformP(CVector(0,0,1));
        v = v2;
        v.normalize();
        assert(testdir.distanceTo(v) < 0.01);



/*l := sqrt(sqr(p.x)+sqr(p.z));
   if l <> 0 then begin
      w := arcsin(p.z/l)/(pi/180);
      //w := arcatn(p.x/p.z)/(pi/180);
      if p.x < 0 then begin
         w := 180-w;

      end;
      w := 90-w;
      y := -w;
   end else begin
      y := 0;
   end;
   with Engine do begin
      PushMatrix;
      LoadIdentity;
      Rotate(0, y, 0);
      Transform(p);
      PopMatrix;
   end;
   Normalizef(@p.x);
   w := arccos(p.z)/(pi/180);
   if p.y < 0 then begin
      w := 360-w;
   end;
   x := w;
   with Engine do begin
      PushMatrix;
      LoadIdentity;
      Rotate(x, 0, 0);
      Transform(p);
      PopMatrix;
   end;
   //ShowMessage(Point3DToStr(p));
   z := 0;*/

        /*X.Wert := (90-a);
   Y.Wert := -b;
   if setz then
   Z.Wert := -c;*/
    return result;
}

bool calcUmkreis(const CVector2& p1, const CVector2& p2, const CVector2& p3, utifloat &R, CVector2 &center)
{
    CVector2 dir1;
    dir1 = p2-p1;
    CVector2 dir2;
    dir2 = p3-p1;

    CVector2 origin1;
    origin1 = p1+dir1*0.5f;
    CVector2 origin2;
    origin2 = p1+dir2*0.5f;

    dir1 = CVector2(dir1.y, -dir1.x);
    dir2 = CVector2(dir2.y, -dir2.x);

    dir1.normalize();
    dir2.normalize();

    bool hassolution = false;
    CVector2 res = intersectLines(origin1, dir1, origin2, dir2, &hassolution);
    R=0; center.setVector(0,0);
    if (!hassolution) {
        return false;
    }
    center = origin1+dir1*res.x;
    R = center.distanceTo(p1);
    assert(absf(center.distanceTo(p2)-R) < 0.05);
    assert(absf(center.distanceTo(p3)-R) < 0.05);

    return hassolution;

}
